------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Clausify
  ( clausify -- :: Form -> [Clause]
  )
 where

import Data
import Set
import List( unzip4 )
import Utils( cross, holes, number )

---------------------------------------------------------------------------------
-- clausify

clausify :: Form -> [Clause]
clausify f = claus True empty 1 f final
 where
  claus topl bound i f k =
    case positive f of
      And gs ->
        list topl topl bound i systemAnd systemAndDef gs k

      Or gs ->
        list topl False bound i systemOr systemOrDef gs k

      All v g ->
        i' `seq` claus topl (v' `insert` bound) i' g' k
       where
        (v',i',g') = (varNameZ % i ::- Unknown, i+1, subst (pair v (Var v')) g)

      -- here, we really should push exi into an or when possible
      Exi v g ->
        i' `seq` claus topl bound i' g' k
       where
        i' = i+1
        t  = (skolemnName % i ::- Unknown) `app` map Var (elements (bound `intersect` free g))
        g1 = All v (Atom (Equal `prd` [Var v,t]) ->. g)
        g2 = subst (pair v t) g
        g' = g2

      -- this can be vastly improved
      Equiv f g ->
        claus topl bound i ((f ->. g) &. (g ->. f)) k

      Not (Atom p) ->
        k i [system [[Neg p]] []]

      Atom p ->
        k i [system [[Pos p]] []]

      rest ->
        error ("clausify: " ++ show rest)

  final i syss = map set (top ++ def)
   where
    (top, def, _) = best syss

  list topl topl' bound i op opDef gs k = combines [] i k gs
   where
    combines syss i k []     = combine syss i k
    combines syss i k (g:gs) = claus topl' bound i g $ \i' sys ->
                                 combines (sys:syss) i' k gs
    
    combine sysss i k
      | number ((> 1).length) sysss > tooMuch =
          combine (take n sysss) i $ \i1 syss1 ->
            combine (drop n sysss) i1 $ \i2 syss2 ->
              combine [syss1, syss2] i2 k

      | topl =
          k i [alt1]

      | otherwise =
          k i' ([alt1] ++ [ best [ opDef p syss | syss <- combos ] | length xs < 4])
     where
      i'     = i+1
      p      = ((skolemnName % i) ::- Unknown) `prd` xs
      xs     = map Var (elements (bound `intersect` free gs))
      n      = length sysss `div` 2
      combos = cross sysss
      alt1   = best [ op syss | syss <- combos ]

---------------------------------------------------------------------------------
-- weights

type Weight =
  ( Int  -- #clauses  in top
  , Int  -- #literals in top
  , Int  -- #clauses  in def
  , Int  -- #literals in def
  )

weight :: [[Lit]] -> [[Lit]] -> Weight
weight top def =
  ( length top
  , sum (map length top)
  , length def
  , sum (map length def)
  )

-- direct operations

weightAnd :: [Weight] -> Weight
weightAnd ws =
  ( sum ctops
  , sum ltops
  , sum cdefs
  , sum ldefs
  )
 where
  (ctops, ltops, cdefs, ldefs) = unzip4 ws

weightOr :: [Weight] -> Weight
weightOr ws =
  ( product ctops
  , sum [ ltop * product ctops' | (ltop,ctops') <- ltops `zip` holes ctops ]
  , sum cdefs
  , sum ldefs
  )
 where
  (ctops, ltops, cdefs, ldefs) = unzip4 ws

-- indirect operations

weightDef :: Weight -> Weight
weightDef (ctop, ltop, cdef, ldef) =
  ( 1
  , 1
  , ctop + cdef
  , ctop + ltop + ldef
  )

weightAndDef :: [Weight] -> Weight
weightAndDef = weightDef . weightAnd

weightOrDef :: [Weight] -> Weight
weightOrDef = weightDef . weightOr

-- comparison

(<=?) :: Weight -> Weight -> Bool
(ctop1, ltop1, cdef1, ldef1) <=? (ctop2, ltop2, cdef2, ldef2) =
  (ctop1 + cdef1, ltop1 + ldef1) <= (ctop2 + cdef2, ltop2 + ldef2)
  -- 3 * (ctop1 + cdef1) + ltop1 + ldef1 <= 3 * (ctop2 + cdef2) + ltop2 + ldef2

---------------------------------------------------------------------------------
-- systems

type System = ([[Lit]], [[Lit]], Weight)

system :: [[Lit]] -> [[Lit]] -> System
system top def = (top, def, weight top def)

-- direct operations

systemAnd :: [System] -> System
systemAnd syss =
  ( concat tops
  , concat defs
  , weightAnd ws
  )
 where
  (tops,defs,ws) = unzip3 syss

systemOr :: [System] -> System
systemOr syss =
  ( map concat (cross tops)
  , concat defs
  , weightOr ws
  )
 where
  (tops,defs,ws) = unzip3 syss

-- indirect operations

systemAndDef :: Atom -> [System] -> System
systemAndDef a syss =
  ( [[Pos a]]
  , [ Neg a : c | top <- tops, c <- top ] ++ concat defs
  , weightAndDef ws
  )
 where
  (tops,defs,ws) = unzip3 syss

systemOrDef :: Atom -> [System] -> System
systemOrDef a syss =
  ( [[Pos a]]
  , [ Neg a : concat cs | cs <- cross tops ] ++ concat defs
  , weightOrDef ws
  )
 where
  (tops,defs,ws) = unzip3 syss

-- comparison

best :: [System] -> System
best []   = error "clausify: best: empty list of systems"
best syss = foldr1 comp syss
 where
  comp sys1@(_, _, w1) sys2@(_, _, w2)
    | w1 <=? w2 = sys1
    | otherwise = sys2

---------------------------------------------------------------------------------
-- the maximum number of combinations to investigate

tooMuch :: Int
tooMuch = 10

---------------------------------------------------------------------------------
-- the end.
