{-# OPTIONS -fglasgow-exts #-}
------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Data where

import Set

import Maybe
  ( fromMaybe
  )

import List
  ( intersperse
  )

import Char
import IO( hFlush, stdout )

import Data.IORef
import System.IO.Unsafe
import Data.Array.IO
import Data.Array.MArray

infixl 5 %
infix  4 ::-

-------------------------------------------------------------------------
-- unique identifier

type Idf = Int

--showsIdf i = showChar '{' . shows i . showChar '}'
showsIdf i = id

-------------------------------------------------------------------------
-- name

data Name = MkName !Idf Appear

instance Show Name where
  showsPrec p (MkName i app) = showsIdf i . showsPrec p app

instance Eq Name where
  MkName x _ == MkName y _ = x == y

instance Ord Name where
  MkName x _ `compare` MkName y _ = x `compare` y

instance Hash Name where
  hash (MkName idf _) = idf

data Appear
  = Extern String
  | Intern String
  | Variant Appear Int
 deriving (Ord, Eq)

instance Show Appear where
  showsPrec p app = normal app
   where
    normal (Extern s)    = showString s
    normal (Intern s)    = showString s . showChar '\''
    normal (Variant a i) = special a . showChar '\'' . shows i
    
    special (Extern s)    = showString s
    special (Intern s)    = showString s
    special (Variant a i) = special a . showChar '\'' . shows i

instance Hash Appear where
  hash (Extern s)    = 37 * hash s + 1
  hash (Intern s)    = 37 * hash s + 2
  hash (Variant x n) = 37 * hash x + n

-------------------------------------------------------------------------
-- symbol

data Symbol
  = !Name ::- !Typing
  | Equal
  | Elt Int
 deriving (Eq, Ord)

typing :: Symbol -> Typing
typing (_ ::- t) = t
typing Equal     = PType [utype,utype]
typing (Elt _)   = FType [] utype

instance Show Symbol where
  showsPrec p (n ::- _) = showsPrec p n
  showsPrec p Equal     = showString "equal"
  showsPrec p (Elt n)   = showString "'" . shows n

{-
instance Eq Symbol where
  (n1 ::- _) == (n2 ::- _) = n1 == n2
  Equal      == Equal      = True
  Elt n1     == Elt n2     = n1 == n2
  _          == _          = False

instance Ord Symbol where
  (n1 ::- _) `compare` (n2 ::- _) = n1 `compare` n2
  (n1 ::- _) `compare` _          = LT

  Equal      `compare` Elt _      = LT                
  Equal      `compare` Equal      = EQ                
  Equal      `compare` (_ ::- _)  = GT                   

  Elt n1     `compare` Elt n2     = n1 `compare` n2   
  Elt _      `compare` _          = GT                
-}

instance Hash Symbol where
  hash (n ::- t) = 173 * hash n + hash t
  hash Equal     = 1
  hash (Elt n)   = 2*n

-------------------------------------------------------------------------
-- type

data Typing
  = VType Type
  | FType [Type] Type
  | PType [Type]
  | Unknown
 deriving (Eq, Ord)  

instance Show Typing where
  showsPrec p Unknown       = showString "?"

  showsPrec p (VType t)     = showsPrec p t

  showsPrec p (FType []  t) = showsPrec p t
  showsPrec p (FType [s] t) = showsPrec p s . showString " -> " . showsPrec p t
  showsPrec p (FType ts  t) = showString "("
                            . foldr (.) id
                              (intersperse (showString ", ")
                              (map (showsPrec p) ts))
                            . showString ") -> "
                            . showsPrec p t

  showsPrec p (PType [])    = showString "Pred"
  showsPrec p (PType [s])   = showsPrec p s . showString " -> Pred"
  showsPrec p (PType ts)    = showString "("
                            . foldr (.) id
                              (intersperse (showString ", ")
                              (map (showsPrec p) ts))
                            . showString ") -> Pred"

instance Hash Typing where
  hash (VType t)    = 4 * hash t 
  hash (FType ts t) = 4 * (hash ts + hash t) + 2
  hash (PType ts)   = 4 * (hash ts) + 3
  hash Unknown      = 1

data Type
  = MkType
  { tname    :: Name
  , tsize    :: Maybe Int
  , equality :: Equality
  }

utype :: Type
utype =
  MkType
  { tname    = name "*"
  , tsize    = Nothing
  , equality = Full
  }
  
tdomain :: Type -> Maybe Int
tdomain t =
  case tsize t of
    Nothing -> Nothing
    Just n  -> case equality t of
                 Safe -> Just n
                 Half -> Just (n+1)
                 Full -> Nothing

instance Eq Type where
  t1 == t2 = tname t1 == tname t2

instance Ord Type where
  t1 `compare` t2 = tname t1 `compare` tname t2

instance Show Type where
  showsPrec _ t = shows (tname t)

instance Hash Type where
  hash t = hash (tname t)

data Equality
  = Safe
  | Half
  | Full
 deriving ( Show, Eq, Ord )
 
-------------------------------------------------------------------------
-- terms

data Term
  = App !Idf !Symbol [Term]
  | Var !Symbol

typ :: Term -> Type
typ (App _ f _) = let FType _ t = typing f in t
typ (Var v)     = let VType t   = typing v in t

instance Eq Term where
  App x _ _ == App y _ _ = x == y
  Var v     == Var w     = v == w
  _         == _         = False

instance Ord Term where
  App x _ _ `compare` App y _ _ = x `compare` y
  Var v     `compare` Var w     = v `compare` w
  Var _     `compare` App _ _ _ = LT
  App _ _ _ `compare` Var _     = GT

instance Show Term where
  showsPrec _ (Var v)
    = shows v
  
  showsPrec _ (App i f [])
    = shows f
    . showsIdf i
  
  showsPrec _ (App i f ts)
    = shows f
    . showsIdf i
    . showChar '('
    . foldr (.) id
      (intersperse (showChar ',') (map shows ts))
    . showChar ')'

instance Hash Term where
  hash (Var v)       = hash v
  hash (App idf _ _) = idf

-------------------------------------------------------------------------
-- atoms

data Atom
  = Pred !Idf Symbol [Term]

instance Eq Atom where
  Pred x _ _ == Pred y _ _ = x == y

instance Ord Atom where
  Pred x _ _ `compare` Pred y _ _ = x `compare` y

instance Show Atom where
  showsPrec _ (Pred i p [])
    = shows p
    . showsIdf i
  
  showsPrec _ (Pred i Equal [x,y])
    = showChar '('
    . shows x
    . showString " ="
    . showsIdf i
    . showString " "
    . shows y
    . showChar ')'
  
  showsPrec _ (Pred i p ts)
    = shows p
    . showsIdf i
    . showChar '('
    . foldr (.) id
      (intersperse (showChar ',') (map shows ts))
    . showChar ')'

-------------------------------------------------------------------------
-- literals

type Lit
  = Signed Atom

data Signed a
  = Pos a
  | Neg a
 deriving (Eq, Ord)

instance Show a => Show (Signed a) where
  showsPrec _ (Pos x) = showString "++" . shows x
  showsPrec _ (Neg x) = showString "--" . shows x
  
instance Functor Signed where
  fmap f (Pos a) = Pos (f a)
  fmap f (Neg a) = Neg (f a)

neg :: Signed a -> Signed a
neg (Pos a) = Neg a
neg (Neg a) = Pos a

sign :: Signed a -> (b -> Signed b)
sign (Pos _) = Pos
sign (Neg _) = Neg

unsign :: Signed a -> a
unsign (Pos a) = a
unsign (Neg a) = a

---------------------------------------------------------------------------------
-- formulas

data Form
  = Not   Form
  | And   [Form]
  | Or    [Form]
  | Equiv Form Form
  | All   Symbol Form
  | Exi   Symbol Form
  | Atom  Atom
 deriving ( Eq, Ord, Show )

true, false :: Form
true  = And []
false = Or []

nt :: Form -> Form
nt (Not f)  = f
nt (And []) = Or []
nt (Or [])  = And []
nt f        = Not f

(&.), (|.), (->.), (<->.) :: Form -> Form -> Form
f      &. Or []  = false
Or []  &. g      = false
And fs &. And gs = And (fs ++ gs)
And fs &. g      = And (g:fs)
f      &. And gs = And (f:gs)
f      &. g      = And [f,g]

f      |. And [] = true
And [] |. g      = true
Or fs  |. Or gs  = Or (fs ++ gs)
Or fs  |. g      = Or (g:fs)
f      |. Or gs  = Or (f:gs)
f      |. g      = Or [f,g]

f      ->. g     = nt f |. g

Or []  <->. g      = nt g
And [] <->. g      = g
f      <->. Or []  = nt f
f      <->. And [] = f
f      <->. g      = Equiv f g

positive :: Form -> Form
positive (Not (Not f))     = positive f
positive (Not (And fs))    = Or  (map nt fs)
positive (Not (Or fs))     = And (map nt fs)
positive (Not (Equiv f g)) = Equiv f (nt g)
positive (Not (All v f))   = Exi v (nt f)
positive (Not (Exi v f))   = All v (nt f)
positive f                 = f

-------------------------------------------------------------------------
-- analyses

isNearPropClause :: Clause -> Bool
isNearPropClause = all isNearPropLit . elements

isNearPropLit :: Lit -> Bool
isNearPropLit (Pos a) = isNearPropAtom a
isNearPropLit (Neg a) = isNearPropAtom a

isNearPropAtom :: Atom -> Bool
isNearPropAtom (Pred _ _ ts) = all isNearPropTerm ts

isNearPropTerm :: Term -> Bool
isNearPropTerm (Var _) = True
isNearPropTerm t       = isGroundTerm t

isGroundTerm :: Term -> Bool
isGroundTerm (Var _)      = False
isGroundTerm (App _ _ ts) = all isGroundTerm ts

-------------------------------------------------------------------------
-- clauses

data CType
  = Assumption
  | Conjecture
 deriving (Eq, Show)

type Clause
  = Set Lit

data QClause
  = Uniq Symbol Clause
 deriving ( Eq )

pclause c =
     "[" ++ show (size (free c)) ++ "] "
  ++ bclause c

tclause c =
     "input_clause(axiom,axiom,["
  ++ tick (concat (intersperse "," (map showL (elements c))))
  ++ "])."
 where
  showL (Pos x) = "++" ++ showA x
  showL (Neg x) = "--" ++ showA x

  showA (Pred _ p [])
    = show p

  showA (Pred _ p ts)
    = show p
   ++ "("
   ++ concat (intersperse "," (map show ts))
   ++ ")"

  tick ('\'':s) = "___" ++ tick s
  tick (c:s)    = c : tick s
  tick ""       = ""

qclause q@(Uniq x c) =
     "[" ++ show (size (free q)) ++ "] unique "
  ++ show x ++ ": " 
  ++ bclause c

bclause c =
     concat (intersperse " | " (map show (elements c)))
  ++ "."

-------------------------------------------------------------------------
-- information

class Abstraction a where
  free     :: a -> Set Symbol
  substK   :: (Symbol :=> Term) -> a -> b -> (a -> b) -> b
  symbols  :: a -> (Set Symbol, Set (Symbol,Int), Set (Symbol,Int))
  subterms :: a -> Set Term

instance Abstraction Term where
  free (Var x)      = unit x
  free (App _ _ ts) = free ts
  
  substK sub (Var x) b k =
    case look x sub of
      Nothing -> b
      Just t  -> k t
  
  substK sub (App _ f ts) b k = substK sub ts b (k . (f `app`))

  symbols (Var x)      = (unit x, empty, empty)
  symbols (App _ f ts) = (vs, unit (f,length ts) `union` fs, ps)
   where
    (vs,fs,ps) = symbols ts

  subterms t =
    t `insert`
      case t of
        App _ _ ts -> subterms ts
        _          -> empty

instance Abstraction Atom where
  free (Pred _ _ ts) = free ts
  
  substK sub (Pred _ p ts) b k = substK sub ts b (k . (p `prd`))

  symbols (Pred _ p ts) = (vs, fs, unit (p,length ts) `union` ps)
   where
    (vs,fs,ps) = symbols ts

  subterms (Pred _ p ts) = subterms ts

instance Abstraction Form where
  free (Not f)     = free f
  free (And fs)    = free fs
  free (Or fs)     = free fs
  free (Equiv f g) = free f `union` free g
  free (All v f)   = v `delete` free f
  free (Exi v f)   = v `delete` free f
  free (Atom p)    = free p
  
  substK sub (Not f)     b k = substK sub f  b (k . Not)
  substK sub (And fs)    b k = substK sub fs b (k . And)
  substK sub (Or fs)     b k = substK sub fs b (k . Or)
  substK sub (Equiv f g) b k = substK sub [f,g] b (k . (\[f,g] -> Equiv f g))
  substK sub (All v f)   b k = substK' v sub f b (\v x -> k (All v x))
  substK sub (Exi v f)   b k = substK' v sub f b (\v x -> k (Exi v x))
  substK sub (Atom p)    b k = substK sub p b (k . Atom)

  symbols (Not f)     = symbols f
  symbols (And fs)    = symbols fs
  symbols (Or fs)     = symbols fs
  symbols (Equiv f g) = symbols [f,g]
  symbols (All v f)   = symbols' v f
  symbols (Exi v f)   = symbols' v f
  symbols (Atom p)    = symbols p

  subterms (Not f)     = subterms f
  subterms (And fs)    = subterms fs
  subterms (Or fs)     = subterms fs
  subterms (Equiv f g) = subterms [f,g]
  subterms (All v f)   = subterms f
  subterms (Exi v f)   = subterms f
  subterms (Atom p)    = subterms p

instance Abstraction a => Abstraction [a] where
  free = foldr union empty . map free
  
  substK sub []     b k = b
  substK sub (x:xs) b k =
    substK sub x
      (substK sub xs b (k . (x:)))
      (\x' -> substK sub xs (k (x':xs)) (k . (x':)))

  symbols [] = (empty, empty, empty)
  symbols xs = foldr1 comb . map symbols $ xs
   where
    comb (vs1,fs1,ps1) (vs2,fs2,ps2) =
      (vs1 `union` vs2, fs1 `union` fs2, ps1 `union` ps2)

  subterms = foldr union empty . map subterms

instance (Ord a, Abstraction a) => Abstraction (a :=> b) where
  free tab = foldr union empty [ free x | (x,_) <- tabulate tab ]
  
  substK sub tab b k = substT (tabulate tab) b (k . table)
   where
    substT []             b k = b
    substT (xy@(x,y):xys) b k =
      substK sub x
        (substT xys b (k . (xy:)))
        (\x' -> let xy' = (x',y) in 
                   substT xys (k (xy':xys)) (k . (xy':)))

  symbols tab = symbols (map fst (tabulate tab))

  subterms tab = subterms (map fst (tabulate tab))

instance Abstraction a => Abstraction (Signed a) where
  free (Pos x) = free x
  free (Neg x) = free x

  substK sub (Pos x) b k = substK sub x b (k . Pos)
  substK sub (Neg x) b k = substK sub x b (k . Neg)

  symbols (Pos x) = symbols x
  symbols (Neg x) = symbols x

  subterms (Pos x) = subterms x
  subterms (Neg x) = subterms x

instance Abstraction QClause where
  free (Uniq v ls) = delete v (free ls)

  substK sub (Uniq v x) b k = substK' v sub x b (\v x -> k (Uniq v x))

  symbols (Uniq v ls) = symbols' v ls

  subterms (Uniq _ ls) = subterms ls

subst :: Abstraction a => (Symbol :=> Term) -> a -> a
subst sub x = substK sub x x id

substK' :: Abstraction a => Symbol -> (Symbol :=> Term) -> a -> b -> (Symbol -> a -> b) -> b
substK' v sub x b k
  | v `notMember` sub && v `notMember` rhs =
      substK sub x b (k v)
  
  | otherwise =
      substK sub (subst (pair v (Var v')) x) b (k v')
 where
  rhs      = free (map snd (tabulate sub))
  (ws,_,_) = symbols x
  (v':_)   = [ v
             | i <- [1..]
             , let v = varNameX % i ::- Unknown
             , v `notMember` ws
             , v `notMember` sub
             , v `notMember` rhs
             ]

symbols' :: Abstraction a => Symbol -> a -> (Set Symbol, Set (Symbol,Int), Set (Symbol,Int))
symbols' v x = (v `insert` vs, fs, ps)
 where
  (vs, fs, ps) = symbols x

-------------------------------------------------------------------------
-- predefined

splitName :: Name
splitName = name' "split"

uniqueName :: Name
uniqueName = name' "uniq"

domainName :: Name
domainName = name' "dom"

skolemnName :: Name
skolemnName = name' "sk"

defName :: Name
defName = name' "def"

varNameX :: Name
varNameX = name' "X"

varNameY :: Name
varNameY = name' "Y"

varNameZ :: Name
varNameZ = name' "Z"

varNameCopy :: Name
varNameCopy = name' "R"

isExternName :: Name -> Bool
isExternName (MkName _ (Extern _)) = True
isExternName _                     = False

isExternSymbol :: Symbol -> Bool
isExternSymbol (n ::- _) = isExternName n
isExternSymbol _         = False

isSplitName :: Name -> Bool
isSplitName (MkName _ (Variant (Intern "split") _)) = True
isSplitName _                                       = False

isSplitSymbol :: Symbol -> Bool
isSplitSymbol (n ::- _) = isSplitName n
isSplitSymbol _         = False

isVarName :: Name -> Bool
isVarName (MkName _ (Extern (c:_))) = isUpper c
isVarName _                         = False

isVarSymbol :: Symbol -> Bool
isVarSymbol (n ::- _) = isVarName n
isVarSymbol _         = False

isVarCopyName :: Name -> Bool
isVarCopyName (MkName _ (Variant (Intern "R") _)) = True
isVarCopyName _                                   = False

isVarCopySymbol :: Symbol -> Bool
isVarCopySymbol (n ::- _) = isVarCopyName n
isVarCopySymbol _         = False

isDefName :: Name -> Bool
isDefName (MkName _ (Variant (Intern "def") _)) = True
isDefName _                                     = False

isDefSymbol :: Symbol -> Bool
isDefSymbol (n ::- _) = isDefName n
isDefSymbol _         = False

-------------------------------------------------------------------------
-- hash-consing

name :: String -> Name
name s = memoMkName (Extern s)

name' :: String -> Name
name' s = memoMkName (Intern s)

(%) :: Name -> Int -> Name
MkName _ a % i = memoMkName (Variant a i)

app :: Symbol -> [Term] -> Term
f `app` ts = memoApp f ts

prd :: Symbol -> [Term] -> Atom
p `prd` ts = memoPred p ts

-------------------------------------------------------------------------
-- dirty stuff

(memoMkName, memoApp, memoPred) =
  unsafePerformIO $
    do idfRef <- newIORef 0
       
       let mkMemo s sq f =
             do tab <- newTable
                return $ \x -> x `sq`
                  unsafePerformIO $
                    do -- putStrLn ("Table: " ++ s)
                       updateTable x (do n <- readIORef idfRef
                                         let n' = n+1
                                         --putStr ("(" ++ show n' ++ ")")
                                         --hFlush stdout
                                         n' `seq` writeIORef idfRef n'
                                         return (f n x)) tab
       
       mkName <- mkMemo "name" seq  MkName
       app    <- mkMemo "app"  seqf (\n (s,ts) -> App n s ts)
       pred   <- mkMemo "pred" seqf (\n (s,ts) -> Pred n s ts)
       return (mkName, curry app, curry pred)
 where
  []     `seqs` a = a
  (x:xs) `seqs` a = x `seq` (xs `seqs` a)

  (s,ts) `seqf` a = s `seq` (ts `seqs` a)
   
-------------------------------------------------------------------------
-- hash tables

class Eq a => Hash a where
  hash :: a -> Int

instance Hash Int where
  hash n = n

instance Hash Char where
  hash c = ord c

instance Hash a => Hash [a] where
  hash = hashf 0
   where
    hashf n []     = n
    hashf n (x:xs) = let n' = 257*n + hash x in n' `seq` hashf n' xs

instance (Hash a, Hash b) => Hash (a,b) where
  hash (x,y) = hash x + 301 * hash y

newtype Table a b
  = MkTable (IORef (HTab a b))

data HTab a b
  = MkHTab !Int !(IOArray Int [(a,b)])

newTable :: IO (Table a b)
newTable =
  do arr <- newArray (0,size) []
     ref <- newIORef (MkHTab 0 arr)
     return (MkTable ref)
 where
  size = 1701

updateTable :: Hash a => a -> IO b -> Table a b -> IO b
updateTable x my (MkTable ref) =
  do MkHTab n arr <- readIORef ref
     let (_,size) = bounds arr
         hashx    = hash x
         i        = hashx `mod` size
     
     xys <- readArray arr i
     case lookup x xys of
       Just y ->
         do return y     

       Nothing ->
         do y <- my
            let add size' arr' =
                  do let i' = hashx `mod` size'
                     abs <- readArray arr' i'
                     writeArray arr' i' ((x,y):abs)
                     writeIORef ref (MkHTab (n+1) arr')
                     return y
            
            if 100 * n < 70 * size then
              do -- putStrLn ("Add: " ++ show n ++ " elements")
                 add size arr
             else
              do let size' = 2*size + 1
                 --putStrLn ("Rehash: " ++ show size ++ " --> " ++ show size')
                 --hFlush stdout
                 arr' <- newArray (0,size') []
                 sequence_
                   [ do abs <- readArray arr i
                        writeArray arr i []
                        sequence_
                          [ do abs <- readArray arr' j
                               writeArray arr' j (ab:abs)
                          | ab@(a,_) <- abs
                          , let j = hash a `mod` size'
                          ]
                   | i <- [0..size]
                   ]
                 --putStrLn "Done."
                 --hFlush stdout
                 add size' arr'
       
{-
newtype Table a b
  = MkTable (IORef (a :=> b))

newTable :: IO (Table a b)
newTable =
  do ref <- newIORef empty
     return (MkTable ref)

updateTable :: Ord a => a -> IO b -> Table a b -> IO b
updateTable x my (MkTable ref) =
  do tab <- readIORef ref
     case look x tab of
       Nothing -> do y <- my
                     writeIORef ref (extend x y tab)
                     return y
       Just y  -> do return y
-}

-------------------------------------------------------------------------
-- the end.
