------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Ground
  ( Loc        -- :: *; Eq
  
  , GAtom(..)  -- :: *; Eq
  , GLit       -- :: *; Eq
  , GArg(..)   -- :: *; Eq
  
  , newLoc     -- :: Int -> Int -> SolverM Loc
  , getArity   -- :: Loc -> SolverM Int
  , addClauses -- :: Maybe Int -> [Int] -> [GLit] -> SolverM ()
  , getLit     -- :: GLit  -> SolverM S.Lit
  , getValue   -- :: GAtom -> SolverM (Maybe Bool)
  )
 where

import Data
import Sat hiding ( var, lit, neg, Lit )
import qualified Sat as S
import Data.IORef
import Data.Array.IO
import qualified Data.Array.MArray as M
import Data.Array.Unboxed
import IO

-------------------------------------------------------------------------
-- types

newtype Vector a
  = MkVector [a]
 deriving (Eq, Show)

vector :: [a] -> Vector a
vector xs = MkVector xs

lub :: Ord a => Vector a -> Vector a -> Vector a
MkVector xs `lub` MkVector ys =
  MkVector (zipWith max xs ys)

instance Ord a => Ord (Vector a) where
  MkVector xs `compare` MkVector ys =
    xs `compare` ys

instance Ix a => Ix (Vector a) where
  range (MkVector xs, MkVector ys) =
    [ MkVector zs | zs <- range' xs ys ]
   where
    range' []     []     = [[]]
    range' (x:xs) (y:ys) =
      [ z:zs | zs <- range' xs ys, z <- range (x,y) ]

  index (MkVector xs, MkVector ys) (MkVector zs) =
    index' xs ys zs
   where
    index' []     []     []     = 0
    index' (x:xs) (y:ys) (z:zs) =
      index (x,y) z + rangeSize (x,y) * index' xs ys zs
    
  inRange (MkVector xs, MkVector ys) (MkVector zs) =
    inRange' xs ys zs
   where
    inRange' []     []     []     = True
    inRange' (x:xs) (y:ys) (z:zs) = inRange (x,y) z && inRange' xs ys zs

data Loc
  = MkLoc !Int (IORef (UArray (Vector Int) Var))
 deriving (Eq)

data GAtom
  = !Loc :@ [GArg]
 deriving (Eq)

type GLit
  = Signed GAtom

data GArg
  = N !Int
  | V !Int
 deriving (Show, Eq)

-------------------------------------------------------------------------
-- operations

newLoc :: Int -> Int -> SolverM Loc
newLoc prio ar =
  do v   <- addVarP prio
     ref <- lift $ newIORef (array (arg,arg) [(arg,v)])
     return (MkLoc prio ref)
 where
  arg = MkVector (replicate ar 1)

getArity :: Loc -> SolverM Int
getArity (MkLoc prio ref) =
  do arr <- lift $ readIORef ref
     let (MkVector xs,_) = bounds arr
     return (length xs)

bumpLoc :: Loc -> Vector Int -> SolverM ()
bumpLoc (MkLoc prio ref) to1 =
  do arr <- lift $ readIORef ref
     let ran@(from,to2) = bounds arr
         to'            = to1 `lub` to2
         ran'           = (from,to')
     
     if inRange ran to'
       then return ()
       else do tab <- sequence
                        [ do v <- if inRange ran arg
                                    then return (arr ! arg)
                                    else addVarP prio
                             return (arg,v)
                        | arg <- range ran'
                        ]
               lift $ writeIORef ref (array ran' tab)

addClauses :: Maybe Int -> [Int] -> [GLit] -> SolverM ()
addClauses ms maxs lits =
  do -- check that all locations have all their SAT-variables
     let arrMax = listArray (1,numVars) maxs :: UArray Int Int
     
         max (N n) = n
         max (V i) = arrMax ! i
     
     sequence_
       [ bumpLoc loc (vector (map max arg))
       | lit <- lits
       , let loc :@ arg = unsign lit
       ]
     
     -- gather the right information about the literals
     lits' <- sequence
                [ do arr <- lift $ readIORef ref
                     return (arr, mkLit, arg)
                | lit <- lits
                , let loc@(MkLoc _ ref) :@ arg = unsign lit
                      mkLit = case lit of
                                Pos _ -> S.lit
                                Neg _ -> S.neg . S.lit
                ]

     -- go through all substitutions
     sub <- lift $ newArray (1,numVars) 0
     
     let enum [] Nothing =
           do subAndAdd
           
         enum [(i,s)] (Just s') =
           if s' <= s then
             do lift $ writeArray sub i s'
                subAndAdd
            else
             do return ()
         
         enum [] (Just _) =
           do return ()
         
         enum ((i,s):ins) ms =
           do sequence_
                [ do lift $ writeArray sub i n
                     enum ins ms'
                | n <- [1..s]
                , let ms' | Just n == ms = Nothing
                          | otherwise    = ms
                ]
         
         get (N n) = return n
         get (V i) = readArray (sub :: IOUArray Int Int) i

         subAndAdd =
           do ls <- lift $ sequence
                      [ do vec <- mapM get arg
                           return (mkLit (arr ! vector vec))
                      | (arr, mkLit, arg) <- lits'
                      ]
              addClause ls
                 
     enum [ (i,s) | (i,s) <- [1..] `zip` maxs ] ms
 where
  numVars = length maxs

getLit :: GLit -> SolverM S.Lit
getLit lit =
  do arr <- lift $ readIORef ref
     let v = arr ! vector [ d | N d <- arg ]
     return (case lit of
               Pos _ -> S.lit v
               Neg _ -> S.neg (S.lit v))
 where
  MkLoc _ ref :@ arg = unsign lit

getValue :: GAtom -> SolverM (Maybe Bool)
getValue (MkLoc _ ref :@ arg) =
  do arr <- lift $ readIORef ref
     if inRange (bounds arr) vec
       then do let v = arr ! vec
               b <- value v
               return (Just b)
       else do return Nothing
 where
  vec = vector [ d | N d <- arg ]

--------------------------------------------------------------------------------
-- the end.
