/* ---------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#include "Satnik.h"
#include <stdio.h>
#include <stdlib.h>

#if (__GNUC__ > 2)
#include <ext/hash_map>
using namespace __gnu_cxx;
#else
#include <hash_map>
#endif

using namespace std;
using namespace Satnik;

typedef unsigned int uint;
typedef void         ExtSolver;

struct ModelInspect {
  vector<int>& model;
  ModelInspect(vector<int>& m) : model(m) {};
  void operator()(Solver& s) { s.copy_model(model); }
};

struct CompareVariables {
  vector<int>& priority;
  CompareVariables(vector<int>& p) : priority(p) {}

  inline bool operator()(const Var& x, const Var& y) const {
    return priority[x] < priority[y];
  }
};

class HSolver {
public:
  Solver            s;
  vector<Lit>       clause;
  vector<Lit>       prefix;
  vector<int>       model;
  vector<int>       priority;

  HSolver(void) : 
    s(), 
    clause() {  }

  uint add_var(int p) { 
    Var v = s.add_vars();
    priority.resize(v+1);
    priority[v] = p;
    return v;
  }

  void clause_add_lit(uint l) { clause.push_back(encode(l)); }
  void clause_begin  (void)   { clause.clear();}
  uint clause_commit (void)   { return s.add_clause(clause) ? 1 : 0; }

  void prefix_clear  (void)   { prefix.clear(); }
  void prefix_add_lit(uint l) { prefix.push_back(encode(l)); }

  uint simplify(void) {
    if ( s.propagate() )
    {
      s.simplify_clauses();
      return 1;
    }
    return 0;
  }

  void dimacs(char* name) { 
    ofstream out;
    out.open(name, ofstream::out);
    if(!out.is_open()) { cout << "Could not open file " << name << "!\n"; exit(-1); }
    s.export_dimacs(out,prefix);
    out.close();
  }

  uint solve(void) {
    ModelInspect m(model);
    return s.solve(s.combine(CompareVariables(priority)),m,prefix) ? 1 : 0; }
  //return s.solve_(m,prefix) ? 1 : 0; }

  uint model_value(Lit l) { return model[index(l)] ? 1 : 0; }

};

extern "C" {
  uint  _lit (uint v)  { return index(lit(v)); }
  uint  _var (uint l)  { return var(encode(l)); }
  uint  _sign(uint l)  { return (uint)sign(encode(l)); }
  uint  _neg (uint l)  { return index(neg(encode(l))); }
  
  HSolver* solver_new(void) { 
    HSolver* s = (HSolver*)malloc(sizeof(HSolver));
    new (s) HSolver();
    
    s->s.set_variable_parameters(70,.998,0);
    s->s.set_polarity_parameters(true,false,true);
    s->s.set_learning_parameters(1000, 0 * s->s.get_nof_constraints(),.9985,.0001);
    s->s.set_restart_parameters (100,1.3);
    s->s.set_verbose(false);
    
    return s;
  }

  void     solver_free(HSolver* s)                    { s->~HSolver(); free(s); }
						     
  Var      solver_add_var  (HSolver* s)               { return s->add_var(0); }
  Var      solver_add_varp (HSolver* s, int p)        { return s->add_var(p); }
						     
  void     solver_clause_begin  (HSolver* s)          { s->clause_begin(); }
  void     solver_clause_add_lit(HSolver* s, uint l)  { s->clause_add_lit(l); }
  uint     solver_clause_commit (HSolver* s)          { return s->clause_commit(); }
						     
  void     solver_prefix_clear  (HSolver* s)          { s->prefix_clear(); }
  void     solver_prefix_add_lit(HSolver* s, uint l)  { s->prefix_add_lit(l); }
						     
  uint     solver_solve (HSolver* s)                  { return s->solve(); }
  void     solver_dimacs (HSolver* s, char*name )     { s->dimacs(name); }
  uint     solver_simplify (HSolver* s)               { return s->simplify(); }
  						     
  uint     solver_model_value (HSolver* s, uint l)    { return s->model_value(encode(l)); }
						     
  uint     solver_get_nof_clauses(HSolver* s)         { return s->s.get_nof_constraints(); }
  uint     solver_get_nof_vars   (HSolver* s)         { return s->s.get_nof_vars(); }
  uint     solver_get_nof_conflicts(HSolver* s)       { return s->s.get_nof_conflicts(); }

  void     solver_set_verbose    (HSolver* s, uint b) { s->s.set_verbose(b); }

  char    *identify (void) { return "Satnik"; }
}
