/* ---------------------------------------------------------------
-- Satnik, a SAT solver                                         --
-- Copyright (c) 2003 Niklas Srensson                          --
-- nik@cs.chalmers.se                                           --
--                                                              --
-- This file is part of Satnik.                                 --
--                                                              --
-- Satnik is free software; you can redistribute it and/or      --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Satnik is distributed in the hope that it will be useful,    --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Satnik; if not, write to the Free         --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#include <fstream>

namespace Satnik {

// the core searching method
template <class C>
Result Solver::search(C &c, int conflicts) {
  vector<Lit> conflict_set;
  vector<Lit> conflict_clause;
  Result      res;
  int         num_prev_conf = num_conf;

  for(;;) {
    num_prop++;
    //cout << "Propagate " << current_level() << '\n';
    if(!propagate()) {
      //cout << "Conflict - ";
      num_conf++; conflict_set.clear(); conflict_clause.clear();
      if(current_level() == root_level) { 
	//cout << "Last branch\n";
	res = Unsatisfiable; break; }


      conflict_set.push_back(neg(assigns.back()));
      conflict->reason(assigns.back(),conflict_set);
      int next_level = analyze(conflict_set, conflict_clause);
      //cout << "Conflict set: " << conflict_set << '\n';
      //cout << "Conflict clause: " << conflict_clause << '\n';
      //cout << "Backtrack to: " << next_level << '\n';

      while(current_level() > next_level) pop();
      //next_decision = 0;

      if(conflict_clause.size() > 1) {
	Clause* c = learn_clause(conflict_clause);
	assign(conflict_clause.front(), (Constraint*)c);
      } else {
	unit_conflicts.push_back(conflict_clause.front());
	assign(conflict_clause.front());
      }

      if((num_conf-num_prev_conf) % clause_compaction_interval == 0) { 
	decay_clause_activity(); 
	delete_learned_clauses(); }
      if((num_conf-num_prev_conf) % variable_reorder_interval == 0) { 
	decay_var_activity(); 
	reorder(c); }

    } else {
      if(conflicts > 0 && (num_conf - num_prev_conf) >= conflicts)  { res = Indeterminate; break; }

      if(current_level() == 0 && simplified_frame < (int)assigns.size()) {
	simplify_clauses();
	simplified_frame = assigns.size(); }
	
      //cout << "Decide\n";
      num_deci++;
      bool random_dec = variable_randomness > 0 && (num_deci % variable_randomness) == 0;
      Var v = random_dec ? get_random_decision() : get_decision();

      bool polarity;
//      if(watches(lit(v,true)).size() == 0)
//	polarity = false;
//      else if(watches(lit(v,false)).size() == 0)
//	polarity = true;
//      else
	polarity = random_dec ? (bool)(rand() / (RAND_MAX /2)) :
	  ((var_polarity[v] > 0) ^ variable_polarity_sign);

      if(v == end_var()) { res = Satisfiable; break; }
      push();
      assign(lit(v, polarity));
    }
  }

  if(res == Satisfiable) {
    for(size_t i = 0; i < problem_constraints.size(); i++) 
      if(!problem_constraints[i]->satisfied(*this)) {
	cout << "No good model!\n";
	cout << *((Clause*)problem_constraints[i]);
	exit(-1);
      } 
  }
	
  return res;
}

// the toplevel solving method + variants
// TODO: explain what is happening here!!
template <class C, class M>
bool Solver::solve(C c, M& m, const vector<Lit>& prefix) {
//  ofstream testing("debug.cnf");
//  export_dimacs(testing);

  // not handled by interface, remove this ???
  if(queue.size() > 0) {
    if(!propagate()) { return false; }
    simplify_clauses();
  }

  if(prefix.size() > 0) {
    push();
    for(size_t i = 0; i < prefix.size(); i++)
      if(!assign(prefix[i])) { 
	queue.clear();
	pop(); return false; }
    if(!propagate()) { pop(); return false; }
  } else simplified_frame = 0;

  root_level    = current_level();
  Result r;

  initialize_order(); 
//  static_clausebased_order(5); 
//  size_t ri = variable_reorder_interval;
//  reorder(c);
//  cout << "static strategy\n";
//  r = search(c,5000);
//  variable_reorder_interval = ri;
//  for(size_t i = 0; i < order.size(); i++)
//    var_activity[order[i]] = 0; 

  reorder(c);

  for(float steps = (float)restart_initial_interval; ; steps *= restart_increase) {
    r = search(c,(int)steps);
    if(r != Indeterminate) break;


    if(verbose)
      cout << "Steps = " << (int)steps << ", Prog = " << progress() 
	   << ", LC = " << learned_clauses.size() 
	   << ", PC = " << problem_constraints.size() << '\n';

    num_restart++;
    while(current_level() > root_level) pop();
  }

  assert(r != Indeterminate);
  if(r == Satisfiable) m(*this);
  while(current_level() > 0) pop();

//  cout << "Unit conflicts: " << unit_conflicts.size() << '\n';

  while(unit_conflicts.size() > 0) {
    Lit l = unit_conflicts.back();
    unit_conflicts.pop_back();
    #ifndef NDEBUG
    bool b = assign(l); assert(b);
    #else
    assign(l);
    #endif
  }

  return r == Satisfiable;
}

};
