------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module SolveInstances where

import qualified Sat as S
import Sat hiding ( neg, lit, Lit, dimacs )
import Ground
import Data hiding ( Type )
import Data.IORef
import Set
import Flags
import List( sortBy, intersperse )
import IO
import Instantiate
import Utils( rjust )
import Output
import AnalysisTypes

-------------------------------------------------------------------------
-- result

data Result
  = Contradiction
  | Inconclusive Int
  | Solution Int
 deriving ( Show, Eq )

-------------------------------------------------------------------------
-- solver

solveInstances :: Flags -> Int -> [(Int,Bool,Symbol,[ClauseSet])] -> IO Result
solveInstances flags minSize css =
  do line <- newIORef False
     ref  <- newIORef (empty,empty)
     
     let getFunLoc (f,n) =
           do (tabf,tabp) <- lift $ readIORef ref
              case look f tabf of
                Nothing  -> do loc <- newLoc prio (n+1)
                               lift $ writeIORef ref (extend f loc tabf,tabp)
                               return loc
                Just loc -> do return loc
          where
           prio = case priority flags of
                    PredConFun | n == 0    -> 2
                               | otherwise -> 3
                    ConPredFun | n == 0    -> 1
                               | otherwise -> 3
                    ConFunPred | n == 0    -> 1
                               | otherwise -> 2
                    _                      -> 0
                               
         getPredLoc (p,n) =
           do (tabf,tabp) <- lift $ readIORef ref
              case look p tabp of
                Nothing  -> do loc <- newLoc prio n
                               lift $ writeIORef ref (tabf,extend p loc tabp)
                               return loc
                Just loc -> do return loc
          where
           prio = case priority flags of
                    PredConFun -> 1
                    ConPredFun -> 2
                    ConFunPred -> 3
                    _          -> 0
     
         printModel k showf showp h =
           do (tabf,tabp) <- lift $ readIORef ref
              sequence_
                ( intersperse (lift $ hPutStrLn h "")
                  ( map snd $
                    sortBy (\(x,_) (y,_) -> x `compare` y)
                    ( [ (show f, printFun f loc)
                      | (f,loc) <- tabulate tabf
                      , isExternSymbol f
                      ]
                   ++ [ (show p, printPred p loc)
                      | (p,loc) <- tabulate tabp
                      , isExternSymbol p
                      ]
                    )
                  )
                )
          where
           enum []     = [[]]
           enum (k:ks) = [ i:is | i <- [1..k], is <- enum ks ]

           isize t =
             case tdomain t of
               Just n  -> n `min` k
               Nothing -> k
           
           showd n = "'" ++ show n

           printFun f loc =
             do n <- getArity loc
                let answer []     xs = return "--"
                    answer (i:is) xs =
                      do Just b <- getValue (loc :@ (map N (xs ++ [i])))
                         if b then
                           return (showd i)
                          else
                           answer is xs

                    ks = [ isize t | let FType ts _ = typing f, t <- ts ]

                sequence_
                  [ do a <- answer [1..k] xs 
                       lift $ hPutStrLn h $ showf f (map showd xs) a
                  | xs <- enum ks
                  ]
{-
                sequence_
                  [ do lift $ hPutStrLn h $ showf f (map showd xs1) a
                  | (i,k') <- [1..] `zip` ks
                  , j <- [k'+1..k]
                  , let xs1 = [ "X" ++ show x | x <- [1..i-1] ]
                           ++ [ showd j ]
                           ++ [ "X" ++ show x | x <- [i+1..n] ]
                        
                        xs2 = [ "X" ++ show x | x <- [1..i-1] ]
                           ++ [ showd (j-1) ]
                           ++ [ "X" ++ show x | x <- [i+1..n] ]
                        
                        a = show f
                         ++ "("
                         ++ concat (intersperse "," xs2)
                         ++ ")"
                  ]
-}
           printPred p loc =
             do n <- getArity loc
                sequence_
                  [ do mb <- getValue (loc :@ (map N xs))
                       lift $ hPutStrLn h $ showp p (map showd xs) (maybe False id mb)
                  | xs <- enum [ isize t | let PType ts = typing p, t <- ts ]
                  ]
         
         showFunUser f xs a =
              show f
           ++ (if not (null xs) then
                "("
             ++ concat (intersperse "," xs)
             ++ ")"
                else ""
              ) 
           ++ " = "
           ++ a
     
         showPredUser p xs b =
              show p
           ++ (if not (null xs) then
                "("
             ++ concat (intersperse "," xs)
             ++ ")"
                else ""
              )
           ++ " : "
           ++ if b then "TRUE" else "FALSE"
         
         tptp = concatMap (\c -> if c == '\'' then "c__" else [c])
     
         showFunTptp f xs a =
              "input_clause(axiom,axiom,[++equal("
           ++ show f
           ++ (if not (null xs) then
                "("
             ++ concat (intersperse "," (map tptp xs))
             ++ ")"
                else ""
              ) 
           ++ ","
           ++ tptp a
           ++ ")])."
     
         showPredTptp p xs b =
              "input_clause(axiom,axiom,["
           ++ (if b then "++" else "--")
           ++ show p
           ++ (if not (null xs) then
                "("
             ++ concat (intersperse "," (map tptp xs))
             ++ ")"
                else ""
              )
           ++ "])."
     
         processClauseSet k i (ForAll cs) =
           do sequence_ [ processClause Nothing k i c | c <- cs ]
         
         processClauseSet k i (ForAllNew k' cs) =
           do sequence_ [ processClause (Just k') k i c | c <- cs ]
         
         processClause ms k i c =
           do if Progress `elem` prints flags then
                let p  = (round (20 * i) - 1) `max` 0  :: Int
                    p' = round (100 * i)      `min` 99 :: Int
                    s  = [ ' ' | p' < 10 ]
                      ++ show p'
                      ++ "%"
                      ++ replicate p '='
                      ++ ">"
                      ++ replicate (19-p) '-'
                      ++ "|"
                 in putStr' (s ++ replicate (length s) '\b')
               else
                return ()
           
              ls' <- mapM processLit ls
              let args = [ isize t | v <- vs, let VType t = typing v ]
              --putStrLn' (show args)
              addClauses ms args ls'
              -- addClauses ms [ k | _ <- vs ] ls'
          where
           ls = elements c
           vs = elements (free c)
           
           isize t =
             case tdomain t of
               Just n  -> n `min` k
               Nothing -> k
           
           processLit l =
             do a <- processAtom (unsign l)
                return (Data.sign l a)
           
           processAtom (Pred _ Equal [App _ f xs, y])
             | not (isElt f) =
             do loc <- getFunLoc (f,length xs)
                return (loc :@ (xs' ++ [y']))
            where
             xs' = map processTerm xs
             y'  = processTerm y
           
             isElt (Elt _) = True
             isElt _       = False
           
           processAtom (Pred _ p xs) =
             do loc <- getPredLoc (p,length xs)
                return (loc :@ xs')
            where
             xs' = map processTerm xs
           
           processTerm (Var v) =
             V (v `ind` vs)
           
           processTerm (App _ (Elt d) []) = N d
         
           x `ind` (y:ys)
             | x == y    = 1
             | otherwise = 1 + (x `ind` ys)             
         
         finalLine =
           do b <- readIORef line
              if b then putStrLn "" else return ()
              writeIORef line False
         
         domains minSize [] =
           do return (Inconclusive minSize)

         domains minSize ((k,check,assump,clauses):rest) =
           do t <- lift $ getTimeSpentS flags
              putStr' (rjust 9 t ++ " |")
              lift $ writeIORef line True
              
              if Instantiated `elem` prints flags
                then putStr' $ unlines $ [ "==> Instantiated problem" ]
                                      ++ map sclause clauses
                                      ++ [ "==> Assumptions" ]
                                      ++ [ show assump ]
                else return ()

              putStr' (rjust 5 (show k) ++ " |")
              let clauses' = flat clauses
              
                  flat []                     = []
                  flat (ForAll cs       : ds) = map (\c -> ForAll [c])       cs ++ flat ds
                  flat (ForAllNew k' cs : ds) = map (\c -> ForAllNew k' [c]) cs ++ flat ds

                  tot = length clauses'
              
              sequence_
                [ processClauseSet k (fromIntegral i / fromIntegral tot) c
                | (i,c) <- [1..] `zip` clauses' 
                ]
              
              loc <- getPredLoc (assump,0)
              assumption <- getLit (Pos (loc :@ []))
              
              simplify
              if contr flags
                then do b <- solve []
                        if b then return () else contradiction
                else do return ()
              
              nVars    <- getNVars
              nClauses <- getNClauses
              t        <- lift $ getTimeSpentS flags
              putStr' ( rjust 9  (show nVars)    ++ " |"
                     ++ rjust 11 (show nClauses) ++ " |"
                     ++ rjust 9  t               ++ " |"
                      )
              
              case dimacs flags of
                Just file -> S.dimacs (size file) [assumption]
                 where
                  size ('%':'d':s) = show k ++ size s
                  size (c:s)       = c : size s
                  size s           = s
                  
                Nothing -> return ()

              r <- if check then
                     if pretend flags then
                       do putStr' (rjust 9 "" ++ " | pretend")
                          return Contradiction
                     else
                       do b <- solve [assumption]
                          n <- getNConflicts
                          putStr' (rjust 9 (show n) ++ " |")
                          if b then
                            do putStr' " yes!"
                               return (Solution k)
                           else
                            do putStr' " no"
                               return Contradiction
                    else
                     do putStr' (rjust 9 "" ++ " | --")
                        return (Inconclusive k)

              putStrLn' ""
              lift $ writeIORef line False
              
              case r of
                Solution k ->
                  do lift finalLine
                     if Model `elem` prints flags then
                       do lift $ putHeader "Model"
                          printModel k showFunUser showPredUser stdout
                      else
                       do return ()
                     
                     case mfile flags of
                       Just f  -> do h <- lift $ openFile f WriteMode
                                     printModel k showFunUser showPredUser h
                                     lift $ hClose h
                       Nothing -> do return ()
                     
                     case cfile flags of
                       Just f  -> do lift $ appendFile f $ unlines $
                                       [ "% Domain:"
                                       , "input_clause(axiom,axiom,["
                                      ++ concat (intersperse "," (map (\i ->
                                            "++equal(X,c__" ++ show i ++ ")") [1..k]))
                                      ++ "])."
                                       ] ++
                                       [ "input_clause(axiom,axiom,[--equal(c__"
                                      ++ show c1
                                      ++ ",c__"
                                      ++ show c2
                                      ++ ")])."
                                       | c1 <- [1..k]
                                       , c2 <- [c1+1..k]
                                       ]
                                     lift $ appendFile f "% Model:\n"
                                     h <- lift $ openFile f AppendMode
                                     printModel k showFunTptp showPredTptp h
                                     lift $ appendFile f "% The end.\n"
                       Nothing -> do return ()
                     
                     return (Solution k)

                _ ->
                  do addClause [S.neg assumption]
                     domains minSize' rest
                 where
                  minSize'
                    | r == Contradiction && minSize == k = minSize+1
                    | otherwise                          = minSize
                
     putHeader "Searching"
     putStrLn "     time | size |    #vars |   #clauses |     time |   #confl | result"
     putStrLn "----------+------+----------+------------+----------+----------+------------"
     
     r <- run $ domains minSize css
     finalLine
     return $      
       case r of
         Just Contradiction -> error "should not happen"
         Just res           -> res
         Nothing            -> Contradiction
 where
  putStrLn' s = lift (putStrLn s >> hFlush stdout)
  putStr'   s = lift (putStr   s >> hFlush stdout)
 
  sclause (ForAll cs) =
    concat (intersperse "\n" (map pclause cs))

  sclause (ForAllNew _ cs) =
    concat (intersperse "\n" (map (("!"++) . pclause) cs))

{-
  pclause c =
       "[" ++ show (size (free c)) ++ "] "
    ++ bclause c

  bclause c =
       concat (intersperse " | " (map show (elements c)))
    ++ "."
-}

-------------------------------------------------------------------------
-- the end.
