(* $Id: bound.ml,v 1.3 2004/12/28 14:06:45 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)




(*** types ***)


type config = Config.config
type choice_point = State.choice_point
type term = Term.term
type literal = Term.literal
type subst = Subst.subst

type complexity = int





(*** bound classes ***)


(* Restarting base class *)
class virtual bound_base
  (__config: config) (__complexity: int) =

object (self: 'a)
  (*** values ***)

  (* environment *)
  val _config = __config

  (* number of abandoned incomplete branches *)
  val mutable _incomplete_branches = 0;

  (* number of restarts *)
  val mutable _restarts = 0;

  (* the current deepening bound. *)
  val mutable _bound : complexity = __complexity

  (* lowest complexity of the dropped candidates. *)
  val mutable _dropped_complexity: int option = None

  (* an incomplete branch has been found and ignored. *)
  val mutable _incomplete = false


  (* named constructor *)
  method virtual private create: config -> complexity -> bound_base


  (*** stats access ***)

  method current = _bound

  method set_restarts value = _restarts <- value

  method set_incomplete_branches value = _incomplete_branches <- value

  method to_string = string_of_int _bound

  method print: unit =
    Print.print_statistic "Incomplete Branches" _incomplete_branches;
    Print.print_statistic "Restarts" _restarts;
    Print.print_statistic "Bound" _bound


  (*** complexitiy access ***)

  method get_complexity literal =
    if Config.iterative_deepening _config = Flags.IT_TermDepth then
      Term_attributes.depth_of_literal literal
    else
      Term_attributes.weight_of_literal literal

  method get_complexity_subst literal subst offset =
    if Config.iterative_deepening _config = Flags.IT_TermDepth then
      Term_attributes.depth_of_literal_subst literal subst offset
    else
      Term_attributes.weight_of_literal_subst literal subst offset

  method compare_complexity (complexity1: complexity) (complexity2: complexity) : int =
    Tools.compare_int complexity1 complexity2




  (*** check ***)

  (* convention: methods ending with _
     take a literal's complexity instead of the literal itself *)

  (* does the literal exceed the deepening bound? *)
  method exceeds complexity : bool =
    self#compare_complexity _bound complexity = -1

  method private exceeds_subst literal subst offset =
    self#exceeds (self#get_complexity_subst literal subst offset)


  (* does this exceeding candidate have the lowest complexity
     of all up to now dropped candidates? *)
  method private exceeds_current_bound_ complexity =
    match _dropped_complexity with
      | None ->
	  (* first exceeding candidate *)
	  true
	  
      | Some dropped_complexity ->
	  (* lowest candidate? *)
	  self#compare_complexity dropped_complexity complexity = 1

  method private exceeds_current_bound literal =
    self#exceeds_current_bound_ (self#get_complexity literal)


  (* for backtracking over an incomplete branch:
     is the choice point of this candidate of any relevance? *)
  method private virtual exceeds_current_choice_point: choice_point -> bool


  method exceeds_current complexity choice_point =
    self#exceeds complexity
    &&
    (
      self#exceeds_current_bound_ complexity
      ||
      self#exceeds_current_choice_point choice_point
    )


  (* the literal is dropped if it exceeds the bound
     and the complexity and choice point are registered for restarting. *)
  method private virtual do_register: complexity -> choice_point -> bool

  method register complexity choice_point =
    self#do_register complexity choice_point

  method register_subst literal subst offset choice_point =
    let complexity =
      self#get_complexity_subst literal subst offset
    in
      self#do_register complexity choice_point




  (*** incompleteness ***)

  method virtual is_branch_incomplete: bool

  method is_derivation_incomplete = _incomplete




  (*** backtracking ***)

  method virtual private do_restart_incomplete: choice_point list

  method restart_incomplete =
    _incomplete <- true;
    _incomplete_branches <- _incomplete_branches + 1;
    self#do_restart_incomplete


  method restart =
    let restart_bound =
      match _dropped_complexity with
	| None ->
	    failwith "Bound.restart"
	      
	| Some dropped_complexity ->
	    dropped_complexity
    in
    let restarted_bound =
      self#create _config restart_bound
    in
      restarted_bound#set_restarts (_restarts + 1);
      (* do we won't to show all found incomplete branches,
	 or only the ones within the current depth bound? *)
      (*restarted_bound#set_incomplete_branches _incomplete_branches;*)
      restarted_bound

  method virtual jump_target : choice_point list

  method virtual backtrack : unit
end



(* Eager Restarting,
   i.e. restarting on the first incomplete branch without backtracking.
   Candidates are eagerly dropped. *)
class bound_eager
  (__config: config) (__bound: int) =

object (self)
  inherit bound_base __config __bound

  val mutable _dropped_choice_point: choice_point option = None

  method private create config bound =
    new bound_eager config bound


  (* choice points are irrelevant *)
  method private exceeds_current_choice_point _ = false


  method private do_register complexity choice_point =
    (* complexity exceeds?. *)
    if self#exceeds complexity then begin
      (* record the minimal dropped complexity *)
      if self#exceeds_current_bound_ complexity then begin
	_dropped_complexity <- Some complexity
      end;

      (* record the oldest dropped choice point *)
      begin
	match _dropped_choice_point with
	  | Some current when
	      State.compare_age current choice_point <= 0 ->
	      ()

	  | _ ->
	      _dropped_choice_point <- Some choice_point
      end;
      
      true
    end
      
    else begin
      false
    end

  (* something dropped since the last restart? *)
  method is_branch_incomplete =
    match _dropped_choice_point with
      | None -> false
      | Some _ -> true

  method private do_restart_incomplete =
    failwith "Bound.bound_eager.do_restart_incomplete"

  method jump_target : choice_point list =
    failwith "Bound.bound_eager.jump_target"

  method backtrack =
    match _dropped_choice_point with
      | Some choice_point when
	  State.is_choice_point_invalid choice_point ->
	  _dropped_choice_point <- None

      | _ ->
	  ()
	  
end






(* Lazy Restarting,
   i.e. backtracking to the first choice point where a candidate has been dropped
   when an incomplete branch has been detected.
   Candidates are eagerly dropped. *)
class bound_lazy
  (__config: config) (__bound: int) =

object (self)
  inherit bound_base __config __bound

  (* the base choice points of the dropped candidates.
     That is, for chronological backtracking the oldest choice point,
     for non-chronological backtracking all choice points
     which are independent of each other.
     This is necessary,
     as when only the oldest choice point was recorded,
     it could be removed in non-chronological backtracking
     while later (also drop) choice points could be kept. *)
  val mutable _dropped_choice_points: choice_point list = []

  method private create config bound =
    new bound_lazy config bound

  (* is this exceeding candidate independent from any other
     already dropped candidates? *)
  method private exceeds_current_choice_point choice_point =
    not (
      List.exists
	(fun dropped_choice_point ->
	   State.depends_on _config choice_point dropped_choice_point
	)
	_dropped_choice_points
    )
	    
  method private do_register complexity choice_point =
    if self#exceeds complexity then begin
      (* get the minimal dropped complexity. *)
      if self#exceeds_current_bound_ complexity then begin
	_dropped_complexity <- Some complexity
      end;
      
      (* get all independent dropped choice points. *)
      if self#exceeds_current_choice_point choice_point then begin
	(* remove choice points dependent on the new
	   dropped candidates' choice point *)
	let independent_choice_points =
	  List.find_all
	    (fun dropped_choice_point ->
	       not (State.depends_on _config dropped_choice_point choice_point)
	    )
	    _dropped_choice_points
	in	    
	  _dropped_choice_points <- choice_point :: independent_choice_points
      end;
      
      true
    end
      
    else begin
      false
    end


  method is_branch_incomplete =
    match _dropped_choice_points with
      | [] -> false
      | _ -> true


  method private do_restart_incomplete =
    (* get the oldest drop choice point. *)
    let retract_choice_point =
      _dropped_choice_points
    in
      _dropped_choice_points <- [];
      retract_choice_point

  method jump_target =
    _dropped_choice_points

  (* remove invalidated dropped choice points. *)
  method backtrack =
    _dropped_choice_points <- List.find_all State.is_choice_point_valid  _dropped_choice_points

end





(* Delayed Restarting,
   i.e. backtracking to the last choice point where a candidate has been dropped.
   candidates are only dropped when an exhausted branch
   is tested for being incomplete. *)
class bound_delayed
  (__config: config) (__bound: int) =

object (self)
  inherit bound_base __config __bound

  (*** values ***)

  (* the youngest choice point corresponding to a drop. *)
  val mutable _dropped_choice_point: choice_point option = None


  method private create config bound =
    new bound_delayed config bound


  (* keep the younger choice point *)
  method private exceeds_current_choice_point choice_point =
    match _dropped_choice_point with
      | None ->
	  true

      | Some dropped_choice_point ->
	  State.compare_age dropped_choice_point choice_point = -1
	    

  method private do_register complexity choice_point =
    if self#exceeds complexity then begin
      (* get the minimal dropped complexity. *)
      if self#exceeds_current_bound_ complexity then begin
	_dropped_complexity <- Some complexity
      end;
      
      (* get the youngest dropped choice point. *)
      if self#exceeds_current_choice_point choice_point then begin
	_dropped_choice_point <- Some choice_point;
      end;
      
      true
    end
      
    else begin
      false
    end


  method is_branch_incomplete =
    match _dropped_choice_point with
      | None -> false
      | _ -> true


  method private do_restart_incomplete =
    let retract_choice_point =
      match _dropped_choice_point with
	| None ->
	    failwith "Bound.bound_delayed.do_restart_incomplete"

	| Some choice_point ->
	    [choice_point]
    in
      _dropped_choice_point <- None;
      
      retract_choice_point


  method jump_target : choice_point list =
    failwith "Bound.bound_delayed.jump_target"

  method backtrack =
    match _dropped_choice_point with
      | Some choice_point when
	  State.is_choice_point_valid choice_point ->
	  _dropped_choice_point <- None

      | _ ->
	  ()

end







(*** signature ***)

class type bound =
object
  (** {2 Bound} *)

  method current: complexity

  (** the deepening bound as a string. *)
  method to_string: string
    
  (** string representation of the current bound (ala {!Print.print_label}). *)
  method print : unit
    


  (** {2 Complexity} *)
    
  (** the complexity value of a literal. *)
  method get_complexity : literal -> complexity
    
  (** [bound_of_literal_subst literal subst offset]
      returns the complexity of the literal assembled by applying
      [subst] with [offset] to [literal]. *)
  method get_complexity_subst : literal -> subst -> int -> complexity
    
  (** usual comparison.
      Amounts to a compare over ints for term depth and term weight. *)
  method compare_complexity : complexity -> complexity -> int


    
  (** {2 Check} *)
    
  (** returns true, if the candidate exceeds the deepening bound. *)
  method exceeds : complexity -> bool
    
  (** returns true, if registering this candidate would have an impact
      wrt. to backtracking over an incomplete branch.
      This is never the case if the candidate does not exceed the depth bound.
	Otherwise, the details depend on the restart strategy ({!Flags.opt_restart}),
      e.g. for eager restarting only the lowest dropped complexity is of relevance.
  *)
  method exceeds_current : complexity -> choice_point -> bool
    
  (** like {!Bound.bound.exceeds},
      and additionally registers the candidate as dropped. *)
  method register : complexity -> choice_point -> bool
      
  (** like {!Bound.bound.register},
      but for the literal assembled by applying
	[subst] with [offset] to [literal]. *)
  method register_subst : literal -> subst -> int -> choice_point -> bool



  (** {2 Incompleteness} *)
    
    
  (** candidates have been dropped in the current derivation branch.
      This implies that the current branch can not be extended towards a model. *)
  method is_branch_incomplete : bool
    
  (** an incomplete branch is part of the derivation,
      but has been ignored while further exploring other derivation branches.
      This implies that the current derivation can not be extended towards a refutation. *)
  method is_derivation_incomplete : bool


  (** {2 Backtracking} *)
    
  (** when abandoning the current incomplete branch
      this determines the potential choice points to backtrack to,
      in order to continue the derivation
      according to the restarting strategy ({!Flags.opt_restart}).
      That is, all these choice points
      are responsible for dropping exceeding candidates
	and have to be invalidated.
  *)
  method restart_incomplete : choice_point list
    
  (** cleans all stored data and starts a fresh derivation with an increased depth bound.
      The new depth bound is the lowest depth bound of the dropped candidates. *)
  method restart : bound

  method jump_target: choice_point list

  (** retracts information invalidated by backtracking some choice points. *)
  method backtrack : unit
end



(*** public interface ***)


let create ?(bound: int = -1) config : bound =
  let bound =
    if bound == -1 then
      (Config.deepening_bound config)
    else
      bound
  in
  match Config.restart config with
    | Flags.RS_Eager ->
	(new bound_eager config bound : bound_eager :> bound)
	  
    | Flags.RS_Lazy ->
	(new bound_lazy config bound : bound_lazy :> bound)
	  
    | Flags.RS_Delayed ->
	(new bound_delayed config bound : bound_delayed :> bound)
