(* $Id: bound.mli,v 1.3 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(** iterative deepening bound *)

(** {6 Types} *)

type config = Config.config
type literal = Term.literal
type choice_point = State.choice_point
type subst = Subst.subst


(** the complexity (of a literal),
    i.e. the value to be checked against the deepening bound. *)
type complexity = int



(**  Here all candidates are checked for {e exceeding} the iterative deepening bound.
     This includes determining if the current derivation branch is {e incomplete},
     i.e. if candidates have been dropped while constructing the branch,
     and (based on {!Flags.opt_restart}) backtracking to a choice point,
     where the branch is not incomplete.
     
     The actual bound criterion is handled internally according
     to {!Flags.opt_iterative_deepening}.
 *)
class type bound =
object
  (** {2 Bound} *)
  
  method current: complexity

  (** the deepening bound as a string. *)
  method to_string: string
    
  (** string representation of the current bound (ala {!Print.print_label}). *)
  method print : unit
    


  (** {2 Complexity} *)
    
  (** the complexity value of a literal. *)
  method get_complexity : literal -> complexity
    
  (** [bound_of_literal_subst literal subst offset]
      returns the complexity of the literal assembled by applying
      [subst] with [offset] to [literal]. *)
  method get_complexity_subst : literal -> subst -> int -> complexity
    
  (** usual comparison.
      Amounts to a compare over ints for term depth and term weight. *)
  method compare_complexity : complexity -> complexity -> int


    
  (** {2 Check} *)
    
  (** returns true, if the candidate exceeds the deepening bound. *)
  method exceeds : complexity -> bool
    
  (** returns true, if registering this candidate would have an impact
      wrt. to backtracking over an incomplete branch.
      This is never the case if the candidate does not exceed the depth bound.
	Otherwise, the details depend on the restart strategy ({!Flags.opt_restart}),
      e.g. for eager restarting only the lowest dropped complexity is of relevance.
  *)
  method exceeds_current : complexity -> choice_point -> bool
    
  (** like {!Bound.bound.exceeds},
      and additionally registers the candidate as dropped. *)
  method register : complexity -> choice_point -> bool
      
  (** like {!Bound.bound.register},
      but for the literal assembled by applying
	[subst] with [offset] to [literal]. *)
  method register_subst : literal -> subst -> int -> choice_point -> bool



  (** {2 Incompleteness} *)
    
    
  (** candidates have been dropped in the current derivation branch.
      This implies that the current branch can not be extended towards a model. *)
  method is_branch_incomplete : bool
    
  (** an incomplete branch is part of the derivation,
      but has been ignored while further exploring other derivation branches.
      This implies that the current derivation can not be extended towards a refutation. *)
  method is_derivation_incomplete : bool


  (** {2 Backtracking} *)
    
  (** when abandoning the current incomplete branch
      this determines the potential choice points to backtrack to,
      in order to continue the derivation
      according to the restarting strategy ({!Flags.opt_restart}).
      That is, all these choice points
      are responsible for dropping exceeding candidates
	and have to be invalidated.
  *)
  method restart_incomplete : choice_point list
    
  (** cleans all stored data and starts a fresh derivation with an increased depth bound.
      The new depth bound is the lowest depth bound of the dropped candidates. *)
  method restart : bound


  (** retracts information invalidated by backtracking some choice points. *)
  method backtrack : unit

  method jump_target: choice_point list
end

(** creates a specialization of object {!Bound.bound}
    based on {!Flags.opt_restart}.
    The initial bound is taken from {!Config.iterative_deepening}
    if not given.
 *)
val create: ?bound:int -> config -> bound
