(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type literal = Term.literal
type flags = Flags.flags


type config = {
  flags: flags;

  start_time: float;

  mutable resolve: bool;
  mutable subsume: bool;
  mutable compact: bool;
  mutable productivity: bool;
  mutable mixed_literals: bool;
  mutable lemma: bool;

  mutable equality: Flags.opt_equality;
  mutable plus_v: bool;
  mutable backtracking: Flags.opt_backtracking;
  mutable iterative_deepening: Flags.opt_iterative_deepening;
  mutable deepening_bound: int;
  mutable lookahead_exceeding: bool;
  mutable restart: Flags.opt_restart;
  mutable random_jump: bool;
  mutable neg_assert_candidates: Flags.opt_neg_assert_candidates;

  mutable time_out_CPU: int;
  mutable time_out_WC: int;

  mutable print_configuration: bool;
  mutable print_statistics: bool;
  mutable print_model_context: bool;
  mutable print_model_context_file: string;
  mutable print_model_DIG: bool;
  mutable print_model_DIG_file: string;
(*  mutable print_model_ARM: bool;*)
  mutable print_derivation_online: bool;
  (*
  mutable print_derivation: bool;
  mutable print_derivation_file: string;
  mutable print_derivation_pruned: bool;
  mutable print_derivation_pruned_file: string;
  mutable print_derivation_dot: string;
  mutable print_derivation_dot_pruned: string;
  *)
  mutable print_derivation_context_unifier: bool;
  mutable print_assert_candidates: bool;
  mutable print_split_candidates: bool;

  mutable problem_file_name: string;

  is_horn: bool;
  contains_equality: bool;

  (* for debugging only.
     some Context functions like is_satisfied, to_DIG
     rely on indexing. *)
  mutable term_indexing: bool;
}


let start_time config = config.start_time

let resolve config = config.resolve
let subsume config = config.subsume
let compact config = config.compact
let productivity config = config.productivity
let mixed_literals config = config.mixed_literals
let lemma config = config.lemma

let equality config = config.equality
let plus_v config = config.plus_v

let default_v config =
  if config.plus_v then
    Term.plus_v
  else
    Term.minus_v

let backtracking config = config.backtracking
let iterative_deepening config = config.iterative_deepening
let deepening_bound config = config.deepening_bound
let lookahead_exceeding config = config.lookahead_exceeding
let restart config = config.restart
let random_jump config = config.random_jump
let neg_assert_candidates config = config.neg_assert_candidates

let time_out_CPU config = config.time_out_CPU
let time_out_WC config = config.time_out_CPU

let print_configuration config = config.print_configuration
let print_statistics config = config.print_statistics
let print_model_context config = config.print_model_context
let print_model_context_file config = config.print_model_context_file
let print_model_DIG config = config.print_model_DIG
let print_model_DIG_file config = config.print_model_DIG_file
(*let print_model_ARM config = config.print_model_ARM*)
let print_derivation_online config = config.print_derivation_online
(*
let print_derivation config = config.print_derivation
let print_derivation_file config = config.print_derivation_file
let print_derivation_pruned config = config.print_derivation_pruned
let print_derivation_pruned_file config = config.print_derivation_pruned_file
let print_derivation_dot config = config.print_derivation_dot
let print_derivation_dot_pruned config = config.print_derivation_dot_pruned
*)
let print_derivation_context_unifier config = config.print_derivation_context_unifier
let print_assert_candidates config = config.print_assert_candidates
let print_split_candidates config = config.print_split_candidates

let is_horn config = config.is_horn
let contains_equality config = config.contains_equality
let term_indexing config = config.term_indexing

let problem_file_name config = config.problem_file_name

let drop_exceeding_eagerly config =
  (not config.lookahead_exceeding)
  &&
  (
    match config.restart with
      | Flags.RS_Eager
      | Flags.RS_Lazy ->
	  true
	    
      | Flags.RS_Delayed ->
	  false
  )



let create
  ~(flags: flags)
  ~(is_horn: bool)
  ~(contains_equality: bool)
  ~(start_time: float)
  : config =
  let get_value get =
    (get flags)#value
  in
  let config =
    {
      start_time = start_time;
      flags = flags;
      resolve = get_value Flags.resolve;
      subsume = get_value Flags.subsume;
      compact = get_value Flags.compact;
      productivity = get_value Flags.productivity;
      mixed_literals = get_value Flags.mixed_literals;
      lemma = false (*get_value Flags.lemma*);
      
      equality = get_value Flags.equality;
      plus_v = get_value Flags.plus_v;
      backtracking = get_value Flags.backtracking;
      iterative_deepening = get_value Flags.iterative_deepening;
      deepening_bound = get_value Flags.deepening_bound;
      lookahead_exceeding = get_value Flags.lookahead_exceeding;
      restart = get_value Flags.restart;
      random_jump = false (*get_value Flags.random_jump*);
      neg_assert_candidates = get_value Flags.neg_assert_candidates;

      time_out_CPU = get_value Flags.time_out_CPU;
      time_out_WC = get_value Flags.time_out_WC;

      print_configuration = get_value Flags.print_configuration;
      print_statistics = get_value Flags.print_statistics;
      print_model_context = get_value Flags.print_model_context;
      print_model_context_file = get_value Flags.print_model_context_file;
      print_model_DIG = get_value Flags.print_model_DIG;
      print_model_DIG_file = get_value Flags.print_model_DIG_file;
(*      print_model_ARM = (* :TODO: get_value Flags.print_model_ARM*) false;*)
      print_derivation_online = get_value Flags.print_derivation_online;
      (*
      print_derivation = get_value Flags.print_derivation;
      print_derivation_file = get_value Flags.print_derivation_file;
      print_derivation_pruned = get_value Flags.print_derivation_pruned;
      print_derivation_pruned_file = get_value Flags.print_derivation_pruned_file;
      print_derivation_dot = get_value Flags.print_derivation_dot;
      print_derivation_dot_pruned = get_value Flags.print_derivation_dot_pruned;
      *)
      print_derivation_context_unifier = get_value Flags.print_derivation_context_unifier;
      print_assert_candidates = get_value Flags.print_assert_candidates;
      print_split_candidates = get_value Flags.print_split_candidates;

      problem_file_name = flags.Flags.file_name;
      
      is_horn = is_horn;
      contains_equality = contains_equality;
      
      term_indexing = true;
    }
  in
    (* horn always uses -v *)
    if config.is_horn then begin
      (Flags.plus_v config.flags)#set_opt false;
      config.plus_v <- false;
    end;

    (* negative assert candidates can only be treated specially in the Horn case. *)
    if not (config.is_horn) then begin
       (Flags.neg_assert_candidates config.flags)#set_opt Flags.NAC_Use;
       config.neg_assert_candidates <- Flags.NAC_Use;
    end;

    (* random jumping makes only sense with lazy restarting. *)
(*    if not (config.restart == Flags.RS_Lazy) then begin
       (Flags.random_jump config.flags)#set_opt false;
       config.random_jump <- false;
    end;*)

    config








let print (config: config) =
  let print_flag = {
    Flags.apply =
      fun flag ->
	(* ignore unit options *)
	match flag#opt_to_string with
	  | [] ->
	      ()
		
	  | _ ->
	      Print.print_label flag#long_name (flag#value_to_string flag#value)
  }
  in
    print_endline ("Configuration:");
    List.iter
      (fun flag ->
	 Flags.process_flag flag print_flag
      )
      config.flags.Flags.flags;
    print_newline ();

    print_endline ("Problem:");
    Print.print_label "Horn" (string_of_bool config.is_horn);
    Print.print_label "Equality" (string_of_bool config.contains_equality);
    print_newline ()
