(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** configuration for a derivation
*)




(** {6 Types} *)

type literal = Term.literal
type flags = Flags.flags

type config





(** {6 Functions} *)


(** {2 Creation} *)

(** [create ~flags ~horn ~equality ~start_time]
    creates a new configuration initialized with [flags].
    The problem is considered to be Horn, if [is-horn] is true,
    to contain equality predicates, if [contains_equality] is true.
    The time (wall clock) when the derivation was started is given with [start_time].
*)
val create: flags:flags -> is_horn:bool -> contains_equality:bool -> start_time:float -> config

(** returns the (wall clock) time when the derivation was started. *)
val start_time : config -> float

(** {2 Options} *)

(** For an explanation of most of these settings see also module [Flags]. *)

(** apply resolve. *)
val resolve : config -> bool

(** apply subsume. *)
val subsume : config -> bool

(** apply compact. *)
val compact : config -> bool

(** only consider productive remainders. *)
val productivity : config -> bool

(** allow context literals containing universal and parametric variables. *)
val mixed_literals : config -> bool

(** compute and store lemmas. *)
val lemma : config -> bool

(** how should equality be handled? *)
val equality : config -> Flags.opt_equality

(** should {e +v} be used instead of {e -v} as the initial context literal? *)
val plus_v : config -> bool

(** returns {e +v} or {e -v}, depending on {!Config.plus_v}. *)
val default_v: config -> literal

(** the chosen backtracking method. *)
val backtracking : config -> Flags.opt_backtracking

(** the chosen iterative deepening method. *)
val iterative_deepening : config -> Flags.opt_iterative_deepening

(** the initial bound for the iterative deepening.
  The current bound is maintained in {!Bound.bound}. *)
val deepening_bound : config -> int

(** apply the close look-ahaed also to Assert candidates exceeding the current deepening bound. *)
val lookahead_exceeding : config -> bool

(** the chosen restarting strategy. *)
val restart : config -> Flags.opt_restart

(** do random jumps. *)
val random_jump : config -> bool

(** how to handle negative Assert candidate literals in the Horn case. *)
val neg_assert_candidates : config -> Flags.opt_neg_assert_candidates


(** can candidates exceeding the current deepening limit
  be immediately dropped after computation?
  (deduced from {!Config.lookahead_exceeding} and {!Config.restart}) *)
val drop_exceeding_eagerly: config -> bool

(** is this problem Horn? *)
val is_horn : config -> bool

(** does this problem contain equality? *)
val contains_equality : config -> bool

(** should term indexing be applied for context checks? *)
val term_indexing : config -> bool

(** the file name of the problem. *)
val problem_file_name: config -> string


(** system cpu timeout in seconds. *)
val time_out_CPU: config -> int

(** system cpu timeout in seconds. *)
val time_out_WC: config -> int


(** {2 Verboseness} *)

(** print the configuration settings. *)
val print_configuration : config -> bool

(** print a statistics after the derivation. *)
val print_statistics : config -> bool

(** print a found model as a context. *)
val print_model_context : config -> bool

(** like {!Config.print_model_context},
    but to the file with the given name. *)
val print_model_context_file : config -> string

(** print a found model as a DIG. *)
val print_model_DIG : config -> bool

(** like {!Config.print_model_DIG},
    but to the file with the given name. *)
val print_model_DIG_file : config -> string

(*(** print a found model as an ARM. *)
val print_model_ARM : config -> bool*)

(** print Assert, Split, and Close applications when they are applied
  during the derivation,
  plus stuff like the depth bound, backtracking, and restarts. *)
val print_derivation_online : config -> bool
(*
(** like {!Config.print_derivation_online},
    but after the derivation is finished. *)
val print_derivation : config -> bool

(** like {!Config.print_derivation},
    but to the file with the given name. *)
val print_derivation_file : config -> string

(** like {!Config.print_derivation},
    but split decisions without any impact on the derivation are omitted. *)
val print_derivation_pruned : config -> bool

(** like {!Config.print_derivation_pruned},
    but to the file with the given name. *)
val print_derivation_pruned_file : config -> string

(** like {!Config.print_derivation_file},
    but in the
    {{:http://www.graphviz.org/cgi-bin/man?dot} graphviz dot format}.

    inference steps are represented by different node colors:
    - Split -> green
    - Assert -> blue
    - Unit Split -> yellow
    - Close -> red
    - Incomplete (kind of Close) -> maroon (red-brown)
*)
val print_derivation_dot : config -> string

(** like {!Config.print_derivation_dot},
    but split decisions without any impact on the derivation are omitted. *)
val print_derivation_dot_pruned : config -> string
*)
(** print the corresponding context unifier when an application
  of Assert, Split, or Close is printed. *)
val print_derivation_context_unifier : config -> bool

(** print each computed applicable assert candidate. *)
val print_assert_candidates : config -> bool

(** print each computed applicable split candidate. *)
val print_split_candidates : config -> bool



(** {2 Representation} *)

(** string representation of the configuration (ala {!Print.print_label}). *)
val print: config -> unit
