(* $Id: context.ml,v 1.46 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)

type config = Config.config
type statistic = Statistic.statistic
type counter = Counter.counter
type symbol = Symbol.symbol
type var = Var.var
type term = Term.term
type clause = Term.clause
type literal = Term.literal
type subst =  Subst.subst
type choice_point = State.choice_point
type state = State.state


type element = {
  el_id: int;
  el_literal: literal;
  el_choice_point: choice_point;
  el_pars: var list;
  el_generation: int;
  el_mandatory_at: int;
  mutable el_compacted: int;
}

(* for representing a model as a DIG
   (Disjunctions of Implicit Generalizations) *)
type dig = {
  (* the positive term *)
  term: term;
  (* the negative exceptions *)
  exceptions: term list;
}

(*** representation ***)

let element_to_string (element: element) : string =
(*  "choice_point:    " ^ State.choice_point_to_string element.el_choice_point ^ "\n"
  ^*) "literal: " ^ Term.literal_to_string element.el_literal ^ "\n"


(* context elements are stored in a term index *)
class index_data =
object
  method is_equal (first: element) (second: element) : bool =
    first == second
      
  method to_string (element: element) : string =
    element_to_string element
end
type predicate_index = element Term_indexing.predicate_index

(* a term index for each sign/predicate combination *)
type index = element Term_indexing.index


(* the context is stored in
- a list for easy backtracking
- an index for fast term checks *)
type context = {
  (* environment *)
  statistic: statistic;
  config: config;
  state: state;

  (* the id counter for the context literals/elements *)
  id_counter: counter;

  (* the max. size of the context over the whole derivation *)
  mutable max_size: int;

  (* the context elements stored in a list *)
  mutable context: element list;

  (* term indexing over the context. an index for each sign/predicate pair *)
  index: index;
}

let get_max_id context =
  match context.context with
    | [] ->
	-1

    | head :: _ ->
	head.el_id

let get_max_prev_id context =
  match context.context with
    | _ :: [] ->
	-1

    | _ :: head :: _ ->
	head.el_id

    | _ ->
	failwith "get_max_prev_id"


let next_el context i =
  if i = -1 then
    let rec do' l =
      match l with
	| [] -> failwith "next_el 0"
      | h :: [] ->
	  h.el_id

      | _ :: l ->
	  do' l
    in
      do' context.context
  else
  let rec do' l =
    match l with
      | x :: y :: tail ->
	  if y.el_id = i then
	    x.el_id
	  else
	    do' (y :: tail)

      | _ ->
	  failwith "next_el"
	    
  in
    do' context.context

(*** globals ***)

(* negative ids are reserved for special global elements *)
let unique_element_id_counter =
  ref (-1)


let get_unique_element_id () : int =
  unique_element_id_counter := !unique_element_id_counter - 1;
  !unique_element_id_counter


(*** constants ***)

let null_element : element =
  {
    el_id = get_unique_element_id ();
    el_literal = Term.null_literal;
    el_pars = [];
    el_generation = 0;
    el_choice_point = State.invalid_choice_point;
    el_compacted = -1;
    el_mandatory_at = -1;
  }






(*** creation ***)

(* the context does not explicitely contain -v/+v,
   this information has to be retrieved from config *)
let create (config: config)
    (statistic: Statistic.statistic) (state: state) : context =

  (* empirical: discrimination trees for Horn, substitution tree for non-Horn *)
  let index =
    if Config.is_horn config then
      Discrimination_tree.create_index (new index_data)
    else
      Substitution_tree.create_index (new index_data)
  in
    {
      config = config;
      statistic = statistic;
      state = state;
      id_counter = Counter.create_with 0;
      max_size = 0;
      context = [];
      index = index;
    }





(*** representation ***)

let print_context ?(show_compacted: bool = false) (context: context) (out: out_channel) : unit =
(*  print_endline (State.active_choice_point_to_string context.state ^ "Context: ");*)
  Print.output_line out (Term.literal_to_string (Config.default_v context.config));
  List.iter
    (fun element ->
       if
	 show_compacted
	 ||
	 match element.el_compacted with
	   | -1 -> true
	   | _ -> false
       then
	 Print.output_line out (Term.literal_to_string element.el_literal);
    )
    context.context




let update_max_size (context: context) : unit =
  context.max_size <- max context.max_size (List.length context.context)


let print_max_size (context: context) : unit =
  update_max_size context;
  Print.print_statistic "Maximum Context Size" context.max_size




(*** iteration ***)

let iter (context: context) (func: element -> unit) : unit =
  List.iter func context.context

let fold (context: context) (func: 'a -> element -> 'a) (acc: 'a) : 'a =
  List.fold_left func acc context.context






(*** access ***)

let is_equal (first: element) (second: element) : bool =
  first == second



let parameters_of_literal (literal: literal) : var list =
  List.find_all
    Var.is_parametric
    (Term.vars_of_literal literal)


(* compact the context with literal *)
let compact (context: context) (index: predicate_index) (literal: literal) : unit =
  (* compact context with the new literal *)
  if not (Config.compact context.config) then begin
    ()
  end

  else begin
    (* find all compacted elements *)
    let compacted_elements =
      if Config.term_indexing context.config then
	index#find_all_instances ~p_preserving:true literal.Term.term
      else
	List.find_all
	  (fun element ->
	     try
	       ignore (Unification.match_literals ~p_preserving:true
			 literal 0
			 element.el_literal 1);
	       true
	     with
	       | Unification.UNIFICATION_FAIL ->
		   false
	  )
	  context.context
    in
      (* mark all uncompacted elements compacted *)
      List.iter
	(fun compacted_element ->
	   if compacted_element.el_compacted >= 0 then
(*	   match compacted_element.el_compacted with
	     | Some _ ->
		 ()

	     | None ->*)
		 (*
	   if not
	     (List.exists
		(fun compacted_at ->
		   State.depends_on context.config (State.active_choice_point context.state) compacted_at
		)
		compacted_element.el_compacted
	     )
	   then*) begin
	     Statistic.inc_compact context.statistic;
(*	     compacted_element.el_compacted <- (State.active_choice_point context.state) :: compacted_element.el_compacted;*)
(*	     compacted_element.el_compacted <- Some (State.active_choice_point context.state);*)
	     compacted_element.el_compacted <- State.id_of_choice_point (State.active_choice_point context.state);
	     
	     (* remove compacted literals from the term index *)
(*	     if Config.term_indexing context.config then begin
	       if not (index#remove compacted_element.el_literal.Term.term (Some compacted_element)) then
		 failwith "Context.compact"
	     end;*)
	   end
	)
	compacted_elements
  end


let add (context: context) (literal: literal) (mandatory_at: int)
  (generation: int): element =

  let new_context_element = {
    el_id = Counter.next context.id_counter;
    el_literal = literal;
    el_choice_point = State.active_choice_point context.state;
    el_pars = parameters_of_literal literal;
    el_generation = generation;
    el_mandatory_at = mandatory_at;
    el_compacted = -1;
  }
  in
(*    print_endline ("EL_ID: " ^ string_of_int new_context_element.el_id);*)
  (* get the index where the literal must be inserted *)
  let index =
    context.index#find new_context_element.el_literal
  in

    (* compact makes only sense for universal literals *)
    begin
      match new_context_element.el_pars with
	| [] -> compact context index literal
	| _ -> ()
    end;

    (* add new literal to context and index *)
    context.context <- new_context_element :: context.context;

    if Config.term_indexing context.config then begin
      index#add new_context_element.el_literal.Term.term new_context_element;
    end;

    if Const.debug then begin
      context.index#iter
	(fun index ->
	   index#iter
	   (fun term element ->
	      if not (Term.term_equal term element.el_literal.Term.term) then begin
		print_endline ("\n\n\nFAIL=====\n");
		print_endline (index#to_string);
		print_endline ("ADD: " ^ Term.literal_to_string new_context_element.el_literal);
		print_endline ("Term             : " ^ Term.term_to_string term);
		print_endline ("But node contains: " ^ element_to_string element);
		failwith "Context.add"
	      end
	   )
	)
    end;
    
    new_context_element








(*** checks ***)

(*
let is_element_mandatory_for_choice_point (config: config) (element: element) (candidate_choice_point : choice_point) : bool =
  (* the candidate's choice point is dependent on the conflicting element *)
  State.depends_on config candidate_choice_point element.el_choice_point
  ||
  (* the element is mandatory in a choice point on which the candidate's choice point is dependent *)
  match element.el_mandatory_at with
    | None ->
	false

    | Some mandatory_choice_point ->
	State.depends_on config candidate_choice_point mandatory_choice_point
*)

let is_element_mandatory_for_choice_point (config: config) (element: element) (candidate_choice_point: int) : bool =
  (* candidate was computed after element was added to the context *)
  candidate_choice_point >= element.el_id
  ||
  (* candidate was computed after element's candidate was computed,
     and element's candidate is mandatory (Assert or Unit Assert)
  *)
  (
    element.el_mandatory_at >= 0
    &&
    candidate_choice_point >= element.el_mandatory_at
  )

(*
  (* the candidate's choice point is dependent on the conflicting element *)
  State.depends_on config candidate_choice_point element.el_choice_point
  ||
  (* the element is mandatory in a choice point on which the candidate's choice point is dependent *)
  match element.el_mandatory_at with
    | None ->
	false

    | Some mandatory_choice_point ->
	State.depends_on config candidate_choice_point mandatory_choice_point
*)


(*** contradictory ***)

let check_contradictory_no_indexing (context: context) (literal: literal) : element option =

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  if literal.Term.sign = element.el_literal.Term.sign then
	    do_check tail

	  else begin
	    try
	      ignore (Unification.unify_terms ~p_preserving:true
			element.el_literal.Term.term 0
			literal.Term.term 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  do_check tail
	  end
  in
    do_check context.context


let check_contradictory_indexing (context: context) (literal: literal) : element option =
  let index =
    context.index#find (Term.request_negated_literal literal)
  in
    index#find_unifiable ~p_preserving:true literal.Term.term


let check_contradictory (context: context) (literal: literal) : element option =
  if Config.term_indexing context.config then
    check_contradictory_indexing context literal
  else
    check_contradictory_no_indexing context literal





(*** subsumed ***)


let check_subsumed_no_indexing (context: context) (literal: literal) : element option =

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  if literal.Term.sign <> element.el_literal.Term.sign then
	    do_check tail

	  else begin
	    try
	      ignore (Unification.match_terms ~p_preserving:true
			element.el_literal.Term.term 0
			literal.Term.term 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  do_check tail
	  end
  in
    do_check context.context


let check_subsumed_indexing (context: context) (literal: literal) : element option =
  let index =
    context.index#find literal
  in
    index#find_generalization ~p_preserving:true literal.Term.term


let check_subsumed (context: context) (literal: literal) : element option =
  if Config.term_indexing context.config then
    check_subsumed_indexing context literal
  else
    check_subsumed_no_indexing context literal







(*** productive ***)


let check_productive_no_indexing (context: context) (producer: literal) (produced: literal) : element option =
  let neg_producer =
    Term.request_negated_literal producer
  and neg_produced =
    Term.request_negated_literal produced
  in

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  (* producer must be ignored *)
	  if Term.are_literals_variants producer element.el_literal then
	    do_check tail
	      
	  (* produced already part of the context -> remainder subsumed/produced by the context *)
	  else if Term.are_literals_variants produced element.el_literal then begin
	    Some element
	  end

	  else begin
	    (* msg *)
	    try
	      ignore (Unification.match_literals producer 0 element.el_literal 1);
	      ignore (Unification.match_literals element.el_literal 0 neg_produced 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  begin
		    (* instance *)
		    try
		      (* produced might be a mixed non-remainder literal - but that's of no importance for match *)
		      ignore (Unification.match_literals element.el_literal 0 produced 1);

		      (* special treatment for -v - any subsuming literal is an instance of -v *)
		      if Term.literal_equal producer (Config.default_v context.config) then begin
			Some element;
		      end


		      (* productivity is currently not implemented for mixed literals *)
		      else if Config.mixed_literals context.config then begin
			None
		      end

		      else begin
			match element.el_pars with
			  | [] ->
			      (* non-parametric, so must be unifiable *)
			      ignore (Unification.unify_literals neg_producer 0 element.el_literal 1);
			      Some element;
			      
			  | _ when
			      not (Config.mixed_literals context.config) ->
			      (* parametric, so must be instance *)
			      ignore (Unification.match_literals neg_producer 0 element.el_literal 1);
			      Some element;

			  | _ ->
			      (* mixed with parameters *)
			      let subst =
				Unification.unify_literals neg_producer 0 element.el_literal 1
			      in
				if
				  List.for_all
				    (fun par ->
					 match Subst.get_ subst par 1 with
					   | None ->
					       true

					   | Some term ->
					       begin
						 (* bound to another term of element *)
						 if term.Subst.st_offset = 1 then
						   false
						 else
						   match term.Subst.st_term with
						     | Term.Var bound_var ->
							 not (Subst.exists
							   (fun binding ->
							      (* binding of another parameter from the element
								 to the same parameter of the producer ->
								 no p-renaming *)
							      (binding.Subst.sb_term == term)
							      &&
							      (binding.Subst.sb_var.Subst.sv_offset == 0)
							      &&
							      (not (Var.equal binding.Subst.sb_var.Subst.sv_var par))
							      &&
							      (Var.is_parametric binding.Subst.sb_var.Subst.sv_var)
							   )
							   subst)

						     | _ ->
							 (* bound to a non-parameter *)
							 false
					       end
				    )
				    element.el_pars
				then
				  Some element
				else
				  raise Unification.UNIFICATION_FAIL
		      end
		    with
		      | Unification.UNIFICATION_FAIL ->
			  do_check tail
		  end
	  end
  in

    do_check context.context



let check_productive_indexing (context: context) (producer: literal) (produced: literal) : element option =

  (* check second part of productivity *)
  let do_check_productive () =
    let index_produced =
      context.index#find produced
    in
      
      (* if the producer is v then any literal more general than produced is shielding *)
      if Term.literal_equal producer (Config.default_v context.config) then begin
	index_produced#find_generalization ~p_preserving:false produced.Term.term
      end
	
      (* for not mixed literals a simple shielding check is sufficient. *)
      else if not (Config.mixed_literals context.config) then begin
	index_produced#find_shielding producer.Term.term produced.Term.term
	end
	
      (* for mixed literals a shieldling literal must also not be shielded itself. *)
      else begin
	let iterator =
	  index_produced#get_shielding_iterator producer.Term.term produced.Term.term
	in
	  (* not shielded *)
	  if iterator#is_empty then
	    None
	      
	  else
	    let index_producer =
	      context.index#find producer
	    in
	    let rec loop iterator =
	      if iterator#is_empty then
		None
		  
	      else
		let shielding_element =
		  iterator#next
		in
		let shield_for_shielding_element =
		  index_producer#find_shielding
		    shielding_element.el_literal.Term.term
		    produced.Term.term
		in
		  match shield_for_shielding_element with
		    | None ->
			Some shielding_element
			  
		    | Some _ ->
			loop iterator
	    in
	      loop iterator
      end
  in
    
  (* check first part of productivity, i.e. msg *)
    let index_producer =
      context.index#find producer
    in
      begin
	match index_producer#find_strongly_shielding producer.Term.term produced.Term.term with
	  | None ->
	      do_check_productive ()
	      
	  | Some element as conflict ->
	      conflict
      end


let check_productive (context: context) (producer: literal) (produced: literal) : element option =

  if not (Config.productivity context.config) then begin
    None
  end

  else if Term.are_terms_variants producer.Term.term produced.Term.term then begin
    None
  end

  (* if the producer is universal and the context is compacted is must produce the instance. *)
  else if not (Term_attributes.is_term_parametric producer.Term.term) then begin
    None
  end

  else begin
    let check =
      if Config.term_indexing context.config then begin
	check_productive_indexing context producer produced
      end
      else begin
	check_productive_no_indexing context producer produced
      end
    in
      begin
	match check with
	  | None ->
	      ()
	      
          | Some _ ->
	      Statistic.inc_filtered_by_productivity context.statistic;
      end;
      check
  end




(* if there are no more general literals in the context,
   and the literal has the same polarity as the default interpretation,
   and if there is no more general literal of opposite polarity,
   then the literal is satisfied. *)
let is_satisfied_v (context: context) (literal: literal) : bool =
  if (Config.default_v context.config).Term.sign = literal.Term.sign then begin
    let index =
      context.index#find (Term.request_negated_literal literal)
    in
      match index#find_generalization ~p_preserving:false literal.Term.term with
	| None ->
	    true
	      
	| Some _ ->
	    false
  end
    
  (* nope, opposite polarity, so the literal is not satisfied. *)
  else begin
    false
  end



(*
  a literal is satisfied by the interpretation induced by the context
  if it is positive and produced by the context,
  or if it is negative and its negation is not produced by the context.
*)
let rec is_satisfied (context: context) (literal: literal) : bool =
  let index =
    context.index#find literal
  in
    
  (* is there a context literal more general than literal,
     i.e. a context literal that potentially produces literal? *)
  let iterator =
    index#get_generalization_iterator ~p_preserving:false literal.Term.term
  in
    (* there is no more general literal - except perhaps for -v *)
    if iterator#is_empty then begin
      is_satisfied_v context literal
    end
	
    (* there are more general literals - is one a producer? *)
    else begin

      let rec loop iterator =
	if iterator#is_empty then
	  (* no producer found *)
	  false
	    
	else
	  let more_general_element =
	    iterator#next
	  in
	    (* is the literal actually in of the context? *)
	    if
	      Term.are_literals_variants
		more_general_element.el_literal
		literal
	    then
	      true

	    else

	    (* p_preserving more general? *)
	    match more_general_element.el_pars with
	      | [] ->
		  (* yes, so the literal is produced for sure. *)
		  true
		    
	      | _ ->
		  (* no, so do the productivity check. *)
		  let is_productive =
		    check_productive
		      context
		      (Term.request_negated_literal more_general_element.el_literal)
		      literal
		  in
		    begin
		      match is_productive with
			| None ->
			    (* ok, for a positive literal ... *)
			    if literal.Term.sign then
			      true

			    (* ... but for a negative literal assure
			       that its negation is not satsified as well *)
			    else begin
			      not (
				is_satisfied
				  context
				  (Term.request_negated_literal literal)
			      )
			    end
			      
			| Some _ ->
			    loop iterator
		    end
      in
	loop iterator
    end
		





(*** representation ***)


let from_file (file_name: string)
  (config: config) (statistic: statistic) (state: state) : context =
  let context =
    create config statistic state
  in

  let in_channel =
    open_in file_name
  in

  let rec read_literals () =
    try
      let line =
	input_line in_channel
      in
      let literal =
	Read_darwin.to_literal line
      in
	ignore (add context literal (-1) 0);
	read_literals ()
    with
      | End_of_file ->
	  ()
  in
    read_literals ();
    close_in in_channel;
    context







(*** Backtracking ***)

(* remove all context literal added in invalid choice points *)
let backtrack_elements (context: context) : unit =

  let remove_from_index (element: element) : unit =
    if Config.term_indexing context.config then begin
(*      (* compacted literals are not part of the index *)
      if
	match element.el_compacted with
	  | -1 -> true
	  | _ -> false
      then*) begin
	let index =
	  context.index#find element.el_literal
	in
	let removed =
	  index#remove
	    element.el_literal.Term.term
	    (Some element)
	in
	  if Const.debug then begin
	    context.index#iter
	      (fun index ->
		 index#iter
		 (fun term element ->
		    if not (Term.term_equal term element.el_literal.Term.term) then begin
		      print_endline ("\n\n\nFAIL=====\n");
		      print_endline (index#to_string);
		      print_endline ("REMOVE: " ^ Term.literal_to_string element.el_literal);
		      print_endline ("Term             : " ^ Term.term_to_string term);
		      print_endline ("But node contains: " ^ element_to_string element);
		      failwith "Context.backtrack"
		    end
		 )
	      )
	  end;
	  
	  if Const.debug && not removed then begin
	    failwith "Context.backtrack index II"
	  end
      end    
    end;
  in
(*
  let pruned =
    State.prune
      context.state
      (fun element -> element.el_choice_point)
      (fun element -> remove_from_index element)
      oldest_retracted
      context.context
  in
    context.context <- pruned
*)
    
  (* all invalidated choice points are the at the beginning of the context *)
  let rec backtrack_chronological () =
    match context.context with
      | [] ->
	  ()
	    
      | element :: tail ->
	  if State.is_choice_point_invalid element.el_choice_point then begin
	    context.context <- tail;
	    remove_from_index element;
	    backtrack_chronological ();
	  end
  in
    backtrack_chronological ()(*;
    Counter.set context.id_counter (get_max_id context)*)


(* uncompact all context literal compacted in invalid choice points *)
let backtrack_compact (context: context) (retracted: int) : unit =
  List.iter
    (fun context_element ->
       if State.is_choice_point_invalid context_element.el_choice_point then begin
	 failwith (
	   "Context.backtrack_compact: invalid context literal "
	   ^ Term.literal_to_string context_element.el_literal
	   ^ " in the context."
	 );
       end

       else 
	 if
	   (context_element.el_compacted > 0)
	   &&
	   (context_element.el_compacted >= retracted)
	 then begin
(*	 match context_element.el_compacted with
	   | Some choice_point when
	       State.is_choice_point_invalid choice_point ->*)

(*	       context_element.el_compacted <- None;*)
	       context_element.el_compacted <- -1;
(*	       List.find_all State.is_choice_point_valid context_element.el_compacted;*)
	       
	       if Config.term_indexing context.config then begin
(*		 match context_element.el_compacted with
		   | [] ->*)
		       (* get the index where the literal must be inserted *)
		       let index =
			 context.index#find context_element.el_literal
		       in
			 index#add context_element.el_literal.Term.term context_element
			   
(*		   | _ ->
		       ()*)
	       end
(*
	   | _ ->
	       ()*)
		 
       end
	 
    )
    context.context
    

    
let backtrack (context: context) (retracted: int) : unit =
  update_max_size context;

  (* first remove all invalid literals *)
  backtrack_elements context;
  (* now undoe the compacting of still valid literals caused by now invalid literals *)
  backtrack_compact context retracted
















(*** DIG ***)


(* in the dig representation a model is represented by positive terms
   with a set of negative terms.
   For -v, this is quite convenient.
   For +v, this implies that per default the most general version
   of a predicate (called schema term) is true,
   if not overridden by the context.
   That is, for the predicate p of arity 3 per default p(u, v, w) is true,
   if not -p(u, v, w) is in the context.
   For a schema term the discrimination into universal and parametric variables
   is of no importance.
*)


(* is this a schema term? *)
let is_schema (term: term) : bool =
  let terms =
    match term with
      | Term.Func (_, terms, _) -> terms
      | _ -> [| |]
  in
    try
      ignore (
	Array.fold_left
	  (fun acc term ->
	     match term with
	       | Term.Var var ->
		   (* check that this variable has not been used before.*)
		   let id =
		     Var.id_of_var var
		   in
		     if List.exists ((==) id) acc then
		       raise Exit
		     else
		       id :: acc
			 
	       | _ ->
		   raise Exit
	  )
	  []
	  terms
      );
      true
    with
      | Exit ->
	  false

(* create the schema term for a symbol. *)
let create_schema (predicate: symbol) : term =
  if Symbol.arity predicate = 0 then
    Term.request_const predicate

  else
    (* create the most general term,
       i.e. a term with a fresh parameter at each position *)
    let terms =
      Array.make (Symbol.arity predicate) Term.null_term
    in
      for i = 0 to Symbol.arity predicate - 1 do
      terms.(i) <- Term.request_var (Var.create_parametric i)
      done;
      Term.request_func (predicate, terms)

(* create the schema term for a term. *)
let create_schema_from_term (term: term) : term =
  match term with
    | Term.Const _ ->
	(* consts are most general *)
	term

    | Term.Func (symbol, _, _) ->
	create_schema symbol

    | Term.Var _ ->
	failwith "Context.create_schema"





(* is term subsumed by any term in terms? *)
let is_subsumed (term: term) (terms: term list) : bool =
  List.exists
    (fun old ->
       try
	 ignore (
	   Unification.match_terms
	     old 0
	     term 1
	 );
	 true
       with
	 | Unification.UNIFICATION_FAIL ->
	     false
    )
    terms

(* return the terms not subsumed by term. *)  
let subsume (term: term) (terms: term list) : term list =
  let subsume =
    [term]
  in
    List.find_all
      (fun term ->
	 not (is_subsumed term subsume)
      )
      terms


(* is term subsumed by any dig in digs? *)
let dig_is_subsumed (term: term) (digs : dig list) : bool =
  List.exists
    (fun dig ->
       (* is the dig term more general? *)
       try
	 ignore (
	   Unification.match_terms
	     dig.term 0
	     term 1
	   );
	 (* yes, is also none of its exceptions more general? *)
	 not (is_subsumed term dig.exceptions)
       with
	 | Unification.UNIFICATION_FAIL ->
	     false
    )
    digs
  

(* return the digs not subsumed by dig *)  
let dig_subsume (dig: dig) (digs: dig list) : dig list =
  let subsume =
    [dig]
  in
    List.find_all
      (fun dig ->
	 not (dig_is_subsumed dig.term subsume)
      )
      digs





(* get the dig representation of all predicates
   which are contained in the context as positive terms.
   for these, the needed schema terms are added as well.
*)
let get_digs (context: context) : dig list =

  (* get all digs *)
  context.index#fold
    (fun acc literal index ->

       (* get all positive elements for this predicate. *)
       if literal.Term.sign then begin
	 let exception_index =
	   context.index#find
	     (Term.request_negated_literal literal)
	 in

	 let terms =
	   index#fold
	     (fun acc term _ ->
		term :: acc
	     )
	     []
	 in
	   
	 (* check if schema term has to be added *)
	 let terms =
	   (* assure that the positive schema literal is added for +v *)
	   if Config.plus_v context.config then begin
	     (* schema already included in positive terms? *)
	     if List.exists is_schema terms then
	       terms

	     (* schema already included in negative terms? *)
	     else if
	       try
		 exception_index#iter
		   (fun term _ ->
		      if is_schema term then raise Exit
		   );
		 false
	       with
		 | Exit ->
		     true
	     then
	       terms

	     (* nope, so explicitly add the schema term *)
	     else begin
	       create_schema_from_term literal.Term.term :: terms
	     end
	   end

	   (* nothing to do for -v *)
	   else begin
	     terms
	   end

	 in

	 (* assemble the digs by adding the exceptions to the terms. *)
	 let digs =
	   List.fold_left
	     (fun acc term ->
		(* redundant, ignore this term. *)
		if dig_is_subsumed term acc then begin
		  acc
		end

		else begin
		  (* get all exceptions of the term. *)
		  let exceptions =
		    List.map
		      (fun element -> element.el_literal.Term.term)
		      (exception_index#find_all_instances ~p_preserving:false term)
		  in

		  (* remove subsumed exceptions. *)
		  let pruned_exceptions =
		    List.fold_left
		      (fun acc current ->
			 if is_subsumed current acc then
			   acc
			     
			 else
			   current :: subsume current acc
		      )
		      []
		      exceptions
		  in

		  (* create the dig *)
		  let dig = {
		    term = term;
		    exceptions = pruned_exceptions;
		  }
		  in
		    (* remove subsumed digs *)
		    dig :: dig_subsume dig acc
		end
	     )
	     []
	     terms
	 in


	   (* add this predicate's digs to all digs *)
	   digs @ acc
       end
	 
       (* ignore negative literals *)
       else begin
	 acc
       end
    )
    []


(* assure that all positive schema terms are added for +v *)
let add_schemas (context: context) (digs: dig list) (clauses: clause list) : dig list =
  if Config.plus_v context.config then begin
    let predicates =
      Equality.get_predicate_symbols clauses
    in

      List.fold_left
	(fun acc predicate ->
	   (* predicate already processed when creating the digs
	      from the context? *)
	   if
	     List.exists
	       (fun dig ->
		  match dig.term with
		    | Term.Const symbol
		    | Term.Func (symbol, _, _) ->
			Symbol.equal predicate symbol
			  
		    | _ ->
			false
	       )
	       digs
	   then
	     digs

	   else begin
	     let schema_term =
	       create_schema predicate
	     in
	     let exception_index =
	       context.index#find
		 (Term.request_literal false schema_term)
	     in
	       (* schema already included in negative terms? *)
	       if
		 try
		   exception_index#iter
		     (fun term _ ->
			if is_schema term then raise Exit
		     );
		   false
		 with
		   | Exit ->
		       true
	       then
		 digs
		   
	       (* add the schema term *)
	       else
		 (* get all exceptions of the term. *)
		 let exceptions =
		   List.map
		     (fun element -> element.el_literal.Term.term)
		     (exception_index#find_all_instances ~p_preserving:false schema_term)
		 in
		 
		 (* remove subsumed exceptions. *)
		 let pruned_exceptions =
		   List.fold_left
		     (fun acc current ->
			if is_subsumed current acc then
			  acc
			    
			else
			  current :: subsume current acc
		     )
		     []
		     exceptions
	       in
		   
		 (* create the dig *)
		 let dig = {
		   term = schema_term;
		   exceptions = exceptions;
		 }
		 in
		   dig :: digs
	   end
	)
	digs
	predicates
  end
    
  (* nothing to do for -v *)
  else begin
    digs
  end



let print_DIG (context: context) (out: out_channel) (clauses: clause list) : unit =
  let digs =
    get_digs context
  in
  let digs =
    add_schemas context digs clauses
  in
    (* finally print the digs.
       parameters have no business here,
       so use the universal form of the literals *)
    List.iter
      (fun dig ->
	 match dig.exceptions with
	   | [] ->
	       Print.output_line out
		 (Term.term_to_string
		    (Term.request_universal_term dig.term))
		 
	   | _ ->
	       Print.output_line out
		 (Term.term_to_string
		    (Term.request_universal_term dig.term)
		  ^ " \\ { ");
	       List.iter
		 (fun term ->
		    Print.output_line out
		      ("  " ^ Term.term_to_string
			 (Term.request_universal_term term));
		 )
		 dig.exceptions;
	       Print.output_line out " }"
      )
      digs
