(* $Id: context.mli,v 1.19 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** context with operations

    Applies the {e Compact} rule of the model evolution calculus.
*)

(** {6 Types} *)

type config = Config.config
type statistic = Statistic.statistic
type counter = Counter.counter
type var = Var.var
type term = Term.term
type literal = Term.literal
type clause = Term.clause
type choice_point = State.choice_point
type state = State.state


(** a context literal *)
type element = {
  el_id: int;
  (** unique id per context literal. *)

  el_literal: literal;
  (** the context literal. *)

  el_choice_point: choice_point;
  (** the choice point in which the context literal was added. *)

  el_pars: var list;
  (** the parameters of the context literal. *)

  el_generation: int;
  (** depth of the tree of context literals
      needed to computed this context literal. *)

  el_mandatory_at: int;
  (** the last context literal used to compute the context unifier
      of this context literal. *)
  (** this context element has been asserted in this choice point
      and must therefore be subsumed by the context in this choice point
      and all choice points depending on it. *)

  mutable el_compacted: int;
  (** the earliest choice point in which a literal was added to the context
      which compacts this literal. *)
}


type context



(** {6 Functions} *)




(** {2 Initialization} *)

(** empties the context and fixes the indexing method based. *)
val create: config -> statistic -> state -> context




(** {2 Constants} *)

(** for initialization usage. *)
val null_element: element





(** {2 Access} *)

(** get an id which is not used for any other context literal/element. *)
val get_unique_element_id: unit -> int


(** [add context literal mandatory_at generation]
  adds a new [literal] to the [context] *)
val add: context -> literal -> int -> int -> element


val is_equal: element -> element -> bool



(** {2 Iteration} *)

val iter: context -> (element -> unit) -> unit

val fold: context -> ('a -> element -> 'a) -> 'a -> 'a



(** {2 Checks} *)

(** [check_contradictory context literal]
  checks if [literal] is contradictory with the [context].
  Returns a contradictory context element, if one exits. *)
val check_contradictory: context -> literal -> element option

(** [check_subsumed context literal]
  checks if [literal] is subsumed by the [context].
  Returns a subsuming context element, if one exits. *)
val check_subsumed: context -> literal -> element option

(** [check_productive context producer produced]
    checks if the negation of [producer]
    is producing [produced] in the [context]
    (that is [producer] is the context literal,
    and [produced] the remainder literal).
    Returns a conflicting context element, if one exists. *)
val check_productive: context -> literal -> literal -> element option

(** [check_valid context literal]
  Returns true, if the literal is satisfied by the [context]. 
    
  :NOTE: this is not complete but merely a prefilter with linear complexity,
  the full test is coNP complete.
 *)
val is_satisfied: context -> literal -> bool



(** {2 Properties} *)

(** is the context literal mandatory for [choice_point]?
  I.e. the literal must be part of the context in [choice_point]
  or there must be a candidate which asserts literal to the context in [choice_point].
*)
val is_element_mandatory_for_choice_point: config -> element -> int -> bool



(** {2 Representation} *)

val element_to_string: element -> string


(** [from_file file_name]
  creates a context from a file.
  The file should contain one literal per line, readable by {!Read_tme}.

  @raise Sys_error on file access failure.

  @raise Failure if the file contains invalid input. *)
val from_file: string ->
  config -> statistic -> state -> context


(** print the (not compacted) context literals, one literal per line.
    Compacted literals are only shown when [show_compacted] is true
    (by default it is false). *)
val print_context: ?show_compacted:bool -> context -> out_channel -> unit

(** prototypical implementation of the DIG representation.
    Works currently only for {e -v} *)
val print_DIG: context -> out_channel -> clause list -> unit


(** string representation of the maximimal context size
    during the derivation (ala {!Print.print_label}). *)
val print_max_size: context -> unit



(** {2 Backtracking} *)

(** removes context elements added in invalid choice points,
    and undoes the corresponding compacting of the context.
    [choice_point] is the oldest choice point retracted
    since the last backtracking.
*)
val backtrack: context -> int -> unit

(* returns-10 if no element exists, its id starting from 1 otherwise *)
val get_max_id: context -> int
val get_max_prev_id: context -> int
val next_el: context -> int -> int
