(* $Id: context_unifier.ml,v 1.35 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type choice_point = State.choice_point
type state = State.state
type packed_subst = Unification.packed_subst

(** when a literal is added to the context
    new context unifiers can be added in several ways. *)
type candidate_type =
  | Close

  | Assert
      (** compute only Assert and Close context unifiers. *)
      
  | Split
      (** compute only Split context unifiers. *)

  | CloseAssert

  | All
      (** compute all context unifiers. *)


type context_partner = {
  cp_element: Context.element;
  mutable cp_partial_context_unifier: packed_subst;
}

type context_partners =
    context_partner array

type input_partner = {
  ip_index: int;
  ip_literal: literal;
  ip_vars: Subst.var list;
  mutable ip_context_partners: context_partner list;
  mutable ip_context_partners_length: int;
(*  mutable ip_context_partners_changed: bool;*)
  mutable ip_resolved: context_partner option;
}

type context_unifier_space = {
  cus_state: state;
  cus_id: int;
  cus_clause: clause;
  cus_shared_vars: Subst.var list;
  cus_input_partners: input_partner array;
  cus_input_partners_ordering: int array;
(*  mutable cus_subsumed: State.choice_point option;*)
  cus_process_close_candidate: (raw_context_unifier -> unit);
  cus_process_assert_candidate: (raw_context_unifier -> bool);
  cus_process_split_candidate: (raw_context_unifier -> unit);
  cus_lemma: bool;
  cus_value: int ref;

(*  mutable cus_backtrack: bool;*)
}


and raw_context_unifier = {
  rcu_space: context_unifier_space;
  rcu_context_partners: context_partners;
}





exception CLOSE of raw_context_unifier





(*** constants ***)


let assert_partner = {
  cp_element = {
    Context.el_id = Context.get_unique_element_id ();
    Context.el_literal = Term.assert_literal;
    Context.el_choice_point = State.valid_choice_point;
    Context.el_pars = [];
    Context.el_generation = 0;
    Context.el_mandatory_at = -1;
    Context.el_compacted = -1;
  };
  cp_partial_context_unifier = [| |];
}

let null_partner = {
  cp_element = {
    Context.el_id = Context.get_unique_element_id ();
    Context.el_literal = Term.null_literal;
    Context.el_choice_point = State.invalid_choice_point;
    Context.el_pars = [];
    Context.el_generation = 0;
    Context.el_mandatory_at = -1;
    Context.el_compacted = -1;
  };
  cp_partial_context_unifier = [| |];
}





(*** misc ***)



let get_input_clause (context_unifier_space: context_unifier_space) : clause =
  context_unifier_space.cus_clause


let get_context_literals (context_partners: context_partners) : literal array =
  (* create a clause where the i.th element is the context literal
     paired to the i.th literal in the input clause *)
(*
  Array.fold_right
    (fun context_literal acc ->
       context_literal.cp_element.Context.el_literal :: acc
    )
    context_partners
    []
  *)
  Array.map
    (fun context_literal ->
       context_literal.cp_element.Context.el_literal
    )
    context_partners




(*** representation ***)


let context_unifier_to_string
  (context_unifier_space: context_unifier_space) (context_partners: context_partners)
  : string =
  let context_literals =
    Array.to_list (
      Array.map
	(fun context_literal ->
	   string_of_int context_literal.cp_element.Context.el_id
	   ^ ": "
	   ^ Term.literal_to_string context_literal.cp_element.Context.el_literal
	)
	context_partners
    )
  in
  (
    if context_unifier_space.cus_lemma then
      "Lemma : "
    else
      "Clause : "
  )
  ^ string_of_int context_unifier_space.cus_id ^ ": " ^ Term.clause_to_string (get_input_clause context_unifier_space)
    ^ "\n"
    ^ "Context: "
    ^ "["
    ^ String.concat ", " context_literals ^ "]"
(*  ^ Term.clause_to_string
    (
      Array.to_list (get_context_literals context_partners)) *)
    
  ^ "\n"


let raw_context_unifier_to_string (raw_context_unifier: raw_context_unifier) : string =
  context_unifier_to_string
    raw_context_unifier.rcu_space
    raw_context_unifier.rcu_context_partners


















(*** misc ***)


(* as long as the choice points of all its context literals are valid,
   the context unifier is valid *)
let is_raw_context_unifier_invalid (raw_context_unifier: raw_context_unifier) : bool =
  Tools.array_exists
    (fun context_partner ->
       State.is_choice_point_invalid context_partner.cp_element.Context.el_choice_point
    )
    raw_context_unifier.rcu_context_partners


let recompute_unifier (raw_context_unifier: raw_context_unifier) : subst =

  let rec recompute_at (i: int) (acc: subst) : subst =
    if i >= Array.length raw_context_unifier.rcu_space.cus_input_partners then
      acc

    else
      let input_index =
	raw_context_unifier.rcu_space.cus_input_partners.(i).ip_index
      in
      let context_partner =
	raw_context_unifier.rcu_context_partners.(input_index)
      in
      let merged_subst =
	try
	  Unification.unify_packed_substs acc context_partner.cp_partial_context_unifier
	with
	  | Unification.UNIFICATION_FAIL ->
	      print_endline ("CONTEXT_UNIFIER.RECOMPUTE_UNIFIER");
	      print_endline (Subst.subst_to_string acc);
	      print_endline (Subst.subst_to_string (Subst.unpack_subst context_partner.cp_partial_context_unifier));
	      failwith "Context_unifier.recompute_unifier"
      in
	recompute_at (i + 1) merged_subst
	  
  in
    recompute_at 0 Subst.empty




(* is this pair producing a remainder literal?
   yes, if one of its parameters is bound to a non-parameter *)
let is_remainder (subst: subst) (pars: var list) (offset: int) =

  let par_offset =
    Subst.context_literal_offset offset
  in

    List.exists
      (fun par ->
	 match Subst.get_ subst par par_offset with
	   | None -> 
	       false
	       
	   | Some term ->
	       begin
		 match term.Subst.st_term with
		   | Term.Var var ->
		       Var.is_universal var

		   | _ -> 
		       true
	       end
      )
      pars





let compute_remainder_states (raw_context_unifier: raw_context_unifier) (subst: subst) : bool array =
  (* no map function as the arrays have a different order *)
  let remainder_states =
    Array.make (Array.length raw_context_unifier.rcu_context_partners) false
  in

    for i = 0 to Array.length raw_context_unifier.rcu_space.cus_input_partners - 1 do
      let input_index =
	raw_context_unifier.rcu_space.cus_input_partners.(i).ip_index
      in
      let context_partner =
	raw_context_unifier.rcu_context_partners.(input_index)
      in
	remainder_states.(input_index) <-
	  is_remainder
	  subst
	  context_partner.cp_element.Context.el_pars
	  input_index;
    done;

    remainder_states




(* a generation of a candidate is
   the max generation of its context_partners + 1 *)
let generation_of_candidate (context_partners: context_partners) : int =
  let max_generation =
    Array.fold_left
      (fun acc context_partner ->
	 Tools.max_int acc context_partner.cp_element.Context.el_generation
      )
      0
      context_partners
  in
    max_generation + 1


let get_creation_choice_point (context_partners: context_partners) : choice_point =
  let youngest_choice_point =
    Array.fold_left
      (fun acc context_partner ->
	 match acc with
	   | Some best_choice_point when
	       State.compare_age best_choice_point context_partner.cp_element.Context.el_choice_point >= 0 ->
	       acc

	   | _ ->
	       Some context_partner.cp_element.Context.el_choice_point
      )
      None
      context_partners
  in
    match youngest_choice_point with
      | None ->
	  failwith "get_creation_choice_point"

      | Some choice_point ->
	  choice_point

let get_creation_cid (context_partners: context_partners) : int =
(*  print_endline ("get_creation_cid");*)
  let youngest_choice_point =
    Array.fold_left
      (fun acc context_partner ->
	 (*
	 print_endline ("acc: " ^ string_of_int acc);
	 print_endline ("cp : " ^ string_of_int context_partner.cp_element.Context.el_id);
	 print_endline ("cp : " ^ Term.literal_to_string context_partner.cp_element.Context.el_literal);
	 *)
	 if acc >= context_partner.cp_element.Context.el_id then
	   acc

	 else
	   context_partner.cp_element.Context.el_id
      )
      0
      context_partners
  in
    youngest_choice_point
      (* if all context literals are _assert_ or -v,
	 all ids are < 0, but then 0 (the root level) is the right level
    match youngest_choice_point with
      | -1 ->
	  print_endline (string_of_int youngest_choice_point);
	  failwith "get_creation_cid"

      | choice_point ->
	  choice_point
      *)

let compare_context_unifiers (first: raw_context_unifier) (second: raw_context_unifier) : int =

  (* as first and second are context unifiers of the same clause
     they must have the same number of context partners *)
  let rec cmp_context_partners (index: int) =
    if index >= Array.length first.rcu_context_partners then begin
      (* a candidate should never be compared with itself *)
      print_endline (raw_context_unifier_to_string first);
      failwith "Context_unifier.compare_context_unifiers";
      0
    end
    else
      let cmp =
	compare
	  first.rcu_context_partners.(index).cp_element.Context.el_id
	  second.rcu_context_partners.(index).cp_element.Context.el_id
      in
	if cmp <> 0 then
	  cmp
	else
	  cmp_context_partners (index + 1)
  in

    if first == second then
      0
    else
      let cmp_clause =
	compare first.rcu_space.cus_id second.rcu_space.cus_id
      in
	if cmp_clause <> 0 then
	  cmp_clause
	else if
	  Const.debug
	  &&
	  (Array.length first.rcu_context_partners != Array.length second.rcu_context_partners)
	then
	  failwith "Context_unifier.cmp_context_partners"
	else
	  cmp_context_partners 0





