(* $Id: context_unifier.mli,v 1.25 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** basic types for and operations on context unifiers

    contains also the types for the search for context unifiers,
    but this is actually done in
    {!Context_unifier_space} and {!Context_unifier_search}.
*)


(** {6 Types} *)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type choice_point = State.choice_point
type state = State.state
type packed_subst = Unification.packed_subst

(** when a literal is added to the context
    new context unifiers can be added in several ways. *)
type candidate_type =
    (* Close is always done first,
       explicit or implicit in broader searches *)
  | Close
      (** compute only Close context unifiers. *)

  | Assert
      (** compute only Assert context unifiers. *)
      
  | Split
      (** compute only Split context unifiers. *)

  | CloseAssert
      (** compute only Close and Assert context unifiers. *)

  | All
      (** compute all context unifiers. *)


(** a context literal paired to an input clause literal. *)
type context_partner = {
  cp_element: Context.element;
  (** the context literal. *)
  mutable cp_partial_context_unifier: packed_subst;
  (** the partial context unifier of this context literal
      and the corresponding input literal. *)
}
    
(** the context literals paired with an input clause.
    A {!Context_unifier.input_partner} is paired with the context literal
    at [context_partners.(input_partner.ip_index)].
    This ensures that [context_partners] remains valid
    even if the [input_partners] of a [context_unifier_space],
    i.e. the literals of a clause,
    are reordered. *)
type context_partners =
    context_partner array


(** a literal of an input clause, with some properties
    and the results of its unification with all context literals. *)
type input_partner = {
  ip_index: int;
  (** the literal is the i.th's literal of the clause
    (the actual order in the datastructure is dynamic during the computation) *)

  ip_literal: literal;
  (** the input literal *)

  ip_vars: Subst.var list;
  (** the variables of the literal *)

  mutable ip_context_partners: context_partner list;
  mutable ip_context_partners_length: int;
(*  mutable ip_context_partners_changed: bool;*)
  (** all current context literals (negated) unifiable with this input literal *)

  (* resolved by this context literal *)
(*  mutable ip_resolved: int;*)

  mutable ip_resolved: context_partner option;
  (** all current (from each other independent) context literals
      resolving this input literal.
      I.e. at most one in chronological backtracking.*)
}

(** represents an input clause and its relationship with the context,
i.e. unifiable literal pairs, ...

  See {!Context_unifier.context_unifier_space}. *)
type context_unifier_space = {
  (** environment *)
  cus_state: state;

  cus_id: int;
  (** unique id per clause *)

  cus_clause: clause;
  (** the input clause *)  

  cus_shared_vars: Subst.var list;
  (** the variables occuring in several clause literals. *)

  cus_input_partners: input_partner array;
  (** the clause literals - may be reordered. *)

  cus_input_partners_ordering: int array;

(*  mutable cus_subsumed: choice_point option;*)
  (** subsumed by context literals from these choice points. *)

  cus_process_close_candidate: (raw_context_unifier -> unit);
  (** to be called on each found closing context unifier
     (see {!Context_unifier_check.check_close}). *)

  cus_process_assert_candidate: (raw_context_unifier -> bool);
  (** to be called on each found assert candidate
     (see {!Context_unifier_check.check_assert}). *)

  cus_process_split_candidate: (raw_context_unifier -> unit);
  (** to be called on each found split candidate
      (see {!Context_unifier_check.check_split}). *)

  cus_lemma: bool;

(*  mutable cus_backtrack: bool;*)

  cus_value: int ref;
}


(** the minimal information needed to recompute a context unifier.

  See {!Context_unifier.context_unifier_space}. *)
and raw_context_unifier = {
  rcu_space: context_unifier_space;
  (** the clause of the context unifier. *)

  rcu_context_partners: context_partners;
  (** the context literals of the context unifier. *)
}



(** raised if {e Close} is triggered by the closing context unifier. *)
exception CLOSE of raw_context_unifier



(** {6 Functions} *)


(** {2 Constants} *)

(** used for initialization. *)
val null_partner: context_partner

(** used as the gap marker in an 'assert context unifier'. *)
val assert_partner: context_partner




  




  

(** {2 Misc} *)

(** has the raw_context_unifier become invalid after backtracking? *)
val is_raw_context_unifier_invalid: raw_context_unifier -> bool

(** recomputes the substitution of the [raw_context_unifier]. *)
val recompute_unifier: raw_context_unifier -> subst

(** [is_remainder context_unifier context_literal_parameters input_literal_index].
  does the context literal,
  which contains the parameters [context_literal_parameters],
  and is paired with the [input_literal_index].th input literal
  in the [context_unifier],
  produce a remainder literal?
  I.e., is one of its parameters is bound to a non-parameter? *)
val is_remainder: subst -> var list -> int -> bool

(** returns true for each literal instantiated by the context unifier
    which is in the remainder.
    The entry at i corresponds to the literal produced with
    the input literal with [ip_index] = i. *)
val compute_remainder_states: raw_context_unifier -> subst -> bool array

(** depth of the dependency tree of the context literals
    needed to compute this context unifier. *)
val generation_of_candidate: context_partners -> int

(** the context literals of the context partners
    in the same order as paired to the clause
    in its initial order. *)
val get_context_literals: context_partners -> literal array

(** the choice point in which this unifier has been computed *)
val get_creation_choice_point: context_partners -> choice_point
val get_creation_cid: context_partners -> int

(** a total ordering on context unifiers
    based on the clause and context literal ids. *)
val compare_context_unifiers: raw_context_unifier -> raw_context_unifier -> int


(** {2 Representation} *)

val raw_context_unifier_to_string: raw_context_unifier -> string

val context_unifier_to_string: context_unifier_space -> context_partners -> string
