(* $Id: context_unifier_check.ml,v 1.6 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type config = Config.config
type bound = Bound.bound
type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type packed_subst = Unification.packed_subst
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type raw_context_unifier = Context_unifier.raw_context_unifier



(*** checks ***)


let check_close
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  : raw_context_unifier option =

  let rec empty_remainder (i: int) : bool =
    if i >= Array.length context_partners then
      true
    else begin
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
	if
	  Context_unifier.is_remainder 
	    subst 
	    context_partner.Context_unifier.cp_element.Context.el_pars
	    input_index
	then
	  false
	else
	  empty_remainder (i + 1)
    end
  in

    if empty_remainder 0 then begin
      let closing_context_unifier = {
	Context_unifier.rcu_space = context_unifier_space;
	Context_unifier.rcu_context_partners = Array.copy context_partners;
      }
      in
	Some closing_context_unifier
    end

    else
      None

let check_assert
  (config: config)
  (bound: bound)
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners) :
  raw_context_unifier option =

  (* the assert literal must be parameter free *)
  let is_assert_valid_at (i : int) : bool =
    let assert_vars =
      Subst.get_bound_vars
	subst
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_vars
    in
      List.for_all 
	(fun var ->
	   Var.is_universal var.Subst.sv_var
	)
	assert_vars
  in

  (* has the unifier an empty remainder
     and leaves the assert literal parameter free? *)
  let rec is_valid_at (i: int) (assert_index: int option) : bool * int option =
    if i >= Array.length context_partners then begin
      true, assert_index
    end

    else begin
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
	(* no parameter in the assert literal? *)
	if context_partner == Context_unifier.assert_partner then begin
	  if is_assert_valid_at i then
	    is_valid_at (i + 1) (Some i)
	  else begin
	    false, assert_index
	  end
	end

	(* empty remainder? *)
	else begin
	  if
	    Context_unifier.is_remainder 
	      subst 
	      context_partner.Context_unifier.cp_element.Context.el_pars
	      input_index
	  then begin
	    false, assert_index
	  end
	
	  (* yes, so continue *)
	  else
	    is_valid_at (i + 1) assert_index
	end
    end
  in

    match is_valid_at 0 None with
      | false, _ ->
	  None

      | true, None ->
	  failwith "Context_unifier_space.check_assert"

      | true, Some assert_index ->
	  let assert_clause_literal =
	    context_unifier_space.Context_unifier.cus_input_partners.(assert_index).Context_unifier.ip_literal
	  in
	    (* grep special Horn candidates, i.e. let positive candidates pass,
	       and negative only when used unrestricted. *)
	    if 
	      (not assert_clause_literal.Term.sign)
	      &&
		(Config.neg_assert_candidates config = Flags.NAC_Ignore)
	    then begin
	      failwith "Context_unifier_check.check_assert: neg_assert";
	    end;
(*
(*	      Config.is_horn config
	      &&*)
	      match Config.neg_assert_candidates config with
		| Flags.HC_Ignore
		| Flags.HC_Lookahead ->
		    not assert_clause_literal.Term.sign

		| Flags.HC_Use ->
		    false
	    then begin
	      if Config.horn_candidates config = Flags.HC_Ignore then
		None

	      else
		let raw_context_unifier = {
		  Context_unifier.rcu_space = context_unifier_space;
		  Context_unifier.rcu_context_partners = Array.copy context_partners;
		}
		in
		  Some raw_context_unifier
	    end

	    (* is the candidate within the deepending bound? *)
	    else*) if
	      Config.drop_exceeding_eagerly config
	      &&
	      bound#register_subst
		assert_clause_literal
		subst
		Subst.input_literal_offset
		(Context_unifier.get_creation_choice_point context_partners) 
	    then
	      None

	      (* either this candidate is obeying the bound
		 or all candidates are kept for later processing. *)
	    else
	      let raw_context_unifier = {
		Context_unifier.rcu_space = context_unifier_space;
		Context_unifier.rcu_context_partners = Array.copy context_partners;
	      }
	      in
		Some raw_context_unifier



let check_split
  (config: config)
  (bound: bound)
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  : raw_context_unifier option =

  let rec min_complexity_at (i: int) (acc: Bound.complexity option) : Bound.complexity option =
    (* all remainder literals exceed the bound? *)
    if i >= Array.length context_partners then begin
      acc
    end

    else
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
      let remainder =
	Context_unifier.is_remainder 
	  subst
	  context_partner.Context_unifier.cp_element.Context.el_pars
	  input_index
      in
	if remainder then begin
	  let clause_literal =
	    context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_literal
	  in
	  let complexity =
	    bound#get_complexity_subst
	      clause_literal
	      subst
	      Subst.input_literal_offset
	  in
	    if bound#exceeds complexity then begin
	      let new_acc =
		match acc with
		  | Some min_complexity when
		      bound#compare_complexity min_complexity complexity <= 0 ->
		      
		  acc
		    
		  | _ ->
		      Some complexity
	      in
		min_complexity_at (i + 1) new_acc
	    end

	    else begin
	      raise Exit
	    end
	end

	else
	  min_complexity_at (i + 1) acc
  in
    
    if Config.drop_exceeding_eagerly config then begin
      try
	match min_complexity_at 0 None with
	  | None ->
	      (* no remainder literal?!?! *)
	      print_endline (Context_unifier.context_unifier_to_string
			       context_unifier_space context_partners);
	      failwith "Context_unifier_check.check_split.min_complexity_at I"

	  | Some complexity ->
	      if
		bound#register
		  complexity
		  (Context_unifier.get_creation_choice_point context_partners) 
	      then
		None
		  
	      else
		(* min complexity is within the bound?!? *)
		failwith "Context_unifier_check.check_split.min_complexity_at II"
      with
	| Exit ->
	    (* valid complexity found *)
	    let raw_context_unifier = {
	      Context_unifier.rcu_space = context_unifier_space;
	      (* the old array is reused and altered, so copy it *)
	      Context_unifier.rcu_context_partners = Array.copy context_partners;
	    }
	    in
	      Some raw_context_unifier
    end	    

    else
      let raw_context_unifier = {
	Context_unifier.rcu_space = context_unifier_space;
	(* the old array is reused and altered, so copy it *)
	Context_unifier.rcu_context_partners = Array.copy context_partners;
      }
      in
	Some raw_context_unifier

