(* $Id: context_unifier_search.ml,v 1.3 2004/12/28 14:06:47 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type counter = Counter.counter
type config = Config.config
type bound = Bound.bound
type var =  Var.var
type literal =  Term.literal
type clause =  Term.clause
type subst =  Subst.subst
type state = State.state
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner
type raw_context_unifier = Context_unifier.raw_context_unifier
type candidate_type = Context_unifier.candidate_type



(* bundle the arguments propagate through all functions. *)
type assemble = {
  (* environment *)
  config: config;

  (* environment *)
  bound: bound;

  (* the clause to search for context unifiers,
     plus some information *)
  context_unifier_space: context_unifier_space;

  ordering: int array;

  (* the pairings of context literals to clause literals
     in the partial context unifier.
     the context literal at index i is paired
     with the clause literal with index ip_index = i.
     depending on the current state of merging the partial context unifiers
     only the first pairs may make sense. *)
  context_partners: context_partners;

  (* the new context literal. *)
  active_partner: context_partner;

  (* the index of the clause literal paired with the new context literal. *)
  fixed_index: int;

  (* closing context unifier found? *)
  mutable close_found: bool;
}


(* the module structure serves just compile time optimization
   of differences of searching for assert or split context unifiers *)

module type Search = 
  sig
    val search_context_unifiers:
      config ->
      bound ->
      context_unifier_space ->
      context_partners ->
      int ->
      context_partner ->
      subst ->
      bool ->
      bool ->
      bool
  end

module type Candidate_type =
sig
  val t: candidate_type
end

module Close : Candidate_type =
struct
  let t = Context_unifier.Close
end

module Assert : Candidate_type =
struct
  let t = Context_unifier.Assert
end

module Split : Candidate_type =
struct
  let t = Context_unifier.Split
end

module CloseAssert : Candidate_type =
struct
  let t = Context_unifier.CloseAssert
end

module All : Candidate_type =
struct
  let t = Context_unifier.All
end


(* functor to build a search module *)
module Make (Candidate_type: Candidate_type) : Search =
struct

  (*** process completed context unifiers ***)
  
  
  (* a context unifier is completed,
     now check if it's valid for Close, Assert, Split,
     and send it to the selection module. *)
  let process_context_unifier
      ~(assert_unifier: bool)
      ~(non_empty_remainder: bool)
      (assemble: assemble)
      (subst: subst)
      : unit =
(*
    print_endline ("check: " );
    print_endline (
    Context_unifier.context_unifier_to_string
      assemble.context_unifier_space
	assemble.context_partners);
*)  
    (* an assert unifier? *)
    if
      not assemble.close_found
      &&
(*      not assemble.context_unifier_space.Context_unifier.cus_lemma &&*)
      match Candidate_type.t with
	| Context_unifier.Assert
	| Context_unifier.CloseAssert
	| Context_unifier.All ->
	    assert_unifier
	      
	| Context_unifier.Close
	| Context_unifier.Split ->
	    false
    then begin
      let candidate =
	Context_unifier_check.check_assert
	  assemble.config
	  assemble.bound
	  subst
	  assemble.context_unifier_space
	  assemble.context_partners 
      in
	match candidate with
	  | None ->
	      ()
	    
	  | Some candidate ->
	      (* propagate the candidate *)
	      if assemble.context_unifier_space.Context_unifier.cus_process_assert_candidate candidate then begin
		(* this assert candidate closes when applied (detected via Selection_assert and lookahead *)
		assemble.close_found <- true;
	      end;
    end

    (* split or close? *)
    else begin

      (* close? *)
      if
	(* close has been checked in the previous search for assert unifiers,
	   or the remainder is already known not to be empty. *)
	(not non_empty_remainder)
(*	&&
	(not assert_unifier)*)
	&&
	(
	  match Candidate_type.t with
	    | Context_unifier.Close
	    | Context_unifier.CloseAssert
	    | Context_unifier.All ->
		true
		  
	    | Context_unifier.Assert
	    | Context_unifier.Split ->
		false
	)
	&&
	begin
	  (* check close *)
	  (* Note: if non_empty_remainder is false,
	     this does not imply that the remainder is empty.
	     After all, non_empty_remainder is just based on testing
	     the extension of the partial context unifier,
	     not the whole remainder.
	     Example: clause {p(x), q(x)} with context literals -p(u), -q(a)
	     does yield non_empty_remainder = false,
	     but has a non-empty remainder.
	  *)
	  let candidate =
	    Context_unifier_check.check_close
	      subst assemble.context_unifier_space assemble.context_partners
	  in
	    match candidate with
	      | None ->
		  false
		    
	      | Some candidate ->
		  (* propagate the candidate *)
		  assemble.close_found <- true;
		  assemble.context_unifier_space.Context_unifier.cus_process_close_candidate candidate;
		  true
	end
      then
	(* yep, closing context unifier found *)
	()
	  
      (* no, so check for split? *)
      else if not assemble.context_unifier_space.Context_unifier.cus_lemma then begin
	match Candidate_type.t with
	  | Context_unifier.Close
	  | Context_unifier.CloseAssert
	  | Context_unifier.Assert ->
	      ()
		
	  | Context_unifier.Split
	  | Context_unifier.All ->
	      let candidate =
		Context_unifier_check.check_split
		  assemble.config
		  assemble.bound
		    subst
		  assemble.context_unifier_space
		  assemble.context_partners
	      in
		begin
		  match candidate with
		    | None ->
			(* dropped at depth bound *)
			()
			  
		    | Some candidate ->
			(* propagate the candidate *)
			assemble.context_unifier_space.Context_unifier.cus_process_split_candidate candidate;
		end
      end;
    end







  (*** search for context unifiers ***)


  (* to avoid extending partial context unifiers with non-empty remainders
     when only assert and close unifiers are targeted,
     the remainder state of every extending context literal is checked for.
     
     if it corresponds to a remainder literal,
     the context unifier has a non-empty remainder,
     and extending the partial context unifier makes no sense.

     this pre-check does actually pay off.
  *)
  let update_non_empty_remainder (config: config) (non_empty_remainder: bool)
      (subst: subst) (context_partner: context_partner) (input_partner: input_partner)
      : bool =
    
    (* Horn never produces non-empty remainders *)
    not (Config.is_horn config)
    &&
      (
	non_empty_remainder
	||
	  Context_unifier.is_remainder 
	  subst
	  context_partner.Context_unifier.cp_element.Context.el_pars
	  input_partner.Context_unifier.ip_index
      )



  (* is the potential context_partner always compacted
     wrt. the new context partner
     when considering the dependency relation? *)
(*
  let is_compacted (config: config) (state: state) (context_partner: context_partner) : bool =
    
    match context_partner.Context_unifier.cp_element.Context.el_compacted with
      | -1 -> false
      | _ ->  true
*)

  let is_compacted (context_partner: context_partner) (current: context_partner) : bool =
    
    context_partner.Context_unifier.cp_element.Context.el_compacted >= 0
    &&
    context_partner.Context_unifier.cp_element.Context.el_compacted < current.Context_unifier.cp_element.Context.el_id

(*
let is_resolved
    (input_partner: input_partner) (context_element: Context.element) : bool =
  
  input_partner.Context_unifier.ip_resolved >= 0
  &&
  input_partner.Context_unifier.ip_resolved < context_element.Context.el_id
*)

(* build all possible unifiers.

   the clause may have been reordered,
   so the literal currently at position i
   may have ip_index <> i.

   the literals are considered from first to last.

   - assert_unifier:
     this is an assert unifier,
     i.e. one literal is not paired with a context literal,
     but with the pseudo context partner Context_unifier.assert_partner

   - non_empty_remainder:
     the partial context unifier produces at least one remainder literal
     (false does _not+ imply that the remainder is empty)

   - current_index:
     try to pair the clause literal currently at this position
     with a context literal at this step.
     invariant: all clause literals at a lower position have already been paired.

   - subst: the partial context unifier
 *)
let rec assemble_context_unifier
    ~(assert_unifier: bool)
    ~(non_empty_remainder: bool)
    (assemble: assemble)
    (current_index: int)
    (subst: subst)
    : unit =

  (* the context unifier is complete. *)
  if current_index >= Array.length assemble.context_unifier_space.Context_unifier.cus_input_partners then begin
    process_context_unifier
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      subst
  end

  (* ignore the fixed literal -
     it has been paired in the initialization with the new context literal. *)
  else if current_index = assemble.fixed_index then begin
    assemble_context_unifier
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      (current_index + 1)
      subst
  end

  else begin
    (* try extend from here to an assert context unifier. *)
    assemble_assert
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      current_index
      subst;
    
    (* and also try to extend from here to a split/close context unifier. *)
    assemble_split
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      current_index
      subst
  end


and assemble_assert
    ~(assert_unifier: bool)
    ~(non_empty_remainder: bool)
    (assemble: assemble)
    (current_index: int)
    (subst: subst)
    : unit =
  
  if
    (* don't build assert unifiers for the split only search *)
    (match Candidate_type.t with
       | Context_unifier.Assert
       | Context_unifier.CloseAssert
       | Context_unifier.All ->
	   true

       | Context_unifier.Close
       | Context_unifier.Split ->
	   false
    )
    &&
    (* this is already an assert unifier *)
    (not assert_unifier)
    &&
    (* no point in trying to build an assert unifier with a non empty remainder *)
    (not non_empty_remainder)
    &&
    (* ignore negative assert literals if wished. *)
    (not (
      (not (assemble.context_unifier_space.Context_unifier.cus_input_partners.(assemble.ordering.(current_index)).Context_unifier.ip_literal.Term.sign))
      &&
      (Config.neg_assert_candidates assemble.config = Flags.NAC_Ignore)
    )
    )
    &&
    (* no closing candidate found yet *)
    (not assemble.close_found)
  then begin
    let input_partner =
      assemble.context_unifier_space.Context_unifier.cus_input_partners.(assemble.ordering.(current_index))
    in
      (* mark this literal as the assert candidate *)
      assemble.context_partners.(input_partner.Context_unifier.ip_index) <-
        Context_unifier.assert_partner;
      
      assemble_context_unifier
	~assert_unifier:true
	~non_empty_remainder:non_empty_remainder
	assemble
	(current_index + 1)
	subst
  end


(* checks if the next literal to choose is resolved.
   raise Not_found otherwise *)
and assemble_resolved_split
    ~(assert_unifier: bool)
    ~(non_empty_remainder: bool)
    (assemble: assemble)
    (current_index: int)
    (subst: subst)
    : unit =
 
  let input_partner =
    assemble.context_unifier_space.Context_unifier.cus_input_partners.(assemble.ordering.(current_index))
  in

  (* is this input literal resolved? *)
  let resolving_partner =
    match input_partner.Context_unifier.ip_resolved with
      | Some resolving_partner when
	  resolving_partner.Context_unifier.cp_element.Context.el_id
	  <
	  assemble.active_partner.Context_unifier.cp_element.Context.el_id
	  ->
	  resolving_partner

      | _ ->
	  raise Not_found
  in
(*
    if 
      (match Candidate_type.t with
	 | Context_unifier.Assert
	 | Context_unifier.All ->
	     true
	 | Context_unifier.Split ->
	     false
      )
    then
      match input_partner.Context_unifier.ip_resolved with
	| None -> raise Not_found
	| Some context_partner -> context_partner
	    

    else
      let active_choice_point =
	State.active_choice_point assemble.context_unifier_space.Context_unifier.cus_state
      in
	match input_partner.Context_unifier.ip_resolved with
	  | None -> raise Not_found

	  | Some resolve_partner ->
	      begin
	     (* resolve does not change the rest of the clause.
		thus, the dependency relation is sufficient for all cases,
		even when Candidate_type.t = Context_unifier.Split,
		i.e. when only split context unfiers are searched for.
		even though resolve may have already been applied
		when searching for assert candidates with context literals,
		which are strictly seen not available yet for the split search,
		this only leads to the loss of some context unifiers.
		and these would have been obsolete later on anyway,
		when resolving leads to new shorter remainders.
	     *)
(*	     State.depends_on
	       assemble.config
	       active_choice_point
	       resolve_partner.Context_unifier.cp_element.Context.el_choice_point
*)
		if

	     match Candidate_type.t with
	       | Context_unifier.Split ->
		   (* resolve only with older context literals. *)
		   (resolve_partner.Context_unifier.cp_element.Context.el_id
		   <
		   assemble.active_partner.Context_unifier.cp_element.Context.el_id)

		   &&

		   State.depends_on
		     assemble.config
		     active_choice_point
		     resolve_partner.Context_unifier.cp_element.Context.el_choice_point

	       | _ ->
		   false
		then
		  resolve_partner
		else
		   raise Not_found
	      end
  in*)
    assemble.context_partners.(input_partner.Context_unifier.ip_index) <- resolving_partner;
    
    (* as this substitution is only used to check for close
       and is recomputed and made admissible,
       before it is actually used for computing a remainder,
       the resolving substitution can be ignored *)
    assemble_context_unifier
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      (current_index + 1)
      subst

  
(* try all context literals of the selected queue 
   in order to extend the current context unifier *)
and assemble_split
    ~(assert_unifier: bool)
    ~(non_empty_remainder: bool)
    (assemble: assemble)
    (current_index: int)
    (subst: subst)
    : unit =

  (* is the current clause literal resolved? *)
  try
    assemble_resolved_split
      ~assert_unifier:assert_unifier
      ~non_empty_remainder:non_empty_remainder
      assemble
      current_index
      subst
  with
    | Not_found ->
	(*
  if assemble.context_unifier_space.Context_unifier.cus_lemma then begin
    ()
  end

  else
	*)
	(* no, so find a context partner *)
	let input_partner =
	  assemble.context_unifier_space.Context_unifier.cus_input_partners.(assemble.ordering.(current_index))
	in
	  List.iter
	    (fun (context_partner: context_partner) ->
	       
	       (* ignore compacted context literals *)
	       if is_compacted
(*		 assemble.config
		 assemble.context_unifier_space.Context_unifier.cus_state*)
		 context_partner
		 assemble.active_partner
	       then
		 ()

	       (* when decoupling the search for split context unifiers
		  we have to assure that only context literals older
		  than the currently processed context literal are used. *)
	       else if
		 match Candidate_type.t with
		   | Context_unifier.Close
		   | Context_unifier.CloseAssert
		   | Context_unifier.All ->
		       false

		   | Context_unifier.Assert
		   | Context_unifier.Split ->
		       (* only consider older context literals *)
		       (context_partner.Context_unifier.cp_element.Context.el_id
			>
			assemble.active_partner.Context_unifier.cp_element.Context.el_id)
		       ||
		       (* or the same context literal occuring earlier in the clause *)
		       (
			 (context_partner.Context_unifier.cp_element.Context.el_id
			  =
			  assemble.active_partner.Context_unifier.cp_element.Context.el_id)
			 &&
			 (assemble.ordering.(current_index) >= assemble.ordering.(assemble.fixed_index))
		       )

	       then
		 ()

	       else begin
		 (* try to extend the unifier *)
		 try
		   let extended_subst =
		     Unification.unify_packed_substs
		       subst
		       context_partner.Context_unifier.cp_partial_context_unifier
		   in
		     
		   let new_non_empty_remainder =
		     update_non_empty_remainder
		       assemble.config
		       non_empty_remainder
		       extended_subst
		       context_partner
		       input_partner
		   in
		     
		     (* a non empty remainder is not interesting for... *)
		     if
		       new_non_empty_remainder
		       &&
		       (
			 (* find only assert and close *)
			 Candidate_type.t = Context_unifier.Close
			 ||
			 Candidate_type.t = Context_unifier.Assert
			 ||
			 Candidate_type.t = Context_unifier.CloseAssert
			 ||
			 (* partial assert context unifier to be extended *)
			 assert_unifier
			 ||
			 (* find only assert and close for lemmas *)
			 assemble.context_unifier_space.Context_unifier.cus_lemma
			 ||
			 (* find only close *)
			 (assemble.close_found)
		       )
		     then
		       ()

		     else if
		       assemble.close_found
		       &&
		       assert_unifier
		     then
		       ()

		     (* finally, extend the partial context unifier and continue *)
		     else begin
		       assemble.context_partners.(input_partner.Context_unifier.ip_index) <- context_partner;
		       
		       assemble_context_unifier
			 ~assert_unifier:assert_unifier
			 ~non_empty_remainder:new_non_empty_remainder
			 assemble
			 (current_index + 1)
			 extended_subst
		     end;
		     
		 with
		   | Unification.UNIFICATION_FAIL ->
		       (* unusable context literal *)
		       ()
	       end;
	    )
	    input_partner.Context_unifier.ip_context_partners



(* searchs all context unifiers and adds the new candidates to candidates.
   one input_partner may be exempted from the search by setting fixed_index to its index
   and initializing partial_context_unifier and subst accordingly *)

(*   one input literal (fixed_index) is already paired with the new context literal.*)

let search_context_unifiers
  (config: config)
  (bound: bound)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  (fixed_index: int)
  (active_partner: context_partner)
  (subst: subst)
  (non_empty_remainder: bool)
  (close_found: bool)
  : bool =

  let assemble = {
    config = config;
    bound = bound;
    context_unifier_space = context_unifier_space;
    ordering = context_unifier_space.Context_unifier.cus_input_partners_ordering;
    context_partners = context_partners;
    fixed_index = fixed_index;
    active_partner = active_partner;
    close_found = close_found;
  }
  in
    assemble_context_unifier
      ~assert_unifier:false
      ~non_empty_remainder:non_empty_remainder
      assemble
      0
      subst;
    assemble.close_found
end


module SearchClose = Make (Close)
module SearchAssert = Make (Assert)
module SearchSplit = Make (Split)
module SearchCloseAssert = Make (CloseAssert)
module SearchAll = Make (All)

