(* $Id: context_unifier_space.ml,v 1.38 2004/12/28 14:06:47 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type counter = Counter.counter
type config = Config.config
type statistic = Statistic.statistic
type bound = Bound.bound
type var =  Var.var
type literal =  Term.literal
type clause =  Term.clause
type subst =  Subst.subst
type state = State.state
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner
type raw_context_unifier = Context_unifier.raw_context_unifier
type problem_literals = Problem_literals.problem_literals
type candidate_type = Context_unifier.candidate_type




(*** globals ***)


(* a unique id per clause represented by a context_unifier_space structure *)
let clause_id_counter : counter =
  Counter.create ()





(*** resolve ***)

(*
(* is this input literal already resolved
   with respect to the new context element? *)
let is_resolved (config: config)
  (input_partner: input_partner) (context_element: Context.element) : bool =

  match input_partner.Context_unifier.ip_resolved with
    | None -> false
    | _ -> true
*)

(*** subsume ***)
(*
(* yes, this special treatment for chronological backtracking is significantly faster *)
let is_already_subsumed (config: config)
  (context_unifier_space: context_unifier_space) (context_element: Context.element) : bool =

  match context_unifier_space.Context_unifier.cus_subsumed with
    | None -> false
    | _ -> true


let mark_subsumed statistic context_unifier_space choice_point =
  match context_unifier_space.Context_unifier.cus_subsumed with
    | None ->
	Statistic.inc_subsume statistic;
	context_unifier_space.Context_unifier.cus_subsumed <- Some (choice_point)

    | _ ->
	()
*)



(*** reorder ***)

(* sort input literals by increasing number of context partners.
   keep resolved literals at the end.
   thus less partial context unifier extending steps need to be done,
   as the first considered literals have fewer context partners
   and thus branching is lower in the top part of the search tree.
*)
(*
(* shift the resolved literal to the end *)
let move_resolved_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let resolved =
    input_partners.(index)
  in
    for i = index + 1 to Array.length input_partners - 1 do
      input_partners.(i - 1) <- input_partners.(i);
    done;
    input_partners.(Array.length input_partners - 1) <- resolved



(* shift the literal with the increased queue
   behind unresolved literals with a shorter queue *)
let move_increased_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in

  let rec swap_with (i : int) : unit =
    (* end of queue *)
    if i >= Array.length input_partners then
      ()

    (* only resolved literal ahead *)
    else if
      match input_partners.(i).Context_unifier.ip_resolved with
	| None ->
	    (* unresolved, so continue *)
	    false

	| _ ->
	    (* resolved, so stop here *)
	    true
    then
      ()

    (* the next literals queue is smaller, so swap *)
    else if
(*      List.length input_partners.(i).Context_unifier.ip_context_partners
      <
      List.length input_partners.(i - 1).Context_unifier.ip_context_partners*)
      input_partners.(i).Context_unifier.ip_context_partners_length
      <
      input_partners.(i - 1).Context_unifier.ip_context_partners_length
    then begin
      let swap =
	input_partners.(i - 1)
      in
	input_partners.(i - 1) <- input_partners.(i);
	input_partners.(i) <- swap;

	swap_with (i + 1)
    end
  in
    swap_with (index + 1)


let move_updated_input_literal (config: config)
  (context_unifier_space: context_unifier_space) (index: int)
  (input_partner: input_partner) (context_partner: context_partner)
  : unit =
  if is_resolved config input_partner context_partner.Context_unifier.cp_element then
    (*move_resolved_input_literal context_unifier_space index*)
    ()
  else
    move_increased_input_literal context_unifier_space index

*)







(*** add a literal ***)

(*
(* pair the new context literal with the clause literals one by one
   and compute all new context unifiers. *)
let add_to_clause
  (config: config)
  (bound: bound)
  (context_unifier_space: context_unifier_space)
  (input_partner: input_partner)
  (context_partner: context_partner)
  (candidate_type: candidate_type)
  (close_found: bool)
  : bool =

  if is_already_subsumed config context_unifier_space context_partner.Context_unifier.cp_element then begin
    false
  end

  else begin
    (* create the inital partial context unifier *)
    let partial_context_unifier =
      Array.make
	(Array.length context_unifier_space.Context_unifier.cus_input_partners)
	Context_unifier.null_partner
    in
      
      partial_context_unifier.(input_partner.Context_unifier.ip_index) <-
      context_partner;
      
      let rec find_i i =
	if i > Array.length context_unifier_space.Context_unifier.cus_input_partners then
	  failwith "find_i"
	else if input_partner == context_unifier_space.Context_unifier.cus_input_partners.(i) then
	  i
	else
	  find_i (i + 1)
      in
      let i =
	find_i 0
      in
      let subst =
	Subst.subst_from_array context_partner.Context_unifier.cp_partial_context_unifier
      in
      let search =
	match candidate_type with
	  | Context_unifier.Assert ->
	      Context_unifier_search.SearchAssert.search_context_unifiers

	  | Context_unifier.Split ->
	      Context_unifier_search.SearchSplit.search_context_unifiers

	  | Context_unifier.All ->
	      Context_unifier_search.SearchAll.search_context_unifiers
      in
      let non_empty_remainder =
	not (Config.is_horn config)
	&&
	Context_unifier.is_remainder 
	  subst
	  context_partner.Context_unifier.cp_element.Context.el_pars
	  input_partner.Context_unifier.ip_index
      in
	(* compute new candidates *)
      let close_found2: bool =
	search
	  config
	  bound
	  context_unifier_space
	  partial_context_unifier
	  i
	  context_partner
	  subst
	  non_empty_remainder
	close_found;
      in
	move_updated_input_literal config context_unifier_space i input_partner context_partner;

	close_found || close_found2
  end





(* checks if the new context literal is contradictory with this unit clause *)
let add_to_unit_clause
  (config: config)
  (statistic: statistic)
  (context_unifier_space: context_unifier_space)
  (context_element: Context.element)
  : unit =

  (* already subsumed? *)
  if is_already_subsumed config context_unifier_space context_element then
    ()

  (* now subsumed? *)
  else if
    Config.subsume config
    &&
    begin
      try
	ignore (Unification.match_literals
		  ~p_preserving:true
		  context_element.Context.el_literal
		  0
		  context_unifier_space.Context_unifier.cus_input_partners.(0).Context_unifier.ip_literal
		  1);
	mark_subsumed
	  statistic
	  context_unifier_space
	  (State.active_choice_point context_unifier_space.Context_unifier.cus_state);
	true
      with
	| Unification.UNIFICATION_FAIL -> 
	    false
    end
  then
    ()

  (* contradictory? *)
  else begin
    let context_literal =
      context_element.Context.el_literal
    in
    let input_literal =
      context_unifier_space.Context_unifier.cus_input_partners.(0)
    in
    let clause_literal =
      input_literal.Context_unifier.ip_literal
    in
      if context_literal.Term.sign = clause_literal.Term.sign then begin
	()
      end

      else begin
	try
	  (* are the clause and the context literal contradictory? *)
	  let subst =
	    Unification.unify_terms
	      ~p_preserving:true
	      clause_literal.Term.term
	      0
	      context_literal.Term.term
	      1
	  in
	    
	  (* yes, so raise CLOSED *)
	  let context_partner = {
	    Context_unifier.cp_element = context_element;
	    Context_unifier.cp_partial_context_unifier = Subst.subst_to_array subst;
	  }
	  in
	  let closing_context_unifier = {
	    Context_unifier.rcu_space = context_unifier_space;
	    Context_unifier.rcu_context_partners = [| context_partner |];
	  }
	  in
	    raise (Context_unifier.CLOSE closing_context_unifier)
	      
	with
	  | Unification.UNIFICATION_FAIL ->
	      ()
      end
  end
*)
      







(*** creation ***)


(* convert the clause into queues. *)
let clause_to_input_partners (problem_literals: problem_literals) (clause: clause) : input_partner array * int array =
  let input_partners =
    Array.mapi
      (fun index literal ->
	 Problem_literals.register_literal problem_literals literal index
      )
      (Array.of_list clause)
  in
  let ordering =
    Array.mapi
      (fun i _ ->
	 i
      )
      input_partners
  in
    Array.sort
      (fun x y ->
	 compare
	   input_partners.(x).Context_unifier.ip_context_partners_length
	   input_partners.(y).Context_unifier.ip_context_partners_length
      ) 
      ordering;
(*
    Array.sort
      (fun queue1 queue2 ->
	 compare
(*	 (List.length queue1.Context_unifier.ip_context_partners)
	 (List.length queue2.Context_unifier.ip_context_partners)*)
	 queue1.Context_unifier.ip_context_partners_length
	 queue2.Context_unifier.ip_context_partners_length
      ) 
      input_partners;
*)
    input_partners, ordering



let create_clause
  (config: config)
  (state: state)
  (statistic: statistic)
  (bound: bound)
  (problem_literals: problem_literals)
  (process_close_candidate: raw_context_unifier -> unit)
  (process_assert_candidate: raw_context_unifier -> bool)
  (process_split_candidate: raw_context_unifier -> unit)
  (clause: clause)
  (is_lemma: bool)
  : context_unifier_space =

  let input_partners, ordering =
    clause_to_input_partners problem_literals clause
  in

  let input_vars =
    Array.map
      (fun partner -> partner.Context_unifier.ip_vars)
      input_partners
  in

  let shared_subst_vars =
    Tools.lists_shared
      Subst.var_equal
      (Array.to_list input_vars)
  in

  let context_unifier_space = {
    Context_unifier.cus_id = Counter.next clause_id_counter;
    Context_unifier.cus_clause = clause;
    Context_unifier.cus_shared_vars = shared_subst_vars;
    Context_unifier.cus_input_partners = input_partners;
    Context_unifier.cus_input_partners_ordering = ordering;
(*    Context_unifier.cus_subsumed = None;*)
    Context_unifier.cus_process_close_candidate = process_close_candidate;
    Context_unifier.cus_process_assert_candidate = process_assert_candidate;
    Context_unifier.cus_process_split_candidate = process_split_candidate;
    Context_unifier.cus_state = state;
    Context_unifier.cus_lemma = is_lemma;
(*    Context_unifier.cus_backtrack = false;*)
    Context_unifier.cus_value = ref 0;
  }
  in
(*
  let do_add_to_clause =
    add_to_clause config bound context_unifier_space
  and do_mark_subsumed =
    mark_subsumed statistic context_unifier_space
  in*)
    Array.iter
      (fun input_partner ->
	 Problem_literals.register_clause_literal
	 problem_literals
 	 input_partner.Context_unifier.ip_literal
	 input_partner.Context_unifier.ip_index
(*	 do_add_to_clause
	 do_mark_subsumed*)
	   context_unifier_space
      )
      input_partners;

    (* if all clause literals are unifiable with -v, we have an initial candidate.
       no split on lemmas, so ignore this.
    *)
    if not is_lemma then begin
      try
	let context_partners =
	  Array.map
	    (fun input_partner ->
	       match input_partner.Context_unifier.ip_context_partners with
		 | [] ->
		     raise Exit
		       
		 | context_partner_v :: [] ->
		     context_partner_v
		       
		 | x :: y :: z ->
		     failwith "Context_unifier_space: more than one element."
	    )
	    input_partners
	in
	let candidate = {
	  Context_unifier.rcu_space = context_unifier_space;
	  Context_unifier.rcu_context_partners = context_partners;
	}
	in
	let subst =
	  Array.fold_left
	    (fun acc input_partner ->
	       match input_partner.Context_unifier.ip_context_partners with
		 | [] ->
		     raise Exit
		       
		 | context_partner_v :: [] ->
		     Unification.unify_packed_substs
		       acc
		       context_partner_v.Context_unifier.cp_partial_context_unifier
		       
		 | x :: y :: z ->
		     failwith "Context_unifier_space: more than one element. II"
	    )
	    Subst.empty
	    input_partners
	in
	  begin
	    match Context_unifier_check.check_split config bound subst context_unifier_space context_partners with
	      | None ->
		  (* dropped at depth bound *)
		  ()
		    
	      | Some candidate ->
		  context_unifier_space.Context_unifier.cus_process_split_candidate candidate;
	  end;
    with
      | Exit ->
	  ()
    end;
    
    context_unifier_space








(* a unit clause must be asserted and checked for contradictory against every context literal.
   but no context unifiers need to be computed, thus no queue is needed. *)
let create_unit_clause
  (config: config)
  (state: state)
  (statistic: statistic)
  (problem_literals: problem_literals)
  (process_close_candidate: raw_context_unifier -> unit)
  (process_assert_candidate: raw_context_unifier -> bool)
  (process_split_candidate: raw_context_unifier -> unit)
  (literal: literal)
  (is_lemma: bool)
  : context_unifier_space =

  let vars =
    List.map
      (fun var ->
	 Subst.make_var var Subst.input_literal_offset
      )
      (Term.vars_of_literal literal)
  in

  let input_partner =
    Problem_literals.register_literal problem_literals literal 0
  in
(*

  let input_partner = {
      Context_unifier.ip_literal = literal;
      Context_unifier.ip_vars = vars;
      Context_unifier.ip_index = 0;
      Context_unifier.ip_context_partners = [];
      Context_unifier.ip_context_partners_length = 0;
      Context_unifier.ip_context_partners_changed = false;
      Context_unifier.ip_resolved = None;
    };
  in*)

  let context_unifier_space =
    {
      Context_unifier.cus_id = Counter.next clause_id_counter;
      Context_unifier.cus_clause = [literal];
      Context_unifier.cus_shared_vars = [];
      Context_unifier.cus_input_partners = [| input_partner |];
      Context_unifier.cus_input_partners_ordering = [| 0 |];
(*      Context_unifier.cus_subsumed = None;*)
      Context_unifier.cus_process_close_candidate = process_close_candidate;
      Context_unifier.cus_process_assert_candidate = process_assert_candidate;
      Context_unifier.cus_process_split_candidate = process_split_candidate;
      Context_unifier.cus_state = state;
      Context_unifier.cus_lemma = is_lemma;
(*    Context_unifier.cus_backtrack = false;*)
    Context_unifier.cus_value = ref 0;
    }
  in
	 Problem_literals.register_clause_literal
	 problem_literals
 	 literal
	 0
(*	 do_add_to_clause
	 do_mark_subsumed*)
	   context_unifier_space;
(*    Problem_literals.register_unit_literal
      problem_literals
      (add_to_unit_clause config statistic context_unifier_space);*)
    
    context_unifier_space



(* creates a context_unifier_space from a clause and returns the initial candidates,
   i.e. a split context unifier computed by using -v *)    
let create
  (config: config)
  (state: state)
  (statistic: statistic)
  (bound: bound)
  (problem_literals: problem_literals)
  (clause: clause)
  (process_close_candidate: raw_context_unifier -> unit)
  (process_assert_candidate: raw_context_unifier -> bool)
  (process_split_candidate: raw_context_unifier -> unit)
  (is_lemma: bool)
  : context_unifier_space =
  match clause with
    | [] ->
	(* e.g. false :- true. in TPTP problem SWC128-1 *)
	(* :TODO: catch in darwin.ml *)
	raise (State.UNSATISFIABLE (clause, [| |]))

    | literal :: [] ->
	create_unit_clause config state statistic problem_literals
	  process_close_candidate process_assert_candidate process_split_candidate literal is_lemma

    | _ ->
	create_clause config state statistic bound problem_literals
	  process_close_candidate process_assert_candidate process_split_candidate clause is_lemma








(*** backtracking ***)
(*
let backtrack_subsumed (context_unifier_space: context_unifier_space) : unit =
  match context_unifier_space.Context_unifier.cus_subsumed with
    | Some choice_point when 
	State.is_choice_point_invalid choice_point ->
	context_unifier_space.Context_unifier.cus_subsumed <- None

    | _ ->
	()
*)

(* sort the newly arranged queues by length, small queues first.

   for chronological backtracking, additionally put resolved queues at the end.
   for non-chronological backtracking, this might even hurt,
   as resolve's effect is dependent on the choice point dependency.
*)
let backtrack_sort (config: config) (context_unifier_space: context_unifier_space) : unit =
  (*
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  Array.sort
    (fun x y ->
       match input_partners.(x).Context_unifier.ip_resolved, input_partners.(y).Context_unifier.ip_resolved with
	 | None, None ->
	     compare
	       input_partners.(x).Context_unifier.ip_context_partners_length
	       input_partners.(y).Context_unifier.ip_context_partners_length
	       
	 | Some _, _ ->
	     -1
	       
	 | _, Some _ ->
	     1
    ) 
    context_unifier_space.Context_unifier.cus_input_partners_ordering
  *)
  (*
  Array.sort
    (fun queue1 queue2 ->
       match queue1.Context_unifier.ip_resolved, queue2.Context_unifier.ip_resolved with
	 | None, None ->
	     compare
	       queue1.Context_unifier.ip_context_partners_length
	       queue2.Context_unifier.ip_context_partners_length
	       
	 | Some _, _ ->
	     -1
	       
	 | _, Some _ ->
	     1
    )
    context_unifier_space.Context_unifier.cus_input_partners
  *)
  (*
  (* try insertion sort *)
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  for i = 1 to Array.length input_partners - 1 do
    let to_move = input_partners.(i) in
    match to_move.Context_unifier.ip_resolved with
      | Some _ ->
	  ()

      | None ->
	  let rec find j =
	    if j = i then
	      ()
		
	    (* insert before resolved literals,
	       and before queues with at least same length *)
	    else if
	      match input_partners.(j).Context_unifier.ip_resolved with
		| Some _ ->
		    true

		| None ->
		    to_move.Context_unifier.ip_context_partners_length
		    <=
		    input_partners.(j).Context_unifier.ip_context_partners_length
	    then begin
	      for k = i downto j + 1 do
		input_partners.(k) <- input_partners.(k - 1);
	      done;

	      input_partners.(j) <- to_move;
	    end

	    else
	      find (j + 1)
	  in
	    find 0
  done
  *)

  (* try insertion sort *)
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let ordering =
    context_unifier_space.Context_unifier.cus_input_partners_ordering
  in
    for i = 1 to Array.length input_partners - 1 do
    let to_move = input_partners.(ordering.(i)) in
    match to_move.Context_unifier.ip_resolved with
      | Some _ ->
	  ()

      | None ->
	  let rec find j =
	    if j = i then
	      ()
		
	    (* insert before resolved literals,
	       and before queues with at least same length *)
	    else if
	      match input_partners.(ordering.(j)).Context_unifier.ip_resolved with
		| Some _ ->
		    true

		| None ->
		    to_move.Context_unifier.ip_context_partners_length
		    <=
		    input_partners.(ordering.(j)).Context_unifier.ip_context_partners_length
	    then begin
	      for k = i downto j + 1 do
		input_partners.(ordering.(k)) <- input_partners.(ordering.(k - 1));
	      done;

	      input_partners.(ordering.(j)) <- to_move;
	    end

	    else
	      find (j + 1)
	  in
	    find 0
  done




  (*
  (* try insertion sort *)
(*  match context_unifier_space.Context_unifier.cus_subsumed with
    | Some _ ->
	()

    | None ->*)
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
(*
    for i = 1 to Array.length input_partners - 1 do
      if input_partners.(i).Context_unifier.ip_context_partners_changed then begin
(*      match input_partners.(i).Context_unifier.ip_resolved with
	| Some _ ->
	    () check not needed if changed is checked first!
	      
	| None ->*)
	    (* shift up *)
	    let rec find_swap j =
	      if j = -1 then
		i
		  
	      else if
		match input_partners.(j).Context_unifier.ip_resolved with
		  | Some _ ->
		      true
			
		  | None ->
		      false
	      then
		find_swap (j - 1)		
		 
	      else if
(*		List.length input_partners.(i).Context_unifier.ip_context_partners
		>=
		List.length input_partners.(j).Context_unifier.ip_context_partners*)
		input_partners.(i).Context_unifier.ip_context_partners_length
		>=
		input_partners.(j).Context_unifier.ip_context_partners_length
	      then
		(j + 1)
		  
	      else begin
		find_swap (j - 1)
	      end
	    in
	      
	    let swap_with =
	      find_swap (i - 1)
	    in
	      if
		(swap_with <> i)
		&&
		match input_partners.(swap_with).Context_unifier.ip_resolved with
		  | Some _ ->
		      false
			
		  | None ->
		      true
	      then begin
		let swap = input_partners.(swap_with) in
		  input_partners.(swap_with) <- input_partners.(i);
		  input_partners.(i) <- swap;
	      end
      end
    done
    *)
    for i = 1 to Array.length input_partners - 1 do
      if input_partners.(i).Context_unifier.ip_context_partners_changed then begin
      match input_partners.(i).Context_unifier.ip_resolved with
	| Some _ ->
	    (*() check not needed if changed is checked first!*)
	    failwith "backtrack_cu"
	      
	| None ->
	let to_move = input_partners.(i) in
	    (* shift up *)
	    let rec swap j =
	      if j = 0 then begin
		input_partners.(j) <- to_move;
	      end

	      else if
		(
		  match input_partners.(j - 1).Context_unifier.ip_resolved with
		    | Some _ ->
			true
			  
		    | None ->
			false
		)
		||
		  (
		  to_move.Context_unifier.ip_context_partners_length
		<
		input_partners.(j - 1).Context_unifier.ip_context_partners_length
		  )
	      then begin
		input_partners.(j) <- input_partners.(j - 1);
		swap (j - 1)
	      end

	      else
		input_partners.(j) <- to_move;
	    in
	      swap i
      end
    done
  *)

let backtrack (config: config) (context_unifier_space: context_unifier_space) : unit =
(*  backtrack_subsumed context_unifier_space;*)
  backtrack_sort config context_unifier_space
