(* $Id: darwin.ml,v 1.60 2005/01/01 11:55:03 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* Sketch of the algorithm:
   i) add candidate to context
   ii) if this triggers a Close:
     - backtrack to a left split state and right split on it
     - remove all now invalidated information
   iii) compute new assert and split candidates
        based on the new context literal
   iv) select a new candidate
   v) if it's a split candidate then create a new choice point
   vi) continue with i)
*)



(*** types ***)

type flags = Flags.flags
type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type choice_point = State.choice_point
type literal = Term.literal
type clause = Term.clause
type state = State.state
type random_jump = Random_jump.random_jump
type raw_context_unifier = Context_unifier.raw_context_unifier
type context = Context.context
type selection = Selection.selection
type selected = Selection_types.selected
type log = Log.log


type derivation = {
  (* problem *)
  dv_clauses: clause list;
(*  mutable dv_lemmas: clause list;*)

  dv_config: config;
  dv_bound: bound;
  dv_statistic: statistic;
  dv_state: state;
  dv_random_jump: random_jump;
  dv_context: context;
  dv_selection: selection;
  dv_log: log;
  dv_in_saved_branch: bool;
}





exception CLOSE = Context_unifier.CLOSE
exception SATISFIABLE = Selection.SATISFIABLE
exception UNSATISFIABLE = State.UNSATISFIABLE
exception NO_SOLUTION = Const.NO_SOLUTION




(*** signal handling ***)

(* raise TERM_SIGNAL on termination signal *)

exception TERM_SIGNAL of string

let handle_term_signal (signal_number: int) : unit =
  if signal_number = Sys.sigint then
    raise (TERM_SIGNAL "\nABORTED: interactive interrupt")

  else if signal_number = Sys.sigterm then
    raise (TERM_SIGNAL "\nABORTED: termination")

  else if signal_number = Sys.sigalrm then
    raise (TERM_SIGNAL "\nWC  TIMEOUT")

  else if signal_number = Sys.sigvtalrm then
    raise (TERM_SIGNAL "\nCPU TIMEOUT")

  else
    raise (TERM_SIGNAL ("\nABORTED: " ^ string_of_int signal_number))


let init_signal_handler () =
  Sys.set_signal Sys.sigint (Sys.Signal_handle handle_term_signal);
  Sys.set_signal Sys.sigterm (Sys.Signal_handle handle_term_signal);
  Sys.set_signal Sys.sigalrm (Sys.Signal_handle handle_term_signal);
  Sys.set_signal Sys.sigvtalrm (Sys.Signal_handle handle_term_signal)






(*** input processing **)


let read_clauses (file_name: string) (default: Flags.opt_input_format) : clause list =
  try
    if Filename.check_suffix file_name ".darwin" then
      Read_darwin.to_clauses_from_file file_name
	
    else if Filename.check_suffix file_name ".tme" then
      Read_tme.to_clauses_from_file file_name
	
    else begin
      match default with
	| Flags.FI_TME ->
	    print_endline ("Defaulting to tme format.");
	    Read_tme.to_clauses_from_file file_name

	| Flags.FI_Darwin ->
	    print_endline ("Defaulting to darwin format.");
	    Read_darwin.to_clauses_from_file file_name
    end
  with
    | Sys_error error_message ->
	print_endline error_message;
	exit 0;
	[]

(* is the claus Horn? *)
let rec is_horn (clause: literal list) (pos_count: int) : bool =
  match clause with
    | [] ->
	true
	  
    | literal :: tail ->
	if literal.Term.sign then begin
	  if pos_count = 0 then
	    is_horn tail 1
	  else
	    false
	end
	else
	  is_horn tail pos_count

(* are the clauses horn? *)
let are_horn (clauses: clause list) : bool =
  List.for_all
    (fun clause ->
       is_horn clause 0
    )
    clauses

(* is a clause empty, or contains only false? *)
let contains_empty_clause (clauses: clause list) : bool =
  List.exists
    (
      List.for_all
	(Term.literal_equal Term.false_literal)
    )
    clauses








(*** create ***)

let create_derivation (clauses: clause list) (lemmas: clause list) (config: config) (bound: bound) (statistic: statistic) (saved_branch: Random_jump.saved_branch) (in_saved_branch: bool)
    : derivation =
  let state =
    State.create config
  in
  let random_jump =
    Random_jump.create config state bound
  in
  let context =
    Context.create config statistic state
  in
  let selection =
    Selection.create clauses lemmas config bound statistic state context saved_branch
  in
  let log =
    Log.create config state context
  in
    {
      dv_clauses = clauses;
(*      dv_lemmas = lemmas;*)

      dv_config = config;
      dv_bound = bound;
      dv_statistic = statistic;
      dv_state = state;
      dv_random_jump = random_jump;
      dv_context = context;
      dv_selection = selection;
      dv_log = log;
      dv_in_saved_branch = in_saved_branch;
    }





(*** add ***)


let add (derivation: derivation) (literal: literal)
  (mandatory_at: int) (generation: int) : unit =

  let context_element =
    Context.add derivation.dv_context literal mandatory_at generation
  in
    State.set_context_id derivation.dv_state context_element.Context.el_literal context_element.Context.el_id;
    Selection.add derivation.dv_selection context_element






(*** inference rules ***)



let assert_rule (derivation: derivation) (selected: selected) : unit =
  if selected.Selection_types.candidate_type = Selection_types.Assert then begin
    Statistic.inc_assert derivation.dv_statistic;
    derivation.dv_log#assert_
      selected.Selection_types.literal selected.Selection_types.raw_context_unifier;
  end
    
  else begin
    Statistic.inc_split derivation.dv_statistic;
    derivation.dv_log#split_unit
      selected.Selection_types.literal selected.Selection_types.raw_context_unifier;
  end;

  (* add to the context *)
  add
    derivation
    selected.Selection_types.literal
(*    (Some*)
       (Context_unifier.get_creation_cid
	  selected.Selection_types.raw_context_unifier.Context_unifier.rcu_context_partners)
(*    )*)
    (Context_unifier.generation_of_candidate
       selected.Selection_types.raw_context_unifier.Context_unifier.rcu_context_partners)

    
let split_left (derivation: derivation) (selected: selected) : unit =
  Statistic.inc_split derivation.dv_statistic;
  derivation.dv_log#split_left
    (State.active_choice_point derivation.dv_state)
    selected.Selection_types.literal
    selected.Selection_types.raw_context_unifier;
  
  (* add to the context *)
  add
    derivation
    selected.Selection_types.literal
    (*None*) (-1)
    (Context_unifier.generation_of_candidate
       selected.Selection_types.raw_context_unifier.Context_unifier.rcu_context_partners)

let split_right (derivation: derivation) (selected: selected) : unit =
  Statistic.inc_split derivation.dv_statistic;
  derivation.dv_log#split_right
    selected.Selection_types.literal
    selected.Selection_types.raw_context_unifier;
(*  print_endline ("SPLIT RIGHT: " ^ Term.literal_to_string selected.Selection_types.literal);*)
(*  derivation.dv_log#split_left
    (State.active_choice_point derivation.dv_state)
    selected.Selection_types.literal
    selected.Selection_types.raw_context_unifier;
*)
  (* add to the context *)
  add
    derivation
    selected.Selection_types.literal
    (*None*) (-1)
    (Context_unifier.generation_of_candidate
       selected.Selection_types.raw_context_unifier.Context_unifier.rcu_context_partners)
(*
(* do a right split on right_split_literal -
   but first backtrack to the corresponding left split. *)
let rec split_right (derivation: derivation)
    (left_split: choice_point) (oldest_retracted: choice_point) (right_split_literal: literal) =
  
  (* backtrack all modules *)
  derivation.dv_bound#backtrack;
  Context.backtrack derivation.dv_context oldest_retracted;
  let selected =
    Selection.backtrack derivation.dv_selection left_split oldest_retracted right_split_literal
  in
    
    derivation.dv_log#split_right
      left_split
      (State.active_choice_point derivation.dv_state)
      selected.Selection_types.literal
      selected.Selection_types.raw_context_unifier;

    (* add to the context *)
    try
      add
	derivation
	selected.Selection_types.literal
	None
	(Context_unifier.generation_of_candidate
	   selected.Selection_types.raw_context_unifier.Context_unifier.rcu_context_partners)
    with
      | CLOSE raw_context_unifier ->
	  close derivation raw_context_unifier
		
      | SATISFIABLE ->
	  check_exhausted derivation
*)

(*
and close (derivation: derivation) (raw_context_unifier: raw_context_unifier) : unit =
  Statistic.inc_close derivation.dv_statistic;
  
  derivation.dv_log#close raw_context_unifier;

  let backtracked_choice_point, oldest_retracted, right_split_literal =
    State.backtrack
      derivation.dv_state
      (Context_unifier.get_context_literals
	 raw_context_unifier.Context_unifier.rcu_context_partners)
      (derivation.dv_log :> State.prune)
  in

    split_right derivation backtracked_choice_point oldest_retracted right_split_literal

*)

let rec close (derivation: derivation) (raw_context_unifier: raw_context_unifier) : unit =
  Statistic.inc_close derivation.dv_statistic;
  let context_literals =
    Context_unifier.get_context_literals
      raw_context_unifier.Context_unifier.rcu_context_partners
  in
  let retracted, backjump, explanation, lemma  =
    State.backtrack
      derivation.dv_state
      raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
      raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
      context_literals
      (derivation.dv_log :> State.prune)
  in
(*
  let skip =
    Random_jump.jump derivation.dv_random_jump
  in
*)
(*    print_endline ("RETRACTED: " ^ State.choice_point_to_string retracted);*)
    derivation.dv_log#close backjump raw_context_unifier;

    (* backtrack all modules *)
    derivation.dv_bound#backtrack;
    Context.backtrack derivation.dv_context (State.id_of_choice_point backjump (*retracted*));
    Selection.backtrack derivation.dv_selection retracted explanation;
    try

      begin
	match lemma with
	  | Some lemma ->

		(* ignore already existing lemma *)
(*	      let lemma =
		Term.order_clause lemma
	      in*)
(*		print_endline ("LEMMA: " ^ Term.clause_to_string lemma);*)

(*
		let rec clauses_equal x y =
		  match x, y with
		    | [], [] ->
			true
			  
		    | hx :: tx, hy :: ty when
			Term.literal_equal hx hy ->
			clauses_equal tx ty
			  
		    | _ ->
			false
		in
		  
		let not_new0 =
		  List.exists
		    (fun old ->
		       clauses_equal
			 old
			 lemma	 
		    )
		    derivation.dv_clauses
		in
		  
		  if not_new0 then begin
		    ();
(*		    print_endline ("LEMMA already in input set - IGNORED\n");*)
		  end
		  else begin
		    
		    let not_new1 =
		      List.exists
			(fun old ->
			   clauses_equal
			     old
			     lemma	 
			)
			derivation.dv_lemmas
		    in
		      
		      if not_new1 then begin
			();
(*			print_endline ("LEMMA learned - IGNORED\n");*)
		      end
		      else begin

*)
			(*	      if List.length derivation.dv_lemmas < 50 then begin*)
			Selection.add_lemma derivation.dv_selection lemma;
(*			derivation.dv_lemmas <- lemma :: derivation.dv_lemmas;*)

			(*
			let b =
			  match explanation with
			    | [] ->
				State.root_choice_point derivation.dv_state
				  
			    | x :: _ ->
				x
			in
			  Selection.add_lemma derivation.dv_selection lemma b;
			*)

			  (*	      end*)



(*			derivation.dv_lemmas <- lemma :: derivation.dv_lemmas;
			
			let b =
			  try
			    List.find
			      State.is_left
			      explanation
			  with
			    | Not_found ->
				State.root_choice_point derivation.dv_state
			in
			let b =
			  match explanation with
			    | [] ->
				State.root_choice_point derivation.dv_state
				  
			    | x :: _ ->
				x
			in
			  Selection.add_lemma derivation.dv_selection lemma b;
*)
(*		      end
		  end*)

	  | None ->
	      ()
      end;

      select derivation

    with
      | CLOSE raw_context_unifier ->
	  if Selection.in_replay derivation.dv_selection then begin
	    failwith "in_replay 1";
	    raise (UNSATISFIABLE ([], [| |]))
	  end;

	  close derivation raw_context_unifier
		
      | SATISFIABLE ->
	  check_exhausted derivation






and check_exhausted derivation =
  (* no incomplete branch, a model has been found. *)
  if not derivation.dv_bound#is_branch_incomplete then begin
    raise SATISFIABLE
  end

  (* no backtracking over incomplete branches, do a complete restart. *)
  else if Config.restart derivation.dv_config = Flags.RS_Eager then begin
    raise SATISFIABLE
  end

  (* abandon this branch and try to continue the derivation. *)
  else begin
(*    failwith "INCOMPLETE";*)
    
   
    let incomplete_choice_points =
      derivation.dv_bound#restart_incomplete 
    in
      
    let retracted =
      State.backtrack_incomplete
	derivation.dv_state
	incomplete_choice_points
    in
(*	
  let skip =
    Random_jump.jump derivation.dv_random_jump
  in
*)
    derivation.dv_log#incomplete retracted;
    derivation.dv_bound#backtrack;
    Context.backtrack derivation.dv_context (State.id_of_choice_point retracted);
    Selection.backtrack derivation.dv_selection retracted [];

	try
	  select derivation
	  (*split_right derivation backtracked_choice_point oldest_retracted right_split_literal*)
        with
	  | CLOSE raw_context_unifier ->
	      if Selection.in_replay derivation.dv_selection then begin
		failwith "in_replay 2";
		raise (UNSATISFIABLE ([], [| |]))
	      end;

	      close derivation raw_context_unifier
		
	  | SATISFIABLE ->
	      check_exhausted derivation

  end



(*** selection ***)

(* endless loop:
   - select a candidate
   - add it to the context
   - backtrack on CLOSED and right split

   loop is exited when
   - UNSATISFIABLE is raised
   - SATISFIABLE is raised and restarting has to be done
*)
and select (derivation: derivation) : unit =
  while true do
    try
      let selected =
	Selection.select derivation.dv_selection
      in
	match selected.Selection_types.candidate_type with
	  | Selection_types.Assert ->
	      assert_rule derivation selected
		
	  | Selection_types.UnitSplit ->
	      assert_rule derivation selected
		
	  | Selection_types.SplitLeft ->
	      if derivation.dv_in_saved_branch then begin
		split_left derivation selected
	      end
	  
	      else begin
		match Random_jump.jump derivation.dv_random_jump with
		  | None ->
		      split_left derivation selected
			
		  | Some jump_target ->
		      begin
			let explanation =
			  State.random_jump derivation.dv_state jump_target
			in
(*			  derivation.dv_log#random_jump;*)
(*			  print_endline ("Random Jump: " ^ State.choice_point_to_string jump_target);*)
			  derivation.dv_bound#backtrack;
			  Context.backtrack derivation.dv_context (State.id_of_choice_point jump_target);
			  Selection.backtrack derivation.dv_selection jump_target explanation;
			  try
			    select derivation
			      
			  with
			    | CLOSE raw_context_unifier ->
	  if Selection.in_replay derivation.dv_selection then begin
	    raise (UNSATISFIABLE ([], [| |]))
	  end;
				close derivation raw_context_unifier
				  
			    | SATISFIABLE ->
				check_exhausted derivation
		      end
	      end

	  | Selection_types.SplitRight ->
	      split_right derivation selected
    with
      | CLOSE raw_context_unifier ->
	  if Selection.in_replay derivation.dv_selection then begin
	    failwith "in_replay 3";
	    raise (UNSATISFIABLE ([], [| |]))
	  end;
	  close derivation raw_context_unifier
	    
      | SATISFIABLE ->
	  check_exhausted derivation
  done




	  



(*** main loop ***)




let print_derivation_start derivation =
  if Config.print_derivation_online derivation.dv_config then
    print_endline ("START OF DERIVATION")

let print_derivation_end derivation =
  if Config.print_derivation_online derivation.dv_config then
    print_endline ("END OF DERIVATION\n")


let print_derivation_stats derivation =
  if Config.print_statistics derivation.dv_config then begin
    print_newline ();
    Statistic.print derivation.dv_statistic;
    Context.print_max_size derivation.dv_context;
    derivation.dv_bound#print;
  end

let print_derivation_post derivation =
  print_derivation_stats derivation;
  Print.print_statistic "Lemmas" (List.length (Selection.get_lemmas derivation.dv_selection));
  print_newline ();
  Print.print_statistic_float "CPU Time" (Sys.time ());
  Print.print_statistic_float "WC  Time" (Unix.time () -. Config.start_time derivation.dv_config);
  derivation.dv_log#finalize



let print_derivation_satisfiable derivation =
  (* print all kinds of models ... *)
  if Config.print_model_context derivation.dv_config then begin
    print_endline ("\nSTART OF MODEL (CONTEXT):");
    Context.print_context derivation.dv_context stdout;
    print_endline ("END OF MODEL");
  end;
  
  if Config.print_model_context_file derivation.dv_config <> "" then begin
    let file_name =
      Config.print_model_context_file derivation.dv_config
    in
      try
	let out =
	  open_out file_name
	in
	  Context.print_context derivation.dv_context out;
	  close_out out
      with
	| Sys_error error_message ->
	    print_endline ("Print Context to " ^ file_name ^ " failed.");
	    print_endline error_message;
  end;
  
  if Config.print_model_DIG derivation.dv_config then begin
    print_endline ("\nSTART OF MODEL (DIG):");
    Context.print_DIG derivation.dv_context stdout derivation.dv_clauses;
    print_endline ("END OF MODEL");
  end;
  
  if Config.print_model_DIG_file derivation.dv_config <> "" then begin
    let file_name =
      Config.print_model_DIG_file derivation.dv_config
    in
      try
	let out =
	  open_out file_name
	in
	  Context.print_DIG derivation.dv_context out derivation.dv_clauses;
	  close_out out
      with
	| Sys_error error_message ->
	    print_endline ("Print DIG to " ^ file_name ^ " failed.");
	    print_endline error_message;
  end




    
let rec restart_incomplete derivation =
  let derivation =
    create_derivation
      derivation.dv_clauses (Selection.get_lemmas derivation.dv_selection) derivation.dv_config derivation.dv_bound#restart derivation.dv_statistic [] false
  in
    restart derivation

and replay derivation =
(*  failwith "replay";*)
  let saved_branch =
    Random_jump.pop derivation.dv_random_jump
  in
  let derivation' =
    create_derivation
      derivation.dv_clauses (Selection.get_lemmas derivation.dv_selection) derivation.dv_config (Bound.create ~bound:derivation.dv_bound#current derivation.dv_config) derivation.dv_statistic saved_branch true
  in
    Random_jump.register derivation.dv_random_jump derivation'.dv_random_jump;
    
    restart derivation'
	

(* (re)start the derivation, and print the result at the end of the derivation. *)
and restart (derivation: derivation) : unit =
  if Config.print_derivation_online derivation.dv_config then begin
    print_endline (
      (State.active_choice_point_to_string derivation.dv_state)
      ^ "Depth bound: " ^ derivation.dv_bound#to_string
    );
  end;
  
  try
    select derivation;
  with
      (* order:
	 - (incomplete or complete) refutation with saved branches -> these need to be explored
	 - incomplete refutation -> restart with increased bound
	 - incomplete, no model found -> try saved branch or restart with increased bound
      *)
      (* still some saved paths to explore. *)
(*    | SATISFIABLE*)
    | UNSATISFIABLE _ when
	not (Random_jump.finished derivation.dv_random_jump) ->
(*	print_endline ("0");*)
	
	replay derivation

      (* no exhausted branch found, but some incomplete ones. *)
    | UNSATISFIABLE _ when
	derivation.dv_bound#is_derivation_incomplete ->
(*	print_endline ("1");*)
	
	restart_incomplete derivation
	  
      (* the current branch is incomplete. *)
    | SATISFIABLE when
	derivation.dv_bound#is_branch_incomplete ->
(*	print_endline ("2");*)

	if not (Random_jump.finished derivation.dv_random_jump) then
	  replay derivation
	else
	  restart_incomplete derivation
	  
    | SATISFIABLE ->
	print_derivation_end derivation;
	
	print_endline ("SATISFIABLE");

	print_derivation_satisfiable derivation;

	print_derivation_post derivation;


    | UNSATISFIABLE (clause, context_literals) ->
	let closing_message =
	  Term.clause_to_string clause
	  ^ "\n"
	  ^ "Context: "
	  ^ Term.clause_to_string (Array.to_list context_literals)
	  ^ "\n"
	in
	  print_endline (closing_message);

	print_derivation_end derivation;
	
	print_endline ("REFUTATION FOUND");
	
	print_derivation_post derivation;
	
    (* Note: this case never happened in current tests,
       but a candidate overflow is bound to happen for longer run times. *)
    | NO_SOLUTION message ->
	print_derivation_end derivation;
	
	print_endline ("NO SOLUTION: " ^ message);
	
	print_derivation_post derivation;
	
    | TERM_SIGNAL signal ->
	print_derivation_stats derivation;
	Print.print_statistic "Lemmas" (List.length (Selection.get_lemmas derivation.dv_selection));
	print_newline ();
	Print.print_statistic_float "CPU Time" (Sys.time ());
	Print.print_statistic_float "WC  Time" (Unix.time () -. Config.start_time derivation.dv_config);

	print_endline (signal)

(*
    | exn ->
	print_derivation_stats derivation;
	Print.print_statistic "Lemmas" (List.length (Selection.get_lemmas derivation.dv_selection));
	print_newline ();
	Print.print_statistic_f "CPU Time" (Sys.time ());
	Print.print_statistic_f "WC  Time" (Unix.time () -. Config.start_time derivation.dv_config);

	print_endline (Printexc.to_string exn)
*)




	  






(* start a derivation *)
let prove (config: config) (clauses: clause list) : unit =
  let derivation : derivation =
    create_derivation clauses [] config (Bound.create config) (Statistic.create config) [] false
  in
    print_endline ("Proving " ^ (*Config.problem_file_name config ^*) " ...");
    print_newline ();

    if Config.print_configuration derivation.dv_config then begin
      Config.print derivation.dv_config;
    end;

    print_derivation_start derivation;

    restart derivation


(* set user defined timeouts *)
let set_timeouts flags =
  (* set the timeout before doing anyting input *)
  let time_out_CPU: int =
    (Flags.time_out_CPU flags)#value
  in
    if time_out_CPU > 0 then begin
      ignore (
	Unix.setitimer Unix.ITIMER_VIRTUAL 
	  {
   	    Unix.it_interval = 0.0;
   	    Unix.it_value  = float_of_int (time_out_CPU);
	  }
      );
    end;

  let time_out_WC: int =
    (Flags.time_out_WC flags)#value
  in
    if time_out_WC > 0 then begin
      ignore (
	Unix.setitimer Unix.ITIMER_REAL
	  {
   	    Unix.it_interval = 0.0;
   	    Unix.it_value  = float_of_int (time_out_WC);
	  }
      );
    end


(* start a derivation based on command line arguments. *)
let () =
  init_signal_handler ();

(*
  let x =
    Read_darwin.to_term "is_a_theorem(equivalent(equivalent(equivalent(_0, equivalent(equivalent(equivalent(_1, _2), equivalent(equivalent(_2, _3), _1)), _3)), _4), equivalent(_4, _0)))"
  and y =
    Read_darwin.to_term "is_a_theorem(equivalent(equivalent(equivalent(equivalent(_0, _1), equivalent(equivalent(_2, _1), _0)), _2), equivalent(equivalent(a, b), equivalent(c, equivalent(equivalent(b, c), a)))))"
  in
  let subst =
    Unification.unify_terms x 0 y 1
  in
  let inst =
    Subst.apply_to_term subst x 0
  in
    print_endline (Subst.subst_to_string subst);
    print_endline (Term.term_to_string inst);
    failwith "done";
*)
(*  
  print_endline (string_of_int ((Gc.get ()).Gc.minor_heap_size));
  Gc.set { (Gc.get()) with Gc.minor_heap_size = 32768 * 4 };
  print_endline (string_of_int ((Gc.get ()).Gc.minor_heap_size));

  print_endline (string_of_int ((Gc.get ()).Gc.space_overhead));
  Gc.set { (Gc.get()) with Gc.space_overhead = 160 };
  print_endline (string_of_int ((Gc.get ()).Gc.space_overhead));
*)
  let start_time =
    Unix.time ()
  in
    try

  let flags =
    Flags.eval_command_line ()
  in

    set_timeouts flags;

  print_endline (Print.version);
  print_newline ();

  print_endline ("Parsing " ^ Flags.file_name flags ^ " ...");

  (* here some parsing messages may be output,
     like defaulting to an input format. *)
  let clauses =
    read_clauses (Flags.file_name flags) (Flags.input_format flags)#value
  in
    print_newline ();

  let is_horn =
    are_horn clauses
  in
  let contains_equality =
    Equality.contains_equality clauses
  in
  let config =
    Config.create ~flags:flags ~is_horn:is_horn ~contains_equality:contains_equality ~start_time:start_time
  in

    if contains_empty_clause clauses then begin
      if Config.print_derivation_online config then begin
	print_endline ("Input contains empty clause.\n");
      end;
      print_endline ("REFUTATION FOUND");
    end
      
    else begin
      let clauses =
	List.map
	  (fun clause ->
(*	     let clause =
	       Term.sort_clause clause
	     in*)
(*
	     let condensed =
	       Term.sort_clause (State.condense clause)
	     in

	       if List.length condensed < List.length clause then begin
		 print_endline ("Clause   : " ^ Term.clause_to_string clause);
		 print_endline ("Condensed: " ^ Term.clause_to_string condensed);
	       end;
		
	       condensed*)
	       clause
	  )
	  clauses
      in

      let clauses =
	if
	  Config.contains_equality config
	  &&
	  Config.equality config = Flags.EQ_Axioms
	then
	  Equality.add_equality clauses
	else
	  clauses
      in
	prove config clauses
    end
  with
    | TERM_SIGNAL signal ->
	print_newline ();
	Print.print_statistic_float "CPU Time" (Sys.time ());
	Print.print_statistic_float "WC  Time" (Unix.time () -. start_time);
	
	print_endline (signal)
