(* $Id: equality.ml,v 1.3 2004/12/28 14:06:48 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



type var = Var.var
type symbol = Symbol.symbol
type term = Term.term
type clause = Term.clause


(* get the predicate symbols (no constants) contained in the clause set. *)
let get_predicate_symbols (clauses: clause list) : symbol list =
  List.fold_left
    (fun acc clause ->
       List.fold_left
         (fun acc literal ->
	    match literal.Term.term with
	      | Term.Const symbol
	      | Term.Func (symbol, _, _) ->
		  (* filter the equality symbol *)
		  if Symbol.equal Symbol.equality symbol then
		    acc
		  else
		    Tools.list_add Symbol.equal acc symbol
		    
	      | _ ->
		  acc
	 )
         acc
         clause
    )
    []
    clauses

    
(* get the function symbols (no constants) contained in the term. *)
let rec get_function_symbols (term: term) (acc: symbol list) : symbol list =
  match term with
    | Term.Func (symbol, terms, _) ->
	Array.fold_left
	  (fun acc term ->
	     get_function_symbols term acc
	  )
	  (Tools.list_add Symbol.equal acc symbol)
	  terms
	    
    | _ ->
	acc


(* get the predicate and function symbols (no constants) contained in the clause set. *)
let get_symbols (clauses: clause list) : symbol list * symbol list =
  List.fold_left
    (fun acc clause ->
       List.fold_left
         (fun ((predicates, functions) as acc) literal ->
	    match literal.Term.term with
	      | Term.Func (symbol, terms, _) ->
		  let new_predicates =
		    (* filter the equality symbol *)
		    if Symbol.equal Symbol.equality symbol then
		      predicates
		    else
		      Tools.list_add Symbol.equal predicates symbol
		  in
		  let new_functions =
		    Array.fold_left
		      (fun acc term ->
			 get_function_symbols term acc
		      )
		      functions
		      terms
		  in
		    (new_predicates, new_functions)
		    
	      | _ ->
		  acc
	 )
         acc
         clause
    )
    ([], [])
    clauses
    

(* basic equality axioms *)
let get_equality_clauses () : clause list =
  [
    (* Reflexivity *)
    Read_tme.to_clause "(X = X).";
    (* Symmetry *)
    Read_tme.to_clause "(Y = X) :- (X = Y).";
    (* Transitivity *)
    Read_tme.to_clause "(X = Z) :- (X = Y), (Y = Z).";
  ]





(* creates for a symbol s permutations of the terms
   s(_0, _1, _2)
   s(_3, _1, _2)
   (_0 = _3)
   where _3 shifts over all positions.
   applies assemble to it to build a clause.
 *)
let do_symbol (symbol: symbol) (acc: clause list)
    (assemble: term -> term -> term -> clause) : clause list =

  let int_to_term (i: int) : term =
    Term.request_var (Var.create_universal i)
  in

  let arity =
    Symbol.arity symbol
  in
  let schema =
    Array.make arity 0
  in
    for i = 0 to Array.length schema - 1 do
      schema.(i) <- i;
    done;
    
    (* a list of 0 ... arity - 1 *)
    let rec do_at (index: int) (acc: clause list) : clause list =
      if index >= arity then
	acc
	    
      else begin
	let left_term =
	  Term.request_func (symbol, Array.map int_to_term schema)
	in

	(* temporarily replace the index variable *)
	schema.(index) <- arity;
	let right_term =
	  Term.request_func (symbol, Array.map int_to_term schema)
	in
	  schema.(index) <- index;
	let equality_term =
	  Term.request_func
	    (Symbol.equality, [| int_to_term index; int_to_term arity |])
	in
	let clause =
	  assemble left_term right_term equality_term
	in
	  print_endline ("N:  " ^ Term.clause_to_string clause);
	  do_at (index + 1) (clause :: acc)
      end
    in
      do_at 0 acc


(*
  the substitution axioms for the predicate p(X, Y, Z) are:
  
  p(_0, _1, _2) :- p(_3, _1, _2)), (_0 = _3).
  p(_0, _1, _2) :- p(_0, _3, _2)), (_0 = _3).
  p(_0, _1, _2) :- p(_0, _1, _3)), (_0 = _3).
*)
let get_predicate_substitutions (symbols: symbol list)
    (acc: clause list) : clause list =

  let assemble left right equal =
    let left_predicate =
      Term.request_literal true left
    and right_predicate =
      Term.request_literal false right
    and equality_predicate =
      Term.request_literal false equal
    in
      [ left_predicate; right_predicate; equality_predicate ]
  in

    List.fold_left
      (fun acc symbol ->
	 do_symbol symbol acc assemble
      )
      acc
      symbols


(*
  the substitution axioms for the function f(X, Y, Z) are:

  (f(_0, _1, _2) = f(_3, _1, _2)) :- (_0 = _3).
  (f(_0, _1, _2) = f(_0, _3, _2)) :- (_1 = _3).
  (f(_0, _1, _2) = f(_0, _1, _3)) :- (_2 = _3).
*)
let get_function_substitutions (symbols : symbol list)
    (acc: clause list) : clause list =

  let assemble left right equal =
    let left_equality_term =
      Term.request_func (Symbol.equality, [| left; right |])
    in
    let left_equality_predicate =
      Term.request_literal true left_equality_term
    in
    let right_equality_predicate =
      Term.request_literal false	equal
    in
      [ left_equality_predicate; right_equality_predicate ]
  in

    List.fold_left
      (fun acc symbol ->
	 do_symbol symbol acc assemble
      )
      acc
      symbols







let contains_equality (clauses: clause list) : bool =
  List.exists
    (List.exists
       (fun literal ->
	  match literal.Term.term with
	    | Term.Func (symbol, _, _) when
		Symbol.equal Symbol.equality symbol ->
		true
		  
	    | _ ->
		false
       )
    )
    clauses



let add_equality (clauses: clause list) : clause list =
  let predicate_symbols, function_symbols =
    get_symbols clauses
  in

  let with_equality =
    get_equality_clauses ()
    @
    get_predicate_substitutions predicate_symbols
    (get_function_substitutions function_symbols clauses)
  in
    (*List.iter (fun x -> print_endline (Term.clause_to_string x)) with_equality;*)
    with_equality
