(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)





(*** types ***)


(*** flag id ***)

type flag_id =
  | FL_Resolve
  | FL_Subsume
  | FL_Compact
  | FL_Productivity
  | FL_MixedLiterals
(* Disable for CASC    
  | FL_Lemma
*)
  | FL_Equality
  | FL_PlusV
  | FL_Backtracking
  | FL_IterativeDeepening
  | FL_DepthBound
  | FL_LookaheadExceeding
  | FL_Restart
(* Disable for CASC    
  | FL_RandomJump
*)
  | FL_NegAssertCandidates

  | FL_InputFormat

  | FL_TimeOutCPU
  | FL_TimeOutWC

  | FL_PrintLevel
  | FL_PrintConfiguration
  | FL_PrintStatistics
  | FL_PrintModelContext
  | FL_PrintModelContextFile
  | FL_PrintModelDIG
  | FL_PrintModelDIGFile
  | FL_PrintModelARM
  | FL_PrintDerivationOnline
      (*
  | FL_PrintDerivation
  | FL_PrintDerivationPruned
  | FL_PrintDerivationFile
  | FL_PrintDerivationPrunedFile
  | FL_PrintDerivationDot
  | FL_PrintDerivationDotPruned
      *)
  | FL_PrintDerivationContextUnifier
  | FL_PrintAssertCandidates
  | FL_PrintSplitCandidates

  | FL_Help    
  | FL_Version



 


(*** flags ***)





(* the actual classes *)

(* flag base class
   abstracts over the internal type of the flag (bool, int, string, variant type, ...)
   and the internal type of the flag, i.e. where a string is used instead of a variant type
   and later on mapped to the variant type
*)
class virtual ['a, 'b] flag_base
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: 'a) =

object (self)
  val _id = __id
  val _short_name = __short_name
  val _long_name = __long_name
  val _description = __description
  val mutable _value : 'a = __default

  method id = _id
  method short_name = _short_name
  method long_name = _long_name
  method description = _description
  method value = _value

  method set value =
    if self#check value then
      self#update value
    else
      self#error value

  method set_opt value = _value <- value
  method private virtual update: 'b -> unit
  method virtual check: 'b -> bool
  method virtual argument_to_string: 'b -> string
  method virtual value_to_string: 'a -> string
  method virtual opt_to_string: string list

    
  method signature : string =
    match self#opt_to_string with
      | [] ->
	  (* unit *)
          ""

      | _type :: [] ->
	  (* a type like bool, int, string, ... *)
          "<" ^ _type ^ ">"

      | _ ->
	  (* enumeration *)
          "[" ^ String.concat "|" self#opt_to_string ^ "]"

  method private error (value: 'b) : unit =
    let msg =
      "wrong argument `" ^ self#argument_to_string value ^ "'; option `"
      ^ self#short_name ^ "'/'" ^ self#long_name
      ^ "' expects a value in " ^ self#signature
    in
      raise (Arg.Bad (msg))
end

class flag_bool
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: bool) =

object (self)
  inherit ['a, 'b] flag_base __id __short_name __long_name __description __default

  method check value = true
  method private update value = _value <- value
  method argument_to_string value = string_of_bool value
  method value_to_string value = string_of_bool value
  method opt_to_string = ["bool"]
end

class flag_int
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: int) =

object (self)
  inherit ['a, 'b] flag_base __id __short_name __long_name __description __default

  method check value = true
  method private update value = _value <- value
  method argument_to_string value = string_of_int value
  method value_to_string value = string_of_int value
  method opt_to_string = ["int"]
end

class flag_unit
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list) =

object (self)
  inherit ['a, 'b] flag_base __id __short_name __long_name __description ()

  method check value = true
  method private update () = ()
  method argument_to_string value = ""
  method value_to_string value = ""
  method opt_to_string = []
end

class flag_string
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: string) =

object (self)
  inherit ['a, 'b] flag_base __id __short_name __long_name __description __default

  method check value = true
  method private update value = _value <- value
  method argument_to_string value = value
  method value_to_string value = value
  method opt_to_string = ["string"]
end


class flag_file
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: string) =

object (self)
  inherit flag_string __id __short_name __long_name __description __default

  method opt_to_string = ["file name"]
end



(* base flag for variant types.
   takes a string as the input value but leaves the internal type unspecified. *)
class virtual ['a] flag_variant
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: 'a) =

object (self)

  method virtual mapping : ('a * string) list

  inherit ['a, 'b] flag_base __id __short_name __long_name __description __default

  method check (value: string) : bool =
    try
      let (opt, name) =
	List.find
	  (fun (opt, name) ->
	     name = value
	  )
	  self#mapping
      in
	true
    with
      | Not_found ->
	  false

  method private update (value: string) : unit =
    try
      let (opt, name) =
	List.find
	  (fun (opt, name) ->
	     name = value
	  )
	  self#mapping
      in
	_value <- opt;
    with
      | Not_found ->
	  failwith "Flag_variant.update"

  method argument_to_string value = value

  method value_to_string _opt : string =
    let (opt, name) =
      List.find
	(fun (opt, name) ->
	   opt = _opt
	)
	self#mapping
    in
      name

  method opt_to_string =
    List.map
      (fun (opt, name) ->
	 name
      )
      self#mapping
end



type opt_equality =
  | EQ_None
  | EQ_Axioms

class flag_equality
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_equality) =

object (self)

  method mapping = [
    (EQ_None, "None");
    (EQ_Axioms, "Axioms");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end


type opt_backtracking =
  | BT_Naive
  | BT_Backjumping

class flag_backtracking
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_backtracking) =

object (self)

  method mapping = [
    (BT_Naive, "Naive");
    (BT_Backjumping, "Backjumping");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end



type opt_iterative_deepening =
  | IT_TermDepth
  | IT_TermWeight

class flag_iterative_deepening
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_iterative_deepening) =

object (self)

  method mapping = [
    (IT_TermDepth, "TermDepth");
    (IT_TermWeight, "TermWeight");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end


type opt_restart =
  | RS_Eager
  | RS_Lazy
  | RS_Delayed

class flag_restart
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_restart) =

object (self)

  method mapping = [
    (RS_Eager, "Eager");
    (RS_Lazy, "Lazy");
    (RS_Delayed, "Delayed");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end


type opt_neg_assert_candidates =
  | NAC_Ignore
  | NAC_Lookahead
  | NAC_Use

class flag_neg_assert_candidates
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_neg_assert_candidates) =

object (self)

  method mapping = [
    (NAC_Ignore, "Ignore");
    (NAC_Lookahead, "Lookahead");
    (NAC_Use, "Use");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end



type opt_input_format =
  | FI_TME
  | FI_Darwin

class flag_input_format
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list)
  (__default: opt_input_format) =

object (self)

  method mapping = [
    (FI_TME, "tme");
    (FI_Darwin, "darwin");
  ]

  inherit ['a] flag_variant __id __short_name __long_name __description __default
end



exception VERSION

let help_short = "short"
let help_long = "long"

class flag_help_short
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list) =

object (self)
  inherit flag_unit __id __short_name __long_name __description

  method private update value = raise (Arg.Help help_short)
end

class flag_help_long
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list) =

object (self)
  inherit flag_unit __id __short_name __long_name __description

  method private update value = raise (Arg.Help help_long)
end



class flag_version
  (__id: flag_id)
  (__short_name: string) (__long_name: string) (__description: string list) =

object (self)
  inherit flag_unit __id __short_name __long_name __description

  method private update value = raise (VERSION)
end




type flag =
  | Bool of flag_bool
  | Int of flag_int
  | String of flag_string
  | Unit of flag_unit
  | Equality of flag_equality
  | Backtracking of flag_backtracking
  | IterativeDeepening of flag_iterative_deepening
  | Restart of flag_restart
  | HornCandidates of flag_neg_assert_candidates
  | InputFormat of flag_input_format


type flags = {
  flags : flag list;
  mutable file_name : string;
}




(* signatures *)

class type ['a, 'b] flag_type =
object
  method id : flag_id
  method short_name : string
  method long_name : string
  method description : string list
  method value : 'a

  method argument_to_string : 'b -> string
  method value_to_string : 'a -> string
  method opt_to_string : string list
  method signature : string

  method set : 'b -> unit
  method set_opt : 'a -> unit
end

type flag_type_bool = (bool, bool) flag_type
type flag_type_int = (int, int) flag_type
type flag_type_unit = (unit, unit) flag_type
type flag_type_string = (string, string) flag_type
type flag_type_equality = (opt_equality, string) flag_type
type flag_type_backtracking = (opt_backtracking, string) flag_type
type flag_type_iterative_deepening = (opt_iterative_deepening, string) flag_type
type flag_type_restart = (opt_restart, string) flag_type
type flag_type_neg_assert_candidates = (opt_neg_assert_candidates, string) flag_type
type flag_type_input_format = (opt_input_format, string) flag_type








(*** flag processing ***)

type 'a apply_to_flag = {
  apply : 'b 'c. ('b, 'c) flag_type -> 'a;
}

let coerce flag_base =
  (flag_base :> ('a, 'b) flag_type)

let process_flag (flag: flag) (apply: 'a apply_to_flag) : 'a =
  match flag with
    | Int                flag_base -> apply.apply (coerce flag_base)
    | Bool               flag_base -> apply.apply (coerce flag_base)
    | String             flag_base -> apply.apply (coerce flag_base)
    | Unit               flag_base -> apply.apply (coerce flag_base)
    | Equality           flag_base -> apply.apply (coerce flag_base)
    | Backtracking       flag_base -> apply.apply (coerce flag_base)
    | IterativeDeepening flag_base -> apply.apply (coerce flag_base)
    | Restart            flag_base -> apply.apply (coerce flag_base)
    | HornCandidates     flag_base -> apply.apply (coerce flag_base)
    | InputFormat        flag_base -> apply.apply (coerce flag_base)






(*** flag access ***)


let short_name_of_flag (flag: flag) : string =
  let apply = {
    apply = fun flag -> flag#short_name
  }
  in
    process_flag flag apply

let get_flag (flags: flag list) (id: flag_id) : flag =
  let apply = {
    apply = fun flag -> id = flag#id
  }
  in
    List.find
      (fun flag ->
	 process_flag flag apply
    )
    flags



(* flag access and coercing to signatue *)

let get_flag_bool (flags: flag list) (id: flag_id) : flag_type_bool =
  match get_flag flags id with
    | Bool flag -> (flag :> flag_type_bool)
    | _ -> failwith "Flags.get_flag_bool"
  
let get_flag_int (flags: flag list) (id: flag_id) : flag_type_int =
  match get_flag flags id with
    | Int flag -> (flag :> flag_type_int)
    | _ -> failwith "Flags.get_flag_int"

let get_flag_unit (flags: flag list) (id: flag_id) : flag_type_unit =
  match get_flag flags id with
    | Unit flag -> (flag :> flag_type_unit)
    | _ -> failwith "Flags.get_flag_unit"

let get_flag_string (flags: flag list) (id: flag_id) : flag_type_string =
  match get_flag flags id with
    | String flag -> (flag :> flag_type_string)
    | _ -> failwith "Flags.get_string_flag"

let get_flag_equality (flags: flag list) (id: flag_id) : flag_type_equality =
  match get_flag flags id with
    | Equality flag -> (flag :> flag_type_equality)
    | _ -> failwith "Flags.get_flag_equality"

let get_flag_backtracking (flags: flag list) (id: flag_id) : flag_type_backtracking =
  match get_flag flags id with
    | Backtracking flag -> (flag :> flag_type_backtracking)
    | _ -> failwith "Flags.get_flag_backtracking"

let get_flag_iterative_deepening (flags: flag list) (id: flag_id) : flag_type_iterative_deepening =
  match get_flag flags id with
    | IterativeDeepening flag -> (flag :> flag_type_iterative_deepening)
    | _ -> failwith "Flags.get_flag_iterative_deepening"

let get_flag_restart (flags: flag list) (id: flag_id) : flag_type_restart =
  match get_flag flags id with
    | Restart flag -> (flag :> flag_type_restart)
    | _ -> failwith "Flags.get_flag_restart"

let get_flag_neg_assert_candidates (flags: flag list) (id: flag_id) : flag_type_neg_assert_candidates =
  match get_flag flags id with
    | HornCandidates flag -> (flag :> flag_type_neg_assert_candidates)
    | _ -> failwith "Flags.get_flag_neg_assert_candidates"

let get_flag_input_format (flags: flag list) (id: flag_id) : flag_type_input_format =
  match get_flag flags id with
    | InputFormat flag -> (flag :> flag_type_input_format)
    | _ -> failwith "Flags.get_flag_input_format"

(* flag access by name *)

let resolve (flags: flags) = get_flag_bool flags.flags FL_Resolve
let subsume (flags: flags) = get_flag_bool flags.flags FL_Subsume
let compact (flags: flags) = get_flag_bool flags.flags FL_Compact
let productivity (flags: flags) = get_flag_bool flags.flags FL_Productivity
let mixed_literals (flags: flags) = get_flag_bool flags.flags FL_MixedLiterals
(*let lemma (flags: flags) = get_flag_bool flags.flags FL_Lemma*)

let equality (flags: flags) = get_flag_equality flags.flags FL_Equality
let plus_v (flags: flags) = get_flag_bool flags.flags FL_PlusV
let backtracking (flags: flags) = get_flag_backtracking flags.flags FL_Backtracking
let iterative_deepening (flags: flags) = get_flag_iterative_deepening flags.flags FL_IterativeDeepening
let deepening_bound (flags: flags) = get_flag_int flags.flags FL_DepthBound
let lookahead_exceeding (flags: flags) = get_flag_bool flags.flags FL_LookaheadExceeding
let restart (flags: flags) = get_flag_restart flags.flags FL_Restart
(*let random_jump (flags: flags) = get_flag_bool flags.flags FL_RandomJump*)
let neg_assert_candidates (flags: flags) = get_flag_neg_assert_candidates flags.flags FL_NegAssertCandidates
let iterative_deepening (flags: flags) = get_flag_iterative_deepening flags.flags FL_IterativeDeepening
let input_format (flags: flags) = get_flag_input_format flags.flags FL_InputFormat

let time_out_CPU (flags: flags) = get_flag_int flags.flags FL_TimeOutCPU
let time_out_WC (flags: flags) = get_flag_int flags.flags FL_TimeOutWC

let print_level (flags: flags) = get_flag_int flags.flags FL_PrintLevel
let print_configuration (flags: flags) = get_flag_bool flags.flags FL_PrintConfiguration
let print_statistics (flags: flags) = get_flag_bool flags.flags FL_PrintStatistics
let print_model_context (flags: flags) = get_flag_bool flags.flags FL_PrintModelContext
let print_model_context_file (flags: flags) = get_flag_string flags.flags FL_PrintModelContextFile
let print_model_DIG (flags: flags) = get_flag_bool flags.flags FL_PrintModelDIG
let print_model_DIG_file (flags: flags) = get_flag_string flags.flags FL_PrintModelDIGFile
let print_model_ARM (flags: flags) = get_flag_bool flags.flags FL_PrintModelARM

let print_derivation_online (flags: flags) = get_flag_bool flags.flags FL_PrintDerivationOnline
(*
let print_derivation (flags: flags) = get_flag_bool flags.flags FL_PrintDerivation
let print_derivation_pruned (flags: flags) = get_flag_bool flags.flags FL_PrintDerivationPruned
let print_derivation_file (flags: flags) = get_flag_string flags.flags FL_PrintDerivationFile
let print_derivation_pruned_file (flags: flags) = get_flag_string flags.flags FL_PrintDerivationPrunedFile
let print_derivation_dot (flags: flags) = get_flag_string flags.flags FL_PrintDerivationDot
let print_derivation_dot_pruned (flags: flags) = get_flag_string flags.flags FL_PrintDerivationDotPruned
*)
let print_derivation_context_unifier (flags: flags) = get_flag_bool flags.flags FL_PrintDerivationContextUnifier

let print_assert_candidates (flags: flags) = get_flag_bool flags.flags FL_PrintAssertCandidates
let print_split_candidates (flags: flags) = get_flag_bool flags.flags FL_PrintSplitCandidates

let file_name (flags: flags) = flags.file_name





(*** another flag type, needs some of the flag access functions above ***)

class flag_verbosity
  (__id: flag_id)
  (__short_name: string) (__long_name: string)
  (__default: int) (__flags: flag list) =

object (self)
  inherit flag_int __id __short_name __long_name [] __default

  val _print_flag_ids = [
    FL_PrintConfiguration;
    FL_PrintStatistics;
    FL_PrintModelContext;
    FL_PrintModelDIG;
    FL_PrintModelARM;
    FL_PrintDerivationOnline;
(*    FL_PrintDerivation;
    FL_PrintDerivationPruned;*)
    FL_PrintDerivationContextUnifier;
    FL_PrintAssertCandidates;
    FL_PrintSplitCandidates;
  ]

  val _levels =
    [|
      ([ ],
       "show the derivation result");
      ([ FL_PrintConfiguration; FL_PrintStatistics ],
       "... the configuration and statistic");
      ([ FL_PrintModelDIG ],
       "... the found model");
      ([ FL_PrintDerivationOnline ],
       "... the derivation steps");
      ([ FL_PrintDerivationContextUnifier ],
       "... the context unifiers underlying the derivation steps");
    |]

  val mutable _flags = __flags

  (* flags do not exists when this object is constructed,
     so they have to be given later on *)
  method set_flags flags = 
    _flags <- flags;
    (* propagate verbosity settings to the new flags *)
    self#update _value


  method flag_ids_for_level level : flag_id list =
    let rec do_flag_ids_for_level i acc =
      if i <= level then
	let (flags, _) =
	  _levels.(i)
	in
	  do_flag_ids_for_level (i + 1) (acc @ flags)
      else
	acc      
    in
      do_flag_ids_for_level 0 []
    
  method description =
    let level_descriptions =
      Array.mapi
	(fun i (flag_ids, description) ->
	   let short_names =
	     List.fold_left
	       (fun acc flag_id ->
		  try
		    short_name_of_flag (get_flag _flags flag_id) :: acc
		  with
		    | Not_found ->
			acc
	       )
	       []
	       flag_ids
	   in
	     "  " ^ string_of_int i ^ " -> "
	     ^ description
	     ^ " (" ^ String.concat "," (List.rev short_names) ^ ")"
	)
	_levels
    in
    let description =
      Array.fold_left
	(fun acc description ->
	   description :: acc
	)
	["Specify the verbosity:"]
	level_descriptions
    in
      List.rev description


  method check (value: int) : bool =
    (value >= 0)
    &&
    (value < Array.length _levels)

  method opt_to_string =
    Array.to_list (
      Array.mapi
        (fun i _ ->
	   string_of_int i
        )
        _levels
    )


  method update (value: int) =
    (* unset all print flags *)
    List.iter
      (fun id ->
	 try
	   (get_flag_bool _flags id)#set false
	 with
	   | Not_found ->
	       (* option deactivated *)
	       ()
      )
      _print_flag_ids;

    (* set the used print flags *)
    List.iter
      (fun id ->
	 try
	   (get_flag_bool _flags id)#set true
	 with
	   | Not_found ->
	       (* option deactivated *)
	       ()
      )
      (self#flag_ids_for_level value)

  (* check that the initial default is valid *)
  initializer self#set _value
end












(*** flags creation ***)


let create (file_name: string) : flags =
  (* verbosity needs access to the flags -
     which are not created yet.
     so first create and keep it here with type flag_verbosity,
     store it coerced to the basic flag type in flags,
     and after the flag creation update it *)
  let verbosity =
    new flag_verbosity
      FL_PrintLevel
      "-pl"
      "--print-level"
      1
      []
  in

  let flags = [
    Bool (
      new flag_bool
	    FL_Resolve
	    "-ar"
	    "--resolve"
	    ["Apply the Resolve inference rule."]
	    true
    );

    Bool (
      new flag_bool
	    FL_Subsume
	    "-as"
	    "--subsume"
	    ["Apply the Subsume inference rule."]
	    true
    );

    Bool (
      new flag_bool
	    FL_Compact
	    "-ac"
	    "--compact"
	    ["Apply the Compact inference rule."]
	    true
    );

    Bool (
      new flag_bool
	    FL_Productivity
	    "-pr"
	    "--productivity"
	    ["Split only on productive remainders."]
	    true
    );

    Bool (
      new flag_bool
	    FL_MixedLiterals
	    "-ml"
	    "--mixed-literals"
	    ["Context literals may contain both universal and parametric variables."]
	    true
    );
(* Disable for CASC    
    Bool (
      new flag_bool
	    FL_Lemma
	    "-lm"
	    "--lemma"
	    ["Computes and uses lemmas, i.e. clauses entailed by the input clause set."]
	    false
    );
*)
    Bool (
      new flag_bool
	    FL_PlusV
	    "-pv"
	    "--plus-v"
	    ["Use +v instead of -v for the initial interpretation."]
	    false
    );
    
    Equality (
      new flag_equality
	   FL_Equality
	   "-eq"
	   "--equality"
	   ["Specifies if the predicate '=' is interpreted as equality";
	    "(e.g. as in (p(a) = p(b)) ), and how equality is handled."]
	   EQ_None
    );

    Backtracking (
      new flag_backtracking
	    FL_Backtracking
	    "-bt"
	    "--backtracking"
	    ["Specifies the backtracking method.";
	     "Note: For Horn problems naive backtracking is always used";
	     "as no backtracking occurs.";
	    ]
	    BT_Backjumping
    );

    IterativeDeepening (
      new flag_iterative_deepening
	   FL_IterativeDeepening
	   "-id"
	   "--iterative-deepening"
	   ["Specifies the criterion for the iterative deepening."]
	   IT_TermDepth
    );

    Int (
      new flag_int
	   FL_DepthBound
	   "-db"
	   "--depth-bound"
	   ["The initial iterative deepening bound."]
	   2
    );

    Bool (
      new flag_bool
	    FL_LookaheadExceeding
	    "-lx"
	    "--lookahead-all"
	    ["Apply the Close lookahead to all Assert candidate literals,";
	     "instead of only to the candidates within the current depth bound."]
	    false
    );


    Restart (
      new flag_restart
	   FL_Restart
	   "-rs"
	   "--restart"
	   ["Specifies how restarting is done when a branch is exhausted";
	    "within the current depth bound.";
	    "If a candidate exceeding the depth bound has been computed:";
	    " - 'Eager' restarts immediately with an increased depth bound.";
	    " - 'Lazy' backtracks to the first (left split) choice point";
	    "   where a candidate has been dropped and continues the derivation";
	    "   with the corresponding right split.";
	    " - 'Delayed' drops a candidate only after an exhausted branch";
	    "   has been found and a candidate is still applicable.";
	    "   Then it backtracks to the most recent (left split) choice point";
	    "   where a candidate has been dropped and then proceeds like 'Lazy'.";
	    "Note: In general 'Lazy' behaves best,";
	    "'Eager' is more performant for unsatisfiable problems,";
	    "while 'Delayed' (slower and higher memory consumption)";
	    "makes only sense for a search for shallow models.";
	   ]
	   RS_Lazy
    );

(* Disable for CASC    
    Bool (
      new flag_bool
	    FL_RandomJump
	    "-rj"
	    "--random-jump"
   	    ["Only applicable in conjunction with lazy restarting.";
	     "In order to increase the likelihood of finding a model quickly";
	     "the search space (tree) is not explored 'in order'.";
	     "Instead, from time to time (based on a given CPU timeout)";
	     "a jump to a different position in the search space is done.";
	     "The search remains complete, as the search is eventually continued";
	     "at all such left positions, if no model is found previously."
	    ]
	    false
    );
*)
    HornCandidates (
      new flag_neg_assert_candidates
	   FL_NegAssertCandidates
	   "-nac"
	   "--neg-assert-candidates"
	   ["Only applicable for Horn problems.";
	    "Specifies how negative Assert candidates are treated:";
	    " - 'Ignore' completely ignores negative candidates.";
	    " - 'Lookahead' uses negative candidates only for the Close lookahead.";
	    " - 'Use' uses negative candidates for all purposes.";
	    "Note: All negative Assert candidates";
	    "can be ignored without loss of completeness."
	   ]
	   NAC_Ignore
    );

    InputFormat (
      new flag_input_format
	   FL_InputFormat
	   "-if"
	   "--input-format"
	   ["Specifies the default format used to parse the input.";
	    "This setting is ignored if the format can be deduced from the file extension.";
	   ]
	   FI_TME
    );



    Int (
      new flag_int
	    FL_TimeOutCPU
	    "-to"
	    "--timeout-cpu"
	    ["Terminates the prover after n seconds of used processor time.";
	     "A value less than or equal to 0 disables the timeout."]
	    0
    );

    Int (
      new flag_int
	    FL_TimeOutWC
	    "-tow"
	    "--timeout-wc"
	    ["Terminates the prover after n seconds of used real (wall clock) time.";
	     "A value less than or equal to 0 disables the timeout."]
	    0
    );

    (* printing *)

    Int (verbosity :> flag_int);

    Bool (
      new flag_bool
	    FL_PrintConfiguration
	    "-pc"
	    "--print-configuration"
	    ["Print the system configuration."]
	    false
    );

    Bool (
      new flag_bool
	    FL_PrintStatistics
	    "-ps"
	    "--print-statistics"
	    ["Print statistics after the derivation terminates."]
	    false
    );

    Bool (
      new flag_bool
	    FL_PrintModelContext
	    "-pmc"
	    "--print-context"
	    ["Print a found model in the context format."]
	    false
    );

    String (
      new flag_file
	    FL_PrintModelContextFile
	    "-pmcf"
	    "--print-context-file"
	    ["Print a found model in the context format to a file."]
	    ""
    );

    Bool (
      new flag_bool
	    FL_PrintModelDIG
	    "-pmd"
	    "--print-DIG"
	    ["Print a found model in the DIG (Disjunctions of Implicit Generalizations) format."]
	    false
    );
    
    String (
      new flag_file
	    FL_PrintModelDIGFile
	    "-pmdf"
	    "--print-DIG-file"
	    ["Print a found model in the DIG format to a file."]
	    ""
    );

    (* :TODO: *)
    (*
    Bool (
      new flag_bool
	    FL_PrintModelARM
	    "-pmc"
	    "--print-model-ARM"
	    ["Print a found model in the ARM (Atomic Representation of Models) format.";
	    "This is (theoretically) not always possible, then a maximal reduced DIG representation is printed."]
	    false
    );
    *)

    Bool (
      new flag_bool
	    FL_PrintDerivationOnline
	    "-pdo"
	    "--print-derivation-online"
	    ["Print the derivation (Assert, Split, Close, ...) in real-time.";
	    (* "This also includes restarts,";
	     "while all other methods only log the derivation from the last restart on."*)
	    ]
	    false
    );
(*
    Bool (
      new flag_bool
	    FL_PrintDerivation
	    "-pd"
	    "--print-derivation"
	    ["Print the derivation after it is finished."]
	    false
    );

    String (
      new flag_file
	   FL_PrintDerivationFile
	   "-pdf"
	   "--print-derivation-file"
	   ["Print the derivation to a file."]
	   ""
    );


    Bool (
      new flag_bool
	    FL_PrintDerivationPruned
	    "-pdp"
	    "--print-derivation-pruned"
	    ["Print the pruned derivation after it is finished.";
	     "i.e. split decisions without any impact on the derivation are omitted."]
	    false
    );

    String (
      new flag_file
	   FL_PrintDerivationPrunedFile
	   "-pdpf"
	   "--print-derivation-pruned-file"
	   ["Print the pruned derivation to a file."]
	   ""
    );

    String (
      new flag_file
	   FL_PrintDerivationDot
	   "-pdot"
	   "--print-derivation-dot"
	   ["Print the derivation in the graphviz dot format to a file."]
	   ""
    );


    String (
      new flag_file
	   FL_PrintDerivationDotPruned
	   "-pdotp"
	   "--print-derivation-dot-pruned"
	   ["Print the pruned derivation in the graphviz dot format to a file."]
	   ""
    );
*)
    Bool (
      new flag_bool
	    FL_PrintDerivationContextUnifier
	    "-pdcu"
	    "--print-context-unifier"
	    ["Modifies any derivation output by additionally printing the context unifier";
	     "for each Assert, Split, and Close application."]
	    false
    );

    Bool (
      new flag_bool
	    FL_PrintAssertCandidates
	    "-pac"
	    "--print-assert-candidates"
	    ["Print each applicable Assert candidate."]
	    false
    );

    Bool (
      new flag_bool
	    FL_PrintSplitCandidates
	    "-psc"
	    "--print-split-candidates"
	    ["Print each applicable Split candidate."]
	    false
    );

    Unit ((
      new flag_help_short
	FL_Help
	"-h"
	"--help"
	["Print a short description of the program usage and options."]
	  ) :> flag_unit
    );

    Unit ((
      new flag_help_long
	FL_Help
	"-help"
	"--more-help"
	["Print a description of the program usage and options."]
	  ) :> flag_unit
    );

    Unit ((
      new flag_version
	FL_Version
	"-v"
	"--version"
	["Print the program version."]
	  ) :> flag_unit
    );
  ]
  in
    verbosity#set_flags flags;

    {
      flags = flags;
      file_name = file_name;
    }












(*** creation of the specification used by the Arg module to evaluate the command line ***)

(* map a flag type to the appropriate setting function *)
let create_spec (flag: flag) : Arg.spec =
  match flag with
    | Int flag ->
	Arg.Int flag#set

    | Bool flag ->
	Arg.Bool flag#set

    | String flag ->
	Arg.String flag#set

    | Unit flag ->
	Arg.Unit flag#set

    | Equality flag ->
	Arg.String flag#set

    | Backtracking flag ->
	Arg.String flag#set

    | IterativeDeepening flag ->
	Arg.String flag#set

    | Restart flag ->
	Arg.String flag#set

    | HornCandidates flag ->
	Arg.String flag#set

    | InputFormat flag ->
	Arg.String flag#set


let create_spec_entry (flag: ('a, 'b) flag_type) (name: string) (update: Arg.spec) =
  (name, update, String.concat "\n" flag#description)

(* traverse all flags and create for each a spec bound to its short and long name,
   e.g. to -v and --version *)
let create_spec_list (flags: flag list) : (Arg.key * Arg.spec * Arg.doc) list =
  let create_spec_entry_short update = {
    apply = fun flag -> create_spec_entry flag flag#short_name update
  }
  in
  let create_spec_entry_long update = {
    apply = fun flag -> create_spec_entry flag flag#long_name update
  }
  in
    List.fold_left
      (fun acc flag ->
	 let spec =
	   create_spec flag
	 in
	   process_flag flag (create_spec_entry_short spec)
	   ::
	   process_flag flag (create_spec_entry_long spec)
	   ::
	   acc
      )
      []
      (List.rev flags)








(*** help messages ***)


(* creates the help for one flag *)
let create_help_for_flag_short (flag: ('a, 'b) flag_type) (signature: string) (default: string) : string =
  " " ^ Print.right_pad flag#short_name 6
  ^ " " ^ Print.right_pad flag#long_name 30
  ^ " " ^ Print.right_pad default 15
  ^ " " ^ signature
  ^ "\n"

let create_help_for_flag_long (flag: ('a, 'b) flag_type) (signature: string) (default: string) : string =
  create_help_for_flag_short flag signature default
  ^ String.concat "\n" (List.map (fun line -> String.make 8 ' ' ^ line) flag#description)
  ^ "\n"
  ^ "\n"

(* creates the help for all flags *)
let create_help (flags: flags) (apply: string apply_to_flag) : string =
  let help_header =
    Print.version
    ^ "\n"
    ^ "\n"
    ^ "Usage: darwin [options] file\n"
    ^ "\n"
  in

  let help_flags =
    List.fold_left
      (fun acc flag ->
         acc ^ process_flag flag apply
      )
      ""
      flags.flags
  in
    help_header ^ help_flags




(*** flags creation and setting by evaluation of the command line *)

let eval_command_line () : flags =
  let flags =
    create ""
  in
  let spec_list =
    create_spec_list flags.flags
  in
    begin
      try
	Arg.parse_argv
	  Sys.argv
	  spec_list
	  (fun name -> flags.file_name <- name)
	  "";

	if flags.file_name = "" then begin
	  raise (Arg.Bad ("problem file not specified"));
	end;
	flags

      with
	| VERSION ->
	    print_endline Print.version;
	    exit 0

	| Arg.Help help ->
	    if help = help_short then
	      let apply = {
		apply = fun flag ->
		  create_help_for_flag_short flag (flag#signature) (flag#value_to_string flag#value)
	      }
	      in
		print_endline (create_help flags apply)

	    else if help = help_long then
	      let apply = {
		apply = fun flag ->
		  create_help_for_flag_long flag (flag#signature) (flag#value_to_string flag#value)
	      }
	      in
		print_endline (create_help flags apply)
		  
	    else
	      failwith "Flags.Arg.Help";

	    exit 0

	| Arg.Bad error_msg ->
	    (* only print the first line of the error msg,
	       the rest is a flag description *)
	    let first_line =
	      try
		let eol =
		  String.index error_msg '\n'
		in
		  String.sub error_msg 0 eol
	      with
		| Not_found ->
		    error_msg
	    in
	      print_endline first_line;
	      print_newline ();
	      print_endline ("Try darwin -h for further information.");
	      exit 0;
    end
