(* $Id: log.ml,v 1.2 2004/12/28 14:06:49 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



type literal = Term.literal
type config = Config.config
type state = State.state
type choice_point = State.choice_point
type context = Context.context
type raw_context_unifier = Context_unifier.raw_context_unifier




(*** public interface ***)

class type log =
object
  method assert_: literal -> raw_context_unifier -> unit
  method split_unit: literal -> raw_context_unifier -> unit
  method split_left: choice_point -> literal -> raw_context_unifier -> unit
  method split_right: (*choice_point -> choice_point ->*)
    literal -> raw_context_unifier -> unit
  method close: choice_point -> raw_context_unifier -> unit
  method incomplete: choice_point -> unit

  method prune: choice_point -> unit
  method is_prune: bool

  method finalize: unit
end








(*** log set ***)

(* contains a set of logs and simply delegates each call to all logs. *)

class log_set
  (protocols: log list) =

object (self)
  val _protocols = protocols

  method iter func =
    List.iter func _protocols

  method exists func =
    List.exists func _protocols

  method assert_ literal raw_context_unifier =
    self#iter
      (fun protocol -> protocol#assert_ literal raw_context_unifier)

  method split_left choice_point literal raw_context_unifier =
    self#iter
      (fun protocol -> protocol#split_left choice_point literal raw_context_unifier)

  method split_right (*left choice_point*) literal raw_context_unifier =
    self#iter
      (fun protocol -> protocol#split_right(* left choice_point*) literal raw_context_unifier)

  method split_unit literal raw_context_unifier =
    self#iter
      (fun protocol -> protocol#split_unit literal raw_context_unifier)

  method close choice_point raw_context_unifier =
    self#iter
      (fun protocol -> protocol#close choice_point raw_context_unifier)

  method incomplete choice_point =
    self#iter
      (fun protocol -> protocol#incomplete choice_point)

  method prune choice_point =
    self#iter
      (fun protocol -> protocol#prune choice_point)

  method is_prune =
    self#exists
      (fun protocol -> protocol#is_prune)

  method finalize =
    self#iter
      (fun protocol -> protocol#finalize)

end





(*** print ***)

(* textual representation of each rule. *)

(* extended: print the context unifier for any rule *)
class print ~(extended: bool) (config: config) (context: context)
   (out: out_channel) =
object (self)

  method private print id string =
    output_string out (State.id_to_string id ^ string ^ "\n")

  method private print_context_unifier context_unifier =
    if extended then
      output_string out (
	Context_unifier.raw_context_unifier_to_string context_unifier
      )

  method private print_remainder context_unifier =
    if extended then
      output_string out (
	Selection_split.remainder_literals_to_string
	  config context context_unifier
      )

  method assert_ id literal (raw_context_unifier: raw_context_unifier) =
    self#print id ("Assert: " ^ Term.literal_to_string literal);
    self#print_context_unifier raw_context_unifier;
    
  method split_unit id literal (raw_context_unifier: raw_context_unifier) =
    self#print id ("Unit Split: " ^ Term.literal_to_string literal);
    self#print_context_unifier raw_context_unifier;
    self#print_remainder raw_context_unifier;

  method split_left id literal (raw_context_unifier: raw_context_unifier) =
    self#print id ("Split Left: " ^ Term.literal_to_string literal);
    self#print_context_unifier raw_context_unifier;
    self#print_remainder raw_context_unifier;

  method split_right id literal (raw_context_unifier: raw_context_unifier) =
    self#print id ("Split Right: " ^ Term.literal_to_string literal);
    self#print_context_unifier raw_context_unifier;
    self#print_remainder raw_context_unifier;
    
  method close id (raw_context_unifier: raw_context_unifier) =
    self#print id ("Close: ");
    self#print_context_unifier raw_context_unifier;

  method incomplete id =
    self#print id ("Bound Reached: ")

  method backtrack id backtracked_id =
    self#print id ("Backtracking to: " ^ State.id_to_string backtracked_id)
(*
  method rename id old_id new_id =
    self#print id 
      ("Renaming: "
       ^ State.id_to_string old_id
       ^ "to "
       ^ State.id_to_string new_id
      )
*)
end













(*** log - pipe ****)

(* immediately write each derivation step *)

type pipe_node = {
  (* a choice point *)
  pi_choice_point: choice_point;

  (* the currently assigned id.
     May change during backtracking and branch reordering.  *)
  mutable pi_id: int;
}

class pipe (print: print) (state: State.state) =
object (self)
  (* the current derivation branch *)
  val mutable _branch: pipe_node list =
    let root =
      {
	pi_choice_point = State.root_choice_point state;
	pi_id = 0;
      }
    in
      [ root ]

  method current =
    match _branch with
      | [] -> failwith "Protocol.pipe.current"
      | current :: _ -> current
    
  method id =
    self#current.pi_id



  method assert_ literal (raw_context_unifier: raw_context_unifier) =
    print#assert_ self#id literal raw_context_unifier
      
  method split_unit literal (raw_context_unifier: raw_context_unifier) =
    print#split_unit self#id literal raw_context_unifier

  method split_left choice_point literal (raw_context_unifier: raw_context_unifier) =
    let node =
      {
	pi_choice_point = choice_point;
	pi_id = self#id + 1;
      }
    in
      _branch <- node :: _branch;
      print#split_left node.pi_id literal raw_context_unifier

  method split_right(* (left: choice_point) (right: choice_point) *)literal (raw_context_unifier: raw_context_unifier) =
    print#split_right self#id literal raw_context_unifier
(*
    (* first backtrack to the left split... *)
  
    let left_node =
      List.find
	(fun node ->
	   State.choice_point_equal left node.pi_choice_point
	)
	_branch
    in
      print#backtrack self#id left_node.pi_id;

      (* ...then reorder the branch... *)
      let new_id =
	self#reorder left_node
      in
      (* ...and finally register the right split. *)
      let right_node =
	{
	  pi_choice_point = right;
	  pi_id = new_id;
	}
      in
	
	_branch <- right_node :: _branch;
	print#split_right new_id literal raw_context_unifier;
*)

  method backtrack choice_point =
    let rec backtrack' branch =
      match branch with
	| [] ->
	    failwith "log.backtrack 1"

	| _ :: [] ->
	    failwith "log.backtrack 2"

	| head :: next :: tail ->
	    if State.choice_point_equal choice_point head.pi_choice_point then begin
	      print#backtrack self#id next.pi_id;
	      _branch <- (next :: tail)
	    end
	    else
	      backtrack' (next :: tail)
    in
      backtrack' _branch;


  method close choice_point (raw_context_unifier: raw_context_unifier) =
    print#close self#id raw_context_unifier;
    self#backtrack choice_point
      
  method incomplete choice_point =
    print#incomplete self#id;
    self#backtrack choice_point


  (* no pruning possible when the derivation is immediately printed
     without any postprocessing *)
  method prune (choice_point: choice_point) =
    ()

  method is_prune =
    false

  method finalize =
    ()
      (*
  (* backtrack to the left split and remove all invalid splits on the way.
     keep and rename the still valid splits (reordering). *)
  method reorder left_node =

    let rec reorder' branch kept : int =
      match branch with
	| [] ->
	    failwith "Log.Pipe.reorder"
	      
	| node :: tail ->
	    (* ok, backtracked to the left split. *)
	    if node == left_node then begin

	      match kept with
		| [] ->
		    (* no kept nodes, so nothing to reorder *)
		    _branch <- branch;
		    left_node.pi_id

		| _ ->
		    (* move the kept nodes over the retracted ones including the left split. *)
		    let leaf_id, new_branch =
		      List.fold_left
			(fun (id, tail) kept ->
			   (* rename *)
			   self#rename kept.pi_id id;
			   kept.pi_id <- id;
			   
			   (* keep and increase counter *)
			   ((id + 1), kept :: tail)
			)
			(left_node.pi_id, tail)
			kept
		    in
		      (* now rename and move the left split. *)
		      self#rename left_node.pi_id leaf_id;
		      left_node.pi_id <- leaf_id;

		      _branch <- left_node :: new_branch;
		      leaf_id
	    end

	    (* still backtracking to do... *)
	    else begin
	      if State.is_choice_point_valid node.pi_choice_point then begin
		reorder' tail (node :: kept)
	      end
		
	      else begin
		reorder' tail kept
	      end
	    end
    in
      reorder' _branch []
      *)
end






















(*
(*** log - process ***)



(* first gather the whole derivation,
   then print it after some potential postprocessing. *)



type extension =
  | Assert of literal * raw_context_unifier
  | Unit of literal * raw_context_unifier
  | Close of raw_context_unifier
  | Incomplete
  | Backtrack of (unit -> int)


(* this represents a choice point (ugly) *)
type process_node = {
  (* the choice point. *)
  pr_choice_point: choice_point;
 
  (* the id at the current status of the derivaiton. *)
  mutable pr_id: int;

  (* a left split? *)
  pr_is_left_split: bool;

  (* the split, or None for the root node *)
  pr_split: (literal * raw_context_unifier) option;

  (* for a left split with a right split - here is the right split. *)
  mutable pr_right_split: process_node option;

  (* for a right split - here is the right split. *)
  pr_left_split: process_node option;

  (* the subsequent rule applications - Assert, Unit Split, Close *)
  mutable pr_extensions: extension list;

  (* used during printing for assigning a temporary id,
     needed for pruning. *)
  mutable pr_effective_id: int;
}

let extend node extension =
  node.pr_extensions <- extension :: node.pr_extensions



class virtual process_print (prune: bool) =
object (self)
  method is_prune = prune

  (* print a header *)
  method virtual prefix: unit

  (* print a footer *)
  method virtual suffix: unit

  (* print_node node parent prints the split of the node. *)
  method virtual print: process_node -> process_node option -> unit
end



(* use a textual representation *)
class process_print_text ~(prune:bool) (print: print) =
object (self)
  inherit process_print prune

  method prefix =
    ()

  method suffix =
    ()

  method private print_extensions (node: process_node) =
    List.iter
      (fun extension ->
	 match extension with
	   | Assert (literal, raw_context_unifier) ->
	       print#assert_ node.pr_effective_id literal raw_context_unifier

	   | Unit (literal, raw_context_unifier) ->
	       print#split_unit node.pr_effective_id literal raw_context_unifier

	   | Close raw_context_unifier ->
	       print#close node.pr_effective_id raw_context_unifier

	   | Incomplete ->
	       print#incomplete node.pr_effective_id

	   | Backtrack left_split ->
	       print#backtrack node.pr_effective_id (left_split ())
      )
      (List.rev node.pr_extensions)

  method print (node: process_node) (parent: process_node option) : unit =
    begin
      match node.pr_split with
	| None ->
	    ()

	| Some (literal, raw_context_unifier) ->
	    if node.pr_is_left_split then
	      print#split_left node.pr_effective_id literal raw_context_unifier
	    else
	      print#split_right node.pr_effective_id literal raw_context_unifier
    end;

    self#print_extensions node

end


(* textual representation, plus header and footer for stdout. *)
class process_print_text_stdout ~(prune:bool) (print: print) =
object (self)
  inherit process_print_text ~prune:prune print

  method prefix =
    print_newline ();
    print_string ("START OF DERIVATION");
    if prune then
      print_string (" (PRUNED)");
    print_newline ();

  method suffix =
    print_string ("END OF DERIVATION");
    if prune then
      print_string (" (PRUNED)");
    print_newline ();
end





(* dot representation, i.e. textual description of a graph. *)
class virtual process_print_dot ~(prune: bool)
  (config: config) (context: context) (out: out_channel)  =
object (self)
  inherit process_print prune

  (*** color definitions ***)
  val color_split = "palegreen"
  val color_assert = "skyblue"
  val color_unit = "khaki1"
  val color_close = "lightcoral"
  val color_incomplete = "maroon3"



  (*** internal printing ***)

  method private print_line string =
    Print.output_line out string

  (* a node's subnode (split, extension) id is made of the node's choice point id,
     plus a sub id for its position in the split/assert/close sequence
     of the node. *)
  method private to_node_id (node: process_node) (id: int) : string =
    "node_"
    ^ string_of_int (State.id_of_choice_point node.pr_choice_point)
    ^ "_" ^ string_of_int id

  (* the id of the leaf node of the node's branch,
     needed to attach an edge to the child node. *)
  method private to_leaf_id (node: process_node) : string =
    self#to_node_id node (List.length node.pr_extensions)



  (* print a dot node. *)
  method private print_node (id: string) (label: string) (color: string) : unit =
    self#print_line (
      id
      ^ " [label = \"" ^ String.escaped label ^ "\", fillcolor = "
      ^ color
      ^ "]")

  (* print a dot edge to an (implicitely created) node *)
  method private print_edge (source: string) (target: string) (label: string) =
    self#print_line (
      source
      ^ " -> "
      ^ target
      ^ " [label = \"" ^ String.escaped label ^ "\"]"
    )




  (*** public ***)
  method prefix =
    (* start graph *)
    self#print_line ("digraph \"" ^ "proof" ^ "\" {");
    self#print_line "nodesep = \"0.2\"";
    self#print_line "ranksep = \"0.2\"";

    (* node defaults *)
    self#print_line "node [fontsize=\"10\", fontname=\"Arial\"]";
    self#print_line "node [shape=box, style = filled, height = \"0.3\"]";
    self#print_line "node [label=\"\", fillcolor = black]";

    (* edge defaults *)
    (* print_line "edge [fontsize=\"10\", fontname=\"Arial\", arrowsize = \"0.5\", arrowhead = \"none\"]" *)
    self#print_line "edge [fontsize=\"10\", fontname=\"Arial\", arrowhead = \"none\"]";

  method suffix =
    (* end graph *)
    self#print_line "}"
end



(* print a compact dot graph, where
   - nodes contain the literal added to the context
   - edges are empty
   that is, a split application has a node for the left split
   and a further node for the right split,
   which are both connected to the same parent node.
*)
class process_print_dot_compact ~(prune: bool)
  (config: config) (context: context) (out: out_channel)  =
object (self)
  inherit process_print_dot ~prune:prune config context out

  (* start from the first extension with id 1 and attach it to the split node. *)
  method private print_extensions (node: process_node) (is_first: bool) =
    let rec print_extensions' extensions id =
      match extensions with
	| [] ->
	    ()
	      
	| extension :: tail ->
	    let source = self#to_node_id node (id - 1)
	    and target = self#to_node_id node id
	    in
	      (* don't print an edge to the previous node
		 if this is in fact the root node of the graph. *)
	    let print_edge () =
	      if not (is_first && id = 1) then
		self#print_edge source target ""
	    in

	      begin
		match extension with
		  | Assert (literal, raw_context_unifier) ->
		      print_edge ();
		      self#print_node target (Term.literal_to_string literal) color_assert;
		      
		  | Unit (literal, raw_context_unifier) ->
		      print_edge ();
		      self#print_node target (Term.literal_to_string literal) color_unit;
		      
		  | Close raw_context_unifier ->
		      print_edge ();
		      self#print_node target "" color_close
		      
		  | Incomplete ->
		      print_edge ();
		      self#print_node target "" color_incomplete
			
		  | Backtrack left_split ->
		      ()
	      end;
	      print_extensions' tail (id + 1)
    in
      print_extensions' (List.rev node.pr_extensions) 1
	

  method print (node: process_node) (parent: process_node option) : unit =
      match node.pr_split with
	| None ->
	    (* the root node *)
	    self#print_extensions node true;
	    
	| Some (literal, raw_context_unifier) ->
	    (* do the edges for left and right split together *)
	    if node.pr_is_left_split then begin
	      let parent =
		match parent with
		  | None -> failwith "Log.Process_print_dot_compact.print"
		  | Some parent -> parent
	      in
	      let parent_node_id =
		self#to_leaf_id parent
	      in
		match node.pr_right_split with
		  | None ->
		      (* for a left split only,
			 don't print the edge if this is the root node of the graph. *)
		      begin
			match parent.pr_split, parent.pr_extensions with
			  | None, [] ->
			      (* the parent node is the empty root choice point *)
			      ()

			  | _ ->
			      (* the parent node consists of at least one rule application *)
			      self#print_edge parent_node_id (self#to_node_id node 0) "";
		      end

		  | Some right_split ->
		      (* for a left and right split, we need a common parent node.
			 if none exists, a default one will be created implicitely. *)
		      self#print_edge parent_node_id (self#to_node_id node 0) "";
		      self#print_edge parent_node_id (self#to_node_id right_split 0) ""
	    end;
	    
	    self#print_node
	      (self#to_node_id node 0)
	      (Term.literal_to_string literal)
	      color_split;
	      
	    self#print_extensions node false;
	    
end



(* print a detailed dot graph, where
   - nodes contain the rule's context unifier
   - and outgoing contain the literal added to the context
   that is, a split application has only one node
   with two outgoing edges.
*)
class process_print_dot_extended ~(prune: bool)
  (config: config) (context: context) (out: out_channel)  =
object (self)
  inherit process_print_dot ~prune:prune config context out

  (* returns the label of the edge rooted in the leaf node of this node. *)
  method private out_edge_label (node: process_node) : string =
    match node.pr_extensions with
      | [] ->
	  begin
	    match node.pr_split with
	      | None ->
		  ""
		    
	      | Some (literal, _) ->
		  Term.literal_to_string literal
	  end
	    
      | leaf :: _ ->
	  begin
	    match leaf with
	      | Assert (literal, _)
	      | Unit (literal, _) ->
		  Term.literal_to_string literal
		    
	      | Close _
	      | Incomplete
	      | Backtrack _ ->
		  ""
	    end;


  (* start from the first extension with id 1 and attach it to the split node.
     parent is None, if this is the root node,
     otherwise the parents node id and outgoing edge label.
  *)
  method private print_extensions (node: process_node) (parent: (string * string) option) : unit =
    let rec print_extensions' extensions id label =
      match extensions with
	| [] ->
	    ()
	    
	| extension :: tail ->
	    let source = self#to_node_id node (id - 1)
	    and target = self#to_node_id node id
	    in
	    let print_edge () =
	      (* don't print an edge to the previous node if it doesn't exist. *)
	      if id = 1 then begin
		match parent with
		  | None ->
		      ()

		  | Some (parent_node_id, parent_label) ->
		      self#print_edge parent_node_id target parent_label;
	      end
		
	      else begin
		self#print_edge source target label;
	      end
	    in
	      begin
	      match extension with
		| Assert (literal, raw_context_unifier) ->
		    print_edge ();
		    self#print_node target
		      (Context_unifier.raw_context_unifier_to_string raw_context_unifier)
		      color_assert;
		    print_extensions' tail (id + 1) (Term.literal_to_string literal)

		| Unit (literal, raw_context_unifier) ->
		    print_edge ();
		    self#print_node target
		      (Context_unifier.raw_context_unifier_to_string raw_context_unifier
		       ^
		       Selection_split.remainder_literals_to_string config context raw_context_unifier)
		      color_unit;
		    print_extensions' tail (id + 1) (Term.literal_to_string literal)
		    
		| Close raw_context_unifier ->
		    print_edge ();
		    self#print_node target
		      (Context_unifier.raw_context_unifier_to_string raw_context_unifier)
		      color_close;
		    print_extensions' tail (id + 1) ""
		      
		      
		| Incomplete ->
		    print_edge ();
		    self#print_node target "" color_incomplete;
		    print_extensions' tail (id + 1) ""
		      
		| Backtrack left_split ->
		    print_extensions' tail (id + 1) ""
	    end;
    in
      print_extensions' (List.rev node.pr_extensions) 1 ""


  method print (node: process_node) (parent: process_node option) : unit =
      match node.pr_split with
	| None ->
	    (* the root node *)
	    self#print_extensions node None;
	    
	| Some (literal, raw_context_unifier) ->
	    if node.pr_is_left_split then begin

	    (* do the edges for left and right split together
	       when the left split is handled. *)
	      let parent =
		match parent with
		  | None -> failwith "Log.Process_print_dot_extended.print parent"
		  | Some parent -> parent
	      in
	      let parent_node_id =
		self#to_leaf_id parent
	      in
		begin
		  match parent.pr_split, parent.pr_extensions with
		    | None, [] ->
			(* the parent node is the empty root choice point *)
			()
			  
		    | _ ->
		      (* the parent node consists of at least one rule application *)
			self#print_edge
			  parent_node_id
			  (self#to_node_id node 0)
			  (self#out_edge_label parent);
		end;


		(* print the left split plus extensions *)
		self#print_node
		  (self#to_node_id node 0)
		  (Context_unifier.raw_context_unifier_to_string raw_context_unifier
		   ^
		   Selection_split.remainder_literals_to_string config context raw_context_unifier)
		  color_split;

		self#print_extensions
		  node
		  (Some (self#to_node_id node 0, Term.literal_to_string literal));
	    end

	    else begin
	      (* attach the first extension to the split node,
		 that is the node created when the left split was handled. *)
	      match node.pr_left_split with
		| None ->
		    failwith "Log.Process_print_dot_extended.print right split"
		      
		  | Some left_split ->
		      self#print_extensions
			node
			(Some (
			  self#to_node_id left_split 0,
			  Term.literal_to_string literal
			))
	    end
end


class process (print: process_print list) (state: State.state) =
  let root = {
    pr_choice_point = State.root_choice_point state;
    pr_id = 0;
    pr_is_left_split = true;
    pr_right_split = None;
      pr_left_split = None;
      pr_split = None;
      pr_extensions = [];
      pr_effective_id = 0;
  }
  in

object (self)
  val mutable _branch: process_node list =
      [ root ]

  val mutable _pruned: choice_point list = []

  val mutable _current = Some root

  method current =
    match _current with
      | None -> failwith "Protocol.Process_base.current"
      | Some current -> current

(*
  method current =
    match _branch with
      | [] -> failwith "Protocol.Process_base.current"
      | current :: _ -> current
*)
  
  method id =
    self#current.pr_id


  method assert_ (literal: literal) (raw_context_unifier: raw_context_unifier) =
    extend self#current (Assert (literal, raw_context_unifier))
    
  method split_unit (literal: literal) (raw_context_unifier: raw_context_unifier) =
    extend self#current (Unit (literal, raw_context_unifier))

  method split_left choice_point literal (raw_context_unifier: raw_context_unifier) =
    let node = {
      pr_choice_point = choice_point;
      pr_id = self#id + 1;
      pr_is_left_split = true;
      pr_right_split = None;
      pr_left_split = None;
      pr_split = Some (literal, raw_context_unifier);
      pr_extensions = [];
      pr_effective_id = 0;
    }
    in
      _branch <- node :: _branch;
      _current <- Some split_left

  method split_right (*(left: choice_point) (right: choice_point)*) (literal: literal) (raw_context_unifier: raw_context_unifier) =
(*
      (* first close the branch by backtracking to the left split. *)
      extend self#current (Backtrack (fun unit -> left_node.pr_effective_id));

      (* then reorder the branch if dynamic backtracking has been used. *)
      let new_id =
	self#reorder left_node
      in
*)

      (* and finally create the right split and register it. *)
      let right_node = {
	pr_choice_point = backjump_node.pr_choice_point;
	pr_id = backjump_node.pr_id;
	pr_is_left_split = false;
	pr_right_split = None;
	pr_left_split = Some self#current;
	pr_split = Some (literal, raw_context_unifier);
	pr_extensions = [];
	pr_effective_id = 0;
      }
      in
	left_node.pr_right_split <- Some right_node;
	
	_branch <- right_node :: _branch;
      _current <- Some right_node

  method backtrack choice_point =
    extend self#current (Backtrack (fun unit -> self#current_.pr_effective_id));
    let backjump_node =
      List.find
	(fun node ->
	   State.is_choice_point_equal node.pr_choice_point choice_point
	)
	_branch
    in
      _current <- Some backjump_node

  method close choice_point (raw_context_unifier: raw_context_unifier) =
    extend self#current (Close raw_context_unifier);
    self#backtrack choice_point
      
  method incomplete choice_point =
    extend self#current Incomplete;
    self#backtrack choice_point

(*
  method rename (source: int) (target: int) =
    ()
*)    
  method prune (choice_point: choice_point) =
    _pruned <- choice_point :: _pruned

  method is_prune : bool =
    List.exists (fun print -> print#is_prune) print

  method finalize =
    List.iter
      (fun print ->
	 print#prefix;

	 (* locally pruned contains the nodes pruned in the current branch *)
	 let rec finalize'
	     (nodes: process_node list) (locally_pruned: process_node list) (parent: process_node option)
	     : unit =
	   match nodes with
	     | [] ->
		 ()
		   
	     | node :: tail ->
		 (* backtrack pruned nodes *)
		 let locally_pruned =
		   let rec prune_local pruned =
		     match pruned with
		       | head :: tail when
			   head.pr_id > node.pr_id ->
			   prune_local tail
			     
		       | _ ->
			   pruned
		   in
		     prune_local locally_pruned
		 in
		   
		   (* a pruned node? *)
		   if
		     print#is_prune
		     &&
		     List.exists (State.choice_point_equal node.pr_choice_point) _pruned
		   then begin
		     (* this node has been pruned, so don't print but keep it. *)
		     finalize' tail (node :: locally_pruned) parent
		   end
		     
		   (* a normal node to print. *)
		   else begin
		     (* adjust the effective id of the node and it's right split *)
		     if node.pr_is_left_split then begin
		       let ajusted_id =
			 node.pr_id - List.length locally_pruned
		       in
			 node.pr_effective_id <- ajusted_id;
			 match node.pr_right_split with
			   | None ->
			       ()

			   | Some right_split ->
			       right_split.pr_effective_id <- ajusted_id;
		     end;
		     
		     print#print node parent;
		     
		     finalize' tail locally_pruned (Some node)
		   end
	 in
	   finalize' (List.rev _branch) [] None;

	   print#suffix
      )
      print;
(*
  method reorder left_node =

    let rec reorder' (derivation) (retracted) (kept) : int =

    let rec reorder' (derivation) (retracted) (kept) : int =
      match derivation with
	| [] ->
	    failwith "Log.Process_base.reorder"
	      
	| node :: tail ->
	    (* ok, backtracked to the left split. *)
	    if node == left_node then begin

	      match kept with
		| [] ->
		    (* no kept nodes, so nothing to reorder *)
		    left_node.pr_id

		| _ ->
		    (* move the kept nodes over the retracted ones including the left split. *)
		    let leaf_id, new_branch =
		      List.fold_left
			(fun (id, tail) kept ->
			   (* rename *)
			   kept.pr_id <- id;
			   
			   (* keep and increase counter *)
			   ((id + 1), kept :: tail)
			)
			(left_node.pr_id, tail)
			kept
		    in

		    (* now rename and move the left split. *)
		    let new_branch =
		      left_node.pr_id <- leaf_id;

		      left_node :: new_branch;
		    in

		    (* and finally rename and move the other retracted choice points. *)
		    let _, new_branch =
		    List.fold_left
		      (fun (id, tail) retracted ->
			 (* only rename a right split when the corresponding
			    left split renamed as well. *)
			 if retracted.pr_is_left_split then begin
			   (* rename the left split *)
			   retracted.pr_id <- id;
			   (* rename the right split *)
			   begin
			     match retracted.pr_right_split with
			       | None -> ()
			       | Some right_split -> right_split.pr_id <- id;
			   end;

			   (* keep and increase counter *)
			   ((id + 1), retracted :: tail)
			 end

			 else begin
			   (* right split - keep node and counter *)
			   (id, retracted :: tail)
			 end
		      )
		      (leaf_id + 1, new_branch)
		      retracted
		    in
		      _branch <- new_branch;
		      leaf_id
	    end

	    else begin
	      if State.is_choice_point_valid node.pr_choice_point then begin
		reorder' tail retracted (node :: kept)
	      end
		
	      else begin
		reorder' tail (node :: retracted) kept
	      end
	    end
    in
      reorder' _branch [] []
*)
end


*)










(***  creation ***)



let create (config: config) (state: state) (context: context) : log =

  let extended =
    Config.print_derivation_context_unifier config
  in

  (* derivations to stdout *)
(*  let stdout_checks = [
    (Config.print_derivation config,
     fun () -> new process_print_text_stdout ~prune:false (new print ~extended:extended config context stdout)
    );

    (Config.print_derivation_pruned config,
     fun () -> new process_print_text_stdout ~prune:true (new print ~extended:extended config context stdout)
    );
  ]
  in

  (* derivations to files *)
  let file_checks = [
    (Config.print_derivation_file config,
     fun out -> new process_print_text ~prune:false (new print ~extended:extended config context out)
    );

    (Config.print_derivation_pruned_file config,
     fun out -> new process_print_text ~prune:true (new print ~extended:extended config context out)
    );

    (Config.print_derivation_dot config,
     fun out ->
       if extended then
	 new process_print_dot_extended ~prune:false config context out
       else
	 new process_print_dot_compact ~prune:false config context out
    );

    (Config.print_derivation_dot_pruned config,
     fun out ->
       if extended then
	 new process_print_dot_extended ~prune:true config context out
       else
	 new process_print_dot_compact ~prune:true config context out
    );

  ]
  in

  (* do stdout checks *)
  let prints : process_print list =
    List.fold_left
      (fun prints (check, print) ->
	 if check then
	   print () :: prints
	 else
	   prints
      )
      []
      stdout_checks
  in

  (* do file checks *)
  let prints =
    List.fold_left
      (fun prints (file_name, print) ->
	 if file_name <> "" then begin
	   try
	     let out =
	       open_out file_name
	     in
	       print out :: prints
	with
	  | Sys_error error_message ->
	      print_endline ("Print derivation to " ^ file_name ^ " failed.");
	      print_endline error_message;
	      prints
	 end

	 else
	   prints
      )
      prints
      file_checks
  in
    
  let logs = [
    (new process (List.rev prints) state :> log)
  ]
  in
*)
  let logs =
    []
  in

  let logs =
    if Config.print_derivation_online config then
      (new pipe (new print ~extended:extended config context stdout) state :> log) :: logs
    else
      logs
  in

  let protocol_set =
    new log_set logs
  in
    (protocol_set :> log)
