(* $Id: problem_literals.ml,v 1.16 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type binding = Subst.binding
type subst = Subst.subst
type choice_point = State.choice_point
type state = State.state
type packed_subst = Subst.packed_subst
type context = Context.context
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner
type context_unifier_space = Context_unifier.context_unifier_space
(*
type candidate_type =
  | Assert
  | Split
  | All
*)
(*** callback function types ***)
(*
type on_add_unit = Context.element -> unit
type on_add = input_partner -> context_partner -> candidate_type -> bool -> bool
type on_subsume =  State.choice_point -> unit
*)



(* binding database *)
module Bindings =
  Weak.Make (
    struct
      type t = binding

      let equal (binding1: binding) (binding2: binding) : bool =
	(Subst.var_equal binding1.Subst.sb_var binding2.Subst.sb_var)
	&&
	(Subst.term_equal binding1.Subst.sb_term binding2.Subst.sb_term)

      let hash (binding: binding) : int =
	let var_hash =
	  Var.id_of_var binding.Subst.sb_var.Subst.sv_var
	and term_hash =
	  match binding.Subst.sb_term.Subst.st_term with
	    | Term.Var var ->
		Var.id_of_var var
		
	    | Term.Const symbol ->
		Symbol.id symbol
		
	    | Term.Func (_, _, hash_value) ->
		hash_value
	in
	  (var_hash + term_hash) * 131

    end
  )



(* input literals with offsets -
   stored in a problem literal which ignores any offset *)
module Elements =
  Map.Make (
    struct
      type t = int

      let compare (first: int) (second: int) : int =
	Tools.compare_int first second
	(*compare first second*)
    end
  )

module SubsumedSpaces =
  Hashtbl.Make (
    struct
      type t = context_unifier_space
	  
      let equal x y =
	x == y
	  
      let hash x =
	x.Context_unifier.cus_id
		
    end
  )



(* data per problem literal with a specific offset *)
type element = {
  (* the problem literal is kept as an input_partner *)
  input_partner: input_partner;

  (* registered subsume call backs (for all clauses containing this literal) *)
(*  mutable callbacks_on_add: on_add list;*)
  mutable spaces: context_unifier_space list;
}

(* data per problem literal without offset *)
type problem_literal = {
  (* variants of the problem literal, i.e. pairing with different offsets *)
  mutable elements: element Elements.t;

  (* registered add call backs (for all clauses containing this literal) *)
(*  mutable callbacks_on_subsume: on_subsume list;*)

  (* which states subsume the literal? *)
  mutable subsumed_at: int (*State.choice_point option*);
}




(* problem literal accessed by literal. *)
module LiteralTable = Term.LiteralTable

(* (bunch of) problem literals accessed by sign and predicate. *)
module Index = Term.LiteralTypeTable


type problem_literals = {
  config: config;
  bound: bound;
  statistic: statistic;
  state: state;
  context: context;

  (* access to problem literals by literal - used during creation *)
  mutable problem_literals: problem_literal LiteralTable.t;

  (* access to problem literals by sign and predicate -
     used when checked against new context literals
     to quickly ignore literals with wrong sign and predicate *)
  index: ((literal * problem_literal) list) Index.t;

  (* all registered call backs for unit clauses *)
(*  mutable callbacks_on_add_unit: on_add_unit list;*)

  (* database for the bindings used in the substitution
     between the problem literals and the context partners (cp_subst).
     this can significantly decrease the memory usage, esp. for horn problems *)
  bindings: Bindings.t;


(*  subsumed_spaces: choice_point SubsumedSpaces.t;*)
  mutable subsumed_spaces: int array;
(*  mutable backtrack_spaces: bool array;*)

  v_par: Var.var;
  v_term: Term.term;
  v_cp_element: Context.element;
}








(*** creation ***)



let create (config: config) bound (statistic : statistic) (state: state) (context: context) : problem_literals =
  (* dummy values for -v/+v:
     just provide a parameter which is bound to a non-parameter term.
     
     as the remainder literals are built from the variables of the input clause
     and not from the parameters of the context literals,
     this binding is never applied.
     it is only checked in order to determine its remainder status,
     so a binding to any non-parameter is sufficient *)
  let v_par = 
    Var.create_parametric 0
  in
  let v_name =
    if Config.plus_v config then
      "__plus_v_constant__"
    else
      "__minus_v_constant__"
  in
  let v_term =
    Term.request_const (Symbol.create_symbol v_name 0)
  in
  let v_cp_element = {
    Context.el_id = Context.get_unique_element_id ();
    Context.el_pars = [v_par];
    Context.el_choice_point = State.root_choice_point state;
    Context.el_literal = Config.default_v config;
    Context.el_generation = 0;
    Context.el_mandatory_at = -1;
    Context.el_compacted = -1;
  }
  in
    {
      config = config;
      bound = bound;
      statistic = statistic;
      state = state;
      context = context;
      problem_literals = LiteralTable.create 16;
      index = Index.create 8;
(*      callbacks_on_add_unit = [];*)
      bindings = Bindings.create 1024;
(*      subsumed_spaces = SubsumedSpaces.create 16;*)
      subsumed_spaces = (*[| |]*) Array.create 32 (-1);
(*      backtrack_spaces = (*[| |]*) Array.create 32 false;*)
      v_par = v_par;
      v_term = v_term;
      v_cp_element = v_cp_element;
    }




(*** binding database ***)  

let register_binding (problem_literals: problem_literals) (binding: binding) : binding =
(*  try
    Bindings.find problem_literals.bindings binding
  with
    | Not_found ->
	Bindings.add problem_literals.bindings binding;*)
	binding


let register_subst (problem_literals: problem_literals) (subst: subst) : packed_subst =
  Array.map (register_binding problem_literals) (Subst.pack_subst subst)







(*** registration ***)






(* creates the initial queue of context partners for a literal.
   i.e. either no initial partner or -v for a positive literal. *)
let create_initial_context_partners (problem_literals: problem_literals)
  (literal: literal) (index: int) : context_partner list =

  if 
    (* in the horn case split never happens,
       so only assert or close context unifiers are of interest.
       thus -v which produces non-empty remainders is not needed. *)
    
    Config.is_horn problem_literals.config
    ||
    (* -v or +v may be used -
       and is only applicable if the literal has the opposite sign.  *)
    literal.Term.sign = Config.plus_v problem_literals.config;
  then
    []
  else begin
    let context_partner =
      (* just ensuring that this context partner is not p-preserving *)
      let subst =
	Subst.set_  ~p_preserving:false ~i_preserving:false
	  Subst.empty
	  problem_literals.v_par (Subst.context_literal_offset index)
	  problem_literals.v_term (Subst.context_literal_offset index);
      in
	{
	  Context_unifier.cp_element = problem_literals.v_cp_element;
	  Context_unifier.cp_partial_context_unifier = register_subst problem_literals subst;
	}
    in
      [context_partner]
  end



let register_to_index (problem_literals: problem_literals)
  (literal: literal) (problem_literal: problem_literal) : unit =

  let existing_problem_literals =
    try
      Index.find problem_literals.index literal
    with
      | Not_found ->
	  []
  in
    Index.replace problem_literals.index literal ((literal, problem_literal) :: existing_problem_literals)


let register_literal (problem_literals: problem_literals)
  (literal: literal) (index: int) : input_partner =

  let problem_literal =
    try
      LiteralTable.find problem_literals.problem_literals literal
    with
      | Not_found ->
	  let new_problem_literal = {
	    elements = Elements.empty;
(*	    callbacks_on_subsume = [];*)
	    subsumed_at = -1 (*None*);
	  }
	  in
	    LiteralTable.add problem_literals.problem_literals literal new_problem_literal;
	    register_to_index problem_literals literal new_problem_literal;
	    new_problem_literal;
  in

  let element =
    try
      Elements.find index problem_literal.elements
    with
      | Not_found ->
	  (* first occurance of this literal/index: create it *)
	  let vars =
	    List.map
	      (fun var ->
		 Subst.make_var var Subst.input_literal_offset
	      )
	      (Term.vars_of_literal literal)
	  in
	    
	  let context_partners =
	    create_initial_context_partners problem_literals literal index
	  in
	  let new_input_partner = {
	    Context_unifier.ip_index = index;
	    Context_unifier.ip_literal = literal;
	    Context_unifier.ip_vars = vars;
	    Context_unifier.ip_context_partners = context_partners;
	    Context_unifier.ip_context_partners_length = List.length context_partners;
(*	    Context_unifier.ip_context_partners_changed = false;*)
	    Context_unifier.ip_resolved = None;
	  }
	  in

	    (* compute all partial context unifiers with all context literals *)
	    Context.iter
	      problem_literals.context
	      (fun context_element ->
		 if Term.literal_equal context_element.Context.el_literal (Config.default_v problem_literals.config) then begin
		   print_endline (Term.literal_to_string context_element.Context.el_literal);
		   failwith "iter -v";
		 end;
		     
		 (* unify *)
		 begin
		   if context_element.Context.el_literal.Term.sign = new_input_partner.Context_unifier.ip_literal.Term.sign then
		     ()

		   else begin
		     try
		       let subst =
			 Unification.unify_terms
			   literal.Term.term
			   Subst.input_literal_offset
			   context_element.Context.el_literal.Term.term
			   (Subst.context_literal_offset index)
		       in
			 
		       let context_partner = {
			 Context_unifier.cp_element = context_element;
			 Context_unifier.cp_partial_context_unifier = register_subst problem_literals subst;
		       }
		       in
			 (* and add it *)
			 new_input_partner.Context_unifier.ip_context_partners <-
			   context_partner :: new_input_partner.Context_unifier.ip_context_partners;
			 new_input_partner.Context_unifier.ip_context_partners_length <-
			   new_input_partner.Context_unifier.ip_context_partners_length + 1;
		     
		     with
		       | Unification.UNIFICATION_FAIL ->
			   ()
		   end
		 end
	      );


	    (* sort these to get a queue suitable for chronological backtracking *)
	    let sorted =
	      List.sort
		(fun x y ->
		   compare
		   y.Context_unifier.cp_element.Context.el_id
		   x.Context_unifier.cp_element.Context.el_id
		)
		new_input_partner.Context_unifier.ip_context_partners
	    in
		new_input_partner.Context_unifier.ip_context_partners <- sorted;

	    let new_element = {
	      input_partner = new_input_partner;
(*	      callbacks_on_add = [];*)
	      spaces = [];
	    }
	    in
	      
	      problem_literal.elements <- Elements.add index new_element problem_literal.elements;
	      new_element
  in
    element.input_partner




let register_clause_literal (problem_literals: problem_literals) (literal: literal) (index: int)
(*  (on_add: on_add) (on_subsume: on_subsume)*) (space: context_unifier_space) : unit  =

  if Array.length problem_literals.subsumed_spaces <= space.Context_unifier.cus_id then begin
    if space.Context_unifier.cus_id >= Sys.max_array_length - 1 then begin
      failwith "register_clause_literal: OVERFLOW";
    end;

    let new_size =
(*      space.Context_unifier.cus_id + 1*)
(*      100000*)
      
      let desired_size = 
	Tools.max_int 32 (2 * (Array.length problem_literals.subsumed_spaces))
      in
	(* obey the hard limit of the max. array size *)
	Pervasives.min Sys.max_array_length desired_size
      
    in
(*      print_endline (string_of_int new_size);*)
    let new_array =
      Array.make new_size (-1)
    in
      Array.blit problem_literals.subsumed_spaces 0 new_array 0 (Array.length problem_literals.subsumed_spaces);
      problem_literals.subsumed_spaces <- new_array;
(*
    let new_array =
      Array.make new_size false
    in
      Array.blit problem_literals.backtrack_spaces 0 new_array 0 (Array.length problem_literals.backtrack_spaces);
      problem_literals.backtrack_spaces <- new_array;*)
  end;

  let problem_literal =
    try
      LiteralTable.find problem_literals.problem_literals literal
    with
      | Not_found ->
	  failwith "register_clause_literal: problem_literal"
  in

  let element =
    try
      Elements.find index problem_literal.elements
    with
      | Not_found ->
	  failwith "register_clause_literal: element"
  in

(*    problem_literal.callbacks_on_subsume <- on_subsume :: problem_literal.callbacks_on_subsume;

    element.callbacks_on_add <- on_add :: element.callbacks_on_add*)
    element.spaces <- space :: element.spaces
    
(*      
let register_unit_literal (problem_literals: problem_literals) (on_add_unit: on_add_unit) : unit =
  problem_literals.callbacks_on_add_unit <- on_add_unit :: problem_literals.callbacks_on_add_unit
*)






(*** context_unifier_space ***)


(* is this input literal already resolved
   with respect to the new context element? *)
(*let is_resolved (config: config)
  (input_partner: input_partner) (context_element: Context.element) : bool =

  match input_partner.Context_unifier.ip_resolved with
    | None -> false
    | _ -> true
*)
(*
let is_resolved
    (input_partner: input_partner) (context_element: Context.element) : bool =
  
  input_partner.Context_unifier.ip_resolved >= 0
  &&
  input_partner.Context_unifier.ip_resolved < context_element.Context.el_id
*)

let is_resolved
    (input_partner: input_partner) (context_element: Context.element) : bool =  
    match input_partner.Context_unifier.ip_resolved with
      | Some resolving_partner when
	  resolving_partner.Context_unifier.cp_element.Context.el_id
	  <
	  context_element.Context.el_id
	  ->
	  true

      | _ ->
	  false


(*** subsume ***)
(*
(* yes, this special treatment for chronological backtracking is significantly faster *)
let is_already_subsumed (config: config)
  (context_unifier_space: context_unifier_space) (context_element: Context.element) : bool =

  match context_unifier_space.Context_unifier.cus_subsumed with
    | None -> false
    | _ -> true


let mark_subsumed statistic context_unifier_space choice_point =
  match context_unifier_space.Context_unifier.cus_subsumed with
    | None ->
	Statistic.inc_subsume statistic;
	context_unifier_space.Context_unifier.cus_subsumed <- Some (choice_point)

    | _ ->
	()
*)



(*** reorder ***)

(* sort input literals by increasing number of context partners.
   keep resolved literals at the end.
   thus less partial context unifier extending steps need to be done,
   as the first considered literals have fewer context partners
   and thus branching is lower in the top part of the search tree.
*)

(* shift the resolved literal to the end *)
let move_resolved_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  print_endline "move_resolved_input_literal";
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let resolved =
    input_partners.(index)
  in
    for i = index + 1 to Array.length input_partners - 1 do
      input_partners.(i - 1) <- input_partners.(i);
    done;
    input_partners.(Array.length input_partners - 1) <- resolved


(*
(* shift the literal with the increased queue
   behind unresolved literals with a shorter queue *)
let move_increased_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in

  let rec swap_with (i : int) : unit =
    (* end of queue *)
    if i >= Array.length input_partners then
      ()

    (* only resolved literal ahead *)
    else if
      match input_partners.(i).Context_unifier.ip_resolved with
	| None ->
	    (* unresolved, so continue *)
	    false

	| _ ->
	    (* resolved, so stop here *)
	    true
    then
      ()

    (* the next literals queue is smaller, so swap *)
    else if
(*      List.length input_partners.(i).Context_unifier.ip_context_partners
      <
      List.length input_partners.(i - 1).Context_unifier.ip_context_partners*)
      input_partners.(i).Context_unifier.ip_context_partners_length
      <
      input_partners.(i - 1).Context_unifier.ip_context_partners_length
    then begin
      let swap =
	input_partners.(i - 1)
      in
	input_partners.(i - 1) <- input_partners.(i);
	input_partners.(i) <- swap;

	swap_with (i + 1)
    end
  in
    swap_with (index + 1)
    *)
(*
let check_cu (context_unifier_space: context_unifier_space) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
*)
(*
(* shift the resolved literal to the end *)
let move_resolved_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  print_endline "move_resolved_input_literal";
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let resolved =
    input_partners.(index)
  in
  let rec do' i =
    if i >= Array.length input_partners then
      Array.length input_partners - 1
      (*index*)

    else
      match input_partners.(i).Context_unifier.ip_resolved with
	| Some _ ->
	    i - 1
	    
	| _ ->
	    do' (i + 1)
  in
  let swap_with =
    do' (index + 1)
  in
    if swap_with != index then begin
  let swap =
    input_partners.(swap_with)
  in
    input_partners.(swap_with) <- resolved;
    input_partners.(index) <- swap
    end
      (*
    for i = index + 1 to Array.length input_partners - 1 do
      input_partners.(i - 1) <- input_partners.(i);
    done;
    input_partners.(Array.length input_partners - 1) <- resolved
*)
*)


let move_increased_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
(*  print_endline "move_increased_input_literal";*)
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let ordering =
    context_unifier_space.Context_unifier.cus_input_partners_ordering
  in

  let rec do' i =
    if i >= Array.length input_partners then
      (*index*)
      i - 1

    (* only resolved literal ahead *)
    else if
      match input_partners.(ordering.(i)).Context_unifier.ip_resolved with
	| None ->
	    (* unresolved, so continue *)
	    false

	| _ ->
	    (* resolved, so stop here *)
	    true
    then
      (*index*)
(*      i - 1*)
      do' (i + 1)

(*
    (* sequence of identical clause literals
       whose pcu queue has been increase simultaneously. *)
    else if
      Term.literal_equal
      input_partners.(index).Context_unifier.ip_literal
      input_partners.(i).Context_unifier.ip_literal
    then
      do' (i + 1)
*)

(*
    else if
(*      List.length input_partners.(i).Context_unifier.ip_context_partners
      <
      List.length input_partners.(i - 1).Context_unifier.ip_context_partners*)
      input_partners.(index).Context_unifier.ip_context_partners_length
      =
      input_partners.(i).Context_unifier.ip_context_partners_length
    then begin
(*      if
	input_partners.(index).Context_unifier.ip_context_partners_length
	>
	  (input_partners.(i).Context_unifier.ip_context_partners_length + 5)
      then*)
	i - 1
(*      else
	index*)
    end
*)
    (* the next literals queue is smaller, so swap *)
    else if
(*      List.length input_partners.(i).Context_unifier.ip_context_partners
      <
      List.length input_partners.(i - 1).Context_unifier.ip_context_partners*)
      input_partners.(ordering.(index)).Context_unifier.ip_context_partners_length
      <=
      input_partners.(ordering.(i)).Context_unifier.ip_context_partners_length
    then begin
(*      if
	input_partners.(index).Context_unifier.ip_context_partners_length
	>
	  (input_partners.(i).Context_unifier.ip_context_partners_length + 5)
      then*)
	i - 1
(*      else
	index*)
    end

    else (* new > compared *)
      do' (i + 1)
  in
  let swap_with =
    do' (index + 1)
  in
    if swap_with != index then begin
      let swap =
	ordering.(swap_with)
      in
	ordering.(swap_with) <- ordering.(index);
	ordering.(index) <- swap
    end



let move_updated_input_literal (config: config)
  (context_unifier_space: context_unifier_space) (index: int)
  (input_partner: input_partner) (context_partner: context_partner)
  : unit =

(*  print_endline "\nmove_updated_input_literal";

  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  for i = 0 to Array.length input_partners - 1 do
    match input_partners.(i).Context_unifier.ip_resolved with
	| None ->
	    print_endline (string_of_int input_partners.(i).Context_unifier.ip_context_partners_length ^ Term.literal_to_string input_partners.(i).Context_unifier.ip_literal)

	| Some _ ->
	    print_endline ("resolved");
  done;
*)
  begin
  if is_resolved input_partner context_partner.Context_unifier.cp_element then
(*    move_resolved_input_literal context_unifier_space index*)
    ()
  else
    move_increased_input_literal context_unifier_space index
(*    ()*)
  end(*;
  
  (* :DEBUG: *)
  print_endline "move_updated_input_literal done";
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  for i = 0 to Array.length input_partners - 1 do
    match input_partners.(i).Context_unifier.ip_resolved with
	| None ->
	    print_endline (string_of_int input_partners.(i).Context_unifier.ip_context_partners_length ^ Term.literal_to_string input_partners.(i).Context_unifier.ip_literal)

	| Some _ ->
	    print_endline ("resolved");
  done;
  for i = 0 to Array.length input_partners - 2 do
(*    let succeed =
      match input_partners.(i).Context_unifier.ip_resolved, input_partners.(i + 1).Context_unifier.ip_resolved with
	| None, None ->
	    input_partners.(i).Context_unifier.ip_context_partners_length
	    <=
	    input_partners.(i + 1).Context_unifier.ip_context_partners_length

	| None, Some _ ->
	    true

	| Some _, Some _ ->
	    true

	| Some _, None ->
	    false
    in*)
    let succeed =
      match input_partners.(i).Context_unifier.ip_resolved, input_partners.(i + 1).Context_unifier.ip_resolved with
	| None, None ->
	    input_partners.(i).Context_unifier.ip_context_partners_length
	    <=
	    input_partners.(i + 1).Context_unifier.ip_context_partners_length

	| None, Some _ ->
	    true

	| Some _, Some _ ->
	    true

	| Some _, None ->
	    true
    in
      if not succeed then begin
	print_endline (Term.clause_to_string context_unifier_space.Context_unifier.cus_clause);
	failwith "move_updated_input_literal"
      end
  done
     *)







(*** add a literal ***)


(* pair the new context literal with the clause literals one by one
   and compute all new context unifiers. *)
let add_to_clause
  (config: config)
  (bound: bound)
  (context_unifier_space: context_unifier_space)
  (input_partner: input_partner)
  (context_partner: context_partner)
  (candidate_type: Context_unifier.candidate_type)
  (close_found: bool)
  : bool =
      if
	candidate_type == Context_unifier.Split
	&&
	context_unifier_space.Context_unifier.cus_lemma
      then
(*  if context_unifier_space.Context_unifier.cus_lemma then*)
    close_found
      else
(*
  if is_already_subsumed config context_unifier_space context_partner.Context_unifier.cp_element then begin
    false
  end

  else*) begin
    (* create the inital partial context unifier *)
    let partial_context_unifier =
      Array.make
	(Array.length context_unifier_space.Context_unifier.cus_input_partners)
	Context_unifier.null_partner
    in
      
    let ordering =
      context_unifier_space.Context_unifier.cus_input_partners_ordering
    in
      partial_context_unifier.(input_partner.Context_unifier.ip_index) <-
      context_partner;
      
      let rec find_i i =
	if i > Array.length context_unifier_space.Context_unifier.cus_input_partners then
	  failwith "find_i"
	else if input_partner == context_unifier_space.Context_unifier.cus_input_partners.(ordering.(i)) then
	  i
	else
	  find_i (i + 1)
      in
      let i =
	find_i 0
      in
      let subst =
	Subst.unpack_subst context_partner.Context_unifier.cp_partial_context_unifier
      in
      let search =
	match candidate_type with
	  | Context_unifier.Close ->
	      Context_unifier_search.SearchClose.search_context_unifiers

	  | Context_unifier.Assert ->
	      Context_unifier_search.SearchAssert.search_context_unifiers

	  | Context_unifier.Split ->
	      Context_unifier_search.SearchSplit.search_context_unifiers

	  | Context_unifier.CloseAssert ->
	      Context_unifier_search.SearchCloseAssert.search_context_unifiers

	  | Context_unifier.All ->
	      if context_unifier_space.Context_unifier.cus_lemma then
		Context_unifier_search.SearchCloseAssert.search_context_unifiers
	      else
		Context_unifier_search.SearchAll.search_context_unifiers
      in
      let non_empty_remainder =
	not (Config.is_horn config)
	&&
	Context_unifier.is_remainder 
	  subst
	  context_partner.Context_unifier.cp_element.Context.el_pars
	  input_partner.Context_unifier.ip_index
      in
	(* compute new candidates *)
      let close_found2: bool =
	search
	  config
	  bound
	  context_unifier_space
	  partial_context_unifier
	  i
	  context_partner
	  subst
	  non_empty_remainder
	close_found;
      in

	begin
	  match candidate_type with
	    | Context_unifier.Close
	    | Context_unifier.CloseAssert
	    | Context_unifier.All ->
		move_updated_input_literal config context_unifier_space i input_partner context_partner;

	    | Context_unifier.Assert
	    | Context_unifier.Split ->
		()

	end;

	close_found || close_found2
  end







(*** resolve ***)

(*
(* is the problem literal already resolved with respect to
   the dependencies on the context literal? *)
let is_already_resolved (problem_literals: problem_literals)
  (input_partner: input_partner) (context_element: Context.element) : bool =

  match input_partner.Context_unifier.ip_resolved with
    | None ->
	false
	  
    | _ ->
	true
*)



(* mark the problem literal as resolved by this context literal *)
let mark_resolved (problem_literals: problem_literals)
  (element: input_partner) (context_partner: context_partner): unit =

  Statistic.inc_resolve problem_literals.statistic;

  element.Context_unifier.ip_resolved <-
    Some (context_partner);

  (* no backtracking in horn - so this stays resolved forever
     and some entries can be removed for garbage collection *)
  if Config.is_horn problem_literals.config then begin
    element.Context_unifier.ip_context_partners <- context_partner :: [];
    element.Context_unifier.ip_context_partners_length <- 1;
  end


(* checks the resolve inference rule against
   the problem literal and the context literal *)
let resolve (problem_literals: problem_literals)
  (input_partner: input_partner) (context_partner: context_partner) : bool =

  if not (Config.resolve problem_literals.config) then
    false

  else if
    context_partner.Context_unifier.cp_element.Context.el_literal.Term.sign
    =
    input_partner.Context_unifier.ip_literal.Term.sign
  then
    false

  else begin
    let context_term =
      context_partner.Context_unifier.cp_element.Context.el_literal.Term.term
    and problem_term =
      input_partner.Context_unifier.ip_literal.Term.term
    in
      (* variants? *)
      if Term.are_terms_variants context_term problem_term then begin
	mark_resolved problem_literals input_partner context_partner;
	(* this is only a variable renaming, so the substitution can be removed.
	   This can vastly improve the speed of merging substitutions. *)
	context_partner.Context_unifier.cp_partial_context_unifier <- [| |];
	true
      end
	
      (* subsumed? this implies Resolve,
	 as the effect on the rest clause can only be a variable remaining *)
      else begin
	try
	  let subst =
	    Unification.match_terms
	      ~p_preserving:true
	      context_term
	      (Subst.context_literal_offset input_partner.Context_unifier.ip_index)
	      problem_term
	      Subst.input_literal_offset
	  in
	    mark_resolved problem_literals input_partner context_partner;
	    (* use this substitution instead of the originally computed -
	       it is subsuming,
	       thus only variables of the context literal are bound,
	       thus merging of substitutions might be more efficient *)
	    context_partner.Context_unifier.cp_partial_context_unifier <- register_subst problem_literals subst;
	    true
	    
	with
	  | Unification.UNIFICATION_FAIL ->
	      false
      end
  end



(*** subsume ***)
(*
let is_space_subsumed problem_literals
  (context_unifier_space: context_unifier_space) : bool =
  problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <> -1
*)
let is_space_currently_subsumed problem_literals
  (context_unifier_space: context_unifier_space) : bool =

  problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <> -1

let is_space_subsumed problem_literals
  (context_unifier_space: context_unifier_space) (context_element: Context.element) : bool =

  problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <> -1
  &&
    (
      problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id)
      <
      context_element.Context.el_id
  )
  
(*
(* yes, this special treatment for chronological backtracking is significantly faster *)
let is_already_subsumed2 problem_literals
  (context_unifier_space: context_unifier_space) (context_element: Context.element) : bool =

  problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <> -1
(*
  begin
    try
      let choice_point: choice_point =
	SubsumedSpaces.find problem_literals.subsumed_spaces context_unifier_space
      in
	true
    with
      | Not_found ->
	  false
  end*)
*)
(*
let is_already_subsumed3 problem_literals
  (context_unifier_space: context_unifier_space) : bool =
  problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <> -1
(*
  begin
    try
      let choice_point: choice_point =
	SubsumedSpaces.find problem_literals.subsumed_spaces context_unifier_space
      in
	true
    with
      | Not_found ->
	  false
  end*)
*)
(*
  match context_unifier_space.Context_unifier.cus_subsumed with
    | None -> false
    | _ -> true*)

let mark_space_subsumed (*statistic*) problem_literals context_unifier_space context_element(* choice_point*) =
  if not (is_space_subsumed problem_literals context_unifier_space context_element) then begin
    Statistic.inc_subsume problem_literals.statistic;
(*    problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <- State.id_of_choice_point choice_point*)
    problem_literals.subsumed_spaces.(context_unifier_space.Context_unifier.cus_id) <- context_element.Context.el_id
  end
  (*
  begin
    try
      let choice_point: choice_point =
	SubsumedSpaces.find problem_literals.subsumed_spaces context_unifier_space
      in
	(*true*) ()
    with
      | Not_found ->
	  Statistic.inc_subsume problem_literals.statistic;
	  SubsumedSpaces.add problem_literals.subsumed_spaces context_unifier_space choice_point;
  end
  *)
(*
  match context_unifier_space.Context_unifier.cus_subsumed with
    | None ->
	Statistic.inc_subsume problem_literals.statistic;
	SubsumedSpaces.add problem_literals.subsumed_spaces context_unifier_space choice_point;
	context_unifier_space.Context_unifier.cus_subsumed <- Some (choice_point)

    | _ ->
	()
*)

let is_problem_literal_subsumed (problem_literals: problem_literals)
  (problem_literal: problem_literal) (context_element: Context.element) : bool =

  if problem_literal.subsumed_at = -1 then
(*  match problem_literal.subsumed_at with
    | -1(*None*) ->*)
	begin
	try
	Elements.iter
	  (fun _ element ->
	     List.iter
	       (fun space ->
		  if not (is_space_subsumed problem_literals space context_element) then
		    raise Exit
	       )
	       element.spaces;
	  )
	  problem_literal.elements;
	true
	with
	  | Exit ->
	      false
	end
(*
    | _ ->
	true*)
  else begin
    problem_literal.subsumed_at < context_element.Context.el_id
  end


(* try subsume with the context literal on the problem literal *)
let does_subsume_problem_literal (literal: literal) (context_element: Context.element) : bool =
  try
    ignore (Unification.match_literals
	      ~p_preserving:true
	      context_element.Context.el_literal
	      0
	      literal
	      1);
    true
  with
    | Unification.UNIFICATION_FAIL -> 
	false



(* try subsume with the context literal on all problem literals *)
let subsume (problem_literals: problem_literals) (problem_literal: problem_literal)
  (literal: literal) (context_element: Context.element) : unit =

  if is_problem_literal_subsumed problem_literals problem_literal context_element then
    ()
      
  else if not (does_subsume_problem_literal literal context_element) then
    ()
      
  else begin
    (* process callbacks *)
    Elements.iter
      (fun _ element ->
	 List.iter
	   (fun space ->
	      mark_space_subsumed problem_literals space (*(State.active_choice_point problem_literals.state)*) context_element
	   )
	   element.spaces;
      )
      problem_literal.elements;
    
    (* mark as subsumed *)
(*    problem_literal.subsumed_at <- (*Some*) State.id_of_choice_point (State.active_choice_point problem_literals.state);*)
    problem_literal.subsumed_at <- context_element.Context.el_id;

    (* no backtracking in horn - so this stays subsumed forever
       and all context partner may be forgotten *)
    if Config.is_horn problem_literals.config then begin
      Elements.iter
	(fun _ element ->
	   element.input_partner.Context_unifier.ip_context_partners <- [];
	   element.input_partner.Context_unifier.ip_context_partners_length <- 0;
	)
	problem_literal.elements
    end;
  end


(*** add ***)


(*
let process_callbacks_on_add_unit (problem_literals: problem_literals) (context_element: Context.element) : unit =
  List.iter
    (fun callback ->
       callback context_element
    )
    problem_literals.callbacks_on_add_unit
*)

(*
let backtrack_context_unifier (config: config) (context_unifier_space: context_unifier_space) : unit =
  (*
  Array.sort
    (fun queue1 queue2 ->
       match queue1.Context_unifier.ip_resolved, queue2.Context_unifier.ip_resolved with
	 | None, None ->
	     compare
	       (List.length queue1.Context_unifier.ip_context_partners)
	       (List.length queue2.Context_unifier.ip_context_partners)
	       
	 | Some _, _ ->
	     -1
	       
	 | _, Some _ ->
	     1
    )
    context_unifier_space.Context_unifier.cus_input_partners
  *)
  (* try insertion sort *)
(*  match context_unifier_space.Context_unifier.cus_subsumed with
    | Some _ ->
	()

    | None ->*)
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
    for i = 1 to Array.length input_partners - 1 do
      if input_partners.(i).Context_unifier.ip_context_partners_changed then begin
      match input_partners.(i).Context_unifier.ip_resolved with
	| Some x ->
	    (*() check not needed if changed is checked first!*)
(*	    print_endline (Term.literal_to_string x.Context_unifier.cp_element.Context.el_literal);*)
(*	    failwith "AHA";*)
	    ()
	      
	| None ->
	    (* shift up *)
	    let rec find_swap j =
	      if j = -1 then
		i
		  
	      else if
		match input_partners.(j).Context_unifier.ip_resolved with
		  | Some _ ->
		      true
			
		  | None ->
		      false
	      then
		find_swap (j - 1)		
		 
	      else if
(*		List.length input_partners.(i).Context_unifier.ip_context_partners
		>=
		List.length input_partners.(j).Context_unifier.ip_context_partners*)
		input_partners.(i).Context_unifier.ip_context_partners_length
		>=
		input_partners.(j).Context_unifier.ip_context_partners_length
	      then
		(j + 1)
		  
	      else begin
		find_swap (j - 1)
	      end
	    in
	      
	    let swap_with =
	      find_swap (i - 1)
	    in
	      if
		(swap_with != i)
		&&
		match input_partners.(swap_with).Context_unifier.ip_resolved with
		  | Some _ ->
		      false
			
		  | None ->
		      true
	      then begin
		let swap = input_partners.(swap_with) in
		  input_partners.(swap_with) <- input_partners.(i);
		  input_partners.(i) <- swap;
	      end
      end
    done
*)

let add_to_element
  (problem_literals: problem_literals)
  (problem_literal: problem_literal)
  (context_literal: literal)
  (context_element: Context.element)
  (candidate_type: Context_unifier.candidate_type)
  (close_found: bool)
  : bool =
  if is_problem_literal_subsumed problem_literals problem_literal context_element(* candidate_type*) then begin
    (*		   print_endline ("RESOLVED");*)
    close_found
  end
  else
  let close_found = ref close_found in
  (* if the first literal * offset fails,
     the literal's other offset combination need not to be checked.
     otherwise the information can be reused *)
  try
    let _ =
      Elements.fold
	(fun index element cache ->
	   
	   match cache with
	       (* first time that this literal is processed with (any index) *)
	     | None ->
		 (* already resolved? *)
		 if is_resolved element.input_partner context_element then begin
		   raise Exit
		 end
		   
		 (* unify *)
		 else begin
		   if context_literal.Term.sign = element.input_partner.Context_unifier.ip_literal.Term.sign then
		     raise Exit
		   else
		   try
		     let subst =
		       Unification.unify_terms
			 element.input_partner.Context_unifier.ip_literal.Term.term
			 Subst.input_literal_offset
			 context_literal.Term.term
			 (Subst.context_literal_offset index)
		     in

			 (* process all callbacks *)
(*			 List.iter
			   (fun space ->
			      if is_already_subsumed2 problem_literals space context_element then begin
				()
			      end
			      else begin
				if space.Context_unifier.cus_backtrack then begin
				  space.Context_unifier.cus_backtrack <- false;
				  backtrack_context_unifier problem_literals.config space;
				end;
			      end
			   )
			   (*element.callbacks_on_add;*)
			   element.spaces;*)

		     (* create the initial context partner *)
		     let context_partner = {
		       Context_unifier.cp_element = context_element;
		       Context_unifier.cp_partial_context_unifier = register_subst problem_literals subst;
		     }
		     in
		       
		       (* and add it *)
		       element.input_partner.Context_unifier.ip_context_partners <-
		       context_partner :: element.input_partner.Context_unifier.ip_context_partners;
		       element.input_partner.Context_unifier.ip_context_partners_length <-
			 element.input_partner.Context_unifier.ip_context_partners_length + 1;
		       
		       (* resolving? *)
		       let resolved =
			 resolve problem_literals element.input_partner context_partner
		       in
			 
			 (* process all callbacks *)
			 List.iter
			   (fun space ->
			      if is_space_subsumed problem_literals space context_element then begin
				()
			      end
			      else begin
(*				if space.Context_unifier.cus_backtrack then begin
				  space.Context_unifier.cus_backtrack <- false;
				  backtrack_context_unifier problem_literals.config space;
				end;*)
			      let ncf: bool =
				add_to_clause problem_literals.config problem_literals.bound space
				(*callback*) element.input_partner context_partner candidate_type !close_found
			      in
				close_found := !close_found || ncf
			      end
			   )
			   (*element.callbacks_on_add;*)
			   element.spaces;
			 
			 Some (element, context_partner, resolved)
		   with
		     | Unification.UNIFICATION_FAIL ->
			 raise Exit
		 end

	     (* literal has already been processed with another index *)
	     | Some (cached_element, cached_context_partner, cached_resolved) ->

(*			 (* process all callbacks *)
			 List.iter
			   (fun space ->
			      if is_already_subsumed2 problem_literals space context_element then begin
				()
			      end
			      else begin
				if space.Context_unifier.cus_backtrack then begin
				  space.Context_unifier.cus_backtrack <- false;
				  backtrack_context_unifier problem_literals.config space;
				end;
			      end
			   )
			   (*element.callbacks_on_add;*)
			   cached_element.spaces;*)

		 (* create the initial context partner *)
		 let new_subst =
		   Subst.replace_offset
		     (Subst.unpack_subst cached_context_partner.Context_unifier.cp_partial_context_unifier)
		     (Subst.context_literal_offset cached_element.input_partner.Context_unifier.ip_index)
		     (Subst.context_literal_offset index)
		 in
		 let new_context_partner = {
		   Context_unifier.cp_element = context_element;
		   Context_unifier.cp_partial_context_unifier = register_subst problem_literals new_subst;
		 }
		 in

		   (* and add it *)
		   element.input_partner.Context_unifier.ip_context_partners <-
		   new_context_partner :: element.input_partner.Context_unifier.ip_context_partners;
		   element.input_partner.Context_unifier.ip_context_partners_length <-
		     element.input_partner.Context_unifier.ip_context_partners_length + 1;

		   if cached_resolved then begin
		     mark_resolved problem_literals element.input_partner new_context_partner;
		   end;

		   (* process all callbacks *)
(*		   List.iter
		     (fun callback ->
			let ncf: bool =
			  callback element.input_partner new_context_partner candidate_type !close_found
			in
			  close_found := !close_found || ncf
		     )
		     element.callbacks_on_add;
*)
			 List.iter
			   (fun space ->
			      if is_space_subsumed problem_literals space context_element then begin
				()
			      end
			      else begin
			      let ncf: bool =
				add_to_clause problem_literals.config problem_literals.bound space
				(*callback*) element.input_partner new_context_partner candidate_type !close_found
			      in
				close_found := !close_found || ncf
			      end
			   )
			   (*element.callbacks_on_add;*)
			   element.spaces;

		   cache
	)
	problem_literal.elements
	None
    in
      !close_found
  with
    | Exit ->
	!close_found





let add_new (problem_literals: problem_literals) (context_element: Context.element) (candidate_type: Context_unifier.candidate_type)
  : unit =
(*  process_callbacks_on_add_unit problem_literals context_element;*)

  (* subsume *)
  if 
    (Config.subsume problem_literals.config)
    &&
    (* only universal literals can subsume input literals as those are always universal *)
    (match context_element.Context.el_pars with
       | [] -> true
       | _ -> false
    )
  then begin
    let subsume_candidates =
      try
	Index.find problem_literals.index context_element.Context.el_literal
      with
	| Not_found ->
	    (* a unit literal and thus not registered *)
	    []
    in
      List.iter
	(fun (literal, problem_literal) ->
	   subsume problem_literals problem_literal literal context_element
	)
	subsume_candidates;
  end;

  (* unify *)
  let unifying_candidates =
    try
      let neg_new_context_literal =
	Term.request_negated_literal context_element.Context.el_literal
      in
	Index.find problem_literals.index neg_new_context_literal
    with
      | Not_found ->
	  (* a unit literal and thus not registered *)
	  []
  in
  let close_found = ref false in
    List.iter
      (fun (literal, problem_literal) ->
	 let cfn =
	 add_to_element
	   problem_literals
	   problem_literal
	   context_element.Context.el_literal
	   context_element
	   candidate_type
	   !close_found
	 in
	   close_found := !close_found || cfn
      )
      unifying_candidates




let add_deferred2 (problem_literals: problem_literals) (context_element: int) (candidate_type: Context_unifier.candidate_type)
  : unit =
(*  print_endline ("BEGIN add_deferred2");*)
  LiteralTable.iter
    (fun _ problem_literal ->
       Elements.iter
       (fun _ element ->
	  List.iter
	  (fun context_partner ->
	     if context_partner.Context_unifier.cp_element.Context.el_id >= context_element then begin
(*	       print_endline ("BEGIN add_deferred2");*)
(*	     if context_partner.Context_unifier.cp_element.Context.el_id >= context_element.Context.el_id then begin*)
	       (*
	       List.iter
		 (fun callback ->
		    let x: bool =
		      callback element.input_partner context_partner candidate_type false
		    in
		      ()
		 )
		 element.callbacks_on_add*)
			 List.iter
			   (fun space ->
			      if is_space_subsumed problem_literals space context_partner.Context_unifier.cp_element then begin
				()
			      end
			      else begin
(*			      print_endline ("BEGIN add_deferred2: " ^ Term.clause_to_string space.Context_unifier.cus_clause);*)
			      let ncf: bool =
				add_to_clause problem_literals.config problem_literals.bound space
				(*callback*) element.input_partner context_partner candidate_type false
			      in
				()
			      end
			   )
			   (*element.callbacks_on_add;*)
			   element.spaces;
	     end	       
	  )
	  element.input_partner.Context_unifier.ip_context_partners
       )
       problem_literal.elements
    )
    problem_literals.problem_literals
(*  print_endline ("END add_deferred2")*)


let add (problem_literals: problem_literals) (context_element: Context.element) (candidate_type: Context_unifier.candidate_type) : unit =
  match candidate_type with
    | Context_unifier.Close
    | Context_unifier.CloseAssert
    | Context_unifier.All ->
	add_new problem_literals context_element candidate_type

    | Context_unifier.Assert
    | Context_unifier.Split ->
	failwith "add deferred"

let add_deferred (problem_literals: problem_literals) (context_element: int) (candidate_type: Context_unifier.candidate_type) : unit =
  match candidate_type with
    | Context_unifier.Close
    | Context_unifier.CloseAssert
    | Context_unifier.All ->
	failwith "add_deferred"

    | Context_unifier.Assert
    | Context_unifier.Split ->
	add_deferred2 problem_literals context_element candidate_type


(*** backtrack ***)


let backtrack (problem_literals: problem_literals) context_id : unit =
(*       let bt_id = State.id_of_choice_point cp in*)
	 for i = 0 to Array.length problem_literals.subsumed_spaces - 1 do
    (*	   let a = problem_literals.subsumed_spaces.(i) in
	   if 
	     (a >= 0)
	     &&
	       (a >= bt_id)
    *)
	   if 
	     (problem_literals.subsumed_spaces.(i) >= 0)
	   then begin
	     
	     if (problem_literals.subsumed_spaces.(i) > context_id)  then begin
	       problem_literals.subsumed_spaces.(i) <- -1;
(*	       problem_literals.backtrack_spaces.(i) <- true;*)
	     end;

	   end
(*	   else begin
	     problem_literals.backtrack_spaces.(i) <- true;
	   end;*)
	 done;

(*
       Array.iteri
	 (fun i subsumed ->
	    if subsumed >= bt_id then
	    if subsumed >= bt_id then
	      problem_literals.subsumed_spaces.(i) <- -1;
	 )
	 problem_literals.subsumed_spaces;*)
(*
       let unsubsumed =
	 SubsumedSpaces.fold
	   (fun space choice_point acc ->
	      if State.is_choice_point_invalid choice_point then begin
		space :: acc
	      end
	      else
		acc
	   )
	   problem_literals.subsumed_spaces
	   []
       in
	 List.iter
	   (SubsumedSpaces.remove problem_literals.subsumed_spaces)
	   unsubsumed;*)

  LiteralTable.iter
    (fun _ problem_literal ->
       if
	 problem_literal.subsumed_at >= 0
	 &&
	 problem_literal.subsumed_at > context_id
       then begin
	 problem_literal.subsumed_at <- -1;
       end;
(*
       (* subsumed *)
       begin
	 match problem_literal.subsumed_at with
	   | Some choice_point when
	       State.is_choice_point_invalid choice_point ->
	       problem_literal.subsumed_at <- None

	   | _ ->
	       ()
       end;
*)

       Elements.iter
         (fun _ element ->
	    (* context_partners *)
	    (*
	    let pruned =
	      State.prune
		problem_literals.state
		(fun context_partner -> context_partner.Context_unifier.cp_element.Context.el_choice_point)
		(fun _ -> ())
		oldest_retracted
		element.input_partner.Context_unifier.ip_context_partners
	    in
	      element.input_partner.Context_unifier.ip_context_partners <- pruned;
	      *)
	    (*
	    if Config.chronological_backtracking problem_literals.config then
	    *)
	      let changed = ref false in
	    begin
	      let rec remove_invalid_context_partner context_partners =
		match context_partners with
		  | context_partner :: tail when
		      State.is_choice_point_invalid context_partner.Context_unifier.cp_element.Context.el_choice_point ->
		      element.input_partner.Context_unifier.ip_context_partners_length <-
			element.input_partner.Context_unifier.ip_context_partners_length - 1;
			changed := true;
			remove_invalid_context_partner tail

		  | _ ->
		      element.input_partner.Context_unifier.ip_context_partners <- context_partners;
(*		      element.input_partner.Context_unifier.ip_context_partners_changed <- !changed*)
	      in
		remove_invalid_context_partner element.input_partner.Context_unifier.ip_context_partners
	    end;
	      (*
		else
	      *)      
(*
	      begin
	      element.input_partner.Context_unifier.ip_context_partners <-
	      List.find_all
		(fun context_partner ->
		   State.is_choice_point_valid context_partner.Context_unifier.cp_element.Context.el_choice_point
		)
		element.input_partner.Context_unifier.ip_context_partners
	    end;
*)		
	    (* resolved *)
	    match element.input_partner.Context_unifier.ip_resolved with
	      | None ->
		  ()

	      | Some context_partner ->
		   if State.is_choice_point_invalid context_partner.Context_unifier.cp_element.Context.el_choice_point then begin
(*		     if not (!changed) then failwith "B";*)

		     element.input_partner.Context_unifier.ip_resolved <- None;
		   end

       )
       problem_literal.elements;

    )
    problem_literals.problem_literals



(* compute assert candidates with all context literals
   added to the context in or after the retract choice point *)
let add_lemma (problem_literals: problem_literals) space : unit =
  (*
    replay for all context_literals added after retract?
  *)
  ()
(*
  InputLiterals.iter
    (fun _ problem_literal ->
       Elements.iter
       (fun _ element ->
	  List.iter
	  (fun context_partner ->
	     if State.compare_age retract context_partner.Context_unifier.cp_element.Context.el_choice_point > 0 then begin
	       (* only recompute context unifiers for literals added at or after
		  the previous point responsible for the conflict. *)
	       ()
	     end

	     else begin
	       List.iter
		 (fun callback ->
		    if space == callback then begin
(*			 print_endline ("USED CONTEXT PARTNER");
			 print_endline (Term.literal_to_string context_partner.Context_unifier.cp_element.Context.el_literal);
			 print_endline (Subst.subst_to_string (Array.to_list context_partner.Context_unifier.cp_partial_context_unifier));*)
(*		      callback#add element.el_input_partner context_partner Context_unifier.Lemma*)
		      ()
		    end
		 )
		 element.el_spaces
	     end
	  )
	  element.el_input_partner.Context_unifier.ip_context_partners
       )
       problem_literal.elements
    )
    problem_literals.problem_literals
*)
