(* $Id: problem_literals.mli,v 1.9 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** checking of problem literals against new context literals

    Contains all literals of the problem set (input clause set).
  
    As the same literal may occur in different clauses
    at the same as well as different positions,
    it is stored paired with the index of its position.
    Identical combinations of literals and positions
    are stored only once.

    Checks for each new context literal its unifiability
    with all problem literals, stores the information
    (substitution, resolved, subsumed, ...),
    and calls the call back functions
    registered by the clauses containing an involved problem literal.

    Applies the {e Resolve} and {e Subsume} rule
    of the model evolution calculus.
    As this module contains no information about clauses,
    {e Resolve} is only applied in a restricted - clause independent - way,
    i.e. as subsumption of the problem literal.
*)


(** {6 Types} *)



type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type subst = Subst.subst
type choice_point = State.choice_point
type state = State.state
type context = Context.context
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner
type problem_literals
type context_unifier_space = Context_unifier.context_unifier_space


(** {2 Callback} *)

(*


(** should be used by unit clauses for registration
  in order to be notified whenever a context literal is added.

  The intention is to raise {!Context_unifier.CLOSE} on a contradiction. *)
type on_add_unit = Context.element -> unit


(** should be used by non-unit clauses for registration
    in order to be notified whenever a unifying context literal is added.
    
    On notification all (as specified by [candidate_type])
    new context unifiers of the clause have to be computed.
*)
type on_add = input_partner -> context_partner -> candidate_type -> bool -> bool


(** should be used by non-unit clauses to register literals
    and be notified whenever a subsuming context literal is added. *)
type on_subsume = State.choice_point -> unit
*)

(** {6 Functions} *)


(** {2 Creation} *)


val create: config -> bound -> statistic -> state ->  context -> problem_literals


(** {2 Registration} *)

(** Clauses have to register their literals
    in order to get notified when a relevant literal is added to the context,
    i.e. a context literal that unifies with one of the clause literals.
*)

(** [register_literal literal index]
    retrieves the input_partner
    for the problem [literal]
    used as the [index].th literal in a clause,
    i.e. the problem literal plus some attributes. *)
val register_literal: problem_literals -> literal -> int -> input_partner
  
(** [register_clause_literal literal index on_add on_subsume]
  registers {!Problem_literals.on_add}
  and {!Problem_literals.on_subsume}
  for the problem [literal]
  used as the [index].th literal in a clause.

  [literal] must have been previously registered
  with {!Problem_literals.register_literal} *)
val register_clause_literal: problem_literals -> literal -> int -> (*on_add -> on_subsume ->*) context_unifier_space -> unit

(** [register_unit_literal on_add_unit]
  registers {!Problem_literals.on_add_unit} *)
(*val register_unit_literal: problem_literals -> on_add_unit -> unit*)


(** {2 Adding} *)

(** checks the new context literal against all stored problem literals,
  stores the results,
  and calls the appropriate registered call back functions.

  Based on [candidate_type] the computation of split candidates might be deferred
  by first computing only assert and then only if necessary split candidates.
  The assert candidates must be computed first, though.
 *)
val add: problem_literals -> Context.element -> Context_unifier.candidate_type -> unit

val add_deferred: problem_literals -> int -> Context_unifier.candidate_type -> unit

(** {2 Backtracking} *)

(** removes information based on invalid context literals.
    [choice_point] is the oldest choice point retracted
    since the last backtracking.
*)
val backtrack: problem_literals -> int -> unit


val add_lemma: problem_literals -> context_unifier_space -> unit

val is_space_currently_subsumed: problem_literals -> context_unifier_space -> bool
