type config = Config.config
type literal = Term.literal
type state = State.state
type choice_point = State.choice_point
type bound = Bound.bound

type replay =
  | Left of literal
      (* replay only the left split -
	 this split was 'above' the destination of a random jump,
	 i.e. it was not jumped over. *)

  | Right of literal
      (* replay only the right split -
	 the left split has been explored before the random jump,
	 i.e. this was already a right split when the random jump happened. *)

  | Split of literal
      (* replay a full split decision -
	 this split has been jumped over when doing the random jump. *)

(* guiding path *)
type saved_branch =
    replay list

type random_jump = {
  config: config;
  state: state;
  bound: bound;
  mutable saved: saved_branch list;
  counter: Counter.counter;
(*  mutable current: saved_branch;*)
}

let check_at = 50
(*
let split_left random_jump literal =
  random_jump.current <- Left literal :: random_jump.current

let split_right random_jump literal =
  random_jump.current <- Left literal :: random_jump.current
*)
let saved_branch_to_string saved_branch =
  String.concat "\n"
    (List.map
       (fun replay ->
	  match replay with
	    | Left literal ->
		"LEFT: " ^ Term.literal_to_string literal;
		
	    | Right literal ->
		"RIGHT: " ^ Term.literal_to_string literal;
		
	    | Split literal ->
		"SPLIT: " ^ Term.literal_to_string literal;
       )
       saved_branch
    )

let create config state bound =
  {
    config = config;
    state = state;
    bound = bound;
    saved = [];
    counter = Counter.create ();
  }


let get_remaining_time_ratio (config: config) : float =
  if Config.time_out_CPU config = 0 then begin
    1.0 (*failwith "random jump no timeout"*)
  end
    
  else begin
    (float_of_int (Config.time_out_CPU config) -. Sys.time ())
    /.
      (float_of_int (Config.time_out_CPU config))
  end


(*
let compute_remaining_search_space (branch: choice_point list) : float =
  let rec compute_remaining_search_space' estimate remaining =
    match remaining with
      | [] ->
	  failwith "compute_remaining_search_space' 1"

      | head :: [] ->
	  if not (State.choice_point_equal (State.root_choice_point state) head) then
	    failwith "compute_remaining_search_space' 2";

	  estimate

      | head :: tail ->
	  let estimate' =
	    if State.is_right_split head then
	      estimate
	    else
	      estimate + 1.0
	  in
	    compute_remaining_search_space' (estimate' / 2.0) tail
  in
    compute_remaining_search_space' branch 0.0
*)


(*
let compute_remaining_search_space (time_ratio: float) (branch: choice_point list) : saved_branch * int =
  let reversed =
    List.rev branch
  in

  let skip =
    ref 0
  in

  let rec compute_remaining_search_space' (space_ratio: float) (decision_level: int) (remaining_branch: choice_point list) =
    match remaining_branch with
      | [] ->
	  (* no branches skipped, so we still have enough time to explore the current search space *)
	  if !skip = 0 then
	    raise Exit

	  else
	    []

      | decision_level :: tail ->
	  (* ignore the root level *)
	  if decision_level = 0 then begin
	    compute_remaining_search_space' space_ratio (decision_level + 1) tail
	  end

	  else begin
	    (* enough time left, just replay decision and continue *)
	    if space_ratio > time_ratio then begin
	      let replay =
		if State.is_right_split head then
		  Right (State.choice_point_literal head)
		else
		  Left (State.choice_point_literal head)
	      in
		replay :: compute_remaining_search_space' estimate level tail
	    end

	    else begin
	      (* not enough time left, skip from here on *)
	      let replay, estimate' =
		if State.is_right_split head then
		  Right (State.choice_point_literal head),
		  estimate

		else begin
		  skip := !skip + 1;
		  
		  Left (State.choice_point_literal head),
		  estimate +. (1.0 /. (2.0 ** (float_of_int level)))
		end
	      in
		replay :: compute_remaining_search_space' estimate' (level + 1) tail
	    end	    
	  end
  in
    try
      compute_remaining_search_space' 0.0 0 branch, !skip
    with
      | Exit ->
	  [], 0
*)	      




let find_random_jump_target (state: state) (time_ratio: float) (branch: choice_point list) : choice_point option =
(*  print_endline ("time_ratio: " ^ string_of_float time_ratio);*)
  let time_ratio' =
    time_ratio +. 0.1
  in

    if time_ratio' > 1.0 then begin
      None
    end

    else  begin
  (* prechek *)
  let reversed =
    List.rev branch
  in

  (* start at root *)
  let rec find_random_jump_target' (space_ratio: float) (depth: int) (remaining_branch: choice_point list) : choice_point option =
    match remaining_branch with
      | [] ->
	  None

      | decision_level :: tail ->
	  (* ignore the root level *)
	  if State.choice_point_equal decision_level (State.root_choice_point state) then begin
	    find_random_jump_target' space_ratio (depth + 1) tail
	  end

	  (* jump at least over 5 decision levels *)
	  else if List.length tail < 5 then begin
	    None
	  end
(*
	  (* jump at least over the current level *)
	  else if State.choice_point_equal decision_level (State.active_choice_point state) then begin
	    None
(*	    find_random_jump_target' space_ratio (depth + 1) tail*)
	  end
*)
	  (* right split - no additional search space here *)
	  else if State.is_right_split decision_level then begin
	    (* actually, there shouldn't be any right splits anymore *)
	    failwith "random jump - right split";
	    find_random_jump_target' space_ratio (depth + 1) tail
	  end

	  else begin
	    (* add the search space of the unexplored right split -
	       the search space is organized as a binary tree *)
	    let space_ratio' =
	      space_ratio +. (1.0 /. (2.0 ** (float_of_int depth)))
	    in
(*	      print_endline ("space_ratio: " ^ string_of_float space_ratio);*)

	      (* not enough time left, skip from here on *)
	      if space_ratio' > time_ratio' then begin
		Some decision_level
	      end

	      (* enough time left, continue search ... *)
	      else begin
		find_random_jump_target' space_ratio' (depth + 1) tail
	      end
	  end
  in
    find_random_jump_target' 0.0 0 reversed
    end


let compute_saved_branch (state: state) (jump_target: choice_point) (branch: choice_point list) : saved_branch =

  let add_right_splits decision_level saved_branch =
    let right_splits =
      State.choice_point_right_literals decision_level
    in
      List.fold_left
	(fun acc split ->
	   Right split :: acc
	)
	saved_branch
	right_splits
  in

  (* start from current level and go backwards to root *)
  let rec compute_saved_branch' (remaining_branch: choice_point list) (saved_branch: saved_branch) =
    match remaining_branch with
      | [] ->
	  saved_branch

      | decision_level :: tail ->
	  (* the root level *)
	  if State.choice_point_equal decision_level (State.root_choice_point state) then begin
	    compute_saved_branch' tail (add_right_splits decision_level saved_branch)
	  end

	  (* jumped over - replay *)
	  else if State.compare_age decision_level jump_target > 0 then begin
(*	  else if State.choice_point_equal decision_level (State.active_choice_point state) then begin*)
	    compute_saved_branch' tail (Split (State.choice_point_literal decision_level) :: (add_right_splits decision_level saved_branch))
	  end

	    (* *)
	  else (*if State.compare_age decision_level jump_target <= 0 then*) begin
	    compute_saved_branch' tail (Left (State.choice_point_literal decision_level) :: (add_right_splits decision_level saved_branch))
	  end
  in
    compute_saved_branch' branch []



let jump (random_jump: random_jump) : choice_point option =
  Counter.inc random_jump.counter;

  if not (Config.random_jump random_jump.config) then begin
    None
  end

  (* only check each check_at steps *)
  else if ((Counter.value random_jump.counter) mod check_at) != 0 then begin
    None
  end

  else begin
(*    print_endline ("CHECK: " ^ string_of_int (Counter.value random_jump.counter));*)
    let branch =
      State.get_branch random_jump.state
    in
(*
    let remaining_search_space_ratio : float =
      compute_remaining_search_space branch
    in
*)
    let remaining_time_ratio : float =
      get_remaining_time_ratio random_jump.config
    in

    let jump_target =
      find_random_jump_target random_jump.state remaining_time_ratio branch
    in

      match jump_target with
	| None ->
	    None

	| Some jump_level ->
	    let jump_level =
	      match random_jump.bound#jump_target with
		| [] ->
		    jump_level

		| other :: _ ->
		    if State.compare_age jump_level other > 0 then
		      other
		    else
		      jump_level
	    in

	    let saved_branch =
	      compute_saved_branch random_jump.state jump_level branch
	    in
	      random_jump.saved <- saved_branch :: random_jump.saved;
	      
	      Statistic.inc_global_debug ();
(*	      print_endline (saved_branch_to_string saved_branch);*)


(*	      failwith ("random jump: " ^ State.choice_point_to_string jump_level);*)
	      
	      jump_target
  end
    

let finished random_jump : bool =
  random_jump.saved == []


let pop random_jump : saved_branch =
  (* pop oldest first *)
  let x =
    List.rev random_jump.saved
  in
    match x with
    | [] ->
	raise Not_found

    | head :: tail ->
	random_jump.saved <- List.rev tail;
	head


let register x y =
  y.saved <- x.saved
