(* $Id: selection.ml,v 1.30 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)

type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type choice_point = State.choice_point
type state = State.state
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected = Selection_types.selected
type problem_literals = Problem_literals.problem_literals
type context = Context.context
type context_element = Context.element


type selection = {
  sel_config: config;
  sel_bound: bound;
  sel_statistic: statistic;
  sel_state: state;
  sel_context: context;

  (* storage for the literals of the input clause set.
     triggers search for context unifiers. *)
  sel_problem_literals: problem_literals;

  (* the context_unifier_spaces build on the input clause set *)
  sel_context_unifier_spaces: context_unifier_space array;
  mutable sel_lemma_spaces: context_unifier_space list;

  (* found closing context unifiers during the last candidate computation. *)
(*  sel_close: raw_context_unifier list ref;*)
  sel_close: raw_context_unifier option ref;

  (* assert candidates from unit clauses of the problem clause set *)
  mutable sel_unit_asserts: (literal * raw_context_unifier) list;

  (* assert candidates *)
  mutable sel_assert: Selection_assert.candidates;

  (* split candidates *)
  mutable sel_split: Selection_split.candidates;

  mutable sel_last_assert: int;
  mutable sel_last_split: int;
  mutable sel_no_splits_yet: bool;

  sel_process_close: raw_context_unifier -> unit;
  sel_process_assert: raw_context_unifier -> bool;
  sel_process_split: raw_context_unifier -> unit;

  mutable sel_bt_counter: int ref;

  mutable sel_saved_branch: Random_jump.saved_branch;
}






exception SATISFIABLE




(*** creation ***)

let create (clauses: clause list) (lemmas: clause list) (config: config) (bound: bound)
  (statistic: statistic) (state: state) (context: context) (saved_branch: Random_jump.saved_branch) : selection =

(*  print_endline ("SAVED\n" ^ Random_jump.saved_branch_to_string saved_branch);*)
(*
  print_endline ("CLAUSES: ");
  List.iter
    (fun clause -> print_endline (Term.clause_to_string clause))
    clauses;

  print_endline ("LEMMAS: ");
  List.iter
    (fun clause -> print_endline (Term.clause_to_string clause))
    lemmas;
*)
  let problem_literals =
    Problem_literals.create config bound statistic state context
  in

  let selection_assert =
    Selection_assert.create config bound statistic state context
  and selection_split =
    Selection_split.create config bound statistic state context
  in

  let close_list =
    ref None
  in

  let process_close_candidate (candidate: raw_context_unifier) : unit =
    (* :TODO: heuristics for which CLOSE to apply. *)
(*    raise (Context_unifier.CLOSE candidate);*)

    let explanation =
      State.get_explanation
	state
	(Context_unifier.get_context_literals candidate.Context_unifier.rcu_context_partners)
    in
      begin
      match explanation with
	| [] ->
(*	    print_endline (Context_unifier.raw_context_unifier_to_string candidate);*)
(*	    print_endline ("C0");*)
	    raise (Context_unifier.CLOSE candidate)
	| head :: [] when
	    State.choice_point_equal (State.active_choice_point state) head->
(*	    print_endline ("C1");*)
	    raise (Context_unifier.CLOSE candidate)

	| _ ->
	    ()
      end;

      begin
	match !close_list with
	  | None ->
	      close_list := Some candidate (*:: !close_list*)

	  | Some old ->
	      let old_explanation =
		State.get_explanation
		  state
		  (Context_unifier.get_context_literals old.Context_unifier.rcu_context_partners)
	      in
		
	      let rec compare_explanations x y =
		match x, y with
		  | [], [] ->
		      0

		  | [], _ ->
		      -1

		  | _, [] ->
		      1

		  | x0 :: t0, x1 :: t1 ->
		      if State.is_right_split x0 then
			compare_explanations t0 y

		      else if State.is_right_split x1 then
			compare_explanations x t1

		      else begin
			let cmp =
			  State.compare_age x0 x1
			in
			  if cmp <> 0 then
			    cmp
			  else
			    compare_explanations t0 t1
		      end
	      in
		if compare_explanations explanation old_explanation < 0 then
		  close_list := Some candidate (*:: !close_list*)		  
      end;



  and process_assert_candidate (candidate: raw_context_unifier) : bool =
    Selection_assert.add selection_assert candidate
  and process_split_candidate (candidate: raw_context_unifier) : unit =
    Selection_split.add selection_split candidate
  in

  (* create a context_unifier_space for every clause,
     gather all candidates computed on creation *)
  let context_unifier_spaces, unit_asserts =
    List.fold_left
      (fun (context_unifier_spaces, unit_asserts) clause ->
	 let new_context_unifier_space =
	   Context_unifier_space.create
	     config
	     state
	     statistic
	     bound
	     problem_literals
	     clause
	     process_close_candidate
	     process_assert_candidate
	     process_split_candidate
	     false
	 in

	 let new_unit_asserts =
	   match clause with
	     | unit_literal :: [] ->
		 let raw_context_unifier = {
		   Context_unifier.rcu_space = new_context_unifier_space;
		   Context_unifier.rcu_context_partners = [| Context_unifier.assert_partner |];
		 }
		 in
		   Statistic.inc_computed_assert_candidates statistic;
		   (unit_literal, raw_context_unifier) :: unit_asserts
		     
	     | _ ->
		 unit_asserts
	 in
	   (new_context_unifier_space :: context_unifier_spaces),
	   new_unit_asserts
      )
      ([], [])
      clauses
  in

  let lemma_spaces, unit_asserts =
    List.fold_left
      (fun (context_unifier_spaces, unit_asserts) clause ->
	 let new_context_unifier_space =
	   Context_unifier_space.create
	     config
	     state
	     statistic
	     bound
	     problem_literals
	     clause
	     process_close_candidate
	     process_assert_candidate
	     process_split_candidate
	     true
	 in
	 let new_unit_asserts =
	   match clause with
	     | unit_literal :: [] ->
		 let raw_context_unifier = {
		   Context_unifier.rcu_space = new_context_unifier_space;
		   Context_unifier.rcu_context_partners = [| Context_unifier.assert_partner |];
		 }
		 in
		   Statistic.inc_computed_assert_candidates statistic;
		   (unit_literal, raw_context_unifier) :: unit_asserts
		     
	     | _ ->
		 unit_asserts
	 in
	   (new_context_unifier_space :: context_unifier_spaces),
	   new_unit_asserts
      )
      ([], unit_asserts)
      lemmas
  in

  (* sort the assert literals of unit clauses by term_weight *)
  let sorted_unit_asserts =
    List.sort
      (fun (x, _) (y, _) ->
	 compare
	   (Term_attributes.weight_of_literal x)
	   (Term_attributes.weight_of_literal y)
      )
      unit_asserts
  in    

  let context_unifier_spaces_array =
    Array.of_list context_unifier_spaces
  in

  let selection =
    {
      sel_config = config;
      sel_bound = bound;
      sel_statistic = statistic;
      sel_state = state;
      sel_context = context;
      sel_problem_literals = problem_literals;
      sel_context_unifier_spaces = context_unifier_spaces_array;
      sel_lemma_spaces = lemma_spaces;
      sel_close = close_list;
      sel_unit_asserts = sorted_unit_asserts;
      sel_assert = selection_assert;
      sel_split = selection_split;
      sel_last_assert = Context.get_max_id context;
      sel_last_split = Context.get_max_id context;
      sel_no_splits_yet = true;
      sel_process_close = process_close_candidate;
      sel_process_assert = process_assert_candidate;
      sel_process_split = process_split_candidate;
      sel_bt_counter = ref 0;
      sel_saved_branch = saved_branch;
    }
  in
    selection





(* if there are more asserts, don't compute any more *)
let max_assert = 20000
(* compute at least these many asserts if possible *)
let min_assert = 10000

(* if there are more splits, don't compute any more *)
let max_split = 10000
(* compute at least these many splits if possible *)
let min_split = 5000


(*** add ***)

let is_defer_split_candidate (selection: selection) : bool =
  (Statistic.get_split selection.sel_statistic = 0)
(*  ||
  (
    (
      (Statistic.get_assert selection.sel_statistic)
      /
      (Statistic.get_split selection.sel_statistic)
    )
    >=
   5
  )*)

(* add the new context literal to all unsubsumed clauses
   and get the new candidates *)
let add (selection: selection) (context_element: Context.element) : unit =
  let no_assert =
    false
      (*
    (* already deferred asserts *)
    (
      selection.sel_last_assert
      <>
      (Context.get_max_prev_id selection.sel_context)
    )
    ||
    (* now too many assert candidates *)
    (
      Selection_assert.size selection.sel_assert
      >
      max_assert
    )
      *)
  in

  let no_split =
    (* already deferred assert *)
    no_assert
    ||
    (* start computing splits only if assert alone does not suffice *)
    selection.sel_no_splits_yet
      (*
    ||
    (* already deferred splits *)
    (
      selection.sel_last_split
      <>
      (Context.get_max_prev_id selection.sel_context)
    )
    ||
    (* now too many split candidates *)
    (
      Selection_split.size selection.sel_split
      >
      max_split
    )
      *)
  in

    if (*false &&*) no_assert then begin
(*      Statistic.inc_global_debug ();*)
(*      print_endline ("no assert");*)
(*      print_endline (string_of_int (Context.get_max_prev_id selection.sel_context));
      failwith ("no assert");*)
      Problem_literals.add selection.sel_problem_literals context_element Context_unifier.Close;
    end

    else if (*false &&*) no_split then begin
(*      print_endline ("no split");*)
(*      Statistic.inc_global_debug ();*)
(*      failwith ("no split");*)
      Problem_literals.add selection.sel_problem_literals context_element Context_unifier.CloseAssert;
      selection.sel_last_assert <- Context.next_el selection.sel_context selection.sel_last_assert;
    end

    else begin
(*      print_endline ("all");*)
      Problem_literals.add selection.sel_problem_literals context_element Context_unifier.All;
      selection.sel_last_assert <- Context.next_el selection.sel_context selection.sel_last_assert;
      selection.sel_last_split <- Context.next_el selection.sel_context selection.sel_last_split;
    end;


    begin
  match !(selection.sel_close) with
    | Some closing(* :: _*) ->	
	selection.sel_close := None (*[]*);
(*	print_endline ("C2");*)
	raise (Context_unifier.CLOSE closing)

    | None (*[]*) ->
	()
    end(*;
    print_endline "ADD DONE"*)
(*
  if is_defer_split_candidate selection then begin
    Problem_literals.add selection.sel_problem_literals context_element Context_unifier.CloseAssert;

    match selection.sel_last_split_search with
      | None ->
	  selection.sel_last_split_search <- Some context_element
	    
      | Some _ ->
	  ()
  end

  else begin
    Problem_literals.add selection.sel_problem_literals context_element Context_unifier.All;
  end
*)




(*** selection ***)


let rec select_unit_assert
    (selection: selection) (candidates: (literal * raw_context_unifier) list)
    : selected option =

  match candidates with
    | [] ->
	selection.sel_unit_asserts <- [];
	None
	
    | (candidate, raw_context_unifier) :: tail ->
	begin
	  match Context.check_subsumed selection.sel_context candidate with
	    | Some _ ->
		select_unit_assert selection tail
		
	    | None ->
		selection.sel_unit_asserts <- tail;
		
		State.apply_assert
		  selection.sel_state
		  candidate
		  raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		  raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
		  0
		  [| |]
		;
		Some {
		  Selection_types.candidate_type = Selection_types.Assert;
		  Selection_types.literal = candidate;
		  Selection_types.raw_context_unifier = raw_context_unifier;
		}
	end


let select (selection: selection) : selected =
  let selected =
    (*  print_endline ("0");*)
    if Selection_assert.has_closing_candidate selection.sel_assert then begin
      match Selection_assert.select selection.sel_assert with
	| Some selected ->
	    selected
	      
	| None ->
	    failwith "Selection.select"
    end
    else
      (* first check if a closing context unifier has been found. *)
      (*  match !(selection.sel_close) with
	  | Some closing(* :: _*) ->	
	  selection.sel_close := None (*[]*);
	  raise (Context_unifier.CLOSE closing)

	  | None (*[]*) ->*)
      (* now try to assert a unit input clause *)
      begin
	(*  print_endline ("1");*)
	match select_unit_assert selection selection.sel_unit_asserts with
	  | Some selected ->
	      selected
		
	  | None ->

	      begin
		let right_split =
		  if Config.neg_assert_candidates selection.sel_config == Flags.NAC_Use then begin
		    None
		  end

		  else begin
		    Selection_split.select_right_split selection.sel_split
		  end

		in
		  match right_split with
		    | Some selected ->
			selected
			  
		    | None ->
			



			(* then try to find an assert literal *)
			begin
			  (*  print_endline ("2");*)
			  let assert_candidate =
			    let rec select () =
			      (*  print_endline ("3");*)

			      if
				selection.sel_last_assert
				<
				  Context.get_max_id selection.sel_context
			      then begin
				(*			      while
							      (
							      selection.sel_last_assert
							      <
							      Context.get_max_id selection.sel_context
							      )
							      &&
							      (
							      Selection_assert.size selection.sel_assert
							      <
							      min_assert
							      )
							      do*)
				(*  print_endline ("4");*)
				selection.sel_last_assert <- Context.next_el selection.sel_context selection.sel_last_assert;
				Problem_literals.add_deferred
				  selection.sel_problem_literals selection.sel_last_assert Context_unifier.Assert;
				selection.sel_last_assert <- Context.get_max_id selection.sel_context;
			      end;
			      let selected =
				Selection_assert.select selection.sel_assert
			      in

				match selected with
				  | Some _ ->
				      selected
					
				  | None ->
				      if
					(
					  selection.sel_last_assert
					  <
					    Context.get_max_id selection.sel_context
					)
					&&
					  (
					    Selection_assert.size selection.sel_assert
					    <
					      min_assert
					  )
				      then begin
					Problem_literals.add_deferred
					  selection.sel_problem_literals selection.sel_last_assert Context_unifier.Assert;
					selection.sel_last_assert <- Context.get_max_id selection.sel_context;
				      end;
				      Selection_assert.select selection.sel_assert
			    in
			      select ()
			  in
			    match assert_candidate with
			      | Some selected ->
				  selected
				    
			      | None ->
				  (* finally try to find a split literal *)
				  let replay =
				    match selection.sel_saved_branch with
				      | [] ->
					  None

				      | head :: _ ->
					  Some head
				  in
				    
				    begin
				      (*  print_endline ("5");*)
				      selection.sel_no_splits_yet <- false;
				      let split_candidate =
					let rec select () =
					  if
					    (
					      selection.sel_last_split
					      <
						Context.get_max_id selection.sel_context
					    )
					    &&
					      (
						Selection_split.size selection.sel_split
						<
						  min_split
					      )
					  then begin
					    selection.sel_last_split <- Context.next_el selection.sel_context selection.sel_last_split;
					    Problem_literals.add_deferred
					      selection.sel_problem_literals selection.sel_last_split Context_unifier.Split;
					    selection.sel_last_split <- Context.get_max_id selection.sel_context;
					    (*			      done;*)
					    (*			      select ();*)
					  end;
					  let selected =
					    Selection_split.select selection.sel_split replay
					  in
					    match selected with
					      | Some _ ->
						  selected
						    
					      | None ->
						  if
						    selection.sel_last_split
						    <
						      Context.get_max_id selection.sel_context
						  then begin
						    (*			      while
									      (
									      selection.sel_last_split
									      <
									      Context.get_max_id selection.sel_context
									      )
									      &&
									      (
									      Selection_split.size selection.sel_split
									      <
									      min_split
									      )
									      do*)
						    (*  print_endline ("6");*)
						    selection.sel_last_split <- Context.next_el selection.sel_context selection.sel_last_split;
						    Problem_literals.add_deferred
						      selection.sel_problem_literals selection.sel_last_split Context_unifier.Split;
						    selection.sel_last_split <- Context.get_max_id selection.sel_context;
						    (*			      done;*)
						    (*			      select ();*)
						  end;
						  Selection_split.select selection.sel_split replay

					in
					  select ()
				      in

					(*

					(* continue the computation of  the split candidates
					  from the last choice point where a split occurred. *)
					  if is_defer_split_candidate selection then begin
					  match selection.sel_last_split_search with
					  | None ->
					  ()
					  
					  | Some context_element ->
					  Problem_literals.add
					  selection.sel_problem_literals context_element Context_unifier.Split;
					  selection.sel_last_split_search <- None;
					  end;
					*)			
					begin
					  (*  print_endline ("7");*)
					  (*			  match Selection_split.select selection.sel_split with*)
					  match split_candidate with
					    | Some selected ->
						selected
						  
					    | None ->
						(* process kept candidates exceeding the depth limit. *)
						Selection_assert.check_exceeding selection.sel_assert;
						Selection_split.check_exceeding selection.sel_split;
						raise SATISFIABLE
					end
				    end
			end  
	      end
      end	
  in
    if
      selected.Selection_types.candidate_type = Selection_types.Assert
      ||
      selected.Selection_types.candidate_type = Selection_types.UnitSplit
    then
      ()

    else begin
      match selection.sel_saved_branch with
	| [] ->
	    ()
	      
	| Random_jump.Split saved_literal :: tail ->
	    (* if a left split, then it must be a recorded on *)
	    if
	      selected.Selection_types.candidate_type = Selection_types.SplitLeft
	      &&
	      Term.are_literals_skolem_variants selected.Selection_types.literal saved_literal
	    then begin
	      selection.sel_saved_branch <- tail
	    end
	      
	    else begin
	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string selected.Selection_types.literal);
	      failwith "Selection.jump 1"
	    end


	| Random_jump.Left saved_literal :: tail ->
	    (* here there must be a right split *)
	    if
	      selected.Selection_types.candidate_type = Selection_types.SplitRight
	      &&
	      Term.are_literals_skolem_variants selected.Selection_types.literal saved_literal
	    then begin
	      selection.sel_saved_branch <- tail
	    end

	    else begin
	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string selected.Selection_types.literal);
	      failwith "Selection.jump 2"
	    end

	| Random_jump.Right saved_literal :: tail ->
	    (* here there must be a right split *)
	    if
	      selected.Selection_types.candidate_type = Selection_types.SplitRight
	      &&
	      Term.are_literals_skolem_variants selected.Selection_types.literal saved_literal
	    then begin
	      selection.sel_saved_branch <- tail
	    end

	    else begin
	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string selected.Selection_types.literal);
	      failwith "Selection.jump 3"
	    end
    end;
    selected


let dec_counter_lemmas selection =
  if !(selection.sel_bt_counter) = 0 then begin
    List.iter
      (fun space ->
	 space.Context_unifier.cus_value := !(space.Context_unifier.cus_value) / 2
      )
      selection.sel_lemma_spaces
  end




(*** backtrack ***)
let backtrack (selection: selection) (retracted: choice_point) (explanation: choice_point list) : unit =
  selection.sel_bt_counter := (!(selection.sel_bt_counter) + 1) mod 100;
  dec_counter_lemmas selection;

  (* backtracking invalidates everything from the last split on,
     so in particular all asserted literals are retracted. *)
  selection.sel_last_assert <- min selection.sel_last_assert (Context.get_max_id selection.sel_context);
  selection.sel_last_split <- min selection.sel_last_split (Context.get_max_id selection.sel_context);
  selection.sel_close := None (*[]*);

  (* context partners - dependent on Context.backtrack for max context id *)
  Problem_literals.backtrack selection.sel_problem_literals (Context.get_max_id selection.sel_context);

  (* partial context unifiers - dependent on Problem_literals.backtrack *)
  Array.iter
    (fun space ->
       if Problem_literals.is_space_currently_subsumed selection.sel_problem_literals space then
	 ()
       else
(*	 space.Context_unifier.cus_backtrack <- true*)
	 Context_unifier_space.backtrack selection.sel_config space
    )
    selection.sel_context_unifier_spaces;

  List.iter
    (fun space ->
       if Problem_literals.is_space_currently_subsumed selection.sel_problem_literals space then
	 ()
       else
(*	 space.Context_unifier.cus_backtrack <- true*)
	 Context_unifier_space.backtrack selection.sel_config space
    )
    selection.sel_lemma_spaces;
(*  List.iter (Context_unifier_space.backtrack selection.sel_config) selection.sel_lemma_spaces;*)

  (* candidates *)
  Selection_assert.backtrack selection.sel_assert;
  Selection_split.backtrack selection.sel_split retracted explanation




let min_lemmas = 50
let max_lemmas = 100

let get_min_lemma_value selection : int =
    match 
      List.fold_left
	(fun acc space ->
	   match acc with
	     | None ->
		 Some !(space.Context_unifier.cus_value)

	     | Some x ->
		 if x <= !(space.Context_unifier.cus_value) then
		   acc
		 else
		   Some !(space.Context_unifier.cus_value)
	)
	None
	selection.sel_lemma_spaces
    with
      | None ->
	  0
	  (*failwith "selection.min_lemma_value"*)

      | Some x ->
	  x



let prune_lemmas selection =
(*  ()*)

  if List.length selection.sel_lemma_spaces >= max_lemmas then begin
  while List.length selection.sel_lemma_spaces > min_lemmas do
    let min =
      get_min_lemma_value selection
    in
(*      print_endline ("Prune: " ^ string_of_int min);*)
      selection.sel_lemma_spaces <-
	List.find_all
	(fun space ->
	   if
	     !(space.Context_unifier.cus_value) > min
	   then
	     true

	   else begin
	     (*print_endline (Term.clause_to_string space.Context_unifier.cus_clause);*)
	     false
	   end
	)
	selection.sel_lemma_spaces
  done
  end


let add_lemma (selection: selection) (lemma: clause) : unit =

  let lemma_length =
    List.length lemma
  in

  let rec clauses_equal x y =
    match x, y with
      | [], [] ->
	  true
	    
      | hx :: tx, hy :: ty when
	  Term.literal_equal hx hy ->
	  clauses_equal tx ty
	    
      | _ ->
	  false
  in

  let clauses_variants x y =
    try
      let x : subst * subst =
      List.fold_left2
	(fun (subst1, subst2) x' y' ->
	   let new_subst1 =
	   Unification.match_literals
	     x' 0
	     y' 1
	   in
	   let new_subst2 =
	   Unification.match_literals
	     y' 0
	     x' 1
	   in
	   let merged_subst1 =
	     Unification.match_substs
	       0
	       subst1
	       new_subst1
	   in
	   let merged_subst2 =
	     Unification.match_substs
	       0
	       subst2
	       new_subst2
	   in
	     (merged_subst1, merged_subst2)
	)
	(Subst.empty, Subst.empty)
	x
	y
      in
	true
    with
      | Unification.UNIFICATION_FAIL ->
	  false
  in
    
    
  let not_new0 =
    Tools.array_exists
      (fun old ->
	 lemma_length = Array.length old.Context_unifier.cus_input_partners
	  &&
	  (
	     clauses_equal
	       old.Context_unifier.cus_clause
	       lemma	 
	    ||
	     clauses_variants
	       old.Context_unifier.cus_clause
	       lemma	 
	  )
      )
      selection.sel_context_unifier_spaces
  in
    
    if not_new0 then begin
      ();
      (*		    print_endline ("LEMMA already in input set - IGNORED\n");*)
    end
    else begin
      
      let not_new1 =
	List.exists
	  (fun old ->
	 lemma_length = Array.length old.Context_unifier.cus_input_partners
	  &&
	  (
	     clauses_equal
	       old.Context_unifier.cus_clause
	       lemma	 
	    ||
	     clauses_variants
	       old.Context_unifier.cus_clause
	       lemma	 
	  )
	  )
	  selection.sel_lemma_spaces
      in
	
	if not_new1 then begin
	  ();
	  (*			print_endline ("LEMMA learned - IGNORED\n");*)
	end
	else begin

(*		    print_endline ("LEMMA: " ^ Term.clause_to_string lemma);
		    print_newline ();*)

	  prune_lemmas selection;



  (* create clause *)
  let new_context_unifier_space =
    Context_unifier_space.create
      selection.sel_config
      selection.sel_state
      selection.sel_statistic
      selection.sel_bound
      selection.sel_problem_literals
      lemma
      selection.sel_process_close
      selection.sel_process_assert
      selection.sel_process_split
      true
  in

  let new_value =
    get_min_lemma_value selection
  in
    new_context_unifier_space.Context_unifier.cus_value := new_value;
    selection.sel_lemma_spaces <- new_context_unifier_space :: selection.sel_lemma_spaces;

  (* unit assert? *)
    begin
      match lemma with
	| unit_literal :: [] ->
	    let raw_context_unifier = {
	      Context_unifier.rcu_space = new_context_unifier_space;
	      Context_unifier.rcu_context_partners = [| Context_unifier.assert_partner |];
	    }
	    in
	      Statistic.inc_computed_assert_candidates selection.sel_statistic;
	      selection.sel_unit_asserts <- (unit_literal, raw_context_unifier) :: selection.sel_unit_asserts
		
      | _ ->
	  ()
    end;


    (* compute candidates *)
    Problem_literals.add_lemma selection.sel_problem_literals new_context_unifier_space;

    
    (* and close if a closing unifier has been found *)
    begin
    match !(selection.sel_close) with
      | None ->
	  ()
	  
      | Some candidate ->
	  selection.sel_close := None;
(*	  print_endline ("C3");*)
	  raise (Context_unifier.CLOSE candidate)

    end
    
		      end
end


let get_lemmas selection =
  List.map
    (fun space ->
       space.Context_unifier.cus_clause
    )
    selection.sel_lemma_spaces


let in_replay selection : bool =
  let x =
  selection.sel_saved_branch != []
  in
    if x then
      print_endline ("SAVED REMAINING\n" ^ Random_jump.saved_branch_to_string selection.sel_saved_branch);
    x;
