(* $Id: selection_assert.ml,v 1.52 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type choice_point = State.choice_point
type state = State.state
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected = Selection_types.selected
type selection_lookahead = Selection_lookahead.selection_lookahead
type context = Context.context




type candidate = {
  (* the context unifier. *)
  ca_raw_context_unifier: raw_context_unifier;

  (* depth or weight, depending on the deepening method. *)
  ca_term_quality: int;
  ca_generation: int;
}



let max_uncompressed = 16000
let max_lookahead = 100000
let max_lookahead_exceeding = 10000

(*** comparison ***)
(*
let compare_candidates (first: candidate) (second: candidate) : int =
  let term_quality =
    compare first.ca_term_quality second.ca_term_quality
  in
    if term_quality <> 0 then
      term_quality
    else
(*      let size =
	compare
	  (Array.length first.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	  (Array.length second.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
      in
	if size <> 0 then
	  size
	else*)
      let generation =
	compare 
	  (Context_unifier.generation_of_candidate
	     first.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	  (Context_unifier.generation_of_candidate
	     second.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
      in
	if generation <> 0 then
	  generation
	else
	  Context_unifier.compare_context_unifiers first.ca_raw_context_unifier second.ca_raw_context_unifier
*)

let compare_candidates (first: candidate) (second: candidate) : int =
    if first.ca_term_quality < second.ca_term_quality then
      -1
    else if first.ca_term_quality > second.ca_term_quality then
      1
    else begin
      if first.ca_generation < second.ca_generation then
	-1
      else if first.ca_generation > second.ca_generation then
	1
      else
	begin
	  Context_unifier.compare_context_unifiers first.ca_raw_context_unifier second.ca_raw_context_unifier
	end
    end

	    
module Valid =
  Heap.Make ( 
    struct
      type t = candidate

      let compare (first: candidate) (second: candidate) : int =
	compare_candidates first second
    end
  )


type compressed_candidates = {
  mutable front: (candidate * literal * int) option;
  candidates: raw_context_unifier option array;
  mutable current_index: int;
}


type candidates = {
  cd_config: config;
  cd_bound: bound;
  cd_statistic: statistic;
  cd_state: state;
  cd_context: context;

  (* freshly computed context unifiers
     which have to be transformed into candidates
     at the beginning of the next candidate selection. *)
(*  mutable cd_unprocessed: raw_context_unifier list;*)

  (* the assert candidates currently valid for adding to the context
     (at least valid when last checked). *)
  mutable cd_valid: Valid.t;

  (* invalid candidates and the choice point of the invalidating context literal. *)
  mutable cd_invalid: (raw_context_unifier * choice_point) list;

  (* candidates added to the context,
     with the assert literal and the choice point of addition.
     have to be reactivated in backtracking. *)
  mutable cd_added: (candidate * literal * choice_point) list;

  (* lookahead for contradictions within assert candidates *)
  mutable cd_lookahead: selection_lookahead;

  (* lookahead for contradictions within assert candidates
     exceeding the deepening bound. *)
  cd_lookahead_exceeding: selection_lookahead;

  (* exceeding candidates with their complexity and creation choice point. *)
  mutable cd_exceeding: (raw_context_unifier * Bound.complexity) list;

  mutable cd_preselected: (candidate * literal * int) option;

  mutable cd_compressed: compressed_candidates list;
}





(*** representation ***)


let candidate_to_string (config: config) (state: state)
  (title: string) (candidate: candidate) (literal: literal) : string =

  State.active_choice_point_to_string state
  ^ title ^ ": " ^ Term.literal_to_string literal ^ "\n"
  ^
  if Config.print_derivation_context_unifier config then
    Context_unifier.raw_context_unifier_to_string candidate.ca_raw_context_unifier
  else
    ""





			      
(*** creation ***)

let create (config: config) (bound: bound) (statistic: statistic) (state: state)
  (context: context) : candidates =

  {
    cd_config = config;
    cd_bound = bound;
    cd_statistic = statistic;
    cd_state = state;
    cd_context = context;
(*    cd_unprocessed = [];*)
    cd_valid = Valid.create ();
    cd_invalid = [];
    cd_added = [];
    cd_lookahead = Selection_lookahead.create max_lookahead;
    cd_lookahead_exceeding = Selection_lookahead.create max_lookahead_exceeding;
    cd_exceeding = [];
    cd_preselected = None;
    cd_compressed = [];
  }



(*** build ***)


(* compute the quality of a term based on the deepening bound criterion. *)
let get_term_quality (candidates: candidates) (literal: literal) : int =
  if Config.iterative_deepening candidates.cd_config = Flags.IT_TermDepth then
    Term_attributes.weight_of_literal literal
  else
    Term_attributes.depth_of_literal literal


(* build the assert literal represented by an assert context unifier *)
let build_assert_literal (candidates: candidates) (raw_context_unifier: raw_context_unifier) : literal * int =
  let subst =
    Context_unifier.recompute_unifier raw_context_unifier
  in

  let input_partners =
    raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners
  in

  (* find the assert literal input partner *)
  let assert_literal_partner, index =
    let rec check_at (index: int) : literal * int =
      if index >= Array.length input_partners then
	failwith "Selection_assert.build_candidate"
      else begin
	let input_partner =
	  input_partners.(index)
	in
	  if raw_context_unifier.Context_unifier.rcu_context_partners.(input_partner.Context_unifier.ip_index)
	    ==
	    Context_unifier.assert_partner
	  then
	    input_partner.Context_unifier.ip_literal, input_partner.Context_unifier.ip_index
	  else
	    check_at (index + 1)
      end
    in
      check_at 0
  in

  let assert_literal =
    Subst.apply_to_literal
      subst
      assert_literal_partner
      Subst.input_literal_offset
  in
    assert_literal, index



(*** add ***)

let remove_empty_compressed candidates =
  candidates.cd_compressed <-
    List.find_all
    (fun compressed ->
       match compressed.front with
	 | None -> false
	 | Some _ -> true
    )
    candidates.cd_compressed


let compress_candidates candidates =
  if Valid.size candidates.cd_valid >= max_uncompressed then begin
    remove_empty_compressed candidates;

    let first =
      Valid.remove_min candidates.cd_valid
    in
  let assert_literal, index =
    build_assert_literal candidates first.ca_raw_context_unifier
  in
    let new_compressed =
      Array.create
	(Valid.size candidates.cd_valid)
	None
    in
    let counter =
      ref 0
    in
      while not (Valid.is_empty candidates.cd_valid) do
	new_compressed.(!counter) <- Some (Valid.remove_min candidates.cd_valid).ca_raw_context_unifier;
	counter := !counter + 1;
      done;
      (*
      Valid.iter_increasing
	(fun candidate ->
	   new_compressed.(!counter) <- candidate.ca_raw_context_unifier;
	)
	candidates.cd_valid;
      *)
      candidates.cd_valid <- Valid.create ();
      candidates.cd_compressed <-
	{
	  front = Some (first, assert_literal, index);
	  candidates = new_compressed;
	  current_index = 0;
	}
	:: candidates.cd_compressed;
      
  end


(* adds a candidate to the valid candidate queue.
   returns the candidate if the close lookahead is successful. *)
let add_valid (candidates: candidates) (candidate: candidate) (literal: literal) : unit =
  compress_candidates candidates;
  try
    Valid.add candidates.cd_valid candidate;
  with
    | Valid.OVERFLOW ->
	raise (Const.NO_SOLUTION "Selection_assert.add_valid: candidate set overflow")







let do_build_candidate (candidates: candidates) (raw_context_unifier: raw_context_unifier) (assert_literal: literal) : candidate =
  let candidate = {
    ca_raw_context_unifier = raw_context_unifier;
    ca_term_quality = get_term_quality candidates assert_literal;
    ca_generation = Context_unifier.generation_of_candidate raw_context_unifier.Context_unifier.rcu_context_partners;
  }
  in
    candidate



(*
  different strategies:
  - immediately drop exceeding candidates
  - register exceeding candidates in lookahead
  - after a branch is exhausted, check if valid exceeding candidates exist
*)

(* build the assert candidate represented by an assert context unifier *)
let build_candidate (candidates: candidates) (raw_context_unifier: raw_context_unifier) :
  (candidate * literal * int) option =
  let assert_literal, index =
    build_assert_literal candidates raw_context_unifier
  in
  let complexity =
    candidates.cd_bound#get_complexity assert_literal
  in
(*
    (* negative Horn candidate meant for lookahead only? *)
    if
      Config.is_horn candidates.cd_config
      &&
      Config.horn_candidates candidates.cd_config = Flags.HC_Lookahead
      &&
      not assert_literal.Term.sign*)
    (* negative candidates meant for lookahead only? *)
    if
      Config.neg_assert_candidates candidates.cd_config = Flags.NAC_Lookahead
      &&
      not assert_literal.Term.sign
    then begin
      if
	Selection_lookahead.check candidates.cd_lookahead assert_literal 
	||
	Selection_lookahead.check candidates.cd_lookahead_exceeding assert_literal
	||
	  (* early check for contradictory assert candidates.
	     later on the corresponding closing context unifier is computed anyway,
	     but this way computation of other assert and split context unifiers
	     is stopped early. *)
	  begin
	    not (Config.is_horn candidates.cd_config)
	    &&
	    match Context.check_contradictory candidates.cd_context assert_literal with
	      | Some _ ->
		  true

	      | None ->
		  false
	  end
      then
	Some (do_build_candidate candidates raw_context_unifier assert_literal, assert_literal, index)

      else begin
	(* lookahead registration *)
	Selection_lookahead.add candidates.cd_lookahead_exceeding assert_literal raw_context_unifier;

	(* ... and silently forget the candidate. *)
	None
      end
	
    end

    (* exceeding candidate? *)
    else if candidates.cd_bound#exceeds complexity then begin
      if
	Config.lookahead_exceeding candidates.cd_config
	&&
	(
	  Selection_lookahead.check candidates.cd_lookahead assert_literal 
	  ||
	  Selection_lookahead.check candidates.cd_lookahead_exceeding assert_literal
	||
	  (* early check for contradictory assert candidates.
	     later on the corresponding closing context unifier is computed anyway,
	     but this way computation of other assert and split context unifiers
	     is stopped early. *)
	  begin
	    not (Config.is_horn candidates.cd_config)
	    &&
	    match Context.check_contradictory candidates.cd_context assert_literal with
	      | Some _ ->
		  true

	      | None ->
		  false
	  end
	)
      then begin
	Some (do_build_candidate candidates raw_context_unifier assert_literal, assert_literal, index)
      end

      else begin

	(* lookahead registration *)
	if Config.lookahead_exceeding candidates.cd_config then begin
	  Selection_lookahead.add candidates.cd_lookahead_exceeding assert_literal raw_context_unifier;
	end;

	(* keep candidates? *)
	begin
	  match Config.restart candidates.cd_config with
	    | Flags.RS_Eager
	    | Flags.RS_Lazy ->
		(* no, drop the candidate *)
		ignore (
		  candidates.cd_bound#register
		    complexity
		    (Context_unifier.get_creation_choice_point raw_context_unifier.Context_unifier.rcu_context_partners));
		
	    | Flags.RS_Delayed ->
		(* keep the exceeding candidates till an open branch is found. *)
		candidates.cd_exceeding <-
		  (raw_context_unifier, complexity) :: candidates.cd_exceeding;
	end;
	
	None
      end
	
    end

    (* regular candidate (obeying the depth bound) *)
    else begin
      (* is the candidate applicable?*)

      match Context.check_subsumed candidates.cd_context assert_literal with
	| Some element ->
	    (* no, not applicable *)

	    (* If the assert literal is permanently subsumed by another assert candidate literal
	       it can be dropped *)
	    if Context.is_element_mandatory_for_choice_point
	      candidates.cd_config
	      element
	      (Context_unifier.get_creation_cid raw_context_unifier.Context_unifier.rcu_context_partners)
	    then
	      (* drop it *)
	      ()
	    else
	      (* put it to the invalid candidates *)
	      candidates.cd_invalid <-
		(raw_context_unifier, element.Context.el_choice_point) :: candidates.cd_invalid;
	    
	    None
		
	| _ ->
	    (* yes, applicable *)
	    let candidate =
	      do_build_candidate candidates raw_context_unifier assert_literal
	    in
	      (* lookahead check *)
	      if
 		Selection_lookahead.check candidates.cd_lookahead assert_literal
		||
 		Selection_lookahead.check candidates.cd_lookahead_exceeding assert_literal
	||
	  (* early check for contradictory assert candidates.
	     later on the corresponding closing context unifier is computed anyway,
	     but this way computation of other assert and split context unifiers
	     is stopped early. *)
	  begin
	    not (Config.is_horn candidates.cd_config)
	    &&
	    match Context.check_contradictory candidates.cd_context assert_literal with
	      | Some _ ->
		  true

	      | None ->
		  false
	  end
	      then begin
		Some (candidate, assert_literal, index)
	      end
	      else begin
		if Config.print_assert_candidates candidates.cd_config then begin
		  print_string (
		    candidate_to_string
		      candidates.cd_config candidates.cd_state
		      "New Assert Candidate" candidate assert_literal);
		end;
		  
		Selection_lookahead.add candidates.cd_lookahead assert_literal candidate.ca_raw_context_unifier;

		add_valid candidates candidate assert_literal;

		None
	      end  
    end


(*
(* build the unprocessed candidates.
   returns a candidate if the close lookahead is successful. *)
let rec build_unprocessed_candidates (candidates: candidates) : (candidate * literal * int) option =
  match candidates.cd_unprocessed with
    | [] ->
	None
	  
    | candidate :: tail ->
	candidates.cd_unprocessed <- tail;
	begin
	  match build_candidate candidates candidate with
	    | None ->
		build_unprocessed_candidates candidates
		  
	    | Some _ as conflicting ->
		conflicting
	end


let add_unprocessed (candidates: candidates) (candidate: raw_context_unifier) : unit =
  candidates.cd_unprocessed <- candidate :: candidates.cd_unprocessed

let add (candidates: candidates) (candidate: raw_context_unifier) : unit =
  Statistic.inc_computed_assert_candidates candidates.cd_statistic;
  add_unprocessed candidates candidate
*)


let add_unprocessed (candidates: candidates) (candidate: raw_context_unifier) : bool =
(*  print_endline ("ADD_UNPROCESSED:\n" ^ Context_unifier.raw_context_unifier_to_string candidate);*)

  begin
    match candidates.cd_preselected with
      | Some _ -> (*failwith "add_unprocessed"*) true
      | _ -> 
begin
  match build_candidate candidates candidate with
    | None ->
	false
	  
    | Some _ as conflicting ->
(*	print_endline "HIT";*)
	candidates.cd_preselected <- conflicting;
	true
end
  end

let add (candidates: candidates) (candidate: raw_context_unifier) : bool =
  Statistic.inc_computed_assert_candidates candidates.cd_statistic;
  add_unprocessed candidates candidate




(*** select ***)

let rec get_best_candidate_uncompressed (candidates: candidates) : (candidate * literal * int) option =

  if Valid.is_empty candidates.cd_valid then begin
    None
  end

  else begin
    let candidate =
      Valid.min candidates.cd_valid
    in
    let assert_literal, index =
      build_assert_literal candidates candidate.ca_raw_context_unifier
    in
      
(*      Selection_lookahead.remove candidates.cd_lookahead assert_literal candidate.ca_raw_context_unifier;*)
      
      if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	(* candidate does no longer exist (due to backtracking) *)
	      let x: candidate =
		Valid.remove_min candidates.cd_valid
	      in
	get_best_candidate_uncompressed candidates;
      end
	
      else begin
	match Context.check_subsumed candidates.cd_context assert_literal with
	  | Some element ->
	      (* candidate is invalid with the context *)
	      if
		Context.is_element_mandatory_for_choice_point
		  candidates.cd_config
		  element
		  (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	      then begin
		()
	      end
		
	      else begin
		candidates.cd_invalid <-
		(candidate.ca_raw_context_unifier, element.Context.el_choice_point) :: candidates.cd_invalid;
	      end;
	      
	      let x: candidate =
		Valid.remove_min candidates.cd_valid
	      in
	      get_best_candidate_uncompressed candidates;
	      
	  | None ->
	      (* valid candidate found *)
	      Some (candidate, assert_literal, index)
      end
  end


let remove_compressed candidates (compressed_candidates: compressed_candidates) : unit =
  match compressed_candidates.candidates.(compressed_candidates.current_index) with
    | None ->
	compressed_candidates.front <- None

    | Some raw_context_unifier ->
	let assert_literal, index =
	  build_assert_literal candidates raw_context_unifier
	in
	let candidate =
	  do_build_candidate candidates raw_context_unifier assert_literal
	in
	  compressed_candidates.candidates.(compressed_candidates.current_index) <- None;
	  if compressed_candidates.current_index < Array.length compressed_candidates.candidates - 1 then begin
	    compressed_candidates.current_index <- compressed_candidates.current_index + 1
	  end;
	  compressed_candidates.front <- Some (candidate, assert_literal, index)


let rec advance_compressed candidates (compressed_candidates: compressed_candidates) : unit =

  match compressed_candidates.front with
    | None ->
	()

    | Some (candidate, assert_literal, index) ->
	  if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	    (* candidate does no longer exist (due to backtracking) *)
	    remove_compressed candidates compressed_candidates;
	    advance_compressed candidates compressed_candidates;
	  end
	    
	  else begin
	    match Context.check_subsumed candidates.cd_context assert_literal with
	      | Some element ->
		  (* candidate is invalid with the context *)
		  if
		    Context.is_element_mandatory_for_choice_point
		      candidates.cd_config
		      element
		      (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		  then begin
		    ()
		  end
		    
		  else begin
		    candidates.cd_invalid <-
		      (candidate.ca_raw_context_unifier, element.Context.el_choice_point) :: candidates.cd_invalid;
		  end;
		  
		  remove_compressed candidates compressed_candidates;
		  advance_compressed candidates compressed_candidates;
	      
	      | None ->
		  (* valid candidate found *)
		  ()
	  end




let rec get_best_candidate (candidates: candidates) : (candidate * literal * int) option =
  (* process all candidate queues till valid candidate is in front *)
  List.iter
    (advance_compressed candidates)
    candidates.cd_compressed;

  (* get best compressed *)
  let best_compressed : compressed_candidates option =
    List.fold_left
      (fun acc compressed ->
	 match acc with
	   | None ->
	       Some compressed

	   | Some old ->
	       begin
		 match old.front, compressed.front with
		   | None, None ->
		       None
			 
		   | None, Some _ ->
		       Some compressed
			 
		   | Some _, None ->
		       acc
			 
		   | Some (x, _, _), Some (y, _, _) ->
		       if compare_candidates x y > 0 then
			 Some compressed
		       else
			 acc
	       end
      )
      None
      candidates.cd_compressed
  in
    begin
      let best_uncompressed =
	get_best_candidate_uncompressed candidates
      in
	match best_uncompressed with
	  | None ->
	      begin
		match best_compressed with
		  | None ->
		      None

		  | Some best ->
		      let x =
		      best.front
		      in
		      remove_compressed candidates best;
		      x
	      end
		
	  | Some (candidate, assert_literal, index) ->
	      begin
		match best_compressed with
		  | None ->
		      let x: candidate =
			Valid.remove_min candidates.cd_valid
		      in
			best_uncompressed
			
		  | Some best ->
		      begin
			match best.front with
			  | None ->
(*			      failwith "bst_frnt"*)
		      let x: candidate =
			Valid.remove_min candidates.cd_valid
		      in
			best_uncompressed


			  | Some (candidate', assert_literal', index') ->
			      if compare_candidates candidate candidate' <= 0 then begin
				let x: candidate =
				  Valid.remove_min candidates.cd_valid
				in
				  best_uncompressed
			      end
				
			      else begin
		      let x =
		      best.front
		      in
		      remove_compressed candidates best;
		      x
			      end
		      end
	      end
    end
    
    (* compare with best uncompressed *)




let select (candidates: candidates) : selected option =
  (*
  let candidate =
    match build_unprocessed_candidates candidates with
      | Some _ as conflicting ->
	  conflicting

      | None ->
	  get_best_candidate candidates
  in
  *)
  let candidate =
    match candidates.cd_preselected with
      | Some _ as conflicting ->
(*	  print_endline "conflicting";*)
	  candidates.cd_preselected <- None;
	  conflicting

      | None ->
	  get_best_candidate candidates
  in

    match candidate with
      | None ->
	  (* all candidates processed, so all lookahead candidates must be invalid now *)
	  candidates.cd_lookahead <- Selection_lookahead.create max_lookahead;

	  None

      | Some (candidate, assert_literal, index) ->
	  (* move the chosen candidate to the used candidates *)
	  if not (Config.is_horn candidates.cd_config) then begin
	    candidates.cd_added <- (candidate, assert_literal, State.active_choice_point candidates.cd_state) :: candidates.cd_added;
	  end;
	  
	  let context_literals =
	    Context_unifier.get_context_literals
	      candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners
	  in
	    State.apply_assert
	      candidates.cd_state
	      assert_literal
	      candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
	      candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
	      index
	      context_literals;
	    Some {
	      Selection_types.candidate_type = Selection_types.Assert;
	      Selection_types.literal = assert_literal;
	      Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
	    }



let check_exceeding candidates =

  let rec check_exceeding exceeding =
    match exceeding with
      | [] ->
	  ()
	    
      | (raw_context_unifier, complexity) :: tail ->
	 if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then begin
	   failwith "Debug: check_exceeding";
	 end;

	  let choice_point =
	    Context_unifier.get_creation_choice_point raw_context_unifier.Context_unifier.rcu_context_partners
	  in
	    if
	      candidates.cd_bound#exceeds_current complexity choice_point
	      &&
	      let assert_literal, _ =
		build_assert_literal candidates raw_context_unifier
	      in
		not (Context.is_satisfied candidates.cd_context assert_literal)
		  
	    then begin
	      let exceeding =
		candidates.cd_bound#register
		  complexity
		  choice_point
	      in
		()
(*
		(* for exhaustive restarting any exceeding candidate is sufficient *)
		if 
		  exceeding
		  && 
		  Config.restart candidates.cd_config = Flags.RS_Exhaustive
		then
		  raise Exit*)
	    end;

	    check_exceeding tail
	      
  in
    try
      check_exceeding candidates.cd_exceeding
    with
      | Exit ->
	  ()




(*** backtrack ***)

(*
let backtrack_unprocessed (candidates: candidates) : unit =
  candidates.cd_unprocessed <-
    List.find_all
      (fun raw_context_unifier ->
	 not (Context_unifier.is_raw_context_unifier_invalid raw_context_unifier)
      )
      candidates.cd_unprocessed
*)

let backtrack_invalid (candidates: candidates) : unit =
  candidates.cd_invalid <-
    List.find_all
      (fun (raw_context_unifier, conflicting_choice_point) ->
	 if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then
	   false
	   
	 else if State.is_choice_point_invalid conflicting_choice_point then begin
	   if add_unprocessed candidates raw_context_unifier then begin
	     failwith "backtrack_invalid";
	   end;
	   false
	 end
	   
	 else
	   true
      )
      candidates.cd_invalid


let backtrack_added (candidates: candidates) : unit =

  let rec do' list =
    match list with
      | [] ->
	  candidates.cd_added <- []

      | (candidate, assert_literal, add_choice_point) :: tail ->
	 if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then
	   do' tail

	 else if State.is_choice_point_invalid add_choice_point then begin
	   ignore (add_valid candidates candidate assert_literal);
	   do' tail
	 end
	   
	 else
	   candidates.cd_added <- list

  in
    do' candidates.cd_added
(*
  candidates.cd_added <-
    List.find_all
      (fun (candidate, assert_literal, add_choice_point) ->
	 if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then
	   false

	 else if State.is_choice_point_invalid add_choice_point then begin
	   ignore (add_valid candidates candidate assert_literal);
	   false
	 end
	   
	 else
	   true
      )
    candidates.cd_added
*)

let backtrack_exceeding (candidates: candidates) : unit =
  candidates.cd_exceeding <-
    List.find_all
      (fun (raw_context_unifier, _) ->
	 not (Context_unifier.is_raw_context_unifier_invalid raw_context_unifier)
      )
      candidates.cd_exceeding



(* removes candidates based on context unifiers
   which do not longer exist after backtracking.

   Furthermore, added and invalid candidates might be reclassified *)
let backtrack (candidates: candidates) : unit =
  begin
    match candidates.cd_preselected with
      | None -> ()
      | Some (candidate, _, _) ->
	 if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then
	   candidates.cd_preselected <- None
	 else
	  failwith "backrack: preselected"
  end;

(*  Selection_lookahead.backtrack candidates.cd_lookahead;*)
  candidates.cd_lookahead <- Selection_lookahead.create max_lookahead;
  Selection_lookahead.backtrack candidates.cd_lookahead_exceeding;

(*  backtrack_unprocessed candidates;*)
  backtrack_invalid candidates;
  backtrack_added candidates;
  backtrack_exceeding candidates








let has_closing_candidate candidates =
  match candidates.cd_preselected with
    | None -> false
    | _ -> true


let size candidates =
  Valid.size candidates.cd_valid
