(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(** choice point management

  The {e explanation} of a context literal
  which stems from the context unifier {e C}
  is the union of the explanations of the context literals used in {e C}.
  That is,
  a context unifier which is solely based on {e -v} resp. {e +v}
  has an empty explanation.

  The {e explanation} of a choice point is the choice point itself
  plus the set of the explanations of the context literals
  associated with this choice point,
  i.e. the split literal and the subsequently asserted literals.
  Note that in contrast to propositional DPLL
  here a Split is not only dependent on itself,
  but is like an Assert dependent on other choice points.

  That is if
    - a choice point {e s} is created by a Split on literal {e L}
      of the context unifier {e C}
    - and the context literals of {e C} were added to the context
      in choice points {e s{_0}}, ..., {e s{_ n}}

  then the explanation
    - of {e L} is the union of the explanation of {e s{_ 0}}, ..., {e s{_n}}.
    - of {e s} is the union of \{{e s}\} and the explanation of {e L}.

  If furthermore (still in state {e s})
    - literal {e K} from context unifier {e D} is asserted
    - and the context literals of {e D} were added to the context
      in choice points {e t{_0}}, ..., {e t{_ n}}

  then the explanation

    - of {e K} is the union of the choice points of {e t{_ 0}}, ..., {e t{_ n}}.
    - of {e s} is the union of the previous explanation of {e s}
      and the explanation of {e K}.
*)

let trace = false

(*
  max. size of regressed literal within regression,
  regression is stopped if the size is exceeded.

  the final lemma might still be very small,
  but its computation is to expensive.
*)
let find_uip = true
let max_regress_size =
    10000
let max_lemma_length = 10


exception MAX_REGRESSED

(*** types ***)

type counter = Counter.counter
type config = Config.config
type literal = Term.literal
type clause = Term.clause

(* explanations are alway kept decreasingly sorted,
   i.e. youngest choice points first, oldest choice points at the end of the list *)
type explanation = choice_point list

and choice_point = {
  (* age, where 0 is oldest *)
  cp_id: int;

  (* the creating split literal. *)
  cp_split_literal: literal;
  mutable cp_right_splits: literal list;

  (* created by a left or a right split? *)
  cp_left_split: bool;

  (* valid, or already invalidated? *)
  mutable cp_valid: bool;
}

type branch_node = {
  (* the context literal *)
  bn_added_literal: literal;

  (* added in this step - needed for removal on backtracking *)
  bn_added_cp: choice_point;

  (* context unifier based on this clause *)
  bn_clause: clause;
  (* and this clause literal, counting from 0 *)
  bn_clause_index: int;
  bn_clause_value: int ref;

  (* context literals of the context unifier *)
  bn_context_literals: literal array;

  (* dependent on these choice points -
     based on bn_context_literals, memoization  *)
  mutable bn_explanation: explanation option;
}


module LiteralTable = Term.LiteralTable


type state = {
  (* environment *)
  st_config: config;

  (* ever increasing choice point id counter. *)
  st_id: counter;

  st_root_choice_point: choice_point;

  mutable st_branch: choice_point list;

  (* the union of the explanations of all context unifiers
     applied by Close. Only valid choice points. *)
  mutable st_closing_explanation: explanation;

  (* mapping from a context literal its branch node information *)
  st_literal_to_branch_node: branch_node LiteralTable.t;

  (* mapping from a context literal to its context id,
     i.e. the level of the context literal if the context is seen as a stack. *)
  st_literal_to_context_id: int LiteralTable.t;
}


class type prune =
object
  method is_prune: bool
  method prune: choice_point -> unit
end



exception UNSATISFIABLE of clause * literal array

(*** constants ***)

let empty_explanation : explanation =
  []

let valid_choice_point : choice_point = {
  cp_id = -1;
  cp_split_literal = Term.null_literal;
  cp_right_splits = [];
  cp_left_split = true;
  cp_valid = true;
}

let invalid_choice_point : choice_point = {
  cp_id = -2;
  cp_split_literal = Term.null_literal;
  cp_right_splits = [];
  cp_left_split = true;
  cp_valid = false;
}


(* does no pruning *)
let prune_dummy : prune =
object
  method is_prune = false
  method prune _ = ()
end




(*** creation ***)

let create (config: config) : state =
  let root_choice_point = {
    cp_id = 0;
    cp_split_literal = Term.null_literal;
    cp_right_splits = [];
    cp_left_split = false; (* needed to avoid backtracking *)
    cp_valid = true;
  }
  in

  let state =
    {
      st_config = config;
      st_id = Counter.create_with 0;
      st_root_choice_point = root_choice_point;
      st_branch = [root_choice_point];
      st_literal_to_branch_node = LiteralTable.create 1024;
      st_literal_to_context_id = LiteralTable.create 1024;
      st_closing_explanation = empty_explanation;
    }
  in

  let root_branch_node = {
    bn_added_literal = Term.null_literal;
    bn_added_cp = root_choice_point;
    bn_clause = [];
    bn_clause_value = ref 0;
    bn_clause_index = -2; (* just some number to fail on if ever used *)
    bn_context_literals = [| |];
    bn_explanation = Some [];
  }
  in
    
    (* add special literals with an empty explanation. *)
    LiteralTable.add state.st_literal_to_branch_node Term.assert_literal root_branch_node;
    LiteralTable.add state.st_literal_to_branch_node (Config.default_v state.st_config) root_branch_node;
    LiteralTable.add state.st_literal_to_branch_node Term.null_literal root_branch_node;
    
    state


let root_choice_point (state: state) : choice_point =
  state.st_root_choice_point


(* the active choice point, i.e. the youngest choice point *)
let active_choice_point (state: state) : choice_point =
  match state.st_branch with
    | [] ->
	failwith "State.active_choice_point"

    | head :: _ ->
	head




(*** representation ***)

let id_to_string (id: int) =
  "[" ^ string_of_int id ^ "] "
  
let choice_point_to_string (choice_point: choice_point) : string =
  id_to_string choice_point.cp_id(* ^ " - " ^ Term.literal_to_string choice_point.cp_split_literal*)

let active_choice_point_to_string (state: state) : string =
  "[" ^ string_of_int (active_choice_point state).cp_id ^ "] "

let explanation_to_string (explanation: explanation) : string =
  "[" ^ String.concat ", " (List.map choice_point_to_string explanation) ^ "]"

let choice_point_to_string_extended (choice_point: choice_point) : string =
  "[" ^ string_of_int choice_point.cp_id ^ "] :"
  ^ "\nValid: " ^ string_of_bool choice_point.cp_valid
  ^ "\nLeft:  " ^ string_of_bool choice_point.cp_left_split
    

let branch_to_string (state: state) : string =
  List.fold_left
    (fun acc choice_point ->
       choice_point_to_string_extended choice_point ^ "\n" ^ acc
    )
    ""
    state.st_branch





(*** Comparison ***)

let choice_point_equal (first: choice_point) (second: choice_point) : bool =
  first == second


let compare_age (first: choice_point) (second: choice_point) : int =
  Tools.compare_int first.cp_id second.cp_id





(*** properties ***)

let id_of_choice_point (choice_point: choice_point) : int =
  choice_point.cp_id

let is_choice_point_valid (choice_point: choice_point) : bool =
  choice_point.cp_valid

let is_choice_point_invalid (choice_point: choice_point) : bool =
  not choice_point.cp_valid

let is_left_split (choice_point: choice_point) : bool =
  choice_point.cp_left_split

let is_right_split (choice_point: choice_point) : bool =
  not choice_point.cp_left_split

let choice_point_literal (choice_point: choice_point) : literal =
  choice_point.cp_split_literal

let choice_point_right_literals (choice_point: choice_point) : literal list =
  choice_point.cp_right_splits

let get_branch (state: state) : choice_point list =
  state.st_branch

let rec contains_explanation_choice_point (explanation: explanation) (choice_point: choice_point) : bool =
  match explanation with
    | [] ->
	false

    | head :: _ when
	head.cp_id < choice_point.cp_id ->
	false

    | head :: _ when
	choice_point_equal choice_point head ->
	true

    | _ :: tail ->
	contains_explanation_choice_point tail choice_point


(* more sophisticated approaches like
   - building and traversing a tree over the choice_point ids
   - or assigning a prime number to each choice_point,
     computing the factor of the primes of all choice_points it depends on,
     and then checking if base divides the factor
   showed to be worse.
*)
let depends_on (config: config) (dependent: choice_point) (base: choice_point) : bool =
  base.cp_id <= dependent.cp_id








(*** explanations ***)


(* explanations are decreasingly sorted *)
let rec merge_explanations (first: explanation) (second: explanation) : explanation =
  if first == second then
    first
  else
    match first, second with
      | [], _ ->
	  second
	    
      | _, [] ->
	  first
	    
      | head1 :: tail1, head2 :: tail2 ->
	  if Const.debug && (is_choice_point_invalid head1 || is_choice_point_invalid head2) then
	    failwith "merge_explanations";
	  
	  begin
	    if head1.cp_id = head2.cp_id then
	      head1 :: merge_explanations tail1 tail2
	    else if head1.cp_id < head2.cp_id then
	      head2 :: merge_explanations first tail2
	    else
	      head1 :: merge_explanations tail1 second

	    (*
	    match compare head1.cp_id head2.cp_id with
	      | -1 ->
		  head2 :: merge_explanations first tail2
		    
	      | 0 ->
		  head1 :: merge_explanations tail1 tail2
		    
	      | _ ->
		  head1 :: merge_explanations tail1 second*)
	  end


let rec get_explanation (state: state) (literals: literal array) : explanation =
  Array.fold_left
    (fun acc literal ->
       let branch_node =
	 try
	   LiteralTable.find state.st_literal_to_branch_node literal
	 with
	   | Not_found ->
	       failwith ("State.get_explanation: " ^ Term.literal_to_string literal)
       in
       let explanation =
	 match branch_node.bn_explanation with
	   | Some explanation ->
	       explanation
		 
	   | None ->
	       let explanation =
		 get_explanation state branch_node.bn_context_literals
	       in
		 branch_node.bn_explanation <- Some explanation;
		 explanation
       in
	 merge_explanations acc explanation
    )
    []
    literals







(*** choice_point access ***)

let apply_assert (state: state) (literal: literal) (clause: clause) (clause_value: int ref) (clause_index: int) (context_literals: literal array) : unit =
  let branch_node = {
    bn_added_literal = literal;
    bn_added_cp = active_choice_point state;
    bn_clause = clause;
    bn_clause_value = clause_value;
    bn_clause_index = clause_index;
    bn_context_literals = context_literals;
    bn_explanation = None;
  }
  in
    (*
    begin
      try
	let x : branch_node =
	  LiteralTable.find state.st_literal_to_branch_node literal
	in
	  failwith ("apply_assert literal " ^ Term.literal_to_string literal);
      with
	| Not_found ->
	    ()
    end;
    *)

    LiteralTable.add state.st_literal_to_branch_node literal branch_node


let apply_split ~(left: bool) (state: state) (literal: literal) (clause: clause) (clause_index: int) (explanation: explanation) : choice_point =
  let new_choice_point = {
    cp_id = Counter.next state.st_id;
    cp_split_literal = literal;
    cp_right_splits = [];
    cp_left_split = left;
    cp_valid = true;
  }
  in

  let branch_node = {
    bn_added_literal = literal;
    bn_added_cp = new_choice_point;
    bn_clause = clause;
    bn_clause_value = ref 0;
    bn_clause_index = clause_index;
    bn_context_literals = [| |];
    bn_explanation = Some (new_choice_point :: explanation);
  }
  in
(*
    begin
      try
	let x : branch_node =
	  LiteralTable.find state.st_literal_to_branch_node literal
	in
	  failwith ("apply_assert literal " ^ Term.literal_to_string literal);
      with
	| Not_found ->
	    ()
    end;
*)
    LiteralTable.add state.st_literal_to_branch_node literal branch_node;

    (* extend the branch *)
    state.st_branch <- new_choice_point :: state.st_branch;
    new_choice_point


let apply_split_left (state: state) (literal: literal) (clause: clause) (clause_index: int) : choice_point =
  apply_split ~left:true state literal clause clause_index []

(*
let apply_split_right (state: state) (literal: literal) (clause: clause) (clause_index: int) (explanation: explanation) : unit =
  let cp =
    apply_split ~left:false state literal clause clause_index explanation
  in
    ()
*)

let apply_split_right (state: state) (literal: literal) (clause: clause) (clause_value: int ref) (clause_index: int) (explanation: explanation) : unit =
  let active =
    active_choice_point state
  in
    active.cp_right_splits <- literal :: active.cp_right_splits;
  let branch_node = {
    bn_added_literal = literal;
    bn_added_cp = active_choice_point state;
    bn_clause = clause;
    bn_clause_value = ref 0;
    bn_clause_index = clause_index;
    bn_context_literals = [| |];
    bn_explanation = Some explanation;
  }
  in

(*
    begin
      try
	let x : branch_node =
	  LiteralTable.find state.st_literal_to_branch_node literal
	in
	  failwith ("apply_assert literal " ^ Term.literal_to_string literal);
      with
	| Not_found ->
	    ()
    end;
*)
    LiteralTable.add state.st_literal_to_branch_node literal branch_node

(*
  let context_literals =
    Array.of_list (
      List.map
	(fun cp ->
	   cp.cp_split_literal
	)
	explanation
    )
  in
(*    print_endline ("SPLIT RIGHT : " ^ Term.literal_to_string literal);*)
    apply_assert
      state
      literal
      clause
      clause_value
      clause_index
      context_literals
*)  


let set_context_id state literal id =
  LiteralTable.add state.st_literal_to_context_id literal id



(*** backtracking ***)

let invalidate (state: state) (choice_point: choice_point) (prune: prune) : unit =

  let rec invalidate' choice_points =
    match choice_points with
      | head :: tail ->
	  if compare_age choice_point head < 0 then begin
	  
	    if head.cp_valid then begin
(*	      print_endline ("INVALIDATE: " ^ choice_point_to_string head);*)
	      head.cp_valid <- false;
	      
	      if prune#is_prune then begin
		prune#prune head
	      end;
	      
	    end;
	    
	    invalidate' tail
	  end
	  else if compare_age choice_point head = 0 then begin
    
	    if head.cp_valid then begin
(*	      print_endline ("INVALIDATE: " ^ choice_point_to_string head);*)
	      head.cp_valid <- false;
	      (*
	      if prune#is_prune then begin
		prune#prune head
	      end;
	      *)
	    end;
	    

	    state.st_branch <- tail;

	  end

	  else begin
	    failwith "State.invalidate I"
	  end
	      
      | [] ->
	  failwith "State.invalidate II"
  in
      invalidate' state.st_branch


(*
(* removes all invalidates choice points (up to oldest_invalidated). *)
let backtrack_branch (state: state) (oldest_invalidated: choice_point) (prune: prune) : unit =

  let rec do_compress_branch (branch: choice_point list) (kept: choice_point list) : choice_point list =
    match branch with
      | head :: tail when
	  compare_age oldest_invalidated head <= 0 ->
	  (* still not gathered all invalid choice points ...*)

	  (*
	  (*:DEBUG:*)
	  if not head.cp_left_split then begin
	    invalidate_choice_point state head prune;
	    do_compress_branch tail kept
	  end else
	  *)

	  if is_choice_point_valid head then begin
	    do_compress_branch tail (head :: kept)
	  end

	  else begin

	    do_compress_branch tail kept
	  end

      | valid_branch ->
	  (* finally, all invalid choice points removed,
	     so readd the kept choice points. *)
	  List.fold_left
	    (fun acc choice_point ->
	       if is_choice_point_valid choice_point then
		 choice_point :: acc
		   
	       else
		 acc
	    )
	    valid_branch
	    kept
  in
  let compressed_branch =
    do_compress_branch state.st_branch []
  in
    (* :DEBUG:
    if List.exists is_choice_point_invalid compressed_branch then begin
      print_endline ("BT: " ^ choice_point_to_string_extended oldest_invalidated);
      List.iter (fun x -> print_endline (choice_point_to_string_extended x)) compressed_branch;
      failwith "State.compress_branch";
    end;
    *)
    state.st_branch <- compressed_branch
*)    



(* remove the invalid literal to choice point mappings. *)
let backtrack_remove_literal_to_branch_nodes (state: state) : unit =
  let invalid_keys =
    LiteralTable.fold
      (fun literal branch_node acc ->
	 if is_choice_point_invalid branch_node.bn_added_cp then begin
	   literal :: acc
	 end
	 else
	   acc
      )
      state.st_literal_to_branch_node
      []
  in
    List.iter
      (LiteralTable.remove state.st_literal_to_branch_node)
      invalid_keys;
    List.iter
      (LiteralTable.remove state.st_literal_to_context_id)
      invalid_keys



(* removes all invalid elements from a list. *)
let prune (state: state) (func: 'a -> choice_point) (apply: 'a -> unit)
    (oldest_retracted: choice_point) (values: 'a list) : 'a list =

  let rec prune' (value: 'a list) (kept: 'a list) : 'a list =
    match value with
      | head :: tail when
	  compare_age oldest_retracted (func head) <= 0 ->

	  (* still not gathered all valid element ... *)
	  if is_choice_point_valid (func head) then
	    prune' tail (head :: kept)

	  else begin
	    apply head;
	    prune' tail kept;
	  end

      | valid ->
	  (* finally, all invalid choice points removed,
	     so readd the kept choice points. *)
	  List.fold_left
	    (fun acc value ->
	       value :: acc
	    )
	    valid
	    kept
  in
    prune' values []

let prune_explanation state oldest explanation =
  prune state (fun x -> x) (fun _ -> ()) oldest explanation

(*
(* removes all invalid choice points from an closing explanation. *)
let prune_explanation (state: state) oldest (explanation: explanation) : explanation =
  if Config.chronological_backtracking state.st_config then
    let rec do_prune_explanation explanation =
      match explanation with
	| head :: tail when
	    is_choice_point_invalid head ->
	    do_prune_explanation tail

	| _ ->
	    explanation
    in
      do_prune_explanation explanation
  else
    List.find_all is_choice_point_valid explanation
*)
(*
(* removes all invalidates choice points (up to oldest_invalidated). *)
let backtrack_branch (state: state) (oldest_invalidated: choice_point) (prune: prune) : unit =

  let rec do_compress_branch (branch: choice_point list) (kept: choice_point list) : choice_point list =
    match branch with
      | head :: tail when
	  compare_age oldest_invalidated head <= 0 ->
	  (* still not gathered all invalid choice points ...*)

	  (*
	  (*:DEBUG:*)
	  if not head.cp_left_split then begin
	    invalidate_choice_point state head prune;
	    do_compress_branch tail kept
	  end else
	  *)

	  if is_choice_point_valid head then begin
	    do_compress_branch tail (head :: kept)
	  end

	  else begin

	    do_compress_branch tail kept
	  end

      | valid_branch ->
	  (* finally, all invalid choice points removed,
	     so readd the kept choice points. *)
	  List.fold_left
	    (fun acc choice_point ->
	       if is_choice_point_valid choice_point then
		 choice_point :: acc
		   
	       else
		 acc
	    )
	    valid_branch
	    kept
  in
  let compressed_branch =
    do_compress_branch state.st_branch []
  in
    (* :DEBUG:
    if List.exists is_choice_point_invalid compressed_branch then begin
      print_endline ("BT: " ^ choice_point_to_string_extended oldest_invalidated);
      List.iter (fun x -> print_endline (choice_point_to_string_extended x)) compressed_branch;
      failwith "State.compress_branch";
    end;
    *)
    state.st_branch <- compressed_branch
*)


(*


(*
   :NOTE:
  this whole depends_on business with one choice_point works only
  for chronological backtracking wrt. Split,
  i.e. 
*)

let get_best_backtrack_choice_point (state: state) (explanation: choice_point list) : choice_point =
  let choice_point =
    List.fold_left
      (fun acc choice_point ->
	 if
	   choice_point != state.st_root_choice_point
	   &&
	   (* left split *)
           choice_point.cp_left_split
	   &&
	   (* not dependend on any other left split of the conflict set *)
	   List.for_all
	     (fun base_for ->
		choice_point == base_for
		||
		(not base_for.cp_left_split)
		||
		not (contains_explanation_choice_point explanation base_for)
	     )
	     choice_point.cp_base_for
	 then begin
	   match acc with
	     | None ->
		 Some choice_point
		   
	     | Some old ->
		 if
		   List.length old.cp_base_for
		   <
		   List.length choice_point.cp_base_for
		 then
		   Some choice_point
		 else
		   acc
	 end
	   
	 else begin
	   acc
	 end
      )
      None
      explanation
  in
    match choice_point with
      | None ->
	  raise UNSATISFIABLE;
	  failwith "State.get_best_backtrack_choice_point";

      | Some x -> x
*)






(*** lemma ***)


let rec get_clause_literal clause i =
  match clause with
    | [] ->
	failwith "get_clause_literal"

    | head :: tail ->
	if i = 0 then
	  head
	else
	  get_clause_literal tail (i - 1)

type to_regress = {
  tr_clause_literal: literal;
  tr_clause_offset: int;
  tr_context_literal_id: int;
  tr_branch_node: branch_node;
  tr_branch_offset: int;
}


module To_regress =
  Set.Make (
    struct
      type t = to_regress
(*	
      let compare x y =
	if x == y then
	  0
	else
	  (* order by context literal *)
	  let cmp_index =
	    compare
	      x.tr_context_literal_id
	      y.tr_context_literal_id
	  in
	    if cmp_index <> 0 then
	      cmp_index
	    else
	      (* discriminate same context literal by its offset, i.e. by variants *)
	      let cmp_offset =
		compare
		  x.tr_branch_offset
		  y.tr_branch_offset
	      in
		if cmp_offset <> 0 then
		  cmp_offset
		else
		  failwith "To_regress.compare"*)

      let compare (x: to_regress) (y: to_regress) : int =
	if x == y then
	  0
	else begin
	  (* order by context literal *)
	  if x.tr_context_literal_id < y.tr_context_literal_id then
	    -1
	  else if x.tr_context_literal_id > y.tr_context_literal_id then
	    1
	  else begin
	    (* discriminate same context literal by its offset, i.e. by variants *)
	    if x.tr_branch_offset < y.tr_branch_offset then
	      -1
	    else if x.tr_branch_offset > y.tr_branch_offset then
	      1
	    else
	      failwith "To_regress.compare"
	  end
	end

    end
  )


let rec regress
    (state: state)
    (* only regress context literals up to this choice point *)
    (regress_stop: choice_point)
    (* clause literal * offset contained in the regressed literal set *)
    (regressed: (literal * int) list)
    (* size of regressed list, passed along for efficiency *)
    (regressed_size: int)
    (* unification constraints on the regressed literal set *)
    (constraints: ((literal * int) * (literal * int)) list)
    (* branch nodes to regress * offset *)
    (to_regress: To_regress.t)
    (* fresh offsets for the regression *)
    (next_offset: int ref)
    :
    (* the regresset set *)
    ((literal * int) list)
    *
    (* the constraints *)
    (((literal * int) * (literal * int)) list)
    =

  (* for each context literal to regress *)
  if trace then begin
    print_endline ("REGRESS");
  end;
  
  (* done *)
  if To_regress.is_empty to_regress then begin
    if trace then begin
      print_endline ("DONE");
    end;
    regressed, constraints
  end

  else begin
    let head =
      To_regress.max_elt to_regress
    in
    let last =
      To_regress.min_elt to_regress
    in
      (* if we reached the UIP, stop right here *)
      if find_uip &&
	(* ignore closing clause *)
	(not (Term.literal_equal head.tr_clause_literal Term.null_literal))
	&&
	(* all literals to regress are paired with the same context literal *)
	(head.tr_context_literal_id = last.tr_context_literal_id)
      then begin

	(* not the split node *)
	if Array.length head.tr_branch_node.bn_context_literals > 0 then begin
(*	  Statistic.inc_global_debug ();*)
	  ()
	end
	else begin
	  ();
	  (*print_endline ("SPLIT");*)
	end;

	if trace then begin
	let uip_literal: literal =
	  LiteralTable.fold
	    (fun literal id acc ->
	       if id = head.tr_context_literal_id then
		 literal
	       else
		 acc
	    )
	    state.st_literal_to_context_id
	    Term.null_literal
	in
	  print_endline ("UIP : " ^ choice_point_to_string head.tr_branch_node.bn_added_cp ^ " " ^ Term.literal_to_string uip_literal);
	end;
	let regressed =
	  To_regress.fold
	    (fun to_regress regressed ->
	       let new_regressed =
		 (to_regress.tr_clause_literal, to_regress.tr_clause_offset)
	       in
		 new_regressed :: regressed
	    )
	    to_regress
	    regressed
	in
	  regressed, constraints;
      end
	
      (* add constraint and regress *)
      else begin
	let tail =
	  To_regress.remove head to_regress
	in
	  
	(* score used clause *)
(*	let next =
	  To_regress.max_elt to_regress
	in*)
	  if
	    To_regress.is_empty tail
	    ||
	    head.tr_branch_node != (To_regress.max_elt tail).tr_branch_node
	  then begin
(*	    print_endline ("INC: ");*)
	    head.tr_branch_node.bn_clause_value := !(head.tr_branch_node.bn_clause_value) + 1;
	  end
	  else begin
(*	    print_endline ("DIFF: ");	    *)
	  end;

	let branch_node = head.tr_branch_node
	and offset = head.tr_branch_offset
	in
	  if trace then begin
	    print_endline ("REGRESSING NODE: " ^ Term.literal_to_string branch_node.bn_added_literal);
	    print_endline (Term.clause_to_string branch_node.bn_clause);
	    print_endline (Term.clause_to_string (Array.to_list branch_node.bn_context_literals));
	  end;
	  let rec check (index: int) (regressed: (literal * int) list) (regressed_size: int)
	      (constraints: ((literal * int) * (literal * int)) list) (to_regress: To_regress.t) =
	    if index >= Array.length branch_node.bn_context_literals then begin
	      regress state regress_stop regressed regressed_size constraints to_regress next_offset
	    end

	    else begin
	      (* process the current context_literal *)
	      let clause_literal =
		get_clause_literal branch_node.bn_clause index
	      in
	      let context_literal =
		branch_node.bn_context_literals.(index)
	      in
	  
		if trace then begin
		  print_endline ("REGRESSING LITERAL: "
				 ^ Term.literal_to_string clause_literal ^ "  --  "
				 ^ Term.literal_to_string context_literal);
		end;
		(* check type of context literal *)
		
		(* the assert literal, paired with nothing - just ignore it *)
		if Term.literal_equal context_literal Term.assert_literal then begin
		  if trace then begin
		    print_endline ("__ASSERT__");
		  end;
		  check (index + 1) ((*(clause_literal, offset) :: *)regressed) regressed_size constraints to_regress
		end
		  
		(* paired with -v - just add clause literal *)
		else if Term.literal_equal context_literal (Config.default_v state.st_config) then begin
		  if trace then begin
		    print_endline ("__V__");
		  end;
		  check (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress
		end
		  
		else begin
		  let branch_node' =
		    try
		      LiteralTable.find state.st_literal_to_branch_node context_literal
		    with
		      | Not_found ->
			  failwith ("State.regress.check: " ^ Term.literal_to_string context_literal)
		  in
		    
		  let explanation =
		    match branch_node'.bn_explanation with
		      | None ->
			  get_explanation state branch_node.bn_context_literals
			    
		      | Some explanation ->
			  explanation
		  in
		    match explanation with
		      | [] ->
			  if trace then begin
			    print_endline ("ROOT ASSERT");
			  end;
			  (* root assert - entailed by the input, just add clause literal *)
			  check (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress
			    
		      | creation_point :: _ when
			  (compare_age regress_stop creation_point > 0) ->
			  if trace then begin
			    print_endline ("OUT OF RANGE");
			  end;
			  (* out of regression range, just add clause literal *)
			  check (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress

		      | _ ->
			  (* (left) split - just add clause literal  *)
			  if
			    (* split *)
			    (Array.length branch_node'.bn_context_literals = 0)
			  then begin
			    
			    if branch_node'.bn_added_cp.cp_left_split then begin
			      if trace then begin
				print_endline ("LEFT SPLIT");
			      end;
			      check (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress
			    end
			      
			    else begin
			      if trace then begin
				print_endline ("RIGHT SPLIT");
			      end;
			      check (index + 1) (((*Term.request_negated_literal*) clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress
			    end
			  end
			    
			  else if regressed_size > max_regress_size then begin
			    if trace then begin
			      print_endline ("MAX_REGRESSED " ^ string_of_int regressed_size);
			    end;
			    raise MAX_REGRESSED;
			    (*check (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) constraints to_regress*)
			  end

			  else begin
			    if trace then begin
			      print_endline ("ASSERT");
			    end;
			    (* assert litearal to regress *)
			    let new_offset =
			      next_offset := !next_offset + 1;
			      !next_offset
			    in
			      if compare_age branch_node.bn_added_cp regress_stop < 0 then
				failwith "UA";

			    let context_literal_id =
			      try
				LiteralTable.find state.st_literal_to_context_id context_literal
			      with
				| Not_found ->
				    failwith "get_context_literal_id"
			    in
			    let new_to_regress = {
			      tr_clause_offset = offset;
			      tr_clause_literal = clause_literal;
			      tr_context_literal_id = context_literal_id;
			      tr_branch_offset = new_offset;
			      tr_branch_node = branch_node';
			    }
			    in
			      check
				(index + 1)
				regressed
				regressed_size
				constraints
				(To_regress.add new_to_regress to_regress)
			      (*
			    let new_constraint =
			      (clause_literal, offset),
			      (get_clause_literal branch_node'.bn_clause branch_node'.bn_clause_index, new_offset)
			    in
			      check
				(index + 1)
				regressed
				regressed_size
				(new_constraint :: constraints)
			      ((branch_node', new_offset) :: to_regress)
			      *)
			  end
		end
	    end
	      
	  in
	  let new_constraints =
	    (* empty clause - start of regression, so no constraint *)
	    if head.tr_clause_literal == Term.null_literal then begin
	      constraints
	    end

	    else begin
	      let new_constraint =
		(head.tr_clause_literal, head.tr_clause_offset),
		(get_clause_literal head.tr_branch_node.bn_clause head.tr_branch_node.bn_clause_index, head.tr_branch_offset)
	      in
		new_constraint :: constraints
	    end
	  in
	    check 0 regressed regressed_size new_constraints tail
	end
  end





      let are_clauses_variants x y : bool =
	let rec do' x' y' subst =
	  match x' with
	    | [] ->
		true

	    | h :: t ->
		begin
		  let rec loop2 processed unprocessed =
		    match unprocessed with
		      | [] ->
			  false
			    
		      | h' :: t' ->
			  if Term.are_literals_weak_variants h h' then begin
			    try
			      let new_subst =
				Unification.match_literals
				  h' 0
				  h 1
			      in
			      let merged_subst =
				Unification.match_substs
				  0
				  subst
				  new_subst
			      in
				do' t (processed @ t') merged_subst
				||
				loop2 (h' :: processed) t'
			    with
			      | Unification.UNIFICATION_FAIL ->
				  loop2 (h' :: processed) t'
			  end
			  else
			    loop2 (h' :: processed) t'
		  in
		    loop2 [] y'
		  
		end
		
	in
	  if List.length x != List.length y then begin
(*		print_endline ("NO VARIANTS I");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
	    false
	  end
	  else begin
	    if do' x y Subst.empty then begin
(*		print_endline ("VARIANTS");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
		true
	    end

	    else begin
(*		print_endline ("NO VARIANTS");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
	      false
	    end
	  end




      let subsumes_greater_smaller x y : bool =
	let rec do' x' subst =
	  match x' with
	    | [] ->
		raise Exit

	    | greater :: tail ->

		begin
		  List.iter
		    (fun smaller ->
		       try
			 let new_subst =
			   Unification.match_literals
			     greater 0
			     smaller 1
			    in
			    let merged_subst =
			      Unification.match_substs
				0
				subst
				new_subst
			    in
			      do' tail merged_subst
		       with
			 | Unification.UNIFICATION_FAIL ->
			     ()
			       
		    )
		    y
		    
		end	  
	in
	  try
	    do' x Subst.empty;
(*		print_endline ("NO SUBSUME");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
	    false
	  with
	    | Exit ->
(*		print_endline ("SUBSUME");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
		true





      let condense (clause: clause) : clause =
	let rec do' condensed to_process =
	    match to_process with
	      | [] ->
		  condensed

	      | head :: tail ->
		  (* :TODO: just keep head if it is ground *)
(*		print_endline ("CONS: " ^ Term.literal_to_string head);*)
		  let condensed' =
		    List.find_all
		      (fun x -> not (Term.literal_equal head x))
		      condensed
		  in

		  if
		    List.exists
		      (fun x ->
			 not (Term.literal_equal head x)
			 &&
			   begin
			     try
			       let x : Subst.subst =
				 Unification.match_literals head 0 x 1
			       in
				 true
			     with
			       | Unification.UNIFICATION_FAIL ->
				   false
			   end
		      )
		      condensed
		    &&
		      
		      subsumes_greater_smaller condensed condensed'
		  then
		    do' condensed' tail
		  else
		    do' condensed tail
	in
	  do' clause clause
      



let (*partition_variants*) filter_variants3 (state: state) (literals: literal list) : literal list =
  let var_rep =
    Hashtbl.create 64
  in
  let literal_rep : Term.var LiteralTable.t =
    LiteralTable.create 128
  in
  let ground_literals : unit LiteralTable.t =
    LiteralTable.create 128
  in

  let rec get_rep var =
    try
      let rep =
	Hashtbl.find var_rep (Var.id_of_var var)
      in
	get_rep rep
    with
      | Not_found ->
	  var
  in


  (* merge equivalence classes of variables,
     i.e. let them point to the same representative *)
  let merge_var_classes vars =
    match vars with
      | [] ->
	  failwith "merge_var_classes"

      | var :: tail ->
	  let rep =
	    get_rep var
(*	    try
	      Hashtbl.find var_rep (Var.id_of_var var)
	    with
	      | Not_found ->
		  Hashtbl.add var_rep (Var.id_of_var var) var;
		  var*)
	  in
	    List.iter
	      (fun var ->
		 let cur_rep =
		   get_rep var
(*		   try
		     Hashtbl.find var_rep (Var.id_of_var var)
		   with
		     | Not_found ->
			 var*)
		 in
		   if not (Var.equal cur_rep rep) then
		     Hashtbl.add var_rep (Var.id_of_var cur_rep) rep
	      )
	      tail
  in

    (* put all terms in different equivalence classes,
       i.e. associate them with representatives *)
    List.iter
      (fun literal ->
	 if LiteralTable.mem ground_literals literal then begin
	   ()
	 end

	 else if LiteralTable.mem literal_rep literal then begin
	   ()
	 end

	 else begin
	   let vars =
	     Term.vars_of_literal literal
	   in
	     match vars with
	       | [] ->
		   if LiteralTable.length ground_literals >= max_lemma_length then begin
		     raise MAX_REGRESSED
		   end;

		   LiteralTable.add ground_literals literal ()
		     
	       | var :: _ ->
		   merge_var_classes vars;
		   LiteralTable.add literal_rep literal var
	 end
      )
      literals;

    (* create clauses corresponding to equivalence classes *)
    let rep_literals =
      Hashtbl.create 128
    in
      LiteralTable.iter
	(fun literal rep ->
	   let real_rep =
	     get_rep rep
	       (*
	     try
	       Hashtbl.find var_rep (Var.id_of_var rep)
	     with
	       | Not_found ->
		   failwith "partition_variants"
	       *)
	   in
	   let current =
	     try
	       Hashtbl.find rep_literals (Var.id_of_var real_rep)
	     with
	       | Not_found ->
		   []
	   in
	     if List.length current >= 2 * max_lemma_length then begin
	       raise MAX_REGRESSED
	     end;

	     Hashtbl.replace rep_literals (Var.id_of_var real_rep) (literal :: current)
	)
	literal_rep;


(*
      (* x with subst and offset 0 equal to y *)
      let literal_match_equal_with_subst (x: literal) subst y =
	let rec do' x' y' =
	  Term.term_equal x' y'
	  ||
	  match x', y' with
	    | Term.Var v0, _ ->
		let binding =
		  Subst.get_ subst v0 0
		in
		  begin
		    match binding with
		      | None ->
			  (* false *)
			  Term.term_equal x' y'
			    
		      | Some term ->
			  Term.term_equal term.Subst.st_term y'
		  end

	    | Term.Const c0, Term.Const c1 ->
		Symbol.equal c0 c1

	    | Term.Func (f0, t0, _), Term.Func (f1, t1, _) ->
		Symbol.equal f0 f1
		&&
		  Tools.array_for_all2
		  do'
		  t0
		  t1

	    | _ ->
		false
		
	in
	  x.Term.sign == y.Term.sign
	  &&
	  do' x.Term.term y.Term.term
      in

      let subsumes_greater_smaller x y : bool =
	let rec do' x' y' subst =
	  match y' with
	    | [] ->
		if
		List.for_all
		  (fun literal ->
		     let a =
		       List.exists
			 (literal_match_equal_with_subst literal subst)
			 y
		     in
		       
		     let instance =
		       Subst.apply_to_literal
			 subst
			 literal
			 0
		     in
		     let b =
			 List.exists
			   (Term.literal_equal instance)
			   y
		     in
		       if a != b then begin
			 print_endline ("A: " ^ string_of_bool a);
			 print_endline ("B: " ^ string_of_bool b);
			 print_endline ("OHO" ^ Term.literal_to_string literal);
		       end;
		       b

		  )
		  x'
		       then
			 raise Exit

	    | smaller :: tail ->
(*		print_endline ("PICK: " ^ Term.literal_to_string smaller);*)
		begin
		  List.iter
		    (fun greater ->
(*		       print_endline ("TRY: " ^ Term.literal_to_string greater);*)
		       if Term.are_literals_variants2 smaller greater then begin
			 try
			   let new_subst =
			     Unification.match_literals
			       greater 0
			       smaller 1
			    in
			    let merged_subst =
			      Unification.match_substs
				0
				subst
				new_subst
			    in
			      (* :TODO: prune x' *)
			      do' x' tail merged_subst
			  with
			    | Unification.UNIFICATION_FAIL ->
				()
		       
		       end
		    )
		    x'

		end
		
	in
	  try
	    do' x y Subst.empty;
(*		print_endline ("NO SUBSUME");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
	    false
	  with
	    | Exit ->
(*		print_endline ("SUBSUME");
		print_endline (Term.clause_to_string (Term.sort_clause x));
		print_endline (Term.clause_to_string (Term.sort_clause y));
		print_newline ();*)
		true
      in
*)



      let clauses =
	Hashtbl.fold
	  (fun _ clause clauses ->
	     let condensed =
	       (condense clause)
	     in
(*	       print_endline ("CLAUSE: " ^ Term.clause_to_string clause);
	       print_endline ("CONDEN: " ^ Term.clause_to_string condensed);*)

	     condensed :: clauses
	  )
	  rep_literals
	  []
      in
      let clauses =
	List.sort
	  (fun x y ->
	     Tools.compare_int (List.length x) (List.length y)
	  )
	  clauses
      in


      let length, clauses =
	List.fold_left
	  (fun ((length, clauses) as acc) clause ->
(*		print_endline ("NEW: " ^ Term.clause_to_string (Term.sort_clause clause));*)
	     let clause =
	       condense clause
	     in
	     if
	       List.exists
		 (fun old_clause ->
		    are_clauses_variants clause old_clause
		 )
		 clauses
	     then
	       acc

	     else begin
	       let new_length =
		 length + List.length clause
	       in
		 if new_length > 2 * max_lemma_length then begin
		   raise MAX_REGRESSED
		 end;

		 (new_length, clause :: clauses)
	     end
	  )
	  (0, [])
	  clauses
      in


(*
      let _, clause =
	LiteralTable.fold
	  (fun ground_literal _ ((length, clause) as acc) ->
(*	     if
	       List.exists
		 (fun literal ->
		    try
		      let x : Subst.subst =
			Unification.match_literals
			  literal 0
			  ground_literal 1
		      in
			true
		    with
		      | Unification.UNIFICATION_FAIL ->
			  false
		 )
		 clause
	     then
	       acc
		 
	     else*)
	       (length + 1, ground_literal :: clause)
	  )
	  ground_literals
	  (length, List.flatten clauses)
      in
*)


      let ground_literals =
	LiteralTable.fold
	  (fun ground_literal _ acc ->
	       ground_literal :: acc
	  )
	  ground_literals
	  []
      in

	
      let clause =
	List.fold_left
	  (fun acc clause ->
	     List.fold_left
	       (fun acc literal ->
		  literal :: acc
	       )
	       acc
	       clause
	  )
	  ground_literals
	  clauses
      in

      let condensed =
	condense clause
      in
(*	print_endline ("\nFINAL");
	print_endline ("Clause: " ^ Term.clause_to_string clause);
	print_endline ("Conden: " ^ Term.clause_to_string condensed);*)
	condensed





let lemma_counter = Counter.create_with 0;;

let get_lemma (state: state) (clause: clause) (clause_value: int ref) (conflict_literals: literal array) (regress_stop: choice_point) : clause option =
(*  print_endline "get_lemma";*)
  if 
(*    (Counter.next lemma_counter <= 90)
   &&*)
      (Config.lemma state.st_config)
  then begin
    try
(*      print_endline ("REGRESS STOP: " ^ choice_point_to_string regress_stop);*)
    if trace then begin
      print_endline ("CLOSING: " ^ string_of_int (Counter.value lemma_counter));
      print_endline (Term.clause_to_string clause);
      print_endline (Term.clause_to_string (Array.to_list conflict_literals));
      print_endline ("REGRESS STOP: " ^ choice_point_to_string regress_stop);
    end;
    let closing_explanation =
      get_explanation state conflict_literals
    in
    let closing_node = {
      bn_added_literal = Term.null_literal;
      bn_added_cp = active_choice_point state;
      bn_clause = clause;
      bn_clause_index = -1;
      bn_context_literals = conflict_literals;
      bn_explanation = Some closing_explanation;
      bn_clause_value = clause_value;
    }
    in
      
    let offset =
      ref 1
    in

    (* regress *)
    let to_regress = {
      tr_clause_literal = Term.null_literal;
      tr_clause_offset = -1;
      tr_context_literal_id = -1;
      tr_branch_node = closing_node;
      tr_branch_offset = 0;
    }
    in
    let regressed_literals, constraints =
      regress
	state
	regress_stop
	[]
	0
	[]
	(To_regress.add to_regress To_regress.empty)
	offset
    in
(*      print_endline "get_lemma 3";*)

    let unifier =
      (*
      let rec please_profile acc constraints' =
	match constraints' with
	  | [] ->
	      acc

	  | ((l1, o1), (l2, o2)) :: tail ->
	   let unifier =
	     Unification.unify_terms
	       l1.Term.term
	       o1
	       l2.Term.term
	       o2
	   in
	   let acc' =
	     Unification.unify_substs
	       acc
	       unifier
	   in
	     please_profile acc' tail
      in
	please_profile Subst.empty constraints
      *)
      List.fold_left
	(fun acc ((l1, o1), (l2, o2)) ->
	   let unifier =
	     Unification.unify_terms
	       l1.Term.term
	       o1
	       l2.Term.term
	       o2
	   in
	     Unification.unify_substs
	       acc
	       unifier	       
	)
	Subst.empty
	constraints
    in
  (*  print_endline "get_lemma 4";*)

    let regressed =
      Subst.apply_to_literals
	unifier
	regressed_literals
    in
(*  print_endline "get_lemma 5";*)
(*	print_endline ("REGRESSED: " ^ Term.clause_to_string regressed);*)
      
    (* remove variants *)
(*
    let variant_free =
(*      filter_variants state regressed*)
      filter_variants state regressed
    in
      let variant_free =
	Term.sort_clause variant_free
      in
*)
    let x' =
      Term.sort_clause (filter_variants3 state regressed)
    in
(*
      print_endline ("OLD: " ^ Term.clause_to_string variant_free);
      print_endline ("NEW: " ^ Term.clause_to_string x');
      print_newline ();
*)    
      let variant_free = x' in

(*
    let variant_free2 =
      filter_variants2 state regressed
    in
      if not (Term.clause_equal variant_free variant_free2) then begin
(*	print_endline ("New: " ^ Term.clause_to_string (Term.sort_clause variant_free));
	print_endline ("Old: " ^ Term.clause_to_string (Term.sort_clause variant_free2));*)
	print_endline ("New: ");
	List.iter (fun x -> print_endline (Term.literal_to_string x)) (Term.sort_clause variant_free);
	print_endline ("Old: ");
	List.iter (fun x -> print_endline (Term.literal_to_string x)) (Term.sort_clause variant_free2);
	failwith "variant_free";
      end;
*)
      (*      	print_endline ("SHORTENED: " ^ Term.clause_to_string variant_free);*)
      (*  print_endline "get_lemma 6";*)
      (*
      let variant_free2 =
	Term.sort_clause variant_free
      in

	if (List.length variant_free2) != (List.length variant_free) then begin
	  failwith "HEHE0";
	end;*)
(*	
	print_endline "get_lemma 7";
	print_endline (Term.clause_to_string variant_free);
	print_endline (Term.clause_to_string variant_free2);
	if List.length variant_free2 > max_lemma_length then begin
	  print_endline "get_lemma 7b";

(*	  None*)
	end;*)
	if List.length variant_free > max_lemma_length then begin
(*	  print_endline "get_lemma 7a";*)

	  None
	end
	else begin

	(* normalize and replace parameters by universal variables *)
      let universal =
	Term.request_universal_clause variant_free
      in
	(*    None*)
	(*
	if List.length universal != List.length variant_free2 then begin
	  failwith "HEHE";
	end;*)

	  if trace || true then begin
	    (*
	      List.iter
	      (fun literal ->
	      print_endline (Term.literal_to_string literal)
	      )
	      universal;
	      print_newline ();*)
	    (*      
		    print_endline ("LEMMA: " ^ Term.clause_to_string universal);
		    print_newline ();*)
	    
	  end;

	(* normalize if its a unit lemma.
	   literals of clauses are automatically normalized when
	   chosen for assert or split by application of the context unifier,
	   unless they are unit, as then no context unifier is computed
	*)
	let universal =
	  match universal with
	    | [ literal ] ->
		[ Subst.apply_to_literal Subst.empty literal 0 ]

	    | _ ->
		universal
	in
	  Some universal;
	  (*None;*)
	end


    (*
    (* negate *)
      let negated =
      List.map
      Term.request_negated_literal
      universal
      in
      let negated =
      Term.order_clause negated
      in
      List.iter
      (fun literal ->
      print_endline (Term.literal_to_string literal)
      )
      universal;
      print_newline ();
    (*	print_endline ("LEMMA: " ^ Term.clause_to_string negated);
      print_newline ();*)

      Some negated
    *)
    with
      | MAX_REGRESSED -> 
(*	  print_endline "get_lemma 8a";*)
	  None
  end
    
  else
    None
    
(*
let get_best_backtrack_choice_point (state: state) (explanation: choice_point list) (prune: prune)
    : choice_point =

  try
    let selected =
      List.find
	(fun choice_point ->
	   choice_point != state.st_root_choice_point
			   &&
			   choice_point.cp_valid
			   &&
			   choice_point.cp_left_split
	)
	explanation
    in
  (* hack:
     the last choice point of the conflict set must always be retracted.
     there might be several closing context unifiers which are completed by the last choice point.
     retracting an arbitraty choice point of the conflict set of only on chosen conflict set
     may leave the other closing context unifiers intact.
     alas, they are not recomputed.
     for that matter, currently Selection eagerly throws CLOSE when finding the first context unifier,
     so the choice point's candidate's are not even fully computed. *)

    begin
      match explanation with
	| head :: _ ->
	    invalidate_choice_point state head prune
	    

	| _ -> ()
    end;


selected
  with
    | Not_found ->

	if prune#is_prune then begin
	  List.iter
	    (fun choice_point ->
	       if choice_point.cp_left_split then begin
		 if is_choice_point_valid choice_point then
		   prune#prune choice_point
		 else
		   failwith "State.unsatisfiable"
	       end
	    )
	    state.st_branch
	end;

	raise UNSATISFIABLE

let get_best_backtrack_choice_point_naive (state: state) (prune: prune) : choice_point =
  get_best_backtrack_choice_point state state.st_branch prune

*)



let backtrack (state: state) (clause: clause) (clause_value: int ref) (context_literals: literal array) (prune: prune) :
    choice_point * choice_point * choice_point list * clause option
    =
  let closing_explanation =
    match Config.backtracking state.st_config with
      | Flags.BT_Naive ->
	  state.st_branch
	  
      | Flags.BT_Backjumping ->
	  get_explanation state context_literals
  in
(*    print_endline ("CLOSING: " ^ explanation_to_string closing_explanation);*)
  let retract, backjump, right_explanation =
    match closing_explanation with
      | [] ->
	  raise (UNSATISFIABLE (clause, context_literals))
	  
      | root :: [] ->
	  if choice_point_equal state.st_root_choice_point root then
	    raise (UNSATISFIABLE (clause, context_literals))
	  else
	    root, state.st_root_choice_point, []

      | retract :: backjump :: tail ->
	  retract, backjump, (backjump :: tail)
  in
(*
  let backjump =
    let rec do' branch =
      match branch with
	| [] ->
	    failwith "do backjump"

	| _ :: [] ->
	    failwith "do backjump 0"

	| x :: y :: tail ->
	    if choice_point_equal y backjump then
	      x
	    else
	      do' (y :: tail)
    in
      do' state.st_branch
  in
*)
  let backjump = retract in

(*    print_endline (choice_point_to_string retract);
    print_endline (choice_point_to_string backjump);*)
    (*
  let retract, right_explanation =
    let rec do' explanation =
      match explanation with
	| [] ->
	    raise (UNSATISFIABLE (clause, context_literals))

	| head :: tail ->
	    if head.cp_left_split then
	      head, tail
	    else
	      do' tail
    in
      do' closing_explanation

  in
    *)
  let lemma =
    get_lemma state clause clause_value context_literals retract
  in

    invalidate state (*retract*) backjump prune;

    (* remove retracted context literals. *)
    backtrack_remove_literal_to_branch_nodes state;

    retract, backjump, right_explanation, lemma



(*
let backtrack_cleanup (state: state)
    (backtrack_choice_point: choice_point) (oldest_invalidated: choice_point)
    (closing_explanation: explanation) (prune: prune)
    : choice_point * choice_point * literal =


  (* remove everything which has now been invalidated. *)
  state.st_branch <- prune_explanation state oldest_invalidated state.st_branch;

  (* warning: for pruning this must only be done after all choice point invalidation
     has been done, which includes backtrack_branch. *)
  state.st_closing_explanation <- prune_explanation state backtrack_choice_point state.st_closing_explanation;

  (* prune invalid choice points from dependencies. *)
  List.iter
    (fun choice_point ->
       choice_point.cp_base_for <- prune_explanation state backtrack_choice_point choice_point.cp_base_for;
       choice_point.cp_dependent_on <- prune_explanation state backtrack_choice_point choice_point.cp_dependent_on;
    )
    state.st_branch;

  (* remove retracted context literals. *)
  backtrack_remove_literal_to_choice_points state;
  
  (* create the right split based on the left split. *)  
  
  (* dependency is dependency of the left split plus the conflict set *)
  let dependency =
    merge_explanations
      (prune_explanation state backtrack_choice_point backtrack_choice_point.cp_dependent_on)
      (prune_explanation state backtrack_choice_point closing_explanation)
  in
  let skolemized_literal = 
    Term.request_negated_literal
      state.st_database
      (Term.request_skolemized_literal state.st_database backtrack_choice_point.cp_split_literal)
  in
  let right_split =
    create_choice_point ~left_split:false state skolemized_literal dependency
  in
    (* and return the toggled left split.
       we need to return the skolemized literal as well,
       as skolemizing the left split literal again yields
       a different skolemized literal. *)
    backtrack_choice_point, oldest_invalidated, skolemized_literal
*)
(*
let backtrack (state: state) (conflict_set: literal list) (prune: prune)
    : choice_point * choice_point * literal =

  let closing_explanation =
    get_explanation state conflict_set
  in

    state.st_closing_explanation <-
      merge_explanations state.st_closing_explanation closing_explanation;

  let backtrack_choice_point =
    match Config.backtracking state.st_config with
      | Flags.BT_Naive ->
	  get_best_backtrack_choice_point_naive state prune

      | Flags.BT_Backjumping ->
	  get_best_backtrack_choice_point state state.st_closing_explanation prune
    
      | Flags.BT_Dynamic ->
	  get_best_backtrack_choice_point state closing_explanation prune
  in
    invalidate_choice_point state backtrack_choice_point prune;

    backtrack_cleanup state backtrack_choice_point backtrack_choice_point
      closing_explanation prune
*)






(* no pruning in an exhausted branch, so use prune_dummy. *)
let backtrack_incomplete (state: state) (incomplete_choice_points: choice_point list)
     : choice_point =

  let retract =
    match incomplete_choice_points with
      | [] ->
	  failwith "State.backtrack_incomplete"

      | head :: _ ->
	  if choice_point_equal state.st_root_choice_point head then
	    raise (UNSATISFIABLE ([], [| |]))
	  else
	    head
  in
(*

  (*
  List.iter (fun x -> print_endline ("INV: " ^ choice_point_to_string_extended x)) incomplete_choice_points;

  List.iter (invalidate_choice_point state) incomplete_choice_points;
  *)
  (* get the oldest incomplete choice point *)
  let rec loop choice_points =
    match choice_points with
      | [] ->
	  failwith "State.backtrack_incomplete"
	  
      | head :: [] ->
	  (*invalidate_choice_point state head prune_dummy;*)
	  (* even the root node incomplete? *)
	  if choice_point_equal state.st_root_choice_point head then
	    raise (UNSATISFIABLE ([], [| |]))

	  (* oldest incomplete choice point found. *)
	  else begin
	    (* it's a left split, backtrack to it. *)
	    if head.cp_left_split then
	      head

	    (* it's a right split, so find another left split to backtrack to. *)
	    else begin
	      failwith "backtrack_incomplete right split"
	      (* :TODO: Loop MGT030-1 *)
	      (*
	      invalidate_subbranch state head prune_dummy;
	      let first_left_split =
		get_best_backtrack_choice_point_naive state prune_dummy
	      in
	      (*print_endline ("LS: " ^ choice_point_to_string first_left_split);*)
		invalidate_choice_point state first_left_split prune_dummy;
		(* prune branch up to the oldest of the retracted choice points. *)
		if compare_age head first_left_split <= 0 then
		  first_left_split, head
		else
		  first_left_split, first_left_split
	      *)
	      let rec find_previous_left_split cps =
		match cps with
		  | [] ->
		      raise (UNSATISFIABLE ([], [| |]))

		  | head' :: tail ->
		      if choice_point_equal head head' then
			tail

		      else
			find_previous_left_split tail
			
	      in
	      let tail =
		find_previous_left_split state.st_branch
	      in
		try
		  List.find
		    (fun cp ->
		       cp.cp_left_split
		    )
		    tail
		with
		  | Not_found ->
		      raise (UNSATISFIABLE ([], [| |]))
	    end
	  end

      | _ :: tail ->
	  (*invalidate_choice_point state head prune_dummy;*)
	  loop tail
  in
    (*
  let backtrack_choice_point, oldest_incomplete_choice_point =
    if Config.restart state.st_config = Flags.RS_Exhaustive then begin
      let first_left_split =
	get_best_backtrack_choice_point_naive state prune_dummy
      in
	invalidate_choice_point state first_left_split prune_dummy;
	first_left_split, first_left_split
    end

    else begin
      loop incomplete_choice_points
    end
  in
    backtrack_cleanup state backtrack_choice_point oldest_incomplete_choice_point
      incomplete_choice_points prune_dummy
    *)

  let retract =
    loop incomplete_choice_points
  in
*)
    invalidate state retract prune_dummy;

    (* remove retracted context literals. *)
    backtrack_remove_literal_to_branch_nodes state;

    retract


let random_jump state jump_target : choice_point list =
  invalidate state jump_target prune_dummy;
  backtrack_remove_literal_to_branch_nodes state;
  state.st_branch
