(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** choice point management

  A {e choice point} represents/manages a Split decision (plus subsequent Asserts)
  and the dependencies (explanation) of this decision on other choice points
  (see the handbook for a more thorough definition).

  The choice points are ordered in a sequence
  which represents the current derivation branch.
  The first choice point of the sequence (the {e root choice point})
  is not really a choice point,
  as it is not created by a split decision,
  and merely extended by the assertion of the unit clauses of the input clause set.
  Splits create new choice points,
  Asserts work on the current choice point,
  i.e. the most recently created choice point (the {e active choice point}).

  A choice point is {e older} than another choice point
  if it comes earlier in the branch,
  that is the root choice point is the oldest one.

  If a choice point becomes invalid (because of backtracking),
  all older choice points, and all dependent context literals and remainders
  become also invalid (transitive closure).
*)



(** {6 Types} *)

type config = Config.config
type literal = Term.literal
type clause = Term.clause

(** see module description. *)
type choice_point

(** manages a derivation branch, i.e. a sequence of choice points,
    and their dependencies. *)
type state



(** notification of choice points irrelevant for the derivation (prunable),
    i.e. retracted left splits not involved in closing any branch.
    That is, by removingthis choice point the derivation
    remains untouched otherwise.
*)
class type prune =
object
  (** should prunable choice points be reported? *)
  method is_prune: bool

  (** notifiy of a pruned choice point. *)
  method prune: choice_point -> unit
end

exception UNSATISFIABLE of clause * literal array



(** {6 Functions} *)


(** {2 Constants } *)

(** a valid choice_point. *)
val valid_choice_point: choice_point

(** an invalid choice point. *)
val invalid_choice_point: choice_point



(** {2 Creation} *)

(** this basically creates a new derivation branch. *)
val create: config -> state

(** [split literal context_literals]
  creates a new choice point initiated by a left split on the new context literal [literal]
  with its context unifier's [context_literals]. *)
val apply_split_left: state -> literal -> clause -> int -> choice_point

val apply_split_right: state -> literal -> clause -> int ref -> int -> choice_point list -> unit

(** [assert_literal literal context_literals]
    extends the current choice point with the new context [literal]
    and the dependencies given by its context unifier's context literals. *)
val apply_assert: state -> literal -> clause -> int ref -> int -> literal array -> unit

val set_context_id: state -> literal -> int -> unit


(** {2 Choice Point Access } *)

(** returns the root choice point. *)
val root_choice_point: state -> choice_point

(** returns the active choice point. *)
val active_choice_point: state -> choice_point






(** {2 Properties } *)

(** the id of the choice point as used by {!State.compare_age} *)
val id_of_choice_point: choice_point -> int

(** is this choice point valid? *)
val is_choice_point_valid: choice_point -> bool

(** is this choice point invalid? *)
val is_choice_point_invalid: choice_point -> bool

(** [depends_on config dependent base].
    does [dependent] depend on [base]?
*)
val depends_on: config -> choice_point -> choice_point -> bool

val is_right_split: choice_point -> bool

val choice_point_literal: choice_point -> literal
(* in reverse order of application *)
val choice_point_right_literals: choice_point -> literal list

val get_branch: state -> choice_point list

(** {2 Comparison} *)

val choice_point_equal: choice_point -> choice_point -> bool

(** compares the age of two choice points.
  returns
    - -1 if the first choice point is older,
    - 1 if the second is older,
    - 0 if the choice points are identical *)
val compare_age: choice_point -> choice_point -> int





(** {2 Backtracking} *)


(** [prune state func apply oldest_retracted values]
    returns [values] without those values
    for which [func] returns an invalid choice point.
    [apply] is applied to the removed values.
    [oldest_rectracted] is the oldest choice point
    retracted since the last pruning was done.
*)
(*val prune: state -> ('a -> choice_point) -> ('a -> unit) -> choice_point ->
  'a list -> 'a list*)

(** [backtrack state conflict_set prune]
    backtracks to a left split choice point
    according to the chosen backtracking method ({!Flags.opt_backtracking}),
    the conflict set (context literals of the closing context unifier),
    and previous backtracking,
    and returns it along with the oldest rectracted choice point
    and the right split (skolemized) literal.
    This choice point and the choice points dependent on it
    are retracted according to the chosen backtracking method.

    @raise UNSATISFIABLE if no left split choice point to backtrack to is left. *)
val backtrack: state -> clause -> int ref -> literal array -> prune -> choice_point * choice_point * choice_point list * clause option

(** to be used when the derivation found an incomplete branch,
    i.e. a branch which is exhausted only under the current deepening bound.
    Behaves similar to {!State.backtrack},
    but instead of the conflict set the list of choice points
    in which candidates have been ignored due to the depth bound
    has to be given.

    The choice point backtracked to depends on the chosen restarting strategy
    ({!Flags.opt_restart}).
*)
val backtrack_incomplete: state -> choice_point list -> choice_point

val random_jump: state -> choice_point -> choice_point list



(** {2 Representation} *)

(** string representation of a choice point, e.g. [0]. *)
val choice_point_to_string: choice_point -> string

(** if used with the id of a choice point this give the same representation
    as using {!State.choice_point_to_string} directly. *)
val id_to_string: int -> string

(** string representation of the active choice point.
    Shortcut for {!State.choice_point_to_string} {!State.active_choice_point}. *)
val active_choice_point_to_string: state -> string

(** extended string representation of the derivation branch,
    including validity and dependency information. *)
val branch_to_string: state -> string

val get_explanation: state -> literal array -> choice_point list

val condense: clause -> clause
