(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(* the term database consists basically of hash tables containing the terms.

   the hash values are for a
   - Var: its var's name hashed
   - Const: its symbol's id
   - Func: its hash_value
   - literal: its enclosed term's hash_value

   terms are simply stored in a weak set.
   as long as a term is referenced outside of the set
   it is kept alive and subsequent creation requests return the same term.
   as soon as the term is not referenced outside of the set
   it may be automatically garbage collected.
*)


(*** types ***)

type var = Var.var
type symbol = Symbol.symbol


type term =
  | Var of var
  | Const of symbol
  | Func of symbol * term array * int

type literal = {
  sign: bool;
  term: term;
}

type clause =
    literal list




(*** hashing ***)

(* the hash value of a term. *)
let hash_of_term (term: term) : int =
  match term with
    | Var var ->
	Var.id_of_var var
	  
    | Const symbol ->
	Symbol.id symbol
	  
    | Func (_, _, hash_value) ->
	hash_value


(* computation of the hash value of a Func term.

   131 is known to work quite well by ... someone Cesare Tinelli knows.
   tests seem to support it. *)
let compute_hash (predicate_symbol: symbol) (arguments: term array) : int =

  let compute_hash' (old_value: int) (new_value: int) =
    (old_value + new_value) * 131
  in

    Array.fold_left 
      (fun old_hash term ->
	 compute_hash' old_hash (hash_of_term term)
      )
      (compute_hash' 0 (Symbol.id predicate_symbol))
      arguments







(*** database types ***)


(* term database *)
module TermTableBase =
  Weak.Make (
    struct
      type t = term

      let rec are_terms_equal_at (index: int) (terms1: term array) (terms2: term array) : bool =
	if index >= Array.length terms1 then
	  true
	else begin
	  (* subterms must have pointer equality
	     as a term is constructed bottom up with this database *)
	  if terms1.(index) == terms2.(index) then
	    are_terms_equal_at (index + 1) terms1 terms2
	  else
	    false
	end


      (* term_equal can not be used, as it uses object identity 
	 which is assured exactly by this module *)
      let equal (term1: term) (term2: term) : bool =
	match term1, term2 with
	  | Var var1, Var var2 ->
	      Var.equal var1 var2

	  | Const const1, Const const2 ->
	      Symbol.equal const1 const2

	  | Func (symbol1, terms1, _), Func (symbol2, terms2, _) ->
	      (Symbol.equal symbol1 symbol2)
	      &&
	      (are_terms_equal_at 0 terms1 terms2)

	  | _ ->
	      false

      let hash (term: term) : int =
	hash_of_term term
    end
  )


(* literal database *)
module LiteralTableBase =
  Weak.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.sign == literal2.sign)
	&&
	(literal1.term == literal2.term)

      let hash (literal: literal) : int =
	hash_of_term literal.term

    end
  )




type database = {
  (* environment *)
  mutable plus_v: bool;

  (* contans the terms: consts, vars, functions *)
  terms: TermTableBase.t;
  (* contains the literals *)
  literals: LiteralTableBase.t;
}






(*** globals ***)

(* this global database contains the global terms *)
let global: database =
  {
    plus_v = false;
    terms = TermTableBase.create 2048;
    literals = LiteralTableBase.create 512;
  }






(*** string representation ***)

let rec term_to_string (term: term) : string =
  match term with
    | Var var ->
	Var.to_string var

    | Const const ->
	Symbol.to_string const

    | Func (symbol, terms, _) ->
	Symbol.to_string symbol
	^ "("
	^ String.concat ", " (List.map term_to_string (Array.to_list terms))
	^ ")"



let literal_to_string (literal: literal) : string =
  let prefix =
    if literal.sign then
      "+" (* "" *)
    else
      "-" (* "~" *)
  in
  let term =
    term_to_string literal.term
  in
    prefix ^ term



let clause_to_string (clause: clause) : string =
  "[" ^ String.concat ", " (List.map literal_to_string clause) ^ "]"





(*** comparison ***)

(* ensure that equal terms have equal pointers *)
let rec debug_term_equal (term1: term) (term2: term) : bool =
  match term1, term2 with
    | Var v1, Var v2 ->
	Var.equal v1 v2

    | Const s1, Const s2 ->
	Symbol.equal s1 s2

    | Func (s1, t1, _), Func (s2, t2, _) ->
	Symbol.equal s1 s2
	&&
	Tools.array_fold2
	  (fun acc term1 term2 ->
	     acc
	     &&
	     debug_term_equal term1 term2
	  )
	  true
	  t1
	  t2

    | _ ->
	false

let rec term_equal (term1: term) (term2: term) : bool =
  (* debug mode: ensure that term equality and pointer equality are the same *)
  if
    Const.debug
    &&
    not (
      (debug_term_equal term1 term2)
      =
      (term1 == term2)
    )
  then begin
    failwith ("Term.term_equal: == is " ^ string_of_bool (term1 == term2) ^ " - " ^ term_to_string term1 ^ " - " ^ term_to_string term2);
  end;

  term1 == term2


let literal_equal (literal1: literal) (literal2: literal) : bool =
  (* debug mode: ensure that literal equality and pointer equality are the same *)
  if
    Const.debug
    &&
    not (
      (
	(literal1.sign == literal2.sign)
	&&
	(debug_term_equal literal1.term literal2.term)
      )
      =
      (literal1 == literal2)
    )
  then begin
    failwith "Term.literal_equal";
  end;

  literal1 == literal2

let clause_equal (clause1: clause) (clause2: clause) : bool =
  Tools.lists_unordered_equal literal_equal clause1 clause2





(*** Hash tables ***)


(* now export tables for further use.
   here equality can be checked by pointer identity. *)

(* access by literal *)
module LiteralTable =
  Hashtbl.Make (
    struct
      type t = literal
	  
      let equal (literal1: literal) (literal2: literal) : bool =
	literal_equal literal1 literal2
	  
      let hash (literal: literal) : int =
	hash_of_term literal.term
		
    end
  )


(* access by literal type, i.e. sign and predicate symbol
   but not the concrete literal. *)
module LiteralTypeTable =
  Hashtbl.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.sign == literal2.sign)
	&&
	match literal1.term, literal2.term with
	  | Func (symbol1, _, _), Func (symbol2, _, _) ->
	      (* same predicate symbol *)
	      Symbol.equal symbol1 symbol2
		
	  | _ ->
	      (* or for var/const same term *)
	      literal1.term == literal2.term

      let hash (literal: literal) : int =
	match literal.term with
	  | Const symbol
	  | Func (symbol, _, _) ->
	      Symbol.id symbol

	  | Var var ->
	      Var.id_of_var var
    end
  )








(*** creation ***)



(* these functions try to find a requested term in the database
   and to return the old instance.

   otherwise a fresh instance is created, stored, and returned. *)


let request_term (term: term) : term =
  try
    TermTableBase.find global.terms term
  with
    | Not_found ->
	TermTableBase.add global.terms term;
 	term


let request_var (var: var) : term =
  let new_term =
    Var var
  in
    request_term new_term

let request_const (symbol: symbol) : term =
  let new_term =
    Const symbol
  in
    request_term new_term


let request_func ((symbol, terms): symbol * term array) : term =
  let new_term =
    Func (symbol, terms, compute_hash symbol terms)
  in
    request_term new_term


let request_literal (sign: bool) (term: term) : literal =
  let new_literal = {
    sign = sign;
    term = term;
  }
  in
    try
      LiteralTableBase.find global.literals new_literal
    with
      | Not_found ->
	  LiteralTableBase.add global.literals new_literal;
 	  new_literal


let request_negated_literal (literal: literal) : literal =
  request_literal (not literal.sign) literal.term


(* does also normalize the remaining variables,
   i.e. removes 'gaps' created by skolemizing some variables *)
let request_skolemized_term (term: term) : term =
  let (var_normalization: (var * term) list ref) =
    ref []
  in
  let (var_skolemization: (var * term) list ref) =
    ref []
  in

  let rec do_skolemize (term: term) : term=
    match term with
      | Var var ->
	  if Var.is_universal var then begin
	    try
	      List.assoc var !var_skolemization
	    with
	      | Not_found ->
		  let fresh_skolem =
		    Symbol.create_skolem ()
		  in
		  let fresh_const =
		    request_const fresh_skolem
		  in
		    var_skolemization := (var, fresh_const) :: !var_skolemization;
		    fresh_const
	  end

	  else begin
	    try
	      List.assoc var !var_normalization
	    with
	      | Not_found ->
		  let fresh_var =
		    (* the next id can be deduced from the number of remapped vars *)
		    Var.clone_renumbered var (List.length !var_normalization)
		  in
		  let fresh_term =
		    request_var fresh_var
		  in
		    var_normalization := (var, fresh_term) :: !var_normalization;
		    fresh_term
	  end

      | Const _ ->
	  term

      | Func (symbol, terms, _) ->
	  request_func (symbol, Array.map do_skolemize terms)

  in
    do_skolemize term


let request_skolemized_literal (literal: literal) : literal =
  request_literal literal.sign (request_skolemized_term literal.term)



let rec do_request_universal_term (term: term) var_mapping : term =
  match term with
    | Var var ->
	if Var.is_parametric var then begin
	  try
	    List.assoc var !var_mapping
	  with
	    | Not_found ->
		let fresh_var =
		  Var.create_universal (List.length !var_mapping)
		in
		let fresh_term =
		  request_var fresh_var
		in
		  var_mapping := (var, fresh_term) :: !var_mapping;
		  fresh_term
	end
	else
	  term
	    
    | Const _ ->
	term
	
    | Func (symbol, terms, _) ->
	request_func (symbol, Array.map (fun t -> do_request_universal_term t var_mapping) terms)

let request_universal_term (term: term) : term =
  let (var_mapping: (var * term) list ref) =
    ref []
  in
    do_request_universal_term term var_mapping

let request_universal_literal (literal: literal) : literal =
  request_literal literal.sign (request_universal_term literal.term)

let request_universal_clause (clause: clause) : clause =
  let (var_mapping: (var * term) list ref) =
    ref []
  in

  let new_clause =
    List.fold_left
      (fun acc literal ->
	 let new_term =
	   do_request_universal_term literal.term var_mapping
	 in
	 let new_literal =
	   request_literal literal.sign new_term
	 in
	   new_literal :: acc
      )
      []
      clause
  in
    List.rev new_clause





let rec replace_in_term (term: term) (old_term: term) (new_term: term) : term =
  if term_equal term old_term then
    new_term

  else if term_equal old_term new_term then
    term

  else begin
    match term with
      | Var _
      | Const _ ->
	  term
	  
      | Func (symbol, terms, _) ->
	  let new_terms =
	    Array.map
	      (fun term ->
		 replace_in_term term old_term new_term)
	      terms
	  in
	    request_func (symbol, new_terms)
  end







(*** special literals ***)


(*** creation ***)

let true_literal : literal =
  request_literal true (request_const (Symbol.create_symbol "true" 0))

let false_literal : literal =
  request_literal true (request_const (Symbol.create_symbol "false" 0))

let assert_literal : literal =
  request_literal true (request_const (Symbol.create_symbol "__assert__" 0))

let null_term : term =
  request_const (Symbol.create_symbol "__null_term__" 0)

let null_literal : literal =
  request_literal true null_term

let minus_v : literal =
  let term =
    request_func (
      Symbol.create_symbol
		    "__-v__" 1,
		    [| request_var (Var.create_parametric 0) |]
    )
  in
    request_literal false term

let plus_v : literal =
  let term =
    request_func (
      Symbol.create_symbol 
			   "__+v__" 1,
			   [| request_var (Var.create_parametric 0) |]
    )
  in
    request_literal true term






(*** decomposition ***)

let rec do_vars_of_term (acc: var list) (term: term) : var list =
  match term with
    | Var var ->
	(* store each variable only once *)
	if List.exists
	  (fun acc_var ->
	     Var.equal acc_var var
	  )
	  acc
	then
	  acc
	else
	  var :: acc

    | Const _ ->
	acc

    | Func (_, terms, _) ->
	Array.fold_left do_vars_of_term acc terms


let vars_of_term (term: term) : var list =
  do_vars_of_term [] term


let vars_of_literal (literal: literal) : var list =
  vars_of_term literal.term





(*** Variants ***)


let are_terms_variants (_term1: term) (_term2: term) : bool =

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (acc: (var * var) list) (term1: term) (term2: term) :
    (var * var) list =
    
    match term1, term2 with
      | Var var1, Var var2 when
	  Var.is_universal var1 = Var.is_universal var2 ->

	  Tools.mapping_extend acc Var.equal Var.equal var1 var2
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->

	  acc
	      
      | Func (symbol1, terms1, _), Func (symbol2, terms2, _) when
	  Symbol.equal symbol1 symbol2 ->

	  Tools.array_fold2 do_variants acc terms1 terms2
	      
      | _ ->
	  raise Exit
  in	    
    (term_equal _term1 _term2)
    ||
    (
      try
	ignore (do_variants [] _term1 _term2);
	true
      with
	| Exit ->
	    false
    )


let are_literals_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.sign = literal2.sign)
  &&
  (are_terms_variants literal1.term literal2.term)





let are_terms_weak_variants (_term1: term) (_term2: term) : bool =

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (acc: (var * var) list) (term1: term) (term2: term) :
    (var * var) list =
    
    match term1, term2 with
      | Var var1, Var var2 ->
	  Tools.mapping_extend acc Var.equal Var.equal var1 var2
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->

	  acc
	      
      | Func (symbol1, terms1, _), Func (symbol2, terms2, _) when
	  Symbol.equal symbol1 symbol2 ->

	  Tools.array_fold2 do_variants acc terms1 terms2
	      
      | _ ->
	  raise Exit
  in	    
    (term_equal _term1 _term2)
    ||
    (
      try
	ignore (do_variants [] _term1 _term2);
	true
      with
	| Exit ->
	    false
    )


let are_literals_weak_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.sign = literal2.sign)
  &&
  (are_terms_weak_variants literal1.term literal2.term)







let are_terms_skolem_variants (_term1: term) (_term2: term) : bool =

  let skolem_mapping =
    ref []
  in

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (acc: (var * var) list) (term1: term) (term2: term) :
    (var * var) list =
    
    match term1, term2 with
      | Var var1, Var var2 when
	  Var.is_universal var1 = Var.is_universal var2 ->

	  Tools.mapping_extend acc Var.equal Var.equal var1 var2
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->

	  acc
	      
      | Const symbol1, Const symbol2 when
	  Symbol.is_skolem symbol1
	  &&
	  Symbol.is_skolem symbol2
	  ->

	  skolem_mapping := Tools.mapping_extend !skolem_mapping Symbol.equal Symbol.equal symbol1 symbol2;
	  acc

      | Func (symbol1, terms1, _), Func (symbol2, terms2, _) when
	  Symbol.equal symbol1 symbol2 ->

	  Tools.array_fold2 do_variants acc terms1 terms2
	      
      | _ ->
	  raise Exit
  in	    
    (term_equal _term1 _term2)
    ||
    (
      try
	ignore (do_variants [] _term1 _term2);
	true
      with
	| Exit ->
	    List.iter
	      (fun (x, y) ->
		 print_endline (Symbol.to_string x ^ " - " ^ Symbol.to_string y)
	      )
	      !skolem_mapping;
	    false
    )


let are_literals_skolem_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.sign = literal2.sign)
  &&
  (are_terms_skolem_variants literal1.term literal2.term)





(*** Ordering ***)

let rec compare_terms x y =
  match x, y with
    (* var first *)
    | Var u, Var v ->
	0
(*	compare (Var.id_of_var u) (Var.id_of_var v)*)

    | Var _, _ ->
	-1

    | _, Var _ ->
	1

    (* const second *)
    | Const a, Const b ->
	Tools.compare_int (Symbol.id a) (Symbol.id b)

    | Const _, _ ->
	-1

    | _, Const _ ->
	1

    (* term last *)
    | Func (f, t0, _), Func (g, t1, _) ->
	if Symbol.equal f g then begin
	  Tools.array_fold2
	    (fun acc x y ->
	       if acc <> 0 then
		 acc
	       else
		 compare_terms x y
	    )
	    0
	    t0
	    t1
	end

	else begin
	  Tools.compare_int (Symbol.id f) (Symbol.id g)
	end
	

(* pos before neg *)
let compare_literals x y =
  if x.sign = y.sign then
    compare_terms x.term y.term
  else if x.sign then
    -1
  else
    1

let sort_clause clause =
  List.sort
    compare_literals
    clause
