(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** terms, literals, clauses

  All terms are stored in a global database
  so that each term exists only once in the system (hash-consing).

  Therefore terms and literals must never be explicitly built using the constructors.
  Instead they must be build by using the appropriate {b request_xxx} function below.

  Fast access is based on a hash value for every term.
  For variables and constants that's their id,
  for terms it is (for efficiency) stored in the term data structure.

  Benefits: low memory usage by term sharing, fast term comparison by pointer equality.

  Drawbacks: term construction implies a term database access, terms store a hash value.
*)


(** {6 Types} *)

type var = Var.var
type symbol = Symbol.symbol


(** a term can be a variable, a constant or a function. *)
type term =
  | Var of var
  | Const of symbol
  | Func of symbol * term array * int
      (** symbol * arguments * hash_value  *)

(** a literal is a signed term - the term is interpreted as a predicate. *)
type literal = {
  sign: bool;
  (** the sign *)

  term: term;
  (** the atom *)
}

(** a clause: a list, not a set as in theory. *)
type clause =
    literal list


(** a specialized [Hashtbl] with literals as keys. *)
module LiteralTable : Hashtbl.S with type key = literal

(** a specialized [Hashtbl] with literal types as keys,
  that is not the concrete literal but it's sign and predicate symbol. *)
module LiteralTypeTable : Hashtbl.S with type key = literal

(** {6 Functions} *)

(** {2 Special Terms} *)


(** the {e true} literal (sign is true). *)
val true_literal: literal

(** the {e false} literal (sign is true). *)
val false_literal: literal

(** to be used for default initialization. *)
val null_term: term

(** to be used for default initialization. *)
val null_literal: literal

(** marks an assert gap in a context unifier ({!Context_unifier}). *)
val assert_literal: literal

(** the {e +v} context literal. *)
val plus_v: literal

(** the {e -v} context literal. *)
val minus_v: literal




(** {2 Request} *)

(** requests a {!Term.term}[.Var] term from a database.

    there might be different instances of the same variable {!Var.var} in the system.
    Still, as the first request to {!Term.request_var} creates a database entry
    for this specific instance of the variable,
    and the following requests return this first stored instance,
    all {!Term.term}[.Var]s built on this variable are identical. *)
val request_var: var -> term

(** requests a {!Term.term}[.Const] term from a database. *)
val request_const: symbol -> term

(** requests a {!Term.term}[.Func] term from a database. *)
val request_func: symbol * term array -> term


(** requests a {!Term.literal} from the database. *)
val request_literal: bool -> term -> literal

(** requests a negation of a literal from the database. *)
val request_negated_literal: literal -> literal

(** requests the skolemization of a literal from the database.
    this is the skolemization in the Model Evolution sense,
    i.e each universal variable is replaced by a fresh constant,
    while parameters are left untouched
    (each occurrence of the same variable is replaced by the same constant).
*)
val request_skolemized_literal: literal -> literal

(** requests the universal version of a term from the database,
    i.e. each parameter occuring in the term is replaced
    by a fresh universal variable.
    (each occurrence of the same parameter is replaced by the same variable).
*)
val request_universal_term: term -> term

(** like {!Term.request_universal_term} for a literal. *)
val request_universal_literal: literal -> literal

(** like {!Term.request_universal_clause} for a clause. *)
val request_universal_clause: clause -> clause

(** [replace_in_term database term sub_term replace_term]
    builds a new term from [term] by replacing
    every occurrence of [sub_term] in [term] by [replace_term]. *)
val replace_in_term: term -> term ->  term -> term




(** {2 Decomposition} *)

(** the variables of the term (each only once). *)
val vars_of_term: term -> var list

(** the variables of the literal (each only once). *)
val vars_of_literal: literal -> var list


  

(** {2 Comparison} *)

val term_equal : term -> term -> bool

val literal_equal : literal -> literal -> bool

(** clauses are not considered as multi-sets for this comparison *)
val clause_equal : clause -> clause -> bool

(** a non-total order on literals. *)
val compare_literals: literal -> literal -> int

(** sorts a clause via {!Term.compare_literals}. *)
val sort_clause: clause -> clause



(** {2 Variants} *)


(** are two terms (universal- and parameter-preserving) variants? *)
val are_terms_variants: term -> term -> bool

(** are two literals (universal- and parameter-preserving) variants? *)
val are_literals_variants: literal -> literal -> bool

(** are two terms variants?
    parameters and variables are not differentiated. *)
val are_literals_weak_variants: literal -> literal -> bool

(** are two terms variants modulo skolem constants??
    the skolem constants of one literal are bijectively mapped
    to the skolem literals of the other literal,
    i.e. f(__0__, __0__) and f(__1__, __1__) are considered to be variants,
    while f(__0__, __0__) and f(__1__, __2__) are not. *)
val are_literals_skolem_variants: literal -> literal -> bool



(** {2 String Representation} *)

(** Example: {b +p(a, _0)} *)
val term_to_string : term -> string

(** Example: {b -p(a, _0)} *)
val literal_to_string : literal -> string

(** Example: {b \{ +p(f(a), _0), -q(b) \}} *)
val clause_to_string : clause -> string
