(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** term unification

  This modules provides the basic unification functions for terms and literals.

  For an explanation of {e offsets}, {e p_preserving}, {e i_preserving} see {!Subst}.*)

type term = Term.term
type literal = Term.literal
type binding = Subst.binding
type subst = Subst.subst
type packed_subst = Subst.packed_subst


(** raised by all functions when the unification fails. *)
exception UNIFICATION_FAIL




(** {6 Functions} *)


(** {2 Matching} *)

(** [match_terms term1 offset1 term2 offset2]
  returns a most general substitution which applied to [term1 offset1]
  produces [term2].

  {b raises} {!Unification.UNIFICATION_FAIL} *)
val match_terms: ?p_preserving:bool -> ?i_preserving:bool -> term -> int -> term -> int -> subst

(** like {!Unification.match_terms} over literals. *)
val match_literals: ?p_preserving:bool -> ?i_preserving:bool -> literal -> int -> literal -> int -> subst





(** {2 Unification} *)

(** [unify_terms term1 offset1 term2 offset2]
  returns a most general substitution which applied to [term1 offset1]
  or [term2 offset2] produces the same term.

  prefers to bind universal variables to parametric ones
  instead of vice versa (see {!Admissible}).
  
  {b raises} {!Unification.UNIFICATION_FAIL} if no unifier exists. *)
val unify_terms: ?p_preserving:bool -> ?i_preserving:bool -> term -> int -> term -> int -> subst


(** like {!Unification.unify_terms} over literals. *)
val unify_literals: ?p_preserving:bool -> ?i_preserving:bool -> literal -> int -> literal -> int -> subst





(** {2 Substitution Merging} *)

(** Merging two substitutions into one substitution.

  Allows to unifiy two literal lists (or clauses)
  separately literal pair by literal pair
  and afterwords combining the computed substitution,
  instead of incrementally computing one substitution.

  Also used in substitution trees.

  Faster if the first substitution is the larger substitution. *)

(** [match_substs subsuming_offset subst1 subst2]
  merges [subst1] and [subst2].

  Terms with [subsuming_offset] have been matched to other terms,
  and the resulting substitutions are now to be merged.

  If a variable is bound to different terms in subst1 and subst2
  the bound terms are unified.
  I.e. if
  - both terms have [subsuming_offset] they are unified ({!Unification.unify_terms})
  - if one term has [subsuming_offset] it is matched
    ({!Unification.match_terms}) to the other
  - if no term has [subsuming_offset] the substitutions are not matcheable
    and {!Unification.UNIFICATION_FAIL} is raised.
*)
val match_substs: ?p_preserving:bool -> ?i_preserving:bool -> int -> subst -> subst -> subst

(** [match_substs subsuming_offset subst1 subst2]
  merges [subst1] and [subst2].

  If a variable is bound to different terms in subst1 and subst2
  the bound terms are unified ({!Unification.unify_terms}).

  if [x -> y] and [y -> x] have to be merged one is dropped
  and the other kept, i.e. this is not considered to be a cyclic binding.

  Examples:
  - [ x -> a] and [ x -> b ] are not mergeable.
  - [ x -> g(y) ] and [ y -> a ] becomes [ x -> g(y); y -> a].
  - [ x -> g(y) ] and [ x -> g(a) ] becomes [ x -> g(y); y -> a].
  - [ x -> y] and [ y -> x ] becomes [x -> y] or [y -> x].

   {b raises} {!Unification.UNIFICATION_FAIL} on not mergeable substitutions. *)
val unify_substs: ?p_preserving:bool -> ?i_preserving:bool -> subst -> subst -> subst

(** like {!Unification.unify_substs},
  but with a [packed_subst] instead of a [subst] as an argument. *)
val unify_packed_substs: ?p_preserving:bool -> ?i_preserving:bool -> subst -> packed_subst -> subst
