#!/usr/bin/env python2.3
# ----------------------------------
#
# pylib_jobs.py
#

"""
Classes for handling jobs that are forked off and return a result via
exist status and a file interface.


Copyright 2005 Stephan Schulz, schulz@eprover.org

This code is part of the support structure for the equational
theorem prover E. Visit

 http://www.eprover.org

for more information.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program ; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307 USA 

The original copyright holder can be contacted as

Stephan Schulz (I4)
Technische Universitaet Muenchen
Institut fuer Informatik
Boltzmannstrasse 3
Garching bei Muenchen
Germany

or via email (address above).
"""

import re
import string
import sys
import time
import os

import pylib_basics
import pylib_io


class task(object):
    """
    A (mostly) virtual class defining the interface to handle tasks,
    i.e. work that is done via forked off processes.
    """
    pid = os.getpid()
    no_host        = 0
    unchecked_host = 1
    ready_host     = 2
    JobRequiresHost = "Job requires host to execute, but none provided"
    
    def __init__(self, ident):
        self.ident = ident
        self.result_filename = None

    def get_filename(self):
        if not self.result_filename:
            self.result_filename = "pytask_%ld_%ld" % self.ident, task.pid
        return self.result_filename

    def vneeds_host(self):
        """
        Virtual function - should return task.no_host for local tasks,
        task.unchecked_host for tasks that try to establish if a host is
        ok, and task.ready_host for tasks that need a host to run on.
        """
        return task.no_host

    def run(self, host = None):
        if self.vneeds_host() and not host:
            raise task.JobRequiresHost
        pid = os.fork()
        if pid:
            return pid
        else:
            # os.setpgid(0,0)
            res = self.vrun(host)
            os._exit(res)

    def vrun(self, host = None):
        """
        Virtual function to be overridden by concrete task classes.
        """
        print "Executing task "+str(self.ident)+" on "+str(host) 
        return 0

    def process_result(self, fork_status):
        """
        Generic result handling: If there is a result file name, read 
        and return its content (or None if the file does not
        exist). Also call virtual function vprocess_result() for
        specialized processing.
        """
        res = None
        status,signal,core=pylib_io.decode_wait_status(fork_status)
        if self.result_filename:
            try:
                fp = pylib_io.flexopen(self.result_filename, "r")
                res = fp.readlines()
            except IOError:
                pass
            
        self.vprocess_result(fork_status,res)
        return res

    def vprocess_result(self,fork_status,res):
        """
        Placeholder for virtual function doing nothing...
        """
        print "Processing  result of task "+str(self.ident)
        

class task_queue(object):
    def __init__(self, hosts = None):
        self.open_tasks = []
        self.running_tasks = {}
        self.unchecked_hosts = []
        if hosts:
            self.unchecked_hosts = hosts
        self.ready_hosts = []
        self.used_hosts = {}

    def add_job(self, job):
        self.open_tasks.append(job)

    def add_jobs(self, jobs):
        for i in jobs:
            self.add_job(i)

    def start_job(self):
        for i in range(len(self.open_tasks)):
            job = self.open_tasks[i]
            pid = None
            host = None
            if job.vneeds_host()==task.no_host:
                pass
            elif job.vneeds_host()==task.unchecked_host:
                try:
                    host = self.unchecked_hosts.pop()
                except IndexError:
                    pass                
            else:
                try:
                    host = self.ready_hosts.pop()
                except IndexError:
                    pass
            try:
                pid = job.run(host)
                self.running_tasks[pid] = job
                del(self.open_tasks[i])
                return pid
            except task.JobRequiresHost:
                pass
        return None

    def start_jobs(self):
        try:
            while self.start_job():
                pass
        except OSError:
            print "Out of processes...."
        return len(self.open_tasks)

    def collect_job(self):
        pid, status = os.wait()
        job = self.running_tasks[pid]
        del(self.running_tasks[pid])
        return job.process_result(status)

    def collect_jobs(self):
        while len(self.running_tasks):
            self.collect_job()

queue = task_queue()

jobs = map(lambda x:task(x), range(200))
queue.add_jobs(jobs)
while queue.start_jobs():
    queue.collect_jobs()
