#include "../ladr/top_input.h"
#include "../ladr/clause2.h"  /* for call to demod_clause */

#define PROGRAM_NAME    "hints_rewriter"
#define PROGRAM_VERSION "2005-A"
#include "../PROGRAM_DATE.h"

static char Help_string[] =
"\nThis program takes a file of demodulators (arg 1) and a\n"
"stream of hints lists (stdin).  Rewritten hints lists are sent to stdout.\n"
"The demodulators are not checked for termination.\n"
"For example,\n\n"
"   hints_rewriter demods < terms.in > terms.out\n\n"
"The file of demodulators contains optional commands\n"
"then a list of demodulators.  The commands can be used to\n"
"declare infix operations and associativity/commutativity.\n"
"Example file of demodulators:\n\n"
"    op(400, infix, ^).\n"
"    op(400, infix, v).\n"
"    assoc_comm(^).\n"
"    assoc_comm(v).\n"
"    clauses(theory).\n"
"    x ^ x = x.\n"
"    x ^ (x v y) = x.\n"
"    x v x = x.\n"
"    x v (x ^ y) = x.\n"
"    end_of_list.\n\n";

int main(int argc, char **argv)
{
  FILE *head_fp;
  Plist demodulators, p;
  Mindex idx;
  Term t;
  BOOL verbose = string_member("verbose", argv, argc);
  int label_att;

  /* counter to label rewritten clauses */
  int count = 0;

  /* handle attributes on read clauses */
  label_att = register_attribute("label", STRING_ATTRIBUTE);

  if (string_member("help", argv, argc) ||
      string_member("-help", argv, argc) ||
      argc < 2) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  init_standard_ladr();
  init_paramod();

  head_fp = fopen(argv[1], "r");
  if (head_fp == NULL)
    fatal_error("demodulator file can't be opened for reading");

  t = read_commands(head_fp, stderr, verbose, KILL_UNKNOWN);

  if (!is_term(t, "clauses", 1))
    fatal_error("clauses(demodulators) not found");

  /* Read list of demodulators. */

  demodulators = read_clause_list(head_fp, stderr, TRUE);

  fclose(head_fp);

  /* AC-canonicalize and index the demodulators. */

  if (assoc_comm_symbols() || comm_symbols())
    idx = mindex_init(DISCRIM_WILD, BACKTRACK_UNIF, 0);
  else
    idx = mindex_init(DISCRIM_BIND, ORDINARY_UNIF, 0);

  for (p = demodulators; p != NULL; p = p->next) {
    /* assume positive equality unit */
    Clause d = p->v;
    Literal lit = d->literals;
    Term alpha = lit->atom->args[0];
    mark_oriented_eq(lit->atom);  /* don not check for termination */
    if (assoc_comm_symbols())
      ac_canonical(lit->atom, -1);
    mindex_update(idx, alpha, INSERT);
  }

  if (verbose)
    fwrite_clause_list(stdout, demodulators, "demodulators");

  /* Read and demodulate lists of clauses. */

  t = read_term(stdin, stderr);

  if (!is_term(t, "clauses", 1))
    fatal_error("hints_rewriter: clauses() not found");

  while (t != NULL)
  {
    Plist hints = read_clause_list(stdin, stderr, TRUE);

    for (p = hints; p != NULL; p = p->next)
    {
      char count_str[8];
      char label_str[20];

      Clause c = p->v;

      demod_clause(c, idx, FALSE);
      /* c now is the demodulated version of c */

      /* append new label attribute */
      count++ ;
      itoa(count, count_str);
      strcpy(label_str, "\"hint ");
      strcat(label_str, count_str);
      strcat(label_str, "\"");

      c -> attributes = set_string_attribute(c -> attributes,
          label_att, new_str_copy(label_str));
    }
    fwrite_clause_list(stdout, hints, "hints");

    zap_plist(hints);

    /* next clause list */

    t = read_term(stdin, stderr);

    if (t != NULL && !is_term(t, "clauses", 1))
       fatal_error("hints_rewriter:  clauses() not found");
  }

  exit(0);

}  /* main */
