#include "top_input.h"

/* Private definitions and types */

typedef struct readlist * Readlist;

struct readlist {
  char *name;      /* name, as it appears in the input file */
  int type;        /* CLASUES, FORMULAS, TERMS */
  Plist *p;        /* *pointer* to the Plist */
  Readlist next;
};

static Readlist Input_lists = NULL;

/*
 * memory management
 */

#define PTRS_READLIST CEILING(sizeof(struct readlist),BYTES_POINTER)
static unsigned Readlist_gets, Readlist_frees;

/*************
 *
 *   Readlist get_readlist()
 *
 *************/

static
Readlist get_readlist(void)
{
  Readlist p = get_cmem(PTRS_READLIST);
  Readlist_gets++;
  return(p);
}  /* get_readlist */

/*************
 *
 *    free_readlist()
 *
 *************/

#if 0
static
void free_readlist(Readlist p)
{
  free_mem(p, PTRS_READLIST);
  Readlist_frees++;
}  /* free_readlist */
#endif

/*************
 *
 *   fprint_top_input_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with the top_input package.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_top_input_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      bytes\n");

  n = sizeof(struct readlist);
  fprintf(fp, "readlist (%4d)      %11u%11u%11u%9.1f K\n",
          n, Readlist_gets, Readlist_frees,
          Readlist_gets - Readlist_frees,
          ((Readlist_gets - Readlist_frees) * n) / 1024.);

}  /* fprint_top_input_mem */

/*************
 *
 *   p_top_input_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with the top_input package.
*/

/* PUBLIC */
void p_top_input_mem()
{
  fprint_top_input_mem(stdout, TRUE);
}  /* p_top_input_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   process_op()
 *
 *************/

static
void process_op(FILE *fout, Term t, int prec, Term type_term, Term symb_term)
{
  if (ARITY(symb_term) != 0) {
    fwrite_term_nl(fout, t);
    fwrite_term_nl(stderr, t);
    fatal_error("symbols in op command must have no arguments");
  }
  else {
    Parsetype pt = NOTHING_SPECIAL;
    if (is_constant(type_term, "infix"))
      pt = INFIX;
    else if (is_constant(type_term, "infix_left"))
      pt = INFIX_LEFT;
    else if (is_constant(type_term, "infix_right"))
      pt = INFIX_RIGHT;
    else if (is_constant(type_term, "prefix"))
      pt = PREFIX;
    else if (is_constant(type_term, "prefix_paren"))
      pt = PREFIX_PAREN;
    else if (is_constant(type_term, "postfix"))
      pt = POSTFIX;
    else if (is_constant(type_term, "postfix_paren"))
      pt = POSTFIX_PAREN;
    else if (is_constant(type_term, "clear"))
      pt = NOTHING_SPECIAL;
    else {
      fwrite_term_nl(fout, t);
      fwrite_term_nl(stderr, t);
      fatal_error("bad parse-type in op command");
    }
    set_parse_type(sn_to_str(SYMNUM(symb_term)), prec, pt);
  }
}  /* process_op */

/*************
 *
 *   flag_handler()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void flag_handler(FILE *fout, Term t, BOOL echo, int unknown_action)
{
  int flag = str_to_flag_id(sn_to_str(SYMNUM(ARG(t,0))));
  if (flag == -1) {
    if (unknown_action == KILL_UNKNOWN) {
      fwrite_term_nl(fout, t);
      fwrite_term_nl(stderr, t);
      fatal_error("flag not recognized");
    }
    else if (unknown_action == WARN_UNKNOWN) {
      bell(stderr);
      fprintf(fout,   "WARNING, flag not recognized: ");
      fwrite_term_nl(fout,   t);
      fprintf(stderr, "WARNING, flag not recognized: ");
      fwrite_term_nl(stderr, t);
    }
    else if (unknown_action == NOTE_UNKNOWN) {
      fprintf(fout,   "NOTE: flag not recognized: ");
      fwrite_term_nl(fout,   t);
    }
  }
  else
    update_flag(fout, flag, is_term(t, "set", 1), echo);
}  /* flag_handler */

/*************
 *
 *   parm_handler()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void parm_handler(FILE *fout, Term t, BOOL echo, int unknown_action)
{
  int val;
  BOOL ok = term_to_int(ARG(t,1), &val);
  if (!ok) {
    int id = str_to_stringparm_id(sn_to_str(SYMNUM(ARG(t,0))));
    if (id == -1) {
      if (unknown_action == KILL_UNKNOWN) {
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("parm not recognized");
      }
      else if (unknown_action == WARN_UNKNOWN) {
	bell(stderr);
	fprintf(fout,   "WARNING, parm not recognized: ");
	fwrite_term_nl(fout,   t);
	fprintf(stderr, "WARNING, parm not recognized: ");
	fwrite_term_nl(stderr, t);
      }
      else if (unknown_action == NOTE_UNKNOWN) {
	fprintf(fout,   "NOTE: parm not recognized: ");
	fwrite_term_nl(fout,   t);
      }
    }
    else {
      char *s = sn_to_str(SYMNUM(ARG(t,1)));
      assign_stringparm(id, s, echo);
    }
  }
  else {
    int parm = str_to_parm_id(sn_to_str(SYMNUM(ARG(t,0))));
    if (parm == -1) {
      if (unknown_action == KILL_UNKNOWN) {
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("parm not recognized");
      }
      else if (unknown_action == WARN_UNKNOWN) {
	bell(stderr);
	fprintf(fout,   "WARNING, parm not recognized: ");
	fwrite_term_nl(fout,   t);
	fprintf(stderr, "WARNING, parm not recognized: ");
	fwrite_term_nl(stderr, t);
      }
      else if (unknown_action == NOTE_UNKNOWN) {
	fprintf(fout,   "NOTE: parm not recognized: ");
	fwrite_term_nl(fout,   t);
      }
    }
    else {
      assign_parm(parm, val, echo);
    }
  }
}  /* parm_handler */

/*************
 *
 *   preliminary_precedence()
 *
 *************/

static
void preliminary_precedence(Plist terms)
{
  Plist strings = NULL;
  Plist p;
  for (p = terms ; p; p = p->next) {
    Term t = p->v;
    if (!CONSTANT(t))
      fatal_error("terms in lex command must be constants (arity deduced from clauses)");
    strings = plist_append(strings, sn_to_str(SYMNUM(t)));
  }
  set_preliminary_precedence(strings);
  zap_plist(strings);
}  /* preliminary_precedence */

/*************
 *
 *   readlist_member()
 *
 *************/

static
Readlist readlist_member(Readlist p, char *name, int type)
{
  if (p == NULL)
    return NULL;
  else if (p->type == type && str_ident(p->name, name))
    return p;
  else
    return readlist_member(p->next, name, type);
}  /* readlist_member */

/*************
 *
 *   readlist_member_wild()
 *
 *************/

static
Readlist readlist_member_wild(Readlist p, int type)
{
  if (p == NULL)
    return NULL;
  else if (p->type == type && str_ident(p->name, ""))
    return p;
  else
    return readlist_member_wild(p->next, type);
}  /* readlist_member */

/*************
 *
 *   accept_list()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void accept_list(char *name, int type, Plist *l)
{
  Readlist p = readlist_member(Input_lists, name, type);
  if (p)
    fatal_error("accept_list, duplicate name/type");
  else {
    p = get_readlist();
    p->name = name;
    p->type = type;
    *l = NULL;
    p->p = l;

    p->next = Input_lists;
    Input_lists = p;
  }
}  /* accept_list */

/*************
 *
 *   symbol_check_and_declare()
 *
 *   Make sure that symbols are being used in reasonable ways.
 *
 *   Look at the clauses and formulas in the Input_lists.
 *   Make sure that (1) no symbol is used as both a relation
 *   and a function symbol, and (2) that no symbol has multiple
 *   arities.  Also tell the symbol package which symbols are
 *   relations and which are functions.
 *
 *************/

static
void symbol_check_and_declare(void)
{
  Ilist fsyms, rsyms;
  Mlist fsyms_multiset = NULL;
  Mlist rsyms_multiset = NULL;
  Readlist p;

  for (p = Input_lists; p; p = p->next) {
    if (p->type == CLAUSES) {
      Mlist f = function_symbols_in_clauses(*(p->p));
      Mlist r = relation_symbols_in_clauses(*(p->p));
      fsyms_multiset = multiset_union(fsyms_multiset, f);
      rsyms_multiset = multiset_union(rsyms_multiset, r);
    }
    else if (p->type == FORMULAS) {
      Mlist f = function_symbols_in_formulas(*(p->p));
      Mlist r = relation_symbols_in_formulas(*(p->p));
      fsyms_multiset = multiset_union(fsyms_multiset, f);
      rsyms_multiset = multiset_union(rsyms_multiset, r);
    }
  }

  /* ??? fsyms_multiset = ilist_cat(fsyms_multiset, skolem_symbols()); */
  
  fsyms = multiset_to_set(fsyms_multiset);  /* set of function symbols */
  rsyms = multiset_to_set(rsyms_multiset);  /* set of relation symbols */

  /* Check if any symbol is used as both a relation and function symbol. */

  function_relation_check(fsyms, rsyms, TRUE);

  /* Check if any symbol is used with multiple arities. */

  arity_check(fsyms, rsyms, TRUE);

  /* Tell the symbol package the functions and relations. */
  /* (needed for ordering the symbols) */

  declare_functions_and_relations(fsyms, rsyms);

  zap_mlist(fsyms_multiset);
  zap_mlist(rsyms_multiset);
  zap_ilist(fsyms);
  zap_ilist(rsyms);
}  /* symbol_check_and_declare */

/*************
 *
 *   read_from_file()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void read_from_file(FILE *fin, FILE *fout, BOOL echo, int unknown_action)
{
  Term t = read_term(fin, fout);

  while (t != NULL) {
    if (is_term(t, "set", 1) || is_term(t, "clear", 1)) {
      /********************************************************** set, clear */
      if (echo)
	fwrite_term_nl(fout, t);
      flag_handler(fout, t, echo, unknown_action);

      /* SPECIAL CASE: this needs action right now! */
      if (is_term(ARG(t,0), "prolog_style_variables", 0)) {
	if (is_term(t,"set",1))
	  set_variable_style(PROLOG_STYLE);
	else
	  set_variable_style(STANDARD_STYLE);
      }
    }
    else if (is_term(t, "assign", 2)) {
      /************************************************************** assign */
      if (echo)
	fwrite_term_nl(fout, t);
      parm_handler(fout, t, echo, unknown_action);
    }
    else if (is_term(t, "assoc_comm", 1) ||
             is_term(t, "commutative", 1)) {
      /************************************************************ AC, etc. */
      Term f = ARG(t,0);
      if (!CONSTANT(f)) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("argument must be symbol only");
      }
      else {
	if (echo)
	  fwrite_term_nl(fout, t);
	if (is_term(t, "assoc_comm", 1))
	  set_assoc_comm(sn_to_str(SYMNUM(f)), TRUE);
	else
	  set_commutative(sn_to_str(SYMNUM(f)), TRUE);
      }
    }
    else if (is_term(t, "op", 3)) {
      /****************************************************************** op */
      /* e.g., op(300, infix, +); */
      Term prec_term = ARG(t,0);
      Term type_term = ARG(t,1);
      Term symb_term = ARG(t,2);
      int prec;
      BOOL ok = term_to_int(prec_term, &prec);
      if (echo)
	fwrite_term_nl(fout, t);

      if (!ok || prec < MIN_PRECEDENCE || prec > MAX_PRECEDENCE) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("bad precedence in op command");
      }
      else if (proper_listterm(symb_term)) {
	while (cons_term(symb_term)) {
	  process_op(fout, t, prec, type_term, ARG(symb_term, 0));
	  symb_term = ARG(symb_term, 1);
	}
      }
      else
	process_op(fout, t, prec, type_term, symb_term);
    }
    else if (is_term(t, "lex", 1)) {
      /***************************************************************** lex */
      Plist p = listterm_to_tlist(ARG(t,0));
      if (p == NULL) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("lex command must contain a proper list, e.g., [a,b,c]");
      }
      else {
	if (echo)
	  fwrite_term_nl(fout, t);
	preliminary_precedence(p);
	zap_plist(p);
      }
    }
    else if (is_term(t, "clauses", 1) ||
	     is_term(t, "formulas", 1) ||
	     is_term(t, "terms", 1)) {
      /***************************************************** list of objects */
      char *name = term_symbol(ARG(t,0));
      Plist objects = NULL;
      int type;
      int echo_id = str_to_flag_id("echo_input");
      BOOL echo_objects = (echo_id == -1 ? TRUE : flag(echo_id));
      int prolog_var_id = str_to_flag_id("prolog_style_variables");

      /* In case the variable style has not been set: */
      if (flag(prolog_var_id))
	set_variable_style(PROLOG_STYLE);
      else
	set_variable_style(STANDARD_STYLE);

      if (is_term(t, "clauses", 1)) {
	type = CLAUSES;
	objects = read_clause_list(fin, fout, FALSE);
	if (echo) {
	  if (echo_objects)
	    fwrite_clause_list(fout, objects, name);
	  else {
	    fprintf(fout, "\n%% ");
	    fwrite_term(fout, t);
	    fprintf(fout, ".  %% not echoed (%d clauses)\n",
		    plist_count(objects));
	  }
	}
      }
      else if (is_term(t, "formulas", 1)) {
	type = FORMULAS;
	objects = read_formula_list(fin, fout);
	if (echo) {
	  if (echo_objects)
	    fwrite_formula_list(fout, objects, name);
	  else {
	    fprintf(fout, "\n%% ");
	    fwrite_term(fout, t);
	    fprintf(fout, ".  %% not echoed (%d formulas)\n",
		    plist_count(objects));
	  }
	}
      }
      else {
	type = TERMS;
	objects = read_term_list(fin, fout);
	if (echo) {
	  if (echo_objects)
	    fwrite_term_list(fout, objects, name);
	  else {
	    fprintf(fout, "\n%% ");
	    fwrite_term(fout, t);
	    fprintf(fout, ".  %% not echoed (%d formulas)\n",
		    plist_count(objects));
	  }
	}
      }

      /* Find the correct list, and append the objects to it. */
      
      {
	Readlist r = readlist_member(Input_lists, name, type);
	if (r == NULL)
	  r = readlist_member_wild(Input_lists, type);
	if (r == NULL) {
	  printf("list head: "); p_term(t);
	  fatal_error("read_lists_from_file: name/type not recognized");
	}
	else {
	  *(r->p) = plist_cat(*(r->p), objects);
	}
      }
    }  /* list of clauses, formulas, or terms */
    else {
      fprintf(fout, "unrecognized command or list: "); p_term(t);
      fprintf(stderr, "unrecognized command or list: "); p_term(t);
      fatal_error("read_from_file, unrecognized command or list");
    }
    zap_term(t);
    t = read_term(fin, fout);
  }
}  /* read_from_file */

/*************
 *
 *   read_all_input()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void read_all_input(int argc, char **argv, FILE *fout,
		    BOOL echo, int unknown_action)
{
  int n = which_string_member("-f", argv, argc);
  if (n == -1)
    read_from_file(stdin, fout, echo, unknown_action);
  else {
    int i;
    for (i = n+1; i < argc; i++) {
      FILE *fin = fopen(argv[i], "r");
      if (fin == NULL) {
	char s[100];
	sprintf(s, "read_all_input, file %s not found", argv[i]);
	fatal_error(s);
      }
      if (echo)
	printf("\n%% Reading from file %s\n\n", argv[i]);
      read_from_file(fin, fout, echo, unknown_action);
      fclose(fin);
    }
  }
  process_standard_options();
  symbol_check_and_declare();
}  /* read_all_input */

/*************
 *
 *   read_commands()
 *
 *************/

/* DOCUMENTATION
This is a legacy routine.
*/

/* PUBLIC */
Term read_commands(FILE *fin, FILE *fout, BOOL echo, int unknown_action)
{
  Term t = read_term(fin, fout);
  BOOL go = (t != NULL);

  while (go) {
    BOOL already_echoed = FALSE;
    /************************************************************ set, clear */
    if (is_term(t, "set", 1) || is_term(t, "clear", 1)) {
      if (echo) {
	fwrite_term_nl(fout, t);
	already_echoed = TRUE;
      }
      flag_handler(fout, t, echo, unknown_action);
      /* SPECIAL CASE: this needs action right now! */
      if (is_term(ARG(t,0), "prolog_style_variables", 0)) {
	if (is_term(t,"set",1))
	  set_variable_style(PROLOG_STYLE);
	else
	  set_variable_style(STANDARD_STYLE);
      }
    }
    else if (is_term(t, "assign", 2)) {
      /************************************************************** assign */
      if (echo) {
	fwrite_term_nl(fout, t);
	already_echoed = TRUE;
      }
      parm_handler(fout, t, echo, unknown_action);
    }
    else if (is_term(t, "assoc_comm", 1) ||
             is_term(t, "commutative", 1)) {
      /************************************************************ AC, etc. */
      Term f = ARG(t,0);
      if (!CONSTANT(f)) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("argument must be symbol only");
      }
      else {
	if (is_term(t, "assoc_comm", 1))
	  set_assoc_comm(sn_to_str(SYMNUM(f)), TRUE);
	else
	  set_commutative(sn_to_str(SYMNUM(f)), TRUE);
      }
    }
    else if (is_term(t, "op", 3)) {
      /****************************************************************** op */
      /* e.g., op(300, infix, +); */
      Term prec_term = ARG(t,0);
      Term type_term = ARG(t,1);
      Term symb_term = ARG(t,2);
      int prec;
      BOOL ok = term_to_int(prec_term, &prec);
      if (!ok || prec < MIN_PRECEDENCE || prec > MAX_PRECEDENCE) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("bad precedence in op command");
      }
      else if (proper_listterm(symb_term)) {
	while (cons_term(symb_term)) {
	  process_op(fout, t, prec, type_term, ARG(symb_term, 0));
	  symb_term = ARG(symb_term, 1);
	}
      }
      else
	process_op(fout, t, prec, type_term, symb_term);
    }
    else if (is_term(t, "lex", 1)) {
      /***************************************************************** lex */
      Plist p = listterm_to_tlist(ARG(t,0));
      if (p == NULL) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("lex command must contain a proper list, e.g., [a,b,c]");
      }
      else {
	preliminary_precedence(p);
	zap_plist(p);
      }
    }
    else {
      /******************************************************** unrecognized */
      /* return this unknown term */
      go = FALSE;
    }

    if (go) {
      if (echo && !already_echoed)
	fwrite_term_nl(fout, t);
      zap_term(t);
      t = read_term(fin, fout);
      go = (t != NULL);
    }
  }
  return t;
}  /* read_commands */

/*************
 *
 *   init_standard_ladr()
 *
 *************/

/* DOCUMENTATION
This routine initializes various LADR packaages.
*/

/* PUBLIC */
void init_standard_ladr(void)
{
  init_wallclock();
  init_paramod();
  init_basic_paramod();
  init_maximal();
  declare_base_symbols();
  declare_standard_parse_types();
  translate_neg_equalities(TRUE);
  init_standard_options();
}  /* init_standard_ladr */

