#include "weight.h"

/* Private definitions and types */

static Plist Rules;

static int Constant_weight;
static int Sk_constant_weight;
static int Not_weight;
static int Or_weight;
static int Prop_atom_weight;
static int Variable_weight;
static int Skolem_penalty;
static int Nest_penalty;
static BOOL Not_rules;  /* any rules for NOT_SYM? */
static BOOL Or_rules;   /* any rules for OR_SYM? */

/* Cache the symnums */

static int Sum_sn;     /* integer arithmetic */
static int Prod_sn;    /* integer arithmetic */
static int Neg_sn;     /* integer arithmetic */
static int Div_sn;     /* integer arithmetic */
static int Max_sn;     /* integer arithmetic */
static int Min_sn;     /* integer arithmetic */
static int Depth_sn;   /* depth */
static int Avar_sn;    /* anonymous variable */

/*************
 *
 *   init_weight()
 *
 *************/

/* DOCUMENTATION
Initialize weighting.  The rules are copied.
*/

/* PUBLIC */
void init_weight(Plist rules,
		 int variable_weight,
		 int constant_weight,
		 int not_weight,
		 int or_weight,
		 int sk_constant_weight,
		 int prop_atom_weight,
		 int skolem_penalty,
		 int nest_penalty)
{
  Plist p;
  Rules = NULL;
  for (p = rules; p; p = p->next) {
    Term rule = copy_term(p->v);
    if (!is_eq_symbol(SYMNUM(rule)))
      fatal_error("init_weight, rule is not equality");
    else {
      term_set_variables(rule, MAX_VARS);
      Rules = plist_append(Rules, rule);
      if (is_term(ARG(rule,0), NOT_SYM, 1))
	Not_rules = TRUE;
      if (is_term(ARG(rule,0), OR_SYM, 2))
	Or_rules = TRUE;
      }
    }

  Variable_weight = variable_weight;
  Constant_weight = constant_weight;
  Not_weight = not_weight;
  Or_weight = or_weight;
  Prop_atom_weight = prop_atom_weight;
  Sk_constant_weight = sk_constant_weight;
  Skolem_penalty = skolem_penalty;
  Nest_penalty = nest_penalty;

  /* Cache symbol numbers. */

  Sum_sn  = str_to_sn("+", 2);
  Prod_sn = str_to_sn("*", 2);
  Div_sn  = str_to_sn("/", 2);
  Max_sn  = str_to_sn("max", 2);
  Min_sn  = str_to_sn("min", 2);
  Depth_sn  = str_to_sn("depth", 1);
  Neg_sn  = str_to_sn("-", 1);
  Avar_sn = str_to_sn("_", 0);

}  /* init_weight */

/*************
 *
 *   apply_depth()
 *
 *************/

static
int apply_depth(Term t, Context subst)
{
  if (VARIABLE(t))
    return term_depth(subst->terms[VARNUM(t)]);
  else if (CONSTANT(t))
    return 0;
  else {
    int depth = 0;
    int i;
    for (i = 0; i < ARITY(t); i++) {
      int d = apply_depth(ARG(t,i), subst);
      depth = IMAX(d, depth);
    }
    return depth + 1;
  }
}  /* apply_depth */

/*************
 *
 *   weight_calc() -- mutually recursive with weight_calc
 *
 *************/

static
int weight_calc(Term b, Context subst)
{
  /* Apply a rule.  Term b is the right side of the rule. 
     The substitution matches the left side of the rule with the
     term being weighed.

     This routine is recursive, applying arithmetic rules to the
     top of b as much as possible.  When a non-arithmetic expression
     is encountered, the substition is applied to it and then
     it is weighed as usual.
   */
  int wt;
  if (VARIABLE(b)) {
    Term t = subst->terms[VARNUM(b)];
    Context subst2 = get_context();
    int wt;
    if (t == NULL) 
      fatal_error("weight_calc, uninstantiated variable");
    wt = weight(t, subst2);
    free_context(subst2);
    return wt;
  }
  else if (SYMNUM(b) == Sum_sn)
    return
      weight_calc(ARG(b,0), subst) +
      weight_calc(ARG(b,1), subst);
  else if (SYMNUM(b) == Prod_sn)
    return
      weight_calc(ARG(b,0), subst) *
      weight_calc(ARG(b,1), subst);
  else if (SYMNUM(b) == Div_sn)
    return
      weight_calc(ARG(b,0), subst) /
      weight_calc(ARG(b,1), subst);
  else if (SYMNUM(b) == Max_sn) {
    int w1 = weight_calc(ARG(b,0), subst);
    int w2 = weight_calc(ARG(b,1), subst);
    return IMAX(w1,w2);
  }
  else if (SYMNUM(b) == Min_sn) {
    int w1 = weight_calc(ARG(b,0), subst);
    int w2 = weight_calc(ARG(b,1), subst);
    return IMIN(w1,w2);
  }
  else if (SYMNUM(b) == Depth_sn)
    return apply_depth(ARG(b,0), subst);
  else if (term_to_int(b, &wt))
    return wt;
  else {
    Term b_prime = apply(b, subst);
    Context subst2 = get_context();
    wt = weight(b, subst2);
    free_context(subst2);
    zap_term(b_prime);
    return wt;
  }
}  /* weight_calc */

/*************
 *
 *   weight() -- mutually recursive with weight_calc
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int weight(Term t, Context subst)
{
  if (VARIABLE(t))
    return Variable_weight;
  else {
    /* Look for a rule to apply. */
    Plist p;
    for (p = Rules; p; p = p->next) {
      Term rule = p->v;
      Term alpha = ARG(rule,0);
      Term beta  = ARG(rule,1);
      Trail tr = NULL;
      if (match_weight(alpha, subst, t, &tr, Avar_sn)) {
	/* We found a rule.  Now calculate the weight. */
	int wt = weight_calc(beta, subst);
	undo_subst(tr);
	return wt;
      }
    }
    /* Nothing matches; return the default. */
    if (CONSTANT(t)) {
      if (skolem_term(t) && Sk_constant_weight != 1)
	return Sk_constant_weight;
      else if (relation_symbol(SYMNUM(t)))
	return Prop_atom_weight;
      else
	return Constant_weight;
    }
    else {
      /* sum of weights of subterms, plus 1 */
      BOOL skolem = (Skolem_penalty != 1 && skolem_term(t));
      int wt = 1;
      int i;
      for (i = 0; i < ARITY(t); i++) {
	int arg_wt = weight(ARG(t, i), subst);
	if (skolem && contains_skolem_function(ARG(t,i)))
	  arg_wt *= Skolem_penalty;

	if (Nest_penalty != 0 &&
	    ARITY(t) <= 2 &&
	    SYMNUM(t) == SYMNUM(ARG(t,i)))
	  wt += Nest_penalty;
	
	wt += arg_wt;
      }
      return wt;
    }
  }
}  /* weight */

/*************
 *
 *   clause_weight()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int clause_weight(Clause c)
{
  if (!Not_rules && !Or_rules) {
    /* There are no rules for OR or NOT, so we don't need to construct a
       Term representation of the clause. */
    int wt = 0;
    Literal lit;
    for (lit = c->literals; lit; lit = lit->next) {
      Context subst = get_context();
      wt += weight(lit->atom, subst);
      free_context(subst);
    }
    wt += (number_of_neg_literals(c) * Not_weight);
    wt += ((number_of_literals(c)-1) * Or_weight);
    return wt;
  }
  else {
    /* Build a temporary Term representation of the clause and weigh that.
       This is done in case there are weight rules for OR or NOT. */
    Term temp = lits_to_term(c->literals);
    Context subst = get_context();
    int wt = weight(temp, subst);
    free_context(subst);
    free_lits_to_term(temp);

    /* If there are no Not_rules, we have already added one for each not;
       so we undo that and add the correct amount.  Same for Or_rules. */
       
    if (!Not_rules)
      wt += (number_of_neg_literals(c) * (Not_weight - 1));
    if (!Or_rules)
      wt += ((number_of_literals(c) - 1) * (Or_weight - 1));
    
    return wt;
  }
}  /* clause_weight */
