#include "xproofs.h"

/* Private definitions and types */

/*************
 *
 *   xx_res2()
 *
 *************/

static
Clause xx_res2(Clause c, int n)
{
  Literal lit = ith_literal(c, n);

  if (lit == NULL ||
      lit->sign == TRUE ||
      !eq_atom(lit->atom))
    return NULL;
  else {
    Context subst = get_context();
    Trail tr = NULL;
    Clause res;

    if (unify(ARG(lit->atom,0), subst, ARG(lit->atom,1), subst, &tr)) {
      Literal l2;
      res = get_clause();
      for (l2 = c->literals; l2; l2 = l2->next)
	if (l2 != lit)
	  append_literal(res, apply_lit(l2,  subst));

      res->attributes = cat_att(res->attributes,
			     inheritable_att_instances(c->attributes, subst));

      res->justification = xxres_just(c, n);
      upward_clause_links(res);
      renumber_variables(res, MAX_VARS);
      undo_subst(tr);
    }
    else {
      res = NULL;
    }
    free_context(subst);
    return res;
  }
}  /* xx_res2 */

/*************
 *
 *   xx_simp2()
 *
 *************/

static
void xx_simp2(Clause c, int n)
{
  Literal lit = ith_literal(c, n);
  if (lit == NULL ||
      lit->sign == TRUE ||
      !eq_atom(lit->atom) ||
      !term_ident(ARG(lit->atom,0), ARG(lit->atom,1)))
    fatal_error("xx_simp2, bad literal");

  zap_term(lit->atom);
  lit->atom = NULL;
  c->literals = remove_null_literals(c->literals);
  c->justification = append_just(c->justification, xx_just(n));
}  /* xx_simp2 */

/*************
 *
 *   factor()
 *
 *************/

static
Clause factor(Clause c, int n1, int n2)
{
  Clause fac;
  Literal l1 = ith_literal(c, n1);
  Literal l2 = ith_literal(c, n2);
  Context subst = get_context();

  Trail tr = NULL;

  if (l1->sign == l2->sign && unify(l1->atom, subst, l2->atom, subst, &tr)) {
    Literal lit;
    fac = get_clause();
    for (lit = c->literals; lit; lit = lit->next)
      if (lit != l2)
	append_literal(fac, apply_lit(lit,  subst));

    fac->attributes = cat_att(fac->attributes,
			      inheritable_att_instances(c->attributes, subst));

    fac->justification = factor_just(c, n1, n2);
    upward_clause_links(fac);
    renumber_variables(fac, MAX_VARS);
    undo_subst(tr);
  }
  else
    fac = NULL;
  free_context(subst);
  return fac;
}  /* factor */

/*************
 *
 *   merge1()
 *
 *************/

static
void merge1(Clause c, int n)
{
  Literal target = ith_literal(c, n);
  Literal prev = ith_literal(c, n-1);
  Literal lit = c->literals;
  BOOL go = TRUE;

  while (go) {
    if (lit->sign == target->sign && term_ident(lit->atom, target->atom))
      go = FALSE;
    else
      lit = lit->next;
  }
  if (lit == target)
    fatal_error("merge1, literal does not merge");
  prev->next = target->next;
  zap_literal(target);
  c->justification = append_just(c->justification, merge_just(n));
}  /* merge1 */

/*************
 *
 *   proof_id_to_clause()
 *
 *************/

static
Clause proof_id_to_clause(Plist proof, int id)
{
  Plist p = proof;
  while (p && ((Clause) p->v)->id != id)
    p = p->next;
  if (p == NULL)
    return NULL;
  else
    return p->v;
}  /* proof_id_to_clause */

/*************
 *
 *   greatest_id_in_proof()
 *
 *************/

static
int greatest_id_in_proof(Plist proof)
{
  if (proof == NULL)
    return INT_MIN;
  else {
    int x = greatest_id_in_proof(proof->next);
    Clause c = proof->v;
    return (c->id > x ? c->id : x);
  }
}  /* greatest_id_in_proof */

/* #define DEBUG_EXPAND */

/*************
 *
 *   expand_proof()
 *
 *************/

/* DOCUMENTATION
Given a proof, which is assumed to be an ordinary Otter proof (not
a copy), return a more detailed proof, entirely new, leaving the
given proof unchanged.
*/

/* PUBLIC */
Plist expand_proof(Plist proof, Uniftype utype, Ilist *pmap,
		   BOOL lex_order_vars)
{
  Plist new_proof = NULL;  /* build it backward, reverse at end */
  int next_id = greatest_id_in_proof(proof) + 1;  /* for numbering new steps */
  Ilist map = NULL;      /* map new IDs to <old-id,n> for intermediate steps */
  int old_id, old_id_n;  /* for the map */
  Plist p;

  for (p = proof; p; p = p->next) {
    Clause c = p->v;
    Clause current = NULL;
    Just j;

    j = c->justification;
    old_id = c->id;
    old_id_n = 0;

#ifdef DEBUG_EXPAND
    printf("\nexpanding: "); fwrite_clause(stdout, c, TRUE, TRUE);
#endif
    if (j->next == NULL &&
	j->type != HYPER_RES_JUST &&
	j->type != UR_RES_JUST) {

      current = copy_clause_ija(c);  /* e.g., binary res with no secondaries */
      map = alist2_insert(map, next_id, old_id, old_id_n++);
      current->id = next_id++;  /* this will be reset below */
      new_proof = plist_prepend(new_proof, current);
    }
    else {
      /* To adjust literal numbers when literals disappear. */
      int merges = 0;
      int unit_deletes = 0;
      int xx_simplify = 0;
      /* primary inference */

      if (j->type == COPY_JUST ||
	  j->type == BACK_DEMOD_JUST ||
	  j->type == BACK_UNIT_DEL_JUST) {
	/* Note that we get "current" directly from the new proof.
	   This prevents an unnecessary "copy" inference. */
	current = proof_id_to_clause(new_proof, j->u.id);
      }
      else if (j->type == HYPER_RES_JUST ||
	       j->type == UR_RES_JUST ||
	       j->type == BINARY_RES_JUST) {
	/* c ncn ncn ncn ... (length is 3m+1) */
	Ilist p = j->u.lst;
	Clause c1 = proof_id_to_clause(proof, p->i);
	int j = 0;  /* literals resolved; subtract from nucleus position */
	p = p->next;
	while (p != NULL) {
	  Clause c3;
	  int n1 = p->i - j;  /* literal number in c1 */
	  int sat_id = p->next->i;
	  if (sat_id == 0)
	    c3 = xx_resolve2(c1, n1);
	  else {
	    Clause c2 = proof_id_to_clause(proof, sat_id);
	    int n2 = p->next->next->i;
	    c3 = resolve2(c1,n1,c2,n2);
	    if (c3 == NULL) {
	      printf("Lit %d: ",n1); fwrite_clause(stdout,c1,TRUE,TRUE);
	      printf("Lit %d: ",n2); fwrite_clause(stdout,c2,TRUE,TRUE);
	      fatal_error("expand_step, clauses don't resolve");
	    }
	  }
	  p = p->next->next->next;
	  map = alist2_insert(map, next_id, old_id, old_id_n++);
	  c3->id = next_id++;
	  new_proof = plist_prepend(new_proof, c3);
	  c1 = c3;
	  j++;
	}
	current = c1;
      }
      else if (j->type == PARA_JUST) {
	Clause from = proof_id_to_clause(proof, j->u.para->from_id);
	Clause into = proof_id_to_clause(proof, j->u.para->into_id);
	Ilist from_pos = j->u.para->from_pos;
	Ilist into_pos = j->u.para->into_pos;
	current = para_pos(from, from_pos, into, into_pos);  /* does just. */
	map = alist2_insert(map, next_id, old_id, old_id_n++);
	current->id = next_id++;
	new_proof = plist_prepend(new_proof, current);
      }
      else if (j->type == FACTOR_JUST) {
	Ilist p = j->u.lst;
	Clause parent = proof_id_to_clause(proof, p->i);
	int lit1 = p->next->i;
	int lit2 = p->next->next->i;

	current = factor(parent, lit1, lit2);
	map = alist2_insert(map, next_id, old_id, old_id_n++);
	current->id = next_id++;
	new_proof = plist_prepend(new_proof, current);
      }
      else if (j->type == XXRES_JUST) {
	Ilist p = j->u.lst;
	Clause parent = proof_id_to_clause(proof, p->i);
	int lit = p->next->i;

	current = xx_res2(parent, lit);
	map = alist2_insert(map, next_id, old_id, old_id_n++);
	current->id = next_id++;
	new_proof = plist_prepend(new_proof, current);
      }
      else if (j->type == NEW_SYMBOL_JUST) {
	Clause parent = proof_id_to_clause(proof, j->u.id);
	current = new_constant(parent);
	map = alist2_insert(map, next_id, old_id, old_id_n++);
	current->id = next_id++;
	new_proof = plist_prepend(new_proof, current);
      }
      else {
	printf("expand_step, unknown primary justification\n");
	new_proof = plist_prepend(new_proof, copy_clause_ija(c));
      }

#ifdef DEBUG_EXPAND
      printf("primary: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif

      /* secondary inferences */

      for (j = j->next; j; j = j->next) {
	if (j->type == DEMOD_JUST) {
	  /* list of pairs: ID, direction */
	  Ilist p;
	  for (p = j->u.lst; p; p = p->next->next) {
	    Clause demod = proof_id_to_clause(proof, p->i);
	    int direction = p->next->i;
	    Ilist from_pos, into_pos;
	    Clause work = copy_clause(current);
	    map = alist2_insert(map, next_id, old_id, old_id_n++);
	    work->id = next_id++;
	    demod1(work, demod, direction, &from_pos, &into_pos,
		   lex_order_vars);
	    work->justification = para_just(PARA_JUST,
					    demod, from_pos,
					    current, into_pos);
	    current = work;
	    new_proof = plist_prepend(new_proof, current);
#ifdef DEBUG_EXPAND
	    printf("demod: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
	  }
	}
	else if (j->type == FLIP_JUST) {
	  Term atom;
	  int n = j->u.id;
	  Clause work = copy_inference(current);
	  current = work;
	  map = alist2_insert(map, next_id, old_id, old_id_n++);
	  current->id = next_id++;
	  atom = ith_literal(current, n)->atom;
	  if (!eq_atom(atom))
	    fatal_error("expand_step, cannot flip nonequality");
	  flip_eq(atom, n);  /* updates justification */
	  new_proof = plist_prepend(new_proof, current);
#ifdef DEBUG_EXPAND
	  printf("flip: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
	}
	else if (j->type == MERGE_JUST) {
	  int n = j->u.id - merges;
	  Clause work = copy_inference(current);
	  current = work;
	  map = alist2_insert(map, next_id, old_id, old_id_n++);
	  current->id = next_id++;
	  merge1(current, n);  /* updates justification */
	  new_proof = plist_prepend(new_proof, current);
#ifdef DEBUG_EXPAND
	  printf("merge: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
	  merges++;
	}
	else if (j->type == UNIT_DEL_JUST) {
	  Ilist p = j->u.lst;
	  int n = p->i - unit_deletes;
	  Clause unit = proof_id_to_clause(proof, p->next->i);
	  Clause work = resolve2(unit,1,current,n);
	  if (work == NULL) {
	    printf("Lit %d: ",n); fwrite_clause(stdout,current,TRUE,TRUE);
	    printf("Lit %d: ",1); fwrite_clause(stdout,unit,TRUE,TRUE);
	    fatal_error("expand_step, clauses don't unit_del");
	  }
	  current = work;
	  map = alist2_insert(map, next_id, old_id, old_id_n++);
	  current->id = next_id++;
	  new_proof = plist_prepend(new_proof, current);
	  unit_deletes++;
#ifdef DEBUG_EXPAND
	  printf("unit_del: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
	}
	else if (j->type == XX_JUST) {
	  int n = j->u.id - xx_simplify;
	  Clause work = copy_inference(current);
	  current = work;
	  map = alist2_insert(map, next_id, old_id, old_id_n++);
	  current->id = next_id++;
	  xx_simp2(current,n);
	  new_proof = plist_prepend(new_proof, current);
	  xx_simplify++;
#ifdef DEBUG_EXPAND
	  printf("xx_simplify: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
	}
	else {
	  printf("expand_step, unknown secondary justification\n");
	  new_proof = plist_prepend(new_proof, current);
	}
      }

      renumber_variables(current, MAX_VARS);

#ifdef DEBUG_EXPAND
      printf("secondary: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif

      }

    /* Okay.  Now current should be identical to c. */
    
    if (current == c)
      fatal_error("expand_proof, current == c");
    else if (!clause_ident(current, c)) {
      fwrite_clause(stdout,c,TRUE,TRUE);
      fwrite_clause(stdout,current,TRUE,TRUE);
      fatal_error("expand step, result is not identical");
    }
    else {
      current->id = c->id;
      next_id--;
      map = alist2_remove(map, next_id);
#ifdef DEBUG_EXPAND
      printf("end: "); fwrite_clause(stdout, current, TRUE, TRUE);
#endif
    }  /* expand */
  }  /* process proof step c */
  *pmap = map;
  return reverse_plist(new_proof);
}  /* expand_proof */

/*************
 *
 *   renumber_proof()
 *
 *************/

/* DOCUMENTATION
We assume that every clause occurs after its parents.
*/

/* PUBLIC */
void renumber_proof(Plist proof, int start)
{
  Ilist map = NULL;         /* map old IDs to new IDs */
  int n = start;            /* for numbering the steps */
  Plist p;

  for (p = proof; p; p = p->next) {
    Clause c = p->v;
    int old_id = c->id;
    c->id = n++;
    map_just(c->justification, map);
    map = alist_insert(map, old_id, c->id);
  }
  zap_ilist(map);
}  /* renumber_proof */

/*************
 *
 *   proof_to_xproof()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist proof_to_xproof(Plist proof, BOOL lex_order_vars)
{
  Ilist map;
  return expand_proof(proof, ORDINARY_UNIF, &map, lex_order_vars);
}  /* proof_to_xproof */

