// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import java.util.*;
import java.net.URI;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;

import org.mathweb.atp.TSTP2X;
import org.mathweb.atp.TimeoutException;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/07/01 12:42:23 $ $Revision: 1.48 $
*/
public class ATP extends BasicService {

  protected boolean verbose=true;
 
  private int defaultTime = 60;
    
  protected TSTP2X tstp2X = new TSTP2X();
  protected X2TSTP x2tstp = new X2TSTP();
  
  protected OWLJDom owlJDom;
  protected OntoReasoner reasoner;
  
  public String name = "ATP";

  /* The MathServ ontology class that is accepted by the service as an input */
  public String acceptedConcept = "TstpProblem";
  
  /* The MathServ ontology class of the OWL individual that is returned by the service */
  public String resultConcept = "TstpCNFRefutation";
  
  /* The MathServ ontology class of the proof object. */
  public String proofConcept = "TstpCnfRefutation";

  /* The MathServ ontology class of the saturation object. */
  public String saturationConcept = "TstpCnfSaturation";
  
  /* The ID of the calculus individual in the MathServ ontology representing the calculus of the prover. */
  public String calculusConcept = "FOResolution";

  private static Logger logger = Logger.getLogger("ATP");
  private static String logFileName = getSystemProperty("atp.log.file", "ATP_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  
  private static int maxOutputLength = 80000;
  
  public ATP()
  {
    super();
    owlJDom = OWLJDom.getInstance();
    reasoner = PellReasoner.getNewReasoner();
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    Enumeration enum = logger.getAllAppenders();
//     for( ; enum.hasMoreElements() ;) {
//       System.out.println("appender: " +enum.nextElement());
//     }
    if (!enum.hasMoreElements()) logger.addAppender(appender);
  }

  
  /** 
   * @param problem a proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    return new String[] {"Failure", "Please override method ATP.callProver." };
  }

  /** 
   * @param provingProblem a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param timeRes the time resource given for the problem, i.e. an individual of the OWL class TimeResource as an XML string.
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
  public String proveProblem(String provingProblem, String timeRes) {
    
    /*  initialise local variables */
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    String probID = null;
    boolean valid = false; 
    int time = 0;
    
    //     System.out.println(provingProblem);
    
    try{
      /*  parse the XML document and get the JDOM root elements */
      Element probRoot = owlJDom.getRootElement(provingProblem);
      String xmlBase = probRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      Element timeRoot = owlJDom.getRootElement(timeRes);
      
      logger.info("... parsing, took me " + (System.currentTimeMillis() - startTime) + " msecs");
      logger.debug("[atp] : " + probRoot + " " + owlJDom);
      //        List cs = probRoot.getChildren();
      //        for(int i=0; i<cs.size();i++) logger.debug(cs.get(i));

      /*  look for potential rdf:IDs of the individual we are looking for. */
      List children = probRoot.getChildren();
      for(int i=0; i<children.size(); i++){
        Element c = (Element) children.get(i);
        String potentialID = owlJDom.valueOf(owlJDom.RDF_ID, c);
        if (potentialID != null) {
          probID = xmlBase + potentialID;
          break;
        }
      }

      if (probID == null) 
        return owlJDom.createFailure(xmlBase, startTime, cpuStart, "Problem has to be of class " + acceptedConcept + "!");
      logger.info("[ATP] prob ID: " + probID);
       
      /*  tell the individual to the ontology reasoner to be able to infer further properties. */
      long indiTime = System.currentTimeMillis();
      valid = reasoner.tellIndividual(provingProblem);
      logger.info("...telling individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
      //        startTime = System.currentTimeMillis();
      //        probID = reasoner.nameOfLastIndividual(acceptedConcept);
      //        logger.info("...looking for ID took me " + (System.currentTimeMillis() - startTime) + " msecs");
       
      //valid = reasoner.isValid();
      //       valid = true;
      //       logger.debug("valid OWL?: " + valid);
      //       logger.debug("right class?: " +reasoner.hasClass(probID,this.acceptedConcept) );
       
      // has the individual got the right type?
      if (valid){
        URI probURI = URI.create(probID);
        /*  ask the reasoner for the formal description of the proving problem. */
        String problem = reasoner.valueOf("formalDescription", probURI);
        logger.info("[ATP] formal: " + problem);

        /*  ask the reasoner for the language the proving problem is written in. */
        String language = reasoner.valueOf("language", probURI);
        logger.info("[ATP] lang: " + language);

        /*  ask the reasoner for  the format (CNF or FOF). */
        String format = reasoner.valueOf("format", probURI); 
        logger.info("[ATP] format: " + format);
         
        time = 0;
        Double num = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(timeRoot);
        if (!num.isNaN()) 
          time = num.intValue();

        /* if no time given, try to extract it from the proving problem */
        if (time == 0) { 
          String timeIndi = reasoner.valueOf("timeRes", probURI);
          logger.debug(timeIndi);
          if (timeIndi == null) 
            time = this.defaultTime;
          else {
            String timeStr = reasoner.valueOf("wallClockTime", timeIndi);
            time =  Integer.parseInt(timeStr);
          }
        } 
        logger.info("[ATP] time: " + time);
        
        String logic = reasoner.valueOf("inLogic", probURI);
        logger.info("[ATP] in logic: " + logic);
         
        /*-----------------------------------------------------*/
        /*  Call the ATP specific method in the subclass of ATP.java */
        /*-----------------------------------------------------*/
        String[] result = this.callProver(problem, language, format, time,  "");
        //         System.out.println(result);
         
        //TODO: extract result status
        String status = result[0];
        if (status == null) status = "Unknown";
        logger.info("[ATP] status : " + status);
        
        if ((format != null) && (format.equals(OWLJDom.MW_NS_STR+"FOF"))) {
          if (status.equals("Unsatisfiable"))
            status = "Theorem";
          else if (status.equals("Satisfiable")) 
            status = "CounterSatisfiable";
        }
        //logger.info("[atp] result: " + result);

        /*  Call the final result, i.e. an OWL individual with all the result info. */
        return createResult(probID, status, startTime, cpuStart, result[1], false);

      } else {
        return owlJDom.createFailure(probID, startTime, cpuStart, "Problem has to be valid RDF description of an individual of OWL class " +  acceptedConcept + "!");
      }
    } catch(TimeoutException ex){
      return createResult(probID, "Timeout", startTime, cpuStart, null, false);
    } catch(Exception ex){
      ex.printStackTrace();
      return owlJDom.createFailure(probID, startTime,  cpuStart,"Error: "+ex.getMessage());
    } finally {
      if ((valid)  && (probID !=null)){
        long indiTime = System.currentTimeMillis();
        reasoner.untellIndividual(probID);
        logger.info("deleting individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
      }
    }
  }
  
  /** 
   * @param provingProblem a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param timeRes the time resource given for the problem, i.e. an individual of the OWL class TimeResource as an XML string.
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class. All proof objects are returned in the XTSTP language, i.e. the XML version of TSTP.
   */
  public String proveProblemXml(String provingProblem, String timeRes) {
    
    /*  initialise local variables */
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    String probID = null;
    boolean valid = false; 
    int time = 0;
     
    //     System.out.println(provingProblem);
     
    try{
      /*  parse the XML document and get the JDOM root elements */
      Element probRoot = owlJDom.getRootElement(provingProblem);
      String xmlBase = probRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      Element timeRoot = owlJDom.getRootElement(timeRes);
       
      logger.info("... parsing, took me " + (System.currentTimeMillis() - startTime) + " msecs");
      logger.debug("[atp] : " + probRoot + " " + owlJDom);

      /*  look for potential rdf:IDs of the individual we are looking for. */
      List children = probRoot.getChildren();
      for(int i=0; i<children.size(); i++){
        Element c = (Element) children.get(i);
        String potentialID = owlJDom.valueOf(owlJDom.RDF_ID, c);
        if (potentialID != null) {
          probID = xmlBase + potentialID;
          break;
        }
      }
      if (probID == null) 
        return owlJDom.createFailure(xmlBase, startTime, cpuStart, "Problem has to be of class " + acceptedConcept + "!");
      logger.info("[ATP] prob ID: " + probID);
       
      /*  tell the individual to the ontology reasoner to be able to infer further properties. */
      long indiTime = System.currentTimeMillis();
      valid = reasoner.tellIndividual(provingProblem);
      logger.info("...telling individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
      //        startTime = System.currentTimeMillis();
      //        probID = reasoner.nameOfLastIndividual(acceptedConcept);
      //        logger.info("...looking for ID took me " + (System.currentTimeMillis() - startTime) + " msecs");
       
      //valid = reasoner.isValid();
      //       valid = true;
      logger.debug("valid OWL?: " + valid);
      //       logger.debug("right class?: " +reasoner.hasClass(probID,this.acceptedConcept) );
       
      // has the individual got the right type?
      if (valid){
        URI probURI = URI.create(probID);
        String problem = reasoner.valueOf("formalDescription", probURI);
        logger.info("[ATP] formal: " + problem);
        String language = reasoner.valueOf("language", probURI);
        logger.info("[ATP] lang: " + language);
        String format = reasoner.valueOf("format", probURI); 
        logger.info("[ATP] format: " + format);
         
        time = 0;
        Double num = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(timeRoot);
        if (!num.isNaN()) 
          time = num.intValue();

        /* if no time given, try to extract it from the proving problem */
        if (time == 0) { 
          String timeIndi = reasoner.valueOf("timeRes", probURI);
          logger.debug(timeIndi);
          if (timeIndi == null) 
            time = this.defaultTime;
          else {
            String timeStr = reasoner.valueOf("wallClockTime", timeIndi);
            time =  Integer.parseInt(timeStr);
          }
        } 
        logger.info("[ATP] time: " + time);
        
        String logic = reasoner.valueOf("inLogic", probURI);
        logger.info("[ATP] in logic: " + logic);
         
        String[] result = this.callProver(problem, language, format, time,  "");
        //         System.out.println(

        //TODO: extract result status
        String status = result[0];
        if (status == null) status = "Unknown";
        logger.info("[ATP] status : " + status);
         
        if ((format != null) && (format.equals(OWLJDom.MW_NS_STR+"FOF")))
            if (status.equals("Unsatisfiable"))
                status = "Theorem";
        
        String xtstp = tstp2X.tstp2XTstp(result[1]);
        //        System.out.println("[atp] xtstp: " + xtstp);

        return createResult(probID, status, startTime, cpuStart, xtstp, true);
      } else {
        return owlJDom.createFailure(probID, startTime, cpuStart, "Problem has to be valid RDF description of an individual of OWL class " +  acceptedConcept + "!");
      }
    } catch(TimeoutException ex){
      return createResult(probID, "Timeout", startTime, cpuStart, null, true);
    } catch(Exception ex){
      ex.printStackTrace();
      return owlJDom.createFailure(probID, startTime,  cpuStart,"Error: "+ex.getMessage());
    } finally {
      if ((valid)  && (probID !=null)){
        long indiTime = System.currentTimeMillis();
        reasoner.untellIndividual(probID);
        logger.info("deleting individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
      }
    }
  }
  

  /** 
   * @param result the result of a theorem prover in the Prolog-style TSTP format.
   * @return The status in the <a href="http://www.cs.miami.edu/~tptp/cgi-bin/DVTPTP2WWW/view_file.pl?Category=Documents&File=SZSOntology">SZS Status Ontology</a> if the result contains a status entry, null otherwise.
   */
  protected String extractStatusFromTSTP(String result){
    String[] lines = result.split("\n");
    for(int i=0; i<lines.length; i++){
      if (lines[i].length()>15){
        //        System.out.println(lines[i].substring(0,15));
        if (lines[i].substring(0,15).equals("% Result     : ")){
          String rest = lines[i].substring(15).trim();
          String[] split = rest.split(" ");
          if (split.length > 1)
            return split[0];
          else 
            return rest;
        }
      }
    }
    return null;
  }
  
  /** 
   * @param probID  a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param stat a valid status from the  <a href="http://www.cs.miami.edu/~tptp/cgi-bin/DVTPTP2WWW/view_file.pl?Category=Documents&File=SZSOntology">SZS Status Ontology</a> which unambiguously indentifies the result of the prover.
   * @param startTime the start wall-clock time measured at the beginning of the prover call.
   * @param cpuStart the start CPU time measured at the beginning of the prover call.
   * @param output the formal (proof) output of the theorem prover.  
   * @param xml a boolean indicating whether the output is an XML string or not. 
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class. 
   */
  public String createResult(String probID, String stat, long startTime, long cpuStart, String output, boolean xml) {
    try{
      /*  create the XML root element rdf:RDF */
      Element rdf = new Element("RDF", OWLJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      rdf.setAttribute("base", OWLJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);

      /*  create the FoAtpResult element */
      Element fo_atp_result = new Element(resultConcept,OWLJDom.MW_NS);
      //      String resultID = OWLJDom.MW_NS_STR+name+ "_" +System.currentTimeMillis();
      String resultID = name + "_" +System.currentTimeMillis();
      fo_atp_result.setAttribute("ID", resultID, OWLJDom.RDF_NS);
      Element result_for = new Element("resultFor",  OWLJDom.MW_NS);
      result_for.setAttribute("resource", probID, OWLJDom.RDF_NS);
      fo_atp_result.addContent(result_for);
      Element status = new Element("status",OWLJDom.MW_NS);
      status.setAttribute("resource", OWLJDom.STAT_NS_STR+stat, OWLJDom.RDF_NS);
      fo_atp_result.addContent(status);

      /* HACK to overcome the ERROR messages of X2TptpExtractor */
      if ((output.length() > 5) && (output.substring(0, 6).equals("ERROR:")))
        output = "% " + output;
      
      /*------------------------------------------------------------------------------------- */
      /*  If the status is known and it is Theorem or Unsatisfiable then add the proof object */
      /*------------------------------------------------------------------------------------- */
      if ((stat.equals("Theorem")) || (stat.equals("Unsatisfiable"))) {
        Element proof = new Element("proof",  OWLJDom.MW_NS);
        Element refutation = null;
        if (xml)
          refutation = new Element("X"+proofConcept,  OWLJDom.MW_NS);
        else 
          refutation = new Element(proofConcept,  OWLJDom.MW_NS);
        
        Element formal = new Element("formalProof",  OWLJDom.MW_NS);
        if ((output != null) && (output.length() < maxOutputLength)) {
          formal.addContent("\n"); 
          formal.addContent(output); 
        } else {
            File outputFile;
          if (xml) outputFile = File.createTempFile("atp-output", ".xml");
          else  outputFile = File.createTempFile("atp-output", ".tstp");
          writeFile(outputFile, output);
          String url = "file://" + outputFile.getAbsolutePath();
          if (xml){
             formal.addContent("<!-- " + url + " -->\n");
            formal.addContent("<!-- The output of the system was bigger than " + maxOutputLength + " characters and has therefore been written in a file available at the above URL. -->");
          } else {
            formal.addContent("% " + url + "\n");
            formal.addContent("%  The output of the system was bigger than  " + maxOutputLength + " characters and has therefore been written in a file available at the above URL.");
          }
        }
        
        refutation.addContent(formal);
        Element calc = new Element("calculus",  OWLJDom.MW_NS);
        calc.setAttribute("resource", OWLJDom.MW_NS_STR+calculusConcept, OWLJDom.RDF_NS);
        Element proof_of = new Element("proofOf",  OWLJDom.MW_NS);
        proof_of.setAttribute("resource", probID, OWLJDom.RDF_NS);
        refutation.addContent(calc);
        refutation.addContent(proof_of);
        proof.addContent(refutation);
        fo_atp_result.addContent(proof);
      } 
      /*---------------------------------------------------------------------------------------------- */
      /*  If the status is CounterSatisfiable, the the prover might have found a saturation derivation */
      /*---------------------------------------------------------------------------------------------- */
      else if (stat.equals("CounterSatisfiable")) {
        Element saturation = new Element("saturation",  OWLJDom.MW_NS);
        Element derivation = null;
        if (xml)
          derivation = new Element("X"+saturationConcept,  OWLJDom.MW_NS);
        else 
          derivation = new Element(saturationConcept,  OWLJDom.MW_NS);
        
        Element formal = new Element("formalSaturation",  OWLJDom.MW_NS);
        if ((output != null) && (output.length() < maxOutputLength)) {
          formal.addContent("\n"); 
          formal.addContent(output); 
        } else {
//           File outputFile;
//           if (xml) outputFile = File.createTempFile("atp-output", ".xml");
//           else  outputFile = File.createTempFile("atp-output", ".tstp");
//           writeFile(outputFile, output);
//           String url = "file://" + outputFile.getAbsolutePath();
          if (xml){
            //              formal.addContent("<!-- " + url + " -->\n");
            formal.addContent("<!-- The output of the system was bigger than " + 
                              maxOutputLength + 
                              " characters and has therefore been written in a file available at the above URL. -->");
          } else {
            //             formal.addContent("% " + url + "\n");
            formal.addContent("%  The output of the system was bigger than  " + 
                              maxOutputLength + 
                              " characters and has therefore been written in a file available at the above URL.");
          }
        }
        
        derivation.addContent(formal);
        Element calc = new Element("calculus",  OWLJDom.MW_NS);
        calc.setAttribute("resource", OWLJDom.MW_NS_STR + calculusConcept, OWLJDom.RDF_NS);
        Element saturation_of = new Element("saturationOf",  OWLJDom.MW_NS);
        saturation_of.setAttribute("resource", probID, OWLJDom.RDF_NS);
        derivation.addContent(calc);
        derivation.addContent(saturation_of);
        saturation.addContent(derivation);
        fo_atp_result.addContent(saturation);
      } 
      
      /*  Add the TimeResource telling the cpu and wc time used up */
      Element time = new Element("time",  OWLJDom.MW_NS);
      Element timeres = OWLJDom.createTimeResource(startTime, cpuStart);
      time.addContent(timeres);
      fo_atp_result.addContent(time);
      rdf.addContent(fo_atp_result); 
      Element newTime = (Element) timeres.clone();
      rdf.addContent(newTime); 
      return owlJDom.newXMLDocument(rdf);
    }
    catch(Exception e){
      e.printStackTrace();
      return owlJDom.createFailure(probID, startTime, cpuStart, e.getMessage());
    }
  }
  
  /** 
      Shuts down the Web Service properly.
   */
  public void shutdown()
  {
    return;
  }

} // end Atp.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


// old JDom stuff
//        Element probRoot = owlJDom.getRootElement(provingProblem);
//        logger.info("... parsing, took me " + (System.currentTimeMillis() - startTime) + " msecs");
//        logger.debug("[atp] : " + probRoot + " " + owlJDom);
// //        List cs = probRoot.getChildren();
// //        for(int i=0; i<cs.size();i++) logger.debug(cs.get(i));
       
//        Element prob = (Element) owlJDom.TSTP_FOF_PROBLEM.selectSingleNode(probRoot);
//        logger.debug("[atp] : " + prob);
//        if (prob == null) 
//          prob = (Element) owlJDom.TSTP_CNF_PROBLEM.selectSingleNode(probRoot);
//        logger.debug("[atp] : " + prob);
//        if (prob == null) 
//          return owlJDom.createFailure("Problem has to be of class TstpFOFProblem or TstpCNFProblem!");
       
//        probID = owlJDom.valueOf(owlJDom.RDF_ID, prob);
//        logger.debug("problem ID: " + probID);
       
       // add the individual as an assertion to the OWL knowledge base
         // "./mw:TimeResource/mw:wallClockTime");
         
         //      logger.debug("prob: " + tstpFOFProblem);
         //         String timeStr = owlJDom.valueOf(owlJDom.PROVING_PROBLEM_TIME_RESOURCE, prob);
         //         String problem = owlJDom.valueOf(owlJDom.PROVING_PROBLEM_FORMAL_DESCRIPTION, prob);
