// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import java.text.ParseException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import edu.cmu.apros.language.*;
import edu.cmu.apros.theory.Assertion;
import edu.cmu.apros.generator.tieralgorithm.TierAlgorithm;

import edu.cmu.apros.proof.Proof;
import edu.cmu.apros.proof.RuleApp;
import edu.cmu.apros.proof.PartialProof;
import edu.cmu.apros.proof.rule.*;
import edu.cmu.apros.proof.util.Occurrence;
import edu.cmu.apros.proof.util.Scope;
import edu.cmu.apros.proof.rule.RuleException;

import edu.cmu.apros.translation.omdoc.*;

import nu.xom.*;

import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;
import org.mathweb.atp.TSTP2X;
import org.mathweb.atp.TimeoutException;

public class Apros extends ATP {

  public OmdocTranslator omdocReader = new OmdocTranslator();
  public OmdocRenderer   omdocWriter = new OmdocRenderer();
  public String logFileName = getSystemProperty("apros.log.file", "APROS_LOG_FILE");
  public static Logger logger = Logger.getLogger("Apros");
  
  public Apros(){
    super();
    name = "Apros";
    acceptedConcept = "FOProvingProblem";
    resultConcept   = "NDProverResult";
    proofConcept    = "NDProof";
    calculusConcept = "AprosNDCalculus";

    logger.setLevel(Level.ALL);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {
      System.out.println("exception creating logger: "+e);
    }
    logger.addAppender(appender);
  }
  
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    try{
      Assertion ass = null;
      if (language.equals("OMDoc")) {
	ass = getConjecture(problem);
      } else 
	if (language.equals("TSTP")){
	  String omdoc = tstp2X.tstp2OMDoc(problem);
          logger.debug("[Apros] omdoc " + omdoc);
          writeFile(msHome + "/etc/data/omdoc/apros.omdoc", omdoc);
	  ass = getConjecture(omdoc);
	} else{
	  throw new Exception("Input format "+language+" not known or not supported.");
	}
      if (ass != null) {
	TierAlgorithm generator = new TierAlgorithm(false);
        logger.debug("OMDoc assertion: " + ass);
	//	  Proof proof = generator.prove(ass);
	generator.initialize(ass);
	
	long startTime = System.currentTimeMillis();
	boolean proofFound = false;
	boolean timeout = false;
	generator.start();
	while (! generator.isProofFound() && !timeout){
	  if (((System.currentTimeMillis() - startTime) / 1000) > time)
	    timeout = true;
	}
	if (timeout){
          generator.requestStop();
	  return new String[] {"Timeout", ""};
	}
	if (generator.isProofFound()) {
	  Proof proof = generator.getProof();
	  String xml = writeProof(proof, ass.getName());
	  return new String[] {"Theorem", xml};
	}
        throw new TimeoutException("Apros could not find a proof within " + time + " seconds.");
      } else {
	throw new Exception("Could not read omdoc problem.");
      }
    } catch(Exception e){
      logger.debug("[Apros] Exception " + e);
      throw e;
    }
  }
  
    
  public Assertion getConjecture(File inputFile) {
    String omdoc = readFile(inputFile);
    return getConjecture(omdoc);
  }
  
  public Assertion getConjecture(String omdoc) {
    try {
      Builder builder = new Builder(false);
      Document document = builder.build(new StringReader(omdoc));
      
      Element omd = document.getRootElement();
      logger.debug("omd " + omd);
      Assertion ass = omdocReader.parseAssertion(omd);
      logger.debug("Ass " + ass.getName());
      return ass;
    }catch (java.text.ParseException e){
      logger.debug("Parse error at line " + e);
    }
    catch (nu.xom.ParseException e) {
      logger.debug("Parse error at line " + e.getLineNumber()
			 + " column " + e.getColumnNumber());
      logger.debug("\nCause: " + e.getCause());
    }
    catch (IOException e) {
      logger.debug("Could not load that file. Stack trace follows.");
      e.printStackTrace();
    }
    return null;
  }

  public String writeProof(Proof proof, String proofOf)
  {
    try {
      proof.check();
      Element proofE = omdocWriter.getElement(proof);
      Attribute proofFor = proofE.getAttribute("for");
      proofE.removeAttribute(proofFor);
      proofFor.setValue(proofOf);
      proofE.addAttribute(proofFor);
      
      Element omdoc = new Element("omdoc", OmdocParser.OMDOC_NS);
      omdoc.insertChild(proofE,0);
      Document doc = new Document(omdoc);
      //      doc.setDocType(new DocType("omdoc", msHome + "/etc/data/import/mathweb/omdoc/dtd/omdoc.dtd"));
      
      String xml;
      ByteArrayOutputStream writer = new ByteArrayOutputStream();
      Serializer s = new Serializer(writer);
      s.setLineSeparator("\n");
      s.setIndent(2);
      s.write(doc);
      return writer.toString();
      // return doc.toXML();
    } catch (RuleException e) {
      logger.debug("in rule exception");
      e.printStackTrace();  //To change body of catch statement use Options | File Templates.
      return "";
    }
    catch (IOException e) {
      logger.debug("I/O exception");
      e.printStackTrace();  //To change body of catch statement use Options | File Templates.
      return "";
    }
  }
  

  public static void main(String[] args) {
    String msHome = System.getProperty("mathserv.home");
    Apros apros = new Apros();
    try {
      System.out.println("Apros server!");
      String ind;
      String result;
      
      ind = apros.readFile(msHome + "/examples/owl/fof2.owl");
      //      ind = apros.readFile(msHome + "/examples/owl/omdoc1.owl"); 
       //     ind = apros.readFile(msHome + "/examples/owl/fof.owl"); 
      OWLJDom owlJDom = OWLJDom.getInstance();
      String timeInd = owlJDom.createTimeIndi(8, 8);
      result = apros.proveProblem(ind, timeInd);
//        ind = apros.readFile(msHome + "/etc/data/omdoc/agatha-fof.omdoc"); 
//       result =  apros.callProver(ind, "OMDoc", "FOF", 200 , "");
      
      System.out.println("result " + result);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }

//     String inputFileName = msHome + "/etc/data/omdoc/b.omdoc";
//     File inputFile = new File(inputFileName);
//     Assertion assertion = apros.getConjecture(inputFile);
//     System.out.println("OMDoc assertion: " + assertion);
    
//     TierAlgorithm generator = new TierAlgorithm(false);
    
//     Proof proof = generator.prove(assertion);
    
//     String xml = apros.writeProof(proof, assertion.getName());
//     apros.writeFile(msHome + "/etc/data/omdoc/b-proof.omdoc", xml);
//     logger.info(xml);
    
  }
  
}//end class Apros
  
