// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.*;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/07/15 19:11:52 $ $Revision: 1.41 $
*/
public class E extends ATP {

  protected boolean verbose=true;

  private String E_HOME = getSystemProperty("e.home", "E_HOME");
  private String[] e_envs = new String[] {"E_HOME="+E_HOME};
  private String eExecutable = findSystemExecutable(E_HOME, "eproof", "EProof.exe");

  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Ep");
  private static String logFileName = getSystemProperty("e.log.file", "E_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
  
  public E()
  {
    super();
    name = "Ep_0_82";
    /* The MathServ ontology class that is accepted by the service as an input */
    acceptedConcept = "TstpProblem";
    /* The MathServ ontology class of the OWL individual that is returned by the service */
    resultConcept = "FoAtpResult";
    /* The MathServ ontology class of the proof object. */
    proofConcept = "TstpCnfRefutation";
    /* The ID of the calculus individual in the MathServ ontology representing the calculus of the prover. */
    calculusConcept = "ep_res_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }  
  
  /** 
   * @param problem a proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    File origFile = null;
    File inputFile = null;
    File outputFile = null;
    try{
      //      origFile = File.createTempFile("eprover", ".orig");
      inputFile = File.createTempFile("eprover", ".in");
      outputFile = File.createTempFile("eprover", ".out");
      
      //      writeFile(origFile, problem);
      
      if ((problem != null) && (problem.length() > 2000000))
        return  new String[] {"Unknown", ""};
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        
        /*  first transform the problem into TPTP CNF form without include statements */
        String toSyntax= "tptp";
        String transform = "none";
        String newProb;

        newProb = tstp2X.tstp2X(problem, toSyntax, transform);
        logger.info("new problem: " + newProb);
        
        /*  write the new problem description in the input file */
        writeFile(inputFile,newProb);
        
        /*  create the shell command to call the prover and execute it */
        //        String eOptions = " --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out --cpu-limit=" + time + " ";
        String eOptions = " --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out ";
        //        String eOptions = " -s --print-statistics -xAuto -tAuto --tstp-in --tstp-out -l4 --memory-limit=100 --cpu-limit=" + time + " ";        

        String command = eExecutable + " " +  eOptions + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        
        logger.info("[E] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.info("[E] starting prover...");
        this.start(command, e_envs, defaultNice, time);
        
        /*  read the output written by the prover call */
        //        logger.info("[E] reading result..." + outputFile.getAbsolutePath());
        String result = this.readFile(outputFile);
        if ((result != null) && (result.length() > 1000000)){
          System.out.println("length: " + result.length());
          return  new String[] {"Unknown", ""};
        }
        
        //logger.info("[E] result: " + result);
        //         logger.info("[E] transforming to TSTP...");
        //         String tstp = x2tstp.x2tstp(result, "E---0.82");
        //         logger.info("[E] TSTP..." + tstp);
        //         return tstp;
        
        /*  split the result in lines and examine them to extract the status */
        String[] split = result.split("\n");
        StringBuffer buf = new StringBuffer();
        String status = null;
        
        for(int i=0; i<split.length; i++){
          //logger.info("[line] " + split[i]);
          if (split[i].length()>2){
            //              logger.info("[line] " + split[i].substring(0,2));
            //              System.out.println("status " + split[i].substring(0,20));
            if ((split[i].substring(0,20).equals("# TSTP exit status: "))){
              //              System.out.println("found status ");
              status = split[i].substring(20);
            }
            
            if ((split[i].substring(0,3).equals("cnf"))){
              buf.append(split[i]);
              buf.append("\n");
            }
          }
        }
        if (status == null) status = "Unknown";
        else if (status.equals("Saturation")) status = "CounterSatisfiable";
        else if (status.equals("Refutation")) status = "Unsatisfiable";
        //        System.out.println("status " + status);
        
        /*  create and return the resulting String array */
        String[] resArray = new String[] {status, buf.toString() };
        //        System.out.println("output " + resArray);
        return resArray;
      } else{
        throw new Exception("Input language "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[E] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[E] Exception " + e); }
    }
    
  }
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    E e = new E();
    OWLJDom owlJDom = OWLJDom.getInstance();
    TSTPCreator creator = new TSTPCreator();

    String timeInd = owlJDom.createTimeIndi(20,20);
    //       String problem = e.readFile(servF1);
    //String servF1 = msHome + "/examples/owl/cnf.owl";
    String servF1 = msHome + "/examples/owl/fof3.owl";
    String ind = e.readFile(servF1);
    
    //    timeInd =  e.readFile(msHome + "/examples/owl/time.owl");
    
    //     ind = e.readFile(msHome + "/examples/owl/cnf3.owl");
    //     ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/LCL/LCL181-3.p");
    //    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
    //    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP123+1.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN052+1.p");
    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG104+1.p");
    try {
      //String language = "TSTP";
      //String result = e.callProver(ind, language, 300, "");
      String result = e.proveProblem(ind, timeInd);
      //String result = e.proveProblemXml(ind, timeInd);
      e.writeFile(msHome + "/examples/owl/e-proof2.owl", result);
      
      System.out.println("result: " + result);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  }
  

} // end E.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


    /**   Special settings for non-auto mode without splitting!: */
//     eOptions = " --tptp-in --tstp-out --forward-context-sr -l4 --print-statistics --destructive-er-aggressive --destructive-er -winvfreqrank -c1 -Ginvfreq -F1 -WSelectMaxLComplex --prefer-initial-clauses -H'(4*Refinedweight(SimulateSOS,1,1,2,1.5,2),3*Refinedweight(PreferNonGoals,1,1,2,1.5,1.5),1*Clauseweight(PreferProcessed,1,1,1),1*FIFOWeight(PreferProcessed))' --memory-limit=192 --delete-bad-limit=150000000 --cpu-limit=" + time + " ";

//       /**
// --forward-context-sr
// --destructive-er-aggressive
// --destructive-er
// -winvfreqrank -c1 -Ginvfreq
// -F1
// -WSelectMaxLComplex
// --prefer-initial-clauses
// -H'(4*Refinedweight(SimulateSOS,1,1,2,1.5,2),3*Refinedweight(PreferNonGoals,1,1,2,1.5,1.5),1*Clauseweigh\
// t(PreferProcessed,1,1,1),1*FIFOWeight(PreferProcessed))'
// --memory-limit=192
// --delete-bad-limit=150000000
//       */

        /*  Command:   eproof --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out %s*/
