// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import com.hp.hpl.jena.util.*;
import com.hp.hpl.jena.vocabulary.ReasonerVocabulary;

import org.jdom.*;

import org.mathweb.onto.*;
import org.mathweb.serv.*;
import org.mathweb.broker.*;
import org.mathweb.owls.*;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.9 $
*/
public class Zenon extends ATP {
  
  public String ZENON_HOME = getSystemProperty("zenon.home", "ZENON_HOME");
  public static String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String[] zenon_envs = new String[] {"ZENON_HOME="+ZENON_HOME};
  public String zenonExecutable = findSystemExecutable(ZENON_HOME, "zenon", "Zenon.exe");

  private static Logger logger = Logger.getLogger("Zenon");
  private static String logFileName = getSystemProperty("zenon.log.file", "ZENON_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
    
  private static String sep = File.separator;

  private static String PROOF_FOUND = "(* PROOF-FOUND *)";
  private static String BEGIN_PROOF = "(* BEGIN-PROOF *)";
  private static String END_PROOF = "(* END-PROOF *)";
  private static String NO_PROOF = "(* NO-PROOF *)";
  private static String TIMEOUT1 = "time limit exceeded";
  private static String TIMEOUT2 = "FINAL WATCH";
  
  private static final int defaultNice = 15;
  
  public Zenon()
  {
    super();
    name = "Zenon_0_3_1";
    acceptedConcept = "TstpFOFProblem";
    resultConcept = "FOScriptResult";
    proofConcept = "CoqProofScript";
    calculusConcept = "zenon_calc";
    
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      inputFile = File.createTempFile("zenon", ".in");
      outputFile = File.createTempFile("zenon", ".out");
      
      logger.debug("problem: " + problem);
      //      System.out.println(inputFile + " " + outputFile);
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        String newProb = tstp2X.tstp2X(problem, 
                                       "oldtptp",
                                       "add_equality:r");
        
        //        System.out.println("new prob " + newProb);
        logger.debug("Zenon problem: " + newProb);

        writeFile(inputFile, newProb);
        
        /* zenon -max-time 300s -ocoq -itptp /project/omega/atp/TPTP-v2.7.0/Problems/GRP/GRP194+1.p */
        /* zenon -p2 -max-time 10s -o out -ocoq -itptp /project/omega/atp/TPTP-v2.7.0/Problems/ALG/ALG092+1.p */
        
        time = time - 1;
        String zenonOptions = " -p0 -ocoq -itptp -max-time";
        String command = zenonExecutable + " " +  zenonOptions + " " + time + "s " 
          + inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[ZENON] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[ZENON] starting prover...");
        this.start(command, zenon_envs, defaultNice, time);
        logger.debug("reading result..." + outputFile.getAbsolutePath());
        String result = this.readFile(outputFile);
        System.out.println(" result..." + result);
        writeFile(msHome + "/etc/data/zenon" + outputFile.getAbsolutePath(), result);
        
        boolean proof = false;
        StringBuffer proofObj = new StringBuffer();
        String status = "Unknown";
        String[] lines = result.split("\n");
        if ((lines.length > 0) && (lines[0].indexOf(TIMEOUT2) > -1))
          return new String[] {"Timeout", ""};
        for(int i=0; i<lines.length; i++){
          if (lines[i].indexOf(TIMEOUT1) > -1)
            return new String[] {"Timeout", ""};
          if (lines[i].indexOf(BEGIN_PROOF) > -1){
            proof = true;
            status = "Theorem";
          } else if (lines[i].indexOf(END_PROOF) > -1){
            proof = false;
          } else if (proof){
            proofObj.append(lines[i]);
            proofObj.append("\n");
          }
        }
        return new String[] {status, proofObj.toString() };
      }
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Zenon] Exception " + e);
      throw e;
    } finally{
      try{ 
        //        if (inputFile != null) inputFile.delete();
        // if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Zenon] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
      Zenon zenon = new Zenon();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      int time = 60;
      
      String filename = argv[0];
      //      filename = msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG017+1.p";
      String owlProblem = creator.tstp2OWLProblem(filename);

      if (argv.length > 1) {
        time = Integer.parseInt(argv[1]);
      } 
      
      String timeInd = owlJDom.createTimeIndi(time,time);
      
      String result = zenon.proveProblem(owlProblem, timeInd);
      System.out.println("result: " + result);
      

//       Calendar rightNow = Calendar.getInstance();
//       String outputFileName = msHome + "/thirdParty/TPTPWorld/rawdata/zenon-"  + spcName+ "-" + time;
//       FileOutputStream outputFile = null; 
      
      
//       String spcFileName = msHome + "/thirdParty/TPTPWorld/NewSPCLists/"+spcName;
      
//       try{
//         outputFile = new FileOutputStream(outputFileName) ;
//         String[] spcLines = zenon.readFile(spcFileName).split("\n");
//         int problems = spcLines.length;
//         String[] probNames = new String[problems];
//         String[] tstpFileNames = new String[problems];
//         for(int i = 0; i< problems; i++){
//           String line = spcLines[i];
//           String[] parts = line.split("   ");
//           if(parts.length>1){
//             String prob = parts[0].trim();
//             String group = prob.substring(0,3);
//             probNames[i]= prob;
//             tstpFileNames[i] = TPTP_HOME+ "/TPTP/Problems/"+ group + "/" + prob + ".p";
//             String owlProblem = creator.tstp2OWLProblem(tstpFileNames[i]);
            
//             String result = zenon.proveProblem(owlProblem, timeInd);
//             Element resultRoot = owlJDom.getRootElement(result);
//             Double wcDouble = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(resultRoot);
//             Double cpuDouble = (Double) OWLJDom.TIME_RESOURCE_CPU.numberValueOf(resultRoot);
//             String statusURI = OWLJDom.STATUS.valueOf(resultRoot);
//             String status = null;
            
//             if ((statusURI == null) || (statusURI.equals(""))) 
//               status = "Unknown";
//             else {
//               String[] answerSplit = statusURI.split("#");
//               if (answerSplit.length > 1) 
//                 status = answerSplit[1].trim();
//               else 
//                 status = statusURI;            
//             }
            
            
//             System.out.println("status: " + status);
//             String content = prob + " " + status + " " + (cpuDouble.doubleValue()/1000.0) + " " + (wcDouble.doubleValue()/1000.0) + "\n";
//             outputFile.write(content.getBytes());
//             outputFile.flush();
//           }
//         }
//       }catch(Exception ex) {
//         ex.printStackTrace();
//       }finally{
//         try{
//           outputFile.close();
//         }catch(Exception ex) {
//           ex.printStackTrace();
//         }
//       }
  }
  
  
  
            //            ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
          //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/PUZ/PUZ001-1.p");
      //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG001-2.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRA/GRA001-1.p");
//       ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SET/SET062+3.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
      
      //      System.out.println(ind);
      
//       try {
//         System.out.println(timeInd);
//         String result = null;
// //         for (int i=0; i< 700; i++) 
// //         {
// //             System.out.println(" call " + i);            
// //             e = new Zenon();
//         result = e.proveProblem(ind, timeInd);
//         //result = e.proveProblemXml(ind, timeInd);
//         System.out.println(result);
        
//         //e.writeFile(msHome + "/examples/owl/zenon-proof2.owl", result);
        
//         //        logger.info("result1: " + result);
        
//         //         result = e.proveProblem(ind);
        
//         //        logger.info("result2: " + result);
//         System.out.println(result);
//       }
//       catch(Exception ex){
//         ex.printStackTrace();
//         System.out.println(ex);
//       }
//   }
  
  
} // end Zenon.java


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
