// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:

package org.mathweb.broker;

import java.util.*;

import salvo.jesus.graph.*;
import salvo.jesus.graph.visual.*;
import salvo.jesus.graph.visual.layout.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.AtomicProcess;
import org.mindswap.owls.service.Service;
import org.mindswap.query.ValueMap;
// new 1.1.0
import org.mindswap.owl.OWLDataValue;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.*;
import org.mathweb.graph.*;
import org.mathweb.broker.ServiceEntry;

import org.mathweb.reasoning.logic.*;

class ServiceExecutor extends BasicService 
{
  private  Logger logger = Logger.getLogger("ServiceExecutor");
  private static OWLJDom owlJDom;
  private OntoReasoner reasoner;

  public ServiceExecutor()
  {
    String msHome = System.getProperty("mathserv.home");
    owlJDom = OWLJDom.getInstance();
    reasoner =  PellReasoner.getInstance();

    logger.setLevel(Level.ALL);
    logger.setLevel(Level.ALL);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(),msHome + "/logs/ServiceExecutor.log",false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }
  

  /*
    Overcome! NOTE: 1) We can NOT handle the case where the query has two inputs with the same type. They are currently indistinguishable for the ServiceExecutor!
          2) We can NOT handle services with several outputs right now!
          3) This is why outputValues maps (only) node names to values
  */  
  public String executeGraph(DirectedGraph graph, Query query, boolean checkPreconds) throws Exception {
    
    long globalStartTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    
    if(checkPreconds) reasoner =  PellReasoner.getNewReasoner();

    logger.info(graph);
    // find the start and finish vertex of the graph
    VertexImpl startVertex = null;
    VertexImpl finishVertex = null;
    int numberServices = 0;
    
    for(Iterator it = graph.getVerticesIterator(); it.hasNext();){
      VertexImpl v = (VertexImpl) it.next();
      //      logger.info("looking for start: " + v + " " +v.getLabel().substring(0,5));
      if ((v.getLabel().length() > 4) && (v.getLabel().substring(0,6).equals("Source"))) {
        startVertex = v; 
      } else if ((v.getLabel().length() > 3) && (v.getLabel().substring(0,4).equals("Sink"))) {
        finishVertex = v; 
      } else numberServices +=1;
    }
    
    logger.info("start vertex: " + startVertex);
    logger.info("finish vertex: " + finishVertex);
    logger.info("number of services: " + numberServices);
    if (startVertex == null) {
      logger.info("couldn't find source node in execution plan!");
      throw new Exception("Plan executor could not find source node in execution plan!");
    }
    
    /* initialize the list of output values given by the Query
       by giving them the name of the Start node + the type (signature) of the input.
    */
    // executed = {Start_1-> startVertex}
    Hashtable executed = new Hashtable();
    executed.put(startVertex.getLabel(), startVertex);
    

    ReasoningProblem queryProblem = query.getProblem();
    List queryInputs = queryProblem.getInputs();
    List queryOutputs = queryProblem.getOutputs();

    // outputValues = { Start -> query.in0}
    Hashtable outputValues = new Hashtable();
    for(int i=0; i < queryInputs.size(); i++){
      Input in = (Input) queryInputs.get(i);
      String iName = in.getName();
      outputValues.put(startVertex.getLabel()+iName, in.getContents());
    }
    logger.info("outputs: " + outputValues);
    
    // reached = {}
    Hashtable reached = new Hashtable();
    
    Hashtable edges = new Hashtable();
    for(Iterator it = graph.getEdgeSet().iterator(); it.hasNext();){

      Edge e = (Edge) it.next();
      edges.put(e.getVertexB()+"."+e.getLabel(), e);
    }
    logger.info("initial edges: " + edges);
    
    //    ArrayList nextEdges = new ArrayList();
    Hashtable inputValues = new Hashtable();
    Hashtable finalResults = new Hashtable();
    
    // while reached <> {Finish} do 
    while(!((reached.size()== 1) && (reached.containsKey(finishVertex.getLabel())))){
//       logger.info("\nnew loop!");
//       logger.info("reached: "+ reached);
//       logger.info("executed: " +executed);
      for (Enumeration edgeKeys =  edges.keys(); edgeKeys.hasMoreElements() ;) {
        String edgeKey = (String) edgeKeys.nextElement();
        AnnotatedEdge e = (AnnotatedEdge) edges.get(edgeKey);
        logger.debug("edge: " + e + ":" + e.getVertexA() + " -> " + e.getVertexB());
        String edgeLabel = e.getLabel();
        VertexImpl v1 = (VertexImpl) e.getVertexA();
        // forall e\in edges /\ {e\in edges | e.A \in executed} do 
        if (executed.containsKey(v1.getLabel())){
          logger.info("edge " + e + " comes out of " + v1.getLabel() + " and that node is executed");
          
          // remove the edge from the set of edges
          VertexImpl v2 = (VertexImpl) e.getVertexB();
          edges.remove(edgeKey);
          // newly reached vertex v2
          logger.info(" edge ends in " + v2.getLabel());
          reached.put(v2.getLabel(),v2);
          
          // try to assign values to input parameters of the service 
          // if it is not the Finish node.
          if (! (v2.getLabel().equals(finishVertex.getLabel()))){
            // now assign values to ALL the ReasoningService inputs
            ReasoningService s = (ReasoningService) v2.getObject();
            List inputs = s.getProblem().getInputs();
            for(int i=0; i<inputs.size(); i++){
              Input in = (Input) inputs.get(i);
              String iName = in.getName();
              // edgelabel is of form  input_name::OWLClass
              String edgeSource = edgeLabel;
              String edgeTarget = "";
              String[] st = edgeLabel.split("::");
              if (st.length > 1){ edgeSource = st[0]; edgeTarget = st[1]; }
              else 
                return owlJDom.createFailure(query.name, globalStartTime, cpuStart,  "The execution graph is not valid.");
              
              // if input_name of the edge is the name of an input of 
              // ReasoningService s, assign the right outputValue
              if (iName.equals(edgeTarget)){
                String value = null;
                Input queryIn;
                Annotation ann = e.getAnnotation();
                System.out.println("annotation: " + "(" + ann.getSource() + ", " +  ann.getTarget() + ")");
                // now find the right output value for that OWLClass:  
                //              logger.info("retrieving output of node " +v1.getLabel());
                value = (String) outputValues.get(v1.getLabel()+ann.getSource());
                if (value == null) {
                logger.debug("couldn't assign value to input " + edgeLabel + " of service instance " + v2.getLabel());
                throw new Exception("Plan executor could not assign values to all the input parameters of service instance " + v2.getLabel());
                } else {
                  // put the input value in the invokation table (key is the node label + input name):
                  inputValues.put(v2.getLabel() + iName, value);
                }
              }
            }  // end of assignment to ReasoningService inputs
            // else this must be an edge to the Finish node, so we simply store the final result
          } else {
            Annotation ann = e.getAnnotation();
            String outParamName = ann.getSource();
            String queryOutName = ann.getTarget();
            finalResults.put(queryOutName, outputValues.get(v1.getLabel()+outParamName));
          }
        } // end of assignment of edges coming out of the fringe of "executed" nodes
      } // end of  while(it.hasNext()) over edges
      
 
      /* for all vertices (services) reached from the executed ones check whether 
         all inputs of the service are bound in the inputValues
         if so, invoke the service and store the result:
      */
      //      logger.info("reached before execution: "+ reached);
      if (finalResults.size() < queryOutputs.size()){
        for (Enumeration nodeKeys =  reached.keys(); nodeKeys.hasMoreElements() ;) {
          String label = (String) nodeKeys.nextElement();
          VertexImpl v = (VertexImpl) reached.get(label);
          ReasoningService serv = (ReasoningService) v.getObject();
          List inputs = serv.getProblem().getInputs();
          List outputs = serv.getProblem().getOutputs();
          Service owlsService = serv.getOwlsService();
          AtomicProcess process = (AtomicProcess) owlsService.getProcess();

          // we assume that the service is executable
          boolean executable = true;
          // a new hashtable for the actual invokation
          Hashtable invocationTable = new Hashtable();
          String value;
          
          // now assign values to ALL the ReasoningService inputs
          for(int k=0; k<inputs.size(); k++){
            Input in = (Input) inputs.get(k);
            String iName = in.getName();
            if (inputValues.containsKey(label + iName)){
              value = (String) inputValues.get(label+iName);
              invocationTable.put(iName,value);
            } else executable = false;
          }
          
          /* if checkPreconds flag is set, then check the preconditions of
           * the service w.r.t. the PellReasoner.
           */
//           if ((checkPreconds) && (!checkPreconditions(serv, invocationTable)))
//             throw new Exception("Preconditions of service " + serv.getName() + " are not fulfilled!");
          
          /* if executable, invoke the service and put its results in the outputValues table
           */
          if (executable) {
            logger.debug("[executor] now executing service... " + serv.getName());
            long startTime = System.currentTimeMillis();
            
            /* if the preconditions ought to be checked and they are not fulfilled return a Failure */
            if ((checkPreconds) && (!checkPreconditions(serv, invocationTable)))
              return owlJDom.createFailure(query.name, globalStartTime, cpuStart,  "The preconditions of service " + serv.getName() + " are not fulfilled with the input parameters bound to " + invocationTable + ".");
             
            ValueMap resultMap= serv.invokeOWLS(invocationTable);
            logger.debug("..execution took me " + (System.currentTimeMillis() - startTime) + " msecs");
            reached.remove(label);
            executed.put(label, v);
            for(int k=0; k<outputs.size(); k++){
              Output out = (Output) outputs.get(k);
              String oName = out.getName();
              // new for 1.1.0 
                           OWLDataValue owlV = resultMap.getDataValue(process.getOutputs().getParameter(oName));
                           value = (String) owlV.getValue();
//               value = (String) resultMap.getValue(process.getOutputs().getParameter(oName));
              logger.debug("[executor] result:\n" + value);
              Element resultRoot = owlJDom.getRootElement(value);
              if (owlJDom.hasChild(resultRoot, "Failure",  OWLJDom.MW_NS)){
                return value;
              }
              outputValues.put(label+oName, value);
            }
//             for(Enumeration en = results.keys(); en.hasMoreElements(); ){
//               String key = (String) en.nextElement();
//               value = (String) results.get(key);
//               logger.info("[executor] result:  " + label + " = " +value);
//               Element resultRoot = owlJDom.getRootElement(value);
//               if (owlJDom.hasChild(resultRoot, "Failure",  OWLJDom.MW_NS)){
//                 return value;
//               }
//               outputValues.put(label, value);
//             }
          }
        } // end of execution of all executable services in "reached"
      }
    } // end of while loop until only the Finish vertex is reached
    
    logger.info("reached main execution after loop: "+ reached);
    logger.info("final result: "+ finalResults);
    
    return createResult(query.name, globalStartTime, cpuStart, new ArrayList(finalResults.values()));
  }
    
  public String createResult(String probID, long startTime, long cpuStart, List results) {
    try{
      Element rdf = new Element("RDF", OWLJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      rdf.setAttribute("base", OWLJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);
      
      Element time = new Element("time",  OWLJDom.MW_NS);
      Element timeres = OWLJDom.createTimeResource(startTime, cpuStart);
      for(int i=0; i<results.size(); i++){
        String res = (String) results.get(i);
        Element resultRoot = owlJDom.getRootElement(res);
        List cs = resultRoot.getChildren();
        for(int k=0; k<cs.size(); k++){
          Element child = (Element) ((Element) cs.get(k)).clone();
          rdf.addContent(child);
        }
      }
      
      rdf.addContent(time); 
      return owlJDom.newXMLDocument(rdf);
    }
    catch(Exception e){
      return owlJDom.createFailure(probID, startTime, cpuStart, e.getMessage());
    }
  }
  

  /* Does it make sense to check preconditions? I guess the checking needs
   * more time than the service realising it itself and failing.
   */  
  private boolean checkPreconditions(ReasoningService serv, Hashtable invocationTable)
  {
    long startTime = System.currentTimeMillis();
    List preconds = serv.getProblem().getPreconds(); //getDynamicPreconds();
    boolean fulfilled = true;
    
    for(int i=0; i<preconds.size(); i++){
      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(i);
      CachedHashCodeObject condition = pre.getCondition();
      if (condition instanceof Literal) {
        if (! checkLiteral((Literal) condition))
          fulfilled = false;
      }
      else if (condition instanceof Conjunct){
        Conjunct conj = (Conjunct) condition;
        List lits = conj.getLiterals();
        for(int k=0; k<lits.size(); k++){
          if (!checkLiteral((Literal) lits.get(k)))
            fulfilled = false;
        }
      }
    }
    logger.debug("Checking preconds took me " + (System.currentTimeMillis() - startTime) + " msecs");
    return fulfilled; 
  }
  
  private boolean checkLiteral(Literal lit)
  {
    Predicate p = lit.getPredicate();
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String pName = p.getName();
    if (!pName.equals("type")){
      return false;
    }
    return true;
  }
  
  public static void main(String argv[]) throws Exception {
    return;
    
  }
    

}
              //                // assign query input to service input of same typea
            //                if (queryIn.getSignature().equals(edgeType)){
            //                  value = queryIn.getContents();
            //                  // remove the input from the available query inputs
            //                  queryInputs.remove(queryIn);
            //                  // finish loop
            //                  qj = queryInputs.size(); 
            //                }
            //              }

  
  
