// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:

package org.mathweb.broker;

import java.io.*;
import java.util.*;

import org.jdom.*;

import salvo.jesus.graph.*;
//import salvo.jesus.graph.visual.*;
//import salvo.jesus.graph.visual.layout.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.broker.ServiceEntry;
import org.mathweb.graph.Annotation;
import org.mathweb.graph.AnnotatedEdge;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;

class ServiceMerger extends BasicService 
{
  private Logger logger = Logger.getLogger("ServiceMerger");
  private String logFileName = getSystemProperty("merger.log.file", "MERGER_LOG_FILE");

  private int counter = 0;
  private HashMap edgeMap = null;

  public class MSet 
  {
    private Set graphs = null;
    private DirectedGraph ggccs = null;
    
    public MSet(Set s, DirectedGraph g) 
    { this.graphs = s; this.ggccs = g; }
    
    public MSet() { this(null, null);}
    
    public DirectedGraph getGgccs() { return ggccs;  }
    public Set getGraphs() { return graphs;  }
  }
  
    //   private class MyComp implements Comparator
//   {
//     public int compare(Object o1, Object o2)
//     {
//       if ((o1 instanceof GraphImpl) && (o2 instanceof GraphImpl)){
//         Graph g1 = (GraphImpl) o1;
//         Graph g2 = (GraphImpl) o2;
//         int v1 = g1.getVerticesCount();
//         int v2 = g2.getVerticesCount();
//         System.out.println("comp: " + v1 + " " + v2);
        
//         if (v1 < v2) return -1;
//         else if (v1 > v2) return 1;
//       }
//       return 0;
//     }
//   }
  
  public ServiceMerger()
  {
    super();

    logger.setLevel(Level.DEBUG);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }
  
  public static List obj2array(Object obj) { 
    if (obj==null) return new ArrayList();
    List tmp = new ArrayList();
    tmp.add(obj);
    return tmp;
  }

  /**
   * Create the union of two sets
   * @param s1 first set
   * @param s2 second set
   * @retrun a new set being the union of the input sets. Both sets have to
   * be either descriptive or listing sets. Mixed mode is not supported.
   */
  public static Set unite(Set s1, Set s2) {
    Set ret = new HashSet(s1);
    ret.addAll(s2);
    return ret;
  }
  
  public boolean subsetNeq(Set s1, Set s2)
  {
    return ((s1.size() < s2.size()) && (s2.containsAll(s1)));
  }
  

  public static Set powerSet(Set s) {
    // Make every member of the submitted set a set of it's own. The
    // algorithm requires this.
    Object[] elem = s.toArray();
    for(int i=0; i<elem.length; i++) {
      if (!(elem[i] instanceof Set)) {
        elem[i] = new HashSet(obj2array(elem[i]));
      }
    }
    
    Set ret = null;
    
    // Recursivly create the powerset
    switch (elem.length) {
      // No magic involved here, the powerset of the empty set is the empty 
      // set by definition.
      case 0: {
        ret = new HashSet(obj2array(new HashSet()));
        break;
      }
      // No magic here either. As the previous case, this is still a
      // definition only used in for one element sets.
      case 1: {
        ret = new HashSet();
        ret.add(new HashSet());
        ret.add(s);
        break;
      }
      // The magic starts here! This is the resucrsion anchor, defining how
      // the powerset of a two element set is derived. All bigger sets trace
      // back to this.
      case 2: {
        ret = new HashSet(obj2array(new HashSet()));
        ret.add(elem[0]);
        ret.add(elem[1]);
        ret.add(unite((Set)elem[0],(Set)elem[1]));
        break;
      }
      /*
       * The powerset of an arbitary big set is derived by the following
       * Algorithm:
       * 1. Split one element of the set (it is a set itself).
       * 2. create the powerset of the remaining sets (recursion starts)
       * 3. create the powerset of the submitted set, by uniting the split off
       *    element with all sets of the powerset created by 2.
       * 4. Create a big union of all sets resulting from 3. and 1.
       * 5. Return this big union.
       */
      default: {
        Object[] tmp = new Object[elem.length-1];
        System.arraycopy(elem,0,tmp,0,tmp.length);
        List tmpList = new ArrayList();
        for (int i=0; i < tmp.length; i++) 
          tmpList.add(tmp[i]);
        
        Set tmp2 = powerSet(new HashSet(tmpList));

        ret = new HashSet(tmp2);

        for (Iterator it = tmp2.iterator(); it.hasNext();) {
          ret.add(unite((Set) elem[tmp.length], (Set) it.next()));
          ret.add(elem[tmp.length]);
        }
        ret.add(elem[tmp.length]);
      }
    }
    return ret;
  }
  
  public List sinks(DirectedGraph g)
  {
    List sinks = new ArrayList();
    Set vs = g.getVertexSet();
    for(Iterator it= vs.iterator(); it.hasNext();){
      Vertex v = (Vertex) it.next();
      List es = g.getOutgoingEdges(v);
      if (es.isEmpty()) sinks.add(v);
    }
    return sinks;
  }
  
  private Object pick(Set objs)
  {
    Iterator it = objs.iterator();
    if (it.hasNext()) return it.next();
    else return null;
  }
  
  public Set compose(Set graphs, Vertex start){
    Set power = powerSet(graphs);
    Set wprime = new HashSet();
    Set newGraphs = new HashSet();
    //    System.out.println("power: " + power);
    
    for(Iterator it = power.iterator(); it.hasNext();){
      Set set = (Set) it.next();
      //      System.out.println("set: " + set);
      
      if (!set.isEmpty()){
        DirectedGraph ggccs = gccs(set, start);
        //        System.out.println("gccs: " + ggccs);
        if (ggccs == null) return null;
        if (ggccs.getVerticesCount() > 1) {
          //          System.out.println("gccs accepted");
          
          MSet ms = new MSet(set, ggccs);
          newGraphs.add(ms);
        }
      }
    }
    
    power = powerSet(newGraphs);
    
    for(Iterator it = power.iterator(); it.hasNext(); ){
      Set msSet = (Set) it.next();
      if (msSet.size() == 2){
        Object[] objs = msSet.toArray();
        MSet ms1 = (MSet) objs[0];
        MSet ms2 = (MSet) objs[1];
        if (subsetNeq(ms1.getGraphs(), ms2.getGraphs()))
          newGraphs.remove(ms1);
        if (subsetNeq(ms2.getGraphs(), ms1.getGraphs()))
          newGraphs.remove(ms2);
      }
    }
    Set result = new HashSet();
    for(Iterator it = newGraphs.iterator(); it.hasNext(); ){
      MSet ms = (MSet) it.next();
      result.add(merge(ms));
    }
    
    return result;
  }
  
  
  public DirectedGraph gccs(Set graphs, Vertex start)    {
    DirectedGraph ggccs = new DirectedGraphImpl();
    try {
      ggccs.add(start);
      
      if (graphs.isEmpty()) return ggccs;

      DirectedGraph g = (DirectedGraph) pick(graphs);
      Set rest = new HashSet(graphs);
      rest.remove(g);
      if (rest.isEmpty()) return ggccs;
      
      Set k;
      Set knew = new HashSet();
      knew.add(start);
      
      do{
        k = knew;
        knew = new HashSet();
        
        for(Iterator it=k.iterator(); it.hasNext();){
          Object obj = it.next();
          //          System.out.println("next sink: " + obj + "::" + obj.getClass().getName());
          Vertex sink = (Vertex) obj;
          if (!sink.getLabel().equals("End")){
            
            List allEdges = g.getOutgoingEdges(sink);
            List adjacent = new ArrayList();
            for(Iterator eit = allEdges.iterator(); eit.hasNext();){
              Edge e = (Edge) eit.next();
              Vertex m = e.getVertexB();
              int length = g.lengthOfLongestPath(sink, m);
              
              //              System.out.println("length of longest path " + sink + " --> " + m + ": " + length);
              
              if ((!(length > 1 )) && (!m.getLabel().equals("End"))){ //allEdges.size())){
                adjacent.add(m);
                //                System.out.println(m + " adjacent to " + sink);
              }
              //  else 
              //               ;illegalEdges.add(e);
            }
            for(Iterator mit = adjacent.iterator(); mit.hasNext();){
              Vertex m = (Vertex) mit.next();
              boolean accepted = true;
              for(Iterator git = rest.iterator(); (git.hasNext() && (accepted == true));){
                DirectedGraph gother = (DirectedGraph) git.next();
                if (!(gother.getVertexSet().contains(m))) accepted = false;
                //TODO: this has to check for all the edges sink-> m in G
                if (gother.getEdge(sink, m) == null) accepted = false;
              }
              if (accepted){
                ggccs.add(m);
                ggccs.addEdge(sink, m);
              knew.add(m);
              }
            }
          }
        }
      } while (! knew.isEmpty());
        
    } catch(Exception e) {
      e.printStackTrace();
      return null;
    }
    
    return ggccs;
  }
  
  
  public DirectedGraph merge(MSet ms)
  {
    edgeMap = new HashMap();
    Set graphs = ms.getGraphs();
    DirectedGraph gm = ms.getGgccs();
    int count = 1;
    Set delete = new HashSet();
    List sinks = sinks(gm);

    try{
      for(Iterator sit = sinks.iterator(); sit.hasNext(); ){
        Vertex sink = (Vertex) sit.next();
        Vertex choice = new VertexImpl();
        choice.setLabel("Choice" + count++);
        gm.add(choice);
        // TODO: handle annotations
      
        for(Iterator git = graphs.iterator(); git.hasNext();){
          DirectedGraph current =  (DirectedGraph) git.next();
          List allEdges = current.getOutgoingEdges(sink);
          
          // go through all outgoing edges in the current graph
          for(Iterator eit = allEdges.iterator(); eit.hasNext();){
            AnnotatedEdge e = (AnnotatedEdge) eit.next();
            Vertex m = e.getVertexB();
            int length = current.lengthOfLongestPath(sink, m);
            //            System.out.println("length of longest path " + sink + " --> " + m + ": " + length);
            
            if (!(length > 1 )){ //allEdges.size())){
              Annotation ann = e.getAnnotation();
              String outSink = ann.getSource();
              String inM  = ann.getTarget();
              gm.add(m);
              AnnotatedEdge newEdge = null;
              String key = sink.toString() + choice + outSink + "in0";
              delete.add(e);
              if (edgeMap.containsKey(key)){
                newEdge = (AnnotatedEdge) edgeMap.get(key);
              } else {
                newEdge = new AnnotatedEdge(sink, choice,  new Annotation(outSink, "in0"));
              }
              edgeMap.put(key, newEdge);
              gm.addEdge(newEdge);
              gm.addEdge(new AnnotatedEdge(choice, m, ann));
            
              //              System.out.println(" new edge " + choice + " --> " +  m);
            }
          }
        }
      }
      this.counter = 1;
      
      for(Iterator git = graphs.iterator(); git.hasNext();){
        DirectedGraph current =  (DirectedGraph) git.next();

        Set nodes = new HashSet(current.getVertexSet());
        Vertex endNode = null;
        
        for(Iterator nodeIt = nodes.iterator(); nodeIt.hasNext();){
          Vertex v = (Vertex) nodeIt.next();
          //          System.out.println("node: " + v);
          
          if (v.getLabel().equals("End")){
            endNode = v;
          }
        }
        
        nodes.removeAll(gm.getVertexSet());
        //        nodes.remove(endNode);
        //        gm.add(endNode);
        
        //        if (endNode != null) nodes.add(endNode);
        
        Set edges = new HashSet(current.getEdgeSet());
        edges.removeAll(delete); 
        
        HashMap map =  copyNodes(nodes);
        Set newEdges = modifyEdges(edges, map);
        
        Set newNodes = new HashSet(map.values());
        for(Iterator nodeIt = newNodes.iterator(); nodeIt.hasNext();){
          Vertex v = (Vertex) nodeIt.next();
          gm.add(v);
        }
        for(Iterator edgeIt = newEdges.iterator(); edgeIt.hasNext();){
          AnnotatedEdge e = (AnnotatedEdge) edgeIt.next();
          Vertex a = e.getVertexA();
          Vertex b = e.getVertexB();
          Annotation ann = e.getAnnotation();
          String out = ann.getSource();
          String in = ann.getTarget();
          
          String key = a.toString() + b + out + in;
          if (!edgeMap.containsKey(key)){
            edgeMap.put(key, e);
            gm.addEdge(e);
          }
        }
      }
      
    } catch(Exception e){
      System.out.println(e);
      e.printStackTrace();
    }
    return gm;
  }
  
  public Set modifyEdges(Set edges, HashMap map)
  {
    Set result = new HashSet();
    for(Iterator it = edges.iterator(); it.hasNext();){
      Object obj = it.next();
      if (obj instanceof AnnotatedEdge){
        AnnotatedEdge e = (AnnotatedEdge) obj;
        Vertex a = e.getVertexA();
        Vertex b = e.getVertexB();
        Vertex newA = (Vertex) map.get(a);
        Vertex newB = (Vertex) map.get(b);
        if (newA != null) a = newA;
        if (newB != null) b = newB;
        Edge ne = new AnnotatedEdge(a, b, e.getAnnotation());
        ne.setFollowVertexLabel(false);
        ne.setLabel(e.getLabel());
        result.add(ne);
      }
    }
    
    return result;
  }
  
  public HashMap copyNodes(Set nodes)
  {
    HashMap result = new HashMap();
    for(Iterator it= nodes.iterator(); it.hasNext();){
      Vertex v = (Vertex) it.next();
      Vertex nv = new VertexImpl();
      nv.setLabel(v.getLabel() + counter++);
      result.put(v, nv);
    }
    return result;
  }
  

  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
     ServiceMerger merger = new ServiceMerger();
    
     Set s = new HashSet();
//     s.add("a");    s.add("b");    s.add("c");  s.add("d");
//     Set ps = ServiceMerger.powerSet(s);
    //    System.out.println("power: " + ps);

    Graph g1= new DirectedGraphImpl();
    Graph g2= new DirectedGraphImpl();
    Graph g3= new DirectedGraphImpl();

    Vertex start = new VertexImpl(); start.setLabel("Start");
    Vertex end = new VertexImpl(); end.setLabel("End");

    Vertex cnf = new VertexImpl(); cnf.setLabel("TrampCNF");
    Vertex otter = new VertexImpl(); otter.setLabel("Otter");
    Vertex spass = new VertexImpl(); spass.setLabel("SPASS");
    Vertex ep = new VertexImpl(); ep.setLabel("Ep");
    Vertex of = new VertexImpl(); of.setLabel("Otterfier");
    Vertex nd = new VertexImpl(); nd.setLabel("NDforFOF");
    Vertex ho = new VertexImpl(); ho.setLabel("HO2FO");

    Edge s2c = new AnnotatedEdge(start, cnf, new Annotation("fof", "fof"));
    Edge s2nd = new AnnotatedEdge(start, nd, new Annotation("fof", "fof"));
    Edge c2o = new AnnotatedEdge(cnf, otter, new Annotation("cnf", "tstp"));
    Edge c2sp = new AnnotatedEdge(cnf, spass , new Annotation("cnf", "tstp"));
    Edge c2ep = new AnnotatedEdge(cnf, ep, new Annotation("cnf", "tstp"));
    Edge c2nd = new AnnotatedEdge(cnf, nd, new Annotation("delta", "delta"));
    Edge o2nd = new AnnotatedEdge(otter, nd, new Annotation("res", "res"));
    Edge ep2of = new AnnotatedEdge(ep, of, new Annotation("res", "oldres"));
    Edge sp2of = new AnnotatedEdge(spass, of, new Annotation("res", "oldres"));
    Edge of2nd = new AnnotatedEdge(of, nd, new Annotation("newres", "res"));
    Edge nd2end = new AnnotatedEdge(nd, end, new Annotation("ndpr", "ndpr"));

    Edge s2ho = new AnnotatedEdge(start, ho, new Annotation("ho", "ho"));
    Edge ho2cnf = new AnnotatedEdge(ho, cnf, new Annotation("fo", "fo"));
    
    g1.add(start); g1.add(end);  g1.add(cnf); g1.add(otter); g1.add(nd);
    g2.add(start); g2.add(end);  g2.add(cnf); g2.add(spass); g2.add(of); g2.add(nd);
    g3.add(start); g3.add(end);  g3.add(cnf); g3.add(ep);    g3.add(of); g3.add(nd);

//     g1.add(ho); g2.add(ho); g3.add(ho);
//     g1.addEdge(s2ho); g1.addEdge(ho2cnf); 
//     g2.addEdge(s2ho); g2.addEdge(ho2cnf); 
//     g3.addEdge(s2ho); g3.addEdge(ho2cnf); 
      g1.addEdge(s2c); g2.addEdge(s2c);     g3.addEdge(s2c);

    g1.addEdge(s2nd);  g1.addEdge(c2o);  g1.addEdge(c2nd); g1.addEdge(o2nd); g1.addEdge(nd2end);
    g2.addEdge(s2nd);  g2.addEdge(c2sp); g2.addEdge(c2nd); g2.addEdge(sp2of); g2.addEdge(of2nd);    
    g2.addEdge(nd2end);
    
    g3.addEdge(s2nd);  g3.addEdge(c2ep); g3.addEdge(c2nd); g3.addEdge(ep2of); g3.addEdge(of2nd);    
    g3.addEdge(nd2end);
    
    s = new HashSet();
    s.add(g1);
    s.add(g2);
    s.add(g3);
    
    Set merged = merger.compose(s, start);
    //    System.out.println("merged: " + merged);
    
    int counter = 1;
    
    for(Iterator mit= merged.iterator(); mit.hasNext();){
      DirectedGraph gr= (DirectedGraph) mit.next();
      System.out.println("merge-graph: " + gr);
      String dot  = ServicePlanner.graph2Dot(gr);
      merger.writeFile(msHome + "/etc/data/graph/graph" + counter + ".dot", dot);
      counter++;
    }
    return;
  }
    

}


      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


