// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
/*
 * Created on Dec 22, 2004
 */
package org.mathweb.onto;

import java.io.*;
import java.util.*;
import java.net.*;

import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.reasoner.*;
import com.hp.hpl.jena.ontology.*;

import com.hp.hpl.jena.reasoner.rulesys.*;
import com.hp.hpl.jena.reasoner.ReasonerRegistry;
import com.hp.hpl.jena.reasoner.dig.*;
import com.hp.hpl.jena.vocabulary.*;


import com.hp.hpl.jena.util.PrintUtil;
import com.hp.hpl.jena.util.FileManager;
import com.hp.hpl.jena.util.LocationMapper;
import com.hp.hpl.jena.util.iterator.*;

import org.mindswap.pellet.jena.PelletReasoner;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import org.mathweb.core.BasicService;
import org.mathweb.util.Queue;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;

public class PellReasoner extends BasicService implements OntoReasoner
{
  //  private static final String debug = "true";
    private static final String debug = "false";
  public URL ontologyFileURL;
  
  private static String MW_NS_STR = OWLJDom.MW_NS_STR;
  
  private static String sep = File.separator;
 
  
  // the dynamic ontology model
  private OntModel ontModel;
  // the static OWL model with the MathWeb ontology
  private Model owlModel;
  // the model for newly told individuals
  private Model indiModel;
  // the model for the individual told last
  private Model lastModel;
  // the inference model over that individual
  private InfModel infModel;

  private List rules;
  private Reasoner owlReasoner;
  //  private Reasoner ruleReasoner;
  
  private Hashtable anonResources;
  private Hashtable valueOfCache;
  private int maxCacheSize = 20;
  
  private static PellReasoner theInstance = (new PellReasoner()).config();

  private static Logger logger;
  
  private String owlFileUrlStr = "file://" + getSystemProperty("mathserv.ontology.owl", "MATHSERV_ONTOLOGY_OWL");
  private String policyConfigFile = getSystemProperty("mathserv.ontology.policy", "MATHSERV_ONTOLOGY_POLICY");
  private LocationMapper locMap = null;
  private FileManager fileManager = null;

  /** 
   * Default constructor
   */
  public PellReasoner(){
    super();
  }

  /** 
   * Constructor which returns a clone of a reasoner.
   */
  private PellReasoner(PellReasoner pell){
    //long startTime = System.currentTimeMillis();
    
    owlModel = pell.getOwlModel();
    
    owlReasoner = new PelletReasoner();
    OntModelSpec spec = new OntModelSpec(OntModelSpec.OWL_DL_MEM );
    spec.setReasoner(owlReasoner);
    
    ontModel = ModelFactory.createOntologyModel(spec, null);
    ontModel.addSubModel(owlModel);

    indiModel = ModelFactory.createDefaultModel();
    lastModel = ModelFactory.createDefaultModel();      

    //    System.out.println("Cloning PellReasoner took me " + (System.currentTimeMillis() - startTime) + " msecs.");
    anonResources = new Hashtable();
    valueOfCache = new Hashtable();
  }
  

  /** 
   * @return configures the OWL reasoner and returns the configured reasoner.
   */
  protected PellReasoner config() {
    try{
      PrintUtil.registerPrefix("mw",MW_NS_STR);
      
      owlReasoner = new PelletReasoner();
      OntModelSpec spec = new OntModelSpec(OntModelSpec.OWL_DL_MEM );
      spec.setReasoner(owlReasoner);
      //      owlReasoner.bindSchema(ontModel);
      
      ontModel = ModelFactory.createOntologyModel(spec, null);
      
      locMap = OntPolicyReader.theInstance.getMapper();
      fileManager = FileManager.get();
      fileManager.setMapper(locMap);
      //      System.out.println("test: " + locMap.altMapping("http://www.mathweb.org/owl/mathweb.owl", null));
      
      //      owlModel = fileManager.loadModel(owlFileUrlStr);
      owlModel = fileManager.loadModel(owlFileUrlStr);
      //ontModel.read(owlFileUrlStr);
      ontModel.addSubModel(owlModel);
      //      owlModel = ontModel;

      //  indiModel is empty
      indiModel = ModelFactory.createDefaultModel();      
      lastModel = ModelFactory.createDefaultModel();      
      anonResources = new Hashtable();
      valueOfCache = new Hashtable();
      //       String ruleSrc = readFile(ontoRulesFileStr);
      //       rules = Rule.parseRules(ruleSrc);
      
      //       ruleReasoner = new GenericRuleReasoner(this.rules);
      //       ruleReasoner.bindSchema(ontModel);
    } catch (Exception e){
      e.printStackTrace();
      
      System.err.println("Something went wrong!");
      System.err.println(e);
    }
    return this;
  }
  
  public boolean tellIndividual(String indiString)
  {
    File inputFile = null;
    try{ 
      inputFile = File.createTempFile("pell-reasoner-indi", ".owl");
      String inputFileName = inputFile.getAbsolutePath();
      writeFile(inputFile, indiString);
      // = fileManager.loadModel("file:" + inputFileName);
      // first remove the old individuals model to avoid copies
      
      lastModel  = ModelFactory.createDefaultModel();      
      lastModel.read("file:" + inputFileName);
      // first remove the old individuals model to avoid copies
      ontModel.removeSubModel(indiModel);
      // then add the new data to the indiModel
      indiModel.add(lastModel);
      // then add the new indiModel again:
      ontModel.addSubModel(indiModel);

      owlReasoner.bindSchema(ontModel);
      
      //      long startTime = System.currentTimeMillis();
      // compute all inferences over the new ontModel
      //      infModel = ModelFactory.createInfModel(owlReasoner, ontModel);
//       System.out.println("... computing huge inference took me " + (System.currentTimeMillis() - startTime) + " msecs");
      
//       System.out.println(" huge inference Model: ");
//       Iterator list = infModel.listStatements();
//       while (list.hasNext()) {
//         System.out.println(" + " + list.next());
//       }

//      if (inputFile != null) inputFile.delete();
      return true;
    } catch (Exception e){
      System.err.println("no solution found!");
      e.printStackTrace();
      return false;
    } finally{
      if (inputFile != null) inputFile.delete();
    }
  }
  
  
  public boolean untellIndividual(String indURI)
  {
    try{ 
      Resource ind;
      if (anonResources.containsKey(indURI)){
        //        System.out.println("anonymous " + indURI);
        ind = ontModel.createResource((AnonId) anonResources.get(indURI));
      } else 
        ind = ontModel.getResource(indURI);

      ontModel.removeSubModel(indiModel);
      // then add the new data to the indiModel
      indiModel.removeAll(ind , (Property) null, (RDFNode) null);
      // then add the new indiModel again:
      ontModel.addSubModel(indiModel);
      return true;
    } catch (Exception e){
      System.err.println("untell of individual failed! " + e);
      e.printStackTrace();
      return false;
    }
  }
  

//   public boolean untellIndividual(String indiID)
//   {
//     File inputFile = null;
//     try{ 
//       inputFile = File.createTempFile("jena-reasoner-indi", ".owl");
//       String inputFileName = inputFile.getAbsolutePath();
//       writeFile(inputFile, indiString);
//       Model old_data = fileManager.loadModel("file:" + inputFileName);
//       indiModel = indiModel.difference(old_data);
//       if (inputFile != null) inputFile.delete();
//       return true; 
//     } catch (Exception e){
//       System.err.println("untell of individual failed! " + e);
//       e.printStackTrace();
//       return false;
//     } finally{
//       if (inputFile != null) inputFile.delete();
//     }
//   }
 
  /** 
   * @param className the name of an OWL class.
   * @return the rdf:ID of the last individual told of this class
   */
  public String nameOfLastIndividual(String owlClass)
  {
    String indID = null;
    Resource cl = ontModel.getResource(MW_NS_STR + owlClass);
    
    ResIterator subjit = lastModel.listSubjects();
    Iterator list;
    while(subjit.hasNext() && indID==null){
      Resource indi = subjit.nextResource();
      Statement s = ontModel.createStatement(indi, RDF.type, cl);
//       list = infModel.listStatements(indi, RDF.type, cl);
//       if (list.hasNext())
      if (ontModel.contains(s));
      indID = indi.getLocalName();
    }
    
//     System.out.println("looking for indis with " + RDF.type + " " + cl + "...");
//     long startTime = System.currentTimeMillis();
//     ResIterator it = infModel.listSubjectsWithProperty(RDF.type, cl);
//     System.out.println("... took me " + (System.currentTimeMillis() - startTime) + " msecs");
//     while(it.hasNext() && indID==null){
//       Resource indi = it.nextResource();
//       System.out.println("resource " + indi + " " + indi.getClass());
//       if(lastModel.containsResource(indi)){
//         indID = indi.getLocalName();
//         System.out.println("found individual " + indID);
//       }
//     }
    return indID;
  }

  

  
  /** 
   * @param individual the rdf:ID of an OWL individual.
   * @param className the name of an OWL class.
   * @return true if the individual is of the given class
   */
  public boolean hasClass(String individual, String className)
  {
    //     Resource ind = infModel.getResource(MW_NS_STR + individual);
    //     Resource cl = infModel.getResource(MW_NS_STR + className);
    //     StmtIterator it = infModel.listStatements(ind, RDF.type, (RDFNode) null); 
    Resource ind = ontModel.getResource(MW_NS_STR + individual);
    Resource cl = ontModel.getResource(MW_NS_STR + className);
    StmtIterator it = ontModel.listStatements(ind, RDF.type, cl);
    //    System.out.println("hasClass " + it.hasNext());
    return it.hasNext();
    //     StmtIterator it = infModel.listStatements(ind, RDF.type, (RDFNode) null);
//     System.out.println("hasClass it " + it);
//     if (it.hasNext()){
//       Statement s = it.nextStatement();
//       Resource cl1 = s.getResource();
//       System.out.println("class " + PrintUtil.print(cl1));
//       return owlModel.contains(cl1, RDFS.subClassOf, cl);
//       //      return false;

//     } else return false;
  }
  

  /** 
   * @param class1 the name of an OWL class
   * @param class2 the name of another OWL class.
   * @return true if class1 is an asserted or inferred subclass of class2
   */
  public boolean subClassOf(String class1, String class2)
  {
    Resource cl1 = ontModel.getResource(MW_NS_STR + class1);
    Resource cl2 = ontModel.getResource(MW_NS_STR + class2);
    if ((cl1 != null) && (cl2 != null))
      return ontModel.contains(cl1, RDFS.subClassOf, cl2);
    else return false;
  }

  /** 
   * @param class1  an OWL class
   * @param class2  another OWL class.
   * @return true if class1 is an asserted or inferred subclass of class2
   */
  public boolean subClassOf(Resource cl1, Resource cl2)
  {
    if ((cl1 != null) && (cl2 != null))
      return ontModel.contains(cl1, RDFS.subClassOf, cl2);
    else return false;
  }

  public void assertLiteral(String property, URI indiURI, URI valueURI)  {
    try{ 
      Property propNode = ontModel.getProperty(MW_NS_STR, property);
      Resource indiRes = ontModel.getResource(indiURI.toString());
      RDFNode valueNode = ontModel.getResource(valueURI.toString());
      
      Statement statement =	ontModel.createStatement(indiRes, propNode, valueNode);
      ontModel.add(statement);
      owlReasoner.bindSchema(ontModel);
    } catch (Exception e){
      System.err.println("could not assert literl!");
      e.printStackTrace();
    }
  }
  
  /** 
   * @param property the name of a MathServ OWL property
   * @param individual the name of a MathServ OWL individual in the current model.
   * @return The string value of 'property' of the 'individual.
   */
  public String valueOf(String property, URI individual)  {
    String key = property + individual.toString();
    if ((key != null) && (valueOfCache.contains(key)))
      return (String) valueOfCache.get(key);
    Resource ind;
    ind = ontModel.getResource(individual.toString());
    
    Property prop = ontModel.getProperty(MW_NS_STR, property);
    String result = valueOf(prop, ind);
    //    System.out.println("valueOf " + key + " " + result);
    if ((key != null) && (result != null) && (valueOfCache.size() < this.maxCacheSize))
      valueOfCache.put(key, result);
    return result;
  }
  
  /** 
   * @param property the name of a MathServ OWL property
   * @param individual the name of a MathServ OWL individual in the current model.
   * @return The string value of 'property' of the 'individual.
   */
  public String valueOf(String property, String individual)
  {
    //Resource ind = infModel.getResource(MW_NS_STR + individual);
    
    Resource ind;
    if (anonResources.containsKey(individual)){
      //      System.out.println("anonymous " + individual);
      ind = ontModel.createResource((AnonId) anonResources.get(individual));
    } else 
      ind = ontModel.getResource(MW_NS_STR + individual);
    
    Property prop = ontModel.getProperty(MW_NS_STR, property);
//     Resource ind;
//     if (anonResources.containsKey(individual)){
//       System.out.println("anonymous " + individual);
//       ind = ontModel.createResource((AnonId) anonResources.get(individual));
//     } else 
//       ind = ontModel.getResource(MW_NS_STR + individual);
    
//     Property prop = ontModel.getProperty(MW_NS_STR, property);
//    System.out.println("valueOf " + ind + " " + prop);
    return valueOf(prop, ind);
  }
  
  public String valueOf(Property prop, Resource ind)
  {
    String result = null;
    try{
      //      System.out.println(" before ontModel.listObjectsOfProp"); 
      Iterator list = ontModel.listObjectsOfProperty(ind, prop);
      //      Iterator list = ontModel.listStatements(ind, prop, (RDFNode) null);
      //      System.out.println(" results? " + list.hasNext());
      if (list.hasNext()) {
        Object o = list.next();
        RDFNode r = (RDFNode) o;
        
        //        System.out.println(" object " + o + " " + o.getClass().getName());
//         Statement s = (Statement) o;
//         RDFNode r = s.getObject();
//         System.out.println(" statement " + s);
//         System.out.println(" value " + r);
//         System.out.println(" value of type " + r.getClass().getName() );
        if (r instanceof Resource){
          Resource res = (Resource) r;
          if (res.isAnon()){
            AnonId id = res.getId();
            result = id.toString();
            anonResources.put(result, id);
          } else {
            //            System.out.println(" value== " + res.getLocalName() + " " + res.getNameSpace());
//             result = res.getLocalName(); 
            result = res.getURI(); 
          }  }
        else if (r instanceof com.hp.hpl.jena.rdf.model.Literal){
          result = ((com.hp.hpl.jena.rdf.model.Literal) r).getString();
        } 
      }
    } catch (Exception e){
      System.err.println("no solution found!");
       e.printStackTrace();
    }
    return result;
  }
  
  /** 
   * @return true if the ontology with the individuals told so far is consistent.
   */
  public boolean isValid()
  {
//     Graph ontGraph = ontModel.getGraph();
//     Model check = ModelFactory.createModelForGraph(ontGraph);
//     check.add(ontModel);
    ValidityReport validity = ontModel.validate();
    return validity.isValid();
  }
  
  public ArrayList getAllSubclasses(String cls){
    ArrayList classes = new ArrayList();
    Resource clsRes = ontModel.getResource(MW_NS_STR + cls);
    ResIterator et = owlModel.listSubjectsWithProperty(RDF.type, clsRes);
    String name = null;
    
    while (et.hasNext()){
      try{
        Object o = et.next();
        Resource r = (Resource) o;
        //        System.out.println(r);
        if (!r.isAnon()){
          name = r.getLocalName();
          classes.add(name);
        }
      }catch(Exception e) {
      }
    }
    return classes;
  }
  
  private String lispName(String name)
  {
    return name.replaceAll("[:]", "-");
  }
  
  public StringBuffer getProdigyClassTree()
  {
    StringBuffer tree = new StringBuffer();
    
    ResIterator et = owlModel.listSubjectsWithProperty(RDF.type, OWL.Class);

    ArrayList cls = new ArrayList();
    ArrayList subs = new ArrayList();

    String[] owlThing = { "OWL-Thing" , ":top-type"};
    subs.add(owlThing);

    int m = 0;
    while (et.hasNext()){
      try{
        Object o = et.next();
        Resource r = (Resource) o;
        //        System.out.println(m + " " +r);
        m++;
        //        if (!r.isAnon()){
        //           Resource newr = ontModel.getResource(MW_NS_STR + r.getLocalName());
        //           if (newr != null)
        cls.add(r);
        //        }
      }catch(Exception e) {
      }
    }
    et = owlModel.listSubjectsWithProperty(RDF.type, OWL.Restriction);
    m = 0;
    while (et.hasNext()){
      try{
        Object o = et.next();
        Resource r = (Resource) o;
        //        System.out.println(m + " " +r);
        m++;
        //        if (!r.isAnon()){
        //           Resource newr = ontModel.getResource(MW_NS_STR + r.getLocalName());
        //           if (newr != null)
        cls.add(r);
      }catch(Exception e) {
      }
    }
    ArrayList root = new ArrayList();
    Queue queue = new Queue();
    String  marker = "*";

    // for every class ony find out whether there is at least one superclass in cls
    for(int i=0 ; i<cls.size();i++){
      Resource r = (Resource) cls.get(i);
      boolean one = false;
      for(int k=0; k< cls.size(); k++){
        Resource r2 = (Resource) cls.get(k);
        //        System.out.println("compare  " + r2 + " <-> " + r);
        //        if ((!r2.isAnon()) && (!r.isAnon()))
        if (owlModel.contains(r, RDFS.subClassOf, r2))
          //if (subClassOf(r, r2))
          { 
            //            System.out.println(r + "<<" + r2);
            one=true;
            k=cls.size();
          }
      }
      
      if (!one)
        root.add(r);
    }
    
    for(int k=0; k<root.size(); k++){
      Resource r = (Resource) root.get(k);
      cls.remove(r);
      queue.enqueue(r);
      //      System.out.println("root  " + r);
      String name;
//       if (r.isAnon()){
//         AnonId id = r.getId();
//         name = lispName(id.toString());
//       }
//       else name = r.getLocalName();
      
//       String[] sub = { "OWL-" +name , "OWL-Thing"};
//       subs.add(sub);
      if (!r.isAnon()){
        name = r.getLocalName();
        String[] sub = { "OWL-" +name , "OWL-Thing"};
        subs.add(sub);
      }
    }
    queue.enqueue(marker);
    
    boolean finished = false;

    while(!finished){
      ArrayList current = new ArrayList();
      boolean marked = false;
      //get current level of nodes out of the queue
      while(!marked){
        Object o = queue.dequeue();
        if ((o instanceof String)){
          String name = (String) o;
          if (name.equals(marker))
            marked = true;
        }
        else if (o instanceof Resource){
          current.add(o);
        }
      }
      for(int k=0; k<current.size(); k++){
        Resource r = (Resource) current.get(k);
        //        System.out.println("node  " + r);
        ArrayList children = new ArrayList();
        
        //Iterator it = cls.iterator();
        //Iterator it = ontModel.listStatements((RDFNode) null, RDFS.subClassOf, r);
        //ResIterator it = ontModel.listSubjectsWithProperty(RDFS.subClassOf, r);
        ResIterator it = owlModel.listSubjectsWithProperty(RDFS.subClassOf, r);
        while(it.hasNext()){
          Resource r2 = (Resource) it.next();
          //          System.out.println("subclass  " + r2);
          //            if (owlModel.contains(r2, RDFS.subClassOf, r)){
          String name1, name2;
//           if (r.isAnon()){
//             AnonId id = r.getId();
//             name2 = lispName(id.toString());
//           } else name2=r.getLocalName();
          
//           if (r2.isAnon()){
//             AnonId id = r2.getId();
//             name1 = lispName(id.toString());
//           } else name1 = r2.getLocalName();
          
//           System.out.println(name1 + "<<" + name2);
//           String[] sub = { "OWL-" +name1, "OWL-" + name2};
//           subs.add(sub);
          if ((!r.isAnon()) && (!r2.isAnon())){
            name2 = r.getLocalName();
            name1 = r2.getLocalName();
            
            //            System.out.println(name1 + "<<" + name2);
            String[] sub = { "OWL-" +name1, "OWL-" + name2};
            subs.add(sub);
            children.add(r2);
          }
        }
        if (!children.isEmpty()){
          for(int s=0; s<children.size(); s++)
            if (!queue.contains(children.get(s)))
              queue.enqueue(children.get(s));
          queue.enqueue(marker);
        }
        
        //    System.out.println("Queue: ");
        //        queue.show();
        
        if (queue.isEmpty())
          finished = true;
      } // for 
    } // while loop
    
    for(int i=0 ; i<subs.size();i++){
      String[] sub = (String[]) subs.get(i);
      tree.append("(ptype-of "+sub[0]+ " "+sub[1]+")\n");
    }
    //    System.out.println(tree);
    return tree;
  }
  

  /** 
   * @return the unique instance of the PellReasoner with the MahtWeb ontology loaded.
   */
  public static PellReasoner getInstance() {
    return theInstance;
  }

  /** 
   * @return a new clone of the unique instance.
   */
  public static PellReasoner getNewReasoner()
  {
    return new PellReasoner(theInstance);
  }
  
  /** 
   * @return the PellReasoner ontology model underlying the reasoner.
   */
  public Model getOntModel()
  {
    return ontModel;
  }

  /** 
   * @return the PellReasoner ontology model underlying the reasoner.
   */
  public Model getOwlModel()
  {
    return owlModel;
  }
  /** 
   * @return the PellReasoner ontology model underlying the reasoner.
   */
  public InfModel getInfModel()
  {
    return ontModel;
  }

  /** 
   * @return the PellReasoner ontology model underlying the reasoner.
   */
  public List getRules()
  {
    return rules;
  }

//   /** 
//    * @return the generic PellReasoner rule reasoner underlying the reasoner.
//    */
//   public Reasoner getRuleReasoner()
//   {
//     return ruleReasoner;
//   }

  /** 
   * @return the generic PellReasoner rule reasoner underlying the reasoner.
   */
  public Reasoner getOwlReasoner()
  {
    return owlReasoner;
  }

  /** 
   * @return the rules of the ontology as planning operators.
   */
  public List rulesToActions()
  {
    List actions = new ArrayList();
    for (int i =0; i< rules.size(); i++){
      Rule r = (Rule) rules.get(i);
      String name = "ontology_" +r.getName();
      ClauseEntry[] cls = r.getBody();
      String subjName="";
      String propName="";
      String objName="";
            
      org.mathweb.reasoning.logic.Literal[] pre= new org.mathweb.reasoning.logic.Literal[cls.length];
      for (int j =0; j< cls.length; j++){
        TriplePattern tp = (TriplePattern) cls[j];
        //        System.out.println("tp: " + tp);
        String tps = tp.toString();
        String[] st= tps.split(" ");
	if (st.length>2) {
          subjName = st[0].substring(1);
          propName = st[1].substring(1);
          objName = st[2].substring(1);
        }
        Predicate p = new Predicate("constraint");
        Argument[] as = new Argument[3];
        // the owl property
        as[0] = new Constant(propName);
        as[1] = new Variable(subjName);
        as[2] = new Variable(objName);
        pre[j] = new org.mathweb.reasoning.logic.Literal(p, as, true);
      }

      cls = r.getHead();
      org.mathweb.reasoning.logic.Literal[] post= new org.mathweb.reasoning.logic.Literal[cls.length];
      for (int j =0; j< cls.length; j++){
        TriplePattern tp = (TriplePattern) cls[j];
        //        System.out.println("tp: " + tp);
        String tps = tp.toString();
        String[] st= tps.split(" ");
	if (st.length>2) {
          subjName = st[0].substring(1);
          propName = st[1].substring(1);
          objName = st[2].substring(1);
        }
        
        Predicate p = new Predicate("constraint");
        Argument[] as = new Argument[3];
        // the owl property
        as[0] = new Constant(propName);
        as[1] = new Variable(subjName);
        as[2] = new Variable(objName);
        post[j] = new org.mathweb.reasoning.logic.Literal(p, as, true);
      }
      Action a = new Action(new Task(name), pre, post, new BindingConstraints());
      
      //      System.out.println("operator: " + a.toLongString());
      actions.add(a);
    }
    return actions;
  }

  public static void main(String argv[]) throws Exception {
    PellReasoner r =  PellReasoner.getInstance();
    
    //    r.tellIndividual(r.readFile(msHome + "/examples/owl/fof.owl"));
    r.tellIndividual(r.readFile(msHome + "/examples/owl/omdoc1.owl")); 
    
    System.out.println("told ind 1 ");
    r.tellIndividual(r.readFile(msHome + "/examples/owl/proof1.owl"));
    System.out.println("told ind 2 ");
    r.tellIndividual(r.readFile(msHome + "/examples/owl/proof2.owl"));
    System.out.println("told ind 3 ");
    System.out.println("language " + r.valueOf("language","test_omdoc_problem"));
    
//     //     //    r.test();
    System.out.println("valueOf(format, test_fof_problem123)  " + r.valueOf("language","test_fof_problem123"));
    System.out.println("valueOf(formalDescription, test_fof_problem123)  " + r.valueOf("formalDescription","test_fof_problem123"));
    System.out.println("subclass " + r.subClassOf("TstpCNFRefutation", "Proof"));
    System.out.println("subclass " + r.subClassOf("TstpFOFProblem", "TstpProblem"));
    System.out.println("subclass " + r.subClassOf("TstpCNFProblem", "TstpProblem"));
    System.out.println("hasClass " + r.hasClass("test_brfp_refutation2", "Proof"));
    System.out.println("hasClass " + r.hasClass("test_fof_problem123", "TstpProblem"));
     System.out.println("valid? " + r.isValid()); 
//    StringBuffer sb = r.getProdigyClassTree();
  }
  
  
}  // end class PellReasoner

/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
