// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.owls;

import java.io.*;
import java.net.URI;
import java.net.URL;
import java.util.*;


import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mindswap.owl.OWLType;
import org.mindswap.owl.OWLCache;
import org.mindswap.owl.OWLReader;
import org.mindswap.owl.OWLOntology;
import impl.owl.OWLCacheImpl;
import org.mindswap.owl.OWLErrorHandler;

import org.mindswap.owl.OWLDataProperty;
import org.mindswap.owls.OWLSFactory;
import org.mindswap.owls.OWLSKnowledgeBase;
// new for 1.1.0 
import org.mindswap.owls.OWLSListFactory;
import org.mindswap.owl.OWLFactory;
import org.mindswap.query.ValueMap;
import org.mindswap.owl.OWLKnowledgeBase;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.grounding.AtomicGrounding;
import org.mindswap.owls.grounding.WSDLAtomicGrounding;
import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.ParameterList;
import org.mindswap.owls.process.InputList;
import org.mindswap.owls.process.OutputList;
import org.mindswap.owls.process.ConditionList;
import org.mindswap.owls.process.ResultList;
import org.mindswap.owls.process.Result;
import org.mindswap.owls.process.Parameter;
import org.mindswap.owls.process.Input;
import org.mindswap.owls.process.Output;
import org.mindswap.owls.process.Condition;
import org.mindswap.owls.profile.Profile;
import org.mindswap.owls.profile.ServiceCategory;
import org.mindswap.owls.service.Service;
import org.mindswap.owls.process.execution.ProcessExecutionEngine;

import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.util.ModelLoader;

import org.mathweb.core.*;
import org.mathweb.onto.*;

import org.mathweb.util.Java2WSDL;
import org.mathweb.reasoning.logic.*;
import org.mathweb.serv.ReasoningService;
import org.mathweb.serv.ReasoningProblem;
import org.mathweb.serv.Query;

public class OWLSParser extends BasicService
{
  public static final Namespace RDF_NS = Namespace.getNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
  public static final Namespace RDFS_NS = Namespace.getNamespace("rdfs","http://www.w3.org/2000/01/rdf-schema#");
  public static final Namespace OM_NS = Namespace.getNamespace("om","http://www.openmath.org/OpenMath");
  public static final Namespace OMDOC_NS = Namespace.getNamespace("omdoc","http://www.openmath.org/omdoc");
  public static final Namespace OWL_NS = Namespace.getNamespace("owl","http://www.w3.org/2002/07/owl#");
  public static final Namespace OWLX_NS = Namespace.getNamespace("owlx","http://www.w3.org/2003/05/owl-xml#");
  public static final Namespace RULEML_NS = Namespace.getNamespace("ruleml","http://www.ruleml.org/inspec#");
  public static final Namespace SWRL_NS = Namespace.getNamespace("swrl","http://www.daml.org/services/owl-s/1.1/generic/swrlx.owl#");
  public static final Namespace WSDL_NS = Namespace.getNamespace("wsdl","http://schemas.xmlsoap.org/wsdl/");
  public static final Namespace WSDL_SOAP_NS = Namespace.getNamespace("wsdlsoap","http://schemas.xmlsoap.org/wsdl/soap/");
  public static final Namespace PROFILE_NS = Namespace.getNamespace("profile" , "http://www.daml.org/services/owl-s/1.1/Profile.owl#");
  public static final Namespace PROCESS_NS = Namespace.getNamespace("process" , "http://www.daml.org/services/owl-s/1.1/Process.owl#");
  
  /**   service descriptions  */

  public static final XPath serviceLocation = newXPath("./wsdl:service/wsdl:port/wsdlsoap:address/@location");
  public static final XPath wsdlName = newXPath("./wsdl:service/@name");
  //  public static final XPath serviceOperation = newXPath("./wsdl:operation/@name");
  

  /**   queries  */
  public static final XPath queryidx = newXPath("./process:AtomicProcess/@rdf:ID");
  public static final XPath queryProfileidx = newXPath("./profile:Profile/@rdf:ID");

  /**     I/O parameters    */
  public static final XPath namex = newXPath("./@name");
  public static final XPath contentsx = newXPath("./rdf:RDF");

  public static final XPath atomClassx = newXPath("./owlx:Class/@owlx:name");
  public static final XPath atomVarx = newXPath("./ruleml:var");
  public static final XPath atomArg1x = newXPath("./swrl:argument1/@rdf:resource");
  public static final XPath atomArg2x = newXPath("./swrl:argument2/@rdf:resource");

  public static final XPath atomPropx = newXPath("./swrl:propertyPredicate/@rdf:resource");

  public static final XPath bodyx = newXPath("./swrl:body");
  public static final XPath headx = newXPath("./swrl:head/*");
  public static final XPath listFirstx = newXPath("./swrl:AtomList/swrl:first/*");
  public static final XPath listRestx = newXPath("./swrl:AtomList/swrl:rest/*");
  
  private ArrayList condNames;
  
  protected SAXBuilder builder;
  protected static boolean expandEntities = false;
  private static Logger logger = Logger.getLogger("OWLSParser");

// new for 1.1.0
  private OWLKnowledgeBase kb;
//   private OWLSKnowledgeBase kb;
  OWLCache cache;
  
  private String owlsDir = getSystemProperty("mathserv.owls.dir", "MATHSERV_OWLS_DIR");
  private String wsdlFileDir = getSystemProperty("mathserv.wsdl.directory", "MATHSERV_WSDL_DIRECTORY");

  private static URI rdfID = URI.create("http://www.w3.org/1999/02/22-rdf-syntax-ns#ID");
  private static URI exprLang = URI.create("http://www.daml.org/services/owl-s/1.1/generic/Expression.owl#expressionLanguage");
  private static URI exprBody = URI.create("http://www.daml.org/services/owl-s/1.1/generic/Expression.owl#expressionBody");
  private static URI servNameURI = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceName");
  private static URI serviceParameter = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceParameter");
  private static URI onParameter = URI.create("http://www.mathweb.org/owl/mathweb.owl#onParameter");
  private static URI inClass = URI.create("http://www.mathweb.org/owl/mathweb.owl#inClass");
  private static URI serviceName = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceName");
    
  private static URI paramValue = URI.create("http://www.daml.org/services/owl-s/1.1/Process.owl#parameterValue");
  
  
  private OWLJDom owlJDom = OWLJDom.getInstance();

  public OWLSParser()
  {
    builder = new SAXBuilder();
    builder.setExpandEntities(expandEntities);
    
    OWLCache cache = new OWLCacheImpl();
    //    System.out.println("owls dir: " + owlsDir);
    // set directory that contains service.idx
    cache.setLocalCacheDirectory(owlsDir);
    cache.setForced(true);
    
// new for 1.1.0 
    kb = OWLFactory.createKB();
    //    kb = OWLSFactory.createKB();
    OWLReader reader = kb.getReader();
    reader.setCache(cache);
    //    reader.setErrorHandler(new Listener());

//     owlsReader = OWLFactory.createOWLSReader("1.1");
    
    logger.setLevel(Level.ALL);
  }
  
  public org.mathweb.serv.Query parseQueryFile(String inputFileName) {
    return parseQuery(readFile(inputFileName));
    
  }
  public org.mathweb.serv.Query parseQuery(String description) {
    org.mathweb.serv.Query query = new org.mathweb.serv.Query();

    File inputFile = null;
    try{ 
      inputFile = File.createTempFile("owls-query", ".owl");
      String inputFileName = inputFile.getAbsolutePath();
      URI uri =  URI.create("file://" + inputFileName);
      writeFile(inputFile, description);
      
//       Model model = ModelLoader.loadModel("file:" + inputFileName);
//        Element root = owlJDom.getRootElement(description);
//       String queryURI = root.getAttributeValue("base", Namespace.XML_NAMESPACE);
//       String queryID = queryidx.valueOf(root);
//       //      System.out.println("processuri: " + queryURI + "#" + queryID);
//       Resource processInfo = model.getResource(queryURI + "#" + queryID);
      //      System.out.println("processInfo: " + processInfo);
      
      Service owlserv = kb.readService(uri);
      Process process = owlserv.getProcess();
      URI procuri = owlserv.getURI();
      ////       String queryURI = procuri.getScheme() + procuri.getUserInfo() + procuri.getAuthority();
      String queryURI = "http://" + procuri.getAuthority() + procuri.getPath();
      //      System.out.println("process: " + queryURI); 

      query.setOwlsService(owlserv);
      
      ReasoningProblem prob = getProblem(process.getInputs(),
                                         process.getOutputs(),
                                         process.getConditions(),
                                         process.getResults(),
                                         process.getLocals(),
                                         process.toString());
      query.setProblem(prob);
      query.name = procuri.toString().substring(7);
      query.setURI(queryURI);

      kb.unload(uri);
      return query;
    } catch (Exception e){
      System.err.println("no solution found!");
      e.printStackTrace();
      return null;
    } finally{
      if (inputFile != null) inputFile.delete();
    }
  }

  public ReasoningService parseService(String description) {
    return parseService(description, null);
  }
  
  public ReasoningService parseService(String description, String newWsdl) {
    File inputFile = null;
    try{
      inputFile = File.createTempFile("owls-service", ".owl");
      writeFile(inputFile, description);
      return parseServiceFile(inputFile.getAbsolutePath(), newWsdl);
    } catch(Exception e){
      logger.debug("[OWLS] Exception " + e);
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
      } catch(Exception e){ logger.debug("[OWLS] Exception " + e); }
    }
    return null;
  }

  public ReasoningService parseServiceFile(String descriptionFileName) {
    return parseServiceFile(descriptionFileName, null);
  }
  
  public ReasoningService parseServiceFile(String descriptionFileName, String newWsdl) {
    ReasoningService serv = new ReasoningService();
    Service owlserv = null;
    URI uri =  URI.create("file://" + descriptionFileName);
    
    try{
      //      System.out.println("filename " + new URI("file://" + descriptionFileName));
      //      Service owlserv = owlsReader.read(new URI("file://" + descriptionFileName));
      //      System.out.println("service " + owlserv);

      owlserv = kb.readService(uri);
      Process process = owlserv.getProcess();
      Profile profile = owlserv.getProfile();

      //      URI servURI = owlserv.getFileURI();
      URI procuri = owlserv.getURI();
      String servURI = "http://" + procuri.getAuthority() + procuri.getPath();
      //      String servURI = owlserv.getURI().toString();
      
      
      serv.name = profile.getServiceName();
      //      serv.name = servURI.toString().substring(7);
      //      serv.name = profile.getProperty(nameProp).getLexicalValue();
//       OWLDataProperty nameProp = kb.getDataProperty(serviceName);
//       serv.name = profile.getProperty(nameProp).getLexicalValue();

      serv.setURI(servURI);

      
      serv.setOwlsService(owlserv);
      
//       ServiceCategory servCat = profile.getCategory();
//       System.out.println("cat: " + servCat);
//       String cat = servCat.toString();
//      serv.directive = cat; 


      ReasoningProblem prob = getProblem(process.getInputs(),
                                         process.getOutputs(),
                                         process.getConditions(),
                                         process.getResults(),
                                         process.getLocals(),
                                         process.toString());
      serv.problem = prob;
      
      //      System.out.println("service: " + serv);
      //      System.out.println("process: " + process.getLabel());
      
      if (newWsdl != null){
        String wsdlFileName = wsdlFileDir + sep + serv.name + ".xml";
        writeFile(wsdlFileName, newWsdl);
        Grounding gr = owlserv.getGrounding();
        List grs = gr.getAtomicGroundings();
        for(int i=0; i< grs.size(); i++){
          AtomicGrounding ag = (AtomicGrounding) grs.get(i);
          if (ag instanceof WSDLAtomicGrounding){
            WSDLAtomicGrounding wag = (WSDLAtomicGrounding) ag;
            wag.setWSDL(URI.create("file:" + wsdlFileName));
          }
        }
      }
      //      kb.unload(uri);
      
      return serv;
    } catch(Exception e){
      System.out.println(" error " + e.getMessage());
      e.printStackTrace();
      return null;
    }
  }

  private ReasoningProblem getProblem(InputList owlins, 
                                      OutputList owlouts, 
                                      ConditionList owlpres, 
                                      ResultList owlress,
                                      ParameterList locals, 
                                      String probName
                                      )      
  {
//     System.out.println("ins: " + owlins);
//     System.out.println("outs: " + owlouts);
//     System.out.println("preconds: " + owlpres);
//     System.out.println("results: " + owlress);
    //    owleffs,
    ReasoningProblem prob = new ReasoningProblem();
    
    ArrayList inputs = new ArrayList(owlins);
    ArrayList outputs = new ArrayList(owlouts);
    ArrayList preconds = new ArrayList(owlpres);
    // new for 1.1.0 
    ConditionList owleffs = OWLSListFactory.createConditionList();
    //    ConditionList owleffs = OWLSFactory.createConditionList();
    
    HashMap locs = new HashMap();
    ArrayList effects = new ArrayList();
    
    for(int i=0; i<owlress.size(); i++){
      Result res = owlress.resultAt(i);
      Object eff = res.getEffect();
      //      System.out.println("result effect: " + eff);
      owleffs.add(eff);
//       ConditionList clist = res.getEffects();
//       owleffs.addAll(clist);
    }
    
    for(int i=0; i<locals.size(); i++) {
      Parameter par = locals.parameterAt(i);
      String name = par.getLocalName();
      OWLType type = par.getParamType();
      String typeStr = type.getURI().getFragment();
      //      System.out.println("local : " + name + "::" + typeStr);
      locs.put(name, typeStr);
    }
    
    for(int i=0; i<inputs.size(); i++){
      Input in = owlins.inputAt(i);
      String name = in.getLocalName();
      String contents = null; 
      OWLDataProperty valueProp = kb.getDataProperty(paramValue);
      //      System.out.println("valueProp: " + valueProp);
      if ((valueProp != null) && (in.hasProperty(valueProp))) {
        contents = in.getProperty(valueProp).getLexicalValue();
      }
      
      //      System.out.println("contents: " + contents);
//       try{
//         Element cont = owlJDom.getRootElement(contents);
//         contents = owlJDom.newXMLDocument((Element) cont.clone());
//       } catch(Exception e){
//         e.printStackTrace();
//         contents = null;
//       }
//       System.out.println("contents: " + contents);
      
      OWLType type = in.getParamType();
      String typeStr = type.getURI().getFragment();
      org.mathweb.serv.Input newIn = new org.mathweb.serv.Input(name, typeStr, contents);
      inputs.set(i, newIn);
    }

    /* TODO: THIS IS A BIG HACK!!! It simply puts all inputs with type TimeResource at the 
       end of the list. We have to fix that. */
    int endIndex = inputs.size()-1;
    for(int i=0; i<inputs.size(); i++){
      org.mathweb.serv.Input in = (org.mathweb.serv.Input) inputs.get(i);
      String typeStr = in.getSignature();
      if (typeStr.equals("TimeResource")){
        Object lastIn = inputs.get(endIndex);
        inputs.set(endIndex, in);
        inputs.set(i, lastIn);
        break;
      }
    }
    
    for(int i=0; i<outputs.size(); i++){
      Output out = owlouts.outputAt(i);
      String name = out.getLocalName();

      OWLType type = out.getParamType();
      String typeStr = type.getURI().getFragment();
      //      System.out.println("output: " + name + "::" + type);
      org.mathweb.serv.Output newOut = new org.mathweb.serv.Output(name, typeStr);
      outputs.set(i, newOut);
    }
    
    condNames = new ArrayList();
    preconds = parseConds(owlpres);
    for(int i=0; i<preconds.size(); i++){
      CachedHashCodeObject s = (CachedHashCodeObject) preconds.get(i);
      org.mathweb.serv.PreCondition p = new org.mathweb.serv.PreCondition(); 
      p.condition = s; 
      p.setName((String) condNames.get(i));
      preconds.set(i,p);
    }
    
    condNames = new ArrayList();
    effects = parseConds(owleffs);
    for(int i=0; i<effects.size(); i++){
      CachedHashCodeObject s = (CachedHashCodeObject) effects.get(i);
      org.mathweb.serv.PostCondition p = new org.mathweb.serv.PostCondition(); 
      p.condition = s; 
      p.setName((String) condNames.get(i));
      effects.set(i,p);
    }
    
    prob.name = probName;
    prob.inputs = inputs;
    prob.outputs = outputs;
    prob.preconds = preconds;
    prob.postconds = effects;
    prob.locals = locs;
    
    return prob;
  }
  

  /**  every pre-condition or post-condition becomes a Cached */
  public ArrayList parseConds(ConditionList pc)
  {
    ArrayList roots = new ArrayList();
    if (!pc.isEmpty())
      for(int i=0; i<pc.size(); i++){ 
        // pre-condition element
        try {
          Condition cond = pc.conditionAt(i);
          String expr = cond.getProperty(kb.getDataProperty(exprBody)).getLexicalValue();
          //          System.out.println("condition: " + expr + expr.getClass().getName());
          StringReader reader = new StringReader(expr);
          Document sDoc = builder.build(reader);
          Element root = sDoc.getRootElement();
          //          System.out.println("root: " + root);
          roots.add(root);
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    return parseRDFConds(roots);
  }

//   private ReasoningProblem parseProblem(Element problem)      {
//     ReasoningProblem prob = new ReasoningProblem();
//     try{
//       prob.name = namex.valueOf(problem);
//       ArrayList inputs = new ArrayList(inputsx.selectNodes(problem));
//       ArrayList outputs = new ArrayList(outputsx.selectNodes(problem));
//       ArrayList preconds = new ArrayList(precondsx.selectNodes(problem));
//       ArrayList postconds = new ArrayList(postcondsx.selectNodes(problem));
//       System.out.println("problem: " +problem);
//       System.out.println("inputs: " +inputs);
      
//       String name = null;
//       String sig = null;
      
//       for(int i=0; i<inputs.size(); i++){
// 	Element e = (Element) inputs.get(i);
// 	name = namex.valueOf(e);
// 	sig = signaturex.valueOf(e);	
// 	// try to split the namespace and the OWL class name from signature
// 	String[] st= sig.split("#");
// 	String cn;
// 	//System.out.println("split: " + st.length + " " +st[0]  + " " +st[1] );
// 	if (st.length>1) cn = st[1];
// 	else cn = sig;
// 	// get the content of the input (in case we are in a query queries)
// 	Element cont = (Element) contentsx.selectSingleNode(e);
// 	System.out.println("input " + name+ "::" + cn+ "= " +cont);
// 	String doc;
// 	if (cont != null) {
// 	  doc = owlJDom.newXMLDocument((Element) cont.clone());
// 	} else doc = "";
// 	Input in = new Input(name, cn, doc);
// 	inputs.set(i,in);
//       }
      
//       for(int i=0; i<outputs.size(); i++){
// 	Element e = (Element) outputs.get(i);
// 	name = namex.valueOf(e);
// 	sig = signaturex.valueOf(e);
// 	String[] st= sig.split("#");
// 	String cn;
// 	if (st.length>1) cn = st[1];
// 	else cn = sig;
// 	Output in = new Output(name,cn);
// 	outputs.set(i,in);
//       }
//       preconds = parseConds(preconds);
//       for(int i=0; i<preconds.size(); i++){
// 	CachedHashCodeObject s = (CachedHashCodeObject) preconds.get(i);
// 	PreCondition p = new PreCondition(); 
// 	p.condition = s; 
// 	preconds.set(i,p);
//       }
//       postconds = parseConds(postconds);
//       for(int i=0; i<postconds.size(); i++){
// 	CachedHashCodeObject s = (CachedHashCodeObject) postconds.get(i);
// 	PostCondition p = new PostCondition(); 
// 	p.condition = s; 
// 	postconds.set(i,p);
//       }
//       prob.inputs=inputs;
//       prob.outputs=outputs;
//       prob.preconds = preconds;
//       prob.postconds = postconds;
//       return prob;
//     } catch(Exception e){
//       e.printStackTrace();
//       return prob;
//     }
//   }
  

  /**  
       @pc a list of root elements, i.e. rdf:RDF elements
  */
  public ArrayList parseRDFConds(ArrayList pc)
  {
    ArrayList allConds = new ArrayList();
    
    if (!pc.isEmpty())
      for(int i=0; i<pc.size(); i++){ 
        // pre-condition element
        Element cond = (Element) pc.get(i);
        ArrayList cs = new ArrayList(cond.getChildren());
        for(int k=0; k<cs.size(); k++){
          cond = (Element) cs.get(k);
          condNames.add(cond.getAttributeValue("ID", RDF_NS));
          //          System.out.println("cond-root: " + cond + " " + cond.getName());
          allConds.add(parseCond(cond));
        }
      }
    return allConds;
  }
  
  public ArrayList parseAllConds(ArrayList pc)
  {
    for(int i=0; i<pc.size(); i++){ 
      Element cond = (Element) pc.get(i);
      pc.set(i, parseCond(cond));
    }
    return pc;
  }
  
  private CachedHashCodeObject parseCond(Element cond)  {
    if (cond.getName().equals("classAtom")){
      return parseClassAtom(cond);
    }
    else if (cond.getName().equals("individualPropertyAtom")){
      return parseIndProperty(cond);
    }
    else if (cond.getName().equals("and")){
      return parseConjunct(cond);
    }
    else if (cond.getName().equals("Imp")){
      return parseImplication(cond);
    }
    return new Constant("Condition of type " + cond.getName() + " not known!");
  }
  
  
  public CachedHashCodeObject parseConjunct(Element cond){
    ArrayList cs = new ArrayList(cond.getChildren());
    ArrayList literals = parseAllConds(cs);
    Conjunct con = new Conjunct(literals);
    return con;
  }
    
  public CachedHashCodeObject parseImplication(Element cond){
    try{
      
      Element body =  (Element) bodyx.selectSingleNode(cond);
      Element head =  (Element) headx.selectSingleNode(cond);
      //      System.out.println("body: " + body);
      //      System.out.println("head: " + head);
      ArrayList bodys = new ArrayList();
      boolean end = false;
      
      while (!end){
        Element first = (Element) listFirstx.selectSingleNode(body);
        //        System.out.println("first: " + first);
        bodys.add(first);
        Element rest =  (Element) listRestx.selectSingleNode(body);
        //        System.out.println("rest: " + rest);
        if (rest == null) end = true;
        else {
          body = rest;
        }
      }
      ArrayList literals = parseAllConds(bodys);
      //      System.out.println("lits: " + literals);
      Literal lit = (Literal) parseCond(head);
      
      return new Implication(lit ,literals);    
    }catch(Exception e){
      e.printStackTrace();
      return null;
    }
  }
  
    
  public CachedHashCodeObject parseClassAtom(Element cond){
    try{
      String cn = atomClassx.valueOf(cond);
      String vn = ((Element)atomVarx.selectSingleNode(cond)).getText();
      Variable var = new Variable(vn.toLowerCase());
    
      //      System.out.println("classAtom: " + cond + " " + cn);
      String[] st= cn.split("#");
      String name, ns;
      //      System.out.println("split: " + st.length + " " +st[0]  + " " +st[1] );
      if (st.length>1){  name = st[1]; ns=st[0];      }
      else {name = cn; ns = OWLJDom.MW_NS.getURI();  }

      Constant classSym = new Constant(name);
      Predicate type = new Predicate("type"); 
      //RDF_NS.getURI());
      Argument[] args = new Argument[2];
      args[0] = var;
      args[1] = classSym;
      return new Literal(type, args, true);
    }catch (org.jdom.JDOMException e) {
      return null;
    }
  }

  public CachedHashCodeObject parseIndProperty(Element cond){
    try{
      //      System.out.println("indProperty: " + cond);
//       System.out.println("propPred: " + propPred);
//       Element e = (Element) propPred;
//       System.out.println("propPred: " + propPred + " " + e.getAttribute("resource", RDF_NS)); 
//       List ats = e.getAttributes();
//       for(int i=0; i< ats.size(); i++){
// 	Attribute a = (Attribute) ats.get(i);
// 	System.out.println("att: " + ats.get(i) + a.getName() + " " + a.getNamespace());
//       }
      Argument[] args = new Argument[2];

      String prop = atomPropx.valueOf(cond);
      //      System.out.println("prop: " + prop);
      String[] st= prop.split("#");
      String name;
      if (st.length>1) name = st[1];
      else name = prop;
      Predicate propSym = new Predicate(name);

      String vn0 = atomArg1x.valueOf(cond);
      String vn1 = atomArg2x.valueOf(cond);
      //      System.out.println("args: " + vn0 + " " + vn1);
      st= vn0.split("#");
      if (st.length>1) args[0] = new Variable(st[1].toLowerCase());
      else  args[0] = new Variable(vn0.toLowerCase());

      st= vn1.split("#");
      if (st.length>1) args[1] = new Variable(st[1].toLowerCase());
      else  args[1] = new Variable(vn1.toLowerCase());
      //      System.out.println("args: " + args[0] + " " + args[1]);

      return new Literal(propSym, args, true);
    }catch (org.jdom.JDOMException e) {
      return null;
    }
  }
  
  
  public static XPath newXPath(String x)
  {
    try{
      XPath xpath =  XPath.newInstance(x);
      xpath.addNamespace(OWLJDom.MW_NS);
      xpath.addNamespace(RDF_NS);
      xpath.addNamespace(RDFS_NS);
      xpath.addNamespace(OMDOC_NS);
      xpath.addNamespace(OWL_NS);
      xpath.addNamespace(OWLX_NS);
      //      xpath.addNamespace(OM_NS);
      //      xpath.addNamespace(RULEML_NS);
      xpath.addNamespace(SWRL_NS);
      xpath.addNamespace(WSDL_NS);
      xpath.addNamespace(WSDL_SOAP_NS);
      xpath.addNamespace(PROFILE_NS);
      xpath.addNamespace(PROCESS_NS);
      return xpath;
    } catch(Exception e){
      return null;
    }
  }

  public static void main(String argv[]) throws Exception {
    OWLSParser c = new OWLSParser();
    ReasoningService[] s  = new ReasoningService[29];
    String msHome = System.getProperty("mathserv.home");
    int counter =1;

    // /*
    s[0] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/OtterService.owl");
    System.out.println(s[0].toProdigyOperator(counter++));

    s[1] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/EXmlService.owl");
    System.out.println(s[1].toProdigyOperator(counter++));

//     s[2] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/TrampCNFService.owl");
//     System.out.println(s[2].toProdigyOperator(counter++));

//     s[3] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/TrampNDService.owl");
//     System.out.println(s[3].toProdigyOperator(counter++));
    
//     s[4] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/OtterfierService.owl");
//     System.out.println(s[4].toProdigyOperator(counter++));
    
//     s[5] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/FlotterCNFService.owl");   
//     System.out.println(s[5].toProdigyOperator(counter++));

//     s[6] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/ProjectNDProof.owl");   
//     System.out.println(s[6].toProdigyOperator(counter++));

//     s[7] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/TSTPAnalyserService.owl");   
//     System.out.println(s[7].toProdigyOperator(counter++));

    
    // example with Java Grounding
//     s[8] = c.parseServiceFile(msHome + "/thirdParty/owl-s-svn/trunk/src/JGroundingTest.owl");   
//     Hashtable map = new Hashtable();
//     map.put("FirstParam", new Integer(100));
//     map.put("SecParam", new Double(100.00));
//     ValueMap resMap = s[8].invokeOWLS(map);
//     System.out.println("result: " + resMap);
        
//     s[9] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/VampireService.owl");
//     System.out.println(s[9].toProdigyOperator(counter++));
 
//     s[10] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/VampireXmlService.owl");
//     System.out.println(s[10].toProdigyOperator(counter++));

//     s[11] = c.parseServiceFile("/home/jzimmer/Flight2.owl");
//     System.out.println(s[11].toProdigyOperator(counter++));

//     for (int i =0; i<4; i++){
//       //       System.out.println("service " + s[i].toString());
//       System.out.println("action: " + s[i].toPlanningOperator().toLongString());
//     }
//      org.mathweb.serv.Query  query;
//      //     query = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/nd-query1.owl");
//      query = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/result-query1.owl");
//      System.out.println("query " + query.toProdigyProblem());
    
  }

}

 
      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

