package org.mathweb.reasoning.logic;

import java.util.*;
import java.io.Serializable;

/**
 * Represents the set of unequality constraints on bindings.
 *
 * An immutable class.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 27, 2003
 */
public class BindingConstraints extends CachedHashCodeObject implements Serializable {
    /** A map from variables to HashSet of variables */
    private HashMap constraints;

    /**
     * Create a new empty constraint list
     */
    public BindingConstraints() {
        constraints = new HashMap();
    }

    /**
     * A a copy constructor
     */
    public BindingConstraints(BindingConstraints c) {
      this.constraints = new HashMap();
      for (Iterator it=c.constraints.keySet().iterator(); it.hasNext(); ) {
	Object key = it.next();
	this.constraints.put(key, ((HashSet)c.constraints.get(key)).clone());
      }
    }


    /**
     * @return True if there are no binding constraints in this object.
     */
    public boolean isEmpty() {
        return constraints.isEmpty();
    }

    protected void computeHashCode() {
        myhashcode = constraints.hashCode();
    }

    public boolean equals(BindingConstraints other) {
        return constraints.equals(other.constraints);
    }

    public boolean equals(Object other){
        try {
            return equals((BindingConstraints)other);
        } catch (ClassCastException e) {
            return false;
        }
    }

    /**
     * @return The number of constraints.
     */
    public int getNumConstraints() {
        return constraints.size();
    }

    /**
     * Add a set of binding constraints to this set.
     *
     * @param other The binding constraints to add.
     * @return These binding constraints combined with the others.
     */
    public BindingConstraints addAll(BindingConstraints other) {
        BindingConstraints b = new BindingConstraints(this);
        
        for (Iterator keyit=other.constraints.keySet().iterator(); keyit.hasNext(); ) {
            Variable key = (Variable) keyit.next();
            HashSet cons = (HashSet) other.constraints.get(key);
            for (Iterator valit=cons.iterator(); valit.hasNext(); ) {
                Argument value = (Argument) valit.next();
                addConstraintMutable(b, key, value);
            }
        }
        return b;
    }

    /**
     * Internal helper function used for adding constraints to the set in a mutable way.
     */
    private static void addConstraintMutable(BindingConstraints b, Argument x, Argument y) {
        if (x == null || y == null)
            throw new IllegalArgumentException("Can not add null variables as constraints");

        if (x.equals(y))
            throw new IllegalArgumentException("Adding constraint <" + x + " != " + y +">");

        // add y to the set of arguments X can not equal
        if (x.isVariable()) {
            HashSet cons = (HashSet) b.constraints.get(x);
            if (cons == null) {
                cons = new HashSet();
                b.constraints.put(x, cons);
            }
            cons.add(y);
        }

        // add x to the set of arguments y can not equal
        if (y.isVariable()) {
            HashSet cons = (HashSet) b.constraints.get(y);
            if (cons == null) {
                cons = new HashSet();
                b.constraints.put(y, cons);
            }
            cons.add(x);
        }        
    }

    /**
     * Add a new constraint to the table so that the two given variables are constrained.
     *
     * @return A copy of the constraint table with the two new constraints added.
     * @exception IllegalArgumentException Thrown if x is the same variable as y.
     */
    public BindingConstraints addConstraint(Argument x, Argument y) {
        BindingConstraints b = new BindingConstraints(this);

        addConstraintMutable(b, x, y);
            
        return b;
    }

    /**
     * @return The set of arguments the given variable can not equal.
     */
    public Set getConstraints(Variable x) {
        HashSet set = (HashSet) constraints.get(x);
        if (set == null)
            return new HashSet();
        return (Set) set.clone();
    }

    /**
     * @return True if x is constrained with respect to y.
     * @exception IllegalArgumentException Thrown if x is the same variable as y.
     */
    public boolean isConstrained(Variable x, Argument y) {
        Set cons = (Set) constraints.get(x);
	if(null == cons)
	    return false;
        return cons.contains(y);
    }

    /**
     * @return True if x != y or y != x or if both are constants true if !x.equals(y)
     */
    public boolean isConstrained(Argument x, Argument y) {
	if(x.getClass() == Variable.class)
	    return isConstrained((Variable)x, y);
	else if(y.getClass() == Variable.class)
	    return isConstrained((Variable)y, x);
	else{ //they're both constants
	    if(x.equals(y))
	      return false;
// 	    if (reasoner != null){
// 	      System.out.println("testing binding constraint subsumption " + x.getName() + " < " +  y.getName() + " " + reasoner.subClassOf(x.getName(), y.getName()));
// 	      if (reasoner.subClassOf(x.getName(), y.getName()))
// 		return false;
// 	      else
// 		return true;
// 	    }
	    return true;
	}
    }

    /**
     * Checks if given assignment is a violation of the binding constrants.
     *
     * @param v The variable to check.
     * @param c The argument to assign to.
     * @param bt The binding table we are doing this assignment in, to check all equal variables.
     */
    public boolean isViolation(Variable v, Argument c, BindingTable bt){
  	HashSet s1 = bt.getEqualitySet(v);
  	HashSet s2 = bt.getEqualitySet(c);
 	s1.add(v);
 	s2.add(c);

   	for(Iterator s1it = s1.iterator(); s1it.hasNext(); ){
 	    Argument s1cur = (Argument)s1it.next();
   	    for(Iterator s2it = s2.iterator(); s2it.hasNext(); ){
   		if(isConstrained(s1cur, (Argument)s2it.next()))
   		    return true;
 	    }
 	}
  	return false;
    }

    /**
     * Apply bindings to this set of binding constraints.
     *
     * @param compressedBindings The bindings to apply.
     * @return The binding constraints with bindings applied.
     */
    public BindingConstraints applyBindings(HashMap compressedBindings){
        // NOTE: if new datamembers get added to this class, make sure to copy them over here
	BindingConstraints newBc = new BindingConstraints();
 	for(Iterator consKeyIt = constraints.keySet().iterator(); consKeyIt.hasNext(); ){
	    Variable consKey = (Variable)consKeyIt.next();
	    Argument newKey = (Argument)compressedBindings.get(consKey);
	    if(null == newKey)
		newKey = consKey;
	    if(newKey.isVariable()){//if v1 != v2, and v1->a, drop the binding altogehter, b/c we don't allow constant to variable bindings
		for(Iterator consValsIt = getConstraints(consKey).iterator(); consValsIt.hasNext(); ){
		    Argument oldVal = (Argument)consValsIt.next();
		    Argument newVal = (Argument)compressedBindings.get(oldVal);
		    if(null == newVal)
			newVal = oldVal;
		    addConstraintMutable(newBc, (Variable)newKey, newVal);
		}
	    }
 	}
	return newBc;
    }

    /**
     * @return an empty BindingTable if I have no constraints or a BT with a mapping for each constraint
     *
     **/
    public BindingTable toBindingTable(BindingConstraints bc){
	BindingTable bt = new BindingTable();
	for(Iterator keyIt = constraints.keySet().iterator(); keyIt.hasNext(); ){
	    Variable key = (Variable)keyIt.next();
	    Set vals = getConstraints(key);
	    for(Iterator valsIt = vals.iterator(); valsIt.hasNext(); ){
		bt = bt.addBinding(key, (Argument)valsIt.next(), bc);
	    }
	}
	return bt;
    }

    public Set getSet() {
	HashSet set = new HashSet();

        for (Iterator it=constraints.keySet().iterator(); it.hasNext(); ) {
            Variable first = (Variable) it.next();
            for (Iterator it2=((HashSet)constraints.get(first)).iterator(); it2.hasNext(); ) {
                Argument second = (Argument) it2.next();
		set.add(new BindingConstraint(first, second));
            }
        }

        return set;
    }

    public String toString() {
        String ret = "";

        for (Iterator it=constraints.keySet().iterator(); it.hasNext(); ) {
            Variable first = (Variable) it.next();
            for (Iterator it2=((HashSet)constraints.get(first)).iterator(); it2.hasNext(); ) {
                Argument second = (Argument) it2.next();
                ret += "\t\t<" + first + " != " + second + ">\n";
            }
        }

        return ret;
    }
}
