package org.mathweb.reasoning.logic;

import java.util.*;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.reasoning.partialorderplanner.NameGen;

/**
 * A BindingTable contains a list of positive and negative bindings.
 * A binding maps a varible to an argument(another variable or a constant).
 *
 * This class is immutable.
 *
 * @author Daniel Terry
 */
public class BindingTable {
  private final HashMap positiveBindings;
  private OntoReasoner reasoner = null;

  /**
   * Construct a new empty binding table.
   */
  public BindingTable(){
    positiveBindings = new HashMap();
  }


  /**
   * Construct a new empty binding table.
   */
  public BindingTable(OntoReasoner reasoner){
    positiveBindings = new HashMap();
    this.reasoner = reasoner;
  }

  /**
   * Copy constructor.
   *
   * @param copyMe The binding table to copy.
   */
  public BindingTable(BindingTable copyMe){
    this.positiveBindings = (HashMap) copyMe.positiveBindings.clone();
    this.reasoner = copyMe.getReasoner();
  }
  
  /**
   * Internal function for adding positive bindings.
   *
   * Doesn't check for consistency with existing bindings.
   */
  private BindingTable addPositiveBinding(Variable v, Argument a){
    BindingTable newBT = new BindingTable(this);
    if(null == newBT.positiveBindings.get(v))
      newBT.positiveBindings.put(v, new Vector(1));
    ((Vector)newBT.positiveBindings.get(v)).add(a);
    return newBT;
  }

  /**
   * Internal function for removing all bindings of variable v.
   *
   * Not recursive.
   */
  private BindingTable removeAllBindings(Variable v){
    BindingTable newBT = new BindingTable(this);
    newBT.positiveBindings.remove(v);
    return newBT;
  }


  /**
   * Calls the appropriate addbinding with the run time types of a1 and a2
   *
   * @param a1 The argument to bind
   * @param a2 The other argument to bind.
   * @param bc The binding constraints that apply.
   * @return The binding table that results from this adding this binding.
   */
  public BindingTable addBinding(Argument a1, Argument a2, BindingConstraints bc){
    if(a1.getClass() == Constant.class){
      if(a2.getClass() == Constant.class)
	return addBinding((Constant)a1, (Constant)a2, bc); //Constant to Constant
      else if(a2.getClass() == Variable.class)
	return addBinding((Variable)a2, (Constant)a1, bc); //Variable to Constant
    }
    else if(a1.getClass() == Variable.class){
      if(a2.getClass() == Constant.class)
	return addBinding((Variable)a1, (Constant)a2, bc); //Constant to Variable (becomes Variable to Constant)
      else if(a2.getClass() == Variable.class)
	return addBinding((Variable)a1, (Variable)a2, bc); //Variable to Variable
    }

    System.out.println("BindingTable::addBinding: ERROR: Argument type is not Variable or Constant");
    return null;
  }

  /**
   * Calls the appropriate addbinding with the run time types of a1 and a2
   *
   * @param a1 The argument to bind
   * @param a2 The other argument to bind.
   * @param bc The binding constraints that apply.
   * @param reasoner An OWL ontology reasoner for subsumption testing
   * @return The binding table that results from this adding this binding.
   */
  public BindingTable addBinding(Argument a1, Argument a2, BindingConstraints bc, OntoReasoner reasoner){
    if(a1.getClass() == Constant.class){
      if(a2.getClass() == Constant.class)
	return addBinding((Constant)a1, (Constant)a2, bc); //Constant to Constant
      else if(a2.getClass() == Variable.class)
	return addBinding((Variable)a2, (Constant)a1, bc); //Variable to Constant
    }
    else if(a1.getClass() == Variable.class){
      if(a2.getClass() == Constant.class)
	return addBinding((Variable)a1, (Constant)a2, bc); //Constant to Variable (becomes Variable to Constant)
      else if(a2.getClass() == Variable.class)
	return addBinding((Variable)a1, (Variable)a2, bc); //Variable to Variable
    }

    System.out.println("BindingTable::addBinding: ERROR: Argument type is not Variable or Constant");
    return null;
  }

  /**
   * @return copy of this if c1.equals(c2) or if c1 is a subconcept of c2
   * modulo the reasoner, null otherwise.
   */
  private BindingTable addBinding(Constant c1, Constant c2, BindingConstraints constraints){
    if(c1.equals(c2))
      return this;
    else {
      if (reasoner != null){
// 	System.out.println("testing subsumption " + c1.getName() + " < " +  c2.getName() + " " + reasoner.subClassOf(c1.getName(), c2.getName()));
	if (reasoner.subClassOf(c1.getName(), c2.getName()))
	  return this;
	else
	  return null;
      }
      else
	return null;
    }
  }

  /**
   * @return copy this with new binding added if it is possible (consistent) or this if the binding already exists or null if it is not possible
   */
  private BindingTable addBinding(Variable v, Constant newBinding, BindingConstraints constraints){
    //Check if the variable already has a binding
    //Argument prevBinding = getPositiveBinding(v);
    Vector prevBindings = getPositiveBindings(v);
    BindingTable newBindings;
	
    //Resolve previous binding then try again
    if(null != prevBindings){
      //Argument prevBinding = (Argument)prevBindings.get(0); //TODO: HERE: change this

      //This if is just a speed up: prevents an extra recursive calls (most common case)
      if(prevBindings.size() == 1 && prevBindings.get(0).equals(newBinding))
	return this;
		
      newBindings = removeAllBindings(v);	//forget that I was ever bound

      for(Iterator prevBs = prevBindings.iterator(); prevBs.hasNext(); ){ //TODO: this recursion is expensive (could use equality set instead)
	newBindings = newBindings.addBinding((Argument)prevBs.next(), newBinding, constraints);
	if(null == newBindings)
	  return null;
      }		

      return newBindings.addBinding(v, newBinding, constraints);

      //This is another way to go about doing it
      // 	    HashMap vEqSet = getEqualitySet(v);
      // 	    for(Iterator eqsIt = vEqSet.iterator(); eqsIt.hasNext(); ){
      // 		Variable eq
      // 	    }

      //recurse through all previous bindings for v, binding them to newBinding (if possible)
      // 	    newBindings = newBindings.addBinding(prevBinding, newBinding, constraints);//TODO: could be circular
      // 	    if(null == newBindings)
      // 		return null;

      // 	    return newBindings.addBinding(v, newBinding, constraints); //next time I won't have any previous bindings
    }
    //If no previous binding and doesn't violate constraints, bind to desired argument
    else {
      if(constraints.isViolation(v, newBinding, this))
	return null;
      return addPositiveBinding(v, newBinding);
    }
  }

  /**
   * @return copy of this with v1->v2 and v2->v1 added
   */
  private BindingTable addBinding(Variable v1, Variable v2, BindingConstraints constraints){
    if(v1.equals(v2))
      return this;
    if(constraints.isViolation(v1, v2, this))
      return null;
    return addPositiveBinding(v1, v2).addPositiveBinding(v2,v1);//TODO: could speed up; this copies whole table twice
  }

  /**
   * @return the Vector of Arguments that the given variable is
   * bound to or null if there are no bindings for it
   */
  public Vector getPositiveBindings(Variable v){
    Vector bindings = (Vector)positiveBindings.get(v);
    if(null == bindings)
      return null;
    return (Vector)bindings.clone();
  }
    
  public Iterator getKeysIter(){
    return positiveBindings.keySet().iterator();
  }
    

  //TODO: this could call getEqualitySet (although might be slower)
  /**
   * @return mapping from every argument I know about to a single argument
   * Note: this function assumes the rest of this class does its job,
   *       specifically it assumes that no variable is bound to another
   *       variable which is already bound to a constant; Also assumes that arguments
   *       are not bound to variables (a->v is bad, but v->a is okay)
   **/
  public HashMap compressBindings(NameGen nameGen){
    HashMap retVal = new HashMap();
    Vector sets = new Vector();
    //Vector constantSets = new Vector();
    for(Iterator keys = getKeysIter(); keys.hasNext(); ){
      Argument curKey = (Argument)keys.next();
      Vector bindings = (Vector)positiveBindings.get(curKey);
      if(null != bindings){
	//Check if if key is bound to constant early (most common case)
	if(bindings.size() == 1 && bindings.get(0).getClass() == Constant.class)
	  retVal.put(curKey, bindings.get(0));
	else {
	  for(Iterator bindingIt = bindings.iterator(); bindingIt.hasNext(); ){
	    Argument binding = (Argument)bindingIt.next();
	    //System.out.println("binding: " + binding);
	    //System.out.println("sets: " + sets);
	    boolean added = false;
	    Set lastAddedSet = null;
	    //for(Iterator setIt = sets.iterator(); setIt.hasNext(); ){
	    for(int i = 0; i < sets.size(); i++){
	      //HashSet curSet = (HashSet)setIt.next();
	      HashSet curSet = (HashSet)sets.get(i);
	      //System.out.println("curSet: " + curSet + ";; curKey: " + curKey);
	      if(added && (curSet.contains(binding) || curSet.contains(curKey))){
		//System.out.println("merging");
		lastAddedSet.addAll(curSet);
		sets.remove(curSet);
		i--;
	      }
	      else if(curSet.contains(binding)){
		curSet.add(curKey);
		lastAddedSet = curSet;
		added = true;
	      }
	      else if(curSet.contains(curKey)){
		curSet.add(binding);
		lastAddedSet = curSet;
		added = true;
	      }
	    }
	    if(!added){
	      //System.out.println(sets);
	      //System.out.println(binding + " not found: creating new set for it and " + curKey);
	      HashSet newSet = new HashSet(2);
	      newSet.add(curKey);
	      newSet.add(binding);
	      sets.add(newSet);
	      //System.out.println("after add: " + sets);
	    }
	  }
	}
      }
    }

    for(Iterator setIt = sets.iterator(); setIt.hasNext(); ){
      HashSet curSet = (HashSet)setIt.next();
      Object newVal;
      for(Iterator elemIt = curSet.iterator(); elemIt.hasNext(); ){
	Object index = elemIt.next();
	Object oldVal = retVal.get(index);
	String newName = index.toString();
	// cut off the @number at the end of instantiated variable names
	String[] st= newName.split("@"); 
	if (st.length>0) newName = st[0];
	newVal = new Variable(newName + "@" + nameGen.getNext());
	//System.out.println("compressBindings: " + index + " --> " +newName);
	//System.out.println("type: " + index.getClass().getName());
	retVal.put(index, newVal);
      }
    }
    return retVal;
  }


  /**
   * @param a The argument to get the things equal to.
   * @return The set of arguments equal to argument a.
   */
  public HashSet getEqualitySet(Argument a){
    return getEqualitySet(a, new HashSet());
  }

  public HashSet getEqualitySet(Argument a, HashSet set){
    if(a.getClass() == Variable.class)
      return getEqualitySet((Variable)a, set);
    else if(a.getClass() == Constant.class)
      return getEqualitySet((Constant)a, set);
    return null;
  }

  /**
   * @return the set of all arguments that v is bound to (including transitive bindings)
   */
  private HashSet getEqualitySet(Variable v, HashSet set){
    if(set.contains(v))
      return set;

    set.add(v);
    Vector bindings = getPositiveBindings(v);
    if(null == bindings)
      return set;

    for(Iterator bindingsIt = bindings.iterator(); bindingsIt.hasNext(); ){
      getEqualitySet((Argument)bindingsIt.next(), set);
    }
    return set;
  }

  /**
   * @return the Ontology reasoner of this binding
   */
  public OntoReasoner getReasoner()
  {
    return reasoner;
  }

  /**
   * Internal function for calculating equality sets.
   *
   * TODO: we should probably make a speed up for this, but its gonna be kinda hard?
   * now it has to look at every variable to see if its bound to c
   */
  private HashSet getEqualitySet(Constant c, HashSet set){
    set.add(c);
    for(Iterator keysIt = getKeysIter(); keysIt.hasNext(); ){
      Variable key = (Variable)keysIt.next();
      if(getPositiveBindings(key).contains(c))
	getEqualitySet(key, set);
    }
    return set;
  }

  public String toString(){
    String s = "";
    Object[] keys = positiveBindings.keySet().toArray();
    for(int i = 0; i < keys.length; i++)
      s += keys[i] + "-->" + positiveBindings.get(keys[i]) + "\n";
    return s;
  }
}
