package org.mathweb.reasoning.logic;

import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import org.mathweb.reasoning.partialorderplanner.NameGen;

/**
 * A Literal is predicate and a list of arguments
 *
 * An immutable object.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 22, 2003
 */
public class Literal extends CachedHashCodeObject {
    final private Predicate predicate;
    final private Argument[] arguments;
    final private boolean truth;

    /**
     * Construct a new literal.
     *
     * @param predicate The predicate
     * @param argument The list of arguments of the literal.
     * @param truth True if this literal is true, false if this literal is false.
     */
    public Literal(Predicate predicate, Argument[] arguments, boolean truth){
	this.predicate = predicate;
	this.arguments = arguments;
	this.truth = truth;
    }

    /**
     * @return True if both this and the other literal are exactly the same and  contain only
     * constants.
     */
    public boolean equalsExactly(Literal other) {
        if (! predicate.equals(other.predicate))
            return false;
        if (truth != other.truth)
            return false;
        if (arguments.length != other.arguments.length)
            return false;
        for (int i = 0; i < arguments.length; i++) {
            if (arguments[i].isVariable() || other.arguments[i].isVariable() || 
                ! arguments[i].equals(other.arguments[i]))
                return false;
        }
        return true;
    }
    
    public boolean equals(Literal l) {
        if (truth != l.truth)
            return false;
        
        if (! predicate.equals(l.predicate))
            return false;
        
        if (arguments.length != l.arguments.length)
            return false;

        for (int i = 0; i < arguments.length; i++) {
            if (! arguments[i].equals(l.arguments[i]))
                return false;
        }
        
        return true;    
    }

    public boolean equals(Object o) {
        try {
            return equals((Literal) o);
        } catch (ClassCastException e) {
            return false;
        }
    }

    protected void computeHashCode() {
        int value = predicate.hashCode();
        value = value ^ (new Boolean(truth)).hashCode();
        for (int i = 0; i < arguments.length; i++)
            value = value ^ arguments[i].hashCode();
        myhashcode = value;
    }

    /**
     * @return The number of arguments.
     */
    public int getNumArguments() {
        return arguments.length;
    }

    /**
     * Accessor for arguments.
     *
     * @param i The number of the argument to get, counting from zero.
     * @return The wanted argument.
     * @exception IllegalArgumentException Thrown if the number of the argument is out of bounds.
     */
    public Argument getArgument(int i) {
        try {
            return arguments[i];
        } catch (ArrayIndexOutOfBoundsException e) {
            throw new IllegalArgumentException("getArgument("+i+") when we have " + arguments.length +" args.");
        }
    }

    public String toString(){
	String s = "";
	if(!truth)
	    s += "NOT ";
	s += predicate + "(";
	for(int i = 0; i < arguments.length; i++){
	    s += arguments[i];
	    if(i != arguments.length-1)
		s += ", ";
	}
	return s + ")";
    }

    /**
     * truth Accessor
     */
    public boolean isTrue(){
	return truth;
    }

    /**
     * @return A copy of this literal with the truth value flipped
     */
    public Literal not() {
        return new Literal(predicate, arguments, ! truth);
    }

    /**
     * Predicate Accessor
     */
    public Predicate getPredicate(){
	return predicate;
    }

    /**
     * Accessor for individual arguments
     *
     * @param index: which argument you want
     * @return the argument at that index
     */
     public Argument getArgumentAt(int index){
 	return arguments[index];
     }

    /**
     * The number of arguments
     */
    public int getArgumentCount(){
	return arguments.length;
    }

    /**
     * @return True if there are no unbound variables in this literal.
     */
    public boolean isBound() {
        for (int i = 0; i < arguments.length; i++) {
            if (arguments[i].isVariable())
                return false;
        }
        return true;
    }

    /**
     * @return The set of unbound Variables in this literal
     */
    public Set getVariables() {
        HashSet set = new HashSet();
        for (int i = 0; i < arguments.length; i++)
            if (arguments[i].isVariable())
                set.add(arguments[i]);
        return set;
    }

    /**
     * @return The set of arguments (vars & constants) in this literal.
     */
    public Set getArguments() {
        HashSet set = new HashSet();
        for (int i = 0; i < arguments.length; i++)
            set.add(arguments[i]);
        return set;
    }
    
    /**
     * @return a copy of this literal some variables replaced by the constants they are bound to
     * Note: this function does not guarantee that the returned Literal
     *       has only constant arguments
     */
    public Literal applyBindings(HashMap compressedBindings){
	Argument[] newArguments = new Argument[arguments.length];

	for(int i = 0; i < newArguments.length; i++){
	    if(arguments[i].getClass() == Variable.class){
		Argument newArgument = (Argument)compressedBindings.get(arguments[i]);
		if(null != newArgument)
		    newArguments[i] = newArgument;
		else
		    newArguments[i] = arguments[i];
	    }
	    else
		newArguments[i] = arguments[i];
	}
	Literal applied = new Literal(predicate, newArguments, truth);

        if (applied.equals(this))
            return this; // save memory!
        return applied;
    }


    /**
     * Renames all the variables in this literal.
     *
     * @param nameGen The namegen used to assign names.
     * @param newNames The new names.
     * @return A copy of this literal with the variables renamed.
     */
    public Literal renameVariables(NameGen nameGen, HashMap newNames){
	Argument[] newArgs = new Argument[arguments.length];
	for(int i = 0; i < arguments.length; i++){
	    if(arguments[i].isVariable()){
		Argument newName = (Argument)newNames.get(arguments[i]);
		if(null == newName){
		    newName = ((Variable)arguments[i]).rename(nameGen.getNext());
		    newNames.put(arguments[i], newName);
		}
		newArgs[i] = newName;
	    }
	    else
		newArgs[i] = arguments[i];
	}
	return new Literal(predicate, newArgs, truth);
    }
}
