package org.mathweb.reasoning.logic;

import java.io.*;
import java.util.*;
//TODO: The loading function could use some error checking to inform the user
//      of times when their file is not correct
//TODO: Could hash on the symbol instead of the file's string; should be consistent
import org.mathweb.reasoning.planning.*;

/**
 * Loads in logic objects (constancs, variables, predicates, literas
 * and soon actions, states, etc.) from a file.  The following is an 
 * example of the file format:
 * 
 * Predicate0 on 
 * Constant0 a
 * Constant1 b
 * Constant2 b
 * Constant3 c
 * Constant4 c
 * Constant5 d
 * Constant6 b
 * Constant7 e
 * Variable0 v1
 * Variable1 v2
 * Variable2 v3
 * Literal0
 *      Predicate0
 *      2
 *		Constant0
 *		Constant1
 * Literal1
 *	Predicate0
 *	2
 *		Constant2
 *		Constant3
 * Literal2
 *	Predicate0
 *	2
 *		Constant4
 *		Constant5
 * Literal3
 *	Predicate0
 *	2
 *		Constant6
 *		Constant7
 * Literal4
 *	Predicate0
 *	2
 *		Variable0
 *		Variable1
 * Literal5
 *	Predicate0
 *      false
 *	2
 *		Variable1
 *		Variable2
 *
 *
 * This file produces the following literals
 *  On(v1,v2)
 *  On(v2,v3)
 *  On(a,b)
 *  On(b,c)
 *  On(c,d)
 *  On(b,e)
 *
 * This is the older version of the parser.  It has been replaced by parser.Parser.
 *
 * @author Daniel Terry.
 */
public class LogicLoader {
    
    HashMap things = new HashMap();
    Vector constants = new Vector();
    Vector variables = new Vector();
    Vector predicates = new Vector();
    public Vector literals = new Vector();
    Vector tasks = new Vector();
    public Vector actions = new Vector();
    public Vector states = new Vector();

    public LogicLoader(String filename){
	FileReader reader;
	StreamTokenizer st;
	String word;
	try {
	    reader = new FileReader(filename);
	    st = new StreamTokenizer(reader);
	    st.nextToken();
	    while(st.ttype != st.TT_EOF){
		word = st.sval;
		if(word.startsWith("Predicate")){
		    Predicate p = loadPredicate(st);
		    things.put(word, p);
		    predicates.add(p);
		}
		else if(word.startsWith("Constant")){
		    Constant c = loadConstant(st);
		    things.put(word, c);
		    constants.add(c);
		}
		else if(word.startsWith("Variable")){
		    Variable v = loadVariable(st);
		    things.put(word, v);
		    variables.add(v);
		}
		else if(word.startsWith("Literal")){
		    Literal l = loadLiteral(st);
		    things.put(word, l);
		    literals.add(l);
		}
		else if(word.startsWith("Task")){
		    Task t = loadTask(st);
		    things.put(word, t);
		    tasks.add(t);
		}
		else if(word.startsWith("Action")){
		    Action a = loadAction(st);
		    things.put(word, a);
		    actions.add(a);
		}
		else if(word.startsWith("State")){
		    State s = loadState(st);
		    things.put(word, s);
		    states.add(s);
		}
		else{
		    System.out.println("ERROR Loading file: " + word + " not recognized");
		    return;
		}

		st.nextToken();
	    }
	    reader.close();
	}
	catch (IOException ioe){
	    System.out.println(ioe);
	}
    }

    /**
     * Returns a vector with the elements of literals in the given range
     * with the start index inclusive and the end index exclusive
     */
    public Vector getLiteralRange(int startIn, int endEx){
	return new Vector(literals.subList(startIn, endEx));
    }
    
    public String toString(){
	String s = "Constants:" + constants + "\n";
	
	s += "Variables:" + variables + "\n";
	s += "Predicates:" + predicates + "\n";
	s += "Literals:" + literals + "\n";
	s += "Tasks:" + tasks + "\n";
	s += "States:" + states + "\n";
	s += "Actions:" + actions + "\n";
	return s;
    }

    private Constant loadConstant(StreamTokenizer st){
	try{
	    st.nextToken();
	    return new Constant(st.sval);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }
    private Variable loadVariable(StreamTokenizer st){
	try{
	    st.nextToken();
	    return new Variable(st.sval);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }

    private Predicate loadPredicate(StreamTokenizer st){
	try{
	    st.nextToken();
	    return new Predicate(st.sval);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }

    private Task loadTask(StreamTokenizer st){
	try{
	    st.nextToken();
	    return new Task(st.sval);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }

    private Literal loadLiteral(StreamTokenizer st){
	try{
	    st.nextToken();
	    Predicate p = (Predicate)things.get(st.sval);
	    st.nextToken();
            boolean sign = true;
            if(st.ttype == StreamTokenizer.TT_NUMBER){
                sign = true;
            } else if (((String)st.sval).equals("true")) {
                sign = true;
                st.nextToken();
            } else {
                sign = false;
                st.nextToken();
            }
	    int argCount = (int)st.nval;
	    Argument[] args = new Argument[argCount];
	    for(int i = 0; i < argCount; i++){
		st.nextToken();
		args[i] = (Argument)things.get(st.sval);
	    }
	    return new Literal(p, args, sign);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }

    private Action loadAction(StreamTokenizer st){
	try{
	    st.nextToken();
	    Task t = (Task)things.get(st.sval);
	    st.nextToken();
	    int preCount = (int)st.nval;
	    Literal[] pres = new Literal[preCount];
	    for(int i = 0; i < preCount; i++){
		st.nextToken();
		pres[i] = (Literal)things.get(st.sval);
	    }
	    st.nextToken();
	    int postCount = (int)st.nval;
	    Literal[] posts = new Literal[postCount];
	    for(int i = 0; i < postCount; i++){
		st.nextToken();
		posts[i] = (Literal)things.get(st.sval);
	    }

            // parse the variable constraints
            st.nextToken();
            int constraintCount = (int)st.nval;
            BindingConstraints constraints = new BindingConstraints();
            for (int i = 0; i < constraintCount; i++) {
                st.nextToken();
                Variable first = (Variable) things.get(st.sval);
                st.nextToken();
                Variable second = (Variable) things.get(st.sval);
                constraints = constraints.addConstraint(first, second);
            }
	    return new Action(t, pres, posts, constraints);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }

    private State loadState(StreamTokenizer st){
	try{
	    st.nextToken();
	    int litCount = (int)st.nval;
	    Literal[] lits = new Literal[litCount];
	    for(int i = 0; i < litCount; i++){
		st.nextToken();
		lits[i] = (Literal)things.get(st.sval);
	    }
	    return new State(lits);
	}
	catch(IOException ioe){
	    System.out.println(ioe);
	}
	return null;
    }



}
