package org.mathweb.reasoning.parser;

import java.util.*;
import java.util.regex.*;
import java.io.*;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;

/**
 * Parser for the input states.
 *
 * Variables are one or more lowercase characters followed by zero 
 * or more numbers.
 *
 * Constants are one or more uppercase characters followed by zero or 
 * more numbers.
 *
 * Characters ~ or ! denote not.
 *
 * states are written as:
 * state { literal1, literal2, literal3 }
 *
 * action {
 * [actionName]
 * pre {
 *   on(x, y), clear(x)
 * }
 * post {
 * ! on(x, y), on(y, TABLE)
 * }
 * constraints {
 * x != y, z != x
 * }
 * }
 *
 * @author Harlan Crystal, Daniel Terry
 * @date November 4, 2003
 */
public class Parser {
    private static final String PREDICATE_REGEX = "[a-zA-Z]+[0-9]*";
    private static final String VARIABLE_REGEX = "[a-z]+[0-9]*";
    public  static final String CONSTANT_REGEX = "[A-Z]+[0-9]*";
    private static final String PREDARG_REGEX = "\\s*(?:"+VARIABLE_REGEX+"|"+CONSTANT_REGEX+")\\s*";
    private static final String PREDARGS_REGEX = makeArgumentsRegex(PREDARG_REGEX);
    private static final String NOT_REGEX = "(?:!|~)";
    private static final String LITERAL_REGEX = "\\s*"+NOT_REGEX+"?\\s*" + 
        PREDICATE_REGEX + "\\s*\\(\\s*" + PREDARGS_REGEX +"\\s*\\)\\s*";
    
    private static final String LITERAL_LIST_REGEX = "\\s*\\{\\s*" + makeArgumentsRegex(LITERAL_REGEX) + "\\s*\\}\\s*";
    private static final String STATE_REGEX = "\\s*(?:state|goal|initial)\\s*\\{"+
        makeArgumentsRegex(LITERAL_REGEX)+"}\\s*";

    private static final String TASK_REGEX = "[a-zA-Z]+[0-9]*[-a-zA-Z]*";
    private static final String CONSTRAINT_REGEX = "\\s*"+VARIABLE_REGEX + "\\s*!=\\s*" + VARIABLE_REGEX +"\\s*";
    private static final String CONSTRAINT_LIST = makeArgumentsRegex(CONSTRAINT_REGEX);
    private static final String ACTION_REGEX = "\\s*action\\s*\\{\\s*("+TASK_REGEX+")\\s*pre\\s*("+LITERAL_LIST_REGEX+")\\s*post\\s*("+ LITERAL_LIST_REGEX+")\\s*constraints\\s*(\\{\\s*" + CONSTRAINT_LIST+ "\\s*\\})\\s*\\}\\s*";


    /** set of states */
    private Vector states;

    /** The index of the initial state */
    private int initialIndex;
    
    /** The index of the goal state */
    private int goalIndex;

    /** set of actions */
    private Vector actions;

    public Parser(String s) throws Exception {
        this.initialIndex = -1;
        this.goalIndex = -1;
        this.states = new Vector();
        this.actions = new Vector();

        while (! s.trim().equals("")) {
            s = s.trim();
            if (s.startsWith("action")) {
                int start = s.indexOf("{");
                if (start == -1)
                    throw new Exception("action without opening brace");
                int stop = indexOfMatchingParen(s.substring(start), '{', '}');
                Action a = parseAction(s.substring(0, stop+start+1));
                if (a == null)
                    throw new Exception("Unparsable action");
                actions.add(a);
                s = s.substring(stop+start+1);
            } else if (s.startsWith("state") || s.startsWith("initial") || s.startsWith("goal")) {
                int start = s.indexOf("{");
                if (start == -1)
                    throw new Exception("action without opening brace");
                int stop = indexOfMatchingParen(s.substring(start), '{', '}');
                State st = parseState(s.substring(0, stop+start+1));
                if (st == null)
                    throw new Exception("Unaparsable state");
                states.add(st);

                if (initialIndex == -1 && s.startsWith("initial"))
                    initialIndex = states.size() - 1;
                else if (initialIndex != -1 && s.startsWith("initial"))
                    throw new Exception("Multiple initial states!");
                else if (goalIndex == -1 && s.startsWith("goal"))
                    goalIndex = states.size() - 1;
                else if (goalIndex != -1 && s.startsWith("goal"))
                    throw new Exception("Multiple goal states");

                s = s.substring(stop+start+1);
            } else {
                throw new Exception("Unknown string start: " + s);
            }
        }
    }

    public static Parser loadFromFile(String filename) throws FileNotFoundException, IOException, Exception {
        BufferedReader in = new BufferedReader(new FileReader(filename));
        String s = "";
        String line = in.readLine();
        while (null != line) {
            s += line + "\n";
            line = in.readLine();
        }
        in.close();
        return new Parser(s);
    }

    /**
     * @return The set of states parsed.
     */
    public Vector getStates() {
        return (Vector) states.clone();
    }

    /**
     * @return The goal state, if there was one.  Otherwise null.
     */
    public State getGoalState() {
        if (goalIndex == -1 || goalIndex >= states.size())
            return null;
        return (State) states.get(goalIndex);
    }

    /**
     * @return The initial state, if there was one.  Otherwise null.
     */
    public State getInitialState() {
        if (initialIndex == -1 || initialIndex >= states.size())
            return null;
        return (State) states.get(initialIndex);
    }

    /**
     * @return The parsed states as an array
     */
    public State [] getStatesArray() {
        State [] sa = new State[states.size()];
        int i = 0;
        for (Iterator it = states.iterator(); it.hasNext(); ) {
            State s = (State) it.next();
            sa[i] = s;
            i++;
        }
        return sa;
    }

    /**
     * @return The set of parsed actions.
     */
    public Vector getActions() {
        return (Vector) actions.clone();
    }

    /**
     * @return The parsed states as an array.
     */
    public Action [] getActionsArray() {
        Action [] aa = new Action[actions.size()];
        int i = 0;
        for (Iterator it = states.iterator(); it.hasNext(); ) {
            Action a = (Action) it.next();
            aa[i] = a;
            i++;
        }
        return aa;
    }

    /** 
     * given a regex of a single argument, generates the regex needed
     * for a list of arguments separated by commas.
     */
    private static String makeArgumentsRegex(String argRegex) {
        String ws = "\\s*";
        String argMulti = "(?:"+argRegex+"\\s*,\\s*)*";
        return "(?:\\s*|"+argMulti+argRegex+")";
    }

    /**
     * Same as makeArgumentsRegex except it captures each argument 
     */
    private static String makeCapturedArgumentsRegex(String argRegex) {
        String ws = "\\s*";
        String argMulti = "(?:("+argRegex+")\\s*,\\s*)*";
        String ret =  "(?:\\s*|"+argMulti+"\\s*("+argRegex+"))";
        System.err.println(ret);
        return ret;
    }

    /**
     * @return A new variable object or null if string does not 
     * describe a variable.
     */
    public static Variable parseVariable(String var) {
        StringTokenizer tok = new StringTokenizer(var);
        if (! tok.hasMoreTokens())
            return null;

        String item = tok.nextToken();
        if (! item.matches(VARIABLE_REGEX))
            return null;
        
        if (tok.hasMoreTokens())
            return null;
        return new Variable(item);
    }

    /**
     * @return A new constant object or null if string does not describe
     * a constant.
     */
    public static Constant parseConstant(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        if (! tok.hasMoreTokens())
            return null;

        String item = tok.nextToken();
        if (! item.matches(CONSTANT_REGEX))
            return null;
        if (tok.hasMoreTokens())
            return null;
        return new Constant(item);
    }

    /**
     * @return A new predicate object or null if string does not describe
     * a predicate
     */
    public static Predicate parsePredicate(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        if (! tok.hasMoreTokens())
            return null;
        
        String item = tok.nextToken();
        if (! item.matches(PREDICATE_REGEX))
            return null;
        if (tok.hasMoreTokens())
            return null;
        return new Predicate(item);
    }

    public static Task parseTask(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        if (! tok.hasMoreTokens())
            return null;

        String item = tok.nextToken();
        if (! item.matches(TASK_REGEX))
            return null;
        if (tok.hasMoreTokens())
            return null;
        return new Task(item);
    }

    /**
     * @return A literal object or null if the string does not describe
     * a literal.
     */
    public static Literal parseLiteral(String s) {
        if (! s.matches(LITERAL_REGEX))
            return null;

        String [] pieces = s.split("\\(|\\)", -1);
        if (pieces.length < 2 || pieces.length > 3) {
            return null;
        }

        // parse out the predicate and the truth value
        String predSection = pieces[0];
        String [] predPieces = predSection.split(NOT_REGEX);
        if (predPieces.length < 1 || predPieces.length > 2 || (predPieces.length == 2 && ! predPieces[0].matches("\\s*")))
            return null;
        boolean truth = true;
        Predicate pred = null;
        if (predPieces.length == 2) {
            truth = false;
            pred = parsePredicate(predPieces[1]);
        } else {
            pred = parsePredicate(predPieces[0]);
        }

        if (pred == null)
            return null;

        // parse out the arguments
        String [] args = pieces[1].split(",", -1);
        if (args.length == 1 && args[0].matches("\\s*"))
            return new Literal(pred, new Argument[0], truth);
        
        Argument [] arguments = new Argument[args.length];
        for (int i = 0; i < args.length; i++) {
            Variable var = parseVariable(args[i]);
            Constant cons = null;
            if (var == null)
                cons = parseConstant(args[i]);
            
            if (var != null)
                arguments[i] = var;
            else if (cons != null)
                arguments[i] = cons;
            else
                return null;
        }
        return new Literal(pred, arguments, truth);
    }

    public static State parseState(String s) {
        if (! s.matches(STATE_REGEX))
            return null;
        
        int firstParen = s.indexOf('{');
        if (firstParen == -1)
            return null;
        Literal [] lits = parseLiteralList(s.substring(firstParen));
        if (lits == null)
            return null;

        return new State(lits);
    }

    /**
     * returns the string from the beginning of input until the character
     * before the final matched paren
     *
     * first char should be open paren
     *
     * return null if such a string can not be found.
     *
     * returned string does not contain the first and last parens
     */
    public static String sectionUntilMatchingParen(String input, 
                                                   char openParen,
                                                   char closeParen) {
        int i = indexOfMatchingParen(input, openParen, closeParen);
        if (i == -1)
            return null;
        return input.substring(1, i);
    }
    public static int indexOfMatchingParen(String input, char openParen,
                                           char closeParen) {
        if (input.charAt(0) != openParen)
            throw new IllegalArgumentException("String doesn't begin with \""+
                                               openParen +"\"");
        int numOpenParen = 1;
        for (int i = 1; i < input.length(); i++) {
            if (input.charAt(i) == openParen)
                numOpenParen++;
            if (input.charAt(i) == closeParen)
                numOpenParen--;
            if (numOpenParen == 0)
                return i;
        }
        return -1;
    }

    public static Literal [] parseLiteralList(String s) {
        if (! s.matches(LITERAL_LIST_REGEX)) 
            return null;

        String [] pieces = s.split("\\{|\\}", -1);
        if (pieces.length < 2 || pieces.length > 3)
            return null;

        // parse out each of the literals
        String args [] = pieces[1].split(",");

        if (args.length == 1 && args[0].matches("\\s*"))
            return new Literal[0];
        Vector literals = new Vector();
        for (int i = 0; i < args.length; i++) {
            String str = args[i];
            Literal lit = parseLiteral(str);
            while (lit == null && i < args.length-1) {
                str += "," + args[i+1];
                lit = parseLiteral(str);
                i++;
            }
            if (lit == null)
                return null;
            literals.add(lit);
        }
        Literal []copy = new Literal[literals.size()];
        for (int i = 0; i < literals.size(); i++)
            copy[i] = (Literal) literals.get(i);

        return copy;
    }

    public static Action parseAction(String s) {
        Matcher m = Pattern.compile(ACTION_REGEX).matcher(s);
        if (! m.matches()) {
            System.out.println("didnt match: " + s);
            return null;
        }
        if (m.groupCount() != 4) {
            System.out.println("Wrong groupcount");
            return null;
        }
        Task task = new Task(m.group(1));
        Literal [] pres = parseLiteralList(m.group(2));
        Literal [] posts = parseLiteralList(m.group(3));
        BindingConstraints constraints = parseBindingConstraints(m.group(4));
        if (task == null || pres == null || posts == null ||
            constraints == null) {
            System.out.println("something was null");
            return null;
        }
        return new Action(task, pres, posts, constraints);
    }
    
    public static BindingConstraints parseBindingConstraints(String s) {
        String regex = "\\{\\s*"+makeArgumentsRegex(CONSTRAINT_REGEX) +"\\s*\\}";
        Matcher m = Pattern.compile(regex).matcher(s);
        if (! m.matches()) {
            return null;
        }
        
        BindingConstraints c = new BindingConstraints();
        
        // cut off first {
        int j = s.indexOf("{");
        int k = s.indexOf("}");
        if (j == -1 || k == -1 || k <= j)
            return null;
        s = s.substring(j+1, k).trim();
        if (s.length() == 0)
            return c;
        String [] pieces = s.split(",");
        for (int i = 0; i < pieces.length; i++) {
            String [] vars = pieces[i].split("!=");
            if (vars.length != 2)
                return null;
            Variable v1 = parseVariable(vars[0]);
            if (v1 == null)
                return null;
            Variable v2 = parseVariable(vars[1]);
            if (v2 == null)
                return null;
            c = c.addConstraint(v1, v2);
        }
        return c;
    }

    public static void main(String [] args) {
        String test = "  on ( )";
        String test2 = "on(B1, A)";
        String test3 = " on  (z,y,x  )  ";
        String test4 = "on(FLOOR)";
        System.out.println(Parser.parseLiteral(test));
        System.out.println(Parser.parseLiteral(test2));
        System.out.println(Parser.parseLiteral(test3));
        System.out.println(Parser.parseLiteral(test4));

        String [] testParen = new String[5];
        testParen[0] = "{hi}";
        testParen[1] = "{hi";
        testParen[2] = "{hi {there}}";
        testParen[3] = "{hi {sup} {whats { good } } }";
        testParen[4] = "{{{{{{hi}}}} hi {} {} {}} {{} {} {{}}}}";
        for (int i = 0; i < testParen.length; i++)
            System.out.println(Parser.sectionUntilMatchingParen(testParen[i], '{', '}'));


        String [] constraintTest = new String[5];
        constraintTest[0] = "x != y";
        constraintTest[1] = "";
        constraintTest[2] = "A != B";
        constraintTest[3] = "a ! = B   ";
        constraintTest[4] = "a != c";
        for (int i = 0; i < constraintTest.length; i++)
            System.out.println(constraintTest[i].matches(Parser.CONSTRAINT_REGEX));

        String state1 = "state{}";
        String state2 = "state{on()}";
        String state3 = "state{   on ()  }";
        String state4 = "state{ on(), on() }";
        String state5 = "state { on(TREE), onTop(x, FLOOR) }";

        String [] actionTest = new String[3];
        actionTest[0] = "action {move pre {on(x, y), on(x, TABLE)} post {on(y, x)} constraints{}}";
        actionTest[1] = "action {move pre {on(x, y), on(x, TABLE)} post {on(y, x)} constraints{x != y}}";
        actionTest[2] = "action {move pre {on(x, y), on(x, TABLE)} post {on(y, x)} constraints{x != y, x != z}}";
        //for (int i = 0; i < actionTest.length; i++)
        //System.out.println(p.parseAction(actionTest[i]));
        try {
            Parser p = new Parser(actionTest[0]+state1+actionTest[1]+state2+actionTest[2]+state3);
            System.out.println("States:");
            for (Iterator it=p.getStates().iterator(); it.hasNext(); ) {
                System.out.println(it.next());
            }
            System.out.println("Actions:");
            for (Iterator it=p.getActions().iterator(); it.hasNext(); ) {
                System.out.println(it.next());
            }
        } catch (Exception e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }
}
