package org.mathweb.reasoning.partialorderplanner;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
import java.util.*;
/**
 * This class represents a casual link, aka a protection interaval, of the form
 * A--p-->B, read "A achieves p for B" where A and B are actions and p is a condition
 * (i.e. a literal).  A and B are refered to as the achiever and the needer respectively
 *
 * This class is immutable.
 * 
 * @author Daniel Terry, Harlan Crystal
 * @date November 2, 2003
 */
public class CausalLink extends CachedHashCodeObject {
    /** The action that is achieving the condition */
    private final Action achiever;
    /** The action that needs the condition to be achieved */
    private final Action needer;
    /** The condition being achieved. */
    private final Literal condition;

    /**
     * Constructs a new causal link.
     *
     * @param achiever The action achieving of the condition.
     * @param needer The action needing the condition.
     * @param condition The needed condition.
     */
    public CausalLink(Action achiever, Action needer, Literal condition){
	this.achiever = achiever;
	this.needer = needer;
	this.condition = condition;
    }

    public boolean equals(CausalLink c) {
        return achiever.equals(c.achiever) && needer.equals(c.needer) && condition.equals(c.condition);
    }

    public boolean equals(Object o) {
        try {
            return equals((CausalLink) o);
        } catch (ClassCastException e) {
            return false;
        }
    }

    protected void computeHashCode() {
        myhashcode = achiever.hashCode();
        myhashcode = myhashcode ^ needer.hashCode();
        myhashcode = myhashcode ^ condition.hashCode();
    }

    /**
     * @return The action achieving the condition.
     */
    public Action getAchiever(){
	return achiever;
    }

    /**
     * @return The action needing the condition.
     */
    public Action getNeeder(){
	return needer;
    }

    /**
     * @return The needed condition.
     */
    public Literal getCondition(){
	return condition;
    }

    /**
     * Apply the given compressed bindings to this causal link.
     *
     * @param cbt The compressed bindings to apply.
     */
    public CausalLink applyBindings(HashMap cbt){
 	CausalLink applied =  new CausalLink(achiever.applyBindings(cbt), needer.applyBindings(cbt), condition.applyBindings(cbt));
        if (applied.equals(this))
            return this; // save memory!
        return applied;
    }
    
    public String toString(){
	return "[" + achiever + "]--[" + condition + "]-->[" + needer + "]";
    }
}
