package org.mathweb.reasoning.partialorderplanner;
import java.util.*;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;

/**
 * This class represents an OpenCondtion, as used by PartialOrderPlan
 * which requires each open condition to know what action it belongs to,
 * so an OpenCondition is an instantiated Literal and the instantiated
 * Action it belongs to.
 *
 * This class is immutable.
 * 
 * @author Daniel Terry, Harlan Crystal
 * @date November 3, 2003
 */
public class OpenCondition extends CachedHashCodeObject {
    /** The condition that is still to be fulfilled. */
    private final Literal condition;
    /** The action that requires this condition to be fulfilled */
    private final Action owner;

    /**
     * Construct a new open condition.
     *
     * @param condition The condition that requires fulfillment.
     * @param owner The action that requires the condition to be fulfilled.
     */
    public OpenCondition(Literal condition, Action owner){
	this.condition = condition;
	this.owner = owner;
    }

    protected void computeHashCode() {
        myhashcode = condition.hashCode();
        myhashcode = myhashcode ^ owner.hashCode();
    }

    public boolean equals(OpenCondition c) {
        return condition.equals(c.condition) && owner.equals(c.owner);
    }

    public boolean equals(Object other) {
        try {
            return equals((OpenCondition) other);
        } catch (ClassCastException e) {
            return false;
        }
    }

    /**
     * @return The condition that must be fulfilled.
     */
    public Literal getCondition(){
	return condition;
    }

    /**
     * @return The action which requires this condition to be fulfilled.
     */
    public Action getOwner(){
	return owner;
    }

    /**
     * Applies the given set of compressed bindings to this object.
     *
     * @param compressedBindingTable The compressed bindings to apply.
     */
    public OpenCondition applyBindings(HashMap compressedBindingTable) {
        OpenCondition applied = new OpenCondition(condition.applyBindings(compressedBindingTable),
                                                  owner.applyBindings(compressedBindingTable));
        if (applied.equals(this))
            return this; // save memory!
        return applied;
    }

    public String toString(){
	return condition + "; \tOwner: " + owner;
    }
}
