package org.mathweb.reasoning.partialorderplanner;
import java.util.*;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
/**
 * This class represents an ordering constraint of the form "A<-B" or "A before B"
 * where this class calls A the "early" action and B the "late" action.  
 *
 * This class is immutable.
 * 
 * @author Daniel Terry, Harlan Crystal
 * @date November 2, 2003
 */
public class OrderConstraint extends CachedHashCodeObject {
    /** The action which occurs first in time */
    private final Action early;
    /** The action which occurs later in time */
    private final Action late;
    
    /**
     * Construct a new order constraint.
     *
     * @param early The action which occurs first in time.
     * @param late The action which occurs later in time.
     */
    public OrderConstraint(Action early, Action late) {
	this.early = early;
	this.late = late;
    }

    protected void computeHashCode() {
        myhashcode = early.hashCode();
        myhashcode = myhashcode ^ late.hashCode();
    }

    public boolean equals(OrderConstraint c) {
        return early.equals(c.early) && late.equals(c.late);
    }

    public boolean equals(Object other) {
        try {
            return equals((OrderConstraint) other);
        } catch (ClassCastException e) {
            return false;
        }
    }

    /**
     * @return The action which occurs earlier in time.
     */
    public Action getEarly(){
	return early;
    }

    /**
     * @return The action which occurs later in time.
     */
    public Action getLate() {
	return late;
    }

    public String toString(){
	return early + " Before " + late;
    }

    /**
     * Applies given bindings to this object.
     *
     * @param cbt The compressed binding table to apply.
     */
    public OrderConstraint applyBindings(HashMap cbt){
	OrderConstraint applied = new OrderConstraint(early.applyBindings(cbt), late.applyBindings(cbt));
        if (applied.equals(this))
            return this; // save memory!
        return applied;
    }

}
