package org.mathweb.reasoning.partialorderplanner;

import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import salvo.jesus.graph.*;

//import org.mathweb.msdl.*;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.parser.Parser;

/**
 * This class represents either a complete or incomplete partial order plan.
 * That is, it contains the following:
 *      -Set of instantiated actions used (from the set of all possible actions)
 *	-Cycle free set of ordering constraints ("A<-B": "A before B")
 *	-Set of causal links, or protection intervals ("A--p-->B: A achieves p for B")
 *	-Set of open preconditions (ie preconditions that are not achieved by some action in the plan)
 * 
 * This class is immutable.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date November 2, 2003
 */
public class PartialOrderPlan extends CachedHashCodeObject {
  public  final HashSet actions;
  public final HashSet orderConstraints;
  public final HashSet causalLinks;
  public final HashSet openConditions; 
  public BindingConstraints bindingConstraints; // Not final b/c when creating new POPs you need to be able to set this (see apply bindings)
  public final Action startA;
  public final Action finishA;

  private static Logger logger = PartialOrderPlanner.logger;

  /**
   * Construct a new partial order plan.
   *
   * @param startA The start action.
   * @param finishA The finish action.
   */
  public PartialOrderPlan(Action startA, Action finishA) {
    actions = new HashSet();
    orderConstraints = new HashSet();
    causalLinks = new HashSet();
    openConditions = new HashSet();
    bindingConstraints = new BindingConstraints();
    this.startA = startA;
    this.finishA = finishA;
    actions.add(startA);
    actions.add(finishA);
  }

  /**
   * Copy constructor.
   *
   * @param copyMe The plan to copy.
   */
  public PartialOrderPlan(PartialOrderPlan copyMe){
    this.actions = (HashSet) copyMe.actions.clone();
    this.orderConstraints = (HashSet) copyMe.orderConstraints.clone();
    this.causalLinks = (HashSet) copyMe.causalLinks.clone();
    this.openConditions = (HashSet) copyMe.openConditions.clone();
    this.bindingConstraints = copyMe.bindingConstraints; /* immutable, so dont need to copy */
    this.startA = copyMe.startA;
    this.finishA = copyMe.finishA;
  }

  /**
   * Used interally for adding binding constraints
   */
  private PartialOrderPlan(PartialOrderPlan copyMe, BindingConstraints newBindings) {
    this.actions = (HashSet) copyMe.actions.clone();
    this.orderConstraints = (HashSet) copyMe.orderConstraints.clone();
    this.causalLinks = (HashSet) copyMe.causalLinks.clone();
    this.openConditions = (HashSet) copyMe.openConditions.clone();
    this.bindingConstraints = newBindings;
    this.startA = copyMe.startA;
    this.finishA = copyMe.finishA;
    actions.add(startA);
    actions.add(finishA);
  }
  
  public Set getOrderingConstraints() {
    return (Set) orderConstraints.clone();
  }

  public Set getCausalLinks() {
    return (Set) causalLinks.clone();
  }

  public Set getOpenPreconditions() {
    return (Set) openConditions.clone();
  }

  protected void computeHashCode() {
    myhashcode = actions.hashCode();
    myhashcode = myhashcode ^ orderConstraints.hashCode();
    myhashcode = myhashcode ^ causalLinks.hashCode();
    myhashcode = myhashcode ^ openConditions.hashCode();
    myhashcode = myhashcode ^ bindingConstraints.hashCode();
  }

  public boolean equals(PartialOrderPlan p) {
    return actions.equals(p.actions) && orderConstraints.equals(p.orderConstraints) && 
      causalLinks.equals(p.causalLinks) && openConditions.equals(p.openConditions) &&
      bindingConstraints.equals(p.bindingConstraints);
  }

  public boolean equals(Object o) {
    try {
      return equals((PartialOrderPlan) o);
    } catch (ClassCastException e) {
      return false;
    }
  }

  /**
   * @return True if this plan has no open preconditions, in other words, is complete.
   */
  public boolean hasNoOpenConditions(){
    return openConditions.isEmpty();
  }

  /**
   * Add an action to the plan.
   *
   * @param a The action to add.
   * @return The plan with action added.
   */
  public PartialOrderPlan addAction(Action a){
    PartialOrderPlan newPlan = new PartialOrderPlan(this, bindingConstraints.addAll(a.getBindingConstraints()));
    newPlan.actions.add(a);
    return newPlan;
  }

  /**
   * Add an ordering constraint to the plan.
   *
   * @param early The action before late.
   * @param late The action after early.
   * @return The plan with the given order constraint added.
   */
  public PartialOrderPlan addOrderConstraint(Action early, Action late) {
    return addOrderConstraint(new OrderConstraint(early, late));
  }
    
  /**
   * Add an ordering constraint to the plan.
   *
   * @param o The order constraint to add.
   * @return The partial order plan with given order constraint added.
   */
  public PartialOrderPlan addOrderConstraint(OrderConstraint o){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    logger.info(" adding new order constr " + o);
    newPlan.orderConstraints.add(o);
    return newPlan;
  }

  /**
   * Add a causal link to the plan.
   *
   * @param c The causal link to add.
   * @return The plan with given causal link added.
   */
  public PartialOrderPlan addCausalLink(CausalLink c){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    logger.info(" adding new causal " + c);
    newPlan.causalLinks.add(c);
    return newPlan;
  }

  /**
   * Add an open condition to the plan.
   *
   * @param o The open condition to add to the plan.
   * @return The plan with given condition added.
   */
  public PartialOrderPlan addOpenCondition(OpenCondition o){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    newPlan.openConditions.add(o);
    return newPlan;
  }

  /**
   * Add multiple conditions to the plan, all owned by the same action.
   *
   * @param ls A vector of Literal objects representing conditions.
   * @param owner The action that requires these conditions to be fulfilled.
   * @return The partial order plan with all these open conditions added.
   */
  public PartialOrderPlan addAllOpenConditions(Vector ls, Action owner){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    for(Iterator it = ls.iterator(); it.hasNext(); ) {
      OpenCondition cond = new OpenCondition((Literal)it.next(), owner);
      logger.info("addding new open cond " + cond);
      newPlan.openConditions.add(cond);
    }
//     for(Iterator it = ls.iterator(); it.hasNext(); )
//       newPlan.openConditions.add(new OpenCondition((Literal)it.next(), owner));
//     return newPlan;
    return newPlan;
  }

  /**
   * Add a binding constraint to the plan.
   *
   * @param x The first argument.
   * @param y The second argument.
   * @return A plan where there is a constraint added saying x != y.
   */
  public PartialOrderPlan addBindingConstraint(Argument x, Argument y) {
    return new PartialOrderPlan(this, bindingConstraints.addConstraint(x, y));
  }

  /**
   * Add a whole set of binding constraints to the plan.
   *
   * @param bc The binding constraints to add.
   * @return The plan with all the given binding constraints added.
   */
  public PartialOrderPlan addAllConstraints(BindingConstraints bc) {
    return new PartialOrderPlan(this, bindingConstraints.addAll(bc));
  }

  /**
   * @return The set of binding constraints associated with this plan.
   */
  public BindingConstraints getBindingConstraints(){
    return bindingConstraints;
  }

  /**
   * See if this plan already contains the given action
   *
   * @param a The action to check.
   * @return True if this action contains action a.
   */
  public boolean hasAction(Action a){
    for(Iterator it = actions.iterator(); it.hasNext(); ){
      Action b = (Action)it.next();
      if(a.equals(b)){
	return true;
      }
    }
    return false;
  }

  /**
   * Checks if given ordering constraint is already in the plan.
   *
   * @param oc The ordering constraint to check.
   * @return True if this plan already contains given ordering constraint.
   */
  public boolean hasOrderConstraint(OrderConstraint oc){
    return orderConstraints.contains(oc);
  }

  public boolean hasCausalLink(CausalLink cl){
    return causalLinks.contains(cl);
  }

  public boolean hasBindingConstraint(BindingConstraint bc){
    return bindingConstraints.getSet().contains(bc);
  }

  public boolean hasOpenCondition(OpenCondition oc){
    return openConditions.contains(oc);
  }

  /**
   * @return The set of actions in this plan.
   */
  public HashSet getActions(){
    return (HashSet)actions.clone();
  }

  /**
   * @return The last element in openCondition.
   * @exception IllegalArgumentException Thrown if there are no open preconditions left.
   */
  public OpenCondition getAnOpenCondition(){
    if(openConditions.isEmpty())
      throw new IllegalArgumentException("There are no more open preconditions, this plan is complete!");
    return (OpenCondition)openConditions.iterator().next();
  }

  /**
   * Removes an open condition from this plan.
   *
   * @param oc The open condition to remove.
   * @return The plan with given opencondition removed.
   */
  public PartialOrderPlan removeOpenCondition(OpenCondition oc){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    newPlan.openConditions.remove(oc);
    return newPlan;
  }

  /**
   * Handles everything involved with adding an action to the plan
   * (e.g. adding ordering constraints, causal links, unifying action with plan, etc.)
   *
   * @param achiever the action to apply, it has not had comBinds applied to it yet
   * @param condition the open condition to achieve, hasn't had compBinds applied to it yet
   * @param needer the action that wants condition to be achieved, has not had compBinds applied to it yet
   * @param compBinds the compressed bindings that unify achiever with the openCondition (and with this entire plan)
   * @return the set of PartialOrderPlans that can result or the empty set if no plans work
   */
  public Set achievePrecondition(Action achiever, Literal condition, Action needer, HashMap compBinds){
    PartialOrderPlan newPlan = new PartialOrderPlan(this);
    newPlan = newPlan.applyBindings(compBinds);
    Set newPlansFin = new HashSet();
    boolean oldAction = false;

    if(actions.contains(achiever))
      oldAction = true;
	    
    achiever = achiever.applyBindings(compBinds);
    condition = condition.applyBindings(compBinds);
    needer = needer.applyBindings(compBinds);

    // avoid loops, i.e. achiever and needer are the same action!
    if (achiever.equalsByOriginalName(needer))
      return new HashSet();
    
    newPlan = newPlan.addAction(achiever);
    
    if(newPlan.causesConflict(new OrderConstraint(achiever, needer))){
      logger.info(" achiever " + achiever + " causes conflict!");
      return new HashSet();
    }
    
    newPlan = newPlan.addOrderConstraint(new OrderConstraint(achiever, needer));
    newPlan = newPlan.addOrderConstraint(new OrderConstraint(achiever, finishA));
    
    //don't add the openConditions if the achiever is already in the local plan
    if(!oldAction)
      newPlan = newPlan.addAllOpenConditions(achiever.copyPreconditions(), achiever);
    
    if(!achiever.equals(startA)) //Dont add Start -> Start
      newPlan = newPlan.addOrderConstraint(new OrderConstraint(startA, achiever));
    
    CausalLink cl = new CausalLink(achiever, needer, condition);
    newPlan = newPlan.addCausalLink(cl);

    Set newPlansNotFin = newPlan.addOrderingsFor(achiever);

    for(Iterator npi = newPlansNotFin.iterator(); npi.hasNext(); ){
      PartialOrderPlan pl = (PartialOrderPlan)npi.next();
      logger.info(" after orderings " + achiever + " " + pl);
      newPlansFin.addAll(pl.addOrderingsFor(cl));
    }
    
    return newPlansFin;
  }

  /**
   * Add ordering constraints for the given Action so that it does
   * not conflict with any existing casual links
   * i.e. Find all causal links X--q-->Y where A causes !q
   * createate successors in which Y<-A and A<-X if they are consistent
   *
   * @return a vector of PartialOrderPlans that have the possible constraints added
   */
  public Set addOrderingsFor(Action a){
    Set set = new HashSet();
    set.add(this);

    for (Iterator it = causalLinks.iterator(); it.hasNext(); ) {
      CausalLink c = (CausalLink) it.next();
      if(!a.equals(c.getNeeder()) && !a.equals(c.getAchiever())){
	Set newSet = new HashSet();
	for (Iterator planit = set.iterator(); planit.hasNext(); ) {
	  PartialOrderPlan plan = (PartialOrderPlan) planit.next();
	  newSet.addAll(resolveThreatsToPlans(plan, c, a));
	}
	set = newSet;
      }
    }
    return set;
  }

  /**
   * Look at all actions and if any conflict with the given causal link
   * (i.e. if any negate the CL's condition) add ordering constraints
   * that order those actions before or after the given CL
   *
   * @return a vector of PartialOrderPlans that have the possible constraints added
   */
  public Set addOrderingsFor(CausalLink cl) {
    Set set = new HashSet();
    set.add(this);
        
    for (Iterator it = actions.iterator(); it.hasNext(); ) {
      Action a = (Action) it.next();
      if(!a.equals(cl.getNeeder()) && !a.equals(cl.getAchiever())){
	HashSet newSet = new HashSet();
	for (Iterator planit = set.iterator(); planit.hasNext(); ) {
	  PartialOrderPlan plan = (PartialOrderPlan) planit.next();
	  newSet.addAll(resolveThreatsToPlans(plan, cl, a));
	}
	set = newSet;
      }
    }
    return set;
  }

  /**
   * Get the set of ordering constraints associated with an action.
   *
   * @param early The action to get the ordering constraints for.
   * @return The set of ordering constraints associated with given action.
   */
  private Set getOCsFor(Action early){
    HashSet ocs = new HashSet();
    for(Iterator ocit = orderConstraints.iterator(); ocit.hasNext(); ){
      OrderConstraint oc = (OrderConstraint)ocit.next();
      if(oc.getEarly().equals(early))
	ocs.add(oc);
    }
    return ocs;
  }

  /**
   * @return True if given ordering constraint causes a conflict with constraints already in the plan.
   */
  public boolean causesConflict(OrderConstraint oc){
    HashSet seen = new HashSet();
    seen.add(oc.getEarly());
    return inConflict(oc.getLate(), seen);
  }

  /**
   * @return returns true if there is an ordering cycle that contains the given action
   */
  private boolean inConflict(Action a, HashSet seen){
    if(seen.contains(a))
      return true;
	
    seen.add(a);
    for(Iterator aOcIt = getOCsFor(a).iterator(); aOcIt.hasNext(); ){
      if(inConflict(((OrderConstraint)aOcIt.next()).getLate(), (HashSet)seen.clone()))
	return true;
    }
    return false;
  }

  /**
   * @return The set of plans that result from resolving the threat a to plan c.  If there are no 
   * threats, it returns the singleton set back with the same old plan.  If the threat are unresolvable,
   * it returns an empty set.
   */
  private HashSet resolveThreatsToPlans(PartialOrderPlan oldplan, CausalLink c, Action a) {
    HashSet plans = new HashSet();
    BindingConstraints bc = resolveThreats(c, a, oldplan.bindingConstraints);
    logger.info(a + " causes threads " + bc);
    if (bc == null) { // there was no conflict
      plans.add(oldplan);
      return plans;
    }

    HashMap compressed = null;

    if (! bc.isEmpty()) {
      PartialOrderPlan patchedUpConstraints = oldplan.addAllConstraints(bc);
      plans.add(patchedUpConstraints);
      compressed = bc.toBindingTable(oldplan.getBindingConstraints()).compressBindings(NameGen.getNameGen());
    } 

    OrderConstraint before = new OrderConstraint(a, c.getAchiever());
    if(!oldplan.causesConflict(before)) {
      PartialOrderPlan newPlan = oldplan.addOrderConstraint(before);
      if(compressed != null)
	newPlan = newPlan.applyBindings(compressed);
      plans.add(newPlan);
    }

    OrderConstraint after = new OrderConstraint(c.getNeeder(), a);
    if(!oldplan.causesConflict(after)){
      PartialOrderPlan newPlan = oldplan.addOrderConstraint(after);
      if(compressed != null)
	newPlan = newPlan.applyBindings(compressed);
      plans.add(newPlan);
    }

    return plans;
  }

  /**
   * A helper function used for finding all the threats that a given action causes for a given 
   * causal link.  It then returns the set of binding constraints that would fix the threat.
   *
   * @return Null, if there are no threats caused by this pair.  Otherwise this returns the set of
   * new constraints that would make this pair fine.  If an constraints set that is empty is returned,
   * there is no set of contraints which make it work so reordering is neccesary.
   */
  private  BindingConstraints resolveThreats(CausalLink c, Action a, BindingConstraints oldbc) {
    // TODO: how does the caller of this combine oldbc with the newly returned bc??
    return a.resolveThreats(c.getCondition(), oldbc, null);
  }

  /**
   * Applies bindings to the plan.
   *
   * TODO: Every copy of the action gets the bindings applied to it (there's a copy in 
   * actions/orderconstraint/causalLinks/openConditions
   * (thats 4 redundant applications for every action in the plan )
   *
   * @param compressedBindings The set of compressed bindings to apply to this plan.
   * @return The plan with the bindings applied.
   */
  public PartialOrderPlan applyBindings(HashMap compressedBindings){
    PartialOrderPlan newPlan = new PartialOrderPlan(startA, finishA);

    for(Iterator it = actions.iterator(); it.hasNext(); ) {
      Action a = (Action) it.next();
      Action a2 = a.applyBindings(compressedBindings);
      newPlan.actions.add(a2);
    }
	
    for(Iterator it = orderConstraints.iterator(); it.hasNext(); ){
      OrderConstraint ocold = (OrderConstraint) it.next();
      OrderConstraint oc = ocold.applyBindings(compressedBindings);
      newPlan.orderConstraints.add(oc);
    }

    for(Iterator it = causalLinks.iterator(); it.hasNext(); ) {
      CausalLink clold = (CausalLink) it.next();
      CausalLink cl = clold.applyBindings(compressedBindings);
      newPlan.causalLinks.add(cl);
    }

    for(Iterator it = openConditions.iterator(); it.hasNext(); ){
      OpenCondition ooc = (OpenCondition)it.next();
      // TODO: ooc.getOwner() has now effectively had its bindings applied twice (once for the copy 
      // in the Actions list and once here --> slow + two identical copies Actions in memory
      OpenCondition noc = ooc.applyBindings(compressedBindings);
      newPlan.openConditions.add(noc);
    }
        
    newPlan.bindingConstraints = bindingConstraints.applyBindings(compressedBindings);
    if (newPlan.equals(this))
      return this; // save memory
    return newPlan;
  }

  public String toString(){
    String retVal = "Partial Order Plan::\n";
    retVal += "\tActions:{\n";
    for(Iterator it = actions.iterator(); it.hasNext(); )
      //retVal += "\t\t" + ((Action)it.next()).toMediumString() + "\n";	
      retVal += "\t\t" + ((Action)it.next()).toLongString() + "\n";	
    //retVal += "\t\t" + ((Action)it.next()) + "\n";	
	
    retVal += "\t}\n\tOrderings:{\n";
    for(Iterator it = orderConstraints.iterator(); it.hasNext(); )
      retVal += "\t\t" + it.next() + "\n";	

    retVal += "\t}\n\tCausalLinks:{\n";
    for(Iterator it = causalLinks.iterator(); it.hasNext(); )
      retVal += "\t\t" + it.next() + "\n";	

    retVal += "\t}\n\tOpen Conditions:{\n";
    for(Iterator it = openConditions.iterator(); it.hasNext(); )
      retVal += "\t\t" + it.next() + "\n";	

    retVal += "\t}\n\tBinding Constraints:{\n";
    retVal += bindingConstraints;
    retVal += "\t}";

    return retVal;
  }

  /**
   * Assuming actions a and b are in this plan, tells if action a is 
   * constrained to occur before action b.
   *
   * @return True if a is constrained to occur before b.  False if not.
   */
  private boolean isBefore(Action a, Action b) {
    return orderConstraints.contains(new OrderConstraint(a, b));
  }

  /**
   * Assuming actions a and b are in this plan, tells if action a is 
   * constrained to occur after action b.
   *
   * @return True if a is constrained to occur after b.  False if not.
   */
  private boolean isAfter(Action a, Action b) {
    return orderConstraints.contains(new OrderConstraint(b, a));
  }

  /**
   * @return A linearized version of the POP in a ListArray. 
   * The start and the finish actions are removed.
   */
  public ArrayList linearize()
  {
    HashSet realActions = actions;
    //while (realActions.contains(startA)) realActions.remove(startA);
    //while (realActions.contains(finishA)) realActions.remove(finishA);
    ArrayList alist = new ArrayList(realActions);
    Action current;
    Action swap;
    logger.info("unsorted: ");
    for(int i=0; i<alist.size(); i++)
      logger.info(alist.get(i));
    
    // simple bubble sort to get linearize version
    for(int i = 0; i< alist.size(); i++){
      current  = (Action) alist.get(i);
      for(int j = i+1; j<alist.size(); j++){
	swap = (Action) alist.get(j);
	if (isBefore(swap, current)){
	  logger.info("action " + swap + " is before " + current);
	  alist.set(j, current);
	  alist.set(i,swap);
	  j=alist.size();
	}
      }
    }
    logger.info("sorted: ");
    for(int i=0; i<alist.size(); i++)
      logger.info(alist.get(i));
    return alist;
  }
  
  private class MyListener implements GraphAddEdgeListener
  {
    public void edgeAdded(GraphAddEdgeEvent e) 
    {
      return;
    }
  }
  
  public DirectedGraph toGraph(Hashtable services)
  {
    ArrayList acts = new ArrayList(actions);
    //linearize();
    Hashtable vertices = new Hashtable();
    //     Graph graph = new DirectedGraphImpl();
    //     graph.addGraphAddEdgeListener(new MyListener());
    DirectedGraph graph = new DirectedGraphImpl();
    //    Graph graph = new GraphImpl();

    // First annotate all vertices with the corresponding ReasoningService and 
    // put them in the vertices Hashtable
    try{
      for(int i = 0; i< acts.size(); i++){
	Action a = (Action) acts.get(i);
	String longName = a.toString();
	String actName = longName;
	// cut off the _number at the end of instantiated action names
	String[] st= longName.split("_"); 
	if (st.length>0) actName = st[0]; 
	VertexImpl v = new VertexImpl(longName);
	v.setObject(services.get(actName));
	if (! vertices.containsKey(longName)) {
	  vertices.put(longName, v);
	  graph.add(v);
	}
      }
      // go over all causal links and introduce a graph edge for them.
      for(Iterator it = causalLinks.iterator(); it.hasNext(); ){
	CausalLink cl = (CausalLink) it.next();
	Literal condition = cl.getCondition();
	if ((condition.getPredicate().getName().equals("type"))){
	  Action achiever = cl.getAchiever();
	  Action needer = cl.getNeeder();
	  if (acts.contains(achiever) && acts.contains(needer)){
	    String achName = achiever.toString();
	    String needName = needer.toString();
	    
	    // cut off the _number at the end of instantiated action names
	    String[] st;
	    
	    // 	   st = achName.split("_");
	    // 	  if (st.length>0) achName = st[0];
	    // 	  st = needName.split("_");
	    // 	  if (st.length>0) needName = st[0];
	    
	    Vertex v1 = (Vertex) vertices.get(achName);
	    Vertex v2 = (Vertex) vertices.get(needName);
	    Edge e = new DirectedEdgeImpl(v1,v2);
	    e.setFollowVertexLabel(false);
	    Argument a0 = condition.getArgument(0);
	    Argument a1 = condition.getArgument(1);
	    String inputName = a0.toString();
	    st = inputName.split("@");
	    if (st.length>0) inputName = st[0];
	    e.setLabel(inputName + "::" + a1.toString());
	    //e.setLabel(inputName);
	    
	    // 	  // NOTE: A service might get two inputs of the same type (OWL class)!
// 	  // so we have to use the name of the inputs to identify them
// 	  ReasoningService target = services.get(v2.getLabel());
// 	  List inputs = target.getProblem().getInputs();
// 	  for(int i=0; i<inputs.size(); i++){
// 	    //      logger.info("in: " + inputs.get(i).getClass().getName());
// 	    Input e = (Input) inputs.get(i);
// 	    String iName = e.getName();
// 	  }
	  
	  List es = graph.getEdges(v1);
	  boolean contains = false;
	  // 	  for(Iterator it2 = es.iterator(); it2.hasNext(); )
// 	    {
// 	      Vertex b = ((Edge) it2.next()).getVertexB();
// 	      if (b.equals(v2))
// 		contains = true;
// 	    }
	  if (!contains) 
	    graph.addEdge(e);	      
	  }
	}
      }
    } catch(Exception e){
      e.printStackTrace();
    }
    return graph;
  }
  
    
  /**
   * @return The number of causal links in this plan.
   */
  public int getNumCausalLinks() {
    return causalLinks.size();
  }

}
