package org.mathweb.reasoning.partialorderplanner;

import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.search.*;
/**
 * This class contains the algorithm for generating partial order plans,
 * which is formulated as a search through the space of possible partial
 * order plans. Adopted by Juergen Zimmer.
 * @author Daniel Terry, Harlan Crystal
 * @date November 2, 2003
 */

/**
   java org.mathweb.reasoning.partialorderplanner.POPTest /home/jzimmer/import/mathweb/agents/src/org/mathweb/reasoning/tests/atp3.txt
*/

public class PartialOrderPlanner extends BasicService {

  public static Logger logger = Logger.getLogger(org.mathweb.reasoning.partialorderplanner.PartialOrderPlanner.class);
  private static String logFileName = getSystemProperty("pop.log.file", "POP_LOG_FILE");

  /** The logic system we exist in */
  private LogicSystem system;
  /** The start action */
  private Action startA;
  /** The finish action */
  private Action finishA;
  private POPNode root;
  Queue fringe = new Queue();
  //Fringe fringe;
  int numGenerated = 1;
  int numFullPlans = 1;
  int maxFullPlans = 100;
  private SearchTree searchTree = null;
  
  /** 
   * Constructor.
   *
   * @param system The logic system that applies.
   * @param initial The initial state.
   * @param goal The goal state.
   */
  public PartialOrderPlanner(LogicSystem system, State initial, State goal){
    this.system = system;
    root = new POPNode(0, createInitialPlan(initial, goal));
    logger.setLevel(Level.OFF);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }

  public void setSearchTree(SearchTree searchTree) {
    this.searchTree = searchTree;
  }

  public void setMaxPlans(int maxPlans){
    this.maxFullPlans = maxPlans;
  }
  

  /**
   * Creates the initial partial order plan.
   *
   * @param initial The initial state.
   * @param goal The goal state.
   */
  private PartialOrderPlan createInitialPlan(State initial, State goal){
    Literal[] initialLiterals = new Literal[initial.getLiteralsCount()];
    Literal[] goalLiterals = new Literal[goal.getLiteralsCount()];

    startA = new Action(new Task("Start"), new Literal[0], initial.copyLiteralsInto(initialLiterals), new BindingConstraints());
    finishA = new Action(new Task("Finish"), goal.copyLiteralsInto(goalLiterals), new Literal[0], new BindingConstraints());

    PartialOrderPlan initPlan = new PartialOrderPlan(startA, finishA);
    initPlan = initPlan.addOrderConstraint(new OrderConstraint(startA, finishA));

    for(int i = 0; i < finishA.getNumPreconditions(); i++)
      initPlan = initPlan.addOpenCondition(new OpenCondition(finishA.getPrecondition(i), finishA));

    logger.info("POPner: initial Plan: \n" + initPlan + "\n");
    return initPlan;
  }

  /**
   * A function for debugging which allows the user to select which open precondition to choose.
   *
   * @param plan The plan to choose from.
   */
  public OpenCondition chooseOpenCondition(PartialOrderPlan plan){
    logger.info("Choose an open precondition");
    Vector ocs = new Vector(plan.getOpenPreconditions());
    int i = 0;
    for(Iterator ocit = ocs.iterator(); ocit.hasNext(); i++){
      logger.info(i + ": " + ocit.next());
    }
    try{
	    
      i = (int)System.in.read();
      System.in.read();
      System.in.read();
      Vector ret = new Vector(1);
      return (OpenCondition)ocs.get(i-(int)'0');
    }catch(Exception ioe){
      logger.info(ioe);
    }
    return null;
  }

  /**
   * A function for debugging which allows the user to select which achieving action to expand.
   *
   * @param actions The collection of achieving actions.
   */
  private Collection chooseAchievingAction(Collection actions){
    logger.info("Choose an achieving action");
    Vector as = new Vector(actions);
    int i = 0;
    for(Iterator asit = as.iterator(); asit.hasNext(); i++){
      logger.info(i + ": " + ((Action)(((Object[])asit.next())[0])).toMediumString());
    }
    try{
      i = (int)System.in.read();
      System.in.read();
      System.in.read();
      Vector ret = new Vector(1);
      ret.add(as.get(i-(int)'0'));
      return ret;
    }catch(Exception ioe){
      logger.info(ioe);
    }
    return null;
  }

  public POPNode getRoot() {
    return root;
  }

  /**
   * Generate all the possible PartialOrderPlans that can be generated by dealing
   * with the next literal in the list of openConditions in the given PartialOrderPlan
   *
   * @param plan the plan from which to remove an open condition
   * @return a list of PartialOrderPlans that result from removing the next open condition or null of none exist
   */
  public Set expandPlan(PartialOrderPlan plan){
    OpenCondition openCondition = plan.getAnOpenCondition();
    PartialOrderPlan startPlan = plan.removeOpenCondition(openCondition);
    HashSet actions = startPlan.getActions();

    Literal p = openCondition.getCondition(); 
    //logger.info("Trying to achieve condition " + p);
    if (searchTree != null)
      searchTree.addNode(plan, p); /*GUI: openCondition*/
    //logger.info(" available actions " + startPlan.getActions());
    
    system = system.renameActions(NameGen.getNameGen());
    Collection achievingActions = system.getAchievingActionsPOP(p, startPlan.getBindingConstraints(), actions);

    Set newPlansFin = new HashSet();
    for (Iterator it = achievingActions.iterator(); it.hasNext(); ){
      Object [] achieverPair = (Object[]) it.next();
      BindingTable achieveBt = (BindingTable)achieverPair[1];
      HashMap compBinds = achieveBt.compressBindings(NameGen.getNameGen());
      Action achiever = ((Action)achieverPair[0]);
      //      logger.info("Achieving action " + achiever);

      Set newPlans = startPlan.achievePrecondition(achiever, p, //openCondition.getCondition(), 
						   openCondition.getOwner(), compBinds);
      newPlansFin.addAll(newPlans);
      if(null != searchTree)
	addSearchTreeNodes(plan, newPlans, openCondition, achiever);/*GUI*/
    }

    return newPlansFin;
  }

  /**
   * Used by GUI
   */
  //     public Set expandPlan(PartialOrderPlan parent, OpenCondition openCondition){
  // 	PartialOrderPlan startPlan = plan.removeOpenCondition(openCondition);
  // 	return expandPlan(startPlan, parent, oc)
  //     }

  //     /**
  //      * Used to generate plans internally (without user intervention)
  //      * @param startPlan already has the openCondition removed from it
  //      * @param parent is startPlan without the oc removed (used for GUI stuff)
  //      */
  //     private Set expandPlan(PartialOrderPlan startPlan, PartialOrderPlan parent, OpenCondition openCondition){
	
  //     }



  /*adds a node for the achieving action and then gives it kids for each plan that action produces*/
  //     private void addSearchTreeNodes(PartialOrderPlan parent, Set pops, OpenCondition openCondition, Action achiever){
  // 	searchTree.addNode(openCondition.getCondition(), achiever); /*GUI: */
  // 	for(Iterator it = pops.iterator(); it.hasNext();){
  // 	    searchTree.addNode(achiever, it.next()); /*GUI*/
  // 	}
  //     }

  /*Each plan node is represented by the action used to achieve it*/
  private void addSearchTreeNodes(PartialOrderPlan parent, Set pops, OpenCondition openCondition, Action achiever){
    if(pops.size() == 0){
      searchTree.addNode(openCondition.getCondition(), achiever);
      searchTree.addNode(achiever, "DeadEnd");
    }
    for(Iterator it = pops.iterator(); it.hasNext();){
      searchTree.addNode(openCondition.getCondition(), achiever, (PartialOrderPlan)it.next()); /*GUI*/
    }
  }

  
  /**
   * @return the winning plan or null if not found yet
   **/
  public PartialOrderPlan generateNext(){
    try {
      POPNode curNode = (POPNode) fringe.getNext();
      //      logger.info("next fringe plan: " + curNode);
      if (curNode.equals(root))
	NameGen.getNameGen().reset();
      Set expandedPlans = expandPlan(curNode.getPlan());

      numGenerated += expandedPlans.size();
      PartialOrderPlan finalPlan = null;
      
      for (Iterator it = expandedPlans.iterator(); it.hasNext(); ) {
	PartialOrderPlan p = (PartialOrderPlan) it.next();
	if (numGenerated > 1000) {
	  //	  logger.info("stopped at: " + curNode.getPlan());
	  return p; //finalPlan = p;
	}
	if (p.hasNoOpenConditions()) {
	  numFullPlans += 1;
	  if (numFullPlans > maxFullPlans) fringe = new Queue();
	  return p; //finalPlan = p;
	}
	else fringe.addToFringe(new POPNode(curNode.getDepth()+1, p));
      }
      if (finalPlan != null) return finalPlan;
    } catch(Exception e){
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Generates a plan to move from the given initial state to the goal state.
   *
   * @param initial The initial state.
   * @param goal The goal state.
   * @return A plan to move from the inital state to the goal state.  If no plan could be found, returns null.
   */
  public PartialOrderPlan generatePlan(){
    PartialOrderPlan p;
    //	fringe = new IterativeDeepeningStack();
    fringe = new Queue();
    fringe.addToFringe(root);

    while(fringe.hasNext()){
      p = generateNext();
            //      logger.info("next: " + p);
      
      if(p != null){
	//logger.info("Completed Plan:\n" + p);
	return p;
      }
    }
    
    return null;
  }


  /**
   * Generates a plan to move from the given initial state to the goal state.
   *
   * @return The next plan.  Returns null if no plan could be found.
   */
  public PartialOrderPlan generateNextPlan(){
    PartialOrderPlan p;
    
    while(fringe.hasNext()){
      p = generateNext();
      //      logger.info("next: " + p);
      
      if(p != null){
	//logger.info("Completed Plan:\n" + p);
	return p;
      }
    }
    return null;
  }
}
