package org.mathweb.reasoning.planning;

import java.util.*;

import org.mathweb.onto.OntoReasoner;
 
import org.mathweb.reasoning.logic.Argument;
import org.mathweb.reasoning.logic.BindingConstraints;
import org.mathweb.reasoning.logic.BindingTable;
import org.mathweb.reasoning.logic.Literal;
import org.mathweb.reasoning.logic.Variable;
import org.mathweb.reasoning.logic.CachedHashCodeObject;
import org.mathweb.reasoning.partialorderplanner.NameGen;

/**
 * An Action represents an instantiated task: it has a symbol to 
 * identify it (a Task) and a list of pre and post conditions
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 22, 2003
 */
public class Action extends CachedHashCodeObject {
    final private Task task; //Symbol; name for this action (eg move)
    final protected Literal[] preconditions;
    final protected Literal[] postconditions;
    final private BindingConstraints notEqualVars;
    final private int actionNum; // unique identifier associated with this action.

    /**
     * Generates a new action.
     *
     * @param task The name of this action.
     * @param notEqualVars A list of pairs of variables constrained to be not equal.
     * @IllegalArgumentException Thrown a variable occurs in the 
     * postconditions but not the preconditions.
     */
    public Action(Task task, Literal [] preconditions, Literal [] postconditions, 
                  BindingConstraints notEqualVars) {
	this.task = task;
	this.preconditions = preconditions;
	this.postconditions = postconditions;
        this.notEqualVars = notEqualVars;
        this.actionNum = 1; // until rename is requested, assume this is the first action of this type so give actionNum = 1
        if (! isConsistent())
            throw new IllegalArgumentException("Variable occurs in the postconditions of an action but not in the preconditions.");
    }

    private Action(Task task, Literal [] preconditions, Literal [] postconditions,
                   BindingConstraints notEqualVars, int actionNum) {
	this.task = task;
	this.preconditions = preconditions;
	this.postconditions = postconditions;
        this.notEqualVars = notEqualVars;
        this.actionNum = actionNum;
    }

    /**
     * Copy constructor.
     */
    public Action(Action a) {
        this.task = a.task;
        this.preconditions = a.preconditions;
        this.postconditions = a.postconditions;
        this.notEqualVars = a.notEqualVars;
        this.actionNum = a.actionNum;
    }

    public Task getTask() {
        return task;
    }

    /**
     * @return The binding constraints of this action.
     */
    public BindingConstraints getBindingConstraints() {
        return notEqualVars;
    }

    /**
     * @return The number of postconditions.
     */
    public int getNumPostconditions() {
        return postconditions.length;
    }

    /**
     * @return The number of preconditions.
     */
    public int getNumPreconditions() {
        return preconditions.length;
    }

    public Literal getPostcondition(int i) {
        return postconditions[i];
    }

    public Literal getPrecondition(int i) {
        return preconditions[i];
    }

// action {
//   ontologyProofAltProof
//   pre {
//     constraint(proofOf, proof1, prob),
//     constraint(alternativeProof, proof2, proof1)	
//       }
//   post {
//     constraint(proofOf, proof2, prob)
//   }
//   constraints {
//   }
// }

    public String toProbString(){
      String s = "action {\n " + task + "_" + actionNum + "\n\tpre{\n";
      for(int i = 0; i < preconditions.length; i++){
	s += "\t\t" + preconditions[i];
	if (i < preconditions.length) s += ",\n";
	else s += "\n";
      }
      
      s += "}\n\t post {\n";
      for(int i = 0; i < postconditions.length; i++){
	s += "\t\t" + postconditions[i];
	if (i < postconditions.length) s += ",\n";
	else s += "\n";
      }
      s += "}\n\tconstraints {\n" + notEqualVars;
      s += "}\n";
      s += "}\n";
      return s;
    }
  
    public String toLongString(){
	String s = "Action: " + task + "_" + actionNum + "\n\tPreconditions:\n";
	for(int i = 0; i < preconditions.length; i++)
	    s += "\t\t" + preconditions[i] + "\n";
	s += "\tPostconditions:\n";
	for(int i = 0; i < postconditions.length; i++)
	    s += "\t\t" + postconditions[i] + "\n";
        s += "\tConstraints:\n" + notEqualVars;
	return s;
    }

    public String toMediumString() {
        String out = task + "_" + actionNum + ": ";
        for (int i=0; i < preconditions.length; i++) {
            out += preconditions[i] + " " + (char)94 + " ";
        }
	out += " ==> ";
        for (int i=0; i < postconditions.length; i++) {
            out += postconditions[i] + " " + (char)94 + " ";
        }
        return out + "||(" + notEqualVars + ")";
    }

    public String toStringShort() {
        String out = task + "_" +actionNum+ " (";
        HashSet set = new HashSet();
        for (int i=0; i < preconditions.length; i++) {
            set.addAll(preconditions[i].getArguments());
        }
        for (int i=0; i < postconditions.length; i++) {
            set.addAll(postconditions[i].getArguments());
        }
        for (Iterator it=set.iterator(); it.hasNext(); ) {
            out += it.next();
            if (it.hasNext())
                out += ", ";
        }
        return out + ")";
    }

    public String toString() {
	return task + "_" + actionNum;
    }

    /**
     * Make a copy of my literals as a vector
     */
    public Vector copyPreconditions(){
	Vector v = new Vector(preconditions.length);
	for(int i = 0; i < preconditions.length; i++)
	    v.add(preconditions[i]);
	return v;
    }

    /**
     * Make a copy of my literals as a vector
     */
    public Vector copyPostconditions(){
	Vector v = new Vector(postconditions.length);
	for(int i = 0; i < postconditions.length; i++)
	    v.add(postconditions[i]);
	return v;
    }

    /**
     * Checks that an action is consistent, in that all variables in the 
     * post conditions occur in the preconditions.
     *
     * @return True if this action is consistent.
     */
    public boolean isConsistent() {
        HashSet vars = new HashSet();
        for (int i = 0; i < postconditions.length; i++)
            vars.addAll(postconditions[i].getVariables());
        
        for (int i = 0; i < preconditions.length; i++)
            vars.removeAll(preconditions[i].getVariables());
	//        return vars.isEmpty();
	return true;
	
    }

    /**
     * @return True if there are no variables in this action.
     */
    public boolean isBound() {
        for (int i = 0; i < preconditions.length; i++)
            if (!preconditions[i].isBound()) 
                return false;

        for (int i = 0; i < postconditions.length; i++)
            if (!postconditions[i].isBound()) 
                return false;

        return true;
    }

    /**
     * @return a copy of this Action with all variables replaced the constants they are bound to
     */
    public Action applyBindings(HashMap compressedBindings){
	//Apply bindings to the preconditions
	Literal[] otherPreconditions = new Literal[preconditions.length];
	for(int i = 0; i < otherPreconditions.length; i++)
	    otherPreconditions[i] = preconditions[i].applyBindings(compressedBindings);

	//Apply bindings to the postconditions
	Literal[] otherPostconditions = new Literal[postconditions.length];
	for(int i = 0; i < otherPostconditions.length; i++)
	    otherPostconditions[i] = postconditions[i].applyBindings(compressedBindings);

	//return new Action(task, otherPreconditions, otherPostconditions, new BindingConstraints());
// 	System.out.println(task + "::notEqualVarnotEqualVars:" + notEqualVars);
// 	System.out.println(task + "::notEqualVarnotEqualVars.applyBindings(bindings):" + notEqualVars.applyBindings(bindings));
 	Action applied = new Action(task, otherPreconditions, otherPostconditions, notEqualVars.applyBindings(compressedBindings), actionNum);
        if (applied.equals(this))
            return this; // save memory!
        return applied;
    }

    protected void computeHashCode() {
        int value = task.hashCode();
        value = value ^ (new Integer(actionNum)).hashCode();
        for (int i = 0; i < preconditions.length; i++)
            value = value ^ preconditions[i].hashCode();
        for (int i = 0; i < postconditions.length; i++)
            value = value ^ postconditions[i].hashCode();
        myhashcode = value;
    }

    /**
     * Checks the equality of two actions.
     *
     * NOTE: does NOT handle alpha equality.  ie, the following case:
     *
     * ie:   pre {pred(x)} post {~pred(x)} == pre {pred(y)} post {~pred(y)}
     */
    public boolean equals(Action a) {
        if (! task.equals(a.task))
            return false;
        if (actionNum != a.actionNum)
            return false;
        if (! notEqualVars.equals(a.notEqualVars))
            return false;
        if (preconditions.length != a.preconditions.length)
            return false;
        for (int i = 0; i < preconditions.length; i++)
            if (! preconditions[i].equals(a.preconditions[i]))
                return false;
        if (postconditions.length != a.postconditions.length)
            return false;
        for (int i = 0; i < postconditions.length; i++)
            if (! postconditions[i].equals(a.postconditions[i]))
                return false;
        return true;
    }

    public boolean equals(Object o) {
        try {
            return equals((Action) o);
        } catch (ClassCastException e) {
            return false;
        }
    }

    public boolean equalsByName(Action a) {
	return task.equals(a.getTask());
    }

    public boolean equalsByOriginalName(Action a) {
      return task.equalsByOriginalName(a.getTask());
    }
  
    /**
     * Find a bindingTable that associates an action with a state:
     * that is, bind the variables in the action's preconditions
     * to constants in the actions
     *
     * @return the BindingTable that binds every variable
     * in the action to a constant in the state or null if no
     * possible binding could be found
     */
//     public Vector bindPreconditionsToState(State s){
//  	Vector constLits = new Vector(s.getLiteralsCount());
//  	s.copyLiteralsInto(constLits);
//  	return Actor.findBindings(copyPreconditions(), constLits, getBindingConstraints());
//     }

    public Vector bindPreconditionsToState(State s, OntoReasoner reasoner){
 	Vector constLits = new Vector(s.getLiteralsCount());
 	s.copyLiteralsInto(constLits);
 	return Actor.findBindings(copyPreconditions(), constLits, getBindingConstraints(), reasoner);
    }

//     /**
//      * Binds according to my own bindingConstraints
//      */
//     public Vector bindPostconditionsToLiteral(Literal l){
// 	return bindPostconditionsToLiteral(l, notEqualVars);
//     }

    /**
     * Binds according to my own bindingConstraints
     */
    public Vector bindPostconditionsToLiteral(Literal l, OntoReasoner reasoner){
      return bindPostconditionsToLiteral(l, notEqualVars, reasoner);
    }

//     /**
//      * Binds according to exteranal bindingConstraints
//      */
//     public Vector bindPostconditionsToLiteral(Literal l, BindingConstraints bc){
//  	BindingTable empty = new BindingTable();
// 	Vector possibleBindings = new Vector();
// 	for(int i = 0; i < postconditions.length; i++){
// 	    if(postconditions[i].isTrue()){
// 		BindingTable bt = Actor.bindLiteral(postconditions[i], l, empty, bc);
// 		if(null != bt)
// 		  possibleBindings.add(bt);
// 	    }
// 	}
// 	return possibleBindings;
//     }

    /**
     * Binds according to exteranal bindingConstraints
     */
    public Vector bindPostconditionsToLiteral(Literal l, BindingConstraints bc, OntoReasoner reasoner){
 	BindingTable empty = new BindingTable(reasoner);
	Vector possibleBindings = new Vector();
	for(int i = 0; i < postconditions.length; i++){
	  if(postconditions[i].isTrue()){
	    BindingTable bt = Actor.bindLiteral(postconditions[i], l, empty, bc);
	    if(null != bt)
	      possibleBindings.add(bt);
	  }
	}
	return possibleBindings;
    }

    /**
     * Renames all the variables in this action.
     *
     * @param nameGen For generating new names.
     * @return The action with all variables renamed to unique new ones.
     */
    public Action renameVariables(NameGen nameGen){
	return renameVariables(nameGen, new HashMap());
    }

    /** 
     * Renames the action's variables and also updates the actionNum so that the renamed action != the 
     * old action.
     *
     * this could be a bit faster if you combine the loop to figure out what vars I use with one which assigns new names
     */
    public Action renameVariables(NameGen nameGen, HashMap newNames){
	Literal[] newPres = new Literal[preconditions.length];
	Literal[] newPosts = new Literal[postconditions.length];
	for(int i = 0; i < newPres.length; i++)
	    newPres[i] = preconditions[i].renameVariables(nameGen, newNames);

	for(int i = 0; i < newPosts.length; i++)
	    newPosts[i] = postconditions[i].renameVariables(nameGen, newNames);

	BindingConstraints newBcs = notEqualVars.applyBindings(newNames);
        return new Action(task, newPres, newPosts, newBcs, nameGen.getNext());
    }

    
//     /**
//      * @param condition The condition to check if this action fulfills as a postcondition.
//      * @param bc The external binding constraints which apply during the check.
//      * @return True if this action fufills this condition by having it in the postconditions.
//      */
//     public boolean doesFulfill(Literal condition, BindingConstraints bc) {
//         for (int i = 0; i < postconditions.length; i++) {
//             BindingTable bt = Actor.bindLiteral(postconditions[i], condition, new BindingTable(), notEqualVars.addAll(bc));
//             if (bt != null)
//                 return true;
//         }
//         return false;
//     }

//     /**
//      * @param condition The condition to check if this action fulfills as a postcondition.
//      * @param bc The external binding constraints which apply during the check.
//      * @return True if this action fufills this condition by having it in the postconditions.
//      */
//     public boolean doesFulfill(Literal condition, BindingConstraints bc, OntoReasoner reasoner) {
//         for (int i = 0; i < postconditions.length; i++) {
//             BindingTable bt = Actor.bindLiteral(postconditions[i], condition, new BindingTable(reasoner), notEqualVars.addAll(bc));
//             if (bt != null)
//                 return true;
//         }
//         return false;
//     }

//     /**
//      * @param condition The positive condition to check if this action deletes.
//      * @param bc The binding constraints that apply during the scan.
//      * @return True if this action threatens this condition by deleting it.
//      */
//     public boolean doesThreaten(Literal condition, BindingConstraints bc) {
//         return doesFulfill(condition.not(), bc);
//     }

//     /**
//      * @param condition The positive condition to check if this action deletes.
//      * @param bc The binding constraints that apply during the scan.
//      * @return True if this action threatens this condition by deleting it.
//      */
//     public boolean doesThreaten(Literal condition, BindingConstraints bc, OntoReasoner reasoner) {
//         return doesFulfill(condition.not(), bc, reasoner);
//     }


    /**
     * Finds all the threats that this action causes for a given literal.  It then returns the set of
     *  binding constraints that would fix the threat.
     *
     * @return Null, if there are no threats caused by this pair.  Otherwise this returns the set of
     * new constraints that would make this pair fine.  If an constraints set that is empty is returned,
     * there is no set of contraints which make it work.
     */
    public BindingConstraints resolveThreats(Literal keepsafe, BindingConstraints oldbc, OntoReasoner reasoner) {
        BindingConstraints bc = null;
        keepsafe = keepsafe.not();
        
        //System.out.println("Action examined: " + a);
        for (int i = 0; i < getNumPostconditions(); i++) {
            Literal possibleThreat = getPostcondition(i);
            if (possibleThreat.equalsExactly(keepsafe) && bc == null) {// make sure to indicate a threat
                return new BindingConstraints();
            } else {
                // is this right?
                BindingTable bt = Actor.bindLiteral(possibleThreat, keepsafe, new BindingTable(reasoner), oldbc);
                if (bt != null) { // there were possible bindings.. add them to the constraints.
                    if (bc == null)
                        bc = new BindingConstraints();
                    for (Iterator keysit = bt.getKeysIter(); keysit.hasNext(); ) {
                        Variable var = (Variable) keysit.next();
                        for (Iterator bit = bt.getPositiveBindings(var).iterator(); bit.hasNext(); ) {
                            Argument arg = (Argument) bit.next();
                            bc = bc.addConstraint(var, arg);
                        }
                    }
                }
            }
        }

        return bc;
    }
}
