package org.mathweb.reasoning.planning;

import java.util.*;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.reasoning.logic.*;
/**
 * Does various methods concerning binding to actions.
 *
 * @author Daniel Terry
 */
public class Actor {
  public static boolean verbose = true; //TODO: remove


    /**
     * Checks to see if the variables in varLit can be bound to constLit
     * and if returns a _copy_ of bound with the new bindings added.
     * A binding is possible if:
     * 1. both literals have the same predicate
     * 2. none of the variables in varLit have already been bound to 
     *    different constants in bound
     * The way of the future
     * TODO: this way is slower than the way of the past because it makes a copy of bound
     * regardless of whether a binding can be found (this can be fixed)     * 
     * @param varLit: Literal whose variables we will bind to constants
     *                Can contain constant or variable arguments
     * @param constLit: Literal whose constants we will bind to variables
     *                  Must contain only constant arguments
     * @param constraints A list of constraints on variables being not equal.
     * @return _copy_ of bound with new bindings added or null if no binding exists
     */
  public static BindingTable bindLiteral(Literal varLit, Literal constLit, BindingTable bound,
					 BindingConstraints constraints){
    //Both literals must have the same predicate
    if(!constLit.getPredicate().equals(varLit.getPredicate())){
      return null;
    }
    
    if (constLit.isTrue() != varLit.isTrue())
      return null;

    //Note the new bindings made
    BindingTable newBindings = new BindingTable(bound);
    for(int i = 0; i < varLit.getArgumentCount(); i++){
      newBindings = newBindings.addBinding(varLit.getArgumentAt(i), constLit.getArgumentAt(i), constraints);
      if(newBindings == null)
	return null;
    }
    return newBindings;
  }
  
//     /**
//      * Find a set of bindingTable that bind the variables in the 
//      * varible literals (varLits) to the constants in constant
//      * literals (constLits)
//      **/
//     public static Vector findBindings(Vector varLits, 
// 			       Vector constLits, BindingConstraints constraints){
// 	return findBindings(varLits, constLits, new BindingTable(), constraints);
//     }

    /**
     * Find a set of bindingTable that bind the variables in the 
     * varible literals (varLits) to the constants in constant
     * literals (constLits)
     **/
    public static Vector findBindings(Vector varLits, 
			       Vector constLits, BindingConstraints constraints, OntoReasoner reasoner){
	return findBindings(varLits, constLits, new BindingTable(reasoner), constraints);
    }

    /**
     * Recursive helper function for findBindings above.
     * Given bindings that have already been made (all in bound), 
     * this function finds all the possible bindingTables
     * that can bind the variables in the varible literals (varLits) 
     * to the constants in the constant literals (constLits)
     **/
    public static Vector findBindings(Vector varLits, 
				      Vector constLits,
				      BindingTable bound, 
                                      BindingConstraints constraints){

	Vector possibleBindings = new Vector();

	//All varible literals have had all of their variables assigned
	if(varLits.size() == 0){
	    possibleBindings.add(bound);
	    return possibleBindings;
	}

	//Get the next unmatched variable literal
	Literal matchMe = (Literal)varLits.lastElement();
	
	//Look at every unmatched constant literal and see if you can
	//match matchMe to it; if so, search the rest for bindings for the
	//rest of the varLits that are consistent with the bindings 
	//in bound and the new ones from matchMe:
	Literal constLit;
	BindingTable newBindings;
	Vector remainingVarLits;
	Vector remainingConstLits;
	for(int i = 0; i < constLits.size(); i++){
	    constLit = (Literal)constLits.get(i);

	    //Check if matchMe can be bound to the next Literal in constLits
	    //if so then get a new BindingTable with bindings from 
	    //the variables in matchMe to the constants in matched
	    newBindings = bindLiteral(matchMe, constLit, bound, constraints);

	    //If that binding worked out, hold on to it and search
	    //through the other varLits with it as a constraint (ie recurse)
	    if(newBindings != null){
		remainingVarLits = (Vector)varLits.clone();
		remainingVarLits.remove(matchMe);
		possibleBindings.addAll(findBindings(remainingVarLits, constLits, newBindings, constraints));
	    }
	}
	return possibleBindings;
    }
}
