package org.mathweb.reasoning.planning;

import java.util.*;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.partialorderplanner.NameGen;
/**
 * Encapsulates all the data that is required for planning.
 * All of the actions that are available to use.
 * 
 * @author Harlan Crystal, Daniel Terry
 * @date October 28, 2003
 */
public class LogicSystem {
  /** The actions that are available to use */
  private Vector actions;
  private OntoReasoner reasoner = null;
  
  /**
   * Constructs a new logic system.
   *
   * @param actions A vector of Action objects that are available to use.
     */
  public LogicSystem(Vector actions) {
    this.actions = (Vector) actions.clone();
  }

  /**
   * Constructs a new logic system with an OWL reasoner.
   * (added by jzimmer 14.7.04)
   * @param actions A vector of Action objects that are available to use.
   * @param reasoner An OWL ontology reasoner for subsumption testing
   */
  public LogicSystem(Vector actions, OntoReasoner reas) {
    this.actions = (Vector) actions.clone();
    this.reasoner = reas;
   }

  /**
   * Renames each actions, so no two actions have the same variable names
   */
  public LogicSystem renameActions(NameGen nameGen){
    Vector newActions = new Vector(actions.size());
    for(Iterator ait = actions.iterator(); ait.hasNext(); ){
      newActions.add(((Action)ait.next()).renameVariables(nameGen));
    }
    return new LogicSystem(newActions, reasoner);
  }


  /**
   * Add an action
   * @param a the action to be added
   * @return void
   **/
  public LogicSystem addAction(Action a){
    LogicSystem newSys = new LogicSystem(((Vector)actions.clone()), reasoner);
    newSys.actions.add(a);
    return newSys;
  }

  /**
   * Generates ALL the states that can result from applying ANY applicable action to given state.
   *
   * @param s The state to expand.
   * @return A vector of States that can result from applying an action to state.
   */
  public Vector generateSuccessors(State s) {
    Vector nextStates = new Vector();

    for(Iterator it = getApplicableActions(s).iterator(); it.hasNext(); ) {
      nextStates.add((new BoundAction((Action)it.next())).applyAction(s));
    }

    return nextStates;
  }

  /**
   * This function returns all possible bound actions, not just all 
   * possible action types
   *
   * Used by basic planner.
   * 
   * @return The Vector of BoundActions which may be applied to this state.
   */
  public Vector getApplicableActions(State s) {
    Vector applicableActions = new Vector();
        
    //Look at every action
    for (Iterator it = actions.iterator(); it.hasNext(); ) {
      Action a = (Action) it.next();
      //Vector bindings = Actor.bindPreconditionsToState(a, s); DT
      Vector bindings = a.bindPreconditionsToState(s, reasoner);

      //Instantiate the action in each way it can be bound to this state
      for(Iterator bindit = bindings.iterator(); bindit.hasNext(); ) {
	Action applied = a.applyBindings((HashMap) ((BindingTable)bindit.next()).compressBindings(null));
	BoundAction bAction = new BoundAction(applied);
	applicableActions.add(bAction);
      }
    }
    return applicableActions;
  }

  /**
   * @return Vector of Object pairs: 
   *     pair[0] = AchievingAction (with pair[1] instantiated on it); 
   *     pair[1] = BindingTable that works with pair[0]
   */
  public Vector getAchievingActions(Literal l){
    Vector applicableActions = new Vector();
    for (Iterator it = actions.iterator(); it.hasNext(); ){
      Action a = (Action) it.next();
      Vector bindings = a.bindPostconditionsToLiteral(l, reasoner);
      //System.out.println("LogicSystem::getAchievingActions: checking Action " + a);
	    
      //Partially instantiate the action in each way it's postconditions 
      //can be bound to this literal
      for(Iterator bindit = bindings.iterator(); bindit.hasNext(); ) {
	Object[] newPair = new Object[2];
	newPair[1] = bindit.next();
	newPair[0] = a.applyBindings((HashMap)((BindingTable)newPair[1]).compressBindings(null));
	//newPair[0] = a.applyBindings((HashMap)((BindingTable)newPair[1]).compressBindings());
	applicableActions.add(newPair);
      }
    }
    return applicableActions;
  }
    
  /**
   * (extended by jzimmer 14.7.04)
   * 
   * @param l The literal to achieve.
   * @param bc The binding constraints.
   * @param set The set of external actions outside the ones provided inside this LogicSystem which we should scan to 
   * see if they achieve the literal.
   * @return The collection of actions/bindings pairs which may achieve literal l.
   */
  private Collection getAchievingActionsBackchaining(Literal l, BindingConstraints bc, Collection set) {
    Collection applicableActions = new Vector();
    for (Iterator it = set.iterator(); it.hasNext(); ){
      Action a = (Action) it.next();
      //System.out.println("LogicSystem::getAchievingActions: checking Action " + a);
      Vector bindings = a.bindPostconditionsToLiteral(l, bc, reasoner);
      //System.out.println("bindings:" + bindings);
	    
      //Partially instantiate the action in each way it's postconditions 
      //can be bound to this literal
      for(Iterator bindit = bindings.iterator(); bindit.hasNext(); ) {
	Object[] newPair = new Object[2];
	newPair[1] = bindit.next();
	newPair[0] = a;
	//newPair[0] = a.applyBindings((HashMap)((BindingTable)newPair[1]).compressBindings());
	applicableActions.add(newPair);
      }
    }
    return applicableActions;
  }

  /**
   * Extended by jzimmer (14.7.04) to integrate reasoning on ontology.
   *
   * @param l The literal to achieve.
   * @param bc The binding constraints.
   * @param set The set of external actions outside the ones provided inside this LogicSystem which we should scan to 
   * see if they achieve the literal.
   * @return The collection of actions/bindings pairs which may achieve literal l.
   */
  private Collection getAchievingActionsBackchainingWithSubsumption(Literal l, BindingConstraints bc, Collection set) {
    Collection applicableActions = new Vector();
    for (Iterator it = set.iterator(); it.hasNext(); ){
      Action a = (Action) it.next();
      //System.out.println("LogicSystem::getAchievingActionsWithSubsumption: checking Action " + a);
      Vector bindings = a.bindPostconditionsToLiteral(l, bc, reasoner);
      //      System.out.println("bindings:" + bindings);
	    
      //Partially instantiate the action in each way it's postconditions 
      //can be bound to this literal
      for(Iterator bindit = bindings.iterator(); bindit.hasNext(); ) {
	Object[] newPair = new Object[2];
	newPair[1] = bindit.next();
	newPair[0] = a;
	//newPair[0] = a.applyBindings((HashMap)((BindingTable)newPair[1]).compressBindings());
	applicableActions.add(newPair);
      }
    }
    return applicableActions;
  }


  /**
   * TODO: factor code: this is almost extacly the same as above
   *       this is used by the PartialOrderPlanner
   * @param l The literal we want to achieve.
   * @param bc The binding constraints.
   * @param fromPlan A set of actions that were already in the plan that we should search as well.
   * @return a vector of pairs; each pair contains a bindingTable and an action
   */
  public Collection getAchievingActionsPOP(Literal l, BindingConstraints bc, Set fromPlan){
//     Collection done = getAchievingActionsBackchaining(l, bc, fromPlan);
//     done.addAll(getAchievingActionsBackchaining(l, bc, this.actions));
    Collection done = getAchievingActionsBackchainingWithSubsumption(l, bc, fromPlan);
    done.addAll(getAchievingActionsBackchainingWithSubsumption(l, bc, this.actions));
    return done;
  }

  public OntoReasoner getReasoner()
  {
    return reasoner;
  }
  
  public String toString() {
    String ret = "{";
    for (Iterator it=actions.iterator(); it.hasNext(); ) 
      ret += it.next();
    return ret +"}";
  }

}
