package org.mathweb.reasoning.planning;
import org.mathweb.reasoning.logic.Literal;
import org.mathweb.reasoning.logic.CachedHashCodeObject;
import java.util.*;

/**
 * A state is a conjunction of literals.  An immutable class.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 22, 2003
 */
public class State extends CachedHashCodeObject {
    private final Literal[] literals;

    /**
     * Generates a new state.
     *
     * @exception IllegalArgumentException Thrown if any of the literals are negative.
     */
    public State(Literal[] literals){
	this.literals = (Literal []) literals.clone();
        for (int i = 0; i < this.literals.length; i++)
            if (! this.literals[i].isTrue())
                throw new IllegalArgumentException("Negative literal in state");
    }

    /**
     * A private constructor we should use internally that doesn't do any of the sanity checks
     */
    private State(Literal [] literals, boolean dummy) {
        this.literals = literals;
    }

    /** 
     * An efficient copy constructor.
     */
    public State(State x) {
        this.literals = (Literal []) x.literals.clone();
    }

    public boolean equals(State other) {
        if (literals.length != other.literals.length)
            return false;
        
        for (int i = 0; i < literals.length; i++) {
            boolean found = false;
            for (int j = 0; !found && j < other.literals.length; j++) {
                if (literals[i].equals(other.literals[j]))
                    found = true;
            }
            if (!found)
                return false;
        }
        return true;
    }

    public boolean equals(Object other) {
        try {
            return equals((State) other);
        } catch (ClassCastException e) {
            return false;
        }
    }
    
    public String toString(){
	String s = "";
	for(int i = 0; i < literals.length; i++) {
	  s += "\t" + literals[i].toString() + " ";
	  if(i != literals.length - 1)
	    s += ", \n";
	  else s += "\n";
	}
	return s;
    }

    /**
     * @return the number of literals in this state
     */
    public int getLiteralsCount(){
	return literals.length;
    }

    /**
     * Removes literals from this state.
     *
     * @param l A negative literal to remove from this state.
     * @return A state just like this one except with positive literals like l removed.
     * @exception IllegalArgumentException Thrown if given literal is not negative.
     */
    public State removeLiteral(Literal l){
        if (l.isTrue())
            throw new IllegalArgumentException("Not designed to delete positive literals");

        Literal flipped = l.not();
        Vector newLiterals = new Vector();
        for (int i = 0; i < literals.length; i++) {
            if (! literals[i].equals(flipped)) {
                newLiterals.add(literals[i]);
                //System.out.println("Added: " + literals[i] + ".  Compared to: " + flipped);
            }
        }
        Literal [] nextLiterals = new Literal[newLiterals.size()];
        return new State((Literal [])newLiterals.toArray(nextLiterals));
    }

    /**
     * Make a copy of this State with the new literal added
     */
    public State addLiteral(Literal l){
	Literal[] newLiterals = new Literal[literals.length+1];
	for(int i = 0; i < literals.length; i++)
	    newLiterals[i] = literals[i];
	newLiterals[newLiterals.length-1] = l;
	return new State(newLiterals);
    }
    
    /**
     * Copies my literals into the given literal array which 
     * needs to be of the correct size
     * @return the array with the copied literals
     */
    public Literal[] copyLiteralsInto(Literal[] newLiterals){
	for(int i = 0; i < literals.length; i++)
	    newLiterals[i] = literals[i];
	return newLiterals;
    }

    /**
     * Copies my literals into the given vector after
     * clearing the given Vector of any elements it might
     * have in it already
     * @return the Vector with the copied literals
     */
    public Vector copyLiteralsInto(Vector newLiterals){
	newLiterals.clear();
	for(int i = 0; i < literals.length; i++)
	    newLiterals.add(literals[i]);
	return newLiterals;
    }

    /**
     * @return True if list contains item
     */
    private static boolean contains(Literal [] list, Literal item) {
        for (int i = 0; i < list.length; i++)
            if (list[i].equals(item))
                return true;
        return false;
    }

    /**
     * Checks if this state is satisfied by state s.
     *
     * In other words, this state is a subset of s.
     *
     * @return True if the state s "satisfies" this state.
     */
    public boolean isSatisfiedBy(State s) {
        for (int i = 0 ; i < literals.length; i++) {
            if (! contains(s.literals, this.literals[i]))
                return false;
        }
        return true;
    }

    protected void computeHashCode() {
        int value = literals[0].hashCode();
        for (int i = 1; i < literals.length; i++)
            value = value ^ literals[i].hashCode();
        myhashcode = value;
    }

    /**
     * Make a copy of my literals
     */
//     public Literal[] copyLiterals(){
// 	Literal[] newLiterals = new Literal[literals.length];
// 	for(int i = 0; i < literals.length; i++)
// 	    newLiterals[i] = literals[i];
// 	return newLiterals;
//     }

    //NOTE: Deprecated: copyLiterals now returns a copy as an array
    /**
     * Make a copy of my literals as a vector
     */
//     public Vector copyLiterals(){
// 	Vector v = new Vector(literals.length);
// 	for(int i = 0; i < literals.length; i++)
// 	    v.add(literals[i]);
// 	return v;
//     }

}
