// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.serv;

import java.util.List;
import java.util.ArrayList;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;

import org.mindswap.owls.OWLSFactory;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.Condition;
import org.mindswap.owls.profile.Profile;
import org.mindswap.owls.profile.ServiceParameter;
import org.mindswap.owls.profile.ServiceCategory;
import org.mindswap.owls.service.Service;
import org.mindswap.owls.process.execution.ProcessExecutionEngine;
import org.mindswap.wsdl.WSDLOperation;
import org.mindswap.wsdl.WSDLParameter;
import org.mindswap.wsdl.WSDLService;

/**
 */
public class Query {

  public String name;
  public String directive;
  public String uri;
  public ReasoningProblem problem;

  private Service owlsService = null;
  private ArrayList paramNames;

  public ReasoningProblem getProblem(){return problem;}
  public void setProblem( ReasoningProblem prob){ problem=prob;}
  public void trim()
  {
    directive = directive.trim();
    problem.trim();
  }

  public String getName(){
    return name;}

  public String getURI(){
    return uri;}
  public void setURI(String newuri){
     uri = newuri;}

  public Service getOwlsService()
  {
    return this.owlsService;
  }

  public void setOwlsService(Service serv)
  {
    this.owlsService = serv;
  }


  public String toString()
  {
    // TODO
    return "Query with problem: "+ problem.toString();
  }

  
  private String getLitString(Literal lit, List paramNames){
    // check whether the both arguments of the literal are variables and whether they 
    // are the inputs of the service. If so, then add it to the list of
    // literals. Otherwise omit it.
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String an0, an1;
    if ((args[0] instanceof Variable) && (args[1] instanceof Variable)){
      String varName0 = args[0].getName().toLowerCase();
      String varName1 = args[1].getName().toLowerCase();
      String type = null;
      if (!paramNames.contains(varName0)) {
        Variable v0 = (Variable) args[0];
        type = v0.getType();
      }
      if (!paramNames.contains(varName1)){
        Variable v1 = (Variable) args[1];
        type = v1.getType();
      }
      return "(" + lit.getPredicate().getName() + " " + name + "#" + varName0 +
          " " + name + "#" + varName1+")";
    }
    return "";
  }

  private ArrayList getCondStrings(CachedHashCodeObject condition, List paramNames){
    ArrayList condStrs = new ArrayList();
    
    if (condition instanceof Literal) {
      Literal lit = (Literal) condition;
      condStrs.add(getLitString(lit, paramNames));
    }
    else if (condition instanceof Conjunct){
      Conjunct conj = (Conjunct) condition;
      List lits = conj.getLiterals(); 
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        // check whether the both arguments of the literal are variables and whether they 
        // are the inputs of the service. If so, then add it to the list of
        // literals. Otherwise omit it.
        condStrs.add(getLitString(lit, paramNames));
      }
    }
    else if (condition instanceof Implication){
      Implication imp = (Implication) condition;
      List lits = imp.getBody();
      Literal head = imp.getHead();
      String impStr = "(if (and\n";
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        String litStr = getLitString(lit, paramNames);
        impStr += "        " + litStr + "\n";
      }
      impStr += "       )\n       ((add " + getLitString(head, paramNames) + ")))";
      condStrs.add(impStr);
    }
    return condStrs;
  }
  


  public String toProdigyProblem()
  {
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    List preconds = problem.getPreconds();
    List postconds= problem.getPostconds(); 
    String constants = "";
    String objects = "";
    String state = "";
    String goal = "";

    this.paramNames = new ArrayList();
    /*
      all inputs have to be bound.
     */
    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      Input e = (Input) inputs.get(i);
      objects += "                (" + name + "#" +e.getName() + " OWL-" +e.getSignature() + ")\n";
      state += "  (bound " + name + "#" + e.getName() + ")\n";
      paramNames.add(e.getName().toLowerCase());
    }

    /*
      all output parameters have to be NOT bound before the operator application.
    */
    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      objects += "                (" + name + "#" + e.getName() + " OWL-" +e.getSignature() + ")\n";
      goal += "  (bound "  + name + "#" + e.getName() + ")\n";
      paramNames.add(e.getName().toLowerCase());
    }

    for(int i=0; i<5; i++){
      constants += "(pinstance-of dummy_time_resource" + i + " OWL-TimeResource)\n";
      state += "  (bound " + "dummy_time_resource" + i + ")\n";
    }
    
    for(int k=0; k<postconds.size(); k++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      PostCondition po = (PostCondition) postconds.get(k);
      CachedHashCodeObject condition = po.getCondition();
      ArrayList strs = getCondStrings(condition, paramNames);
      
      for (int s=0; s<strs.size(); s++) { 
        String str = (String) strs.get(s);
        goal += "                " +str + "\n";          
      }
    }
    return constants + "(setf (current-problem)\n"+
      "            (create-problem\n"+
      "              (name " + this.name + ")\n"+
      "              (objects \n" + objects + "             )\n"+
      "              (state (and " + state + "              ))\n" +
      "              (goal  (and " + goal  + "))))\n";
  }
  
// (setf (current-problem)
//       (create-problem
// ;;       (name foo)
//        (objects
//  	    (hopr1  OWL-HOProblem)
// 	    (hores1  OWL-HOATPResult)
// 	  )
//        (state
// 	    (and
// 	      (bound hopr1)
// 	    )
// 	  )   
//        (goal
// 	     (and
// 	       (bound hores1)
// 	        (resultFor hores1 hopr1)
// 	     )
// 	  )))

  public State getInitialState()
  {
    List inputs = problem.getInputs();
    List preconds= problem.getPreconds();
    List inputNames = new ArrayList();

    // keeps the literals of the pre and post of an action
    ArrayList ls = new ArrayList(); 
    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      
      Input e = (Input) inputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getSignature())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
      inputNames.add(e.getName().toLowerCase());
    }
     // add  literals for pre-conditions to List of literals
    for(int i=0; i<preconds.size(); i++){
      //      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(i);
      List l = pre.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }

    //  create an array out of the List of Literals 
    org.mathweb.reasoning.logic.Literal[] pre= new org.mathweb.reasoning.logic.Literal[ls.size()];
    Object[] ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      // the  PreCondition is overridden by a Literal!!!
      pre[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    State st = new State(pre);
    //    System.out.println("initial state: " + st);
    
    return st;
  }

  public State getGoalState()
  {
    List inputs = problem.getInputs();
    List inputNames = new ArrayList();
    List outputs = problem.getOutputs();
    List postconds= problem.getPostconds(); 
    ArrayList ls = new ArrayList();

    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      Input e = (Input) inputs.get(i);
      inputNames.add(e.getName().toLowerCase());
    }

    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getSignature())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
    }
    for(int i=0; i<postconds.size(); i++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      PostCondition po = (PostCondition) postconds.get(i);
      List l =po.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }
    org.mathweb.reasoning.logic.Literal[] post = new org.mathweb.reasoning.logic.Literal[ls.size()];
    Object[] ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      post[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    return new State(post);
  }  
}
