// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.trans;

import java.io.*;
import java.net.URI;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;
import org.mathweb.atp.TSTP2X;
import org.mathweb.onto.OntoReasoner;
//import org.mathweb.onto.JenaReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/24 16:23:37 $ $Revision: 1.3 $
*/
public class Flotter extends BasicService {
  
  public String FLOTTER_HOME = getSystemProperty("flotter.home", "FLOTTER_HOME");
  public String[] flotter_envs = new String[] {"FLOTTER_HOME="+FLOTTER_HOME};
  public String flotterExecutable = findSystemExecutable(FLOTTER_HOME, "spass", "SPASS.exe");
  public String dfg2tptpExecutable = findSystemExecutable(FLOTTER_HOME, "dfg2tptp", "DFG2TPTP.exe");
  private String logFileName = getSystemProperty("flotter.log.file", "FLOTTER_LOG_FILE");

  protected TSTP2X tstp2X = new TSTP2X();

  private static OWLJDom owlJDom;
  private static OntoReasoner reasoner;
  private static String sep = File.separator;
    
  private static final int defaultTime = 100;

  private static Logger logger = Logger.getLogger("Flotter");
  
  public Flotter()
  {
    super();
    owlJDom = OWLJDom.getInstance();
    reasoner = PellReasoner.getNewReasoner();

    logger.setLevel(Level.ALL);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }
  
  public String tstpFOF2CNF(String fofProblem)
  {
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();

    File inputFile = null;
    File outputFile = null;
    File tptpFile = null;
    String probID = null;
    boolean told = false; 

    try{
      reasoner.tellIndividual(fofProblem);
      told = true;
      
      Element fofRoot = owlJDom.getRootElement(fofProblem);
      String xmlBase = fofRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      
      logger.debug("[flotter] : " + fofRoot); 
      
      Element prob = (Element) owlJDom.TSTP_FOF_PROBLEM.selectSingleNode(fofRoot);
      if (prob == null) 
        return owlJDom.createFailure(probID, startTime, cpuStart, "The proving problem given to tstpCNF2CNF must be an individual of the concept TstpFOFProblem");
      
      probID = xmlBase + owlJDom.valueOf(OWLJDom.RDF_ID, prob);
      URI probURI = URI.create(probID);
      logger.debug("[flotter] prob ID: " + probID);
      String fofFormal = reasoner.valueOf("formalDescription", probURI);
      logger.debug("[flotter] FOF: " + fofFormal);
      if ((fofFormal == "") || (fofFormal == null))
        throw new Exception("FOF problem is of wrong type!");
      
    // first we use tptp2X to generate DFG syntax
      String dfgProb = tstp2X.tstp2X(fofFormal, 
                                     "dfg",
                                     "rm_equality:rstfp");
      String logic = reasoner.valueOf("inLogic", probURI);
      inputFile = File.createTempFile("flotter", ".in");
      outputFile = File.createTempFile("flotter", ".out");
      tptpFile = File.createTempFile("flotter", ".tptp");
      
      // now we call flotter on the DFG file
      writeFile(inputFile, dfgProb);
      String command = flotterExecutable + " -PGiven=0 -PProblem=0 -Flotter " + 
        inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath();
      logger.debug("[flotter] command: " + command);
      String[] commands = {"sh", "-c", command};
      logger.debug("[flotter] starting translation...");
      this.exec(commands, flotter_envs, false, true);
      
      // now use tptp2X again translate DFG to TPTP
      command = dfg2tptpExecutable + " " + outputFile.getAbsolutePath() +
        " " + tptpFile.getAbsolutePath();
      logger.debug("[flotter] command: " + command);
      commands = new String[] {"sh", "-c", command};
      logger.debug("[flotter] starting translation...");
      this.exec(commands, flotter_envs, false, true);
      logger.debug("reading result..." + tptpFile.getAbsolutePath());
      String result = this.readFile(tptpFile);
      String timeRes = reasoner.valueOf("timeRes", probURI);
      int time;
      logger.debug(timeRes);
      if (timeRes == null) 
        time = this.defaultTime;
      else {
        String timeStr = reasoner.valueOf("wallClockTime", timeRes);
        time =  Integer.parseInt(timeStr);
      }
      
      return createCNFProblem(result, probURI, time, startTime, cpuStart,  logic);
      
    } catch(Exception e){
      logger.debug("[flotter] Exception " + e);
      return owlJDom.createFailure(probID, startTime, cpuStart,"A problem occured during the Flotter tstpFOF2CNF transformation\n" + e);
    } finally{
      //      if ((told) && (probID !=null))  reasoner.untellIndividual(probID);

      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
        if (tptpFile != null) tptpFile.delete();
      } catch(Exception e){ logger.debug("[flotter] Exception " + e); }
    }
  }
  
  private String createCNFProblem(String output, URI probURI, int timeRes, long startTime, long cpuStart, String logic){
    try{
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", OWLJDom.MW_NS_STR, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element prob = new Element("TstpCNFProblem",OWLJDom.MW_NS);
      String resultID = "flotter_CNF_for_" +probURI.getFragment() +System.currentTimeMillis();
      prob.setAttribute("ID", resultID, OWLJDom.RDF_NS);
      Element format = new Element("format",OWLJDom.MW_NS);
      format.setAttribute("resource",  OWLJDom.MW_NS_STR+"TSTP", OWLJDom.RDF_NS);
      Element formalProb = new Element("formalDescription",OWLJDom.MW_NS);
      formalProb.addContent(output);
      Element probLogic = new Element("inLogic",OWLJDom.MW_NS); 
      probLogic.setAttribute("resource", logic, OWLJDom.RDF_NS);
      Element satEquiv = new Element("satEquivalentTo", OWLJDom.MW_NS);
      satEquiv.setAttribute("resource", probURI.toString(), OWLJDom.RDF_NS);

      Element time = new Element("timeRes",OWLJDom.MW_NS);
      Element timeres = new  Element("TimeResource",OWLJDom.MW_NS);
      Element wcTime = new  Element("wallClockTime",OWLJDom.MW_NS);
      Integer iTime = new Integer(timeRes);
      wcTime.addContent(iTime.toString());
      wcTime.setAttribute("datatype", OWLJDom.RDFS_NS_STR+"positiveInteger", OWLJDom.RDF_NS);
      timeres.addContent(wcTime);
      time.addContent(timeres);

      Element timeUsed = OWLJDom.createTimeResource(startTime, cpuStart);

      prob.addContent(formalProb);
      prob.addContent(probLogic);
      prob.addContent(satEquiv);
      prob.addContent(time);

      rdf.addContent(prob);
      rdf.addContent(timeUsed);

      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating the OWL answer: " + e.getMessage();
    }
  }


  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Flotter e = new Flotter();
      String servF1 = msHome + "/examples/owl/fof2.owl";
      
      String ind = e.readFile(servF1);
      System.out.println("problem: " + ind);
        
      try {
        String result = e.tstpFOF2CNF(ind);
        logger.info("result2: " + result);
      }
      catch(Exception ex){
        ex.printStackTrace();
      }
      
  }

} // end Flotter.java


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
