// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.URI;

// import javax.xml.rpc.ServiceException;
// import javax.xml.rpc.server.ServiceLifecycle;
// import javax.xml.rpc.server.ServletEndpointContext;

import javax.xml.transform.Source;
import javax.xml.transform.URIResolver;
import javax.xml.transform.Transformer;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;
import org.mathweb.atp.TSTP2X;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.Omega;
import org.mathweb.util.OmdocCreator;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.4 $
*/
public class HoSetDefsElim extends BasicService //implements ServiceLifecycle 
{
  private static Logger logger = Logger.getLogger("HoSetDefsElim");
  private static String logFileName = getSystemProperty("hosetdefselim.log.file", "HOSETDEFSELIM_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  public String omdoc2postXslFile = getSystemProperty("omdoc.post.xsl", "OMDOC_POST_XSL");
  public String xslHomeDir = getSystemProperty("omdoc.xsl.home", "XSL_HOME");

  /** TSTP trafo stuff */
  private static TSTP2X tstp2X = new TSTP2X();
  
  /** Ontology stuff */
  private static OWLJDom owlJDom;
  private static OntoReasoner reasoner;
  
  private static Omega theInstance = new Omega("HoDefsElim").startOmega(); 

  private static String proofConcept = "TstpCnfBrFPRefutation";
  private static String servXmlBase = "http://www.mathweb.org/trans/HoSetDefsElim#";
  private String cnfID;
  
  private boolean busy;
  //  public ServletEndpointContext endpointContext;

  private static final int defaultTime = 100;
  
  private Transformer transformer;
  private int counter = 0;

  public HoSetDefsElim()
  {
    mkTempDir("hoSetDefsElimService");
    
    owlJDom = OWLJDom.getInstance();
    //    reasoner = PellReasoner.getNewReasoner();
    
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    logger.removeAllAppenders();
    logger.addAppender(appender);
    
    busy = false;
  }
  
 
  public void destroy()
  {
    logger.debug("destroy() method being called!");
  }

  class MyResolver implements URIResolver
  {
    public Source resolve(String href, String base) {
      System.out.println("href: " + href);
      System.out.println("base: " + base);
      File xslFile = new File(xslHomeDir + "/" + href);
      return new StreamSource(xslFile);
    }
  }
  

  private synchronized void busy(boolean b)   { this.busy = b; }
    
  /**
   * Eliminates all definitions in a proving problem in the theory http://www.ags.uni.sb.de/~omega/theories#typed-set.thy.
   * The resulting problem is in theory  http://www.ags.uni.sb.de/~omega/theories#base.thy
   * 
   * @param postProblem A string containing the OMDoc problem, i.e. an individual of the OWL class OmdocProvingProblem.
   *   
   * @return An OmdocProvingProblem resulting from the elimination of all definitions.
   */
  public synchronized String eliminateDefs(String omdocProblem) throws TransformerException
  {
    logger.debug("[hoSetDefsElim] : " + omdocProblem); 
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    
    if (busy) return owlJDom.createFailure(null, startTime,  cpuStart,"Sorry, Omega is too busy right now! Call again later.");
    //    if (busy) return new String[] {owlJDom.createFailure(probID, startTime,  cpuStart,"Sorry, Tramp is too busy right now! Call again later."), null };
    while(!theInstance.isReady()){
      if ((System.currentTimeMillis() - startTime) > 20000)
        break;
    }
    if (!theInstance.isReady()) 
      return  owlJDom.createFailure(null, startTime, cpuStart, "Sorry, Omega is not ready yet! Call again later.");
    
    String probID = null;
    boolean told = false;
    try{
      busy(true);
      reasoner = PellReasoner.getNewReasoner();
      reasoner.tellIndividual(omdocProblem);
      told = true;
      Element omdocRoot = owlJDom.getRootElement(omdocProblem);
      String xmlBase = omdocRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      
      logger.debug("[hoSetDefsElim] : " + omdocRoot); 
      if (owlJDom.hasChild(omdocRoot, "OmdocProvingProblem",  OWLJDom.MW_NS))        {
        Element prob = (Element) owlJDom.OMDOC_PROBLEM.selectSingleNode(omdocRoot);
        if (prob == null) 
          return owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to eliminateDefs must be an individual of the concept OmdocProvingProblem");
        
        probID = xmlBase + owlJDom.valueOf(OWLJDom.RDF_ID, prob);
        logger.debug("[hoSetDefsElim] prob ID: " + probID);
        URI probURI = URI.create(probID);
        String omdocFormal = reasoner.valueOf("formalDescription", probURI);
        logger.debug("[hoSetDefsElim] FOF: " + omdocFormal);
        if ((omdocFormal == "") || (omdocFormal == null))
          throw new Exception("Omdoc problem is of wrong type!");
        
        Element formalRoot = owlJDom.getRootElement(omdocFormal);

        writeFile(msHome + "/etc/data/omdoc/HoSet-omdoc"+(++counter), omdocFormal);

        StringWriter writer = new StringWriter();
        
        System.out.println("omdoc: \n" + omdocFormal);
        System.out.println("omdoc root: \n" + formalRoot);

        String omdoc2postXslStr = readFile(omdoc2postXslFile);
        StringReader omdoc2postXsl = new StringReader(omdoc2postXslStr);
        StreamSource xsltStream = new StreamSource(omdoc2postXsl);
        
        TransformerFactory tFactory = TransformerFactory.newInstance();
        tFactory.setURIResolver(new MyResolver());
        transformer = tFactory.newTransformer(new StreamSource(omdoc2postXsl));
        
        System.out.println("trafo: " + transformer.getClass().getName());
        //transformer = tFactory.newTransformer(xsltSource);
        
        transformer.transform(new StreamSource(new StringReader(omdocFormal)), 
                              new StreamResult(writer));
        writeFile(msHome + "/etc/data/omdoc/HoSet-post"+counter, writer.toString());
        
        String postProb = writer.toString();
        System.out.println("post: \n" + postProb);

        String result = theInstance.callHoSetDefsElim(postProb);
        
        String logic = "fooLogic"; //reasoner.valueOf("inLogic", probURI);

        String finalRes = createResult(result, probURI, startTime, cpuStart);
//         logger.debug("final result : " +finalRes);
        return finalRes;
      } else{
//         return  new String[] {owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstp2Cnf must be an individual of the concept TstpFOFProblem"), null  };
        return  owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstp2Cnf must be an individual of the concept TstpFOFProblem");
        
      }
    } catch(Exception e){
      e.printStackTrace();
      logger.debug("[hoSetDefsElim] " +e);
      return  owlJDom.createFailure(probID, startTime, cpuStart, "Some error occurred: " + e.getMessage());
//       return  new String[] {owlJDom.createFailure(probID, startTime, cpuStart, "Some error occurred: " + e.getMessage()), null
//       };
    }
    finally {
      if ((told) && (probID !=null)){
        logger.debug("untelling indi");
        reasoner.untellIndividual(probID);
      }
      busy(false);
    }
  }

  private String createResult(String cnfOut, URI probURI, long startTime, long cpuStart){
    try{
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", servXmlBase, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element prob = new Element("TstpCNFProblem",OWLJDom.MW_NS);
      cnfID = "omega_FOF_for_" +probURI.getFragment() +System.currentTimeMillis();
      prob.setAttribute("ID", cnfID, OWLJDom.RDF_NS);
      Element format = new Element("format",OWLJDom.MW_NS);
      format.setAttribute("resource",  OWLJDom.MW_NS_STR+"TSTP", OWLJDom.RDF_NS);
      Element formalProb = new Element("formalDescription",OWLJDom.MW_NS);
      formalProb.addContent(cnfOut);
      Element probLogic = new Element("inLogic",OWLJDom.MW_NS); 
      //      probLogic.setAttribute("resource", logic, OWLJDom.RDF_NS);
      Element satEquiv = new Element("satEquivalentTo", OWLJDom.MW_NS);
      satEquiv.setAttribute("resource", probURI.toString(), OWLJDom.RDF_NS);

//       Element time = new Element("timeRes",OWLJDom.MW_NS);
//       Element timeres = new  Element("TimeResource",OWLJDom.MW_NS);
//       Element wcTime = new  Element("wallClockTime",OWLJDom.MW_NS);
//       Element cpuTime = new  Element("cpuTime",OWLJDom.MW_NS);
//       Integer iTime = new Integer(timeRes + 1);
//       Integer jTime = new Integer(timeRes);
//       wcTime.addContent(iTime.toString());
//       wcTime.setAttribute("datatype", OWLJDom.RDFS_NS_STR+"integer", OWLJDom.RDF_NS);
//       cpuTime.addContent(jTime.toString());
//       cpuTime.setAttribute("datatype", OWLJDom.RDFS_NS_STR+"integer", OWLJDom.RDF_NS);
//       timeres.addContent(wcTime);
//       timeres.addContent(cpuTime);
//       time.addContent(timeres);
      
      prob.addContent(formalProb);
      prob.addContent(probLogic);
      prob.addContent(satEquiv);
      //      prob.addContent(time);

      rdf.addContent(prob);

      Element timeUsed = OWLJDom.createTimeResource(startTime, cpuStart);
      rdf.addContent(timeUsed);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating the OWL answer: " + e.getMessage();
    }
  }

  public void waitMillis(long millis)
  {
    long startTime = System.currentTimeMillis();
    while ((System.currentTimeMillis() - startTime) < millis);
  }

  public void shutdown()
  {
    if (theInstance != null)
      theInstance.shutdown();
    removeTempDir();
  }
  
  public void test () {
    try {
      String setProb = readFile(msHome + "/thirdParty/omega/xmp/test-bool-prop-set88.post");
      String result = null;
//       while(!theInstance.isReady())
//         System.out.println("not ready");
      
      System.out.println("HoSetDefsElim POST prob: "+ setProb);
      result = theInstance.callHoSetDefsElim(setProb);
      System.out.println(result);
      //       String[] cnf;
      //       cnf = tstpFOF2CNF(fofProb);
//       writeFile(msHome + "/examples/owl/cnf2.owl", cnf[0]);
//       writeFile(msHome + "/examples/owl/delta2.owl", cnf[1]);
//       logger.debug("[HoSetDefsElim] generated cnf: " + cnf[0]);
//       cnf = tstpFOF2CNF(fofProb);
//       logger.debug("[HoSetDefsElim] 2nd generated cnf: " + cnf[0]);
//       String cnf;
//       cnf = tstpFOF2CNF(fofProb);
//       writeFile(msHome + "/examples/owl/cnf2.owl", cnf);
//       logger.debug("[HoSetDefsElim] generated cnf: " + cnf);
//       cnf = tstpFOF2CNF(fofProb);
//       logger.debug("[HoSetDefsElim] 2nd generated cnf: " + cnf);
     }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    HoSetDefsElim serv = new HoSetDefsElim();
    OmdocCreator creator = new OmdocCreator();
    String omdocFile = "/project/home-nb/jzimmer/import/mathserve/thirdParty/omega/xmp/set/BOOL-PROP-SET88-1.omdoc";
    String owlProb = creator.omdocFile2OWLProblem(omdocFile);
    String result = serv.eliminateDefs(owlProb);
    
    serv.shutdown();
    System.exit(0);
  }
  
} // end HoSetDefsElim.java

     
 
