// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;

import org.mathweb.atp.TSTP2X;
import org.mathweb.atp.Otter;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
//import org.mathweb.onto.OWLReasoner;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.7 $
*/
public class Prex extends BasicService {

  static boolean verbose=true;
    
  /** Lisp executables and environments */
  public String ACL_HOME = getSystemProperty("acl.home", "ACL_HOME");
  public String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME};
  public String aclExecutable = findSystemExecutable(ACL_HOME, "lisp", "alisp");
  
  public String prexDump = getSystemProperty("prex.dump", "PREX_DUMP");
  public String prexExecutable = getSystemProperty("prex.exec", "PREX_EXEC");
  public String prexXtermExecutable = getSystemProperty("prex.exec.xterm", "PREX_EXEC_XTERM");
  public String prexEmacs = getSystemProperty("prex.emacs.exec", "PREX_EMACS_EXEC");
  public String prexIlisp = getSystemProperty("prex.ilisp", "PREX_ILISP");
  public String emacsGeometry = getSystemProperty("prex.emacs.geometry", "PREX_EMACS_GEOMETRY");

  /** TSTP trafo stuff */
  private static TSTP2X tstp2X = new TSTP2X();
  
  /** Ontology stuff */
  private static OWLJDom owlJDom;
  private static OntoReasoner reasoner;
  
  /** For communication with Lisp process */
  private static String sep = File.separator;
  private int serverPort = 0;
  private int defaultPort = 12345;
  private CommandThread lispThread;
  
  private WebServer xmlRpcServer;
  private String lispServerURL = "";
  private org.apache.xmlrpc.XmlRpcClient lispServerConnection;
  
  public static final int defaultNice = 15;
  public Throwable exception;
  public boolean started = false;
  public boolean ready = false;
  public boolean busy = false;
  private int cnfCounter = 1;
  private int ndCounter = 1;
  
  public Prex()
  {
    super();
    mkTempDir("prex");
    owlJDom = OWLJDom.getInstance();
    reasoner = PellReasoner.getNewReasoner();
  }
  
  /**
   * Create an ND proof for a TSTP FOF problem given a FO-ATP result.
   * 
   * @param fofProblem A string containing an OWL TstpFofProblem.
   * @param foAtpResult A string containing the OWL individual with a .
   * @return If successful, an ND proof for the FOF problem.
   */
  public String nlForND(String ndProof)
  {
    String fofID = null;
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();

    try{
      String fofProblem=ndProof;
      
      String foAtpResult=ndProof;
      
      reasoner.tellIndividual(fofProblem);

      Element fofRoot = owlJDom.getRootElement(fofProblem);
      Element resRoot = owlJDom.getRootElement(foAtpResult);

      System.out.println("[prexND] : " + fofRoot + " " + resRoot + OWLJDom.MW_NS); 

      if ((owlJDom.hasChild(fofRoot, "TstpFOFProblem",  OWLJDom.MW_NS)) &&
          (owlJDom.hasChild(resRoot, "FoAtpResult",  OWLJDom.MW_NS)))
        {
          fofID = owlJDom.valueOf(OWLJDom.RDF_ID, fofRoot);
          String fofFormal = owlJDom.valueOf(OWLJDom.PROVING_PROBLEM_FORMAL_DESCRIPTION,fofRoot);
//           System.out.println("[prex] prob ID: " + fofID);
//           System.out.println("[prex] tstp: " + fofFormal);
          
          String fofPost = tstp2X.tstp2Post(fofFormal);
          //           System.out.println("[prex] POST: " + fofPost);
          Element proof = (Element) owlJDom.getSingleNode(OWLJDom.FO_ATP_RESULT_PROOF, resRoot);
          String actualProof = owlJDom.valueOf(OWLJDom.PROOF_FORMAL_PROOF, proof);
          //           System.out.println("[prex] TSTP: " + actualProof);
          
          String nlProof = callPrexFof2ND(fofPost, actualProof);
          busy = false;
          return createNDProof(nlProof, fofID);
        }else{
          busy = false;
          return owlJDom.createFailure(fofID, startTime, cpuStart, "The proving problem given must be an individual of the concept TstpFofProblem");
        }
    } catch(Exception e) {
      e.printStackTrace();
      busy = false;
      return owlJDom.createFailure(fofID, startTime, cpuStart, e.getMessage());
    } finally {
      busy = false;
    }
  }
  
  private String callPrexFof2ND(String postProblem, String tstpProof) throws Exception
  {
    try{
      if (busy) return "Sorry, Prex is too busy right now! Call again later.";
      busy = true;
      String tempDirPath = tempDir.getAbsolutePath();
      String postFileName = tempDirPath + sep + "prex"+ ndCounter + ".post";
      String tstpFileName = tempDirPath + sep + "prextstp"+ ndCounter + ".tstp";
      String outputFileName = tempDirPath + sep + "prex"+ ndCounter + ".twega"; 
      ndCounter++;
      writeFile(postFileName,postProblem);
      //writeFile(tstpFileName,"\"" + tstpProof + "\"");
      writeFile(tstpFileName,tstpProof);
      Vector params = new Vector();
      params.addElement(postFileName);
      callLispCommand("read-problem", params);
      params = new Vector();
      params.addElement(tstpFileName);
      String nd = (String) callLispCommand("transform-tstp-proof-from-file", params);
      System.out.println("Prex ND: " + nd);
      busy = false; 
      return nd;
    } catch (XmlRpcException e) {
      System.out.println("Prex error " + e.code + ":" + e.getMessage());
      throw new Exception("Prex error (code " + e.code + "): " + e.getMessage());
    }
  }

  private String createNDProof(String output, String probID){
    try{
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      Element proof = new Element("TwegaNDProof",OWLJDom.MW_NS);
      Element format = new Element("format",OWLJDom.MW_NS);
      format.setAttribute("resource", "Twega", OWLJDom.RDF_NS);
      Element proofOf = new Element("proofOf",OWLJDom.MW_NS);
      proofOf.setAttribute("resource", probID, OWLJDom.RDF_NS);
      Element formalProof = new Element("formalProof",OWLJDom.MW_NS);
      formalProof.addContent(output);
      Element proofLogic = new Element("proofLogic",OWLJDom.MW_NS);
      proofLogic.setAttribute("resource", "fol_plus_equ", OWLJDom.RDF_NS);

      proof.addContent(proofOf);
      proof.addContent(proofLogic);
      proof.addContent(format);
      proof.addContent(formalProof);
      rdf.addContent(proof);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating the OWL answer: " + e.getMessage();
    }
  }
  
  
  
  /** The XML-RPC server for the lisp process */
  private class DumbHandler implements XmlRpcHandler {
    public Object execute ( String methodName, Vector params ) 
    {
      try{ 
        System.out.println("RPC call method: " + methodName);
        System.out.println("args: " + params);
        for (int i=0; i<params.size(); i++){
          System.out.println("args " + params.get(i) + " " + params.get(i).getClass());
        }
        if (methodName.equals("initLispServer")) {
          lispServerURL = (String) params.get(0);
          System.out.println("lisp server: " + lispServerURL);
          lispServerConnection =  new org.apache.xmlrpc.XmlRpcClient(lispServerURL);
          return "T";
        }
        else if (methodName.equals("ready")) {
          System.out.println("lisp server says that it is ready!");
          ready = true;
          return "T";
        }
      } catch (MalformedURLException e){
        System.out.println("Trouble with RPC call " + e);
        e.printStackTrace();
      } catch(Exception e){
        e.printStackTrace();
      }
      return null;
    }
  }
  
  private void startXmlRpcServer() throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = defaultPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        System.out.println("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ( "$default", new DumbHandler());
        success = true;
      }catch(IOException e){
        e.printStackTrace();
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        System.out.println("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  public void startPrexEmacs(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String ilisp = readFile(prexIlisp);
        String tempDirPath = tempDir.getAbsolutePath();
        String inputFileName = tempDirPath + sep + "prex.el";
        String emacsStartup = ilisp + "\n" +
          "(setq loui-program " + 
          "\"" + prexExecutable + " " + aclExecutable + " " + prexDump + " " + serverURL + "\") ";
        writeFile(inputFileName, emacsStartup);
        
        if (serverPort != 0) {
          String command2 = prexEmacs + " " +
            "-bg white -fg black -debug-init -display localhost:0.0 -name \"Prex Service (Emacs)\" " +
            "-geometry " + emacsGeometry + " -font 8x13" +
            " -l " + inputFileName + " -f loui";
          System.out.println("prex command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          started = true;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Prex:" +e);
      }
    }
  }

  public void startPrexXterm(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String tempDirPath = tempDir.getAbsolutePath();
        
        if (serverPort != 0) {
          String command2 =  prexXtermExecutable + " " + aclExecutable + " " + prexDump + " " + serverURL + " localhost:0.0 ";
          System.out.println("prex command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          started = true;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Prex:" +e);
      }
    }
  }

  public synchronized void startPrex(){
      if (!started){
        started = true;
        try {
          startXmlRpcServer();
          String serverURL = "http://localhost:" + serverPort;
          String tempDirPath = tempDir.getAbsolutePath();
          String outputFileName = tempDirPath + sep + "prex" + ".out";
          
          if (serverPort != 0) {
            String command2 = aclExecutable + " -I " + prexDump + 
              " -qq -e \"(progn (ags::rpc~start-server '(\\\"" + serverURL + 
              "\\\" \\\"initLispServer\\\" \\\"foo\\\")) " +
              "(setf excl::*restart-app-function* (lambda () (omega::prex~top))))\" > " + outputFileName + " 2>&1";
            System.out.println("prex command: " + command2);
            lispThread = runThreaded(command2, acl_envs);
          }
        }
        catch(ClassNotFoundException e) {
          System.err.println("Something went wrong when starting Prex:" +e);
        }
      }
  }
 
  public boolean ready()
  {
    return this.ready;
  }
  
  
  public Object callLispCommand(String command, Vector params) throws XmlRpcException
  {
    try{
      Object result = lispServerConnection.execute(command, params);
      //      System.out.println("Result: " + result + " " + result.getClass());
      return result;
    } catch (MalformedURLException e){
      System.out.println("Trouble with RPC call " + e);
      e.printStackTrace();
    } catch( IOException e) {
      System.out.println("Trouble with RPC call " + e);
      e.printStackTrace();
    }
    return null;
      
  }
  
  
  public void shutdown()
  {
    try {
      Vector params = new Vector();
      callLispCommand("shutdown", params);
      waitMillis(3000);
      xmlRpcServer.removeHandler("$default");
      xmlRpcServer.shutdown();
      lispThread.shutdown();
      removeTempDir();
    } catch (XmlRpcException e) {
      System.out.println("Trouble with RPC call " + e);
      e.printStackTrace();
    }
  }
  
  public void waitMillis(long millis)
  {
    long t1 = System.currentTimeMillis();
    while ((System.currentTimeMillis() - t1) < millis);
  }
  
  public void test () {
    try {
      String fofProb = readFile(msHome + "/examples/owl/fof.owl");
      String tstpProof = readFile(msHome + "/examples/owl/otterresult.owl");
      //      System.out.println("[Prex]  " + fofProb + tstpProof); 
      

            //       Vector params = new Vector();
      //       params.addElement(probFile);
      //       callLispCommand("read-problem", params);
      //       params = new Vector();
      //       Object result = callLispCommand("clausify", params);
      //       String prob = (String) result;
      //        String delta = (String) result.get(1);
      //OtterService o = new OtterService();
      //String n = o.callProver(prob, "TSTP", 300, "");
      //      System.out.println("[Prex] result " + prob); 
      //      System.out.println("[Prex] final result1: " + nd);
//       nd = ndForFOF(fofProb, tstpProof);
//       System.out.println("[Prex] final result2: " + nd);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    Prex e = new Prex();
    e.startPrexEmacs();
    //e.startPrexXterm();
    //e.startPrex();
    while (! e.ready()) {
      e.waitMillis(1000);
    }
    e.test();
    
  }
  
} // end PrexAgent.java

      
