// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;

import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.TrampXmlRpcHandler;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.22 $
*/
public class Tramp extends BasicService {

  /** Lisp executables and environments */
  private String ACL_HOME = getSystemProperty("acl.home", "ACL_HOME");
  private String aclExecutable = findSystemExecutable(ACL_HOME, "lisp", "alisp");
  
  private String trampDump = getSystemProperty("tramp.dump", "TRAMP_DUMP");
  private String trampExecutable = getSystemProperty("tramp.exec", "TRAMP_EXEC");
  private String trampXtermExecutable = getSystemProperty("tramp.exec.xterm", "TRAMP_EXEC_XTERM");
  private String trampEmacs = getSystemProperty("tramp.emacs.exec", "TRAMP_EMACS_EXEC");
  private String trampIlisp = getSystemProperty("tramp.ilisp", "TRAMP_ILISP");
  private String DISPLAY = getSystemProperty("tramp.display", "TRAMP_DISPLAY");
  private String emacsGeometry = getSystemProperty("tramp.emacs.geometry", "TRAMP_EMACS_GEOMETRY");
  private String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME, "DISPLAY="+DISPLAY};


  private static Logger logger = Logger.getLogger("Tramp");
  private static String logFileName = getSystemProperty("tramp.log.file", "TRAMP_LOG_FILE");

  private static String lispLogFileName = getSystemProperty("acl.log.file", "ACL_LOG_FILE");

  /** For communication with Lisp process */
  private static String sep = File.separator;
  private int serverPort = 0;
  private int defaultPort = 12345;
  private CommandThread lispThread;
  
  private WebServer xmlRpcServer;
  private String lispServerURL = "";
  private org.apache.xmlrpc.XmlRpcClient lispServerConnection;
  
  private static final int defaultNice = 15;
  private Throwable exception;
  private boolean started = false;
  private boolean ready = false;
  private int cnfCounter = 1;
  private int ndCounter = 1;
  private static int logfileCounter = 1;
  private static int lisplogfileCounter = 1;
  private String user = "foo";
  
  public Tramp(String usr)
  {
    mkTempDir("tramp");
    //    owlJDom = OWLJDom.getInstance();

    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    FileAppender appender = newAppender(logFileName + usr + (logfileCounter++) + ".log");
    logger.removeAllAppenders();
    logger.addAppender(appender);

    xmlRpcServer = null;
    lispServerConnection = null;
    started = false;
    ready = false;
    user = usr;
    //    startTramp();
  }
  
  
  public String[] callTrampClausifier(String postProb) throws Exception {
    String cnfProb = null;
    File probFile = null;
    try{
      String tempDirPath = tempDir.getAbsolutePath();
      String probFileName = tempDirPath + sep + "problem" + cnfCounter + ".post";
      cnfCounter ++;
      probFile = new File(probFileName);
      writeFile(probFile, postProb);
      Vector params = new Vector();
      params.addElement(probFileName);
      
      callLispCommand("read-problem", params);
      params = new Vector();
      Object result = callLispCommand("clausify", params);
      System.out.println("result: " + result + result.getClass().getName());
      Vector resVec = (Vector) result;
      
      cnfProb = (String) resVec.get(0);
      String delta = (String) resVec.get(1);
      logger.debug("[Tramp] result " + cnfProb); 
      //        logger.debug("[Tramp] delta relation: " + delta);
      return new String[] {cnfProb, delta};
    } catch (XmlRpcException e) {
      logger.debug("Tramp error " + e.code + ":" + e.getMessage());
      throw new Exception("Tramp error (code " + e.code + "): " + e.getMessage());
    } finally{
      try{ 
        if (probFile != null) probFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
  }
  
  public String callTrampFof2ND(String postProblem, String tstpProof) throws Exception
  {
    File postFile = null;
    File tstpFile = null;
    File outputFile = null;
    try{
      String tempDirPath = tempDir.getAbsolutePath();
      String postFileName = tempDirPath + sep + "tramp"+ ndCounter + ".post";
      String tstpFileName = tempDirPath + sep + "tramptstp"+ ndCounter + ".tstp";
      String outputFileName = tempDirPath + sep + "tramp"+ ndCounter + ".twega"; 
      ndCounter++;
      postFile = new File(postFileName);
      tstpFile = new File(tstpFileName);
      outputFile = new File(outputFileName);
      writeFile(postFileName,postProblem);
      writeFile(tstpFileName,tstpProof);

      // first read in the POST problem:
      Vector params = new Vector();
      params.addElement(postFileName);
      callLispCommand("read-problem", params);

      // then call the proof transformation command:
      params = new Vector();
      params.addElement(tstpFileName);
      String nd = (String) callLispCommand("transform-tstp-proof-from-file", params);
      logger.debug("Tramp ND: " + nd);
      callLispCommand("show-pds", params);
      return nd;
    } catch (XmlRpcException e) {
      logger.debug("Tramp error " + e.code + ":" + e.getMessage());
      throw new Exception("Tramp error (code " + e.code + "): " + e.getMessage());
    } finally{
      try{ 
        if (postFile != null) postFile.delete();
        if (tstpFile != null) tstpFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
  }
  
  
  private void startXmlRpcServer() throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = defaultPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        logger.debug("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ( "$default", new TrampXmlRpcHandler(this));
        success = true;
      }catch(IOException e){
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        logger.debug("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  public Tramp startTrampEmacs(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String ilisp = readFile(trampIlisp);
        String tempDirPath = tempDir.getAbsolutePath();
        String inputFileName = tempDirPath + sep + "tramp.el";
        String emacsStartup = ilisp + "\n" +
          "(setq loui-program " + 
          "\"" + trampExecutable + " " + aclExecutable + " " + trampDump + " " + serverURL + "\") ";
        writeFile(inputFileName, emacsStartup);
        
        if (serverPort != 0) {
          String command2 = trampEmacs + " " +
            "-bg white -fg black -debug-init -display " + DISPLAY + " -name \"Tramp Service (Emacs)\" " +
            "-geometry " + emacsGeometry + " -font 8x13" +
            " -l " + inputFileName + " -f loui";
          logger.debug("tramp command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          logger.debug("after execution");
          started= true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Tramp:" +e);
        logger.fatal("Something went wrong when starting Tramp:" +e);
      }
    }
    return this;
  }

  public Tramp startTrampXterm(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String tempDirPath = tempDir.getAbsolutePath();
        
        if (serverPort != 0) {
          String command2 =  trampXtermExecutable + " " + aclExecutable + " " + trampDump + " " + serverURL + " localhost:0.0 ";
          logger.debug("tramp command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          started = true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Tramp:" +e);
        logger.fatal("Something went wrong when starting Tramp:" +e);
      }
    }
    return this;
  }

  public Tramp startTramp() {
      if (!started){
        try {
          startXmlRpcServer();
          String serverURL = "http://localhost:" + serverPort;
          String tempDirPath = tempDir.getAbsolutePath();
          String outputFileName = tempDirPath + sep + "tramp" + ".out";
          
          if (serverPort != 0) {
            String command2 = aclExecutable + " -I " + trampDump + 
              " -qq -e \"(progn (ags::rpc~start-server '(\\\"" + serverURL + 
              "\\\" \\\"initLispServer\\\" \\\""+ user + "\\\")) " +
              "(setf excl::*restart-app-function* (lambda () (omega::tramp~top))))\" > " + lispLogFileName + user + (lisplogfileCounter++) + ".log 2>&1";
            logger.debug("tramp command: " + command2);
            lispThread = runThreaded(command2, acl_envs);
          }
          started = true;
          return this;
        }
        catch(Exception e) {
          System.err.println("Something went wrong when starting Tramp:" +e);
          logger.fatal("Something went wrong when starting Tramp:" +e);
        }
      }
      return this;
  }
 
  public boolean isStarted() { return this.started; }
  public boolean isReady() { return this.ready; }
  public void setReady() { this.ready = true; }
  
  public void setLispServerConnection(String url)
  {
    try{
      lispServerURL = url;
      lispServerConnection = new org.apache.xmlrpc.XmlRpcClient(url);
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
  }
  

  public Object callLispCommand(String command, Vector params) throws XmlRpcException
  {
    try{
      Object result = lispServerConnection.execute(command, params);
      //      logger.debug("Result: " + result + " " + result.getClass());
      return result;
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    } catch(IOException e) {
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
    return null;
  }
  
  public void shutdown()
  {
    try { 
      Vector params = new Vector();
      if (lispServerConnection != null) {
        logger.debug("shutting down lisp process!");
        callLispCommand("shutdown", params);
        logger.debug("waiting 4 secs!");
        waitMillis(4000);
        logger.debug("waited 4 secs for lisp process!");
        xmlRpcServer.removeHandler("$default");
        xmlRpcServer.shutdown();
        lispThread.shutdown();
        logger.debug("lisp process terminated!");
      }
    } catch (XmlRpcException e) {
      logger.debug("Trouble with RPC call " + e);
      e.printStackTrace();
    } finally {
      removeTempDir();
    }
  }
  
  
  public void waitMillis(long millis)
  {
    long t1 = System.currentTimeMillis();
    while ((System.currentTimeMillis() - t1) < millis);
  }
  
  public void test () {
    try {
      //      String fofProb = readFile(msHome + "/examples/owl/fof2.owl");
//       String[] cnf;
//       cnf = callTrampClausifier(fofProb);
//       logger.debug("[Tramp] generated cnf: " + cnf);
      
//       nd = ndForFOF(fofProb, tstpProof);
//       logger.debug("[Tramp] generated ND proof: " + nd);
      
      //       Vector params = new Vector();
      //       params.addElement(probFile);
      //       callLispCommand("read-problem", params);
      //       params = new Vector();
      //       Object result = callLispCommand("clausify", params);
      //       String prob = (String) result;
      //        String delta = (String) result.get(1);
      //OtterService o = new OtterService();
      //String n = o.callProver(prob, "TSTP", 300, "");
      //      logger.debug("[Tramp] result " + prob); 
//       logger.debug("[Tramp] final result2: " + nd);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    Tramp e = new Tramp("test");
    //    e.startTrampEmacs();
    //e.startTrampXterm();
    //    e.startTramp();
    while (! e.isReady()) {
      System.out.println("not ready");
      e.waitMillis(1000);
    }
    e.test();
    e.shutdown();
    System.exit(0);
        
  }
  
} // end Tramp.java

      
