// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.URI;

import javax.xml.rpc.ServiceException;
import javax.xml.rpc.server.ServiceLifecycle;
import javax.xml.rpc.server.ServletEndpointContext;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;
import org.mathweb.atp.TSTP2X;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.Tramp;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/06/24 19:11:55 $ $Revision: 1.24 $
*/
public class TrampCNFService extends BasicService //implements ServiceLifecycle 
{
  private static Logger logger = Logger.getLogger("TrampCNFService");
  private static String logFileName = getSystemProperty("tramp.cnf.log.file", "OTTER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  /** TSTP trafo stuff */
  private static TSTP2X tstp2X = new TSTP2X();
  
  /** Ontology stuff */
  private static OWLJDom owlJDom;
  private static OntoReasoner reasoner;
  
  private static Tramp theInstance = new Tramp("CNF");//(new Tramp("CNF")).startTramp();

  private static String proofConcept = "TstpCnfBrFPRefutation";
  private static String servXmlBase = "http://www.mathweb.org/trans/Tramp#";
  private String cnfID;
  
  private boolean busy;
  public ServletEndpointContext endpointContext;

  private static final int defaultTime = 100;
  
  public TrampCNFService()
  {
    mkTempDir("trampService");
    
    owlJDom = OWLJDom.getInstance();
    //    reasoner = PellReasoner.getNewReasoner();
    
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    logger.removeAllAppenders();
    logger.addAppender(appender);
    
    busy = false;
  }
  
//   public void init(Object context) throws ServiceException
//   {
//     logger.debug("context : " + context.getClass().getName());
//     endpointContext = (ServletEndpointContext) context;
//     if (theInstance == null) {
//     theInstance = new Tramp("CNF");
//     theInstance.startTramp();
//       //  theInstance.startTrampEmacs();
//     }
//   }
  
  
  public void destroy()
  {
    logger.debug("destroy() method being called!");
//     if (theInstance != null)
//       theInstance.shutdown();
  }

  private synchronized void busy(boolean b)   { this.busy = b; }
    
  /**
   * Create the Tramp-specific CNF for a TSTP FOF problem.
   * 
   * @param tstpProblem A string containing the TSTP problem description.
   * @return The CNF for the problem.
   */
  public synchronized String tstpFOF2CNF(String fofProblem)
  {
    logger.debug("[trampCNF] : " + fofProblem); 
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    
    if (busy) return owlJDom.createFailure(null, startTime,  cpuStart,"Sorry, Tramp is too busy right now! Call again later.");
    //    if (busy) return new String[] {owlJDom.createFailure(probID, startTime,  cpuStart,"Sorry, Tramp is too busy right now! Call again later."), null };
    
    
    theInstance.startTramp(); 
    while(!theInstance.isReady()){
      if ((System.currentTimeMillis() - startTime) > 20000)
        break;
    }
    if (!theInstance.isReady()) 
      return  owlJDom.createFailure(null, startTime, cpuStart, "Sorry, Tramp is not ready yet! Call again later.");
    
    String probID = null;
    boolean told = false;
    try{
      busy(true);
      reasoner = PellReasoner.getNewReasoner();
      reasoner.tellIndividual(fofProblem);
      told = true;
      Element fofRoot = owlJDom.getRootElement(fofProblem);
      String xmlBase = fofRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      
      logger.debug("[trampCNF] : " + fofRoot); 
      if (owlJDom.hasChild(fofRoot, "TstpFOFProblem",  OWLJDom.MW_NS))        {
        Element prob = (Element) owlJDom.TSTP_FOF_PROBLEM.selectSingleNode(fofRoot);
        if (prob == null) 
          return owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstpFOF2CNF must be an individual of the concept TstpFOFProblem");
        //           return  new String[] {owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstp2CNF must be an individual of the concept TstpFOFProblem"), null
        //           };
        
        probID = xmlBase + owlJDom.valueOf(OWLJDom.RDF_ID, prob);
        logger.debug("[trampCNF] prob ID: " + probID);
        URI probURI = URI.create(probID);
        String fofFormal = reasoner.valueOf("formalDescription", probURI);
        logger.debug("[trampCNF] FOF: " + fofFormal);
        if ((fofFormal == "") || (fofFormal == null))
          throw new Exception("FOF problem is of wrong type!");
        String fofPost = tstp2X.tstp2Post(fofFormal);
        logger.debug("[trampCNF] POST: " + fofPost);
        String logic = reasoner.valueOf("inLogic", probURI);
        
        String[] result = theInstance.callTrampClausifier(fofPost);

        String timeRes = reasoner.valueOf("timeRes", probURI);
        logger.debug("timeRes : " +timeRes);
        int time;
        if (timeRes == null) 
          time = this.defaultTime;
        else {
          String timeStr = reasoner.valueOf("wallClockTime", timeRes);
          time =  Integer.parseInt(timeStr);
        }
        String finalRes = createResult(result[0], result[1], probURI, time, startTime, cpuStart, logic);
        logger.debug("final result : " +finalRes);
        return finalRes;
      } else{
//         return  new String[] {owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstp2Cnf must be an individual of the concept TstpFOFProblem"), null  };
        return  owlJDom.createFailure(probID, startTime,  cpuStart,"The proving problem given to tstp2Cnf must be an individual of the concept TstpFOFProblem");
        
      }
    } catch(Exception e){
      e.printStackTrace();
      logger.debug("[trampCNF] " +e);
      return  owlJDom.createFailure(probID, startTime, cpuStart, "Some error occurred: " + e.getMessage());
//       return  new String[] {owlJDom.createFailure(probID, startTime, cpuStart, "Some error occurred: " + e.getMessage()), null
//       };
    }
    finally {
      if ((told) && (probID !=null)){
        logger.debug("untelling indi");
        reasoner.untellIndividual(probID);
      }
      busy(false);
    }
  }

  private String createResult(String cnfOut, String deltaOut, URI probURI, int timeRes, long startTime, long cpuStart, String logic){
    try{
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", servXmlBase, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element prob = new Element("TstpCNFProblem",OWLJDom.MW_NS);
      cnfID = "tramp_CNF_for_" +probURI.getFragment() +System.currentTimeMillis();
      prob.setAttribute("ID", cnfID, OWLJDom.RDF_NS);
      Element format = new Element("format",OWLJDom.MW_NS);
      format.setAttribute("resource",  OWLJDom.MW_NS_STR+"TSTP", OWLJDom.RDF_NS);
      Element formalProb = new Element("formalDescription",OWLJDom.MW_NS);
      formalProb.addContent(cnfOut);
      Element probLogic = new Element("inLogic",OWLJDom.MW_NS); 
      probLogic.setAttribute("resource", logic, OWLJDom.RDF_NS);
      Element satEquiv = new Element("satEquivalentTo", OWLJDom.MW_NS);
      satEquiv.setAttribute("resource", probURI.toString(), OWLJDom.RDF_NS);

      Element time = new Element("timeRes",OWLJDom.MW_NS);
      Element timeres = new  Element("TimeResource",OWLJDom.MW_NS);
      Element wcTime = new  Element("wallClockTime",OWLJDom.MW_NS);
      Element cpuTime = new  Element("cpuTime",OWLJDom.MW_NS);
      Integer iTime = new Integer(timeRes + 1);
      Integer jTime = new Integer(timeRes);
      wcTime.addContent(iTime.toString());
      wcTime.setAttribute("datatype", OWLJDom.RDFS_NS_STR+"integer", OWLJDom.RDF_NS);
      cpuTime.addContent(jTime.toString());
      cpuTime.setAttribute("datatype", OWLJDom.RDFS_NS_STR+"integer", OWLJDom.RDF_NS);
      timeres.addContent(wcTime);
      timeres.addContent(cpuTime);
      time.addContent(timeres);
      
      prob.addContent(formalProb);
      prob.addContent(probLogic);
      prob.addContent(satEquiv);
      prob.addContent(time);

      Element delta = new Element("DeltaRelation",OWLJDom.MW_NS);
      String resultID = "tramp_DeltaRel_for_" +probURI.getFragment() +System.currentTimeMillis();
      delta.setAttribute("ID", resultID, OWLJDom.RDF_NS);

      Element toFOF = new Element("toFOF",OWLJDom.MW_NS);
      toFOF.setAttribute("resource", probURI.toString(), OWLJDom.RDF_NS);

      Element relatesCNF = new Element("relatesCNF",OWLJDom.MW_NS);
      relatesCNF.setAttribute("resource", cnfID, OWLJDom.RDF_NS);

      Element relation = new Element("relation",OWLJDom.MW_NS); 
      relation.addContent(deltaOut);
      
      delta.addContent(relatesCNF);
      delta.addContent(toFOF);
      delta.addContent(relation);

      rdf.addContent(prob);
      rdf.addContent(delta);

      Element timeUsed = OWLJDom.createTimeResource(startTime, cpuStart);
      rdf.addContent(timeUsed);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating the OWL answer: " + e.getMessage();
    }
  }

  public void waitMillis(long millis)
  {
    long startTime = System.currentTimeMillis();
    while ((System.currentTimeMillis() - startTime) < millis);
  }

  public void shutdown()
  {
    if (theInstance != null)
      theInstance.shutdown();
    removeTempDir();
  }
  
  public void test () {
    try {
      String fofProb = readFile(msHome + "/examples/owl/fof2.owl");
      //String fofProb = readFile(msHome + "/src/b");
      logger.debug("[Tramp]  " + fofProb); 
 //       String[] cnf;
//       cnf = tstpFOF2CNF(fofProb);
//       writeFile(msHome + "/examples/owl/cnf2.owl", cnf[0]);
//       writeFile(msHome + "/examples/owl/delta2.owl", cnf[1]);
//       logger.debug("[TrampCNFService] generated cnf: " + cnf[0]);
//       cnf = tstpFOF2CNF(fofProb);
//       logger.debug("[TrampCNFService] 2nd generated cnf: " + cnf[0]);
      String cnf;
      cnf = tstpFOF2CNF(fofProb);
      writeFile(msHome + "/examples/owl/cnf2.owl", cnf);
//       logger.debug("[TrampCNFService] generated cnf: " + cnf);
//       cnf = tstpFOF2CNF(fofProb);
//       logger.debug("[TrampCNFService] 2nd generated cnf: " + cnf);
     }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    TrampCNFService e = new TrampCNFService();
    e.waitMillis(3000);
    
    e.test();
    e.shutdown();
    System.exit(0);
  }
  
} // end TrampCNFService.java

     
 
