// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.util.Iterator;
import java.net.URI;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import javax.xml.transform.TransformerException;
import javax.xml.rpc.ServiceException;
// import javax.xml.rpc.server.ServiceLifecycle;
// import javax.xml.rpc.server.ServletEndpointContext;
// import javax.xml.rpc.handler.MessageContext;
//import javax.servlet.http.HttpSession;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;

import org.mathweb.atp.TSTP2X;

import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.Tramp;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/07/05 16:00:18 $ $Revision: 1.22 $
*/
public class TrampNDService extends BasicService //implements ServiceLifecycle 
{

  private static Logger logger = Logger.getLogger("TrampNDService");
  private static String logFileName = getSystemProperty("tramp.nd.log.file", "OTTER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  /** TSTP trafo stuff */
  private static TSTP2X tstp2X = new TSTP2X();
  
  /** Ontology stuff */
  private static OWLJDom owlJDom;
  private static OntoReasoner reasoner;
  
  private static Tramp theInstance = (new Tramp("ND")).startTramp();
  
  private static String proofConcept = "TstpCnfBrFPRefutation";
  private static String servXmlBase = "http://www.mathweb.org/trans/Tramp#";
  
  private boolean busy;
  
  public static final int defaultNice = 15;
  //  public ServletEndpointContext endpointContext;

  public TrampNDService()
  {
    mkTempDir("trampService");
    //    theInstance = new Tramp("ND");
    
    owlJDom = OWLJDom.getInstance();
    reasoner = PellReasoner.getNewReasoner();

    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    logger.removeAllAppenders();
    logger.addAppender(appender);

    busy = false;
  }
  
//   public void init(Object context) throws ServiceException
//   {
//     logger.debug("context : " + context.getClass().getName());    
//     //    endpointContext = (ServletEndpointContext) context;

// //   MessageContext ctx = endpointContext.getMessageContext();
// //     Iterator it = ctx.getPropertyNames();
// // //     while(it.hasNext()){
// // //       logger.debug("property: " + it.next());
// // //     }
// //     String[] props = new String[] {"attachments.directory", "home.dir", "transport.http.servletLocation", "transport.http.servletRequest", "jws.classDir", "realpath", "remoteaddr", "servletEndpointContext", "transport.http.servlet", "configPath", "transport.url", "transport.http.servletPathInfo", "path"};
// //     Object obj;
// //     for(int k=0; k<props.length; k++){
// //        obj =  ctx.getProperty(props[k]);
// //        logger.debug(obj.getClass().getName() + " " + obj);
// //     }
    
    
// //     if (theInstance == null) {
// //       theInstance = new Tramp("ND");
// //       //      waitMillis(400);
// //       //      theInstance.startTramp();
// //       //    theInstance.startTrampEmacs();
// //     }
//   }
  
  public void destroy()
  {
    logger.debug("destroy() method being called!");
    if (theInstance != null)
      theInstance.shutdown();
    return;
  }
  
  /**
   * Mark this service instance as busy or not.
   * 
   * @param b If true, mark the service busy.
   */
  private synchronized void busy(boolean b)   { this.busy = b; }
    
   
  /**
   * Create an ND proof for a TSTP FOF problem given a FO-ATP result.
   * 
   * @param fofProblem A string containing an OWL TstpFofProblem.
   * @param foAtpResult A string containing the OWL individual with a .
   * @return If successful, an ND proof for the FOF problem.
   */
  public synchronized String ndForFOF(String fofProblem, String foAtpResult, String deltaRel)
  {
    logger.debug(fofProblem);
    logger.debug(foAtpResult);
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    String probID = null;
    String resultID = null;
    boolean told1 = false;
    boolean told2 = false;
    
    // if this class is busy, return directly
    if (busy) return owlJDom.createFailure(null, startTime, cpuStart, "Sorry, Tramp is too busy right now! Call again later.");

    // otherwise wait 4 seconds until Tramp main class is ready (Lisp is started)
    //    theInstance.startTramp();
    
    while(!theInstance.isReady()){
      logger.info("[trampND] tramp not ready " + (System.currentTimeMillis() - startTime)/1000);
      if ((System.currentTimeMillis() - startTime) > 4000)  break;
    }
    if (!theInstance.isReady()) return owlJDom.createFailure(null, startTime, cpuStart, "Sorry, Tramp is not ready yet! Call again later.");
    
    try{
      busy(true);
      //      reasoner = PellReasoner.getNewReasoner();
      reasoner.tellIndividual(fofProblem); told1 = true;
      reasoner.tellIndividual(foAtpResult); told2 = true;
      Element fofRoot = owlJDom.getRootElement(fofProblem);
      Element resRoot = owlJDom.getRootElement(foAtpResult);
      String fofBase = fofRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      String resBase = resRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      
      logger.debug("[trampND] : " + fofRoot + " " + resRoot + OWLJDom.MW_NS); 
      
      if ((owlJDom.hasChild(fofRoot, "TstpFOFProblem",  OWLJDom.MW_NS)) &&
          (owlJDom.hasChild(resRoot, "FOBrFPResult",  OWLJDom.MW_NS)))
        {
          
          // first translate the TSTP FOF problem into POST:
          Element prob = (Element) owlJDom.TSTP_FOF_PROBLEM.selectSingleNode(fofRoot);
          probID = fofBase + owlJDom.valueOf(OWLJDom.RDF_ID, prob);
          URI probURI = URI.create(probID);
          logger.debug("[trampND] prob ID: " + probID);
          String fofFormal = reasoner.valueOf("formalDescription", probURI);
          if (fofFormal == "") throw new Exception("FOF problem is of wrong type!");
          String fofPost = tstp2X.tstp2Post(fofFormal);
          logger.debug("[trampND] POST: " + fofPost);
          
          // then extract the TSTP proof from the FO ATP Result:
          
          //          Element atpResult = (Element) owlJDom.FO_ATP_RESULT.selectSingleNode(resRoot);
          logger.debug("[trampND] ATP root: " + resRoot);
          Element atpResult = resRoot.getChild("FOBrFPResult", OWLJDom.MW_NS);
          if (atpResult == null) throw new Exception("ATP result is of wrong type!");
          resultID = resBase + owlJDom.valueOf(OWLJDom.RDF_ID, atpResult);
          URI resultURI = URI.create(resultID);
          logger.debug("[trampND] result ID: " + resultID);
          String proof = reasoner.valueOf("proof", resultURI);
          String actualProof = reasoner.valueOf("formalProof", proof);
          //          logger.debug("[trampND]  proof: " + proof);
          //          logger.debug("[trampND]  formal: " + actualProof);
          
//           Element proof = (Element) owlJDom.getSingleNode(OWLJDom.FO_ATP_RESULT_PROOF, atpResult);
//           Element refutation = (Element) owlJDom.getSingleNode(OWLJDom.BRFP_REFUTATION,proof);
//          String actualProof = owlJDom.valueOf(OWLJDom.PROOF_FORMAL_PROOF, refutation);
//          logger.debug("[trampND] cnf proof: " + actualProof);
          if (actualProof == null) throw new Exception("Refutation proof is of wrong type!");
          
          String ndProof = theInstance.callTrampFof2ND(fofPost, actualProof);
          logger.debug("[trampND] nd proof: " + ndProof);
          busy(false);
          return createNDProof(ndProof, probURI, startTime, cpuStart);
        }else{
          return owlJDom.createFailure(probID, startTime, cpuStart, "The proving problem given to ndForFOF must be an individual of the concept TstpFOFProblem");
        }
    } catch(JDOMException e){
      return owlJDom.createFailure(probID, startTime,  cpuStart,"Could not parse XML: " + e);
    } catch(TransformerException e){
      return owlJDom.createFailure(probID, startTime,  cpuStart,"Could not transform TSTP problem into POST format:\n " + e);
    } catch(Exception e) {
      e.printStackTrace();
      return owlJDom.createFailure(probID, startTime,  cpuStart, e.toString());
    } finally {
      if ((told1) && (probID !=null))
        reasoner.untellIndividual(probID);
      if ((told2) && (resultID !=null))
        reasoner.untellIndividual(resultID);
      busy(false);
    }
  }

  private String createNDProof(String output, URI probURI, long startTime, long cpuStart){
    try{
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", servXmlBase, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element proof = new Element("TwegaNDProof",OWLJDom.MW_NS);
      String resultID = "tramp_ND_proof_for_" +probURI.getFragment() +System.currentTimeMillis();
      proof.setAttribute("ID", resultID, OWLJDom.RDF_NS);
      Element language = new Element("language",OWLJDom.MW_NS);
      language.setAttribute("resource", OWLJDom.MW_NS_STR+"Twega", OWLJDom.RDF_NS);
      Element proofOf = new Element("proofOf",OWLJDom.MW_NS);
      proofOf.setAttribute("resource", probURI.toString(), OWLJDom.RDF_NS);
      Element formalProof = new Element("formalProof",OWLJDom.MW_NS);
      formalProof.addContent(output);
      Element proofLogic = new Element("proofLogic",OWLJDom.MW_NS);
      proofLogic.setAttribute("resource", OWLJDom.MW_NS_STR+"fol_plus_equ", OWLJDom.RDF_NS);
      Element calc = new Element("calculus",OWLJDom.MW_NS);
      calc.setAttribute("resource", OWLJDom.MW_NS_STR+"OmegaNDCalculus", OWLJDom.RDF_NS);
      
      Element time = OWLJDom.createTimeResource(startTime, cpuStart);

      proof.addContent(calc);
      proof.addContent(proofOf);
      proof.addContent(proofLogic);
      proof.addContent(language);
      proof.addContent(formalProof);
      rdf.addContent(proof);
      rdf.addContent(time);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating the OWL answer: " + e.getMessage();
    }
  }
  
  public void waitMillis(long millis)
  {
    long startTime = System.currentTimeMillis();
    while ((System.currentTimeMillis() - startTime) < millis);
  }


  public void shutdown()
  {
    if (theInstance != null)
      theInstance.shutdown();
    removeTempDir();
  }
  
  public void test () {
    String msHome = System.getProperty("mathserv.home");
    try {
      String fofProb, tstpProof;
      fofProb = readFile(msHome + "/examples/owl/fof2.owl");
      tstpProof = readFile(msHome + "/examples/owl/otter-proof2.owl");
      String deltaRel = "";
     //tstpProof = readFile(msHome + "/examples/owl/spass-proof2.owl");
      //tstpProof = readFile(msHome + "/examples/owl/e-proof2.owl");
      System.out.println("[Tramp]  " + fofProb + tstpProof); 
      String nd;
 
      nd = ndForFOF(fofProb, tstpProof, deltaRel);
      System.out.println("[Tramp] generated ND proof: " + nd);
      writeFile(msHome + "/examples/owl/nd2.owl", nd);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    TrampNDService e = new TrampNDService();
    e.waitMillis(3000);
    
    e.test();
    e.shutdown();
    System.exit(0);
  }
  
} // end TrampNDService.java

      

