/*--------------------------------------------------------------*/
/*      infer.c
/*
/*
/*      functions:
/*         int binary_resolvents()     ................... #1
/*         int binary_factors()        ................... #2
/*         int hash_resolvent()        ................... #2a
/*         int function_eval()         ................... #2b
/*         standardize()               ................... #3
/*         clear_var_substitution_tble()  ................ #4
/*         clean_var_names().............................. #5
/*--------------------------------------------------------------*/
 
#include <stdio.h>

#include "const.h"
#include "clause.h"

/*--------------------------------------------------------------*/
/* externally defined functions                                 */
/*--------------------------------------------------------------*/
 
extern int unify();                 /* unify two literals       */
extern int install_clause();        /* installation of clauses  */
extern clause *generate_clause();   /* generate a new clause    */
extern int eval_fxns();
extern int calc_lit_hc();
extern void calc_clause_hashcode();
extern clause *copy_clause();
extern void repair_literal();
extern void replace_vars();
extern void normalize_literal();
extern void print_clause2();

/*--------------------------------------------------------------*/
/* externally defined constants                                 */
/*--------------------------------------------------------------*/

extern clause *clist[CSPACE_CLAUSES];
extern int *var_substitution_tbl[];    /* variable subs table   */
extern int ply, nextc, iter_depth;
extern int s_nod;
extern int o_nc, o_debug, o_print;
extern int number_preds, equal_pred;
extern int alt_hx, alt_hy;
extern int alt_HX, alt_HY;
extern int v1_count, v2_count;
extern int hr_hashcodes[3][50];
extern int o_maxterm;

/*--------------------------------------------------------------*/
/* internally defined functions                                 */
/*--------------------------------------------------------------*/

int hash_resolvent();
int binary_resolvents();
int binary_factors();
int function_eval();
void standardize();
void clear_var_substitution_tble();
void clean_var_names();

/*------------------------------------------------------------------*/
/*      binary_resolvents()         ................... #1
/*
/*      binary res is an inference rule that produces a new clause
/*      from two parent clauses.
/*
/*      Let C = {Li}n and D = {Mj}m (n,m = number of lit'ls in clause).
/*      If Li and Mj are unifiable with mgu
/*      s, then the binary resolvent is (C - {Li})s + (D - {Mj})s.
/*
/*      The first step in the resolution process is standardization of
/*      the variables in the two clauses. MAXVARS is maximum number of
/*      distinct vars in a clause. By convention, the variables of a
/*      clause start at either 0 or MAXVARS. Thus, we need only st'dize
/*      50% of the time (with luck, the two clauses will already be
/*      standardized).
/*
/*      Since variables of a clause are linked together, we need only
/*      follow the chain to st'dize the variables apart. Naturally, we
/*      use the nvar field so that we do the min work possible; i.e.
/*      we standardize the clause with the least number of variables.
/*
/*      parameters:     c1,c2 : clauses to resolve
/*                      m : [UNIT]|[MERGE]
/*
/*      called by:  calc_inferences()                search.c
/*
/*      calls:      standardize()                    infer.c
/*                  clear_var_substitution_tble()    unify.c
/*                  unify()                          unify.c
/*                  generate_clause()                gen.c
/*                  install_clause()                 search.c
/*
/*      returns:    # of clauses produced
/*------------------------------------------------------------------*/
 
int binary_resolvents(c1,c2,m)
clause *c1,*c2;
int m;
{
literal *l1, *l2;
int id1, id2;
clause *new_cl;
int ix1, ix2;	
int numb_resolvents;

/*----------------------------------------------*/
/* check resolution restrictions                */
/*----------------------------------------------*/

id1 = c1->c_bits; id2 = c2->c_bits;
 
/*--------------------------------------------------------------*/
/* if extended-search opt. is on, and c2 is not a unit          */
/* clause, return                                               */
/*--------------------------------------------------------------*/

if(m & UNIT) {
   if(!(id2 & UNIT)) {return(0);}}
		
/*--------------------------------------------------------------*/
/* if MERGE opt. is on, and neither c1 or c2 is a merge, return	*/
/*--------------------------------------------------------------*/
 
if(m & MERGE) {
   if(!(id2 & MERGE)) {
      if(!(id1 & MERGE)) {return(0);}}}
					
/*----------------------------------------------*/
/* if one of the clauses is not ACTIVE, return  */
/*----------------------------------------------*/
 
if(!(id1 & id2 & ACTIVE)) {return(0);}

/*----------------------------------------------*/
/* if at ply 1 and both clauses are equality    */
/* substitution axioms, don't form resolvent.   */
/* Also, if NC is on and neither clause         */
/* is in the NC, return                         */
/*----------------------------------------------*/

if(ply == 1) {
   if((id1 & EQCLS) && (id2 & EQCLS)) {
      return(0);}}

if(ply == 1) {
   if(o_nc) {
      if(!((id1 | id2) & NC)) {return(0);}}}

numb_resolvents = 0;       /* count of the # of binary resolvents */
standardize(c1,c2);

ix1 = 0;
for(l1 = c1->c_lit ; l1 != 0 ; l1 = l1->l_next) {
  if(!(l1->l_id & ELIMLIT)) {
    if(ix1 > 1)  {
       return(numb_resolvents);} 
    ix1++;    
    ix2 = 0;
    for(l2 = c2->c_lit ; l2 != 0; l2 = l2->l_next) {
      if(!(l2->l_id & ELIMLIT)) {
        ix2++;
        id1 = l1->l_id & IDMASK;          /* predicate IDs  */
        id2 = l2->l_id & IDMASK;
 
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/

        if(id1 == id2) {
          if(((l1->l_id ^ l2->l_id) & ~VARMASK) == NEGMASK) {
            clear_var_substitution_tble(c1->c_bits,c1->c_nv);
            clear_var_substitution_tble(c2->c_bits,c2->c_nv);
            if(unify(&l1->l_id,&l2->l_id)) {
              if(new_cl=generate_clause(c1,c2,l1,l2,ix1,ix2)) {
                if(install_clause(new_cl)) {
                  numb_resolvents++;}}}}}}}}}
return(numb_resolvents);
}
 
/*-------------------------------------------------------------*/
/*      binary_factors()      ................... #2
/*
/*      factoring is an inference rule that produces a new clause
/*      from a given non-unit cl. A typical example of factoring
/*      is the following:
/*
/*                     p(x) | p(y)
/*
/*      can be factored into p(x). What we do is attempt to find an
/*      mgu such that we collapse two literals of a clause into a
/*      singleton. A factored cl is a merge if parent is a merge
/*
/*      binary_factors will generate all factors of a given cause.
/*      Since the new clause will have only one parent, we use the
/*      left parent (and NULL the right parent).
/*
/*      parameters: c : non-unit clause to factor
/*
/*      called by:  calc_inferences()               search.c
/*
/*      calls:      clear_var_substitution_tble()   unify.c
/*                  unify()                         unify.c
/*                  generate_clause()               gen.c
/*                  install_clause()                search.c
/*
/*      returns:    the # of factors produced
/*-------------------------------------------------------------*/
 
int binary_factors(c)
clause *c;
{
literal *l1, *l2;
clause *new_cl;
int ix1, ix2;
int numb_factors;
	
/*---------------------------------------------------*/
/* if at ply 1 and nc strategy is ON and clause c    */
/* is not in the negated conclusion, return          */
/*---------------------------------------------------*/

if(o_nc) {  
   if(ply == 1) {  
      if(!(c->c_bits & NC))  {return(0);}}}
 
if(!(c->c_bits & NC) && ply <= iter_depth)  {return(0);}

ix1 = ix2 = 0;
numb_factors = 0;                    /* # of factors produced  */

for(l1 = c->c_lit; l1->l_next != 0; l1 = l1->l_next) {
   if(ix1 > 1) {return(numb_factors);}
   if(!(l1->l_id & ELIMLIT)) {
      ix1++;
      ix2 = ix1;
      for(l2 = l1->l_next; l2 != 0; l2=l2->l_next) {
         if(!(l2->l_id & ELIMLIT)) {
            ix2++;
            if((l1->l_id & ~VARMASK) != (l2->l_id & ~VARMASK)) {
	       continue;}
            clear_var_substitution_tble(c->c_bits,c->c_nv);
            if(unify(&l1->l_id,&l2->l_id)) {
               new_cl = generate_clause(c,0,l1,0,ix1,ix2);
               if(new_cl) {
                  if(install_clause(new_cl)) {
                     numb_factors++;}}}}}}}
return(numb_factors);
}

/*--------------------------------------------------------------*/
/* hash_resolvent
/*
/*--------------------------------------------------------------*/

int hash_resolvent(c,lit_old,c_index,constant_val)
clause *c;
literal *lit_old;
int c_index, constant_val;
{
clause *c_new;
literal *lit,*lit2;
int n_index;
int i;

hr_hashcodes[0][ply] = alt_HX;
hr_hashcodes[1][ply] = alt_HY;
hr_hashcodes[2][ply] = c_index;

c_new = copy_clause(c,c_index,0);
repair_literal(c_new,c_index);
c_new->c_change = constant_val;
replace_vars(c_new,constant_val);

n_index = 0;
for(lit = c_new->c_lit ; lit!= 0; lit = lit->l_next) {
   n_index++;
   if(n_index == c_index) {lit2 = lit;break;}}

normalize_literal(lit,c_new);

for(lit = c_new->c_lit ; lit!= 0; lit = lit->l_next) {
   calc_lit_hc(lit,c_new);}
calc_clause_hashcode(c_new);

if(install_clause(c_new)) {
   lit2->l_X = alt_HX;lit2->l_Y = alt_HY;
   alt_HX = 0;
   return(1);}

alt_HX = 0;
return(0);
}

/*--------------------------------------------------------------*/
/* function_eval()     ............................#4
/*
/* see whether clause can be simplified by a function evaluation
/*
/* called by:
/*
/* calls:
/*
/* returns: a 1 if success or a 0 if unsuccessful
/*--------------------------------------------------------------*/

int function_eval(c)
clause *c;
{
literal *lit;
int lits_modified;

lits_modified = 0;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(eval_fxns(c,lit)) {
       calc_lit_hc(lit,c);
       lits_modified++;}}
if(lits_modified != 0) {return(1);}
return(0);
}


/*--------------------------------------------------------------*/
/*      standardize(c1,c2)            ................... #5
/*
/*      renames variables
/*
/*      called by:  binary_resolvents()      infer.c
/*
/*      calls:      nobody
/*
/*      returns:    nothing
/*--------------------------------------------------------------*/
 
void standardize(c1,c2)
clause *c1,*c2;
{
int i, a, *p;
literal *l;
 
/*---------------------------------------------*/
/* if variable in one clauses are numbered     */
/* beginning at 0 in one clauses and 200 in    */
/* the other, var are already standardized,    */
/* and the program should return               */
/*---------------------------------------------*/
 
if((c1->c_bits & VID) != (c2->c_bits & VID)) {return;}
 
/*---------------------------------------------*/
/* renumber the variables in clause c2         */
/*---------------------------------------------*/
 
if(!(c2->c_bits & VID)) {
   c2->c_bits |= VID;
   i = MAX_VARS;}        /* if now 0, add 200  */
else {
   c2->c_bits &= ~VID;
   i = -MAX_VARS;}      /* otherwise, sub 200  */
   
/*----------------------------------------------*/
/* now pass through variable chain while        */
/* renumbering variables                        */
/*----------------------------------------------*/
 
for(l = c2->c_lit ; l != 0 ; l = l->l_next) {
   p = &l->l_id;
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {
         a += *p++ & ARGMASK;}
      else {
         *p += i;
         p++;}}}
 
}
 
 
/*-------------------------------------------------------------*/
/*      clear_var_substitution_tble() .................... #6
/*
/*      vstclr is used to zero a section of the vst table (i.e.
/*      set each entry to NULL). bres, fact fxns call this fxn
/*      when they wants to perform a unification
/*
/*      calling arguments:
/*                  i = c_base (ptr to first variable in cl
/*                  n = c_nv (number of distinct var in cl
/*
/*      called by:  binary_resolutions()    infer.c
/*                  binary_factors()        infer.c
/*                  bres_ply_1()            elim.c
/*                  bres_replace()          elim.c
/*                  bres_link()             elim.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*--------------------------------------------------------------*/
 
void clear_var_substitution_tble(i,n)
int i,n;
{
int **p, j;
 
if(!(i & VID)) {j = 0;}
else {j = MAX_VARS;}
for(p = &var_substitution_tbl[j]; n > 0; n--) {
   *p++ = 0;}
}
 
/*--------------------------------------------------------------*/
/* clean_var_names()
/*--------------------------------------------------------------*/

void clean_var_names(c)
clause *c;
{
literal *l;
int i, a, *p;

c->c_bits = c->c_bits &= ~VID;

for(l = c->c_lit; l != 0; l = l->l_next) {
   p = &l->l_id;
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {
         a += *p++ & ARGMASK;}
      else {
         *p &= 0x0000003F;
         p++;}}}
 
return;
}
