/*-------------------------------------------------------*/
/*       pass2.c
/*       copyrighted by Monty Newborn 2005
/*
/*      functions:
/*
/*          functions involved in saving the proof in the
/*          cl_array
/*
/*      init_cl_array() ......................#1
/*      save_line()     ......................#2
/*      fix_fxn_evals().......................#2a
/*      elim_base_clause_hc().................#3
/*      elim_same_cont_hc() ..................#4
/*      test_for_base_match().................#5
/*      test_for_cont_match().................#6
/*      stack_hash_codes()  ..................#7
/*      print_hash_code_list()  ..............#8
/*      print_cl_array()    ..................#9
/*      on_hash_code_list() ..................#10
/*      alt_on_hash_code_list()...............#11
/*      delete_found_hcs()....................#12
/*      delete_alt_hcs()......................#13
/*      move_down_renumber()..................#14
/*
/*          function that transfers the proof from the bottom of
/*          cl_array to the top
/*
/*      rearrange_cl_array()..................#15
/*
/*          functions involved in the reconstruction of the proof
/*          from information in the cl_array
/*
/*      bres_proof_line().....................#16
/*      factor_proof_line()...................#17
/*      instantiate_clause()..................#17b
/*      substitute_string()...................#17a
/*      install_proof_line()..................#18
/*      infer_proof_line()....................#19
/*      clean_proof().........................#19a
/*      search_proof_line() ..................#20
/*
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*
/*      The functions carried out in this file can be broken down into
/*      four categories.  They all deal with placing information in the
/*      cl_array and manipulating it so that the proof resides there
/*      as a set of four-tuples.  Each tuple contains the four
/*      arguments necessary to form a resolvent or a factor
/*
/*          1.  clears cl_array, hash_code_list
/*
/*                  init_cl_array()
/*
/*          2.  place lines of proof in cl array along with hash codes
/*              of missing literals.
/*
/*                 save_line()
/*                    delete_found_hash_codes()
/*                    elim_base_clause_hc()
/*                       test_for_base_match()
/*                    elim_same_cont_hc()
/*                       test_for_cont_match()
/*                    move_down_renumber()
/*                 stack_hash_codes()
/*                 on_hash_code_list()
/*                 alt_on_hash_code_list()
/*                 print_cl_array()
/*                 print_hash_code_list()
/*
/*          3.  when the search is done, reorganize the information
/*              in the cl array by reordering clauses and adding
/*              inferences that yielded the hash codes of the missing
/*              literals.
/*
/*                 rearrange_cl_array()
/*
/*          4.  Reconstruct proof by carrying out the inferences
/*              as indicated by the information in the cl array.
/*              each row of the cl array correponds to one inference.
/*
/*                 search_proof_line()
/*                    infer_proof_line()
/*                       factor_proof_line()
/*                          clear_var_substitution_tble()
/*                          unify() 
/*                          generate_clause()
/*                          install_proof_line()
/*                       bres_proof_line()
/*                          standardize()               
/*                          clear_var_substitution_tble()
/*                          unify()
/*                          generate_clause()
/*                          install_proof_line()
/*                          print_cl_array()      
/*                          print_clause()       
/*
/*-------------------------------------------------------------*/
 
#include <stdio.h>
 
#include "const.h"
#include "clauseS.h"

/*-------------------------------------------------------------*/
/* externally defined functions                                */
/*-------------------------------------------------------------*/
 
extern void fatal();                       /* theo.c           */

extern void print_clause();                /* dump.c           */
extern void print0();
extern void stat_dump();
extern void print_continuation();

extern clause *generate_clause();          /* gen.c            */
extern clause *hgenerate_clause();         /* gen.c            */
extern int test_equality();
extern int eval_fxns();
extern clause *copy_clause();
extern void print_sub_array();
extern void hgen_normalize();

extern void standardize();                 /* infer.c          */
extern void standardize0();                 /* infer.c          */
extern void clear_var_substitution_tble();

extern int unify();                        /* unify.c          */
 
extern void calc_clause_hashcode();        /* hash.c           */
extern void calc_alt_hash();
extern void normalize_literal();
extern void replace_vars();

/*-------------------------------------------------------------*/
/* externally define constants                                 */
/*-------------------------------------------------------------*/
 
extern char *OutPtr;
 
extern int *clause_space[];
extern clause *clist[];
extern dnode depth[];
extern char *fidtbl[];

extern int sub_array[65536][6];
extern int sub_array_last;
extern int *var_substitution_tbl[];

extern int ply, base, org_base, nc, nextc;
extern int s_nod, s_nod0, iter_depth;
extern int magic_hashx, magic_hashy;
extern int alt_hx, alt_hy;
extern int equal_clist_index, equal_pred;
extern int o_screen, o_print;
extern int o_debug, big_bug;
extern int extra_pass, phase;
extern int o_maxterm, o_maxvar, o_maxlit;
extern int o_learn;
extern int number_const_hash;
extern int number_constants;
extern int rows_less_nfxns;
extern int var_array[64], var_array_ptr;
extern int hr_hashcodes[3][50];
extern int v1_count, v2_count;
extern int s_starttime;
 
extern int this_computer;
extern int clist0;

/*-------------------------------------------------------------*/
/* internally defined functions                                */
/*-------------------------------------------------------------*/

void init_cl_array();
void save_line();
int fix_fxn_evals();
int elim_base_clause_hc();
int elim_same_cont_hc();
int test_for_base_match();
int test_for_cont_match();
void stack_hash_codes();
void print_hash_code_list();
void print_cl_array();
int on_hash_code_list();
int alt_on_hash_code_list();
void delete_found_hcs();
void delete_alt_hcs();
void move_down_renumber();
void rearrange_cl_array();
void bres_proof_line();
void factor_proof_line();
void hash_resolve();
void instantiate_clause();
void substitute_string();
void install_proof_line();
void infer_proof_line();
void clean_proof();
void search_proof_line();

/*------------------------------------------------------*/
/* the hash code list saves  hash values of literals    */
/* for which the complement must be found in pass2.     */
/* hash_code_list_active is a 32-bit vector the i-th bit*/
/* of which is a 0 or 1 to indicate whether the i-th    */
/* element on hash_code_list is still to be found       */
/*------------------------------------------------------*/
 
int o_maxterm_tempy, o_maxvar_tempy,o_maxlit_tempy;
int hash_code_listx[256], hash_code_listy[256], hash_code_listz[256];
int hash_code_listu[256];
int hash_code_list_top, hash_code_list_active[256]; 
int hash_code_active_number;
int current_proof_top, proof_length, deleted_clauses;
int clarray_top;
int killer_hashx, killer_hashy, killer_hash_ptr;
 
/*------------------------------------------------------*/
/* cl[PROOF_ARRAY_SIZE][14] stores info on resolt's and */
/* factors that must be performed to obtain the proof.  */
/* a proof can be at most PROOF_ARRAY_SIZE int.         */
/*------------------------------------------------------*/
 
int cl[PROOF_ARRAY_SIZE][14];
int xl[PROOF_ARRAY_SIZE][14];
int zz;

/*---------------------------------------------------------------*/
/*      init_cl_array()                    ................#1
/*
/*      initialize several variables and clears the cl array,
/*      and hash_code_list.  executed before search begins.
/*
/*      called by:      main()              theo.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_cl_array()
{       
int i,j;
	
hash_code_list_top      = 0;  /* counter of lits repby hash vals */
hash_code_active_number = 0;
proof_length            = 0;
current_proof_top       = 0;
clarray_top             = 0;
	
for(i = 0; i < PROOF_ARRAY_SIZE ; i++) {
   for(j = 0; j < 14 ; j++)   {cl[i][j] = 0;}}
for(i = 0; i < 256  ; i++) {
   hash_code_list_active[i] = 0;
   hash_code_listu[i] = 0;
   hash_code_listx[i] = hash_code_listy[i] = hash_code_listz[i] = 0;}
return;
}
 
/*---------------------------------------------------------------*/
/*      save_line()                       ................#2
/*
/*      this function is called to save the path of inferences
/*      that led to the contradiation and to each unit clause
/*      whose hash code was previously saved on hash_code_list.
/*      Inferences are stored in the cl-array.
/*
/*      called by:  main()                  theo.c
/*                  install_clause()        search.c
/*
/*      calls:      print_cl_array()        pass2.c
/*                  print_clause()          tdump.c
/*                  delete_found_hcs()      pass2.c
/*                  elim_base_clause_hc()   pass2.c
/*                  elim_same_cont_hc()     pass2.c
/*                  move_down_renumber()    pass2.c
/*                  delete_alt_hcs()        pass2.c
/*                  normalize_literal()     hash.c
/*                  calc_lit_hc()           hash.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void save_line(c_killer,to_kill)
clause *c_killer;
int to_kill;
 
/*----------------------------------------------*/
/* c_killer (killer clause) is 0 in phase 1     */
/* and is a pointer to the unit clause ending   */
/* the path which is being saved in phase 2     */
/*----------------------------------------------*/
 
{
clause *c;
literal *l, *alit, *litt;
int i, ii, ij, j, k, l_index, cid, number_of_inserts;
int top_ptr, line_length;
int clpar, cl2par, clindex, crindex, ply1, killer, cr;
int hr_lit_elim;
int real_index;
int index_ctr;
int hxtemp,hytemp;
int hash_baseX, hash_baseY;
int p_size;
	
/*---------------------------------------------------------------*/
/*
/*      cl array structure
/*
/*      cl[i][0]......bits 28-32 ptr to element on hash_code_list
/*                    that corresponds to the hashcode of the
/*                    missing literal
/*                    bits 16-26 the index of the clause
/*                    bits 05-15 the index of the left parent clause
/*                    bits 00-04 the index of the left literal
/*      cl[i][1]......bits 05-16 the index of the rght parent clause
/*                    bits 00-04 the index of the right literal
/*           ....or   bits 00-31 the l->l_X hash code if literal
/*                    is missing
/*      cl[i][2]......0 if no literal is missing from clause
/*                    l->l_Y if literal is missing
/*
/*--------------------------------------------------------------*/
				
/*--------------------------------------*/
/* print a "." to indicate a line is    */
/* being saved                          */
/*--------------------------------------*/
 
//if(o_debug == 0) {print0(".");}
 
/*--------------------------------------*/
/* print line if debug option is on     */
/*--------------------------------------*/

if(o_debug > 0) {
   printf(
   "\nSAVE LINE PLY:%d NODES:%d ITER:%d\n",
                                ply,s_nod - s_nod0,iter_depth);
   OutPtr += sprintf(OutPtr,
   "\nSAVE LINE PLY:%d NODES:%d ITER:%d\n",
                                ply,s_nod - s_nod0,iter_depth);
   stat_dump();

   if(!proof_length) {
      for(i = 1; i <= ply; i++)  {
	 c = clist[depth[i].d_tc];
         print_clause(c);
         if(o_debug > 1 && c->c_rpar != 0) { 
	    print_clause(c->c_rpar);}}}
   else {
      for(i = 1; i < ply; i++)  {
	 c = clist[depth[i].d_tc];
         print_clause(c);
         if(o_debug > 1 && c->c_rpar != 0) {
	    print_clause(c->c_rpar);}}
	 
      print_clause(c_killer);
      if(o_debug > 1 && c_killer->c_rpar != 0) {
	 print_clause(c_killer->c_rpar);}}} 
 
/*----------------------------------------*/
/* scan main line and hash_code_list for  */
/* missing clauses whose hash codes       */
/* have already been found                */
/*----------------------------------------*/

delete_found_hcs(killer_hashx,killer_hashy);             

/*----------------------------------------*/
/* delete variants of clause              */
/*----------------------------------------*/

if(phase == 2 && !to_kill) {
   delete_alt_hcs(c_killer);}

top_ptr = 0;
number_of_inserts = 0;

ply1 = ply;

if(proof_length) {ply1--;}

for(i = 1 ; i <= ply1 ; i++) {

/*--------------------------------------*/
/* transform clauses in clist to rows   */      
/* in the cl-array                      */
/*--------------------------------------*/

   c = clist[depth[i].d_tc];
   cid = c->c_id + number_of_inserts; 
   clpar = c->c_indices >> LEFT_INDEX_SHIFT;
   hr_lit_elim = 0;
   if((c->c_indices & RIGHT_INDEX_MASK) == (c->c_indices >> 6)
	   && c->c_rpar == c->c_lpar) {

/*--------------------------------------*/
/* clause was instantiated first.       */
/* names to change are in c->c_change.  */
/*--------------------------------------*/

      hr_lit_elim = clpar;
      cl[top_ptr][0] = c->c_id << 16;
      cl[top_ptr][1] = 0;
      cl[top_ptr][2] = 0;
      cl[top_ptr][3] = 0;
      cl[top_ptr][5] = c->c_change;
      cl[top_ptr][6] = 0;
      cl[top_ptr][7] = 0;
      cl[top_ptr][8] = 0;
      cl[top_ptr][9] = 0;
      cl[top_ptr][10] = 0;
      cl[top_ptr++][11] = 0;

/*-------------------------------------*/
/* add a second row                    */
/*-------------------------------------*/

       cl[top_ptr][0]=(c->c_id << 16)+(c->c_lpar->c_id<<4)+clpar;
       cl[top_ptr][1] = 
                (c->c_rpar->c_id<<4)+(c->c_indices&RIGHT_INDEX_MASK);
       cl[top_ptr][2] = 0;
       cl[top_ptr][3] = 0;
       cl[top_ptr][5] = -1;
       cl[top_ptr][6] = 0;
       cl[top_ptr][7] = 0;
       cl[top_ptr][8] = hr_hashcodes[2][i];
       cl[top_ptr][9] = -hr_hashcodes[0][i];
       cl[top_ptr][10] = -hr_hashcodes[1][i];
       cl[top_ptr][11] = 0;
       index_ctr = 0;
       for(litt = c->c_lit; litt != 0; litt = litt->l_next) {
          if(++index_ctr == (c->c_indices & RIGHT_INDEX_MASK)) {break;}}
       if(!elim_base_clause_hc(top_ptr,litt)) {
          elim_same_cont_hc(top_ptr,i);} 
       top_ptr = fix_fxn_evals(c,top_ptr);
       top_ptr++;
       goto add_row;}

/*-------------------------------------*/
/* add row for basic binary resolution */
/* or binary factor.                   */
/*-------------------------------------*/

   cl[top_ptr][0] = (c->c_id << 16) + (c->c_lpar->c_id << 4) + clpar;
   cl[top_ptr][3] = 0;
   cl[top_ptr][5] = 0;
   cl[top_ptr][6] = 0;
   cl[top_ptr][7] = 0;
   cl[top_ptr][11] = 0;
   if(c->c_rpar) {
      cl[top_ptr][1] = 
          (c->c_rpar->c_id << 4) + (c->c_indices & RIGHT_INDEX_MASK);}
   else {
      cl[top_ptr][1] = c->c_indices & RIGHT_INDEX_MASK;}
      
   cl[top_ptr][2] = 0;
   top_ptr = fix_fxn_evals(c,top_ptr);
   top_ptr++;

/*------------------------------------------*/
/* now take care of each ELIMLIT literal    */
/* add a row to the cl array for each       */
/* literal in clause c that is represented  */
/* by its hash code.                        */
/*------------------------------------------*/

add_row:
   l_index = 0;
   real_index = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      l_index++;
      if(!(l->l_id & ELIMLIT)) real_index++;
      if(l->l_id & ELIMLIT && l_index != hr_lit_elim) {  
         number_of_inserts++;
	 cid = (cl[top_ptr - 1][0] >> 16);            
	 clpar = l_index;               
	 for(j = top_ptr - 1; cl[j][2] != 0; j--) {
	    cl2par = (cl[j][0] & INDEX_MASK);
	    if(cl2par < clpar) {clpar--;}}
			
	 cl[top_ptr][0] = ((cid + 1)<<16) + (cid<<4) 
				   + real_index + 1;
         cl[top_ptr][3] = 0;
         cl[top_ptr][5] = 0;
         cl[top_ptr][6] = 0;
         cl[top_ptr][7] = 0;
         cl[top_ptr][11] = 1;
         cl[top_ptr][4] = 0;
	 cl[top_ptr][1] = hxtemp = -l->l_X;
	 cl[top_ptr][2] = hytemp = -l->l_Y;
	 cl[top_ptr][13] = (l->l_id & CONSMASK) >> 20;
	 cl[top_ptr][13] += ((l->l_id & CONSMASK2) >> 8);
	 if(!elim_base_clause_hc(top_ptr,l)) {
	    if(!elim_same_cont_hc(top_ptr,i)) {
	       //normalize_literal(l,c); 
	       calc_lit_hc(l,c);        
	       cl[top_ptr][1] = -l->l_X;
	       cl[top_ptr][2] = -l->l_Y;
	       cl[top_ptr][13] = (l->l_id & CONSMASK) >> 20;
	       cl[top_ptr][13] += ((l->l_id & CONSMASK2) >> 8);
	       if(!elim_base_clause_hc(top_ptr,l)) {
	          elim_same_cont_hc(top_ptr,i);}}}
         if(cl[top_ptr][2] != -1) {
	    l->l_X = -hxtemp;
	    l->l_Y = -hytemp;
            cl[top_ptr][1] = hxtemp;
	    cl[top_ptr][2] = hytemp;
	    cl[top_ptr][13] = (l->l_id & CONSMASK2) >> 8;
	    cl[top_ptr][13] += ((l->l_id & CONSMASK) >> 20);}
         top_ptr++;}}}
 
/*----------------------------------------------*/
/* for secondary lines, add unit clause         */
/*----------------------------------------------*/

if(phase == 2) {
   c = c_killer;
   hr_lit_elim = 0;
   cid = c->c_id + number_of_inserts; 
   if(((c->c_indices & RIGHT_INDEX_MASK) == 
	   (c->c_indices >> 6)) && c->c_rpar == c->c_lpar) {
      clpar = c->c_indices >> LEFT_INDEX_SHIFT;
      hr_lit_elim = clpar;
      cl[top_ptr][0]= c->c_id << 16;
      cl[top_ptr][1] = 0;
      cl[top_ptr][2] = 0;
      cl[top_ptr][3] = 0;
      cl[top_ptr][5] = c->c_change;
      cl[top_ptr][6] = 0;
      cl[top_ptr][7] = 0;
      cl[top_ptr][8] = 0;
      cl[top_ptr][9] = 0;
      cl[top_ptr][10] = 0;
      cl[top_ptr++][11] = 0;

/*---------------------------------*/
/* add a second row.               */
/*---------------------------------*/

      cl[top_ptr][0]=(c->c_id << 16)+(c->c_lpar->c_id<<4)+clpar;
      cl[top_ptr][1] = 
                (c->c_rpar->c_id<<4)+(c->c_indices&RIGHT_INDEX_MASK);
      cl[top_ptr][2] = 0;
      cl[top_ptr][3] = 0;
      cl[top_ptr][5] = -1;
      cl[top_ptr][6] = 0;
      cl[top_ptr][7] = 0;
      cl[top_ptr][8] = hr_hashcodes[2][i];
      cl[top_ptr][9] = -hr_hashcodes[0][i];
      cl[top_ptr][10] = -hr_hashcodes[1][i];
      cl[top_ptr][11] = 0;
      index_ctr = 0;
      for(litt = c->c_lit; litt != 0; litt = litt->l_next) {
         if(++index_ctr == (c->c_indices & RIGHT_INDEX_MASK)) {break;}}
      if(!elim_base_clause_hc(top_ptr,litt)) {
         elim_same_cont_hc(top_ptr,i);} 
      top_ptr = fix_fxn_evals(c,top_ptr);
      top_ptr++;
      goto add_row2;}
	
   clpar = c->c_indices >> LEFT_INDEX_SHIFT;
   cl[top_ptr][0] = (c->c_id << 16) + (c->c_lpar->c_id << 4) + clpar;
   cl[top_ptr][2] = 0;
   cl[top_ptr][3] = 0;
   cl[top_ptr][4] = 0;
   cl[top_ptr][5] = 0;
   cl[top_ptr][6] = 0;
   cl[top_ptr][7] = 0;
   cl[top_ptr][11] = 0;
   if(c->c_rpar) {
      cl[top_ptr][1] = 
         (c->c_rpar->c_id << 4) + (c->c_indices & RIGHT_INDEX_MASK);}
   else {cl[top_ptr][1] = c->c_indices & RIGHT_INDEX_MASK;}
   top_ptr = fix_fxn_evals(c,top_ptr);
   top_ptr++;
 
/*------------------------------------------*/
/* now take care of missing literals from   */
/* killer clause of secondary line          */
/*------------------------------------------*/
 
 add_row2:
   l_index = 0;
   real_index = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      l_index++;
      if(!(l->l_id & ELIMLIT)) real_index++;
      if((l->l_id & ELIMLIT) && hr_lit_elim != l_index) {  
         number_of_inserts++;
	 cid = cl[top_ptr - 1][0] >> 16;
	 clpar = l_index;
	 if(top_ptr) {
	    for(j = top_ptr - 1; cl[j][2] != 0; j--) {
	           cl2par = (cl[j][0] & INDEX_MASK);
	           if(cl2par < clpar) {clpar--;}}}
         cl[top_ptr][0] = ((cid + 1)<< 16) + (cid<< 4) 
				     + real_index + 1;
         cl[top_ptr][3] = 0; 
         cl[top_ptr][5] = 0; 
         cl[top_ptr][6] = 0; 
         cl[top_ptr][7] = 0; 
         cl[top_ptr][4] = 0; 
	 cl[top_ptr][1] = hxtemp = -l->l_X;
	 cl[top_ptr][2] = hytemp = -l->l_Y;
         cl[top_ptr][11] = 1; 
	 cl[top_ptr][13] = (l->l_id & CONSMASK) >> 20;
	 cl[top_ptr][13] += (l->l_id & CONSMASK2) >> 8;
	 if(!elim_base_clause_hc(top_ptr,l)) {
	    if(!elim_same_cont_hc(top_ptr,i)) {
	       //normalize_literal(l,c); 
	       calc_lit_hc(l,c);        
	       cl[top_ptr][1] = -l->l_X;
	       cl[top_ptr][2] = -l->l_Y;
	       cl[top_ptr][13] = (l->l_id & CONSMASK2) >> 8;
	       cl[top_ptr][13] += (l->l_id & CONSMASK) >> 20;
	       if(!elim_base_clause_hc(top_ptr,l)) {
	          elim_same_cont_hc(top_ptr,i);}}}

         if(cl[top_ptr][2] != -1) {
	    l->l_X = -hxtemp;
	    l->l_Y = -hytemp;
            cl[top_ptr][1] = hxtemp;
	    cl[top_ptr][2] = hytemp;}
	 top_ptr++;}}
   cl[top_ptr - 1][3] = killer_hash_ptr;}
 
/*--------------------------------------------------*/
/* if main line ends when complement found in hash  */
/* table, add extra line for this magic resolvent   */
/*--------------------------------------------------*/

if(!proof_length) {
   c = clist[depth[ply].d_tc];
   if(c->c_nlit == 1) {
      if(!proof_length && magic_hashx) {
         cl[top_ptr][0] = ((cl[top_ptr - 1][0] >> 16) + 1) <<16; 
         cl[top_ptr][0]= cl[top_ptr][0]+(cl[top_ptr][0]<<4)+1;
         cl[top_ptr][1] = -magic_hashx;
         cl[top_ptr][2] = -magic_hashy;
	 cl[top_ptr][3] = i;
	 cl[top_ptr][4] = 0;
	 cl[top_ptr][5] = 0;
	 cl[top_ptr][6] = 0;
	 cl[top_ptr][7] = 0;
	 cl[top_ptr][11] = 1;
	 for(alit = c->c_lit ; ; alit = alit->l_next) {
	    if(!(alit->l_id & ELIMLIT)) break;}
         if(!elim_base_clause_hc(top_ptr,alit)) {
	    elim_same_cont_hc(top_ptr,i);}              
         top_ptr++;}}}
line_length = top_ptr;
 
/*---------------------------------------------------*/
/* renumber parents; first right then left           */
/*---------------------------------------------------*/

for(i = 1; i < line_length; i++) {
 
/*----------------------------------*/
/* renumber right parents           */
/*----------------------------------*/
 
   if(!cl[i][2]) {
      cr = (cl[i][1] >> 4) & PARENT_MASK;
      crindex = cl[i][1] & INDEX_MASK;
      if(cr > base + 1) {
         k = base + 1; 
         for(j = base + 1;  ; j++) {
            if(!cl[j - base - 1][2] && !cl[j - base - 1][4]) {
               k++; 
               if(k > cr) {break;}}} 
	  cl[i][1] = (j << 4) + crindex;}}

/*----------------------------------*/
/* renumber left parents            */
/*----------------------------------*/

   killer = cl[i][3];
   cid = ((cl[i-1][0]) >> 16) + 1;
   clpar = cid - 1;
   clindex = (cl[i][0] & INDEX_MASK);
   cl[i][0] =  (cid<<16) + (clpar<<4) + clindex;}
   cl[i][3] = killer;
 
/*----------------------------------*/
/* unmark hashes to base clauses    */
/*----------------------------------*/

for(i = 1 ; i < line_length ; i++) {
   if(cl[i][2] == -1) cl[i][2] = 0;}
 
proof_length += line_length;

current_proof_top = PROOF_ARRAY_SIZE_M1 - proof_length;
move_down_renumber(line_length);

p_size = PROOF_ARRAY_SIZE;

for(i = clarray_top; i < p_size; i++) {
   if(cl[i][4] != 0) {
      if(cl[i][6] == killer_hashx && cl[i][7] == killer_hashy) {
	 cl[i][7] = 0;
         cl[i][6] = p_size - proof_length +line_length - 1;}}}

for(i = p_size-1; i >= p_size-proof_length; i--) {
   if(cl[i][6] != 0 && cl[i][7] != 0) {
      j = cl[i][4] & 0x0000FFFF;
      for(ij = 0; ij <= base; ij++) {
         c = clist[ij];
	 if(c->c_X == cl[i][6] && c->c_Y == cl[i][7]) {
            cl[i][6] = ij + 1;
	    cl[i][7] = 0;}}}}

if(o_debug) {
   print_cl_array(p_size - proof_length,0);
   if(o_debug > 1) getchar();}

//printf("C%d: Save line %d Lines:%d To Find:%d N: %d T%d\n",
//this_computer,proof_length,
//hash_code_list_top,hash_code_active_number,s_nod,time_now(0) - s_starttime);
return;
}
 
/*---------------------------------------------------------------*/
/*     fix_fxn_evals()                   .................#3
/*
/*     adds clauses based on function evaluations
/*
/*     called by:  save_line()              pass2.c
/*
/*     calls:
/*
/*     returns:  NIL
/*---------------------------------------------------------------*/

int fix_fxn_evals(c,t_ptr)
clause *c;
int t_ptr;
{
literal *lit;
int a, *p, lit_end, cid, lit_number, lit_number2;
int i, s_ptr;
int delete_number, resol_lit;

resol_lit = 0;
if(((c->c_indices & RIGHT_INDEX_MASK) == 
	   (c->c_indices >> 6)) && c->c_rpar == c->c_lpar) {
   resol_lit = c->c_indices & RIGHT_INDEX_MASK;}

lit_number2 = 1;
lit_number = 1;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(lit->l_id & EVALMASK) {
      p = &lit->l_id;
      for(a = *p++ & ARGMASK; a > 0; a--) {
         if(*p & FCNMASK) {a += *p++ & ARGMASK;}
         else {p++;}}

      lit_end = (int) lit->l_next;
      if(!lit_end) {lit_end = (int) c->c_end;}

      while((unsigned int) p < (unsigned int) lit_end) {
         if(*p == 0) {p++;continue;}
	 cid = cl[++t_ptr - 1][0] >> 16;
	 cl[t_ptr][0] = cid + 1 << 16;
         cl[t_ptr][1] = lit_number2;
	 cl[t_ptr][2] = 0;
	 cl[t_ptr][3] = 0;
	 cl[t_ptr][4] = *p;

/*--------------------------------------------*/
/* determine # characters to delete           */
/*--------------------------------------------*/

         delete_number = 0;
         s_ptr = cl[t_ptr][4] & 0x0000FFFF;
         cl[t_ptr][6] = sub_array[s_ptr][3];
	 cl[t_ptr][7] = sub_array[s_ptr][4];
	 cl[t_ptr][12] = sub_array[s_ptr][5]; 
         i = 0;
         while(s_ptr != 0) {
            if(sub_array[i][1] == s_ptr) {
               s_ptr = i;
               i = 0;
               delete_number++;}
            else if(sub_array[i][2] == s_ptr) {
               s_ptr = i;
               i = 0;}
            else i++;}
         cl[t_ptr][5] = delete_number;
         p++;}}
   if(lit_number2 == resol_lit) resol_lit = 0;
   else lit_number2++;
   lit_number++;}
return(t_ptr);
}



/*---------------------------------------------------------------*/
/*      elim_base_clause_hc()              ................#4
/*
/*      fix up cl entries which have hash codes to base clauses
/*
/*      parameters: top_ptr = row in cl_array with missing literal
/*
/*      called by:  save_line()                     pass2.c
/*
/*      calls:      test_for_base_match()           pass2.c
/*                  calc_alt_hash()                 hash.c
/*                  test_equality()                 pass2.c
/*                  normalize_literal()             hash.c
/*
/*      returns:    0 => no match; 1 => match
/*---------------------------------------------------------------*/
 
int elim_base_clause_hc(top_ptr,ll)
int top_ptr;
literal *ll;
{
clause *c;
literal *l;
int i, k1, k2, m, varbl, varbl2;

for(i = 0 ; i <= base ; i++) {
   c = clist[i];
   if(c->c_nlit == 1) { 
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) break;}
      normalize_literal(l,c); 

/*----------------------------------------------*/
/* test hash code for match with base clause    */
/*----------------------------------------------*/
 
      if(test_for_base_match(i,l->l_X,l->l_Y,top_ptr)) {
         if(o_debug > 0) {
            printf("\nHash code (HX: %d HY: %d) same as base clause:",
			     l->l_X,l->l_Y);
            OutPtr += sprintf(OutPtr,
		   "\nHash code (HX: %d HY: %d) same as base clause:",
			     l->l_X,l->l_Y);
	    print_clause(c);}
	 return(1);}
 
      if(c->c_nv == 1)  {
	 for(k1 = 1; k1 <= number_constants; k1++) {
	    calc_alt_hash(l,0,-1,k1,0);
	    if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
               if(o_debug) {
                  printf("\nHash code same as instance of base clause:");
                  OutPtr += sprintf(OutPtr,
		       "\nHash code same as instance of base clause:");
                  print_clause(c);}
               return(1);}}
 
         for(m = 0; m <= 3; m++) {
            calc_alt_hash(l,0,-1,1 - rows_less_nfxns + m,0);
	    if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
               if(o_debug) {
                  printf("\nHC %d %d same as instance of base clause:",alt_hx,alt_hy);
                  print_clause(c);}
	       return(1);}}}
 
/*----------------------------------------------*/
/* if base clause is not a ground instance      */
/* check alternate hash codes                   */
/*----------------------------------------------*/
 
      if(c->c_nv >= 2) {
	 for(varbl = c->c_nv - 2; varbl < c->c_nv ; varbl++) {
	    for(k1 = 1; k1 <= number_constants; k1++) {
	       calc_alt_hash(l,varbl,-1,k1,0);
	       if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
               if(o_debug) {
                  printf("\nHC %d %d same as instance of base clause:",alt_hx,alt_hy);
                  print_clause(c);}
               return(1);}}}}

      if(c->c_nv >= 2 && c->c_nv < 6) {
	 for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
            for(varbl2 = varbl + 1 ; varbl2 < c->c_nv; varbl2++) {
               for(k1 = 1; k1 <= number_const_hash; k1++) {
	          for(k2 = 1; k2 <= number_const_hash; k2++) {
	             calc_alt_hash(l,varbl,varbl2,k1,k2);
	             if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
                        if(o_debug) {
                           printf("\nHC %d %d same as instance of base clause:",alt_hx,alt_hy);
                           print_clause(c);}
	            return(1);}}}}}}}}

if(equal_pred && (ll->l_id & 0x8000FFFF) == (equal_pred | 0x80000000)) { 
   if(test_equality(ll)) {
      if(equal_clist_index != -2) {
         cl[top_ptr][1] = ((equal_clist_index + 1) << 4) + 1;}
      else cl[top_ptr][1] = 0;
      cl[top_ptr][2] = -1;
      cl[top_ptr][11] = 2;  /* indicates equal not present! */
      if(o_debug) {
         printf(
	    "\nEquality Predicate (Clause %d) eliminated literal",
	       equal_clist_index);
	 printf(" in clause %d\n",top_ptr);
         OutPtr += sprintf(OutPtr,
	    "\nEquality Predicate (Clause %d) eliminated literal",
	       equal_clist_index);
	 OutPtr += sprintf(OutPtr," in clause %d\n",top_ptr);}
      return(1);}}

return(0);
}
 
/*---------------------------------------------------------------*/
/*      elim_same_cont_hc()               ................#5
/*
/*      fix up cl entries which have hash codes to clauses on same
/*      continuation
/*
/*      parameters: top_ptr = row in cl_array with missing literal
/*                  pli     = length of current continuation - 1
/*
/*      called by:  save_line()                     pass2.c
/*
/*      calls:      test_for_cont_match()           pass2.c
/*                  calc_alt_hash()                 hash.c
/*
/*      returns:    0 => not elim; 1 => eliminated 
/*------------------------------------------------------------------*/
 
int elim_same_cont_hc(top_ptr,pli)
int top_ptr, pli;
{
clause *c;
literal *l;
int i, m, k1, k2, varbl, varbl2;
 
for(i = 1 ; i < pli ; i++) { 
   c = clist[depth[i].d_tc]; 
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l->l_id & ELIMLIT; l = l->l_next) { };
			
/*--------------------------------------------------*/
/* test hash code for match with contuation clause  */
/*--------------------------------------------------*/

      normalize_literal(l,c); 
      if(test_for_cont_match(i,l->l_X,l->l_Y,top_ptr)) {
         if(o_debug) {
            printf("\nHash code (HX: %d HY: %d) same as clause on line:"
		      ,l->l_X,l->l_Y);
            OutPtr += sprintf(OutPtr,
		   "\nHash code (HX: %d HY: %d) same as clause on line:"
		      ,l->l_X,l->l_Y);
            print_clause(c);}
         return(1);}

/*----------------------------------------------*/
/* if base clause is not a ground instance      */
/* check alternate hash codes                   */
/*----------------------------------------------*/
 
      if(c->c_nv == 1) {
         for(k1 = 1; k1 <= number_constants; k1++) {
	    calc_alt_hash(l,0,-1,k1,0);  
	    if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)) {
               if(o_debug) {
                  printf("\nHC %d %d same as clause on line:",
		      l->l_X,l->l_Y);
                  OutPtr += sprintf(OutPtr,"\nHC %d %d same as clause on line:",
		      l->l_X,l->l_Y);
                  print_clause(c);}
	       return(1);}}
 
         for(m = 0; m <= 3; m++) {
            calc_alt_hash(l,0,-1,1 - rows_less_nfxns + m,0);  
            if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
               if(o_debug) {
                  printf("\nHC %d %d same as clause on line:",
			  l->l_X,l->l_Y);
                  OutPtr += sprintf(OutPtr,"\nHC %d %d same as clause on line:",
			  l->l_X,l->l_Y);
                  print_clause(c);}
               return(1);}}}
 
       if(c->c_nv >= 2) {
            for(varbl = c->c_nv - 2; varbl < c->c_nv; varbl++) {
	           for(k1 = 1; k1 <= number_constants ; k1++) {
	              calc_alt_hash(l,varbl,-1,k1,0);
	              if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)) {
             if(o_debug) {
                printf("\nHC %d %d same as instance of clause on line:",
                               alt_hx,alt_hy);
              OutPtr += sprintf(OutPtr,"\nHC %d %d same as instance of clause on line:",
                               alt_hx,alt_hy);
              print_clause(c);}
             return(1);}}}}
       
         if(c->c_nv >= 2 && c->c_nv < 6) {  
            for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
	         for(varbl2 = varbl + 1; varbl2 < c->c_nv; varbl2++) {
	              for(k1 = 1; k1 <= number_const_hash ; k1++) {
	                 for(k2 = 1; k2 <= number_const_hash; k2++) {
	                    calc_alt_hash(l,varbl,varbl2,k1,k2);
	                    if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)){
                               if(o_debug) {
                                  printf("\nHC %d %d same as instance of clause on line:",
                                  alt_hx,alt_hy);
                                  OutPtr += sprintf(OutPtr,
                                     "\nHC %d %d same as instance of clause on line:",
                                   alt_hx,alt_hy);
                                 print_clause(c);}
                               return(1);}}}}}}}}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      test_for_base_match()             ................#6
/*
/*      parameters: c_index = the index of the base clause
/*                              under comparison
/*                  hc1,hc2 = hashcodes of missing literal
/*                  top_ptr = row in cl_array with missing lit
/*
/*      called by:  elim_base_clause_hc()             pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    0 if no match found, 1 if match found
/*-------------------------------------------------------------*/
 
int test_for_base_match(c_index,hc1,hc2,top_ptr)
int c_index, hc1, hc2, top_ptr;
{

if(cl[top_ptr][1] == hc1 && cl[top_ptr][2] == hc2) {
   cl[top_ptr][1] = ((c_index +1) << 4) + 1;
   cl[top_ptr][2] = -1;
   return(1);}
if(cl[top_ptr][9] == hc1 && cl[top_ptr][10] == hc2) {
   cl[top_ptr][1] = ((c_index +1) << 4) + 1;
   cl[top_ptr][2] = -1;
   cl[top_ptr][8] = -1;
   cl[top_ptr][9] = cl[top_ptr][10] = 0;
   return(1);}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      test_for_cont_match()              ................#7
/*
/*      called by:  elim_cont_clause_hc()          pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    0 => no match; 1 => match
/*-------------------------------------------------------------*/
 
int test_for_cont_match(c_index,hc1,hc2,top_ptr)
int c_index, hc1, hc2, top_ptr;
{
clause *c;
literal *l;
int additional_rows, j, litend, a, *p;

if((cl[top_ptr][1] == hc1 && cl[top_ptr][2] == hc2) ||
   (cl[top_ptr][9] == hc1 && cl[top_ptr][10] == hc2)) {
   additional_rows = 0;
   for(j = 1; j <= c_index; j++) {
      c = clist[depth[j].d_tc];
      for(l = c->c_lit; l != 0; l = l->l_next) {

/*-----------------------------------------*/
/* add rows for fxn evaluations            */
/*-----------------------------------------*/

         if(l->l_id & EVALMASK) {
            if(l->l_next == 0) {litend = (int) c->c_end;}
            else {litend = (int) l->l_next;}

            p = &l->l_id;
            for(a = *p++ & ARGMASK; a > 0; a--) {
               if(*p & FCNMASK) {a += *p++ & ARGMASK;}
               else {p++;}}
            while((int) p != litend) {
               if(*p != 0) {additional_rows++;}
               p++;}} 

/*-----------------------------------------*/
/* add rows for hash resolutions           */
/*-----------------------------------------*/

         if(l->l_id & ELIMLIT) {
            additional_rows++;}}}

   c_index = c_index + additional_rows; 
   cl[top_ptr][1] = ((c_index + base + 1) << 4) + 1;
   cl[top_ptr][2] = -1;
   if(cl[top_ptr][9] == hc1 && cl[top_ptr][10] == hc2) {
      cl[top_ptr][8] = -1;
      cl[top_ptr][9] = cl[top_ptr][10] = 0;}
   return(1);}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      stack_hash_codes()            ....................#8
/*
/*      put hash codes of lits that must be found on next pass
/*      on hash_code_list
/*
/*      called by:   main()          theo.c
/*
/*      calls:       print_hash_code_list()     pass2.c
/*
/*      returns:     nil
/*-------------------------------------------------------------*/
 
void stack_hash_codes()
{
int i, j, m;

/*-------------------------------------------------------------*/
/* if hash code for missing lit in clause, stack it for        */
/* latter discovery. do not stack it if it is already on       */
/* hash_code_list                                              */
/*-------------------------------------------------------------*/
								
for(i = current_proof_top ; i < PROOF_ARRAY_SIZE ; i++) {
   if(cl[i][2])  {                

/*----------------------------------------------------------*/
/* literal in this clause is represented by its hash code   */
/*----------------------------------------------------------*/
      for(j = 0; j <= hash_code_list_top ; j++)  { 
 
/*----------------------------------------------------------*/
/* if already on hash_code_list, do not add a second time   */
/*----------------------------------------------------------*/
 
	     if(cl[i][1] == hash_code_listx[j]) {
	        if(cl[i][2] == hash_code_listy[j]) {
		   if(hash_code_list_active[j] == 0) {

		      if(o_debug > 0) {
         printf("\nOn list; reactivate hashcode %d <%d %d> from clause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);
         OutPtr += sprintf(OutPtr,
		"\nOn list, reactivate hashcode %d <%d %d> from clause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);}

                      hash_code_active_number++; 
		      hash_code_list_active[j] = 1;}
	           break;}}
	     if(!hash_code_listx[j]) {
			
/*----------------------------------------------------------*/
/* add hash code of literal at end of hash_code_list, inc   */
/* the top of hash_code_list and fix up the j-th bit of the */
/* hash_code_list_active word                               */
/*----------------------------------------------------------*/

	        hash_code_listx[j] = cl[i][1];
	        hash_code_listy[j] = cl[i][2];
		hash_code_listu[j] = s_nod - s_nod0;

                if(o_debug > 0) {
printf("\nNot on list; add hashcode %d <%d %d> from clause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);
OutPtr += sprintf(OutPtr,"\nNot on list, add hashcode %d <%d %d> from clause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);}

	        hash_code_list_top++; 
		hash_code_active_number++;
	        hash_code_list_active[j] = 1;
	        break;}}}}

/*----------------------------------------------------------*/
/* now add hash codes of hr_hashcodes to find               */
/*----------------------------------------------------------*/

for(i = current_proof_top ; i < PROOF_ARRAY_SIZE ; i++) {
   if(cl[i][8])  {                
	
/*----------------------------------------------------------*/
/* literal in this clause is resolved away by hr_resolutn   */
/*----------------------------------------------------------*/

      for(j = 0; j <= hash_code_list_top ; j++)  { 
 
/*----------------------------------------------------------*/
/* if already on hash_code_list, do not add a second time   */
/*----------------------------------------------------------*/
 
	     if(cl[i][9] == hash_code_listx[j]) {
	        if(cl[i][10] == hash_code_listy[j]) {

		   if(cl[i][8] > 0) {
		   if(hash_code_list_active[j] == 0) {

		      if(o_debug > 0) {
         printf("\nOn list; reactivate hashcode %d <%d %d> from HRclause %d\n",
         j,hash_code_listx[j],hash_code_listy[j],i);
         OutPtr += sprintf(OutPtr,
		"\nOn list, reactivate hashcode %d <%d %d> from HRclause %d\n",
         j,hash_code_listx[j],hash_code_listy[j],i);}

                      hash_code_active_number++;
		      hash_code_list_active[j] = 1;}}
	           break;}}
	     if(!hash_code_listx[j]) {
			
/*----------------------------------------------------------*/
/* add hash code of literal at end of hash_code_list, inc   */
/* the top of hash_code_list and fix up the j-th bit of the */
/* hash_code_list_active word                               */
/*----------------------------------------------------------*/

	        hash_code_listx[j] = cl[i][9];
	        hash_code_listy[j] = cl[i][10];
		hash_code_listu[j] = s_nod - s_nod0;

                if(o_debug > 0) {
         printf("\nNot on list; add hashcode %d <%d %d> from HRclause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);
         OutPtr += sprintf(OutPtr,
		"\nNot on list; add hashcode %d <%d %d> from HRclause %d\n",
         hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],i);}

	        hash_code_list_top++; 
		hash_code_active_number++;
	        hash_code_list_active[j] = 1;
	        break;}}}}

/*----------------------------------------------------------*/
/* now add hash codes of equal predicates to find           */
/*----------------------------------------------------------*/

for(i = current_proof_top ; i < PROOF_ARRAY_SIZE ; i++) {
   if(cl[i][5] && cl[i][7])  {                

/*----------------------------------------------------------*/
/* literal in this clause is represented by its hash code   */
/*----------------------------------------------------------*/
      for(j = 0; j <= hash_code_list_top ; j++)  { 
 
/*----------------------------------------------------------*/
/* if already on hash_code_list, do not add a second time   */
/*----------------------------------------------------------*/
 
             m = cl[i][4] & 0x0000FFFF;
	     if(sub_array[m][3] == hash_code_listx[j]) {
	        if(sub_array[m][4] == hash_code_listy[j]) {

		   if(hash_code_list_active[j] == 0) {

                if(o_debug > 0) {
printf("\nOn list; reactivate hashcode %d <%d %d> from SA %d used by clause %d\n",
   j,hash_code_listx[j],hash_code_listy[j],m,i);
OutPtr += sprintf(OutPtr,
       "\nOn list; reactivate hashcode %d <%d %d> from SA %d used by clause %d\n",
   j,hash_code_listx[j],hash_code_listy[j],m,i);}


 		      hash_code_active_number++; 
		      hash_code_list_active[j] = 1;}
	           break;}}
	     if(!hash_code_listx[j]) {
			
/*----------------------------------------------------------*/
/* add hash code of literal at end of hash_code_list, inc   */
/* the top of hash_code_list and fix up the j-th bit of the */
/* hash_code_list_active word                               */
/*----------------------------------------------------------*/

	        hash_code_listx[j] = sub_array[m][3];
	        hash_code_listy[j] = sub_array[m][4];
		hash_code_listu[j] = s_nod - s_nod0;

                if(o_debug > 0) {
printf("\nNot on list; add hashcode %d <%d %d> from SA %d used by clause %d\n",
   hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],m,i);
OutPtr += sprintf(OutPtr,
       "\nNot on list; add hashcode %d <%d %d> from SA %d used by clause %d\n",
   hash_code_list_top+1,hash_code_listx[j],hash_code_listy[j],m,i);}

	        hash_code_list_top++; 
		hash_code_active_number++;
	        hash_code_list_active[j] = 1;
	        break;}}}}
clarray_top = current_proof_top + 1;

if(o_debug) {
   print_hash_code_list();}
return;
}
 
/*-------------------------------------------------------------*/
/*      print_hash_code_list()           ................#9
/*
/*      used as a debugging tool; prints the contents of the
/*      missing literal hash_code_list
/*
/*      called by:      print_cl_array()             pass2.c
/*                      stack_hash_codes()           pass2.c
/*                      halted_search_options()      tdump.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*-------------------------------------------------------------*/
 
void print_hash_code_list()
{
int i, j, to_find_count;

print0("\nHASH CODE LIST");

to_find_count = 0;

for(i = 0; i < hash_code_list_top ; i++) {
   if(hash_code_list_active[i]) {to_find_count++;}}

printf("   To Discover[D]: %d  Found[F]: %d",
   to_find_count,hash_code_list_top - to_find_count);
OutPtr += sprintf(OutPtr,"   To Discover[D]: %d  Found[F]: %d",
   to_find_count,hash_code_list_top - to_find_count);

for(i = 0; i < hash_code_list_top ; i++) {
   print0("\n");
   if(hash_code_list_active[i]) {
      printf("D[%d]:%d %d L%d N%d ",
	 i+1,hash_code_listx[i],hash_code_listy[i],hash_code_listz[i],
	 hash_code_listu[i]);
      OutPtr += sprintf(OutPtr,"D[%d]:%d %d L%d N%d ",
	 i+1,hash_code_listx[i],hash_code_listy[i],hash_code_listz[i],
	 hash_code_listu[i]);}
   else {
      printf("F[%d]:%d %d L%d N%d ",
	 i+1,hash_code_listx[i],hash_code_listy[i],hash_code_listz[i],
	 hash_code_listu[i]);
      OutPtr += sprintf(OutPtr,"F[%d]:%d %d L%d N%d ",
	 i+1,hash_code_listx[i],hash_code_listy[i],hash_code_listz[i],
	 hash_code_listu[i]);}}

print0("\nEND_HASH_CODE_LIST\n\n");
}                       
		
 
/*-------------------------------------------------------------*/
/*      print_cl_array()                ................#10
/*
/*      called when debugging program
/*
/*      called by:  save_line()                   pass2.c
/*
/*      calls:      print_hash_code_list()        pass2.c
/*
/*      CH == cl[i][5]
/*      H  == cl[i][6] and cl[i][7]
/*      HR == cl[i][8]
/*      H  == cl[i][9], cl[i][10]
/*      R  == if cl[i][11] == 1
/*      E  == if cl[i][11] == 2
/*      X  == cl[i][12] if cl[i][12] != 0
/*      S  == cl[i][13] if cl[i][13] != 0
/*      returns:    nil
/*-------------------------------------------------------------*/
 
void print_cl_array(top_of_proof,specified_row)
int top_of_proof, specified_row;
{       
int i,ttop1, ttop2;
 
print0("\nCL_ARRAY");
		
if(specified_row == 0) {
   ttop1 = top_of_proof;
   ttop2 = PROOF_ARRAY_SIZE;}
else {
   ttop1 = specified_row;
   ttop2 = ttop1 + 1;}
   
for(i = ttop1; i < ttop2 ; i++) {
   if(!cl[i][0]) {break;}
   if(!cl[i][2]) {         
      printf(
      "\n%d  L %4d,%d  R %4d,%d H %2d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                 /* block  */
      (cl[i][0] >> 4) & PARENT_MASK,              /* l_par   */
      (cl[i][0] & INDEX_MASK),                    /* index1  */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      cl[i][3],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      printf(" HR %d H %d,%d",cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {printf(" R");}
      if(cl[i][11] == 2) {printf(" E");}
      if(cl[i][12] != 0) {printf(" X%d",cl[i][12]);}
      if(cl[i][13] != 0) {printf(" S%x",cl[i][13]);}
      OutPtr += sprintf(OutPtr,
      "\n%d  L %4d,%d  R %4d,%d  H %2d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par */
      (cl[i][0] & INDEX_MASK),                      /* index1 */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      cl[i][3],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      OutPtr += sprintf(OutPtr," HR %d H %d,%d",cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {OutPtr += sprintf(OutPtr," R");}
      if(cl[i][11] == 2) {OutPtr += sprintf(OutPtr," E");}
      if(cl[i][12] != 0) {OutPtr += sprintf(OutPtr," X%d",cl[i][12]);}
      if(cl[i][13] != 0) {OutPtr += sprintf(OutPtr," S%x",cl[i][13]);}}
 
   else if(cl[i][2] && phase != 3) {              
      printf(
      "\n%d  L %4d,%d  H %d %d  H %2d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block  */
      (cl[i][0] >> 4) & PARENT_MASK,                /* l_par   */
      (cl[i][0] & INDEX_MASK),                      /* index1  */
      cl[i][1],
      cl[i][2],
      cl[i][3],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      printf(" HR %d H %d,%d",cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {printf(" R");}
      if(cl[i][11] == 2) {printf(" E");}
      if(cl[i][12] != 0) {printf(" X%d",cl[i][12]);}
      if(cl[i][13] != 0) {printf(" S%x",cl[i][13]);}
			
      OutPtr += sprintf(OutPtr,
      "\n%d  L %4d,%d  H %d %d  H %d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block   */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par   */
      (cl[i][0] & INDEX_MASK),                       /* index1  */
      cl[i][1],
      cl[i][2],
      cl[i][3],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      OutPtr += sprintf(OutPtr," HR %d H %d,%d",
		   cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {OutPtr += sprintf(OutPtr," R");}
      if(cl[i][11] == 2) {OutPtr += sprintf(OutPtr," E");}
      if(cl[i][12] != 0) {OutPtr += sprintf(OutPtr," X%d",cl[i][12]);}
      if(cl[i][13] != 0) {OutPtr += sprintf(OutPtr," S%x",cl[i][13]);}}
			
   else {
      printf(
      "\n%d  L %4d,%d  R %4d,%d H %d  I: %d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                    /* block   */
      (cl[i][0] >> 4) & PARENT_MASK,                  /* l_par   */
      (cl[i][0] & INDEX_MASK),                        /* index1  */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      cl[i][3],
      cl[i][2],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      printf(" HR %d H %d,%d",cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {printf(" R");}
      if(cl[i][11] == 2) {printf(" E");}
      if(cl[i][12] != 0) {printf(" X%d",cl[i][12]);}
      if(cl[i][13] != 0) {printf(" S%x",cl[i][13]);}
 
      OutPtr += sprintf(OutPtr,
      "\n%d  L %4d,%d  R %4d,%d H %d  I: %d  ST %8x CH %x H %d,%d",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block     */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par     */
      (cl[i][0] & INDEX_MASK),                       /* index1    */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      cl[i][3],
      cl[i][2],
      cl[i][4],
      cl[i][5],
      cl[i][6],
      cl[i][7]);
      OutPtr += sprintf(OutPtr," HR %d H %d,%d",
		   cl[i][8],cl[i][9],cl[i][10]);
      if(cl[i][11] == 1) {OutPtr += sprintf(OutPtr," R");}
      if(cl[i][11] == 2) {OutPtr += sprintf(OutPtr," E");}
      if(cl[i][12] != 0) {OutPtr += sprintf(OutPtr," X%d",cl[i][12]);}
      if(cl[i][13] != 0) {OutPtr += sprintf(OutPtr," S%x",cl[i][13]);}}}
 
print0("\n");
if(specified_row != 0) return;
if(phase != 1) {print_hash_code_list();} 
return;         
}
 
/*--------------------------------------------------------------*/
/*      on_hash_code_list(hcx,hcy)         ................#11
/*
/*      see if complement of hcx,hcy is on hash_code_list
/*
/*      called by:   install_clause()        search.c
/*
/*      calls:       nobody
/*
/*      returns:     0 if hash code not on hash_code_list
/*                   1 if hash code on hash_code_list
/*-------------------------------------------------------------*/
 
int on_hash_code_list(hcx,hcy)
int hcx, hcy;
{
int i;
 
for( i = 0; i <= hash_code_list_top; i++) {
   if(hcx == hash_code_listx[i] && hcy == hash_code_listy[i]) {
      if(hash_code_list_active[i] == 1) {
			
/*----------------------------------------------------------*/
/*  if found on hash_code_list, delete by setting           */
/*  hash_code_list_active[i] to 0.                          */
/*----------------------------------------------------------*/
		
	     hash_code_list_active[i] = 0;
	     hash_code_listu[i] = s_nod - s_nod0;
	     hash_code_listz[i] = PROOF_ARRAY_SIZE - proof_length - 1;
	     hash_code_active_number--;
	     killer_hashx = hcx; 
	     killer_hashy = hcy;
	     killer_hash_ptr = i + 1;

             if(o_debug) {
                printf("\nHash code (HX: %d HY: %d) %d found\n",
			hcx,hcy,i + 1);
                OutPtr += sprintf(OutPtr,
	               "\nHash code (HX: %d HY: %d) %d found\n"
		       ,hcx,hcy,i + 1);
                printf("Remaining to find: %d\n",hash_code_active_number);
		OutPtr += sprintf(OutPtr,
		       "Remaining to find: %d\n",hash_code_active_number);}
	     return(i + 1);}}
   if(!hash_code_active_number) {return(0);}} 
return(0);
}
 
/*---------------------------------------------------------------*/
/*      alt_on_hash_code_list()    .......................#12
/*
/*      called by:   install_clause()            search.c
/*
/*      calls:       on_hash_code_list()         pass2.c
/*                   calc_alt_hash()             hash.c
/*
/*      returns:     0 if hash code not on hash_code_list
/*                   2 if hash code on hash_code_list
/*---------------------------------------------------------------*/
 
int alt_on_hash_code_list(l,c)
literal *l;
clause *c;
{
int m, k1, k2, varbl, varbl2;
int found;
 
found = 0;

if(c->c_nv == 1) { 
   for(k1 = 1; k1 <= number_constants; k1++) {
      calc_alt_hash(l,0,-1,k1,0);            
      if(on_hash_code_list(alt_hx,alt_hy)) {
         found = 2;}}
 
   for(m = 0; m <= 3; m++) {
      calc_alt_hash(l,0,-1,1 - rows_less_nfxns + m,0);     
      if(on_hash_code_list(alt_hx,alt_hy)) {
         found = 2;}}}

if(c->c_nv >= 2) {
   for(varbl = c->c_nv - 2; varbl < c->c_nv ; varbl++) {
      for(k1 = 1; k1 <= number_constants; k1++) {
	     calc_alt_hash(l,varbl,-1,k1,0);
	     if(on_hash_code_list(alt_hx,alt_hy)) {
	        found = 2;}}}}

if(c->c_nv  >= 2  && c->c_nv < 6) {
   for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
      for(varbl2 = varbl + 1; varbl2 < c->c_nv; varbl2++) {
	     for(k1 = 1; k1 <= number_const_hash; k1++) {
	        for(k2 = 1; k2 <= number_const_hash; k2++) {
	           calc_alt_hash(l,varbl,varbl2,k1,k2);
	           if(on_hash_code_list(alt_hx,alt_hy)) {
		       found = 2;}}}}}}

return(found);
}
 
/*---------------------------------------------------------------*/
/*      delete_found_hcs()               ................#13
/*
/*      deletes hash codes from the line just found for which the
/*      complement's hash code has been previously found and placed
/*      in the hash_array
/*
/*      called by:   save_line()          pass2.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void delete_found_hcs(hx,hy)
int hx,hy;
{
int m;

for(m = clarray_top; m < PROOF_ARRAY_SIZE; m++) {
   if(cl[m][1] == hx && cl[m][2] == hy) {
      cl[m][2] = 0;
      cl[m][1] = (current_proof_top << 4) + 1;}
   if(cl[m][8] > 0 && cl[m][9] == hx && cl[m][10] == hy) {
      cl[m][1] = (current_proof_top << 4) + 1;
      cl[m][8] = -cl[m][8];}
   if(cl[m][6] == hx && cl[m][7] == hy) {
      cl[m][6] = current_proof_top;
      cl[m][7] = 0;}}

on_hash_code_list(hx,hy);

return;
}

/*--------------------------------------------------------------*/
/*      delete_alt_hcs(c)   ................#14
/*
/*      calculates var in the formula, and for each constant
/*      in the list of constants for theorem, calculate a hash
/*      code and add it to the clause hash table
/*
/*      called by:      install_clause()               search.c
/*
/*      calls:          calc_alt_hash()                hash.c
/*                      delete_found_hcs()             pass2.c
/*
/*      returns:        0;2 if contradiction found
/*---------------------------------------------------------------*/
 
void delete_alt_hcs(c) 
clause *c;
{
int m, const1, const2, varbl, varbl2;
literal *l;
 
/*-----------------------------------------------*/
/* if there is one variable, replace it with     */
/* each constant and calculate the resulting     */
/* hash code.  add the hash code to the clause   */
/* hash table.                                   */
/*-----------------------------------------------*/

for(l = c->c_lit; (l->l_id & ELIMLIT) != 0; l = l->l_next) { };

if(c->c_nv == 1) {
   for(const1 = 1 ; const1 <= number_constants ; const1++) {
      calc_alt_hash(l,0,-1,const1,0);    
      delete_found_hcs(alt_hx,alt_hy);}
 
/*-----------------------------------------------*/
/* calc hashcode with single variable replaced   */
/* with the variable y.  most likely, but not    */
/* for sure, this will produce a new hashcode    */
/*-----------------------------------------------*/
 
   for(m = 0; m <= 3; m++) {
      calc_alt_hash(l,0,-1,1 - rows_less_nfxns + m,0);
      delete_found_hcs(alt_hx,alt_hy);}}
 
/*-----------------------------------------------*/
/* if there are two variables, replace each one  */
/* with a constant and calculate the hashcode    */
/*-----------------------------------------------*/
 
if(c->c_nv >= 2) {
   for(varbl = c->c_nv - 2; varbl < c->c_nv; varbl++) {
     for(const1 = 1; const1 <= number_constants; const1++) {
	    calc_alt_hash(l,varbl,-1,const1,0);
	    delete_found_hcs(alt_hx,alt_hy);}}}
			
/*------------------------------------------------*/
/* replace every pair of variables with every     */
/* pair of constants, calculate the hashcode      */
/* and add the hashcode to the literal hash table */
/*------------------------------------------------*/ 
 
if(c->c_nv >= 2 && c->c_nv < 6) {
  for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
   for(varbl2 = varbl + 1 ; varbl2 < c->c_nv; varbl2++) {
     for(const1 = 1; const1 <= number_const_hash; const1++) {
       for(const2 = 1; const2 <= number_const_hash; const2++) {
          calc_alt_hash(l,varbl,varbl2,const1,const2);
          delete_found_hcs(alt_hx,alt_hy);}}}}}
return;
}
 
/*---------------------------------------------------------------*/
/*      move_down_renumber()                ................#15
/*
/*      adds the continuation stored in the top rows of cl_array to
/*      those in the bottom where the previously-found parts of
/*      the proof have been stored.
/*
/*      called by:      save_line()                pass2.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void move_down_renumber(path_length)
int path_length;
{
int cid, clpar, clindex, crpar, crindex, j, m;
int t_ptr, b_ptr, diff, hash_pointer, old_bottom_ptr;
 
b_ptr = PROOF_ARRAY_SIZE - proof_length;
old_bottom_ptr = b_ptr + path_length;
 
for(t_ptr = 0 ; t_ptr < path_length ; t_ptr++) {
   if(cl[t_ptr][2]) {
      for(j = 0; j <= hash_code_list_top; j++) {
	     if(cl[t_ptr][1] == hash_code_listx[j] && 
	        cl[t_ptr][2] == hash_code_listy[j] &&
                s_nod - s_nod0 > hash_code_listu[j] &&
			   hash_code_list_active[j] == 0) {
	        cl[t_ptr][1] = (hash_code_listz[j]<< 4) + 1;
	        cl[t_ptr][2] = 0;
                break;}}}}

for(t_ptr = 0 ; t_ptr < path_length ; t_ptr++) {
   hash_pointer = cl[t_ptr][3];
   cid = b_ptr;
   clpar = (cl[t_ptr][0] >> 4) & PARENT_MASK;
   if(clpar > base + 1) {
      diff = base + 2 + t_ptr - clpar;
      clpar = b_ptr - diff;}
   clindex = (cl[t_ptr][0] & INDEX_MASK);
	
   if(!cl[t_ptr][2]) {
      crpar = (cl[t_ptr][1] >> 4) & PARENT_MASK;
      if(crpar > base + 1 && crpar<PROOF_ARRAY_SIZE - proof_length) {
	     diff = base + 2 + t_ptr - crpar;
	     crpar = b_ptr - diff;}
      crindex = (cl[t_ptr][1] & INDEX_MASK);
		
      cl[b_ptr][0] = 
	   (cid << 16) + (clpar << 4) + clindex;
      cl[b_ptr][3] = hash_pointer;
      cl[b_ptr][1] = (crpar << 4) + crindex;
      cl[b_ptr][2] = 0;
      cl[b_ptr][5] = cl[t_ptr][5];
      cl[b_ptr][6] = cl[t_ptr][6];
      cl[b_ptr][7] = cl[t_ptr][7];
      cl[b_ptr][8] = cl[t_ptr][8];
      cl[b_ptr][9] = cl[t_ptr][9];
      cl[b_ptr][10] = cl[t_ptr][10];
      cl[b_ptr][11] = cl[t_ptr][11];
      cl[b_ptr][12] = cl[t_ptr][12];
      cl[b_ptr][13] = cl[t_ptr][13];
      cl[b_ptr][4] = cl[t_ptr][4];}
	
   else {
      cl[b_ptr][0] = 
	    (cid << 16) + (clpar << 4) + clindex;
      cl[b_ptr][3] = hash_pointer;
      cl[b_ptr][1] = cl[t_ptr][1];
      cl[b_ptr][2] = cl[t_ptr][2];
      cl[b_ptr][5] = cl[t_ptr][5];
      cl[b_ptr][6] = cl[t_ptr][6];
      cl[b_ptr][7] = cl[t_ptr][7];
      cl[b_ptr][8] = cl[t_ptr][8];
      cl[b_ptr][9] = cl[t_ptr][9];
      cl[b_ptr][10] = cl[t_ptr][10];
      cl[b_ptr][11] = cl[t_ptr][11];
      cl[b_ptr][12] = cl[t_ptr][12];
      cl[b_ptr][13] = cl[t_ptr][13];
      cl[b_ptr][4] = cl[t_ptr][4];}
	
   cl[t_ptr][0] = cl[t_ptr][1] = cl[t_ptr][2] = cl[t_ptr][4] = cl[t_ptr][5] = 0;
   cl[t_ptr][6] = cl[t_ptr][7] = 0;	
   cl[t_ptr][8] = cl[t_ptr][9] = cl[t_ptr][10] = cl[t_ptr][11] = 0;	
   cl[t_ptr][12] = cl[t_ptr][13] = 0;
   b_ptr++;}
		
for(m = b_ptr; m < PROOF_ARRAY_SIZE ; m++) {
   if(cl[m][2] == -1) {
      cl[m][1] = ((old_bottom_ptr - 1) << 4) + 1;
      cl[m][2] = 0;}}
return;
}
 
/*-------------------------------------------------------------*/
/*      rearrange_cl_array()            ................#16
/*
/*      rearranges proof in the top rows of the cl_array when this
/*      function is entered, the proof is in the bottom of cl_array
/*      (rows from "PROOF_ARRAY_SIZE - proof_length" 
/*      to PROOF_ARRAY_SIZE).
/*
/*      called by:     main()           theo.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*-------------------------------------------------------------*/
 
void rearrange_cl_array()
{
int clause_id, m;
int start_index, bottom_of_block, j, k;
int not_ready, rpar, rpar_gone;
int cid, clpar, clindex, crpar, crindex;
int smallest, smallest_ptr, c0;
 
/*--------------------------------------*/
/* determine order of blocks            */
/*--------------------------------------*/

clause_id = 1;

while(clause_id <= proof_length) {
   start_index = PROOF_ARRAY_SIZE - proof_length;
   get_next_blocks:
   for(m = start_index ; m < PROOF_ARRAY_SIZE ; m++) {
      if(cl[m][3] || m == PROOF_ARRAY_SIZE_M1) {
         bottom_of_block = m;
         break;}}
   not_ready = 0;
   for(j = start_index; j <= bottom_of_block; j++) {

   /*-------------------------------------------------*/
   /* if a substitution, replace with clause          */
   /*-------------------------------------------------*/

      if((cl[j][1] >> 4) == 0 && (cl[j][1] & 0x0000000F) != 0
	 && (cl[j][0] &  0x0000000F) == 0) {
          cl[j][1] = (cl[j][6] << 4) + cl[j][1];}
      if(cl[j][2]) {not_ready++;}
      if((cl[j][1]>>4) > base+1 && ((cl[j][1]>>4) < start_index 
			|| (cl[j][1]>>4) > bottom_of_block)) {
         rpar = cl[j][1] >> 4;
         rpar_gone = cl[rpar][2];
         if(!rpar_gone) {
            not_ready++;}}} 
   if(!not_ready) {
      for(j = start_index; j <= bottom_of_block; j++) {
	     cl[j][2] = clause_id++;}}
   start_index = m + 1;
   if(m < PROOF_ARRAY_SIZE) {
      goto get_next_blocks;}}
	
/*----------------------------------*/
/* now move up to top of array      */
/*----------------------------------*/

cl[1599][11] = 0;
k = 0;
for(j = PROOF_ARRAY_SIZE - proof_length ; j < PROOF_ARRAY_SIZE ; j++) {
   cid = cl[j][2] + base + 1;
   clpar = (cl[j][0] >> 4) & PARENT_MASK;
   if(clpar > base + 1) {
      clpar = cid - 1;}
   clindex = cl[j][0] & INDEX_MASK;
   crpar = (cl[j][1] >> 4) & PARENT_MASK;
   if(crpar > base + 1) {
      crpar =  cl[crpar][2] + base + 1;}
   crindex = cl[j][1] & INDEX_MASK;
	
   cl[j][0] = (cid << 16) + (clpar << 4) + clindex;
   cl[j][1] = (crpar << 4) + crindex;}

for(j=PROOF_ARRAY_SIZE-proof_length ;j < PROOF_ARRAY_SIZE_M1 ;j++ ) { 
   smallest = cl[j][2];
   smallest_ptr = j;
   for(m = j + 1 ; m < PROOF_ARRAY_SIZE ; m++) {
      if(cl[m][2] < smallest) {
	     smallest = cl[m][2];
	     smallest_ptr = m;}}
   c0 = cl[j][0]; cl[j][0] = cl[smallest_ptr][0]; 
   cl[smallest_ptr][0] = c0;
   c0 = cl[j][1]; cl[j][1] = cl[smallest_ptr][1]; 
   cl[smallest_ptr][1] = c0;
   c0 = cl[j][2]; cl[j][2] = cl[smallest_ptr][2]; 
   cl[smallest_ptr][2] = c0;
   c0 = cl[j][3]; cl[j][3] = cl[smallest_ptr][3]; 
   cl[smallest_ptr][3] = c0;
   c0 = cl[j][4]; cl[j][4] = cl[smallest_ptr][4]; 
   cl[smallest_ptr][4] = c0;
   c0 = cl[j][5]; cl[j][5] = cl[smallest_ptr][5]; 
   cl[smallest_ptr][5] = c0;
   c0 = cl[j][6]; cl[j][6] = cl[smallest_ptr][6]; 
   cl[smallest_ptr][6] = c0;
   c0 = cl[j][7]; cl[j][7] = cl[smallest_ptr][7]; 
   cl[smallest_ptr][7] = c0;
   c0 = cl[j][8]; cl[j][8] = cl[smallest_ptr][8]; 
   cl[smallest_ptr][8] = c0;
   c0 = cl[j][9]; cl[j][9] = cl[smallest_ptr][9]; 
   cl[smallest_ptr][9] = c0;
   c0 = cl[j][10]; cl[j][10] = cl[smallest_ptr][10]; 
   cl[smallest_ptr][10] = c0;
   c0 = cl[j][11]; cl[j][11] = cl[smallest_ptr][11]; 
   cl[smallest_ptr][11] = c0;
   c0 = cl[j][12]; cl[j][12] = cl[smallest_ptr][12]; 
   cl[smallest_ptr][12] = c0;
   c0 = cl[j][13]; cl[j][13] = cl[smallest_ptr][13]; 
   cl[smallest_ptr][13] = c0;}

k = 0;
for(j = PROOF_ARRAY_SIZE - proof_length ;j < PROOF_ARRAY_SIZE ;j++) {
   cl[k][0] = cl[j][0]; cl[j][0] = 0;
   cl[k][1] = cl[j][1]; cl[j][1] = 0;
   cl[k][2] = cl[j][2]; cl[j][2] = 0;
   cl[k][3] = cl[j][3]; cl[j][3] = 0;
   cl[k][4] = cl[j][4]; cl[j][4] = 0;
   cl[k][5] = cl[j][5]; cl[j][5] = 0;
   cl[k][6] = cl[j][6]; cl[j][6] = 0;
   cl[k][7] = cl[j][7]; cl[j][7] = 0;
   cl[k][8] = cl[j][8]; cl[j][8] = 0;
   cl[k][9] = cl[j][9]; cl[j][9] = 0;
   cl[k][10] = cl[j][10]; cl[j][10] = 0;
   cl[k][11] = cl[j][11]; cl[j][11] = 0;
   cl[k][12] = cl[j][12]; cl[j][12] = 0;
   cl[k][13] = cl[j][13]; cl[j][13] = 0;
   k++;}

if(o_debug > 0) {
   printf("\nDone with rearrange_cl_array\n");
   print_cl_array(0,0);}

return;
}               
 
/*-------------------------------------------------------------*/
/*      bres_proof_line(c1,c2,l1_int,l2_int)...............#17
/*
/*      parameters:  c1,c2:  pointers to the clause to resolve
/*                   ix1,ix2: first and second indices of factor
/*
/*      called by:   infer_proof_line()             pass2.c
/*
/*      calls:       standardize()                  infer.c
/*                   clear_var_substitution_tble()  infer.c
/*                   unify()                        unify.c
/*                   generate_clause()              gen.c
/*                   install_proof_line()           pass2.c
/*                   printf_cl_array()              pass2.c
/*                   print_clause()                 tdump.c
/*
/*      returns:     nil (exits from program on error)
/*-------------------------------------------------------------*/
 
void bres_proof_line(ci1,ci2,ix1,ix2,hr)
int ci1, ci2;
int ix1, ix2, hr;
{
literal *l1, *l2;
int i;
clause *c, *c1, *c2;
int iz1, iz2;
int tc;

c1 = clist[ci1 - 1];
c2 = clist[ci2 - 1];

iz1 = 0;
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   if(l1->l_id & ELIMLIT) continue;
   iz1++;
   if(iz1 == ix1) break;}

iz2 = 0;
for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
   if(l2->l_id & ELIMLIT) continue;
   iz2++;
   if(iz2 == ix2) break;}

standardize(c1,c2);             /* standardize variables apart  */
clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);
if(unify(&l1->l_id,&l2->l_id)) {
   if(hr == 0) {
      if(c = generate_clause(c1,c2,l1,l2,ix1,ix2)) { 
         install_proof_line(c);
         clist[nextc - 1]->c_id = nextc;
         return;}}
   else {
      if(c = hgenerate_clause(c1,c2,l1,l2,ix1,ix2)) { 
	 c->c_nv = c1->c_nv;
         install_proof_line(c);
         clist[nextc - 1]->c_id = nextc;
	 if(phase == 3) {
	    hgen_normalize(c);}
         return;}}}
      
/*--------------------------------------------*/
/* at this point, if the clauses cannot be    */
/* installed, there is a bug, most likely     */
/* a hashing error.                           */
/*--------------------------------------------*/
 
else {
   big_bug = 1;
   tc = this_computer;this_computer = 0;
   printf("\nBIG BUG!!..#nodes=%d\n",s_nod);fflush(stdout);
   OutPtr += sprintf(OutPtr,"\nBIG BUG!!..#nodes=%d\n",s_nod); 
   print_cl_array(0,0);  fflush(stdout);
   printf("Type CR to continue\n");fflush(stdout);
   getchar(); 
/*  for(i = 0; i <= base; i++) {
      print_clause(clist[i]);  
      if((i == org_base) && (org_base != base)) {
	     print0("Inferred clauses:\n");}    
      if(i == (org_base - nc))  { 
	     print0("Negated conclusion:\n");}} 
*/
   printf("Type CR to continue\n");fflush(stdout);
// getchar();	
   printf("Proof:\n");fflush(stdout);
	
/*----------------------------------*/
/* write out the proof clauses      */
/*----------------------------------*/
	
   for(i = 1; i <= ply; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_lpar->c_id != c->c_id - 1 && i != 1) {
	     printf("\n");fflush(stdout);}
      if(c->c_lpar != 0) print_clause(c->c_lpar);fflush(stdout);
      if(c->c_rpar != 0) print_clause(c->c_rpar);fflush(stdout);
      print_clause(c);fflush(stdout);
      printf("=========Type CR to continue===========\n\n");
      fflush(stdout);}
 
   printf("\n");fflush(stdout);
   print_clause(c1);fflush(stdout);print_clause(c2);fflush(stdout);
   printf("Type CR to continue\n");fflush(stdout);
// zz = getchar(); 
   fatal(" ");}
}
 
/*---------------------------------------------------------------*/
/*      factor_proof_line(c1,ix1,ix2)     ................#18
/*
/*      parameters:   c1:  pointer to the clause to factor
/*                    ix1,ix2: first and second indices of factor
/*
/*      called by:    infer_proof_line()                pass2.c
/*
/*      calls:        clear_var_substitution_tble()     infer.c
/*                    unify()                           unify.c
/*                    generate_clause()                 gen.c
/*                    install_proof_line()              pass2.c
/*
/*      returns:      nil
/*---------------------------------------------------------------*/
 
void factor_proof_line(c1,ix1,ix2)
int  c1;                            /* index of clause to factor    */
int ix1, ix2;                       /* index of literals to factor  */
{
literal *l1, *l2;
clause *c;                          /* clause to factor             */
clause *new_clause;                 /* facotred clause (new clause) */
int lit_index;
int iz1, iz2;
	
c = clist[c1 - 1];                  /* clause to factor             */
	
/*----------------------------------------------*/
/* scan for literals corresponding to ix1, ix2  */
/*----------------------------------------------*/
 
iz1 = 0;
for(l1 = c->c_lit; l1 != 0; l1 = l1->l_next) {
   if(l1->l_id & ELIMLIT) continue;
   iz1++;
   if(iz1 == ix1) break;}

if(ix1 != 0) {
   iz2 = 0;
   for(l2 = c->c_lit; l2 != 0; l2 = l2->l_next) {
      if(l2->l_id & ELIMLIT) continue;
      iz2++;
      if(iz2 == ix2) break;}}
else {
   l2 = l1;
   ix2 = ix1;}
			
clear_var_substitution_tble(c->c_bits,c->c_nv);
	
if(unify(&l1->l_id,&l2->l_id)) {
   new_clause = generate_clause(c,0,l1,0,ix1,ix2);    
   if(c->c_bits & MERGE) {new_clause->c_bits |= MERGE;}
   install_proof_line(new_clause);
   clist[nextc - 1]->c_id = nextc;}
 
return;
}

/*---------------------------------------------------------------*/
/* hash_resolve() ..................................... #19
/*
/* called by:
/*
/* calls:
/*
/* returns:
/*---------------------------------------------------------------*/

void hash_resolve(c1,c2,c1_index,c2_index)
int c1, c2, c1_index, c2_index;
{
clause *c, *c_new;
literal *l1, *l2, *l3, *llit;
int new_constant, new_constant2, *pp, aa, indexx, l_index;

c = clist[c1 - 1];
c_new = copy_clause(c,0,c1_index);

if(c2_index == 0) {
   c_new->c_indices = (c1_index<<6) + c1_index;
   c_new->c_rpar = clist[c1 - 1];}
else { 
   c_new->c_indices = (c1_index<<6) + c2_index;
   c_new->c_rpar = clist[c2 - 1];}

c_new->c_lpar = clist[c1 - 1];
c_new->c_nlit = c->c_nlit - 1;
c_new->c_nv = c->c_nv;
c_new->c_bits = (c->c_bits | ACTIVE) & ~MERGE;

/*--------------------------------------------------*/
/* replace constant into entire clause              */
/*--------------------------------------------------*/

if((new_constant = (cl[ply - 1][13] & 0x000000FF)) != 0) {
l_index = 0;
for(llit = c_new->c_lit;llit != 0;llit = llit->l_next) {
   l_index++;
   if(l_index == cl[ply - 1][0] & INDEX_MASK) {
      llit->l_id += (new_constant & 0x000000FF) << 20;}
   pp = &llit->l_id;
   for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
      if(*pp & FCNMASK) {aa += *pp++ & ARGMASK;}
      else {
         if(*pp == 0) {
            *pp++ = (new_constant << 6) | NEGMASK;}}}
    calc_lit_hc(llit,c);}}

/*--------------------------------------------------*/
/* replace 2nd constant into entire clause          */
/*--------------------------------------------------*/

if((new_constant2 = ((cl[ply - 1][13] & 0x00000F00) >> 8)) != 0) {
l_index = 0;
for(llit = c_new->c_lit;llit != 0;llit = llit->l_next) {
   l_index++;
   if(l_index == cl[ply - 1][0] & INDEX_MASK) {
      llit->l_id += new_constant2 << 16;}
   pp = &llit->l_id;
   for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
      if(*pp & FCNMASK) {aa += *pp++ & ARGMASK;}
      else {
         if(*pp == 1) {*pp++ = (new_constant2 << 6) | NEGMASK;}}}
    calc_lit_hc(llit,c);}}

/*---------------------------------------------------------------*/
/* treat "equalish" here
/*---------------------------------------------------------------*/

c_new->c_change = 0;
calc_clause_hashcode(c_new);
normalize_literal(c_new->c_lit,c_new);
install_proof_line(c_new);
clist[nextc - 1]->c_id = nextc;
}

/*---------------------------------------------------------------*/
/*     instantiate_clause()     ................#20
/*
/*     called by:     infer_proof_line()               pass2.c
/*
/*     calls:
/*
/*     returns:       nil
/*---------------------------------------------------------------*/

void instantiate_clause(c_index)
int c_index;
{
clause *c, *c_new;
int substitution_info;
int var1, var2, con1, con2;
static char var_[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";

c = clist[c_index - 1];
substitution_info = cl[ply - 1][5];
var1 = substitution_info & 0x000000FF;
con1 = (substitution_info >> 8) & 0x000000FF;
var2 = (substitution_info >> 16) & 0x000000FF;
con2 = (substitution_info >> 24) & 0x000000FF;

c_new = copy_clause(c,0,0);

v1_count = 1;
if(var2 != 0) v2_count = 1;
if(substitution_info) replace_vars(c_new,substitution_info);
//if(substitution_info) normalize_literal(c_new->c_lit,c_new);
//install_proof_line(c_new);
clist[nextc - 1]->c_id = nextc;
c_new->c_lpar = clist[nextc - 1];
c_new->c_change = substitution_info;
c_new->c_indices = 0;
c_new->c_bits = ACTIVE;
install_proof_line(c_new);
return;}

/*---------------------------------------------------------------*/
/*      substitute_string()     ................#21
/*
/*      called by:    infer_proof_line()                pass2.c
/*
/*      calls:        clear_var_substitution_tble()     infer.c
/*                    install_proof_line()              pass2.c
/*
/*      returns:      nil
/*---------------------------------------------------------------*/
 
void substitute_string(c1)
int c1;
{
int lit_index, lit_ptr, *string_ptr;
int lit_end, delete_number, delete_char_position;
int replace_with;
literal *lit;
clause *c, *c_new;
int *arg_scan, *arg_scan2;
int s_ptr;
int i, ii;
int c2, ply_temp;
int to_fix, fixed;

/*----------------------------------*/
/* get clause to fix up             */
/*----------------------------------*/


ply_temp = ply;
c2 = c1;

to_fix = 0;
while((cl[ply_temp - 1][0] & INDEX_MASK) == 0) {
   c2--;
   to_fix++;
   ply_temp--;}
c2++;

c = clist[c2 - 1];
c_new = copy_clause(c,0,0);

c_new->c_lpar = clist[nextc - 1];
c_new->c_rpar = clist[(cl[ply - 1][1] >> 4) - 1];
c_new->c_indices = ((cl[ply - 1][1] & 0x0000000F) << 6) + 1;
c_new->c_bits |= EVALMASK;
		
/*----------------------------------*/
/* determine literal to repair      */
/*----------------------------------*/

fixed = 0;
while(to_fix > 0) {
   to_fix--;
   fixed = 1;
   lit_index = cl[ply_temp][1] & 0x0000000F;
   lit_ptr = 1;

/*-----------------------------------*/
/* scan for literal to modify        */
/*-----------------------------------*/

   for(lit = c_new->c_lit; lit != 0; lit = lit->l_next) {
      if(lit_ptr++ == lit_index) {break;}}

   lit_end = (int) lit->l_next;
   if((int) lit_end == 0) {lit_end = (int) c_new->c_end;}

   delete_char_position = (cl[ply_temp][4] & 0xFF000000) >> 24;
   s_ptr = cl[ply_temp][4] & 0x0000FFFF;
   arg_scan = &lit->l_id;
   for(i = 1; ;i++) {
      arg_scan++;
      if(i == delete_char_position) {break;}}

   replace_with = (cl[ply_temp][4] &0x00FF0000) >> 16;

   if(replace_with == 0x000000FF) {goto now_delete;}
   if(replace_with & 0x00000080) {
      replace_with = ((replace_with & 0x0000007F) << 5) | 0x80000000;}
   *arg_scan++ = replace_with;
now_delete:
   arg_scan2 = arg_scan;

   delete_number = cl[ply_temp][5];
   while(delete_number-- > 1) {arg_scan2++;}
   while((unsigned int) arg_scan2 < (unsigned int) lit_end) {
      *arg_scan++ = *arg_scan2++;}
   ply_temp++;}

if(cl[ply_temp][4] == 0 && cl[ply_temp][11] == 0 && c_new->c_nlit == 1) {
   for(lit = c_new->c_lit; lit != 0; lit = lit->l_next) {
      if((lit->l_id & ELIMLIT) == 0) break;}
   normalize_literal(lit,c_new);}
install_proof_line(c_new);
clist[nextc - 1]->c_id = nextc;
return;}

/*-----------------------------------------------------------*/
/*      install_proof_line(c)             .............#22
/*
/*      install clause c on the clist
/*
/*      called by:  bres_proof_line()       pass2.c
/*                  fact_proof_line()       pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*-----------------------------------------------------------*/
 
void install_proof_line(c)
clause *c;
{
dnode *d;
literal *lit;
	
d = &depth[ply];                       /* insert at current ply */
c->c_id = base + ply + 1 ;            /* give clause an ID     */
d->d_tc = nextc++;

clist[nextc] = c->c_end;
 
if((unsigned int) c->c_end > (unsigned int) d->d_rc) {
   d->d_rc = c->c_end;}

if(c->c_lpar->c_bits & NC) {c->c_bits |= NC;}
if(c->c_rpar != 0 && (c->c_rpar->c_bits & NC) != 0) {c->c_bits |= NC;}

clist[d->d_tc] = c;
return;
}
  
/*---------------------------------------------------------------*/
/*      infer_proof_line()            ................#23
/*
/*      determine next inference (factor or binary resol) as indicated
/*      by the cl_array
/*
/*      called by:  search_proof_line()   pass2.c
/*
/*      calls:      bres_proof_line()     pass2.c
/*                  factor_proof_line()   pass2.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void infer_proof_line()
{
int c1, c2;                      /* index of clauses to resolve  */
int c1_index, c2_index;          /* index of literals to resolve */

/*----------------------*/
/* decode cl info       */
/*----------------------*/
	
c1       = (cl[ply - 1][0] >> 4) & PARENT_MASK;
c1_index = (cl[ply - 1][0] & INDEX_MASK);
c2           = (cl[ply - 1][1] >> 4) & PARENT_MASK;
c2_index = (cl[ply - 1][1] & INDEX_MASK);


//printf("INFER PROOF LINE: ");
//print_clause(clist[c1 -1]);
//if(c2 != 0) print_clause(clist[c2-1]);

if(cl[ply - 1][11] > 0 && clist[c1 - 1]->c_nlit > 1) {
//printf("BRES....\n");
   bres_proof_line(c1,c2,c1_index,c2_index,cl[ply - 1][11]);
   return;}

if(cl[ply - 1][5] == -1) {
//printf("HR...\n");
   hash_resolve(c1,c2,c1_index,c2_index);
   return;}

if(c2 == 0 & c2_index == 0) {
//printf("INSTSNTIATE...\n");
   instantiate_clause(c1);
   return;}

if(c1_index == 0) {
//printf("SUB STIRNG...\n");
   substitute_string(c1,c1_index);
   return;}

else if(c2 == 0) {
//printf("FSCTOR...\n");
   factor_proof_line(c1,c1_index,c2_index);
   return;}

else {
//printf("BRES2....\n");
   bres_proof_line(c1,c2,c1_index,c2_index,cl[ply - 1][11]);}
return;
}                                                       

/*---------------------------------------------------------------*/
/* clean_proof()                        ................#24
/*---------------------------------------------------------------*/

void clean_proof()
{
int i, j, jj, k, m, n, y, clx, new_cls;
int clid, left_parent, right_parent, root;
int delete_cls, hash_number, clause_number, clause_number2;

/*---------------------------------------*/
/* determine which clasues can be deleted*/
/* if i can be deleted, cl[i][12] points */
/* to replacement clause.                */
/*---------------------------------------*/

return;
for(i = 0; i < proof_length; i++) {cl[i][12] = 0;}

for(i = 0; i < proof_length - 1; i++) {
   if(cl[i][3] == 0) {continue;}
   hash_number = cl[i][3];
   clause_number = (cl[i][0] >> 16) & CIDMASK2;
   for(j = i + 1; j < proof_length - 1; j++) {
      if(cl[j][3] == hash_number) {

/*---------------------------------------*/
/* another line with same hash_number    */
/* mark line for deletion by setting     */
/* cl[j][2] to 0 for clauses on line     */
/* to be deleted.                        */
/*---------------------------------------*/

         cl[j][2] = 0;
	 clause_number2 = (cl[j][0] >> 16) & CIDMASK2;
	 k = j - 1;
         while(cl[k][3] == 0) {cl[k--][2] = 0;}

/*----------------------------------------*/
/* replace parent of inferences that are  */
/* clause_number2 with clause_number.     */
/*----------------------------------------*/

         for(jj = j + 1; jj < proof_length; jj++) {
            if(((cl[jj][0] >> 4) & PARENT_MASK) == clause_number2) {
               cl[jj][0] = ((cl[jj][0] & 0xFFFF000F) | 
				       (clause_number << 4));}
            if(((cl[jj][1] >> 4) & PARENT_MASK) == clause_number2) {
               cl[jj][1] = ((cl[jj][1] & 0xFFFF000F) | 
					(clause_number << 4));}}}}}

/*---------------------------------------*/
/* renumber parents of other clauses     */
/*---------------------------------------*/

if(o_debug > 0) {
   printf("Done with Clean Array 1\n");
   print_cl_array(0,0);}

delete_cls = 0;
for(i = proof_length - 1; i > 0 ; i--) {
   if(cl[i][2] != 0) {continue;}
   root = cl[i][0] >> 16;
   for(j = i; j < proof_length - delete_cls - 1; j++) {
      left_parent = (cl[j+1][0] & 0x0000FFF0) >> 4; 
      if(left_parent >= root) {left_parent--;}
      cl[j][0] = (cl[j][0] & 0xFFFF0000) | (left_parent << 4);
      cl[j][0] = cl[j][0] | (cl[j+1][0] & 0x0000000F);

      right_parent = (cl[j+1][1] & 0x0000FFF0) >> 4; 
      if(right_parent >= root) {right_parent--;}
      cl[j][1] = (cl[j][1] & 0xFFFF0000) | (right_parent << 4);
      cl[j][1] = cl[j][1] | (cl[j+1][1] & 0x0000000F);
      for(k = 2; k < 14; k++) {cl[j][k] = cl[j + 1][k];}}
   for(k = 0; k < 14; k++ ) {cl[j][k] = 0;}
   delete_cls++;}

if(o_debug > 0) {
   printf("Done with Clean Array 2\n");
   print_cl_array(0,0);}

proof_length = proof_length - delete_cls;
deleted_clauses = 0;
for(i = proof_length; i > 0; i--) {
   for(j = i - 1; j >= 0; j--) {
        if((cl[i][0] & 0x0000FFFF) == (cl[j][0] & 0x0000FFFF) 
	       && cl[i][1] == cl[j][1]) {
	    cl[i][12] = cl[j][0] >> 16;
	    deleted_clauses++;
            m = i; n = j;
top2:       m++;n++;
            if((cl[m][0] & 0x0000000F) == (cl[n][0] & 0x0000000F)
	       && cl[m][1] == cl[n][1] 
	       &&(cl[m][0]&0x0000FFF0)>>4==(cl[m-1][0]&0xFFFF0000)>>16
	       &&(cl[n][0]&0x0000FFF0)>>4==(cl[n-1][0]&0xFFFF0000)>>16){
	       cl[m][12] = cl[n][0] >> 16;
	       deleted_clauses++;
	       goto top2;}
	    else break;}}}

if(o_debug > 0) {
   printf("Done with Clean Array 3\n");
   print_cl_array(0,0);}

for(i = proof_length; i > 0; i--) {
   if(((cl[i][0] & 0x0000FFF0)>>4) == (cl[i][0]>>16) - 1) {
      if(cl[i][12] == 0 && cl[i - 1][12] != 0) {
         new_cls = cl[i - 1][12] - base - 1;
         while(cl[new_cls - 1][12] != 0 && new_cls >= 0) {
            new_cls = cl[new_cls - 1][12] - base - 1;}
         cl[i][0] = (cl[i][0]&0xFFFF000F) | ((new_cls+base+1) << 4);}}
clxx:
   clx = (cl[i - 1][1] & 0x0000FFF0) >> 4;
   if(clx <= base) y = 0;
   else y = cl[clx - base - 2][12];
   if(y != 0) {
      cl[i - 1][1] = (cl[i - 1][1] & 0xFFFF000F) + (y << 4);
      goto clxx;}} 

/*---------------------------------------*/
/* final clean -- delete redundant cls   */
/*---------------------------------------*/

/*if(o_debug > 0) {
   printf("Done with Clean Array 4\n");
   print_cl_array(0,0);}

for(i = proof_length - 2; i > 0 ; i--) {
   if(cl[i][12] != 0) {
      for(j = i + 1; j <= proof_length; j++) {
         left_parent = (cl[j][0] & 0x0000FFF0) >> 4;
	 if(left_parent > i + base + 1) {
	    cl[j][0] = cl[j][0] - 16;}
	 cl[j - 1][0] = cl[j][0] - 65536;
         right_parent = ((cl[j][1] >> 4) & PARENT_MASK);
	 if(right_parent > i + base + 1) {cl[j][1] = cl[j][1] - 16;}
         cl[j - 1][1] = cl[j][1];}
      for(j = i; j <= proof_length - 1; j++) {
	 for(k = 2; k < 14; k++) {cl[j][k] = cl[j + 1][k];}}
      for(k = 0; k < 14; k++) {cl[proof_length][k] = 0;}}}
*/

if(o_debug > 0) {
   printf("Done with Clean Array\n");
   print_cl_array(0,0);}

}

/*---------------------------------------------------------------*/
/*      search_proof_line()             ................#25
/*
/*      search_proof_line generates a single path of inferences as
/*      guided by the cl[] array.  This is performed after phases
/*      1 and 2 are complete
/*
/*      called by:  main()                  theo.c
/*
/*      calls:      infer_proof_line()      pass2.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void search_proof_line()
{
dnode *d;
int o_maxterm_tempy, o_maxvar_tempy,o_maxlit_tempy;

rearrange_cl_array();
clean_proof();

depth[0].d_tc = base;               /* initial generating clause    */
depth[0].d_ec = base;               /* top clause at level 0 = base */
depth[0].d_bc = 0;                  /* bottom clause at level 0 = 0 */
	
nextc = base +1;                       /* start storing proof here  */

/*--------------------------------------------*/
/* determine where to begin storing pointers  */
/* to clauses in the proof                    */
/*--------------------------------------------*/ 

clist[nextc] = depth[0].d_rc = clist[base]->c_end;
 
/*----------------------------------------------*/
/* do not place a limit on the number of terms  */
/* in a literal in the proof                    */
/*----------------------------------------------*/
 
o_maxterm_tempy = o_maxterm;
o_maxterm = 100;
o_maxvar_tempy = o_maxvar;
o_maxlit_tempy = o_maxlit;
o_maxvar = 32;
o_maxlit = 16;

/*----------------------------------------------*/
/* pick off resolutions stored in the cl-array  */
/* carry them out to make sure that a hash err  */
/* has not led to an illegal resolution         */
/*----------------------------------------------*/

for(ply = 1; ply <= proof_length; ply++) {
   d = &depth[ply];
   nextc  = depth[ply - 1].d_ec + 1;  
   d->d_rc = 0;
   d->d_tc = d->d_bc = nextc;
   infer_proof_line();
   d->d_ec = depth[ply - 1].d_ec + 1;

/*----------------------------*/
/* break on null clause       */
/*----------------------------*/

   if(clist[d->d_tc]->c_nlit == 0) {
      o_maxterm = o_maxterm_tempy;
/*      o_maxvar = o_maxvar_tempy;
      o_maxlit = o_maxlit_tempy;
*/
      if(o_debug > 0) {print_cl_array(0,0);}
      return;}}
}
