------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Instantiate where

import Data hiding ( Type )
import List hiding ( insert, delete, union )
import Set
import AnalysisTypes
import Flags
import Trace

-------------------------------------------------------------------------
-- clause sets

data ClauseSet
  = ForAll [Clause]
  | ForAllNew Int [Clause]
 deriving (Show, Eq)
 
-------------------------------------------------------------------------
-- instantiate

instantiate :: Flags -> [(Type,Int)] -> [Clause] -> [QClause] -> [(Symbol,[ClauseSet])]
instantiate flags predefs cs qcs =
  [ (assump, ForAll symcs : cs)
  | ((assump,cs),symcs) <- parts 1 [] patterns nonGroundCs
                     `zip` symmetries
  ]
 where
  (groundCs, nonGroundCs) = partition isGround (sortBy siz cs)
   where
    c1 `siz` c2 = size c1 `compare` size c2

  iqcs = [1..] `zip` qcs
  
  (_, fs1, ps) = symbols cs
  (_, fs2, _)  = symbols qcs
  ncons        = size (set [ c | (c@(Elt _),0) <- elements (fs1 `union` fs2)])

  isGround c = size (free c) == 0
  
  con k = Elt k `app` []
  
  symmetries :: [[Clause]]
  symmetries =
    transp
    [ symmForType tp predef
    | (tp,predef) <- predefs
    ]
   where
    transp []  = repeat []
    transp xss = concat [ x | x:_ <- xss ] : transp [ xs | _:xs <- xss ]
   
    symmForType tp predef =
      zipWith (\k f -> f k) [1..]
        -- do not use symmetries before predef size
         $ replicate predef (\_ -> [])

        -- constant-triangle
        ++ [ \k ->
               [ set [ Pos (t `eq` con i) | i <- [1..k] ]
               ] ++
               [ set (Neg (t `eq` con k) :
                 [ Pos (t' `eq` con (k-1))
                 | t' <- ts
                 ])
               | kanin flags
               , let n = k-2-predef
	             allCons' | n >= 0    = drop n allCons
		              | otherwise = []
               , (t,ts) <- drop 1 allCons' `zip` tail (inits allCons')
               ]
           | t <- allCons
           ]

        -- predicate-symmetries
        ++ [ \k -> [] | not (null allUnitPreds)]
        ++ concat
           [ repeat $ \k ->
               [ set [ Neg (p `prd` [con k])
                     , Pos (p `prd` [con (k-1)])
                     ]
               ]
           | (p:_) <- [allUnitPreds]
           ]

        -- function-symmetries
        ++ concat
           [ repeat $ \k ->
               [ c
               | k > n
               , c <- [ set [ Pos ((f `app` [con (k-j) | j <- [1..n]]) `eq` con i)
                            | i <- [1..k]
                            ]
                      ] ++
                      [ set ( Neg ((f `app` [con (k-1)]) `eq` con k)
                            : Neg ((f `app` [con (k-2)]) `eq` con (k-1))
                            : [ Pos ((f `app` [con j])   `eq` con (k-2))
                              | j <- [numCons..k-2]
                              , j > 0
                              ]
                            )
                      | k > ((numCons + 2) `max` 3)
                      , n == 1
                      ]
               ]
           | ((f,n):_) <- [allFuns]
           , n > 0
           ]
     where
      numCons = predef + length allCons
      
      allCons =
        [ t
        | Uniq v c <- qcs
        , [Pos (Pred _ Equal [t@(App _ f []), Var v'])] <- [elements c]
        , v == v'
        , typ t == tp
        ]

      allFuns =
        sortBy cmp
        [ (f,length xs)
        | Uniq v c <- qcs
        , [Pos (Pred _ Equal [t@(App _ f (xs@(_:_))), Var v'])] <- [elements c]
        , v == v'
        , typ t == tp
        ]
       where
        (_,x) `cmp` (_,y) = x `compare` y

      allUnitPreds =
        [ p
        | (p,_) <- elements ps
        , typing p == PType [tp]
        ]

  parts :: Int -> [Term] -> [[Bool]] -> [Clause] -> [(Symbol,[ClauseSet])]
  parts k oldCons (pat:pats) cs =
    (dom, clauses) : parts (k+1) allCons pats (extra ++ cs)
   where
    dom  = domainName % k ::- PType []
    domk = dom `prd` []
    
    newCon  = con k
    allCons = newCon : oldCons
    
    qclauses =
      [ c
      | (i,qc) <- iqcs
      , c      <- atMostOne i qc
      ]

    extra =
      [ c
      | c <- qclauses
      , not (isGround c)
      ]
    
    clauses =
      [ ForAll
        ( -- constant equalities
             [ unit (Pos (newCon `eq` newCon)) ]
          ++ concat
             [ [ unit (Neg (c `eq` newCon))
               , unit (Neg (newCon `eq` c))
               ]
             | c <- oldCons
             ]

          -- ground clauses (only if k == 1)
          ++ [ c
             | k == 1
             , c <- groundCs
             ]
        )
      , ForAllNew k
        ( -- instantiate clauses
          cs
        )
      , ForAll
        ( -- uniqueness clauses
             [ c
             | qc <- qcs
             , c  <- atLeastOne qc
             ]

          ++ qclauses
        )
      ]
  
    atLeastOne :: QClause -> [Clause]
    atLeastOne (Uniq v c) =
      [ set $
          pre
          [ l
          | a <- allCons
          , l <- elements (subst (pair v a) c)
          ]
      | let VType t = typing v
      , pre <- case tdomain t of
                 Just k' | k' == k -> [id]
                         | k  > k' -> []
                 _                 -> [(Neg domk :)]
      ]
  
    atMostOne :: Int -> QClause -> [Clause]
    atMostOne i (Uniq v c) =
      [ set [neg l, a]
      | let VType t = typing v
      , case tdomain t of
          Just k' | k > k' -> False
          _                -> True
      , l <- ls
      , a <- pattern i
      ]
     where
      ls      = elements (subst (pair v newCon) c)
      vs      = elements (v `delete` free c)
      uni i j = (uniqueName % i % j ::- PType (map typ xs)) `prd` xs
      xs      = map Var vs

      pattern i =
        [ sgn (uni i j)
        | (j,b) <- [0..] `zip` pat
        , let sgn = if b then Pos else Neg
        ]

  a `eq` b = Equal `prd` [a,b]
  
-- unique non-empty non-overlapping infinite list of 0/1 patterns

patterns :: [[Bool]]
patterns = pats 4
 where
  pats n = init base ++ map (last base ++) (pats (n+1))
   where
    base = bits n

  bits 0 = [[]]
  bits k = [ False : bs | bs <- bits (k-1) ]
        ++ [ True  : bs | bs <- bits (k-1) ]

-------------------------------------------------------------------------
-- the end.
