------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Sat
  ( SolverM       -- :: * -> *; Monad
  , try           -- :: SolverM a -> SolverM (Maybe a)
  , lift          -- :: IO a -> SolverM a
  , run           -- :: SolverM a -> IO (Maybe a)
  , contradiction -- :: SolverM a

  , simplify      -- :: SolverM ()
  , solve         -- :: Maybe GLit -> SolverM Bool
  , dimacs        -- :: String -> Maybe GLit -> SolverM ()
  , value         -- :: GLit   -> SolverM Bool
  
  , getNClauses   -- :: SolverM Int
  , getNVars      -- :: SolverM Int
  , getNConflicts -- :: SolverM Int

  , setVerbose    -- :: Bool -> SolverM ()

  , identify      -- :: IO ()

  , Loc        -- :: *; Eq
  
  , GAtom(..)  -- :: *; Eq
  , GLit       -- :: *; Eq
  , GArg(..)   -- :: *; Eq
  
  , newLoc     -- :: Int -> SolverM Loc
  , getArity   -- :: Loc -> SolverM Int
  , addClauses -- :: Maybe Int -> [Int] -> [GLit] -> SolverM ()
  )
 where

import Data hiding ( sign )

import Foreign
import Foreign.C.String
import CTypes
import IO ( hFlush, stdout )
import Maybe ( fromMaybe, maybe )


newLoc :: Int -> SolverM Loc
newLoc p = lift $ do
  ptr  <- loc_new (fromIntegral p)
  fptr <- newForeignPtr loc_free ptr
  return (Loc fptr)


getArity :: Loc -> SolverM Int
getArity (Loc l) = lift $ withForeignPtr l (fmap fromIntegral . loc_arity)
  

addClauses :: Maybe Int -> [Int] -> [GLit] -> SolverM ()
addClauses f d ls = do
  r <- withSolver (addClauses_ f d ls)
  if r == 0
    then contradiction
    else return ()
  
addClauses_ f d ls s = do
    solver_clause_begin s
    mapM_ (sign addLit) ls
    mapM_ (solver_clause_add_size s . fromIntegral) d
    solver_clause_commit s (maybe 0 (\x -> fromIntegral (x-1)) f)

 where addArg (N i) = solver_clause_add_lit_con s (fromIntegral i)
       addArg (V i) = solver_clause_add_lit_var s (fromIntegral i)
       addLit (Loc l :@ args) b = do
         withForeignPtr l (flip (solver_clause_add_lit s) b)
	 mapM_ addArg args

-------------------------------------------------------------------------
-- types

newtype Loc = Loc (ForeignPtr ()) deriving Eq

data GAtom
  = !Loc :@ [GArg]
 deriving (Eq)

type GLit
  = Signed GAtom

data GArg
  = N !Int
  | V !Int
 deriving (Show, Eq)

newtype Solver
  = MkSolver (Ptr ())

-------------------------------------------------------------------------
-- misc

sign :: (GAtom -> CUInt -> a) -> GLit -> a
sign f (Pos a) = f a 1
sign f (Neg a) = f a 0

-------------------------------------------------------------------------
-- solver monad

newtype SolverM a
  = MkSolverM (forall b . Solver -> IO b -> (a -> IO b) -> IO b)

instance Monad SolverM where
  return x =
    MkSolverM (\_s _fail ok -> ok x)
    
  MkSolverM m >>= k =
    MkSolverM (\s fail ok ->
      m s fail (\y ->
        let MkSolverM m' = k y in
          m' s fail ok
      )
    )

-------------------------------------------------------------------------
-- primitive operations

try :: SolverM a -> SolverM (Maybe a)
try (MkSolverM m) =
  MkSolverM (\s _fail ok ->
    m s (ok Nothing) (ok . Just)
  )

solver :: SolverM Solver
solver = MkSolverM (\s _fail ok -> ok s)

lift :: IO a -> SolverM a
lift io = MkSolverM (\_s _fail ok -> io >>= ok)

withSolver :: (Solver -> IO a) -> SolverM a
withSolver f = MkSolverM (\s _fail ok -> f s >>= ok)

contradiction :: SolverM a
contradiction = MkSolverM (\_s fail _ok -> fail)

run :: SolverM a -> IO (Maybe a)
run (MkSolverM m) =
  do s <- solver_new
     let fail =
           do solver_free s
              return Nothing
         
         ok a =
           do solver_free s
              return (Just a)
     
      in m s fail ok

-------------------------------------------------------------------------
-- derived operations

simplify :: SolverM ()
simplify =
  do r <- withSolver solver_simplify
     if r == 0
       then contradiction
       else return ()

declAssumption :: Maybe GLit -> SolverM ()
declAssumption Nothing    = withSolver $ \s -> solver_lit_begin s nullPtr 0
declAssumption (Just lit) = withSolver $ \s -> sign (declAtom s) lit

declAtom :: Solver -> GAtom -> CUInt -> IO ()
declAtom s (Loc l :@ args) b = do
  withForeignPtr l (flip (solver_lit_begin s) b)
  mapM_ (solver_lit_add_con s) [ fromIntegral d | N d <- args ]

solve :: Maybe GLit -> SolverM Bool
solve ml = do 
  declAssumption ml
  r <- withSolver solver_solve
  return (r == 1)


dimacs :: String -> Maybe GLit -> SolverM ()
dimacs file ml = do
  declAssumption ml
  withSolver $ \s -> withCString file (solver_dimacs s)

value :: GLit -> SolverM Bool
value = sign value_
value_ (Loc l :@ args) sign = 
  withSolver $ \s -> 
    do withForeignPtr l (flip (solver_lit_begin s) sign)
       mapM_ (solver_lit_add_con s) [ fromIntegral d | N d <- args ]
       r <- solver_read s
       return (r == 1)

getNClauses, getNVars, getNConflicts :: SolverM Int
getNClauses   = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_clauses s
getNVars      = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_vars s
getNConflicts = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_conflicts s

setVerbose :: Bool -> SolverM ()
setVerbose b = withSolver $ \s -> solver_set_verbose s (if b then 1 else 0)

identify :: IO String
identify =
  do cs <- prim_identify
     peekCString cs

--------------------------------------------------------------------------------
-- foreign imports

foreign import ccall "static Wrapper.h " solver_new               :: IO Solver
foreign import ccall "static Wrapper.h " solver_free              :: Solver -> IO ()

foreign import ccall "static Wrapper.h "  loc_new                  :: CInt   -> IO (Ptr ())
foreign import ccall "static Wrapper.h &" loc_free                 :: FunPtr (Ptr () -> IO ())
foreign import ccall "static Wrapper.h "  loc_arity                :: Ptr () -> IO CUInt
					      
foreign import ccall "static Wrapper.h " solver_clause_begin       :: Solver -> IO ()
foreign import ccall "static Wrapper.h " solver_clause_add_lit     :: Solver -> (Ptr ()) -> CUInt -> IO ()
foreign import ccall "static Wrapper.h " solver_clause_add_lit_var :: Solver -> CInt -> IO ()
foreign import ccall "static Wrapper.h " solver_clause_add_lit_con :: Solver -> CInt -> IO ()
foreign import ccall "static Wrapper.h " solver_clause_add_size    :: Solver -> CInt -> IO ()
foreign import ccall "static Wrapper.h " solver_clause_commit      :: Solver -> CInt -> IO CUInt

foreign import ccall "static Wrapper.h " solver_lit_begin          :: Solver -> (Ptr ()) -> CUInt -> IO ()
foreign import ccall "static Wrapper.h " solver_lit_add_con        :: Solver -> CInt -> IO ()
foreign import ccall "static Wrapper.h " solver_read               :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_solve              :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_dimacs             :: Solver -> CString -> IO ()
foreign import ccall "static Wrapper.h " solver_simplify           :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_get_nof_clauses   :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_get_nof_vars      :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_get_nof_conflicts :: Solver -> IO CUInt
foreign import ccall "static Wrapper.h " solver_set_verbose       :: Solver -> CUInt -> IO ()
foreign import ccall "static Wrapper.h identify" prim_identify :: IO CString

--------------------------------------------------------------------------------
-- the end.
