#include "ac_redun.h"

static Ilist C_symbols = NULL;   /* commutative */
static Ilist A_symbols = NULL;   /* associative */
static Ilist A2_symbols = NULL;  /* associate2, x(yz) = y(xz) */
static Ilist AC_symbols = NULL;  /* C and (A or A2) */

/* Private definitions and types */

/*************
 *
 *   same_top()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
BOOL same_top(Term t1, Term t2)
{
  if (VARIABLE(t1) || VARIABLE(t2))
    return term_ident(t1, t2);
  else
    return SYMNUM(t1) == SYMNUM(t2);
}  /* same_top */

/*************
 *
 *   commutativity()
 *
 *************/

/* DOCUMENTATION
If the atom is commutativity, return the symnum of the operation;
otherwise return 0 (which is never a symnum).
*/

/* PUBLIC */
int commutativity(Term atom)
{
  if (eq_atom(atom)) {
    Term alpha = ARG(atom, 0);
    Term beta  = ARG(atom, 1);
    if (ARITY(alpha) == 2 && same_top(alpha, beta)) {
      Term a = ARG(alpha,0);
      Term b = ARG(alpha,1);
      Term c = ARG(beta,0);
      Term d = ARG(beta,1);
      if (VARIABLE(a) && VARIABLE(b) &&
	  VARNUM(a) != VARNUM(b) &&
	  term_ident(a,d) && term_ident(b,c))
	return SYMNUM(alpha);
    }
  }
  return 0;
}  /* commutativity */

/*************
 *
 *   associativity()
 *
 *************/

/* DOCUMENTATION
If the atom is associativity, f(f(x,y),z) = f(x,f(y,z)),
return the symnum of the operation;
otherwise return 0 (which is never a symnum).
*/

/* PUBLIC */
int associativity(Term atom)
{
  if (eq_atom(atom)) {
    Term alpha = ARG(atom, 0);
    Term beta  = ARG(atom, 1);
    if (ARITY(alpha) == 2 &&
	same_top(alpha, beta) &&
	same_top(alpha, ARG(alpha,0)) &&
	same_top(alpha, ARG(beta,1))) {

      Term a = ARG(ARG(alpha,0),0);
      Term b = ARG(ARG(alpha,0),1);
      Term c = ARG(alpha,1);
      Term d = ARG(beta,0);
      Term e = ARG(ARG(beta,1),0);
      Term f = ARG(ARG(beta,1),1);
	
      if (VARIABLE(a) && VARIABLE(b) && VARIABLE(c) &&
	  VARNUM(a) != VARNUM(b) &&
	  VARNUM(b) != VARNUM(c) &&
	  VARNUM(a) != VARNUM(c) &&
	  term_ident(a,d) && term_ident(b,e) && term_ident(c,f))
	return SYMNUM(alpha);
    }
  }
  return 0;
}  /* associativity */

/*************
 *
 *   associativity2()
 *
 *************/

/* DOCUMENTATION
If the atom is associativity2, f(x,f(y,z)) = f(y,f(x,z)),
return the symnum of the operation;
otherwise return 0 (which is never a symnum).
*/

/* PUBLIC */
int associativity2(Term atom)
{
  if (eq_atom(atom)) {
    Term alpha = ARG(atom, 0);
    Term beta  = ARG(atom, 1);
    if (ARITY(alpha) == 2 &&
	same_top(alpha, beta) &&
	same_top(alpha, ARG(alpha,1)) &&
	same_top(alpha, ARG(beta,1))) {

      Term a = ARG(alpha,0);
      Term b = ARG(ARG(alpha,1),0);
      Term c = ARG(ARG(alpha,1),1);
      Term d = ARG(beta,0);
      Term e = ARG(ARG(beta,1),0);
      Term f = ARG(ARG(beta,1),1);
	
      if (VARIABLE(a) && VARIABLE(b) && VARIABLE(c) &&
	  VARNUM(a) != VARNUM(b) &&
	  VARNUM(b) != VARNUM(c) &&
	  VARNUM(a) != VARNUM(c) &&
	  term_ident(a,e) && term_ident(b,d) && term_ident(c,f))
	return SYMNUM(alpha);
    }
  }
  return 0;
}  /* associativity2 */

/*************
 *
 *   can_compare()
 *
 *   Total lexicographic order on terms: var < constant < complex.
 *   Argument sn is the greatest symbol at the root.
 *
 *************/

static
Ordertype can_compare(Term a, Term b, int sn)
{
  if (VARIABLE(a) || VARIABLE(b))
    return term_compare_vcp(a, b);
  else if (SYMNUM(a) == sn && SYMNUM(b) == sn)
    return term_compare_vcp(a, b);
  else if (SYMNUM(a) == sn)
    return GREATER_THAN;
  else if (SYMNUM(b) == sn)
    return LESS_THAN;
  else
    return term_compare_vcp(a, b);
}  /* can_compare */

/*************
 *
 *   canon() -- yet another canonicalizer.
 *
 *************/

static
Term canon(Term t, int flag)
{
  if (term_flag(t, flag))
    return t;  /* already canonicalized */
  else if (VARIABLE(t))
    return t;
  else {
    int sn = SYMNUM(t);
    int i;
    for (i = 0; i < ARITY(t); i++)
      ARG(t,i) = canon(ARG(t,i), flag);

    if (ilist_member(AC_symbols, sn)) {
      Term a = ARG(t,0);
      Term b = ARG(t,1);
      if (same_top(t, b) && can_compare(a, ARG(b,0), sn) == GREATER_THAN) {
	Term b0 = ARG(b,0);
	ARG(t,0) = b0;
	ARG(b,0) = a;
	term_flag_clear(b, flag);
	ARG(t,1) = canon(b, flag);
	t = canon(t, flag);
      }
      else if (can_compare(a, b, sn) == GREATER_THAN) {
	ARG(t,0) = b;
	ARG(t,1) = a;
	t = canon(t, flag);
      }
    }
    else if (ilist_member(C_symbols, sn)) {
      Term a = ARG(t,0);
      Term b = ARG(t,1);
      if (can_compare(a, b, sn) == GREATER_THAN) {
	ARG(t,0) = b;
	ARG(t,1) = a;
      }
    }
    term_flag_set(t, flag);  /* mark as canonicalized */
    return t;
  }
}  /* canon */

/*************
 *
 *   cac_redundant()
 *
 *************/

static
BOOL cac_redundant(Term atom)
{
  if (!eq_atom(atom))
    return FALSE;
  else if (VARIABLE(ARG(atom,0)) || VARIABLE(ARG(atom,1)))
    return FALSE;
  else if (SYMNUM(ARG(atom,0)) != SYMNUM(ARG(atom,1)))
    return FALSE;
  else if (symbol_count(ARG(atom,0)) != symbol_count(ARG(atom,1)))
    return FALSE;
  else {
    int flag = claim_term_flag();  /* for marking canonicalized subterms */
    BOOL cac_redund;
    Term a = copy_term(ARG(atom, 0));
    Term b = copy_term(ARG(atom, 1));

    a = canon(a, flag);
    b = canon(b, flag);

    cac_redund = term_ident(a,b);
    zap_term(a);
    zap_term(b);
    release_term_flag(flag);
    return cac_redund;
  }
}  /* cac_redundant */

/*************
 *
 *   cac_redun_process()
 *
 *************/

static
void cac_redun_process(Clause c)
{
  BOOL null_literals = FALSE;
  Literal lit;
  int i = 0;
  for (lit = c->literals; lit; lit = lit->next) {
    i++;
    /* At first, I tried to simplify neg literals as well.
       This caused a problem (I don't remember what it was).
     */
    if (lit->sign && cac_redundant(lit->atom)) {
#if 0
      if (!term_ident(ARG(lit->atom,0),ARG(lit->atom,1))) {
	printf("AC or C redundant: ");
	p_term(lit->atom);
      }
#endif
      if (lit->sign) {
	/* replace literal with TRUE (which will cause clause to be deleted) */
	zap_term(lit->atom);
	lit->atom = get_rigid_term(TRUE_SYM, 0);
	lit->atom->container = c;
      }
      else {
	/* remove literal */
	c->justification = append_just(c->justification, xx_just(i));
	zap_term(lit->atom);
	lit->atom = NULL;
	null_literals = TRUE;
      }
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
}  /* cac_redun_process */

/*************
 *
 *   cac_redundancy()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void cac_redundancy(Clause c, BOOL print)
{
  if (pos_eq_unit(c)) {

    /* If the equation is comm, assoc, or assoc2, make a note and return. */

    int sn;
    sn = commutativity(c->literals->atom);
    if (sn != 0) {
      if (!ilist_member(C_symbols, sn)) {
	C_symbols = ilist_append(C_symbols, sn);
	if ((ilist_member(A_symbols, sn) || ilist_member(A2_symbols, sn)) &&
	    !ilist_member(AC_symbols, sn)) {
	  if (print)
	    printf("\n%% Operation %s is associative-commutative; redundancy checks enabled.\n", sn_to_str(sn));
	  AC_symbols = ilist_append(AC_symbols, sn);
	}
	else {
	  if (print)
	    printf("\n%% Operation %s is commutative; redundancy checks enabled.\n", sn_to_str(sn));
	}
      }
      return;
    }

    sn = associativity(c->literals->atom);
    if (sn != 0) {
      if (!ilist_member(A_symbols, sn)) {
	A_symbols = ilist_append(A_symbols, sn);
	if (ilist_member(C_symbols, sn) && !ilist_member(AC_symbols, sn)) {
	  if (print)
	    printf("\n%% Operation %s is associative-commutative; redundancy checks enabled.\n", sn_to_str(sn));
	  AC_symbols = ilist_append(AC_symbols, sn);
	}
      }
      return;
    }

    sn = associativity2(c->literals->atom);
    if (sn != 0) {
      if (!ilist_member(A2_symbols, sn)) {
	A2_symbols = ilist_append(A2_symbols, sn);
	if (ilist_member(C_symbols, sn) && !ilist_member(AC_symbols, sn)) {
	  if (print)
	    printf("\n%% Operation %s is associative-commutative; redundancy checks enabled.\n", sn_to_str(sn));
	  AC_symbols = ilist_append(AC_symbols, sn);
	}
      }
      return;
    }
  }  /* pos_eq_unit */

  /* The equation is not comm, assoc, or assoc2.  If there is at least
     one commutative operation, check for redundancy. */

  if (C_symbols)
    cac_redun_process(c);  

}  /* cac_redundancy */

