#include "clause3.h"

/* Private definitions and types */

/*************
 *
 *   copy_clause_ija()
 *
 *
 *************/

/* DOCUMENTATION
Copy a clause, including ID, justification, and attributes.
Do not call delete_clause() with clauses built this way,
because if delete_clause() receives a clause with an ID,
it assumes that clause is in the clause ID table.
Use zap_clause_ija() instead.
*/

/* PUBLIC */
Clause copy_clause_ija(Clause c)
{
  Clause d = copy_clause_with_flags(c);
  d->id = c->id;
  d->justification = copy_justification(c->justification);
  d->attributes = copy_attributes(c->attributes);
  return d;
}  /* copy_clause_ija */

/*************
 *
 *   zap_clause_ija()
 *
 *************/

/* DOCUMENTATION
Delete a clause that was created with copy_clause_ija().
*/

/* PUBLIC */
void zap_clause_ija(Clause c)
{
    zap_just(c->justification);
    zap_clause(c);  /* this zaps attributes */
}  /* zap_clause_ija */

/*************
 *
 *   copy_clauses_ija()
 *
 *************/

/* DOCUMENTATION
Copy a Plist of clauses.  This does a deep copy, using copy_clause_ija().
*/

/* PUBLIC */
Plist copy_clauses_ija(Plist p)
{
  Plist a;
  Plist b = NULL;   /* the new list */

  for (a = p; a; a = a->next) {
    Clause old = a->v;
    Clause new = copy_clause_ija(old);
    b = plist_prepend(b, new);  /* build it backward */
  }
  return reverse_plist(b);
}  /* copy_clauses_ija */

/*************
 *
 *   zap_clauses_ija()
 *
 *************/

/* DOCUMENTATION
Delete (deep) a Pist of clauses.
*/

/* PUBLIC */
void zap_clauses_ija(Plist p)
{
  if (p != NULL) {
    zap_clauses_ija(p->next);
    zap_clause_ija(p->v);
    free_plist(p);
  }
}  /* zap_clauses_ija */

/*************
 *
 *   clause_member_plist()
 *
 *************/

/* DOCUMENTATION
Is a Clause a member of a Plist?  A deep identity check is done
(clause_ident).
*/

/* PUBLIC */
Clause clause_member_plist(Plist p, Clause c)
{
  while (p) {
    if (clause_ident(c, p->v))
      return p->v;
    p = p->next;
  }
  return NULL;
}  /* clause_member_plist */

/*************
 *
 *   has_input_just()
 *
 *************/

/* DOCUMENTATION
Does a clause have justtification "input"?
*/

/* PUBLIC */
BOOL has_input_just(Clause c)
{
  return c->justification && c->justification->type == INPUT_JUST;
}  /* has_input_just */

/*************
 *
 *   has_copy_just()
 *
 *************/

/* DOCUMENTATION
Does a clause have justtification "copy"?
*/

/* PUBLIC */
BOOL has_copy_just(Clause c)
{
  return c->justification && c->justification->type == COPY_JUST;
}  /* has_copy_just */

/*************
 *
 *   initial_clause()
 *
 *************/

/* DOCUMENTATION
Is (was) the clause part of the initial sos (after processing input clauses,
before starting search)/
*/

/* PUBLIC */
BOOL initial_clause(Clause c)
{
  return c->initial;
}  /* initial_clause */

/*************
 *
 *   input_clauses()
 *
 *************/

/* DOCUMENTATION
Given a Plist of clauses, return the Plist of input clauses (in the
same order).
*/

/* PUBLIC */
Plist input_clauses(Plist a)
{
  if (a == NULL)
    return NULL;
  else {
    Plist c = input_clauses(a->next);
    if (has_input_just(a->v))
      return plist_prepend(c, a->v);
    else
      return c;
  }
}  /* input_clauses */

/*************
 *
 *   intersect_clauses()
 *
 *************/

/* DOCUMENTATION
Intersect 2 Plists of clauses.
The order of the result is determined by the order of the first list.
A deep identity check is done (clause_ident).
*/

/* PUBLIC */
Plist intersect_clauses(Plist a, Plist b)
{
  if (a == NULL)
    return NULL;
  else {
    Plist c = intersect_clauses(a->next, b);
    if (clause_member_plist(b, a->v))
      return plist_prepend(c, a->v);
    else
      return c;
  }
}  /* intersect_clauses */

/*************
 *
 *   max_clause_weight()
 *
 *************/

/* DOCUMENTATION
Given a Plist of clauses, return the weight of the heaviest clause.
The weight field of the clause is used.
*/

/* PUBLIC */
int max_clause_weight(Plist p)
{
  if (p == NULL)
    return INT_MIN;
  else {
    int tail_max = max_clause_weight(p->next);
    Clause c = p->v;
    return IMAX(c->weight,tail_max);
  }
}  /* max_clause_weight */

/*************
 *
 *   nonneg_clauses()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist nonneg_clauses(Plist clauses)
{
  Plist nonneg = NULL;
  Plist p;
  for (p = clauses; p; p = p->next)
    if (!negative_clause(p->v))
      nonneg = plist_append(nonneg, p->v);
  return nonneg;
}  /* nonneg_clauses */

