#include "clauseid.h"

/* Private definitions and types */

#define CLAUSE_ID_TAB_SIZE  50000

static Plist     Clause_id_tab[CLAUSE_ID_TAB_SIZE];
static unsigned  Clause_id_count = 0;

/*************
 *
 *   next_clause_id()
 *
 *************/

static
int next_clause_id(void)
{
  return ++Clause_id_count;
}  /* next_clause_id */

/*************
 *
 *   clause_ids_assigned()
 *
 *************/

/* DOCUMENTATION
What is the most recently assigned clause ID?
*/

/* PUBLIC */
int clause_ids_assigned(void)
{
  return Clause_id_count;
}  /* clause_ids_assigned */

/*************
 *
 *   assign_clause_id(c)
 *
 *************/

/* DOCUMENTATION
This routine assigns a unique identifier to the id field of a clause.
It also inserts the clause into a hash table so that given an id
number, the corresponding clause can be retrieved quickly (see
find_clause_by_id()).
*/

/* PUBLIC */
void assign_clause_id(Clause c)
{
  int i;

  if (c->id > 0) {
    p_clause(c);
    fatal_error("assign_clause_id, clause already has ID.");
  }
  c->id = next_clause_id();
  i = c->id % CLAUSE_ID_TAB_SIZE;
  Clause_id_tab[i] = insert_clause_into_plist(Clause_id_tab[i], c, TRUE);
}  /* assign_clause_id */

/*************
 *
 *     unassign_clause_id(c)
 *
 *************/

/* DOCUMENTATION
This routine removes a clause from the ID hash table and resets
the ID of the clause to 0.  A fatal error occurs if the clause
has not been assigned an ID.
*/

/* PUBLIC */
void unassign_clause_id(Clause c)
{
  int i;
  Plist p1, p2;

  i = c->id % CLAUSE_ID_TAB_SIZE;
  for (p2 = NULL, p1 = Clause_id_tab[i];
       p1 && ((Clause) p1->v)->id < c->id;
       p2 = p1, p1 = p1->next);
  if (p1 == NULL || ((Clause) p1->v)->id != c->id) {
    p_clause(c);
    fatal_error("unassign_clause_id, cannot find clause.");
  }
  if (p2)
    p2->next = p1->next;
  else
    Clause_id_tab[i] = p1->next;
  c->id = 0;
  free_plist(p1);
}  /* unassign_clause_id */

/*************
 *
 *   has_clause_id()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause has been assigned an ID.
*/

/* PUBLIC */
BOOL has_clause_id(Clause c)
{
  return c->id > 0;
}  /* has_clause_id */

/*************
 *
 *     find_clause_by_id(id)
 *
 *     Given a clause ID, retrieve the clause (or NULL).
 *
 *************/

/* DOCUMENTATION
This routine retrieves the clause with the given ID number
(or NULL, if there is no such clause).
*/

/* PUBLIC */
Clause find_clause_by_id(int id)
{
  int i;
  Plist p1;

  i = id % CLAUSE_ID_TAB_SIZE;
  for (p1 = Clause_id_tab[i];
       p1 && ((Clause) p1->v)->id < id;
       p1 = p1->next);
  if (p1 != NULL && ((Clause) p1->v)->id == id)
    return(p1->v);
  else
    return(NULL);
}  /* find_clause_by_id */

/*************
 *
 *     fprint_clause_id_tab(fp)
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) all the clauses in the ID hash table.
*/

/* PUBLIC */
void fprint_clause_id_tab(FILE *fp)
{
  int i;
  Plist p;

  fprintf(fp, "\nID clause table:\n");
  for (i=0; i<CLAUSE_ID_TAB_SIZE; i++)
    for (p = Clause_id_tab[i]; p; p = p->next)
      fprint_clause(fp, p->v);
  fflush(fp);
}  /* fprint_clause_id_tab */

/*************
 *
 *     p_clause_id_tab(tab)
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) all the clauses in the ID hash table.
*/

/* PUBLIC */
void p_clause_id_tab()
{
  fprint_clause_id_tab(stdout);
}  /* p_clause_id_tab */

/*************
 *
 *   insert_clause_into_plist()
 *
 *************/

/* DOCUMENTATION
This routine inserts a clause into a sorted (by ID) Plist of clauses.
Boolean paramemeter "increasing" tells whether the list is increasing
or decreasing.
The updated Plist is returned.
If the clause is already there, nothing happens.
*/

/* PUBLIC */
Plist insert_clause_into_plist(Plist p, Clause c, BOOL increasing)
{
  Plist prev, curr, new;
  prev = NULL;
  curr = p;
  while (curr != NULL && (increasing ? ((Clause) curr->v)->id < c->id
	                             : ((Clause) curr->v)->id > c->id)) {
    prev = curr;
    curr = curr->next;
  }
  if (curr == NULL || ((Clause) curr->v)->id != c->id) {
    new = get_plist();
    new->v = c;
    new->next = curr;
    if (prev != NULL)
      prev->next = new;
    else
      p = new;
  }
  return p;
}  /* insert_clause_into_plist */
