#include "clausify.h"

/* Private definitions and types */

/*************
 *
 *   formula_to_literal()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Literal representation.
If the formula is not an ATOM_FORM or the negation of an ATOM_FORM,
NULL is returned.
<P>
The returned literal is an entirely new copy,
and the given formula is not changed.
*/

/* PUBLIC */
Literal formula_to_literal(Formula f)
{
  if (f->type == NOT_FORM && f->kids[0]->type == ATOM_FORM)
    return new_literal(0, copy_term(f->kids[0]->atom));
  else if (f->type == ATOM_FORM)
    return new_literal(1, copy_term(f->atom));
  else
    return NULL;
}  /* formula_to_literal */

/*************
 *
 *   formula_to_clause()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Clause representation.
If f is not an atom, literal, or disjunction of literals, the
returned clause will be NULL or not well formed.
<P>
The returned clause is an entirely new copy, and the given formula
is not changed.
*/

/* PUBLIC */
Clause formula_to_clause(Formula f)
{
  Clause c = get_clause();
  if (f->type == ATOM_FORM || f->type == NOT_FORM)
    append_literal(c, formula_to_literal(f));
  else if (f->type == OR_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      append_literal(c, formula_to_literal(f->kids[i]));
  }
  else {
    fatal_error("formula_to_clause, formula not ATOM, NOT, or OR");
  }
  upward_clause_links(c);
  return c;
}  /* formula_to_clause */

/*************
 *
 *   formula_to_clauses()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Plist of Clauses
representation.
If f is not an atom, literal, or disjunction of literals, or a
conjunction of those things, the clauses in the returned list may be NULL
or not well formed.
<P>
The returned clauses are entirely new copies, and the given formula
is not changed.
*/

/* PUBLIC */
Plist formula_to_clauses(Formula f)
{
  Plist lst = NULL;
  if (f->type == AND_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      lst = plist_append(lst, formula_to_clause(f->kids[i]));
  }
  else if (f->type == OR_FORM || f->type == NOT_FORM || f->type == ATOM_FORM)
    lst = plist_append(lst, formula_to_clause(f));
  return lst;
}  /* formula_to_clauses */

/*************
 *
 *   clausify_formula()
 *
 *************/

/* DOCUMENTATION
This routine translates a Formula f into a Clist of clauses.
The translation includes Skolemization, so the result should
be unsatisfiable iff f is unsatisfiable.  The variables in
each clause are renumbered.  If there are more than MAX_VARS
variables in a clause, a fatal error occurs.
<P>
Formula f is not changed
*/

/* PUBLIC */
Plist clausify_formula(Formula f)
{
  Formula g;
  Plist clauses, p;

  g = clausify_prepare(formula_copy(f));
  clauses = formula_to_clauses(g);

  for (p = clauses; p; p = p->next) {
    Clause c = p->v;
    renumber_variables(c, MAX_VARS);
    c->justification = clausify_just();
  }
  zap_formula(g);
  return clauses;
}  /* clausify_formula */

/*************
 *
 *   clausify_formulas()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist clausify_formulas(Plist formulas)
{
  Plist clauses = NULL;
  Plist p;
  for (p = formulas; p; p = p->next)
    clauses = plist_cat(clauses, clausify_formula(p->v));
  return clauses;
}  /* clausify_formulas */

/*************
 *
 *   vars_to_names()
 *
 *************/

static
Term vars_to_names(Term t)
{
  if (VARIABLE(t)) {
    Term a;
    char *s1 = malloc(10);
    char *s2 = malloc(10);
    Variable_style v = variable_style();
    s2 = itoa(VARNUM(t), s2);

    switch (v) {
    case INTEGER_STYLE:   s1 = strcpy(s1, "");      break;
    case STANDARD_STYLE:  s1 = strcpy(s1, "var_");  break;
    case PROLOG_STYLE:    s1 = strcpy(s1, "VAR_");  break;
    }
    s1 = strcat(s1, s2);

    a = get_rigid_term(s1, 0);
    free_term(t);
    free(s1);
    free(s2);
    return a;
  }
  else {
    int i;
    for (i = 0; i < ARITY(t); i++)
      ARG(t,i) = vars_to_names(ARG(t,i));
    return t;
  }
}  /* vars_to_names */

/*************
 *
 *   literal_to_formula()
 *
 *************/

static
Formula literal_to_formula(Literal lit)
{
  Formula a = formula_get(0, ATOM_FORM);
  a->atom = vars_to_names(copy_term(lit->atom));
  if (lit->sign)
    return a;
  else {
    Formula n = formula_get(1, NOT_FORM);
    n->kids[0] = a;
    return n;
  }
}  /* literal_to_formula */

/*************
 *
 *   clause_to_formula()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Formula clause_to_formula(Clause c)
{
  if (c->literals == NULL) {
    Formula f = formula_get(0, ATOM_FORM);
    f->atom = get_rigid_term(FALSE_SYM, 0);
    return f;
  }
  else {
    Literal lit;
    int i;
    Formula f = formula_get(number_of_literals(c), OR_FORM);
    for (lit = c->literals, i = 0; lit; lit = lit->next, i++) {
      f->kids[i] = literal_to_formula(lit);
    }
    return f;
  }
}  /* clause_to_formula */

