#include "features.h"

/* Private definitions and types */

static Ilist Feature_symbols;  /* list of featured symbols (symnums)*/

/* The following are work arrays, indexed by symnum, used for calculating
   the features of a clause.  They are allocated by init_features() and
   left in place throughout the process.
*/

static int *Pos_occurrences;
static int *Neg_occurrences;
static int *Pos_maxdepth;
static int *Neg_maxdepth;

/*************
 *
 *   init_features()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_features(Ilist fsyms, Ilist rsyms)
{
  int n = greatest_symnum() + 1;

  Feature_symbols = ilist_cat(ilist_copy(rsyms), ilist_copy(fsyms));
  Pos_occurrences = calloc(n, sizeof(int));
  Neg_occurrences = calloc(n, sizeof(int));
  Pos_maxdepth    = calloc(n, sizeof(int));
  Neg_maxdepth    = calloc(n, sizeof(int));
}  /* init_features */

/*************
 *
 *   fill_in_arrays()
 *
 *************/

static
void fill_in_arrays(Term t, BOOL sign, int depth)
{
  if (!VARIABLE(t)) {
    int i;
    if (sign) {
      Pos_occurrences[SYMNUM(t)]++;
      Pos_maxdepth[SYMNUM(t)] = IMAX(depth, Pos_maxdepth[SYMNUM(t)]);
    }
    else {
      Neg_occurrences[SYMNUM(t)]++;
      Neg_maxdepth[SYMNUM(t)] = IMAX(depth, Neg_maxdepth[SYMNUM(t)]);
    }
    for (i = 0; i < ARITY(t); i++)
      fill_in_arrays(ARG(t,i), sign, depth+1);
  }
}  /* fill_in_arrays */

/*************
 *
 *   features()
 *
 *************/

/* DOCUMENTATION
Given a clause, build the feature vector.

Features:
  positive literals
  negative literals
  foreach relation symbol
     positive occurrences
     negative occurrences
  foreach function symbol
     positive occurrences
     negative occurrences
     positive maxdepth
     negative maxdepth
*/

/* PUBLIC */
Ilist features(Clause c)
{
  Ilist f = NULL;
  Ilist p;
  Literal lit;

  /* Build it backwards, then reverse it. */

  f = ilist_prepend(f, positive_literals(c));
  f = ilist_prepend(f, negative_literals(c));

  for (lit = c->literals; lit; lit = lit->next) {
    fill_in_arrays(lit->atom, lit->sign, 0);
  }

  for (p = Feature_symbols; p; p = p->next) {
    f = ilist_prepend(f, Pos_occurrences[p->i]);
    f = ilist_prepend(f, Neg_occurrences[p->i]);

    if (function_symbol(p->i)) {
      f = ilist_prepend(f, Pos_maxdepth[p->i]);
      f = ilist_prepend(f, Neg_maxdepth[p->i]);
    }

    Pos_occurrences[p->i] = 0;
    Neg_occurrences[p->i] = 0;
    Pos_maxdepth[p->i] = 0;
    Neg_maxdepth[p->i] = 0;
  }
  f = reverse_ilist(f);
#if 0
  printf("Features for clause "); f_clause(c);
  p_features(f);
#endif
  return f;
}  /* features */

/*************
 *
 *   feature_length()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int feature_length(void)
{
  int n = 2;  /* pos lits, neg lits */
  Ilist p;
  for (p = Feature_symbols; p; p = p->next) {
    n += 2;
    if (function_symbol(p->i))
      n += 2;
  }
  return n;
}  /* feature_length */

/*************
 *
 *   features_less_or_equal()
 *
 *************/

/* DOCUMENTATION
Return TRUE if Ilists c and d are thr same length and
each member of c is <= the corresponding member of d.
*/

/* PUBLIC */
BOOL features_less_or_equal(Ilist c, Ilist d)
{
  while (c && d && c->i <= d->i) {
    c = c->next;
    d = d->next;
  }
  return !c && !d;
}  /* features_less_or_equal */

/*************
 *
 *   p_features()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void p_features(Ilist f)
{
  Ilist p;
  printf("  pos_lits=%d, neg_lits=%d\n", f->i, f->next->i);
  f = f->next->next;

  for (p = Feature_symbols; p; p = p->next) {
    printf("  symbol %s: ", sn_to_str(p->i));
    printf("pos_occ=%d, neg_occ=%d", f->i, f->next->i);
    f = f->next->next;
    if (function_symbol(p->i)) {
      printf(", pos_max=%d, neg_max=%d", f->i, f->next->i);
      f = f->next->next;
    }
    printf("\n");
  }
}  /* p_features */

