#include "hints.h"

/* Private definitions and types */

static Lindex Hints_idx = NULL;        /* FPA index for hints */
static Clist Redundant_hints = NULL;   /* list of hints not indexed */

/*************
 *
 *   init_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_hints(Uniftype utype)
{
  Hints_idx = lindex_init(FPA, utype, 10, FPA, utype, 10);
  Redundant_hints = clist_init("redundant_hints");
}  /* init_hints */

/*************
 *
 *   done_with_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void done_with_hints(void)
{
  if (!lindex_empty(Hints_idx) || !clist_empty(Redundant_hints))
    printf("ERROR: Hints index not empty!\n");
  lindex_destroy(Hints_idx);
  Hints_idx = NULL;
  clist_free(Redundant_hints);
  Redundant_hints = NULL;
}  /* done_with_hints */

/*************
 *
 *   redundant_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int redundant_hints(void)
{
  return clist_length(Redundant_hints);
}  /* redundant_hints */

/*************
 *
 *   index_hint()
 *
 *************/

/* DOCUMENTATION
Index a clause C as a hint (make sure to call init_hints first).
If the clause is equivalent to a previously indexed hint H, any
labels on C are copied to H, and C is not indexed.
*/

/* PUBLIC */
void index_hint(Clause c, int label_at)
{
  Clause h = back_subsume_one(c, Hints_idx);  /* look for h subsumed by c */
  if (h && subsumes(h, c)) {
    int i = 1;
    char *s;
    while ((s = get_string_attribute(c->attributes, label_at, i))) {
      h->attributes = set_string_attribute(h->attributes, label_at, s);
      i++;
    }
    clist_append(c, Redundant_hints);
  }
  else
    lindex_update(Hints_idx, c, INSERT);
}  /* index_hint */

/*************
 *
 *   unindex_hint()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void unindex_hint(Clause c)
{
  if (!clist_member(c, Redundant_hints))
    lindex_update(Hints_idx, c, DELETE);
  else
    clist_remove(c, Redundant_hints);
}  /* unindex_hint */

/*************
 *
 *   find_hint()
 *
 *************/

static
Clause find_hint(Clause c)
{
  return back_subsume_one(c, Hints_idx);
}  /* find_hint */

/*************
 *
 *   adjust_weight_with_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void adjust_weight_with_hints(Clause c,
			      int bsub_add_wt_at,
			      int bsub_wt_at,
			      int label_at,
			      BOOL degrade)
{
  Clause hint = find_hint(c);

  if (hint != NULL) {

    /* This code allows both bsub_wt and bsub_add_wt to be used. */
    /* I don't think this will happen much in practice. */

    int bsub_wt = get_int_attribute(hint->attributes, bsub_wt_at, 1);
    int bsub_add_wt = get_int_attribute(hint->attributes, bsub_add_wt_at, 1);

    if (bsub_wt != INT_MAX)
      c->weight = bsub_wt;
    if (bsub_add_wt != INT_MAX)
      c->weight += bsub_add_wt;
    
    /* If the hint has label attributes, copy them to the clause. */
    
    {
      int i = 0;
      char *s = get_string_attribute(hint->attributes, label_at, ++i);
      while (s) {
	if (!string_attribute_member(c->attributes, label_at, s))
	  c->attributes = set_string_attribute(c->attributes, label_at, s);
	s = get_string_attribute(hint->attributes, label_at, ++i);
      }
    }

    /* Veroff's hint degradation strategy.  Cut in half the bsub_add_wt. */

    if (degrade && bsub_add_wt != INT_MAX) {
      int new = bsub_add_wt / 2;
      replace_int_attribute(hint->attributes, bsub_add_wt_at, new, 1);
      /*
      printf("degraded hint %d from %d to %d\n",hint->id,bsub_add_wt,new);
      */
    }
  }
}  /* adjust_weight_with_hints */
