#include "maximal.h"

/* Private definitions and types */

int Maximal_flag     = -1;  /* termflag to mark maximal literals */

/*************
 *
 *   init_maximal()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_maximal(void)
{
  if (Maximal_flag != -1)
    fatal_error("init_maximal, called more than once");
  Maximal_flag = claim_term_flag();
}  /* init_maximal */

/*************
 *
 *   greater_literals()
 *
 *************/

static
BOOL greater_literals(Literal l1, Literal l2)
{
  /* Is l1 > l2? */
  /* equality > nonequality  */
  /* they must be of the same sign */
  if (l1->sign != l2->sign)
    return FALSE;
  else {
    Term t1 = l1->atom;
    Term t2 = l2->atom;
    Ordertype p = sym_precedence(SYMNUM(t1), SYMNUM(t2));
    if (p == GREATER_THAN)
      return TRUE;
    else if (p == LESS_THAN)
      return FALSE;
    else
      return term_greater(t1, t2);  /* LPO, RPO, KBO */
  }
}  /* greater_literals */

/*************
 *
 *   maximal_literal_test()
 *
 *************/

/* DOCUMENTATION
Test if a literal is maximal in a clause (w.r.t. others literals
of the same sign).
This version does not use a flag.
*/

/* PUBLIC */
BOOL maximal_literal_test(Clause c, Literal lit)
{
  /* If there is a greater literal of the same sign, return FALSE. */
  Literal l2 = c->literals;
  BOOL maximal = TRUE;
  while (l2 && maximal) {
    if (lit != l2 && lit->sign == l2->sign && greater_literals(l2, lit))
      maximal = FALSE;
    else
      l2 = l2->next;
  }
  return maximal;
}  /* maximal_literal_test */

/*************
 *
 *   mininal_neg_literal_test()
 *
 *************/

/* DOCUMENTATION
Test if a literal is a minimal negative literal.
This version does not use a flag.
*/

/* PUBLIC */
BOOL minimal_neg_literal_test(Clause c, Literal lit)
{
  if (lit->sign)
    return FALSE;
  else {
    /* If there is a lesser negative literal, return FALSE. */
    Literal l2 = c->literals;
    BOOL minimal = TRUE;
    while (l2 && minimal) {
      if (lit != l2 && l2->sign == FALSE && greater_literals(lit, l2))
	minimal = FALSE;
      else
	l2 = l2->next;
    }
    return minimal;
  }
}  /* minimal_neg_literal_test */

/*************
 *
 *   mark_maximal_literals()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void mark_maximal_literals(Clause c)
{
  Literal lit;

  if (Maximal_flag == -1)
    fatal_error("mark_maximal_literals, init_maximal() was not called");
  
  for (lit = c->literals; lit; lit = lit->next) {
    if (maximal_literal_test(c, lit))
      /* Note: this is a property of the literal, but we mark the atom. */
      term_flag_set(lit->atom, Maximal_flag);
  }
}  /* mark_maximal_literals */

/*************
 *
 *   maximal_literal_check()
 *
 *************/

/* DOCUMENTATION
Check if a literal is maximal in the clause that contains it.
This only checks a flag.  It does not compute maximality.
*/

/* PUBLIC */
BOOL maximal_literal_check(Literal lit)
{
  return term_flag(lit->atom, Maximal_flag);
}  /* maximal_literal_check */

/*************
 *
 *   maximal_literals()
 *
 *************/

/* DOCUMENTATION
Return the number of maximal literals.  This checks a flag only.
*/

/* PUBLIC */
int maximal_literals(Clause c)
{
  int n = 0;
  Literal lit;
  for (lit = c->literals; lit; lit = lit->next) {
    if (maximal_literal_check(lit))
      n++;
  }
  return n;
}  /* maximal_literals */

/*************
 *
 *   first_maximal_literal_of_sign()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Literal first_maximal_literal_of_sign(Clause c, BOOL sign)
{
  Literal lit = c->literals;
  while (lit && !(lit->sign = sign &&
		  maximal_literal_check(lit)))
    lit = lit->next;
  return lit;
}  /* first_maximal_literal_of_sign */



