#include "paramod.h"
#include "resolve.h"

/* Private definitions and types */

static BOOL  Ordered_inference  = FALSE;  /* not implemented yet (literals) */
static BOOL  Ordered_instance   = FALSE;  /* non-oriented from lits */

static int Ordered_instance_prunes = 0;  /* counter */
static int Basic_prunes = 0;             /* counter */

/*************
 *
 *   paramodulation_options()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */

void paramodulation_options(BOOL ordered_inference,
			    BOOL ordered_instance,
			    BOOL basic_paramodulation)
{
  Ordered_inference = ordered_inference;
  Ordered_instance = ordered_instance;
  set_basic_paramod(basic_paramodulation);
  Ordered_instance_prunes = 0;
  Basic_prunes = 0;
}  /* paramodulation_options */

/*************
 *
 *   para_instance_prunes()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int para_instance_prunes()
{
  return Ordered_instance_prunes;
}  /* para_instance_prunes */

/*************
 *
 *   basic_paramodulation_prunes()
 *
 *************/

/* DOCUMENTATION
How many paramodulants were killed because they failed the "basic" test.
*/

/* PUBLIC */
int basic_paramodulation_prunes(void)
{
  return Basic_prunes;
}  /* basic_paramodulation_prunes */

/*************
 *
 *   basic_check()
 *
 *************/

static
BOOL basic_check(Term into_term)
{
  if (basic_paramod() && nonbasic_term(into_term)) {
    Basic_prunes++;
    return FALSE;
  }
  else
    return TRUE;
}  /* basic_check */

/*************
 *
 *   apply_lit_para()
 *
 *************/

static
Literal apply_lit_para(Literal lit, Context c)
{
  if (basic_paramod())
    return new_literal(lit->sign, apply_basic(lit->atom, c));
  else
    return new_literal(lit->sign, apply(lit->atom, c));
}  /* apply_lit_para */

/*************
 *
 *   apply_substitute_para()
 *
 *************/

static
Term apply_substitute_para(Term t, Term beta, Context from_subst,
			   Term into, Context into_subst)
{
  if (basic_paramod())
    return apply_basic_substitute(t, beta, from_subst, into, into_subst);
  else
    return apply_substitute(t, beta, from_subst, into, into_subst);
}  /* apply_substitute_para */

/*************
 *
 *   paramodulate()
 *
 *   Construct a paramodulant.
 *
 *************/

static
Clause paramodulate(Literal from_lit,
		    Term alpha, Context from_subst,
		    Term into,  Context into_subst)
{
  Clause from_clause = alpha->container;
  Clause into_clause = into->container;
  Clause p = get_clause();
  Term beta = NULL;
  Literal lit;
  for (lit = from_clause->literals; lit != NULL; lit = lit->next) {
    if (lit != from_lit)
      append_literal(p, apply_lit_para(lit, from_subst));
    else
      beta = (ARG(lit->atom,0)==alpha ? ARG(lit->atom,1) : ARG(lit->atom,0));
  }
  for (lit = into_clause->literals; lit != NULL; lit = lit->next) {
    append_literal(p, new_literal(lit->sign,
				  apply_substitute_para(lit->atom,
							beta, from_subst,
							into, into_subst)));
  }
  inherit_attributes(from_clause, from_subst, into_clause, into_subst, p);
  upward_clause_links(p);
  return p;
}  /* paramodulate */

/*************
 *
 *   para_right()
 *
 *   Should we paramoulate with the right side?
 *   If it is oriented, then NO;
 *   else if it is a renamable_flip unit, then NO;
 *   else YES.
 *
 *************/

static
BOOL para_right(Term atom)
{
  /* Assume atom is an eq_atom. */
  if (oriented_eq(atom))
    return FALSE;
  else if (renamable_flip_eq(atom) && unit_clause(atom->container))
    return FALSE;
  else
    return TRUE;
}  /* para_right */

/*************
 *
 *   ordered_instance_check()
 *
 *************/

static
BOOL ordered_instance_check(Term atom, int side, Context subst)
{
  BOOL ok = TRUE;
  /* This checks eqality atoms only.
     It does not check that literals are maximal in clauses. */
  
  if (!oriented_eq(atom)) {
    /* Check instance of atom; fail if the side we're using is smaller. */
    Term a1 = apply(ARG(atom, 0), subst);
    Term a2 = apply(ARG(atom, 1), subst);
    if ((side == 1 && term_greater(a2, a1)) ||  /* LPO, RPO, KBO */
	(side == 2 && term_greater(a1, a2)))
      ok = FALSE;
    zap_term(a1);
    zap_term(a2);
    Ordered_instance_prunes++;
  }
  return ok;
}  /* ordered_instance_check */

/*************
 *
 *   para_into()
 *
 *   Paramodulate from a given side of a given literal into a given
 *   term and/or its subterms.  The position vectors are for the
 *   justification only.  The parameter "where" should be
 *   PARA_ALL, PARA_TOP_ONLY, or PARA_ALL_EXCEPT_TOP.
 *
 *************/

static
void para_into(Literal from_lit, Term alpha, Context cf,
	       Ilist from_vec, int from_side,
	       Literal into_lit, Term into, Context ci,
	       Ilist into_vec, int into_side,
	       Para_loc where,
	       void (*proc_proc) (Clause))
{
  if (!VARIABLE(into) && basic_check(into)) {
    int i;
    if (COMPLEX(into) && (where == PARA_ALL || where == PARA_ALL_EXCEPT_TOP)) {
      Ilist g = ilist_prepend(into_vec, 0);
      for (i = 0; i < ARITY(into); i++) {
	g->i += 1;
	para_into(from_lit, alpha, cf, from_vec, from_side,
		  into_lit, ARG(into,i), ci, g, into_side,
		  PARA_ALL, proc_proc);
      }
      free_ilist(g);  /* Frees first node only; into_vec is unchanged. */
    }
    if (where == PARA_ALL || where == PARA_TOP_ONLY) {
      Trail tr = NULL;
      if (unify(alpha, cf, into, ci, &tr)) {
	if (!Ordered_instance ||
	    ordered_instance_check(from_lit->atom, from_side, cf)) {
	  Clause p = paramodulate(from_lit, alpha, cf, into, ci);
	  p->justification = para_just_rev_copy(PARA_JUST,
						alpha->container, from_vec,
						into->container, into_vec);
	  (*proc_proc)(p);
	}
	undo_subst(tr);
      }
    }
  }
}  /* para_into */

/*************
 *
 *   para_lits_into_noneq()
 *
 *************/

static
void para_lits_into_noneq(Literal flit, Term alpha, Context cf,
			  Literal ilit, Context ci,
			  void (*proc_proc) (Clause))
{
  if (!VARIABLE(alpha)) {
    /* Position vectors are constructed backward. */
    Ilist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Ilist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Term a = ilit->atom;
    int i;
    int from_side = (ARG(flit->atom,0) == alpha ? 1 : 2);

    from_vec->next->i = literal_number(alpha->container, flit);
    from_vec->i = from_side;  /* arg of flit */
    into_vec->next->i = literal_number(ilit->atom->container, ilit);
    /* The atom is not equality, so do nothing special. */
    for (i = 0; i < ARITY(a); i++) {
      into_vec->i += 1;  /* increment arg number */
      para_into(flit, alpha, cf, from_vec, from_side,
		ilit, ARG(a,i), ci, into_vec, -1,
		PARA_ALL, proc_proc);
    }
    zap_ilist(from_vec);
    zap_ilist(into_vec);
  }
}  /* para_lits_into_noneq */

/*************
 *
 *   para_lits_into_eq()
 *
 *************/

static
void para_lits_into_eq(Literal flit, Term alpha, Context cf,
		       Literal ilit, Term into_arg, Context ci,
		       Para_loc where,
		       void (*proc_proc) (Clause))
{
  if (!VARIABLE(alpha) && !VARIABLE(into_arg)) {
    int from_side = (ARG(flit->atom,0) == alpha ? 1 : 2);
    int into_side = (ARG(ilit->atom,0) == into_arg ? 1 : 2);
  
    /* Position vectors are constructed backward. */

    Ilist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Ilist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
  
    from_vec->next->i = literal_number(alpha->container, flit);
    from_vec->i = from_side;  /* arg of flit */
    into_vec->next->i = literal_number(into_arg->container, ilit);
    into_vec->i = into_side;  /* arg of ilit */

    para_into(flit, alpha, cf, from_vec, from_side,
	      ilit, into_arg, ci, into_vec, into_side,
	      where, proc_proc);

    zap_ilist(from_vec);
    zap_ilist(into_vec);
  }
}  /* para_lits_into_eq */

/*************
 *
 *   para_from_into()
 *
 *************/

/* DOCUMENTATION
Paramodulate from one clause into another (non-backtrack unification version).
<P>
For oriented equality atoms, we go from left sides only
and into both sides.
For nonoriented equality atoms, we go from and into both sides.
<P>
"top_check" is used to prevent redundancy.  If paramodulating
C->D and then D->C, use top_check for the second call.
*/

/* PUBLIC */
void para_from_into(Clause from, Context cf,
		    Clause into, Context ci,
		    BOOL top_check,
		    void (*proc_proc) (Clause))
{
  if (Ordered_inference && !positive_clause(from))
    return;
  else {
    BOOL into_positive = positive_clause(into);
    Literal flit;
    for (flit = from->literals; flit; flit = flit->next) {
      if (Ordered_inference && !maximal_literal_check(flit))
	; /* do nothing */
      else if (pos_eq(flit)) {
	Term f1 = ARG(flit->atom,0);
	Term f2 = ARG(flit->atom,1);
	Literal ilit;
	for (ilit = into->literals; ilit; ilit = ilit->next) {
	  BOOL ok;
	  if (!Ordered_inference)
	    ok = TRUE;
	  else if (into_positive)
	    ok = maximal_literal_check(ilit);
	  else
	    ok = ilit->sign == FALSE;

	  if (!ok)
	    ;  /* don't go into this literal */

	  else if (eq_atom(ilit->atom)) {
	    Term i1 = ARG(ilit->atom,0);
	    Term i2 = ARG(ilit->atom,1);
	    Para_loc where;

	    if (top_check && ilit->sign &&
		oriented_eq(ilit->atom) && oriented_eq(flit->atom))
	      where = PARA_ALL_EXCEPT_TOP;
	    else
	      where = PARA_ALL;

	    para_lits_into_eq(flit,f1,cf,ilit,i1,ci,where,proc_proc);      /* L->L */
	    if (para_right(flit->atom))
	      para_lits_into_eq(flit,f2,cf,ilit,i1,ci,PARA_ALL,proc_proc); /* R->L */
	    para_lits_into_eq(flit,f1,cf,ilit,i2,ci,PARA_ALL,proc_proc);   /* L->R */
	    if (para_right(flit->atom))
	      para_lits_into_eq(flit,f2,cf,ilit,i2,ci,PARA_ALL,proc_proc); /* R->R */
	  }
	  else {
	    para_lits_into_noneq(flit,f1,cf,ilit,ci,proc_proc);            /* L ->  */
	    if (para_right(flit->atom))
	      para_lits_into_noneq(flit,f2,cf,ilit,ci,proc_proc);          /* R ->  */
	  }
	}
      }
    }
  }
}  /* para_from_into */

/*************
 *
 *   para_pos()
 *
 *************/

/* DOCUMENTATION
Construct a paramodulant from the given data.  A fatal error
occurs if it does not exist.  In building the justification,
the position vectors are copied.
*/

/* PUBLIC */
Clause para_pos(Clause from, Ilist from_pos, Clause into, Ilist into_pos)
{
  Literal from_lit = ith_literal(from, from_pos->i);
  Literal into_lit = ith_literal(into, into_pos->i);
  BOOL left_to_right = from_pos->next->i == 1;
  Term alpha = ARG(from_lit->atom, left_to_right ? 0 : 1);
  Term into_term = term_at_pos(into_lit->atom, into_pos->next);
  Context cf = get_context();
  Context ci = get_context();
  Trail tr = NULL;
  Clause paramodulant;

  BOOL ok = unify(alpha, cf, into_term, ci, &tr);
  if (!ok)
    fatal_error("para_pos, terms do not unify");

  paramodulant = paramodulate(from_lit, alpha, cf, into_term, ci);
  paramodulant->justification = para_just(PARA_JUST,
					  from, copy_ilist(from_pos),
					  into, copy_ilist(into_pos));
  renumber_variables(paramodulant, MAX_VARS);
  undo_subst(tr);
  free_context(cf);
  free_context(ci);
  return paramodulant;
}  /* para_pos */

