#include "resolve.h"

/* Private definitions and types */

static BOOL  Ordered_inference  = FALSE;
static BOOL  Ordered_instance   = FALSE;
static int   Lit_selection      = LIT_SELECTION_MAXIMAL;

static int   Nucleus_limit      = INT_MAX;  /* limit num of clashable lits */
static BOOL  Initial_nuclei     = FALSE;    /* nuclei must be input clauses  */

static int   Ordered_instance_prunes = 0;  /* counter */

/*************
 *
 *   resolution_options()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */

void resolution_options(BOOL ordered_inference,
			BOOL ordered_instance,
			char *lit_selection,
			BOOL initial_nuclei,
			int nucleus_limit)
{
  Ordered_inference = ordered_inference;
  Ordered_instance = ordered_instance;
  Nucleus_limit = nucleus_limit;
  Initial_nuclei = initial_nuclei;

  if (str_ident(lit_selection, "first"))
    Lit_selection = LIT_SELECTION_FIRST;
  else if (str_ident(lit_selection, "all"))
    Lit_selection = LIT_SELECTION_ALL;
  else if (str_ident(lit_selection, "maximal"))
    Lit_selection = LIT_SELECTION_MAXIMAL;
  else if (str_ident(lit_selection, "first_maximal"))
    Lit_selection = LIT_SELECTION_FIRST_MAXIMAL;
  else
    fatal_error("resolution_options, lit_selection not known");

  Ordered_instance_prunes = 0;
}  /* resolution_options */

/*************
 *
 *   res_instance_prunes()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int res_instance_prunes()
{
  return Ordered_instance_prunes;
}  /* res_instance_prunes */

/*************
 *
 *   unit_check()
 *
 *************/

static
BOOL unit_check(Literal lit)
{
  return unit_clause(lit->atom->container);
}  /* unit_check */

/*************
 *
 *   pos_hyper_sat_test()
 *
 *************/

static
BOOL pos_hyper_sat_test(Literal lit)
{
  Clause c = lit->atom->container;
  if (positive_clause(c))
    return Ordered_inference ? maximal_literal_check(lit) : TRUE;
  else
    return FALSE;
}  /* pos_hyper_sat_test */

/*************
 *
 *   neg_hyper_sat_test()
 *
 *************/

static
BOOL neg_hyper_sat_test(Literal lit)
{
  Clause c = lit->atom->container;
  if (negative_clause(c))
    return Ordered_inference ? maximal_literal_check(lit) : TRUE;
  else
    return FALSE;
}  /* neg_hyper_sat_test */

/*************
 *
 *   hyper_sat_atom()
 *
 *************/

static
void hyper_sat_atom(BOOL flipped, Literal slit, Term atom, BOOL positive,
		    Lindex idx, void (*proc_proc) (Clause))
{
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Term fnd_atom;
  Mindex_pos mate_pos;
  fnd_atom = mindex_retrieve_first(atom,
				   positive ? idx->neg : idx->pos,
				   UNIFY,
				   sat_subst, nuc_subst,
				   FALSE, &mate_pos);
  while (fnd_atom) {
    /* it must be in a nucleus because of the index used above */
    Clause nuc = fnd_atom->container;
    Literal nuc_lit = atom_to_literal(fnd_atom);
    Clash p = NULL;
    Clash first = NULL;
    Literal nlit;
    for (nlit = nuc->literals; nlit; nlit = nlit->next) {
      p = append_clash(p);
      if (first == NULL)
	first = p;
      p->nuc_lit = nlit;
      p->nuc_subst = nuc_subst;
      if (nlit == nuc_lit) {
	p->sat_lit = slit;
	p->sat_subst = sat_subst;
	p->clashable = TRUE;
	p->clashed = TRUE;
	p->flipped = flipped;
      }
      else {
	p->clashable = positive ? !nlit->sign : nlit->sign;
	if (p->clashable) {
	  p->mate_index = positive ? idx->pos : idx->neg;
	  p->sat_subst = get_context();
	}
      }  /* for each literal of nucleus */
    }  /* if we have a nuc */
    clash(first,
	  positive ? pos_hyper_sat_test : neg_hyper_sat_test,
	  HYPER_RES_JUST,
	  proc_proc);
    zap_clash(first);
    fnd_atom = mindex_retrieve_next(mate_pos);
  }  /* for each mate */
  free_context(sat_subst);
  free_context(nuc_subst);
}  /* hyper_sat_atom */

/*************
 *
 *   hyper_satellite()
 *
 *************/

static
void hyper_satellite(Clause c, BOOL positive, Lindex idx,
		     void (*proc_proc) (Clause))
{
  /* Assume c is positive. */
  Literal slit;
  for (slit = c->literals; slit; slit = slit->next) {
    if (!Ordered_inference || maximal_literal_check(slit)) {
      hyper_sat_atom(FALSE, slit, slit->atom, positive, idx, proc_proc);
      if (pos_eq(slit)) {
	Term flip = top_flip(slit->atom);
	hyper_sat_atom(TRUE, slit, flip, positive, idx, proc_proc);
	zap_top_flip(flip);
      }
    }  /* if sat is ok */
  }  /* for each literal of satellite */
}  /* hyper_satellite */

/*************
 *
 *   hyper_nucleus()
 *
 *************/

static
void hyper_nucleus(Clause c, BOOL positive, Lindex idx,
		   void (*proc_proc) (Clause))
{
  /* Assume c is nonpositive. */
  Clash p = NULL;
  Clash first = NULL;
  Literal lit;
  Context nuc_subst = get_context();
  for (lit = c->literals; lit; lit = lit->next) {
    p = append_clash(p);
    if (first == NULL)
      first = p;
    p->clashable = positive ? !lit->sign : lit->sign;
    p->nuc_lit = lit;
    p->nuc_subst = nuc_subst;
    if (p->clashable) {
      p->mate_index = positive ? idx->pos : idx->neg;
      p->sat_subst = get_context();
    }
  }
  clash(first,
	positive ? pos_hyper_sat_test : neg_hyper_sat_test,
	HYPER_RES_JUST,
	proc_proc);
  free_context(nuc_subst);
  zap_clash(first);  /* This also frees satellite contexts. */
}  /* hyper_nucleus */

/*************
 *
 *   hyper_resolution()
 *
 *************/

/* DOCUMENTATION
Hyperresolution.
*/

/* PUBLIC */
void hyper_resolution(Clause c, BOOL positive, Lindex idx,
		      void (*proc_proc) (Clause))
{
  if (positive ? positive_clause(c) : negative_clause(c))
    hyper_satellite(c, positive, idx, proc_proc);
  else
    hyper_nucleus(c, positive, idx, proc_proc);
}  /* hyper_resolution */

/*************
 *
 *   target_check()
 *
 *************/

static
BOOL target_check(Literal lit, int target_constraint)
{
  if (target_constraint == ALL_RES)
    return TRUE;
  else if (target_constraint == POS_RES)
    return lit->sign;
  else if (target_constraint == NEG_RES)
    return !lit->sign;
  else {
    fatal_error("target_check, constraint out of range");
    return FALSE;  /* to please the compiler */
  }
}  /* target_check */

/*************
 *
 *   ur_sat_atom()
 *
 *************/

static
void ur_sat_atom(BOOL flipped, Clause c, int target_constraint,
		 Term sat_atom, Lindex idx,
		 void (*proc_proc) (Clause))

{
  /* Assume C is a unit. */
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Term fnd_atom;
  Mindex_pos mate_pos;
  Literal slit = c->literals;

  fnd_atom = mindex_retrieve_first(sat_atom, (slit->sign?idx->neg:idx->pos),
				   UNIFY, sat_subst, nuc_subst,
				   FALSE, &mate_pos);
  while (fnd_atom) {
    Clause nuc = fnd_atom->container;
    int numlits = number_of_literals(nuc);
    if (numlits > 1 && numlits - 1 <= Nucleus_limit &&
	(!Initial_nuclei || nuc->initial)) {
      Literal fnd_lit = atom_to_literal(fnd_atom);
      Literal target;
      for (target=nuc->literals; target; target=target->next) {
	if (target != fnd_lit && target_check(target, target_constraint)) {
	    
	  Clash p = NULL;
	  Clash first = NULL;
	  Literal nlit;
	  for (nlit = nuc->literals; nlit; nlit = nlit->next) {
	    p = append_clash(p);
	    if (first == NULL)
	      first = p;
	    p->nuc_lit = nlit;
	    p->nuc_subst = nuc_subst;
	    if (nlit == fnd_lit) {
	      p->sat_lit = slit;
	      p->sat_subst = sat_subst;
	      p->clashable = TRUE;
	      p->clashed = TRUE;
	      p->flipped = flipped;
	    }
	    else {
	      p->clashable = (nlit != target);
	      if (p->clashable) {
		p->mate_index = (nlit->sign ? idx->neg : idx->pos);
		p->sat_subst = get_context();
	      }
	    }
	  }  /* for each literal of nucleus */
	  clash(first, unit_check, UR_RES_JUST, proc_proc);
	  zap_clash(first);
	}  
      }  /* for each target */
    }  /* if we have a nuc */
    fnd_atom = mindex_retrieve_next(mate_pos);
  }  /* for each mate */
  free_context(sat_subst);
  free_context(nuc_subst);
}  /* ur_sat_atom */

/*************
 *
 *   ur_satellite()
 *
 *************/

static
void ur_satellite(Clause c, int target_constraint, Lindex idx,
		  void (*proc_proc) (Clause))

{
  Term atom = c->literals->atom;
  ur_sat_atom(FALSE, c, target_constraint, atom, idx, proc_proc);
  /* if equality, try with the flip */
  if (eq_atom(atom)) {
    Term flip = top_flip(atom);
    ur_sat_atom(TRUE, c, target_constraint, flip, idx, proc_proc);
    zap_top_flip(flip);
  }
}  /* ur_satellite */

/*************
 *
 *   ur_nucleus()
 *
 *************/

static
void ur_nucleus(Clause c, int target_constraint, Lindex idx,
		 void (*proc_proc) (Clause))
{
  if (number_of_literals(c) - 1 > Nucleus_limit ||
      (Initial_nuclei && !c->initial))
    return;
  else {
    Literal target;
    for (target = c->literals; target; target = target->next) {
      if (target_check(target, target_constraint)) {
	Clash p = NULL;
	Clash first = NULL;
	Literal lit;
	Context nuc_subst = get_context();
	for (lit = c->literals; lit; lit = lit->next) {
	  p = append_clash(p);
	  if (first == NULL)
	    first = p;
	  p->clashable = (lit != target);
	  p->nuc_lit = lit;
	  p->nuc_subst = nuc_subst;
	  if (p->clashable) {
	    p->mate_index = (lit->sign ? idx->neg : idx->pos);
	    p->sat_subst = get_context();
	  }
	}
	clash(first, unit_check, UR_RES_JUST, proc_proc);
	free_context(nuc_subst);
	zap_clash(first);  /* This also frees satellite contexts. */
      }
    }
  }
}  /* ur_nucleus */

/*************
 *
 *   ur_resolution()
 *
 *************/

/* DOCUMENTATION
Unit-resulting resolution.
*/

/* PUBLIC */
void ur_resolution(Clause c, int target_constraint, Lindex idx,
		   void (*proc_proc) (Clause))
{
  if (unit_clause(c))
    ur_satellite(c, target_constraint, idx, proc_proc);
  else
    ur_nucleus(c, target_constraint, idx, proc_proc);
}  /* ur_resolution */

/*************
 *
 *   instantiate_clause()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause instantiate_clause(Clause c, Context subst)
{
  Clause d = get_clause();
  Literal lit;
  for (lit = c->literals; lit; lit = lit->next)
    append_literal(d, apply_lit(lit, subst));
  return d;
}  /* instantiate_clause */

/*************
 *
 *   binary_parent_test()
 *
 *************/

static
BOOL binary_parent_test(Literal lit, BOOL positive_res)
{
  Clause c = lit->atom->container;
  if (positive_res && lit->sign) {
    if (!positive_clause(c))
      return FALSE;
    else
      return Ordered_inference ? maximal_literal_check(lit) : TRUE;
  }
  else if (!positive_res && !lit->sign) {
    if (!negative_clause(c))
      return FALSE;
    else
      return Ordered_inference ? maximal_literal_check(lit) : TRUE;
  }
  else {
    /* nucleus */
    if (!Ordered_inference)
      return TRUE;
    if (Lit_selection == LIT_SELECTION_MAXIMAL)
      return maximal_literal_check(lit);
    else if (Lit_selection == LIT_SELECTION_ALL)
      return TRUE;
    else if (Lit_selection == LIT_SELECTION_FIRST_MAXIMAL)
      return first_maximal_literal_of_sign(c, !positive_res) == lit;
    else if (Lit_selection == LIT_SELECTION_FIRST)
      return first_literal_of_sign(c, !positive_res) == lit;
    else {
      fatal_error("binary_parent_test, unknown Lit_selection");
      return FALSE;
    }
  }
}  /* binary_parent_test */

/*************
 *
 *   ordered_instance_test()
 *
 *************/

static
BOOL ordered_instance_test(Literal lit, Context subst)
{
  if (!Ordered_inference || !Ordered_instance)
    return TRUE;
  else {
    Clause c = lit->atom->container;
    if (positive_clause(c)) {
      if (maximal_literals(c) == 1)
	return TRUE;
      else {
	int n = literal_number(lit->atom->container, lit);
	Clause d = instantiate_clause(lit->atom->container, subst);
	BOOL result = maximal_literal_test(d, ith_literal(d, n));
	zap_clause(d);
	Ordered_instance_prunes++;
	return result;
      }
    }
    else
      return TRUE;
  }
}  /* ordered_instance_test */

/*************
 *
 *   bin_res()
 *
 *************/

static
void bin_res(BOOL flipped,
	     Literal l1, Context s1,
	     Literal l2, Context s2,
	     void (*proc_proc) (Clause))
{
  Clause r = get_clause();
  Clause nuc =  l1->atom->container;
  Clause sat =  l2->atom->container;
  Ilist j  = NULL;
  Literal l3;
  int i;
  int n = 0;

  /* Include literals in the nucleus. */
  for (l3 = nuc->literals, i=1; l3; l3 = l3->next, i++) {
    if (l3 == l1)
      n = i;  /* index of resolved literal */
    else
      append_literal(r, apply_lit(l3, s1));
  }
  j = ilist_append(j, nuc->id);
  j = ilist_append(j, n);
  
  /* Include literals in the satellite. */
  for (l3 = sat->literals, i=1; l3; l3 = l3->next, i++) {
    if (l3 == l2)
      n = i;  /* index of resolved literal */
    else
      append_literal(r, apply_lit(l3, s2));
  }
  
  j = ilist_append(j, sat->id);
  j = ilist_append(j, flipped ? -n : n);
  
  inherit_attributes(nuc, s1, sat, s2, r);

  r->justification = resolve_just(j, BINARY_RES_JUST);
  upward_clause_links(r);
  (*proc_proc)(r);
}  /* bin_res */

/*************
 *
 *   xx_res()
 *
 *************/

static
void xx_res(Literal lit, void (*proc_proc) (Clause))
{
  Term alpha = ARG(lit->atom,0);
  Term beta  = ARG(lit->atom,1);
  Context subst = get_context();
  Trail tr = NULL;

  if (unify(alpha, subst, beta, subst, &tr)) {
    Clause parent = lit->atom->container;
    int n = literal_number(parent, lit);
    Clause c = get_clause();
    Literal l;
    c->justification = xxres_just(parent, n);
    for (l = parent->literals; l; l = l->next) {
      if (l != lit)
	append_literal(c, apply_lit(l, subst));
    }
    undo_subst(tr);
    upward_clause_links(c);
    c->attributes = inheritable_att_instances(parent->attributes, subst);
    
    (*proc_proc)(c);
  }
  free_context(subst);
}  /* xx_res */

/*************
 *
 *   binary_resolution()
 *
 *************/

/* DOCUMENTATION
Binary resolution.
*/

/* PUBLIC */
void binary_resolution(Clause c,
		       BOOL positive_res,
		       Lindex idx,
		       void (*proc_proc) (Clause))
{
  Context nuc_subst = get_context();
  Context sat_subst = get_context();
  Literal lit;
  for (lit = c->literals; lit; lit = lit->next) {
    if (binary_parent_test(lit, positive_res)) {
      Term sat_atom;
      Mindex_pos mate_pos;
      sat_atom = mindex_retrieve_first(lit->atom,
				       lit->sign ? idx->neg : idx->pos,
				       UNIFY, nuc_subst, sat_subst,
				       FALSE, &mate_pos);
      while (sat_atom) {
	Literal slit = atom_to_literal(sat_atom);
	if (binary_parent_test(slit, positive_res)) {
	  if (ordered_instance_test(lit, nuc_subst) &&
	      ordered_instance_test(slit, sat_subst))
	    bin_res(FALSE, lit, nuc_subst, slit, sat_subst, proc_proc);
	}
	sat_atom = mindex_retrieve_next(mate_pos);
      }
      /* if equality, try for resolution with the flip. */
      if (eq_atom(lit->atom)) {
	Term flip = top_flip(lit->atom);
	sat_atom = mindex_retrieve_first(flip,
					 lit->sign ? idx->neg : idx->pos,
					 UNIFY, nuc_subst, sat_subst,
					 FALSE, &mate_pos);
	while (sat_atom) {
	  Literal slit = atom_to_literal(sat_atom);
	  if (binary_parent_test(slit, positive_res)) {
	    if (ordered_instance_test(lit, nuc_subst) &&
		ordered_instance_test(slit, sat_subst))
	      bin_res(TRUE, lit, nuc_subst, slit, sat_subst, proc_proc);
	  }
	  sat_atom = mindex_retrieve_next(mate_pos);
	}
	zap_top_flip(flip);
      }
      /* try for resolution with x=x */
      if (neg_eq(lit)) {
	xx_res(lit, proc_proc);
      }
    }
  }
  free_context(nuc_subst);
  free_context(sat_subst);
}  /* binary_resolution */

/*************
 *
 *   binary_factors()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void binary_factors(Clause c, void (*proc_proc) (Clause))
{
  Literal l1;
  int i = 1;
  Context subst = get_context();
  for (l1 = c->literals; l1; l1 = l1->next, i++) {
    Literal l2;
    int j = i+1;
    for (l2 = l1->next; l2; l2 = l2->next, j++) {

      Trail tr = NULL;
      if (l1->sign == l2->sign &&
	  /* maximal_literal_check(l1) && maximal_literal_check(l2) &&  */
	  unify(l1->atom,subst,l2->atom,subst,&tr)) {
	Clause f = get_clause();
	Literal l3;
	f->justification = factor_just(c, i, j);
	for (l3 = c->literals; l3; l3 = l3->next) {
	  if (l3 != l2)
	    append_literal(f, apply_lit(l3, subst));
	}
	undo_subst(tr);
	upward_clause_links(f);
	f->attributes = cat_att(f->attributes,
				inheritable_att_instances(c->attributes,
							  subst));
	(*proc_proc)(f);
      }
    }
  }
  free_context(subst);
}  /* binary_factors */

/*************
 *
 *   merge_literals0()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void merge_literals0(Clause c)
{
  Literal l1;
  int n1;
  BOOL null_literals = FALSE;
  for (l1 = c->literals, n1 = 1; l1; l1 = l1->next, n1++) {
    if (l1->atom) {
      Literal l2;
      int n2;
      for (l2 = l1->next, n2 = n1+1; l2; l2 = l2->next, n2++) {
	if (l2->atom) {
	  if (l1->sign == l2->sign && term_ident(l1->atom, l2->atom)) {
	    zap_term(l2->atom);
	    c->justification = append_just(c->justification, merge_just(n2));
	    l2->atom = NULL;  /* remove it below */
	    null_literals = TRUE;
	  }
	}
      }
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
}  /* merge_literals0 */

/*************
 *
 *   merge_literals()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void merge_literals(Clause c)
{
  Literal l3;
  int n;
  BOOL null_literals = FALSE;
  for (l3 = c->literals, n = 1; l3; l3 = l3->next, n++) {
    Literal l2;
    for (l2 = c->literals; l2 != l3; l2 = l2->next) {
      if (l2->atom && l3->atom) {
	if (l3->sign == l2->sign && term_ident(l3->atom, l2->atom)) {
	  zap_term(l3->atom);
	  c->justification = append_just(c->justification, merge_just(n));
	  l3->atom = NULL;  /* remove it below */
	  null_literals = TRUE;
	}
      }
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
}  /* merge_literals */

/*************
 *
 *   copy_inference()
 *
 *************/

/* DOCUMENTATION
This makes a "copy" inference; that is, a copy of the clause
in which the justification is "copy".
All attributes are copied (not just the inheritible attributes).
An ID is not assigned.
*/

/* PUBLIC */
Clause copy_inference(Clause c)
{
  Clause new = copy_clause(c);
  /* Don't copy the justification; build a "copy" justification. */
  new->justification = copy_just(c);
#if 1
  new->attributes = copy_attributes(c->attributes);
#else
  /* We could use the following to copy inheritable attributes only. */
  new->attributes = inheritable_att_instances(c->attributes, NULL);
#endif
  return new;
}  /* copy_inference */

/*************
 *
 *   resolve2()
 *
 *************/

/* DOCUMENTATION
Resolve, if possible, two clauses on the literals (specified
by literals, counting from 1).
Renumber vars, include justification, transfer inheritable
attributes, but do not assign an ID.
<P>
if n2 < 0, then the literal is abs(n2), and it should be flipped.
*/

/* PUBLIC */
Clause resolve2(Clause c1, int n1, Clause c2, int n2)
{
  Clause res;
  Literal l1 = ith_literal(c1, n1);
  Literal l2 = ith_literal(c2, abs(n2));
  Term a1 = l1->atom;
  Term a2 = l2->atom;
  Context s1 = get_context();
  Context s2 = get_context();
  Trail tr = NULL;
  Term a2x;

  if (n2 < 0)
    a2x = top_flip(a2);
  else
    a2x = a2;

  if (l1->sign != l2->sign && unify(a1, s1, a2x, s2, &tr)) {
    Literal lit;
    res = get_clause();
    for (lit = c1->literals; lit; lit = lit->next)
      if (lit != l1)
	append_literal(res, apply_lit(lit,  s1));
    for (lit = c2->literals; lit; lit = lit->next)
      if (lit != l2)
	append_literal(res, apply_lit(lit,  s2));

    inherit_attributes(c1, s1, c2, s2, res);
    res->justification = binary_res_just(c1, n1, c2, n2);
    upward_clause_links(res);
    renumber_variables(res, MAX_VARS);
    undo_subst(tr);
  }
  else
    res = NULL;

  if (n2 < 0)
    zap_top_flip(a2x);

  free_context(s1);
  free_context(s2);
  return res;
}  /* resolve2 */

/*************
 *
 *   resolve3()
 *
 *************/

/* DOCUMENTATION
Similar to resolve2(), but literals are given instead of integers.
*/

/* PUBLIC */
Clause resolve3(Clause c1, Literal l1, Clause c2, Literal l2)
{
  return resolve2(c1, literal_number(c1, l1),
		  c2, literal_number(c2, l2));
}  /* resolve3 */

/*************
 *
 *   xx_resolve2()
 *
 *************/

/* DOCUMENTATION
Resolve, if possible, a clause with x=x.
Renumber vars, include justification, transfer inheritable
attributes, but do not assign an ID.
*/

/* PUBLIC */
Clause xx_resolve2(Clause c, int n)
{
  Clause res;
  Literal l = ith_literal(c, n);
  Context s = get_context();
  Trail tr = NULL;

  if (neg_eq(l) &&
      unify(ARG(l->atom,0), s,
	    ARG(l->atom,1), s, &tr)) {
    Literal lit;
    res = get_clause();
    for (lit = c->literals; lit; lit = lit->next)
      if (lit != l)
	append_literal(res, apply_lit(lit,  s));

    res->attributes = inheritable_att_instances(c->attributes, s);
    res->justification = xxres_just(c, n);
    upward_clause_links(res);
    renumber_variables(res, MAX_VARS);
    undo_subst(tr);
  }
  else
    res = NULL;
  free_context(s);
  return res;
}  /* xx_resolve2 */

