#ifndef TP_SYMBOLS_H
#define TP_SYMBOLS_H

#include "string.h"
#include "strbuf.h"
#include "order.h"
#include "glist.h"

/* INTRODUCTION
This collection of routines manages the global symbol table.
Each symbol is a pair (string,arity) and has a unique ID number.
For example, ("f",2), sometimes written f/2, is a different symbol
from f/3.
<P>
These symbols are used mostly as constant, function, and predicate
symbols, but they can be used for variables and other things as
well.
<P>
The symbol table routines call malloc() and free() directly.
The LADR memory package (tp_alloc()) is not used.
*/

/* Public definitions */

/* maximum number of chars in string part of symbol (includes '\0') */

#define MAX_NAME      51

/* parse/print properties of symbols */

typedef enum {NOTHING_SPECIAL,
	      INFIX,         /* xfx */
	      INFIX_LEFT ,   /* yfx */
	      INFIX_RIGHT,   /* xfy */
	      PREFIX_PAREN,  /* fx  */
	      PREFIX,        /* fy  */
	      POSTFIX_PAREN, /* xf  */
	      POSTFIX        /* yf  */
             } Parsetype;

#define MIN_PRECEDENCE      1
#define MAX_PRECEDENCE    999

/* Function/relation properties of symbols */

typedef enum { UNSPECIFIED_SYMBOL,
	       FUNCTION_SYMBOL,
	       RELATION_SYMBOL
             } Symbol_type;

/* Unification properties of symbols */

typedef enum { EMPTY_THEORY,
	       COMMUTE,
	       ASSOC_COMMUTE
             } Unif_theory;

/* LRPO status */

typedef enum { LRPO_LR_STATUS,
	       LRPO_MULTISET_STATUS
             } Lrpo_status;

/* Variable style */

typedef enum { STANDARD_STYLE,      /* x,y,z,... */
	       PROLOG_STYLE,        /* A,B,C,... */
	       INTEGER_STYLE        /* 0,1,2,... */
             } Variable_style;

/* Logic symbols when in Term form */

#define TRUE_SYM    "$T"
#define FALSE_SYM   "$F"
#define AND_SYM     "&"
#define OR_SYM      "|"
#define NOT_SYM     "-"
#define IFF_SYM     "<->"
#define IMP_SYM     "->"
#define IMPBY_SYM   "<-"
#define ALL_SYM     "all"
#define EXISTS_SYM  "exists"
#define QUANT_SYM   "$quantified"

/* Other symbols when in Term form */

#define ATTRIB_SYM  "#"    /* operator for attaching attributes */
#define EQ_SYM      "="    /* for inference rules that involve equality */
#define NEQ_SYM     "!="   /* abbreviation for negation of EQ_SYM */

typedef struct symbol * Symbol;

/* End of public definitions */

/* Public function prototypes from symbols.c */

int str_to_sn(char *str, int arity);

void set_preliminary_precedence(Plist strings);

BOOL exists_preliminary_precedence();

void all_symbols_lrpo_status(Lrpo_status status);

void fprint_syms(FILE *fp);

void p_syms(void);

char *sn_to_str(int symnum);

int sn_to_arity(int symnum);

int sn_to_occurrences(int symnum);

BOOL is_symbol(int symnum, char *str, int arity);

BOOL is_eq_symbol(int symnum);

void set_lex_val(int symnum, int lex_val);

void set_skolem(int symnum);

void set_unfold_symbol(int symnum);

BOOL is_unfold_symbol(int symnum);

void set_assoc_comm(char *str, BOOL set);

void set_commutative(char *str, BOOL set);

BOOL assoc_comm_symbols(void);

BOOL comm_symbols(void);

BOOL is_assoc_comm(int sn);

BOOL is_commutative(int sn);

void fprint_sym(FILE *fp, int symnum);

void sprint_sym(String_buf sb, int symnum);

void p_sym(int symnum);

int sn_to_lex_val(int sn);

Lrpo_status sn_to_lrpo_status(int sn);

void set_parse_type(char *str, int precedence, Parsetype type);

BOOL binary_parse_type(char *str, int *precedence, Parsetype *type);

BOOL unary_parse_type(char *str, int *precedence, Parsetype *type);

int special_parse_type(char *str);

BOOL variable_name(char *s);

BOOL str_exists(char *str);

void skolem_check(BOOL flag);

int next_skolem_symbol(int arity);

int fresh_symbol(char *prefix, int arity);

int greatest_symnum(void);

Ordertype sym_precedence(int symnum_1, int symnum_2);

void set_variable_style(Variable_style style);

Variable_style variable_style(void);

Ilist syms_with_lex_val(void);

void set_symbol_type(int symnum, Symbol_type type);

Symbol_type get_symbol_type(int symnum);

BOOL function_symbol(int symnum);

BOOL relation_symbol(int symnum);

void lex_order_3(Ilist fsyms, Ilist rsyms,
		 Mlist fsyms_multiset, Mlist rsyms_multiset);

Ilist sort_by_lex_val(Ilist p);

Ilist current_fsym_precedence();

Ilist current_rsym_precedence();

void declare_functions_and_relations(Ilist fsyms, Ilist rsyms);

Ilist skolem_symbols(void);

int not_symnum(void);

int or_symnum(void);

void declare_base_symbols(void);

int function_or_relation_sn(char *str);

void set_kb_weight(int symnum, int weight);

BOOL zero_wt_kb(void);

int sn_to_kb_wt(int symnum);

Ilist not_in_preliminary_precedence(Ilist syms);

void assign_greatest_precedence(int symnum);

void new_constant_properties(int sn);

Ilist all_function_symbols(void);

Ilist all_relation_symbols(void);

void print_fsym_precedence(FILE *fp);

void print_rsym_precedence(FILE *fp);

void print_kb_weights(FILE *fp);

void set_lrpo_status(int symnum, Lrpo_status status);

BOOL is_skolem(int symnum);

void skolem_reset(void);

void decommission_skolem_symbols(void);

int gen_new_symbol(char *prefix, int arity, Ilist syms);

void lex_insert_after_initial_constants(Ilist syms);

#endif  /* conditional compilation of whole file */
