#include "tlist.h"

/* Private definitions and types */

/*************
 *
 *   tlist_member()
 *
 *************/

/* DOCUMENTATION
This function checks if a term is a member of a Plist.
The function term_ident(t1,t2) is used.
*/

/* PUBLIC */
BOOL tlist_member(Term t, Plist lst)
{
  if (lst == NULL)
    return FALSE;
  else if (term_ident(lst->v, t))
    return TRUE;
  else
    return tlist_member(t, lst->next);
}  /* tlist_member */

/*************
 *
 *   zap_tlist()
 *
 *************/

/* DOCUMENTATION
Zap each term in the list and free the list.  If you want to free the
list only, without zapping the terms, use zap_plist(p) instead.
*/

/* PUBLIC */
void zap_tlist(Plist p)
{
  if (p != NULL) {
    zap_tlist(p->next);
    zap_term(p->v);
    free_plist(p);
  }
}  /* zap_tlist */

/*************
 *
 *   tlist_remove()
 *
 *************/

/* DOCUMENTATION
Remove a term from a Plist.  Term_ident() is used, and the
term in the list is zapped.
*/

/* PUBLIC */
Plist tlist_remove(Term t, Plist p)
{
  if (p == NULL)
    return NULL;
  else {
    p->next = tlist_remove(t, p->next);

    if (term_ident(t, p->v)) {
      Plist next = p->next;
      zap_term(p->v);
      free_plist(p);
      return next;
    }
    else
      return p;
  }
}  /* tlist_remove */

/*************
 *
 *   tlist_union()
 *
 *************/

/* DOCUMENTATION
Return the union of two Plists of terms.  If the inputs
are sets, the output is a set.  Do not refer the inputs
after the call.  Duplicates are zapped.
*/

/* PUBLIC */
Plist tlist_union(Plist a, Plist b)
{
  if (a == NULL)
    return b;
  else if (tlist_member(a->v, b)) {
    Plist c = tlist_union(a->next,b );
    zap_term(a->v);
    free_plist(a);
    return c;
  }
  else {
    a->next = tlist_union(a->next, b);
    return a;
  }
}  /* tlist_union */

/*************
 *
 *   constants_in_term()
 *
 *************/

/* DOCUMENTATION
Given a term t, return a plist containing the set of constants in t.
*/

/* PUBLIC */
Plist constants_in_term(Term t, Plist p)
{
  if (VARIABLE(t))
    return p;
  else if (CONSTANT(t)) {
    if (tlist_member(t, p))
      return p;
    else
      return plist_prepend(p, t);
  }
  else {
    int i;
    for (i = 0; i < ARITY(t); i++)
      p = constants_in_term(ARG(t, i), p);
    return p;
  }
}  /* constants_in_term */

/*************
 *
 *   tlist_copy()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist tlist_copy(Plist p)
{
  if (p == NULL)
    return NULL;
  else {
    Term t = p->v;
    Plist new = get_plist();
    new->v = copy_term(t);
    new->next = tlist_copy(p->next);
    return new;
  }
}  /* tlist_copy */

