#include "actions.h"

/* Private definitions and types */

/* These lists are for rules such as
     given = 13 -> set(print_kept).
   This is a slow version, because the rules are parsed
   each time they might be applied, that is, at each
   given, generated, and kept clause.
   We might have to change to a more efficient data structure.
*/

static Plist Given_rules     = NULL;
static Plist Generated_rules = NULL;
static Plist Kept_rules      = NULL;

/*************
 *
 *   init_action()
 *
 *************/

/* DOCUMENTATION
This is one of two ways to set up an action.
<P>
Also see register_action(), for which the rule Term is not yet built.
*/

/* PUBLIC */
void init_action(Term t)
{
  Term rule = copy_term(t);
  Term trigger, stat;

  if (!is_term(rule, "->", 2))
    fatal_error("init_actions, bad rule");
  trigger = ARG(rule,0);
  if (!is_term(trigger, EQ_SYM, 2))
    fatal_error("init_actions, bad trigger");
  stat = ARG(trigger,0);

  if (is_constant(stat, "given"))
    Given_rules = plist_append(Given_rules, rule);
  else if (is_constant(stat, "generated"))
    Generated_rules = plist_append(Generated_rules, rule);
  else if (is_constant(stat, "kept"))
    Kept_rules = plist_append(Kept_rules, rule);
  else
    fatal_error("init_action, bad statistic");
}  /* init_action */

/*************
 *
 *   init_actions()
 *
 *************/

/* DOCUMENTATION
The actions are copied.
*/

/* PUBLIC */
void init_actions(Plist rules)
{
  /* rule:          trigger -> action
     Example:     given=100 -> assign(max_weight, 25)
  */
  Plist p;
  for (p = rules; p; p = p->next)
    init_action(p->v);
}  /* init_actions */

/*************
 *
 *   register_action()
 *
 *************/

/* DOCUMENTATION
This is one of two ways to set up an action.
<P>
Also see init_action(), which takes the rule as a Term.
*/

/* PUBLIC */
void register_action(char *stat, char *val, char *op, char *arg1, char *arg2)
{
  Term action, rule;
  
  if (str_ident(op, "set") || str_ident(op, "clear"))
    action = term1(op, term0(arg1));
  else if (str_ident(op, "assign"))
    action = term2(op, term0(arg1), term0(arg2));
  else {
    fatal_error("register_action, operation must be set, clear, or assign");
    action = NULL;  /* to satisfy the compiler */
  }

  rule = term2("->",
	       term2("=",
		     term0(stat),
		     term0(val)),
	       action);
  
  if (str_ident(stat, "given"))
    Given_rules = plist_append(Given_rules, rule);
  else if (str_ident(stat, "generated"))
    Generated_rules = plist_append(Generated_rules, rule);
  else if (str_ident(stat, "kept"))
    Kept_rules = plist_append(Kept_rules, rule);
  else
    fatal_error("register_action, bad statistic");

  printf("\n%% action: "); fwrite_term_nl(stdout, rule);
  
}  /* register_action */

/*************
 *
 *   statistic_actions()
 *
 *************/

/* DOCUMENTATION
Given a statistic and a value, look in the rules
(given to init_actions) and apply as appropriate.

*/

/* PUBLIC */
void statistic_actions(char *stat, int n)
{
  Plist p = NULL;
  if (str_ident(stat, "given"))
    p = Given_rules;
  else if (str_ident(stat, "generated"))
    p = Generated_rules;
  else if (str_ident(stat, "kept"))
    p = Kept_rules;
  else
    fatal_error("statistic_actions, unknown stat");
  
  for (; p; p = p->next) {
    Term rule = p->v;
    Term trigger = ARG(rule,0);
    Term tval = ARG(trigger,1);
    Term action = ARG(rule,1);
    int ival;
    if (!term_to_int(tval, &ival))
      fatal_error("statistic_actions, bad integer in trigger");
    if (n == ival) {
      printf("\n%% Applying rule: ");
      fwrite_term_nl(stdout, rule);
      if (is_term(action, "set", 1) || is_term(action, "clear", 1)) {
	flag_handler(stdout, action, TRUE, KILL_UNKNOWN);
      }
      else if (is_term(action, "assign", 2)) {
	parm_handler(stdout, action, TRUE, KILL_UNKNOWN);
      }
      else {
	fatal_error("statistic_action, unknown action");
      }
    }
  }
}  /* statistic_actions */

/*************
 *
 *   proof_action()
 *
 *   Given a clause and an attribute ID; for each attribute
 *   of that type, look for a "trigger -> action".
 *
 *   The trigger here is "in_proof", and the actions supported are exit,
 *   set(flag), clear(flag), assign(parm, i), assign(stringparm, string).
 *
 *************/

static
void proof_action(Clause c, int attribute_id, void (*exit_proc)(int))
{
  int i = 0;
  Term t = get_term_attribute(c->attributes, attribute_id, ++i);
  while (t) {
    if (is_term(t, "->", 2) &&
	is_constant(ARG(t,0), "in_proof")) {
      Term action = ARG(t,1);
      if (is_constant(action, "exit"))
	exit_proc(ACTION_EXIT);
      else if (is_term(action, "set", 1) || is_term(action, "clear", 1)) {
	printf("proof action: ");
	fwrite_term_nl(stdout, action);
	flag_handler(stdout, action, TRUE, KILL_UNKNOWN);
      }
      else if (is_term(action, "assign", 2)) {
	printf("proof action: ");
	fwrite_term_nl(stdout, action);
	parm_handler(stdout, action, TRUE, KILL_UNKNOWN);
      }
      else {
	fprintf(stderr, "\nUnknown proof action: ");
	fwrite_term_nl(stderr, t);
      }
    }
    t = get_term_attribute(c->attributes, attribute_id, ++i);
  }
}  /* proof_action */

/*************
 *
 *   actions_in_proof()
 *
 *************/

/* DOCUMENTATION
Quick and dirty, until a general mechanism gets designed ...
If the proof contains any #action(in_proof -> <action>) attributes,
perform the action.

Actions supported here:  exit, set, clear, assign.
*/

/* PUBLIC */
void actions_in_proof(Plist proof,
		      Prover_attributes att,
		      void (*exit_proc)(int))
{
  Plist p;
  for (p = proof; p; p = p->next) {
    proof_action(p->v, att->action, exit_proc);
    proof_action(p->v, att->action2, exit_proc);
  }
}  /* actions_in_proof */

