#define PROVER_NAME     "Autosketches"
#define PROVER_VERSION  "9"
#include "../PROGRAM_DATE.h"

#include "../ladr/nonport.h"
#include "../ladr/top_input.h"
#include "../ladr/xproofs.h"
#include "utilities.h"
#include "prover_exits.h"
#include "search.h"

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  printf("----- %s %s, %s -----\n", PROVER_NAME, PROVER_VERSION, PROGRAM_DATE);
  printf("Process %d was started by %s on %s,\n%s",
	 my_process_id(), username(), hostname(), get_date());
	 
  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".\n");
}  // print_banner

/*************
 *
 *   attach_sketch_hint_labels()
 *
 *   For each clause in a Plist, if it has no labels, append one,
 *   e.g., label(S4_H5), meaning sketch 4 hint 5.
 *
 *************/

static
void attach_sketch_hint_labels(int sketch_count, Plist hints)
{
  Plist p;
  int hint_count;
  for (p = hints, hint_count = 1; p; p = p->next, hint_count++) {
    Clause c = p->v;
    if (!exists_attribute(c->attributes, label_att())) {
      char *s;
      String_buf sb = get_string_buf();
      sb_append(sb, "S");
      sb_append_int(sb, sketch_count);
      sb_append(sb, "_H");
      sb_append_int(sb, hint_count);
      s = sb_to_malloc_string(sb);
      zap_string_buf(sb);
      c->attributes = set_string_attribute(c->attributes, label_att(), s);
      free(s);
    }
  }
}  /* attach_sketch_hint_labels */

/*************
 *
 *    main -- autosketches!
 *
 *************/

int main(int argc, char **argv)
{
  Prover_input input;

  int use_expanded_proofs_flag;     // special-purpose option
  Plist assumptions;                // special-purpose list

  /************************ General initialization ***************************/

  print_banner(argc, argv);         // local routine

  /************************ Initialize our own options and lists *************/

  use_expanded_proofs_flag = init_flag("use_expanded_proofs", TRUE);
  accept_list("assumptions",  CLAUSES,  &assumptions);

  /************************ Read the input ***********************************/

  input = std_prover_init_and_input(argc, argv,
			   TRUE,           // echo input to stdout
			   KILL_UNKNOWN);  // unknown flags/parms are fatal

  // tell search() to return expanded proofs as well as ordinary proofs
  input->xproofs = flag(use_expanded_proofs_flag);
  
  /******************** Search for a proof *******************************/

  {
    int sketch_count = 0;
    double total_user_seconds = 0.0;    // running total for all searches
    double total_system_seconds = 0.0;  // running total for all searches

    Plist original_sos   = input->sos;
    Plist original_hints = input->hints;
    
    Plist assumptions_work = copy_plist(assumptions);       // shallow
    Plist hints_work       = copy_plist(original_hints);    // shallow

    while (TRUE) {  // exit from within loop

      Prover_results results;

      Plist sos_work = plist_cat(copy_plist(original_sos),
				 copy_plist(assumptions_work));  // shallow

      fprintf(stdout,"\nStarting a search with assumptions:\n\n");
      fwrite_clause_list(stdout, assumptions_work, "assumptions");

      // note that sos_work and hints_work change as we iterate

      input->sos   = sos_work;
      input->hints = hints_work;

      results = forking_search(input);

      total_user_seconds   += results->user_seconds;
      total_system_seconds += results->system_seconds;

      if (results->proofs == NULL) {
	printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
	       total_user_seconds, total_system_seconds);
	exit_with_message(stdout, results->return_code);
      }

      else if (assumptions_work == NULL) {
	printf("The preceding proof was found with no assumptions, "
	       "so we are done.\n");
	printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
	       total_user_seconds, total_system_seconds);
	exit_with_message(stdout, results->return_code);
      }

      zap_plist(sos_work);      // shallow zap
	
      // append proof to hints

      {
	Plist assumps_in_proof, inputs_in_proof, new_hints;

	sketch_count++;

	if (flag(use_expanded_proofs_flag))
	  new_hints = copy_clauses_ija(results->xproofs->v); // deep copy
	else
	  new_hints = copy_clauses_ija(results->proofs->v);  // deep copy

	attach_sketch_hint_labels(sketch_count, new_hints);  // label(S3_H45)

	hints_work = plist_cat(hints_work, new_hints); // shallow, uses up args

	// get assumptions that occur as input clauses in the proof

	inputs_in_proof = input_clauses(results->proofs->v);       // shallow
	assumps_in_proof = intersect_clauses(assumptions_work,
					     inputs_in_proof);  // shallow

	if (assumps_in_proof == NULL) {
	  printf("\nThe following assumptions were available for\n"
		 "the preceding proof, but they were not used.\n"
		 "Therefore we are done.\n\n");
	  fwrite_clause_list(stdout,assumptions_work,"assumptions_not_used");
	  printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
		 total_user_seconds, total_system_seconds);
	  exit_with_message(stdout, MAX_PROOFS_EXIT);
	}
	else {
	  Clause assump = assumps_in_proof->v;  // remove this one
	  printf("\nThe preceding proof uses the following assmumptions.\n"
		 "The first of these will be removed.\n\n");
	  fwrite_clause_list(stdout,assumps_in_proof,"assumptions_in_proof");
	  assumptions_work = plist_remove(assumptions_work, assump);
	}
	zap_plist(inputs_in_proof);   // shallow zap
	zap_plist(assumps_in_proof);  // shallow zap
      }
      zap_prover_results(results);
    }  // end of while loop
  }

  exit(1);  // to satisfy the compiler (won't be called)

}  // main
