#define PROVER_NAME     "Autosketches"
#define PROVER_VERSION  "5 (learn)"
#include "../PROGRAM_DATE.h"

#include "../ladr/nonport.h"
#include "../ladr/top_input.h"
#include "../ladr/xproofs.h"
#include "utilities.h"
#include "prover_exits.h"
#include "search.h"

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  printf("----- %s %s, %s -----\n", PROVER_NAME, PROVER_VERSION, PROGRAM_DATE);
  printf("Process %d was started by %s on %s,\n%s",
	 my_process_id(), username(), hostname(), get_date());
	 
  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".\n");
}  // print_banner

/*************
 *
 *   attach_sketch_hint_labels()
 *
 *   For each clause in a Plist, if it has no labels, append one,
 *   e.g., label(S4_H5), meaning sketch 4 hint 5.
 *
 *************/

static
void attach_sketch_hint_labels(int sketch_count, Plist hints)
{
  Plist p;
  int hint_count;
  for (p = hints, hint_count = 1; p; p = p->next, hint_count++) {
    Clause c = p->v;
    if (!exists_attribute(c->attributes, label_att())) {
      char *s;
      String_buf sb = get_string_buf();
      sb_append(sb, "S");
      sb_append_int(sb, sketch_count);
      sb_append(sb, "_H");
      sb_append_int(sb, hint_count);
      s = sb_to_malloc_string(sb);
      zap_string_buf(sb);
      c->attributes = set_string_attribute(c->attributes, label_att(), s);
      free(s);
    }
  }
}  /* attach_sketch_hint_labels */

/*************
 *
 *   ith_clause_on_plist()
 *
 *   Get the i-th clause on a plist of clauses.
 *
 *   Note:  Should generalize this to any plist.
 *
 *************/

static
Clause ith_clause_on_plist(Plist p, int i)
{
   Clause temp_cl;
 
   if (p == NULL || i < 1)
      return NULL;
   else if (i == 1)
   {
      temp_cl = p->v;
      return temp_cl;
   }
   else
      return ith_clause_on_plist(p->next, i-1);
}  /* ith_clause_on_plist */

/*************
 *
 *    main -- autosketches!
 *
 *************/

int main(int argc, char **argv)
{
  Prover_input input;
  int use_expanded_proofs_flag;     // special-purpose option
  int aggressive_sketches_flag;     // special-purpose option, BV
  Plist assumptions;                // special-purpose list

  /************************ General initialization ***************************/

  print_banner(argc, argv);         // local routine

  /************************ Initialize our own options and lists *************/

  use_expanded_proofs_flag = init_flag("use_expanded_proofs", TRUE);
  aggressive_sketches_flag = init_flag("aggressive_sketches", FALSE);  // BV
  accept_list("assumptions",  CLAUSES,  &assumptions);

  /********************* Read the input **************************************/

  input = std_prover_init_and_input(argc, argv,
			   TRUE,           // echo input to stdout
			   KILL_UNKNOWN);  // unknown flags/parms are fatal

  // tell search() to return expanded proofs as well as ordinary proofs
  input->xproofs = flag(use_expanded_proofs_flag);
  
  /************************ Search for a proof *******************************/

  { // main search block (Bill's style)

    // declarations and definitions local to main search block

    int sketch_count = 0;
    int list_size;
    double total_user_seconds = 0.0;    // running total for all searches
    double total_system_seconds = 0.0;  // running total for all searches

    Plist original_sos   = input->sos;
    Plist original_hints = input->hints;

    Plist assumptions_work = copy_plist(assumptions);    // shallow
    Plist hints_work       = copy_plist(original_hints); // shallow
    
    Clause last_assump = NULL;   // want to remember between iterations

    int done = FALSE;

    while (!done)
    {
      Prover_results results;

      Plist sos_work = plist_cat(copy_plist(original_sos),
				 copy_plist(assumptions_work));  // shallow

      sketch_count++ ;

      fprintf(stdout,"\n***********************************************\n");
      fprintf(stdout,"AUTO SKETCHES ITERATION %d\n", sketch_count);
      fprintf(stdout,"***********************************************\n\n");

      list_size = plist_count(assumptions_work);
      fprintf(stdout,"\nStarting a search with %d assumptions:\n\n", list_size);
      fwrite_clause_list(stdout, assumptions_work, "assumptions");

      // note that sos_work and hints_work change as we iterate
      
      input->sos   = sos_work;
      input->hints = hints_work;

      results = forking_search(input);

      total_user_seconds   += results->user_seconds;
      total_system_seconds += results->system_seconds;

      // how to proceed depends on the results of the most recent search

      /*
         proof found:
            if list of available assumptions empty
               stop - final proof found
            else
               if list of assumptions actually used is empty
                  stop - final proof found
               else
                  append most recent proof to hints
                  delete last assumption used in proof and search again
         proof not found:
            if first iteration (all assumptions available)
               stop with failure
            else if last iteration (working list of assumptions is empty)
               start over with all but most recently deleted assumption
               note:  make permanent change to input list of assumptions
               note:  this case is highly experimental
            else (current working list of assumptions is not empty)
               add most recent deleted back in and delete something else
      */

      // ***********
      // proof found
      // ***********

      if (results->proofs != NULL)
      {
        // no assumptions available, so done
        if (assumptions_work == NULL)
        {
	  printf("The preceding proof was found with no assumptions, "
	    "so we are done.\n");
          fprintf(stdout,"\n***********************************************\n");
          fprintf(stdout,"%d AUTO SKETCHES ITERATIONS USED\n", sketch_count);
          fprintf(stdout,"***********************************************\n\n");
	  printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
	     total_user_seconds, total_system_seconds);
	  exit_with_message(stdout, results->return_code);
        }

        // prepare for next iteration
        else
        {
          // find list of assumptions actually used

	  Plist proof, assumps_in_proof, inputs_in_proof;

	  proof = results->proofs->v;  // use first proof only

          // get assumptions that occur as input clauses in the proof

	  inputs_in_proof = input_clauses(proof);       // shallow
	  assumps_in_proof = intersect_clauses(assumptions_work,
	    				       inputs_in_proof);  // shallow

          // no assumptions used, so done
	  if (assumps_in_proof == NULL)
          {
	     printf("\nThe following assumptions were available for\n"
		    "the preceding proof, but they were not used.\n"
		    "Therefore we are done.\n\n");
	     fwrite_clause_list(stdout,assumptions_work,"assumptions_not_used");
             printf("\n***********************************************\n");
             printf("%d AUTO SKETCHES ITERATIONS USED\n", sketch_count);
             printf("***********************************************\n\n");
	     printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
		    total_user_seconds, total_system_seconds);
	     exit_with_message(stdout, MAX_PROOFS_EXIT);
	  }
	  else
          {
	    Plist new_hints;

	    if (flag(use_expanded_proofs_flag))
	      new_hints = copy_clauses_ija(results->xproofs->v); // deep copy
	    else
	      new_hints = copy_clauses_ija(results->proofs->v);  // deep copy

	    attach_sketch_hint_labels(sketch_count, new_hints);  // label(S3_H45)
 
	    hints_work = plist_cat(hints_work, new_hints); // shallow, args lost

            list_size = plist_count(assumps_in_proof);

  	    printf("\nThe preceding proof (%d) uses the following %d assumptions.\n"
		 "The last of these will be removed.\n\n", sketch_count, list_size);
	    fwrite_clause_list(stdout,assumps_in_proof,"assumptions_in_proof");

            // modify to account for aggressive sketches search
            // (use only assumptions that were used in previous proof)
	    if (flag(aggressive_sketches_flag))
            {
 	      zap_plist(assumptions_work);   // shallow zap
              assumptions_work = copy_plist(assumps_in_proof);
	    }

            // remove last assumption from list used in proof
            last_assump  = ith_clause_on_plist(assumps_in_proof,list_size);
  	    assumptions_work = plist_remove(assumptions_work, last_assump);
          }

	  zap_plist(inputs_in_proof);   // shallow zap
	  zap_plist(assumps_in_proof);  // shallow zap
        }
      }   // case: proof found

      // ***************
      // proof not found
      // ***************

      else
      {
         // first iteration, so stop with failure
         // if (sketch_count == 1)
         if (last_assump == NULL)
         {
            fprintf(stdout,"\n***********************************************\n");
            fprintf(stdout,"NO PROOF EVEN WITH FULL ASSUMPTIONS LIST\n");
            fprintf(stdout,"***********************************************\n\n");
	    printf("\nTotal user_CPU=%.2f, system_CPU=%.2f.\n",
	           total_user_seconds, total_system_seconds);
	    exit_with_message(stdout, results->return_code);
         }
         
         // last iteration (no assumptions)
         //    start all over with all but most recent
         else if (plist_count(assumptions_work) == 0)
         {
           fprintf(stdout,"\n***********************************************\n");
           fprintf(stdout,"NO PROOF WITHOUT ANY ASSUMPTIONS\n");
           fprintf(stdout,"***********************************************\n\n");
           // make permanent change to original assumptions list
  	   assumptions = plist_remove(assumptions, last_assump);
           
           // reset as if new start
           last_assump = NULL;

 	   zap_plist(assumptions_work);   // shallow zap
           assumptions_work = copy_plist(assumptions);  // shallow
         }

         else
         {
           // insert most recently deleted assumption at beginning
           // of working assumption list and try again, it should
           // get proof right away (same set of assumptions as previous
           // iteration, but possibly a different proof) and then
           // will try deleting something different
           //
           // IMPORTANT: need to prevent this from becoming an infinite
           // loop of failed assumptions!
           plist_prepend(assumptions_work, last_assump);
           fprintf(stdout,"\n***********************************************\n");
           fprintf(stdout,"RETRY WITH DIFFERENT ASSUMPTION DELETED\n");
           fprintf(stdout,"***********************************************\n\n");
         }

      }   // case: proof not found

      zap_plist(sos_work);      // shallow zap
      zap_prover_results(results);

    }  // while loop

  }  // main search block

  exit(1);  // to satisfy the compiler (won't be called)

}  // main
