#include "pred_elim.h"

#include "../ladr/clause2.h"
#include "../ladr/subsume.h"
#include "../ladr/resolve.h"

/* Private definitions and types */

/*************
 *
 *   rsym_occurrences() - number of occurrences of a pred symbol in a clause
 *
 *************/

static
int rsym_occurrences(int symbol, Clause c)
{
  int n = 0;
  Literal lit;
  for (lit = c->literals; lit; lit = lit->next)
    if (SYMNUM(lit->atom) == symbol)
      n++;
  return n;
}  /* rsym_occurrences */

/*************
 *
 *   arg_check()
 *
 *************/

static
BOOL arg_check(int symbol, Clause c)
{
  Literal lit;
  for (lit = c->literals; lit; lit = lit->next) {
    if (!lit->sign &&
	SYMNUM(lit->atom) == symbol &&
	!args_distinct_vars(lit->atom))
      return FALSE;
  }
  return TRUE;
}  /* arg_check */

/*************
 *
 *   eliminable_relation()
 *
 *************/

static
BOOL eliminable_relation(int symbol, Plist clauses, BOOL equality)
{
  /* Eliminable if no clause has more than one occurrence.
     Additional constraint for equality: all negative occurrences
     have unique vars as args.
  */
  Plist p;
  for (p = clauses; p; p = p->next) {
    int n = rsym_occurrences(symbol, p->v);
    if (n > 1)
      return FALSE;
    else if (equality && !arg_check(symbol, p->v))
      return FALSE;
  }
  return TRUE;
}  /* eliminable_relation */

/*************
 *
 *   eliminable_relations()
 *
 *************/

static
Ilist eliminable_relations(Plist clauses, BOOL equality)
{
  Ilist rsyms = rsym_set_in_clauses(clauses);
  Ilist eliminable = NULL;
  Ilist p;
  for (p = rsyms; p; p = p->next) {
    if (!is_eq_symbol(p->i) &&          /* don't allow equalities */
	sn_to_arity(p->i) != 0 &&      /* don't allow prop atoms */
	eliminable_relation(p->i, clauses, equality))

      eliminable = ilist_append(eliminable, p->i); 
  }
  zap_ilist(rsyms);
  return eliminable;
}  /* eliminable_relations */

/*************
 *
 *   resolve_on_symbol()
 *
 *************/

static
Plist resolve_on_symbol(int symbol, Clause c1, Clause c2)
{
  Plist resolvents = NULL;
  Literal l1, l2;
  for (l1 = c1->literals; l1; l1 = l1->next)
    for (l2 = c2->literals; l2; l2 = l2->next) {
      if (l1->sign != l2->sign &&
	  SYMNUM(l1->atom) == symbol &&
	  SYMNUM(l2->atom) == symbol) {
	Clause res = resolve3(c1, l1, c2, l2);
	if (res) {
	  if (tautology(res))
	    delete_clause(res);
	  else {
	    assign_clause_id(res);
	    resolvents = plist_append(resolvents, res);
	  }
	}
      }
    }
  return resolvents;
}  /* resolve_on_symbol */

/*************
 *
 *   gen_given()
 *
 *************/

static
Plist gen_given(int symbol, Clause given, Plist usable)
{
  Plist new = NULL;
  Plist p;
  for (p = usable; p; p = p->next) {
    Clause c = p->v;
    Plist resolvents = resolve_on_symbol(symbol, given, c);
    new = plist_cat(new, resolvents);
  }
  return new;
}  /* gen_given */

/*************
 *
 *   subsumed_by_member()
 *
 *************/

static
BOOL subsumed_by_member(Clause c, Plist p)
{
  if (p == NULL)
    return FALSE;
  else if (subsumes(p->v, c))
    return TRUE;
  else
    return subsumed_by_member(c, p->next);
}  /* subsumed_by_member */

/*************
 *
 *   incorporate_new_clauses()
 *
 *************/

static
Plist incorporate_new_clauses(Plist sos, Plist new)
{
  Plist p;
  for (p = new; p; p = p->next) {
    if (!subsumed_by_member(p->v, sos)) {
      sos = plist_append(sos, p->v);
#ifdef GEN_DEBUG
      printf("kept: "); f_clause(p->v);
#endif
    }
  }
  zap_plist(new);
  return sos;
}  /* incorporate_new_clauses */

/*************
 *
 *   elim_relation()
 *
 *************/

static
Plist elim_relation(int symbol, Plist sos, Clist disabled)
{
  /* this does a naive given-clause loop */
  Plist usable = NULL;
  while (sos) {
    Plist new;
    Clause given = sos->v;
    sos = plist_pop(sos);
    usable = plist_append(usable, given);
#ifdef GEN_DEBUG
    printf("\ngiven: "); f_clause(given);
#endif
    new = gen_given(symbol, given, usable);
    sos = incorporate_new_clauses(sos, new);
  }
  /* partition usable into clauses with and without symbol */
  {
    Plist without = NULL;
    Plist p;
    for (p = usable; p; p = p->next) {
      if (rsym_occurrences(symbol, p->v) > 0)
	clist_append(p->v, disabled);
      else
	without = plist_append(without, p->v);
    }
    zap_plist(usable);
    return without;
  }
}  /* elim_relation */

/*************
 *
 *   predicate_elimination()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void predicate_elimination(Clist clauses, Clist disabled, BOOL print)
{
  int n = 0;
  Plist clauses2 = prepend_clist_to_plist(NULL, clauses);
  BOOL equality = equality_in_clauses(clauses2);  /* eq => different method */
  Ilist syms = eliminable_relations(clauses2, equality);
  clist_remove_all_clauses(clauses);
  if (print)
    printf("\n%% Predicate elimination:");
  while (syms) {
    n++;
    /* use first symbol, discard rest, get new list */
    if (print)
      printf(" %s/%d", sn_to_str(syms->i), sn_to_arity(syms->i));
    clauses2 = elim_relation(syms->i, clauses2, disabled);
    zap_ilist(syms);
    syms = eliminable_relations(clauses2, equality);
  }
  if (print)
    printf("%s\n", n > 0 ? "." : " (none).");

  clist_append_plist(clauses, clauses2);
  zap_plist(clauses2);
}  /* predicate_elimination */
