#define PROVER_NAME     "Prover9"
#define PROVER_VERSION  "July-2005B"
#include "../PROGRAM_DATE.h"

#include "../ladr/nonport.h"
#include "../ladr/top_input.h"
#include "../ladr/clausify.h"
#include "../ladr/clause3.h"

#include "utilities.h"
#include "search.h"
#include "prover_exits.h"

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  printf("----- %s %s, %s -----\n", PROVER_NAME, PROVER_VERSION, PROGRAM_DATE);
  printf("Process %d was started by %s on %s,\n%s",
	 my_process_id(), username(), hostname(), get_date());
	 
  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".\n");
}  // print_banner

/*************
 *
 *   number_of_clauses()
 *
 *************/

static
int number_of_clauses(Prover_input p)
{
  return
    plist_count(p->usable) + plist_count(p->sos) +
    plist_count(p->goals) + plist_count(p->demods);
}  /* number_of_clauses */

/*************
 *
 *   conjoin_all_formulas() -- formulas are not copied!
 *
 *************/

static
Formula conjoin_all_formulas(Prover_input p)
{
  Formula f;
  Plist all = NULL;
  all = plist_cat2(all, p->fof_usable);
  all = plist_cat2(all, p->fof_sos);
  all = plist_cat2(all, p->fof_goals);
  f = formulas_to_conjunction(all);
  zap_plist(all);
  return f;
}  /* conjoin_all_formulas */

/*************
 *
 *   reduce_problem()
 *
 *************/

static
Formula reduce_problem(Formula f)
{
  /* Formula f is a negated conjecture.  We have to un-negate it,
     reduce it, then re-negate it.  If the reduction succeeds, the
     returned formula is always a flat disjuncion, even if it has
     0 (meaning FALSE) or 1 members. */
  f = miniscope_formula(negate(f), 500);  /* bogo_tick limit, about 50/sec */
  if (f != NULL)
    f = make_disjunction(nnf(negate(f)));
  return f;
}  /* reduce_problem */

/*************
 *
 *   foffer()
 *
 *   Attempt to reduce the problem to independent subproblems and
 *   refute all of the subproblems.  If we fail to reduce the problem,
 *   or we decide not to run the subproblems, return -1; otherwise,
 *   return the ordinary "search" return code from the last subproblem
 *   that was run.
 *
 *************/

static
int foffer(Prover_input input)
{
  double children_user_seconds = 0.0;    // running total for all children
  double children_system_seconds = 0.0;  // running total for all children
  int i, nnf_size, cnf_max;
  int n_max, c_max;
  int max_seconds = parm(input->options->max_seconds);
  int return_code;

  Formula f0 = conjoin_all_formulas(input);  // f0 is the original
  Formula f  = formula_copy(f0);

  // Now f is the negation of the conjecture.

  f = nnf2(f, DISJUNCTION);      // we do this only to get the sizes
  nnf_size = formula_size(f);    // number of Formula nodes
  cnf_max = cnf_max_clauses(f);  // maximum number of clauses in CNF

  printf("\n%% Attempting problem reduction;"
	 " original problem has <nnf_size,cnf_max> = <%d,%d>.\n",
	 nnf_size, cnf_max);
  fflush(stdout);

  // Problem reduction returns an equivalent formula (or NULL if it fails).

  f = reduce_problem(f);

  if (f == NULL) {
    printf("\n%% Problem reduction (%.2f seconds) failed; "
	   "reverting to ordinary search.\n", user_seconds());
    fflush(stdout);
    return -1;
  }
  
  // Formula f is disjunction, even if it has 0 or 1 members; f is still
  // the denial, so members of disjunction are (denied) subproblems.

  if (f->arity == 1) {
    printf("%% Problem reduction (%.2f sec) gives one subproblem (<%d,%d>);\n"
	   "%%   reverting to ordinary search with original formulas.\n",
	   user_seconds(), formula_size(f), cnf_max_clauses(f));
    fflush(stdout);
    return -1;
  }

  // Print sizes of subproblems and get maximum sizes.

  n_max = c_max = 0;
  printf("%% Problem reduction (%.2f sec) gives %d independent subproblems: (",
	 user_seconds(), f->arity);
  for (i = 0; i < f->arity; i++) {
    int n = formula_size(f->kids[i]);
    int c = cnf_max_clauses(f->kids[i]);
    if (!flag(input->options->quiet))
      printf(" <%d,%d>", n, c);
    n_max = IMAX(n_max,n);
    c_max = IMAX(c_max,c);
  }
  printf(" ).\n");
  fflush(stdout);

  // If both maximums are greater than originals, revert to ordinary search.

  printf("%% Max nnf_size of subproblems is %d; max cnf_max is %d.\n",
	 n_max, c_max);
  if (n_max > nnf_size && c_max > cnf_max) {
    printf("%% Reverting to ordinary search with original formulas.\n");
    fflush(stdout);
    return -1;
  }

  // Attack the subproblems.

  skolem_check(FALSE);  /* don't check that Skolem symbols are new */
  
  for (i = 0, return_code = MAX_PROOFS_EXIT;
       i < f->arity && return_code == MAX_PROOFS_EXIT;
       i++) {

    Prover_results results;
    
    if (flag(input->options->print_subproblems)) {
      printf("\n%% Subproblem %d of %d (negated):  ", i+1, f->arity);
      p_formula(f->kids[i]);
    }
    else
      printf("\n%% Starting Subproblem %d of %d.\n", i+1, f->arity);
    fflush(stdout);

    input->sos = clausify_formula(f->kids[i]);

    if (max_seconds != INT_MAX) {
      assign_parm(input->options->max_seconds,
		  max_seconds - 
		    (int) (children_user_seconds + user_seconds()),
		  FALSE);
      printf("\n%% Max_seconds is %d for this subproblem.\n",
	     parm(input->options->max_seconds));
      fflush(stdout);
    }

    results = forking_search(input);
    return_code = results->return_code;
    children_user_seconds   += results->user_seconds;
    children_system_seconds += results->system_seconds;

    zap_prover_results(results);
    zap_clauses_ija(input->sos);
    input->sos = NULL;
    decommission_skolem_symbols();
    skolem_reset();
  }
  
  if (return_code == MAX_PROOFS_EXIT)
    printf("\n%% All %d subproblems have been proved, so we are done.\n", i);
  else
    printf("\n%% Search failed on subproblem %d.\n", i);

  printf("\nTotal user_CPU=%.2f, system_CPU=%.2f, wall_clock=%d.\n",
	 children_user_seconds + user_seconds(),
	 children_system_seconds + system_seconds(),
	 wallclock());

  fflush(stdout);
  return return_code;
}  /* foffer */

/*************
 *
 *    main -- basic prover + FOF problem reduction
 *
 *************/

int main(int argc, char **argv)
{
  Prover_input input;
  int return_code;
  BOOL ordinary_search;

  print_banner(argc, argv);

  /***************** Initialize and read the input ***************************/

  input = std_prover_init_and_input(argc, argv,
			    TRUE,           // echo input to stdout
			    KILL_UNKNOWN);  // unknown flags/parms are fatal

  /***************** Search for a proof **************************************/

  if (flag(input->options->fof_reduction) && number_of_clauses(input) == 0) {
    // try problem reduction
    return_code = foffer(input);
    ordinary_search = (return_code == -1);
  }
  else {
    ordinary_search = TRUE;
    return_code = -1;  // to please the compiler
  }

  if (ordinary_search) {
    Prover_results results;
    std_clausify(input, TRUE);
    results = search(input);             // ordinary search for a proof
    return_code = results->return_code;
  }

  /***************** Print result message and exit ***************************/

  if (return_code == MAX_PROOFS_EXIT) {
    printf("\nTHEOREM PROVED\n");
    if (!flag(input->options->quiet))
      fprintf(stderr, "\nTHEOREM PROVED\n");
  }
  else {
    // Note that we fail if we found less than max_proofs proofs.
    printf("\nSEARCH FAILED\n");
    if (!flag(input->options->quiet))
      fprintf(stderr, "\nSEARCH FAILED\n");
  }

  exit_with_message(stdout, return_code);
  exit(1);  // to satisfy the compiler (won't be called)
}  // main
