#include "search.h"

// system includes

#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <setjmp.h>  /* Yikes! */

// LADR includes

#include "../ladr/sos.h"
#include "../ladr/paramod.h"
#include "../ladr/nonport.h"
#include "../ladr/hints.h"
#include "../ladr/xproofs.h"
#include "../ladr/ac_redun.h"
#include "../ladr/weight.h"
#include "../ladr/int_code.h"

// Local includes

#include "prover_exits.h"
#include "utilities.h"
#include "forward_subsume.h"
#include "index_lits.h"
#include "demodulate.h"
#include "actions.h"
#include "control_sos.h"
#include "pred_elim.h"
#include "unfold.h"

// Private definitions and types

static jmp_buf Jump_env;   // for setjmp/longjmp

static Prover_options Opt;
static struct prover_attributes Att;
static struct prover_stats Stats;
static struct prover_clocks Clocks;

static struct {

  Clist sos;
  Clist usable;
  Clist goals;
  Clist demods;
  Clist hints;

  Plist actions;
  Plist weights;
  Plist kb_weights;

  Clist limbo;
  Clist disabled;
  Plist empties;

  Lindex clashable_idx;

  BOOL use_clash_idx;
  BOOL no_fapl;

  int number_of_input_clauses;

  BOOL initialized;
  double start_time;
  int start_ticks;

  int return_code;

} Glob;

enum {
  STATS_NONE,
  STATS_SOME,
  STATS_LOTS,
  STATS_ALL
};

/*************
 *
 *    init_prover_options()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Prover_options init_prover_options(void)
{
  Prover_options p = calloc(1, sizeof(struct prover_options));
  // FLAGS:
  //   internal name                    external name            default

  p->binary_resolution      = init_flag("binary_resolution",      FALSE);
  p->neg_binary_resolution  = init_flag("neg_binary_resolution",  FALSE);
  p->hyper_resolution       = init_flag("hyper_resolution",       FALSE);
  p->neg_hyper_resolution   = init_flag("neg_hyper_resolution",   FALSE);
  p->ur_resolution          = init_flag("ur_resolution",          FALSE);
  p->pos_ur_resolution      = init_flag("pos_ur_resolution",      FALSE);
  p->neg_ur_resolution      = init_flag("neg_ur_resolution",      FALSE);
  p->factor                 = init_flag("factor",                 FALSE);
  p->paramodulation         = init_flag("paramodulation",         FALSE);

  p->ordered_inference      = init_flag("ordered_inference",      TRUE);
  p->ordered_instance       = init_flag("ordered_instance",       FALSE);
  p->para_units_only        = init_flag("para_units_only",        FALSE);
  p->basic_paramodulation   = init_flag("basic_paramodulation",   FALSE);
  p->initial_nuclei         = init_flag("initial_nuclei",         FALSE);

  p->process_initial_sos    = init_flag("process_initial_sos",    TRUE);
  p->back_demod             = init_flag("back_demod",             FALSE);
  p->lex_dep_demod          = init_flag("lex_dep_demod",          TRUE);
  p->lex_dep_demod_sane     = init_flag("lex_dep_demod_sane",     FALSE);
  p->safe_unit_conflict     = init_flag("safe_unit_conflict",     FALSE);
  p->back_subsume           = init_flag("back_subsume",           TRUE);
  p->degrade_hints          = init_flag("degrade_hints",          TRUE);
  p->reuse_goals            = init_flag("reuse_goals",            FALSE);
  p->cac_redundancy         = init_flag("cac_redundancy",         TRUE);
  p->unit_deletion          = init_flag("unit_deletion",          FALSE);
  p->back_unit_deletion     = init_flag("back_unit_deletion",     FALSE);
  p->dont_flip_input        = init_flag("dont_flip_input",        FALSE);

  p->echo_input             = init_flag("echo_input",             TRUE);
  p->quiet                  = init_flag("quiet",                  FALSE);
  p->print_subproblems      = init_flag("print_subproblems",      TRUE);
  p->print_initial_clauses  = init_flag("print_initial_clauses",  TRUE);
  p->print_given            = init_flag("print_given",            TRUE);
  p->print_gen              = init_flag("print_gen",              FALSE);
  p->print_kept             = init_flag("print_kept",             FALSE);
  p->print_labeled          = init_flag("print_labeled",          FALSE);
  p->print_proofs           = init_flag("print_proofs",           TRUE);
  p->standard_proofs        = init_flag("standard_proofs",        TRUE);
  p->x_proofs               = init_flag("x_proofs",               FALSE);
  p->r_proofs               = init_flag("r_proofs",               FALSE);
  p->rx_proofs              = init_flag("rx_proofs",              FALSE);
  p->xr_proofs              = init_flag("xr_proofs",              FALSE);
  p->default_output         = init_flag("default_output",         TRUE);

  p->fof_reduction          = init_flag("fof_reduction",          FALSE);
  p->predicate_elimination  = init_flag("predicate_elimination",  FALSE);
  p->inverse_order          = init_flag("inverse_order",          FALSE);
  p->unfold_eq              = init_flag("unfold_eq",              FALSE);
  p->fold_eq                = init_flag("fold_eq",                FALSE);
  p->sort_initial_sos       = init_flag("sort_initial_sos",       FALSE);

  p->breadth_first          = init_flag("breadth_first",          FALSE);
  p->input_sos_first        = init_flag("input_sos_first",        TRUE);

  p->automatic              = init_flag("auto",                   FALSE);
  p->auto1                  = init_flag("auto1",                  FALSE);
  p->auto_inference         = init_flag("auto_inference",         FALSE);
  p->hands_off_options      = init_flag("hands_off_options",      FALSE);
  p->casc_2005              = init_flag("casc_2005",              FALSE);

  // PARMS:
  //  internal name               external name      default    min      max

  p->max_given =        init_parm("max_given",       INT_MAX,      0,INT_MAX);
  p->max_kept =         init_parm("max_kept",        INT_MAX,      0,INT_MAX);
  p->max_proofs =       init_parm("max_proofs",            1,     -1,INT_MAX);
  p->max_megs =         init_parm("max_megs",            200,      0,INT_MAX);
  p->max_seconds =      init_parm("max_seconds",     INT_MAX,      0,INT_MAX);

  p->new_constants =    init_parm("new_constants",         0,      0,INT_MAX);
  p->para_lit_limit =   init_parm("para_lit_limit",  INT_MAX,      1,INT_MAX);
  p->nucleus_limit =    init_parm("nucleus_limit",   INT_MAX,      0,INT_MAX);

  p->unfold_eq_limit =  init_parm("unfold_eq_limit",      -1,     -1,INT_MAX);
  p->fold_denial_max =  init_parm("fold_denial_max",       0,      0,INT_MAX);

  p->pick_given_ratio = init_parm("pick_given_ratio",      0,      0,INT_MAX);
  p->age_part         = init_parm("age_part",              1,      0,INT_MAX);
  p->weight_part      = init_parm("weight_part",           2,      0,INT_MAX);
  p->weight_neg_part  = init_parm("weight_neg_part",       2,      0,INT_MAX);


  p->max_weight =       init_parm("max_weight",      INT_MAX,INT_MIN,INT_MAX);
  p->lex_dep_demod_lim =init_parm("lex_dep_demod_lim",INT_MAX,     0,INT_MAX);
  p->max_literals =     init_parm("max_literals",    INT_MAX,      1,INT_MAX);
  p->demod_limit =      init_parm("demod_limit",        1000,     -1,INT_MAX);

  p->variable_weight =  init_parm("variable_weight",       1,INT_MIN,INT_MAX);
  p->constant_weight =  init_parm("constant_weight",       1,INT_MIN,INT_MAX);
  p->not_weight =       init_parm("not_weight",            0,INT_MIN,INT_MAX);
  p->or_weight =        init_parm("or_weight",             0,INT_MIN,INT_MAX);
  p->sk_constant_weight=init_parm("sk_constant_weight",    1,INT_MIN,INT_MAX);
  p->prop_atom_weight = init_parm("prop_atom_weight",      1,INT_MIN,INT_MAX);
  p->skolem_penalty =   init_parm("skolem_penalty",        1,      0,INT_MAX);
  p->nest_penalty =     init_parm("nest_penalty",          0,      0,INT_MAX);
  p->bsub_hint_add_wt = init_parm("bsub_hint_add_wt",  -1000,INT_MIN,INT_MAX);
  p->bsub_hint_wt =     init_parm("bsub_hint_wt",    INT_MAX,INT_MIN,INT_MAX);
  p->default_weight =   init_parm("default_weight",  INT_MAX,INT_MIN,INT_MAX);

  p->sos_limit =        init_parm("sos_limit",       INT_MAX,      0,INT_MAX);
  p->min_sos_limit =    init_parm("min_sos_limit",         0,      0,INT_MAX);
  p->lrs_interval =     init_parm("lrs_interval",         50,      1,INT_MAX);
  p->lrs_seconds =      init_parm("lrs_seconds",     INT_MAX,      1,INT_MAX);
  p->lrs_ticks =        init_parm("lrs_ticks",       INT_MAX,      1,INT_MAX);

  p->report =           init_parm("report",               -1,     -1,INT_MAX);

  // STRINGPARMS:
  // (internal-name, external-name, number-of-strings, str1, str2, ... )
  // str1 is always the default

  p->order = init_stringparm("order", 3,
			     "lpo",
			     "rpo",
			     "kbo");

  p->literal_selection = init_stringparm("literal_selection", 4,
					 "maximal",
					 "all",
					 "first_maximal",
					 "first");

  p->stats = init_stringparm("stats", 4,
			     "lots",
			     "all",
			     "some",
			     "none");

  // Flag and parm Dependencies.  These cause other flags and parms
  // to be changed.  The changes happen immediately and can be undone
  // by later settings in the input.
  // DEPENDENCIES ARE NOT APPLIED TO DEFAULT SETTINGS!

  flag_flag_dependency(p->paramodulation,     TRUE, p->back_demod,       TRUE);
  flag_parm_dependency(p->para_units_only,    TRUE, p->para_lit_limit,      1);
  flag_flag_dependency(p->back_unit_deletion, TRUE, p->unit_deletion,    TRUE);

  flag_flag_dependency(p->ur_resolution, TRUE,  p->pos_ur_resolution, TRUE);
  flag_flag_dependency(p->ur_resolution, TRUE,  p->neg_ur_resolution, TRUE);

  flag_parm_dependency(p->lex_dep_demod, FALSE, p->lex_dep_demod_lim, 0);
  flag_parm_dependency(p->lex_dep_demod,  TRUE, p->lex_dep_demod_lim, INT_MAX);

  flag_parm_dependency(p->breadth_first,    TRUE,  p->age_part,        1);
  flag_parm_dependency(p->breadth_first,    TRUE,  p->weight_part,     0);
  flag_parm_dependency(p->breadth_first,    TRUE,  p->weight_neg_part, 0);

  parm_flag_dependency(p->new_constants, p->x_proofs, FALSE);   // no support
  parm_flag_dependency(p->new_constants, p->rx_proofs, FALSE);  // no support

  flag_parm_dependency(p->unfold_eq,    TRUE,  p->unfold_eq_limit, INT_MAX);
  flag_parm_dependency(p->unfold_eq,   FALSE,  p->unfold_eq_limit,      -1);
  flag_flag_dependency(p->unfold_eq,    TRUE,  p->fold_eq,           FALSE);
  flag_flag_dependency(p->fold_eq,      TRUE,  p->unfold_eq,         FALSE);

  parm_flag_dependency(p->fold_denial_max, p->x_proofs, FALSE);   // no support
  parm_flag_dependency(p->fold_denial_max, p->rx_proofs, FALSE);  // no support

  parm_parm_dependency(p->pick_given_ratio, p->age_part,          1);
  parm_parm_dependency(p->pick_given_ratio, p->weight_part, INT_MIN); // copy
  parm_parm_dependency(p->pick_given_ratio, p->weight_neg_part,   0);

  flag_flag_dependency(p->default_output, TRUE, p->quiet,               FALSE);
  flag_flag_dependency(p->default_output, TRUE, p->echo_input,           TRUE);
  flag_flag_dependency(p->default_output, TRUE, p->print_initial_clauses,TRUE);
  flag_flag_dependency(p->default_output, TRUE, p->print_given,          TRUE);
  flag_flag_dependency(p->default_output, TRUE, p->print_subproblems,    TRUE);
  flag_flag_dependency(p->default_output, TRUE, p->print_proofs,         TRUE);
  flag_flag_dependency(p->default_output, TRUE, p->standard_proofs,      TRUE);
  flag_stringparm_dependency(p->default_output, TRUE, p->stats,        "lots");

  flag_flag_dependency(p->default_output, TRUE, p->print_kept,          FALSE);
  flag_flag_dependency(p->default_output, TRUE, p->print_gen,           FALSE);

  // automatic (auto1)

  flag_flag_dependency(p->automatic, TRUE, p->auto1,             TRUE);

  flag_flag_dependency(p->auto1, TRUE, p->auto_inference,        TRUE);
  flag_flag_dependency(p->auto1, TRUE, p->fof_reduction,         TRUE);
  flag_flag_dependency(p->auto1, TRUE, p->predicate_elimination, TRUE);
  flag_flag_dependency(p->auto1, TRUE, p->unfold_eq,             TRUE);
  flag_flag_dependency(p->auto1, TRUE, p->inverse_order,         TRUE);

  flag_parm_dependency(p->auto1, TRUE, p->new_constants,            1);
  flag_parm_dependency(p->auto1, TRUE, p->lex_dep_demod_lim,       11);
  flag_flag_dependency(p->auto1, TRUE, p->lex_dep_demod_sane,    TRUE);

  flag_parm_dependency(p->auto1, TRUE, p->max_weight,             100);

  flag_parm_dependency(p->auto1, TRUE, p->age_part,                 1);
  flag_parm_dependency(p->auto1, TRUE, p->weight_part,              2);
  flag_parm_dependency(p->auto1, TRUE, p->weight_neg_part,          2);

  flag_parm_dependency(p->auto1, TRUE, p->sos_limit,            10000);

  flag_flag_dependency(p->auto1, TRUE, p->x_proofs,              FALSE);
  flag_stringparm_dependency(p->auto1, TRUE, p->stats,          "lots");

  flag_parm_dependency(p->auto1, TRUE, p->max_megs,               200);
  flag_flag_dependency(p->auto1, TRUE, clocks_id(),             FALSE);

  // casc_2005  (also triggered by --casc_2005 on the command line)

  flag_flag_dependency(p->casc_2005, TRUE, p->auto1,                 TRUE);

  flag_parm_dependency(p->casc_2005, TRUE, p->fold_denial_max,          3);

  flag_parm_dependency(p->casc_2005, TRUE, p->max_weight,             200);
  flag_parm_dependency(p->casc_2005, TRUE, p->nest_penalty,             1);
  flag_parm_dependency(p->casc_2005, TRUE, p->skolem_penalty,           3);
  flag_parm_dependency(p->casc_2005, TRUE, p->sk_constant_weight,       0);
  flag_parm_dependency(p->casc_2005, TRUE, p->prop_atom_weight,         5);

  flag_flag_dependency(p->casc_2005, TRUE, p->sort_initial_sos,       TRUE);
  flag_parm_dependency(p->casc_2005, TRUE, p->sos_limit,           INT_MAX);
  flag_parm_dependency(p->casc_2005, TRUE, p->lrs_ticks,              3000);

  flag_parm_dependency(p->casc_2005, TRUE, p->max_megs,                400);

  flag_stringparm_dependency(p->casc_2005, TRUE, p->stats,          "some");
  flag_flag_dependency(p->casc_2005, TRUE, p->echo_input,            FALSE);
  flag_flag_dependency(p->casc_2005, TRUE, p->quiet,                  TRUE);
  flag_flag_dependency(p->casc_2005, TRUE, p->print_subproblems,     FALSE);
  flag_flag_dependency(p->casc_2005, TRUE, p->print_initial_clauses, FALSE);
  flag_flag_dependency(p->casc_2005, TRUE, p->print_given,           FALSE);

  return p;
  
}  // init_prover_options

/*************
 *
 *   init_prover_attributes()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_prover_attributes(void)
{
  Att.label            = register_attribute("label",        STRING_ATTRIBUTE);
  Att.bsub_hint_add_wt = register_attribute("bsub_hint_add_wt",INT_ATTRIBUTE);
  Att.bsub_hint_wt     = register_attribute("bsub_hint_wt",    INT_ATTRIBUTE);
  Att.answer           = register_attribute("answer",         TERM_ATTRIBUTE);

  Att.action           = register_attribute("action",         TERM_ATTRIBUTE);
  Att.action2          = register_attribute("action2",        TERM_ATTRIBUTE);

  declare_term_attribute_inheritable(Att.answer);
  declare_term_attribute_inheritable(Att.action2);
}  // Init_prover_attributes

/*************
 *
 *   label_att()
 *
 *************/

/* DOCUMENTATION
Return the attribute identifier for label attributes.
*/

/* PUBLIC */
int label_att(void)
{
  return Att.label;
}  /* label_att */

/*************
 *
 *   update_stats()
 *
 *************/

static
void update_stats(void)
{
  Stats.demod_attempts = demod_attempts() + fdemod_attempts();
  Stats.demod_rewrites = demod_rewrites() + fdemod_rewrites();
  Stats.res_instance_prunes = res_instance_prunes();
  Stats.para_instance_prunes = para_instance_prunes();
  Stats.basic_para_prunes = basic_paramodulation_prunes();
  Stats.sos_removed = control_sos_removed();
  Stats.nonunit_fsub = nonunit_fsub_tests();
  Stats.nonunit_bsub = nonunit_bsub_tests();
  Stats.usable_size = Glob.usable ? Glob.usable->length : 0;
  Stats.sos_size = Glob.sos ? Glob.sos->length : 0;
  Stats.demodulators_size = Glob.demods ? Glob.demods->length : 0;
  Stats.goals_size = Glob.goals ? Glob.goals->length : 0;
  Stats.limbo_size = Glob.limbo ? Glob.limbo->length : 0;
  Stats.disabled_size = Glob.disabled ? Glob.disabled->length : 0;
  Stats.hints_size = Glob.hints ? Glob.hints->length : 0;
  Stats.kbyte_usage = bytes_palloced() / 1000;
}  /* update_stats */

/*************
 *
 *   fprint_prover_stats()
 *
 *************/

static
void fprint_prover_stats(FILE *fp, struct prover_stats s, int stats_level)
{
  fprintf(fp,"\nGiven=%u. Generated=%u. Kept=%u. Proofs=%u.\n",
	  s.given, s.generated, s.kept, s.proofs);
  fprintf(fp,"Usable=%u. Sos=%u. Demods=%u. Goals=%u. Limbo=%d, "
	  "Disabled=%u. Hints=%u.\n",
	  s.usable_size, s.sos_size, s.demodulators_size,
	  s.goals_size, s.limbo_size, s.disabled_size, s.hints_size);
  
  if (stats_level >= STATS_LOTS) {

    fprintf(fp,"Weight_deleted=%u. Literals_deleted=%u.\n",
	    s.weight_deleted, s.literals_deleted);
    fprintf(fp,"Forward_subsumed=%u. Back_subsumed=%u.\n",
	    s.subsumed, s.back_subsumed);
    fprintf(fp,"Sos_limit_deleted=%u. Sos_displaced=%u. Sos_removed=%u.\n",
	    s.sos_limit_deleted, s.sos_displaced, s.sos_removed);
    fprintf(fp,"New_demodulators=%u (%u lex), Back_demodulated=%u. Back_unit_deleted=%u.\n",
	    s.new_demodulators, s.new_lex_demods, s.back_demodulated, s.back_unit_deleted);
    fprintf(fp,"Demod_attempts=%u. Demod_rewrites=%u.\n",
	    s.demod_attempts, s.demod_rewrites);
    fprintf(fp,"Res_instance_prunes=%u. Para_instance_prunes=%u. Basic_paramod_prunes=%u.\n",
	    s.res_instance_prunes, s.para_instance_prunes, s.basic_para_prunes);
    fprintf(fp,"Nonunit_fsub_feature_tests=%u. ", s.nonunit_fsub);
    fprintf(fp,"Nonunit_bsub_feature_tests=%u.\n", s.nonunit_bsub);
  }

  fprintf(fp,"Megabytes=%.2f.\n", s.kbyte_usage / 1000.0);

#if 1
  fprintf(fp,"User_CPU=%.2f, System_CPU=%.2f, Wall_clock=%d.\n",
	  user_seconds(), system_seconds(), wallclock());
#else
  {
    double user_sec = user_seconds();
    fprintf(fp,"User_CPU=%.2f, System_CPU=%.2f, Wall_clock=%d, "
	    "Mega_mem_calls/sec=%.2f, "
	    "Mega_next_calls/sec=%.2f, "
	    "Mega_sub_calls/sec=%.2f.\n",
	    user_sec, system_seconds(), wallclock(),
	    user_sec == 0 ? 0.0 : mega_mem_calls() / user_sec,
	    user_sec == 0 ? 0.0 : mega_next_calls() / user_sec,
	    user_sec == 0 ? 0.0 : mega_sub_calls() / user_sec);
  }
#endif
}  /* fprint_prover_stats */

/*************
 *
 *   fprint_prover_clocks()
 *
 *************/

/* DOCUMENTATION
Given an arroy of clock values (type double) indexed by
the ordinary clock indexes, print a report to a file.
*/

/* PUBLIC */
void fprint_prover_clocks(FILE *fp, struct prover_clocks clks)
{
  if (clocks_enabled()) {
    fprintf(fp, "\n");
    fprint_clock(fp, clks.pick_given);
    fprint_clock(fp, clks.infer);
    fprint_clock(fp, clks.preprocess);
    fprint_clock(fp, clks.demod);
    fprint_clock(fp, clks.unit_del);
    fprint_clock(fp, clks.redundancy);
    fprint_clock(fp, clks.conflict);
    fprint_clock(fp, clks.weigh);
    fprint_clock(fp, clks.hints);
    fprint_clock(fp, clks.subsume);
    fprint_clock(fp, clks.back_subsume);
    fprint_clock(fp, clks.back_demod);
    fprint_clock(fp, clks.back_unit_del);
    fprint_clock(fp, clks.index);
    fprint_clock(fp, clks.disable);
  }
}  /* fprint_prover_clocks */

/*************
 *
 *   fprint_all_stats()
 *
 *************/

static
void fprint_all_stats(FILE *fp)
{
  int stats_level;
  if (!Opt || stringparm(Opt->stats, "all"))
    stats_level = STATS_ALL;
  else if (stringparm(Opt->stats, "lots"))
    stats_level = STATS_LOTS;
  else if (stringparm(Opt->stats, "some"))
    stats_level = STATS_SOME;
  else if (stringparm(Opt->stats, "none"))
    stats_level = STATS_NONE;
  else
    stats_level = STATS_ALL;
  
  update_stats();
  fprint_prover_stats(fp, Stats, stats_level);

  fprint_prover_clocks(fp, Clocks);

  if (stats_level == STATS_ALL) {
    print_memory_stats(fp);
    p_sos_dist();
  }
  fflush(fp);
}  // fprint_all_stats

/*************
 *
 *   exit_string()
 *
 *************/

static
char *exit_string(int code)
{
  char *message;
  switch (code) {
  case MAX_PROOFS_EXIT:  message = "max_proofs";  break;
  case FATAL_EXIT:       message = "fatal_error"; break;
  case SOS_EMPTY_EXIT:   message = "sos_empty";   break;
  case MAX_MEGS_EXIT:    message = "max_megs";    break;
  case MAX_SECONDS_EXIT: message = "max_seconds"; break;
  case MAX_GIVEN_EXIT:   message = "max_given";   break;
  case MAX_KEPT_EXIT:    message = "max_kept";    break;
  case ACTION_EXIT:      message = "action";      break;
  case SIGSEGV_EXIT:     message = "SIGSEGV";     break;
  case SIGINT_EXIT:      message = "SIGINT";      break;
  default: message = "???";
  }
  return message;
}  /* exit_string */

/*************
 *
 *   exit_with_message()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void exit_with_message(FILE *fp, int code)
{
  int proofs = Glob.initialized ? Stats.proofs : -1;
  
  if (proofs == -1)
    fprintf(fp, "\nExiting.\n");
  else if (proofs == 0)
    fprintf(fp, "\nExiting with failure.\n");
  else
    fprintf(fp, "\nExiting with %d proof%s.\n",
	    proofs, proofs == 1 ? "" : "s");
  
  if (!(Opt && flag(Opt->quiet))) {
    fprintf(stderr, "\n------ process %d exit (%s) ------\n",
	    my_process_id(), exit_string(code));
    bell(stderr);
  }

  fprintf(fp, "\nProcess %d exit (%s) %s",
	  my_process_id(), exit_string(code), get_date());
  
  fflush(fp);
  exit(code);
}  /* exit_with_message */

/*************
 *
 *   report()
 *
 *************/

static
void report(void)
{
  double seconds = user_seconds();
  printf("\n%% Report at %.2f seconds, %s", seconds, get_date());
  fprint_all_stats(stdout);
  printf("\n%% end of report\n");
  if (!flag(Opt->quiet))
    fprintf(stderr, "A report (%.2f seconds) has been sent to the output.\n",
	    seconds);
}  /* report */

/*************
 *
 *   possible_report()
 *
 *************/

static
void possible_report(void)
{
  static int Next_report;
  int runtime;

  if (Next_report == 0)
    Next_report = parm(Opt->report);

  runtime = user_time() / 1000;

  if (runtime >= Next_report) {
    report();
    while (runtime >= Next_report)
      Next_report += parm(Opt->report);
  }
}  /* possible_report */

/*************
 *
 *   prover_sig_handler()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void prover_sig_handler(int condition)
{
  printf("\nprover_sig_handler: condition %d", condition);
  switch (condition) {
  case SIGSEGV:
    fprint_all_stats(stdout);
    exit_with_message(stdout, SIGSEGV_EXIT);
    break;
  case SIGINT:
    fprint_all_stats(stdout);
    exit_with_message(stdout, SIGINT_EXIT);
    break;
  case SIGUSR1: report(); break;
  default: fatal_error("prover_sig_handler, unknown signal");
  }
}  /* prover_sig_handler */

/*************
 *
 *   done_with_search()
 *
 *************/

static
void done_with_search(int return_code)
{
  fprint_all_stats(stdout);
  /* If we need to return 0, we have to encode it as something else. */
  longjmp(Jump_env, return_code == 0 ? INT_MAX : return_code);
}  /* done_with_search */

/*************
 *
 *   exit_if_over_limit()
 *
 *************/

static
void exit_if_over_limit(void)
{
  /* max_megs is handled elsewhere */

  if (Stats.kept > parm(Opt->max_kept))
    done_with_search(MAX_KEPT_EXIT);
  else if (Stats.given > parm(Opt->max_given))
    done_with_search(MAX_GIVEN_EXIT);
  else if (user_seconds() > parm(Opt->max_seconds))
    done_with_search(MAX_SECONDS_EXIT);
}  // exit_if_over_limit

/*************
 *
 *   max_megs_exit()
 *
 *   This is intended to be called from the memory package.
 *
 *************/

static
void max_megs_exit(void)
{
  done_with_search(MAX_MEGS_EXIT);
}  /* max_megs_exit */

/*************
 *
 *   inferences_to_make()
 *
 *************/

static
BOOL inferences_to_make(void)
{
  return Glob.sos->first != NULL;
}  // inferences_to_make

/*************
 *
 *   index_clashable()
 *
 *   Insert/delete a clause into/from resolution index.
 *
 *************/

static
void index_clashable(Clause c, Indexop operation)
{
  if (Glob.use_clash_idx) {
    clock_start(Clocks.index);
    lindex_update(Glob.clashable_idx, c, operation);
    clock_stop(Clocks.index);
  }
}  /* index_clashable */

/*************
 *
 *   disable_clause()
 *
 *************/

static
void disable_clause(Clause c)
{
  // Assume c is in Usable, Sos, Goals, or Demodulators
  // Also, c may be in Demodulators.
  //
  // Unindex c according to which lists it is on and
  // the flags that are set, remove c from the lists,
  // and append c to Disabled.  Make sure you don't
  // have a Clist_pos for c during the call, because
  // it will be freed during the call.

  clock_start(Clocks.disable);

  if (clist_member(c, Glob.demods)) {
    index_demodulator(c, demodulator_type(c,
					  parm(Opt->lex_dep_demod_lim),
					  flag(Opt->lex_dep_demod_sane)),
		      DELETE, Clocks.index);
    clist_remove(c, Glob.demods);
  }
  
  if (clist_member(c, Glob.usable)) {
    index_literals(c, DELETE, Clocks.index, Glob.no_fapl);
    index_back_demod(c, DELETE, Clocks.index, flag(Opt->back_demod));
    index_clashable(c, DELETE);
    clist_remove(c, Glob.usable);
  }
  else if (clist_member(c, Glob.goals)) {
    index_goal(c, DELETE, Clocks.index);
    index_back_demod(c, DELETE, Clocks.index, flag(Opt->back_demod));
    clist_remove(c, Glob.goals);
  }
  else if (clist_member(c, Glob.sos)) {
    index_literals(c, DELETE, Clocks.index, Glob.no_fapl);
    index_back_demod(c, DELETE, Clocks.index, flag(Opt->back_demod));
    remove_from_sos(c, Glob.sos, pick_set(c, Opt));
  }
  else if (clist_member(c, Glob.limbo)) {
    index_literals(c, DELETE, Clocks.index, Glob.no_fapl);
    clist_remove(c, Glob.limbo);
  }
  else
    fatal_error("disable_clause: where IS the clause?");

  /* cannot delete "unused" c, because it might really have been used (huh?) */

  compress_clause(c);
  clist_append(c, Glob.disabled);
  clock_stop(Clocks.disable);
}  // disable_clause

/*************
 *
 *   free_search_memory()
 *
 *   This frees memory so that we can check for memory leaks.
 *
 *************/

/* DOCUMENTATION
This is intended for debugging only.
*/

/* PUBLIC */
void free_search_memory(void)
{
  // Demodulators

  while (Glob.demods->first) {
    Clause c = Glob.demods->first->c;
    index_demodulator(c, demodulator_type(c,
					  parm(Opt->lex_dep_demod_lim),
					  flag(Opt->lex_dep_demod_sane)),
		      DELETE, Clocks.index);
    clist_remove(c, Glob.demods);
    if (c->containers == NULL)
      delete_clause(c);
  }
  clist_free(Glob.demods);
  
  destroy_demodulation_index();

  // Lightest

  // zap_sos();

  // Usable, Sos, Limbo

  while (Glob.usable->first)
    disable_clause(Glob.usable->first->c);
  clist_free(Glob.usable);
  Glob.usable = NULL;

  while (Glob.goals->first)
    disable_clause(Glob.goals->first->c);
  clist_free(Glob.goals);
  Glob.goals = NULL;

  while (Glob.sos->first)
    disable_clause(Glob.sos->first->c);
  clist_free(Glob.sos);
  Glob.sos = NULL;

  while (Glob.limbo->first)
    disable_clause(Glob.limbo->first->c);
  clist_free(Glob.limbo);
  Glob.limbo = NULL;

  destroy_literals_index();
  destroy_back_demod_index();
  lindex_destroy(Glob.clashable_idx);
  Glob.clashable_idx = NULL;

  delete_clist(Glob.disabled);
  Glob.disabled = NULL;

  if (Glob.hints->first) {
    Clist_pos p;
    for(p = Glob.hints->first; p; p = p->next)
      unindex_hint(p->c);
    done_with_hints();
  }
  delete_clist(Glob.hints);
  Glob.hints = NULL;

}  // free_search_memory

/*************
 *
 *   handle_proof_and_maybe_exit()
 *
 *************/

static
void handle_proof_and_maybe_exit(Clause empty_clause)
{
  Term answer = get_term_attribute(empty_clause->attributes, Att.answer, 1);
  Plist proof = get_clause_ancestors(empty_clause);
  Plist p;

  Glob.empties = plist_append(Glob.empties, empty_clause);

  Stats.proofs++;

  uncompress_clauses(proof);

  if (!flag(Opt->quiet)) {
    bell(stderr);
    fprintf(stderr, "-------- Proof %d -------- ", Stats.proofs);
    if (answer != NULL)
      fwrite_term_nl(stderr, answer);
    else
      fprintf(stderr, "\n");
  }

  printf("\n-------- Proof %d -------- (%.2f + %.2f seconds) ",
	 Stats.proofs, user_seconds(), system_seconds());
	 
  if (answer != NULL)
    fwrite_term_nl(stdout, answer);
  else
    fprintf(stdout, "\n");

  if (flag(Opt->print_proofs)) {
    if (flag(Opt->standard_proofs)) {
      printf("\n-------- PROOF --------\n");
      printf("Length of proof is %d.\n", proof_length(proof));
      printf("Maximum clause weight is %d.\n", max_clause_weight(proof));
      for (p = proof; p; p = p->next)
	fwrite_clause(stdout, p->v, TRUE, TRUE);
      printf("\n-------- end of proof  -------\n");
    }
    if (flag(Opt->x_proofs) || flag(Opt->xr_proofs)) {
      Ilist map;
      Plist xproof = expand_proof(proof, ORDINARY_UNIF, &map);
      if (flag(Opt->x_proofs)) {
	printf("\n-------- XPROOF -------- \n");
	printf("Length of xproof is %d.\n", proof_length(xproof));
	for (p = xproof; p; p = p->next)
	  fwrite_clause_jmap(stdout, p->v,TRUE,TRUE,map);

	printf("\n-------- end of xproof -------- \n");
      }
      if (flag(Opt->xr_proofs)) {
	Plist xrproof = copy_clauses_ija(xproof);
	renumber_proof(xrproof, 1);
	printf("\n-------- XRPROOF -------- \n");
	printf("Length of xrproof is %d.\n", proof_length(xproof));
	for (p = xrproof; p; p = p->next)
	  fwrite_clause(stdout, p->v, TRUE, TRUE);
	printf("\n-------- end of xrproof -------- \n");
	zap_clauses_ija(xrproof);
      }
      zap_ilist(map);
      zap_clauses_ija(xproof);
    }
    if (flag(Opt->r_proofs) || flag(Opt->rx_proofs)) {
      Plist rproof = copy_clauses_ija(proof);
      renumber_proof(rproof, 1);
      if (flag(Opt->r_proofs)) {
	printf("\n-------- RPROOF -------- \n");
	printf("Length of rproof is %d.\n", proof_length(rproof));
	for (p = rproof; p; p = p->next)
	  fwrite_clause(stdout, p->v, TRUE, TRUE);
	printf("\n-------- end of rproof -------- \n");
      }
      if (flag(Opt->rx_proofs)) {
	Ilist map;
	Plist rxproof = expand_proof(rproof, ORDINARY_UNIF, &map);
	printf("\n-------- RXPROOF -------- \n");
	printf("Length of rxproof is %d.\n", proof_length(rxproof));
	for (p = rxproof; p; p = p->next)
	  fwrite_clause_jmap(stdout, p->v,TRUE,TRUE,map);
	printf("\n-------- end of rxproof -------- \n");
	zap_clauses_ija(rxproof);
      }
      zap_clauses_ija(rproof);
    }
  }  /* print_proofs */

  actions_in_proof(proof, &Att, done_with_search);

  if (parm(Opt->max_proofs) != -1 && Stats.proofs>=parm(Opt->max_proofs))
    done_with_search(MAX_PROOFS_EXIT);  /* does not return */

  else if (!flag(Opt->reuse_goals)) {
    /* Disable all goal clauses in the proof.  This is safe,
       because the only indexes referring to the goals are
       back_demod and index_literals.
    */
    Plist goals = clauses_in_clist(proof, Glob.goals);
    Plist p;
    for (p = goals; p; p = p->next) {
      printf("\n%% Disabling goal clause %d.\n", ((Clause) p->v)->id);
      disable_clause(p->v);
    }
    zap_plist(goals);
  }
  
  zap_plist(proof);
}  // handle_proof_and_maybe_exit

/*************
 *
 *   cl_process()
 *
 *   Process a newly inferred (or input) clause.
 *
 *   It is likely that a pointer to this routine was passed to
 *   an inference rule, and that inference rule called this routine
 *   with a new clause.
 *
 *   If this routine decides to keep the clause, it is appended
 *   to the Limbo list rather than to Sos.  Clauses in Limbo have been
 *   kept, but operations that can delete clauses (back subsumption,
 *   back demoulation) have not yet been applied.  The Limbo list
 *   is processed after the inference rule is finished.
 *
 *   Why use the Limbo list?  Because we're not allowed to delete
 *   clauses (back subsumption, back demodulation, back unit deletion)
 *   while inferring clauses.
 *
 *   Why not infer the whole batch of clauses, and then process them?
 *   Because there can be too many.  We have to do demodulation and
 *   subsumption right away, and get kept clauses indexed for
 *   forward demodulation and forward subsumption right away,
 *   so they can be used on the next inferred clause.
 *
 *   There are 2 versions:
 *
 *     cl_process (the ordinary version)
 *         calls cl_process_simplify, cl_process_keep
 *
 *     cl_process_goal (this does fewer operations)
 *         calls cl_process_simplify, cl_process_keep
 *
 *************/

/* First, some helper functions. */

static
void cl_process_simplify(Clause c)
{
  Stats.generated++;
  statistic_actions("generated", Stats.generated);
  if (flag(Opt->print_gen))
    fwrite_clause(stdout, c, TRUE, TRUE);

  if (!clist_empty(Glob.demods)) {
    clock_start(Clocks.demod);
    demodulate_clause(c, parm(Opt->demod_limit), !flag(Opt->quiet));
    if (flag(Opt->print_gen)) {
      printf("demod: ");
      fwrite_clause(stdout, c, TRUE, TRUE);
    }
    clock_stop(Clocks.demod);
  }

  simplify_literals(c);  // Remove any instances of x!=x.
  
  orient_equalities(c, TRUE);
  merge_literals(c);

  if (flag(Opt->unit_deletion)) {
    clock_start(Clocks.unit_del);
    unit_deletion(c);
    clock_stop(Clocks.unit_del);
  }

  if (flag(Opt->cac_redundancy)) {
    clock_start(Clocks.redundancy);
    cac_redundancy(c, !flag(Opt->quiet));  // Simplify C or AC redundant literals to $T
    clock_stop(Clocks.redundancy);
  }
}  // cl_process_simplify

static
void cl_process_keep(Clause c)
{
  Stats.kept++;
  statistic_actions("kept", clause_ids_assigned());  /* Note different stat */
  if (!c->normal_vars)
    renumber_variables(c, MAX_VARS);
  if (c->id == 0)
    assign_clause_id(c);  // unit conflict may have already assigned ID
  mark_parents_as_used(c);
  mark_maximal_literals(c);
  if (flag(Opt->print_kept)) {
    printf("kept: ");
    fwrite_clause(stdout, c, TRUE, TRUE);
  }
  else if (flag(Opt->print_labeled) &&
	   get_string_attribute(c->attributes, Att.label, 1)) {
    printf("\nkept (wt=%d): ", c->weight);
    fwrite_clause(stdout, c, TRUE, TRUE);
  }
}  // cl_process_keep

static
void cl_process_conflict(Clause c)
{
  if (number_of_literals(c) == 1) {
    Plist empties, p;
    if (!c->normal_vars)
      renumber_variables(c, MAX_VARS);
    clock_start(Clocks.conflict);
    empties = unit_conflict(c, parm(Opt->max_proofs) - Stats.proofs);
    clock_stop(Clocks.conflict);
    for (p = empties; p; p = p->next) {
      assign_clause_id(p->v);
      mark_parents_as_used(p->v);
      handle_proof_and_maybe_exit(p->v);
    }
    zap_plist(empties);  /* shallow zap */
  }
}  // cl_process_conflict

static
BOOL cl_process_weight_ok(Clause c)
{
  clock_start(Clocks.weigh);
  c->weight = clause_weight(c);
  clock_stop(Clocks.weigh);
  if (!clist_empty(Glob.hints)) {
    clock_start(Clocks.hints);
    if (!c->normal_vars)
      renumber_variables(c, MAX_VARS);
    adjust_weight_with_hints(c,
			     Att.bsub_hint_add_wt,
			     Att.bsub_hint_wt,
			     Att.label,
			     flag(Opt->degrade_hints));
    clock_stop(Clocks.hints);
  }
  if (c->weight > parm(Opt->default_weight) && 
      c->weight <= parm(Opt->max_weight))
    c->weight = parm(Opt->default_weight);

  if (c->used)
    return TRUE; // (e.g., by unit conflict) keep
  else if (c->weight > parm(Opt->max_weight)) {
    Stats.weight_deleted++;
    return FALSE;  // discard
  }
  else if (c->weight > 50 &&
	   clist_number_of_weight(Glob.limbo, c->weight) >= 3) {
    Stats.sos_limit_deleted++;
    return FALSE;  // discard
  }
  else if (!sos_keep(c, Glob.sos, Opt)) {
    Stats.sos_limit_deleted++;
    return FALSE;  // discard
  }
  else
    return TRUE;  // keep
}  // cl_process_weight_ok

static
void cl_process_new_demod(Clause c)
{
  if (flag(Opt->back_demod)) {
    int type = demodulator_type(c,
				parm(Opt->lex_dep_demod_lim),
				flag(Opt->lex_dep_demod_sane));
    if (type != NOT_DEMODULATOR) {
      if (flag(Opt->print_kept)) {
	char *s;
	switch(type) {
	case ORIENTED:     s = ""; break;
	case LEX_DEP_LR:   s = " (lex_dep_lr)"; break;
	case LEX_DEP_RL:   s = " (lex_dep_rl)"; break;
	case LEX_DEP_BOTH: s = " (lex_dep_both)"; break;
	default:           s = " (?)";
	}
	printf("    new demodulator%s: %d.\n", s, c->id);
      }
      clist_append(c, Glob.demods);
      index_demodulator(c, type, INSERT, Clocks.index);
      Stats.new_demodulators++;
      if (type != ORIENTED)
	Stats.new_lex_demods++;
    }
  }
}  // cl_process_new_demod

static
void cl_process_goal(Clause c)
{
  cl_process_simplify(c);  // same as ordinary cl_process
  if (number_of_literals(c) == 0) {
    assign_clause_id(c);
    mark_parents_as_used(c);
    handle_proof_and_maybe_exit(c);
  }
  else {
    // skip tautology, weight, forward subsumption, hints.
    cl_process_conflict(c);
    cl_process_keep(c);
    // index and append to Goals
    index_goal(c, INSERT, Clocks.index); // back unit del, conflict
    index_back_demod(c, INSERT, Clocks.index, flag(Opt->back_demod));
    clist_append(c, Glob.goals);
  }
}  // cl_process_goal

static
void cl_process(Clause c)
{
  // If the infer_clock is running, stop it and restart it when done.

  BOOL infer_clock_stopped = FALSE;
  if (clock_running(Clocks.infer)) {
    clock_stop(Clocks.infer);
    infer_clock_stopped = TRUE;
  }
  clock_start(Clocks.preprocess);

  exit_if_over_limit();
  if (parm(Opt->report) > 0)
    possible_report();

  cl_process_simplify(c);

  if (number_of_literals(c) == 0) {
    assign_clause_id(c);
    mark_parents_as_used(c);
    handle_proof_and_maybe_exit(c);
  }
  else if (tautology(c) || eq_tautology(c)) {
    Stats.subsumed++;
    delete_clause(c);  // tautology
  }
  else if (number_of_literals(c) > parm(Opt->max_literals)) {
    Stats.literals_deleted++;
    delete_clause(c);  // too many literals
  }
  else {
    if (flag(Opt->safe_unit_conflict))
      cl_process_conflict(c);  // marked as used if conflict

    if (!cl_process_weight_ok(c))
      delete_clause(c);  // too heavy
    else {
      Clause subsumer;
      clock_start(Clocks.subsume);
      subsumer = forward_subsumption(c);
      clock_stop(Clocks.subsume);
      if (subsumer != NULL && !c->used) {
	Stats.subsumed++;
	delete_clause(c);  // subsumed
      }
      else {
	cl_process_keep(c);
	if (!flag(Opt->safe_unit_conflict))
	  cl_process_conflict(c);
	cl_process_new_demod(c);
	// We insert c into the literal index now so that it will be
	// available for unit conflict and forward subsumption while
	// it's in limbo.  (It should not be back subsumed while in limbo.
	// See fatal error in limbo_process).
	index_literals(c, INSERT, Clocks.index, Glob.no_fapl);
	clist_append(c, Glob.limbo);
      }  // not subsumed
    }  // not weight deleted
  }  // not empty clause or tautology
  
  clock_stop(Clocks.preprocess);
  if (infer_clock_stopped)
    clock_start(Clocks.infer);
}  // cl_process

/*************
 *
 *   back_demod()
 *
 *   For each clause that can be back demodulated, make a copy,
 *   disable the original, cl_process the copy (as if it
 *   had just been inferred).
 *
 *************/

static
void back_demod(Clause demod)
{
  Plist results, p, prev;

  clock_start(Clocks.back_demod);
  results = back_demodulatable(demod,
		       demodulator_type(demod,
					parm(Opt->lex_dep_demod_lim),
					flag(Opt->lex_dep_demod_sane)));
  clock_stop(Clocks.back_demod);
  p = results;
  while(p != NULL) {
    Clause old = p->v;
    if (!clist_member(old, Glob.disabled)) {
      Clause new;
      BOOL goal = clist_member(old, Glob.goals);
      if (flag(Opt->basic_paramodulation))
	new = copy_clause_with_flag(old, nonbasic_flag());
      else
	new = copy_clause(old);
      Stats.back_demodulated++;
      if (flag(Opt->print_kept))
	printf("        %d back demodulating %d.\n", demod->id, old->id);
      new->justification = back_demod_just(old);
      new->attributes = inheritable_att_instances(old->attributes, NULL);
      disable_clause(old);
      if (goal)
	cl_process_goal(new);
      else
	cl_process(new);
    }
    prev = p;
    p = p->next;
    free_plist(prev);
  }
}  // back_demod

/*************
 *
 *   back_unit_deletion()
 *
 *   For each clause that can be back unit deleted, make a copy,
 *   disable the original, cl_process the copy (as if it
 *   had just been inferred).
 *
 *************/

static
void back_unit_deletion(Clause unit)
{
  Plist results, p, prev;

  clock_start(Clocks.back_unit_del);
  results = back_unit_deletable(unit);
  clock_stop(Clocks.back_unit_del);
  p = results;
  while(p != NULL) {
    Clause old = p->v;
    if (!clist_member(old, Glob.disabled)) {
      Clause new;
      BOOL goal = clist_member(old, Glob.goals);
      if (flag(Opt->basic_paramodulation))
	new = copy_clause_with_flag(old, nonbasic_flag());
      else
	new = copy_clause(old);
      Stats.back_unit_deleted++;
      if (flag(Opt->print_kept))
	printf("        %d back unit deleting %d.\n", unit->id, old->id);
      new->justification = back_unit_deletion_just(old);
      new->attributes = inheritable_att_instances(old->attributes, NULL);
      disable_clause(old);
      if (goal)
	cl_process_goal(new);
      else
	cl_process(new);
    }
    prev = p;
    p = p->next;
    free_plist(prev);
  }
}  // back_unit_deletion

/*************
 *
 *   limbo_process()
 *
 *   Apply back subsumption and back demodulation to the Limbo
 *   list.  Since back demodulated clauses are cl_processed,
 *   the Limbo list can grow while it is being processed.
 *
 *   If this prover had a hot-list, or any other feature that
 *   generates clauses from newly kept clauses, it probably would
 *   be done here.
 *
 *   The Limbo list operates as a queue.  An important property
 *   of the Limbo list is that if A is ahead of B, then A does
 *   not simplify or subsume B.  However, B can simplify or subsume A.
 *
 *************/

static
void limbo_process(void)
{
  while (Glob.limbo->first) {
    Clause c = Glob.limbo->first->c;

    // factoring

    if (flag(Opt->factor))
      binary_factors(c, cl_process);

    // new constants

    if (parm(Opt->new_constants) > Stats.new_constants) {
      Clause new = new_constant(c);
      if (new) {
	Stats.new_constants++;
	if (!flag(Opt->quiet)) {
	  printf("\n%% New constant: ");
	  f_clause(new);
	}
	cl_process(new);
      }
    }

    // fold denial

    if (parm(Opt->fold_denial_max) > 1 &&
	Glob.number_of_input_clauses > 3  &&
	(has_input_just(c) || has_copy_just(c))) {
      Clause new = fold_denial(c, parm(Opt->fold_denial_max));
      if (new) {
	Stats.new_constants++;
	if (!flag(Opt->quiet)) {
	  printf("\n%% Fold denial: ");
	  f_clause(new);
	}
	cl_process(new);
      }
    }

    // back subsumption

    if (flag(Opt->back_subsume)) {
      Plist subsumees;
      clock_start(Clocks.back_subsume);
      subsumees = back_subsumption(c);
      while (subsumees != NULL) {
	Clause d = subsumees->v;
	if (clist_member(d, Glob.limbo))  // See comment in cl_process.
	  fatal_error("back subsume limbo clause");
	Stats.back_subsumed++;
	if (flag(Opt->print_kept))
	  printf("    %d back subsumes %d.\n", c->id, d->id);
	disable_clause(d);
	subsumees = plist_pop(subsumees);
      }
      clock_stop(Clocks.back_subsume);
    }

    // back demodulation

    if (clist_member(c, Glob.demods)) {
      if (flag(Opt->print_kept))
	printf("    starting back demodulation with %d.\n", c->id);
      back_demod(c);  // This calls cl_process on rewritable clauses.
    }

    // back unit_deletion

    if (flag(Opt->back_unit_deletion) && unit_clause(c)) {
      back_unit_deletion(c);  // This calls cl_process on rewritable clauses.
    }

    // move to Sos and index for back demodulation

    clist_remove(c, Glob.limbo);

    if (clist_length(Glob.sos) >= parm(Opt->sos_limit)) {
      sos_displace(Glob.sos, Opt, disable_clause);
      Stats.sos_displaced++;
    }

    insert_into_sos(c, Glob.sos, pick_set(c, Opt));
    index_back_demod(c, INSERT, Clocks.index, flag(Opt->back_demod));
  }
}  // limbo_process

/*************
 *
 *   given_infer()
 *
 *   Make given_clause inferences according to the flags that are set.
 *   Inferred clauses are sent to cl_process().
 *
 *   We could process the Limbo list after each inference rule,
 *   and this might improve performance in some cases.  But it seems
 *   conceptually simplier if we process the Limbo clauses after
 *   all of the inferences have been made.
 *
 *************/

static
void given_infer(Clause given)
{
  clock_start(Clocks.infer);

  if (flag(Opt->binary_resolution))
    binary_resolution(given,
		      TRUE,
		      Glob.clashable_idx,
		      cl_process);

  if (flag(Opt->neg_binary_resolution))
    binary_resolution(given,
		      FALSE,
		      Glob.clashable_idx,
		      cl_process);

  if (flag(Opt->hyper_resolution))
    hyper_resolution(given, TRUE, Glob.clashable_idx, cl_process);

  if (flag(Opt->neg_hyper_resolution))
    hyper_resolution(given, FALSE, Glob.clashable_idx, cl_process);

  if (flag(Opt->pos_ur_resolution))
    ur_resolution(given, POS_RES, Glob.clashable_idx, cl_process);

  if (flag(Opt->neg_ur_resolution))
    ur_resolution(given, NEG_RES, Glob.clashable_idx, cl_process);

  if (flag(Opt->paramodulation) &&
      number_of_literals(given) <= parm(Opt->para_lit_limit)) {
    // This paramodulation does not use indexing.
    Context cf = get_context();
    Context ci = get_context();
    Clist_pos p;
    for (p = Glob.usable->first; p != NULL; p = p->next) {
      if (number_of_literals(p->c) <= parm(Opt->para_lit_limit)) {
	para_from_into(given, cf, p->c, ci, FALSE, cl_process);
	para_from_into(p->c, cf, given, ci, TRUE,  cl_process);
      }
    }
    free_context(cf);
    free_context(ci);
  }
  clock_stop(Clocks.infer);
}  // given_infer

/*************
 *
 *   make_inferences()
 *
 *   Assume that there are inferences to make.
 *
 *   If we had the option of using the pair algorithm instead of
 *   the given algorithm, we could make that decision here.
 *
 *************/

static
void make_inferences(void)
{
  Clause given_clause;

  clock_start(Clocks.pick_given);
  given_clause = get_given_clause(Glob.sos, Stats.given, Opt);
  clock_stop(Clocks.pick_given);

  if (given_clause != NULL) {
    statistic_actions("given", Stats.given);
    Stats.given++;

    if (Stats.given == 500) {
      int ratio = (Stats.back_subsumed == 0 ?
		   INT_MAX :
		   Stats.kept / Stats.back_subsumed);
      if (ratio > 20) {
	clear_flag(Opt->back_subsume, TRUE);
	printf("\n%% Back_subsumption disabled, ratio of kept"
	       " to back_subsumed is %d (%.2f of %.2f sec).\n",
	       ratio, clock_seconds(Clocks.back_subsume), user_seconds());
	fflush(stdout);
      }
    }
    
    if (Stats.given > parm(Opt->max_given))
      done_with_search(MAX_GIVEN_EXIT);

    if (flag(Opt->print_given)) {
      printf("\ngiven #%d (wt=%d): ", Stats.given, given_clause->weight);
      fwrite_clause(stdout, given_clause, TRUE, TRUE);
    }
    clist_append(given_clause, Glob.usable);
    index_clashable(given_clause, INSERT);
    given_infer(given_clause);
  }
}  // make_inferences

/*************
 *
 *   orient_input_eq()
 *
 *   This is designed for input clauses, and it's a bit tricky.  If any
 *   equalities are flipped, we make the flip operations info inferences
 *   so that proofs are complete.  This involves replacing and hiding
 *   (disabling) the original clause.
 *
 *************/

static
Clause orient_input_eq(Clause c)
{
  Clause new = copy_inference(c);
  orient_equalities(new, TRUE);
  if (clause_ident(c, new)) {
    delete_clause(new);
    /* the following puts "oriented" marks on c */
    orient_equalities(c, TRUE);
    return c;
  }
  else {
    /* Replace c with new in Usable. */
    assign_clause_id(new);
    mark_parents_as_used(new);
    clist_swap(c, new);
    clist_append(c, Glob.disabled);
    return new;
  }
}  /* orient_input_eq */

/*************
 *
 *   auto1_inference()
 *
 *   This looks at the clauses and decides which inference rules to use.
 *
 *************/

static
void auto1_inference(Clist sos, Clist usable, Clist goals, 
		    Prover_options opt)
{
  BOOL print = !flag(opt->quiet);
  Plist clauses;
  BOOL equality, nonunits, horn;
  int number_of_clauses;

  clauses = prepend_clist_to_plist(NULL, sos);
  clauses = prepend_clist_to_plist(clauses, usable);
  clauses = prepend_clist_to_plist(clauses, goals);

  equality          = pos_equality_in_clauses(clauses);
  nonunits          = !unit_clauses(clauses);
  horn              = horn_clauses(clauses);
  number_of_clauses = plist_count(clauses);

  if (print)
    printf("\n%% Auto inference settings:\n");

  if (equality) {
    if (print)
      printf("  %% set(paramodulation).  %% (positive equality literals)\n");
    set_flag(opt->paramodulation, print);
  }

  if (nonunits) {
    if (horn) {
      if (equality) {
	if (print)
	  printf("  %% set(hyper_resolution).  %% (nonunit Horn with equality)\n");
	set_flag(opt->hyper_resolution, print);
	if (print)
	  printf("  %% set(neg_ur_resolution).  %% (nonunit Horn with equality)\n");
	set_flag(opt->neg_ur_resolution, print);
	if (print)
	  printf("  %% assign(para_lit_limit, 2).  %% (nonunit Horn with equality)\n");
	assign_parm(opt->para_lit_limit, 2, print);
      }
      else {
	int diff = neg_pos_depth_difference(clauses);
	if (diff > 0) {
	  if (print)
	    printf("  %% set(hyper_resolution).  %% (HNE depth_diff=%d)\n", diff);
	  set_flag(opt->hyper_resolution, print);
	}
	else {
	  if (print)
	    printf("  %% set(neg_binary_resolution).  %% (HNE depth_diff=%d)\n", diff);
	  set_flag(opt->neg_binary_resolution, print);
	  if (print)
	    printf("  %% clear(ordered_inference).  %% (HNE depth_diff=%d)\n", diff);
	  clear_flag(opt->ordered_inference, print);
	  if (print)
	    printf("  %% set(ur_resolution).  %% (HNE depth_diff=%d)\n", diff);
	  set_flag(opt->ur_resolution, print);
	}
      }
    }
    else {
      // there are nonhorn clauses
      if (print)
	printf("  %% set(binary_resolution).  %% (non-Horn)\n");
      set_flag(opt->binary_resolution, print);
      if (number_of_clauses < 100) {
	if (print)
	  printf("  %% set(neg_ur_resolution).  %% (less than 100 clauses)\n");
	set_flag(opt->neg_ur_resolution, print);
      }
    }
  }
  zap_plist(clauses);
}  /* auto1_inference */

/*************
 *
 *   reasonable_options()
 *
 *   This looks at the clauses and sets reasonable options.
 *
 *************/

static
void reasonable_options(Clist sos, Clist usable, Clist goals,
			Prover_options opt)
{
  BOOL print    = !flag(opt->quiet);
  Plist sos2    = prepend_clist_to_plist(NULL, sos);
  Plist usable2 = prepend_clist_to_plist(NULL, usable);
  Plist goals2  = prepend_clist_to_plist(NULL, goals);

  BOOL changes = FALSE;
  BOOL horn = (horn_clauses(sos2) &&
	       horn_clauses(usable2) &&
	       horn_clauses(goals2));

  if (print)
    printf("\n%% Reasonable options, based on the structure of the clauses:\n");

  if (!horn) {
    if (print)
      printf("  %% set(factor), because non-Horn\n");
    set_flag(opt->factor, print);
    if (print)
      printf("  %% set(back_unit_deletion), because non-Horn\n");
    set_flag(opt->back_unit_deletion, print);
    changes = TRUE;
  }

  if (horn && (neg_nonunit_in_clauses(sos2) ||
	       neg_nonunit_in_clauses(usable2) ||
	       neg_nonunit_in_clauses(goals2))) {
    if (print)
	printf("  %% set(back_unit_deletion), because Horn and neg nonunits.\n");
    set_flag(opt->back_unit_deletion, print);
    changes = TRUE;
  }

  if (horn && flag(opt->hyper_resolution) && !flag(opt->back_subsume)) {
    Glob.no_fapl = TRUE;
  }

  if (!changes && print)
    printf("  %% (nothing changed)\n");

  zap_plist(sos2);
  zap_plist(usable2);
  zap_plist(goals2);
}  /* reasonable_options */

/*************
 *
 *   init_search()
 *
 *************/

static
void init_search(void)
{
  // Initialize clocks.
  
  Clocks.pick_given    = clock_init("pick_given");
  Clocks.infer         = clock_init("infer");
  Clocks.preprocess    = clock_init("preprocess");
  Clocks.demod         = clock_init("demod");
  Clocks.unit_del      = clock_init("unit_deletion");
  Clocks.redundancy    = clock_init("redundancy");
  Clocks.conflict      = clock_init("conflict");
  Clocks.weigh         = clock_init("weigh");
  Clocks.hints         = clock_init("hints");
  Clocks.subsume       = clock_init("subsume");
  Clocks.back_subsume  = clock_init("back_subsume");
  Clocks.back_demod    = clock_init("back_demod");
  Clocks.back_unit_del = clock_init("back_unit_del");
  Clocks.index         = clock_init("index");
  Clocks.disable       = clock_init("disable");

  init_actions(Glob.actions);
  init_weight(Glob.weights,
	      parm(Opt->variable_weight),
	      parm(Opt->constant_weight),
	      parm(Opt->not_weight),
	      parm(Opt->or_weight),
	      parm(Opt->sk_constant_weight),
	      parm(Opt->prop_atom_weight),
	      parm(Opt->skolem_penalty),
	      parm(Opt->nest_penalty));

  // Term ordering

  if (stringparm(Opt->order, "lpo")) {
    assign_order_method(LPO_METHOD);
    all_symbols_lrpo_status(LRPO_LR_STATUS);
    set_lrpo_status(str_to_sn(EQ_SYM, 2), LRPO_MULTISET_STATUS);
  }
  else if (stringparm(Opt->order, "rpo")) {
    assign_order_method(RPO_METHOD);
    all_symbols_lrpo_status(LRPO_MULTISET_STATUS);
  }
  else if (stringparm(Opt->order, "kbo")) {
    assign_order_method(KBO_METHOD);
  }

  symbol_order(Glob.usable, Glob.sos, Glob.demods, Glob.goals, TRUE);

  if (Glob.kb_weights) {
    if (!stringparm(Opt->order, "kbo")) {
      assign_stringparm(Opt->order, "kbo", TRUE);
      printf("%% assign(order, kbo), because KB weights were given.\n");
    }
    init_kb_weights(Glob.kb_weights);
    print_kb_weights(stdout);
  }
  else if (stringparm(Opt->order, "kbo")) {
    auto_kb_weights(Glob.usable, Glob.sos);
    print_kb_weights(stdout);
  }

  if (!flag(Opt->quiet)) {
    print_rsym_precedence(stdout);
    print_fsym_precedence(stdout);
  }

  if (flag(Opt->inverse_order)) {
    if (exists_preliminary_precedence()) {  // lex command
      if (!flag(Opt->quiet))
	printf("%% Skipping inverse_order, because there is a lex command.\n");
    }
    else if (stringparm(Opt->order, "kbo")) {
      if (!flag(Opt->quiet))
	printf("%% Skipping inverse_order, because term ordering is KBO.\n");
    }
    else {
      inverse_order(Glob.sos);
      if (!flag(Opt->quiet)) {
	printf("%% After inverse_order: ");
	print_fsym_precedence(stdout);
      }
    }
  }

  if (parm(Opt->unfold_eq_limit) >= 0)
    unfold_eq_defs(Glob.sos, parm(Opt->unfold_eq_limit), 3,
		   !flag(Opt->quiet));
  else if (flag(Opt->fold_eq)) {
    fold_eq_defs(Glob.sos);
    if (!flag(Opt->quiet)) {
      printf("%% After fold_eq: ");
      print_fsym_precedence(stdout);
    }
  }

  // Automatic settings
  
  if (flag(Opt->auto_inference))
    auto1_inference(Glob.sos, Glob.usable, Glob.goals, Opt);

  if (!flag(Opt->hands_off_options))
    reasonable_options(Glob.sos, Glob.usable, Glob.goals, Opt);

  // Adjust options

  if (parm(Opt->max_proofs) == -1)
    assign_parm(Opt->max_proofs, INT_MAX, FALSE);
  
  if (parm(Opt->new_constants) == -1)
    assign_parm(Opt->new_constants, INT_MAX, FALSE);

  // Tell packages about options and other things.

  resolution_options(flag(Opt->ordered_inference),
		     flag(Opt->ordered_instance),
		     stringparm1(Opt->literal_selection),
		     flag(Opt->initial_nuclei),
		     parm(Opt->nucleus_limit));
		     
  paramodulation_options(flag(Opt->ordered_inference),
			 flag(Opt->ordered_instance),
			 flag(Opt->basic_paramodulation));

  set_max_megs(parm(Opt->max_megs));
  set_max_megs_proc(max_megs_exit);

}  /* init_search */

/*************
 *
 *   index_and_process_input_clauses()
 *
 *************/

static
void index_and_process_input_clauses(void)
{
  Clist_pos p;
  Clist temp_sos;

  // Index Usable clauses if hyper, UR, or binary-res are set.

  Glob.use_clash_idx = (flag(Opt->binary_resolution) ||
			flag(Opt->neg_binary_resolution) ||
			flag(Opt->hyper_resolution) ||
			flag(Opt->neg_hyper_resolution) ||
			flag(Opt->pos_ur_resolution) ||
			flag(Opt->neg_ur_resolution));

  // Allocate and initialize indexes (even if they won't be used).

  init_literals_index();  // fsub, bsub, fudel, budel, ucon

  init_demodulator_index(DISCRIM_BIND, ORDINARY_UNIF, 0);

  init_back_demod_index(FPA, ORDINARY_UNIF, 10);

  Glob.clashable_idx = lindex_init(FPA, ORDINARY_UNIF, 10,
				   FPA, ORDINARY_UNIF, 10);

  init_hints(ORDINARY_UNIF);

  // Do Sos last, in case we PROCESS_INITIAL_SOS.

  ////////////////////////////////////////////////////////////////////////////
  // Usable

  for (p = Glob.usable->first; p != NULL; p = p->next) {
    Clause c = p->c;
    mark_maximal_literals(c);
    if (Opt->dont_flip_input)
      orient_equalities(c, FALSE);
    else
      c = orient_input_eq(c);  /* this replaces c if any flipping occurs */
    index_literals(c, INSERT, Clocks.index, Glob.no_fapl);
    index_back_demod(c, INSERT, Clocks.index, flag(Opt->back_demod));
    index_clashable(c, INSERT);
  }

  ////////////////////////////////////////////////////////////////////////////
  // Demodulators

  for (p = Glob.demods->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (!pos_eq_unit(c))
      fatal_error("input demodulator is not equation");
    else {
      int type;
      if (flag(Opt->dont_flip_input))
	orient_equalities(c, FALSE);  /* don't allow flips */
      else
	c = orient_input_eq(c);  /* this replaces c if any flipping occurs */
      if (c->justification->next != NULL) {
	printf("\nNOTE: input demodulator %d has been flipped.\n", c->id);
	fprintf(stderr, "\nNOTE: input demodulator %d has been flipped.\n",
		c->id);
	bell(stderr);
      }
      type = demodulator_type(c,
			      parm(Opt->lex_dep_demod_lim),
			      flag(Opt->lex_dep_demod_sane));
      if (flag(Opt->dont_flip_input) &&
	  type != ORIENTED &&
	  !renamable_flip_eq(c->literals->atom)) {
	type = ORIENTED;  /* let the user beware */
	mark_oriented_eq(c->literals->atom);
	fprintf(stderr, "\nWARNING: demodulator does not satisfy term order\n");
	printf("\nWARNING: demodulator does not satisfy term order: ");
	f_clause(c);
      }
      else if (type == NOT_DEMODULATOR) {
	Term a = ARG(c->literals->atom,0);
	Term b = ARG(c->literals->atom,1);
	printf("bad input demodulator: "); f_clause(c);
	if (term_ident(a, b))
	  fatal_error("input demodulator is instance of x=x");
	else if (!variables_subset(a, b) && !variables_subset(b, a))
	  fatal_error("input demoulator does not have var-subset property");
	else
	  fatal_error("input demoulator not allowed");
      }
      index_demodulator(c, type, INSERT, Clocks.index);
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  // Hints
  
  if (Glob.hints->first) {
    for (p = Glob.hints->first; p != NULL; p = p->next) {
      Clause h = p->c;
      orient_equalities(h, TRUE);
      // If the hint already has a weight attribute, it overrides the
      // following, because it is earlier in the list of attributes.
      if (parm(Opt->bsub_hint_add_wt) != 0)
	h->attributes = set_int_attribute(h->attributes,
					  Att.bsub_hint_add_wt,
					  parm(Opt->bsub_hint_add_wt));
      if (parm(Opt->bsub_hint_wt) != INT_MAX)
	h->attributes = set_int_attribute(h->attributes,
					  Att.bsub_hint_wt,
					  parm(Opt->bsub_hint_wt));
      index_hint(h, Att.label);  // This is handled in the hints package.
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  // Sos

  // Move Sos to a temporary list, then process that temporary list,
  // putting the clauses back into Sos in the "correct" way, either
  // by calling cl_process() or doing it here.

  temp_sos = Glob.sos;                    // move Sos to a temporary list
  name_clist(temp_sos, "temp_sos");       // not really necessary
  Glob.sos = clist_init("sos");           // get a new Sos list

  if (flag(Opt->process_initial_sos)) {
    // Do not apply max_weight or max_literals while processing initial sos.
    int max_weight = parm(Opt->max_weight);
    int max_literals = parm(Opt->max_literals);
    assign_parm(Opt->max_weight, INT_MAX, TRUE);
    assign_parm(Opt->max_literals, INT_MAX, TRUE);
    
    // Make a "copy" inference and treat as if inferred.
    while (temp_sos->first) {
      Clause c = temp_sos->first->c;
      Clause new = copy_inference(c);
      clist_remove(c, temp_sos);
      clist_append(c, Glob.disabled);
      cl_process(new);
    }
    // This will put processed clauses back into Sos.
    limbo_process();  // back subsumption and back demodulation.

    // Restore max_weight and max_literals.
    assign_parm(Opt->max_weight, max_weight, TRUE);
    assign_parm(Opt->max_literals, max_literals, TRUE);

    // The following prevents trivial "copy" inferences in proofs.
    for (p = Glob.sos->first; p; p = p->next)
      fix_input_just(p->c);
  }
  else {
    /* not applying full processing initial sos */
    while (temp_sos->first) {
      Clause c = temp_sos->first->c;
      clist_remove(c, temp_sos);
      if (Opt->dont_flip_input)
	orient_equalities(c, FALSE);
      else
	c = orient_input_eq(c);
      mark_maximal_literals(c);
      c->weight = clause_weight(c);
      if (!clist_empty(Glob.hints)) {
	clock_start(Clocks.hints);
	adjust_weight_with_hints(c,
				 Att.bsub_hint_add_wt,
				 Att.bsub_hint_wt,
				 Att.label,
				 flag(Opt->degrade_hints));
	clock_stop(Clocks.hints);
      }
      insert_into_sos(c, Glob.sos, pick_set(c, Opt));
      index_literals(c, INSERT, Clocks.index, Glob.no_fapl);
      index_back_demod(c, INSERT, Clocks.index, flag(Opt->back_demod));
    }
  }

  clist_zap(temp_sos);  // free the temporary list

  // mark all sos clauses as "initial"

  for (p = Glob.sos->first; p; p = p->next)
    p->c->initial = TRUE;

  ////////////////////////////////////////////////////////////////////////////
  // Goals.  Make a "copy" inference and call cl_process_goal.

  if (!clist_empty(Glob.goals)) {
    Clist_pos p;
    Clist temp = Glob.goals;           // move Goals to a temporary list
    name_clist(temp, "temp");          // not really necessary
    Glob.goals = clist_init("goals");  // get a new Goals list
    while (temp->first) {
      Clause c = temp->first->c;
      Clause new = copy_inference(c);
      clist_remove(c, temp);
      clist_append(c, Glob.disabled);
      cl_process_goal(new);
    }
    clist_zap(temp);  // free the empty list

    // The following prevents trivial "copy" inferences in proofs.
    for (p = Glob.goals->first; p; p = p->next)
      fix_input_just(p->c);
  }

  ////////////////////////////////////////////////////////////////////////////

  if (flag(Opt->print_initial_clauses)) {
    printf("\n%% Clauses after input processing:\n");
    fwrite_clause_clist(stdout,Glob.usable,      TRUE,TRUE);
    fwrite_clause_clist(stdout,Glob.sos,         TRUE,TRUE);
    fwrite_demod_clist(stdout,Glob.demods,TRUE,TRUE);
    fwrite_clause_clist(stdout,Glob.goals,       TRUE,TRUE);
  }
  if (Glob.hints->length > 0)
    printf("\n%% %d hints processed (%d are redundant).\n",
	   Glob.hints->length, redundant_hints());
}  // index_and_process_input_clauses

/*************
 *
 *   fatal_setjmp()
 *
 *************/

static
void fatal_setjmp(void)
{
  int return_code = setjmp(Jump_env);
  if (return_code != 0)
    fatal_error("longjmp called outside of search");
}  /* fatal_setjmp */

/*************
 *
 *   collect_prover_results()
 *
 *************/

static
Prover_results collect_prover_results(BOOL xproofs)
{
  Plist p;
  Prover_results results = calloc(1, sizeof(struct prover_results));

  for (p = Glob.empties; p; p = p->next) {
    Plist proof = get_clause_ancestors(p->v);
    uncompress_clauses(proof);
    results->proofs = plist_append(results->proofs, proof);
    if (xproofs) {
      Plist xproof = proof_to_xproof(proof);
      results->xproofs = plist_append(results->xproofs, xproof);
    }
  }
  update_stats();  /* puts package stats into Stats */
  results->stats = Stats;  /* structure copy */
  results->user_seconds = user_seconds();
  results->system_seconds = system_seconds();
  results->return_code = Glob.return_code;
  return results;
}  /* collect_prover_results */

/*************
 *
 *   zap_prover_results()
 *
 *************/

/* DOCUMENTATION
Free the dynamically allocated memory associated with a Prover_results.
*/

/* PUBLIC */
void zap_prover_results(Prover_results results)
{
  Plist a, b;  /* results->proofs is a Plist of Plist of clauses */
  for (a = results->proofs; a; a = a->next) {
    for (b = a->v; b; b = b->next) {
      Clause c = b->v;
      /* There is a tricky thing going on with the ID.  If you try
	 to delete a clause with an ID not in the clause ID table,
	 a fatal error occurs.  If IDs in these clauses came from
	 a child process, they will not be in the table.  Setting
	 the ID to 0 gets around that problem.
       */
      c->id = 0;
      delete_clause(c);  /* zaps justification, attributes */
    }
  }
  free(results);
}  /* zap_prover_results */

/*************
 *
 *   search()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Prover_results search(Prover_input p)
{
  int return_code = setjmp(Jump_env);
  if (return_code != 0) {
    // we just landed from longjmp(); fix return code and return
    Glob.return_code = (return_code == INT_MAX ? 0 : return_code);
    fatal_setjmp();  /* This makes longjmps cause a fatal_error. */
    return collect_prover_results(p->xproofs);
  }
  else {
    // search for a proof

    printf("\n%% Entering prover search function.\n");

    Opt = p->options;          // put options into a global variable
    Glob.initialized = TRUE;   // this signifies that Glob is being used

    Glob.start_time  = user_seconds();
    Glob.start_ticks = bogo_ticks();

    if (flag(Opt->sort_initial_sos) && plist_count(p->sos) <= 100)
      p->sos = sort_plist(p->sos,
			  (Ordertype (*)(void*, void*)) clause_compare_m4);

    // Copy clauses and term lists into Glob.

    Glob.usable = copy_clauses_to_clist(p->usable, "usable", TRUE);
    Glob.sos    = copy_clauses_to_clist(p->sos, "sos", TRUE);
    Glob.demods = copy_clauses_to_clist(p->demods,"demodulators",TRUE);
    Glob.goals  = copy_clauses_to_clist(p->goals, "goals", TRUE);
    Glob.hints  = copy_clauses_to_clist(p->hints, "hints", TRUE);

    Glob.weights    = tlist_copy(p->weights);
    Glob.kb_weights = tlist_copy(p->kb_weights);
    Glob.actions    = tlist_copy(p->actions);

    Glob.number_of_input_clauses = clause_ids_assigned();
    printf("%% Assigned IDs to %d clauses.\n", Glob.number_of_input_clauses);
    
    // Allocate auxiliary clause lists.

    Glob.limbo    = clist_init("limbo");
    Glob.disabled = clist_init("disabled");
    Glob.empties  = NULL;

    // predicate elimination

    if (flag(p->options->predicate_elimination) &&
	clist_empty(Glob.usable) && clist_empty(Glob.goals)) {
      predicate_elimination(Glob.sos, Glob.disabled,
			    !flag(Opt->quiet));
    }

    // structure_of_clauses(Glob.sos);  // just outputs some data on clauses
    // structure_of_variables(Glob.sos);  // just outputs some data on clauses

    // printf("%% Sos before input processing:\n");
    // fwrite_clause_clist(stdout,Glob.sos,      TRUE,TRUE);

    init_search(); // init clocks, ordering, auto-mode, init packages
    
    index_and_process_input_clauses();

    printf("\n%% Starting search at %.2f seconds.\n", user_seconds());
    fflush(stdout);
    Glob.start_time = user_seconds();

    // ****************************** Main Loop ******************************

    while (inferences_to_make()) {

      // make_inferences: each inferred clause is cl_processed, which
      // does forward demodulation and subsumption; if the clause is kept
      // it is put on the Limbo list, and it is indexed so that it can be
      // used immediately with subsequent newly inferred clauses.

      make_inferences();

      // limbo_process: this applies back subsumption and back demodulation,
      // and moves the clauses to the Sos list.

      limbo_process();

      // adjust_sos_limit: maybe adjust the sos limit size; if it is
      // reduced, low priority sos clauses may be disabled.

      adjust_sos_limit(Glob.sos, Stats.given,
		       (int) (user_seconds() - Glob.start_time),
		       bogo_ticks() - Glob.start_ticks,
		       Opt, disable_clause);

    }  // ************************ end of main loop ************************

    fprint_all_stats(stdout);
    fatal_setjmp();  /* This makes longjmps cause a fatal_error. */
    Glob.return_code = SOS_EMPTY_EXIT;
    return collect_prover_results(p->xproofs);
  }
}  /* search */

/*************
 *
 *   forking_search()
 *
 *************/

/* DOCUMENTATION
This is similar to search(), except that a child process is created
to do the search, and the child sends its results to the parent on
a pipe.

<P>
The parameters and results are the same as search().
As in search(), the Plists lists of objets (the parameters) are not changed.
*/

/* PUBLIC */
Prover_results forking_search(Prover_input input)
{
  Prover_results results;

  int rc;
  int fd[2];          /* pipe for child -> parent data */

  rc = pipe(fd);
  if (rc != 0)
    fatal_error("forking_search: pipe creation failed");

  fflush(stdout);
  fflush(stderr);
  rc = fork();
  if (rc < 0)
    fatal_error("forking_search: fork failed");
  
  if (rc == 0) {

    /*********************************************************************/
    /* This is the child process.  Search, send results to parent, exit. */
    /*********************************************************************/

    BOOL print = !flag(input->options->quiet);

    int to_parent = fd[1];  /* fd for writing data to parent */
    close(fd[0]);           /* close "read" end of pipe */

    fprintf(stdout,"\n%% Child search process %d started.\n", my_process_id());
    /* fprintf(stderr,"\n%% Child search process %d started.\n\n", my_process_id()); */

    /* child processes cannot do some things */

    if (parm(input->options->new_constants) > 0) {
      if (print)
	printf("  %% assign(new_constants, 0).  %% (child process cannot do this)\n");
      assign_parm(input->options->new_constants, 0, print);
    }

    if (parm(input->options->fold_denial_max) > 0) {
      if (print)
	printf("  %% assign(fold_denial_max, 0).  %% (child process cannot do this)\n");
      assign_parm(input->options->fold_denial_max, 0, print);
    }

    /* search */
    
    results = search(input);

    /* send results to the parent */

    {
      /* Format of data (all integers) sent to parent:
	---------------------- 
	number-of-proofs
	  number-of-steps
	    [clauses-in-proof]
	  number-of-steps
	    [clauses-in-proof]
          ...
        [same for xproofs]

	stats  (MAX_STATS of them)
	user_milliseconds
	system_milliseconds
	return_code
	---------------------- 
      */

      Ibuffer ibuf = ibuf_init();
      Plist p, a;

      /* collect proofs */

      ibuf_write(ibuf, plist_count(results->proofs));  /* number of proofs */
      for (p = results->proofs; p; p = p->next) {
	ibuf_write(ibuf, plist_count(p->v));  /* steps in this proof */
	for (a = p->v; a; a = a->next)
	  put_clause_to_ibuf(ibuf, a->v);
      }
      
      /* collect xproofs */

      ibuf_write(ibuf, plist_count(results->xproofs));  /* number of xproofs */
      for (p = results->xproofs; p; p = p->next) {
	ibuf_write(ibuf, plist_count(p->v));  /* steps in this proof */
	for (a = p->v; a; a = a->next)
	  put_clause_to_ibuf(ibuf, a->v);
      }
      
      {
	/* collect stats (shortcut: handle stats struct as sequence of ints) */
	int *x = (void *) &(results->stats);
	int n = sizeof(struct prover_stats) / sizeof(int);
	int i;
	for (i = 0; i < n; i++)
	  ibuf_write(ibuf, x[i]);
      }

      /* collect clocks */
      ibuf_write(ibuf, (int) (results->user_seconds * 1000));
      ibuf_write(ibuf, (int) (results->system_seconds * 1000));
      /* collect return_code */
      ibuf_write(ibuf, results->return_code);

      /* write the data to the pipe */

      rc = write(to_parent,
		 ibuf_buffer(ibuf),
		 ibuf_length(ibuf) * sizeof(int));
      if (rc != ibuf_length(ibuf) * sizeof(int))
	fatal_error("forking_search, incomplete write from child to parent");
      rc = close(to_parent);
      
      ibuf_free(ibuf);  /* not necessary, because we're going to exit now */
    }

    /* child exits */

    // printf("child symbols: "); p_syms(); fflush(stdout);

    exit_with_message(stdout, results->return_code);
    
    return NULL;  /* won't happen */

  }  /* end of child code */

  else {
  
    // printf("parent symbols: "); p_syms();  fflush(stdout);

    /*********************************************************************/
    /* This is the parent process.  Get results from child, then return. */
    /*********************************************************************/

    int from_child = fd[0];  /* fd for reading data from child */
    close(fd[1]);            /* close "write" end of pipe */

    /* read results from child (read waits until data are available) */

    {
      Ibuffer ibuf = fd_read_to_ibuf(from_child);
      int num_proofs, num_steps, i, j;

      results = calloc(1, sizeof(struct prover_results));
      
      /* read proofs */

      num_proofs = ibuf_read(ibuf);
      for (i = 0; i < num_proofs; i++) {
	Plist proof = NULL;
	num_steps = ibuf_read(ibuf);
	for (j = 0; j < num_steps; j++) {
	  Clause c = get_clause_from_ibuf(ibuf);
	  proof = plist_prepend(proof, c);  /* build backward */
	}
	results->proofs = plist_append(results->proofs, reverse_plist(proof));
      }

      /* read xproofs */

      num_proofs = ibuf_read(ibuf);
      for (i = 0; i < num_proofs; i++) {
	Plist proof = NULL;
	num_steps = ibuf_read(ibuf);
	for (j = 0; j < num_steps; j++) {
	  Clause c = get_clause_from_ibuf(ibuf);
	  proof = plist_prepend(proof, c);  /* build backward */
	}
	results->xproofs = plist_append(results->xproofs, reverse_plist(proof));
      }

      {
	/* read stats (shortcut: handle stats struct as sequence of ints) */
	int *x = (void *) &(results->stats);
	int n = sizeof(struct prover_stats) / sizeof(int);
	int i;
	for (i = 0; i < n; i++)
	  x[i] = ibuf_read(ibuf);
      }

      /* read clocks */
      results->user_seconds = ibuf_read(ibuf) / 1000.0;
      results->system_seconds = ibuf_read(ibuf) / 1000.0;
      /* read return_code */
      results->return_code = ibuf_read(ibuf);
    }

    /* Wait for child to exit and get the exit code.  We should not
       have to wait long, because we already have its results. */

    {
      int child_status, child_exit_code;
      wait(&child_status);
      if (!WIFEXITED(child_status))
	fatal_error("forking_search: child terminated abnormally");
      child_exit_code = WEXITSTATUS(child_status);
      results->return_code = child_exit_code;
    }

    rc = close(from_child);

    return results;
  }  /* end of parent code */
}  /* forking_search */
