#ifndef TP_SEARCH_H
#define TP_SEARCH_H

#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <setjmp.h>  /* Yikes! */

#include "../ladr/clock.h"
#include "../ladr/glist.h"

/* INTRODUCTION
*/

/* Public definitions */

// Options

typedef struct prover_options * Prover_options;

struct prover_options {
  // These are integer IDs assigned by init_flag(), init_parm, init_stringparm.

  // Flags (Boolean options)
  int 
    binary_resolution,       // inference rules
    neg_binary_resolution,
    hyper_resolution,
    neg_hyper_resolution,
    ur_resolution,
    pos_ur_resolution,
    neg_ur_resolution,
    factor,
    paramodulation,

    ordered_inference,       // restrictions on inference rules
    ordered_instance,
    para_units_only,
    basic_paramodulation,
    initial_nuclei,

    process_initial_sos,     // processing generated clauses
    back_demod,
    lex_dep_demod,
    lex_dep_demod_sane,
    safe_unit_conflict,
    reuse_goals,
    back_subsume,
    unit_deletion,
    back_unit_deletion,
    cac_redundancy,
    degrade_hints,
    dont_flip_input,

    echo_input,              // output
    quiet,
    print_subproblems,
    print_initial_clauses,
    print_given,
    print_gen,
    print_kept,
    print_labeled,
    print_proofs,
    standard_proofs,
    x_proofs,
    r_proofs,
    rx_proofs,
    xr_proofs,
    default_output,

    fof_reduction,           // preprocessing
    predicate_elimination,
    inverse_order,
    unfold_eq,
    fold_eq,
    sort_initial_sos,

    breadth_first,          // selecting given clause
    input_sos_first,

    automatic,            // auto is a reserved symbol in C!
    auto1,
    auto_inference,
    hands_off_options,
    casc_2005;

  // Parms (Integer options)

  int
    max_given,             // search limits
    max_kept,
    max_proofs,
    max_megs,
    max_seconds,

    new_constants,        // inference
    para_lit_limit,
    nucleus_limit,

    unfold_eq_limit,      // fold / unfold
    fold_denial_max,

    pick_given_ratio,      // select given clause
    age_part,
    weight_part,
    weight_neg_part,

    max_weight,            // processing generated clauses
    lex_dep_demod_lim,
    max_literals,
    demod_limit,

    variable_weight,        // weighting parameters
    constant_weight,
    not_weight,
    or_weight,
    sk_constant_weight,
    prop_atom_weight,
    skolem_penalty,
    nest_penalty,
    bsub_hint_add_wt,
    bsub_hint_wt,
    default_weight,

    sos_limit,             // sontrol size of SOS
    min_sos_limit,
    lrs_interval,
    lrs_seconds,
    lrs_ticks,

    report;

  // Stringparms (string options)

  int
    order,               // LPO, RPO, KBO
    literal_selection,
    stats;
};

// Clocks

struct prover_clocks {
  Clock pick_given,
    infer,
    preprocess,
    demod,
    redundancy,
    unit_del,
    conflict,
    weigh,
    hints,
    subsume,
    back_subsume,
    back_demod,
    back_unit_del,
    index,
    disable;
};

// Statistics

typedef struct prover_stats * Prover_stats;

struct prover_stats {
  unsigned  given,
    generated,
    kept,
    proofs,
    weight_deleted,
    literals_deleted,
    subsumed,
    back_subsumed,
    sos_limit_deleted,
    sos_displaced,
    sos_removed,
    new_demodulators,
    new_lex_demods,
    back_demodulated,
    back_unit_deleted,
    demod_attempts,
    demod_rewrites,
    res_instance_prunes,
    para_instance_prunes,
    basic_para_prunes,
    nonunit_fsub,
    nonunit_bsub,
    usable_size,
    sos_size,
    demodulators_size,
    disabled_size,
    hints_size,
    goals_size,
    limbo_size,
    kbyte_usage,
    new_constants;
};

// Search input

typedef struct prover_input * Prover_input;

struct prover_input {
  // clause lists
  Plist usable, sos, demods, goals, hints;
  // formula lists
  Plist fof_usable, fof_sos, fof_goals;
  // term lists
  Plist actions, weights, kb_weights;
  // ordinary options
  Prover_options options;
  // extra options
  BOOL xproofs;  // tell search() to return xproofs as well as ordinary proofs
};

// Search results

typedef struct prover_results * Prover_results;

struct prover_results {
  Plist proofs;
  Plist xproofs;
  struct prover_stats stats;
  double user_seconds, system_seconds;
  int return_code;
};

/* End of public definitions */

/* Public function prototypes from search.c */

Prover_options init_prover_options(void);

void init_prover_attributes(void);

int label_att(void);

void fprint_prover_clocks(FILE *fp, struct prover_clocks clks);

void exit_with_message(FILE *fp, int code);

void prover_sig_handler(int condition);

void free_search_memory(void);

void zap_prover_results(Prover_results results);

Prover_results search(Prover_input p);

Prover_results forking_search(Prover_input input);

#endif  /* conditional compilation of whole file */
