#include "utilities.h"

#include <sys/wait.h>
#include <unistd.h>
#include "../ladr/clausify.h"
#include "../ladr/top_input.h"
#include "../ladr/hash.h"
#include "../ladr/di_tree.h"
#include "../ladr/ioutil.h"
#include "../ladr/flatterm.h"
#include "../ladr/resolve.h"
#include "../ladr/demod.h"
#include "../ladr/clause3.h"

/* Private definitions and types */

/*************
 *
 *   print_memory_stats()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void print_memory_stats(FILE *fp)
{
  fprintf(fp, "\nMegs malloced by palloc(): %d.\n", megs_malloced());
  
  fprint_strbuf_mem(fp, TRUE);
  fprint_parse_mem(fp, FALSE);
  fprint_hash_mem(fp, FALSE);
  fprint_term_mem(fp, FALSE);
  fprint_attrib_mem(fp, FALSE);
  fprint_glist_mem(fp, FALSE);

  fprint_just_mem(fp, FALSE);
  fprint_formula_mem(fp, FALSE);
  fprint_clause_mem(fp, FALSE);
  fprint_clist_mem(fp, FALSE);

  fprint_unify_mem(fp, FALSE);
  fprint_btm_mem(fp, FALSE);
  fprint_btu_mem(fp, FALSE);

  fprint_fpa_mem(fp, FALSE);
  fprint_fpalist_mem(fp, FALSE);
  fprint_discrim_mem(fp, FALSE);
  fprint_discrimb_mem(fp, FALSE);
  fprint_discrimw_mem(fp, FALSE);
  fprint_flatterm_mem(fp, FALSE);
  fprint_mindex_mem(fp, FALSE);
  fprint_lindex_mem(fp, FALSE);
  fprint_clash_mem(fp, FALSE);
  fprint_di_tree_mem(fp, FALSE);

  memory_report(fp);
}  // print_memory_stats

/*************
 *
 *   fsym_collect()
 *
 *************/

static
void fsym_collect(Ilist *table, Term t, int depth)
{
  if (VARIABLE(t))
    return;
  else {
    int i;
    table[SYMNUM(t)] = ilist_prepend(table[SYMNUM(t)], depth);
    for (i = 0; i < ARITY(t); i++)
      fsym_collect(table, ARG(t,i), depth+1);
  }
}  /* fsym_collect */

/*************
 *
 *   inverse_axiom()
 *
 *************/

static
BOOL inverse_axiom(Clause c, int *f2, int *f1, int *f0)
{
  if (number_of_literals(c) != 1 || !pos_eq(c->literals))
    return FALSE;
  else {
    Term alpha = ARG(c->literals->atom,0);
    Term beta  = ARG(c->literals->atom,1);
    if (CONSTANT(alpha))
      { Term t = alpha; alpha = beta; beta = t; }
    if (!CONSTANT(beta) || ARITY(alpha) != 2)
      return FALSE;
    else {
      Term a0 = ARG(alpha,0);
      Term a1 = ARG(alpha,1);
      if (VARIABLE(a0))
	{ Term t = a0; a0 = a1; a1 = t; }
      if (!VARIABLE(a1) || ARITY(a0) != 1)
	return FALSE;
      else if (!term_ident(ARG(a0,0),a1))
	return FALSE;
      else {
	*f2 = SYMNUM(alpha);
	*f1 = SYMNUM(a0);
	*f0 = SYMNUM(beta);
	return TRUE;
      }
    }
  }
}  /* inverse_axiom */

/*************
 *
 *   fsym_report()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void fsym_report(Ilist fsyms, Plist clauses)
{
  Ilist *table = calloc(greatest_symnum()+1  , sizeof(Ilist));
  Plist p;

  for (p = clauses; p; p = p->next) {
    Clause c = p->v;
    Literal lit;
    int f0, f1, f2;
    for (lit = c->literals; lit; lit = lit->next)
      fsym_collect(table, lit->atom, 0);
    if (inverse_axiom(c, &f2, &f1, &f0))
      { printf("Inverse axiom: "); f_clause(c); }
  }

  {
    Ilist a;
    printf("Symbols:\n");
    for (a = fsyms; a; a = a->next) {
      Ilist b;
      int n = 0;
      printf("  Symbol %22s/%d: ", sn_to_str(a->i), sn_to_arity(a->i));
      for (b = table[a->i]; b; b = b->next) {
	printf(" %d", b->i);
	n += b->i;
      }
      printf("   (%d)\n", n);
    }
  }
  free(table);
}  /* fsym_report */

/*************
 *
 *   inverse_order()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void inverse_order(Clist clauses)
{
  Clist_pos p;
  Ilist binary = NULL;
  Ilist unary = NULL;
    
  for (p = clauses->first; p; p = p->next) {
    int f2, f1, f0;
    if (inverse_axiom(p->c, &f2, &f1, &f0)) {
      binary = ilist_append(binary, f2);
      unary = ilist_append(unary, f1);
    }
  }
  lex_insert_after_initial_constants(unary);
  lex_insert_after_initial_constants(binary);
  zap_ilist(unary);
  zap_ilist(binary);
}  /* inverse_order */

/*************
 *
 *   symbol_order()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void symbol_order(Clist usable, Clist sos, Clist demods,
		  Clist goals, BOOL echo)
{
  Ilist fsyms, rsyms;
  Mlist fsyms_multiset, rsyms_multiset;
  Plist nonneg;

  Plist all = NULL;
  all = prepend_clist_to_plist(all, usable);
  all = prepend_clist_to_plist(all, sos);
  all = prepend_clist_to_plist(all, demods);
  all = prepend_clist_to_plist(all, goals);

  fsyms = fsym_set_in_clauses(all);
  rsyms = rsym_set_in_clauses(all);

  // fsym_report(fsyms, all);

  nonneg = nonneg_clauses(all);
  zap_plist(all);

  fsyms_multiset = function_symbols_in_clauses(nonneg);
  rsyms_multiset = relation_symbols_in_clauses(nonneg);

  lex_order_3(fsyms, rsyms, fsyms_multiset, rsyms_multiset);

  if (echo)  {
    Ilist p, missing_fsyms, missing_rsyms;

    // print any symbols missing from the lex command
    missing_fsyms = not_in_preliminary_precedence(fsyms);
    if (missing_fsyms && !ilist_subset(fsyms, missing_fsyms)) {
      fprintf(stderr, "WARNING, function symbols not in lex command.\n");
      printf("\n%% WARNING, function symbols not in lex command:");
      for (p = missing_fsyms; p; p = p->next)
	printf(" %s", sn_to_str(p->i));
      printf(".\n");
    }
    zap_ilist(missing_fsyms);

    missing_rsyms = not_in_preliminary_precedence(rsyms);
    if (missing_rsyms && !ilist_subset(rsyms, missing_rsyms)) {
      fprintf(stderr, "WARNING, relation symbols not in lex command.\n");
      printf("\n%% WARNING, relation symbols not in lex command:");
      for (p = missing_rsyms; p; p = p->next)
	printf(" %s", sn_to_str(p->i));
      printf(".\n");
    }
    zap_ilist(missing_rsyms);
      
  }

  init_features(fsyms, rsyms);  // feature-vector subsumption for nonunits
  zap_ilist(fsyms);
  zap_ilist(rsyms);
}  /* symbol_order */

/*************
 *
 *   unary_symbols()
 *
 *************/

/* DOCUMENTATION
Given an Ilist, return the sublist that represent unary symbols.
*/

/* PUBLIC */
Ilist unary_symbols(Ilist a)
{
  Ilist p;
  Ilist b = NULL;

  for (p = a; p; p = p->next) {
    if (sn_to_arity(p->i) == 1)
      b = ilist_append(b, p->i);
  }
  return b;
}  /* unary_symbols */

/*************
 *
 *   auto_kb_weights()
 *
 *************/

/* DOCUMENTATION
If there is exactly one unary function symbol in usable+sos,
give it kb_weight=0 and highest precedence.
*/

/* PUBLIC */
void auto_kb_weights(Clist usable, Clist sos)
{
  Plist clauses = NULL;
  Ilist fsyms, unaries;

  clauses = prepend_clist_to_plist(clauses, usable);  /* shallow */
  clauses = prepend_clist_to_plist(clauses, sos);     /* shallow */

  fsyms = fsym_set_in_clauses(clauses);
  unaries = unary_symbols(fsyms);

  if (ilist_count(unaries) == 1) {
    int symnum = unaries->i;
    set_kb_weight(symnum, 0);
    assign_greatest_precedence(symnum);
    printf("\n%% Assigning unary symbol %s kb_weight 0 and highest"
	   " precedence (%d).\n", sn_to_str(symnum), sn_to_lex_val(symnum));
  }

  zap_ilist(unaries);
  zap_ilist(fsyms);
  zap_plist(clauses);  /* shallow */
}  /* auto_kb_weights */

/*************
 *
 *   process_command_line_args()
 *
 *************/

static
void process_command_line_args(int argc, char **argv, Prover_options opt)
{
  extern char *optarg;
  int c;

  /* getopt() option string

     No colons after option:  no argument.
     One colon after option:  argument required.
     Two colon after options: argument optional. (GNU extension! Don't use it!)

     Initial colon: unrecognized options will not cause errors.
  */
  
  while ((c = getopt(argc, argv,":t:f")) != EOF) {
    switch (c) {
    case 't':
      {
	int n = atoi(optarg);
	int id = opt->max_seconds;
	printf("\n%% From the command line: assign(%s, %d).\n",
	       parm_id_to_str(id), n);
	assign_parm(id, n, TRUE);
	break;
      }
    case 'f':  /* input files */
      /* do nothing --- this is handled elsewhere */
      break;
    default:
      break;
      /* fatal_error("unrecognized command-line option"); */
    }
  }
}  /* process_command_line_args */

/*************
 *
 *   std_prover_init_and_input()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Prover_input std_prover_init_and_input(int argc, char **argv,
				       BOOL echo, int unknown_action)
{
  Prover_input p = calloc(1, sizeof(struct prover_input));

  init_standard_ladr();

  p->options = init_prover_options();

  if (string_member("--casc_2005", argv, argc)) {
    printf("\n%% From the command line: set(casc_2005).\n");
    set_flag(p->options->casc_2005, FALSE);
  }

  init_prover_attributes();

  signal(SIGINT,  prover_sig_handler);
  signal(SIGUSR1, prover_sig_handler);
  signal(SIGSEGV, prover_sig_handler);

  // Tell the top_input package what lists to accept and where to put them.

  accept_list("usable",       CLAUSES,  &(p->usable));
  accept_list("sos",          CLAUSES,  &(p->sos));
  accept_list("demodulators", CLAUSES,  &(p->demods));
  accept_list("goals",        CLAUSES,  &(p->goals));
  accept_list("hints",        CLAUSES,  &(p->hints));

  accept_list("usable",       FORMULAS, &(p->fof_usable));
  accept_list("sos",          FORMULAS, &(p->fof_sos));
  accept_list("goals",        FORMULAS, &(p->fof_goals));

  accept_list("actions",      TERMS,    &(p->actions));
  accept_list("weights",      TERMS,    &(p->weights));
  accept_list("kb_weights",   TERMS,    &(p->kb_weights));

  // Read commands such as set, clear, op, lex.
  // Read lists, filling in variables given to the accept_list calls.

  read_all_input(argc, argv, stdout, echo, unknown_action);

  process_command_line_args(argc, argv, p->options);

  if (echo)
    printf("\n%% Finished reading the input.\n");

  return p;
}  /* std_prover_init_and_input */

/*************
 *
 *   std_clausify()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void std_clausify(Prover_input p, BOOL echo)
{
  // Clausify any formulas and append to the corresponding clause lists.

  p->usable = plist_cat(p->usable, clausify_formulas(p->fof_usable));
  p->sos    = plist_cat(p->sos,    clausify_formulas(p->fof_sos));
  p->goals  = plist_cat(p->goals,  clausify_formulas(p->fof_goals));

}  /* std_clausify */

/*************
 *
 *   neg_pos_depth_diff()
 *
 *************/

static
int neg_pos_depth_diff(Clause c)
{
  Literal lit;
  int max_pos = 0;
  int max_neg = 0;
  for (lit = c->literals; lit; lit = lit->next) {
    int depth = term_depth(lit->atom);
    if (lit->sign)
      max_pos = IMAX(max_pos, depth);
    else
      max_neg = IMAX(max_neg, depth);
  }
  return max_neg - max_pos;
}  /* neg_pos_depth_diff */

/*************
 *
 *   neg_pos_wt_diff()
 *
 *************/

static
int neg_pos_wt_diff(Clause c)
{
  Literal lit;
  int max_pos = 0;
  int max_neg = 0;
  for (lit = c->literals; lit; lit = lit->next) {
    int wt = symbol_count(lit->atom);
    if (lit->sign)
      max_pos = IMAX(max_pos, wt);
    else
      max_neg = IMAX(max_neg, wt);
  }
  return max_neg - max_pos;
}  /* neg_pos_wt_diff */

/*************
 *
 *   neg_pos_depth_difference()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int neg_pos_depth_difference(Plist sos)
{
  Plist p;
  int neg = 0;
  int pos = 0;
  for (p = sos; p; p = p->next) {
    Clause c = p->v;
    if (mixed_clause(c)) {
      Literal lit;
      int max_pos = 0;
      int max_neg = 0;
      for (lit = c->literals; lit; lit = lit->next) {
	int depth = term_depth(lit->atom);
	if (lit->sign)
	  max_pos = IMAX(max_pos, depth);
	else
	  max_neg = IMAX(max_neg, depth);
      }
      neg += max_neg;
      pos += max_pos;
      // printf("max_neg=%d, max_pos=%d, ", max_neg, max_pos); f_clause(c);
    }
  }
  // printf("neg=%d, pos=%d\n", neg, pos);
  return neg - pos;
}  /* neg_pos_depth_difference */

/*************
 *
 *   structure_of_clauses()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void structure_of_clauses(Clist clauses)
{
  int num_pos = 0;
  int pos_wt = 0;
  int pos_depth = 0;

  int neg_wt = 0;
  int neg_depth = 0;
  int num_neg = 0;

  int mix_wt = 0;
  int mix_depth = 0;
  int num_mix = 0;

  int mix_wt_diff = 0;
  int mix_depth_diff = 0;

  int num_nonhorn = 0;
  int nonhorn = 0;

  int num_nonunit = 0;
  int nonunit = 0;

  Clist_pos p;
  for (p = clauses->first; p; p = p->next) {
    Clause c = p->c;
    int wt = clause_symbol_count(c);
    int depth = clause_depth(c);
    if (positive_clause(c)) {
      num_pos++;
      pos_wt += wt;
      pos_depth += depth;
    }
    else if (negative_clause(c)) {
      num_neg++;
      neg_wt += wt;
      neg_depth += depth;
    }
    else {
      num_mix++;
      mix_wt += wt;
      mix_depth += depth;
      mix_wt_diff += neg_pos_wt_diff(c);
      mix_depth_diff += neg_pos_depth_diff(c);
    }

    if (!horn_clause(c)) {
      num_nonhorn++;
      nonhorn += number_of_pos_literals(c) - 1;
    }

    if (!unit_clause(c)) {
      num_nonunit++;
      nonunit += number_of_literals(c);
    }

  }
  {
    double pw = num_pos == 0 ? 0 : pos_wt / (double) num_pos;
    double pd = num_pos == 0 ? 0 : pos_depth / (double) num_pos;

    double nw = num_neg == 0 ? 0 : neg_wt / (double) num_neg;
    double nd = num_neg == 0 ? 0 : neg_depth / (double) num_neg;

    double mw = num_mix == 0 ? 0 : mix_wt / (double) num_mix;
    double md = num_mix == 0 ? 0 : mix_depth / (double) num_mix;

    double mwd = num_mix == 0 ? 0 : mix_wt_diff / (double) num_mix;
    double mdd = num_mix == 0 ? 0 : mix_depth_diff / (double) num_mix;

    double nh = num_nonhorn == 0 ? 0 : nonhorn / (double) num_nonhorn;
    double nu = num_nonunit == 0 ? 0 : nonunit / (double) num_nonunit;

    printf("\n%% Struc"
	   " (N %2d %2d %2d)"
	   " (WT %5.2f %5.2f %5.2f %5.2f)"
	   " (DP %5.2f %5.2f %5.2f %5.2f)"
	   " (NH %2d %5.2f)"
	   " (NU %2d %5.2f)"
	   "\n",
	   num_pos, num_neg, num_mix,
	   pw, nw, mw, mwd,
	   pd, nd, md, mdd,
	   num_nonhorn, nh, num_nonunit, nu);
  }
}  /* structure_of_clauses */

/*************
 *
 *   p_term_list()
 *
 *************/

static
void p_term_list(Plist terms)
{
  Plist p;
  printf("[");
  for (p = terms; p; p = p->next) {
    fwrite_term(stdout, p->v);
    printf("%s", p->next ? "," : "");
  }
  printf("]");
}  /* p_term_list */

/*************
 *
 *   plist_size_of_diff()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int plist_size_of_diff(Plist a, Plist b)
{
  if (a == NULL)
    return 0;
  else {
    int n = plist_size_of_diff(a->next, b);
    return (!plist_member(b, a->v) ? n+1 : n);
  }
}  /* plist_size_of_diff */

/*************
 *
 *   structure_of_variables()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void structure_of_variables(Clist clauses)
{
  int n = 0;
  int np = 0;
  int nn = 0;

  Clist_pos p;
  for (p = clauses->first; p; p = p->next) {
    Clause c = p->c;
    if (mixed_clause(c)) {
      Plist pvars = NULL;
      Plist nvars = NULL;
      Literal lit;
      for (lit = c->literals; lit; lit = lit->next) {
	if (lit->sign)
	  pvars = set_of_vars(lit->atom, pvars);
	else
	  nvars = set_of_vars(lit->atom, nvars);
      }
      printf("\n");
      f_clause(c);

      printf(" nvars (%d): ", plist_size_of_diff(nvars, pvars));
      p_term_list(nvars); 
      printf(" pvars (%d): ", plist_size_of_diff(pvars, nvars));
      p_term_list(pvars); 
      n++;
      np += plist_count(pvars);
      nn += plist_count(nvars);
      zap_plist(nvars);
      zap_plist(pvars);
    }
  }
  printf("\nnn=%d, np=%d, n=%d\n", nn, np, n);
}  /* structure_of_variables */

/*************
 *
 *   clause_compare_m4()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Ordertype clause_compare_m4(Clause a, Clause b)
{
  /* pos < neg < mixed */
  BOOL a_pos = positive_clause(a);
  BOOL b_pos = positive_clause(b);
  BOOL a_neg = negative_clause(a);
  BOOL b_neg = negative_clause(b);
  
  if (a_pos && !b_pos)
    return LESS_THAN;
  else if (!a_pos && b_pos)
    return GREATER_THAN;
  else if (a_neg && !b_neg)
    return LESS_THAN;
  else if (!a_neg && b_neg)
    return GREATER_THAN;
  else {
    /* now both pos, both neg, or both mixed */
    /* fewer symbols < more symbols */
    int na = clause_symbol_count(a);
    int nb = clause_symbol_count(b);
    if (na > nb)
      return GREATER_THAN;
    else if (na < nb)
      return LESS_THAN;
    else {
      /* fewer literals < more literals */
      na = number_of_literals(a);
      na = number_of_literals(b);
      if (na > nb)
	return GREATER_THAN;
      else if (na < nb)
	return LESS_THAN;
      else {
	/* shallower < deeper */
	na = clause_depth(a);
	na = clause_depth(b);
	if (na > nb)
	  return GREATER_THAN;
	else if (na < nb)
	  return LESS_THAN;
	else 
	  return SAME_AS;  /* should we go further? */
      }
    }
  }
}  /* clause_compare_m4 */

/*************
 *
 *   bogo_ticks()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int bogo_ticks(void)
{
  return mega_mem_calls();
}  /* bogo_ticks */
