//
// File:         BufferedAllocationStatistics.hpp
// Description:  Raw memory allocation.
// Created:      Mar 2, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Apr 28, 2004, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Completely redefined. 
//=====================================================================
#ifndef BUFFERED_ALLOCATION_STATISTICS_H
#define BUFFERED_ALLOCATION_STATISTICS_H
//=====================================================================
#include <climits> 
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
//=====================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BUFFERED_ALLOCATION_STATISTICS
#  define DEBUG_NAMESPACE "BufferedAllocationStatistics<MaxSmallObjectSize>"
#endif 
#include "debugMacros.hpp"
//=====================================================================
namespace BK 
{

  template <ulong MaxSmallObjectSize>
  class BufferedAllocationStatistics
  {
  public:
    BufferedAllocationStatistics() 
    {
      reset();
    };
  
    ~BufferedAllocationStatistics()
    { 
      BK_CORRUPT(*this);
    };

    void reset()
    {  
      CALL("reset()");
      _bufferIsBeingUsed = false;
      _sizeOfRemainingNonbufferMemory = ULONG_MAX;

      for (ulong size = 0UL; size <= MaxSmallObjectSize; ++size)
	{
	  _small[size].reset(size);
	};

      _firstUsedSmallSize = 0;
      _lastUsedSmallSize = 0;

      _sizeOfFreeLarge = 0UL;
      _sizeOfAllocatedLarge = 0UL;
      _sizeOfActiveBufferDeficitInLarge = 0UL;
      _sizeOfSuspendedBufferDeficitInLarge = 0UL;
      _sizeOfBufferProficitInLarge = 0UL;

      _totalSizeOfAllocated = 0L;
      _totalSizeOfActiveBufferDeficit = 0L;

    }; // void reset()

    void setSizeOfNonbufferMemory(ulong size) 
    {
      _sizeOfRemainingNonbufferMemory = size;
    };

    void suspendActiveBufferDeficit()
    {
      CALL("suspendActiveBufferDeficit()");
      if (_bufferIsBeingUsed)
	{
	  DOP(ulong debugSuspended = 0UL);
	  for (SmallSizeStatistics* s = _firstUsedSmallSize; 
	       s; 
		   s = s->nextUsed)
	    {
	      s->suspendedBufferDeficit += s->activeBufferDeficit;
	      DOP(debugSuspended += (s->activeBufferDeficit * s->size));
	      s->activeBufferDeficit = 0UL;
	    };
	  _sizeOfSuspendedBufferDeficitInLarge += _sizeOfActiveBufferDeficitInLarge;
	  DOP(debugSuspended += _sizeOfActiveBufferDeficitInLarge);
	  _sizeOfActiveBufferDeficitInLarge = 0UL;
	  ASSERT(_totalSizeOfActiveBufferDeficit == debugSuspended);
	  _totalSizeOfActiveBufferDeficit = 0UL;
	};
      ASSERT(checkIntegrity());
    };

    ulong free(ulong size) const { return _small[size].free; };

    ulong allocated(ulong size) const { return _small[size].allocated; };
  
    ulong sizeOfFreeLarge() const
    {
      return _sizeOfFreeLarge;
    };
    ulong sizeOfAllocatedLarge() const
    {
      return _sizeOfAllocatedLarge;
    };

    ulong totalSizeOfAllocated() const
    {
      return _totalSizeOfAllocated;
    };

    ulong totalSizeOfBufferDeficit() const
    {
      CALL("totalSizeOfBufferDeficit() const");
      return totalSizeOfActiveBufferDeficit() + totalSizeOfSuspendedBufferDeficit();
    };
  
  
    ulong totalSizeOfActiveBufferDeficit() const
    {
      return _totalSizeOfActiveBufferDeficit;
    };

    ulong totalSizeOfSuspendedBufferDeficit() const
    {
      CALL("totalSizeOfSuspendedBufferDeficit() const");
      ulong res = _sizeOfSuspendedBufferDeficitInLarge;
      for (SmallSizeStatistics* s = _firstUsedSmallSize; s; s = s->nextUsed)
	{
	  res += s->suspendedBufferDeficit;
	};
      return res;
    };

  

    // registered status change

    void smallFreeToAlloc(ulong size)
    {
      CALL("smallFreeToAlloc(ulong size)");
      ASSERT(size <= MaxSmallObjectSize);
      ASSERT(_small[size].free > 0UL);
      if (_bufferIsBeingUsed)
	{
	  ASSERT(_small[size].used);
	  if (_small[size].bufferProficit)
	    {
	      --_small[size].bufferProficit;
	    }
	  else
	    if (_small[size].activeBufferDeficit || 
		_small[size].suspendedBufferDeficit)
	      {
		++_small[size].activeBufferDeficit;
		_totalSizeOfActiveBufferDeficit += size;
	      }
	    else
	      {
		// check large sizes for proficit
		if (_sizeOfBufferProficitInLarge >= size)
		  {
		    _sizeOfBufferProficitInLarge -= size;
		  }
		else
		  {
		    // check other small sizes for proficit
		    SmallSizeStatistics* otherUsed = _small[size].nextUsed;
		  check_other_used:
		    if (otherUsed)
		      {
			if (otherUsed->bufferProficit)
			  {
			    ASSERT(otherUsed->size > size);
			    --otherUsed->bufferProficit;
			    
			    // transfer remainder
			    ASSERT(!_small[otherUsed->size - size].activeBufferDeficit);
			    ASSERT(!_small[otherUsed->size - size].suspendedBufferDeficit);
			    ++_small[otherUsed->size - size].bufferProficit;
			    makeUsed(otherUsed->size - size);
			  }
			else
			  {
			    otherUsed = otherUsed->nextUsed;
			    goto check_other_used;
			  };
		      }
		    else
		      {
			++_small[size].activeBufferDeficit;
			_totalSizeOfActiveBufferDeficit += size;
		      };
		  };
	      };
	};
      --_small[size].free;
      ++_small[size].allocated;
      _totalSizeOfAllocated += size;
      ASSERT(checkIntegrity());
    }; // void smallFreeToAlloc(ulong size)

    void largeFreeToAlloc(ulong size)
    {
      CALL("largeFreeToAlloc(ulong size)");
      ASSERT(size > MaxSmallObjectSize);
      if (_bufferIsBeingUsed)
	{
	  if (_sizeOfBufferProficitInLarge >= size)
	    {
	      _sizeOfBufferProficitInLarge -= size;
	    }
	  else
	    {
	      ulong freshDeficit = size - _sizeOfBufferProficitInLarge;
	      _sizeOfBufferProficitInLarge = 0UL;
	      _sizeOfActiveBufferDeficitInLarge += freshDeficit;
	      _totalSizeOfActiveBufferDeficit += freshDeficit;
	    };
	};
      _sizeOfFreeLarge -= size;
      _sizeOfAllocatedLarge += size;
      _totalSizeOfAllocated += size;
      ASSERT(checkIntegrity());
    }; // void largeFreeToAlloc(ulong size)

    void smallAllocToFree(ulong size)
    {
      CALL("smallAllocToFree(ulong size)");
      ASSERT(size <= MaxSmallObjectSize);
      if (_bufferIsBeingUsed)
	{
	  makeUsed(size);
	  // adjust buffer proficit/deficit
	  if (_small[size].bufferProficit) 
	    // ^ implies that there is no deficit in smaller sizes
	    {
	      ++_small[size].bufferProficit;
	    }
	  else
	    if (_small[size].activeBufferDeficit)
	      {
		--_small[size].activeBufferDeficit;
		_totalSizeOfActiveBufferDeficit -= size;
	      }
	    else 
	      if (_small[size].suspendedBufferDeficit)
		{
		  --_small[size].suspendedBufferDeficit;
		}
	      else
		{
		  ulong remainingProficit = size;

		  // try to cover some active deficit in smaller sizes
		  
		  SmallSizeStatistics* otherUsed = _small[size].previousUsed;
		  while (otherUsed && remainingProficit)
		    {
		      while (otherUsed->size <= remainingProficit &&
			     otherUsed->activeBufferDeficit)
			{
			  --otherUsed->activeBufferDeficit;
			  _totalSizeOfActiveBufferDeficit -= otherUsed->size;
			  remainingProficit -= otherUsed->size;
			};
		      otherUsed = otherUsed->previousUsed;
		    };

		  // try to cover some suspended deficit in smaller sizes
		  otherUsed = _small[size].previousUsed;
		  while (otherUsed && remainingProficit)
		    {
		      while (otherUsed->size <= remainingProficit &&
			     otherUsed->suspendedBufferDeficit)
			{
			  ASSERT(!otherUsed->activeBufferDeficit);
			  --otherUsed->suspendedBufferDeficit;
			  remainingProficit -= otherUsed->size;
			};
		      otherUsed = otherUsed->previousUsed;
		    };
		  
		  if (remainingProficit)
		    {
		      ASSERT(!_small[remainingProficit].activeBufferDeficit);
		      ASSERT(!_small[remainingProficit].suspendedBufferDeficit);
		      ++_small[remainingProficit].bufferProficit;
		      makeUsed(remainingProficit);
		    };
		  
		};
	      
	}; // if (_bufferIsBeingUsed)

      ++_small[size].free;
      --_small[size].allocated;
      _totalSizeOfAllocated -= size;
      ASSERT(checkIntegrity());
    }; // void smallAllocToFree(ulong size)

    void largeAllocToFree(ulong size)
    {
      CALL("largeAllocToFree(ulong size)");
      ASSERT(size > MaxSmallObjectSize);
      if (_bufferIsBeingUsed)
	{
	  ulong remainingProficit = _sizeOfBufferProficitInLarge + size;
	  if (remainingProficit <= _sizeOfActiveBufferDeficitInLarge)
	    {
	      _sizeOfActiveBufferDeficitInLarge -= remainingProficit;
	      _totalSizeOfActiveBufferDeficit -= remainingProficit;
	      remainingProficit = 0UL;
	    }
	  else // remainingProficit > _sizeOfActiveBufferDeficitInLarge
	    {
	      remainingProficit -= _sizeOfActiveBufferDeficitInLarge;
	      _totalSizeOfActiveBufferDeficit -= _sizeOfActiveBufferDeficitInLarge;
	      _sizeOfActiveBufferDeficitInLarge = 0UL;
	      if (remainingProficit <= _sizeOfSuspendedBufferDeficitInLarge)
		{
		  _sizeOfSuspendedBufferDeficitInLarge -= remainingProficit;
		  remainingProficit = 0UL;
		}
	      else // remainingProficit > _sizeOfSuspendedBufferDeficitInLarge
		{
		  remainingProficit -= _sizeOfSuspendedBufferDeficitInLarge;
		  _sizeOfSuspendedBufferDeficitInLarge = 0UL;
		      
		  // now try to cover some active deficit in small sizes
		  SmallSizeStatistics* smallSize = _lastUsedSmallSize;
		  while (smallSize && remainingProficit)
		    {
		      while (smallSize->size <= remainingProficit &&
			     smallSize->activeBufferDeficit)
			{
			  --smallSize->activeBufferDeficit;
			  _totalSizeOfActiveBufferDeficit -= smallSize->size;
			  remainingProficit -= smallSize->size;
			};
		      smallSize = smallSize->previousUsed;
		    };

		      
		  // now try to cover some suspended deficit in small sizes
		  smallSize = _lastUsedSmallSize;
		  while (smallSize && remainingProficit)
		    {
		      while (smallSize->size <= remainingProficit &&
			  smallSize->suspendedBufferDeficit)
			{
			  ASSERT(!smallSize->activeBufferDeficit);
			  --smallSize->suspendedBufferDeficit;
			  remainingProficit -= smallSize->size;
			};
		      smallSize = smallSize->previousUsed;
		    };
		};

	    };
	  _sizeOfBufferProficitInLarge = remainingProficit;
	}; // if (_bufferIsBeingUsed)

      _sizeOfFreeLarge += size;
      _sizeOfAllocatedLarge -= size;
      _totalSizeOfAllocated -= size;
      ASSERT(checkIntegrity());
    }; // void largeAllocToFree(ulong size)
    

    void smallRawToAlloc(ulong size)
    {
      CALL("smallRawToAlloc(ulong size)");
      ASSERT(checkIntegrity());
      ASSERT(size <= MaxSmallObjectSize);
      if (_bufferIsBeingUsed)
	{
	use_buffer:
	  makeUsed(size);
	  if (_small[size].bufferProficit)
	    {
	      --_small[size].bufferProficit;
	    }
	  else
	    if (_small[size].activeBufferDeficit || 
		_small[size].suspendedBufferDeficit)
	      {
		++_small[size].activeBufferDeficit;
		_totalSizeOfActiveBufferDeficit += size;
	      }
	    else
	      {
		// check large sizes for proficit
		if (_sizeOfBufferProficitInLarge >= size)
		  {
		    _sizeOfBufferProficitInLarge -= size;
		  }
		else
		  {
		    // check other small sizes for proficit
		    SmallSizeStatistics* otherUsed = _small[size].nextUsed;
		  check_other_used:
		    if (otherUsed)
		      {
			if (otherUsed->bufferProficit)
			  {
			    ASSERT(otherUsed->size > size);
			    --otherUsed->bufferProficit;
			    
			    // transfer remainder
			    ASSERT(!_small[otherUsed->size - size].activeBufferDeficit);
			    ASSERT(!_small[otherUsed->size - size].suspendedBufferDeficit);
			    ++_small[otherUsed->size - size].bufferProficit;
			    makeUsed(otherUsed->size - size);
			  }
			else
			  {
			    otherUsed = otherUsed->nextUsed;
			    goto check_other_used;
			  };
		      }
		    else
		      {
			++_small[size].activeBufferDeficit;
			_totalSizeOfActiveBufferDeficit += size;
		      };
		  };
	      };
	  
	  
	}
      else // !_bufferIsBeingUsed
	{
	  if (size > _sizeOfRemainingNonbufferMemory)
	    {
	      startUsingBuffer();
	      goto use_buffer;
	    }
	  else
	    _sizeOfRemainingNonbufferMemory -= size;
	};

      ++_small[size].allocated;
      _totalSizeOfAllocated += size;
      ASSERT_IN(checkIntegrity(),"END");
    }; // void smallRawToAlloc(ulong size)

    void largeRawToAlloc(ulong size)
    {
      CALL("largeRawToAlloc(ulong size)");
      ASSERT(size > MaxSmallObjectSize);
      if (_bufferIsBeingUsed)
	{
	use_buffer:
	  if (_sizeOfBufferProficitInLarge >= size)
	    {
	      _sizeOfBufferProficitInLarge -= size;
	    }
	  else
	    {
	      ulong freshDeficit = size - _sizeOfBufferProficitInLarge;
	      _sizeOfBufferProficitInLarge = 0UL;
	      _sizeOfActiveBufferDeficitInLarge += freshDeficit;
	      _totalSizeOfActiveBufferDeficit += freshDeficit;
	    };
	}
      else // !_bufferIsBeingUsed
	{
	  if (size > _sizeOfRemainingNonbufferMemory)
	    {
	      startUsingBuffer();
	      goto use_buffer;
	    }
	  else
	    _sizeOfRemainingNonbufferMemory -= size;
	};

      _sizeOfAllocatedLarge += size;
      _totalSizeOfAllocated += size;
      ASSERT(checkIntegrity());
    }; // void largeRawToAlloc(ulong size)

    void smallRawToFree(ulong size)
    {
      CALL("smallRawToFree(ulong size)");
      smallRawToAlloc(size);
      smallAllocToFree(size);
    }; // void smallRawToFree(ulong size)

    void largeRawToFree(ulong size)
    {
      CALL("largeRawToFree(ulong size)");     
      largeRawToAlloc(size);
      largeAllocToFree(size);
    }; // void largeRawToFree(ulong size)

    void smallFreeToRaw(ulong size)
    {
      CALL("smallFreeToRaw(ulong size)");
      ASSERT(size <= MaxSmallObjectSize);
      ASSERT(_small[size].free);
      if (!_bufferIsBeingUsed)
	{
	  _sizeOfRemainingNonbufferMemory += size;
	};
      --_small[size].free;
      ASSERT(checkIntegrity());
    }; // void smallFreeToRaw(ulong size)

    void largeFreeToRaw(ulong size)
    {
      CALL("largeFreeToRaw(ulong size)");
      ASSERT(size > MaxSmallObjectSize);
      ASSERT(_sizeOfFreeLarge >= size); 
     
      if (!_bufferIsBeingUsed)
	{
	  _sizeOfRemainingNonbufferMemory += size;
	};
      _sizeOfFreeLarge -= size;
      ASSERT(checkIntegrity());
    }; // void largeFreeToRaw(ulong size)
  

    void largeFreeToSmallFree(ulong size)
    {
      CALL("largeFreeToSmallFree(ulong size)");
      ASSERT(size <= MaxSmallObjectSize);
      ASSERT(_sizeOfFreeLarge >= size); 
      if (_bufferIsBeingUsed)
	{
	  makeUsed(size);
	  
	  // adjust buffer proficit/deficit
	  if (_small[size].bufferProficit) 
	    // ^ implies that there is no deficit in smaller sizes
	    {
	      ++_small[size].bufferProficit;
	    }
	  else
	    if (_small[size].activeBufferDeficit)
	      {
		--_small[size].activeBufferDeficit;
		_totalSizeOfActiveBufferDeficit -= size;
	      }
	    else 
	      if (_small[size].suspendedBufferDeficit)
		{
		  --_small[size].suspendedBufferDeficit;
		}
	      else
		{
		  ulong remainingProficit = size;

		  // try to cover some active deficit in smaller sizes
		  
		  SmallSizeStatistics* otherUsed = _small[size].previousUsed;
		  while (otherUsed && remainingProficit)
		    {
		      while (otherUsed->size <= remainingProficit &&
			     otherUsed->activeBufferDeficit)
			{
			  --otherUsed->activeBufferDeficit;
			  _totalSizeOfActiveBufferDeficit -= otherUsed->size;
			  remainingProficit -= otherUsed->size;
			};
		      otherUsed = otherUsed->previousUsed;
		    };

		  // try to cover some suspended deficit in smaller sizes
		  otherUsed = _small[size].previousUsed;
		  while (otherUsed && remainingProficit)
		    {
		      while (otherUsed->size <= remainingProficit &&
			     otherUsed->suspendedBufferDeficit)
			{
			  ASSERT(!otherUsed->activeBufferDeficit);
			  --otherUsed->suspendedBufferDeficit;
			  remainingProficit -= otherUsed->size;
			};
		      otherUsed = otherUsed->previousUsed;
		    };
		  
		  if (remainingProficit)
		    {
		      ASSERT(!_small[remainingProficit].activeBufferDeficit);
		      ASSERT(!_small[remainingProficit].suspendedBufferDeficit);
		      ++_small[remainingProficit].bufferProficit;
		      makeUsed(remainingProficit);
		    };
		  
		};
	      
	  if (_sizeOfBufferProficitInLarge >= size)
	    {
	      _sizeOfBufferProficitInLarge -= size;
	    }
	  else
	    {
	      ulong freshDeficit = size - _sizeOfBufferProficitInLarge;
	      _sizeOfBufferProficitInLarge = 0UL;
	      _sizeOfActiveBufferDeficitInLarge += freshDeficit;
	      _totalSizeOfActiveBufferDeficit += freshDeficit;
	    };

	}; // if (_bufferIsBeingUsed)

      ++_small[size].free;
      _sizeOfFreeLarge -= size;
      ASSERT(checkIntegrity());
    }; // void largeFreeToSmallFree(ulong size)

    // debugging output

#ifndef NO_DEBUG
    bool checkIntegrity()
    {
      if (_small[0].used ||
	  _small[0].free ||
	  _small[0].allocated || 
	  _small[0].activeBufferDeficit ||
	  _small[0].suspendedBufferDeficit ||
	  _small[0].bufferProficit || 
	  _small[0].previousUsed || 
	  _small[0].nextUsed)
	{
	  DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 1\n");
	  return false;
	};
      if (_bufferIsBeingUsed)
	{
	  ulong allocated = _sizeOfAllocatedLarge;
	  ulong activeDeficit = _sizeOfActiveBufferDeficitInLarge;
	  bool someDeficit = false;
	  for (ulong size = 1UL; size <= MaxSmallObjectSize; ++size)
	    {
	      if (_small[size].used)
		{
		  if (someDeficit && _small[size].bufferProficit)
		    {
		      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 2\n");
		      return false;
		    };
		       
		  if (_small[size].bufferProficit &&
		      (_small[size].activeBufferDeficit ||
		       _small[size].suspendedBufferDeficit))
		    {
		      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 3\n");
		      return false;
		    };
		  if (_sizeOfBufferProficitInLarge >= size && 
		      (_small[size].activeBufferDeficit ||
		       _small[size].suspendedBufferDeficit))
		    {
		      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 4\n");
		      return false;
		    };
		  if (_small[size].previousUsed)
		    {
		      if (_small[size].previousUsed->nextUsed != _small + size)
			{
			  DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 5\n");
			  return false;
			};
		    }
		  else
		    if (_firstUsedSmallSize != _small + size)
		      {
			DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 6\n");
			return false;
		      };
		  if (_small[size].nextUsed)
		    {
		      if (_small[size].nextUsed->previousUsed != _small + size)
			{
			  DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 7\n");
			  return false;
			};
		    }
		  else
		    if (_lastUsedSmallSize != _small + size)
		      {
			DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 8\n");
			return false;
		      };
		  allocated += size * _small[size].allocated;
		  activeDeficit += size * _small[size].activeBufferDeficit;
		  someDeficit = 
		    someDeficit || 
		    _small[size].activeBufferDeficit ||
		    _small[size].suspendedBufferDeficit;
		}
	      else // !_small[size].used
		if (_small[size].free || 
		    _small[size].allocated ||
		    _small[size].activeBufferDeficit ||
		    _small[size].suspendedBufferDeficit ||
		    _small[size].bufferProficit ||
		    _small[size].previousUsed ||
		    _small[size].nextUsed)
		  {
		    DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 9\n");
		    return false;
		  };
	    }; // for (ulong size = 1UL; size <= MaxSmallObjectSize; ++size)

	  if (allocated != _totalSizeOfAllocated) 
	    {
	      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 10\n");
	      return false;
	    };
	  if (activeDeficit != _totalSizeOfActiveBufferDeficit) 
	    {
	      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 11\n");
	      return false;
	    };
	}
      else // !_bufferIsBeingUsed
	{     
	  if (_totalSizeOfActiveBufferDeficit) 
	    {
	      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 12\n");
	      return false;
	    };
	  ulong allocated = _sizeOfAllocatedLarge;
	  for (ulong size = 1UL; size <= MaxSmallObjectSize; ++size)
	    {
	      if (_small[size].used || 
		  _small[size].activeBufferDeficit || 
		  _small[size].suspendedBufferDeficit ||
		  _small[size].bufferProficit || 
		  _small[size].previousUsed ||
		  _small[size].nextUsed)
		{
		  DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 13\n");
		  return false;
		};
	      allocated += size * _small[size].allocated;
	    }; // for (ulong size = 1UL; size <= MaxSmallObjectSize; ++size)

	  if (allocated != _totalSizeOfAllocated) 
	    {
	      DMSG("BK::BufferedAllocationStatistics::checkIntegrity()> problem 14\n");
	      return false;
	    };
 
	};
      return true;
    }; // bool checkIntegrity()

    ostream& output(ostream& str,ulong size) const
    {
      str << "SIZE " << size << ":\n";
      str << "  FREE " << _small[size].free << "\n";
      str << "  ALLOCATED " << _small[size].allocated << "\n";
      str << "  ACTIVE BUFFER DEFICIT " << _small[size].activeBufferDeficit << "\n";
      str << "  SUSPENDED BUFFER DEFICIT " << _small[size].suspendedBufferDeficit << "\n";
      str << "  BUFFER PROFICIT " << _small[size].bufferProficit << "\n\n";
      return str;
    };
    ostream& output(ostream& str) const
    {
      str << "TOTAL SIZE OF ALLOCATED " << totalSizeOfAllocated() << "\n";
      str << "TOTAL SIZE OF ACTIVE BUFFER DEFICIT " 
	  << totalSizeOfActiveBufferDeficit() << "\n";
      str << "LARGE PIECES:\n";
      str << "  SIZE OF FREE " << _sizeOfFreeLarge << "\n";
      str << "  SIZE OF PERSISTENT " << _sizeOfAllocatedLarge << "\n";
      str << "  SIZE OF ACTIVE BUFFER DEFICIT " << _sizeOfActiveBufferDeficitInLarge << "\n";
      str << "  SIZE OF SUSPENDED BUFFER DEFICIT " << _sizeOfSuspendedBufferDeficitInLarge << "\n";    
      str << "  SIZE OF BUFFER PROFICIT " << _sizeOfBufferProficitInLarge << "\n\n";     
      return str;
    };
#endif

  private:

    class SmallSizeStatistics
    {
    public:
      SmallSizeStatistics() {};
      ~SmallSizeStatistics() {};
      void reset(ulong sz)
      {
	size = sz;
	used = false;
	free = 0L;
	allocated = 0L;
	activeBufferDeficit = 0L;
	suspendedBufferDeficit = 0L;
	bufferProficit = 0L;
	previousUsed = 0;
	nextUsed = 0;
      };
    public:
      ulong size;
      bool used;
      ulong free;
      ulong allocated;
      ulong activeBufferDeficit;
      ulong suspendedBufferDeficit;
      ulong bufferProficit;
      SmallSizeStatistics* previousUsed;
      SmallSizeStatistics* nextUsed;
    }; // class SmallSizeStatistics

  private:
    void startUsingBuffer()
    {
      CALL("startUsingBuffer()");
      ASSERT(!_bufferIsBeingUsed);
      _bufferIsBeingUsed = true;
      _sizeOfRemainingNonbufferMemory = 0UL;

      SmallSizeStatistics* previousUsedSmallSize = 0;
      for (ulong size = 0UL; size <= MaxSmallObjectSize; ++size)
	{
	  if (_small[size].free || _small[size].allocated)
	    {
	      _small[size].used = true;
	      _small[size].bufferProficit = _small[size].free;
	      if (!_firstUsedSmallSize) _firstUsedSmallSize = _small + size;
	      if (previousUsedSmallSize)
		previousUsedSmallSize->nextUsed = _small + size;
	      _small[size].previousUsed = previousUsedSmallSize;
	      previousUsedSmallSize = _small + size;
	    };
	};
      _lastUsedSmallSize = previousUsedSmallSize;
      
      _sizeOfBufferProficitInLarge = _sizeOfFreeLarge;
    }; // void startUsingBuffer()

    void makeUsed(ulong size)
    {
      CALL("makeUsed(ulong size)");
      ASSERT(size);
      ASSERT(size <= MaxSmallObjectSize);
      ASSERT(_bufferIsBeingUsed);

      if (!_small[size].used)
	{
	  _small[size].used = true;
	  if (!_firstUsedSmallSize)
	    {
	      ASSERT(!_lastUsedSmallSize);
	      _firstUsedSmallSize = _small + size;
	      _lastUsedSmallSize = _small + size;
	    };
	  if (_firstUsedSmallSize > _small + size)
	    {
	      _firstUsedSmallSize->previousUsed = _small + size;
	      _small[size].nextUsed = _firstUsedSmallSize;
	      _firstUsedSmallSize = _small + size;
	    }
	  else
	    if (_lastUsedSmallSize < _small + size)
	      {
		_lastUsedSmallSize->nextUsed = _small + size;
		_small[size].previousUsed = _lastUsedSmallSize;
		_lastUsedSmallSize = _small + size; 
	      }
	    else // somewhere between _firstUsedSmallSize and _lastUsedSmallSize
	      {
		SmallSizeStatistics* previous;
		SmallSizeStatistics* next = _firstUsedSmallSize;
		do
		  {
		    previous = next;
		    next = next->nextUsed;
		  }
		while (next < _small + size);
		ASSERT(previous < _small + size);
		ASSERT(next > _small + size); 
		ASSERT(previous->nextUsed == next);
		ASSERT(next->previousUsed == previous);
		previous->nextUsed = _small + size;
		_small[size].previousUsed = previous;
		next->previousUsed = _small + size;
		_small[size].nextUsed = next;
	      };
	}; // if (!_small[size].used)
    }; // void makeUsed(ulong size)

  private:
    // mode
    bool _bufferIsBeingUsed;
    ulong _sizeOfRemainingNonbufferMemory;
  
    // statistics on small sizes
    
    SmallSizeStatistics _small[MaxSmallObjectSize + 1];
    SmallSizeStatistics* _firstUsedSmallSize;
    SmallSizeStatistics* _lastUsedSmallSize;

    // statistics on large sizes
 
    ulong _sizeOfFreeLarge;
    ulong _sizeOfAllocatedLarge;
    ulong _sizeOfActiveBufferDeficitInLarge;
    ulong _sizeOfSuspendedBufferDeficitInLarge;
    ulong _sizeOfBufferProficitInLarge;
    
    // fast access to derived values
    ulong _totalSizeOfAllocated;
    ulong _totalSizeOfActiveBufferDeficit;
  }; // class BufferedAllocationStatistics<MaxSmallObjectSize>


}; // namespace BK



//=====================================================================
#endif
