//
// File:         ObjectCache.hpp
// Description:  Caching reusable objects. 
// Created:      Oct 15, 2003, Alexandre Riazanov
// Status:       draft
// mail:         riazanov@cs.man.ac.uk
//==========================================================================
#ifndef OBJECT_CACHE_H
#define OBJECT_CACHE_H
//==========================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "OnRequestMemoryRecycler.hpp"
//==========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_CACHE
#  define DEBUG_NAMESPACE "ObjectCache<Alloc,PageSize,C>"
#endif
#include "debugMacros.hpp"
//==========================================================================

namespace BK 
{
  template <class Alloc,ulong PageSize,class C>
  class ObjectCache : public OnRequestMemoryRecycler
  {
  public:
    ObjectCache();
    ~ObjectCache();
    void reset() { reclaimAllObjects(); };
    C* reserveObject();
    C* tryToReserveObject();
    void reclaimAllObjects();
  private:

    class Page
    {
    public:
      Page() : _next(0) {};
      ~Page() {};
      Page*& next() { return _next; };
      C* objects() { return _objects; };
      C* endOfObjects() { return _objects + PageSize; };
      void* operator new(size_t) 
      {
	CALL("Page::operator new(size_t)");
	return Alloc::allocate(sizeof(Page));
      };
      void* operator new(size_t,bool) 
      {
	CALL("Page::operator new(size_t,bool)");
	// may return 0
	return Alloc::tryToAllocate(sizeof(Page));
      };
      void operator delete(void* obj) 
      {
	CALL("Page::operator delete(void* obj)");
	Alloc::deallocate(obj,sizeof(Page));
      };
    private:
      Page* _next;
      C _objects[PageSize];
    }; // class ObjectCache<Alloc,PageSize,C>::Page
    
  private:
    bool recycleChunkOfUnusedMem(size_t minLength)
    {
      CALL("recycleChunkOfUnusedMem(size_t minLength)");
      if ((minLength > sizeof(Page)) ||
	  (!_currentPage->next())) 
	return false;
      Page* p = _currentPage->next();
      ASSERT(p != &_mainPage);
      _currentPage->next() = p->next();
      delete p;
      return true;
    };
      
    bool recycleSomeUnusedMem()
    {
      CALL("recycleSomeUnusedMem()");
      if (!_currentPage->next())
	return false;
      Page* p = _currentPage->next();
      ASSERT(p != &_mainPage);
      _currentPage->next() = p->next();
      delete p;
      return true;
    };

    bool recycleAllUnusedMem()
    {
      CALL("recycleAllUnusedMem()");
      Page* p = _currentPage->next();
      if (!p) return false;
      do
	{
	  ASSERT(p != &_mainPage);
	  Page* tmp = p->next();	  
	  delete p;
	  p = tmp;
	}
      while (p);
      return true;
    };
  private:
    Page _mainPage;
    Page* _currentPage;
    C* _nextAvailableObject;
    C* _endOfCurrentPage;
  }; // class ObjectCache<class Alloc,ulong PageSize,class C>
}; // namespace BK


//==========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_CACHE
#  define DEBUG_NAMESPACE "ObjectCache<Alloc,PageSize,C>"
#endif
#include "debugMacros.hpp"
//==========================================================================

namespace BK 
{
  template <class Alloc,ulong PageSize,class C>
  inline
  ObjectCache<Alloc,PageSize,C>::ObjectCache() :
    _currentPage(&_mainPage),
    _nextAvailableObject(_mainPage.objects()),
    _endOfCurrentPage(_mainPage.endOfObjects())
  {
    CALL("constructor ObjectCache()");
  };
  
  template <class Alloc,ulong PageSize,class C>
  inline
  ObjectCache<Alloc,PageSize,C>::~ObjectCache()
  {
    CALL("destructor ~ObjectCache()");
    if (DestructionMode::isThorough())
      {
	Page* p1 = _mainPage.next();
	while (p1)
	  {
	    Page* p2 = p1->next();
	    delete p1;
	    p1 = p2;
	  };
      };
  };
  
  
  template <class Alloc,ulong PageSize,class C>
  inline
  C* ObjectCache<Alloc,PageSize,C>::reserveObject()
  {
    CALL("reserveObject()");
    if (_nextAvailableObject == _endOfCurrentPage)
      {
	if (!_currentPage->next())
	  _currentPage->next() = new Page();
	_currentPage = _currentPage->next();
	_nextAvailableObject = _currentPage->objects();
	_endOfCurrentPage = _currentPage->endOfObjects();
      };
    C* res = _nextAvailableObject;
    ++_nextAvailableObject;
    return res;
  };
  
  
  template <class Alloc,ulong PageSize,class C>
  inline
  C* ObjectCache<Alloc,PageSize,C>::tryToReserveObject()
  {
    CALL("tryToReserveObject()");
    if (_nextAvailableObject == _endOfCurrentPage)
      {
	if (!_currentPage->next())
	  {
	    Page* newPage = new(true) Page();
	    if (!newPage) return static_cast<C*>(0);
	    _currentPage->next() = newPage;
	  };
	_currentPage = _currentPage->next();
	_nextAvailableObject = _currentPage->objects();
	_endOfCurrentPage = _currentPage->endOfObjects();
      };
    C* res = _nextAvailableObject;
    ++_nextAvailableObject;
    return res;
  };
  
  template <class Alloc,ulong PageSize,class C>
  inline
  void ObjectCache<Alloc,PageSize,C>::reclaimAllObjects()
  {
    CALL("reclaimAllObjects()");
    _currentPage = &_mainPage;
    _nextAvailableObject = _mainPage.objects();
    _endOfCurrentPage = _mainPage.endOfObjects();
  };
  

}; // namespace BK


//==========================================================================
#endif
