//
// File:         ObjectPoolWithMemory.hpp
// Description:  An implementation of the abstraction ObjectPool
//               parameterised by an allocator.
// Created:      Oct 15, 2003, Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//==========================================================================
#ifndef OBJECT_POOL_WITH_MEMORY_H
#define OBJECT_POOL_WITH_MEMORY_H
//==========================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
//==========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL_WITH_MEMORY
#  define DEBUG_NAMESPACE "ObjectPoolWithMemory<class Alloc,class C>"
#endif
#include "debugMacros.hpp"
//==========================================================================

namespace BK 
{

template <class Alloc,class C>
class ObjectPoolWithMemory : public ObjectPool<C>
{
public:
  ObjectPoolWithMemory() : _free(0) 
  {

  };
  ~ObjectPoolWithMemory();
  void prepareNObjects(ulong n)
  {
    CALL("prepareNObjects(ulong n)");
    OnRequestMemoryRecycler::AutoLock 
      memRecyclingLock(static_cast<OnRequestMemoryRecycler*>(this));
    while (n)
      {
	Entry* e = 
	  ::new (Alloc::allocate(sizeof(Entry))) Entry();
	e->nextFree() = _free;
	_free = e;
	--n;
      };
  };

private:
  typedef typename ObjectPool<C>::Entry Entry;
private:  
  Entry* reserveEntry()
  {
    CALL("reserveEntry()");
    OnRequestMemoryRecycler::AutoLock 
      memRecyclingLock(static_cast<OnRequestMemoryRecycler*>(this));
    Entry* res;
    if (_free)
      {
	res = _free;
	_free = _free->nextFree();
      }
    else
      {
	res = 
	  ::new (Alloc::allocate(sizeof(Entry))) Entry();
      };
    return res;
  };

  Entry* tryToReserveEntry()
  {
    CALL("reserveEntry()");
    OnRequestMemoryRecycler::AutoLock 
      memRecyclingLock(static_cast<OnRequestMemoryRecycler*>(this));
    Entry* res;
    if (_free)
      {
	res = _free;
	_free = _free->nextFree();
      }
    else
      {
	void* mem = Alloc::tryToAllocate(sizeof(Entry));
	if (!mem) return static_cast<Entry*>(0);
	res = ::new (mem) Entry();
      };
    return res;
  };

  

  void recycleEntry(Entry* e)
  {
    CALL("recycleEntry(Entry* e)");

    OnRequestMemoryRecycler::AutoLock 
      memRecyclingLock(static_cast<OnRequestMemoryRecycler*>(this));
    e->nextFree() = _free;
    _free = e;
  };

  bool recycleChunkOfUnusedMem(size_t minLength)
  {
    CALL("recycleChunkOfUnusedMem(size_t minLength)");
    if ((!_free) || (minLength > sizeof(Entry)))
      return false;
    Entry* e = _free;
    _free = _free->nextFree();
    e->~Entry();
    Alloc::deallocate(static_cast<void*>(e),sizeof(Entry));
    return true;
  };

  bool recycleSomeUnusedMem()
  {
    CALL("recycleSomeUnusedMem()");
    if (!_free) return false;
    Entry* e = _free;
    _free = _free->nextFree();
    e->~Entry();
    Alloc::deallocate(static_cast<void*>(e),sizeof(Entry));
    return true;
  };

  bool recycleAllUnusedMem()
  {
    CALL("recycleAllUnusedMem()");
    bool res = (_free != 0);
    while (_free)
      {
	Entry* e = _free;
	_free = e->nextFree();
	e->~Entry();
	Alloc::deallocate(static_cast<void*>(e),sizeof(Entry));
      };
    return res;
  };
  
private:
  Entry* _free;
}; // class ObjectPoolWithMemory<class Alloc,class C>

}; // namespace BK


//==========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL_WITH_MEMORY
#  define DEBUG_NAMESPACE "ObjectPoolWithMemory<class Alloc,class C>"
#endif
#include "debugMacros.hpp"
//==========================================================================

namespace BK 
{
  template <class Alloc,class C>
  inline
  ObjectPoolWithMemory<Alloc,C>::~ObjectPoolWithMemory()
  {
    CALL("destructor ~ObjectPoolWithMemory()");


    if (DestructionMode::isThorough())
      {
	OnRequestMemoryRecycler::AutoLock 
	  memRecyclingLock(static_cast<OnRequestMemoryRecycler*>(this));
	while (_free)
	  {
	    Entry* e = _free;
	    _free = e->nextFree();
	    e->~Entry();


	    Alloc::deallocate(static_cast<void*>(e),sizeof(Entry));
	  };

      };
  };
  
  
}; // namespace BK

//==========================================================================
#endif
