//
// File:         OnRequestMemoryRecycler.hpp
// Description:  Abstract class for on-request memory recycling discipline. 
// Created:      Oct 09, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk 
// Status:       Sketch.
//============================================================================
#ifndef ON_REQUEST_MEMORY_RECYCLER_H
#define ON_REQUEST_MEMORY_RECYCLER_H
//============================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ON_REQUEST_MEMORY_RECYCLER
#  define DEBUG_NAMESPACE "OnRequestMemoryRecycler"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace BK 
{
  class OnRequestMemoryRecycler
  {
  public:
    class AutoLock
    {
    public:
      AutoLock(OnRequestMemoryRecycler* r) : _recycler(r)
      {
	if (_recycler) _recycler->lockMemoryRecycling();
      };
      ~AutoLock()
      {
	if (_recycler) _recycler->unlockMemoryRecycling();
      };
    private:
      OnRequestMemoryRecycler* _recycler;
    };
  public:
    OnRequestMemoryRecycler() : _locked(0UL) {};
    virtual ~OnRequestMemoryRecycler() {};
    void lockMemoryRecycling() { ++_locked; };
    void unlockMemoryRecycling() { if (_locked) --_locked; };
    bool memoryRecyclingIsLocked() const { return _locked != 0UL; };
    ulong numberOfMemoryRecyclingLocks() const { return _locked; };
    bool recycleChunkOfUnusedMemory(size_t minLength)
    {
      CALL("recycleChunkOfUnusedMemory(size_t minLength)");
      if (memoryRecyclingIsLocked()) return false;
      AutoLock lock(this);
      return recycleChunkOfUnusedMem(minLength);
    };
    bool recycleSomeUnusedMemory()
    {
      CALL("recycleSomeUnusedMemory()");
      if (memoryRecyclingIsLocked()) return false;
      AutoLock lock(this);
      return recycleSomeUnusedMem();
    };
    bool recycleAllUnusedMemory()
    {
      CALL("recycleAllUnusedMemory()");
      if (memoryRecyclingIsLocked()) return false;
      AutoLock lock(this);
      return recycleAllUnusedMem();
    };
  protected:
    virtual bool recycleChunkOfUnusedMem(size_t minLength);
    virtual bool recycleSomeUnusedMem();
    virtual bool recycleAllUnusedMem();
  private:
    ulong _locked;
  }; // class OnRequestMemoryRecycler

}; // namespace BK

//============================================================================
#endif


