
/*
 * File Flattening.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Flattening.hpp"
#include "Unit.hpp"
#include "Position.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"



void Flattening::flatten (Unit& u)
{
  ASS(u.unitType() == FORMULA);

  Formula f(u.formula());
  Flattening flat;
  flat.flatten(f,Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_FLATTEN,UnitList(u));
    inf->setInfo(flat._info);
    u = Unit(f,inf);
  }
} 



void Flattening::flatten (Formula& f, const Position& pos)
{
  switch (f.connective()) {
  case ATOM:
    return;

  case AND:
  case OR: 
    {
      FormulaList flattenedArgs (f.args());
      if (flatten(flattenedArgs,f.connective(),pos,0)) {
	_info = new InferenceInfo(IR_FLATTEN,pos,_info);
      }
      else if (flattenedArgs == f.args()) { 
	return;
      }
      f = Formula (f.connective(), flattenedArgs);
      return;
    }

  case IMP:
  case IFF:
  case XOR: 
    {
      Formula flattenedLeft (f.left());
      flatten(flattenedLeft,Position(0,pos));
      Formula flattenedRight (f.right());
      flatten(flattenedRight,Position(1,pos));
      if (flattenedLeft == f.left() && flattenedRight == f.right()) {
	return;
      }
      f = Formula (f.connective(), flattenedLeft, flattenedRight);
      return;
    }
    
  case NOT: 
    {
      Formula flattenedArg (f.uarg());
      flatten(flattenedArg,Position(0,pos));
      if (flattenedArg == f.uarg()) {
	return;
      }
      f = Formula (NOT, flattenedArg);
      return;
    }
    
  case FORALL:
  case EXISTS: 
    {
      Formula flattenedArg (f.qarg());
      flatten(flattenedArg,Position(0,pos));
      if (flattenedArg.connective() == f.connective()) {
	
	VarList vs (f.vars());
	vs.append (flattenedArg.vars());
	f = Formula (f.connective(), vs, flattenedArg.qarg());
	_info = new InferenceInfo(IR_FLATTEN,pos,_info);
	return;
      }
      
      if (flattenedArg == f.qarg()) {
	return;
      }
      f = Formula (f.connective(), f.vars(), flattenedArg);
      return;
    }

  default:
    ASS( false );
    return;
  }
} 



bool Flattening::flatten (FormulaList& fs, 
			  Connective c, 
			  const Position& pos,
			  int k)
{
  if (fs.isEmpty()) {
    return false;
  }
  Formula flattenedHead (fs.head()); 
  flatten(flattenedHead,Position(k,pos));
  FormulaList flattenedTail(fs.tail());
  bool flattened = flatten(flattenedTail,c,pos,k+1);

  if (flattenedHead.connective() == c) { 
    fs = flattenedHead.args();
    fs.append(flattenedTail);
    return true;
  }
  
  if (flattenedHead == fs.head() && flattenedTail == fs.tail()) {
    return false;
  }
  fs = FormulaList (flattenedHead, flattenedTail);

  return flattened;
} 


