
/*
 * File FunctionDefinition.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "FunctionDefinition.hpp"
#include "Problem.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "VarCounter.hpp"
#include "TermVarIterator.hpp"
#include "TermFunIterator.hpp"
#include "Inference.hpp"
#include "Substitution.hpp"

#define UNTOUCHED 0
#define UNFOLDED 1
#define BLOCKED 2
#define REMOVED 3


struct FunctionDefinition::Def {
  
  Unit unit;
  
  int fun;
  
  Term lhs;
  
  Term rhs;
  
  int mark;
  
  Def () : mark(REMOVED) {}
#if DEBUG_PREPRO
  string toString () const
  {
    return lhs.toString() + " = " + rhs.toString() + "\n";
  }
#endif
}; 



FunctionDefinition::FunctionDefinition (Problem& problem,
					const Signature& signature)
  : _problem(problem),
    _numberOfFunctions(signature.noOfFuns()),
    _defs(new Def[_numberOfFunctions]),
    _counter (new int[_numberOfFunctions]),
    _removed(0)
{
  TRACER("FunctionDefinition::FunctionDefinition");

  if (! (_defs && _counter)) {
    ALLOCATION_FAILURE;
  }
  for (int i = signature.noOfFuns()-1; i >= 0; i--) {
    _counter[i] = 0;
  }
} 


FunctionDefinition::~FunctionDefinition ()
{
  TRACER("FunctionDefinition::~FunctionDefinition");
  delete [] _defs;
  delete [] _counter;
}


bool FunctionDefinition::isFunctionDefinition (const Clause& c)
{
  Term lhs;
  Term rhs;

  return isFunctionDefinition(c, lhs, rhs);
} 



bool FunctionDefinition::isFunctionDefinition (const Clause& c, 
					       Term& lhs, 
					       Term& rhs)
{
  TRACER("FunctionDefinition::isFunctionDefinition (const Clause& c...)");

  LiteralList lits (c.literals());

  if (! lits.hasLength(1)) {
    return false;
  }

  
  Literal l(lits.head());
  
  if (l.isNegative()) {
    return false;
  }

  return isFunctionDefinition(l.atom(), lhs, rhs);
} 



bool FunctionDefinition::isFunctionDefinition (const Atom& a, 
					       Term& lhs, 
					       Term& rhs)
{
  TRACER ( "FunctionDefinition::isFunctionDefinition(const Atom&...)" );

  if (! a.isEquality()) {
    return false;
  }

  
  Term l (a.args().head());
  Term r (a.args().second());
  if (defines(l,r)) {
    lhs = l;
    rhs = r;
    return true;
  }

  if (defines(r,l)) {
    lhs = r;
    rhs = l;
    return true;
  }
  
  return false;
} 



bool FunctionDefinition::defines (const Term& lhs, const Term& rhs)
{
  TRACER("FunctionDefinition::defines");
  if (lhs.tag() != COMPOUND) {
    return false;
  }
  if (occurs(lhs.functor(),rhs)) {
    return false;
  }
  if (lhs.args().isEmpty() && 
      rhs.tag() == COMPOUND &&
      rhs.args().isNonEmpty()) { 
    return false;
  }

  
  
  
  VarCounter counter;
  Iterator<Term> ls (lhs.args());
  while (ls.hasNext()) {
    Term v (ls.next());
    if (v.tag() != VAR) {
      return false;
    }
    Var w (v.var());
    if (counter.getCount(w) != 0) { 
      return false;
    }
    counter.incCount(w);
  }

  
  
  
  
  
  TermVarIterator vs(rhs);
  while (vs.hasNext()) {
    Var v = vs.next();
    switch (counter.getCount(v)) {
    case 0: 
      return false;

    case 1: 
      counter.incCount(v);
      break;

    default: 
      break;
    }
  }

  return true;
} 



bool FunctionDefinition::occurs (const FunctionSymbol& f, const Term& t)
{
  TRACER ("FunctionDefinition::occurs");

  TermFunIterator funs(t);
  while (funs.hasNext()) {
    if (f == funs.next()) {
      return true;
    }
  }
  return false;
} 



int FunctionDefinition::removeAllDefinitions ()
{
  TRACER("FunctionDefinition::removeAllDefinitions");

  
  
  UnitChain::DelIterator us(_problem.giveUnits());
  while (us.hasNext()) {
    Unit u (us.next());
    Clause c(u.clause());
    Term l;
    Term r;
    if (isFunctionDefinition(c,l,r)) {
      int n = l.functor().number();
      Def& def = _defs[n];
      if (def.mark != REMOVED) {
	
	count(c);
	continue;
      }
      _removed++;
      
      def.mark = UNTOUCHED;
      def.unit = u;
      def.fun = n;
      def.lhs = l;
      def.rhs = r;
      us.del();
    }
    else { 
      count(c);
    }
  }
  if (_removed == 0) {
    return 0;
  }

  
  
  while (! unfoldAllDefs()) {
  }

  
  UnitChain::DelIterator cs(_problem.giveUnits());
  while (cs.hasNext()) {
    Unit u(cs.next());
    LiteralList ls(u.clause().literals());
    UnitList parents;
    apply(ls,parents);
    if (ls != u.clause().literals()) { 
      Clause newClause(ls);
      parents.push(u);
      Unit newUnit(newClause,
		   new Inference(IR_DEFINITION_UNFOLDING,parents));
      cs.replace(newUnit);
    }
  }

  return _removed;
} 



void FunctionDefinition::count (const Clause& c)
{
  TRACER("FunctionDefinition::count");

  Iterator<Literal> lits(c.literals());
  while (lits.hasNext()) {
    TermFunIterator funs(lits.next());
    while (funs.hasNext()) {
      _counter[funs.next().number()]++;
    }
  }
} 


bool FunctionDefinition::unfoldAllDefs ()
{
  TRACER("FunctionDefinition::unfoldAllDefs");

  bool result = true;
  for (int i = _numberOfFunctions-1; i >= 0; i--) {
    Def& def = _defs[i];
    if(_counter[i] != 0) { 
      result = unfold(def) && result;
    }
  }
  return result;
} 


bool FunctionDefinition::unfold (Def& def)
{
  TRACER("FunctionDefinition::unfold (Def&...)");

  switch (def.mark) {
  case UNTOUCHED:
    {
      
      TermFunIterator funs(def.rhs);
      while (funs.hasNext()) {
	if (_defs[funs.next().number()].mark == BLOCKED) { 
	  
	  def.mark = REMOVED; 
	  
	  _problem.addUnit(def.unit);
	  
	  count(def.unit.clause());
	  return false;
	}
      }

      def.mark = BLOCKED;
      Term newRHS(def.rhs);
      UnitList parents; 
      bool result = unfold(newRHS,parents);
      if (def.rhs != newRHS) { 
	
	parents.push(def.unit);
	Atom newEquality(def.lhs,newRHS);
	Literal newLit(true,newEquality);
	LiteralList newList(newLit);
	Clause newClause(newList);
	Unit newUnit(newClause,
		     new Inference(IR_DEFINITION_UNFOLDING,parents));
	def.unit = newUnit;
	def.rhs = newRHS;
      }
      def.mark = UNFOLDED;
      return result;
    }

  case REMOVED:
  case UNFOLDED:
    return true;

#if DEBUG_PREPRO
  case BLOCKED:
    ASS(false);
#endif
  }
} 



bool FunctionDefinition::unfold (Term& t,UnitList& parents)
{
  TRACER("FunctionDefinition::unfold (Term&...)");

  switch (t.tag()) {
  case VAR:
  case NUMERIC:
    return true;
  case COMPOUND:
    break;
  }
  TermList newArgs(t.args());
  bool result = unfold(newArgs,parents);
  Def& def = _defs[t.functor().number()];
  
  result = unfold(def) && result;

  switch (def.mark) {
  case UNFOLDED: 
    {
      
      
      
      
      
      Substitution s;
      ASS(def.lhs.args().length() == newArgs.length());
      Iterator<Term> xs(def.lhs.args());
      Iterator<Term> ts(newArgs);
      while(xs.hasNext()) {
	s.bind(xs.next().var(),ts.next());
      }
      
      Term r(def.rhs);
      s.apply(r);
      t = r;
      if (! parents.member(def.unit)) {
	parents.push(def.unit);
      }
    }
    return result;

  case REMOVED:
    if (newArgs != t.args()) { 
      t = Term(t.functor(),newArgs);
    }
    return result;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool FunctionDefinition::unfold (TermList& ts,UnitList& parents)
{
  TRACER("FunctionDefinition::unfold (TermList&...)");

  if (ts.isEmpty()) {
    return true;
  }
  Term hd(ts.head());
  bool r1 = unfold(hd,parents);
  TermList tl(ts.tail());
  bool r2 = unfold(tl,parents);
  if (ts.head() != hd || ts.tail() != tl) {
    ts = TermList(hd,tl);
  }
  return r1 && r2;
} 



void FunctionDefinition::apply (Term& t,UnitList& parents)
{
  TRACER("FunctionDefinition::apply (Term&...)");

  switch (t.tag()) {
  case VAR:
  case NUMERIC:
    return;
  case COMPOUND:
    break;
  }
  TermList newArgs(t.args());
  apply(newArgs,parents);
  Def& def = _defs[t.functor().number()];

  switch (def.mark) {
  case UNFOLDED: 
    {
      
      
      
      
      
      Substitution s;
      ASS(def.lhs.args().length() == newArgs.length());
      Iterator<Term> xs(def.lhs.args());
      Iterator<Term> ts(newArgs);
      while(xs.hasNext()) {
	s.bind(xs.next().var(),ts.next());
      }
      
      Term r(def.rhs);
      s.apply(r);
      t = r;
      if (! parents.member(def.unit)) {
	parents.push(def.unit);
      }
    }
    return;

  case REMOVED:
    if (newArgs != t.args()) { 
      t = Term(t.functor(),newArgs);
    }
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void FunctionDefinition::apply (TermList& ts,UnitList& parents)
{
  TRACER("FunctionDefinition::apply (TermList&...)");

  if (ts.isEmpty()) {
    return;
  }
  Term hd(ts.head());
  apply(hd,parents);
  TermList tl(ts.tail());
  apply(tl,parents);
  if (ts.head() != hd || ts.tail() != tl) {
    ts = TermList(hd,tl);
  }
} 



void FunctionDefinition::apply (LiteralList& ts,UnitList& parents)
{
  TRACER("FunctionDefinition::apply (LiteralList&...)");

  if (ts.isEmpty()) {
    return;
  }
  Literal hd(ts.head());
  apply(hd,parents);
  LiteralList tl(ts.tail());
  apply(tl,parents);
  if (ts.head() != hd || ts.tail() != tl) {
    ts = LiteralList(hd,tl);
  }
} 



void FunctionDefinition::apply (Literal& l,UnitList& parents)
{
  TRACER("FunctionDefinition::apply (LiteralList&...)");

  Atom a(l.atom());
  TermList ts(a.args());
  apply(ts,parents);
  if (ts != a.args()) {
    l = Literal(l.sign(),Atom(a.functor(),ts));
  }
} 



bool FunctionDefinition::isFunctionDefinition (const Formula& g, 
					       Term& lhs, 
					       Term& rhs)
{
  TRACER ( "Definition::isFunctionDefinition" );

  Formula f(g);
  
  while (f.connective() == FORALL) {
    f = f.qarg();
  }

  return f.connective() == ATOM &&
         isFunctionDefinition(f.atom(), lhs, rhs);
} 


bool FunctionDefinition::isFunctionDefinition (const Formula& f)
{
  Term lhs;
  Term rhs;

  return isFunctionDefinition(f, lhs, rhs);
} 

