
/*
 * File Inference.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "XML.hpp"
#include "Tracer.hpp"



string Inference::toString (InferenceRule r)
{
  switch (r) {
  case IR_INPUT:
    return "input";
  case IR_CHOICE_AXIOM:
    return "choice axiom";
  case IR_MONOTONE_REPLACEMENT:
    return "monotone replacement";
  case IR_FORALL_ELIMINATION:
    return "forall-elimination";
  case IR_NOT_AND:
    return "not-and";
  case IR_NOT_OR:
    return "not-or";
  case IR_NOT_IMP:
    return "not-implies";
  case IR_NOT_IFF:
    return "not-iff";
  case IR_NOT_XOR:
    return "not-xor";
  case IR_NOT_NOT:
    return "not-not";
  case IR_NOT_FORALL:
    return "not-forall";
  case IR_NOT_EXISTS:
    return "not-exists";
  case IR_IMP_TO_OR:
    return "implies-to-or";
  case IR_IFF_TO_AND:
    return "equivalence-to-and";
  case IR_XOR_TO_AND:
    return "xor-to-and";
  case IR_RECTIFY:
    return "rectify";
  case IR_CLOSURE:
    return "universal closure";
  case IR_NORMALIZE:
    return "normalize";
  case IR_ENNF:
    return "ENNF transformation";
  case IR_NNF:
    return "NNF transformation";
  case IR_DUMMY_QUANTIFIER_REMOVAL:
    return "dummy quantifier removal";
  case IR_FORALL_AND:
    return "forall-and";
  case IR_EXISTS_OR:
    return "exists-or";
  case IR_QUANTIFIER_SWAP:
    return "quantifier swap";
  case IR_FORALL_OR:
    return "forall-or";
  case IR_EXISTS_AND:
    return "exists-and";

  case IR_PERMUT:
    return "permutation";
  case IR_FLATTEN:
    return "flattening";
  case IR_REORDER_EQ:
    return "equality reordering";
  case IR_HALF_EQUIV:
    return "equivalence to implication";
  case IR_MINISCOPE:
    return "miniscoping";
  case IR_SKOLEMIZE:
    return "skolemization";
  case IR_CLAUSIFY:
    return "cnf transformation";
  case IR_REMOVE_DUPLICATE_LITERALS:
    return "remove duplicate literals";
  case IR_KERNEL:
    return "kernel inference";
  case IR_RESOLUTION:
    return "resolution";
  case IR_EQUALITY_PROXY_REPLACEMENT:
    return "equality proxy replacement";
  case IR_EQUALITY_PROXY_AXIOM1:
    return "equality proxy axiom E(x,x)";
  case IR_EQUALITY_PROXY_AXIOM2:
    return "equality proxy axiom ~E(x,y) \\/ x=y";
  case IR_DEFINITION_UNFOLDING:
    return "definition unfolding";

#if DEBUG_PREPRO
  default:
    ASS(false);
    return 0;
#endif
  }
} 



Inference::Inference (InferenceRule r)
  : _rule (r),
    _info(0)
{
}


Inference::Inference (InferenceRule r, const Unit& premise)
  : _rule (r),
    _premises (premise),
    _info(0)
{
}


Inference::Inference (const UnitList& premises,
		      VampireKernel::Clause::BackgroundFlagIterator& flags)
  : _rule (IR_KERNEL),
    _premises (premises),
    _info(new KernelInferenceInfo(flags))
{
}



Inference::Inference (InferenceRule r, 
		      const Unit& premise, 
		      InferenceInfo* info)
  : _rule (r),
    _premises (premise),
    _info(info)
{
}


Inference::Inference (InferenceRule r, const UnitList& premises)
  : _rule (r),
    _premises (premises),
    _info(0)
{
}


Inference::~Inference ()
{
  TRACER("Inference::~Inference");

  
} 


XML::Element* Inference::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("inference");
  xml->addAttribute(new XML::TextAttribute("rule",
					   toString(_rule)));

  Iterator<Unit> ps(premises());
  while (ps.hasNext()) {
    Unit u (ps.next());
    XML::EmptyElement* premise = new XML::EmptyElement("premise");
    premise->addAttribute(new XML::LongAttribute("number",u.number()));
    if (u.numberInRefutation()) {
      premise->addAttribute(new XML::IntegerAttribute("refutation_number",
						     u.numberInRefutation()));
    }
    xml->addChild(premise);
  }
  if (info()) {
    XML::DeepElement* details = new XML::DeepElement("details");
    xml->addChild(details);
    for (InferenceInfo* inf = info(); inf; inf = inf->next()) {
      details->addChild(inf->toXML());
    }
  }

  return xml;
} 

