
/*
 * File Kernel.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <climits>
#include "Kernel.hpp"
#include "Problem.hpp"
#include "Tracer.hpp"
#include "Int.hpp"
#include "Timer.hpp"
#include "RenamingToNew.hpp"
#include "Signature.hpp"
#include "Options.hpp"
#include "GlobAlloc.hpp"



int Kernel::passClauses (const Problem& problem,
			 const Options& opts)
{
  TRACER("Kernel::passClauses");

  _kernel.beginInputClauses();
  int lastClauseNumber = 0;

  UnitChain::Iterator units (problem.units());

  while (units.hasNext()) {
    
    Unit u (units.next());
    ASS(u.unitType() == CLAUSE);
    RenamingToNew renaming;

    
    int number = u.number();
    if (number > lastClauseNumber) {
      lastClauseNumber = number;
    }
    
    
    _kernel.beginInputClause (number, 
			     u.giveAway(),
			     u.inputType() != AXIOM,
			     opts.sos() && u.inputType() == AXIOM);

    
    Iterator<Literal> lits (u.clause().literals());
    while (lits.hasNext()) {
      Literal lit (lits.next());
      _kernel.beginLiteral(lit.isPositive(),
			  lit.atom().functor().kernelNumber());

      Iterator<Term> args (lit.atom().args());
      while ( args.hasNext() ) {
	passTerm(args.next(), renaming);
      }

      _kernel.endLiteral();		  
    }

    _kernel.endInputClause();
  }

  _kernel.endInputClauses();

  return lastClauseNumber;
} 



void Kernel::passTerm (const Term& t, 
		       RenamingToNew& renaming)
{
  TRACER("Kernel::passTerm");

  switch (t.tag()) 
    {
    case VAR: 
      _kernel.receiveVariable(renaming.findOrBindToNew(t.var()).toInt());
      return;

    case COMPOUND: 
      {
	
	
	_kernel.receiveFunction (t.functor().kernelNumber());
	Iterator<Term> ts (t.args());
	while (ts.hasNext()) {
	  passTerm(ts.next(), renaming);
	}
	return;
      }
    case NUMERIC:
      _kernel.receiveNumericConstant(t.number());
      return;
    }
} 



void Kernel::passOptions (VampireKernel::Options& kernel,
			  const Options& vampire)
{
  
  TRACER("Kernel::passOptions");

  kernel.age_weight_ratio_numerator = vampire.ageWeightRatio().age;
  kernel.age_weight_ratio_denominator = vampire.ageWeightRatio().weight;
  kernel.allocation_buffer_price = vampire.allocationBufferPrice ();
  kernel.allocation_buffer_size = vampire.allocationBufferSize ();
  

  switch (vampire.backwardDemodulation()) 
    {
    case Options::DEMODULATION_OFF:
      kernel.backward_demodulation = 0;
      break;
    case Options::DEMODULATION_PREORDERED:
      kernel.backward_demodulation = 1;
      break;
    case Options::DEMODULATION_ALL:
      kernel.backward_demodulation = 4;
      break;
    }
  switch (vampire.backwardSubsumption()) 
    {
    case Options::SUBSUMPTION_OFF:
      kernel.backward_subsumption = false;
      break;
    case Options::SUBSUMPTION_MULTISET:
      kernel.backward_subsumption = true;
      kernel.bs_set_mode = false;
      break;
    case Options::SUBSUMPTION_SET:
      kernel.backward_subsumption = true;
      kernel.bs_set_mode = true;
      break;
    }
  if (vampire.basicStrategy()) {
    kernel.prohibit_non_basic_superposition = true;
    kernel.prohibit_non_basic_fd = true;
    kernel.prohibit_non_basic_bd = true;
    kernel.reduce_non_basic = true;
    kernel.input_non_basicness = true;
  }
  else {
    kernel.prohibit_non_basic_superposition = false;
    kernel.prohibit_non_basic_fd = false;
    kernel.prohibit_non_basic_bd = false;
    kernel.reduce_non_basic = false;
    kernel.input_non_basicness = false;
  }
  kernel.bd_on_splitting_branches = vampire.bdOnSplittingBranches ();

  kernel.elim_def = vampire.elimDef ();
  kernel.simplify_by_eq_res = vampire.equalityResolutionWithDeletion();

  kernel.normalize_by_commutativity = vampire.fastDemodulationByCommutativity();
  kernel.fd_on_splitting_branches = vampire.fdOnSplittingBranches ();
  switch (vampire.forwardDemodulation()) 
    {
    case Options::DEMODULATION_OFF:
      kernel.forward_demodulation = 0;
      break;
    case Options::DEMODULATION_PREORDERED:
      kernel.forward_demodulation = 1;
      break;
    case Options::DEMODULATION_ALL:
      kernel.forward_demodulation = 4;
      break;
    }
  switch (vampire.forwardSubsumption ()) 
    {
    case Options::SUBSUMPTION_OFF:
      kernel.forward_subsumption = false;
      break;
    case Options::SUBSUMPTION_MULTISET:
      kernel.forward_subsumption = true;
      kernel.fs_set_mode = false;
      break;
    case Options::SUBSUMPTION_SET:
      kernel.forward_subsumption = true;
      kernel.fs_set_mode = true;
      break;
    }
  kernel.simplify_by_fs = vampire.forwardSubsumptionResolution ();
  kernel.fs_use_sig_filters = vampire.forwardSubsumptionSignatureFilters();

  kernel.static_neg_eq_splitting = vampire.groundInequalitySplitting ();

  
  kernel.inherited_negative_selection = vampire.inheritedNegativeSelection ();
  
  

  switch (vampire.literalComparisonMode()) 
    {
    case Options::NORMAL:
      kernel.literal_comparison_mode = 0;
      kernel.header_precedence_kinky = false;
      break;
    case Options::KINKY:
      kernel.literal_comparison_mode = 2;
      kernel.header_precedence_kinky = true;
      break;
    case Options::PREDICATE:
      kernel.literal_comparison_mode = 2;
      kernel.header_precedence_kinky = false;
      break;
    }
  
  kernel.first_mem_check = vampire.lrsFirstMemoryCheck();
  kernel.first_time_check = vampire.lrsFirstTimeCheck();
  kernel.lrs_softness = vampire.lrsSoftness();

  switch (vampire.mainAlg()) 
    {
    case Options::LRS:
      kernel.main_alg = 0;
      kernel.lrs = 1;
      break;
    case Options::OTTER:
      kernel.main_alg = 0;
      kernel.lrs = 0;
      break;
    case Options::DISCOUNT:
      kernel.main_alg = 3;
      kernel.lrs = 0;
      break;
    }
  kernel.main_selected_lit_demodulation = vampire.mainPremiseDemodulation();
  kernel.main_premise_subsumption = vampire.mainPremiseSubsumption();
  int max = vampire.maxActive();
  kernel.max_active =  max ? max : LONG_MAX;
  max = vampire.maxInferenceDepth();
  kernel.max_inference_depth = max ? max : LONG_MAX;
  max = vampire.maxPassive();
  kernel.max_passive = max ? max : LONG_MAX;
  kernel.max_skolem = vampire.maxSkolem();
  max = vampire.maxWeight();
  kernel.max_weight = max ? max : LONG_MAX;
  kernel.memory_limit = vampire.memoryLimit() * 1000 - 
                        GlobalAllocator::getUsedMemoryInKilobytes();

  
  

  
  
  kernel.no_inferences_between_rules= vampire.noInferencesBetweenRules();
  kernel.nongoal_penalty_coefficient = vampire.nongoalWeightCoefficient();
  

  kernel.orphan_murder = vampire.orphanMurder();
  
  
  

  kernel.paramodulation = vampire.paramodulation ();
  kernel.paramodulate_from_variables = vampire.paramodulationFromVariables ();
  
  kernel.proof = false;

  

  kernel.selected_pos_eq_penalty_coefficient = vampire.selectedEqualityWeightCoefficient();    
  kernel.selection = vampire.selection();
  kernel.show_active = vampire.showActive();
  kernel.show_gen = vampire.showGenerated();
  kernel.show_input = vampire.showInput();
  kernel.show_opt = vampire.showOptions();
  kernel.show_passive = vampire.showPassive();
  kernel.show_prepro = vampire.showPreprocessed();
  kernel.show_profile = vampire.showProfile();
  kernel.show_weight_limit_changes = vampire.showWeightLimitChanges();
  kernel.side_selected_lit_demodulation = vampire.sidePremiseDemodulation();
  kernel.side_premise_subsumption = vampire.sidePremiseSubsumption();
  switch (vampire.simplificationOrdering()) 
    {
    case Options::KBO:
      kernel.simplification_ordering = VampireKernel::Options::StandardKBOrdering;
      break;
    case Options::KBO_NONREC:
      kernel.simplification_ordering = VampireKernel::Options::NonrecursiveKBOrdering;
      break;
    }
  kernel.sim_back_sup = vampire.simultaneousBackwardSuperposition();
  
  switch (vampire.splitting ()) 
    {
    case Options::SPLIT_OFF:
      kernel.static_splitting = false;
      kernel.dynamic_splitting = false;
      break;

    case Options::SPLIT_PREPRO_ONLY:
      kernel.static_splitting = true;
      kernel.dynamic_splitting = false;
      break;

    case Options::SPLIT_ON:
      kernel.static_splitting = true;
      kernel.dynamic_splitting = true;
      break;
    }
  switch (vampire.splittingStyle ()) 
    {
    case Options::BLOCKING:
      kernel.splitting_style = 2;
      break;
    case Options::PARALLEL:
      kernel.splitting_style = 1;
      break;
    }

  
  
  

  
  kernel.term_weighting_scheme = VampireKernel::Options::NonuniformConstantTermWeighting;

  
  string jobId = Int::toString(vampire.testNumber());
  kernel.job_id = jobId.c_str();
  
  switch(vampire.transitivityChaining()) 
    {
    case Options::TC_NONE:
      kernel.chaining_for_transitive = false;
      break;
    case Options::TC_ONLY_NONVARIABLES:
      kernel.chaining_for_transitive = true;
      kernel.chaining_through_variables = false;
      break;
    case Options::TC_FULL:
      kernel.chaining_for_transitive = true;
      kernel.chaining_through_variables = true;
      break;
    }

  

  
  kernel.tabulation = 0;
  
  kernel.std_output = 0;



} 


void Kernel::passSignature (Signature& sig)
{
  TRACER ("Kernel::passSignature");

  if (Options::globalOptions.mode() == Options::KIF) {
    sig.addArithmetic (); 
  }

  _kernel.beginInputSignature();
  
  
  Signature::PredIterator preds (sig);
  while (preds.hasNext()) {
    PredicateSymbol& pred = static_cast<PredicateSymbol&>(preds.next());
    ulong symbolId; 
    _kernel.registerInputPredicate(pred.name(),
				  pred.arity(),
				  sig.isAnswerPredicate(pred),
				  symbolId);
    sig.registerKernelNumber(pred,symbolId);
    if (pred.isEquality()) {
      continue;
    }
    _kernel.setSymbolWeightConstPart(symbolId,pred.weight());
    _kernel.setSymbolPrecedence(symbolId,pred.precedence()); 
    if (pred.negativePrecedence()) {
      
      _kernel.setHeaderPrecedence(true,symbolId,pred.precedence()+1);
      _kernel.setHeaderPrecedence(false,symbolId,pred.negativePrecedence()+1);
    }
  }
  
  Signature::FunIterator funs (sig);
  while (funs.hasNext()) {
    ulong symbolId; 
    Symbol& fun = funs.next();
    _kernel.registerInputFunction(fun.name(),
				 fun.arity(),
				 symbolId);
    sig.registerKernelNumber(fun,symbolId);
  }

  _kernel.endInputSignature();

  if (Options::globalOptions.mode() == Options::KIF) {
    
    _kernel.interpretFunction ("AdditionFn", 2, 
			      "floating_point_arithmetic", "binary_plus");
    _kernel.interpretFunction ("SubtractionFn", 2,
			      "floating_point_arithmetic", "binary_minus");
    
    _kernel.interpretFunction ("MultiplicationFn", 2,
			      "floating_point_arithmetic", "multiply");
    _kernel.interpretFunction ("DivisionFn", 2,
			      "floating_point_arithmetic", "divide");
    
    
    _kernel.interpretPredicate (">", 2,
			       "floating_point_arithmetic", "greater");
    _kernel.interpretPredicate (">=", 2,
			       "floating_point_arithmetic", "greater_or_equal");
    _kernel.interpretPredicate ("<=", 2,
			       "floating_point_arithmetic", "less_or_equal");
    _kernel.interpretPredicate ("<", 2,
			       "floating_point_arithmetic", "less");
    _kernel.interpretPredicate ("equal", 2,
			       "floating_point_arithmetic", "equal");
  }
} 



void Kernel::getClause (const VampireKernel::Clause* clause, 
			Clause& result) const
{
  TRACER("Kernel::getClause");

  LiteralList literals;
  getLiteralList(clause->literals(), literals);
  result = Clause(literals);
} 



void Kernel::getLiteral (const VampireKernel::Literal* literal, 
			 Literal& result) const
{
  TRACER("Kernel::getLiteral");

  TermList ts;
  getTermList(literal->firstArgument(), ts);
  ulong sym = literal->predicate();
  Atom a (Signature::sig.getPredicateByItsKernelNumber(sym,
						       _kernel.symbolArity(sym)),
	  ts);

  result = Literal(literal->isPositive(),a);
} 



void Kernel::getLiteralList (const VampireKernel::Literal* literal,
			     LiteralList& result) const
{
  TRACER("Kernel::getLiteralList");

  if (! literal) {
    return;
  }

  Literal l;
  getLiteral(literal,l);
  LiteralList ls;
  getLiteralList(literal->next(),ls);
  result = LiteralList(l,ls);
} 



void Kernel::getTerm (const VampireKernel::Subterm* term, 
		      Term& result) const
{
  TRACER("Kernel::getTerm");
  
  if (term->isVariable()) {
    result = Term(Var(term->variableNumber()+1));
    return;
  }

  
  ulong symbol = term->topFunction();
  if (_kernel.isNumericConstantId(symbol)) {
    result = Term(_kernel.numericConstantValue(symbol));
    return;
  }

  TermList ts;
  getTermList(term->firstArgument(),ts);
  result = Term(Signature::sig.getFunctionByItsKernelNumber(symbol,
							    _kernel.symbolArity (symbol)),
		ts);
} 



void Kernel::getTermList (const VampireKernel::Subterm* terms, 
			  TermList& result) const
{
  TRACER("Kernel::getTermList");

  if (! terms) {
    return;
  }

  Term t;
  getTerm(terms,t);
  TermList ts;
  getTermList(terms->next(),ts);

  result = TermList(t,ts);
} 



RunResult Kernel::rerun ()
{
  TRACER("rerun");

  try 
    {
      _kernel.lookForAnotherAnswer();
    }
  catch (const VampireKernel::Emergency& e)
    {
      switch (e.type())
	{
	case VampireKernel::Emergency::AllocationFailure:
	  
	  return UNKNOWN;
	case VampireKernel::Emergency::TimeLimit:
	  
	  return UNKNOWN;
	};
    };
  

  RunResult result = UNKNOWN;
  switch (_kernel.terminationReason())
    {
    case VampireKernel::UnknownTerminationReason:
      ICP("ICP0"); 
      BK::RuntimeError::report("Kernel session terminated for unknown reason");
    case VampireKernel::RefutationFound:
      result = REFUTATION;
      break;
    case VampireKernel::Satisfiable:
      result = Options::globalOptions.sos() ? UNKNOWN : SATISFIABLE;
      break;
    case VampireKernel::EmptyPassive:
      break;
    case VampireKernel::MaxPassive:
      break;
    case VampireKernel::MaxActive:
      break;
    #ifdef DEBUG_PREPRO
    default:
      ASS (false);
    #endif
    };
  
  return result;
} 



RunResult Kernel::run (long timeLimitInDeciseconds,
		       VampireKernel::Options& kernelOptions, 
		       Problem& problem,
		       const Options& options,
		       Timer& timer,
		       XML::DeepElement* log)
{
  TRACER ("Kernel::run");

  bool weightIncrement = options.weightIncrement();
  int weightLimit = 10;

  long initialTime = timer.elapsedDeciseconds();

  for (;;) {
    bool emergency = false;
    try 
      {
	
	
	BK::GlobAlloc::reset ();
	
	if (weightIncrement) {
	  kernelOptions.max_weight = weightLimit;
	}
	_kernel.openSession(kernelOptions);
	
	passSignature(Signature::sig);
	
	
	int lastClauseNumber = passClauses(problem, options);






 	_kernel.saturate(lastClauseNumber+1,
 			 timeLimitInDeciseconds + initialTime - timer.elapsedDeciseconds());
      }
    catch (const VampireKernel::Emergency& e)
      {
	emergency = true;
	switch (e.type())
	  {
	  case VampireKernel::Emergency::AllocationFailure:
	    
	    break;

	  case VampireKernel::Emergency::TimeLimit:
	    
	    return UNKNOWN;
	  };
      };
    

    if (! emergency) {
      switch (_kernel.terminationReason()) {
      case VampireKernel::RefutationFound:
	return REFUTATION;

      case VampireKernel::Satisfiable:
	return options.sos() ? UNKNOWN :SATISFIABLE;

      default:
	break;
      }
    }
    if (weightIncrement) {
      BK::DestructionMode::makeFast();
      _kernel.closeSession();
      BK::DestructionMode::makeThorough();
      weightLimit = (weightLimit * 6) / 5;

    }
    else {
      return UNKNOWN;
    }
  }
} 
