
/*
 * File LaTeX.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Inference.hpp"
#include "Int.hpp"
#include "LaTeX.hpp"
#include "Options.hpp"
#include "Refutation.hpp"
#include "Unit.hpp"


string LaTeX::toLaTeX (const Refutation& ref)
{
  string result = "\\documentclass[fleqn]{article}\n"
                  "\\usepackage{fullpage,latexsym}\n"
                  "\\input{VampireProofMacros}\n"
                  "\\begin{document}\n"
                  "\\begin{VampireProof}\n";

  Iterator<Unit> units(ref.refutation());
  while (units.hasNext()) {
    result += toLaTeXAsInference(units.next());
  }

  return result + "\\end{VampireProof}\n"
	          "\\end{document}\n";
} 



string LaTeX::toLaTeX (const Formula& f)
{
  static string names [] =
    { "", " \\Vand ", " \\Vor ", " \\Vimp ", " \\Viff ", " \\Vxor ", 
      "\\neg ", "\\forall ", "\\exists "};

  Connective c = f.connective();
  string con = names[(int)c];
  switch (c) {
  case ATOM:
    return toLaTeX(f.atom());

  case AND:
  case OR: 
    {
      string result = toLaTeX(f.args().head());
      Iterator<Formula> arg (f.args().tail());
      while (arg.hasNext()) {
	result += con + toLaTeX(arg.next(),c);
      }
      return result;
    }

  case IMP:
  case IFF:
  case XOR:
    return toLaTeX(f.left(),c) + con + toLaTeX(f.right(),c);

  case NOT:
    return con + toLaTeX(f.uarg(),c);

  case FORALL:
  case EXISTS:
    {
      string result = string("(") + con;
      Iterator<Var> vs(f.vars());
      while (vs.hasNext()) {
	result += string(" ") + toLaTeX(vs.next());
      }
      result += ")";
      return result + toLaTeX(f.qarg(),c);
    }
#if DEBUG_PREPRO
  default:
    ASS (false);
#endif
  }
} 


string LaTeX::toLaTeX (const Formula& f, Connective outer)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toLaTeX(f) + ")" :
         toLaTeX(f);
} 



string LaTeX::toLaTeX (const Clause& c)
{
  if (c.isEmpty()) {
    return "\\VEmptyClause";
  }

  string result = toLaTeX(c.literals().head());

  Iterator<Literal> lits (c.literals().tail());
  while (lits.hasNext()) {
    result += string(" \\Vor ") + toLaTeX(lits.next());
  }
  return result;
} 



string LaTeX::toLaTeX (const Literal& l)
{
  if (l.isPositive()) {
    return toLaTeX(l.atom());
  }
  Atom a (l.atom());
  if (a.isEquality()) {
    TermList ts (a.args());
    return toLaTeX(ts.head()) + " \\neq " + toLaTeX(ts.second());
  }
  return string("\\neg") + toLaTeX(a);
} 



string LaTeX::toLaTeX (const Symbol& s)
{
  const string& name = s.name();
  if (name == "$`") { 
    return "`";
  }
  if (name == "$qvars") { 
    return "";
  }
  if (name == "$answer") { 
    return "\\mathit{answer}=";
  }

  return string("\\mathit{") + name + '}';
} 



string LaTeX::toLaTeX (const Term& t)
{
  switch (t.tag()) {
  case VAR:
    return toLaTeX(t.var());

  case COMPOUND:
    return toLaTeX(t.functor()) + toLaTeX(t.args());

  case NUMERIC:
    return Int::toString(t.number());

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



string LaTeX::toLaTeX (const TermList& terms)
{
  if (terms.isEmpty()) {
    return "";
  }
  string result = string("(") + toLaTeX(terms.head());
  Iterator<Term> ts (terms.tail());
  while (ts.hasNext()) {
    result += ",";
    result += toLaTeX(ts.next());
  }
  return result + ")";
} 



string LaTeX::toLaTeX (const Atom& a)
{
  TermList ts(a.args());
  if (a.isEquality()) {
    return toLaTeX(ts.head()) + "=" + toLaTeX(ts.second());
  }
  return toLaTeX(a.functor()) + toLaTeX(ts);
} 


string LaTeX::toLaTeX (const Unit& u)
{
  string prefix = Int::toString(u.number()) + ". ";
  string postfix = string("(") + Unit::toString(u.inputType());
  if (u.name() != "") {
    postfix += string(",name: ") + u.name();
  }
  postfix += ")";

  switch (u.unitType()) {
  case CLAUSE:
    return prefix + toLaTeX(u.clause()) + postfix;
  case FORMULA:
    return prefix + toLaTeX(u.formula()) + postfix;
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



string LaTeX::toLaTeXAsInference (const Unit& u)
{
  UnitList parents (u.inference()->premises());
  string result = "\\noindent\\hrulefill~$";

  if (parents.isNonEmpty()) {
    result += "(";
    Iterator<Unit> ps (parents);
    while (ps.hasNext()) {
      result += Int::toString(ps.next().number());
      if (ps.hasNext()) {
	result += ',';
      }
    }
    result += ")\\rightarrow ";
  }
  result += Int::toString(u.number()) + "$~\\noindent\\hrulefill\\[\n";

  
  result += "\\begin{VampireInference}\n";

  if (parents.isNonEmpty()) {
    Iterator<Unit> ps (parents);
    while (ps.hasNext()) {
      Unit premise (ps.next());
      result += "\\begin{VampirePremise}%\n~~";
      switch (premise.unitType()) {
      case CLAUSE:
	result += toLaTeX(premise.clause());
	break;
      case FORMULA:
	result += toLaTeX(premise.formula());
	break;
      }
      result += "\\end{VampirePremise}\n";
      if (ps.hasNext()) {
	result += "\\VPremiseSeparator\n";
      }
    }
    result += "\\VConclusionSeparator\n";
  }

  
  result += "\\begin{VampireConclusion}\n~~";
  switch (u.unitType()) {
  case CLAUSE:
    result += toLaTeX(u.clause());
    break;
  case FORMULA:
    result += toLaTeX(u.formula());
    break;
  }

  return result + "\\end{VampireConclusion} \\end{VampireInference}\n\\]\n";
} 



string LaTeX::toLaTeX (Var v)
{
  return string("x_{") + Int::toString(v.toInt()) + "}";
} 



void LaTeX::output (const Refutation& ref,const Options& options)
{
  string fileName = options.latexOutput();
  if (fileName == "off") {
    return;
  }
  string refutation = toLaTeX(ref);
  if (fileName == "on") {
    cout << refutation;
    return;
  }
  ofstream stream(fileName.c_str());
  if (stream) {
    stream << refutation;
  }
  else {
    cerr << "Cannot open file " << fileName << " for LaTeX output\n";
  }
} 

