
/*
 * File LiteralOrderingFinder.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "LiteralOrderingFinder.hpp"
#include "Problem.hpp"
#include "Options.hpp"
#include "Signature.hpp"
#include "Symbol.hpp"
#include "Sort.hpp"
#include "Int.hpp"
#include "Tracer.hpp"
#include "Options.hpp"


class LiteralOrderingFinder::Counter
{
public:
  Counter();
  
  int occurrencesInAxioms;
  
  int occurrencesInConjecture;
  
  int occurrencesInAssumptions;
  
  int positiveOccurrences;
  
  int negativeOccurrences;
  
  int neutralOccurrences;
  
  Symbol* symbol;
  
  int number;
}; 


class LiteralOrderingFinder::ArityComparator 
{
public:
  ArityComparator (int byArity, const Counter* counters);
  bool lessThan(const Symbol*,const Symbol*);
private:
  
  int _byArity;
  
  const Counter* _counters;
};



LiteralOrderingFinder::LiteralOrderingFinder (const Problem& problem, 
					      const Options& options, 
					      Signature& signature)
  : _problem(problem),
    _options(options),
    _signature(signature),
    _lastSymbol(0),
    _functions(new Counter[signature.noOfFuns()]),
    _predicates(new Counter[signature.noOfPreds()])
{
  if (! _functions || ! _predicates) {
    ALLOCATION_FAILURE;
  }
}


LiteralOrderingFinder::~LiteralOrderingFinder ()
{
  TRACER("LiteralOrderingFinder::~LiteralOrderingFinder");

  delete [] _predicates;
  delete [] _functions;
} 



void LiteralOrderingFinder::registerOccurrences ()
{
  TRACER("LiteralOrderingFinder::registerOccurrences");

  Chain<Unit>::Iterator units(_problem.units());
  while (units.hasNext()) {
    registerOccurrences(units.next());
  }
} 



void LiteralOrderingFinder::registerOccurrences (const Unit& unit)
{
  TRACER("LiteralOrderingFinder::registerOccurrences (const Unit& unit)");

  switch (unit.unitType()) {
  case CLAUSE:
    registerOccurrences(unit.clause(),unit.inputType());
    return;

  case FORMULA:
    registerOccurrences(unit.formula(),unit.inputType(),1);
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void LiteralOrderingFinder::registerOccurrences (const Formula& formula,
						 InputType it,
						 int polarity)
{
  TRACER("LiteralOrderingFinder::registerOccurrences (const Formula&...");

  switch (formula.connective()) {
  case ATOM:
    registerOccurrences(formula.atom(),it,polarity);
    return;

  case AND:
  case OR:
    {
      Iterator<Formula> fs(formula.args());
      while (fs.hasNext()) {
	registerOccurrences(fs.next(),it,polarity);
      }
    }
    return;

  case IMP:
    registerOccurrences(formula.left(),it,-polarity);
    registerOccurrences(formula.right(),it,polarity);
    return;

  case IFF:
  case XOR:
    registerOccurrences(formula.left(),it,0);
    registerOccurrences(formula.right(),it,0);
    return;

  case NOT:
    registerOccurrences(formula.uarg(),it,-polarity);
    return;

  case FORALL:
  case EXISTS:
    registerOccurrences(formula.qarg(),it,polarity);
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void LiteralOrderingFinder::registerOccurrences (const Atom& atom,
						 InputType it,
						 int polarity)
{
  TRACER("LiteralOrderingFinder::registerOccurrences (const Atom&...");

  registerOccurrence(const_cast<PredicateSymbol&>(atom.functor()),it,polarity);

  Iterator<Term> ts(atom.args());
  while (ts.hasNext()) {
    registerOccurrences(ts.next(),it,polarity);
  }
} 



void LiteralOrderingFinder::registerOccurrences (const Term& t,
						 InputType it,
						 int polarity)
{
  TRACER("LiteralOrderingFinder::registerOccurrences (const Term&");

  switch (t.tag()) {
  case VAR:
  case NUMERIC:
    return;

  case COMPOUND:
    {
      registerOccurrence(const_cast<FunctionSymbol&>(t.functor()),it,polarity);
      Iterator<Term> ts(t.args());
      while (ts.hasNext()) {
	registerOccurrences(ts.next(),it,polarity);
      }
    }
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void LiteralOrderingFinder::registerOccurrences (const Clause& c,
						 InputType it)
{
  TRACER("LiteralOrderingFinder::registerOccurrences (const Clause&");

  Iterator<Literal> ls(c.literals());
  while (ls.hasNext()) {
    const Literal& lit = ls.next();
    registerOccurrences(lit.atom(),it,lit.sign() ? 1 : -1);
  }
} 



void LiteralOrderingFinder::registerOccurrence (Symbol& s, 
						InputType it, 
						int polarity)
{
  TRACER("LiteralOrderingFinder::registerOccurrence (Symbol&");

  Counter& c = s.isFunction() ? 
    _functions[s.number()] :
    _predicates[s.number()];

  c.symbol = &s;
  if (c.number == -1) {
    
      c.number = ++_lastSymbol;
  }

  switch (polarity) {
  case -1:
    c.negativeOccurrences++;
    break;
  case 0:
    c.neutralOccurrences++;
    break;
  case 1:
    c.positiveOccurrences++;
    break;
  }

  switch (it) {
  case AXIOM:
    c.occurrencesInAxioms++;
    break;
  case ASSUMPTION:
    c.occurrencesInAssumptions++;
    break;
  case CONJECTURE:
    c.occurrencesInConjecture++;
    break;
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void LiteralOrderingFinder::findOrdering ()
{
  TRACER("LiteralOrderingFinder::findOrdering");

  registerOccurrences();

  
  int comparatorMode = 0;
  switch (_options.symbolPrecedence()) {
  case Options::BY_ARITY:
    comparatorMode = 1;
    break;
  case Options::BY_OCCURRENCE:
    comparatorMode = 0;
    break;
  case Options::BY_REVERSE_ARITY:
    comparatorMode = -1;
    break;
  }

  ArityComparator funComparator(comparatorMode,_functions);
  int noOfFuns = _signature.noOfFuns();
  Sort<Symbol*,ArityComparator> fsort(noOfFuns,funComparator);
  Signature::FunIterator funs(_signature);
  while (funs.hasNext()) {
    fsort.add(&funs.next());
  }
  fsort.sort();
  for (int i = noOfFuns-1;i >= 0; i--) {
    Symbol* s = fsort[i];
    s->setPrecedence(i+1);
  }

  ArityComparator predComparator(comparatorMode,_predicates);
  int noOfPreds = _signature.noOfPreds();
  Sort<Symbol*,ArityComparator> psort(noOfPreds,predComparator);
  Signature::PredIterator preds(_signature);
  while (preds.hasNext()) {
    psort.add(&preds.next());
  }
  psort.sort();
  for (int j = noOfPreds-1;j >= 0; j--) {
    Symbol* s = psort[j];
    switch (_options.literalComparisonMode()) {
    case Options::KINKY:
      s->setPrecedence(j+1);
      s->setNegativePrecedence(2*noOfPreds-j);
      break;

    case Options::PREDICATE:
    case Options::NORMAL:
      s->setNegativePrecedence(2*j+2); 
                                       
      s->setPrecedence(2*j+1);
      break;
    }
  }
} 



bool LiteralOrderingFinder::ArityComparator::lessThan (const Symbol* s1,
						       const Symbol* s2)
{
  TRACER("LiteralOrderingFinder::ArityComparator::lessThan");

  ASS(s1);
  ASS(s2);

  if (s1->isEqualityProxy()) {
    return false;
  }
  if (s2->isEqualityProxy()) {
    return true;
  }


  
  switch (Int::compare(s1->arity()*_byArity,
		       s2->arity()*_byArity)) {
  case -1:
    return true;
  case 0:
    break;
  case 1:
    return false;
  }

  const Counter& c1 = _counters[s1->number()];
  const Counter& c2 = _counters[s2->number()];

  
  int occ1 = c1.occurrencesInAxioms +
             c1.occurrencesInConjecture +
             c1.occurrencesInAssumptions;
  int occ2 = c2.occurrencesInAxioms +
             c2.occurrencesInConjecture +
             c2.occurrencesInAssumptions;

  switch (Int::compare(c2.occurrencesInAxioms*occ1,
		       c1.occurrencesInAxioms*occ2)) {
  case -1:
    return true;
  case 0:
    break;
  case 1:
    return false;
  }

  
  switch (Int::compare(occ2,occ1)) {
  case -1:
    return true;
  case 0:
    break;
  case 1:
    return false;
  }

  
  switch (Int::compare(c1.occurrencesInConjecture,
		       c2.occurrencesInConjecture)) {
  case -1:
    return true;
  case 0:
    break;
  case 1:
    return false;
  }

  
  switch (Int::compare(c1.negativeOccurrences,
		       c2.negativeOccurrences)) {
  case -1:
    return true;
  case 0:
    break;
  case 1:
    return false;
  }

  
  switch (Int::compare(c1.number,c2.number)) {
  case -1:
    return true;
  default:
    return false;
  }
} 



LiteralOrderingFinder::ArityComparator::ArityComparator (int byArity,
							 const Counter* counters)
  : _byArity(byArity),
    _counters(counters)
{
}



LiteralOrderingFinder::Counter::Counter()
  : occurrencesInAxioms(0),
    occurrencesInConjecture(0),
    occurrencesInAssumptions(0),
    positiveOccurrences(0),
    negativeOccurrences(0),
    neutralOccurrences(0),
    symbol(0),
    number(-1)
{
}
