
/*
 * File Matching.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Matching.hpp"
#include "Map.hpp"
#include "Substitution.hpp"
#include "Formula.hpp"


bool Matching::match (const Formula& f1, const Formula& f2, Substitution& sbst)
{
  Matching matcher(sbst);
  return matcher.match(f1,f2);
} 


bool Matching::match (const Formula& f1, const Formula& f2)
{
  if (f1.connective() != f2.connective()) {
    return false;
  }

  switch (f1.connective()) {
  case ATOM:
    return match(f1.atom(),f2.atom());

  case AND:
  case OR:
    return match(f1.args(),f2.args());

  case IMP:
  case IFF:
  case XOR:
    return match(f1.left(),f2.left()) &&
           match(f1.right(),f2.right());

  case NOT:
    return match(f1.uarg(),f2.uarg());

  case FORALL:
  case EXISTS:
    {
      VarList vs1(f1.vars());
      VarList vs2(f2.vars());
      if (vs1.length() != vs2.length()) {
	return false;
      }

      Iterator<Var> vi1 (vs1);
      Iterator<Var> vi2 (vs2);
      while (vi1.hasNext()) {
	Var v1(vi1.next());
	Var v2(vi2.next());
	_sbst.bind(v1,Term(v2));
	_bound1.incCount(v1);
	_bound2.incCount(v2);
      }
      if (! match(f1.qarg(),f2.qarg())) {
	return false;
      }
      
      vi1.reset(vs1);
      vi2.reset(vs2);
      while (vi1.hasNext()) {
	Var v1(vi1.next());
	_sbst.undoBinding(v1);
	_bound1.decCount(v1);
	_bound2.decCount(vi2.next());
      }
      return true;
    }
#if DEBUG_PREPRO
  default:
    ASS (false);
#endif
  }
} 



bool Matching::match (const FormulaList& fs1,
		      const FormulaList& fs2)
{
  if (fs1.isEmpty()) {
    return fs2.isEmpty();
  }
  
  if (fs2.isEmpty()) {
    return false;
  }

  return match(fs1.head(),fs2.head()) && 
         match(fs1.tail(),fs2.tail());
} 



bool Matching::match (const Term& t1, const Term& t2)
{
  if (t1.tag() != t2.tag()) {
    return false;
  }

  
  switch (t1.tag()) {
  case VAR:
    {
      Var v(t1.var());
      Term bound;
      if (!_sbst.isBound(v,bound)) { 
      
	if (hasBoundVariables(t2)) {
	  return false;
	}
	_sbst.bind(v,t2);
	return true;
      }
      
      if (_bound1.getCount(v) != 0) { 
	return bound.equals(t2);
      }
      
      return bound.equals(t2) && ! hasBoundVariables(t2);
    }

  case COMPOUND:
    return t1.functor() == t2.functor() &&
           match(t1.args(), t2.args());

  case NUMERIC:
    return t1.number() == t2.number();
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool Matching::match (const TermList& ts1,
		      const TermList& ts2)
{
  if (ts1.isEmpty()) {
    return ts2.isEmpty();
  }
  
  if (ts2.isEmpty()) {
    return false;
  }

  return match(ts1.head(),ts2.head()) && 
         match(ts1.tail(),ts2.tail());
} 



bool Matching::match (const Atom& a1, const Atom& a2)
{
  return a1.functor() == a2.functor() &&
         match(a1.args(), a2.args());
} 



bool Matching::hasBoundVariables (const Term& t) const
{
  switch (t.tag()) {
  case VAR:
    return _bound2.getCount(t.var()) != 0;

  case COMPOUND:
    {
      Iterator<Term> ts (t.args());
      while (ts.hasNext()) {
	if (hasBoundVariables(ts.next())) {
	  return true;
	}
      }
      return false;
    }

  case NUMERIC:
    return false;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
