
/*
 * File Miniscope.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Formula.hpp"
#include "Miniscope.hpp"
#include "Assertion.hpp"
#include "VampireException.hpp"
#include "VarCounter.hpp"


void Miniscope::miniscope (Unit& u)
{
  ASS(u.unitType() == FORMULA);
  ASS(Assertion::isInNNF(u.formula()));
  ASS(Assertion::isFlat(u.formula()));

  Formula f(u.formula());
  Position p;
  
  InferenceInfo* infs = 0;
  miniscope(f,infs,p);
  if (! f.equals(u.formula())) {
    u = Unit(f,new Inference(IR_MINISCOPE,u,infs));
  }
} 


void Miniscope::miniscope (Formula& f,
			   InferenceInfo*& inf, 
			   const Position& pos)
{
  TRACER("Miniscope::miniscope (Formula& f, ... )");

  switch (f.connective()) {
  case ATOM:
    return;

  case NOT:
    return;

  case AND:
  case OR: 
    {
      FormulaList nas(f.args()); 
      miniscope(nas, f.connective(), inf, pos, 0, f.args());
      if (nas == f.args()) { 
	return;
      }
      f = Formula(f.connective(), nas);
      return;
    }

  case EXISTS: 
  case FORALL: 
    {
      Formula down(f.qarg());
      Position pdown(0, pos);
      miniscope(down, inf, pdown);
      VarList vs; 
      VarList removed;
      findOccurring(down, f.vars(), vs, removed);
      if (! removed.isEmpty()) { 
	
	
	
	inf = new InferenceInfoTermList(IR_DUMMY_QUANTIFIER_REMOVAL,
					pos,
					inf,
					TermList(removed));
	if (vs.isEmpty()) { 
	  f = down;
	  return;
	}
      }

      Formula g(down);
      topMiniscope(g, f.connective(), vs, inf, pos);

      if (down != g) {  
	f = g;
	return;
      }
      
      if (removed.isNonEmpty() || down != f.qarg()) { 
	
	f = Formula (f.connective(), vs, down);
      }
      return;
    }

#if DEBUG_PREPRO
  default:
    ASS( false );
#endif
  }
} 



void Miniscope::miniscope (FormulaList& fs, 
			   Connective c, 
			   InferenceInfo*& inf,
			   const Position& pos,
			   int argNumber,
			   const FormulaList& firstArg)
{
  TRACER("void Miniscope::miniscope (FormulaList fs, ...)");

  if (fs.isEmpty ()) {
    return;
  }

  Formula g(fs.head());
  
  ASS (g.connective() != c);

  miniscope(g, inf, Position(argNumber,pos));
  FormulaList gs (fs.tail()); 
  miniscope(gs, c, inf, pos, argNumber+1, firstArg);

  if (g == fs.head() && gs == fs.tail()) { 
    return;
  }

  if (g.connective() == c) { 
    

    
    FormulaList args;
    appendN (firstArg, fs, argNumber, args);
    inf = new InferenceInfo(IR_FLATTEN,pos,inf);
    fs = g.args();
    fs.append(gs);
    return;
  }

  
  fs = FormulaList (g,gs);
} 



void Miniscope::findOccurring (const Formula& f, const VarList& in, 
			       VarList& out, VarList& removedVars)
{
  TRACER ("Miniscope::findOccurring");
  ASS (in.isNonEmpty());

  
  
  
  
  VarCounter counter;

  markOccurring(f, counter);
  splitVarList (in, out, removedVars, counter);
} 



void Miniscope::splitVarList (const VarList& vs, 
			      VarList& yes,
			      VarList& no,
			      VarCounter& counter)
{
  TRACER("Miniscope::splitVarList");

  if (vs.isEmpty()) {
    return;
  }

  
  splitVarList (vs.tail(), yes, no, counter);

  Var v (vs.head());
  if (counter.getCount(v) == 1) { 
    if (yes == vs.tail()) { 
      yes = vs;
      return;
    }
    
    yes.push(v);
    return;
  }

  
  no.push(v);
} 


void Miniscope::markOccurring (const Formula& f, VarCounter& counter)
{
  TRACER("Miniscope::markOccurring(const Formula& ...)");

  switch (f.connective()) {
  case ATOM:
    markOccurring(f.atom().args(), counter);
    return;

  case AND:
  case OR: 
    {
      Iterator<Formula> fs (f.args());
      while (fs.hasNext()) {
	markOccurring(fs.next(), counter);
      }
      return;
    }

  case NOT:
    markOccurring(f.uarg(), counter);
    return;

  case FORALL:
  case EXISTS:
    {
      
      Iterator<Var> vs(f.vars());
      while (vs.hasNext()) {
	Var v (vs.next());
	switch (counter.getCount(v)) {
	case 0: 
	  counter.setCount(v,2); 
	  break;
	case 1: 
	  break;
	default: 
	  counter.incCount(v);
	  break;
	}
      }
      markOccurring(f.qarg(), counter);
      
      vs.reset(f.vars());
      while (vs.hasNext()) {
	Var v (vs.next());
	switch (counter.getCount(v)) {
	case 1: 
	  break;
	case 2: 
	  counter.setCount(v,0); 
	  break;
	default: 
	  counter.decCount(v);
	  break;
	}
      }
      return;
    }

#if DEBUG_PREPRO
  default:
    ASS( false );
#endif
  }
} 



void Miniscope::markOccurring (const TermList& terms, VarCounter& counter)
{
  TRACER("Miniscope::markOccurring (const TermList& ...)");

  Iterator<Term> ts(terms);
  while (ts.hasNext()) {
    markOccurring(ts.next(),counter);
  }
} 



void Miniscope::markOccurring (const Term& t, VarCounter& counter)
{
  TRACER("Miniscope::markOccurring (const Term& ...)");

  switch (t.tag()) {
  case VAR: 
    {
      Var v (t.var());
      if (counter.getCount(v) == 0) { 
	
	counter.incCount(v);
      }
      return;
    }

  case COMPOUND:
    markOccurring(t.args(),counter);
    return;

  case NUMERIC:
    return;
  }
} 



int Miniscope::maxVar (const VarList& vars)
{
  TRACER("Miniscope::maxVar");
  ASS (vars.isNonEmpty());

  int m = vars.head().toInt();

  Iterator<Var> vs (vars.tail());
  while (vs.hasNext()) {
    int v = vs.next().toInt();
    if (v > m) {
      m = v;
    }
  }

  return m;
} 



void Miniscope::appendN (const FormulaList& fst, 
			 const FormulaList& snd,
			 int N,
			 FormulaList& result)
{
  ASS (N >= 0);

  if (N == 0) {
    result = snd;
    return;
  }

  
  appendN (fst.tail(), snd, N-1, result);
  result.push (fst.head());
} 



void Miniscope::topMiniscope (Formula& F, 
			      Connective Q, 
			      const VarList& vs, 
			      InferenceInfo*& inf, 
			      const Position& pos)
{
  TRACER("Miniscope::topMiniscope(Formula& ...)");

  switch (F.connective()) {
  case ATOM: 
  case NOT:
    return;
      
  case AND:
  case OR: 
    {
      bool isAnd = (F.connective() == AND);
      if (isAnd == (Q == FORALL)) { 
	
	
	inf = new InferenceInfo(isAnd ? IR_FORALL_AND : IR_EXISTS_OR,pos,inf);
	FormulaList args(F.args());
	topMiniscope(args, Q, vs, inf, pos, 0);
	F = Formula (F.connective(), args);
	return;
      }
      
      
      
      
      Miniscope cnt (Q, vs, F, pos, inf);
      if (cnt.noMiniscoping()) {
	return;
      }

      cnt.pushOnes();

      
      while (cnt.pushMany()) {
      }
      cnt.putResult(F);
      return;
    }

  case FORALL:
  case EXISTS: 
    if (F.connective() == Q) { 
      
      
      
      Formula f(F.qarg());
      inf = new InferenceInfo(IR_QUANTIFIER_SWAP,pos,inf);
      Position pdown (0, pos);
      topMiniscope(f,Q,vs,inf,pdown);
      if (f == F.qarg()) { 
	InferenceInfo* next = inf->next();
	inf->destroy();
	inf = next;
	inf = new InferenceInfo(IR_FLATTEN,pos,inf);
	VarList newvars(vs);
	newvars.append(F.vars());
	F = Formula(Q, newvars, f);
	return;
      }
      
      if (f.connective() == Q) {
	inf = new InferenceInfo(IR_FLATTEN,pos,inf);
	VarList newvars(F.vars());
	newvars.append(f.vars());
	F = Formula(Q, newvars, f.qarg());
	return;
      }
      
      F = Formula(Q, F.vars(), f);
      return;
    }
    
    return;
    
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void Miniscope::topMiniscope (FormulaList& fs,
			      Connective Q, 
			      const VarList& X,
			      InferenceInfo*& inf, 
			      const Position& pos,
			      int index)
{
  TRACER("Miniscope::topMiniscope(FormulaList& ...)");

  if (fs.isEmpty()) {
    return;
  }

  
  FormulaList gs(fs.tail());
  topMiniscope(gs, Q, X, inf, pos, index+1);

  VarList V;
  VarList removed;
  findOccurring(fs.head(), X, V, removed);
  if (! removed.isEmpty()) {
    inf = new InferenceInfoTermList(IR_DUMMY_QUANTIFIER_REMOVAL,
				    Position(index,pos),
				    inf,
				    TermList(removed));
  }

  
  if (V.isEmpty()) { 
    fs = FormulaList (fs.head(), gs);
    return;
  }

  
  Formula f(fs.head()); 
  topMiniscope(f, Q, V, inf, Position(index,pos));

  if (f == fs.head()) { 
    
    fs = FormulaList(Formula(Q,V,f), gs);
    return;
  }

  
  fs = FormulaList (f,gs);
} 



Miniscope::Miniscope (Connective Q, 
		      const VarList& V, 
		      const Formula& formula,
		      const Position& position,
		      InferenceInfo*& infs)
  :
  _quant(Q),
  _formula(formula),
  _nfs(formula.args().length()),
  _realNfs(_nfs),
  _nvs(V.length()),
  _occ(new bool* [_nvs]),
  _mem(new bool [_nfs * _nvs]),
  _index(0),
  _vars(new Var[_nvs]),
  _N(new int [_nvs]),
  _formulas(new Formula [_nfs]),
  _position(position),
  _inferences(infs)
{
  TRACER ("Miniscope::Miniscope");

  ASS(_nfs > 1 && _nvs > 0);

  if (! (_occ && _mem && _vars && _N && _formulas) ) {
    ALLOCATION_FAILURE;
  }
  
  for (int k = _nfs * _nvs - 1; k >= 0; k--) {
    _mem [k] = false;
  }

  
  _maxV = maxVar(V);

  
  _index = new bool* [_maxV+1];
  if (! _index) {
    ALLOCATION_FAILURE;
  }
  
  for (int v = _maxV; v >= 0; v--) {
    _index[v] = false;
  }
  
  
  bool* nxt = _mem;
  Var* nv = _vars;
  bool** nocc = _occ;
  Iterator<Var> vs (V);
  while (vs.hasNext()) {
    int v = vs.next().toInt();
    _index [v] = nxt;
    *nocc = nxt;
    nocc++;
    nxt += _nfs;
    nv->set(v);
    nv++;
  }

  int j = 0; 
  Iterator<Formula> fs (formula.args());
  while (fs.hasNext()) {
    Formula f(fs.next());
    count(f, j);
    _formulas[j] = f;
    j++;
  }
  
  delete [] _index;
  _index = 0;

  
  
  bool* onm = _mem;
  int* Nn = _N; 
  for (int n = 0; n < _nvs; n++) {
    Nn[0] = 0;
    for (int m = 0; m < _nfs; m++) {
      if (*onm) { 
        Nn[0] ++;
      }
      onm ++;
    }
    Nn++;
  }
} 



Miniscope::~Miniscope ()
{
  TRACER ("Miniscope::~Miniscope");

  delete [] _mem;
  delete [] _occ;
  delete [] _vars;
  delete [] _N;
  delete [] _formulas;
} 



void Miniscope::count (const Formula& f, int j)
{
  TRACER ("Miniscope::count(const Formula& f, int j)");

  switch (f.connective()) {
  case ATOM: 
    {
      Iterator<Term> ts (f.atom().args());
      while (ts.hasNext()) {
	count(ts.next(), j);
      }
      return;
    }

  case AND:
  case OR: 
    {
      Iterator<Formula> fs (f.args());
      while (fs.hasNext()) {
	count(fs.next(), j);
      }
      return;
    }

  case NOT:
    count(f.uarg(), j);
    return;
    
  case FORALL:
  case EXISTS:
    count(f.qarg(), j);
    return;

  default:
    ASS(false);
    return;
  }
} 



void Miniscope::count (const Term& t, int j)
{
  TRACER ("Miniscope::count(const Term& t, int j)");

  switch (t.tag()) {
  case VAR:
    {
      int v = t.var().toInt();
      if (v > _maxV) {
	return;
      }
	
      bool* vi = _index[v];
      if (! vi) {
	return;
      }

      vi[j] = true;
      return;
    }
    
  case NUMERIC:
    return;

  case COMPOUND: 
    {
      Iterator<Term> ts (t.args());
      while (ts.hasNext()) {
	count(ts.next(), j);
      }
      return;
    }
  }
} 


bool Miniscope::noMiniscoping () const
{
  TRACER("Miniscope::noMiniscoping");

  for (int i = _nvs-1; i >= 0; i--) {
    if ( _N[i] < _nfs ) {
      return false;
    }
  }

  return true;
} 



void Miniscope::pushOnes ()
{
  TRACER ("Miniscope::pushOnes");

  
  for (int v = 0; v < _nvs; v++) {
    if (_N[v] == 1) {
      
      int k;
      for (k = _nfs-1; k >= 0; k--) {
	if (_occ[v][k]) { 
	  pushOne(v, k);
	  break;
	}
      }
      ASS (k >=0);
    }
  }
} 



void Miniscope::pushOne (int v, int k)
{
  TRACER ("Miniscope::pushOne");

  
  
  IntList vs;
  occurInSameFormulas(v,vs);
  VarList ws;
  decodeVars(vs,ws);
  deleteVars(vs);
  Formula f(_formulas[k]);
  IntList toppledIndexes(k);
  IntList toppled;
  decodeFormulas(toppledIndexes, toppled);
  _inferences = new InferenceInfoTermList(_quant == FORALL ? IR_FORALL_OR :
					                     IR_EXISTS_AND,
					  _position,
					  _inferences,
					  TermList(ws));
  
  
  VarList xs;
  collectVars(xs);
  Position newPos(_position);
  if (xs.isNonEmpty()) {
    newPos.push(0);
  }
  newPos.push(k);
  topMiniscope(f,_quant,ws,_inferences,newPos);
  if (f == _formulas[k]) { 
    _formulas[k] = Formula(_quant,ws,f);
  }
  else {
    _formulas[k] = f;
  }
} 



void Miniscope::occurInSameFormulas (int v, IntList& vs) const
{
  TRACER("Miniscope::occurInSameFormulas (Var v, VarList& vs)");

  IntList ws; 

  for (int w = _nvs-1; w > v; w--) {
    if (_N[w] == 0) { 
      continue;
    }
    if (occurInSameFormulas(v,w)) {
      ws.push (w);
    }
  }

  vs = IntList(v,ws);
} 



bool Miniscope::occurInSameFormulas (int v, int w) const
{
  TRACER("Miniscope::occurInSameFormulas (int v, int w)");

  bool* V = _occ[v];
  bool* W = _occ[w];

  for (int i = _nfs-1; i >= 0; i--) {
    if (*V != *W) {
      return false;
    }
    V++;
    W++;
  }

  return true;
} 



void Miniscope::decodeVars (const IntList& in, VarList& out) const
{
  TRACER ("Miniscope::decodeVars");

  if (in.isEmpty()) {
    return;
  }

  
  decodeVars (in.tail(), out);
  out.push(_vars[in.head()]);
} 



void Miniscope::deleteVars (const IntList& vs)
{
  TRACER("Miniscope::deleteVars");

  Iterator<int> ws (vs);
  while (ws.hasNext()) {
    _N[ws.next()] = 0;
  }
} 



void Miniscope::putResult (Formula& res) 
{
  TRACER("Miniscope::putResult");

  VarList ws;
  
  collectVars(ws);

  FormulaList fs;
  
  for (int k = _nfs-1; k >= 0; k--) {
    Formula f(_formulas[k]);
    if (! f.isNull()) {
      fs = FormulaList(f,fs);
    }
  }
  ASS (fs.length() > 1);

  res = Formula(_formula.connective(),fs);
  if (ws.isNonEmpty()) {
    res = Formula(_quant,ws,res);
  }
} 


void Miniscope::collectVars (VarList& ws) 
{
  TRACER("Miniscope::collectVars");

  
  for (int v = _nvs-1; v >= 0; v--) {
    if (_N[v] != 0) { 
      ws.push(_vars[v]);
    }
  }
} 


void Miniscope::decodeFormulas (const IntList& in, IntList& out) const
{
  TRACER("Miniscope::decodeFormulas/2");

  decodeFormulas (in, out, 0, 0);
} 



void Miniscope::decodeFormulas (const IntList& in, IntList& out,
				int inIndex, int outIndex) const
{
  TRACER("Miniscope::decodeFormulas/4");

  if (in.isEmpty()) {
    return;
  }
  int i = in.head();
  while (inIndex != i) {
    if (! _formulas[i].isNull()) {
      outIndex ++;
    }
    inIndex++;
  }
  
  ASS (! _formulas[i].isNull());
  out = IntList (outIndex);

  decodeFormulas(in.tail(), const_cast<IntList&>(out.tail()), 
		 inIndex+1, outIndex+1);
} 



bool Miniscope::pushMany ()
{
  TRACER ("Miniscope::pushMany");

  if (_realNfs < 3) { 
    return false;
  }

  
  
  int v = 0; 
  int m = _realNfs; 
  
  
  for (int w = 0; w < _nvs; w++) {
    if (_N[w] != 0 && _N[w] < m) {
      m = _N[w];
      v = w;
    }
  }
  
  if (m == _realNfs) { 
    return false;
  }
  
  IntList vs;
  occurInSameFormulas(v,vs);

  
  
  FormulaList G;
  int lastDeleted = 0; 
  bool* occs = _occ[v]; 
  for (int i = _nfs-1; i >= 0; i--) {
    Formula f (_formulas[i]);
    if (f.isNull()) { 
      continue;
    }
    if (occs[i]) { 
      G.push(f);
      if (lastDeleted != 0) {
	
	_formulas[lastDeleted].makeNull();
        _realNfs --;
	
	for (int w = _nvs-1; w >= 0; w--) {
	  if (_N[w] != 0) { 
	    if (_occ[w][lastDeleted]) {
	      if (_occ[w][i]) { 
		_N[w] --;
		ASS (_N[w] > 0);
	      }
	      else { 
		_occ[w][i] = true;
	      }
	    }
	    
	  }
	}
      }
      
      lastDeleted = i;
    }
  }

  
  VarList ws;
  decodeVars(vs,ws);
  deleteVars(vs);

  
  Formula F (_quant, ws, 
	     Formula( _formula.connective(), G));
  _formulas[lastDeleted] = F;

  _inferences = new InferenceInfoTermList(_quant == FORALL ? IR_FORALL_OR :
					                     IR_EXISTS_AND,
					  _position,
					  _inferences,
					  TermList(ws));

  
  return true;
} 

