
/*
 * File Options.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <string>
#include <cstdlib>
#include <iostream>
#include <climits>

#include "Tracer.hpp"
#include "Options.hpp"
#include "Random.hpp"
#include "XML.hpp"
#include "Int.hpp"
#include "VampireException.hpp"


string Opt::_demodulationValues[] =
  {"all","off","preordered"};
string Opt::_subsumptionValues[] =
  {"multiset","off","set"};
string Opt::_literalComparisonModeValues[] =
  {"kinky","normal","predicate"};
string Opt::_inputSyntaxValues[] =
  {"kif","tptp"};
string Opt::_functionDefinitionEliminationValues[] =
  {"all","dummy","no"};

string Opt::_mainAlgValues[] =
  {"discount","lrs","otter"};
string Opt::_modeValues[] = 
  {"casc-j2","kif","spider","vampire"};
string Opt::_outputSyntaxValues[] =
  {"kif","latex","native","tptp","xml"};
string Opt::_simplificationOrderingValues[] =
  {"KBO","nonrecursive_KBO"};
string Opt::_splittingValues[] =
  {"off","on","prepro_only"};
string Opt::_splittingStyleValues[] =
  {"blocking","parallel"};
string Opt::_statisticsValues[] =
  {"brief","detailed","none"};
string Opt::_symbolPrecedenceValues[] =
  {"arity", "occurrence", "reverse_arity"};
string Opt::_termWeightingSchemeValues[] =
  {"nonuniform_constant","uniform"};
string Opt::_transitivityChainingValues[] =
  {"full","none","only_nonvariables"};


Options::Options ()
  :
  _ageWeightRatio (Opt::AGE_WEIGHT_RATIO, AgeWeight(1), _allOptions),
  _allocationBufferPrice (Opt::ALLOCATION_BUFFER_PRICE, 2.0, _allOptions),
  _allocationBufferSize (Opt::ALLOCATION_BUFFER_SIZE, 10, _allOptions),
  _arityCheck (Opt::ARITY_CHECK, true, _allOptions),

  _backwardDemodulation (Opt::BACKWARD_DEMODULATION, DEMODULATION_ALL, _allOptions),
  _backwardSubsumption (Opt::BACKWARD_SUBSUMPTION, SUBSUMPTION_MULTISET, _allOptions),
  _basicStrategy (Opt::BASIC_STRATEGY, false, _allOptions),
  _bdOnSplittingBranches (Opt::BD_ON_SPLITTING_BRANCHES, true, _allOptions),

  _elimDef (Opt::ELIM_DEF, 0, _allOptions),
  _equalityProxy (Opt::EQUALITY_PROXY, false, _allOptions),
  _equalityResolutionWithDeletion (Opt::EQUALITY_RESOLUTION_WITH_DELETION, false, _allOptions),

  _fastDemodulationByCommutativity(Opt::FAST_DEMODULATION_BY_COMMUTATIVITY, true, _allOptions),
  _fdOnSplittingBranches (Opt::FD_ON_SPLITTING_BRANCHES, true, _allOptions),
  _forwardDemodulation (Opt::FORWARD_DEMODULATION, DEMODULATION_ALL, _allOptions),
  _forwardSubsumption (Opt::FORWARD_SUBSUMPTION, SUBSUMPTION_MULTISET, _allOptions),
  _forwardSubsumptionResolution (Opt::FORWARD_SUBSUMPTION_RESOLUTION, true, _allOptions),
  _forwardSubsumptionSignatureFilters (Opt::FORWARD_SUBSUMPTION_SIGNATURE_FILTERS, false, _allOptions),
  _functionDefinitionElimination (Opt::FUNCTION_DEFINITION_ELIMINATION, FDE_DUMMY, _allOptions),
  _groundInequalitySplitting (Opt::GROUND_INEQUALITY_SPLITTING, 3, _allOptions),

  _include (Opt::INCLUDE, "", _allOptions),
  _inheritedNegativeSelection (Opt::INHERITED_NEGATIVE_SELECTION, false, _allOptions),
  _inputFile (Opt::INPUT_FILE, "", _allOptions),
  _inputSyntax (Opt::INPUT_SYNTAX, INPUT_TPTP, _allOptions),

  _latexOutput(Opt::LATEX_OUTPUT,"off",_allOptions),
  _literalComparisonMode (Opt::LITERAL_COMPARISON_MODE, NORMAL, _allOptions),
  _logFile (Opt::LOG_FILE, "", _allOptions),
  _lrsFirstMemoryCheck (Opt::LRS_FIRST_MEMORY_CHECK, 100, _allOptions),
  _lrsFirstTimeCheck (Opt::LRS_FIRST_TIME_CHECK, 5, _allOptions),
  _lrsSoftness (Opt::LRS_SOFTNESS, 1.0, _allOptions),

  _mainAlg (Opt::MAIN_ALG, LRS, _allOptions),
  _mainPremiseDemodulation(Opt::MAIN_PREMISE_DEMODULATION,0,_allOptions),
  _mainPremiseSubsumption(Opt::MAIN_PREMISE_SUBSUMPTION,0,_allOptions),
  _maxActive (Opt::MAX_ACTIVE, 0, _allOptions),
  _maxInferenceDepth (Opt::MAX_INFERENCE_DEPTH, 0, _allOptions),
  _maxPassive (Opt::MAX_PASSIVE, 0, _allOptions),
  _maxSkolem (Opt::MAX_SKOLEM, 4096, _allOptions),
  _maxWeight (Opt::MAX_WEIGHT, 0, _allOptions),
  _memoryLimit (Opt::MEMORY_LIMIT, 300, _allOptions),
  _miniscope (Opt::MINISCOPE, true, _allOptions),
  _mode (Opt::MODE, VAMPIRE, _allOptions),

  _noInferencesBetweenRules (Opt::NO_INFERENCES_BETWEEN_RULES, false, _allOptions),
  _nongoalWeightCoefficient (Opt::NONGOAL_WEIGHT_COEFFICIENT, 1.0, _allOptions),
  _normalize(Opt::NORMALIZE, false, _allOptions),

  _orphanMurder (Opt::ORPHAN_MURDER, false, _allOptions),
  _outputEqualityInfix (Opt::OUTPUT_EQUALITY_INFIX, false, _allOptions),
  _outputMessages (Opt::OUTPUT_MESSAGES, true, _allOptions),
  _outputSyntax (Opt::OUTPUT_SYNTAX, OS_NATIVE, _allOptions),

  _paramodulation (Opt::PARAMODULATION, true, _allOptions),
  _paramodulationFromVariables (Opt::PARAMODULATION_FROM_VARIABLES, true, _allOptions),
  _proof (Opt::PROOF, true, _allOptions),
  _proofChecking (Opt::PROOF_CHECKING, false, _allOptions),

  _randomSeed (Opt::RANDOM_SEED, Random::getSeed(), _allOptions),

  _selectedEqualityWeightCoefficient (Opt::SELECTED_EQUALITY_WEIGHT_COEFFICIENT, 1.0, _allOptions),
  _selection (Opt::SELECTION, P_4, _allOptions),
  _showActive (Opt::SHOW_ACTIVE, false, _allOptions),
  _showGenerated (Opt::SHOW_GENERATED, false, _allOptions),
  _showInput (Opt::SHOW_INPUT, false, _allOptions),
  _showOptions (Opt::SHOW_OPTIONS, false, _allOptions),
  _showPassive (Opt::SHOW_PASSIVE, false, _allOptions),
  _showPreprocessed (Opt::SHOW_PREPROCESSED, false, _allOptions),
  _showProfile (Opt::SHOW_PROFILE, false, _allOptions),
  _showWeightLimitChanges (Opt::SHOW_WEIGHT_LIMIT_CHANGES, false, _allOptions),
  _sidePremiseDemodulation(Opt::SIDE_PREMISE_DEMODULATION,0,_allOptions),
  _sidePremiseSubsumption(Opt::SIDE_PREMISE_SUBSUMPTION,0,_allOptions),
  _simplificationOrdering (Opt::SIMPLIFICATION_ORDERING, KBO, _allOptions),
  _simultaneousBackwardSuperposition(Opt::SIMULTANEOUS_BACKWARD_SUPERPOSITION,false,_allOptions),
  _sos (Opt::SOS, false, _allOptions),
  _splitting (Opt::SPLITTING, SPLIT_PREPRO_ONLY, _allOptions),
  _splittingStyle (Opt::SPLITTING_STYLE, PARALLEL, _allOptions),
  _statistics (Opt::STATISTICS, STATISTICS_DETAILED, _allOptions),
  _symbolPrecedence (Opt::SYMBOL_PRECEDENCE, BY_ARITY, _allOptions),
  _symbolWeightByArity (Opt::SYMBOL_WEIGHT_BY_ARITY, 0, _allOptions),

  _termWeightingScheme (Opt::TERM_WEIGHTING_SCHEME, TWS_UNIFORM, _allOptions),
  _testId (Opt::TEST_ID, "unspecified_test", _allOptions),
  _testNumber (Opt::TEST_NUMBER, Random::getInteger(), _allOptions),
  _timeLimit (Opt::TIME_LIMIT, 0, _allOptions),
  _transitivityChaining(Opt::TRANSITIVITY_CHAINING,TC_NONE,_allOptions),

  _unusedDefinitionRemoval (Opt::UNUSED_DEFINITION_REMOVAL, true, _allOptions),

  _weightIncrement (Opt::WEIGHT_INCREMENT, false, _allOptions)
{
} 



void Options::correct (int argc, char* argv [])
{
  TRACER ( "Options::correct" );

  char** last = argv + argc;
  char** current = argv + 1;

  if (argc == 3) { 
    readFromTestId(argv[1]);
    setInputFile(argv[2]);
    return;
  }

  while (current != last) {
    Opt* opt = find(*current); 
    if (opt) { 
      current++; 
      if (current == last) { 
	USER_ERROR(string("error in command line after ") + current[-1]);
      }
      
      opt->readValueOrFail(*current);
      
      current++;
    }
    
    else if (current+1 != last) { 
      USER_ERROR(string("Error in command line: ") + 
		 *current + 
		 " is not a valid option");
    }
    else { 
      setInputFile (*current);

      const long maxTime = LONG_MAX;
      if (timeLimitInDeciseconds() > maxTime) {
	setTimeLimitInDeciseconds (maxTime);
      }
      else { 
	setTimeLimitInSeconds(timeLimitInDeciseconds());
      }

      return;
    }
  }
  
  USER_ERROR("input file name missing");
} 


bool Option<bool>::readValue (const char* val)
{
  TRACER("Option<bool>::readValue");

  if (! strcmp(val,"on")) {
    _value = true;
  }
  else if (! strcmp(val,"off")) {
    _value = false;
  }
  else { 
    return false;
  }

  _changed = true;
  return true;
} 



void Opt::output (ostream& str) const
{
  TRACER("Opt::output");

  str << name() << " = ";
  toStream (str);
  str << '\n';
} 



void Option<bool>::toStream (ostream& str) const
{
  TRACER("Option<bool>::toStream");

  str << (_value ? "on" : "off");
} 


bool Option<const char*>::readValue (const char* val)
{
  TRACER("Option<const char*>::readValue");
  _value = val;

  _changed = true;
  return true;
} 



bool Option<string>::readValue (const char* val)
{
  TRACER("Option<string>::readValue");

  _value = val;
  _changed = true;
  return true;
} 


void Option<const char*>::toStream (ostream& str) const
{
  TRACER("Option<const char*>::toStream");
  str << (_value ? _value : "(undefined)");
} 



void Option<string>::toStream (ostream& str) const
{
  TRACER("Option<string>::toStream");

  str << _value;
} 


bool Option<int>::readValue (const char* val)
{
  TRACER("Option<int>::readValue");

  char* endptr = 0;
  _value = (int)(strtol (val, &endptr, 10));

  if (*endptr) { 
    return false;
  }

  _changed = true;
  return true;
} 



void Option<int>::toStream (ostream& str) const
{
  TRACER("Option<int>::toStream");

  str << _value;
} 


bool Option<long>::readValue (const char* val)
{
  TRACER("Option<long>::readValue");

  char* endptr = 0;
  _value = strtol (val, &endptr, 10);

  if (*endptr) { 
    return false;
  }

  _changed = true;
  return true;
} 



void Option<long>::toStream (ostream& str) const
{
  TRACER("Option<long>::toStream");

  str << _value;
} 


bool Option<float>::readValue (const char* val)
{
  TRACER("Option<float>::readValue");

  char* endptr = 0;
  _value = (float)(strtod (val, &endptr));

  if (*endptr) { 
    return false;
  }

  _changed = true;
  return true;
} 



void Option<float>::toStream (ostream& str) const
{
  TRACER("Option<float>::toStream");

  str << _value;
} 



bool Option<Options::Demodulation>::readValue (const char* val)
{
  TRACER("Option<Options::Demodulation>::readValue");

  int found = find(val, _demodulationValues,
		   sizeof(_demodulationValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::Demodulation)found;
  _changed = true;
  return true;
} 



void Option<Options::Demodulation>::toStream (ostream& str) const
{
  TRACER("Option<Options::Demodulation>::toStream");

  str << _demodulationValues[_value];
} 



XML::Element* Option<Options::Demodulation>::toXML () const
{
  TRACER("Option<Options::Demodulation>::toXML ()");

  return new XML::TextElement(name(),_demodulationValues[_value]);
} 



bool Option<Options::LiteralComparisonMode>::readValue (const char* val)
{
  TRACER("Option<Options::LiteralComparisonMode>::readValue");

  int found = find(val, _literalComparisonModeValues,
		   sizeof(_literalComparisonModeValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::LiteralComparisonMode)found;
  _changed = true;
  return true;
} 



void Option<Options::LiteralComparisonMode>::toStream (ostream& str) const
{
  TRACER("Option<Options::LiteralComparisonMode>::toStream");

  str << _literalComparisonModeValues[_value];
} 



XML::Element* Option<Options::LiteralComparisonMode>::toXML () const
{
  TRACER("Option<Options::LiteralComparisonMode>::toXML ()");

  return new XML::TextElement(name(),_literalComparisonModeValues[_value]);
} 



XML::Element* Option<const char*>::toXML () const
{
  TRACER("Option<const char*>::toXML ()");

  return new XML::TextElement(name(),_value ? _value : "(undefined)");
} 



XML::Element* Option<string>::toXML () const
{
  TRACER("Option<string>::toXML ()");

  return new XML::TextElement(name(),_value);
} 



bool Option<Options::Subsumption>::readValue (const char* val)
{
  TRACER("Option<Options::Subsumption>::readValue");

  int found = find(val, _subsumptionValues,
		   sizeof(_subsumptionValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::Subsumption)found;
  _changed = true;
  return true;
} 



void Option<Options::Subsumption>::toStream (ostream& str) const
{
  TRACER("Option<Options::Subsumption>::toStream");

  str << _subsumptionValues[_value];
} 



XML::Element* Option<Options::Subsumption>::toXML () const
{
  TRACER("Option<Options::Subsumption>::toXML ()");

  return new XML::TextElement(name(),_subsumptionValues[_value]);
} 



bool Option<Options::Selection>::readValue (const char* val)
{
  TRACER("Option<Options::Selection>::readValue");

  if (! strcmp (val,"0")) {
    _value = Options::P_0;
  }
  else if (! strcmp (val,"1")) {
    _value = Options::P_1;
  }
  else if (! strcmp (val,"2")) {
    _value = Options::P_2;
  }
  else if (! strcmp (val,"3")) {
    _value = Options::P_3;
  }
  else if (! strcmp (val,"4")) {
    _value = Options::P_4;
  }
  else if (! strcmp (val,"5")) {
    _value = Options::P_5;
  }
  else if (! strcmp (val,"6")) {
    _value = Options::P_6;
  }
  else if (! strcmp (val,"7")) {
    _value = Options::P_7;
  }
  else if (! strcmp (val,"8")) {
    _value = Options::P_8;
  }
  else if (! strcmp (val,"9")) {
    _value = Options::P_9;
  }
  else if (! strcmp (val,"10")) {
    _value = Options::P_10;
  }
  else if (! strcmp (val,"11")) {
    _value = Options::P_11;
  }
  else if (! strcmp (val,"12")) {
    _value = Options::P_12;
  }
  else if (! strcmp (val,"13")) {
    _value = Options::P_13;
  }
  else if (! strcmp (val,"14")) {
    _value = Options::P_14;
  }
  else if (! strcmp (val,"15")) {
    _value = Options::P_15;
  }
  else if (! strcmp (val,"16")) {
    _value = Options::P_16;
  }
  else if (! strcmp (val,"17")) {
    _value = Options::P_17;
  }
  else if (! strcmp (val,"-2")) {
    _value = Options::N_2;
  }
  else if (! strcmp (val,"-3")) {
    _value = Options::N_3;
  }
  else if (! strcmp (val,"-4")) {
    _value = Options::N_4;
  }
  else if (! strcmp (val,"-5")) {
    _value = Options::N_5;
  }
  else if (! strcmp (val,"-6")) {
    _value = Options::N_6;
  }
  else if (! strcmp (val,"-7")) {
    _value = Options::N_7;
  }
  else if (! strcmp (val,"-9")) {
    _value = Options::N_9;
  }
  else if (! strcmp (val,"-10")) {
    _value = Options::N_10;
  }
  else if (! strcmp (val,"-11")) {
    _value = Options::N_11;
  }
  else if (! strcmp (val,"-12")) {
    _value = Options::N_12;
  }
  else if (! strcmp (val,"-13")) {
    _value = Options::N_13;
  }
  else if (! strcmp (val,"-14")) {
    _value = Options::N_14;
  }
  else if (! strcmp (val,"-15")) {
    _value = Options::N_15;
  }
  else if (! strcmp (val,"-16")) {
    _value = Options::N_16;
  }
  else if (! strcmp (val,"-17")) {
    _value = Options::N_17;
  }
  else if (! strcmp (val,"1002")) {
    _value = Options::P_1002;
  }
  else if (! strcmp (val,"1003")) {
    _value = Options::P_1003;
  }
  else if (! strcmp (val,"1004")) {
    _value = Options::P_1004;
  }
  else if (! strcmp (val,"1005")) {
    _value = Options::P_1005;
  }
  else if (! strcmp (val,"1006")) {
    _value = Options::P_1006;
  }
  else if (! strcmp (val,"1007")) {
    _value = Options::P_1007;
  }
  else if (! strcmp (val,"1008")) {
    _value = Options::P_1008;
  }
  else if (! strcmp (val,"1009")) {
    _value = Options::P_1009;
  }
  else if (! strcmp (val,"1010")) {
    _value = Options::P_1010;
  }
  else if (! strcmp (val,"1011")) {
    _value = Options::P_1011;
  }
  else if (! strcmp (val,"1012")) {
    _value = Options::P_1012;
  }
  else if (! strcmp (val,"1013")) {
    _value = Options::P_1013;
  }
  else if (! strcmp (val,"1014")) {
    _value = Options::P_1014;
  }
  else if (! strcmp (val,"-1005")) {
    _value = Options::N_1005;
  }
  else if (! strcmp (val,"-1009")) {
    _value = Options::N_1009;
  }
  else if (! strcmp (val,"-1012")) {
    _value = Options::N_1012;
  }
  else if (! strcmp (val,"-1014")) {
    _value = Options::N_1014;
  }
  else { 
    return false;
  }

  _changed = true;
  return true;
} 



void Option<Options::Selection>::toStream (ostream& str) const
{
  TRACER("Option<Options::Selection>::toStream");

  str << (int)_value;
} 



Opt* Options::findByShortName (const char* name)
{
  int found;

  
  static StringInt sn[] = {
    {"awr",Opt::AGE_WEIGHT_RATIO},
    {"bas",Opt::BASIC_STRATEGY},
    {"bd",Opt::BACKWARD_DEMODULATION},
    {"bs",Opt::BACKWARD_SUBSUMPTION},
    {"ep",Opt::EQUALITY_PROXY},
    {"fde",Opt::FUNCTION_DEFINITION_ELIMINATION},
    {"fsr",Opt::FORWARD_SUBSUMPTION_RESOLUTION},
    {"i",Opt::INCLUDE},
    {"l",Opt::LOG_FILE},
    {"lcm",Opt::LITERAL_COMPARISON_MODE},
    {"ls",Opt::LRS_SOFTNESS},
    {"m",Opt::MEMORY_LIMIT},
    {"ma",Opt::MAIN_ALG},
    {"n",Opt::NORMALIZE},
    {"nwc",Opt::NONGOAL_WEIGHT_COEFFICIENT},
    {"p",Opt::PROOF},
    {"pc",Opt::PROOF_CHECKING},
    {"s",Opt::SELECTION},
    {"sos",Opt::SOS},
    {"sp",Opt::SPLITTING},
    {"t",Opt::TIME_LIMIT},
    {"wi",Opt::WEIGHT_INCREMENT} };
	
  static StringIntMap shortNames(sn,sizeof(sn)/sizeof(Options::StringInt));
  if (shortNames.find(name,found)) {
    return _allOptions[found];
  }
  return 0;
} 



Opt* Options::find (const char* name) 
{
  TRACER("Options::find");

  if (*name != '-' || ! name[1]) {
    return 0;
  }

  if (name[1] != '-') { 
    return findByShortName(name+1);
  }


  static StringInt nm[] = {
    {"age_weight_ratio",Opt::AGE_WEIGHT_RATIO},
    {"allocation_buffer_price",Opt::ALLOCATION_BUFFER_PRICE},
    {"allocation_buffer_size",Opt::ALLOCATION_BUFFER_SIZE},
    {"arity_check",Opt::ARITY_CHECK},
    {"backward_demodulation",Opt::BACKWARD_DEMODULATION},
    {"backward_subsumption",Opt::BACKWARD_SUBSUMPTION},
    {"basic_strategy",Opt::BASIC_STRATEGY},
    {"bd_on_splitting_branches",Opt::BD_ON_SPLITTING_BRANCHES},
    {"elim_def",Opt::ELIM_DEF},
    {"equality_proxy",Opt::EQUALITY_PROXY},
    {"equality_resolution_with_deletion",Opt::EQUALITY_RESOLUTION_WITH_DELETION},
    {"fast_demodulation_by_commutativity",Opt::FAST_DEMODULATION_BY_COMMUTATIVITY},
    {"fd_on_splitting_branches",Opt::FD_ON_SPLITTING_BRANCHES},
    {"forward_demodulation",Opt::FORWARD_DEMODULATION},
    {"forward_subsumption",Opt::FORWARD_SUBSUMPTION},
    {"forward_subsumption_resolution",Opt::FORWARD_SUBSUMPTION_RESOLUTION},
    {"forward_subsumption_signature_filters",Opt::FORWARD_SUBSUMPTION_SIGNATURE_FILTERS},
    {"function_definition_elimination",Opt::FUNCTION_DEFINITION_ELIMINATION},
    {"ground_inequality_splitting",Opt::GROUND_INEQUALITY_SPLITTING},
    {"include",Opt::INCLUDE},
    {"inherited_negative_selection",Opt::INHERITED_NEGATIVE_SELECTION},
    {"input_file",Opt::INPUT_FILE},
    {"input_syntax",Opt::INPUT_SYNTAX},
    {"latex_output",Opt::LATEX_OUTPUT},
    {"literal_comparison_mode",Opt::LITERAL_COMPARISON_MODE},
    {"log_file",Opt::LOG_FILE},
    {"lrs_first_memory_check",Opt::LRS_FIRST_MEMORY_CHECK},
    {"lrs_first_time_check",Opt::LRS_FIRST_TIME_CHECK},
    {"lrs_softness",Opt::LRS_SOFTNESS},
    {"main_alg",Opt::MAIN_ALG},
    {"main_premise_demodulation",Opt::MAIN_PREMISE_DEMODULATION},
    {"main_premise_subsumption",Opt::MAIN_PREMISE_SUBSUMPTION},
    {"max_active",Opt::MAX_ACTIVE},
    {"max_inference_depth",Opt::MAX_INFERENCE_DEPTH},
    {"max_passive",Opt::MAX_PASSIVE},
    {"max_skolem",Opt::MAX_SKOLEM},
    {"max_weight",Opt::MAX_WEIGHT},
    {"memory_limit",Opt::MEMORY_LIMIT},
    {"miniscope",Opt::MINISCOPE},
    {"mode",Opt::MODE},
    {"no_inferences_between_rules",Opt::NO_INFERENCES_BETWEEN_RULES},
    {"nongoal_weight_coefficient",Opt::NONGOAL_WEIGHT_COEFFICIENT},
    {"normalise",Opt::NORMALIZE},
    {"normalize",Opt::NORMALIZE},
    {"orphan_murder",Opt::ORPHAN_MURDER},
    {"output_equality_infix",Opt::OUTPUT_EQUALITY_INFIX},
    {"output_messages",Opt::OUTPUT_MESSAGES},
    {"output_syntax",Opt::OUTPUT_SYNTAX},
    {"paramodulation_from_variables",Opt::PARAMODULATION},
    {"paramodulation",Opt::PARAMODULATION_FROM_VARIABLES},
    {"proof",Opt::PROOF},
    {"proof_checking",Opt::PROOF_CHECKING},
    {"random_seed",Opt::RANDOM_SEED},
    {"selected_equality_weight_coefficient",Opt::SELECTED_EQUALITY_WEIGHT_COEFFICIENT},
    {"selection",Opt::SELECTION},
    {"show_active",Opt::SHOW_ACTIVE},
    {"show_generated",Opt::SHOW_GENERATED},
    {"show_input",Opt::SHOW_INPUT},
    {"show_options",Opt::SHOW_OPTIONS},
    {"show_passive",Opt::SHOW_PASSIVE},
    {"show_preprocessed",Opt::SHOW_PREPROCESSED},
    {"show_profile",Opt::SHOW_PROFILE},
    {"show_weight_limit_changes",Opt::SHOW_WEIGHT_LIMIT_CHANGES},
    {"side_premise_demodulation",Opt::SIDE_PREMISE_DEMODULATION},
    {"side_premise_subsumption",Opt::SIDE_PREMISE_SUBSUMPTION},
    {"simplification_ordering",Opt::SIMPLIFICATION_ORDERING},
    {"simultaneous_backward_superposition",Opt::SIMULTANEOUS_BACKWARD_SUPERPOSITION},
    {"sos",Opt::SOS},
    {"splitting",Opt::SPLITTING},
    {"splitting_style",Opt::SPLITTING_STYLE},
    {"statistics",Opt::STATISTICS},
    {"symbol_precedence",Opt::SYMBOL_PRECEDENCE},
    {"symbol_weight_by_arity",Opt::SYMBOL_WEIGHT_BY_ARITY},
    {"term_weighting_scheme",Opt::TERM_WEIGHTING_SCHEME},
    {"test_id",Opt::TEST_ID},
    {"test_number",Opt::TEST_NUMBER},
    {"time_limit",Opt::TIME_LIMIT},
    {"transitivity_chaining",Opt::TRANSITIVITY_CHAINING},
    {"unused_definition_removal",Opt::UNUSED_DEFINITION_REMOVAL},
    {"weight_increment",Opt::WEIGHT_INCREMENT} };
	

  static StringIntMap names(nm,sizeof(nm)/sizeof(Options::StringInt));
  int found;
  if (names.find(name+2,found)) {
    return _allOptions[found];
  }
  return 0;
} 



void Options::kif ()
{
  TRACER("Options::kif");

  
  setUnusedDefinitionRemoval (false); 
  setArityCheck (false); 
  setMiniscope (false);

  
  setNongoalWeightCoefficient (2.0);

  
  setSos(true);
  setOutputMessages(false);
  setBackwardSubsumption(SUBSUMPTION_MULTISET);
  setLogFile("");
  setInputSyntax(INPUT_KIF);
  setOutputSyntax(OS_KIF);
} 



void Options::fullSplitting ()
{
  TRACER("Options::fullSplitting");

  setSplitting (SPLIT_ON);
  setSplittingStyle (PARALLEL);
  setFdOnSplittingBranches (true);
  setBdOnSplittingBranches (true);
} 



string Options::includeFileName (string relativeName)
{
  TRACER("Options::includeFileName");
  ASS (relativeName[0] == '\'');

  
  int lastPosition = relativeName.find_last_of('\'');
  ASS (lastPosition != 0);

  
  string truncatedRelativeName = relativeName.substr(1,lastPosition-1);

  if (truncatedRelativeName[0] == '/') { 
    return truncatedRelativeName;
  }

  
  
  
  string dir = include ();
  if (dir == "") { 
    
    dir = getenv("TPTP_DIR");
    if (dir == "") { 
      dir = ".";
    }
  }
  

  return dir + "/" + truncatedRelativeName;
} 



void Options::output (ostream& str) const
{
  str << (void*)this << "\n" << flush;
  TRACER("Options::output");

  if (! showOptions()) {
    return;
  }
  if (outputSyntax() == OS_XML) {
    XML::Element* xml = toXML();
    xml->write(str);
    xml->destroy();
    return;
  }

  str << "=========== Options ==========\n";








  str << "======= End of options =======\n";
} 


 
string Options::problemName () const
{
  TRACER("Options::problemName");

  string inputFileName = inputFile();
  if (inputFileName == "") {
    return "unknown";
  }

  int length = inputFileName.length() - 1;
  const char* name = inputFileName.c_str();

  int b = length - 1; 
  while (b >= 0 && name[b] != '/') {
    b--;
  }
  b++;

  int e = length-1; 
  while (e >= b && name[e] != '.') {
    e--;
  }
  if (e < b) {
    e = length-1;
  }

  return inputFileName.substr(b,e-b);
} 



string Opt::_optionNames[] = 
  {"age_weight_ratio",
   "allocation_buffer_price",
   "allocation_buffer_size",
   "arity_check",

   "backward_demodulation",
   "backward_subsumption",
   "basic_strategy",
   "bd_on_splitting_branches",

   "elim_def",
   "equality_proxy",
   "equality_resolution_with_deletion",

   "fast_demodulation_by_commutativity",
   "fd_on_splitting_branches",
   "forward_demodulation",
   "forward_subsumption",
   "forward_subsumption_resolution",
   "forward_subsumption_signature_filters",
   "function_definition_elimination",

   "ground_inequality_splitting",

   "include",
   "inherited_negative_selection",
   "input_file",
   "input_syntax",

   "latex_output",
   "literal_comparison_mode",
   "log_file",
   "lrs_first_memory_check",
   "lrs_first_time_check",
   "lrs_softness",

   "main_alg",
   "main_premise_demodulation",
   "main_premise_subsumption",
   "max_active",
   "max_inference_depth",
   "max_passive",
   "max_skolem",
   "max_weight",
   "memory_limit",
   "miniscope",
   "mode",

   "no_inferences_between_rules",
   "nongoal_weight_coefficient",
   "normalize",

   "orphan_murder",
   "output_equality_infix",
   "output_messages",
   "output_syntax",

   "paramodulation_from_variables",
   "paramodulation",
   "proof",
   "proof_checking",

   "random_seed",

   "selected_equality_weight_coefficient",
   "selection",
   "show_active",
   "show_generated",
   "show_input",
   "show_options",
   "show_passive",
   "show_preprocessed",
   "show_profile",
   "show_weight_limit_changes",
   "side_premise_demodulation",
   "side_premise_subsumption",
   "simplification_ordering",
   "simultaneous_backward_superposition",
   "sos",
   "splitting",
   "splitting_style",
   "statistics",
   "symbol_precedence",
   "symbol_weight_by_arity",

   "term_weighting_scheme",
   "test_id",
   "test_number",
   "time_limit",
   "transitivity_chaining",

   "unused_definition_removal",

   "weight_increment"
}; 



int Opt::numberOfOptions ()
{
  TRACER("Opt::numberOfOptions");

  return NUMBER_OF_OPTIONS;
} 



Opt::Opt (Tag tag, Opt** registry)
  : _tag (tag), 
    _changed (false)
{
  TRACER("Opt::Opt (Tag tag, Opt** registry)");

  registry[tag] = this;
} 



XML::Element* Options::toXML () const
{
  TRACER("Options::toXML ()");

  XML::DeepElement* xml = new XML::DeepElement("options");

  int numberOfOptions = Opt::numberOfOptions();
  for (int i = 0; i < numberOfOptions; i++) {
    xml->addChild(_allOptions[i]->toXML());
  }

  return xml;
} 



XML::Element* Option<int>::toXML () const
{
  TRACER("Option<int>::toXML ()");

  return new XML::IntegerElement(name(),_value);
} 



XML::Element* Option<Options::Selection>::toXML () const
{
  TRACER("Option<Selection>::toXML ()");

  return new XML::IntegerElement(name(),_value);
} 



XML::Element* Option<long>::toXML () const
{
  TRACER("Option<long>::toXML ()");

  return new XML::LongElement(name(),_value);
} 



XML::Element* Option<float>::toXML () const
{
  TRACER("Option<float>::toXML ()");

  return new XML::FloatElement(name(),_value);
} 


XML::Element* Option<bool>::toXML () const
{
  TRACER("Option<bool>::toXML ()");

  return new XML::TextElement(name(),_value ? "on" : "off");
} 



int Opt::find (const char* what, string* where, int numberOfElements)
{
  TRACER("Opt::find");

  int from = 0;
  int to = numberOfElements;
  while (from < to) {
    int mid = (from + to) / 2;
    int comp = where[mid].compare(what);
    if (comp == 0) { 
      return mid;
    }
    if (comp > 0) {
      to = mid;
    }
    else {
      from = mid+1;
    }
  }
  
  return -1;
} 



bool Option<Options::InputSyntax>::readValue (const char* val)
{
  TRACER("Option<Options::InputSyntax>::readValue");

  int found = find(val, _inputSyntaxValues,
		   sizeof(_inputSyntaxValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::InputSyntax)found;
  _changed = true;
  return true;
} 



bool Option<Options::FunctionDefinitionElimination>::readValue (const char* val)
{
  TRACER("Option<Options::FunctionDefinitionElimination>::readValue");

  int found = find(val, _functionDefinitionEliminationValues,
		   sizeof(_functionDefinitionEliminationValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::FunctionDefinitionElimination)found;
  _changed = true;
  return true;
} 



void Option<Options::InputSyntax>::toStream (ostream& str) const
{
  TRACER("Option<Options::InputSyntax>::toStream");

  str << _inputSyntaxValues[_value];
} 



void Option<Options::FunctionDefinitionElimination>::toStream (ostream& str) const
{
  TRACER("Option<Options::FunctionDefinitionElimination>::toStream");

  str << _functionDefinitionEliminationValues[_value];
} 



XML::Element* Option<Options::InputSyntax>::toXML () const
{
  TRACER("Option<Options::InputSyntax>::toXML ()");

  return new XML::TextElement(name(),_inputSyntaxValues[_value]);
} 



XML::Element* Option<Options::FunctionDefinitionElimination>::toXML () const
{
  TRACER("Option<Options::FunctionDefinitionElimination>::toXML ()");

  return new XML::TextElement(name(),_functionDefinitionEliminationValues[_value]);
} 



bool Option<Options::MainAlg>::readValue (const char* val)
{
  TRACER("Option<Options::MainAlg>::readValue");

  int found = find(val, _mainAlgValues,
		   sizeof(_mainAlgValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::MainAlg)found;
  _changed = true;
  return true;
} 



void Option<Options::MainAlg>::toStream (ostream& str) const
{
  TRACER("Option<Options::MainAlg>::toStream");

  str << _mainAlgValues[_value];
} 



XML::Element* Option<Options::MainAlg>::toXML () const
{
  TRACER("Option<Options::MainAlg>::toXML ()");

  return new XML::TextElement(name(),_mainAlgValues[_value]);
} 



bool Option<Options::Mode>::readValue (const char* val)
{
  TRACER("Option<Options::Mode>::readValue");

  int found = find(val, _modeValues,
		   sizeof(_modeValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::Mode)found;
  _changed = true;
  return true;
} 



void Option<Options::Mode>::toStream (ostream& str) const
{
  TRACER("Option<Options::Mode>::toStream");

  str << _modeValues[_value];
} 



XML::Element* Option<Options::Mode>::toXML () const
{
  TRACER("Option<Options::Mode>::toXML ()");

  return new XML::TextElement(name(),_modeValues[_value]);
} 


bool Option<Options::OutputSyntax>::readValue (const char* val)
{
  TRACER("Option<Options::OutputSyntax>::readValue");

  int found = find(val, _outputSyntaxValues,
		   sizeof(_outputSyntaxValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::OutputSyntax)found;
  _changed = true;
  return true;
} 



void Option<Options::OutputSyntax>::toStream (ostream& str) const
{
  TRACER("Option<OutputSyntax>::toStream");
  str << _outputSyntaxValues[_value];
} 



XML::Element* Option<Options::OutputSyntax>::toXML () const
{
  TRACER("Option<OutputSyntax>::toXML ()");

  return new XML::TextElement(name(),_outputSyntaxValues[_value]);
} 



bool Option<Options::SimplificationOrdering>::readValue (const char* val)
{
  TRACER("Option<Options::SimplificationOrdering>::readValue");

  int found = find(val, _simplificationOrderingValues,
		   sizeof(_simplificationOrderingValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::SimplificationOrdering)found;
  _changed = true;
  return true;
} 



void Option<Options::SimplificationOrdering>::toStream (ostream& str) const
{
  TRACER("Option<Options::SimplificationOrdering>::toStream");

  str << _simplificationOrderingValues[_value];
} 



XML::Element* Option<Options::SimplificationOrdering>::toXML () const
{
  TRACER("Option<Options::SimplificationOrdering>::toXML ()");

  return new XML::TextElement(name(),_simplificationOrderingValues[_value]);
} 



bool Option<Options::Splitting>::readValue (const char* val)
{
  TRACER("Option<Options::Splitting>::readValue");

  int found = find(val, _splittingValues,
		   sizeof(_splittingValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::Splitting)found;
  _changed = true;
  return true;
} 



void Option<Options::Splitting>::toStream (ostream& str) const
{
  TRACER("Option<Options::Splitting>::toStream");

  str << _splittingValues[_value];
} 



XML::Element* Option<Options::Splitting>::toXML () const
{
  TRACER("Option<Options::Splitting>::toXML ()");

  return new XML::TextElement(name(),_splittingValues[_value]);
} 



bool Option<Options::SplittingStyle>::readValue (const char* val)
{
  TRACER("Option<Options::SplittingStyle>::readValue");

  int found = find(val, _splittingStyleValues,
		   sizeof(_splittingStyleValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::SplittingStyle)found;
  _changed = true;
  return true;
} 



void Option<Options::SplittingStyle>::toStream (ostream& str) const
{
  TRACER("Option<Options::SplittingStyle>::toStream");

  str << _splittingStyleValues[_value];
} 



XML::Element* Option<Options::SplittingStyle>::toXML () const
{
  TRACER("Option<Options::SplittingStyle>::toXML ()");

  return new XML::TextElement(name(),_splittingStyleValues[_value]);
} 



bool Option<Options::Statistics>::readValue (const char* val)
{
  TRACER("Option<Options::Statistics>::readValue");

  int found = find(val, _statisticsValues,
		   sizeof(_statisticsValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::Statistics)found;
  _changed = true;
  return true;
} 



void Option<Options::Statistics>::toStream (ostream& str) const
{
  TRACER("Option<Options::Statistics>::toStream");

  str << _statisticsValues[_value];
} 



XML::Element* Option<Options::Statistics>::toXML () const
{
  TRACER("Option<Options::Statistics>::toXML ()");

  return new XML::TextElement(name(),_statisticsValues[_value]);
} 



bool Option<Options::SymbolPrecedence>::readValue (const char* val)
{
  TRACER("Option<Options::SymbolPrecedence>::readValue");

  int found = find(val, _symbolPrecedenceValues,
		   sizeof(_symbolPrecedenceValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::SymbolPrecedence)found;
  _changed = true;
  return true;
} 



void Option<Options::SymbolPrecedence>::toStream (ostream& str) const
{
  TRACER("Option<Options::SymbolPrecedence>::toStream");

  str << _symbolPrecedenceValues[_value];
} 



XML::Element* Option<Options::SymbolPrecedence>::toXML () const
{
  TRACER("Option<Options::SymbolPrecedence>::toXML ()");

  return new XML::TextElement(name(),_symbolPrecedenceValues[_value]);
} 



bool Option<Options::TermWeightingScheme>::readValue (const char* val)
{
  TRACER("Option<Options::TermWeightingScheme>::readValue");

  int found = find(val, _termWeightingSchemeValues,
		   sizeof(_termWeightingSchemeValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::TermWeightingScheme)found;
  _changed = true;
  return true;
} 



void Option<Options::TermWeightingScheme>::toStream (ostream& str) const
{
  TRACER("Option<Options::TermWeightingScheme>::toStream");

  str << _termWeightingSchemeValues[_value];
} 



XML::Element* Option<Options::TermWeightingScheme>::toXML () const
{
  TRACER("Option<Options::TermWeightingScheme>::toXML ()");

  return new XML::TextElement(name(),_termWeightingSchemeValues[_value]);
} 



bool Option<Options::TransitivityChaining>::readValue (const char* val)
{
  TRACER("Option<Options::TransitivityChaining>::readValue");

  int found = find(val, _transitivityChainingValues,
		   sizeof(_transitivityChainingValues)/sizeof(string));
  if (found == -1) {
    return false;
  }

  _value = (Options::TransitivityChaining)found;
  _changed = true;
  return true;
} 



void Option<Options::TransitivityChaining>::toStream (ostream& str) const
{
  TRACER("Option<Options::TransitivityChaining>::toStream");

  str << _transitivityChainingValues[_value];
} 



XML::Element* Option<Options::TransitivityChaining>::toXML () const
{
  TRACER("Option<Options::TransitivityChaining>::toXML ()");

  return new XML::TextElement(name(),_transitivityChainingValues[_value]);
} 



const string& Options::modeName ()
{
  TRACER("Options::modeName");

  return Opt::modeValues()[mode()];
} 



Options::StringIntMap::StringIntMap (StringInt* map,int length)
  : _map(map),
    _length(length)
{
}


bool Options::StringIntMap::find (const char* what,int& found)
{
  TRACER("Options::StringIntMap::find");

  int from = 0;
  int to = _length;
  while (from < to) {
    int mid = (from + to) / 2;
    int comp = _map[mid].str.compare(what);
    if (comp == 0) { 
      found = _map[mid].integer;
      return true;
    }
    if (comp > 0) {
      to = mid;
    }
    else {
      from = mid+1;
    }
  }
  
  return -1;
} 



void Options::setAgeWeightRatio (int weight)
{
  AgeWeight aw(weight);
  setAgeWeightRatio(aw);
} 



bool Option<Options::AgeWeight>::readValue (const char* val)
{
  TRACER("Option<AgeWeight>::readValue");

  
  bool found = false;
  int colonIndex = 0;
  while (val[colonIndex]) {
    if (val[colonIndex] == ':') {
      found = true;
      break;
    }
    colonIndex++;
  }

  int age;
  int weight;

  if (found) {
    char num[20];
    strncpy(num,val,colonIndex);
    num[colonIndex] = 0;
    char* endptr = 0;
    age = (int)(strtol (num, &endptr, 10));

    if (*endptr) { 
      return false;
    }

    strcpy(num,val+colonIndex+1);
    endptr = 0;
    weight = (int)(strtol (num, &endptr, 10));

    if (*endptr) { 
      return false;
    }
  }
  else {
    char* endptr = 0;
    weight = (int)(strtol (val, &endptr, 10));

    if (*endptr) { 
      return false;
    }
    age = 1;
  }

  Options::AgeWeight aw(age,weight);
  setValue(aw);

  return true;
} 



void Option<Options::AgeWeight>::toStream (ostream& str) const
{
  TRACER("Option<AgeWeight>::toStream");

  Options::AgeWeight aw = getValue();





  str << aw.age << ':' << aw.weight;
} 



XML::Element* Option<Options::AgeWeight>::toXML () const
{
  TRACER("Option<Options::AgeWeight>::toXML ()");

  Options::AgeWeight aw = getValue();
  string str = Int::toString(aw.age) + ':' + Int::toString(aw.weight);

  return new XML::TextElement(name(),str);
} 



void Options::readFromTestId (string testId)
{
  setNormalize(true);
  setTestId(testId);

  string ma(testId,0,3); 
  if (ma == "dis") {
    setMainAlg(DISCOUNT);
  }
  else if (ma == "lrs") {
    setMainAlg(LRS);
  }
  else if (ma == "ott") {
    setMainAlg(OTTER);
  }
  else {
  error: USER_ERROR("bad test id " + testId);
  }

  
  unsigned index = testId.find_last_of('_');
  if (index == string::npos) { 
    goto error;
  }
  _timeLimit.readValueOrFail(testId.substr(index+1).c_str());
  
  setTimeLimitInSeconds (timeLimitInDeciseconds());

  testId = testId.substr(3,index-2);
  switch (testId[0]) {
  case '+':
    testId = testId.substr(1);
    break;
  case '-':
    break;
  default:
    goto error;
  }

  index = testId.find('_');
  _selection.readValueOrFail(testId.substr(0,index).c_str());
  testId = testId.substr(index+1);

  if (testId == "") {
    goto error;
  }

  index = testId.find('_');
  _ageWeightRatio.readValueOrFail(testId.substr(0,index).c_str());
  testId = testId.substr(index+1);
  
  while (testId != "") {
    unsigned index1 = testId.find('=');
    if (index1 == string::npos) {
      goto error;
    }
    index = testId.find('_');
    if (index1 > index) {
      goto error;
    }
    Opt* opt = findByShortName(testId.substr(0,index1).c_str());
    if (! opt) {
      goto error;
    }
    string value = testId.substr(index1+1,index-index1-1);
    if (opt == &_literalComparisonMode) {
      if (value == "p") {
	value = "predicate";
      }
      else if (value == "k") {
	value = "kinky";
      }
    }
    opt->readValueOrFail(value.c_str());
    testId = testId.substr(index+1);
  }
} 



string Options::AgeWeight::toString () const
{
  if (age == 1) {
    return Int::toString(weight);
  }
  return Int::toString(age) + ':' + Int::toString(weight);
} 


void Opt::readValueOrFail (const char* value)
{
  if (! readValue(value)) {
    USER_ERROR((string)"incorrect value  " + value + " for " + name());
  }
}

