
/*
 * File Options.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Options__
#define __Options__


namespace XML {
  class Element;
};


#include <string>
#include "Allocator.hpp"


class Opt {
 public:
  
  enum Tag {
    AGE_WEIGHT_RATIO = 0,
    ALLOCATION_BUFFER_PRICE = 1,
    ALLOCATION_BUFFER_SIZE = 2,
    ARITY_CHECK = 3,

    BACKWARD_DEMODULATION = 4,
    BACKWARD_SUBSUMPTION = 5,
    BASIC_STRATEGY = 6,
    BD_ON_SPLITTING_BRANCHES = 7,

    ELIM_DEF = 8,
    EQUALITY_PROXY = 9,
    EQUALITY_RESOLUTION_WITH_DELETION = 10,

    FAST_DEMODULATION_BY_COMMUTATIVITY = 11,
    FD_ON_SPLITTING_BRANCHES = 12,
    FORWARD_DEMODULATION = 13,
    FORWARD_SUBSUMPTION = 14,
    FORWARD_SUBSUMPTION_RESOLUTION = 15,
    FORWARD_SUBSUMPTION_SIGNATURE_FILTERS = 16,
    FUNCTION_DEFINITION_ELIMINATION = 17,

    GROUND_INEQUALITY_SPLITTING = 18,

    INCLUDE = 19,
    INHERITED_NEGATIVE_SELECTION = 20,
    INPUT_FILE = 21,
    INPUT_SYNTAX = 22,

    LATEX_OUTPUT = 23,
    LITERAL_COMPARISON_MODE = 24,
    LOG_FILE = 25,
    LRS_FIRST_MEMORY_CHECK = 26,
    LRS_FIRST_TIME_CHECK = 27,
    LRS_SOFTNESS = 28,

    MAIN_ALG = 29,
    MAIN_PREMISE_DEMODULATION = 30,
    MAIN_PREMISE_SUBSUMPTION = 31,
    MAX_ACTIVE = 32,
    MAX_INFERENCE_DEPTH = 33,
    MAX_PASSIVE = 34,
    MAX_SKOLEM = 35,
    MAX_WEIGHT = 36,
    MEMORY_LIMIT = 37,
    MINISCOPE = 38,
    MODE = 39,

    NO_INFERENCES_BETWEEN_RULES = 40,
    NONGOAL_WEIGHT_COEFFICIENT = 41,
    NORMALIZE = 42,

    ORPHAN_MURDER = 43,
    OUTPUT_EQUALITY_INFIX = 44,
    OUTPUT_MESSAGES = 45,
    OUTPUT_SYNTAX = 46,

    PARAMODULATION = 47,
    PARAMODULATION_FROM_VARIABLES = 48,
    PROOF = 49,
    PROOF_CHECKING = 50,

    RANDOM_SEED = 51,

    SELECTED_EQUALITY_WEIGHT_COEFFICIENT = 52,
    SELECTION = 53,
    SHOW_ACTIVE = 54,
    SHOW_GENERATED = 55,
    SHOW_INPUT = 56,
    SHOW_OPTIONS = 57,
    SHOW_PASSIVE = 58,
    SHOW_PREPROCESSED = 59,
    SHOW_PROFILE = 60,
    SHOW_WEIGHT_LIMIT_CHANGES = 61,
    SIDE_PREMISE_DEMODULATION = 62,
    SIDE_PREMISE_SUBSUMPTION = 63,
    SIMPLIFICATION_ORDERING = 64,
    SIMULTANEOUS_BACKWARD_SUPERPOSITION = 65,
    SOS = 66,
    SPLITTING = 67,
    SPLITTING_STYLE = 68,
    STATISTICS = 69,
    SYMBOL_PRECEDENCE = 70,
    SYMBOL_WEIGHT_BY_ARITY = 71,

    TERM_WEIGHTING_SCHEME = 72,
    TEST_ID = 73,
    TEST_NUMBER = 74,
    TIME_LIMIT = 75,
    TRANSITIVITY_CHAINING = 76,

    UNUSED_DEFINITION_REMOVAL = 77,

    WEIGHT_INCREMENT = 78,

    NUMBER_OF_OPTIONS = 79 
  };

  Opt (Tag t, Opt** arrayOfOpts);

  
  virtual bool readValue (const char* value) = 0;
  void readValueOrFail (const char* value);
  
  virtual void toStream (ostream& str) const = 0;
  
  string name () const { return _optionNames[_tag]; }
  
  bool changed () const { return _changed; }
  void output (ostream& str) const;
  virtual XML::Element* toXML () const = 0;

  
  static int numberOfOptions();
  
  static string* optionNames() { return _optionNames; }
  
  static string* modeValues() { return _modeValues; }
  static int find (const char* what, string* where, int numberOfElements);

  
  void* operator new (size_t);
 protected:
  const Tag _tag;
  bool _changed;
  static string _optionNames [];
  static string _demodulationValues[];
  static string _subsumptionValues[];
  static string _literalComparisonModeValues[];
  static string _inputSyntaxValues[];
  static string _functionDefinitionEliminationValues[];
  static string _mainAlgValues[];
  static string _modeValues[];
  static string _outputSyntaxValues[];
  static string _simplificationOrderingValues[];
  static string _splittingValues[];
  static string _splittingStyleValues[];
  static string _statisticsValues[];
  static string _symbolPrecedenceValues[];
  static string _termWeightingSchemeValues[];
  static string _transitivityChainingValues[];
}; 



template <class C>
class Option :
  public Opt {
 public:
  
  inline Option (Opt::Tag t, const C val, Opt** registry)
    : Opt (t, registry),
      _value (val)
  {}
  virtual ~Option() {}
  virtual bool readValue (const char* val);
  virtual void toStream (ostream& str) const;
  virtual XML::Element* toXML () const;

  
  inline C getValue () const 
  { 
    return _value; 
  }

  
  inline void setValue (C newVal)
  { 
    _value = newVal; 
    _changed = true; 
  }

 private:
  C _value;
}; 



class Options 
{
 public:
  
  class StringInt {
   public:
    string str;
    int integer;
  };

  
  class StringIntMap {
  public:
    StringIntMap (StringInt* map,int length);
    bool find (const char* name,int& found);
  private:
    StringInt* _map;
    int _length;
  };

  
  struct AgeWeight {
    
    int age;
    
    int weight;
    
    explicit AgeWeight (int w) : age(1), weight(w) {}
    
    AgeWeight (int a, int w) : age(a), weight(w) {}
    string toString () const;
  };

  
  enum Mode {
    CASC_J2 = 0,
    KIF = 1,
    SPIDER = 2,
    VAMPIRE = 3
  };

  
  enum FunctionDefinitionElimination {
    FDE_ALL = 0,
    FDE_DUMMY = 1,
    FDE_NO = 2
  };

   
  enum Statistics {
    
    STATISTICS_BRIEF = 0,
    
    STATISTICS_DETAILED = 1,
    
    STATISTICS_NONE = 2
  };

  
  enum MainAlg {
    DISCOUNT = 0,
    LRS = 1,
    OTTER = 2
  };

  
  enum Splitting {
    SPLIT_OFF = 0,
    SPLIT_ON = 1,
    SPLIT_PREPRO_ONLY = 2
  };

  
  enum TransitivityChaining {
    TC_FULL = 0,
    TC_NONE = 1,
    TC_ONLY_NONVARIABLES
  };

  
  enum Subsumption {
    SUBSUMPTION_MULTISET = 0,
    SUBSUMPTION_OFF = 1,
    SUBSUMPTION_SET = 2
  };

  
  enum SplittingStyle {
    BLOCKING = 0,
    PARALLEL = 1
  };


  enum Selection {
    P_0 = 0,
    P_1 = 1,
    P_2 = 2,
    P_3 = 3,
    P_4 = 4,
    P_5 = 5,
    P_6 = 6,
    P_7 = 7,
    P_8 = 8,
    P_9 = 9,
    P_10 = 10,
    P_11 = 11,
    P_12 = 12,
    P_13 = 13,
    P_14 = 14,
    P_15 = 15,
    P_16 = 16,
    P_17 = 17,
    P_1002 = 1002,
    P_1003 = 1003,
    P_1004 = 1004,
    P_1005 = 1005,
    P_1006 = 1006,
    P_1007 = 1007,
    P_1008 = 1008,
    P_1009 = 1009,
    P_1010 = 1010,
    P_1011 = 1011,
    P_1012 = 1012,
    P_1013 = 1013,
    P_1014 = 1014,
    N_2 = -2,
    N_3 = -3,
    N_4 = -4,
    N_5 = -5,
    N_6 = -6,
    N_7 = -7,
    N_9 = -9,
    N_10 = -10,
    N_11 = -11,
    N_12 = -12,
    N_13 = -13,
    N_14 = -14,
    N_15 = -15,
    N_16 = -16,
    N_17 = -17,
    N_1005 = -1005,
    N_1009 = -1009,
    N_1012= -1012,
    N_1014= -1014
  };

  enum LiteralComparisonMode {
    KINKY = 0,
    NORMAL = 1,
    PREDICATE = 2
  };

  enum Demodulation {
    DEMODULATION_ALL = 0,
    DEMODULATION_OFF = 1,
    DEMODULATION_PREORDERED = 2
  };

  enum SymbolPrecedence {
    BY_ARITY = 0,
    BY_OCCURRENCE = 1,
    BY_REVERSE_ARITY = 2
  };

  enum SimplificationOrdering {
    KBO = 0,
    KBO_NONREC = 1
  };

  enum TermWeightingScheme {
    TWS_NONUNIFORM_CONSTANT = 0,
    TWS_UNIFORM =1
  };

  enum InputSyntax {
    INPUT_KIF = 0,
    INPUT_TPTP = 1
  };

  
  enum OutputSyntax {
    OS_KIF = 0,
    OS_NATIVE = 1,
    OS_TPTP = 2,
    OS_XML = 3
  };


 public:
  Options ();
  void correct (int argc, char* argv []);
  void output (ostream&) const;
  void readFromTestId (string testId);

  string problemName () const;

  
  OutputSyntax outputSyntax () const
    { return _outputSyntax.getValue(); }
  string testId () const 
    { return _testId.getValue(); }
  int testNumber () const 
    { return _testNumber.getValue(); }
  Statistics statistics () const 
    { return _statistics.getValue(); }
  bool proof () const 
    { return _proof.getValue(); }
  bool proofChecking () const 
    { return _proofChecking.getValue(); }
  bool miniscope () const 
    { return _miniscope.getValue (); }
  bool normalize () const 
    { return _normalize.getValue(); }
  Mode mode () const 
    { return _mode.getValue(); }
  string include () const 
    { return _include.getValue(); }
  string logFile () const 
    { return _logFile.getValue(); }
  string inputFile () const 
    { return _inputFile.getValue(); }
  int symbolWeightByArity () const 
    { return _symbolWeightByArity.getValue(); }
  int randomSeed () const 
    { return _randomSeed.getValue(); }
  InputSyntax inputSyntax () const 
    { return _inputSyntax.getValue(); }
  bool outputMessages () const 
    { return _outputMessages.getValue(); }
  bool showGenerated () const 
    { return _showGenerated.getValue(); }
  bool showPassive () const 
    { return _showPassive.getValue(); }
  bool showActive () const 
    { return _showActive.getValue(); }
  bool showProfile () const 
    { return _showProfile.getValue(); }
  bool showPreprocessed () const 
    { return _showPreprocessed.getValue(); }
  bool showInput () const 
    { return _showInput.getValue(); }
  bool showOptions () const 
    { return _showOptions.getValue(); }
  bool showWeightLimitChanges () const 
    { return _showWeightLimitChanges.getValue(); }
  bool unusedDefinitionRemoval () const
    { return _unusedDefinitionRemoval.getValue (); }
  bool weightIncrement () const
    { return _weightIncrement.getValue (); }
  bool outputEqualityInfix () const 
    { return _outputEqualityInfix.getValue (); }
  bool arityCheck () const 
    { return _arityCheck.getValue (); }
  MainAlg mainAlg () const 
    { return _mainAlg.getValue(); }
  Splitting splitting () const 
    { return _splitting.getValue(); }
  SplittingStyle splittingStyle () const 
    { return _splittingStyle.getValue(); }
  Selection selection () const 
    { return _selection.getValue(); }
  string latexOutput () const 
    { return _latexOutput.getValue(); }
  LiteralComparisonMode literalComparisonMode () const 
    { return _literalComparisonMode.getValue(); }
  bool inheritedNegativeSelection () const 
    { return _inheritedNegativeSelection.getValue(); }
  bool forwardSubsumptionResolution () const 
    { return _forwardSubsumptionResolution.getValue(); }
  Demodulation forwardDemodulation () const 
    { return _forwardDemodulation.getValue(); }
  bool fdOnSplittingBranches () const 
    { return _fdOnSplittingBranches.getValue(); }
  Demodulation backwardDemodulation () const 
    { return _backwardDemodulation.getValue(); }
  bool basicStrategy () const 
    { return _basicStrategy.getValue(); }
  bool bdOnSplittingBranches () const 
    { return _bdOnSplittingBranches.getValue(); }
  Subsumption backwardSubsumption () const 
    { return _backwardSubsumption.getValue(); }
  Subsumption forwardSubsumption () const 
    { return _forwardSubsumption.getValue(); }
  bool orphanMurder () const 
    { return _orphanMurder.getValue(); }
  int elimDef () const 
    { return _elimDef.getValue(); }
  int lrsFirstTimeCheck () const 
    { return _lrsFirstTimeCheck.getValue(); }
  float lrsSoftness () const 
    { return _lrsSoftness.getValue(); }
  int lrsFirstMemoryCheck () const 
    { return _lrsFirstMemoryCheck.getValue(); }
  int maxInferenceDepth () const 
    { return _maxInferenceDepth.getValue(); }
  SymbolPrecedence symbolPrecedence () const 
    { return _symbolPrecedence.getValue(); }
  bool noInferencesBetweenRules () const 
    { return _noInferencesBetweenRules.getValue(); }
  int timeLimitInDeciseconds () const 
    { return _timeLimit.getValue(); }
  int memoryLimit () const 
    { return _memoryLimit.getValue(); }
  int allocationBufferSize () const 
    { return _allocationBufferSize.getValue(); }
  float allocationBufferPrice () const 
    { return _allocationBufferPrice.getValue(); }
  int groundInequalitySplitting () const
    { return _groundInequalitySplitting.getValue(); }
  int maxSkolem () const 
    { return _maxSkolem.getValue(); }
  long maxActive () const 
    { return _maxActive.getValue(); }
  long maxPassive () const 
    { return _maxPassive.getValue(); }
  int maxWeight () const 
    { return _maxWeight.getValue(); }
  AgeWeight ageWeightRatio () const 
    { return _ageWeightRatio.getValue(); }
  bool paramodulation () const 
    { return _paramodulation.getValue(); }
  bool paramodulationFromVariables () const 
    { return _paramodulationFromVariables.getValue(); }
  bool equalityProxy () const 
    { return _equalityProxy.getValue(); }
  bool equalityResolutionWithDeletion () const 
    { return _equalityResolutionWithDeletion.getValue(); }
  TermWeightingScheme termWeightingScheme () const 
    { return _termWeightingScheme.getValue (); }
  SimplificationOrdering simplificationOrdering () const 
    { return _simplificationOrdering.getValue(); }
  float nongoalWeightCoefficient () const 
    { return _nongoalWeightCoefficient.getValue(); }
  float selectedEqualityWeightCoefficient () const 
    { return _selectedEqualityWeightCoefficient.getValue(); }
  bool sos () const 
    { return _sos.getValue(); }
  bool fastDemodulationByCommutativity () const 
    { return _fastDemodulationByCommutativity.getValue(); }
  bool forwardSubsumptionSignatureFilters () const 
    { return _forwardSubsumptionSignatureFilters.getValue(); }
  FunctionDefinitionElimination functionDefinitionElimination () const 
    { return _functionDefinitionElimination.getValue(); }
  bool mainPremiseDemodulation () const 
    { return _mainPremiseDemodulation.getValue() != 0; }
  bool mainPremiseSubsumption () const 
    { return _mainPremiseSubsumption.getValue() != 0; }
  bool sidePremiseDemodulation () const 
    { return _sidePremiseDemodulation.getValue() !=0; }
  bool sidePremiseSubsumption () const 
    { return _sidePremiseSubsumption.getValue() != 0; }
  bool simultaneousBackwardSuperposition () const 
    { return _simultaneousBackwardSuperposition.getValue(); }
  TransitivityChaining transitivityChaining () const 
    { return _transitivityChaining.getValue(); }

  
  void setFdOnSplittingBranches (bool newVal)
    { _fdOnSplittingBranches.setValue(newVal); }
  void setBasicStrategy (bool newVal)
    { _basicStrategy.setValue(newVal); }
  void setBdOnSplittingBranches (bool newVal)
    { _bdOnSplittingBranches.setValue(newVal); }
  void setSplitting (Splitting newVal) 
    { _splitting.setValue(newVal); }
  void setSplittingStyle (SplittingStyle newVal) 
    { _splittingStyle.setValue(newVal); }
  void setStatistics (Statistics newVal) 
    { _statistics.setValue(newVal); } 
  void setMiniscope (bool newVal)
    { _miniscope.setValue (newVal); }
  void setShowOptions (bool newVal) 
    { _showOptions.setValue(newVal); }
  void setNormalize (bool newVal) 
    { _normalize.setValue(newVal); }
  void setLogFile (const string& newVal) 
    { _logFile.setValue(newVal); }
  void setArityCheck (bool newVal) 
    { _arityCheck.setValue (newVal); }
  void setOutputSyntax (OutputSyntax newVal) 
    { _outputSyntax.setValue(newVal); }
  void setInputSyntax (InputSyntax newVal) 
    { _inputSyntax.setValue(newVal); }
  void setBackwardSubsumption (Subsumption newVal) 
    { _backwardSubsumption.setValue(newVal); }
  void setOutputMessages (bool newVal) 
    { _outputMessages.setValue(newVal); }
  void setNongoalWeightCoefficient (float newVal) 
    { _nongoalWeightCoefficient.setValue(newVal); }
  void setSos (bool newVal) 
    { _sos.setValue(newVal); }
  void setUnusedDefinitionRemoval (bool newVal) 
    { _unusedDefinitionRemoval.setValue (newVal); }
  void setWeightIncrement (bool newVal) 
    { _weightIncrement.setValue (newVal); }
  void setLrsSoftness (float newVal) 
    { _lrsSoftness.setValue(newVal); }
  void setMainAlg (MainAlg alg) 
    { _mainAlg.setValue(alg); }
  void setMemoryLimit (int newVal) 
    { _memoryLimit.setValue (newVal); }
  void setSelection (Selection s) 
    { _selection.setValue (s); }
  void setLiteralComparisonMode (LiteralComparisonMode lcm) 
    { _literalComparisonMode.setValue(lcm); }
  void setLatexOutput (string newVal) 
    { _latexOutput.setValue(newVal); }
  void setInputFile (const string& newVal) 
    { _inputFile.setValue(newVal); }
  void setTestId (string newVal) 
    { _testId.setValue (newVal); }
  void setTimeLimitInSeconds (int newVal) 
    { _timeLimit.setValue (10*newVal); }
  void setTimeLimitInDeciseconds (int newVal) 
    { _timeLimit.setValue (newVal); }
  void setElimDef (int newVal) 
    { _elimDef.setValue(newVal); }
  void setAgeWeightRatio (AgeWeight newVal) 
    { _ageWeightRatio.setValue(newVal); }
  void setAgeWeightRatio (int weight);
  void setSymbolPrecedence (SymbolPrecedence newVal)
    { _symbolPrecedence.setValue(newVal); }
  void setForwardDemodulation (Demodulation newVal) 
    { _forwardDemodulation.setValue(newVal); }
  void setMaxWeight (long newVal)
    { _maxWeight.setValue(newVal); }
  void setSimplificationOrdering (SimplificationOrdering newVal)
    { _simplificationOrdering.setValue(newVal); }

  
  void kif ();
  void fullSplitting ();

  
  string includeFileName (string relativeName); 
  bool testNumberChanged () const { return _testNumber.changed(); }
  XML::Element* toXML() const;
  const string& modeName();

 private:
  Option <AgeWeight> _ageWeightRatio;
  Option <float> _allocationBufferPrice;
  Option <int> _allocationBufferSize;
  Option <bool> _arityCheck;

  Option <Demodulation> _backwardDemodulation;
  Option <Subsumption> _backwardSubsumption;
  Option<bool> _basicStrategy;
  Option<bool> _bdOnSplittingBranches;

  Option <int> _elimDef;
  Option <bool> _equalityProxy;
  Option <bool> _equalityResolutionWithDeletion;

  Option<bool> _fastDemodulationByCommutativity;
  Option<bool> _fdOnSplittingBranches;
  Option <Demodulation> _forwardDemodulation;
  Option <Subsumption> _forwardSubsumption;
  Option <bool> _forwardSubsumptionResolution;
  Option <bool> _forwardSubsumptionSignatureFilters;
  Option<FunctionDefinitionElimination> _functionDefinitionElimination;
  Option <int> _groundInequalitySplitting;

  Option <string> _include;
  Option <bool> _inheritedNegativeSelection;
  Option <string> _inputFile;
  Option <InputSyntax> _inputSyntax;

  Option <string> _latexOutput;
  Option <LiteralComparisonMode> _literalComparisonMode;
  Option <string> _logFile;
  Option <int> _lrsFirstMemoryCheck;
  Option <int> _lrsFirstTimeCheck;
  Option <float> _lrsSoftness;

  Option <MainAlg> _mainAlg;
  Option <int> _mainPremiseDemodulation;
  Option <int> _mainPremiseSubsumption;
  Option <long> _maxActive;
  Option <int> _maxInferenceDepth;
  Option <long> _maxPassive;
  Option <int> _maxSkolem;
  Option <int> _maxWeight;
  Option <int> _memoryLimit;
  Option <bool> _miniscope;
  Option <Mode> _mode;

  Option <bool> _noInferencesBetweenRules;
  Option <float> _nongoalWeightCoefficient;
  Option <bool> _normalize;

  Option <bool> _orphanMurder;
  Option <bool> _outputEqualityInfix;
  Option <bool> _outputMessages;
  Option <OutputSyntax> _outputSyntax;

  Option <bool> _paramodulation;
  Option <bool> _paramodulationFromVariables;
  Option <bool> _proof;
  Option <bool> _proofChecking;

  Option <int> _randomSeed;

  Option <float> _selectedEqualityWeightCoefficient;
  Option <Selection> _selection;
  Option <bool> _showActive;
  Option <bool> _showGenerated;  
  Option <bool> _showInput;
  Option <bool> _showOptions;
  Option <bool> _showPassive;  
  Option <bool> _showPreprocessed;
  Option <bool> _showProfile;
  Option <bool> _showWeightLimitChanges;  
  Option <int> _sidePremiseDemodulation;
  Option <int> _sidePremiseSubsumption;
  Option <SimplificationOrdering> _simplificationOrdering;
  Option <bool> _simultaneousBackwardSuperposition;
  Option <bool> _sos;
  Option <Splitting> _splitting;
  Option <SplittingStyle> _splittingStyle;
  Option <Statistics> _statistics;
  Option <SymbolPrecedence> _symbolPrecedence;
  Option <int> _symbolWeightByArity;

  Option <TermWeightingScheme> _termWeightingScheme;
  Option <string> _testId;
  Option <int> _testNumber;
  
  Option <int> _timeLimit;
  Option <TransitivityChaining> _transitivityChaining;

  Option <bool> _unusedDefinitionRemoval;

  Option <bool> _weightIncrement;

  Opt* find (const char*); 
  Opt* findByShortName (const char*); 
  
  Opt* _allOptions[Opt::NUMBER_OF_OPTIONS];

public:
  
  static Options globalOptions;
}; 


inline
ostream& operator << (ostream& str, const Options& opts)
{
  opts.output(str);

  return str;
} 


#endif

