
/*
 * File Preprocess.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "CNF.hpp"
#include "ClausalDefinition.hpp"
#include "Compare.hpp"
#include "Definition.hpp"
#include "EqualityAxioms.hpp"
#include "EqualityProxy.hpp"
#include "FunctionDefinition.hpp"
#include "Flattening.hpp"
#include "LiteralOrderingFinder.hpp"
#include "Miniscope.hpp"
#include "NNF.hpp"
#include "Normalisation.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "TheoryFinder.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "PureLiteral.hpp"
#include "Rectify.hpp"
#include "Signature.hpp"
#include "Skolem.hpp"
#include "Sort.hpp"
#include "SymCounter.hpp"
#include "Tautology.hpp"
#include "TPTP.hpp"
#include "Tracer.hpp"
#include "XML.hpp"



void Preprocess::preprocess (Problem& problem,
			     const Options& opt,
			     const Property& property)
{
  TRACER("Preprocess::preprocess (Problem& problem, Options& opt)");

  
  if (opt.arityCheck()) {
    Signature::sig.arityCheck();
  }
  
  
  if (opt.normalize()) {
    Compare::createSymbolCounter(problem);
  }

  UnitChain::DelIterator us (problem.giveUnits());
  while (us.hasNext()) {
    Unit u (us.next());
    Unit v(u);
    preprocess1(v,opt,problem);
    if (v != u) {
      us.replace(v);
    }
  }

  if (Signature::sig.hasAnswerAtom()) {
    problem.setAnswerLiteral(Literal(true,
				     Signature::sig.getAnswerAtom()));
  }

  
  if (opt.normalize()) {
    normalise(problem);
  }





  
  
  switch (opt.mode()) 
    {
    case Options::KIF: 
      break;

    default:
      removeEqualityAxioms(problem);
      if (problem.hasFormulas() && 
 	  opt.unusedDefinitionRemoval()) {
	Definition::removeUnusedDefinitions(problem);
      }
      break;
    }

  UnitChain::DelIterator us2 (problem.giveUnits());
  while (us2.hasNext()) {
    Unit u (us2.next());
    Unit v (u);
    preprocess2(v,opt);
    if (v != u) {
      us2.replace(v);
    }
  }

  LiteralOrderingFinder lof(problem,opt,Signature::sig);
  lof.findOrdering();

  CNF::clausify (problem);
   if (opt.functionDefinitionElimination() == Options::FDE_ALL &&
       property.hasProp(PR_HAS_FUNCTION_DEFINITIONS)) {
    FunctionDefinition fd(problem,Signature::sig);
    fd.removeAllDefinitions();
  }

  
  UnitChain::DelIterator units (problem.giveUnits());
  while (units.hasNext()) {
    Unit newUnit;
    switch (Tautology::isTautology(units.next(),newUnit)) {
    case -1:
      units.replace(newUnit);
      break;

    case 0:
      break;

    case 1:
      units.del();
      break;
    }
  }

  PureLiteral pure(problem,Signature::sig);
  pure.findAndRemove();

  ClausalDefinition cdef(problem,Signature::sig);
  cdef.findAndReplace();

  if (opt.equalityProxy() && property.hasProp(PR_HAS_X_EQUALS_Y)) {
    EqualityProxy proxy(Signature::sig);
    proxy.apply(problem);
  }


} 



void Preprocess::preprocess1 (Unit& u, 
			      const Options& opt, 
			      Problem& problem)
{
  switch (u.unitType()) {
  case CLAUSE:
    if (opt.normalize()) {
      Normalisation::normalise(u);
    }
    return;
    
  case FORMULA:
    {
      
      VarList answerVariables;
      Rectify::rectify(u,answerVariables);
      problem.setOriginalAnswerVariables(answerVariables);

      Formula f (u.formula());
      
      Flattening::flatten(u);
      
      if (opt.normalize()) {
	Normalisation::normalise(u);
      }
      return;
    }

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void Preprocess::preprocess2 (Unit& unit, const Options& opt)
{
  switch (unit.unitType()) {
  case CLAUSE:
    return;
    
  case FORMULA:
    NNF::ennf(unit);
    
    Flattening::flatten(unit);
    
    
    

    
    NNF::nnf(unit);
    
    Flattening::flatten(unit);
    
    if (opt.miniscope()) {
      Miniscope::miniscope(unit);
    }
    Skolem::skolemise(unit);
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



void Preprocess::normalise (Problem& problem)
{
  TRACER ( "Preprocess::normalise" );
  UnitChain& units = problem.giveUnits();

  
  Compare cmp;
  Sort <Unit,Compare> sort (units.length(),cmp);

  
  UnitChain::DelIterator unit1 (units);
  while (unit1.hasNext()) {
    sort.add(unit1.next());
    unit1.del();
  }

  
  sort.sort ();

  
  for (int i = sort.length()-1;i >= 0;i--) {
    units.addFirst(sort[i]);
  }
} 



int Preprocess::removeEqualityAxioms (Problem& problem)
{
  TRACER("Preprocess::removeEqualityAxioms");
  int removed = 0;

  UnitChain::DelIterator us (problem.giveUnits());
  while (us.hasNext()) {
    if (EqualityAxioms::isEqualityAxiom(us.next())) {
      us.del();
      removed++;
    }
  }

  return removed;
} 











