
/*
 * File Pretty.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Formula.hpp"
#include "Pretty.hpp"
#include "Int.hpp"


#define INC 2


string Pretty::toString (const Formula& f)
{
  return toString(f,0);
} 


string Pretty::toString (const Formula& f, int n)
{
  if (isOneLine(f)) {
    return f.toString();
  }

  Connective c(f.connective());
  string con = Formula::toString(c);
  
  switch (c) {
  case AND:
  case OR:
    {
      FormulaList fs(f.args());
      string result = toStringInScopeOf(fs.head(),c,n);
      Iterator<Formula> forms(fs.tail());
      while (forms.hasNext()) {
	result += indent(n) + con + ' ' + 
	          toStringInScopeOf(forms.next(),c,n+INC);
      }
      return result;
    }

  case IMP:
  case IFF:
  case XOR:
    return toStringInScopeOf(f.left(),c,n+INC) +
           indent(n) + con + ' ' +
           toStringInScopeOf(f.right(),c,n+INC);

  case NOT:
    return con + toStringInScopeOf(f.uarg(),c,n);

  case FORALL:
  case EXISTS:
    {
      string result = string("(") + con;
      Iterator<Var> vs(f.vars());
      while (vs.hasNext()) {
	result += string(" ") + vs.next().toString();
      }
      result += ")";
      return result + 
	     toStringInScopeOf(f.qarg(),c,n+quantifierPrintLength(c,f.vars()));
    }

#if DEBUG_PREPRO
  case ATOM:
  default:
    ASS(false);
#endif
  }
} 



string Pretty::indent (int n)
{
  static char* source = 0;
  
  static int sourceLength = 0;

  if (sourceLength < n+2) { 
    
    sourceLength = Int::max(n+2,sourceLength+80);
    delete [] source;
    source = new char[sourceLength];
    if (! source) {
      ALLOCATION_FAILURE;
    }
    
    source[0] = '\n';
    for (int i = sourceLength-1;i > 0;i--) {
      
      source[i] = ' ';
    }
  }

  source[n+1] = '\0';
  string result = string(source);
  source[n+1] = ' ';

  return result;
} 



string Pretty::toStringInScopeOf (const Formula& f,
				  Connective outer,
				  int n)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toString(f,n+1) + ")" :
         toString(f,n);
} 



bool Pretty::isOneLine (const Formula& f)
{
  switch (f.connective()) {
  case ATOM:
    return true;

  case AND:
  case OR:
    {
      Iterator<Formula> fs(f.args());
      while (fs.hasNext()) {
	if (! isBuiltUsingUnaryConnectives(fs.next())) {
	  return false;
	}
      }
      return true;
    }

  case IMP:
  case IFF:
  case XOR:
    return isBuiltUsingUnaryConnectives(f.left()) &&
           isBuiltUsingUnaryConnectives(f.right());

  case NOT:
  case FORALL:
  case EXISTS:
    {
      Formula g(f);
      stripUnaryConnectives(g);
      switch (g.connective()) {
      case ATOM:
	return true;

      case AND:
      case OR:
	{
	  Iterator<Formula> gs(g.args());
	  while (gs.hasNext()) {
	    if (! isLiteral(gs.next())) {
	      return false;
	    }
	  }
	  return true;
	}

      case IMP:
      case IFF:
      case XOR:
	return isLiteral(g.left()) &&
	       isLiteral(g.right());

#if DEBUG_PREPRO
      default:
	ASS(false);
#endif
      }
    }

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool Pretty::isLiteral (const Formula& f)
{
  switch (f.connective()) {
  case ATOM:
    return true;

  case NOT:
    return f.uarg().connective() == ATOM;

  default:
    return false;
  }
} 


void Pretty::stripUnaryConnectives (Formula& f)
{
  for (;;) {
    switch (f.connective()) {
    case NOT:
      f = f.uarg();
      break;

    case FORALL:
    case EXISTS:
      f = f.qarg();
      break;

    default:
      return;
    }
  }
} 



bool Pretty::isBuiltUsingUnaryConnectives (const Formula& f)
{
  Formula g(f);
  stripUnaryConnectives(g);
  return g.connective() == ATOM;
} 



int Pretty::quantifierPrintLength (Connective Q,const VarList& vs)
{
  int result = 2 + Formula::toString(Q).length();
  Iterator<Var> ws(vs);
  while (ws.hasNext()) {
    result += 1 + ws.next().toString().length();
  }
  return result;
} 
