
/*
 * File Property.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "Int.hpp"
#include "Property.hpp"
#include "Problem.hpp"
#include "VarCounter.hpp"
#include "FunctionDefinition.hpp"


Property::Property (const Problem& p)
  : _problem(p),
    _goalClauses (0),
    _axiomClauses (0),
    _equalityAtoms (0),
    _atoms (0),
    _goalFormulas (0),
    _axiomFormulas (0),
    _subformulas (0),
    _terms (0),
    _unitGoals (0),
    _unitAxioms (0),
    _hornGoals (0),
    _hornAxioms (0),
    _equationalClauses (0),
    _pureEquationalClauses (0),
    _groundUnitAxioms (0),
    _positiveAxioms (0),
    _groundPositiveAxioms (0),
    _groundGoals (0),
    _maxFunArity (0), 
    _maxPredArity (0),
    _totalNumberOfVariables (0),
    _maxVariablesInClause (0),
    _props(0)
{
} 



void Property::scanProblem ()
{
  TRACER("Property::scan (Problem& p)");

  UnitChain::Iterator us (_problem.units());
  while (us.hasNext()) {
    scan (us.next());
  }
  
  if (formulas() > 0) { 
    if (_equalityAtoms == 0) {
      _category = FNE;
    }
    else {
      _category = FEQ;
    }
  }
  
  else if (_maxFunArity == 0) { 
    if (_pureEquationalClauses == clauses()) { 
      if (clauses() == unitClauses()) { 
	_category = UEQ;
      }
      else {
	_category = PEQ;
      }
    }
    else {
      _category = EPR;
    }
  }
  
  else if (_equationalClauses == 0 ) { 
    if (clauses() == hornClauses()) { 
      _category = HNE;
    }
    else {
      _category = NNE;
    }
  }
  
  else if (_pureEquationalClauses == clauses()) { 
    if (clauses() == unitClauses()) { 
      _category = UEQ;
    }
    else {
      _category = PEQ;
    }
  }
  
  else if (clauses() == hornClauses()) { 
    _category = HEQ;
  }
  else {
    _category = NEQ;
  }
} 



void Property::scan (const Unit& u)
{
  TRACER("Property::scan(const Unit&)");

  switch (u.unitType()) 
    {
    case CLAUSE:
      {
	const Clause& c = u.clause();
	scan (c, u.inputType() == AXIOM);
	if ((! hasProp(PR_HAS_X_EQUALS_Y)) && hasXEqualsY(c)) {
	  addProp(PR_HAS_X_EQUALS_Y);
	}
	if ((! hasProp(PR_HAS_FUNCTION_DEFINITIONS)) && 
	    FunctionDefinition::isFunctionDefinition(c)) {
	  addProp(PR_HAS_FUNCTION_DEFINITIONS);
	}
	return;
      }

    case FORMULA:
      {
	if (u.inputType() == AXIOM) {
	  _axiomFormulas ++;
	}
	else {
	  _goalFormulas++;
	}
	const Formula& f = u.formula();
	scan(f);
	if (! hasProp(PR_HAS_X_EQUALS_Y)) {
	  VarCounter vc;
	  if (hasXEqualsY(f,vc,1)) {
	    addProp(PR_HAS_X_EQUALS_Y);
	  }
	}
	if ((! hasProp(PR_HAS_FUNCTION_DEFINITIONS)) && 
	    FunctionDefinition::isFunctionDefinition(f)) {
	  addProp(PR_HAS_FUNCTION_DEFINITIONS);
	}
	return;
      }

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
} 



void Property::scan (const Clause& clause, bool isAxiom)
{
  TRACER("void Property::scan (const Clause& clause, bool isAxiom)");

  int positiveLiterals = 0;
  int negativeLiterals = 0;
  int equationalLiterals = 0;
  int groundLiterals = 0;
  _variablesInThisClause = 0;

  Iterator<Literal> lits (clause.literals());
  while (lits.hasNext()) {
    Literal literal (lits.next());
    if (literal.isPositive()) {
      positiveLiterals ++;
    }
    else {
      negativeLiterals ++;
    }

    Atom atom (literal.atom());
    if (atom.isEquality()) {
      equationalLiterals++;
    }

    bool isGround = true;
    scan (atom, isGround);

    if (isGround) {
      groundLiterals++;
    }
  }
  int literals = positiveLiterals + negativeLiterals;
  _atoms += literals;

  if ( equationalLiterals > 0 ) {
    _equationalClauses ++;
    _equalityAtoms += equationalLiterals;
  }
  if ( literals == equationalLiterals ) {
    _pureEquationalClauses ++;
  }

  if (isAxiom) {
    _axiomClauses ++;
    if ( literals == 1) {
      _unitAxioms ++;
      if ( groundLiterals == 1) {
	_groundUnitAxioms ++;
      }
    }
    if (positiveLiterals <= 1) {
      _hornAxioms ++;
    }
    if (negativeLiterals == 0) {
      _positiveAxioms ++;
      if (literals == groundLiterals) {
	_groundPositiveAxioms ++;
      }
    }
  }
  else {
    _goalClauses ++;
    if ( literals == 1) {
      _unitGoals ++;
    }
    if (positiveLiterals <= 1) {
      _hornGoals ++;
    }
    if (literals == groundLiterals) {
      _groundGoals ++;
    }
  }

  _totalNumberOfVariables += _variablesInThisClause;
  if (_variablesInThisClause > _maxVariablesInClause) {
    _maxVariablesInClause = _variablesInThisClause;
  }
} 



void Property::scan (const Formula& formula)
{
  TRACER("void Property::scan (const Formula&)");

  _subformulas++;
  switch (formula.connective()) 
    {
    case ATOM:
      {
	_atoms++;
	Atom a (formula.atom());
	if (a.isEquality()) {
	  _equalityAtoms++;
	}
	bool dummy = false;
	scan(a,dummy);
	return;
      }
      
    case AND:
    case OR:
      {
	Iterator<Formula> args (formula.args());
	while (args.hasNext()) {
	  scan(args.next());
	}
	return;
      }

    case IMP:
    case IFF:
    case XOR:
      scan(formula.left());
      scan(formula.right());
      return;

    case NOT:
      scan(formula.uarg());
      return;

    case FORALL:
    case EXISTS:
      scan(formula.qarg());
      return;

#if DEBUG_PREPRO
    default:
      ASS (false);
#endif
    }
} 



void Property::scan (const Atom& atom, bool& isGround)
{
  TRACER("void Property::scan (const Atom& atom, bool& isGround)");

  if (! atom.isEquality()) {
    int arity = atom.functor().arity();
    if (arity > _maxPredArity) {
      _maxPredArity = arity;
    }
  }

  Iterator<Term> args (atom.args());
  while ( args.hasNext() ) {
    scan (args.next(), isGround);
  }
} 



void Property::scan (const Term& term, bool& isGround)
{
  TRACER("void Property::scan (const Term& term, bool& isGround)");

  _terms ++;

  switch (term.tag()) 
    {
    case VAR:
      isGround = false;
      _variablesInThisClause++;
      return;

    case NUMERIC:
      return;

    case COMPOUND: 
      {
	int arity = term.functor().arity();
	if (arity > _maxFunArity) {
	  _maxFunArity = arity;
	}

	Iterator<Term> args (term.args());
	while ( args.hasNext() ) {
	  scan (args.next(), isGround);
	}
	return;
      }

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
} 


 
string Property::categoryString() const
{
  TRACER("string Property::categoryString() const");

  switch (_category)
    {
    case NEQ:
      return "NEQ";
    case HEQ:
      return "HEQ";
    case PEQ:
      return "PEQ";
    case HNE:
      return "HNE";
    case NNE:
      return "NNE";
    case FEQ:
      return "FEQ";
    case FNE: 
      return "FNE";
    case EPR:
      return "EPR";
    case UEQ:
      return "UEQ";
#if DEBUG_PREPRO
    default:
      ASS(false);
      return "";
#endif
    }
} 



string Property::toString () const
{
  string result("TPTP class: "); 
  result += categoryString() + "\n"; 

  if (clauses() > 0) { 
    result += "Clauses: "; 
    result += Int::toString(clauses());
    result += " (";
    result += Int::toString(_unitAxioms+_unitGoals);
    result += " unit, ";
    result += Int::toString(_goalClauses);
    result += " goal, ";
    result += Int::toString(_equationalClauses);
    result += " equational)\n";

    result += "Variables: "; 
    result += Int::toString(_totalNumberOfVariables);
    result += " (";
    result += Int::toString(_maxVariablesInClause);
    result += " maximum in a single clause)\n";
  }

  if (formulas() > 0) { 
    result += "Formulas: "; 
    result += Int::toString(formulas());
    result += " (";
    result += Int::toString(_goalFormulas);
    result += " goal)\n";
    result += "Subformulas: "; 
    result += Int::toString(_subformulas);
    result += "\n";
  }

  result += "Atoms: "; 
  result += Int::toString(_atoms);
  result += " (";
  result += Int::toString(_equalityAtoms);
  result += " equality)\n";
    
  return result;
} 



bool Property::hasXEqualsY (const Clause& c)
{
  Iterator<Literal> ls (c.literals());
  while (ls.hasNext()) {
    Literal l (ls.next());
    if (l.isPositive() && isXEqualsY(l.atom())) {
      return true;
    }
  }
  return  false;
} 



bool Property::isXEqualsY (const Atom& a)
{
  TRACER("Property::isXEqualsY (const Atom& a)");

  if (! a.isEquality()) {
    return false;
  }
  
  const TermList& ts = a.args();
  const Term& t1 = ts.head();
  if (t1.tag() != VAR) {
    return false;
  }
  const Term& t2 = ts.second();
  return t2.tag() == VAR &&
         t1.var() != t2.var();
} 


 
bool Property::hasXEqualsY (const Formula& f, VarCounter& vc, int polarity)
{
  switch (f.connective()) {
  case ATOM:
    {
      if (polarity == -1) { 
	return false;
      }
      const Atom& a = f.atom();
      if (! a.isEquality()) {
	return false;
      }
      
      const TermList& ts = a.args();
      const Term& t1 = ts.head();
      if (t1.tag() != VAR) {
	return false;
      }
      const Term& t2 = ts.second();
      return t2.tag() == VAR &&
	     t1.var() != t2.var() &&
	     vc.getCount(t1.var()) == 0 && 
             vc.getCount(t2.var()) == 0;
    }
    
  case AND:
  case OR:
    {
      Iterator<Formula> fs(f.args());
      while (fs.hasNext()) {
	if (hasXEqualsY(fs.next(),vc,polarity)) {
	  return true;
	}
      }
      return false;
    }

  case IMP:
    return hasXEqualsY(f.left(),vc,-polarity) ||
           hasXEqualsY(f.right(),vc,polarity);

  case IFF:
  case XOR:
    return hasXEqualsY(f.left(),vc,0) ||
           hasXEqualsY(f.right(),vc,0);

  case NOT:
    return hasXEqualsY(f.uarg(),vc,-polarity);

  case FORALL:
    
    if (polarity == -1) {
      Iterator<Var> vs(f.vars());
      while (vs.hasNext()) {
	vc.incCount(vs.next());
      }
    }
    return hasXEqualsY(f.qarg(),vc,polarity);

  case EXISTS:
    
    if (polarity == 1) {
      Iterator<Var> vs(f.vars());
      while (vs.hasNext()) {
	vc.incCount(vs.next());
      }
    }
    return hasXEqualsY(f.qarg(),vc,polarity);

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
