
/*
 * File Renaming.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Int.hpp"
#include "Clause.hpp"
#include "Renaming.hpp"


Renaming::Renaming ()
  : _direct(0),
    _inverse(0),
    _nextDirect(0),
    _nextInverse(0)
{
} 


Renaming::~Renaming ()
{
  delete [] _direct;
  delete[] _inverse;
} 



bool Renaming::isVariant (const Clause& c1, const Clause& c2)
{
  Renaming ren;
  return ren.match(c1.literals(),c2.literals());
} 



bool Renaming::match (const LiteralList& ls1, const LiteralList& ls2)
{
  if (ls1.isEmpty()) {
    return ls2.isEmpty();
  }

  return ls2.isNonEmpty() &&
         match(ls1.head(),ls2.head()) &&
         match(ls1.tail(),ls2.tail());
} 



bool Renaming::match (const TermList& ts1, const TermList& ts2)
{
  if (ts1.isEmpty()) {
    return ts2.isEmpty();
  }

  return ts2.isNonEmpty() &&
         match(ts1.head(),ts2.head()) &&
         match(ts1.tail(),ts2.tail());
} 



bool Renaming::match (const Literal& l1, const Literal& l2)
{
  return l1.sign() == l2.sign() &&
         match(l1.atom(),l2.atom());
} 



bool Renaming::match (const Atom& a1, const Atom& a2)
{
  return a1.functor() == a2.functor() &&
         match(a1.args(),a2.args());
} 



bool Renaming::match (const Term& t1, const Term& t2)
{
  if (t1.tag() != t2.tag()) {
    return false;
  }

  switch (t1.tag()) {
  case VAR:
    return addBinding(t1.var().toInt(),t2.var().toInt());

  case COMPOUND:
    return t1.functor() == t2.functor() &&
           match(t1.args(),t2.args());

  case NUMERIC:
    return t1.number() == t2.number();

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool Renaming::addBinding (int v1,int v2)
{
  if (v1 >= _nextDirect) {
    expandToFit(v1,_direct,_nextDirect);
  }
  else if (_direct[v1] == v2) { 
    return true;
  }
  else if (_direct[v1] != -1) { 
    return false;
  }
  
  _direct[v1] = v2;

  if (v2 >= _nextInverse) {
    expandToFit(v2,_inverse,_nextInverse);
    _inverse[v2] = v1;
    return true;
  }
  if (_inverse[v2] != -1) { 
    return false;
  }
  
  _inverse[v2] = v1;
  return true;
} 



void Renaming::expandToFit (int v,int*& varArray,int& ln)
{
  
  int newLength = Int::max(ln+8,v+1);

  
  int* newArray = new int[newLength];
  if (! newArray) {
    ALLOCATION_FAILURE;
  }
  
  for (int i = 0;i < ln;i++) {
    newArray[i] = varArray[i];
  }
  for (int k = ln;k < newLength;k++) {
    newArray[k] = -1;
  }
  delete [] varArray;
  varArray = newArray;
  ln = newLength;
} 
