
/*
 * File Run.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */




#include <signal.h>
#include <ctime>


#include "VampireKernel.hpp"
#include "GlobAlloc.hpp"




#include "antlr/TokenStreamSelector.hpp"


antlr::TokenStreamSelector* selector;
#include "KIFLexer.hpp"
#include "KIFParser.hpp"
#include "XMLLexer.hpp"
#include "XMLParser.hpp"
#include "TPTPLexer.hpp"
#include "TPTPParser.hpp"


#include "Run.hpp"
#include "XML.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Statistics.hpp"
#include "VampireMode.hpp"
#include "KIFMode.hpp"
#include "SpiderMode.hpp"
#include "CASCMode.hpp"

bool Run::saved;


string Run::toString (RunResult krr)
{
  TRACER("Run::toString (RunResult krr)");

  switch (krr) {
  case REFUTATION:
    return "refutation";
  case SATISFIABLE:
    return "satisfiable";
  case UNKNOWN:
    return "unknown";
#if DEBUG_PREPRO
  default:
    ASS(false);
    return "";
#endif
  }
}



void runtimeErrorHook ()
{
  TRACER("runtimeErrorHook");

  if (VampireKernel::currentSession()) {
    cout << "% Trying to shut down current kernel session...\n";
    VampireKernel::currentSession()->reportError();
  }

  XML::Element* err = new XML::TextElement("error",
					   "some module detected inconsistency in run time");
  err->addAttribute(new XML::TextAttribute("type","runtime"));
  Run::log->addChild(err);
  Run::saveLog();
  exit(EXIT_SUCCESS);
} 



void debugSystemInternalErrorHook ()
{  
  TRACER("debugSystemInternalErrorHook");

  if (VampireKernel::currentSession()) {
    cout << "% Trying to shut down current kernel session...\n";
    VampireKernel::currentSession()->reportError();
  }

  XML::Element* err = new XML::TextElement("error",
					   "debugging system detected internal inconsistency");
  err->addAttribute(new XML::TextAttribute("type","debug"));
  Run::log->addChild(err);
  Run::saveLog();
  exit(EXIT_SUCCESS);
} 



void Run::saveLog () 
{
  TRACER("Run::saveLog");

  if (saved) {
    return;
  }
  saved = true;
  string logFile (Options::globalOptions.logFile());
  if (logFile != "") {
    log->append(logFile);
  }
} 



void Run::terminateBySignal (string signalDescription,int exitCode)
{
  TRACER("Run::terminateBySignal (string signalDescription)");

  cancelSignalHandling();

  VampireKernel* kernel = VampireKernel::currentSession ();
  if (kernel) {
    Statistics statistics (*kernel);
    XML::Element* xmlStat = statistics.toXML();
    if (xmlStat) {
      log->addChild(xmlStat);
    }
    XML::Element* err = new XML::TextElement("aborted","");
    err->addAttribute(new XML::TextAttribute("signal",signalDescription));
    log->addChild(err);
  }
  saveLog();
  exit(exitCode);
} 


void Run::cancelSignalHandling ()
{
  TRACER("cancelSignalHandling");

  signal(SIGTERM,SIG_DFL);
#ifndef _MSC_VER
  signal(SIGQUIT,SIG_DFL);
  signal(SIGHUP,SIG_DFL);
  signal(SIGXCPU,SIG_DFL);
#endif
  signal(SIGINT,SIG_DFL);
#ifndef DEBUG_ERROR_SIGNALS
  signal(SIGILL,SIG_DFL);
  signal(SIGFPE,SIG_DFL);
  signal(SIGSEGV,SIG_DFL);
#ifndef _MSC_VER
  signal(SIGBUS,SIG_DFL);
  signal(SIGTRAP,SIG_DFL);
#endif
  signal(SIGABRT,SIG_DFL);
#endif
} 



void handleSignal (int sigNum)
{
  
  
  static bool handled = false;
  string signalDescription = Run::signalToString(sigNum);

  switch (sigNum)
    {
    case SIGTERM:
# ifndef _MSC_VER
    case SIGQUIT:
    case SIGHUP:
    case SIGXCPU:
      if (handled) {
	exit(0);
      }
      handled = true;
      cout << "% Aborted by signal " << signalDescription << "\n";
      if (VampireKernel::currentSession()) {
	cout << "% Trying to shut down current kernel session...\n";
      }
      
      cout << "% Bailing out now: aborted by interrupt signal " 
	   << signalDescription << "\n";
      Run::terminateBySignal (signalDescription,EXIT_FAILURE);
      return;
# endif

    case SIGINT:
      {
	if (VampireKernel::currentSession())
	  {
	    VampireKernel::currentSession()->suspend();
	  }
	else
	  {
	    cout << "% No kernel session running at the moment.\n";
	  };
	cout << "Continue? (y/n)\n";
	
	char a;
	cin >> a;
	
	if ((a == 'n') || (a == 'N'))
	  {
	    cout << "% Terminated by the user.\n";
	    Run::terminateBySignal (signalDescription,EXIT_SUCCESS);
	    return;
	  }
	else
	  {
            if (VampireKernel::currentSession())
	      VampireKernel::currentSession()->resume();
	    cout << "\nContinuing...\n";
	  }
      }
      return;      

    case SIGILL:
    case SIGFPE:
    case SIGSEGV:

# ifndef _MSC_VER
    case SIGBUS:
    case SIGTRAP:
# endif
    case SIGABRT: 
      {
	if (handled) {
	  exit(0);
	}
	handled = true;
	cout << "% Error signal " << signalDescription << " caught\n";
	VampireKernel* kernel = VampireKernel::currentSession ();
	if (kernel) {
	  cout << "% Trying to shut down current kernel session...\n";
	  kernel->reportErrorSignal(sigNum);
	}
#if DEBUG_PREPRO
	MyTracer::printStack (cout);
#endif
	cout << "% Bailing out now: aborted by error signal " << signalDescription << "\n";
	Run::terminateBySignal (signalDescription,EXIT_FAILURE);
	return;
      }

    default:
      break;
    };
} 


void Run::setSignalHandlers()
{
  signal(SIGTERM,handleSignal);
#ifndef _MSC_VER
  signal(SIGQUIT,handleSignal);
  signal(SIGHUP,handleSignal);
  signal(SIGXCPU,handleSignal);
#endif
  signal(SIGINT,handleSignal);
#ifndef DEBUG_ERROR_SIGNALS
  signal(SIGILL,handleSignal);
  signal(SIGFPE,handleSignal);
  signal(SIGSEGV,handleSignal);
#ifndef _MSC_VER
  signal(SIGBUS,handleSignal);
  signal(SIGTRAP,handleSignal);
#endif
  signal(SIGABRT,handleSignal);
#endif
} 


string Run::signalToString (int sigNum)
{
  switch (sigNum)
    {
    case SIGTERM: return "SIGTERM";
# ifndef _MSC_VER
    case SIGQUIT: return "SIGQUIT";
    case SIGHUP: return "SIGHUP";
    case SIGXCPU: return "SIGXCPU";
    case SIGBUS: return "SIGBUS";
    case SIGTRAP: return "SIGTRAP";
# endif
    case SIGINT: return "SIGINT";
    case SIGILL: return "SIGILL";
    case SIGFPE: return "SIGFPE";
    case SIGSEGV: return "SIGSEGV";
    case SIGABRT: return "SIGABRT";
    default: return "UNKNOWN SIGNAL";
    };
} 



void Run::run (UnitList& units, Timer& timer)
{
  time_t jobTime = time(static_cast<time_t*>(0));
  tm* gmt = gmtime(&jobTime); 

  if (gmt->tm_year > 104 || 
      gmt->tm_mon > 6 ||
      gmt->tm_mday > 8
      ) { 
    cout << 
      "\n"
      "*************** WARNING *****************************\n"
      "* This software is only intended for the            *\n"
      "* CASC-J2 competition. Its use in any other form is *\n"
      "* not permitted. All questions related to the use,  *\n"
      "* licensing and distribution of Vampire should be   *\n"
      "* addressed to Andrei Voronkov.                     *\n"
      "*****************************************************\n\n";
  }






  switch (Options::globalOptions.mode()) {
  case Options::KIF:
    KIFMode::run(units, timer);
    break;

  case Options::VAMPIRE:
    VampireMode::run(units, timer);
    break;

  case Options::SPIDER:
    SpiderMode::run(units, timer);
    break;

  case Options::CASC_J2:
    CASCMode::run(units, timer);
    break;
  }
} 



void Run::reportAllocationFailure ()
{
  switch (Options::globalOptions.mode()) {
  case Options::KIF:
    KIFMode::reportAllocationFailure();
    break;

  case Options::VAMPIRE:
    VampireMode::reportAllocationFailure();
    break;

  case Options::SPIDER:
    SpiderMode::reportAllocationFailure();
    break;

  case Options::CASC_J2:
    CASCMode::reportAllocationFailure();
    break;
  }
} 



void Run::readInput (string inputFileName, 
		     FormulaList& formulas,
		     UnitList& units)
{
  TRACER("Run::readInput");

  std::ifstream inFile (inputFileName.c_str());
  if(!inFile) {
    USER_ERROR(string("cannot open input file ") + inputFileName);
  };

  switch (Options::globalOptions.inputSyntax()) {
  case Options::INPUT_KIF: 
    {
      KIFLexer lexer(inFile); 
      KIFParser parser (lexer);
      parser.start(formulas);
      return;
    }

  case Options::INPUT_TPTP: 
    {
      TPTPLexer lexer(inFile); 
      TPTPParser parser (lexer);
      parser.units(units);
      return;
    }
  }
} 
