
/*
 * File Signature.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <string.h>
#include <stdio.h>


#include "Tracer.hpp"
#include "Int.hpp"
#include "Signature.hpp"
#include "SymCounter.hpp"
#include "XML.hpp"
#include "Symbol.hpp"


Signature::Signature () :
  _noOfPreds (0),
  _noOfFuns (0),
  _constants (0),
  _propositionalSymbols (0),
  _maxFunArity (0),
  _maxPredArity (0),
  _answerPredicate (0),
  _kernelSymbols (new Symbol*[1022]),
  _maxNumberOfKernelSymbols (1022),
  _lastSkolem (0),
  _skolemPrefix("sk")
{
  TRACER("Signature::Signature");

  if (! _kernelSymbols) {
    ALLOCATION_FAILURE;
  }

  
  for (int i = 0; i < 1022; i++) {
    _kernelSymbols[i] = 0;
  }

  
  for ( int j = NO_OF_BUCKETS - 1; j >= 0; j-- ) {
    _fbuckets [j] = SymbolList::empty ();
    _pbuckets [j] = SymbolList::empty ();
  }

  
  _equality = &createPred("equal", 2);
} 



void Signature::arityCheck () const 
{
  TRACER("Signature::arityCheck");

  
  for ( int i = NO_OF_BUCKETS - 1; i >= 0; i-- ) {
    
    bool curIsFun = true;
    
    SymbolList* cur = _fbuckets [i];

    for (;;) { 
      if ( cur->isEmpty() ) {
        if (curIsFun) {
          cur = _pbuckets [i];
          if ( cur->isEmpty() ) {
            break; 
          }
          curIsFun = false;
        }
        else {  
          break;
        }
      }
      
      Symbol* curSym = cur->head ();

      SymbolList* rest = cur;
      bool restIsFun = curIsFun;

      for (;;) { 
        rest = rest->tail ();
        if ( rest->isEmpty() ) { 
          if (! restIsFun) { 
            break;
          }
          
          rest = _pbuckets [i];
          if ( rest->isEmpty() ) { 
            break; 
          }
          restIsFun = false;
        }
        
        
        if (curSym->name() == rest->head()->name()) { 
          if (curIsFun) {
            if (restIsFun) { 
              USER_ERROR(string("Function symbol ") +
			 curSym->name() +
			 string(" is used with two different arities ") + Int::toString(curSym->arity()) +
			 " and " + 
			 Int::toString(rest->head()->arity()));
            }
            else {  
	      USER_ERROR(string("Symbol ") + 
			 curSym->name() +
			 " is used both as a function symbol and a predicate symbol");
	    }
          }
          else { 
            USER_ERROR(string("Predicate symbol ") + 
		       curSym->name() +
		       " is used with two different arities " +
		       Int::toString(curSym->arity()) +
		       " and " + 
		       Int::toString(rest->head()->arity()));
          }
        }
      }
      cur = cur->tail ();
    }
  }
  
} 


Signature::~Signature () 
{
  TRACER("Signature::~Signature");

  
  for ( int i = NO_OF_BUCKETS - 1; i >= 0; i-- ) {
    SymbolList::Iterator fit (_fbuckets [i]);
    while (fit.hasNext()) {
      delete fit.next();
    }
    _fbuckets [i]->destroy();
    SymbolList::Iterator pit (_pbuckets [i]);
    while (pit.hasNext()) {
      delete pit.next();
    }
    _pbuckets [i]->destroy();
  }
} 



Symbol& Signature::create (const string& str, 
			   int arity, 
			   SymbolList** buckets, 
			   bool isPred )
{
  TRACER("Signature::create");

  int bucket = hash(str);
  for (SymbolList* l = buckets[bucket]; ! l->isEmpty (); l = l->tail()) {
    Symbol* sym = l->head();
    if (arity == sym->arity() &&         
         str == sym->name()) {  
      return *(l->head());
    }
  }

  if (isPred) { 
    if (arity == 0) {
      _propositionalSymbols++;
    }
    else if (arity > _maxPredArity) {
      _maxPredArity = arity;
    }
  }
  else { 
    if (arity == 0) {
      _constants++;
    }
    else if (arity > _maxFunArity) {
      _maxFunArity = arity;
    }
  }

  Symbol* newf; 
  if (isPred) {
    newf = new PredicateSymbol (str,arity,_noOfPreds);
  }
  else {
    newf = new FunctionSymbol (str,arity,_noOfFuns);
  }
  buckets[bucket] = new SymbolList(newf, buckets [bucket]);

  return *newf;
} 



int Signature::hash (const string& str)
{
  TRACER("Signature::hash");

  int result = 0;
  int length = str.length();

  for (int i = 0; i < length; i++ ) 
    result = ( result * 10 + str[i] ) % NO_OF_BUCKETS;
  
  return result;
} 



Signature::Iterator::Iterator (SymbolList** buckets)
  :
  _firstBucket (buckets),
  _currentBucket (buckets + NO_OF_BUCKETS - 1),
  _currentList (*_currentBucket)
{
  TRACER("Signature::Iterator::Iterator");
} 
    


bool Signature::Iterator::hasNext ()
{
  TRACER("Signature::Iterator::hasNext");

  for (;;) {
    if ( _currentList->isNonEmpty() ) {
      return true;
    }
    
    if (_currentBucket == _firstBucket) {
      return false;
    }
    
    _currentBucket--;
    _currentList = *_currentBucket;
  }
} 
    


Symbol& Signature::Iterator::next ()
{
  TRACER("Signature::Iterator::next");
  ASS( _currentList->isNonEmpty() );

  Symbol* result = _currentList->head ();
  _currentList = _currentList->tail ();

  return *result;
} 
    


Symbol* Signature::exists (const string& name, SymbolList** buckets)
{
  TRACER("Signature::exists");

  int bucket = hash (name);
  for (SymbolList* l = buckets[bucket];
       ! l->isEmpty ();
       l = l->tail()) {
    Symbol* sym = l->head();
    if (name == sym->name()) {  
      return sym;
    }
  }

  return 0;
} 


Symbol* Signature::exists (const string& name, 
			   int arity, 
			   SymbolList** buckets)
{
  TRACER("Signature::exists");

  int bucket = hash (name);
  for (SymbolList* l = buckets[bucket]; ! l->isEmpty (); l = l->tail()) {
    Symbol* sym = l->head();
    if (name == sym->name() && arity == sym->arity()) {  
      return sym;
    }
  }

  return 0;
} 



void Signature::createAnswerAtom (const TermList& ts)
{
  TRACER("Signature::createAnswerAtom");

  _answerPredicate = &createPred ("$answer", ts.length ());
  _answerAtom = Atom(*_answerPredicate, ts);
} 



void Signature::addArithmetic ()
{
  TRACER("Signature::addArithmetic");

  createPred ( ">", 2);
  createPred ( "<", 2);
  createPred ( ">=", 2);
  createPred ( "<=", 2);

  createFun ( "+", 2);
  createFun ( "-", 2);
  createFun ( "-", 1);
  createFun ( "*", 2);
} 


string Signature::summary () 
{
  TRACER("Signature::summary");

  return string("Predicate symbols: ") +
    Int::toString(_noOfPreds) +
    " (propositional " +
    Int::toString(_propositionalSymbols) +
    ", maximal arity: " +
    Int::toString(_maxPredArity) +
    ")\n" +
    "Function symbols: " +
    Int::toString(_noOfFuns) +
    " (constants " +
    Int::toString(_constants) +
    ", skolem: " +
    Int::toString(_lastSkolem+1) +
    ", maximal arity: " +
    Int::toString(_maxFunArity) +
    ")\n";
} 



XML::Element* Signature::toXML () const
{
  TRACER("Signature::toXML");

  XML::DeepElement* signature = new XML::DeepElement("signature");
  XML::DeepElement* functions = new XML::DeepElement("functions");
  signature->addChild(functions);
  XML::DeepElement* predicates = new XML::DeepElement("predicates");
  signature->addChild(predicates);
  FunIterator fs(const_cast<Signature&>(*this));
  while (fs.hasNext()) {
    functions->addChild(fs.next().toXML());
  }
  PredIterator ps(const_cast<Signature&>(*this));
  while (ps.hasNext()) {
    predicates->addChild(ps.next().toXML());
  }

  return signature;
} 



void Signature::registerKernelNumber (Symbol& symbol, unsigned long n)
{
  TRACER("Signature::registerKernelNumber");

  ASS(&symbol != 0);

  expandKernelSymbolsToFit(n);

  _kernelSymbols[n] = &symbol;
  symbol.setKernelNumber(n);
} 



const FunctionSymbol& Signature::getFunctionByItsKernelNumber (unsigned long n,
							       int arity)
{
  TRACER("Signature::getFunctionByItsKernelNumber");

  expandKernelSymbolsToFit(n);
  Symbol* symbol = _kernelSymbols[n];
  if (symbol && symbol->arity() == arity && symbol->isFunction()) {
    return *(static_cast<FunctionSymbol*>(symbol));
  }

  static int lastFun = 0;
  static char name [12] = {'f', '_'}; 
  string nm;
  do {
    sprintf (name+2, "%d", ++lastFun);
    nm = name;
  }
  while (existsFun (nm));

  
  FunctionSymbol& result = createFun(nm, arity);
  _kernelSymbols[n] = &result;
  return result;
} 



const PredicateSymbol& Signature::getPredicateByItsKernelNumber (unsigned long n,
								 int arity)
{
  TRACER("Signature::getPredicateByItsKernelNumber");

  expandKernelSymbolsToFit(n);
  Symbol* symbol = _kernelSymbols[n];
  if (symbol && symbol->arity() == arity && ! symbol->isFunction()) {
    return *(static_cast<PredicateSymbol*>(symbol));
  }

  static int lastPred = 0;
  static char name [12] = {'p', '_'}; 
  string nm;
  do {
    sprintf (name+2, "%d", ++lastPred);
    nm = name;
  }
  while (existsPred (nm));

  
  PredicateSymbol& result = createPred(nm, arity);

  _kernelSymbols[n] = &result;

  return result;
} 



void Signature::expandKernelSymbolsToFit (unsigned long n)
{
  TRACER("Signature::expandKernelSymbolsToFit");

  if (_maxNumberOfKernelSymbols > (long)n) {
    return;
  }

  int newLength = Int::max(_maxNumberOfKernelSymbols + 1024,n+1);
  Symbol** newArray = new Symbol* [newLength];
  if (! newArray) {
    ALLOCATION_FAILURE;
  }
  
  for (int i = 0;i < _maxNumberOfKernelSymbols;i++) {
    newArray[i] = _kernelSymbols[i];
  }
  
  for (int j = _maxNumberOfKernelSymbols;j < newLength;j++) {
    newArray[j] = 0;
  }
  delete [] _kernelSymbols;
  _kernelSymbols = newArray;
  _maxNumberOfKernelSymbols = newLength;
} 



const Atom& Signature::getAnswerAtom () const
{
  ASS(hasAnswerAtom());

  return _answerAtom;
} 



FunctionSymbol& Signature::newSkolemFunction (int arity)
{
  TRACER("Signature::newSkolemFunction");

  for (;;) {
    string name(_skolemPrefix + Int::toString(_lastSkolem++));
    FunctionSymbol* f = existsFun (name,arity);
    if (! f) {
      FunctionSymbol& result = createFun(name, arity);
      result.addProperty(Symbol::SKOLEM);
      return result;
    }
  }
} 



void Signature::setSkolemPrefix (const string& prefix)
{
  _skolemPrefix = prefix;
} 


