
/*
 * File Skolem.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Skolem.hpp"
#include "Unit.hpp"
#include "Tracer.hpp"
#include "Assertion.hpp"
#include "Signature.hpp"
#include "Position.hpp"
#include "InferenceInfo.hpp"
#include "Inference.hpp"
#include "Rectify.hpp"



void Skolem::skolemise (Unit& u)
{
  TRACER("Skolem::skolemise (Unit& u)");

  VarList dummy;
  Rectify::rectify(u,dummy);

  ASS(u.unitType() == FORMULA);
  ASS(Assertion::isInNNF(u.formula()));

  Formula f(u.formula());
  Skolem skol;
  skol.skolemiseNNF(f,VarList(),Position(),Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_SKOLEMIZE,u);
    inf->setInfo(skol._info);
    u = Unit(f,inf);
  }
} 


void Skolem::skolemiseNNF (Formula& f, 
			   const VarList& vs,
			   const Position& positionWithERemoved,
			   const Position& positionWithQRemoved)
{
  TRACER("Skolem::skolemiseNNF (Formula& f,...)");

  switch (f.connective()) {
  case ATOM: 
    {
      Atom a (f.atom());
      _subst.apply(a);
      if (a == f.atom()) {
	return;
      }
      f = Formula(a);
    }
    return;

  case AND:
  case OR: 
    {
      FormulaList fs (f.args());
      skolemiseNNF(fs, vs, positionWithERemoved, positionWithQRemoved, 0);
      if (fs == f.args()) {
	return;
      }
      f = Formula(f.connective(), fs);
    }
    return;

  case NOT: 
    {
      Atom a(f.uarg().atom());
      _subst.apply(a);
      if (a == f.uarg().atom()) {
	return;
      }
      f = Formula (NOT, Formula(a));
    }
    return;

  case FORALL: 
    {
      VarList ws(f.vars());
      VarList newVs(vs);
      newVs.append(ws);
      f = f.qarg();
      skolemiseNNF(f, newVs, 
		   Position(0,positionWithERemoved),
		   positionWithQRemoved);
      _info = new InferenceInfoTermList(IR_SKOLEMIZE,
					positionWithQRemoved,
					_info,
					TermList(ws));
    }
    return;

  case EXISTS: 
    {
      TermList skolemTerms;
      
      VarList xs(f.vars());
      makeSkolemTerms(xs, TermList(vs), vs.length(), skolemTerms);
      
      f = f.qarg();
      skolemiseNNF(f, vs, 
		   positionWithERemoved,
		   positionWithQRemoved);
      
      _info = new InferenceInfoTermList(IR_SKOLEMIZE,
					positionWithERemoved,
					_info,
					skolemTerms);
      Iterator<Var> ws (xs);
      while (ws.hasNext()) {
	_subst.undoBinding (ws.next());
      }
    }
    return;

#if DEBUG_PREPRO
  case IMP:
  case IFF:
  case XOR:
  default:
    ASS( false );
#endif
  }
} 



void Skolem::skolemiseNNF (FormulaList& fs,
			   const VarList& skolemArgs,
			   const Position& positionWithERemoved,
			   const Position& positionWithQRemoved,
			   int pos) 
{
  TRACER("skolemiseNNF (FormulaList& fs,...)");

  if (fs.isEmpty()) {
    return;
  }

  Formula h (fs.head());
  skolemiseNNF(h, skolemArgs, 
	       Position(pos,positionWithERemoved), 
	       Position(pos,positionWithQRemoved));

  FormulaList t (fs.tail());
  skolemiseNNF(t, skolemArgs, 
	       positionWithERemoved, 
	       positionWithQRemoved, pos+1);

  if (h == fs.head() && t == fs.tail()) {
    return;
  }

  fs = FormulaList(h,t);
} 



void Skolem::makeSkolemTerms (const VarList& vars, 
			      const TermList& skolemArgs,
			      int arity,
			      TermList& skolemTerms)
{
  if (vars.isEmpty()) {
    return;
  }
  Term skolemTerm(Signature::sig.newSkolemFunction(arity), skolemArgs);
  _subst.bind(vars.head(), skolemTerm);
  makeSkolemTerms(vars.tail(), skolemArgs, arity, skolemTerms);
  skolemTerms.push(skolemTerm);
} 


