
/*
 * File Unit.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <string.h>


#include "Tracer.hpp"
#include "Unit.hpp"
#include "XML.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Int.hpp"


long Unit::_lastNumber = 0;



UnitList::UnitList () 
  : Lst<Unit> ()
{
} 



UnitList::UnitList (const Unit &hd)
  : Lst<Unit> (hd)
{
} 


UnitList::UnitList (const Unit &hd, const UnitList& tl)
  :
  Lst<Unit> (hd,tl)
{
} 



UnitList::UnitList (LstData<Unit>* d)
  :
  Lst<Unit> (d)
{
} 



UnitList::UnitList (const UnitList& ts)
  :
  Lst<Unit> (ts)
{
} 



UnitList::UnitList (InputType tp, const FormulaList& formulas)
{
  FormulaList fs (formulas);
  while (fs.isNonEmpty()) {
    push (Unit (0, tp, fs.head()));
    fs.pop();
  }
} 



void Unit::operator = (const Unit& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} 


void Unit::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} 



string Unit::toString(InputType t)
{
  switch (t) {
  case AXIOM:
    return "axiom";
  case CONJECTURE:
    return "conjecture";
  case ASSUMPTION:
    return "assumption";
#if DEBUG_PREPRO
  default:
    ASS(false);
    return 0;
#endif
  }
} 


string Unit::toString(UnitType t)
{
  switch (t) {
  case CLAUSE:
    return "clause";
  case FORMULA:
    return "formula";
#if DEBUG_PREPRO
  default:
    ASS(false);
    return 0;
#endif
  }
} 


string Unit::toString(DefType dt)
{
  switch (dt) {
  case DT_NONE:
    return "none";
  case DT_PRED:
    return "pred";
  case DT_FUN:
    return "fun";
#if DEBUG_PREPRO
  default:
    ASS(false);
    return 0;
#endif
  }
} 



ostream& operator << (ostream& str, InputType t)
{
  str << Unit::toString(t);
  return str;
} 


ostream& operator << (ostream& str, UnitType t)
{
  switch ( t ) {
    case CLAUSE:
      str << "input_clause";
      break;

    case FORMULA:
      str << "input_formula";
      break;
  }

  return str;
} 


Unit::Data::Data (const string& name, InputType intype, const Clause& c)
  :
  _counter (1),
  _untype (CLAUSE),
  _intype (intype),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _name (name),
  _clause (c),
  _defType (DT_NONE),
  _inference (new Inference(IR_INPUT))
{
  TRACER( "Unit::Data::Data-1" );
} 


Unit::Data::Data (const string& name, InputType intype, const Formula& f)
  :
  _counter (1),
  _untype (FORMULA),
  _intype (intype),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _name (name),
  _formula (f),
  _defType (DT_NONE),
  _inference (new Inference(IR_INPUT))
{
  TRACER( "Unit::Data::Data-2" );
} 



Unit::Data::Data (const Formula& f, Inference* inf)
  : _counter(1),
    _untype(FORMULA),
    _intype(inf->premises().maxInputType()),
    _number(++_lastNumber),
    _numberInRefutation(0),
    _formula(f),
    _defType(DT_NONE),
    _inference(inf)
{
  TRACER( "Unit::Data::Data-3" );
}



Unit::Data::Data (const Clause& c,Inference* inf)
  : _counter (1),
    _untype (CLAUSE),
    _intype(inf->premises().maxInputType()),
    _number (++ _lastNumber),
    _numberInRefutation(0),
    _clause (c),
    _defType (DT_NONE),
    _inference (inf)
{
  TRACER( "Unit::Data::Data-4" );
}



Unit::Data::Data (InferenceRule ir, const Formula& f, const Unit& parent)
  :
  _counter (1),
  _untype (FORMULA),
  _intype (parent.inputType()),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _formula (f),
  _defType (DT_NONE),
  _inference (new Inference(ir,parent))
{
  TRACER( "Unit::Data::Data-5" );
} 



Unit::Data::Data (InferenceRule ir, const Clause& c, const Unit& parent)
  :
  _counter (1),
  _untype (CLAUSE),
  _intype (parent.inputType()),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _clause (c),
  _defType (DT_NONE),
  _inference (new Inference(ir,parent))
{
  TRACER( "Unit::Data::Data-6" );
} 



Unit::Data::Data (const Clause& c,
		  const UnitList& parents,
		  long number,
		  VampireKernel::Clause::BackgroundFlagIterator& flag)
  :
  _counter(1),
  _untype(CLAUSE),
  _intype(parents.maxInputType()),
  _number(number),
  _numberInRefutation(0),
  _clause(c),
  _defType(DT_NONE),
  _inference (new Inference(parents,flag))
{
  TRACER( "Unit::Data::Data-7" );
} 



Unit::Data::Data (const Formula& f, const UnitList& parents)
  :
  _counter(1),
  _untype(FORMULA),
  _intype(parents.maxInputType()),
  _number(++ _lastNumber),
  _numberInRefutation(0),
  _formula(f),
  _defType(DT_NONE),
  _inference(new Inference(IR_KERNEL,parents))
{
  TRACER( "Unit::Data::Data-8" );
} 



Unit::Data::~Data ()
{
  TRACER( "Unit::Data::~Data" );

  ASS (_counter == 0);
  delete _inference;
} 








XML::Element* Unit::toXML() const
{
  XML::DeepElement* unit = new XML::DeepElement("unit");
  unit->addAttribute(new XML::TextAttribute("input_type",
					    toString(inputType())));
  unit->addAttribute(new XML::LongAttribute("number",number()));
  if (numberInRefutation()) {
    unit->addAttribute(new XML::IntegerAttribute("refutation_number",
						 numberInRefutation()));
  }

  unit->addChild(new XML::TextElement("is_definition",
					   toString(defType())));

  if (name() != "") {
    unit->addChild(new XML::TextElement("name",name()));
  }

  unit->addChild(inference()->toXML());

  if (unitType() == CLAUSE) {
    
    unit->addChild(new XML::TextElement("clause",clause().toString()));

  }
  else { 
    
    unit->addChild(new XML::TextElement("formula",formula().toString()));

  }

  return unit;
} 



string Unit::toString () const
{
  string prefix = Int::toString(number()) + ". ";
  string postfix = string(" (") + toString(inputType());
  if (name() != "") {
    postfix += string(",name=") + name();
  }
  postfix += ")";

  switch (unitType()) {
  case CLAUSE:
    return prefix + clause().toString() + postfix;
  case FORMULA:
    return prefix + formula().toString() + postfix;
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



string Unit::toStringAsInference () const
{
  static string stars = "***********";
  UnitList parents (inference()->premises());

  string result = stars + " [";
  Iterator<Unit> ps1 (parents);
  while (ps1.hasNext()) {
    result += Int::toString(ps1.next().number());
    if (ps1.hasNext()) {
      result += ",";
    }
  }

  if (parents.isNonEmpty()) {
    result += "->";
  }
  Inference* inf = inference();
  string infString = inf->rule() == IR_KERNEL ?
    static_cast<const KernelInferenceInfo*>(inf->info())->rulesAsString() :
    Inference::toString(inf->rule());

  result += Int::toString(number()) + ", " + 
            infString + "] " + 
            stars + "\n";

  Iterator<Unit> ps (parents);
  while (ps.hasNext()) {
    result += "  ";
    Unit nxt (ps.next());
    switch (nxt.unitType()) {
    case CLAUSE:
      result += nxt.clause().toString();
      break;
    case FORMULA:
      
      result += nxt.formula().toString();
      break;
    }
    result += "\n";
  }

  if (parents.isNonEmpty()) {
    result += "-----------------------------\n  ";
  }

  switch (unitType()) {
  case CLAUSE:
    result += clause().toString();
    break;
  case FORMULA:
    result += formula().toString();
    break;
  }
  return result + "\n";
} 



InputType Unit::max (InputType type1, InputType type2)
{
  return type1 > type2 ? type1 : type2;
}



InputType UnitList::maxInputType () const
{
  InputType result = AXIOM;
  Iterator<Unit> us(*this);
  while (us.hasNext()) {
    switch (us.next().inputType()) {
    case AXIOM:
      break;
    case ASSUMPTION:
      result = ASSUMPTION;
      break;
    case CONJECTURE:
      return CONJECTURE;
    }
  }

  return result;
} 
