//
// File:         AutoSymbolAndHeaderPrecedence.hpp
// Description:  Automatic assignment of precedences to the functors and predicates of the signature.
// Created:      Apr 24, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      May 20, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in
//               headerPrecedence(bool negative,ulong predNum,long& precedence)
//====================================================
#ifndef AUTO_FUNCTOR_PRECEDENCE_H
//=================================================
#define AUTO_FUNCTOR_PRECEDENCE_H
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "Tuple.hpp"
#include "GSortTree.hpp"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_AUTO_FUNCTOR_PRECEDENCE
 #define DEBUG_NAMESPACE "AutoSymbolAndHeaderPrecedence"
#endif 
#include "debugMacros.hpp"
//=================================================  

namespace VK
{
class AutoSymbolAndHeaderPrecedence
{
private:
  typedef BK::Tuple2<ulong,ulong> ArityFuncPair; 
public:
  typedef BK::GSortTree<BK::GlobAlloc,ArityFuncPair,AutoSymbolAndHeaderPrecedence> ArityFuncSortTree;  
  // can not make it private because some of the nested typedefs/classes 
  // have static data members

 public:
  AutoSymbolAndHeaderPrecedence();
  ~AutoSymbolAndHeaderPrecedence();
    
  void setDependenceOnArity(long typeOfDependence) { _dependenceOnArity = typeOfDependence; };

  void setInversePolarityFlag(bool flag) { _inversePolarity = flag; };

  void setKinkyHeaderPrecedenceFlag(bool flag) { _kinkyHeaderPrecedence = flag; };


  bool symbolPrecedenceIsReserved(long precedence,ulong& symNum);

  bool headerPrecedenceIsReserved(long precedence,bool& negative,ulong& predNum);

  void assignFixedSymbolPrecedence(ulong symNum,long precedence);

  void assignFixedHeaderPrecedence(bool negative,ulong predNum,long precedence);

  void assignMaximalSymbolPrecedence(ulong symNum);

  static long maximalSymbolPrecedence()
    {
      return LONG_MAX/8;
    };

  static long minimalSymbolPrecedence()
    {
      return -(LONG_MAX/8);
    };

  static long maximalHeaderPrecedence()
    {
      return LONG_MAX/8;
    };

  static long minimalHeaderPrecedence()
    {
      return -(LONG_MAX/8);
    };

  void assignArity(ulong symNum,ulong arity);
  
  void assignAllPrecedences(); 

  bool symbolPrecedence(ulong symNum,long& precedence)
  {  
    if (!_registeredSymbol.sub(symNum)) return false;
    precedence = _symbolPrecedence.sub(symNum);
    return true; 
  };
  
  bool headerPrecedence(bool negative,ulong predNum,long& precedence) 
  {
   CALL("headerPrecedence(bool negative,ulong predNum,long& precedence)"); 

   //DF; cout << "PN = " << predNum << "\n";

   if (!_registeredSymbol.sub(predNum)) return false;
   if (negative)
     {
       precedence = _negativeHeaderPrecedence.sub(predNum);
     }
   else
     {
       precedence = _positiveHeaderPrecedence.sub(predNum);
     };
   return true;
  }; // bool headerPrecedence(bool negative,ulong predNum,long& precedence) 


 private: 
  typedef BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,128UL> BoolArray;
  typedef BK::Array<BK::GlobAlloc,long,BK::FastRelocation<long>,128UL> LongArray;
  typedef BK::Array<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,128UL> ULongArray;
 private: 
  long normalisedSymbolPrecedence(long inputPrecedence) 
  {
    return inputPrecedence*2; 
  };
  long normalisedHeaderPrecedence(long inputPrecedence) 
  {
    CALL("normalisedHeaderPrecedence(long inputPrecedence)");
    ASSERT(inputPrecedence != 0L);
    return inputPrecedence * 4; 
  };
  long previousNormalisedSymbolPrecedence(long normalisedPrecedence) 
  { 
    return normalisedPrecedence - 2; 
  };
  void assignSmallerSymbolPrecedence(ulong symNum); 
  static long absval(long num) { if (num >= 0) return num; return -num; }; 
  static long maximalOf(long x,long y) { if (x > y) return x; return y; };
 private: 
  long _dependenceOnArity;
  bool _inversePolarity;
  bool _kinkyHeaderPrecedence;
  long _someSymbolPrecedenceSet;
  long _someHeaderPrecedenceSet;
  long _minimalFixedSymbolPrecedence;
  long _maximalFixedSymbolPrecedence;
  long _minimalFixedHeaderPrecedence;
  long _maximalFixedHeaderPrecedence;
  ulong _minSymbolNumber;
  ulong _maxSymbolNumber;
  LongArray _symbolPrecedence; 
  BoolArray _symbolPrecedenceIsFixed;
  LongArray _positiveHeaderPrecedence; 
  BoolArray _positiveHeaderPrecedenceIsFixed;
  LongArray _negativeHeaderPrecedence; 
  BoolArray _negativeHeaderPrecedenceIsFixed;
  ULongArray _arity;
  BoolArray _registeredSymbol;
}; // class AutoSymbolAndHeaderPrecedence

}; // namespace VK

//==================================================
#endif
