//
// File:         Chaining.cpp
// Description:  Ordered chaining for transitive predicates.  
// Created:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Computation of main premise residual size was optimised.
//==================================================================
#include "Chaining.hpp"
#include "Clause.inl"
#include "RuntimeError.hpp"
#include "prefix.hpp"
#include "variables.hpp"
#include "Term.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
//===================================================================


Chaining::Chaining(OR_INDEX* unifIndex) :
  OPEN_BR(1,"br 1") DCOMMA
  _statistics(0),
  _transitivity(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _queryVarIndex(0UL),
  _storedVarIndex(1UL),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _pseudoHeader((TERM::Arity)1,(TERM::Functor)0,Signature::PositivePolarity),
  _dummyClauseCollector(DOP("Chaining::_dummyClauseCollector")),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false)  
{
  _residualQuerySizeFormula.SetIndex(_queryVarIndex);
  _unificationMachine.SetIndex(_storedVarIndex);
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; // Chaining::Chaining()

Chaining::~Chaining()
{  
  if (DestructionMode::isThorough())
    { 
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
      ASSERT(_storedConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
    };
}; // Chaining::~Chaining()

bool Chaining::loadQuery(Clause* cl)
{
  CALL("loadQuery(Clause* cl)");
  _query = cl;
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      const Clause* transLaw = 
	_transitivity->transitivityLaw(lit->functor());
      if (transLaw)
	{
	  // at least one eligible literal
	  _outerContinuation = LoadForwardPositive;
	  // collect residual query size formula
	  _residualQuerySizeFormula.Reset();
	  ulong remainingSelected = cl->LitList()->numOfSelectedLits();
	  for (LiteralList::Iterator iter1(cl->LitList()); 
	       iter1.notEnd(); 
	       iter1.next())
	    {
	      if (remainingSelected)
		{
		  if (!_transitivity->transitivityLaw(iter1.currentLiteral()->functor()))
		    {
		      // residual literal
		      _residualQuerySizeFormula.Compile(iter1.currentLiteral());
		    };
		  --remainingSelected;
		}
	      else //  nonselected literal
		if (!iter1.currentLiteral()->isAnswerLiteral())
		  {
		    // residual literal
		    _residualQuerySizeFormula.Compile(iter1.currentLiteral());
		  };
	    };
	  
	  
	  if (_queryPremise.Load(_queryVarIndex,_query)) return true;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	}; // if (transLaw) 
      --numOfSelectedLits;
    };
  return false;
}; // bool Chaining::loadQuery(Clause* cl)

bool Chaining::nextInference()
{
  CALL("nextInference()");
  switch (_outerContinuation)
    {
    case LoadForwardPositive:      
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardPositive:
      if (nextForwardPositive())
	{
	  _outerContinuation = NextForwardPositive;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextBackwardPositive:
      if (nextBackwardPositive())
	{
	  _outerContinuation = NextBackwardPositive;
	  return true;
	};
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardNegative1:
      if (nextForwardNegative1())
	{
	  _outerContinuation = NextForwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextBackwardNegative1:
      if (nextBackwardNegative1())
	{
	  _outerContinuation = NextBackwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardNegative2:
      if (nextForwardNegative2()) 
	{
	  _outerContinuation = NextForwardNegative2;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here
      

    case NextBackwardNegative2:
      if (nextBackwardNegative2())
	{
	  _outerContinuation = NextBackwardNegative2;
	  return true;
	};
      return false;

#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
      return false;
#else
#  ifdef _SUPPRESS_WARNINGS_
      RuntimeError::report("Wrong _outerContinuation value in Chaining::nextInference()");
      return false;
#  endif
#endif
    };

  ICP("ICP1");
  RuntimeError::report("Illegal control point in Chaining::nextInference()");
  return false;
}; // bool Chaining::nextInference()


bool Chaining::nextForwardPositive()
{
  CALL("nextForwardPositive()");
  _forwardMode = true;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	  
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Positive());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
      _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      
      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      // first inference without factoring
      if (!buildResolvent(_queryTail,
			  _currentCandidate->chainingTail(),
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,
			      _currentCandidate->chainingTail(),
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardPositive()



bool Chaining::nextBackwardPositive()
{
  CALL("nextBackwardPositive()");
  _forwardMode = false;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	 
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};  
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Positive());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize > _clauseSizeLimit)
	{ 
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 


      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;        
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};


      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _queryTail,
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _queryTail,
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardPositive()


bool Chaining::nextForwardNegative1()
{
  CALL("nextForwardNegative1()");
  _forwardMode = true;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      // non-singleton variable restriction
      // does not apply to this instance of the rule 
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  _newLitHeader.Inversepolarity();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg1ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 


      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};

      if (subsumeMainPremise()) goto try_next_leaf;  
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      // first inference without factoring
      if (!buildResolvent(_queryTail,
			  _currentCandidate->chainingTail(),
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,
			      _currentCandidate->chainingTail(),
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardNegative1()


bool Chaining::nextBackwardNegative1()
{
  CALL("nextBackwardNegative1()");
  _forwardMode = false;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Positive())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Negative());
  
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      
      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _queryTail,
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _queryTail,
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardNegative1()



bool Chaining::nextForwardNegative2()
{
  CALL("nextForwardNegative2()");
  _forwardMode = true;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      // non-singleton variable restriction
      // does not apply to this instance of the rule 
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  _newLitHeader.Inversepolarity();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);

	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _queryTail,
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _queryTail,
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardNegative2()


bool Chaining::nextBackwardNegative2()
{
  CALL("nextBackwardNegative2()");
  _forwardMode = false;
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.lastSelected()) return false;
  lit = _queryPremise.nthSelectedLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Positive())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if (_headTailComparison == Less)
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if ((_headTailComparison != Greater) &&
      (_headTailComparison != Equal))
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreater();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since
	  // there may be several factoring inferences
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  
	  
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      if (!_storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      // first inference without factoring
      if (!buildResolvent(_queryTail,
			  _currentCandidate->chainingTail(),
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,
			      _currentCandidate->chainingTail(),
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardNegative2()



bool Chaining::posChainingLitSelectedIn(Clause* cl)
{
  CALL("posChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Positive()) return true;
	};
      --numOfSelectedLits;
    };
  // no appropriate literal
  return false;
}; // bool Chaining::posChainingLitSelectedIn(Clause* cl)

bool Chaining::negChainingLitSelectedIn(Clause* cl)
{
  CALL("negChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Negative()) return true;
	};
      --numOfSelectedLits;
    };
  // no appropriate literal
  return false;
}; // bool Chaining::negChainingLitSelectedIn(Clause* cl)



bool Chaining::buildResolvent(PrefixSym* newLitArg1,
			      const TERM& newLitArg2,
			      bool inheritFromQuery)
{
  CALL("buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2,..)");
  ASSERT(newLitArg2.isVariable() || newLitArg2.IsReference()); 

  if ((!_query->Active())
      || (!_storedPremise.TheClause()->Active()))
    return false;

  if (_residualQuerySize + 
      _storedPremise.TheClause()->practicallyMinResidualSize() > 
      _clauseSizeLimit)
    {
      if (_forwardMode)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.discardedBySize);
	}
      else
	{
	  ++(_statistics->inferences.deduction.backwardChaining.discardedBySize);
	};

      return false;
    }; 

  _clauseCollector->openNewClause(((_forwardMode) ? 
				   ClauseBackground::ForwTransChain 
				   :
				   ClauseBackground::BackTransChain 
				   ),
				  _query,
				  _storedPremise.TheClause(),
				  const_cast<Clause*>(_currentTransitivityLaw));
  _clauseCollector->announceNewClause();

  LiteralProperties::SLModifier newLitSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier querySLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier storedSLMod = LiteralProperties::DefaultSLModifier;
  
  if (_queryPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_queryPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  querySLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_queryPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  querySLMod = LiteralProperties::PostponeSelection;	  
	};
    };

  if (_storedPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_storedPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  storedSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_storedPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  storedSLMod = LiteralProperties::PostponeSelection;
	};
    };


  if (!_queryPremise.collectUnmarkedLiterals(*_clauseCollector,querySLMod))
    return false;
  if (!_storedPremise.collectUnmarkedLiterals(*_clauseCollector,storedSLMod))
    return false;

  LiteralProperties newLitProp;
  if (inheritFromQuery)
    {
      newLitProp.inheritFrom(_queryPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    }
  else
    { 
      newLitProp.inheritFrom(_storedPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    };
  newLitProp.setDeploymentMode(LiteralProperties::DefaultDeploymentMode);

  return
    _clauseCollector->openNewNonpropLiteral(_newLitHeader,newLitProp) &&
    _clauseCollector->pushTermWithUnifier(newLitArg1) &&
    _clauseCollector->pushTermWithUnifier(_storedVarIndex,newLitArg2) &&
    _clauseCollector->closeNonpropLiteral() &&
    _clauseCollector->finaliseClause();
}; // bool Chaining::buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2,..)



bool Chaining::buildResolvent(const TERM& newLitArg1,
			      PrefixSym* newLitArg2,
			      bool inheritFromQuery)
{
  CALL("buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg,..)");
  
  if ((!_query->Active())
      || (!_storedPremise.TheClause()->Active()))
    return false;

  if (_residualQuerySizeFormula.Compute() + 
      _storedPremise.TheClause()->practicallyMinResidualSize() > 
      _clauseSizeLimit)
    {
      if (_forwardMode)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.discardedBySize);
	}
      else
	{
	  ++(_statistics->inferences.deduction.backwardChaining.discardedBySize);
	};

      return false;
    }; 

  _clauseCollector->openNewClause(((_forwardMode) ? 
				   ClauseBackground::ForwTransChain 
				   :
				   ClauseBackground::BackTransChain 
				   ),
				  _query,
				  _storedPremise.TheClause(),
				  const_cast<Clause*>(_currentTransitivityLaw));
  _clauseCollector->announceNewClause();


  LiteralProperties::SLModifier newLitSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier querySLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier storedSLMod = LiteralProperties::DefaultSLModifier;
  
  if (_queryPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (inheritFromQuery)
	newLitSLMod = LiteralProperties::AccelerateSelection; 
 
      if (_queryPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  querySLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_queryPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  querySLMod = LiteralProperties::PostponeSelection;	  
	};
    };

  if (_storedPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {      
      if (!inheritFromQuery)
	newLitSLMod = LiteralProperties::AccelerateSelection; 
 
      if (_storedPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  storedSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_storedPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  storedSLMod = LiteralProperties::PostponeSelection;
	};
    };


  if (!_queryPremise.collectUnmarkedLiterals(*_clauseCollector,querySLMod))
    return false;
  if (!_storedPremise.collectUnmarkedLiterals(*_clauseCollector,storedSLMod))
    return false;

  LiteralProperties newLitProp;
  if (inheritFromQuery)
    {
      newLitProp.inheritFrom(_queryPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    }
  else
    { 
      newLitProp.inheritFrom(_storedPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    };
  newLitProp.setDeploymentMode(LiteralProperties::DefaultDeploymentMode);


  return 
    _clauseCollector->openNewNonpropLiteral(_newLitHeader,newLitProp) &&
    _clauseCollector->pushTermWithUnifier(_storedVarIndex,newLitArg1) &&
    _clauseCollector->pushTermWithUnifier(newLitArg2) &&
    _clauseCollector->closeNonpropLiteral() &&
    _clauseCollector->finaliseClause();
    
}; // bool Chaining::buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg2,..)



bool Chaining::nextFactoring()
{
  CALL("nextFactoring()");
  
  if (_queryPremise.WasFactored())
   {
    if (_queryPremise.GetNextFactor())
     {
      return true;
     }
    else 
     {
      _queryPremise.ResetFactoring();
      if (_storedPremise.WasFactored())
       {
        return _storedPremise.GetNextFactor();
       }
      else
       {
        return _storedPremise.GetFirstFactor(_storedPremise.CutLit()+1);
       };
     };
   }
  else 
   {
    if (_queryPremise.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (_storedPremise.WasFactored())
       {
        return _storedPremise.GetNextFactor();
       }
      else
       {
        return _storedPremise.GetFirstFactor(_storedPremise.CutLit()+1);
       };
     };
   };
}; // bool Chaining::nextFactoring()

bool Chaining::subsumeMainPremise()
{
  CALL("subsumeMainPremise()"); 
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_query->hasSubsumedPremiseInstance()) return false;
  _query->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 

  if (collectMainPremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause)) 
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _query);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _query->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  if (_forwardMode)
    {
      ++(_statistics->inferences.deduction.forwardChaining.mainPremisesSubsumed);
    }
  else
    {
      ++(_statistics->inferences.deduction.backwardChaining.mainPremisesSubsumed);
    };
 discarded:
  _query->unblockForSubsumption();
  return true;
}; // bool Chaining::subsumeMainPremise()

bool Chaining::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_query->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_queryVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; // bool Chaining::collectMainPremiseInstance()



bool Chaining::subsumeSidePremise()
{
  CALL("subsumeSidePremise()"); 
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_currentCandidate->TheClause()->hasSubsumedPremiseInstance()) return false;
  _currentCandidate->TheClause()->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 

  if (collectSidePremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause))  
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _currentCandidate->TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _currentCandidate->TheClause()->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  if (_forwardMode)
    {
      ++(_statistics->inferences.deduction.forwardChaining.sidePremisesSubsumed);
    }
  else
    {
      ++(_statistics->inferences.deduction.backwardChaining.sidePremisesSubsumed);
    }; 
 discarded:
  _currentCandidate->TheClause()->unblockForSubsumption();
  return true;
}; // bool Chaining::subsumeSidePremise()

bool Chaining::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");
  for(LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_storedVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; // bool Chaining::collectSidePremiseInstance()




bool VK::Chaining::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      if (_forwardMode)
		{
		  ++(_statistics->inferences.deduction.forwardChaining.mainPremisesDemodulated);
		}
	      else
		{
		  ++(_statistics->inferences.deduction.backwardChaining.mainPremisesDemodulated);
		};
	      
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() &&
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {	  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);   
	      ASSERT(simplifyingClause != _query);
	      ASSERT(simplifyingClause != _currentCandidate->TheClause());
	      
	      if (simplifyingClause)
		{	
		  _query->protectParamodulationsInto();
		  _currentCandidate->TheClause()->protectParamodulationsInto();
		  if (_forwardMode)
		    {
		      ++(_statistics->inferences.deduction.forwardChaining.mainPremisesDemodulated);
		    }
		  else
		    {
		      ++(_statistics->inferences.deduction.backwardChaining.mainPremisesDemodulated);
		    };
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (_forwardMode)
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
		}
	      else
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
		};
	      return tautology;
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
      if (collectSidePremiseSubtermsAvailableForParamodulationInto() &&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{		    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	  ASSERT(simplifyingClause != _query);
	  ASSERT(simplifyingClause != _currentCandidate->TheClause());

	  if (simplifyingClause)
	    {	
	      _query->protectParamodulationsInto();
	      _currentCandidate->TheClause()->protectParamodulationsInto();
	      if (_forwardMode)
		{
		  ++(_statistics->inferences.deduction.forwardChaining.sidePremisesDemodulated);
		}
	      else
		{
		  ++(_statistics->inferences.deduction.backwardChaining.sidePremisesDemodulated);
		};
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	 
	  bool tautology = 
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  return tautology; 
	};
    }
  else
    return false;
}; // bool VK::Chaining::demodulateSelectedLits()



bool 
VK::Chaining::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");
  LiteralProperties prop;
  ulong numOfSelectedLits = _query->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_query->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_queryVarIndex,
								      TERM::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_queryVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_queryVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::Chaining::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::Chaining::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _currentCandidate->TheClause()->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_storedVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_storedVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_storedVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::Chaining::collectSidePremiseSubtermsAvailableForParamodulationInto()




//=================================================================

