//
// File:         clause.hpp
// Description:  Representation of stored clauses.
// Created:      Feb 13, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
//====================================================
#ifndef CLAUSE_H
//===========================================================================
#define CLAUSE_H
#include <iostream>
#include <cstddef>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#ifdef DEBUG_ALLOC_OBJ_TYPE
#  include "ClassDesc.hpp"       
#endif
#include "GlobAlloc.hpp"
#include "GList.hpp"
#include "BitWord.hpp"
#include "Term.hpp"
#include "ClauseBackground.hpp"
#include "LiteralList.hpp"
#include "ExpandingStack.hpp"
#include "DefinedHeaderTable.hpp"
#include "SymbolInfo.hpp"
using namespace std;
//===========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
//================================================= 
namespace VK
{                           
class Clause     
{
 public:
  class Context;
  class Traversal;
  class OpenUnshared; 
  class NewNumber;
 private:
  class ClauseList {};  
 public: 
  typedef BK::GList<BK::GlobAlloc,Clause*,ClauseList> List;
  enum MainSet
  {
    MainSetFree = 0UL, // fixed 
    MainSetNew = 1UL,
    MainSetPassive = 2UL,
    MainSetReservedPassive = 3UL, 
    MainSetActive = 4UL,
    MainSetDefinitions = 5UL,

    // end of active sets

    MainSetTrash = 6UL,
    MainSetHistory = 7UL,
    MainSetAnswers = 8UL,

    // static clause sets
    
    MainSetInputQuery = 9UL,
    MainSetInputSupport = 10UL,
    MainSetPreprocessingHistory = 11UL,
    MainSetNameIntro
  };
  enum FlagNumber
  {
    MainSetPieceBegin = 0UL,
    MainSetPieceEnd = 3UL, // allows values 0UL - 15UL

    FlagLiteralsShared = 4UL,
    FlagInForwardSubsumptionIndex = 5UL,
    FlagInBackwardSubsumptionIndex = 6UL,    
    FlagInForwardDemodulationIndex = 7UL,
    FlagInBackDemodulationIndex = 8UL, 
    FlagInInferenceDispatcherAsActive = 9UL,
    FlagInInferenceDispatcherAsDefinition = 10UL,
    FlagInClauseNumIndex = 11UL,
    FlagInQueueForBackSimplification = 12UL,
    FlagInQueueForIntegrationIntoBackwardDemodIndex = 13UL,
    FlagHasSubsumedPremiseInstance = 14UL,
    FlagBlockedForSubsumption = 15UL,
    FlagBlockedForDemodulationFrom = 16UL,
    FlagParamodulationsIntoProtected = 17UL,
    FlagIsNameIntro = 18UL,
    FlagAllMaximalLiteralsSelected = 19UL,
    FlagContainsAnswerLiterals = 20UL, 
    FlagHasChildren = 21UL,

    ProtectionCounterBegin = 27UL, // allows values 0UL - 31UL
    ProtectionCounterEnd = 31UL
  };
 public:
  Clause(LiteralList* lits);

  // for making arrays of dummy clauses in KEPT and CLAUSE_SET 
  Clause();
  
  ~Clause() {};
  void* Clause::operator new(size_t);
  void Clause::operator delete(void* obj);

  void clearFlags();
  void setMainSet(const MainSet& ms);
  MainSet mainSet() const { return (MainSet)_flags.piece(MainSetPieceBegin,MainSetPieceEnd); };
 
  bool flag(const FlagNumber& fn) const { return _flags.bit(fn); };
  
  void setFlag(const FlagNumber& fn) { _flags.set(fn); };
  void clearFlag(const FlagNumber& fn) { _flags.clear(fn); };

#ifndef NO_DEBUG 
  bool checkObj() const;
#endif  

  Clause* Next() const { return next; };
  void SetNext(Clause* c) { next = c; };
  Clause* Previous() const { return previous; };
  void SetPrevious(Clause* p) { previous = p; };
  LiteralList* LitList() const;
  const TERM* firstLiteral() const;
  const ClauseBackground& background() const;
  ClauseBackground& background();
  void setBackground(const ClauseBackground& a) { _background = a; };
  void* inputClauseOrigin() const;
  void setInputClauseOrigin(void* orig);
  ulong Number() const;

  void SetNumber(long n) { number = n; };
  long inferenceDepth() const { return _background.inferenceDepth(); };
  long weight() const;
  long weightBase() const { return _weightBase; };
  long size() const;  // logically different from weight()!
  void AssignMaxPriority();
  bool MaxPriority() const;
  long NumOfPosEq() const;
  ulong NumOfEqLits() const;
  bool containsEquality() const;
  long numOfPosEq() const { return static_cast<long>(num_of_pos_eq); };
  void SetNumOfPosEq(long n) { num_of_pos_eq = n; };
  long practicallyMinResidualSize() const;
  void setPracticallyMinResidualSize(long mcs) {  _practicallyMinResidualSize = mcs; };
  ulong numOfAllLiterals() const;
  ulong numOfNonanswerLiterals() const;
  void setSize(long s) { _size = s; };
  void setWeightBase(long w) { _weightBase = w; };
  bool isInputClause() const;
  bool isInputOrPreprocessedInputClause() const;
  bool isSubgoal() const { return _background.isSubgoal(); };
  bool isBuiltInTheoryFact() const { return _background.isBuiltInTheoryFact(); };
  bool isSupportClause() const { return _background.isSupport(); };
  bool isNonSubsuming() const { return _background.isNonSubsuming(); };
  bool isNonSubsumable() const { return _background.isNonSubsumable(); };
  void markAsQuery() { _background.markAsQuery(); };
  void markAsBuiltInTheoryFact() { _background.markAsBuiltInTheoryFact(); };
  bool Active() const;
  bool isProtected() const;
  void protect();
  void unprotect();

  bool hasSubsumedPremiseInstance() const;
  void registerPremiseInstanceSubsumption();
  bool paramodulationsIntoProtected() const;
  void protectParamodulationsInto();
  bool isBlockedForSubsumption() const;
  void blockForSubsumption();
  void unblockForSubsumption();
  bool isBlockedForDemodulationFrom() const;
  void blockForDemodulationFrom();
  void unblockForDemodulationFrom();
  void MakeVIP() { _background.MakeVIP(); };
  bool IsVIP() const { return _background.IsVIP(); };  
  void MarkAsName();
  bool IsSplitName() const;
  void markAsNameIntro() { setFlag(FlagIsNameIntro); };
  bool isNameIntro() const { return flag(FlagIsNameIntro); };
  ulong NumberOfLiteral(const TERM* lit) const;
  bool Empty() const;
  bool isAnswer() const;
  bool Unit() const;
  bool PositiveUnitEquality() const;
  bool PreorderedPositiveUnitEquality() const;
  bool posEqSelected() const;
  bool MakeDefinitionOf(const TERM& definedHeader);
  bool DefinitionUnfoldingSelected(const DefinedHeaderTable* definedHeaderTable) const;
  bool Definition(const DefinedHeaderTable* definedHeaderTable,TERM& definedHeader) const;
  bool HornClause() const;
  bool IsRule() const; // <=> contains both positive and negative nonsplitting literals 
  bool isSymmetryLaw() const;
  bool isTotalityLaw() const;
  bool isCommutativityLaw(ulong& func) const;
  bool isTransitivityLaw() const;
  bool isTransitivityLaw(ulong& pred) const;
  bool ContainsEliminated(const SymbolInfo* predElimTable) const;
  bool singletonVar(ulong var) const;
  void Destroy();
  void DestroyUnshared();
  static void InsertAfter(Clause* c,Clause* db);
  static void InsertBefore(Clause* c,Clause* db);
  static void Remove(Clause* c);
  static Clause* MakeEmptyClause();
  Clause* unsharedCopy() const;
  bool IsOrphan() const;
  bool allMaximalSelected() const;

  //=========== output: ======================
  ostream& outputBackground(ostream& str) const;
  ostream& output(ostream& str) const;
  ostream& outputAsPrologTerm(ostream& str) const;     
  

 private:  // structure
  Clause* next;
  Clause* previous;
  LiteralList* litList; 
  ClauseBackground _background;
  long number;
  long _size;
  long _weightBase;
  long _practicallyMinResidualSize;
  BK::BitWord _flags;   
  ulong num_of_pos_eq;
}; // class Clause
}; // namespace VK

namespace std
{
inline ostream& operator<<(ostream& str,const VK::Clause& cl)
{
  return cl.output(str);
}; 

inline ostream& operator<<(ostream& str,const VK::Clause* cl)
{
  return cl->output(str);
}; 
}; // namespace std


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::Traversal"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{
class Clause::Traversal
{
 public:
  enum State 
  { 
    EndOfClause,
    LitHeader,
    Func,
    Var               
  };
 public: 
  Traversal() : termTrav(), iter() { CALL("constructor Traversal()"); };
  Traversal(const Clause* c);   
  ~Traversal() { CALL("destructor ~Traversal()"); };
  void reset(const Clause* c);  
  const State& state() const { return currState; };
  void next();
  void after();
  const TERM& symbol() const { return termTrav.symbol(); };
  const TERM& term() const { return termTrav.term(); };
 private:
  State currState;
  TERM::Traversal termTrav;
  const TERM* nextLit;
  LiteralList::Iterator iter;
}; // class Clause::Traversal 

}; // namespace VK


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::OpenUnshared"
#endif
#include "debugMacros.hpp"
//=================================================

 
namespace VK
{
class Clause::OpenUnshared
{
 public:
  OpenUnshared();
  ~OpenUnshared();
  void reset();
  bool litHeader(const TERM& hd,const LiteralProperties& prop);
  bool functor(const TERM& f);
  void variable(const TERM& v) { _currLit.variable(v); };
  
  void ancestor(Clause* cl);
  void usedRule(ClauseBackground::Rule r) { _background.UsedRule(r); }; 
  void markAsSubgoal() { _background.markAsSubgoal(); };
  void markAsSupport() { _background.markAsSupport(); };
  Clause* result();
  void recycle();
 private: 
  ulong _length;
  ulong _size;
  ulong _numOfPosEq;
  TERM::OpenUnshared _currLit;
  LiteralProperties _currLitProperties;
  BK::ExpandingStack<BK::GlobAlloc,TERM*,BK::FastRelocation<TERM*>,32> _literals;
  BK::ExpandingStack<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32> _literalProperties;
  OpenClauseBackground _background;
}; // class Clause::OpenUnshared

}; // namespace VK


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::NewNumber"
#endif 
#include "debugMacros.hpp"
//=================================================

namespace VK
{
class Clause::NewNumber
{
 public:
  NewNumber() : _nextNumber(0UL) {};
  NewNumber(ulong n) : _nextNumber(n) {};
  void reset(ulong n) { _nextNumber = n; };
  ulong nextNumber() const { return _nextNumber; };
  ulong generate() { ++_nextNumber; return _nextNumber - 1; };
 private:
  ulong _nextNumber;
}; // class Clause::NewNumber

}; // namespace VK




//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::Context"
#endif 
#include "debugMacros.hpp"
//=================================================

namespace VK
{
  
  class Clause::Context
  {
  public:
    class AutoActivator
      {
      public:
	AutoActivator()  : _previous(Context::getCurrent())
	{
	  Context::setCurrent(static_cast<Context*>(0));
	};
	AutoActivator(Context* c) : _previous(Context::getCurrent())
	{
	  Context::setCurrent(c);
	};
	~AutoActivator()
	{
	  Context::setCurrent(_previous);
	};
      private:
	Context* _previous;
    }; // class AutoActivator
  public:
    Context() :
#ifdef DEBUG_ALLOC_OBJ_TYPE
      _classDesc("Clause",sizeof(Clause),sizeof(Clause)),
#endif
      _selectedPosEqPenaltyCoefficient(1.0),
      _useSelectedPosEqPenaltyCoefficient(false),
      _nongoalPenaltyCoefficient(1.0),
      _useNongoalPenaltyCoefficient(false),
      _showLiteralProperties(false)
    {
    };
    ~Context() {};
    
    
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
   
    void setSelectedPosEqPenaltyCoefficient(const float& c) 
    { 
      _selectedPosEqPenaltyCoefficient = c;
      _useSelectedPosEqPenaltyCoefficient = (c != 1.0);
    }
    void setNongoalPenaltyCoefficient(const float& c) 
    { 
      _nongoalPenaltyCoefficient = c; 
      _useNongoalPenaltyCoefficient = (c != 1.0);
    };
    void setShowLiteralProperties(bool fl) { _showLiteralProperties = fl; };
    
    long computeWeight(long weightBase,
		       bool subgoal,
		       bool posEqSelected) const;
		      
  private:
#ifdef DEBUG_ALLOC_OBJ_TYPE
    BK::ClassDesc _classDesc;
#endif
    float _selectedPosEqPenaltyCoefficient; 
    bool _useSelectedPosEqPenaltyCoefficient;
    float _nongoalPenaltyCoefficient;
    bool _useNongoalPenaltyCoefficient;
    bool _showLiteralProperties;
    static Context* _current;
    friend class Clause;
  }; // class Clause::Context

}; // namespace VK

//===========================================================================
#endif
