//
// File:         clause.inl
// Description:  Representation of stored clauses.
// Created:      Feb 13, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
//=====================================================================
#ifndef CLAUSE_INL
#  define CLAUSE_INL
//======================================================================
#include "Clause.hpp"
#include "LiteralProperties.inl"
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
//======================================================================
namespace VK
{ 
  inline
  Clause::Clause(LiteralList* lits) : 
    litList (lits), 
    num_of_pos_eq(0)
  {
    setMainSet(MainSetFree);      
  };

  // for making arrays of dummy clauses in KEPT and CLAUSE_SET 
  
  inline
  Clause::Clause()
  {
    next = this;
    previous = this;  
    litList = 0;
    setMainSet(MainSetFree);
  };
  
  
  
  
  inline
  void* Clause::operator new(size_t)
  {
    CALL("operator new(size_t)");
#ifdef DEBUG_ALLOC_OBJ_TYPE 
    ALWAYS(Context::getCurrent()->_classDesc.registerAllocated(sizeof(Clause)));
    return BK::GlobAlloc::allocate(sizeof(Clause),&Context::getCurrent()->_classDesc);
#else
    return BK::GlobAlloc::allocate(sizeof(Clause));
#endif
  };
  
  
  inline
  void Clause::operator delete(void* obj)
  { 
    CALL("operator delete(void* obj)");
#ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(Context::getCurrent()->_classDesc.registerDeallocated(sizeof(Clause)));
    BK::GlobAlloc::deallocate(obj,sizeof(Clause),&Context::getCurrent()->_classDesc);
#else
    BK::GlobAlloc::deallocate(obj,sizeof(Clause));
#endif
  };

  
  inline
  void Clause::clearFlags()
  {
    _flags.clear();
  }; 

  
  inline
  void Clause::setMainSet(const MainSet& ms) 
  {
    CALL("setMainSet(const MainSet& ms)");
    ASSERT(!(ms >> 4));   
    _flags.setPiece(MainSetPieceBegin,MainSetPieceEnd,ms);      
    ASSERT(mainSet() == ms);
  };

  inline 
  void Clause::protect()
  {
    CALL("Clause::protect()");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    ASSERT(protCounter < 31UL);
    ++protCounter;
    _flags.setPiece(ProtectionCounterBegin,ProtectionCounterEnd,protCounter);
  };

  inline 
  void Clause::unprotect()
  {
    CALL("Clause::protect()");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    ASSERT(protCounter > 0UL);
    --protCounter;
    _flags.setPiece(ProtectionCounterBegin,ProtectionCounterEnd,protCounter);
  };
  
  inline 
  bool Clause::isProtected() const
  {
    CALL("isProtected() const");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    return protCounter != 0UL;
  };
    

  inline
  LiteralList* Clause::LitList() const 
  {
    CALL("LitList() const");
    ASSERT(checkObj()); 
    return litList; 
  };
  
  inline 
  const TERM* 
  Clause::firstLiteral() const 
  { 
    return  LitList()->firstLit(); 
  };

  
  inline
  const ClauseBackground& Clause::background() const 
  {
    CALL("background() const"); 
    ASSERT(checkObj()); 
    return _background; 
  }; 

  
  inline
  ClauseBackground& Clause::background()
  {
    CALL("background()"); 
    ASSERT(checkObj()); 
    return _background; 
  };  
  
  
  inline
  void* Clause::inputClauseOrigin() const 
  {
    CALL("inputClauseOrigin() const");
    ASSERT(isInputClause());
    return _background.inputClauseOrigin();
  };

  
  inline
  void Clause::setInputClauseOrigin(void* orig)
  {
    CALL("setInputClauseOrigin(void* orig)");
    ASSERT(isInputClause());
    _background.setInputClauseOrigin(orig);
  };

  
  inline
  ulong Clause::Number() const 
  {
    CALL("Number() const");
    ASSERT(checkObj()); 
    return number;
  };

  
  inline
  long Clause::weight() const 
  {
    CALL("weight() const"); 
    ASSERT(checkObj());
    return Context::getCurrent()->computeWeight(_weightBase,isSubgoal(),posEqSelected());
  }; // long weight() const 


  

  
  inline
  long Clause::size() const  // logically different from weight()!
  {
    CALL("size() const"); 
    ASSERT(checkObj());
    if (_size < 0) return ((-1)*_size);
    return _size;
  };

    
  
  inline
  void Clause::AssignMaxPriority() 
  { 
    if (_size > 0) _size = (-1)*_size; 
  };
  
  inline
  bool Clause::MaxPriority() const 
  {
    CALL("MaxPriority() const"); 
    ASSERT(checkObj()); 
    return (_size < 0); 
  };    
  
  inline
  long Clause::NumOfPosEq() const 
  {
    CALL("NumOfPosEq() const"); 
    ASSERT(checkObj()); 
    return num_of_pos_eq; 
  };

  
  inline
  ulong Clause::NumOfEqLits() const 
  {
    CALL("NumOfEqLits() const");   
    ASSERT(checkObj()); 
    ulong res = 0;

    for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
      if (iter.currentElement()->isEquality()) res++;

    return res;       
  };
  
  
  inline
  bool Clause::containsEquality() const 
  {
    CALL("containsEquality() const"); 
    ASSERT(checkObj()); 
    for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
      if (iter.currentElement()->isEquality()) return true;
    return false;
  }; // bool containsEquality() const 

  
  inline
  long Clause::practicallyMinResidualSize() const 
  {
    CALL("practicallyMinResidualSize() const"); 
    ASSERT(checkObj());
    return _practicallyMinResidualSize; 
  };

  
  inline
  ulong Clause::numOfAllLiterals() const 
  { 
    CALL("numOfAllLiterals() const");
    ASSERT(checkObj()); 
    return LitList()->length();
  };

  
  inline
  ulong Clause::numOfNonanswerLiterals() const
  {
    CALL("numOfNonanswerLiterals() const");
    ASSERT(checkObj());
    if (flag(FlagContainsAnswerLiterals))
      {
	ulong res = 0;
	for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
	  if (!iter.currentLiteral()->isAnswerLiteral()) res++; 
	ASSERT(res < numOfAllLiterals());
	return res;  
      }
    else
      return numOfAllLiterals();
  };

  
  inline
  bool Clause::isInputClause() const
  {
    CALL("isInputClause()");
    return _background.isInput();
  };

  inline
  bool Clause::isInputOrPreprocessedInputClause() const
  {
    CALL("isInputOrPreprocessedInputClause() const");
    return _background.isInput() || _background.isPrepro();
  };
    
  
  inline
  bool Clause::Active() const
  {
    CALL("Active() const");
    return  mainSet() < MainSetTrash; 
  };
 
 
  

  
  inline
  bool Clause::hasSubsumedPremiseInstance() const
  {
    CALL("hasSubsumedPremiseInstance() const");
    return flag(FlagHasSubsumedPremiseInstance);
  };

  
  inline
  void Clause::registerPremiseInstanceSubsumption()
  {
    CALL("registerPremiseInstanceSubsumption()");
    setFlag( FlagHasSubsumedPremiseInstance);
  };

  
  inline
  bool Clause::paramodulationsIntoProtected() const
  {
    CALL("paramodulationsIntoProtected() const");
    return flag(FlagParamodulationsIntoProtected);
  };
  
  
  inline
  void Clause::protectParamodulationsInto()
  {
    CALL("protectParamodulationsInto()");
    setFlag(FlagParamodulationsIntoProtected);
  };



  
  inline
  bool Clause::isBlockedForSubsumption() const
  {
    CALL("isBlockedForSubsumption() const");
    return flag(FlagBlockedForSubsumption);
  };
  
  
  inline
  void Clause::blockForSubsumption()
  {
    CALL("blockForSubsumption()");
    ASSERT(!isBlockedForSubsumption());
    setFlag(FlagBlockedForSubsumption);
  };

  
  inline
  void Clause::unblockForSubsumption()
  {
    CALL("unblockForSubsumption()");
    ASSERT(isBlockedForSubsumption());
    clearFlag(FlagBlockedForSubsumption);
  };


  
  inline
  bool Clause::isBlockedForDemodulationFrom() const
  {
    CALL("isBlockedForDemodulationFrom() const");
    return flag(FlagBlockedForDemodulationFrom);
  };
  
  
  inline
  void Clause::blockForDemodulationFrom()
  {
    CALL("blockForDemodulationFrom()");
    // may be already blocked
    setFlag(FlagBlockedForDemodulationFrom);
  };

  
  inline
  void Clause::unblockForDemodulationFrom()
  {
    CALL("unblockForDemodulationFrom()");
    // may be already unblocked
    clearFlag(FlagBlockedForDemodulationFrom);
  };

  
  inline
  void Clause::MarkAsName()
  {
    CALL("MarkAsName()");
    ASSERT(checkObj());
    _background.MarkAsName();
  };

  
  inline
  bool Clause::IsSplitName() const 
  {
    CALL("IsSplitName() const");
    ASSERT(checkObj());
    return (_background.properties().bit(ClauseBackground::Split) 
	    && _background.properties().bit(ClauseBackground::Name));
  };

  
  inline
  ulong Clause::NumberOfLiteral(const TERM* lit) const
  {
    CALL("NumberOfLiteral(TERM* lit) const");
    ASSERT(checkObj());
    ulong i = 0;
    
    for (LiteralList::Iterator iter(LitList());;iter.next())
      {
	ASSERT(i < LitList()->length());
	if (iter.currentLiteral() == lit) return i;
	i++; 
      };
    ICP("0");
  }; // ulong NumberOfLiteral(const TERM* lit)

  
  inline
  bool Clause::Empty() const 
  { 
    CALL("Empty() const"); 
    return (!(numOfAllLiterals())); 
  };
  
  
  inline
  bool Clause::isAnswer() const 
  {
    CALL("isAnswer() const");
    ASSERT(LitList()->numOfSelectedLits() || LitList()->containsOnlyAnswerLiterals());
    return !LitList()->numOfSelectedLits();
  };
   

  
  inline
  bool Clause::Unit() const 
  { 
    CALL("Unit() const"); 
    // Returns true if the clause contains exactly one non-answer literal.  
    if (flag(FlagContainsAnswerLiterals)) 
      {
	ulong n = 0UL;
	for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
	  if (!iter.currentLiteral()->isAnswerLiteral())
	    {
	      if (n) return false;
	      n++;                
	    };
	ASSERT((n == 0UL) || (n == 1UL)); 
	return n != 0UL; 
      } 
    else 
      return (numOfAllLiterals() == 1);
  };

  
  inline
  bool Clause::PositiveUnitEquality() const
  { 
    // Only applicable if literal selection has been
    // applied to the clause.
    if (!Unit()) return false;    
    const TERM* lit = LitList()->firstElem()->literal();
    return (lit->Positive()) && (lit->IsEquality());
  };

  
  inline
  bool Clause::PreorderedPositiveUnitEquality() const
  { 
    // Only applicable if literal selection has been
    // applied to the clause.
    if (!Unit()) return false;    
    const TERM* lit = LitList()->firstElem()->literal();
    return (lit->Positive()) && (lit->IsEquality()) && (lit->IsOrderedEq());
  };

  
  inline
  bool Clause::posEqSelected() const 
  {
    CALL("posEqSelected() const");
    ulong numOfSel = LitList()->numOfSelectedLits();
    for (LiteralList::Iterator iter(LitList()); numOfSel; iter.next())
      {
	if (iter.currentElement()->isEquality() && 
	    iter.currentElement()->positive() &&
	    iter.currentElement()->properties().useEqForParamodulationFrom())
	  return true; 
	numOfSel--;
      };
    return false;
  };

  
  inline
  bool Clause::isTransitivityLaw() const
  {
    ulong dummy;
    return isTransitivityLaw(dummy);
  };

  
  inline
  bool Clause::ContainsEliminated(const SymbolInfo* predElimTable) const
  {
    CALL("ContainsEliminated(const SymbolInfo* predElimTable) const");
    for (LiteralList::Iterator iter(LitList()); iter.notEnd(); iter.next())
      if (predElimTable->isToBeEliminated(iter.currentElement()->functor())) 
	return true;
    return false;
  }; 

  
  inline
  bool Clause::IsOrphan() const 
  {
    CALL("IsOrphan() const");
    ASSERT(checkObj()); 
    for (const ClauseBackground::ClauseList* cl = _background.ancestors(); 
	 cl; 
	 cl = cl->tl())
      if (!cl->hd()->Active()) return true;
    return false;
  };

  
  inline
  bool Clause::allMaximalSelected() const
  {
    CALL("allMaximalSelected() const");
    return flag(FlagAllMaximalLiteralsSelected); 
  };



inline bool Clause::DefinitionUnfoldingSelected(const DefinedHeaderTable* definedHeaderTable) const
{
  CALL("DefinitionUnfoldingSelected(const DefinedHeaderTable* definedHeaderTable) const");
  LiteralList::Iterator iter(LitList());
  ulong sel = LitList()->numOfSelectedLits();
  while (sel) 
    {
      ASSERT(iter.notEnd());
      TERM header = *(iter.currentLiteral());
      header.Inversepolarity();
      if (definedHeaderTable->isDefined(header.HeaderNum())) return true; 
      iter.next();
      sel--;
    };
  return false;
}; // bool Clause::DefinitionUnfoldingSelected(const DefinedHeaderTable* definedHeaderTable) const 

inline void Clause::InsertAfter(Clause* c,Clause* db)
{
  Clause* next_cl = db->next;
  c->next = next_cl;
  next_cl->previous = c;
  db->next = c;
  c->previous = db;
}  // Clause::InsertAfter()


inline void Clause::InsertBefore(Clause* c,Clause* db)
{
  Clause* previous_cl = db->previous;
  previous_cl->next = c;
  c->previous = previous_cl;
  c->next = db;
  db->previous = c;
}  // Clause::InsertBefore()



inline void Clause::Remove(Clause* c)
{
  Clause* previous_cl = c->previous;
  Clause* next_cl = c->next;

  previous_cl->next = next_cl;
  next_cl->previous = previous_cl;
};  // Clause::Remove()

inline Clause* Clause::MakeEmptyClause()   
{
  return new Clause(new (0UL) LiteralList(0UL,0UL));
};

inline bool Clause::IsRule() const
{
  CALL("IsRule() const");
  // Returns true if and only if 
  // the clause contains both positive
  // and negative non-splitting non-answer literals.

  ASSERT(!Empty());
  if (Unit()) return false;
  LiteralList::Iterator iter(LitList());
  while (iter.notEnd() && 
	 (iter.currentLiteral()->IsSplitting()) ||
	 iter.currentLiteral()->isAnswerLiteral())
    {
      iter.next();
    };
  if (iter.end()) return false; // only splitting or answer literals 
 
  TERM::Polarity firstLitPol = iter.currentLiteral()->polarity();
  iter.next();
  while (iter.notEnd())
    {
      if ((!iter.currentLiteral()->IsSplitting()) 
	  && (!iter.currentLiteral()->isAnswerLiteral()) 
	  && (iter.currentLiteral()->polarity() != firstLitPol)) 
	return true; 
      iter.next();   
    };
  return false; // same polarity everywhere 
}; // bool Clause::IsRule() const



inline bool Clause::singletonVar(ulong var) const
{
  CALL("singletonVar(ulong var) const");
  TERM::Iterator litIter;
  ulong count = 0UL;
  for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
    {
      litIter.Reset(iter.currentLiteral());
      if (litIter.NextVar())
	{
	  do
	    {
	      if (litIter.symbolRef().var() == var)
		{
		  ++count;
		  if (count > 1UL) return false;
		};
	    }
	  while (litIter.NextVar());
	};
    };
  return true;
}; // bool Clause::singletonVar(ulong var) const


#ifndef NO_DEBUG 

inline
bool Clause::checkObj() const
{
#ifndef DEBUG_ALLOC
  return true;
#else 
  BK::ObjDesc* objDesc = BK::GlobAlloc::hostObj(static_cast<const void*>(this));
  if (!objDesc) 
    { 
      DMSG("Clause::checkObj()> Null pointer to object descriptor.\n");  
      return false; 
    }; 
  bool res = true;   
#ifdef DEBUG_ALLOC_OBJ_SIZE
  if (objDesc->size !=  sizeof(Clause)) 
    {
      DMSG("Clause::checkObj()> objDesc->size !=  sizeof(Clause)\n");
      res = false;
    };
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
  if (objDesc->status != BK::ObjDesc::Persistent) 
    {
      DMSG("Clause::checkObj()> objDesc->status != ObjDesc::Persistent\n");
      DMSG("Clause::checkObj()> clause # "); DMSG(number); DMSG('\n');  
      res = false;
    };
#endif
#ifdef DEBUG_ALLOC_OBJ_TYPE
  if (objDesc->classDesc != &Context::getCurrent()->_classDesc) 
    {
      DMSG("Clause::checkObj()> objDesc->classDesc != &Context::getCurrent()->_classDesc\n");
      res = false;
    };
#endif
  if (!res) 
    {
      DMSG("Clause::checkObj()> host object ");
      DMSG(*objDesc); 
      DMSG("\n");
    }; 
  return res;
#endif
}; // bool Clause::checkObj() const
#endif  


}; // namespace VK



//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause::Context"
#endif 
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
  inline
  long Clause::Context::computeWeight(long weightBase,
				      bool subgoal,
				      bool posEqSelected) const
  {
    CALL("computeWeight(..)");
    long wt = weightBase; 
    if (_useSelectedPosEqPenaltyCoefficient && posEqSelected)
      {
	wt = (long)(wt * _selectedPosEqPenaltyCoefficient);
      };

    if (_useNongoalPenaltyCoefficient && !subgoal)
      {
	wt = (long)(wt * _nongoalPenaltyCoefficient);	  
      };

    if (wt > (long)VampireKernelConst::MaxClauseSize) 
      {
	wt = VampireKernelConst::MaxClauseSize;
      }
    else 
      if (wt < 1)
	{
	  wt = 1;
	};
      
    return wt;
  }; // long Clause::Context::computeWeight(..)

}; // namespace VK


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause::Traversal"
#endif 
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
  
  inline
  Clause::Traversal::Traversal(const Clause* c) : termTrav(), iter()
  {
    CALL("constructor Traversal(const Clause* c)");
    ASSERT(!c->isBuiltInTheoryFact());
    reset(c);  
  };   

  inline
  void Clause::Traversal::reset(const Clause* c)
  {
    CALL("reset(const Clause* c)");
    if (c->Empty()) { currState = EndOfClause; }
    else 
      {
	currState = LitHeader;
                 
	nextLit = c->LitList()->firstLit(); 
 
	termTrav.reset(nextLit);

	iter.reset(c->LitList());   
	iter.next();
      };
  };  

  inline
  void Clause::Traversal::next()
  {
    CALL("next()");
    termTrav.next();   
    if (termTrav.state() == TERM::Traversal::End) // end of literal
      if (iter.end()) { currState = EndOfClause; }
      else 
	{
	  nextLit = iter.currentLiteral();
	  currState = LitHeader;
	  termTrav.reset(nextLit);
	  iter.next();
	}
    else currState = (termTrav.state() == TERM::Traversal::Func) ? Func : Var;
  };

  
  inline
  void Clause::Traversal::after()
  {
    CALL("after()");
    termTrav.after();
    if (termTrav.state() == TERM::Traversal::End) // end of literal
      if (iter.end()) { currState = EndOfClause; }
      else 
	{
	  nextLit = iter.currentElement()->literal();
	  currState = LitHeader;
	  termTrav.reset(nextLit);
	  iter.next();
	}
    else currState = (termTrav.state() == TERM::Traversal::Func) ? Func : Var;   
  };

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause::OpenUnshared"
#endif 
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
  
  
  inline
  Clause::OpenUnshared::OpenUnshared() : 
    _literals(DOP("Clause::OpenUnshared::_literals")),
    _literalProperties(DOP("Clause::OpenUnshared::_literalProperties"))
    {
      CALL("constructor OpenUnshared()");
      reset(); 
    };
  
  inline
  Clause::OpenUnshared::~OpenUnshared() 
  { 
    CALL("destructor ~OpenUnshared()");
    if (BK::DestructionMode::isThorough())
      {
	recycle();
      };
  };


  
  inline
  void Clause::OpenUnshared::reset()
    {
      _length = 0UL;
      _size = 0UL;
      _numOfPosEq = 0UL;
      _currLit.reset();
      _literals.reset();
      _literalProperties.reset();
      _background.Reset();
    };
   
  inline
  bool Clause::OpenUnshared::litHeader(const TERM& hd,const LiteralProperties& prop)
    {
      CALL("litHeader(const TERM& hd,const LiteralProperties& prop)");
      if (_length >= VampireKernelConst::MaxlNumOfLits) 
	{ 
	  recycle(); 
	  return false; 
	};   
      if (_length) 
	{
	  _literals.pushSafe(_currLit.result().First());  
	  _literalProperties.pushSafe(_currLitProperties);
	};
      _length++;
      _size += (1 + hd.arity());
      if (hd.Positive() && hd.IsEquality()) _numOfPosEq++;
      _currLit.reset();
      _currLitProperties = prop;
      if (!_currLit.litHeader(hd)) 
	{ 
	  recycle(); 
	  return false; 
	}; 
      return true;
    }; // bool Clause::OpenUnshared::litHeader(const TERM& hd,const LiteralProperties& prop)
  

  
  inline
  bool Clause::OpenUnshared::functor(const TERM& f)
    {
      CALL("functor(const TERM& f)");
      _size += f.arity();
      if (!_currLit.functor(f)) 
	{ 
	  recycle(); 
	  return false; 
	}; 
      return true;
    };

  
  inline
  void Clause::OpenUnshared::ancestor(Clause* cl) 
    { 
      _background.pushAncestor(cl,cl->inferenceDepth()); 
      if (cl->isSubgoal()) _background.markAsSubgoal();
    };


  
  inline
  Clause* Clause::OpenUnshared::result()  
    {
      CALL("result()");
      LiteralList* lits = new(_length) LiteralList(_length,_length);
      if (_length) // nonempty clause
	{
	  _literals.pushSafe(_currLit.result().First());  
	  _literalProperties.pushSafe(_currLitProperties);
	  LiteralList::Iterator iter(lits);
	  while (_literals)
	    {
	      ASSERT(_literalProperties);
	      ASSERT(iter.notEnd()); 
	      iter.currentElement()->setLiteral(_literals.pop());
	      iter.currentElement()->properties() = 
		_literalProperties.pop();
	      iter.next();
	    };
	  ASSERT(!_literalProperties);
	};
      Clause* res = new Clause(lits);
      _background.writeNewBackgroundTo(res->background());
      res->setSize(_size);
      res->setWeightBase(0);
      res->SetNumOfPosEq(_numOfPosEq);
      res->setPracticallyMinResidualSize(0);   
      return res;
    };

  
  inline
  void Clause::OpenUnshared::recycle() 
    {
      CALL("recycle()");
      while (_literals) _literals.pop()->DestroyLit();
      _currLit.recycle();    
    };
}; // namespace VK


//======================================================================
#endif
