//
// File:         CustomLiteralSelection.cpp
// Description:  Flexible schemes for literal selection.
// Created:      Jun 03, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//============================================================================
#include "CustomLiteralSelection.hpp"
#include "LiteralProperties.inl"
using namespace BK;
using namespace VK;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::~CustomLiteralSelection()
{
};

void 
CustomLiteralSelection::select(LitList& lits,
			       LitList& selectedLits,
			       LitList& unselectedLits,
			       bool& allMaximalSelected,
			       const LitListStatistics& stat)
{
  CALL("select(..)");    
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));    
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  DOP(ulong debugTotalNumOfLits = 
      lits.length() + 
      selectedLits.length() +
      unselectedLits.length());

  ulong numOfLits = lits.length();
  ASSERT(numOfLits);
  if (numOfLits == 1UL)
    {
      lits.popTo(selectedLits);
      allMaximalSelected = true;
      return;
    };
  allMaximalSelected = false;
  _select(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
  ASSERT(lits.isEmpty());
  ASSERT(debugTotalNumOfLits == selectedLits.length() + unselectedLits.length());
}; // void CustomLiteralSelection::select(..)


long CustomLiteralSelection::literalSize(TemporaryLiteral& lit)
{
  CALL("literalSize(TemporaryLiteral& lit)");
  return static_cast<long>(lit.size());
};

long CustomLiteralSelection::literalPriority(TemporaryLiteral& lit)
{
  CALL("literalPriority(TemporaryLiteral& lit)");
  return lit.properties().selectabilityPriority();
};



long CustomLiteralSelection::numOfDifferentVariables(TemporaryLiteral& lit)
{
  CALL("numOfDifferentVariables(TemporaryLiteral& lit)");
  return static_cast<long>(lit.numOfDifferentVariables());
};

bool CustomLiteralSelection::isGround(TemporaryLiteral& lit)
{
  CALL("isGround(TemporaryLiteral& lit)");
  return lit.isGround();
};

bool CustomLiteralSelection::isNegativeEquality(TemporaryLiteral& lit)
{
  CALL("isNegativeEquality(TemporaryLiteral& lit)");
  return lit.isNegative() && lit.isEquality();
};


bool CustomLiteralSelection::isNotPositiveEquality(TemporaryLiteral& lit)
{
  CALL("isNotPositiveEquality(TemporaryLiteral& lit)");
  return (!lit.isEquality()) || lit.isNegative();
};

bool CustomLiteralSelection::isNotAvailableForParamodFrom(TemporaryLiteral& lit)
{
  CALL("isNotAvailableForParamodFrom(TemporaryLiteral& lit)");
  return 
    !lit.isEquality() || 
    lit.isNegative() ||
    !lit.properties().useEqForParamodulationFrom();
};




bool CustomLiteralSelection::oneLiteralSelected(LitListElement* firstSelected,
						LitListElement* firstUnselected)
{
  CALL("oneLiteralSelected(..)");
  ASSERT(firstSelected);
  return !firstSelected->next(); 
};

bool 
CustomLiteralSelection::noNegativeWithHighPriorityUnselected(LitListElement* firstSelected,
						LitListElement* firstUnselected)
{
  CALL("noNegativeWithHighPriorityUnselected(..)");
  long maxPriorityOfSelected = 
    firstSelected->literal()->properties().selectabilityPriority();
  LitListElement* el = firstSelected->next();
  while (el)
    {
      long litPriority = el->literal()->properties().selectabilityPriority();
      if (litPriority > maxPriorityOfSelected) 
	maxPriorityOfSelected = litPriority;
      el = el->next();
    };

  el = firstUnselected;
  
  while (el)
    {
      if (el->literal()->isNegativeForSelection() &&
	  (el->literal()->properties().selectabilityPriority() > 
	   maxPriorityOfSelected))
	return false;
      el = el->next();
    };

  return true;
}; // bool CustomLiteralSelection::noNegativeWithHighPriorityUnselected(..)





bool CustomLiteralSelection::noBigNegativeUnselected(LitListElement* firstSelected,
						      LitListElement* firstUnselected)
{
  CALL("noBigNegativeUnselected(..)");
  ulong maxSizeOfSelected = firstSelected->literal()->size();
  LitListElement* el = firstSelected->next();
  while (el)
    {
      ulong litSize = el->literal()->size();
      if (litSize > maxSizeOfSelected) maxSizeOfSelected = litSize;
      el = el->next();
    };

  el = firstUnselected;
  
  while (el)
    {
      if (el->literal()->isNegativeForSelection() &&
	  (el->literal()->size() > maxSizeOfSelected))
	return false;
      el = el->next();
    };

  return true;
}; // bool CustomLiteralSelection::noBigNegativeUnselected(..)






bool CustomLiteralSelection::noGroundNegativeUnselected(LitListElement* firstSelected,
							LitListElement* firstUnselected)
{
  CALL("noGroundNegativeUnselected(..)");
  while (firstUnselected)
    {
      if (firstUnselected->literal()->isNegativeForSelection() &&
	  firstUnselected->literal()->isGround())
	return false;
      firstUnselected = firstUnselected->next();
    };
  return true;
}; // bool CustomLiteralSelection::noGroundNegativeUnselected(..)


bool CustomLiteralSelection::noNegativeEqualitiesUnselected(LitListElement* firstSelected,
							     LitListElement* firstUnselected)
{
  CALL("noNegativeEqualitiesUnselected(..)");
  while (firstUnselected)
    {
      if (firstUnselected->literal()->isEquality() &&
	  firstUnselected->literal()->isNegativeForSelection())
	return false;
      firstUnselected = firstUnselected->next();
    };
  return true;
}; // bool CustomLiteralSelection::noNegativeEqualitiesUnselected(..)

bool CustomLiteralSelection::noNegativeWithManyDiffVarsUnselected(LitListElement* firstSelected,
								   LitListElement* firstUnselected)
{
  CALL("noNegativeWithManyDiffVarsUnselected(..)");
  ulong maxNumOfDiffVars = firstSelected->literal()->numOfDifferentVariables();
  LitListElement* el = firstSelected->next();
  while (el)
    {
      ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
      if (numOfDiffVars > maxNumOfDiffVars) maxNumOfDiffVars = numOfDiffVars;
      el = el->next();
    };

  el = firstUnselected; 
  
  while (el)
    {
      if (el->literal()->isNegativeForSelection() &&
	  (el->literal()->numOfDifferentVariables() > maxNumOfDiffVars))
	return false;
      el = el->next();
    };

  return true;
}; // bool CustomLiteralSelection::noNegativeWithManyDiffVarsUnselected(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ComplexLiteralCharacteristic"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::ComplexLiteralCharacteristic::~ComplexLiteralCharacteristic()
{  
}; 


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::InverseVarDepthProfile"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::InverseVarDepthProfile::~InverseVarDepthProfile()
{  
}; 

void CustomLiteralSelection::InverseVarDepthProfile::computeFor(TemporaryLiteral& lit)
{
  CALL("computeFor(TemporaryLiteral& lit)");
  lit.collectVarDepthProfile(_data);
}; // void CustomLiteralSelection::InverseVarDepthProfile::computeFor(TemporaryLiteral& lit)

Comparison CustomLiteralSelection::InverseVarDepthProfile::compareWith(ComplexLiteralCharacteristic* c)
{
  CALL("compareWith(ComplexLiteralCharacteristic* c)");
  return 
    _data.compareByInverselyInducedOrder((static_cast<InverseVarDepthProfile*>(c))->_data);
}; // Comparison CustomLiteralSelection::InverseVarDepthProfile::compareWith(ComplexLiteralCharacteristic* c)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::Filter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::Filter::~Filter()
{  
}; 

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::MaxValFilter::~MaxValFilter()
{
};


void 
CustomLiteralSelection::MaxValFilter::selectOne(LitList& lits,
						LitList& selectedLits,
						LitList& unselectedLits)
{
  CALL("selectOne(..)");  
  LitHandle cand;
  
  lits.popTo(cand);
  long candVal = _val(*cand);
  long nextVal;
  while (!lits.isEmpty())
    {
      nextVal = _val(*(lits.last()->literal()));
      
      if (nextVal > candVal)
	{
	  unselectedLits.collect(cand);
	  lits.popTo(cand);
	  candVal = nextVal;
	}
      else
	lits.popTo(unselectedLits);
    };
  selectedLits.collect(cand);
}; // void CustomLiteralSelection::MaxValFilter::selectOne(..)

void 
CustomLiteralSelection::MaxValFilter::selectAll(LitList& lits,
						LitList& selectedLits,
						LitList& unselectedLits)
{
  CALL("selectAll(..)");
  long candVal = _val(*(lits.last()->literal()));
  long nextVal;
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      nextVal = _val(*(el->literal()));
      if (nextVal != candVal)
	{
	  if (nextVal > candVal)
	    {
	      candVal = nextVal;
	      do 
		{
		  lits.popTo(unselectedLits);
		}
	      while (lits.last() != el);
	      el = el->previous();
	    }
	  else // nextVal < candVal
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    };
	}
      else
	el = el->previous();
    };

  // all remaining literals have maximal _val
  ASSERT(!lits.isEmpty());
  selectedLits.collectAllFrom(lits);
}; // void CustomLiteralSelection::MaxValFilter::selectAll(..)

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MinValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::MinValFilter::~MinValFilter()
{
};


void 
CustomLiteralSelection::MinValFilter::selectOne(LitList& lits,
						LitList& selectedLits,
						LitList& unselectedLits)
{
  CALL("selectOne(..)");  
  LitHandle cand;
  
  lits.popTo(cand);
  long candVal = _val(*cand);
  long nextVal;
  while (!lits.isEmpty())
    {
      nextVal = _val(*(lits.last()->literal()));
      
      if (nextVal < candVal)
	{
	  unselectedLits.collect(cand);
	  lits.popTo(cand);
	  candVal = nextVal;
	}
      else
	lits.popTo(unselectedLits);
    };
  selectedLits.collect(cand);
}; // void CustomLiteralSelection::MinValFilter::selectOne(..)

void 
CustomLiteralSelection::MinValFilter::selectAll(LitList& lits,
						LitList& selectedLits,
						LitList& unselectedLits)
{
  CALL("selectAll(..)");
  long candVal = _val(*(lits.last()->literal()));
  long nextVal;
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      nextVal = _val(*(el->literal()));
      if (nextVal != candVal)
	{
	  if (nextVal < candVal)
	    {
	      candVal = nextVal;
	      do 
		{
		  lits.popTo(unselectedLits);
		}
	      while (lits.last() != el);
	      el = el->previous();
	    }
	  else // nextVal < candVal
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    };
	}
      else
	el = el->previous();
    };

  // all remaining literals have maximal _val
  ASSERT(!lits.isEmpty());
  selectedLits.collectAllFrom(lits);
}; // void CustomLiteralSelection::MinValFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxComplexValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::MaxComplexValFilter::~MaxComplexValFilter()
{
};


void 
CustomLiteralSelection::MaxComplexValFilter::selectOne(LitList& lits,
						       LitList& selectedLits,
						       LitList& unselectedLits)
{
  CALL("selectOne(..)");  
  ASSERT(!lits.isEmpty());
  ComplexLiteralCharacteristic* candVal = _tmpVal1;
  ComplexLiteralCharacteristic* bufferVal = _tmpVal2;
 next_candidate:
  ASSERT(!lits.isEmpty());
  candVal->computeFor(*(lits.last()->literal()));
  LitListElement* tmp;
  for (LitListElement* el = lits.last()->previous(); 
       el; 
       /* nothing here */)
    {
      bufferVal->computeFor(*(el->literal()));
      switch (candVal->compareWith(bufferVal))
	{
	case Less:
	  lits.popTo(unselectedLits);
	  goto next_candidate;

	case Equal:
	  // as below
	case Greater:
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	  break;
	  
	case Incomparable:
	  el = el->previous();
	  break;
	}; // switch (candVal->compareWith(bufferVal))
    }; // for 
  // lits.last() is OK
  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits); // only one has to be selected
}; // void CustomLiteralSelection::MaxComplexValFilter::selectOne(..)

void 
CustomLiteralSelection::MaxComplexValFilter::selectAll(LitList& lits,
						       LitList& selectedLits,
						       LitList& unselectedLits)
{
  CALL("selectAll(..)");
  ASSERT(!lits.isEmpty());
  ComplexLiteralCharacteristic* candVal = _tmpVal1;
  ComplexLiteralCharacteristic* bufferVal = _tmpVal2;
 next_candidate:
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1UL)
    {
      lits.popTo(selectedLits);
      return;
    };
  candVal->computeFor(*(lits.last()->literal()));
  LitListElement* tmp;
  for (LitListElement* el = lits.last()->previous(); 
       el; 
       /* nothing here */)
    {
      bufferVal->computeFor(*(el->literal()));
      switch (candVal->compareWith(bufferVal))
	{
	case Less:
	  lits.popTo(unselectedLits);
	  goto next_candidate;

	case Equal:
	  el = el->previous();
	  break;

	case Greater:
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	  break;
	  
	case Incomparable:
	  el = el->previous();
	  break;
	}; // switch (candVal->compareWith(bufferVal))
    }; // for ..
  // lits.last() is OK
  lits.popTo(selectedLits);
  if (lits.isEmpty()) return;
  goto next_candidate;
}; // void CustomLiteralSelection::MaxComplexValFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPOrAllFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::AllPOrAllFilter::~AllPOrAllFilter()
{
};


void 
CustomLiteralSelection::AllPOrAllFilter::selectOne(LitList& lits,
						   LitList& selectedLits,
						   LitList& unselectedLits)
{
  CALL("selectOne(..)"); 
  LitListElement* el = lits.last();
  while (el && (!(_prop(*(el->literal())))))
    {
      el = el->previous();      
    };
  
  if (el)
    {
      lits.passTo(el,selectedLits);
    }
  else // select any one
    {
      lits.popTo(selectedLits);
    };
  lits.passAllTo(unselectedLits);
}; // void CustomLiteralSelection::AllPOrAllFilter::selectOne(..)

void 
CustomLiteralSelection::AllPOrAllFilter::selectAll(LitList& lits,
						   LitList& selectedLits,
						   LitList& unselectedLits)
{
  CALL("selectAll(..)");
   
  LitListElement* el = lits.last();
  while (el && (!(_prop(*(el->literal())))))
    {
      el = el->previous();      
    };
  
  if (el) 
    {
      // select all with the property
      while (lits.last() != el)
	lits.popTo(unselectedLits);
      lits.popTo(selectedLits);
      while (!lits.isEmpty())
	{
	  if (_prop(*(lits.last()->literal())))
	    {
	      lits.popTo(selectedLits);
	    }
	  else
	    lits.popTo(unselectedLits);
	};
    }
  else // select all
    lits.passAllTo(selectedLits);
}; // void CustomLiteralSelection::AllPOrAllFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPThroughF1OrAllThroughF2"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::AllPThroughF1OrAllThroughF2::~AllPThroughF1OrAllThroughF2()
{
};


void 
CustomLiteralSelection::AllPThroughF1OrAllThroughF2::selectOne(LitList& lits,
							       LitList& selectedLits,
							       LitList& unselectedLits)
{
  CALL("selectOne(..)"); 
  LitListElement* el = lits.last();
  while (el && !(_prop(*(el->literal()))))
    {
      el = el->previous();      
    };
  
  if (el)
    {
      // unselect all without the property
      while (lits.last() != el)
	lits.popTo(unselectedLits);
      el = el->previous();
      while (el)
	{
	  if (!(_prop(*(el->literal()))))
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else
	    el = el->previous();
	};
      
      // apply _filter1      
      _filter1->selectOne(lits,selectedLits,unselectedLits);
    }
  else // apply _filter2
    {
      _filter2->selectOne(lits,selectedLits,unselectedLits);
    };
}; // void CustomLiteralSelection::AllPThroughF1OrAllThroughF2::selectOne(..)

void 
CustomLiteralSelection::AllPThroughF1OrAllThroughF2::selectAll(LitList& lits,
							       LitList& selectedLits,
							       LitList& unselectedLits)
{
  CALL("selectAll(..)");
  LitListElement* el = lits.last();
  while (el && !(_prop(*(el->literal()))))
    {
      el = el->previous();      
    };
  
  if (el)
    {
      // unselect all without the property
      while (lits.last() != el)
	lits.popTo(unselectedLits);
      el = el->previous();
      while (el)
	{
	  if (!(_prop(*(el->literal()))))
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else
	    el = el->previous();
	};
      
      // apply _filter1      
      _filter1->selectAll(lits,selectedLits,unselectedLits);
    }
  else // apply _filter2
    {
      _filter2->selectAll(lits,selectedLits,unselectedLits);
    };
}; // void CustomLiteralSelection::AllPThroughF1OrAllThroughF2::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::CompositionFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::CompositionFilter::~CompositionFilter()
{
};


void 
CustomLiteralSelection::CompositionFilter::selectOne(LitList& lits,
						     LitList& selectedLits,
						     LitList& unselectedLits)
{
  CALL("selectOne(..)");  
  if (_filters.empty()) 
    {
      lits.popTo(selectedLits);
      lits.passAllTo(unselectedLits);
      return;
    };

  LitListElement* lastSelected = selectedLits.last();
  
  Filter** lastFilter = _filters.last();
  Filter** filt = _filters.begin();
  while (filt != lastFilter)
    {
      (*filt)->select(lits,selectedLits,unselectedLits);
      while (selectedLits.last() != lastSelected)
	selectedLits.popTo(lits);
      ++filt;
    };
  
  // apply the last filter
  (*lastFilter)->selectOne(lits,selectedLits,unselectedLits);
}; // void CustomLiteralSelection::CompositionFilter::selectOne(..)

void 
CustomLiteralSelection::CompositionFilter::selectAll(LitList& lits,
						     LitList& selectedLits,
						     LitList& unselectedLits)
{
  CALL("selectAll(..)");
  if (_filters.empty()) 
    { 
      lits.passAllTo(selectedLits);
      return; 
    };
  
  LitListElement* lastSelected = selectedLits.last();
  
  Filter** lastFilter = _filters.last();
  Filter** filt = _filters.begin();
  while (filt != lastFilter)
    {
      (*filt)->select(lits,selectedLits,unselectedLits);
      while (selectedLits.last() != lastSelected)
	selectedLits.popTo(lits);
      ++filt;
    };
  
  // apply the last filter
  (*lastFilter)->select(lits,selectedLits,unselectedLits);
}; // void CustomLiteralSelection::CompositionFilter::selectAll(..)

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::OneOfEachPolarityFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::OneOfEachPolarityFilter::~OneOfEachPolarityFilter()
{
};


void 
CustomLiteralSelection::OneOfEachPolarityFilter::selectOne(LitList& lits,
						       LitList& selectedLits,
						       LitList& unselectedLits)
{
  CALL("selectOne(..)");
  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits);
}; // void CustomLiteralSelection::OneOfEachPolarityFilter::selectOne(..)

void 
CustomLiteralSelection::OneOfEachPolarityFilter::selectAll(LitList& lits,
							   LitList& selectedLits,
							   LitList& unselectedLits)
{
  CALL("selectAll(..)");
  bool posSelected = false;
  bool negSelected = false;
  while (!lits.isEmpty())
    {
      if (lits.last()->literal()->isPositiveForSelection())
	{
	  if (posSelected)
	    {
	      lits.popTo(unselectedLits);
	    }
	  else
	    {
	      posSelected = true;
	      lits.popTo(selectedLits);
	    };
	}
      else // negative literal
	if (negSelected)
	  {
	    lits.popTo(unselectedLits);
	  }
	else
	  {
	    negSelected = true;
	    lits.popTo(selectedLits);
	  };
    };
}; // void CustomLiteralSelection::OneOfEachPolarityFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::SelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::SelectionAcceptability::~SelectionAcceptability()
{  
}; 


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================


CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::~ConjOfPrimitiveSelectionAcceptability()
{
}; 


bool CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::accepts(LitListElement* firstSelected,
									    LitListElement* firstUnselected)
{
  CALL("accepts(..)");

  Primitive* endOfPrim = _primitives.end();
  for (Primitive* prim = _primitives.begin(); 
       prim != endOfPrim;
       ++prim)
    {
      if (!((*prim)(firstSelected,firstUnselected))) return false;
    };
  return true;
}; //  bool CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::accepts()




//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "TotalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


void 
TotalLiteralSelection::_select(LitList& lits,
			       LitList& selectedLits,
			       LitList& unselectedLits,
			       bool& allMaximalSelected,
			       const LitListStatistics& stat)
{
  CALL("_select(..)");
  lits.passAllTo(selectedLits);
  allMaximalSelected = true;
}; // void TotalLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaximalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

void 
MaximalLiteralSelection::_select(LitList& lits,
				 LitList& selectedLits,
				 LitList& unselectedLits,
				 bool& allMaximalSelected,
				 const LitListStatistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = true;
  selectMaximal(lits,selectedLits,unselectedLits);
}; // void MaximalLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "HyperLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
HyperLiteralSelection::_select(LitList& lits,
			       LitList& selectedLits,
			       LitList& unselectedLits,
			       bool& allMaximalSelected,
			       const LitListStatistics& stat)
{
  CALL("_select(..)");
  if (stat.numOfNegNonanswerLits())
    {
      allMaximalSelected = false;
      if (stat.numOfPosNonanswerLits())
	{
	  // mixed polarity, unselect all positive
	  unselectPositive(lits,unselectedLits,stat.numOfPosNonanswerLits());
	};
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL)
	{
	  lits.popTo(selectedLits);
	}
      else
	_filterForNegative->select(lits,selectedLits,unselectedLits);      
    }
  else // only positive literals
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,unselectedLits);
    };
}; // void HyperLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaxWithCorrectionLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
MaxWithCorrectionLiteralSelection::_select(LitList& lits,
					   LitList& selectedLits,
					   LitList& unselectedLits,
					   bool& allMaximalSelected,
					   const LitListStatistics& stat)
{
  CALL("_select(..)");  
  if (stat.numOfNegNonanswerLits())
    { 
      LitListElement* lastSelected = selectedLits.last();
      LitListElement* lastUnselected = unselectedLits.last();
      selectMaximal(lits,selectedLits,unselectedLits);
      if (!_maxSelectionAcceptability->accepts(((lastSelected) ? lastSelected->next() : selectedLits.first()),
					       ((lastUnselected) ? lastUnselected->next() : unselectedLits.first())))
	{      
	  allMaximalSelected = false;
	  // undo selection for negative literals
	  LitListElement* el = unselectedLits.last();
	  while (el != lastUnselected)
	    {
	      if (el->literal()->isNegativeForSelection())
		{
		  LitListElement* tmp = el->previous();
		  unselectedLits.passTo(el,lits);
		  el = tmp;
		}
	      else
		el = el->previous();
	    };

	  while (selectedLits.last() != lastSelected)
	    {
	      if (selectedLits.last()->literal()->isNegativeForSelection())
		{
		  selectedLits.popTo(lits);
		}
	      else
		selectedLits.popTo(unselectedLits);
	    };

	  ASSERT(!lits.isEmpty());
	  
	  // select appropriate negative 	
	  
	  _filterForNegative->select(lits,selectedLits,unselectedLits);
	}
      else // maximal selection accepted
	{
	  allMaximalSelected = true;
	};
     
    }
  else // only positive literals
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,unselectedLits);
    };
}; // void MaxWithCorrectionLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "BestOnlyLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
BestOnlyLiteralSelection::_select(LitList& lits,
				  LitList& selectedLits,
				  LitList& unselectedLits,
				  bool& allMaximalSelected,
				  const LitListStatistics& stat)
{
  CALL("_select(..)");
  LitListElement* cand = 0;
  if (stat.numOfNegNonanswerLits())
    {
      if (stat.numOfPosNonanswerLits())
	{
	  LitListElement* lastSelected = selectedLits.last();
	  LitListElement* lastUnselected = unselectedLits.last();
	  selectMaximal(lits,selectedLits,unselectedLits);

	  // now try to beat the maximal selection
	  
	  // light-weight test first: is a negative literal selected?
	  LitListElement* el;
	  for (el = (lastSelected) ? lastSelected->next() : selectedLits.first(); 
	       el; 
	       el = el->next())
	    if (el->literal()->isNegativeForSelection()) 
	      goto undo_maximal_selection;
	  
	  // all selected are positive
	  
	  for (cand = (lastUnselected) ? lastUnselected->next() : unselectedLits.first();
	       cand;
	       cand = cand->next())	    
	    {
	      if (cand->literal()->isNegativeForSelection())		
		for (el = (lastSelected) ? lastSelected->next() : selectedLits.first();
		     el; 
		     el = el->next())
		  {
		    Comparison cmp = 
		      _litOrdering->compare(*(el->literal()),*(cand->literal()));
		    if ((cmp == BK::Less) || (cmp == BK::Equal))
		      goto undo_maximal_selection;
		  };
	    };

	  // maximal selection is OK
	  allMaximalSelected = true;
	  return;

	undo_maximal_selection:
	  el = (lastUnselected) ? lastUnselected->next() : unselectedLits.first();
	  while (el)
	    {
	      if (el->literal()->isNegativeForSelection())
		{
		  LitListElement* tmp = el->next();
		  unselectedLits.passTo(el,lits);
		  el = tmp;
		}
	      else
		el = el->next();
	    };

	  while (selectedLits.last() != lastSelected)
	    {
	      if (selectedLits.last()->literal()->isNegativeForSelection())
		{
		  selectedLits.popTo(lits);
		}
	      else
		selectedLits.popTo(unselectedLits);
	    };

	}; // if (stat.numOfPosNonanswerLits())
      // only negative literals in lits if we get here
      allMaximalSelected = false;
      selectOneMaximal(_litOrdering,lits,selectedLits,unselectedLits);
    }
  else // positive fact
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,unselectedLits);
    };
  
}; // void BestOnlyLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "ArbitraryLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
ArbitraryLiteralSelection::_select(LitList& lits,
				   LitList& selectedLits,
				   LitList& unselectedLits,
				   bool& allMaximalSelected,
				   const LitListStatistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = false;
  _filter->select(lits,selectedLits,unselectedLits);
}; // void ArbitraryLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
CompleteBidirectionalLiteralSelection::_select(LitList& lits,
					       LitList& selectedLits,
					       LitList& unselectedLits,
					       bool& allMaximalSelected,
					       const LitListStatistics& stat)
{
  CALL("_select(..)");

  if (stat.numOfNegNonanswerLits())
    {
      allMaximalSelected = false;
      if (stat.numOfPosNonanswerLits())
	{
	  // rule (mixed polarity lit. list)
	  if (lits.length() == 2UL)
	    {
	      lits.passAllTo(selectedLits);
	      return;
	    };

	  // collect all positive literals in the buffer
	  ASSERT(_buffer.isEmpty());
	  ASSERT(lits.length() > 2UL);
	  LitListElement* el = lits.first();
	check_lit:
	  
	  if (el->literal()->isPositiveForSelection())
	    {
	      LitListElement* tmp = el->next();
	      lits.passTo(el,_buffer);
	      if (tmp)
		{
		  el = tmp;
		  goto check_lit;
		};
	    }
	  else
	    {
	      el = el->next();
	      if (el) goto check_lit;
	    };
	  
	  // all positive ones are in _buffer now
	  _filterForNegativeInRules->select(lits,selectedLits,unselectedLits);
	  _filterForPositiveInRules->select(_buffer,selectedLits,unselectedLits);
	}
      else // negative fact
	{
	  _filterForNegativeFacts->select(lits,selectedLits,unselectedLits);
	};
    }
  else // positive fact
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,unselectedLits);
    };  
}; // void CompleteBidirectionalLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "IncompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
IncompleteBidirectionalLiteralSelection::_select(LitList& lits,
						 LitList& selectedLits,
						 LitList& unselectedLits,
						 bool& allMaximalSelected,
						 const LitListStatistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = false;
  
  if (stat.numOfNegNonanswerLits())
    {
      if (stat.numOfPosNonanswerLits())
	{
	  // rule (mixed polarity lit. list)
	  if (lits.length() == 2UL)
	    {
	      lits.passAllTo(selectedLits);
	      return;
	    };
	  
	  // collect all positive literals in the buffer
	  ASSERT(_buffer.isEmpty());
	  ASSERT(lits.length() > 2UL);
	  LitListElement* el = lits.first();
	check_lit:
	  
	  if (el->literal()->isPositiveForSelection())
	    {
	      LitListElement* tmp = el->next();
	      lits.passTo(el,_buffer);
	      if (tmp)
		{
		  el = tmp;
		  goto check_lit;
		};
	    }
	  else
	    {
	      el = el->next();
	      if (el) goto check_lit;
	    };
	  
	  // all positive ones are in _buffer now
	  _filterForNegativeInRules->select(lits,selectedLits,unselectedLits);
	  _filterForPositiveInRules->select(_buffer,selectedLits,unselectedLits);
	}
      else // negative fact
	{
	  _filterForNegativeFacts->select(lits,selectedLits,unselectedLits);
	};
    }
  else // positive fact
    {
      _filterForPositiveFacts->select(lits,selectedLits,unselectedLits);
    };  
}; // void IncompleteBidirectionalLiteralSelection::_select(..)



//============================================================================


