//
// File:         ForwardProcessing.cpp
// Description:  Retention tests and simplifications on new clauses.
// Created:      Nov 11, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       stable
// Revised:      Jun 15, 2004, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in rewriteTopOfComplexSubtermByDemodulation(..).
//               markInputNonBasicPositions() was not called on a literal
//               after rewriting.
//=================================================================
#include "ForwardProcessing.hpp"
#include "ForwardSubsumptionFrontEnd.hpp"
#include "BuiltInTheoryDispatcher.hpp"
#include "NegativeEqualitySplitting.hpp"
#include "PropositionalSplitting.hpp"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_PROCESSING
#  define DEBUG_NAMESPACE "ForwardProcessing"
#endif
#include "debugMacros.hpp"
//===================================================================


ForwardProcessing::ForwardProcessing() :
  OPEN_BR(1,"br 1") DCOMMA
  _preprocessingInProgress(false),
  _equalityPossible(true),
  _forwardSubsumption(0),
  _forwardSubsumptionResolution(false),
  _minLengthOfSubsumptionResolutionTarget(0UL),
  _forwardDemodulation(0),
  _onBranchForwardDemodulation(false),
  _destructiveEqResolution(false),
  _propositionalSplitting(0),
  _negativeEqualitySplitting(0),
  _commutativityTableForNormalisation(0),
  _commutativityTableForEqTaut(0),
  _commutativityTableForPropTaut(0),
  _commutativityTableForTrivialEqRes(0),
  _commutativityTableForIdenticalLitFactoring(0),
  _weedModuloC(false),
  _prohibitNonBasicDemodulation(false),
  _reduceNonBasic(false),
  _useInputNonBasicness(false),
  _useRunTimeNonBasicness(false),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _clauseLengthLimit(VampireKernelConst::MaxlNumOfLits),
  _literalSizeLimit(VampireKernelConst::MaxTermSize),
  _inferenceDepthLimit(ULONG_MAX),
  _statistics(0),
  _internalClauseQueue(DOP("ForwardProcessing::_internalClauseQueue")),
  _internalClauseCollector(DOP("ForwardProcessing::_internalClauseCollector")),
  _splittingBranches(DOP("ForwardProcessing::_splittingBranches")),
  _ancestorReceiver(0)
{ 
  CALL("constructor ForwardProcessing()");
  _ancestorReceiver.connectTo(this);
  _internalClauseCollector.setOutputQueue(&_internalClauseQueue);
  _internalClauseCollector.setPreprocessingInProgress(_preprocessingInProgress);
};


ForwardProcessing::~ForwardProcessing()
{
  CALL("destructor ~ForwardProcessing()");
}; // ForwardProcessing::~ForwardProcessing()
  
void ForwardProcessing::reset()
{
  CALL("ForwardProcessing::reset()");
  _preprocessingInProgress = false;
  _equalityPossible = true;
  _forwardSubsumption = 0;
  _forwardSubsumptionResolution = false;
  _minLengthOfSubsumptionResolutionTarget = 0UL;
  _forwardDemodulation = 0;
  _onBranchForwardDemodulation = false;
  _destructiveEqResolution = false;
  _propositionalSplitting = 0;
  _negativeEqualitySplitting = 0;
  _commutativityTableForNormalisation = 0;
  _commutativityTableForEqTaut = 0;
  _commutativityTableForPropTaut = 0;
  _commutativityTableForTrivialEqRes = 0;
  _commutativityTableForIdenticalLitFactoring = 0;
  _weedModuloC = false;
  _prohibitNonBasicDemodulation = false; 
  _reduceNonBasic = false;
  _useInputNonBasicness = false;
  _useRunTimeNonBasicness = false;
  _clauseSizeLimit = VampireKernelConst::MaxClauseSize;
  _clauseLengthLimit = VampireKernelConst::MaxlNumOfLits;
  _literalSizeLimit = VampireKernelConst::MaxTermSize;
  _inferenceDepthLimit = ULONG_MAX;
  _statistics = 0;
  _currentClause.clear();
  _internalClauseQueue.reset();
  _internalClauseCollector.reset();
  _splittingBranches.reset();

  _internalClauseCollector.setOutputQueue(&_internalClauseQueue);
  _internalClauseCollector.setPreprocessingInProgress(_preprocessingInProgress);
}; // void ForwardProcessing::reset()


bool ForwardProcessing::processCurrent()
{
  CALL("processCurrent()");
  
 check_inference_depth:
  if (_currentClause->background().inferenceDepth() > 
      static_cast<long>(_inferenceDepthLimit))
    {
      registerRejected(InferenceTooDeep);
      return false;
    };
  
 weed:
  if (!weedCurrent())
    {
      return false; // tautology
    };  

 check_empty:
  if (_currentClause->literals()->isEmpty())
    {
      registerRetained();
      return true;
    };

  if (subsumeCurrent())
    {
      return false;
    };  

  if (_useInputNonBasicness) _currentClause->markInputNonBasicPositions();

  //DF; _currentClause->outputComplete(cout << "AGGGGGGAAA ") << "\n";
  
  if (reduceNonBasicInCurrentByCommutativity())
    {
      return false; // redundant, not simplified!
    };

  if (reduceNonBasicInCurrent())
    {
      return false; // redundant, not simplified!
    };


  if (simplifyCurrentByForwardSubsumptionResolution())
    goto check_empty;
  
  if (normaliseCurrentByCommutativity())
    goto weed;
  
  if (demodulateIntoCurrent())
    goto check_inference_depth;

  if (simplifyCurrentByBuiltInTheories())
    goto check_inference_depth;

  if (simplifyCurrentByDestructiveEqualityResolution())
    {
      goto weed;
    };

  if (splitNegativeEqualitiesInCurrent())
    goto weed;

  if (applyPropositionalSplittingToCurrent())
    goto weed;
  
  if (discardCurrentAsTooComplex())
    {
      return false;
    };

  registerRetained();
  return true;
}; // bool ForwardProcessing::processCurrent()



bool ForwardProcessing::processCurrentNoEquality()
{
  CALL("processCurrentNoEquality()");
 
 check_inference_depth:
  if (_currentClause->background().inferenceDepth() > 
      static_cast<long>(_inferenceDepthLimit))
    {
      registerRejected(InferenceTooDeep);
      return false;
    };
  
 weed:
  if (!weedCurrent())
    return false; // tautology
  
 check_empty:
  if (_currentClause->literals()->isEmpty())
    {
      registerRetained();
      return true;
    };

  if (subsumeCurrent())
    return false;
  
  if (simplifyCurrentByForwardSubsumptionResolution())
    goto check_empty;
  
  if (simplifyCurrentByBuiltInTheories())
    goto check_inference_depth;

  if (applyPropositionalSplittingToCurrent())
    goto weed;

  if (discardCurrentAsTooComplex())
    return false;

  registerRetained();
  return true;
}; // bool ForwardProcessing::processCurrentNoEquality()




inline bool ForwardProcessing::weedCurrent()
{
  CALL("weedCurrent()");
  ASSERT(!_currentClause.isNull());
  bool tautology;
  bool eqTautology;
  if (_weedModuloC)
    {
      if (!_currentClause->weededModuloCommutativity())
	{	  
	  TempLitListExtras::weed(*(_currentClause->literals()),
				  tautology,
				  eqTautology,
				  _commutativityTableForEqTaut,
				  _commutativityTableForPropTaut,
				  _commutativityTableForTrivialEqRes,
				  _commutativityTableForIdenticalLitFactoring,
				  static_cast<BK::Receiver<const Clause*>*>(&_ancestorReceiver));
	  _currentClause->markAsWeededModuloCommutativity();
	}
      else
	tautology = false;
    }
  else
    {
      if (!_currentClause->weeded())
	{
	  TempLitListExtras::weed(*(_currentClause->literals()),
				  tautology,
				  eqTautology);
	  _currentClause->markAsWeeded();
	}
      else
	tautology = false;
    };
  if (tautology)
    {
      if (eqTautology)
	{
	  registerRejected(EquationalTautology);
	}
      else
	{
	  registerRejected(PropositionalTautology);
	};
      ASSERT(!_currentClause->literals()->isEmpty());
      return false; // discarded
    };
  return true; // retained
}; // bool ForwardProcessing::weedCurrent()

inline bool ForwardProcessing::subsumeCurrent()
{
  CALL("subsumeCurrent()");
  ASSERT(!_currentClause.isNull());
  if (_currentClause->testedForSubsumption() ||
      _currentClause->isNonSubsumable() ||
      !_forwardSubsumption)
    return false;
  bool res = 
    (_forwardSubsumption->subsume(*(_currentClause->literals())) != 0);
  
  if (res) 
    {
      registerRejected(Subsumed);
    }
  else
    _currentClause->markAsTestedForSubsumption();   
  return res;
}; // bool ForwardProcessing::subsumeCurrent()

Clause* ForwardProcessing::canBeSubsumed(TemporaryClause& cl)
{
  CALL("canBeSubsumed(TemporaryClause& cl)");
  if (!_forwardSubsumption) return 0;
  return 
    _forwardSubsumption->subsume(*(cl.literals()));
}; // Clause* ForwardProcessing::canBeSubsumed(TemporaryClause& cl)

inline bool ForwardProcessing::simplifyCurrentByForwardSubsumptionResolution()
{
  CALL("simplifyCurrentByForwardSubsumptionResolution()");
  ASSERT(!_currentClause.isNull());
  if (_currentClause->simplifiedBySubsumptionResolution() || 
      (!_forwardSubsumption) ||
      (!_forwardSubsumptionResolution) ||
      (_currentClause->literals()->length() < _minLengthOfSubsumptionResolutionTarget))
    return false;
  ulong res = 
    _forwardSubsumption->simplifyBySubsumptionResolution(*(_currentClause->literals()),
							 _currentClause->background());
  _currentClause->markAsSimplifiedBySubsumptionResolution();
  if (res)
    {
      _currentClause->background().UsedRule(ClauseBackground::ForwSubsRes);
      _currentHasBeenSimplifiedBySubsumptionResolution = true;
      _statistics->inferences.simplifying.forwardSubsumptionResolution += res;
    };
  return res != 0UL;
}; // bool ForwardProcessing::simplifyCurrentByForwardSubsumptionResolution()

inline bool ForwardProcessing::normaliseCurrentByCommutativity()
{
  CALL("normaliseCurrentByCommutativity()");
  ASSERT(!_currentClause.isNull());
  if (_currentClause->normalisedByCommutativity() ||
      !_commutativityTableForNormalisation)
    return false;
  bool res = false;
  for (LitListElement* el = _currentClause->literals()->first();
       el;
       el = el->next())
    {
      if (el->literal()->arity())
	{	  
	  const Clause* commutativityLaw;
	  const bool* nonBasicArgs;
	  bool literalChanged = false;
	  Flatterm* literalBody = el->literal()->flatterm();
	  for (Flatterm* currentSubterm = literalBody->lastSymbol();
	       currentSubterm != literalBody;
	       currentSubterm = currentSubterm->previous())
	    {
	      if (currentSubterm->isComplex() && 
		  (!_prohibitNonBasicDemodulation || 
		   !currentSubterm->isNonBasicPosition()) &&
		  ((commutativityLaw = 
		    _commutativityTableForNormalisation->commutativityLaw(currentSubterm->functor()))) &&
		  SimplificationOrdering::current()->compareWithTransposition(currentSubterm) == Greater &&
		  (!_useInputNonBasicness ||
                   !((nonBasicArgs = currentSubterm->symbolRef().nonBasicArguments())) ||
		   ((!nonBasicArgs[0] || currentSubterm->arg2()->isNonBasicPosition()) && 
                    (!nonBasicArgs[1] || currentSubterm->arg1()->isNonBasicPosition()))))
		{		
		  currentSubterm->BinSwapArgs();
		  literalChanged = true;        
		  _currentClause->background().pushAncestor(const_cast<Clause*>(commutativityLaw),
							    commutativityLaw->inferenceDepth());      
		  ++(_statistics->inferences.simplifying.forwardDemodulation);
		};
	    }; // for (Flatterm* currentSubterm ..

	  if (literalChanged)
	    {
	      res = true;
	      if (el->literal()->isEquality())
		el->literal()->makeEqUnordered();
	    };
	};
    }; // for (LitListElement* el ..
  
  if (res)
    {
      _currentClause->background().UsedRule(ClauseBackground::ForwDemod);
      _currentClause->background().MakeVIP();
      _currentClause->registerIntraLiteralModification();
    };
  _currentClause->markAsNormalisedByCommutativity();
  return res;
}; // bool ForwardProcessing::normaliseCurrentByCommutativity()



bool ForwardProcessing::reduceNonBasicInCurrentByCommutativity()
{
  CALL("reduceNonBasicInCurrentByCommutativity()");
  ASSERT(!_currentClause.isNull());
  if (!_reduceNonBasic || 
      !_commutativityTableForNormalisation ||
      _currentClause->testedForReductionOfNonBasicByCommutativity())
    return false;
    
  for (LitListElement* el = _currentClause->literals()->first();
       el;
       el = el->next())
    {
      if (el->literal()->arity())
	{	  
	  const Clause* commutativityLaw;
	  Flatterm* literalBody = el->literal()->flatterm();
	  for (Flatterm* currentSubterm = literalBody->lastSymbol();
	       currentSubterm != literalBody;
	       currentSubterm = currentSubterm->previous())
	    {
	      if (currentSubterm->isNonBasicPosition() && 
		  currentSubterm->isComplex() &&
		  ((commutativityLaw = 
		    _commutativityTableForNormalisation->commutativityLaw(currentSubterm->functor()))))
		{		  
		  if (SimplificationOrdering::current()->compareWithTransposition(currentSubterm) == Greater)
		    {
		      registerRejected(ReducedInNonBasicPosition);
		      return true;
		    };
		};
	    }; // for (Flatterm* currentSubterm ..
	};
    }; // for (LitListElement* el ..  

  return false; 
}; // bool ForwardProcessing::reduceNonBasicInCurrentByCommutativity()

bool ForwardProcessing::reduceNonBasicInCurrent()
{
  CALL("reduceNonBasicInCurrent()");
  if (!_reduceNonBasic || 
      !_forwardDemodulation ||
      _currentClause->testedForReductionOfNonBasic())
    return false;
  
  _splittingBranches.reset(); // on-branch demodulation is excluded

  for (LitListElement* el = _currentClause->literals()->first(); 
       el; 
       el = el->next())
    {
      Flatterm* litBody = el->literal()->flatterm();
      if (litBody->arity())
	{
	  Flatterm::Iterator iter(litBody);
	  iter.NextSym(); // skipping the header
	  do
	    {
	      if (iter.CurrPos()->isNonBasicPosition() && 
		  iter.CurrPos()->isComplex() &&
		  _forwardDemodulation->Rewrite(iter.CurrPos(),_splittingBranches))
		{
		  //DF; cout << "RREEEDDDUCED BY " << _forwardDemodulation->Rule() << "\n";

		  _forwardDemodulation->abandonCurrentQuery();
		  registerRejected(ReducedInNonBasicPosition);

		  return true;
		};
	      iter.NextSym();
	    }
	  while (!iter.EndOfTerm());
	};
    }; 
  return false;
}; // bool ForwardProcessing::reduceNonBasicInCurrent()


inline bool ForwardProcessing::demodulateIntoCurrent()
{
  CALL("demodulateIntoCurrent()");
  ASSERT(!_currentClause.isNull());
  if (_currentClause->normalisedByDemodulation() ||
      (!_forwardDemodulation))
    return false;
  
  bool res = false;

  if (_onBranchForwardDemodulation)
    {
      // collect _splittingBranches 
      _splittingBranches.reset();
      DOP(_splittingBranches.unfreeze());
      for (const LitListElement* el = _currentClause->literals()->first(); 
	   el; 
	   el = el->next())
	if (el->literal()->header().IsSplitting()) 
	  _splittingBranches.addSafe(el->literal()->header().HeaderNum());
      DOP(_splittingBranches.freeze()); 
    };

  LitHandle spareLiteral;  
  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(spareLiteral))
    return false;
      
  for (LitListElement* el = _currentClause->literals()->first(); 
       el; 
       el = el->next())
    if (normaliseLitByDemodulation(el->literal(),
				   spareLiteral))
      res = true;
    
  if (res)
    { 
      _currentClause->background().UsedRule(ClauseBackground::ForwDemod);
      _currentClause->background().MakeVIP();
      _currentClause->registerIntraLiteralModification();
    };
  _currentClause->markAsNormalisedByDemodulation();
  return res;
}; // bool ForwardProcessing::demodulateIntoCurrent()



inline
bool 
ForwardProcessing::normaliseLitByDemodulation(LitHandle& lit,
					      LitHandle& spareLit)
{
  CALL("normaliseLitByDemodulation(..)");
  ASSERT(!spareLit.isNull());
  ulong arity = lit->arity();
  if (!arity) return false;
  bool res = false;
  ulong argNum = 0UL;
  const Flatterm* arg = lit->flatterm()->next();
  ulong argPos;
  do
    {
      argPos = lit->flatterm()->subtermPosition(arg);
      if (arg->isComplex() &&
	  normaliseComplexSubtermByDemodulation(lit,
						arg,
						spareLit))
	{
	  res = true;
	  // note that lit->flatterm() has changed
	  arg = lit->flatterm()->subtermIn(argPos);
	};
      arg = arg->after();
      ++argNum;
    }
  while (argNum < arity);
  return res;
}; // bool ForwardProcessing::normaliseLitByDemodulation(..)



inline
bool 
ForwardProcessing::normaliseComplexSubtermByDemodulation(LitHandle& lit,
							 const Flatterm* subterm,
							 LitHandle& spareLit)
{
  CALL("normaliseComplexSubtermByDemodulation(..)");
  ASSERT(subterm->isComplex());
  bool res = false;
  const ulong subtermPosition = lit->flatterm()->subtermPosition(subterm);
  ulong modifiedPos = subtermPosition;
  DOP(ulong endOfSubtermPosition = lit->flatterm()->subtermPosition(subterm->after()));
  if (rewriteComplexSubtermByDemodulation(lit,
					  subterm,
					  modifiedPos,
					  spareLit))
    {
      ASSERT(modifiedPos >= subtermPosition);
      ASSERT(modifiedPos < endOfSubtermPosition);
      res = true;
      subterm = lit->flatterm()->subtermIn(subtermPosition);
      DOP(endOfSubtermPosition = lit->flatterm()->subtermPosition(subterm->after()));
      while (subterm->isComplex() &&
	     rewriteComplexSubtermByDemodulation(lit,
						 subterm,
						 modifiedPos,
						 spareLit))
	{
	  ASSERT(modifiedPos >= subtermPosition);
	  ASSERT(modifiedPos < endOfSubtermPosition);
	  subterm = lit->flatterm()->subtermIn(subtermPosition);
	  DOP(endOfSubtermPosition = lit->flatterm()->subtermPosition(subterm->after()));
	};
    };
  return res;
}; // bool ForwardProcessing::normaliseComplexSubtermByDemodulation(..)


inline
bool 
ForwardProcessing::rewriteComplexSubtermByDemodulation(LitHandle& lit,
						       const Flatterm* subterm,
						       ulong& modifiedPosition,
						       LitHandle& spareLit)
{
  CALL("rewriteComplexSubtermByDemodulation(..)");
  ASSERT(subterm->isComplex());
  if (_prohibitNonBasicDemodulation &&
      subterm->isNonBasicPosition())
    return false;
  const Flatterm* litBody = lit->flatterm();
  const ulong subtermPosition = litBody->subtermPosition(subterm);  
  if (rewriteTopOfComplexSubtermByDemodulation(lit,
					       subterm,
					       spareLit))
    {
      modifiedPosition = subtermPosition;
      return true;
    };
  
  // try eligible arguments of subterm 
  ulong arity = subterm->arity();
  if (!arity) return false;

  const Flatterm* arg = subterm->next();
  for (ulong argNum = 0UL; argNum < arity; ++argNum)
    {
      if (arg->isComplex())
	{
	  ulong endOfArgPos = litBody->subtermPosition(arg->after());
	  if (endOfArgPos >= modifiedPosition)
	    {
	      // this argument is eligible
	      if (rewriteComplexSubtermByDemodulation(lit,
						      arg,
						      modifiedPosition,
						      spareLit))
		return true;
	    };
	};
      arg = arg->after();
    };
  return false;
}; // bool ForwardProcessing::rewriteComplexSubtermByDemodulation(..)


inline
bool 
ForwardProcessing::rewriteTopOfComplexSubtermByDemodulation(LitHandle& lit,
							    const Flatterm* subterm,
							    LitHandle& spareLit)
{
  CALL("rewriteTopOfComplexSubtermByDemodulation(..)");
  ASSERT(subterm->isComplex());
  ASSERT(lit->flatterm()->containsSubterm(subterm));
  ASSERT(!spareLit.isNull());
  if (_forwardDemodulation->Rewrite(subterm,_splittingBranches))
    {

      //DF; _currentClause->outputComplete(cout << "OOPPSS ") << "\n";
  

      if (!_useInputNonBasicness ||
	  _forwardDemodulation->RHSTerm().isVariable() ||
	  checkSubstSafetyWRTInputNonBasicness(_forwardDemodulation->RHSTerm().First(),
					       false,
					       _forwardDemodulation->substitutionCache()))
	{
	  TERM header = lit->header();
	  if (header.IsEquality()) header.MakeUnordered();
	  ASSERT(header.arity());

	  spareLit->pushNonpropHeader(header,lit->properties());
	  Flatterm::Iterator iter;
	  iter.Reset(lit->flatterm());
	  iter.NextSym(); // skipping the header
	  TERM sym;
	  if (_useRunTimeNonBasicness)
	    {
	      while (iter.NotEndOfTerm())
		{
		  if (iter.CurrPos() == subterm)
		    {
		      do
			{
			  sym = _forwardDemodulation->RHSSymbol();
		  
			  if (sym.isVariable())
			    {
			      spareLit->pushVar(sym,true);
			    }
			  else // functor
			    if (sym.arity())
			      {
				if (!spareLit->pushNonconstFun(sym,
							       _forwardDemodulation->RHSSubstPosInSubstitution()))
				  {
				    _forwardDemodulation->abandonCurrentQuery();
				    return false;
				  };
			      }
			    else // constant
			      spareLit->pushConst(sym,
						  _forwardDemodulation->RHSSubstPosInSubstitution());
			}
		      while (_forwardDemodulation->RHSNext());
		      iter.after();
		    }
		  else
		    {
		      sym = iter.CurrPos()->Symbol();
        
		      if (sym.isVariable())
			{
			  spareLit->pushVar(sym,true);
			}
		      else // functor
			if (sym.arity())
			  {
			    if (!spareLit->pushNonconstFun(sym,
							   iter.CurrPos()->isSubstitutionPosition())) 
			      {
				_forwardDemodulation->abandonCurrentQuery();
				return false;
			      };
			  }
			else // constant
			  spareLit->pushConst(sym,
					      iter.CurrPos()->isSubstitutionPosition());
		      iter.NextSym();
		    }; 
		}; // while (iter.NotEndOfTerm())
	    }
	  else // !_useRunTimeNonBasicness
	    {

	      while (iter.NotEndOfTerm())
		{
		  if (iter.CurrPos() == subterm)
		    {
		      do
			{
			  sym = _forwardDemodulation->RHSSymbol();
		  
			  if (sym.isVariable())
			    {
			      spareLit->pushVar(sym);
			    }
			  else // functor
			    if (sym.arity())
			      {
				if (!spareLit->pushNonconstFun(sym))
				  {
				    _forwardDemodulation->abandonCurrentQuery();
				    return false;
				  };
			      }
			    else // constant
			      spareLit->pushConst(sym);
			}
		      while (_forwardDemodulation->RHSNext());
		      iter.after();
		    }
		  else
		    {
		      sym = iter.CurrPos()->Symbol();
        
		      if (sym.isVariable())
			{
			  spareLit->pushVar(sym);
			}
		      else // functor
			if (sym.arity())
			  {
			    if (!spareLit->pushNonconstFun(sym)) 
			      {
				_forwardDemodulation->abandonCurrentQuery();
				return false;
			      };
			  }
			else // constant
			  spareLit->pushConst(sym);
		      iter.NextSym();
		    }; 
		}; // while (iter.NotEndOfTerm())
	    };

	  spareLit->endOfLiteral();

	  if (_useInputNonBasicness) spareLit->markInputNonBasicPositions();

	  lit.swapWith(spareLit);
	  Clause* simplifier = _forwardDemodulation->Rule(); 
	  ASSERT(simplifier->Active());  

	  _currentClause->background().pushAncestor(simplifier,
						    simplifier->inferenceDepth());     
	  simplifier->setFlag(Clause::FlagHasChildren);

      
	  if (_forwardDemodulation->SplittingBranchesUsed())
	    {
	      _currentHasBeenSimplifiedByOnBranchDemodulation = true;
	      ++(_statistics->inferences.simplifying.onBranchForwardDemodulation);
	    }
	  else
	    {
	      _currentHasBeenSimplifiedByDemodulation = true;
	      ++(_statistics->inferences.simplifying.forwardDemodulation);
	    };
	  _forwardDemodulation->abandonCurrentQuery();
	  return true;
	};
      _forwardDemodulation->abandonCurrentQuery();
    }; // if (_forwardDemodulation->Rewrite(subterm,_splittingBranches))
  return false;
}; // bool ForwardProcessing::rewriteTopOfComplexSubtermByDemodulation(..)



bool 
ForwardProcessing::checkSubstSafetyWRTInputNonBasicness(const TERM* complexTerm,
							bool alreadyInputNonBasic,
							const ForwardMatchingSubstitutionCache& subst)
{
  CALL("checkSubstSafetyWRTInputNonBasicness(..)");

  if (alreadyInputNonBasic)
    {
      for (const TERM* arg = complexTerm->Args();
	   !arg->IsNull();
	   arg = arg->Next())
	{
	  if (arg->isVariable())
	    {
	      if (!subst.map(arg->var())->isNonBasicPosition()) return false;
	    }
	  else 
	    {
	      if (!checkSubstSafetyWRTInputNonBasicness(arg->First(),
							true,
							subst))
		return false;
	    };
	};
    }
  else // !alreadyInputNonBasic
    {
      const bool* nonBasicArgs = complexTerm->nonBasicArguments();
      if (nonBasicArgs)
	{
	  ulong argNum = 0UL;
	  for (const TERM* arg = complexTerm->Args();
	       !arg->IsNull();
	       arg = arg->Next())
	    {
	      if (nonBasicArgs[argNum])
		{
		  if (arg->isVariable())
		    {
		      if (!subst.map(arg->var())->isNonBasicPosition()) 
			{
			  return false;
			};
		    }
		  else
		    {
		      if (!checkSubstSafetyWRTInputNonBasicness(arg->First(),
								true,
								subst))
			return false;
		    };
		}
	      else
		{
		  if (!arg->isVariable() &&
		      !checkSubstSafetyWRTInputNonBasicness(arg->First(),
							    false,
							    subst))
		    return false;
		};
	      ++argNum;
	    };
	}
      else 
	{ 
	  for (const TERM* arg = complexTerm->Args();
	       !arg->IsNull();
	       arg = arg->Next())
	    {
	      if (!arg->isVariable() &&
		  !checkSubstSafetyWRTInputNonBasicness(arg->First(),
							false,
							subst))
		return false;
	    };
	};
    };
  return true;
}; // bool ForwardProcessing::checkSubstSafetyWRTInputNonBasicness(..)


Clause* ForwardProcessing::canBeRewrittenByDemodulation(TemporaryClause& cl)
{
  CALL("canBeRewrittenByDemodulation(TemporaryClause& cl)");
  if (!_forwardDemodulation) return 0;
  _splittingBranches.reset(); // on-branch demodulation is excluded
  for (LitListElement* el = cl.literals()->first(); 
       el; 
       el = el->next())
    {
      if (el->literal()->arity())
	{ 
	  Flatterm::Iterator iter;
	  iter.Reset(el->literal()->flatterm());
	  iter.NextSym(); // skipping the header
	  do
	    {
	      if (iter.CurrPos()->isComplex() &&
		  _forwardDemodulation->Rewrite(iter.CurrPos(),_splittingBranches))
		{
		  Clause* res = _forwardDemodulation->Rule();
		  _forwardDemodulation->abandonCurrentQuery();
		  return res;
		};
	      iter.NextSym();
	    }
	  while (iter.NotEndOfTerm());
	};
    };
  return 0;
}; // Clause* ForwardProcessing::canBeRewrittenByDemodulation(TemporaryClause& cl)


inline bool ForwardProcessing::simplifyCurrentByBuiltInTheories()
{
  CALL("implifyCurrentByBuiltInTheories()");
  ASSERT(!_currentClause.isNull());

  if ((!BuiltInTheoryDispatcher::current()) || 
      _currentClause->simplifiedByBuiltInTheories())
    return false;
  
  bool res = false;
  LitHandle spareLiteral;
  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(spareLiteral))
    return false;

  for (LitListElement* el = _currentClause->literals()->first(); 
       el; 
       el = el->next())
    if (simplifyLitByByBuiltInTheories(el->literal(),spareLiteral))
      {
	res = true;
	++(_statistics->inferences.simplifying.builtInTheory); 
	// ^ should move to BuiltInTheoryDispatcher
      };
  if (res)
    {
      _currentClause->background().UsedRule(ClauseBackground::SimplificationByBuiltInTheories);     
      _currentClause->registerIntraLiteralModification();
      _currentHasBeenSimplifiedByBuiltInTheories = true;
    };
  _currentClause->markAsSimplifiedByBuiltInTheories();
  return res;
}; // bool ForwardProcessing::simplifyCurrentByBuiltInTheories()


inline
bool ForwardProcessing::simplifyLitByByBuiltInTheories(LitHandle& lit,
						       LitHandle& spareLit)
{
  CALL("simplifyLitByByBuiltInTheories(..)");
  ASSERT(!lit.isNull());
  ASSERT(!spareLit.isNull());
  bool res = false;
  Clause* simplifier;
  while ((simplifier = BuiltInTheoryDispatcher::current()->simplify(*lit,*spareLit))) // assignment, not a comparison 
    {
      res = true;
      _currentClause->background().pushAncestor(simplifier,
						simplifier->inferenceDepth());
      simplifier->setFlag(Clause::FlagHasChildren);

      ASSERT(spareLit->flatterm()->CheckIntegrity());

      spareLit->properties() = lit->properties();

      if (_useRunTimeNonBasicness)
	spareLit->unmarkAllSubstitutionPositions();

      lit.swapWith(spareLit);      
    };    

  return res;
}; // bool ForwardProcessing::simplifyLitByByBuiltInTheories(..)




inline bool ForwardProcessing::simplifyCurrentByDestructiveEqualityResolution()
{
  CALL("simplifyCurrentByDestructiveEqualityResolution()");
  ASSERT(!_currentClause.isNull());
  if ((!_destructiveEqResolution) ||
      _currentClause->simplifiedByDestructiveEqualityResolution())
    return false;
  BasicTempLitList resolvedLit(_currentClause->literals()->elementPool());
  ulong var;
#if (defined _SUPPRESS_WARNINGS_) || (defined DEBUG_NAMESPACE)
  var = ULONG_MAX;
#endif
  const Flatterm* instance;
  DOP(ulong debugNumOfLiterals = _currentClause->literals()->length());
  DLC_INIT(debugNumOfLiterals);
  for (LitListElement* el = _currentClause->literals()->last();
       el;
       /* nothing here */)
    if (el->literal()->isNegative() && 
	el->literal()->isEquality() &&
	el->literal()->isUnorderedEquality())
      {
	DLC_INC;
	ASSERT(resolvedLit.isEmpty());
	instance = 0;
	if (el->literal()->arg1()->isVariable())
	  {
	    var = el->literal()->arg1()->Var();
	    if (!el->literal()->arg2ContainsVariable(var))
	      instance = el->literal()->arg2();
	  }
	else
	  if (el->literal()->arg2()->isVariable())
	    {
	      var = el->literal()->arg2()->Var();
	      if (!el->literal()->arg1ContainsVariable(var))
		instance = el->literal()->arg1();
	    };
	if (!instance)
	  {
	    el = el->previous();
	  }
	else // el->literal() is eligible
	  {
	    LitListElement* tmp = el->previous();
	    _currentClause->literals()->passTo(el,resolvedLit);
	    el = tmp;	    
	    ulong instanceDepth = instance->Depth();
	    long instanceSize = instance->size();
	    if (TempLitListExtras::instantiationIsPossible(*(_currentClause->literals()),
							   var,
							   instanceDepth,
							   instanceSize,
							   _useInputNonBasicness &&
                                                           !instance->isVariable() &&
							   !instance->isNonBasicPosition()))
	      {

		
		//DF; _currentClause->outputComplete(cout << "---- BEFORE ");
		//DF;  resolvedLit.output(cout << " WITH ") << "\n";

		LitHandle spareLiteral;
		if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(spareLiteral))
		  return false;
		for (LitListElement* oldEl = _currentClause->literals()->first();
		     oldEl;
		     oldEl = oldEl->next())
		  {
		    if (oldEl->literal()->containsVariable(var))
		      {
			oldEl->literal()->instantiateVar(*spareLiteral,
							 var,
							 instance,
							 _useRunTimeNonBasicness  &&
							 !instance->isVariable() &&
							 !instance->isNonBasicPosition());
			spareLiteral->properties() = oldEl->literal()->properties();
			oldEl->literal().swapWith(spareLiteral);
			_currentClause->registerIntraLiteralModification();
		      };
		  };
		_currentClause->background().UsedRule(ClauseBackground::EqResSimp);
		_currentClause->background().MakeVIP();
		ASSERT(debugNumOfLiterals == _currentClause->literals()->length() + 1);
		_currentHasBeenSimplifiedByDestructiveEqualityResolution = true;
		++(_statistics->inferences.simplifying.destructiveEqualityResolution);
		//DF; _currentClause->outputComplete(cout << "---- AFTER ") << "\n"; 

		return true;
	      }
	    else // instantiation impossible
	      resolvedLit.popTo(*(_currentClause->literals()));
	  };
      }
    else
      el = el->previous();
  ASSERT(debugNumOfLiterals == _currentClause->literals()->length());
  _currentClause->markAsSimplifiedByDestructiveEqualityResolution();
  return false;
}; // bool ForwardProcessing::simplifyCurrentByDestructiveEqualityResolution()

inline bool ForwardProcessing::splitNegativeEqualitiesInCurrent()
{
  CALL("splitNegativeEqualitiesInCurrent()");
  ASSERT(!_currentClause.isNull());

  if ((!_negativeEqualitySplitting) ||
      _currentClause->lockedForNegativeEqualitySplitting())
    return false;
  ulong numOfSplitOffArgs;
  ulong numOfNewNames;
  bool res = 
    _negativeEqualitySplitting->splitOffAppropriateNegEqArguments(*_currentClause,
							     &_internalClauseCollector,
							     numOfSplitOffArgs,
							     numOfNewNames);

  if (res)
    {
      _currentClause->registerIntraLiteralModification();
      _currentHasBeenTransformedByNegativeEqualitySplitting = true;
      ++(_statistics->inferences.negativeEqualitySplitting.accomplished);
      _statistics->inferences.negativeEqualitySplitting.argumentsSplitOff +=
	numOfSplitOffArgs;
      _statistics->inferences.negativeEqualitySplitting.introducedNames +=
	numOfNewNames;
    }
  else
    _currentClause->markAsLockedForNegativeEqualitySplitting();
  return res;
}; // bool ForwardProcessing::splitNegativeEqualitiesInCurrent()

inline bool ForwardProcessing::applyPropositionalSplittingToCurrent()
{
  CALL("applyPropositionalSplittingToCurrent()");
  ASSERT(!_currentClause.isNull());
  if ((!_propositionalSplitting) ||
      _currentClause->lockedForPropositionalSplitting())
    return false;
  
  ulong numOfComponents;
  ulong numOfNewNames;
  bool res = 
    _propositionalSplitting->split(*_currentClause,
				   &_internalClauseCollector,
				   numOfComponents,
				   numOfNewNames);
  if (res)
    {
      _currentClause->registerIntraLiteralModification();
      _currentHasBeenTransformedByPropositionalSplitting = true;
      ++(_statistics->inferences.propositionalSplitting.accomplished);
      
      _statistics->inferences.propositionalSplitting.components +=
	numOfComponents;
      _statistics->inferences.propositionalSplitting.introducedComponentNames +=
	numOfNewNames;
    }
  else
    _currentClause->markAsLockedForPropositionalSplitting();
  return res;
}; // bool ForwardProcessing::applyPropositionalSplittingToCurrent()

bool ForwardProcessing::discardCurrentAsTooComplex()
{
  CALL("discardCurrentAsTooComplex()");
  ASSERT(!_currentClause.isNull());
  if (_currentClause->literals()->length() > _clauseLengthLimit)
    {
      registerRejected(TooManyLiterals);
      return true;
    };
  ulong clauseSize = 0UL;  
  for (LitListElement* el = _currentClause->literals()->first();
       el;
       el = el->next())
    {
      ulong litSize = el->literal()->size();
      if (litSize > _literalSizeLimit)
	{
	  registerRejected(LiteralTooLarge);
	  return true;
	};
	
      clauseSize += litSize;
      if (clauseSize > _clauseSizeLimit)
	{
	  registerRejected(ClauseTooLarge);
	  return true;
	};
    }; 
  
  return false;
}; // bool ForwardProcessing::discardCurrentAsTooComplex()


//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_PROCESSING
#  define DEBUG_NAMESPACE "ForwardProcessing::AncestorReceiver"
#endif
#include "debugMacros.hpp"
//===================================================================

void ForwardProcessing::AncestorReceiver::receive(const Clause* const & anc)
{
  CALL("receive(const Clause* const & anc)");
  _forwardProcessing->_currentClause->background().pushAncestor(const_cast<Clause*>(anc),
								anc->inferenceDepth());      
}; // void ForwardProcessing::AncestorReceiver::receive(const Clause*& anc)


//=================================================================
