//
// File:         InferenceDispatcher.cpp
// Description:  Interface to several inference engines.
// Created:      Mar 5, 2000, 20:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "InferenceDispatcher.hpp"
#include "Clause.inl"
#include "SymbolInfo.hpp"
//====================================================

VK::InferenceDispatcher::InferenceDispatcher() : 
  binResFlag(false),
  OPEN_BR(1,"br 1") DCOMMA  
  _paramodulationFlag(false),
  _paramodulateFromVariables(false),
  _noInfBetweenRulesFlag(false),
  _binRes(&_ordResIndex),  
  _forwardSuperposition(&_ordResIndex), 
  _equalityFactoring(0L), 
  _backwardSuperposition(&_ordResIndex), 
  _chaining(&_ordResIndex),
  _query(0),
  _clauseCollector(0),
  _forwardProcessing(0)
{
  CALL("constructor InferenceDispatcher()");
  unlock();
}; // VK::InferenceDispatcher::InferenceDispatcher()

VK::InferenceDispatcher::~InferenceDispatcher() 
{
};

void VK::InferenceDispatcher::reset()
{
  CALL("reset()");
  BK::destroy(*this);
  BK::init(*this);
}; // void VK::InferenceDispatcher::reset()


void VK::InferenceDispatcher::newQuery(Clause* cl) 
{ 
  CALL("newQuery(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->isAnswer());
  _query = cl; 
  if (builtInTransitivityLaw(_query))
    {
      _noMoreInferences = true;
    }
  else
    {
      _noMoreInferences = false;
      _continuation = FirstBinRes;
    }; 
}; // void VK::InferenceDispatcher::newQuery(Clause* cl) 


void VK::InferenceDispatcher::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked());
  _query->unprotect();
  unlock();
}; // void VK::InferenceDispatcher::abandonCurrentQuery()



bool VK::InferenceDispatcher::nextInference()
{
  CALL("nextInference()");
  ASSERT(isLocked());

  if (_noMoreInferences) return false;

  switch (_continuation)
    { 
    case FirstBinRes:
      COP("FirstBinRes");
      if (!binResFlag) 
	goto first_forw_sup;   
      ASSERT(Unifier::current()->NoBoundVariables());
      if (!_binRes.Load(_query))
	goto first_chaining;  
      _continuation = NextBinRes;  
      // no break here, proceed with the next case 

    case NextBinRes:
      COP("NextBinRes");
      if (_binRes.NextInference()) return true;
      Unifier::current()->UnbindAll(); 
      // no break here, try chaining now
       
    case FirstChaining:
      COP("FirstChaining");
    first_chaining:
      ASSERT(Unifier::current()->NoBoundVariables()); 
      if ((!_chainingTable) ||
	  (!_chaining.loadQuery(_query)))
	goto first_forw_sup; 
      _continuation = NextChaining;
      // no break here, make the first inference

    case NextChaining:
      COP("NextChaining");
      ASSERT(_chainingTable);
      if (_chaining.nextInference())
	{
	  return true;
	};
      // no break here, try forward superposition

    case FirstForwSup:
    first_forw_sup:
    COP("FirstForwSup");
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_paramodulationFlag) return false; // all the rules below deal with equality
    if ((!_forwardSuperposition.Load(_query))
	|| (!_forwardSuperposition.NextInference()))
      goto first_eq_factor;
    _continuation = NextForwSup;
    return true; 

    case NextForwSup:
      COP("NextForwSup");
      ASSERT(_paramodulationFlag);
      if (_forwardSuperposition.NextInference()) return true;
      Unifier::current()->UnbindAll(); 
      // no break here, try equality factoring here

    case FirstEqFactor:
    first_eq_factor:
    COP("first_eq_factor");
    ASSERT(_paramodulationFlag);
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_equalityFactoring.Load(_query)) goto first_back_sup; 
    _continuation = NextEqFactor;            
    // no break here, proceed with the next case

    case NextEqFactor: 
      COP("NextEqFactor");
      ASSERT(_paramodulationFlag);
      if (_equalityFactoring.NextInference()) return true;
      Unifier::current()->UnbindAll();   
      // no break here, try backward superposition

    case FirstBackSup:
    first_back_sup:
    COP("FirstBackSup");
    ASSERT(_paramodulationFlag);
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_backwardSuperposition.SetQuery(_query)) 
      {       
	return false; 
      };     
    _continuation = NextBackSup;      
    // no break here, proceed with the next case 

    case NextBackSup:
      COP("NextBackSup");
      ASSERT(_paramodulationFlag);
      if (_backwardSuperposition.NextInference()) return true;
      Unifier::current()->UnbindAll();
      return false; 
    };
  ICP("ICP1"); 
  return false;   
}; // bool VK::InferenceDispatcher::nextInference()

void VK::InferenceDispatcher::makeActive(Clause* cl)
{
  CALL("makeActive(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsDefinition));
  if (!builtInTransitivityLaw(cl))
    {
      _ordResIndex.Integrate(cl);
    };
  cl->setFlag(Clause::FlagInInferenceDispatcherAsActive);  
  cl->unprotect();
  unlock();
}; // void VK::InferenceDispatcher::makeActive(Clause* cl)

void VK::InferenceDispatcher::remove(Clause* cl) 
{
  CALL("remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsDefinition));
  if (!builtInTransitivityLaw(cl))
    {
      _ordResIndex.Remove(cl); 
    };
  cl->clearFlag(Clause::FlagInInferenceDispatcherAsActive);   
  cl->unprotect();
  unlock();
}; // void VK::InferenceDispatcher::remove(Clause* cl) 


void VK::InferenceDispatcher::integrateDefinition(Clause* cl,const TERM& definedHeader)
{
  CALL("integrateDefinition(Clause* cl,const TERM& definedHeader)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsDefinition));
  _ordResIndex.Integrate(cl); // temporary
  cl->setFlag(Clause::FlagInInferenceDispatcherAsDefinition); 
  cl->unprotect();
  unlock();
}; // void VK::InferenceDispatcher::integrateDefinition(Clause* cl,const TERM& definedHeader)

void VK::InferenceDispatcher::removeDefinition(Clause* cl)
{
  CALL("removeDefinition(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  ASSERT(cl->flag(Clause::FlagInInferenceDispatcherAsDefinition));
  _ordResIndex.Remove(cl); // temporary
  cl->clearFlag(Clause::FlagInInferenceDispatcherAsDefinition);
  cl->unprotect();
  unlock();
}; // void VK::InferenceDispatcher::removeDefinition(Clause* cl)

inline
bool VK::InferenceDispatcher::builtInTransitivityLaw(const Clause* cl)
{
  CALL("builtInTransitivityLaw(const Clause* cl)");
  if (!_chainingTable) return false;
  ulong transPred;
  if (!cl->isTransitivityLaw(transPred)) return false;
  return (_chainingTable->transitivityLaw(transPred) != 0);
}; // bool VK::InferenceDispatcher::builtInTransitivityLaw(const Clause* cl)

//====================================================




