//
// File:         Input.hpp
// Description:  Input signature, terms and clauses.
// Created:      Jun 04, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#ifndef INPUT_H
#define INPUT_H
//======================================================================
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include <cstring>
#include <cstdio>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "DestructionMode.hpp"
#include "GList.hpp"
#include "Malloc.hpp"
#include "VampireKernel.hpp"
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
class Input
{
 public:
  class Symbol;
  class Signature;
  class Term;
  class Literal;
  class Clause;
  class LiteralDefinition;
  class Interpretation;
  typedef BK::GList<BK::Malloc,Term*,Input> TermList;
  typedef BK::GList<BK::Malloc,Literal*,Input> LiteralList;
  typedef BK::GList<BK::Malloc,Clause*,Input> ClauseList;
  typedef BK::GList<BK::Malloc,LiteralList*,Input> LiteralListList;
  typedef BK::GList<BK::Malloc,LiteralDefinition*,Input> LiteralDefList;
  typedef BK::GList<BK::Malloc,Interpretation,Input> InterpretationList;
 public:
  static void destroyTermList(TermList* lst);
  static void destroyLiteralList(LiteralList* lst);
  static void destroyLiteralListList(LiteralListList* lst);
  static void destroyClauseList(ClauseList* lst);
  static void destroyLiteralDefList(LiteralDefList* lst);
};
}; // namespace VK


#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const VK::Input::ClauseList* lst);
};
#endif

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Interpretation"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
  class Input::Interpretation
    {
    public:
      Interpretation();   
      Interpretation(const char* theory,const char* nativeSymbol);
      ~Interpretation();
      void replace(const char* theory,const char* nativeSymbol)
      {
	BK::destroy(*this);
	BK::init(*this,theory,nativeSymbol);
      };
      const char* theoryName() const { return _theoryName; };
      const char* nativeSymbolName() const { return _nativeSymbolName; };
    private:
      char* _theoryName;
      char* _nativeSymbolName;
    };

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Symbol"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
class Input::Symbol
{
 public:
  enum Type { Variable, Function, Predicate };
  const Type& type() const { return _type; };
  bool isVariable() const { return _type == Variable; };
  const char* printName() const { return _printName; };
  ulong number() const { return _number; };
  void assignNumber(ulong n) { _number = n; };
  ulong arity() const 
    {
      CALL("arity() const");
      return _arity; 
    };
  ulong variableNumber() const { return _normalisedVariableNumber; };
  bool isEquality() const 
  {
    return 
      strcmp(_printName,"equal") == 0 &&
      _type == Predicate &&
      _arity == 2UL;
  };

  bool precedenceAssigned() const { return _precedenceAssigned; };
  long precedence() const 
    {
      CALL("precedence() const");
      ASSERT(precedenceAssigned());
      return _precedence;
    };

  bool positiveHeaderPrecedenceAssigned() const 
  { 
    return _positiveHeaderPrecedenceAssigned; 
  };

  long positiveHeaderPrecedence() const 
    {
      CALL("positiveHeaderPrecedence() const");
      ASSERT(positiveHeaderPrecedenceAssigned());
      return _positiveHeaderPrecedence;
    };

  bool negativeHeaderPrecedenceAssigned() const 
  { 
    return _negativeHeaderPrecedenceAssigned; 
  };

  long negativeHeaderPrecedence() const 
    {
      CALL("negativeHeaderPrecedence() const");
      ASSERT(negativeHeaderPrecedenceAssigned());
      return _negativeHeaderPrecedence;
    };


  void addInterpretation(const char* theory,const char* nativeSymbol)
    {
      CALL("addInterpretation(const char* theory,const char* nativeSymbol)");
      _interpretations = 
	new InterpretationList(_interpretations);
      _interpretations->hd().replace(theory,nativeSymbol);
    };

  const InterpretationList* interpretations() const { return _interpretations; };

  bool isAnswer() const { return _isAnswer; };

  bool weightAssigned() const { return _weightAssigned; };
  VampireKernel::TermWeightType weightConstPart() const { return _weightConstPart; };


  void setWeightConstPart(VampireKernel::TermWeightType c)
    {
      _weightConstPart = c;
      _weightAssigned = true;
    };

  Symbol* next() { return _next; };

  bool nthArgInPositiveIsSubstitutionArg(ulong argNum) const
  {
    return 
      !_nonSubstitutionArgumentInPositive ||
      !_nonSubstitutionArgumentInPositive[argNum];
  };
  bool nthArgInNegativeIsSubstitutionArg(ulong argNum) const
  {
    return 
      !_nonSubstitutionArgumentInNegative ||
      !_nonSubstitutionArgumentInNegative[argNum];
  };


 private:
  Symbol(const Type& t,const char* nm,ulong a) :
    _type(t),
    _arity(a),
    _precedenceAssigned(false),
    _positiveHeaderPrecedenceAssigned(false),
    _negativeHeaderPrecedenceAssigned(false),
    _isAnswer(false),
    _weightConstPart(1L),
    _weightAssigned(false),
    _interpretations(0),
    _nonSubstitutionArgumentInPositive(0),
    _nonSubstitutionArgumentInNegative(0),
    _variableTimeStamp(0UL),
    _normalisedVariable(0),
    _smallerNames(0),
    _greaterNames(0),
    _smallerPrecedences(0),
    _greaterPrecedences(0),
    _otherHeaderPrecedences(0),
    _next(0)
    {
      CALL("constructor Symbol(const Type& t,const char* nm,ulong a)");
      _printName = static_cast<char*>(BK::Malloc::allocate(strlen(nm) + 1));
      strcpy(_printName,nm);
    };
  ~Symbol()
    {
      CALL("destructor ~Symbol()");
      if (BK::DestructionMode::isThorough())
	{
	  BK::Malloc::deallocate(static_cast<void*>(_printName),strlen(_printName) + 1);
	  InterpretationList::destroyList(_interpretations);
	  if (_nonSubstitutionArgumentInPositive)
	    BK::Malloc::deallocate(static_cast<void*>(_nonSubstitutionArgumentInPositive),
				   _arity * sizeof(bool));
	  if (_nonSubstitutionArgumentInNegative)
	    BK::Malloc::deallocate(static_cast<void*>(_nonSubstitutionArgumentInNegative),
				   _arity * sizeof(bool));
	};
    };
  
  void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::Malloc::allocate(sizeof(Symbol));
    };
  
  void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
	  BK::Malloc::deallocate(obj,sizeof(Symbol));
    };

  void setType(const Type& t) { _type = t; };

  ulong& normalisedVariableNumber() 
    {
      CALL("normalisedVariableNumber()");
      ASSERT(_type == Variable);
      return _normalisedVariableNumber; 
    };
  ulong& variableTimeStamp() 
    {
      CALL("variableTimeStamp()");
      ASSERT(_type == Variable);
      return _variableTimeStamp; 
    };
  Symbol*& normalisedVariable() 
    {
      CALL("normalisedVariable()");
      ASSERT(_type == Variable);
      return _normalisedVariable; 
    };
  Symbol*& smallerNames() { return _smallerNames; };
  Symbol*& greaterNames() { return _greaterNames; };
  Symbol*& smallerPrecedences() { return _smallerPrecedences; };
  Symbol*& greaterPrecedences() { return _greaterPrecedences; };
  
  void assignPrecedence(long prec) 
    {
      CALL("assignPrecedence(long prec)");
      ASSERT(!precedenceAssigned());
      _precedenceAssigned = true;
      _precedence = prec;
    };

  void assignHeaderPrecedence(bool positive,long prec) 
    {
      CALL("assignHeaderPrecedence(bool positive,long prec)");
      if (positive)
	{
	  ASSERT(!positiveHeaderPrecedenceAssigned());
	  _positiveHeaderPrecedenceAssigned = true;
	  _positiveHeaderPrecedence = prec;
	}
      else
	{
	  ASSERT(!negativeHeaderPrecedenceAssigned());
	  _negativeHeaderPrecedenceAssigned = true;
	  _negativeHeaderPrecedence = prec;
	};
    };

  
  void markAsAnswer() { _isAnswer = true; };
  void markNonSubstitutionArgumentInPositive(ulong argNum)
  {
    CALL("markNonSubstitutionArgumentInPositive(ulong argNum)");
    ASSERT(argNum < _arity);
    ASSERT(_type != Variable);
    if (!_nonSubstitutionArgumentInPositive)
      {
	_nonSubstitutionArgumentInPositive = 
	  static_cast<bool*>(BK::Malloc::allocate(_arity * sizeof(bool)));
	for (ulong i = 0UL; i < _arity; ++i)
	  _nonSubstitutionArgumentInPositive[i] = false;
      };
    _nonSubstitutionArgumentInPositive[argNum] = true;
  };
  
  void markNonSubstitutionArgumentInNegative(ulong argNum)
  {
    CALL("markNonSubstitutionArgumentInNegative(ulong argNum)");
    ASSERT(argNum < _arity);
    ASSERT(_type != Variable);
    if (!_nonSubstitutionArgumentInNegative)
      {
	_nonSubstitutionArgumentInNegative = 
	  static_cast<bool*>(BK::Malloc::allocate(_arity * sizeof(bool)));
	for (ulong i = 0UL; i < _arity; ++i)
	  _nonSubstitutionArgumentInNegative[i] = false;
      };
    _nonSubstitutionArgumentInNegative[argNum] = true;
  };
  
  void setNext(Symbol* s) { _next = s; };

 private:
  Type _type;
  char* _printName;
  ulong _number;
  ulong _arity;
  bool _precedenceAssigned;
  long _precedence;
  bool _positiveHeaderPrecedenceAssigned;
  long _positiveHeaderPrecedence;
  bool _negativeHeaderPrecedenceAssigned;
  long _negativeHeaderPrecedence;
  bool _isAnswer;
  VampireKernel::TermWeightType _weightConstPart;
  bool _weightAssigned;
  InterpretationList* _interpretations;
  bool* _nonSubstitutionArgumentInPositive;
  bool* _nonSubstitutionArgumentInNegative;
  ulong _normalisedVariableNumber;
  ulong _variableTimeStamp;
  Symbol* _normalisedVariable;
  Symbol* _smallerNames;
  Symbol* _greaterNames;
  Symbol* _smallerPrecedences;
  Symbol* _greaterPrecedences;
  Symbol* _otherHeaderPrecedences;
  Symbol* _next;
  friend class Input::Signature;  
  friend class Input::Term;  
  friend class Input::Literal;  
}; // class Input::Symbol

}; // namespace VK

#ifndef NO_DEBUG_VIS
namespace std
{
inline
ostream& operator<<(ostream& str,const VK::Input::Symbol* s)
{
  return str << s->printName();
};
};
#endif

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Signature"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
class Input::Signature
{
 public:
  Signature() : 
    _symbols(0),
    _normalisedVariables(0),
    _nextNormalisedVariable(&_normalisedVariables),
    _nextNormalisedVariableNumber(0UL),
    _currentVariableTimeStamp(0UL),
    _precedences(0),
    _headerPrecedences(0),
    _symbolList(0)
    {
    };
  ~Signature()
    {
      CALL("destructor ~Signature()");
      if (BK::DestructionMode::isThorough())
	{
	  destroyTree(_symbols);
	  destroyTree(_normalisedVariables);
	};
    };
  
  void resetVariables() 
    {
      _nextNormalisedVariable = &_normalisedVariables;
      ++_currentVariableTimeStamp;
    };

  Symbol* registerVariable(const char* name);
  Symbol* registerFunction(const char* name,ulong arity);
  Symbol* registerPredicate(const char* name,ulong arity);
  Symbol* registerSymbol(bool predicate,const char* name,ulong arity);
  bool registerPrecedence(bool predicate,const char* name,ulong arity,long prec);
  bool registerHeaderPrecedence(bool positive,const char* predicateName,ulong arity,long prec);
  bool registerAnswerPredicate(const char* name,ulong arity);
  Symbol* symbolList() { return _symbolList; };

 private:
  Symbol* nextNormalisedVariable();
  static const char* normalisedVariableName(ulong varNum);
  static void destroyTree(Symbol* tree);
  static Symbol** lookUp(bool predicate,const char* name,ulong arity,Symbol** tree);
  static Symbol** lookUpPrecedence(long precedence,Symbol** tree);
  static bool findHeaderPrecedence(long precedence,Symbol* list);
 private:
  Symbol* _symbols;
  Symbol* _normalisedVariables;
  Symbol** _nextNormalisedVariable;
  ulong _nextNormalisedVariableNumber;
  ulong _currentVariableTimeStamp;
  Symbol* _precedences;
  Symbol* _headerPrecedences;
  Symbol* _symbolList;
}; // class Input::Signature

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Term"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
class Input::Term
{
 public:
  Term(Symbol* t) : _top(t), _arguments(0) {};
  Term(Symbol* t,TermList* args) : _top(t), _arguments(args) {};
  ~Term()
    {
      CALL("destructor ~Term()");
      if (BK::DestructionMode::isThorough())
	{
	  destroyTermList(_arguments);
	};
    };
  void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::Malloc::allocate(sizeof(Term));
    };
  void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
      BK::Malloc::deallocate(obj,sizeof(Term));
    };
  const Symbol* top() const { return _top; };
  const TermList* arguments() const { return _arguments; };
 private:
  Term() {};
  TermList* arguments() { return _arguments; };
  Symbol* top() { return _top; };
  void collectNonSubstitutionArguments()
  {
    if (!_top->isVariable())
      {
	ulong argNum = 0UL;
	for (TermList* arg = _arguments; arg; (arg = arg->tl(), ++argNum))
	  if (!arg->hd()->_top->isVariable())
	    {
	      arg->hd()->collectNonSubstitutionArguments();
	      _top->markNonSubstitutionArgumentInPositive(argNum);
	    };
      };
  };
 private:
  Symbol* _top;
  TermList* _arguments;
  friend class Input::Literal;
}; // class Input::Term
}; // namespace VK

#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const VK::Input::Term* t);
};
#endif

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Literal"
#endif
#include "debugMacros.hpp"
//======================================================================

namespace VK
{
class Input::Literal
{
public:
  typedef VampireKernel::LiteralAnnotation Annotation;
public:
  Literal(bool positive,const Annotation& ann,Term* at) : 
    _isPositive(positive),
    _annotation(ann),
    _atom(at)
    {
    };


  ~Literal() 
    {
      CALL("destructor ~Literal()");
      if (BK::DestructionMode::isThorough())
	{
	  delete _atom;
	};
    };   
  void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::Malloc::allocate(sizeof(Literal));
    };
  void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
      BK::Malloc::deallocate(obj,sizeof(Literal));
    };
  bool isPositive() const { return _isPositive; };
  const Annotation& annotation() const { return _annotation; };
  const Term* atom() const { return _atom; };
  Term* atom() { return _atom; };
  bool isEquality() const { return _atom->top()->isEquality(); };

 private:
  Literal() {}; 
  void collectNonSubstitutionArguments()
  {
    CALL("collectNonSubstitutionArguments()");
    ulong argNum = 0UL;
    for (TermList* arg = atom()->arguments(); 
	 arg; 
	 (arg = arg->tl(), ++argNum))
      if (!arg->hd()->top()->isVariable())
	{
	  arg->hd()->collectNonSubstitutionArguments();
	  if (_isPositive)
	    {
	      atom()->top()->markNonSubstitutionArgumentInPositive(argNum);
	    }
	  else
	    atom()->top()->markNonSubstitutionArgumentInNegative(argNum);
	};
  };
 private:
  bool _isPositive;
  Annotation _annotation;
  Term* _atom;
  friend class Input::Clause;
}; // class Input::Literal
}; // namespace VK

#ifndef NO_DEBUG_VIS
namespace std
{
inline
ostream& operator<<(ostream& str,const VK::Input::Literal* lit)
{
  if (!lit->isPositive()) str << "~";
  return str << lit->atom();
};
};
#endif

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Clause"
#endif
#include "debugMacros.hpp"
//======================================================================


namespace VK
{
class Input::Clause
{
 public:
  typedef VampireKernel::ClauseAnnotation Annotation;
  enum Category { Unclassified, Axiom, Conjecture, Hypothesis };
 public:
  Clause(const char* nm,const Annotation& ann,LiteralList* lits) : 
    _literals(lits), 
    _category(Unclassified),
    _annotation(ann)
    {
      _name = static_cast<char*>(BK::Malloc::allocate(strlen(nm) + 1));
      strcpy(_name,nm);
    };
  ~Clause()
    {
      CALL("destructor ~Clause()");
      if (BK::DestructionMode::isThorough())
	{
	  destroyLiteralList(_literals);
	  BK::Malloc::deallocate(static_cast<void*>(_name),strlen(_name) + 1);
	};
    };
  void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::Malloc::allocate(sizeof(Clause));
    };
  void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
      BK::Malloc::deallocate(obj,sizeof(Clause));
    }; 
  const LiteralList* literals() const { return _literals; };
  const Category& getCategory() const { return _category; };
  void setCategory(const Category& cat) { _category = cat; };
  const char* name() const { return _name; };
  const Annotation& annotation() const { return _annotation; };
  void collectNonSubstitutionArguments() 
  {
    for (LiteralList* lit = _literals; lit; lit = lit->tl())
      lit->hd()->collectNonSubstitutionArguments();
  };
 private:
  Clause() {};
 private:
  LiteralList* _literals;
  Category _category;
  char* _name;
  Annotation _annotation;
}; // class Input::Clause

}; // namespace VK



#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const VK::Input::Clause* cl);
};
#endif


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::LiteralDefinition"
#endif
#include "debugMacros.hpp"
//======================================================================


namespace VK
{
class Input::LiteralDefinition
{
 public:
  enum Category { Unclassified, Axiom, Conjecture, Hypothesis };
 public:
  LiteralDefinition(const char* nm,
		    Literal* head,
		    LiteralListList* tail) : 
    _head(head),
    _tail(tail),
    _category(Unclassified) 
    {
      _name = static_cast<char*>(BK::Malloc::allocate(strlen(nm) + 1));
      strcpy(_name,nm);
    };
  ~LiteralDefinition()
    {
      CALL("destructor ~LiteralDefinition()");
      if (BK::DestructionMode::isThorough())
	{ 
	  delete _head;
	  destroyLiteralListList(_tail);
	  BK::Malloc::deallocate(static_cast<void*>(_name),strlen(_name) + 1);
	};
    };
  void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::Malloc::allocate(sizeof(LiteralDefinition));
    };
  void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
      BK::Malloc::deallocate(obj,sizeof(LiteralDefinition));
    }; 
  const Literal* head() const { return _head; };
  const LiteralListList* tail() const { return _tail; };
  const Category& getCategory() const { return _category; };
  void setCategory(const Category& cat) { _category = cat; };
  const char* name() const { return _name; };
 private:
  LiteralDefinition() {};
 private:
  Literal* _head;
  LiteralListList* _tail;
  Category _category;
  char* _name;
}; // class Input::LiteralDefinition

}; // namespace VK



#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const VK::Input::LiteralDefinition* ld);
};
#endif




//======================================================================
#endif
