//
// File:         LiteralSelection.cpp
// Description:  Defines various literal selection functions
// Created:      Dec 12, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Note:         Was a part of NewClause.
// Revised:      May 15, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               A number of bugs fixed, mostly in the bidirectional selection
//               functions.  
// Revised:      Jun 01, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fixed in function1(..).
//============================================================================
#include "VampireKernelDebugFlags.hpp"
#include "LiteralSelection.hpp"
#include "Term.hpp"
#include "SymbolInfo.hpp"
#include "Signature.hpp"
#include "TermWeightType.hpp"
#include "TemporaryLiteral.inl"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif 
#include "debugMacros.hpp"
//=================================================


using namespace BK;
using namespace Gem;
using namespace VK;

const LiteralSelection* LiteralSelection::_currentObject; 
CustomLiteralSelection* LiteralSelection::_customSelection; 

LiteralSelection::LiteralSelection() :
  _allWithMaxPriorityFilter(CustomLiteralSelection::literalPriority),
  _oneOfMaxSizeFilter(CustomLiteralSelection::literalSize),
  _allOfMaxSizeFilter(CustomLiteralSelection::literalSize),
  _oneOfMinSizeFilter(CustomLiteralSelection::literalSize),
  _allOfMinSizeFilter(CustomLiteralSelection::literalSize),
  _allNegEqOrAllFilter(CustomLiteralSelection::isNegativeEquality),
  _oneNegEqOrAnyoneFilter(CustomLiteralSelection::isNegativeEquality),
  _allExceptPosEqOrAllFilter(CustomLiteralSelection::isNotAvailableForParamodFrom),
  _allWithMaxNumOfDiffVarsFilter(CustomLiteralSelection::numOfDifferentVariables),
  _allWithMostShallowVarsFilter(&_inverseVarDepthProfile1,&_inverseVarDepthProfile2),
  _complexFilter3(&_complexCharacteristic1_1,&_complexCharacteristic1_2),
  _complexFilter5(&_complexCharacteristic1_1,&_complexCharacteristic1_2),
  _complexLiteralOrdering1_1(10UL,1.1f,0.0f),
  _complexLiteralOrdering1_2(10UL,2.0f,0.0f),
  _complexLiteralOrdering1_3(10UL,2.0f,-10.0f)
{      
  _literalComparisonMode = 0L;
  _standardLiteralOrdering = 0;
  _basicSelFun = function1; 
  _splittingIsEnabled = false;
  _splittingStyle = 0L;
      
  _oneOfEachPolarityFilter.setSelectOne(false);

  
  _allWithMaxPriorityFilter.setSelectOne(false);
  _oneOfMaxSizeFilter.setSelectOne(true);
  _allOfMaxSizeFilter.setSelectOne(false);
  _oneOfMinSizeFilter.setSelectOne(true);
  _allOfMinSizeFilter.setSelectOne(false);

  _allNegEqOrAllFilter.setSelectOne(false);
  _oneNegEqOrAnyoneFilter.setSelectOne(true);
  _allExceptPosEqOrAllFilter.setSelectOne(false);


  _oneBigNegEqOrOneBigWithMaxPriorityFilter.setSelectOne(true);
  _oneBigNegEqOrOneBigWithMaxPriorityFilter.resetFilters();
  _oneBigNegEqOrOneBigWithMaxPriorityFilter.addFilter(&_allWithMaxPriorityFilter);
  _oneBigNegEqOrOneBigWithMaxPriorityFilter.addFilter(&_allNegEqOrAllFilter);
  _oneBigNegEqOrOneBigWithMaxPriorityFilter.addFilter(&_oneOfMaxSizeFilter);
 
  _oneBigNegEqOrOneBigFilter.setSelectOne(true);
  _oneBigNegEqOrOneBigFilter.resetFilters();
  _oneBigNegEqOrOneBigFilter.addFilter(&_allNegEqOrAllFilter);
  _oneBigNegEqOrOneBigFilter.addFilter(&_oneOfMaxSizeFilter);

  _oneSmallPreferablyNegEqFilter.setSelectOne(true);
  _oneSmallPreferablyNegEqFilter.resetFilters();
  _oneSmallPreferablyNegEqFilter.addFilter(&_allOfMinSizeFilter);
  _oneSmallPreferablyNegEqFilter.addFilter(&_oneNegEqOrAnyoneFilter);


  _allWithMaxNumOfDiffVarsFilter.setSelectOne(false);
  _allWithMostShallowVarsFilter.setSelectOne(false);

  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.setSelectOne(true);
  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.resetFilters();
  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.addFilter(&_allWithMaxPriorityFilter);
  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.addFilter(&_allWithMostShallowVarsFilter);
  _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter.addFilter(&_oneOfMaxSizeFilter);

  _oneOfMaxSizeWithMaxPriorityFilter.setSelectOne(true);
  _oneOfMaxSizeWithMaxPriorityFilter.resetFilters();
  _oneOfMaxSizeWithMaxPriorityFilter.addFilter(&_allWithMaxPriorityFilter);
  _oneOfMaxSizeWithMaxPriorityFilter.addFilter(&_oneOfMaxSizeFilter);
  

  _complexFilter1.setSelectOne(true);
  _complexFilter1.resetFilters();
  _complexFilter1.addFilter(&_allWithMaxPriorityFilter);
  _complexFilter1.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter1.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _complexFilter1.addFilter(&_allWithMostShallowVarsFilter);
  _complexFilter1.addFilter(&_oneOfMaxSizeFilter);
      
  _complexFilter2.setSelectOne(true);
  _complexFilter2.resetFilters();
  _complexFilter2.addFilter(&_allWithMaxPriorityFilter);
  _complexFilter2.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter2.addFilter(&_allOfMaxSizeFilter);
  _complexFilter2.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _complexFilter2.addFilter(&_allWithMostShallowVarsFilter);

  _complexFilter3.setSelectOne(false);

  _complexFilter4.setSelectOne(true);
  _complexFilter4.resetFilters();
  _complexFilter4.addFilter(&_allWithMaxPriorityFilter);
  _complexFilter4.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter4.addFilter(&_complexFilter3);
  _complexFilter4.addFilter(&_oneOfMaxSizeFilter);

  _complexFilter5.setSelectOne(true);
      

  _maxSelectionAcceptability1.resetPrimitives();
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noNegativeWithHighPriorityUnselected);
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noBigNegativeUnselected);
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesUnselected);

  _maxSelectionAcceptability2.resetPrimitives();
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeWithHighPriorityUnselected);
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeWithManyDiffVarsUnselected);
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesUnselected);


}; // LiteralSelection()





void 
LiteralSelection::select(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected) const
{
  CALL("select(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected) const");

  LitListStatistics stat; 
  allMaximalSelected = false;
  TempLitListExtras::adjustSelectionLayers(lits);
  BasicTempLitList nonselectableLits(unselectedLits.elementPool());
  bool allRemainingAreSelected;
  unselectNonselectable(lits,nonselectableLits,allRemainingAreSelected);
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));

  if (allRemainingAreSelected) 
    {
      selectedLits.collectAllFrom(lits);
      goto mark_and_return;  
    };
  stat.collectFor(lits);  
  _currentObject = this;

  if (_splittingIsEnabled)
    {      
      switch (_splittingStyle)
	{
	case 1L: 
	  selectWithNonblockingSplitting(lits,selectedLits,unselectedLits,allMaximalSelected,stat); 
	  goto mark_and_return;
	case 2L: 
	  selectWithBlockingSplitting(lits,selectedLits,unselectedLits,allMaximalSelected,stat); 
	  goto mark_and_return;
	default: 
	  RuntimeError::report("Wrong splitting style number."); 
	  goto mark_and_return;
	};
    };
 
  // no splitting
 
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  if (_stepResolutionLiteralSelectionFlag)
    {
      bool containsNextStateLiterals = false;
      bool containsCurrentStateLiterals = false;
      LitListElement* el = lits.first();
    check_lit:
      if (el)
	{
	  if (isNextStepLiteral(*(el->literal()))) 
	    {
	      containsNextStateLiterals = true;
	    }
	  else
	    if (isCurrentStepLiteral(*(el->literal()))) 
	      containsCurrentStateLiterals = true;
	  if (containsNextStateLiterals && containsCurrentStateLiterals)
	    {
	      unselectCurrentStateLiterals(lits,selectedLits,unselectedLits);
	      ASSERT(!lits.isEmpty());
	      if (lits.length() == 1) 
		{ 
		  lits.popTo(selectedLits); 
		  goto mark_and_return; 
		};
	      stat.collectFor(lits);
	    }
	  else
	    {
	      el = el->next();
	      goto check_lit;
	    };
	};	
    };
#endif

  _basicSelFun(lits,selectedLits,unselectedLits,allMaximalSelected,stat);

 mark_and_return:
  if (!nonselectableLits.isEmpty())
    {
      allMaximalSelected = false;
      unselectedLits.collectAllFrom(nonselectableLits);
    };
  mark(selectedLits,unselectedLits);
}; // void LiteralSelection::select(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected)



void 
LiteralSelection::selectForSupport(LitList& lits,
				   LitList& selectedLits,
				   LitList& unselectedLits) const
{
  CALL("selectForSupport(..)");
  LitListStatistics stat; 
  TempLitListExtras::adjustSelectionLayers(lits);
  BasicTempLitList nonselectableLits(unselectedLits.elementPool());
  bool allRemainingAreSelected;
  unselectNonselectable(lits,nonselectableLits,allRemainingAreSelected);
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));

  if (allRemainingAreSelected) 
    {
      selectedLits.collectAllFrom(lits);
      goto mark_and_return;  
    };

  stat.collectFor(lits);

  if (stat.numOfNegNonanswerLits() || stat.numOfNoneqNonanswerLits())
    {
      // select everything apart from positive equalities
      while (!lits.isEmpty())
	{
	  if (lits.last()->literal()->isEquality() && 
	      lits.last()->literal()->isPositive() &&
	      lits.last()->literal()->properties().useEqForParamodulationFrom())
	    {
	      lits.popTo(unselectedLits);
	    }
	  else
	    lits.popTo(selectedLits);
	};

    }
  else // only positive equalities left
    {
      _currentObject = this;
      bool dummyBool;
      _basicSelFun(lits,selectedLits,unselectedLits,dummyBool,stat);
    };
  
 mark_and_return:
  unselectedLits.collectAllFrom(nonselectableLits);
  mark(selectedLits,unselectedLits);
}; // void LiteralSelection::selectForSupport(..) const



bool 
LiteralSelection::selectForElimination(LitList& lits,LitList& selectedLits,
                                       LitList& unselectedLits,  
                                       const VK::SymbolInfo* eliminationTable) const
{
  CALL("selectForElimination(..) const");
  TempLitListExtras::adjustSelectionLayers(lits);
  LitListElement* el = lits.last();
  while (el)
    {
      if (eliminationTable->isToBeEliminated(el->literal()->header().functor()))
	{
	  ASSERT(!el->literal()->isAnswerLiteral());	 
	  lits.passTo(el,selectedLits);
#ifdef DEBUG_NAMESPACE
	  for (LitListElement* debugEl = lits.first(); 
	       debugEl;
	       debugEl = debugEl->next())
	    {
	      ASSERT(!eliminationTable->isToBeEliminated(debugEl->literal()->header().functor()));
	    };
#endif
	  lits.passAllTo(unselectedLits);
	  mark(selectedLits,unselectedLits);
	  return true;
	};        
      
      el = el->previous();
    };
  return false; 
}; // bool LiteralSelection::selectForElimination(LitList& lits,...




inline 
void 
LiteralSelection::selectWithNonblockingSplitting(LitList& lits,LitList& selectedLits,
						 LitList& unselectedLits,
						 bool& allMaximalSelected,
						 LitListStatistics& stat) const
{
  CALL("selectWithNonblockingSplitting(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  allMaximalSelected = false;
  if (lits.length() > (stat.numOfNegSplittingBranches() + stat.numOfPosSplittingBranches()))
    {
      // a nonsplitting literal exist, all splitting branches are unselected
      DLC_INIT(lits.length());
      for (LitListElement* el = lits.first(); el; /* nothing here */)
	{
	  DLC_INC;
	  if (el->literal()->header().IsSplitting())     
	    {	      
	      LitListElement* tmp = el->next();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else
	    el = el->next();
	};
      

      ASSERT(!lits.isEmpty());
      if (lits.length() == 1) 
	{
	  lits.popTo(selectedLits);
	  allMaximalSelected = true;
	}
      else // lits.length() > 1
	{
	  LitListStatistics newStat;
	  newStat.collectFor(lits);
	  _basicSelFun(lits,selectedLits,unselectedLits,allMaximalSelected,newStat); 
	};
    }
  else // all literals are splitting branches, select the youngest among them  
    {
      LitListElement* youngestBranch = lits.last();
      ASSERT(youngestBranch);
      ASSERT(youngestBranch->literal()->isSplittingBranch());
      DLC_INIT(lits.length() - 1);
      for (LitListElement* el = youngestBranch->previous();
	   el;
	   el = youngestBranch->previous())
	{
	  DLC_INC;
	  ASSERT(el->literal()->isSplittingBranch());
	  if (el->literal()->header().functor() > 
	      youngestBranch->literal()->header().functor())
	    {
	      lits.popTo(unselectedLits);
	      youngestBranch = el;
	    }
	  else
	    {
	      lits.passTo(el,unselectedLits);
	    };

	};
      ASSERT(lits.length() == 1UL);
      ASSERT(lits.last() == youngestBranch);
      lits.popTo(selectedLits);
    };
}; // void LiteralSelection::selectWithNonblockingSplitting(..)


inline 
void 
LiteralSelection::selectWithBlockingSplitting(LitList& lits,LitList& selectedLits,
					      LitList& unselectedLits,
					      bool& allMaximalSelected,
					      LitListStatistics& stat) const
{
  CALL("selectWithBlockingSplitting(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  allMaximalSelected = false;
  LitListElement* el;
  if (stat.numOfNegSplittingBranches()) // a negative branch exists, select the youngest among the negative branches
    {
      LitListElement* youngestNegBranch = lits.last();
      while ((!youngestNegBranch->literal()->isSplittingBranch()) ||
	     (!youngestNegBranch->literal()->isNegativeForSelection()))
	{
	  lits.popTo(unselectedLits);
	  youngestNegBranch = lits.last();
	  ASSERT(youngestNegBranch);
	};
      DLC_INIT(lits.length() - 1);
      for (el = youngestNegBranch->previous();
	   el;
	   el = youngestNegBranch->previous())
	{
	  DLC_INC;
	  if (el->literal()->isSplittingBranch() &&
	      el->literal()->isNegativeForSelection())
	    {
	      if (el->literal()->header().functor() > 
		  youngestNegBranch->literal()->header().functor())
		{
		  lits.popTo(unselectedLits);
		  youngestNegBranch = el; 
		}
	      else
		{
		  lits.passTo(el,unselectedLits);
		};

	    }
	  else
	    lits.passTo(el,unselectedLits);
	}; // for (el = youngestNegBranch->previous();...
      ASSERT(lits.length() == 1UL);
      ASSERT(lits.last() == youngestNegBranch);
      lits.popTo(selectedLits);
      // that's it
    }
  else // no negative splitting branches 
    {
      if (lits.length() > stat.numOfPosSplittingBranches()) 
	{	      	  
	  // a nonsplitting literal exists, all splitting branches are unselected
	  DLC_INIT(lits.length());
	  for (el = lits.first(); el; /* nothing here */)
	    {
	      DLC_INC;
	      if (el->literal()->header().IsSplitting())
		{
		  LitListElement* tmp = el->next();
		  lits.passTo(el,unselectedLits);
		  el = tmp;
		}
	      else
		el = el->next();
	    };

	  ASSERT(!lits.isEmpty());
	  if (lits.length() == 1) 
	    {
	      lits.popTo(selectedLits);
	      allMaximalSelected = true;
	    }
	  else // lits.length() > 1
	    { 
              LitListStatistics newStat;
	      newStat.collectFor(lits);	  
	      _basicSelFun(lits,selectedLits,unselectedLits,allMaximalSelected,newStat); 
	    };

	}
      else // lits contains only positive splitting branches, select the youngest among them
	{

	  LitListElement* youngestBranch = lits.last();
	  ASSERT(youngestBranch);
	  ASSERT(youngestBranch->literal()->isSplittingBranch());
	  ASSERT(youngestBranch->literal()->isPositive());
	  DLC_INIT(lits.length() - 1);
	  for (LitListElement* el = youngestBranch->previous();
	       el;
	       el = youngestBranch->previous())
	    {
	      DLC_INC;
	      ASSERT(el->literal()->isSplittingBranch());
	      ASSERT(el->literal()->isPositive());
	      if (el->literal()->header().functor() > 
		  youngestBranch->literal()->header().functor())
		{
		  lits.popTo(unselectedLits);
		  youngestBranch = el;
		}
	      else
		{
		  lits.passTo(el,unselectedLits);
		};

	    };
	  ASSERT(lits.length() == 1UL);
	  ASSERT(lits.last() == youngestBranch);
	  lits.popTo(selectedLits);
	  allMaximalSelected = true;
	};
    };
};  // void LiteralSelection::selectWithBlockingSplitting(..)




inline 
bool 
LiteralSelection::satisfactorySelection1(LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("satisfactorySelection1(LitList& selectedLits,LitList& unselectedLits) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(selectedLits));
  if (selectedLits.length() > 1UL) return false;
  ASSERT(selectedLits.length() == 1UL); 
  ulong litSize = selectedLits.first()->literal()->size();
  long priority = 
    selectedLits.first()->literal()->properties().selectabilityPriority();
  for (const LitListElement* el = unselectedLits.last();
       el;
       el = el->previous())
    {
      if ((!el->literal()->isAnswerLiteral()) &&
	  el->literal()->isNegativeForSelection() && 
	  (el->literal()->properties().selectabilityPriority() > priority ||
	   (el->literal()->size() > litSize &&
	    el->literal()->properties().selectabilityPriority() >= priority)))
	return false;  
    };
  return true;
}; // bool LiteralSelection::satisfactorySelection1(LitList& selectedLits,LitList& unselectedLits)


inline 
bool 
LiteralSelection::satisfactorySelection2(LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("satisfactorySelection2(LitList& selectedLits,LitList& unselectedLits) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(selectedLits));
  if (selectedLits.length() > 1) return false;
  ASSERT(selectedLits.length() == 1UL); 
  LitListElement* lit = selectedLits.first();
  ulong litSize = lit->literal()->size();
  long priority = lit->literal()->properties().selectabilityPriority();
  if (lit->literal()->isPositiveForSelection())
    {
      if (lit->literal()->isEquality() &&
	  lit->literal()->properties().useEqForParamodulationFrom()) 
	return false;
      // maxLit is a positive and is not equality or not available for paramod. from
      for (const LitListElement* el = unselectedLits.last();  
	   el;
	   el = el->previous())
	{
	  if (!el->literal()->isAnswerLiteral() &&
	      el->literal()->isNegativeForSelection())
	    {
	      if (el->literal()->isEquality() || 
		  el->literal()->properties().selectabilityPriority() > priority ||
		  (el->literal()->size() < litSize &&
		   el->literal()->properties().selectabilityPriority() >= priority)) 
		return false;
	    };
	};
      return true;
    }
  else // !lit->literal()->isPositiveForSelection()
    return false;

}; // bool LiteralSelection::satisfactorySelection2(LitList& selectedLits,LitList& unselectedLits)



inline 
bool 
LiteralSelection::satisfactorySelection3(LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("satisfactorySelection3(LitList& selectedLits,LitList& unselectedLits) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(selectedLits));
  if (selectedLits.length() > 1UL) return false;
  ASSERT(selectedLits.length() == 1UL); 
  LitListElement* lit = selectedLits.first();
  long priority = lit->literal()->properties().selectabilityPriority();
  if (lit->literal()->isPositiveForSelection())
    {
      if (lit->literal()->isEquality() &&
	  lit->literal()->properties().useEqForParamodulationFrom()) 
	return false;
      // lit is a positive and is not equality or not available for paramod. from
      ulong numOfDiffVars = lit->literal()->numOfDifferentVariables();
      static TemporaryLiteral::VarDepthProfile maxVarDepthProfile;
      lit->literal()->collectVarDepthProfile(maxVarDepthProfile);
      ulong litSize = lit->literal()->size();

      DLC_INIT(unselectedLits.length());
      for (LitListElement* el = unselectedLits.last();
	   el;
	   el = el->previous())
	{
	  DLC_INC;
	  if (!el->literal()->isAnswerLiteral() && 
	      el->literal()->isNegativeForSelection() &&
	      el->literal()->properties().selectabilityPriority() >= priority)
	    {
	      if (el->literal()->numOfDifferentVariables() > numOfDiffVars) return false;
	      if (el->literal()->properties().selectabilityPriority() > priority) return false;
	      if (el->literal()->numOfDifferentVariables() == numOfDiffVars)
		{
		  // compare the depths of variable occurences
		  static TemporaryLiteral::VarDepthProfile varDepthProfile;
		  el->literal()->collectVarDepthProfile(varDepthProfile);
		  switch (maxVarDepthProfile.compareByInverselyInducedOrder(varDepthProfile))
		    {
		    case Less: return false;
		    case Equal: 
		      // compare the sizes
		      if (litSize < el->literal()->size()) return false;
		      break;
#ifdef DEBUG_NAMESPACE
		    case Greater: break; 
		    default: ICP("ICP0"); break;
#else 
#ifdef _SUPPRESS_WARNINGS_
		    default: break;
#endif 
#endif         
		    };
		};
	    };
	};
    }
  else // !lit->literal()->isPositiveForSelection()
    return false;
  return true;
}; // bool LiteralSelection::satisfactorySelection3(LitList& selectedLits,LitList& unselectedLits)



inline 
void 
LiteralSelection::selectNegativeWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("selectNegativeWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const");

  // lits contains at least 1 negative non-answer literal
  // find first such literal
  LitListElement* cand = lits.last();
  while (cand->literal()->isPositiveForSelection() || 
	 cand->literal()->isAnswerLiteral())
    { 
      lits.popTo(unselectedLits);
      cand = lits.last();
    };

  // now find a better one
  ulong candSize = cand->literal()->size();
  long candPriority = cand->literal()->properties().selectabilityPriority();
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = cand->previous();
       el;
       el = cand->previous())
    {
      DLC_INC;
      if (el->literal()->isNegativeForSelection() && 
	  !el->literal()->isAnswerLiteral() && 
	  (el->literal()->properties().selectabilityPriority() > candPriority ||
	   (el->literal()->size() > candSize && 
	    el->literal()->properties().selectabilityPriority() == candPriority))) 
	{
	  cand = el;
	  candPriority = cand->literal()->properties().selectabilityPriority();
	  lits.popTo(unselectedLits);
	}
      else
	lits.passTo(el,unselectedLits);
    };
  ASSERT(lits.length() == 1UL);
  lits.popTo(selectedLits);
}; // void LiteralSelection::selectNegativeWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits)



inline 
void 
LiteralSelection::selectNegativeLeastProlific(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("selectNegativeLeastProlific(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const");

  // lits contains at least 1 negative non-answer literal
  // find first such literal
  LitListElement* cand = lits.last();
  while (cand->literal()->isPositiveForSelection() || 
	 cand->literal()->isAnswerLiteral())
    { 
      lits.popTo(unselectedLits);
      cand = lits.last();
    };

  long candPriority = cand->literal()->properties().selectabilityPriority();


  // now find a better one
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = cand->previous();
       el;
       el = cand->previous())
    {
      DLC_INC;
      if (el->literal()->isNegativeForSelection() && 
	  !el->literal()->isAnswerLiteral() && 
	  (el->literal()->properties().selectabilityPriority() > candPriority ||
	   (el->literal()->preferredNegative(*(cand->literal())) &&
	    el->literal()->properties().selectabilityPriority() == candPriority)))
	{
	  cand = el;
	  candPriority = cand->literal()->properties().selectabilityPriority();
	  lits.popTo(unselectedLits);
	}
      else
	lits.passTo(el,unselectedLits);
    };
  ASSERT(lits.length() == 1UL);
  lits.popTo(selectedLits);
}; // void LiteralSelection::selectNegativeLeastProlific(LitList& lits,LitList& selectedLits,LitList& unselectedLits)  



inline 
void 
LiteralSelection::selectNegativeMaximallyNongroundWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const
{
  CALL("selectNegativeMaximallyNongroundWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits) const");
  
  // lits may contain answer literals! 
  // First, get rid of all positive and answer literals.
  LitListElement* el;
  DLC_INIT(lits.length());
  for (el = lits.first();
       el;
       /* nothing here */)
    {
      DLC_INC;
      if (el->literal()->isPositiveForSelection() ||
	  el->literal()->isAnswerLiteral())
	{	
	  LitListElement* tmp = el->next();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	el = el->next();
    };

  // lits contains at least 1 negative literal
  ASSERT(!lits.isEmpty());
  unselectLitsWithLowerPriority(lits,unselectedLits);
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };

  // get rid of the literals that do not have the maximal number of different variables

  ulong maxNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
  
  DLC_RESETL(lits.length());
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
      if (numOfDiffVars > maxNumOfDiffVars)
	{
	  do 
	    {
	      lits.popTo(unselectedLits);
	    }
	  while (lits.last() != el);
	  el = el->previous();
	}
      else
	if (numOfDiffVars < maxNumOfDiffVars)
	  {
	    LitListElement* tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();
    }; // for (el = lits.last()->previous();...


  // now all the literals have the maximal number of different variables 
  if (lits.length() == 1) { lits.popTo(selectedLits); return; }; 

  // additionally, compare the multisets of depths of variable occurences    
 
  static TemporaryLiteral::VarDepthProfile varDepthProfile1;
  static TemporaryLiteral::VarDepthProfile varDepthProfile2;
  TemporaryLiteral::VarDepthProfile* maxVarDepthProfile = &varDepthProfile1;
  TemporaryLiteral::VarDepthProfile* litVarDepthProfile = &varDepthProfile2; 
  lits.last()->literal()->collectVarDepthProfile(*maxVarDepthProfile);
  LitListElement* tmp;
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      el->literal()->collectVarDepthProfile(*litVarDepthProfile); 
      // compare the profiles
      switch (maxVarDepthProfile->compareByInverselyInducedOrder(*litVarDepthProfile))
	{
	case Less:
	  maxVarDepthProfile = litVarDepthProfile;
	  litVarDepthProfile = 
	    ((litVarDepthProfile == &varDepthProfile1) ? 
	     &varDepthProfile2 
	     : 
	     &varDepthProfile1);
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);
	  el = el->previous();
	  break;   

	case Greater:
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	  break;

	case Equal:
	  el = el->previous();
	  break;
     
#ifdef DEBUG_NAMESPACE
	default : ICP("ICP0"); return;
#else
#ifdef _SUPPRESS_WARNINGS_
	default: RuntimeError::report("Illegal control point in  LiteralSelection::selectNegativeMaximallyNongroundWithMaximalWeight(..)");
	  return;
#endif
#endif 
	};
    };

  if (lits.length() == 1) { lits.popTo(selectedLits); return; }; 

  // select literals with greater size

  long maxSize = lits.last()->literal()->size();
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      long litSize = el->literal()->size();
      if (litSize > maxSize)
	{
	  maxSize = litSize;
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);     
	  el = el->previous();
	}
      else 
	if (litSize < maxSize)
	  {
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();
    };
  // now all the literals in lits are of the greatest size
 
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
  
  // now select on "don't care" basis
  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits);
}; // void LiteralSelection::selectNegativeMaximallyNongroundWithMaximalWeight(LitList& lits,LitList& selectedLits,LitList& unselectedLits)



inline 
void
LiteralSelection::unselectAnswerLiterals(LitList& lits,
					 LitList& selectedLits,
					 LitList& unselectedLits,
					 ulong numOfAnswerLits) const
{
  CALL("unselectAnswerLiterals(..)");
  LitListElement* el = lits.first();
  while (numOfAnswerLits)
    {
      ASSERT(el);
      if (el->literal()->isAnswerLiteral())
	{
	  LitListElement* tmp = el->next();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
          --numOfAnswerLits;
	}
      else
	el = el->next();
    };
}; // void LiteralSelection::unselectAnswerLiterals(..)


inline 
void 
LiteralSelection::unselectNonselectable(LitList& lits,
					LitList& unselectedLits,
					bool& allRemainingShouldBeSelected)
{
  CALL("unselectNonselectable(..)");
  if (lits.isEmpty()) { allRemainingShouldBeSelected = true; return; };

  LitListElement* el = lits.last();
  while (el &&
	 (el->literal()->isAnswerLiteral() || 
	  el->literal()->properties().selectionLayer() != 0UL))
    {
      el = el->previous();
      lits.popTo(unselectedLits);
    };

  if (lits.length() <= 1UL) 
    {
      allRemainingShouldBeSelected = true; 
      return; 
    };

  ASSERT(el);
  LiteralProperties maxSelectability = el->literal()->properties();

  el = el->previous();
  ASSERT(el);
  do
    {
      if (el->literal()->isAnswerLiteral() || 
	  el->literal()->properties().selectionLayer() != 0UL)
	{
	  LitListElement* tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	{
	  Comparison cmp = 
	    compareSelectability(el->literal()->properties(),maxSelectability);
	  if (cmp == Greater)
	    {
	      maxSelectability = el->literal()->properties();
	      do
		{
		  lits.popTo(unselectedLits);
		}
	      while (lits.last() != el);
	      el = el->previous();
	    }
	  else
	    if (cmp == Less)
	      {
		LitListElement* tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      {
		ASSERT(cmp == Equal);
		el = el->previous();
	      };
	};
    }
  while (el);

  allRemainingShouldBeSelected = 
    lits.length() <= 1UL ||
    maxSelectability.selectabilityCategory() == LiteralProperties::HighSelectability; 
}; // void LiteralSelection::unselectNonselectable(..) 


inline
BK::Comparison
LiteralSelection::compareSelectability(const LiteralProperties& prop1,
				       const LiteralProperties& prop2)
{
  CALL("compareSelectability(..)");
  LiteralProperties::SelectabilityCategory cat1 =
    prop1.selectabilityCategory();
  LiteralProperties::SelectabilityCategory cat2 =
    prop2.selectabilityCategory();  

  if ((cat1 == LiteralProperties::DefaultSelectability) &&
      (cat2 == LiteralProperties::DefaultSelectability))
    return Equal;

  if (cat1 == LiteralProperties::DefaultSelectability)
    {
      ASSERT(cat2 != LiteralProperties::DefaultSelectability);
      if (cat2 == LiteralProperties::LowSelectability)
	return Greater;
      ASSERT(cat2 == LiteralProperties::HighSelectability);
      return Less;
    }
  else if (cat1 == LiteralProperties::LowSelectability)
    {
      if (cat2 != LiteralProperties::LowSelectability)
	return Less;
      return BK::compare(prop1.selectabilityPriority(),
			 prop2.selectabilityPriority());
    }
  else
    {
      ASSERT(cat1 == LiteralProperties::HighSelectability);
      if (cat2 != LiteralProperties::HighSelectability)
	return Greater;
      return BK::compare(prop1.selectabilityPriority(),
			 prop2.selectabilityPriority());
    };
}; // BK::Comparison LiteralSelection::compareSelectability(..)



inline
bool 
LiteralSelection::unselectLitsWithLowerPriority(LitList& lits,
						LitList& unselectedLits)
{
  CALL("unselectLitsWithLowerPriority(..)");
  if (lits.length() <= 1UL) return false;
  ulong originalLength = lits.length();
  LitListElement* el = lits.last();
  long maxPriority = el->literal()->properties().selectabilityPriority();
  el = el->previous();
  ASSERT(el);
  do
    {
      long pri = el->literal()->properties().selectabilityPriority();
      if (pri == maxPriority)
	{
	  el = el->previous();
	}
      else if (pri < maxPriority)
	{
	  LitListElement* tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else // pri > maxPriority
	{
	  maxPriority = pri;
	  ASSERT(lits.last() != el);
	  do
	    {
	      lits.popTo(unselectedLits);
	    }
	  while (lits.last() != el);
	  el = el->previous();
	};

    }
  while (el);
  ASSERT(!lits.isEmpty());
  return lits.length() != originalLength; // something unselected
}; // bool LiteralSelection::unselectLitsWithLowerPriority(..)
  


inline
bool 
LiteralSelection::unselectPosLitsWithLowerPriority(LitList& lits,
						   LitList& unselectedLits)
{
  CALL("unselectPosLitsWithLowerPriority(..)");
  LitListElement* el = lits.last();
  while (el && el->literal()->isNegativeForSelection())
    el = el->previous();  
  if (!el || !el->previous()) return false;
  
  ulong originalLength = lits.length();
  long maxPriority = el->literal()->properties().selectabilityPriority();
  el = el->previous();
  ASSERT(el);
  do
    {
      if (el->literal()->isNegativeForSelection())
	{
	  el = el->previous();
	}
      else
	{
	  long pri = el->literal()->properties().selectabilityPriority();
	  if (pri == maxPriority)
	    {
	      el = el->previous();
	    }
	  else if (pri < maxPriority)
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else // pri > maxPriority
	    {
	      maxPriority = pri;
	      ASSERT(lits.last() != el);
	      for (LitListElement* el1 = lits.last(); el1 != el; /* nothing here */)
		{
		  if (el1->literal()->isPositiveForSelection())
		    {
		      LitListElement* tmp = el1->previous();
		      lits.passTo(el1,unselectedLits);
		      el1 = tmp;
		    }
		  else
		    el1 = el1->previous();
		};
	      el = el->previous();
	    };
	};
    }
  while (el);
  ASSERT(!lits.isEmpty());
  return lits.length() != originalLength; // something unselected
}; // bool LiteralSelection::unselectPosLitsWithLowerPriority(..)
  



inline
bool 
LiteralSelection::unselectNegLitsWithLowerPriority(LitList& lits,
						   LitList& unselectedLits)
{
  CALL("unselectNegLitsWithLowerPriority(..)");
  LitListElement* el = lits.last();
  while (el && el->literal()->isPositiveForSelection())
    el = el->previous();  
  if (!el || !el->previous()) return false;

  ulong originalLength = lits.length();
  long maxPriority = el->literal()->properties().selectabilityPriority();
  el = el->previous();
  ASSERT(el);
  do
    {
      if (el->literal()->isPositiveForSelection())
	{
	  el = el->previous();
	}
      else
	{
	  long pri = el->literal()->properties().selectabilityPriority();
	  if (pri == maxPriority)
	    {
	      el = el->previous();
	    }
	  else if (pri < maxPriority)
	    {
	      LitListElement* tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else // pri > maxPriority
	    {
	      maxPriority = pri;
	      ASSERT(lits.last() != el);
	      for (LitListElement* el1 = lits.last(); el1 != el; /* nothing here */)
		{
		  if (el1->literal()->isNegativeForSelection())
		    {
		      LitListElement* tmp = el1->previous();
		      lits.passTo(el1,unselectedLits);
		      el1 = tmp;
		    }
		  else
		    el1 = el1->previous();
		};
	      el = el->previous();
	    };
	};
    }
  while (el);
  ASSERT(!lits.isEmpty());
  return lits.length() != originalLength; // something unselected
}; // bool LiteralSelection::unselectNegLitsWithLowerPriority(..)
  

inline 
void 
LiteralSelection::mark(LitList& selectedLits,LitList& unselectedLits)
{
  CALL("mark(LitList& selectedLits,LitList& unselectedLits)");
  LitListElement* el;
  for (el = selectedLits.first();
       el;
       el = el->next())
    {
      el->literal()->markAsSelected();
    };
  for (el = unselectedLits.first();
       el;
       el = el->next())
    {
      el->literal()->markAsUnselected();
    };	 
}; // void LiteralSelection::mark(LitList& selectedLits,LitList& unselectedLits)



#ifdef VKERNEL_FOR_STEP_RESOLUTION

inline
bool 
LiteralSelection::isNextStepLiteral(const TemporaryLiteral& lit)
{
  return lit.arity() && 
    (!lit.isEquality()) && 
    lit.flatterm()->next()->arity();
};


inline
bool 
LiteralSelection::isCurrentStepLiteral(const TemporaryLiteral& lit)
{
  return lit.arity() && 
    (!lit.isEquality()) && 
    (!lit.flatterm()->next()->arity());
};

void 
inline
LiteralSelection::unselectCurrentStateLiterals(LitList& lits,
					       LitList& selectedLits,
					       LitList& unselectedLits) const
{
  CALL("unselectCurrentStateLiterals(LitList& lits,...) const");
  
  LitListElement* el = lits.first();
  while (el)
    {
      if (isCurrentStepLiteral(*(el->literal())))
	{	  
	  LitListElement* tmp = el->next();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	el = el->next();
    };
  
}; // LiteralSelection::unselectCurrentStateLiterals(LitList& lits,...) const
#endif








void 
LiteralSelection::function1(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");

  /*
    Selects all maximal literals.
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  
  DLC_DECL(0UL);

  allMaximalSelected = true; 
  LitListElement* cand;
  TERM max_header; 

  if (_currentObject->_literalComparisonMode == 3L)
    {
      // temporary patch
      ((LiteralSelection*)_currentObject)->_maximalSelection.select(lits,
								    selectedLits,
								    unselectedLits,
								    allMaximalSelected,
								    stat);
      return;
    };

  if (_currentObject->_literalComparisonMode == 0L) goto compare_atoms;

  // filter out literals with nonmaximal headers
  max_header = lits.last()->literal()->header();
  LitListElement* el;
  LitListElement* tmp;
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      if ((_currentObject->_literalComparisonMode == 2L) 
	  && (el->literal()->header().functor() == max_header.functor()))
	{     
	  el = el->previous();
	}
      else
	{
	  switch (Signature::current()->compareHeaderPrecedences(el->literal()->header().polarity(),
								 el->literal()->header().functor(),
								 max_header.polarity(),
								 max_header.functor())) 
	    {
	    case Greater :
	      max_header = el->literal()->header();
	      do 
		{ 
		  lits.popTo(unselectedLits); 
		} 
	      while (lits.last() != el);
	      el = el->previous();
	      break;
 
	    case Less :
	      tmp = el->previous();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	      break;

	    case Equal :
	      ASSERT((el->literal()->header() == max_header) || 
		     (el->literal()->header().IsEquality() &&  
		      max_header.IsEquality() && 
		      (el->literal()->header().polarity() == max_header.polarity())));
	      el = el->previous();
	      break;

	    case Incomparable : 
	      ICP("E0"); 
	      el = el->previous();
	      break; 
	    };
	};
    };
 
  // all remaining literals have the maximal header
  // or _currentObject->_literalComparisonMode == 0L     

 compare_atoms:
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };

  cand = lits.last();
  ASSERT((_currentObject->_literalComparisonMode != 1L) 
	 || (cand->literal()->header() == max_header) 
	 || (cand->literal()->header().IsEquality() 
	     &&  max_header.IsEquality() 
	     && (cand->literal()->header().polarity() == max_header.polarity()))); 
  ASSERT((_currentObject->_literalComparisonMode != 2L) 
         || (cand->literal()->header().functor() == max_header.functor())
         || (cand->literal()->header().IsEquality() 
	     &&  max_header.IsEquality())); 

  
  el = cand->previous();
  DLC_RESETL(lits.length() - 1);
  while (el)
    {
      DLC_INC;
      ASSERT((_currentObject->_literalComparisonMode != 1L) 
	     || (el->literal()->header() == max_header) 
	     || (el->literal()->header().IsEquality() 
		 &&  max_header.IsEquality() 
		 && (el->literal()->header().polarity() == max_header.polarity()))); 
      ASSERT((_currentObject->_literalComparisonMode != 2L) 
	     || (cand->literal()->header().functor() == max_header.functor())
	     || (cand->literal()->header().IsEquality() 
		 &&  max_header.IsEquality())); 
      
      switch (el->literal()->compare(*(cand->literal())))
	{
	case Greater:
	  lits.popTo(unselectedLits);
	  goto compare_atoms;
	case Less:
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	  break; 
	case Incomparable: 
	  el = el->previous();
	  break;

#ifdef DEBUG_NAMESPACE
    
	case Equal:
	  ICP("ICP0"); break;
	default: ICP("ICP1"); break;
#else  
#ifdef _SUPPRESS_WARNINGS_
        default: break;
#endif
#endif
	};
      
    };  
  ASSERT(lits.last() == cand);
  lits.popTo(selectedLits);
  if (lits.isEmpty()) return;
  goto compare_atoms;
  ASSERT(!selectedLits.isEmpty());
}; // void LiteralSelection::function1(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function2(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function2(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
 
  /* 
     If lits contains negative literals, one of them is selected.
     Negative literals with greater weights are prefered.
     If there is no negative literal, we proceed with selection 1.  
  */
 
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));

  if (stat.numOfNegNonanswerLits())
    {
      allMaximalSelected = false;
      _currentObject->selectNegativeWithMaximalWeight(lits,selectedLits,unselectedLits);
    }
  else // no negative literals
    { 
      function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
    };  
}; // void LiteralSelection::function2(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)

void 
LiteralSelection::function3(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function3(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");

  /*
    If lits contains negative literals, one of them is selected.
    To choose between negative literals we heuristically estimate the number of possible
    inferences and a literal promising less inferences is selected.
    If there is no negative literal, we proceed with selection 1. 
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));

  if (stat.numOfNegNonanswerLits())
    {
      allMaximalSelected = false;
      _currentObject->selectNegativeLeastProlific(lits,selectedLits,unselectedLits);
    }
  else // no negative literals
    { 
      function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
    };    
}; // void LiteralSelection::function3(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)

void 
LiteralSelection::function4(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function4(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
 
  /* 
     We first select maximal literals. Then we try to improve selection 
     by selecting a negative non-answer literal. This is based on the assumption that 
     selecting one literal is better than selecting several literals and 
     selecting a negative literal with greater weight is better than selecting 
     a maximal literal.
  */ 
 

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));

  function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat); 
  if (stat.numOfNegNonanswerLits()) // quality of selection can be questioned 
    {
      if (!_currentObject->satisfactorySelection1(selectedLits,unselectedLits))
	{
	  lits.collectAllFrom(unselectedLits);
	  lits.collectAllFrom(selectedLits);
	  _currentObject->selectNegativeWithMaximalWeight(lits,selectedLits,unselectedLits);
	  allMaximalSelected = false;  
	};
    };
}; // void LiteralSelection::function4(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function5(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function5(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /* 
     We first select maximal literals. Then we try to improve selection by selecting
     a negative literal. This is based on the assumption that selecting one literal
     is better than selecting several literals and selecting a negative literal 
     with less possible inferences (heuristicaly estimated) is better than 
     selecting a maximal literal.
  */ 
 
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));


  function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
  if (stat.numOfNegNonanswerLits()) // quality of selection can be questioned 
    {
      if (!_currentObject->satisfactorySelection2(selectedLits,unselectedLits))
	{
	  lits.collectAllFrom(unselectedLits);
	  lits.collectAllFrom(selectedLits);
	  _currentObject->selectNegativeLeastProlific(lits,selectedLits,unselectedLits);
	  allMaximalSelected = false;
	};
    };
}; // void LiteralSelection::function5(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)



void 
LiteralSelection::function6(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function6(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
 
  /* 
     If lits contains negative literals, one of them is selected.
     Among the negative literals we always select a literal 
     with the maximal number of different variables. Among such literals 
     we prefer those having smaller depths of variable occurences.
     Negative literals with greater weights are prefered.
     If there is no negative literal, we proceed with selection 1.  
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));

  if (stat.numOfNegNonanswerLits())
    {
      allMaximalSelected = false;
      _currentObject->selectNegativeMaximallyNongroundWithMaximalWeight(lits,selectedLits,unselectedLits);
    }
  else // no negative literals
    { 
      function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
    };  
}; // void LiteralSelection::function6(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function7(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function7(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
 
  /*  
      We first select maximal literals. Then we try to improve selection 
      by selecting a negative literal. This is based on the assumption that 
      selecting one literal is better than selecting several literals and 
      selecting a negative literal with more different variables is 
      better than selecting 
      a maximal literal. If we decide to use negative selection, among 
      the negative literals
      we chose those having more different variable. Among these 
      literals we chose 
      those with smaller depths of variable occurences and then 
      select those having greater size.   
  */ 
 
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));

  function1(lits,selectedLits,unselectedLits,allMaximalSelected,stat); 
  if (stat.numOfNegNonanswerLits()) // quality of selection can be questioned 
    {
      if (!_currentObject->satisfactorySelection3(selectedLits,unselectedLits))
	{
	  lits.collectAllFrom(unselectedLits);
	  lits.collectAllFrom(selectedLits);
	  _currentObject->selectNegativeMaximallyNongroundWithMaximalWeight(lits,selectedLits,unselectedLits);
	  allMaximalSelected = false;  
	};
    };
}; // void LiteralSelection::function7(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)



void 
LiteralSelection::function1002(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1002(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Arbitrary selection (almost always makes everything incomplete). 
    If lits contains a literal which is not a positive equality available for paramod. from,
    we never select such an equality. 
    Among the literals available for selection we always select a literal 
    with the maximal number of different variables. Among such literals we prefer those having 
    smaller depths of variable occurences. Among such literals we select those with maximal size.
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  allMaximalSelected = false;

  if (unselectLitsWithLowerPriority(lits,unselectedLits))
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  bool selectPositiveEq = 
    (!stat.numOfNegNonanswerLits()) && 
    (!stat.numOfNoneqNonanswerLits());

  if (!selectPositiveEq)
    {
      while (lits.last()->literal()->isEquality() && 
	     lits.last()->literal()->isPositive() &&
	     lits.last()->literal()->properties().useEqForParamodulationFrom())
	lits.popTo(unselectedLits);
    };
  
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
 
  ulong maxNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();

  LitListElement* el;
  LitListElement* tmp;
  DLC_INIT(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      if (!selectPositiveEq && 
	  el->literal()->isEquality() && 
	  el->literal()->isPositive() &&
	  el->literal()->properties().useEqForParamodulationFrom())
	{
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;   
	}
      else
	{
	  // compare the number of different variables 
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars > maxNumOfDiffVars)
	    {
	      maxNumOfDiffVars = numOfDiffVars;
	      do 
		{ 
		  lits.popTo(unselectedLits); 
		} 
	      while (lits.last() != el);     
	      el = el->previous();
	    }
	  else 
	    if (numOfDiffVars < maxNumOfDiffVars)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp; 
	      }
	    else
	      el = el->previous();
	};
    };
  // now all the literals have the maximal number of different variables 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };

  // additionally, compare the multisets of depths of variable occurences    
 
  static TemporaryLiteral::VarDepthProfile varDepthProfile1;
  static TemporaryLiteral::VarDepthProfile varDepthProfile2;
  TemporaryLiteral::VarDepthProfile* maxVarDepthProfile = &varDepthProfile1;
  TemporaryLiteral::VarDepthProfile* litVarDepthProfile = &varDepthProfile2; 
  lits.last()->literal()->collectVarDepthProfile(*maxVarDepthProfile);
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      ASSERT(selectPositiveEq || 
	     !el->literal()->isEquality() || 
	     el->literal()->isNegative() ||
	     !el->literal()->properties().useEqForParamodulationFrom());
      el->literal()->collectVarDepthProfile(*litVarDepthProfile); 
      // compare the profiles
      switch (maxVarDepthProfile->compareByInverselyInducedOrder(*litVarDepthProfile))
	{
	case Less:
	  maxVarDepthProfile = litVarDepthProfile;
	  litVarDepthProfile = 
	    ((litVarDepthProfile == &varDepthProfile1) ? 
	     &varDepthProfile2 
	     : 
	     &varDepthProfile1);
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);  
	  el = el->previous();
	  break;   

	case Greater:
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	  break;

	case Equal: 
	  el = el->previous();
	  break;
     
#ifdef DEBUG_NAMESPACE
	default : ICP("ICP0"); return;
#else
#ifdef _SUPPRESS_WARNINGS_
	default: RuntimeError::report("Illegal control point in  LiteralSelection::function1002(..)");
	  return;
#endif
#endif 
	};
    };


  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };

  // select literals with greater size 
  long maxSize = lits.last()->literal()->size();
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      ASSERT(selectPositiveEq || 
	     !el->literal()->isEquality() || 
	     el->literal()->isNegative() ||
	     !el->literal()->properties().useEqForParamodulationFrom());

      long litSize = el->literal()->size();
      if (litSize > maxSize)
	{
	  maxSize = litSize;
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);   
	  el = el->previous();
	}
      else 
	if (litSize < maxSize)
	  {
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();
    };
  // now all the literals in lits are of the greatest size 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
  
  // now select on "don't care" basis
  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits);
}; // void LiteralSelection::function1002(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function1003(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1003(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Arbitrary selection (almost always makes everything incomplete). 
    If lits contains a literal which is not a positive equality available for paramod. from,
    we never select a positive equality. 
    Among the literals available for selection we always select a literal 
    of the maximal size. Among several such literals we select by the function 1002.        
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  

  allMaximalSelected = false;


  if (unselectLitsWithLowerPriority(lits,unselectedLits))
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  bool selectPositiveEq = 
    (!stat.numOfNegNonanswerLits()) && 
    (!stat.numOfNoneqNonanswerLits());


  if (!selectPositiveEq)
    {
      while (lits.last()->literal()->isEquality() && 
	     lits.last()->literal()->isPositive() &&
	     lits.last()->literal()->properties().useEqForParamodulationFrom())
	lits.popTo(unselectedLits);
    };

  LitListElement* tmp;
  long maxSize = lits.last()->literal()->size();
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      if (!selectPositiveEq && 
	  el->literal()->isEquality() && 
	  el->literal()->isPositive() &&
	  el->literal()->properties().useEqForParamodulationFrom())
	{
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	{
	  long litSize = el->literal()->size();
	  if (litSize > maxSize)
	    {
	      maxSize = litSize;
	      do { lits.popTo(unselectedLits); } while (lits.last() != el);   
	      el = el->previous();
	    }
	  else 
	    if (litSize < maxSize)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      el = el->previous();
	};
    };
  // now all the literals in lits are of the greatest size 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
  LitListStatistics newStat;
  newStat.collectFor(lits);
 
  function1002(lits,selectedLits,unselectedLits,allMaximalSelected,newStat);
  allMaximalSelected = false;
}; // void LiteralSelection::function1003(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)




void 
LiteralSelection::function1004(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1004(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Arbitrary selection (almost always makes everything incomplete). 
    If lits contains a literal which is not a positive equality avail. for paramod. from,
    we never select such an equality. 
    Among the literals available for selection we always select a literal 
    with the minimal number of different variables.
    Among such literals we select those having minimal total number of variable
    occurences. Among such literals we select one of the greatest size. 
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  allMaximalSelected = false;
 
  if (unselectLitsWithLowerPriority(lits,unselectedLits))
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  bool selectPositiveEq = 
    (!stat.numOfNegNonanswerLits()) && 
    (!stat.numOfNoneqNonanswerLits());

  if (!selectPositiveEq)
    {
      while (lits.last()->literal()->isEquality() && 
	     lits.last()->literal()->isPositive() &&
	     lits.last()->literal()->properties().useEqForParamodulationFrom())
	lits.popTo(unselectedLits);
    };
 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 
  ulong minNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
  LitListElement* el;
  LitListElement* tmp;
  DLC_INIT(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {      
      DLC_INC;
      ASSERT(el->next());
      if (!selectPositiveEq && 
	  el->literal()->isEquality() && 
	  el->literal()->isPositive() &&
	  el->literal()->properties().useEqForParamodulationFrom())
	{
	  tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	{
	  // compare the number of different variables 
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars < minNumOfDiffVars)
	    {
	      minNumOfDiffVars = numOfDiffVars;
	      do { lits.popTo(unselectedLits); } while (lits.last() != el);   
	      el = el->previous();
	    }
	  else 
	    if (numOfDiffVars > minNumOfDiffVars)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      el = el->previous();
	};
    };
  // now all the literals have the minimal number of different variables 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };

  // compare the total number of variable occurences  
  TermWeightType  minNumOfVarOccurences =  
    lits.last()->literal()->numberOfVariableOccurences();
   
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {      
      DLC_INC;
      ASSERT(el->next());
      ASSERT(selectPositiveEq || 
	     !el->literal()->isEquality() ||
	     el->literal()->isNegative() ||
	     !el->literal()->properties().useEqForParamodulationFrom());

      // compare the number of variable occurences 
      TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
      if (numOfVarOccurences < minNumOfVarOccurences)
	{
	  minNumOfVarOccurences = numOfVarOccurences;
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);
	  el = el->previous();
	}
      else 
	if (numOfVarOccurences > minNumOfVarOccurences)
	  {
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();
    };
  // now all the literals have the minimal number of variable occurences 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 
  // now select literals with maximal size
  long maxSize = lits.last()->literal()->size();
  
  DLC_RESETL(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      ASSERT(selectPositiveEq || 
	     !el->literal()->isEquality() || 
	     el->literal()->isNegative() ||
	     !el->literal()->properties().useEqForParamodulationFrom());
      long litSize = el->literal()->size();
      if (litSize > maxSize)
	{
	  maxSize = litSize;
	  do { lits.popTo(unselectedLits); } while (lits.last() != el);  
	  el = el->previous();
	}
      else 
	if (litSize < maxSize)
	  {
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();
    };
  // now all the literals in lits are of the greatest size 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 
  // select on "don't care" basis 

  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits);
}; // void LiteralSelection::function1004(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)



void 
LiteralSelection::function1005(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1005(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Arbitrary selection (almost always makes everything incomplete). 
    If lits contains a literal which is not a positive equality avail. for paramod. from,
    we never select a positive equality. 
    Among the literals available for selection we always select
    either a "maximally ground" negative literal, or a "maximally nonground" positive literal. 
    "Maximally ground" means having minimal total number of variable occurences 
    among literals having minimal number of different variables. "Maximally non ground" 
    means having maximal total number of variable occurences among the literals 
    having maximal number of different variables. 
    Among such literals we select one with the maximal size.
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits));
  allMaximalSelected = false; 
 

  if (unselectLitsWithLowerPriority(lits,unselectedLits))
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  bool selectPositiveEq = 
    (!stat.numOfNegNonanswerLits()) && 
    (!stat.numOfNoneqNonanswerLits());

  if (!selectPositiveEq)
    {
      while (lits.last()->literal()->isEquality() && 
	     lits.last()->literal()->isPositive() &&
	     lits.last()->literal()->properties().useEqForParamodulationFrom())
	{
	  lits.popTo(unselectedLits);
	};
    };

  ASSERT(!lits.isEmpty());

  if (lits.length() == 1) { lits.popTo(selectedLits); return; };


  ulong minNegLitNumOfDiffVars;
  TermWeightType minNegLitNumOfVarOccurences;
  ulong maxPosLitNumOfDiffVars;
  TermWeightType maxPosLitNumOfVarOccurences;
  if (lits.last()->literal()->isNegativeForSelection()) 
    {
      minNegLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
      minNegLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxPosLitNumOfDiffVars = 0;
      maxPosLitNumOfVarOccurences = 0;
    }
  else
    {
      minNegLitNumOfDiffVars = VampireKernelConst::MaxNumOfVariables;
      minNegLitNumOfVarOccurences = VampireKernelConst::MaxTermSize;
      maxPosLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables(); 
      maxPosLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
    }; 



  LitListElement* el;
  LitListElement* tmp;
  DLC_INIT(lits.length() - 1);
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {     
      DLC_INC;
      ASSERT(el->next());
      if (!selectPositiveEq && 
	  el->literal()->isEquality() && 
	  el->literal()->isPositive() &&
	  el->literal()->properties().useEqForParamodulationFrom())
	{
          tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	{
	  if (el->literal()->isNegativeForSelection())
	    {
	      ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	      if (numOfDiffVars < minNegLitNumOfDiffVars)
		{
		  minNegLitNumOfDiffVars = numOfDiffVars;
		  minNegLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
         
		  LitListElement* anotherEl = lits.last();
		  while (anotherEl != el)
		    {
		      if (anotherEl->literal()->isNegativeForSelection())
			{
			  ASSERT(anotherEl->literal()->numOfDifferentVariables() > numOfDiffVars);
			  tmp = anotherEl->previous();
			  lits.passTo(anotherEl,unselectedLits);
			  anotherEl = tmp;
			}
		      else
			anotherEl = anotherEl->previous();
		    };
		  el = el->previous();
		}
	      else
		if (numOfDiffVars > minNegLitNumOfDiffVars)
		  {
		    tmp = el->previous();
		    lits.passTo(el,unselectedLits);
		    el = tmp;
		  }
		else // numOfDiffVars == minNegLitNumOfDiffVars 
		  {
		    TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		    if (numOfVarOccurences < minNegLitNumOfVarOccurences)
		      {
			minNegLitNumOfVarOccurences = numOfVarOccurences; 
			LitListElement* anotherEl = lits.last();
			while (anotherEl != el)
			  {
			    if (anotherEl->literal()->isNegativeForSelection())
			      {
				ASSERT(anotherEl->literal()->numOfDifferentVariables() == minNegLitNumOfDiffVars);
				ASSERT(anotherEl->literal()->numberOfVariableOccurences() > numOfVarOccurences);
				tmp = anotherEl->previous();
				lits.passTo(anotherEl,unselectedLits);
				anotherEl = tmp;
			      }
			    else
			      anotherEl = anotherEl->previous();
			  };
			
			el = el->previous();
		      }
		    else
		      if (numOfVarOccurences > minNegLitNumOfVarOccurences)
			{
			  tmp = el->previous();
			  lits.passTo(el,unselectedLits);
			  el = tmp;
			}
		      else
			el = el->previous();
		  };
	    }
	  else // !el->literal()->isNegativeForSelection()
	    {
	      ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	      if (numOfDiffVars > maxPosLitNumOfDiffVars)
		{
		  maxPosLitNumOfDiffVars = numOfDiffVars;
		  maxPosLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();

		  LitListElement* anotherEl = lits.last();
		  while (anotherEl != el)
		    {
		      if (!anotherEl->literal()->isNegativeForSelection())
			{
			  ASSERT(anotherEl->literal()->numOfDifferentVariables() < numOfDiffVars);
			  tmp = anotherEl->previous();
			  lits.passTo(anotherEl,unselectedLits);
			  anotherEl = tmp;
			}
		      else
			anotherEl = anotherEl->previous();
		      
		    };
		  el = el->previous();
		}
	      else
		if (numOfDiffVars < maxPosLitNumOfDiffVars)
		  {
		    tmp = el->previous();
		    lits.passTo(el,unselectedLits);
		    el = tmp;
		  }
		else // numOfDiffVars == maxPosLitNumOfDiffVars 
		  {
		    TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		    if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
		      {
			maxPosLitNumOfVarOccurences = numOfVarOccurences; 

			LitListElement* anotherEl = lits.last();
			while (anotherEl != el)
			  {
			    if (!anotherEl->literal()->isNegativeForSelection())
			      {
				ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxPosLitNumOfDiffVars);
				ASSERT(anotherEl->literal()->numberOfVariableOccurences() < numOfVarOccurences);
				tmp = anotherEl->previous();
				lits.passTo(anotherEl,unselectedLits);
				anotherEl = tmp;
			      }
			    else
			      anotherEl = anotherEl->previous();
			  };
			el = el->previous();
		      }
		    else
		      if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
			{
			  tmp = el->previous();
			  lits.passTo(el,unselectedLits);
			  el = tmp;
			}
		      else
			el = el->previous();
		  };
	    };


	};
    };
  // now all negative literals are maximally ground
  // and all positive literals are maximally nonground
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; }; 


  // now select literals with maximal size
  long maxSize = lits.last()->literal()->size();
  DLC_RESETL(lits.length());
  for (el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(selectPositiveEq || 
	     !el->literal()->isEquality() || 
	     el->literal()->isNegative() ||
	     !el->literal()->properties().useEqForParamodulationFrom());
      long litSize = el->literal()->size();
      if (litSize > maxSize)
	{
	  maxSize = litSize;
	  do { lits.popTo(unselectedLits); } while (lits.last() != el); 
	  el = el->previous();    
	}
      else 
	if (litSize < maxSize)
	  {         
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	  }
	else
	  el = el->previous();  
    };
  // now all the literals in lits are of the greatest size 
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 
  // select on "don't care" basis 



  lits.popTo(selectedLits);
  lits.passAllTo(unselectedLits);
}; // void LiteralSelection::function1005(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat) 


void 
LiteralSelection::function1006(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1006(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Bidirectional selection. 
    In a rule, i.e. a clause containing both 
    positive and negative literals we always select one positive and one negative
    literal. If we have several literals of the same polarity, we select 
    a bigger of the maximally nonground ones. In a fact, i.e. a clause with all literals 
    of the same polarity, we select a bigger maximally nonground literal.     
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits)); 
  
  allMaximalSelected = false;

  bool posUnselected = unselectPosLitsWithLowerPriority(lits,unselectedLits);
  bool negUnselected = unselectNegLitsWithLowerPriority(lits,unselectedLits);

  if (posUnselected || negUnselected)
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  ulong maxNegLitNumOfDiffVars;
  TermWeightType maxNegLitNumOfVarOccurences;
  ulong maxNegLitSize;
  ulong maxPosLitNumOfDiffVars;
  TermWeightType maxPosLitNumOfVarOccurences;
  ulong maxPosLitSize;
  if (lits.last()->literal()->isNegativeForSelection()) 
    {
      maxNegLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
      maxNegLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxNegLitSize = lits.last()->literal()->size(); 
      maxPosLitNumOfDiffVars = 0;
      maxPosLitNumOfVarOccurences = 0;
      maxPosLitSize = 0;
    }
  else
    {
      maxNegLitNumOfDiffVars = 0;
      maxNegLitNumOfVarOccurences = 0; 
      maxNegLitSize = 0;
      maxPosLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables(); 
      maxPosLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxPosLitSize = lits.last()->literal()->size(); 
    }; 
  LitListElement* tmp;

  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      if (el->literal()->isNegativeForSelection())
	{
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars > maxNegLitNumOfDiffVars)
	    {
	      maxNegLitNumOfDiffVars = numOfDiffVars;
	      maxNegLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxNegLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (anotherEl->literal()->isNegativeForSelection())
		    {
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() < numOfDiffVars);
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars < maxNegLitNumOfDiffVars)
	      { 
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == maxNegLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences > maxNegLitNumOfVarOccurences)
		  {
		    maxNegLitSize = el->literal()->size();
		    maxNegLitNumOfVarOccurences = numOfVarOccurences; 

		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (anotherEl->literal()->isNegativeForSelection())
			  {
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxNegLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() < numOfVarOccurences);
			    
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		    };

		    el = el->previous();
		  }
		else
		  if (numOfVarOccurences < maxNegLitNumOfVarOccurences)
		    {
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence ==  maxNegLitNumOfVarOccurences
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxNegLitSize)
			{
			  maxNegLitSize = litSize;


			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (anotherEl->literal()->isNegativeForSelection())
				{
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxNegLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == maxNegLitNumOfVarOccurences);
				  
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };
			  el = el->previous();
			}            
		      else
			if (litSize < maxNegLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();
		    };
	      };
	}
      else // !el->literal()->isNegativeForSelection()
	{
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars > maxPosLitNumOfDiffVars)
	    {
	      maxPosLitNumOfDiffVars = numOfDiffVars;
	      maxPosLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxPosLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (!anotherEl->literal()->isNegativeForSelection())
		    {		      
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() < numOfDiffVars);
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars < maxPosLitNumOfDiffVars)
	      {          
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == maxPosLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
		  {
		    maxPosLitSize = el->literal()->size();
		    maxPosLitNumOfVarOccurences = numOfVarOccurences; 

		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (!anotherEl->literal()->isNegativeForSelection())
			  {
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxPosLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() < numOfVarOccurences);
			    
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		      };
		    el = el->previous();
		  }
		else
		  if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
		    {
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence == maxPosLitNumOfVarOccurences 
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxPosLitSize)
			{
			  maxPosLitSize = litSize;
			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (!anotherEl->literal()->isNegativeForSelection())
				{				  
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxPosLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == maxPosLitNumOfVarOccurences);
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };
			  el = el->previous();
			}            
		      else
			if (litSize < maxPosLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();
		    };
	      };
	};
    }; // for (..
  // all negative literals are maximally nonground and have the same size
  // all positive literals are maximally nonground and have the same size
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
  bool positiveHasBeenSelected = false;
  bool negativeHasBeenSelected = false;
  while (!lits.isEmpty())
    {
      if (lits.last()->literal()->isNegativeForSelection())
	{
	  if (negativeHasBeenSelected) 
	    {
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      negativeHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	}
      else 
	{
	  if (positiveHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      positiveHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	};
    };

}; // void LiteralSelection::function1006(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function1007(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1007(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Bidirectional selection. 
    In a rule, i.e. a clause containing both 
    positive and negative literals we always select one positive and one negative
    literal. If we have several literals of the same polarity, we select 
    those with maximal size. In a fact, i.e. a clause with all literals 
    of the same polarity, we select one of the biggest literals.     
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits)); 
  
  allMaximalSelected = false;

  bool posUnselected = unselectPosLitsWithLowerPriority(lits,unselectedLits);
  bool negUnselected = unselectNegLitsWithLowerPriority(lits,unselectedLits);

  if (posUnselected || negUnselected)
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  ulong maxNegLitSize;
  ulong maxPosLitSize;
  if (lits.last()->literal()->isNegativeForSelection()) 
    {
      maxNegLitSize = lits.last()->literal()->size(); 
      maxPosLitSize = 0;
    }
  else
    {
      maxNegLitSize = 0;
      maxPosLitSize = lits.last()->literal()->size(); 
    }; 
  LitListElement* tmp;
  
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {      
      DLC_INC;
      ASSERT(el->next());
      if (el->literal()->isNegativeForSelection())
	{
	  ulong litSize = el->literal()->size();
	  if (litSize > maxNegLitSize)
	    {
	      maxNegLitSize = litSize;

	      
	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (anotherEl->literal()->isNegativeForSelection())
		    {
		      ASSERT(anotherEl->literal()->size() < litSize);	
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }            
	  else
	    if (litSize < maxNegLitSize)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      el = el->previous();
	}
      else // !el->literal()->isNegativeForSelection()
	{
	  ulong litSize = el->literal()->size();
	  if (litSize > maxPosLitSize)
	    {
	      maxPosLitSize = litSize;


	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (!anotherEl->literal()->isNegativeForSelection())
		    {
		      ASSERT(anotherEl->literal()->size() < litSize);
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }            
	  else
	    if (litSize < maxPosLitSize)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      el = el->previous();
	};
    }; // for (..
  // all negative literals have maximal size
  // all positive literals have maximal size
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 

  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
  bool positiveHasBeenSelected = false;
  bool negativeHasBeenSelected = false;
  while (!lits.isEmpty())
    {
      if (lits.last()->literal()->isNegativeForSelection())
	{
	  if (negativeHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      negativeHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	}
      else 
	{
	  if (positiveHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      positiveHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	};
    };

}; // void LiteralSelection::function1007(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function1008(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1008(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Bidirectional selection. 
    In a rule, i.e. a clause containing both 
    positive and negative literals we always select one positive and one negative
    literal. If we have several literals of the same polarity, we select 
    those that are maximally ground. In a fact, i.e. a clause with all literals 
    of the same polarity, we select one maximally ground literal.     
  */


  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits)); 

  allMaximalSelected = false;


  bool posUnselected = unselectPosLitsWithLowerPriority(lits,unselectedLits);
  bool negUnselected = unselectNegLitsWithLowerPriority(lits,unselectedLits);

  if (posUnselected || negUnselected)
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  ulong minNegLitNumOfDiffVars;
  TermWeightType minNegLitNumOfVarOccurences;
  ulong maxNegLitSize;
  ulong minPosLitNumOfDiffVars;
  TermWeightType minPosLitNumOfVarOccurences;
  ulong maxPosLitSize;
  if (lits.last()->literal()->isNegativeForSelection()) 
    {
      minNegLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
      minNegLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxNegLitSize = lits.last()->literal()->size(); 
      minPosLitNumOfDiffVars = VampireKernelConst::MaxNumOfVariables;
      minPosLitNumOfVarOccurences = VampireKernelConst::MaxTermSize;
      maxPosLitSize = 0;
    }
  else
    {
      minNegLitNumOfDiffVars = VampireKernelConst::MaxNumOfVariables;
      minNegLitNumOfVarOccurences = VampireKernelConst::MaxTermSize; 
      maxNegLitSize = 0;
      minPosLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables(); 
      minPosLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxPosLitSize = lits.last()->literal()->size(); 
    }; 

  LitListElement* tmp;
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      if (el->literal()->isNegativeForSelection())
	{
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars < minNegLitNumOfDiffVars)
	    {
	      minNegLitNumOfDiffVars = numOfDiffVars;
	      minNegLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxNegLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (anotherEl->literal()->isNegativeForSelection())
		    {		      
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() > numOfDiffVars);
		       tmp = anotherEl->previous();
		       lits.passTo(anotherEl,unselectedLits);
		       anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars > minNegLitNumOfDiffVars)
	      { 
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == minNegLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences < minNegLitNumOfVarOccurences)
		  {
		    maxNegLitSize = el->literal()->size();
		    minNegLitNumOfVarOccurences = numOfVarOccurences; 


		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (anotherEl->literal()->isNegativeForSelection())
			  {
			    
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == minNegLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() > numOfVarOccurences);
			    
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		      };
		    el = el->previous();
		  }
		else
		  if (numOfVarOccurences > minNegLitNumOfVarOccurences)
		    {
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence ==  minNegLitNumOfVarOccurences
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxNegLitSize)
			{
			  maxNegLitSize = litSize;

			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (anotherEl->literal()->isNegativeForSelection())
				{
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == minNegLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == minNegLitNumOfVarOccurences);
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };	
			  el = el->previous();		  
			}            
		      else
			if (litSize < maxNegLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();	
		    };
	      };
	}
      else // !el->literal()->isNegativeForSelection()
	{

	  COP("COP300");

	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars < minPosLitNumOfDiffVars)
	    {
	      minPosLitNumOfDiffVars = numOfDiffVars;
	      minPosLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxPosLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (!anotherEl->literal()->isNegativeForSelection())
		    {
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() > numOfDiffVars)
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars > minPosLitNumOfDiffVars)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == minPosLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences < minPosLitNumOfVarOccurences)
		  {
		    minPosLitNumOfVarOccurences = numOfVarOccurences; 
		    maxPosLitSize = el->literal()->size();

		    
		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (!anotherEl->literal()->isNegativeForSelection())
			  {
			    
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == minPosLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() > numOfVarOccurences);
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		      };
		    el = el->previous();

		  }
		else
		  if (numOfVarOccurences > minPosLitNumOfVarOccurences)
		    {
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence == minPosLitNumOfVarOccurences 
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxPosLitSize)
			{
			  maxPosLitSize = litSize;


			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (!anotherEl->literal()->isNegativeForSelection())
				{
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == minPosLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == minPosLitNumOfVarOccurences);
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };
			  el = el->previous();
			}            
		      else
			if (litSize < maxPosLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();
		    };
	      };
	};
    }; // for (..
  // all negative literals are maximally ground and have the same size
  // all positive literals are maximally ground and have the same size
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 

  COP("COP500");


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
  bool positiveHasBeenSelected = false;
  bool negativeHasBeenSelected = false;
  while (!lits.isEmpty())
    {
      if (lits.last()->literal()->isNegativeForSelection())
	{
	  if (negativeHasBeenSelected) 
	    {
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      negativeHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	}
      else 
	{
	  if (positiveHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      positiveHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	};
    };

}; // void LiteralSelection::function1008(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


void 
LiteralSelection::function1009(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("function1009(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)");
  /*
    Bidirectional selection. 
    In a rule, i.e. a clause containing both 
    positive and negative literals we always select one positive and one negative
    literal. If we have several positive literals, we select 
    those that are maximally nonground. If we have several negative literals, 
    we select those that are maximally ground. 
    In a positive fact, i.e. a clause with only positive literals, 
    we select one maximally nonground literal. 
    In a negative fact, i.e. a clause with only negative literals, 
    we select one maximally ground literal.           
  */

  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  ASSERT(!TempLitListExtras::containsSplittingBranches(lits)); 

  allMaximalSelected = false;


  bool posUnselected = unselectPosLitsWithLowerPriority(lits,unselectedLits);
  bool negUnselected = unselectNegLitsWithLowerPriority(lits,unselectedLits);

  if (posUnselected || negUnselected)
    {
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1UL) { lits.popTo(selectedLits); return; };
      stat.collectFor(lits);
    };

  ulong minNegLitNumOfDiffVars;
  TermWeightType minNegLitNumOfVarOccurences;
  ulong maxNegLitSize;
  ulong maxPosLitNumOfDiffVars;
  TermWeightType maxPosLitNumOfVarOccurences;
  ulong maxPosLitSize;
  if (lits.last()->literal()->isNegativeForSelection()) 
    {
      minNegLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables();
      minNegLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxNegLitSize = lits.last()->literal()->size(); 
      maxPosLitNumOfDiffVars = 0;
      maxPosLitNumOfVarOccurences = 0;
      maxPosLitSize = 0;
    }
  else
    {
      minNegLitNumOfDiffVars = VampireKernelConst::MaxNumOfVariables;
      minNegLitNumOfVarOccurences = VampireKernelConst::MaxTermSize; 
      maxNegLitSize = 0;
      maxPosLitNumOfDiffVars = lits.last()->literal()->numOfDifferentVariables(); 
      maxPosLitNumOfVarOccurences = lits.last()->literal()->numberOfVariableOccurences(); 
      maxPosLitSize = lits.last()->literal()->size(); 
    }; 

  LitListElement* tmp;
  
  DLC_INIT(lits.length() - 1);
  for (LitListElement* el = lits.last()->previous();
       el;
       /* nothing here */)
    {
      DLC_INC;
      ASSERT(el->next());
      if (el->literal()->isNegativeForSelection())
	{
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars < minNegLitNumOfDiffVars)
	    {
	      minNegLitNumOfDiffVars = numOfDiffVars;
	      minNegLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxNegLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (anotherEl->literal()->isNegativeForSelection())
		    {
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() > numOfDiffVars);
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars > minNegLitNumOfDiffVars)
	      { 
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == minNegLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences < minNegLitNumOfVarOccurences)
		  {
		    minNegLitNumOfVarOccurences = numOfVarOccurences; 
		    maxNegLitSize = el->literal()->size();

		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (anotherEl->literal()->isNegativeForSelection())
			  {
			    
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == minNegLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() > numOfVarOccurences);
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		      };
		    el = el->previous();
		  }
		else
		  if (numOfVarOccurences > minNegLitNumOfVarOccurences)
		    {
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence ==  minNegLitNumOfVarOccurences
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxNegLitSize)
			{
			  maxNegLitSize = litSize;

			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (anotherEl->literal()->isNegativeForSelection())
				{
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == minNegLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == minNegLitNumOfVarOccurences);
				  
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };
			  el = el->previous();
			}            
		      else
			if (litSize < maxNegLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();
			  
		    };
	      };
	}
      else // !el->literal()->isNegativeForSelection()
	{
	  ulong numOfDiffVars = el->literal()->numOfDifferentVariables();
	  if (numOfDiffVars > maxPosLitNumOfDiffVars)
	    {
	      maxPosLitNumOfDiffVars = numOfDiffVars;
	      maxPosLitNumOfVarOccurences = el->literal()->numberOfVariableOccurences();
	      maxPosLitSize = el->literal()->size();

	      LitListElement* anotherEl = lits.last();
	      while (anotherEl != el)
		{
		  if (!anotherEl->literal()->isNegativeForSelection())
		    {		      
		      ASSERT(anotherEl->literal()->numOfDifferentVariables() < numOfDiffVars);
		      tmp = anotherEl->previous();
		      lits.passTo(anotherEl,unselectedLits);
		      anotherEl = tmp;
		    }
		  else
		    anotherEl = anotherEl->previous();
		};
	      el = el->previous();
	    }
	  else
	    if (numOfDiffVars < maxPosLitNumOfDiffVars)
	      {
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else // numOfDiffVars == maxPosLitNumOfDiffVars 
	      {
		TermWeightType numOfVarOccurences = el->literal()->numberOfVariableOccurences();
		if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
		  {
		    maxPosLitNumOfVarOccurences = numOfVarOccurences; 
		    maxPosLitSize = el->literal()->size();

		    LitListElement* anotherEl = lits.last();
		    while (anotherEl != el)
		      {
			if (!anotherEl->literal()->isNegativeForSelection())
			  {
			    ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxPosLitNumOfDiffVars);
			    ASSERT(anotherEl->literal()->numberOfVariableOccurences() < numOfVarOccurences);
			    tmp = anotherEl->previous();
			    lits.passTo(anotherEl,unselectedLits);
			    anotherEl = tmp;
			  }
			else
			  anotherEl = anotherEl->previous();
		      };
		    el = el->previous();
		  }
		else
		  if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
		    {          
		      tmp = el->previous();
		      lits.passTo(el,unselectedLits);
		      el = tmp;
		    }
		  else // numOfVarOccurence == maxPosLitNumOfVarOccurences 
		    {
		      ulong litSize = el->literal()->size();
		      if (litSize > maxPosLitSize)
			{
			  maxPosLitSize = litSize;

			  LitListElement* anotherEl = lits.last();
			  while (anotherEl != el)
			    {
			      if (!anotherEl->literal()->isNegativeForSelection())
				{
				  
				  ASSERT(anotherEl->literal()->numOfDifferentVariables() == maxPosLitNumOfDiffVars);
				  ASSERT(anotherEl->literal()->numberOfVariableOccurences() == maxPosLitNumOfVarOccurences);
				  tmp = anotherEl->previous();
				  lits.passTo(anotherEl,unselectedLits);
				  anotherEl = tmp;
				}
			      else
				anotherEl = anotherEl->previous();
			    };
			  el = el->previous();
			}            
		      else
			if (litSize < maxPosLitSize)
			  {
			    tmp = el->previous();
			    lits.passTo(el,unselectedLits);
			    el = tmp;
			  }
			else
			  el = el->previous();
		    };
	      };
	};
    }; // for (..
  // all negative literals are maximally nonground and have the same size
  // all positive literals are maximally nonground and have the same size
  ASSERT(!lits.isEmpty());
  if (lits.length() == 1) { lits.popTo(selectedLits); return; };
 

  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
  bool positiveHasBeenSelected = false;
  bool negativeHasBeenSelected = false;
  while (!lits.isEmpty())
    {
      if (lits.last()->literal()->isNegativeForSelection())
	{
	  if (negativeHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      negativeHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	}
      else 
	{
	  if (positiveHasBeenSelected) 
	    { 
	      lits.popTo(unselectedLits);
	    }
	  else 
	    {
	      positiveHasBeenSelected = true;
	      lits.popTo(selectedLits);
	    };
	};
    };

}; // void LiteralSelection::function1009(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)





void 
LiteralSelection::customBasicSelFun(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)
{ 
  CALL("customBasicSelFun(..)");
  _customSelection->select(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
}; // void LiteralSelection::customBasicSelFun(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected,LitListStatistics& stat)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection::ComplexLiteralCharacteristic1"
#endif 
#include "debugMacros.hpp"
//=================================================

LiteralSelection::ComplexLiteralCharacteristic1::~ComplexLiteralCharacteristic1()
{
};


void LiteralSelection::ComplexLiteralCharacteristic1::computeFor(TemporaryLiteral& lit)
{
  CALL("computeFor(TemporaryLiteral& lit)");
  _positive = lit.isPositiveForSelection();
  _priority = lit.properties().selectabilityPriority();
  _numOfDifferentVars = lit.numOfDifferentVariables();
  _numOfVarOccurences = lit.numberOfVariableOccurences();
}; 

Comparison LiteralSelection::ComplexLiteralCharacteristic1::compareWith(CustomLiteralSelection::ComplexLiteralCharacteristic* c)
{
  CALL("compareWith(ComplexLiteralCharacteristic* c)");
  if (_positive != (static_cast<ComplexLiteralCharacteristic1*>(c))->_positive)
    return Incomparable;
  
  if (_priority != (static_cast<ComplexLiteralCharacteristic1*>(c))->_priority)
    if (_priority > (static_cast<ComplexLiteralCharacteristic1*>(c))->_priority)
      {
	return Greater;
      }
    else
      return Less;
    

  if (_positive)
    {
      if (_numOfDifferentVars > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Greater;
      if (_numOfDifferentVars < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Less;
      
      if (_numOfVarOccurences > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Greater;
      if (_numOfVarOccurences < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Less;
    }
  else // !_positive
    {      
      if (_numOfDifferentVars > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Less;
      if (_numOfDifferentVars < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Greater;
      
      if (_numOfVarOccurences > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Less;
      if (_numOfVarOccurences < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Greater;
    };
      
  return Equal;
}; // Comparison LiteralSelection::ComplexLiteralCharacteristic1::compareWith(ComplexLiteralCharacteristic* c)

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection::ComplexLiteralOrdering1"
#endif 
#include "debugMacros.hpp"
//=================================================



LiteralSelection::ComplexLiteralOrdering1::~ComplexLiteralOrdering1()
{
};

Comparison LiteralSelection::ComplexLiteralOrdering1::compare(TemporaryLiteral& lit1,
							      TemporaryLiteral& lit2) const
{
  CALL("compare(..) const"); 
  long pri1 = lit1.properties().selectabilityPriority();
  long pri2 = lit2.properties().selectabilityPriority();
  if (pri1 != pri2)
    if (pri1 > pri2)
      {
	return Greater;
      }
    else
      return Less;
      
  if (lit1.isEquality())
    {
      if (lit2.isEquality())
	{
	  // both are equalities
	  if (lit1.isNegative())
	    {
	      if (lit2.isNegative())
		{
		  // both are negative equalities
		  if (lit1.size() == 3UL) // exceptional case
		    return Greater;
		  if (lit2.size() == 3UL) // exceptional case
		    return Less;
		  
		  float resDiff = 
		    lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
		    lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
		  if (resDiff > 0.0) return Greater;
		  if (resDiff < 0.0) return Less;
		  		  
		  if (lit1.numberOfVariableOccurences() >
		      lit2.numberOfVariableOccurences())
		    return Less;
		  
		  if (lit1.numberOfVariableOccurences() <
		      lit2.numberOfVariableOccurences())
		    return Greater;
		  
		  if (lit1.numOfDifferentVariables() >
		      lit2.numOfDifferentVariables())
		    return Less;
		  
		  if (lit1.numOfDifferentVariables() <
		      lit2.numOfDifferentVariables())
		    return Greater;

		  return Incomparable;
		}
	      else
		return Greater;
	    }
	  else
	    if (lit2.isNegative())
	      {
		return Less;
	      }
	    else // both are positive equalities 
	      {
		float resDiff = 
		  lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
		  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
		if (resDiff > 0.0) return Greater;
		if (resDiff < 0.0) return Less;
		if (lit1.numOfDifferentVariables() >
		    lit2.numOfDifferentVariables())
		  return Greater;
		
		if (lit1.numOfDifferentVariables() <
		    lit2.numOfDifferentVariables())
		  return Less;
		  
		return Incomparable;	
	      };

	}
      else // lit1 is eq, lit2 is not
	{
	  if (lit1.isNegative())
	    {
	      return Greater;
	    }
	  else
	    {
	      if (lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) > 
		  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff)) 
		return Greater;
	      return Less;
	    };
	};
    }
  else
    if (lit2.isEquality())
      {
	// lit1 is not an eq, lit2 is
	if (lit2.isNegative())
	  {
	    return Less;
	  }
	else
	  {
	    if (lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) < 
		lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff)) 
	      return Less;
	    return Greater;
	  };

      }
    else // neither is equality
      {
	float resDiff = 
	  lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
	  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
	if (resDiff > 0.0) return Greater;
	if (resDiff < 0.0) return Less;
	if (lit1.isPositiveForSelection())
	  {
	    if (lit2.isPositiveForSelection())
	      {
		// both are positive
		if (lit1.numOfDifferentVariables() >
		    lit2.numOfDifferentVariables())
		  return Greater;
		
		if (lit1.numOfDifferentVariables() <
		    lit2.numOfDifferentVariables())
		  return Less;		  
		return Incomparable;	
	      }
	    else 
	      return Greater;	      
	  }
	else
	  if (lit2.isPositiveForSelection())
	    {
	      return Less;
	    }
	  else // both are negative
 	    {
	      if (lit1.numOfDifferentVariables() >
		  lit2.numOfDifferentVariables())
		return Less;
	      
	      if (lit1.numOfDifferentVariables() <
		  lit2.numOfDifferentVariables())
		return Greater;

	      return Incomparable;
	    };	
      };
}; // Comparison LiteralSelection::ComplexLiteralOrdering1::compare(..)

Comparison LiteralSelection::ComplexLiteralOrdering1::compareLayers(const TemporaryLiteral& lit1,
								    const TemporaryLiteral& lit2) const
{
  CALL("compareLayers(..) const");
  long pri1 = lit1.properties().selectabilityPriority();
  long pri2 = lit2.properties().selectabilityPriority();
  if (pri1 != pri2)
    if (pri1 > pri2)
      {
	return Greater;
      }
    else
      return Less;
      
  if (lit1.isEquality() && lit1.isNegative())
    {
      if ((!lit2.isEquality()) || lit2.isPositive())
	return Greater;
    }
  else
    if (lit2.isEquality() && lit2.isNegative())
      return Less;
  return Equal;
}; // Comparison LiteralSelection::ComplexLiteralOrdering1::compareLayers(..) const



//=================================================


