//
// File:         LiteralSelection.hpp
// Description:  Defines various literal selection functions
// Created:      Dec 12, 2001
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         Was a part of NewClause.
// Revised:      Dec 15, 2001. Bug fix in setSplittingFlag.
//============================================================================
#ifndef LITERAL_SELECTION_H
//=============================================================================

#define LITERAL_SELECTION_H 
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "StandardLiteralOrdering.hpp"
#include "CustomLiteralSelection.hpp"
class SymbolInfo;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
class LiteralSelection
{
public:
  typedef TemporaryLiteral::Extras::Handle LitHandle;
  typedef BasicTempLitList LitList;
  typedef BasicTempLitList::Element LitListElement;
  typedef TempLitListExtras::Characteristics LitListStatistics;

 public:
  LiteralSelection();
  ~LiteralSelection() {};


  void setFunNum(long num);
  void setLiteralComparisonMode(long mode) { _literalComparisonMode = mode; };
  void setStandardLiteralOrdering(StandardLiteralOrdering* ord) 
  { 
    _standardLiteralOrdering = ord;
    _totalSelection.setStandardLiteralOrdering(ord);
    _maximalSelection.setStandardLiteralOrdering(ord);
    _hyperSelection.setStandardLiteralOrdering(ord);
    _maxWithCorrectionSelection.setStandardLiteralOrdering(ord);
    _arbitrarySelection.setStandardLiteralOrdering(ord);
    _completeBidirectSelection.setStandardLiteralOrdering(ord);
    _incompleteBidirectSelection.setStandardLiteralOrdering(ord);
    _bestOnlySelection.setStandardLiteralOrdering(ord);
  };

  void setSplittingFlag(bool fl) 
    { 
      _splittingIsEnabled = (_splittingIsEnabled || fl);
      //  _splittingIsEnabled can only be changed from false to true.
    };
  void setSplittingStyle(long splittingStyle) { _splittingStyle = splittingStyle; };


#ifdef VKERNEL_FOR_STEP_RESOLUTION
  void setStepResolutionLiteralSelection(bool fl)
  {
    _stepResolutionLiteralSelectionFlag = fl;    
  };
#endif



    void 
      select(LitList& lits,LitList& selectedLits,
	     LitList& unselectedLits,bool& allMaximalSelected) const;
    
    void 
      selectForSupport(LitList& lits,
		       LitList& selectedLits,
		       LitList& unselectedLits) const;
    
  bool 
    selectForElimination(LitList& lits,LitList& selectedLits,
			 LitList& unselectedLits,  
			 const SymbolInfo* eliminationTable) const;

private:

  // Literals of different polarities are incomparable.
  // In positive literals, the more different variables
  // and var. occurences, the bigger the characteristic value.
  // In negative literals the other way around.

  class ComplexLiteralCharacteristic1 : public CustomLiteralSelection::ComplexLiteralCharacteristic
  {
  public:
    ComplexLiteralCharacteristic1() {};
    ~ComplexLiteralCharacteristic1();
    void computeFor(TemporaryLiteral& lit);
    BK::Comparison compareWith(CustomLiteralSelection::ComplexLiteralCharacteristic* c);
  private:
    bool _positive;
    long _priority;
    ulong _numOfDifferentVars;
    ulong _numOfVarOccurences;
  }; // class ComplexLiteralCharacteristic1

  class ComplexLiteralOrdering1 : public LiteralOrdering
  {
  public:
    ComplexLiteralOrdering1(ulong resistanceDepthCoeff,
			    float resistanceConstCoeff,
			    float resistanceVarCoeff) :
      _resistanceDepthCoeff(resistanceDepthCoeff),
      _resistanceConstCoeff(resistanceConstCoeff),
      _resistanceVarCoeff(resistanceVarCoeff)
    {
      _isStratified = true;
    };
    ~ComplexLiteralOrdering1();
    BK::Comparison compare(TemporaryLiteral& lit1,
			   TemporaryLiteral& lit2) const;
    BK::Comparison compareLayers(const TemporaryLiteral& lit1,
				 const TemporaryLiteral& lit2) const;
  private:
    ComplexLiteralOrdering1() {};
  private:
    ulong _resistanceDepthCoeff;
    float _resistanceConstCoeff;
    float _resistanceVarCoeff;
  }; // class ComplexLiteralOrdering1


private:


  void 
    selectWithNonblockingSplitting(LitList& lits,LitList& selectedLits,
				   LitList& unselectedLits,bool& allMaximalSelected,
	                           LitListStatistics& stat) const;

  void 
    selectWithBlockingSplitting(LitList& lits,LitList& selectedLits,
				LitList& unselectedLits,bool& allMaximalSelected,
	                        LitListStatistics& stat) const;


  static 
    void 
    function1(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat);

  static 
    void 
    function2(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat);

  static 
    void 
    function3(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat);

  static 
    void 
    function4(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat); 

  static 
    void 
    function5(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat); 

  static 
    void 
    function6(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat);  
  
  static 
    void 
    function7(LitList& lits,LitList& selectedLits,
	      LitList& unselectedLits,bool& allMaximalSelected,
	      LitListStatistics& stat);


  static 
    void 
    function1002(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1003(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1004(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1005(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1006(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1007(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1008(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);

  static 
    void 
    function1009(LitList& lits,LitList& selectedLits,
		 LitList& unselectedLits,bool& allMaximalSelected,
		 LitListStatistics& stat);  


  static 
    void 
    customBasicSelFun(LitList& lits,LitList& selectedLits,
		      LitList& unselectedLits,bool& allMaximalSelected,
		      LitListStatistics& stat);  





  inline
    bool 
    satisfactorySelection1(LitList& selectedLits,LitList& unselectedLits) const;

  inline
    bool 
    satisfactorySelection2(LitList& selectedLits,LitList& unselectedLits) const;
  
  inline
    bool
    satisfactorySelection3(LitList& selectedLits,LitList& unselectedLits) const;
   

  inline
    void 
    selectNegativeWithMaximalWeight(LitList& lits,LitList& selectedLits,
				    LitList& unselectedLits) const;

  inline
    void 
    selectNegativeLeastProlific(LitList& lits,LitList& selectedLits,
				LitList& unselectedLits) const;  
  
  inline
    void 
    selectNegativeMaximallyNongroundWithMaximalWeight(LitList& lits,
						      LitList& selectedLits,
						      LitList& unselectedLits) const;

  inline 
    void 
    unselectAnswerLiterals(LitList& lits,
			   LitList& selectedLits,
			   LitList& unselectedLits,
                           ulong numOfAnswerLits) const;

  inline 
  static
  void 
  unselectNonselectable(LitList& lits,
			LitList& unselectedLits,
			bool& allRemainingShouldBeSelected);

  inline
  static
  BK::Comparison
  compareSelectability(const LiteralProperties& prop1,
		       const LiteralProperties& prop2);

  inline
  static
  bool 
  unselectLitsWithLowerPriority(LitList& lits,
				LitList& unselectedLits);
  
  inline
  static
  bool 
  unselectPosLitsWithLowerPriority(LitList& lits,
				   LitList& unselectedLits);
  inline
  static
  bool 
  unselectNegLitsWithLowerPriority(LitList& lits,
				   LitList& unselectedLits);
  


  inline 
  static
  void 
  mark(LitList& selectedLits,LitList& unselectedLits);

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  static 
  inline
  bool 
  isNextStepLiteral(const TemporaryLiteral& lit);

  static
  inline
  bool 
  isCurrentStepLiteral(const TemporaryLiteral& lit);

  void 
  inline
  unselectCurrentStateLiterals(LitList& lits,
			       LitList& selectedLits,
			       LitList& unselectedLits) const;
#endif



 private:
  static const LiteralSelection* _currentObject; 
  // To communicate the current object to the static selection functions.
  // They are made static to be represented by pointers.   
  void (*_basicSelFun)(LitList&,LitList&,LitList&,bool&,LitListStatistics&);
  long _literalComparisonMode;
  StandardLiteralOrdering* _standardLiteralOrdering;
  bool _splittingIsEnabled;
  long _splittingStyle;

  static CustomLiteralSelection* _customSelection; 


  // components for custom selection schemes
  CustomLiteralSelection::OneOfEachPolarityFilter _oneOfEachPolarityFilter;
  CustomLiteralSelection::MaxValFilter _allWithMaxPriorityFilter;
  CustomLiteralSelection::MaxValFilter _oneOfMaxSizeFilter;
  CustomLiteralSelection::MaxValFilter _allOfMaxSizeFilter;
  CustomLiteralSelection::MinValFilter _oneOfMinSizeFilter;
  CustomLiteralSelection::MinValFilter _allOfMinSizeFilter;
  CustomLiteralSelection::AllPOrAllFilter _allNegEqOrAllFilter;
  CustomLiteralSelection::AllPOrAllFilter _oneNegEqOrAnyoneFilter;
  CustomLiteralSelection::AllPOrAllFilter _allExceptPosEqOrAllFilter;
  CustomLiteralSelection::CompositionFilter _oneBigNegEqOrOneBigWithMaxPriorityFilter;
  CustomLiteralSelection::CompositionFilter _oneBigNegEqOrOneBigFilter;
  CustomLiteralSelection::CompositionFilter _oneSmallPreferablyNegEqFilter;
  CustomLiteralSelection::MaxValFilter _allWithMaxNumOfDiffVarsFilter;
  
  CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile1;
  CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile2;

  CustomLiteralSelection::MaxComplexValFilter _allWithMostShallowVarsFilter;
  CustomLiteralSelection::CompositionFilter _oneOfMaxSizeMaxNongroundWithMaxPriorityFilter;
  CustomLiteralSelection::CompositionFilter _oneOfMaxSizeWithMaxPriorityFilter;

  ComplexLiteralCharacteristic1 _complexCharacteristic1_1;
  ComplexLiteralCharacteristic1 _complexCharacteristic1_2;
  

  CustomLiteralSelection::CompositionFilter _complexFilter1;
  // Only literals with maximal priority are considered for selection. 
  // If possible, no positive equalities avail. for paramod from are selected.
  // From avail. literals we select max. nonground.
  // From remaining literals we select one of max. size.

  CustomLiteralSelection::CompositionFilter _complexFilter2;
  // Only literals with maximal priority are considered for selection.
  // If possible, no positive equalities avail. for paramod from are selected.
  // From avail. literals we select those of max. size.
  // From remaining literals we select one max. nonground.

  CustomLiteralSelection::MaxComplexValFilter _complexFilter3;
  // Select all maximal literals w.r.t. ComplexLiteralCharacteristic1.

  CustomLiteralSelection::CompositionFilter _complexFilter4;
  // Only literals with maximal priority are considered for selection.
  // First, try to get rid of positive equalities avail. for paramod from if
  // literals of other kinds exist. 
  // Then use _complexFilter3, then finish with _oneOfMaxSizeFilter.

  CustomLiteralSelection::MaxComplexValFilter _complexFilter5;
  // Same as _complexFilter3, but always selects 1 literal. 

  CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability1;
  CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability2;
  CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability3;

  ComplexLiteralOrdering1 _complexLiteralOrdering1_1;
  ComplexLiteralOrdering1 _complexLiteralOrdering1_2;
  ComplexLiteralOrdering1 _complexLiteralOrdering1_3;

  // custom selection schemes
  TotalLiteralSelection _totalSelection;
  MaximalLiteralSelection _maximalSelection;
  HyperLiteralSelection _hyperSelection;
  MaxWithCorrectionLiteralSelection _maxWithCorrectionSelection;
  ArbitraryLiteralSelection _arbitrarySelection;
  CompleteBidirectionalLiteralSelection _completeBidirectSelection;
  IncompleteBidirectionalLiteralSelection _incompleteBidirectSelection;
  BestOnlyLiteralSelection _bestOnlySelection;


#ifdef VKERNEL_FOR_STEP_RESOLUTION
  bool _stepResolutionLiteralSelectionFlag;
#endif
}; // class LiteralSelection

}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
inline 
void 
LiteralSelection::setFunNum(long num)
{
  switch (num) 
    {
    case 0L : // every literal is selected  
      _basicSelFun = customBasicSelFun;
      _customSelection = &_totalSelection;
      return;
    case 1L : 
      _basicSelFun = function1; 
      return;
    case 2L : 
      _basicSelFun = function2;     
      return;
    case 3L : 
      _basicSelFun = function3; 
      return;
    case 4L : 
      _basicSelFun = function4; 
      return;
    case 5L : 
      _basicSelFun = function5;
      return;
    case 6L : 
      _basicSelFun = function6; 
      return;
    case 7L : 
      _basicSelFun = function7; 
      return;

    case 8L : // simulates 1L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maximalSelection;
      return;

    case 9L : // simulates 2L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeWithMaxPriorityFilter);
      return;

    case 10L : // simulates 4L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigWithMaxPriorityFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;

    case 11L : // simulates 6L      
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundWithMaxPriorityFilter);
      return;

    case 12L : // simulates 7L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundWithMaxPriorityFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability2);
      return;

    case 13L : // experimental
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_1);
      return;


    case 14L : // complete analogue of 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _completeBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _completeBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeWithMaxPriorityFilter);
      return;

    case 15L : // complete analogue of 1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;

    case 16L : // similar to 13L, but constants have much higher resistance 
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_2);
      return;


    case 17L : // similar to 16L, but variables have very low (negative) resistance 
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_3);
      return;


    case 1002L :
      _basicSelFun = function1002; 
      return;
    case 1003L :
      _basicSelFun = function1003; 
      return;
    case 1004L :
      _basicSelFun = function1004; 
      return;
    case 1005L :
      _basicSelFun = function1005; 
      return;
    case 1006L :
      _basicSelFun = function1006; 
      return;
    case 1007L :
      _basicSelFun = function1007; 
      return;
    case 1008L :
      _basicSelFun = function1008; 
      return;    
    case 1009L :
      _basicSelFun = function1009; 
      return;          



    case 1010L : // simulates 1002L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter1);
      return;

    case 1011L : // simulates 1003L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter2);
      return;

    case 1012L : // simulates 1005L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;

    
    case 1013L : // simulates 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_oneOfMaxSizeWithMaxPriorityFilter);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeWithMaxPriorityFilter);
      return;

    case 1014L : // simulates 1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;




     

    case -2L : 
      _basicSelFun = function2;
      return;
    case -3L : 
      _basicSelFun = function3;
      return;
    case -4L : 
      _basicSelFun = function4;
      return;
    case -5L : 
      _basicSelFun = function5;
      return;
    case -6L : 
      _basicSelFun = function6;
      return;
    case -7L : 
      _basicSelFun = function7;
      return;


    case -9L : // simulates -2   
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeWithMaxPriorityFilter);
      return;


    case -10L : // simulates -4       
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigWithMaxPriorityFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;


    case -11L : // simulates -6L      
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundWithMaxPriorityFilter);
      return;

    case -12L : // simulates -7L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundWithMaxPriorityFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability2);
      return;

    case -13L : // experimental
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_1);
      return;

    case -14L : // complete analogue of 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _completeBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeWithMaxPriorityFilter);
      _completeBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeWithMaxPriorityFilter);
      return;

    case -15L : // complete analogue of -1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;


    case -16L : // similar to -13L, but constants have much higher resistance 
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_2);
      return;


    case -17L : // similar to -16L, but variables have very low (negative) resistance 
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_3);
      return;



    case -1005L :
      _basicSelFun = function1005; 
      return;
    case -1009L :
      _basicSelFun = function1009; 
      return;



    case -1012L : // simulates -1005L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;

    case -1014L : // simulates -1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;



    default: BK::RuntimeError::report("Wrong selection number.");
    };
}; // void LiteralSelection::setFunNum(long num)


}; // namespace VK


//======================================================================
#endif
