//
// File:         Master.cpp
// Description:  Main loop.
// Created:      Mar 5, 2000, 20:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.hpp"
#include "Clause.hpp"
#include "ClauseBackground.hpp"
#include "BuiltInTheoryDispatcher.hpp"
#include "PagedPassiveClauseStorage.inl"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#  define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.hpp"
//====================================================


using namespace BK;
using namespace Gem;
using namespace VK;

/**********
Master::Master(ComponentName* componentNaming) :
  _memoryReleaseInProgress(false),
  _continuation(Continuation_Init),
  _terminated(false),
  _proof(0),
  _reservedPassive(Clause::MainSetReservedPassive,"_reservedPassive"),
  _active(Clause::MainSetActive,"_active"),
  _history(Clause::MainSetHistory,"_history"),
  _answers(Clause::MainSetAnswers,"_answers"),
  _trash(Clause::MainSetTrash,"_trash"),
  _queueForBackSimplification(DOP("_queueForBackSimplification")),
  _queueForIntegrationIntoBackwardDemodIndex(DOP("_queueForIntegrationIntoBackwardDemodIndex")),
  _backSubsumption(&_clauseNumIndex),

  _literalSharing(&_sharingIndex),

  _clauseAssembler(&_literalSharing,&_litSelection),

  
  _newClauseQueue(DOP("Master::_newClauseQueue")),
  _inputClauseCollector(DOP("Master::_inputClauseCollector")),
  _reanimatedClauseCollector(DOP("Master::_reanimatedClauseCollector")),
  _mainClauseCollector(DOP("Master::_mainClauseCollector")),
  _propositionalSplitting(componentNaming),
  _negativeEqualitySplitting(componentNaming),
  _sharedReanimatedClause(0)
{
  CALL("constructor Master()");  
  linkObjects();
  setStrategy();
}; // Master::Master()

Master::~Master()
{
  CALL("destructor ~Master()");
  if (DestructionMode::isThorough())
    {	
      ASSERT(!_queueForBackSimplification.isLocked());
      ASSERT(!_queueForIntegrationIntoBackwardDemodIndex.isLocked());
      if (_infDispatcher.isLocked()) _infDispatcher.abandonCurrentQuery();
      if (backDemodulationIsLocked()) abandonBackDemodulationQuery();
      ASSERT(!_forwardDemodulation.isLocked());
      ASSERT(!_backSubsumption.isLocked());
      ASSERT(!_forwardSubsumption.isLocked());
      ASSERT(!_clauseNumIndex.isLocked());
      ASSERT(!_clauseAssembler.sharingIsLocked());

      // sets of persistent clauses
      switch (_parameters.main_alg)
	{
	case 0: destroyPassive0(); break;
	case 3: destroyPassive3(); break;
	default: ICP("ICP0"); break; 
	}; 

      Clause* cl;
      while (_reservedPassive.Nonempty())
	{
	  cl = _reservedPassive.Top();
	  enqueueForDisposal(cl);
	};
      _reservedPassive.Purge(); 

      while (_active.Nonempty())
	{
	  cl = _active.Top();
	  enqueueForDisposal(cl);
	};
      _active.Purge(); 

      while (_answers.Nonempty())
	{
	  cl = _answers.Top();
	  enqueueForDisposal(cl);
	};
      _answers.Purge(); 
  
      ASSERT(_answers.clean());

      if (_sharedReanimatedClause)
	{
	  ASSERT(_sharedReanimatedClause->mainSet() == Clause::MainSetFree);
	  enqueueForDisposal(_sharedReanimatedClause);
	};


      // queue for disposal 
      // not after _history.DestroyAllClauses()! 
      disposeObsolete();
      _history.DestroyAllClauses();  // not before disposeObsolete()!
    };
}; // Master::~Master()
***********/

    
void Master::init(ComponentName* componentNaming)
{
  CALL("init(ComponentName* componentNaming)");

  BK::init(options);
  _memoryReleaseInProgress = false;
  BK::init(_timer);
  _continuation = Continuation_Init;
  _terminated = false;
  _proof = 0;

  BK::init(_parameters);
  BK::init(_statistics);
  BK::init(_unifier);
  BK::init(_passive);
  BK::init(_pagedPassive);
  BK::init(_reservedPassive,Clause::MainSetReservedPassive,"_reservedPassive");
  BK::init(_active,Clause::MainSetActive,"_active");
  BK::init(_history,Clause::MainSetHistory,"_history");
  BK::init(_answers,Clause::MainSetAnswers,"_answers");
  BK::init(_trash,Clause::MainSetTrash,"_trash");

#ifdef DEBUG_NAMESPACE
  BK::init(_queueForBackSimplification,"_queueForBackSimplification");
  BK::init(_queueForIntegrationIntoBackwardDemodIndex,"_queueForIntegrationIntoBackwardDemodIndex");
#else
  BK::init(_queueForBackSimplification);
  BK::init(_queueForIntegrationIntoBackwardDemodIndex);   
#endif

  BK::init(_infDispatcher);
  BK::init(_forwardSubsumption);
  _backSubsumption.init(&_clauseNumIndex);
  BK::init(_forwardDemodulation);
  BK::init(_backDemodulation);

 
  BK::init(_sharingIndex);
  BK::init(_literalSharing,&_sharingIndex);

  //_litSelection.init();
  BK::init(_litSelection);

  BK::init(_standardLiteralOrdering);
  BK::init(_clauseAssembler,&_literalSharing,&_litSelection);



#ifdef DEBUG_NAMESPACE

  BK::init(_newClauseQueue,"Master::_newClauseQueue");
  BK::init(_inputClauseCollector,"Master::_inputClauseCollector");
  BK::init(_reanimatedClauseCollector,"Master::_reanimatedClauseCollector");
  BK::init(_mainClauseCollector,"Master::_mainClauseCollector");
#else
  BK::init(_newClauseQueue);
  BK::init(_inputClauseCollector);
  BK::init(_reanimatedClauseCollector);
  BK::init(_mainClauseCollector);
#endif
  BK::init(_newInputClause);
  BK::init(_retainedClause);
  BK::init(_reanimatedClause);
  BK::init(_propositionalSplitting,componentNaming);
  BK::init(_negativeEqualitySplitting,componentNaming);

  BK::init(_forwardProcessing);
  BK::init(_forwardProcessingForSupport);


  BK::init(_clauseNumIndex);

  _sharedReanimatedClause = 0;
  linkObjects();
  setStrategy();
}; // void init(ComponentName* componentNaming)


void Master::destroy()
{
  CALL("destroy()");
  if (DestructionMode::isThorough())
    {	
      ASSERT(!_queueForBackSimplification.isLocked());
      ASSERT(!_queueForIntegrationIntoBackwardDemodIndex.isLocked());
      if (_infDispatcher.isLocked()) _infDispatcher.abandonCurrentQuery();
      if (backDemodulationIsLocked()) abandonBackDemodulationQuery();
      ASSERT(!_forwardDemodulation.isLocked());
      ASSERT(!_backSubsumption.isLocked());
      ASSERT(!_forwardSubsumption.isLocked());
      ASSERT(!_clauseNumIndex.isLocked());
      ASSERT(!_clauseAssembler.sharingIsLocked());
      

      // sets of persistent clauses
      switch (_parameters.main_alg)
	{
	case 0: destroyPassive0(); break;
	case 3: destroyPassive3(); break;
	default: ICP("ICP0"); break; 
	}; 

      Clause* cl;
      while (_reservedPassive.Nonempty())
	{
	  cl = _reservedPassive.Top();
	  enqueueForDisposal(cl);
	};
      _reservedPassive.Purge(); 

      while (_active.Nonempty())
	{
	  cl = _active.Top();
	  enqueueForDisposal(cl);
	};
      _active.Purge(); 

      while (_answers.Nonempty())
	{
	  cl = _answers.Top();
	  enqueueForDisposal(cl);
	};
      _answers.Purge(); 
  
      ASSERT(_answers.clean());

      if (_sharedReanimatedClause)
	{
	  ASSERT(_sharedReanimatedClause->mainSet() == Clause::MainSetFree);
	  enqueueForDisposal(_sharedReanimatedClause);
	};


      // queue for disposal 
      // not after _history.DestroyAllClauses()! 
      disposeObsolete();
      _history.DestroyAllClauses();  // not before disposeObsolete()!
    };
    
  BK::destroy(_clauseNumIndex);

  BK::destroy(_forwardProcessingForSupport);
  BK::destroy(_forwardProcessing);
  BK::destroy(_negativeEqualitySplitting);
  BK::destroy(_propositionalSplitting);
  BK::destroy(_reanimatedClause);
  BK::destroy(_retainedClause);
  BK::destroy(_newInputClause);
  BK::destroy(_mainClauseCollector);
  BK::destroy(_reanimatedClauseCollector);
  BK::destroy(_inputClauseCollector);
  BK::destroy(_newClauseQueue);

  BK::destroy(_clauseAssembler);
  BK::destroy(_standardLiteralOrdering);
 
  //_litSelection.destroy();
  BK::destroy(_litSelection);

  BK::destroy(_literalSharing);
  BK::destroy(_sharingIndex);

  BK::destroy(_backDemodulation);
  BK::destroy(_forwardDemodulation);
  _backSubsumption.destroy();
  BK::destroy(_forwardSubsumption);
  BK::destroy(_infDispatcher);



  BK::destroy(_queueForIntegrationIntoBackwardDemodIndex);
  BK::destroy(_queueForBackSimplification);
  BK::destroy(_trash);
  BK::destroy(_answers);
  BK::destroy(_history);
  BK::destroy(_active);
  BK::destroy(_reservedPassive);
  BK::destroy(_pagedPassive);
  BK::destroy(_passive);
  BK::destroy(_unifier);
  BK::destroy(_parameters);
  BK::destroy(_statistics);
  BK::destroy(_timer);
  BK::destroy(options);
  BK_CORRUPT(*this);
}; // void Master::destroy()

void Master::reset() 
{
  CALL("reset()");
  _unifier.reset();
  Unifier::setCurrent(&_unifier);

  // sets of persistent clauses
  switch (_parameters.main_alg)
    {
    case 0: destroyPassive0(); break;
    case 3: destroyPassive3(); break;
    default: ICP("ICP0"); break; 
    }; 

  Clause* cl;
  while (_reservedPassive.Nonempty())
    {
      cl = _reservedPassive.Top();
      enqueueForDisposal(cl);
    };
  _reservedPassive.Purge(); // to make it clean()

  while (_active.Nonempty())
    {
      cl = _active.Top();
      enqueueForDisposal(cl);
    };
  _active.Purge(); // to make it clean()


  while (_answers.Nonempty())
    {
      cl = _answers.Top();
      enqueueForDisposal(cl);
    };
  _answers.Purge(); // to make it clean()
  
  ASSERT(_answers.clean());

  if (_sharedReanimatedClause)
    {
      ASSERT(_sharedReanimatedClause->mainSet() == Clause::MainSetFree);
      enqueueForDisposal(_sharedReanimatedClause);
    };

  // queue for disposal 
  // not after _history.DestroyAllClauses()! 
  disposeObsolete();
  _history.DestroyAllClauses();  // not before disposeObsolete()!

  _queueForBackSimplification.reset();
  _queueForIntegrationIntoBackwardDemodIndex.reset();

  // indices, inference engines and forward processing   


  _infDispatcher.reset();
  _forwardSubsumption.reset();
  _backSubsumption.reset();
  _forwardDemodulation.reset();
  _backDemodulation.reset();
  _sharingIndex.reset();
  _literalSharing.reset();
  _clauseAssembler.reset();
  
  _newClauseQueue.reset();
  _inputClauseCollector.reset();
  _reanimatedClauseCollector.reset();
  _mainClauseCollector.reset();
  _newInputClause.clear();
  _retainedClause.clear();
  _reanimatedClause.clear();
  _propositionalSplitting.reset();
  _negativeEqualitySplitting.reset();
  _forwardProcessing.reset();
  _forwardProcessingForSupport.reset();
  
  _clauseNumIndex.reset();

  options.reset(); 
  resetStatistics();
  _statistics.reset();
  _sharedReanimatedClause = 0;
  linkObjects();
  setStrategy();
}; //  void Master::reset() 

void Master::linkObjects()
{
  CALL("linkObjects()");
  _inputClauseCollector.setOutputQueue(&_newClauseQueue);
  _inputClauseCollector.setStatistics(&_statistics);
  _reanimatedClauseCollector.setOutputQueue(0);
  _reanimatedClauseCollector.setStatistics(&_statistics);
  _mainClauseCollector.setOutputQueue(&_newClauseQueue);
  _mainClauseCollector.setStatistics(&_statistics);
  _forwardProcessing.setStatistics(&_statistics);
  _forwardProcessingForSupport.setStatistics(&_statistics);
  _infDispatcher.setStatistics(&_statistics);
  _infDispatcher.setClauseCollector(&_mainClauseCollector);
  _infDispatcher.setForwardProcessing(&_forwardProcessing);
  _backDemodulation.setClauseCollector(&_mainClauseCollector);
}; // void Master::linkObjects()

void Master::endOfInput()
{
  CALL("endOfInput()");
  // destroy obsolete clauses
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());

  disposeObsolete(); 

  ASSERT(_trash.Empty());

}; // void Master::endOfInput()


void Master::mainLoop0()
{
  CALL("mainLoop0()");

  Clause* current;

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop0_ProcessNewInference:
      cancelTermination();
      goto process_new_inference;
    case Continuation_MainLoop0_BackSimplify:
      cancelTermination();
      goto back_simplify;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::mainLoop0()");
    };

  ASSERT(!_terminated);
  ASSERT(!_proof);

  _timer.reset(); 
  _timer.start();
  _timerIsOn = true;
  _externalTickFlag = false;
  _startMemory = (float)GlobAlloc::occupiedByObjects(); 
  // ^ mainly for LRS on memory


 select_current:

  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      _timer.stop();
      _timerIsOn = false;
      return; 
    }; 

  current = select0(); 
  if (!current) 
    { 
      terminate(EmptyPassive);
      _timer.stop();
      _timerIsOn = false;
      return; 
    };


  if ((_parameters.orphan_murder)
      && (!current->IsVIP())
      && (current->IsOrphan()))
    {      
      ++_statistics.clauses.retained.passiveOrphansMurdered;
      disposeClause(current);  
      estimateResources();
      goto select_current;
    }; 
  
  makeActive0(current); 


  
  if (_parameters.outputStreamForPassiveProfiles &&
      (_active.Total() % _parameters.showPassiveProfileInterval == 0))
    {
      _passive.outputProfile(*_parameters.outputStreamForPassiveProfiles);
    };

  eliminateAllocationBufferDeficit0();  

  _infDispatcher.newQuery(current);
 next_inference:
  if (_infDispatcher.nextInference())
    {
    process_new_inference:
      processCollected0();
      if (_proof) 
	{ 
	  terminate(Proof); 
	  _timer.stop();
	  _timerIsOn = false;
	  _continuation = Continuation_MainLoop0_ProcessNewInference;
	  return; 
	};
      _continuation = Continuation_Init;

      if (_passive.Total() >= _parameters.maxPassive) 
	{ 
	  terminate(MaxPassive);  
	  _infDispatcher.abandonCurrentQuery();
	  _timer.stop();
	  _timerIsOn = false;
	  return; 
	};

      internalTick0();

    back_simplify:
      backSimplify();
      if (_terminated) 
	{ 
	  _timer.stop();
	  _timerIsOn = false;
	  _continuation = Continuation_MainLoop0_BackSimplify; 
	  return; 
	};
      _continuation = Continuation_Init;
      integrateIntoBackDemIndexFromQueue();
      goto next_inference;
    };
  // no more inferences with current
  _infDispatcher.abandonCurrentQuery();
  
  internalTick0(); 
    

  // destroy obsolete clauses
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());

  disposeObsolete(); 
  ASSERT(_trash.Empty());
  estimateResources();
  if ((_parameters.lrs == 2) || (_parameters.lrs == 3))
    {
      recycleUnreachablePassive0();
    };  
  goto select_current; 
}; // void Master::mainLoop0()

void Master::backSimplify()
{
  CALL("backSimplify()");  
  Clause* backSimplified;
  bool someActiveSimplified;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_BackSimplify:
      // same as below
    case Continuation_EliminateDefinitions_BackSimplify:
      // same as below
    case Continuation_MainLoop0_BackSimplify:
      someActiveSimplified = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backSimplify()");
    };

  ASSERT((_parameters.backward_demodulation > 0) || 
	 _queueForBackSimplification.empty());
  someActiveSimplified = false;
  bool onBranch;
  if (_parameters.backward_demodulation > 0)
    {
    next_simplifier:
      if (_queueForBackSimplification.dequeue(_currentSimplifier))
	{ 
	  ASSERT(_currentSimplifier->flag(Clause::FlagInQueueForBackSimplification));
	  _currentSimplifier->clearFlag(Clause::FlagInQueueForBackSimplification);
	  if (backDemodulationQuery(_currentSimplifier))
	    { 
	    next_back_demodulated:
	      if (nextBackDemodulated(onBranch))
		{         
		  if (collectBackSimplified())
		    {  
		      _currentSimplifier->setFlag(Clause::FlagHasChildren);
		      backSimplified = _backDemodulation.simplified();
		      

		      someActiveSimplified = someActiveSimplified 
			|| (backSimplified->mainSet() == Clause::MainSetActive) 
			|| (backSimplified->mainSet() == Clause::MainSetDefinitions);  
		      
		      if (onBranch)
			{
			  ++_statistics.inferences.simplifying.onBranchBackwardDemodulation;
			  if ((backSimplified->mainSet() == Clause::MainSetActive) ||
			      (backSimplified->mainSet() == Clause::MainSetDefinitions))
			    {
			      ++_statistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
			    }
			  else
			    {
			      ++_statistics.clauses.retained.passiveSimplifiedByOnBranchDemodulation;
			    };
			}
		      else
			{
			  ++_statistics.inferences.simplifying.backwardDemodulation;
			  if ((backSimplified->mainSet() == Clause::MainSetActive) ||
			      (backSimplified->mainSet() == Clause::MainSetDefinitions))
			    {
			      ++_statistics.clauses.retained.activeSimplifiedByDemodulation;
			    }
			  else
			    {
			      ++_statistics.clauses.retained.passiveSimplifiedByDemodulation;
			    };
			};

		      backSimplified->setFlag(Clause::FlagHasChildren);

		      enqueueForDisposal(backSimplified);        

		    process_new_inference:
		      processCollected0(); // not before enqueueForDisposal(backSimplified)!

		      if (_proof) 
			{ 
			  terminate(Proof); 
			  return; 
			};
		            
		      _continuation = Continuation_Init;

		      if (_passive.Total() >= _parameters.maxPassive) 
			{ 
			  terminate(MaxPassive); 
			  abandonBackDemodulationQuery();
			  return; 
			}; 
		    };
		  internalTick0();
		  goto next_back_demodulated;
		}; // if (nextBackDemodulated()) 
	      abandonBackDemodulationQuery();
	    }; // if (backDemodulationQuery(_currentSimplifier))
	  
	  goto next_simplifier;
	}; // if (_queueForBackSimplification.dequeue(_currentSimplifier))
      if (_parameters.orphan_murder && someActiveSimplified) killActiveOrphans0();
    };
}; // void Master::backSimplify()


void Master::eliminateDefinitions()
{
  CALL("Master::eliminateDefinitions()");
  ASSERT(!_parameters.main_alg);
  setPreprocessingInProgress(true);
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_EliminateDefinitions_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    case Continuation_EliminateDefinitions_BackSimplify:
      cancelTermination();
      goto back_simplify;

    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::eliminateDefinitions()");
    };

  _timer.reset();
  _timer.start();
  Clause* cl;
 select:
  _passive.ResetIterator();
 try_next:
  cl = _passive.IterNext();
  if (cl) 
    {
      if (cl->ContainsEliminated(_parameters.pred_elim_table))
	{
	  _passive.Remove(cl);
      
	  // perform resolutions
	  makeActive0(cl);
	  _infDispatcher.newQuery(cl);
	check_next_inference:
	  if (_infDispatcher.nextInference())
	    {
	    process_collected_input:
	      processCollected0();
	      if (_proof) 
		{ 
		  terminate(Proof); 
		  _continuation = Continuation_EliminateDefinitions_ProcessCollectedInput;
		  _timer.stop();
		  _timerIsOn = false;
		  return;
		};
	      _continuation = Continuation_Init;

	    back_simplify:
	      backSimplify();
	      if (_terminated) 
		{ 
		  _continuation = Continuation_EliminateDefinitions_BackSimplify;    
		  _timer.stop();
		  _timerIsOn = false;
		  return;
		};
	      _continuation = Continuation_Init;
	      goto check_next_inference;
	    };

	  _infDispatcher.abandonCurrentQuery();

	  // destroy obsolete clauses
	  ASSERT(_queueForBackSimplification.empty());
	  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());
	  disposeObsolete(); 

	  // use next
	  goto select;     
	} 
      else // !cl->ContainsEliminated(_parameters.pred_elim_table)
	goto try_next;      
    }; // if (cl)

  // all done, collect results
  ASSERT(!_proof); 
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());

  // all _active are obsolete now, make them _history
  Clause* act;
  while (_active.Nonempty())
    {
      act = _active.Top();  
      enqueueForDisposal(act);
    };
  disposeObsolete();    
  _timer.stop();
  _timerIsOn = false;
}; // void Master::eliminateDefinitions()




void Master::inputClause0(Clause* cl)
{
  CALL("inputClause0(Clause* cl)");
  if (cl->isSupportClause())
    {
      inputSupportClause0(cl);
    }
  else
    inputQueryClause0(cl);
}; // void Master::inputClause0(Clause* cl)

void Master::inputSupportClause0(Clause* cl)
{
  CALL("inputSupportClause0(Clause* cl)");

  ASSERT(cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputSupportClause0_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputSupportClause0(Clause* cl)");
    };

  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;

  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
    {
      _inputClauseCollector.addAncestor(cl);
      _inputClauseCollector.markAsSupport();
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{
	  ASSERT(!_newInputClause.isNull());
	  _forwardProcessingForSupport.load(_newInputClause);
	process_collected_input:
	  while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	    {
	      ++_statistics.clauses.retained.total;
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  ASSERT(sharedRetainedClause);
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputSupportClause0_ProcessCollectedInput;
		      return;
		    };
		  sharedRetainedClause->SetNumber(_clauseNumberGenerator->generate());      
		  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTRODUCE_CLAUSE_ID(sharedRetainedClause);
		  if (_parameters.useClauseNumIndex) 
		    {
		      sharedRetainedClause->protect();
		      _clauseNumIndex.insert(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		    };
		  if (_parameters.showPassive) _parameters.showPassive(sharedRetainedClause);
		  integratePassive(sharedRetainedClause);
		  makeActive0(sharedRetainedClause);
		};
	    }; // while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue(); 
	}; // if (_inputClauseCollector.finaliseClause(_newInputClause))

    }; // if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
}; // void Master::inputSupportClause0(Clause* cl)


void Master::inputQueryClause0(Clause* cl)
{
  CALL("inputQueryClause0(Clause* cl)");
  ASSERT(!cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    case Continuation_InputQueryClause0_BackSimplify:
      cancelTermination();
      goto back_simplify;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputQueryClause0(Clause* cl)");
    };

  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;

  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
    {
      _inputClauseCollector.addAncestor(cl);
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{	  
	  ASSERT(!_newInputClause.isNull());
	  _passive.MaxPriorityOn();
	  _forwardProcessing.load(_newInputClause);

	process_collected_input:
	  while (_forwardProcessing.anotherRetained(_retainedClause))
	    {
	      ++_statistics.clauses.retained.total;
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputQueryClause0_ProcessCollectedInput;
		      return;
		    };
		  makePassive0(sharedRetainedClause); 
		  if (_parameters.backward_subsumption) backSubsumeBy(sharedRetainedClause);
		  if (_parameters.backward_demodulation > 0 &&
		      eligibleAsBackwardSimplifier(sharedRetainedClause))
		    {
		      ASSERT(!sharedRetainedClause->flag(Clause::FlagInQueueForBackSimplification));
		      sharedRetainedClause->protect();
		      _queueForBackSimplification.enqueue(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		      sharedRetainedClause->setFlag(Clause::FlagInQueueForBackSimplification);
		    }; 
		  integratePassive(sharedRetainedClause);
		};
	    }; // while (_forwardProcessing.anotherRetained(_retainedClause))
	  _continuation = Continuation_Init;
	  _passive.MaxPriorityOff(); 
	  
	back_simplify:
	  backSimplify();
	  if (_terminated)
	    {
	      _timer.stop();
	      _timerIsOn = false;
	      _continuation = Continuation_InputQueryClause0_BackSimplify;
	      return; 
	    };
	  ASSERT(_queueForBackSimplification.empty());
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue();	  
	}; // if (_inputClauseCollector.finaliseClause(_newInputClause))
    }; // if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
  
  _timer.stop();
  _timerIsOn = false;
}; // void Master::inputQueryClause0(Clause* cl)

void Master::processCollected0()
{
  CALL("processCollected0()");
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_BackSimplify:
      // same as below
    case Continuation_EliminateDefinitions_BackSimplify:
      // same as below
    case Continuation_EliminateDefinitions_ProcessCollectedInput:
      // same as below
    case Continuation_MainLoop0_ProcessNewInference:
      // same as below
    case Continuation_MainLoop0_BackSimplify:
      goto process_retained;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::processCollected0()");
    };
  
  while (!_newClauseQueue.isEmpty())
    {
      _forwardProcessing.load(_newClauseQueue);

    process_retained:  
      while (_forwardProcessing.anotherRetained(_retainedClause))
	{
	  ++_statistics.clauses.retained.total;
	  sharedRetainedClause = 
	    _clauseAssembler.assembleShared(*_retainedClause);
	  if (!sharedRetainedClause)
	    {
	      ++_statistics.clauses.retained.passiveRecycled;
	    }
	  else
	    {
	      sharedRetainedClause->setMainSet(Clause::MainSetNew);
	      if (sharedRetainedClause->isAnswer()) 
		{
		  rememberAnswer0(sharedRetainedClause);
		  _proof = sharedRetainedClause; 
		  return; 
		}; 
	      
	      _continuation = Continuation_Init;
	      
	      
	      makePassive0(sharedRetainedClause);
	      
	      if (sharedRetainedClause->weight() > _parameters.maxWeight)
		{        
		  enqueueForDisposal(sharedRetainedClause);
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  if (_parameters.backward_subsumption) 
		    if (backSubsumeBy(sharedRetainedClause))
		      if (_parameters.orphan_murder) 
			killActiveOrphans0();
		  
		  
		  if (_parameters.backward_demodulation > 0 &&
		      eligibleAsBackwardSimplifier(sharedRetainedClause))
		    {
		      ASSERT(!sharedRetainedClause->flag(Clause::FlagInQueueForBackSimplification));
		      sharedRetainedClause->protect();
		      _queueForBackSimplification.enqueue(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		      
		      sharedRetainedClause->setFlag(Clause::FlagInQueueForBackSimplification);
		    };
		  
		  integratePassive(sharedRetainedClause);
		  
		  eliminateAllocationBufferDeficit0();
		};
	    };
	}; // while (_forwardProcessing.anotherRetained(_retainedClause))
    }; // while (!_newClauseQueue.isEmpty())
}; // void Master::processCollected0()




Clause* Master::select0()
{
  CALL("select0()");
  Clause* selected;
 try_next_passive:
  selected = _passive.Next();
  if (!selected)
    {// _passive is empty, try _reservedPassive 
      _parameters.lrs = 0;
      updateWeightLimit(VampireKernelConst::MaxClauseSize);
      loadReservedClauses();    
      selected = _passive.Next();
      if (selected) 
	{
	  ASSERT(selected->mainSet() == Clause::MainSetPassive);
	  selected->setMainSet(Clause::MainSetFree);
	};  
      return selected;
    };
  if (selected->practicallyMinResidualSize() > _parameters.maxWeight)
    { 
      makeReserved(selected); 
      goto try_next_passive;         
    };
  if (selected)     
    {
      ASSERT(selected->mainSet() == Clause::MainSetPassive);
      selected->setMainSet(Clause::MainSetFree);
    }; 
  return selected; 
}; // Clause* Master::select0()  

void Master::makeActive0(Clause* cl)
{
  CALL("makeActive0(Clause* cl)");
  ++(_statistics.clauses.retained.selected);
  cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 

 
  _active.Insert(cl);

  _infDispatcher.makeActive(cl);
  if (_parameters.showActive) _parameters.showActive(cl); 
}; // void Master::makeActive0(Clause* cl)

void Master::makePassive0(Clause* cl)
{
  CALL("makePassive0(Clause* cl)");
  cl->SetNumber(_clauseNumberGenerator->generate());
  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTRODUCE_CLAUSE_ID(cl);
  cl->setMainSet(Clause::MainSetPassive);
  _passive.Insert(cl);
  if (_parameters.useClauseNumIndex) 
    {
      cl->protect();
      _clauseNumIndex.insert(cl);
      cl->unprotect();
    };
  if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::makePassive0(Clause* cl)


void Master::rememberAnswer0(Clause* cl)
{
  CALL("rememberAnswer0(Clause* cl)");
  cl->SetNumber(_clauseNumberGenerator->generate());
  _answers.Insert(cl);
  if (_parameters.useClauseNumIndex) 
    {
      cl->protect();
      _clauseNumIndex.insert(cl);
      cl->unprotect();
    };
  if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::rememberAnswer0(Clause* cl)

void Master::integratePassive(Clause* cl)
{
  CALL("integratePassive(Clause* cl)");
  // see also removeFromIndices(Clause* cl)

  if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
  if (_parameters.backward_subsumption) integrateIntoBackwardSubsumptionIndex(cl);
  if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
  if (_parameters.backward_demodulation > 0) 
    {
      if (backDemodulationIsLocked())
	{
	  cl->protect();
	  _queueForIntegrationIntoBackwardDemodIndex.enqueue(cl);
	  cl->unprotect();
	  cl->setFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	}
      else 
	{
	  cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	  integrateIntoBackDemodulationIndex(cl);
	};
    };
}; // void Master::integratePassive(Clause* cl) 


void Master::integrateIntoBackDemIndexFromQueue()
{
  CALL("integrateIntoBackDemIndexFromQueue()");
  ASSERT(!backDemodulationIsLocked());
  Clause* cl;
  while (_queueForIntegrationIntoBackwardDemodIndex.dequeue(cl))
    { 
      ASSERT(cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));
      ASSERT(cl->mainSet() == Clause::MainSetPassive);
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
      integrateIntoBackDemodulationIndex(cl);
    };
}; // void Master::integrateIntoBackDemIndexFromQueue()




void Master::removeFromIndices(Clause* cl)
{
  CALL("removeFromIndices(Clause* cl)");
  // see also integratePassive(Clause* cl)
  if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
  if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
}; // void Master::removeFromIndices(Clause* cl)

void Master::makeReserved(Clause* cl)
{
  CALL("makeReserved(Clause* cl)");
  _reservedPassive.Insert(cl);
}; // void Master::makeReserved(Clause* cl)


bool Master::canBeEnqueuedForDisposal(Clause* cl)
{
  CALL("canBeEnqueuedForDisposal(Clause* cl)");
  if (cl->mainSet() == Clause::MainSetTrash) return true;
    
  if (cl->flag(Clause::FlagInQueueForBackSimplification) &&
      _queueForBackSimplification.isLocked())
    return false;

  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex) &&
      _queueForIntegrationIntoBackwardDemodIndex.isLocked())
    return false;

  return true;
}; // bool Master::canBeEnqueuedForDisposal(Clause* cl)


void Master::enqueueForDisposal(Clause* cl)
{
  CALL("enqueueForDisposal(Clause* cl)"); 
  ASSERT(canBeEnqueuedForDisposal(cl));
  if (cl->mainSet() == Clause::MainSetTrash) return;
  switch (cl->mainSet()) 
    {
    case Clause::MainSetFree: break;
    case Clause::MainSetNew:
      ICP("ICP0");
      break;
    case Clause::MainSetPassive: 
      _passive.Remove(cl); 
      cl->setMainSet(Clause::MainSetFree);
      break;
    case Clause::MainSetReservedPassive: _reservedPassive.Remove(cl); break;    
    case Clause::MainSetActive: _active.Remove(cl); break;
    case Clause::MainSetAnswers: _answers.Remove(cl); break;
#ifdef DEBUG_NAMESPACE
    default: 
      ICP("ICP1"); 
      return;
#else
    default: 
      RuntimeError::report("Wrong clause affiliation.");
#endif
    };

  ASSERT(cl->mainSet() == Clause::MainSetFree);
      
 
  if (cl->flag(Clause::FlagInQueueForBackSimplification))
    {
      ASSERT(!_queueForBackSimplification.isLocked());
      _queueForBackSimplification.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForBackSimplification);
    };
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex))
    {
      ASSERT(!_queueForIntegrationIntoBackwardDemodIndex.isLocked());
      _queueForIntegrationIntoBackwardDemodIndex.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
    };

  bool canBeDestroyed = !cl->isProtected();
  if (!_infDispatcher.isLocked())
    {
      if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
      if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
    }
  else
    {
      if ((cl->flag(Clause::FlagInInferenceDispatcherAsActive)) 
	  || (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)))
	canBeDestroyed = false;
    };

  if (!backDemodulationIsLocked())
    {
      if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl); 
    }
  else
    if (cl->flag(Clause::FlagInBackDemodulationIndex)) canBeDestroyed = false;

  if (!_forwardDemodulation.isLocked())
    {
      if (cl->flag(Clause::FlagInForwardDemodulationIndex)) 
	removeFromForwardDemodulationIndex(cl);
    }
  else
    if (cl->flag(Clause::FlagInForwardDemodulationIndex)) 
      canBeDestroyed = false;

  if (!_backSubsumption.isLocked() && !_clauseNumIndex.isLocked())
    {
      if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) 
	removeFromBackwardSubsumptionIndex(cl);
    }
  else 
    if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) 
      canBeDestroyed = false;
  

  if (!_forwardSubsumption.isLocked())
    {
      if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) 
	removeFromForwardSubsumptionIndex(cl);
    }
  else
    if (cl->flag(Clause::FlagInForwardSubsumptionIndex))
      canBeDestroyed = false;

  if (cl->flag(Clause::FlagHasChildren)) canBeDestroyed = false;

  if (cl->flag(Clause::FlagInClauseNumIndex) &&
      _clauseNumIndex.isLocked())
   canBeDestroyed = false;
  
  bool shared = cl->flag(Clause::FlagLiteralsShared);
  if (shared && _clauseAssembler.sharingIsLocked())
   canBeDestroyed = false;

  if (canBeDestroyed)
    {
      if (shared)
	{
	  _clauseAssembler.removeFromSharing(cl);
	  cl->clearFlag(Clause::FlagLiteralsShared);
	};
      if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
      if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
    }
  else // must wait in the queue
    {
      _trash.Insert(cl);
    };
}; // void Master::enqueueForDisposal(Clause* cl)

void Master::disposeObsolete()
{
  CALL("disposeObsolete()"); 
  Clause* disposedClause;
  while (_trash.Nonempty())
    { 
      disposedClause = _trash.Pop();
      ASSERT(disposedClause->mainSet() == Clause::MainSetFree);
      disposeClause(disposedClause);
    };
}; // void Master::disposeObsolete()

void Master::disposeClause(Clause* cl)
{
  CALL("disposeClause(Clause* cl)");
  ASSERT(cl->mainSet() == Clause::MainSetFree);
  if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
  if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
  removeFromIndices(cl);

  ASSERT(!cl->flag(Clause::FlagInQueueForBackSimplification));
  ASSERT(!cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));

  if (cl->flag(Clause::FlagHasChildren))
    {
      // must be in history
      _history.Insert(cl); 
    }
  else // can be completely destroyed
    {
      ASSERT(!cl->isProtected());
      bool shared = cl->flag(Clause::FlagLiteralsShared);
      if (shared) 
	{
	  _clauseAssembler.removeFromSharing(cl);
	  cl->clearFlag(Clause::FlagLiteralsShared);
	};
      if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
      if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
    };

}; // void Master::disposeClause(Clause* cl)


void Master::loadReservedClauses()
{
  CALL("loadReservedClauses()");
  Clause* cl;
  while (!_reservedPassive.Empty())
    {
      cl = _reservedPassive.Pop();
      ASSERT(cl->mainSet() == Clause::MainSetFree); 
      cl->setMainSet(Clause::MainSetPassive);
      _passive.Insert(cl); 
    };
}; // void Master::loadReservedClauses()

bool Master::collectBackSimplified()
{
  CALL("collectBackSimplified()");
  return _backDemodulation.collectInference();  
}; // bool Master::collectBackSimplified()


bool Master::backSubsumeBy(Clause* cl)
{
  CALL("backSubsumeBy(Clause* cl)");
  if (cl->isNonSubsuming()) return false;
  Clause* subsumedClause;
  bool usedSubsumed = false;
  if (_backSubsumption.query(cl))
    {
      while (_backSubsumption.subsumeNext())
	{ 
	  ++_statistics.inferences.simplifying.backwardSubsumption;

	  cl->MakeVIP();
	  subsumedClause = _backSubsumption.subsumed();
	  ASSERT(subsumedClause != cl);
	  ASSERT(subsumedClause->Active()); 
	  usedSubsumed = (usedSubsumed 
			  || (subsumedClause->mainSet() == Clause::MainSetActive) 
			  || (subsumedClause->mainSet() == Clause::MainSetDefinitions));     
	  if (usedSubsumed)
	    {
	      ++_statistics.clauses.retained.activeSubsumed;
	    }
	  else
	    {
	      ++_statistics.clauses.retained.passiveSubsumed;
	    };
	  enqueueForDisposal(subsumedClause);     
	};
      _backSubsumption.abandonCurrentQuery();
    };
  return (bool)usedSubsumed;
}; // bool Master::backSubsumeBy(Clause* cl)

void Master::estimateResources1()
{
  CALL("estimateResources1()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();

  bool noTimeCheck = 
    used_time < 
    (_parameters.timeLimitForLRS * 
     _parameters.first_time_check);
  bool noMemoryCheck = 
    used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimitForLRS   - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable1(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());

  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources1()



void Master::estimateResources2()
{
  CALL("estimateResources2()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimitForLRS * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimitForLRS - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable2(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources2()


void Master::estimateResources3()
{
  CALL("estimateResources3()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();

  bool noTimeCheck = 
    used_time < _parameters.timeLimitForLRS * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimitForLRS - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable3(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
}; // void Master::estimateResources3()




void Master::estimateResources4()
{
  CALL("estimateResources4()");
  ASSERT(_parameters.lrs == 4);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();

  bool noTimeCheck = 
    used_time < _parameters.timeLimitForLRS * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (_halfLongMax) 
    : 
    ((long)(((_parameters.timeLimitForLRS - used_time)*moved_from_passive)/used_time));

  long reachableByMemory = (noMemoryCheck)? 
    (_halfLongMax)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;

  long newWeightLimit = _parameters.maxWeight;

  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && 
	  (newWeightLimit < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  // weight limit increases 
	  ++newWeightLimit;
	  if (newWeightLimit > _parameters.maxMaxWeight) 
	    {
	      newWeightLimit = _parameters.maxMaxWeight;
	    };
	};
    }
  else // currentlyUnreachable > 0
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  // weight limit decreases 
	  if (newWeightLimit > _passive.MaxPresentWeight()) 
	    {
	      newWeightLimit = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (newWeightLimit > _passive.MinPresentWeight()) 
	      { 
		--newWeightLimit;
	      };
	};
    };

  _parameters.unreachable = currentlyUnreachable;

  updateWeightLimit(newWeightLimit);

  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
}; // void Master::estimateResources4()



void Master::estimateResources5()
{
  CALL("estimateResources5()");
  ASSERT(_parameters.lrs == 5);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();

  bool noTimeCheck = 
    used_time < _parameters.timeLimitForLRS * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (_halfLongMax) 
    : 
    ((long)(((_parameters.timeLimitForLRS - used_time)*moved_from_passive)/used_time));

  long reachableByMemory = (noMemoryCheck)? 
    (_halfLongMax)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;

  long newWeightLimit = _parameters.maxWeight;

  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && 
	  (newWeightLimit < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  // weight limit increases 
	  ++newWeightLimit;
	  if (newWeightLimit > _parameters.maxMaxWeight) 
	    {
	      newWeightLimit = _parameters.maxMaxWeight;
	    };
	};
    }
  else // currentlyUnreachable > 0
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  // weight limit decreases 
	  if (newWeightLimit > _passive.MaxPresentWeight()) 
	    {
	      newWeightLimit = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (newWeightLimit >= _passive.MinPresentWeight())
	      { 
		--newWeightLimit;
	      };
	};
    };

  _parameters.unreachable = currentlyUnreachable;
  
  updateWeightLimit(newWeightLimit);

  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
}; // void Master::estimateResources5()


void Master::updateWeightLimit(long newLimit)
{
  CALL("updateWeightLimit(long newLimit)");
  _parameters.previousMaxWeight = _parameters.maxWeight;
  _parameters.maxWeight = newLimit;
  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
    {
      _parameters.maxWeight = _parameters.maxMaxWeight;
    };

  _infDispatcher.setClauseSizeLimit(_parameters.maxWeight);
  _mainClauseCollector.setSizeLimit(_parameters.maxWeight);
  _forwardProcessing.setClauseSizeLimit(_parameters.maxWeight); 
  _clauseAssembler.setMaxWeight(_parameters.maxWeight); 

  if (_parameters.showWeightChange && 
      (_parameters.previousMaxWeight != _parameters.maxWeight))
    _parameters.showWeightChange(_parameters.previousMaxWeight,
				 _parameters.maxWeight); 
}; // void Master::updateWeightLimit(long newLimit)


bool Master::completeStrategyUsed()
{
  return !(_parameters.incompleteAPriori
	   || _statistics.completenessViolated()
	   || _parameters.chaining_for_transitive
	   || (_parameters.selection >= 1000)
	   || (_parameters.selection <= -1000)
	   || (_parameters.literal_comparison_mode == 3UL)
	   || BuiltInTheoryDispatcher::current());      
}; // bool Master::completeStrategyUsed()


Clause* Master::goodActiveCandidateForDisposal()
{
  CALL("goodActiveCandidateForDisposal()");
  CLAUSE_SET::Iter iter(_active);
  Clause* cl;
  // first, find any candidate
  do
    {
      cl = iter.clause();
      if (!cl) return static_cast<Clause*>(0);
      iter.next();
    }
  while (cl->isInputOrPreprocessedInputClause() ||
	 !canBeEnqueuedForDisposal(cl));
  
  Clause* currentCandidate = cl;

  ulong numOfTrials = 100UL;
  ulong numOfImprovements = 5UL;

  while (!iter.isEnd())
    {
      cl = iter.clause();
      iter.next();
      if (!cl->isInputOrPreprocessedInputClause() &&
	  canBeEnqueuedForDisposal(cl))
	{
	  if (cl->weight() > currentCandidate->weight())
	    {
	      currentCandidate = cl;
	      --numOfImprovements;
	      if (!numOfImprovements) return currentCandidate;
	    };
	};
      --numOfTrials;
      if (!numOfTrials) return currentCandidate; 
    };
  
  return currentCandidate; 
}; // Clause* Master::goodActiveCandidateForDisposal()


Clause* Master::goodPassiveCandidateForDisposal0()
{
  CALL("goodPassiveCandidateForDisposal0()");

  // find first candidate
  Clause* cl = _passive.TheWorstUnblockedClause();
  while (cl &&
	 (cl->isInputOrPreprocessedInputClause() ||
	  !canBeEnqueuedForDisposal(cl)))
    {
      cl = _passive.theWorstUnblockedClauseAfter(cl);
    };  
  return cl; 
}; // Clause* Master::goodPassiveCandidateForDisposal0()


void Master::recycleUnreachablePassive0()
{
  CALL("recycleUnreachablePassive0()");
  Clause* cl;
 recycle_next:
  if (_passive.UnreachableKept() <= (_passive.CurrentlyKept()/5)) 
    return; // nothing to recycle
  cl = _passive.TheMostUnreachableClause();    
  if (!cl) return;  
  ASSERT(cl->mainSet() == Clause::MainSetPassive);
  _passive.Cancel(cl);
  cl->setMainSet(Clause::MainSetFree);
  enqueueForDisposal(cl);
  ++_statistics.clauses.retained.passiveRecycled;
  goto recycle_next;
}; // void Master::recycleUnreachablePassive0()


bool Master::releaseALotOfMemory0()
{
  CALL("releaseALotOfMemory0()");

  
  bool res = false;

  // first, recycle about 10% of passive clauses
  ulong numOfClausesToRecycle =
    _passive.Size() / 10;
  Clause* cl;

  while (numOfClausesToRecycle &&
	 ((cl = goodPassiveCandidateForDisposal0())))
    {
      _passive.Cancel(cl);
      cl->setMainSet(Clause::MainSetFree);
      enqueueForDisposal(cl);
      ++_statistics.clauses.retained.passiveRecycled;
      --numOfClausesToRecycle;
      res = true;  
      BK::GlobalStopFlag::check();
    };
 

#ifdef CHANGE_WEIGHT_LIMIT_AFTER_MAJOR_RECYCLING
  if (res)
    {
      _parameters.maxMaxWeight = _passive.MaxPresentWeight();
      updateWeightLimit(_passive.MaxPresentWeight());
    };
#endif


#ifdef DONT_RECYCLE_ACTIVE
  // now recycle about 10% of active clauses if possible  
  numOfClausesToRecycle = _active.Size() / 10;

  while (numOfClausesToRecycle)
    {
      Clause* cand = goodActiveCandidateForDisposal();
      if (!cand) return res;
      --numOfClausesToRecycle;
      res = true;
      enqueueForDisposal(cand);
      ++_statistics.clauses.retained.activeRecycled;
      BK::GlobalStopFlag::check();
    };
#endif

  return res;
}; // bool Master::releaseALotOfMemory0()

void Master::killActiveOrphans0()
{
  CALL("killActiveOrphans0()");
  ASSERT(!_active.mainIteratorIsEngaged());
  _active.resetMainIterator();
  for (Clause* cl = _active.Next(); cl; cl = _active.Next())
    if ((cl->Active())
	&& (!cl->IsVIP())
	&& (cl->IsOrphan()))
      {
	enqueueForDisposal(cl);
	++_statistics.clauses.retained.activeOrphansMurdered;
      };
}; // void Master::killActiveOrphans0()


void Master::eliminateAllocationBufferDeficit0()
{
  CALL("eliminateAllocationBufferDeficit0()");
  if (BK::GlobAlloc::activeBufferDeficit())
    {
      long disposalBound = 
	(long)(BK::GlobAlloc::occupiedByObjects() - 
	       (_parameters.allocation_buffer_price * 
		BK::GlobAlloc::activeBufferDeficit()));
      do
	{
	  Clause* cl = _passive.TheWorstUnblockedClause();    
	  if (!cl) return;  

	  ASSERT(cl->mainSet() == Clause::MainSetPassive);
	  ++_statistics.clauses.retained.passiveRecycled;
	  _passive.Cancel(cl);
	  cl->setMainSet(Clause::MainSetFree);
	  enqueueForDisposal(cl);
	}
      while (BK::GlobAlloc::activeBufferDeficit() && 
	     BK::GlobAlloc::occupiedByObjects() > disposalBound);

      if (BK::GlobAlloc::activeBufferDeficit()) 
	BK::GlobAlloc::annulBufferDeficit();
    };
}; // void Master::eliminateAllocationBufferDeficit0() 





void Master::destroyPassive0() 
{ 
  CALL("destroyPassive0()");
  _passive.DestroyAllClauses();
 
  _passive.ResetIterator();
  for (Clause* cl = _passive.IterNext(); cl; cl = _passive.IterNext())
    enqueueForDisposal(cl);
}; // void Master::destroyPassive0() 



void Master::initParameters()
{
  CALL("Master::initParameters()");

  _parameters.incompleteAPriori = options.incompleteAPriori;
 
  // main algorithm

  _parameters.main_alg = options.main_alg;

  // inference rules
  _parameters.binary_resolution = options.binary_resolution;
  _parameters.no_inferences_between_rules = options.no_inferences_between_rules;
  _parameters.splitting = options.splitting; 
  _parameters.splitting_style = options.splitting_style;
  _parameters.small_basic_comp_in_hyp_splitting = options.small_basic_comp_in_hyp_splitting;
  _parameters.neg_eq_splitting = options.neg_eq_splitting;  
  _parameters.neg_eq_spl_var_count_limit = options.neg_eq_spl_var_count_limit;  
  _parameters.paramodulation = options.paramodulation;  
  _parameters.paramodulate_from_variables = options.paramodulate_from_variables;
  _parameters.sim_back_sup = options.sim_back_sup;
  _parameters.forw_sup_no_dup_red = options.forw_sup_no_dup_red;
  _parameters.back_sup_no_dup_red = options.back_sup_no_dup_red;
  _parameters.prohibit_non_basic_superposition = options.prohibit_non_basic_superposition;
  _parameters.prohibit_non_basic_fd = options.prohibit_non_basic_fd;
  _parameters.prohibit_non_basic_bd = options.prohibit_non_basic_bd;
  _parameters.reduce_non_basic = options.reduce_non_basic;
  _parameters.input_non_basicness = options.input_non_basicness;
  _parameters.run_time_non_basicness = options.run_time_non_basicness;
  _parameters.selection = options.selection;
  _parameters.literal_comparison_mode = options.literal_comparison_mode;
  _parameters.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches
  _parameters.handle_equality = options.handle_equality;
  _parameters.forward_subsumption = options.forward_subsumption;
  _parameters.fs_set_mode = options.fs_set_mode; 
  _parameters.fs_use_sig_filters = options.fs_use_sig_filters; 
  _parameters.simplify_by_fs = options.simplify_by_fs;
  _parameters.main_premise_subsumption_expiry =
    (options.forward_subsumption) ? 
    options.main_premise_subsumption * (options.timeLimit/100)
    :
    0.0;
    
  _parameters.side_premise_subsumption_expiry =
    (options.forward_subsumption) ?  
    options.side_premise_subsumption * (options.timeLimit/100)
    :
    0.0;
      
  _parameters.main_selected_lit_demodulation_expiry = 
    (options.forward_demodulation) ?
    options.main_selected_lit_demodulation * (options.timeLimit/100)
    :
    0.0;
      
  _parameters.side_selected_lit_demodulation_expiry = 
    (options.forward_demodulation) ?
    options.side_selected_lit_demodulation * (options.timeLimit/100)
    :
    0.0;
      
  _parameters.forward_demodulation = options.forward_demodulation;
  _parameters.fd_on_splitting_branches = options.fd_on_splitting_branches;
  _parameters.normalize_by_commutativity = options.normalize_by_commutativity;
  _parameters.eq_taut_mod_C = options.eq_taut_mod_C;
  _parameters.prop_taut_mod_C = options.prop_taut_mod_C;
  _parameters.trivial_eq_res_mod_C = options.trivial_eq_res_mod_C;
  _parameters.identical_lit_factoring_mod_C = options.identical_lit_factoring_mod_C;
  _parameters.chaining_for_transitive = options.chaining_for_transitive;
  _parameters.chaining_through_variables = options.chaining_through_variables;
  _parameters.backward_demodulation = options.backward_demodulation;
  _parameters.bd_on_splitting_branches = options.bd_on_splitting_branches;
  _parameters.simplify_by_eq_res = options.simplify_by_eq_res;
  _parameters.backward_subsumption = options.backward_subsumption;
  _parameters.bs_set_mode = options.bs_set_mode;
  _parameters.orphan_murder = options.orphan_murder;
  _parameters.lrs = options.lrs;
  _parameters.pred_elim_table = options.pred_elim_table;

  _parameters.orient_pos_equalities_in_support = options.orient_pos_equalities_in_support;
  _parameters.orient_neg_equalities_in_support = options.orient_neg_equalities_in_support;
 
  _parameters.all_pos_eq_mode = options.all_pos_eq_mode;
  _parameters.all_neg_eq_mode = options.all_neg_eq_mode;
  

  _parameters.answer_predicate_table = options.answer_predicate_table;
  _parameters.answer_predicate_table_size = options.answer_predicate_table_size;
  // numeric parameters  
  _parameters.age_weight_ratio_numerator = options.age_weight_ratio_numerator;
  _parameters.age_weight_ratio_denominator = options.age_weight_ratio_denominator;
  _parameters.best_good_weight_ratio_numerator = options.best_good_weight_ratio_numerator;
  _parameters.best_good_weight_ratio_denominator = options.best_good_weight_ratio_denominator;

  _parameters.timeLimit = options.timeLimit;
  _parameters.timeLimitForLRS = 
    (long)(((float)options.timeLimit) * options.lrs_softness); 
  _parameters.memoryLimit = options.memoryLimit;
  _parameters.allocation_buffer_price = options.allocation_buffer_price;
  _parameters.maxActive = options.maxActive;
  _parameters.maxPassive = options.maxPassive;
  _parameters.maxWeight = (options.maxWeight <= ((long)VampireKernelConst::MaxClauseSize)) ? options.maxWeight : VampireKernelConst::MaxClauseSize;
  _parameters.previousMaxWeight = options.maxWeight;
  _parameters.maxMaxWeight = options.maxWeight;
  _parameters.maxInferenceDepth = options.maxInferenceDepth;
  _parameters.first_time_check = options.first_time_check; 
  _parameters.first_mem_check = options.first_mem_check;   
  _parameters.reachable = LONG_MAX;
  _parameters.unreachable = (-1)*_halfLongMax;     
  // interface parameters
  _parameters.outputStreamForGenerated = options.outputStreamForGenerated;
  _parameters.showPassive = options.showPassive;
  _parameters.showActive = options.showActive;
  _parameters.showLitProp = options.showLitProp;
  _parameters.outputStreamForPassiveProfiles = options.outputStreamForPassiveProfiles;
  _parameters.showPassiveProfileInterval = options.showPassiveProfileInterval;
  _parameters.showWeightChange = options.showWeightChange;
  _parameters.showUsagePace = options.showUsagePace;
  // auxilliary
  _parameters.useClauseNumIndex = options.backward_subsumption;

  
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _parameters.step_resolution_literal_selection = options.step_resolution_literal_selection;
#endif
  
}; // void Master::initParameters()



void Master::passParametersToObjects()
  {
    // clause sets
    _passive.setAgeWeightRatio(_parameters.age_weight_ratio_numerator,
			       _parameters.age_weight_ratio_denominator);
    _passive.setBestGoodWeightRatio(_parameters.best_good_weight_ratio_numerator,
				    _parameters.best_good_weight_ratio_denominator);
    _pagedPassive.setAgeWeightRatio(_parameters.age_weight_ratio_numerator,
				    _parameters.age_weight_ratio_denominator);
    _pagedPassive.setBestGoodWeightRatio(_parameters.best_good_weight_ratio_numerator,
					 _parameters.best_good_weight_ratio_denominator);


    // literal selection etc.

    TemporaryLiteral::Context::getCurrent()->setInversePolarityForSelection(_parameters.selection < 0);
    TemporaryLiteral::Context::getCurrent()->setChainingTable(_parameters.chaining_for_transitive);
    TemporaryLiteral::Context::getCurrent()->setShowLiteralProperties(_parameters.showLitProp);
    TemporaryLiteral::Context::getCurrent()->setAllPosEqMode(_parameters.all_pos_eq_mode);
    TemporaryLiteral::Context::getCurrent()->setAllNegEqMode(_parameters.all_neg_eq_mode);

    _standardLiteralOrdering.setMode(_parameters.literal_comparison_mode);

    _litSelection.setSplittingFlag(_parameters.splitting);
    _litSelection.setSplittingStyle(_parameters.splitting_style);
    _litSelection.setFunNum(_parameters.selection);
    _litSelection.setLiteralComparisonMode(_parameters.literal_comparison_mode);
    _litSelection.setStandardLiteralOrdering(&_standardLiteralOrdering);
    
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    _litSelection.setStepResolutionLiteralSelection(_parameters.step_resolution_literal_selection);
#endif
       
    // clause sharing 
    _clauseAssembler.setInheritNegativeSelection(_parameters.inherited_negative_selection);
    _clauseAssembler.setEqualityIsPossibleFlag(_parameters.handle_equality);
    _clauseAssembler.setEliminatedPredicateTable(_parameters.pred_elim_table);
    _clauseAssembler.setMaxMaxWeight(_parameters.maxMaxWeight);
    _clauseAssembler.setMaxWeight(_parameters.maxWeight);

    // temp clause collection and forward processing
    
    _inputClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
    _inputClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
    _inputClauseCollector.setSizeLimit(_parameters.maxWeight);

    _reanimatedClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
    _reanimatedClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);
    _reanimatedClauseCollector.setSizeLimit(VampireKernelConst::MaxClauseSize);

    _mainClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
    _mainClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
    _mainClauseCollector.setSizeLimit(_parameters.maxWeight);

    _mainClauseCollector.setMarkSubstitutionPositions(_parameters.run_time_non_basicness);

    // _propositionalSplitting

    if (_parameters.splitting)
      {
	switch (_parameters.splitting_style)
	  {
	  case 1L:
	    _propositionalSplitting.setStyle(PropositionalSplitting::HyperParallel);
	    _propositionalSplitting.setSmallBasicComponentInHyper(_parameters.small_basic_comp_in_hyp_splitting);
	    break;
	  case 2L:
	    _propositionalSplitting.setStyle(PropositionalSplitting::BinaryBlocking);
	    break;
	    
	  default:
	    BK::RuntimeError::report("Wrong splitting style number.");
	  };
      }; // if (_parameters.splitting)

    // _negativeEqualitySplitting

    if (_parameters.neg_eq_splitting > 0L)
      {
	_negativeEqualitySplitting.setMinimalEligibleArgSize(_parameters.neg_eq_splitting + 1);
	_negativeEqualitySplitting.setMaximalEligibleVarCount(_parameters.neg_eq_spl_var_count_limit);
      }
    else
      {
	_negativeEqualitySplitting.setMinimalEligibleArgSize(ULONG_MAX);
      };

    // _forwardProcessing


    _forwardProcessing.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
    _forwardProcessing.setShowCompleteClauses(_parameters.outputStreamForGenerated);  

    _forwardProcessing.setEqualityPossible(_parameters.handle_equality);
    if (_parameters.forward_subsumption)
      {
	_forwardProcessing.setForwardSubsumption(&_forwardSubsumption);
      }
    else
      _forwardProcessing.setForwardSubsumption(0);
    _forwardProcessing.setForwardSubsumptionResolution(_parameters.simplify_by_fs);

    if (_parameters.forward_demodulation > 0)
      {
	_forwardProcessing.setForwardDemodulation(&_forwardDemodulation);
	_forwardProcessing.setOnBranchForwardDemodulation(_parameters.fd_on_splitting_branches);
      }
    else
      _forwardProcessing.setForwardDemodulation(0);
    _forwardProcessing.setDestructiveEqResolution(_parameters.simplify_by_eq_res); 
    if (_parameters.splitting)
      {
	_forwardProcessing.setPropositionalSplitting(&_propositionalSplitting);
      }
    else 
      _forwardProcessing.setPropositionalSplitting(0);


    if (_parameters.neg_eq_splitting > 0L)
      {
	_forwardProcessing.setNegativeEqualitySplitting(&_negativeEqualitySplitting);
      }
    else 
      _forwardProcessing.setNegativeEqualitySplitting(0);
    _forwardProcessing.setCommutativityTableForNormalisation(_parameters.normalize_by_commutativity);
    _forwardProcessing.setCommutativityTableForEqTaut(_parameters.eq_taut_mod_C);
    _forwardProcessing.setCommutativityTableForPropTaut(_parameters.prop_taut_mod_C);
    _forwardProcessing.setCommutativityTableForTrivialEqRes(_parameters.trivial_eq_res_mod_C);
    _forwardProcessing.setCommutativityTableForIdenticalLitFactoring(_parameters.identical_lit_factoring_mod_C);
    
    _forwardProcessing.setProhibitNonBasicDemodulation(_parameters.prohibit_non_basic_fd ||
						       _parameters.reduce_non_basic);
    _forwardProcessing.setReduceNonBasic(_parameters.reduce_non_basic);
    _forwardProcessing.setUseInputNonBasicness(_parameters.input_non_basicness);
    _forwardProcessing.setUseRunTimeNonBasicness(_parameters.run_time_non_basicness);
    
    _forwardProcessing.setClauseSizeLimit(_parameters.maxWeight);
    _forwardProcessing.setInferenceDepthLimit(_parameters.maxInferenceDepth);


    // _forwardProcessingForSupport


    _forwardProcessingForSupport.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
    _forwardProcessingForSupport.setShowCompleteClauses(_parameters.outputStreamForGenerated);  

    _forwardProcessingForSupport.setEqualityPossible(_parameters.handle_equality);
    _forwardProcessingForSupport.setForwardSubsumption(0);
    _forwardProcessingForSupport.setForwardSubsumptionResolution(0UL);
    _forwardProcessingForSupport.setForwardDemodulation(0);
    _forwardProcessingForSupport.setDestructiveEqResolution(false); 
    _forwardProcessingForSupport.setPropositionalSplitting(0);
    _forwardProcessingForSupport.setNegativeEqualitySplitting(0);
    _forwardProcessingForSupport.setCommutativityTableForNormalisation(0);
    _forwardProcessingForSupport.setClauseSizeLimit(_parameters.maxWeight);
    _forwardProcessingForSupport.setInferenceDepthLimit(_parameters.maxInferenceDepth);

    // _forwardSubsumption      
    _forwardSubsumption.SetSetMode(_parameters.fs_set_mode);
    _forwardSubsumption.SetUseSigFilters(_parameters.fs_use_sig_filters);
    _forwardSubsumption.setSubsumptionResolutionMinLength(_parameters.simplify_by_fs);
    // _backSubsumption

    _backSubsumption.setSetMode(_parameters.bs_set_mode);

    // forward demodulation
  
    _forwardDemodulation.SetDegree(_parameters.forward_demodulation);
    _forwardDemodulation.SetFDOnSplittingBranchesFlag(_parameters.fd_on_splitting_branches);
    _forwardDemodulation.setDemodulateFromSupport(_parameters.orient_pos_equalities_in_support);

    // _backDemodulation
    _backDemodulation.setClauseCollector(&_mainClauseCollector);
    _backDemodulation.setDegree(_parameters.backward_demodulation);
    _backDemodulation.setBDOnSplittingBranchesFlag(_parameters.bd_on_splitting_branches);
    _backDemodulation.setDemodulateFromSupport(_parameters.orient_pos_equalities_in_support);
    _backDemodulation.setProhibitDemodulationIntoInputNonBasic(_parameters.input_non_basicness &&
							       (_parameters.prohibit_non_basic_bd ||
								_parameters.reduce_non_basic));
    _backDemodulation.setCheckSubstSafetyWRTInputNonBasicness(_parameters.input_non_basicness);

    // _infDispatcher
    _infDispatcher.setClauseCollector(&_mainClauseCollector);
    _infDispatcher.setForwardProcessing(&_forwardProcessing);
    _infDispatcher.binResFlag = _parameters.binary_resolution;
    _infDispatcher.setNoInfBetweenRulesFlag(_parameters.no_inferences_between_rules);
    _infDispatcher.setParamodulationFlag(_parameters.paramodulation);
    _infDispatcher.setParamodulateFromVariablesFlag(_parameters.paramodulate_from_variables);
    _infDispatcher.setSimBackSup(_parameters.sim_back_sup);
    _infDispatcher.setForwSupNoDuplicateRedexes(_parameters.forw_sup_no_dup_red);
    _infDispatcher.setBackSupNoDuplicateRedexes(_parameters.back_sup_no_dup_red);
    _infDispatcher.setUseInputNonBasicness(_parameters.input_non_basicness &&
					   _parameters.prohibit_non_basic_superposition);
    _infDispatcher.setChainingTable(_parameters.chaining_for_transitive);
    _infDispatcher.setChainingThroughVariablesFlag(_parameters.chaining_through_variables);

    _infDispatcher.setMainPremiseSubsumption(_parameters.main_premise_subsumption_expiry > 0.001);
    _infDispatcher.setSidePremiseSubsumption(_parameters.side_premise_subsumption_expiry > 0.001);
    _infDispatcher.setMainSelectedLitDemodulation(_parameters.main_selected_lit_demodulation_expiry > 0.001);
    _infDispatcher.setSideSelectedLitDemodulation(_parameters.side_selected_lit_demodulation_expiry > 0.001);
    _infDispatcher.setOrientEqualitiesInSupport(_parameters.orient_pos_equalities_in_support,
						_parameters.orient_pos_equalities_in_support);
    _infDispatcher.setClauseSizeLimit(_parameters.maxWeight);
    
    

  }; // void Master::passParametersToObjects()





//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#  define DEBUG_NAMESPACE "Master::Options"
#endif                  
#include "debugMacros.hpp"
//====================================================


void Master::Options::reset()
{
  incompleteAPriori = false;
  // main algorithm
  main_alg = 0;
 
  // inference rules
  binary_resolution = true;
  no_inferences_between_rules = false;
  splitting = false;  
  splitting_style = 1L;
  small_basic_comp_in_hyp_splitting = false;
  neg_eq_splitting = 0L; 
  neg_eq_spl_var_count_limit = 0L; 
  paramodulation = false;
  paramodulate_from_variables = true;
  sim_back_sup = 0;
  forw_sup_no_dup_red = false;
  back_sup_no_dup_red = false;
  prohibit_non_basic_superposition = false;
  prohibit_non_basic_fd = true;
  prohibit_non_basic_bd = true;
  reduce_non_basic = false;
  input_non_basicness = false;
  run_time_non_basicness = false;
  selection = 1;
  literal_comparison_mode = 1;
  inherited_negative_selection = false;

  // strategy switches
  handle_equality = false;
  forward_subsumption = true;
  fs_set_mode = false;
  fs_use_sig_filters = false;
  simplify_by_fs = 0;
  main_premise_subsumption = 0;
  side_premise_subsumption = 0;
  main_selected_lit_demodulation = 0;
  side_selected_lit_demodulation = 0;
  forward_demodulation = 0;
  fd_on_splitting_branches = false;
  normalize_by_commutativity = 0;
  eq_taut_mod_C = 0;
  prop_taut_mod_C = 0;
  trivial_eq_res_mod_C = 0;
  identical_lit_factoring_mod_C = 0;
  chaining_for_transitive = 0;
  chaining_through_variables = true;
  backward_demodulation = 0;
  bd_on_splitting_branches = false;
  simplify_by_eq_res = false;
  backward_subsumption = false;
  bs_set_mode = false;
  orphan_murder = 0;
  lrs = 0;
  pred_elim_table = 0;

  orient_pos_equalities_in_support = true;
  orient_neg_equalities_in_support = true;

  all_pos_eq_mode = LiteralProperties::DefaultEqualityMode;
  all_neg_eq_mode = LiteralProperties::DefaultEqualityMode;
  

  answer_predicate_table = 0;
  answer_predicate_table_size = 0UL;
  // numeric options
  age_weight_ratio_numerator = 1;
  age_weight_ratio_denominator = 0;
  best_good_weight_ratio_numerator = 1;
  best_good_weight_ratio_denominator = 0;
  timeLimit = LONG_MAX/10;
  memoryLimit = (float)LONG_MAX;
  allocation_buffer_price = 1.0;
  maxActive = LONG_MAX;
  maxPassive = LONG_MAX;
  maxWeight = VampireKernelConst::MaxClauseSize;
  maxInferenceDepth = LONG_MAX;
  lrs_softness = 1.0;
  first_time_check = 1.0;
  first_mem_check = 1.0;
  // interface options
  outputStreamForGenerated = 0;
  showPassive = 0;
  showActive = 0;
  showLitProp = false;
  outputStreamForPassiveProfiles = 0;
  showPassiveProfileInterval = LONG_MAX;
  showWeightChange = 0;
  showUsagePace = 0;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  step_resolution_literal_selection = false;
#endif
}; // void Master::Options::reset()



//====================================================
