//
// File:         PagedPassiveClauseStorage.hpp
// Description:  Page-queue based storage for passive clauses. 
// Created:      Apr 27, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//====================================================
#ifndef PAGED_PASSIVE_CLAUSE_STORAGE_H
//=================================================
#define PAGED_PASSIVE_CLAUSE_STORAGE_H
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "BitWord.hpp"
#include "PageQueueDisciplineStorage.hpp"
#include "Term.hpp"
#include "ClauseBackground.hpp"
#include "Clause.hpp"
#include "LiteralProperties.hpp"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.hpp"
//=================================================  

namespace VK
{
  class PagedPassiveClauseStorage
  {
  public: 
    class ClauseProfile
    {
    public:
      ClauseProfile() {};
      ~ClauseProfile() {};
      long clauseNumber() const { return _clauseNumber; };
      long clauseSize() const { return _clauseSize; };
      long clauseWeight() const { return _clauseWeight; };
      const BK::BitWord& rules() const { return _rules; };
      ulong numOfLiterals() const { return _numOfLiterals; };
      ulong numOfAncestors() const { return _numOfAncestors; };
      ostream& output(ostream& str) const;
    private:
      long _clauseNumber;
      long _clauseSize;
      long _clauseWeight;
      BK::BitWord _rules;
      ulong _numOfLiterals;
      ulong _numOfAncestors;
      friend class PagedPassiveClauseStorage;
    }; // class ClauseProfile
  
    class ClauseLiterals
    {
    public:
      ClauseLiterals() {};
      ~ClauseLiterals() {};
      bool nextLiteral();
      bool nextSymbol(TERM& sym);
      const LiteralProperties& currentLiteralProperties() const;
      void reset();
    
    private:
      const TERM* _beginning;
      ulong _numberOfLiterals;
      const TERM* _cursor;
      ulong _numOfRemainingLiterals;
      ulong _numOfRemainingSymbols;
      LiteralProperties _currentLiteralProperties;
      friend class PagedPassiveClauseStorage;
    }; // class ClauseLiterals

    class ClauseAncestors
    {
    public:
      ClauseAncestors() {};
      ~ClauseAncestors() {};
      bool nextAncestor(Clause*& anc);
      void reset();
    private:
      Clause** _beginning;
      ulong _numberOfAncestors;
      Clause** _cursor;
      ulong _numOfRemainingAncestors;
      friend class PagedPassiveClauseStorage;
    }; // class ClauseAncestors

    class StoredClause
    {
    public:   
      StoredClause() {};
      ~StoredClause() {};
      ClauseProfile& profile() { return _profile; };
      ClauseLiterals& literals() { return _literals; };
      ClauseAncestors& ancestors() { return _ancestors; };

      long clauseNumber() const { return _profile.clauseNumber();  };
      long clauseSize() const { return _profile.clauseSize();  };
      const BK::BitWord& rules() const { return _profile.rules();  };
      ulong numOfLiterals() const { return  _profile.numOfLiterals(); };
      ulong numOfAncestors() const { return  _profile.numOfAncestors(); };   

      void resetLiterals() { _literals.reset(); };
      bool nextLiteral() { return _literals.nextLiteral(); };
      bool nextSymbol(TERM& sym) { return _literals.nextSymbol(sym); };

      void resetAncestors() { _ancestors.reset(); };
      bool nextAncestor(Clause*& anc) { return _ancestors.nextAncestor(anc); };

      bool isVIP() const { return rules().bit(ClauseBackground::VIP); };    
      bool isOrphan();
      ostream& output(ostream& str);    
    
    private: 
      ClauseProfile _profile;
      ClauseLiterals _literals;
      ClauseAncestors _ancestors;
      friend class PagedPassiveClauseStorage;
    }; // class StoredClause

    class StoredClauseHandle
    {
    public:
      StoredClauseHandle() : _memory(0) {};
      ~StoredClauseHandle() {};
      operator bool() const { return _memory != 0; };
      void clear() { _memory = 0; };
      void set(void* mem) { _memory = mem; };
      void* memory() const { return _memory; };
    private:
      void* _memory;
      friend class PagedPassiveClauseStorage;
    }; // class StoredClauseHandle 

    class Iterator
    {
    public:
      Iterator() {};
      ~Iterator() {};
      void reset(PagedPassiveClauseStorage& storage);
      bool nextClause();
      StoredClause& currentClause() { return _currentClause; };   
    private:
      PagedPassiveClauseStorage* _storage;
      StoredClause _currentClause;
      long _nextCategoryNumber; 
      void* _nextObject; 
    }; // class Iterator  


    class ConstIterator
    {
    public:
      ConstIterator() {};
      ~ConstIterator() {};
      void reset(const PagedPassiveClauseStorage& storage);
      bool nextClause();
      const StoredClause& currentClause() { return _currentClause; };   
    private:
      const PagedPassiveClauseStorage* _storage;
      StoredClause _currentClause;
      long _nextCategoryNumber; 
      const void* _nextObject; 
    }; // class ConstIterator  
 

  public:
    PagedPassiveClauseStorage();
    ~PagedPassiveClauseStorage();
    void reset(); 
  
    static ulong pageSize() // bytes 
    { 
      return PageSize*sizeof(void*); 
    }; 

    void setAgeWeightRatio(long numerator,long denominator);
    void setBestGoodWeightRatio(long numerator,long denominator);
    void maximalPriorityOn() { _maximalPriorityFlag = true; };
    void maximalPriorityOff() { _maximalPriorityFlag = false; };

    // Properties 
 
    bool empty() const { return _maxNonemptyCatNum < _minNonemptyCatNum; };
    bool nonempty() const { return _maxNonemptyCatNum >= _minNonemptyCatNum; };
    long totalNumberOfClauses() const { return _totalNumberOfClauses; };
    long numberOfSelectedClauses() const { return _numberOfSelectedClauses; };
    long numberOfCancelledClauses() const { return _numberOfCancelledClauses; };
    long numberOfCurrentlyKeptClauses() const 
    { 
      return _totalNumberOfClauses - _numberOfSelectedClauses - _numberOfCancelledClauses;
    };
    long maxPresentWeight() const { return _maxNonemptyCatNum; };

    // Writing new clauses

    void  openClause(StoredClauseHandle& handle,long clNum,long clSize,
		     long clWeight,const BK::BitWord rules,ulong numOfLits,
		     ulong numOfAncestors); 
    bool openClause(long clNum,long clSize,long clWeight,
		    const BK::BitWord rules,ulong numOfLits,ulong numOfAncestors);
    void openLiteral(const LiteralProperties& prop);
    void writeSymbol(const TERM& sym);
    void closeLiteral();
    void writeAncestor(const Clause* ancestor);
    void endOfClause();


    // Selection and popping the selected clause. 
    bool select(StoredClauseHandle& selectedClauseHandle);  

    bool select(StoredClause& selectedClause);

    void popSelected();
       
    // Cancelling clauses at the end of the queue and recycling pages. 

    bool cancelWorstClause();

    bool recyclePage()
    {
      CALL("recyclePage()");
      return _freePages.recyclePage();
    }; // bool recyclePage()

    static void loadClause(const StoredClauseHandle& handle,StoredClause& clause);

    ostream& outputProfile(ostream& str) const;
  private: 
    enum
    {
      PageSize = 16384UL // in machine words, <=> 64Kb
    };
    // Make sure that PageSize qualifies for large object in the allocator.
    // It should also be big enough to accomodate any large dynamically 
    // allocated object in the Vampire kernel. 
    typedef BK::PageQueueDisciplineStorage<BK::GlobAlloc,PageSize,PagedPassiveClauseStorage> Storage;
  private:
    long categoryNum(long clauseWeight);
    static ulong sizeInWords(ulong sizeInBytes);
    static void selectFrom(Storage* cat,StoredClauseHandle& selectedClauseHandle);
    static void readFromMemory(const void* memory,ClauseProfile& prof,ClauseLiterals& lits,ClauseAncestors& ancestors);

  private:
    enum 
    {
      MaximalCategoryNumber = VampireKernelConst::MaxClauseSize
    };
  private:
    // parameters 
    long _ageWeightRatioNumerator;   
    long _ageWeightRatioDenominator;   
    long _bestGoodWeightNumerator;
    long _bestGoodWeightDenominator;
    bool _maximalPriorityFlag; 

    // statistics
    long _minNonemptyCatNum; 
    long _maxNonemptyCatNum;
    long _totalSelectedByAgeOrWeight;
    long _totalSelectedByWeight;

    long _totalNumberOfClauses;
    long _numberOfSelectedClauses;  
    long _numberOfCancelledClauses; 
    long _numberOfCurrentlyKept[MaximalCategoryNumber + 1];

    // storage
    Storage::FreePages _freePages;
    Storage _category[MaximalCategoryNumber + 1];

    Storage* _currentCategory;
    long _selectedCategoryNumber;
    StoredClauseHandle _selectedClauseHandle;
    bool _lastSelectedByAge;
    bool _lastSelectedByWeight;
    void** _cursor;
    TERM* _currentLiteral;
    ulong _currentLiteralSize;
#ifdef DEBUG_NAMESPACE
    bool _debugClauseOpenForWriting; 
    void** _debugEndOfCurrentClause;
    ulong _debugNumberOfLiteralsToWrite;
    ulong _debugNumberOfAncestorsToWrite;
#endif  
    friend class Iterator;
    friend class ConstIterator;
  }; // class PagedPassiveClauseStorage

}; // namespace VK
//====================================================
#endif
