//
// File:         PagedPassiveClauseStorage.inl
// Description:  Page-queue based storage for passive clauses. 
// Created:      Apr 27, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//====================================================
#ifndef PAGED_PASSIVE_CLAUSE_STORAGE_INL
#  define PAGED_PASSIVE_CLAUSE_STORAGE_INL
//=================================================
#include "PagedPassiveClauseStorage.hpp"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.hpp"
//=================================================  

namespace VK
{
inline
void PagedPassiveClauseStorage::reset() 
{ 
  CALL("reset()"); 
  BK::destroy(*this); 
  BK::init(*this); 
}; 

inline
bool PagedPassiveClauseStorage::openClause(long clNum,long clSize,long clWeight,
					   const BK::BitWord rules,ulong numOfLits,ulong numOfAncestors)
{
  CALL("openClause(long clNum,long clSize,long clWeight,const BK::BitWord rules,ulong numOfLits,ulong numOfAncestors)");
  StoredClauseHandle handle;
  openClause(handle,clNum,clSize,rules,numOfLits,numOfAncestors);
  return (bool)handle;
}; // bool PagedPassiveClauseStorage::openClause(long clNum,long clSize,long clWeight,const BK::BitWord rules,ulong numOfLits,ulong numOfAncestors)
 
inline
void PagedPassiveClauseStorage::openLiteral(const LiteralProperties& prop)
{
  CALL("openLiteral(const LiteralProperties& prop)");
  ASSERT(_debugClauseOpenForWriting); 
  ASSERT(_cursor + 2*sizeInWords(sizeof(TERM)) < _debugEndOfCurrentClause);
  ASSERT(_debugNumberOfLiteralsToWrite);
  _currentLiteral = (TERM*)_cursor;
  _currentLiteralSize = 0UL;
  _cursor += sizeInWords(sizeof(TERM));
  ((TERM*)_cursor)->Make(prop.content());
  _cursor += sizeInWords(sizeof(TERM));
}; // void PagedPassiveClauseStorage::openLiteral(const LiteralProperties& prop)

inline
void PagedPassiveClauseStorage::writeSymbol(const TERM& sym) 
{
  CALL("writeSymbol(const TERM& sym)");
  ASSERT(_debugNumberOfLiteralsToWrite);
  ASSERT(_cursor + sizeInWords(sizeof(TERM)) <= _debugEndOfCurrentClause);
  *((TERM*)_cursor) = sym;
  _cursor += sizeInWords(sizeof(TERM));
  _currentLiteralSize++;
}; // void PagedPassiveClauseStorage::writeSymbol(const TERM& sym) 
  
inline
void PagedPassiveClauseStorage::closeLiteral()
{ 
  CALL("closeLiteral()");
  ASSERT(_debugNumberOfLiteralsToWrite);
  DOP(_debugNumberOfLiteralsToWrite--);
  _currentLiteral->Make(_currentLiteralSize);       
}; // void closeLiteral()

inline
void PagedPassiveClauseStorage::writeAncestor(const Clause* ancestor)
{
  CALL("writeAncestor(const Clause* ancestor)");
  ASSERT(_debugClauseOpenForWriting); 
  ASSERT(!_debugNumberOfLiteralsToWrite);
  ASSERT(_debugNumberOfAncestorsToWrite);
  ASSERT(_cursor + 1 <= _debugEndOfCurrentClause);
  *_cursor = (void*)ancestor;     
  _cursor++;
  DOP(_debugNumberOfAncestorsToWrite--);
}; // void writeAncestor(const Clause* ancestor)

inline
void PagedPassiveClauseStorage::endOfClause()
{
  CALL("endOfClause()");
  ASSERT(_debugClauseOpenForWriting);  
  DOP(_debugClauseOpenForWriting = false);
  ASSERT(!_debugNumberOfLiteralsToWrite);
  ASSERT(!_debugNumberOfAncestorsToWrite);
  ASSERT(!empty()); 
  ASSERT(_category[_minNonemptyCatNum].nonempty());
  ASSERT(_category[_maxNonemptyCatNum].nonempty());
}; // void endOfClause()

inline
void PagedPassiveClauseStorage::loadClause(const StoredClauseHandle& handle,StoredClause& clause)
{
  CALL("loadClause(const StoredClauseHandle& handle,StoredClause& clause)");
  ASSERT((bool)handle);
  readFromMemory(handle.memory(),
		 clause.profile(),
		 clause.literals(),
		 clause.ancestors());
}; // void loadClause(const StoredClauseHandle& handle,StoredClause& clause)


inline
long PagedPassiveClauseStorage::categoryNum(long clauseWeight) 
{
  if (clauseWeight > (long)MaximalCategoryNumber) return MaximalCategoryNumber;
  if (clauseWeight < 0L) return 0L;
  return clauseWeight;
};



inline
ulong PagedPassiveClauseStorage::sizeInWords(ulong sizeInBytes) 
{
  return (sizeInBytes + sizeof(void*) - 1) / sizeof(void*);
}; // ulong sizeInWords(ulong sizeInBytes)  

inline
void PagedPassiveClauseStorage::selectFrom(Storage* cat,StoredClauseHandle& selectedClauseHandle)
{
  CALL("selectFrom(Storage* cat,StoredClauseHandle& selectedClauseHandle)");
  ASSERT(cat);
  ASSERT(cat->nonempty());      
  selectedClauseHandle.set(Storage::objectContent(cat->head()));
}; // void selectFrom(Storage* cat,StoredClauseHandle& selectedClauseHandle)

}; // namespace VK 

//=========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::ClauseLiterals"
#endif 
#include "debugMacros.hpp"
//=========================================================================


namespace VK
{
inline
bool PagedPassiveClauseStorage::ClauseLiterals::nextLiteral() 
{
  CALL("nextLiteral()");
  ASSERT(!_numOfRemainingSymbols);
  if (_numOfRemainingLiterals)
    {
      _numOfRemainingLiterals--;
      _numOfRemainingSymbols = _cursor->Content();
      _cursor++;
      _currentLiteralProperties.make(_cursor->Content());
      _cursor++;
      return true;
    }
  else return false;   
};

 
inline
bool PagedPassiveClauseStorage::ClauseLiterals::nextSymbol(TERM& sym) 
{
  CALL("nextSymbol(TERM& sym)");
  if (_numOfRemainingSymbols) 
    {
      _numOfRemainingSymbols--;
      sym = *_cursor;
      _cursor++;
      return true;
    }
  else return false;
};




inline
void PagedPassiveClauseStorage::ClauseLiterals::reset() 
{
  CALL("reset()");    
  _cursor = _beginning;
  _numOfRemainingLiterals = _numberOfLiterals;
  _numOfRemainingSymbols = 0;
};
    
inline
const LiteralProperties& 
PagedPassiveClauseStorage::ClauseLiterals::currentLiteralProperties() const
{
  CALL("currentLiteralProperties() const");
  return _currentLiteralProperties;
};

}; // namespace VK 

//=========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::ClauseAncestors"
#endif 
#include "debugMacros.hpp"
//=========================================================================

namespace VK
{

inline
bool PagedPassiveClauseStorage::ClauseAncestors::nextAncestor(Clause*& anc) 
{
  CALL("nextAncestor(Clause*& anc)");
  if (_numOfRemainingAncestors)
    {
      _numOfRemainingAncestors--;
      anc = *_cursor;
      _cursor++;
      return true;
    }
  else return false;
};

inline
void PagedPassiveClauseStorage::ClauseAncestors::reset()
{
  CALL("reset()");    
  _cursor = _beginning;
  _numOfRemainingAncestors = _numberOfAncestors;
};

}; // namespace VK 


//=========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::StoredClause"
#endif 
#include "debugMacros.hpp"
//=========================================================================

namespace VK
{
inline
bool PagedPassiveClauseStorage::StoredClause::isOrphan() 
{
  CALL("isOrphan()");
  resetAncestors();
  Clause* ancestor;
  while (nextAncestor(ancestor)) 
    if (!ancestor->Active()) return true;
  return false;  
};

}; // namespace VK 


//=========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::Iterator"
#endif 
#include "debugMacros.hpp"
//=========================================================================



namespace VK
{

inline
void PagedPassiveClauseStorage::Iterator::reset(PagedPassiveClauseStorage& storage)
{
  CALL("reset(PagedPassiveClauseStorage& storage)");
  _storage = &storage;
  _nextCategoryNumber = _storage->_minNonemptyCatNum + 1;
  _nextObject = _storage->_category[_storage->_minNonemptyCatNum].head();
};  
 

inline
bool PagedPassiveClauseStorage::Iterator::nextClause() 
{
  CALL("nextClause()");
  while (!_nextObject)
    {
      // try next category
      if (_nextCategoryNumber > _storage->_maxNonemptyCatNum) return false;      
      _nextObject = _storage->_category[_nextCategoryNumber].head();
      _nextCategoryNumber++; 
    };    

  PagedPassiveClauseStorage::readFromMemory(PagedPassiveClauseStorage::Storage::objectContent(_nextObject),
					    _currentClause.profile(),
					    _currentClause.literals(),
					    _currentClause.ancestors());
  _nextObject = Storage::nextObject(_nextObject);  
  return true;
}; // bool PagedPassiveClauseStorage::Iterator::nextClause() 

}; // namespace VK 


//=========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::ConstIterator"
#endif 
#include "debugMacros.hpp"
//=========================================================================

namespace VK
{
inline
void PagedPassiveClauseStorage::ConstIterator::reset(const PagedPassiveClauseStorage& storage)
{
  CALL("reset(const PagedPassiveClauseStorage& storage)");
  _storage = &storage;
  _nextCategoryNumber = _storage->_minNonemptyCatNum + 1;
  _nextObject = _storage->_category[_storage->_minNonemptyCatNum].head();
};  
 

inline
bool PagedPassiveClauseStorage::ConstIterator::nextClause() 
{
  CALL("nextClause()");
  while (!_nextObject)
    {
      // try next category
      if (_nextCategoryNumber > _storage->_maxNonemptyCatNum) return false;      
      _nextObject = _storage->_category[_nextCategoryNumber].head();
      _nextCategoryNumber++; 
    };    

  PagedPassiveClauseStorage::readFromMemory(PagedPassiveClauseStorage::Storage::objectContent(_nextObject),
					    _currentClause.profile(),
					    _currentClause.literals(),
					    _currentClause.ancestors());
  _nextObject = Storage::nextObject(_nextObject);  
  return true;
}; // bool PagedPassiveClauseStorage::ConstIterator::nextClause() 

}; // namespace VK 




//====================================================
#endif
