//
// File:         Parser.cpp
// Description:  TPTP parser front end.
// Created:      Jun 04, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#include <cstdlib>
#include <iostream>
#include "Malloc.hpp"
#include "Parser.hpp"
using namespace std;
using namespace BK;
using namespace VK;
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PARSER
 #define DEBUG_NAMESPACE "Parser"
#endif
#include "debugMacros.hpp"
//======================================================================


bool Parser::parseFile(const char* tptpDir,const char* fileName)
{
  CALL("parseFile(const char* tptpDir,const char* fileName)");
  PR_FILE_PARS_RESULT* parsRes = new PR_FILE_PARS_RESULT();  
  ::ParseFile(fileName,parsRes);
  if (parsRes->Error())
    {      
      cout << "Error in " << fileName << " ";
      cout << "[" << parsRes->ErrorLineNum() << "/" << parsRes->ErrorCharNum() << "] : ";
      cout << parsRes->ErrorMessage() << "\n";
      return false;
    };


 for (PR_LITERAL_LIST* lst = parsRes->Literals(); lst; lst = lst->Next())
  if (!(processInputItem(lst->Literal(),tptpDir,fileName))) return false;
  
 delete parsRes;
 return true;   
}; // bool Parser::parseFile(const char* tptpDir,const char* fileName)
  

bool Parser::processInputItem(PR_LITERAL* item,
			      const char* tptpDir,
			      const char* fileName)
{
  CALL("processInputItem(PR_LITERAL* item,..)");
  switch (itemSort(item))
    {
    case TPTPIncludeDirective:
      {
	if (PR_TERM_LIST::Length(item->Arguments()) != 1)
	  {
	    cout << "Error in " << fileName << " "
		 << "[" << item->Header()->LineNum() << "/" 
		 << item->Header()->CharNum() << "] : "
		 << "Wrong arity. " << "\n";
	    return false; 
	  };
	if (item->Arguments()->Term()->Tag() != PR_TERM::ATOM_STRING)
	  {
	    cout << "Error in " << fileName << " "
		 << "[" << item->Header()->LineNum() << "/" 
		 << item->Header()->CharNum() << "] : "
		 << "String constant expected. " << "\n";
	    return false;
	  };
      
	const char* includedFileName = item->Arguments()->Term()->Symbol()->Symbol();
	char* fullFileName = 
	  static_cast<char*>(Malloc::allocate(strlen(tptpDir) + strlen(includedFileName) + 2));
	strcpy(fullFileName,tptpDir);
	fullFileName[strlen(tptpDir)] = '/';
	strcpy(fullFileName + strlen(tptpDir) + 1,includedFileName);
	fullFileName[strlen(tptpDir)  + strlen(includedFileName) + 1] = (char)0;
	bool result = parseFile(tptpDir,fullFileName);
	Malloc::deallocate(static_cast<void*>(fullFileName),strlen(tptpDir) + strlen(includedFileName) + 2);
	return result;
      };

    case TPTPInputClause:      
      if (!readClause(item)) 
	{	  
	  cout << "Error in " << fileName << "\n"
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read input clause. " << "\n";
	  return false;
	};
      return true;

    case TPTPInputLiteralDefinition:      
      if (!readLiteralDefinition(item)) 
	{	  
	  cout << "Error in " << fileName << "\n"
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read input literal definition. " << "\n";
	  return false;
	};
      return true;

    case TPTPInputFormula:
      cout << "Error in " << fileName << " "
	   << "[" << item->Header()->LineNum() << "/" 
	   << item->Header()->CharNum() << "] : "
	   << "Full first-order logic syntax is not supported.\n" 
	   << "Use a clausifier.\n";
      return false;


    case TPTPExtensionSymbolPrecedence:
      if (!readSymbolPrecedence(item))
	{
	  cout << "Error in " << fileName << " "
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read symbol precedence. " << "\n";
	  return false;
	};
      return true;

    case TPTPExtensionHeaderPrecedence:
      if (!readHeaderPrecedence(item))
	{
	  cout << "Error in " << fileName << " "
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read header precedence. " << "\n";
	  return false;
	};
      return true;

    case TPTPExtensionSymbolWeight:
      if (!readSymbolWeight(item))
	{
	  cout << "Error in " << fileName << " "
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read symbol weight. " << "\n";
	  return false;
	};
      return true;


    case TPTPExtensionSymbolInterpretation:
      if (!readSymbolInterpretation(item))
	{
	  cout << "Error in " << fileName << " "
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read symbol interpretation. " << "\n";
	  return false;
	};
      return true;


      
      
    case TPTPExtensionAnswerPredicateDeclaration:
      if (!readAnswerPredicateDeclaration(item))
	{
	  cout << "Error in " << fileName << " "
	       << "[" << item->Header()->LineNum() << "/" 
	       << item->Header()->CharNum() << "] : "
	       << "Can not read answer predicate declaration. " << "\n";
	  return false;
	};
      return true;

    default:      
      cout << "Error in " << fileName << " "
	   << "[" << item->Header()->LineNum() << "/" 
	   << item->Header()->CharNum() << "] : "
	   << "Unknown input item type. " << "\n";
      return false;
   };


}; // bool Parser::processInputItem(PR_LITERAL* item,..)


Parser::ItemSort Parser::itemSort(PR_LITERAL* item)
{
  CALL("itemSort(PR_LITERAL* item)");
  const char* top = item->Header()->Symbol();
  if (!(strcmp(top,"include"))) return TPTPIncludeDirective;
  if (!(strcmp(top,"input_clause"))) return TPTPInputClause;
  if (!(strcmp(top,"input_literal_definition"))) return TPTPInputLiteralDefinition;
  if (!(strcmp(top,"input_formula"))) return TPTPInputFormula;
  if (!(strcmp(top,"symbol_precedence"))) return TPTPExtensionSymbolPrecedence;
  if (!(strcmp(top,"header_precedence"))) return TPTPExtensionHeaderPrecedence;
  if (!(strcmp(top,"symbol_weight"))) return TPTPExtensionSymbolWeight;
  if (!(strcmp(top,"interpret"))) return TPTPExtensionSymbolInterpretation;
  if (!(strcmp(top,"answer"))) return TPTPExtensionAnswerPredicateDeclaration;
  return UnknownItemSort;
}; // Parser::ItemSort Parser::itemSort(PR_LITERAL* item)

Parser::ItemType Parser::itemType(PR_TERM* it)
{
  CALL("itemType(PR_TERM* it)");  
  if (it->Tag() != PR_TERM::ATOM_FUNCTOR) return UnknownItemType;
  const char* str = it->Symbol()->Symbol(); 
  if (!(strcmp(str,"axiom"))) return TPTPAxiom;
  if (!(strcmp(str,"conjecture"))) return TPTPConjecture;
  if (!(strcmp(str,"hypothesis"))) return TPTPHypothesis;
  return UnknownItemType;
}; //  Parser::ItemType Parser::itemType(PR_TERM* it)



bool Parser::readClause(PR_LITERAL* item)
{
  CALL("readClause(PR_LITERAL* item)");
  ASSERT(itemSort(item) == TPTPInputClause);
  if (PR_TERM_LIST::Length(item->Arguments()) != 3) return false; 
  if (PR_TERM_LIST::Nth(item->Arguments(),0)->Tag() != PR_TERM::LIST) return false;
  if ((PR_TERM_LIST::Nth(item->Arguments(),2)->Tag() != PR_TERM::ATOM_FUNCTOR) && 
      (PR_TERM_LIST::Nth(item->Arguments(),2)->Tag() != PR_TERM::ATOM_NUMBER))
    return false;
      
  const char* clauseName = 
    PR_TERM_LIST::Nth(item->Arguments(),2)->Symbol()->Symbol();

  Input::Clause::Annotation annotation;
  
  PR_TERM* clauseTypeTerm = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* clauseTypeBody;
  PR_TERM* clauseTypeAnnotation;

  if (clauseTypeTerm->Tag() == PR_TERM::ATOM_FUNCTOR)
    {
      clauseTypeBody = clauseTypeTerm;
      clauseTypeAnnotation = 0;
    }
  else  
    if (clauseTypeTerm->Tag() == PR_TERM::COMPLEX && 
	!strcmp(clauseTypeTerm->Symbol()->Symbol(),"<@"))
      {
	// annotated
	if (PR_TERM_LIST::Length(clauseTypeTerm->Arguments()) != 2)
	  {
	    cout << "Can not read input clause type with annotation.\n"; 
	    return false;
	  };
	clauseTypeBody = PR_TERM_LIST::Nth(clauseTypeTerm->Arguments(),1);
	clauseTypeAnnotation = PR_TERM_LIST::Nth(clauseTypeTerm->Arguments(),0);
      }
    else
      {
	cout << "Can not read input clause type.\n"; 
	return false;
      };

  ItemType clauseType = itemType(clauseTypeBody);
  if (clauseType == UnknownItemType) return false;
  
  if (clauseTypeAnnotation && 
      !readClauseAnnotation(clauseTypeAnnotation,annotation))
    {
      cout << "Can not read input clause annotation.\n"; 
      return false;
    };
  

  _signature.resetVariables();
  Input::LiteralList* literals;
  if (!readLiterals(PR_TERM_LIST::Nth(item->Arguments(),0)->Arguments(),literals)) 
    return false;

  Input::Clause* clause = new Input::Clause(clauseName,annotation,literals);
  switch (clauseType)
    {
    case TPTPAxiom: clause->setCategory(Input::Clause::Axiom); break;
    case TPTPConjecture: clause->setCategory(Input::Clause::Conjecture); break;
    case TPTPHypothesis: clause->setCategory(Input::Clause::Hypothesis); break;
    case UnknownItemType: clause->setCategory(Input::Clause::Unclassified); break;
    };
  
  _clauses = new Input::ClauseList(clause,_clauses);
  return true;
}; // bool Parser::readClause(PR_LITERAL* item)



bool Parser::readClauseAnnotation(PR_TERM* annotation,
				  Input::Clause::Annotation& result)
{
  CALL("readClauseAnnotation(PR_TERM* annotation,Input::Clause::Annotation& result)");
  if (annotation->Tag() != PR_TERM::LIST) return false;
  bool nonSubsuming = false;
  bool nonSubsumable = false;
  for (PR_TERM_LIST* lst = annotation->Arguments();
       lst;
       lst = lst->Next())
    {
      PR_TERM* entry = lst->Term();
      switch (entry->Tag())
	{
	case PR_TERM::ATOM_FUNCTOR: 
	  if (!strcmp(entry->Symbol()->Symbol(),"non_subsuming"))
	    {
	      if (nonSubsuming)
		{
		  cout << "Duplicate 'non_subsuming' entry in clause annotation.\n";
		  return false;
		};
	      nonSubsuming = true;
	    }
	  else if (!strcmp(entry->Symbol()->Symbol(),"non_subsumable"))
	    {
	      if (nonSubsumable)
		{
		  cout << "Duplicate 'non_subsumable' entry in clause annotation.\n";
		  return false;
		};
	      nonSubsumable = true;
	    }
	  else
	    {
	      cout << "Can not read input clause annotation entry:\n"
		   << "must be 'non_subsuming' or 'non_subsumable'.\n";
	      return false;
	    };
	  break;
	default:
	  cout << "Can not read input clause annotation entry:\n"
	       << "must be 'non_subsuming' or 'non_subsumable'.\n";
	  return false;
	};
    };
  
  if (nonSubsuming) result.markAsNonSubsuming();
  if (nonSubsumable) result.markAsNonSubsumable();

  return true;
}; // bool Parser::readClauseAnnotation(PR_TERM* annotation,Input::Clause::Annotation& result)



bool Parser::readLiteralDefinition(PR_LITERAL* item)
{
  CALL("readLiteralDefinition(PR_LITERAL* item)");
  ASSERT(itemSort(item) == TPTPInputLiteralDefinition);
  if (PR_TERM_LIST::Length(item->Arguments()) != 4) return false; 
  if ((PR_TERM_LIST::Nth(item->Arguments(),3)->Tag() != PR_TERM::ATOM_FUNCTOR) && 
      (PR_TERM_LIST::Nth(item->Arguments(),3)->Tag() != PR_TERM::ATOM_NUMBER))
    return false;
  const char* litDefName =
    PR_TERM_LIST::Nth(item->Arguments(),3)->Symbol()->Symbol();
  ItemType litDefType = itemType(PR_TERM_LIST::Nth(item->Arguments(),2));
  if (litDefType == UnknownItemType) return false;
  
  _signature.resetVariables();

  Input::Literal* head;
    
  if (!readLiteral(PR_TERM_LIST::Nth(item->Arguments(),1),head))
    return false;
  

  if (PR_TERM_LIST::Nth(item->Arguments(),0)->Tag() != PR_TERM::LIST) 
    {
      delete head;
      return false;
    };

  Input::LiteralListList* tail = 0;
  
  for (PR_TERM_LIST* lst = PR_TERM_LIST::Nth(item->Arguments(),0)->Arguments();
       lst;
       lst = lst->Next())
    {
      if (lst->Term()->Tag() != PR_TERM::LIST) 
	{	  
	  delete head;
	  Input::destroyLiteralListList(tail);
	  return false;
	};

      Input::LiteralList* literals;
      if (!readLiterals(lst->Term()->Arguments(),literals))
	{	  
	  delete head;
	  Input::destroyLiteralListList(tail);
	  return false;
	};
      tail = new Input::LiteralListList(literals,tail);      
    };

  Input::LiteralDefinition* literalDef = 
    new Input::LiteralDefinition(litDefName,head,tail);
  
  switch (litDefType)
    {
    case TPTPAxiom: literalDef->setCategory(Input::LiteralDefinition::Axiom); break;
    case TPTPConjecture: literalDef->setCategory(Input::LiteralDefinition::Conjecture); break;
    case TPTPHypothesis: literalDef->setCategory(Input::LiteralDefinition::Hypothesis); break;
    case UnknownItemType: literalDef->setCategory(Input::LiteralDefinition::Unclassified); break;
    };
  
  _literalDefinitions = 
    new Input::LiteralDefList(literalDef,_literalDefinitions);
  return true; 
}; // bool Parser::readLiteralDefinition(PR_LITERAL* item)



bool Parser::readLiterals(PR_TERM_LIST* lst,Input::LiteralList*& result)
{
  CALL("readLiterals(PR_TERM_LIST* lst,Input::LiteralList*& result)");
  if (!lst) { result = 0; return true; };
  Input::LiteralList* tail;
  if (!readLiterals(lst->Next(),tail)) return false; 
  
  Input::Literal* lit;
  if (!readLiteral(lst->Term(),lit))
    {
      Input::destroyLiteralList(tail);
      return false;
    };
  result = new Input::LiteralList(lit,tail);
  return true;
}; // bool Parser::readLiterals(PR_TERM_LIST* lst,Input::LiteralList*& result)

bool Parser::readLiteral(PR_TERM* lit,Input::Literal*& result)
{
  CALL("readLiteral(PR_TERM* lit,Input::Literal*& result)");
  Input::Literal::Annotation annotation;
  PR_TERM_LIST* annotationEntries;
  PR_TERM* litBody;

  if (strcmp(lit->Symbol()->Symbol(),"<@") == 0)
    {
      PR_TERM* markupTerm = 
	PR_TERM_LIST::Nth(lit->Arguments(),0);
      litBody = 
	PR_TERM_LIST::Nth(lit->Arguments(),1);
      if (markupTerm->Tag() != PR_TERM::LIST)
	{
	  cout << "Can not read literal annotation " 
	       << "in [" << lit->Symbol()->LineNum() << "/" 
	       << lit->Symbol()->CharNum() << "].\n";
	  return false;
	}; 
      annotationEntries = markupTerm->Arguments();
    }
  else
    {
      annotationEntries = 0;
      litBody = lit;
    };

  if (strcmp(litBody->Symbol()->Symbol(),"++") &&
      strcmp(litBody->Symbol()->Symbol(),"+++") &&
      strcmp(litBody->Symbol()->Symbol(),"--") &&
      strcmp(litBody->Symbol()->Symbol(),"---"))
    {
      cout << "Can not read literal polarity " 
	   << "in [" << lit->Symbol()->LineNum() << "/" 
	   << lit->Symbol()->CharNum() << "].\n";
      return false;
    };

  bool preselected = 
    (strcmp(litBody->Symbol()->Symbol(),"+++") == 0) ||
    (strcmp(litBody->Symbol()->Symbol(),"---") == 0);
    

  if (!readLiteralAnnotation(annotationEntries,preselected,annotation))
    {
      cout << "Can not read annotated literal " 
	   << "in [" << lit->Symbol()->LineNum() << "/" 
	   << lit->Symbol()->CharNum() << "].\n";
      return false;
    };

  bool isPositive = 
    (strcmp(litBody->Symbol()->Symbol(),"++") == 0) ||
    (strcmp(litBody->Symbol()->Symbol(),"+++") == 0);

  Input::Term* atom;
  if (!readAtom(litBody->Arguments()->Term(),atom)) return false;
  result = new Input::Literal(isPositive,annotation,atom);
  return true;
}; // bool Parser::readLiteral(PR_TERM* lit,Input::Literal*& result)

bool Parser::readAtom(PR_TERM* at,Input::Term*& result)
{
  CALL("readAtom(PR_TERM* at,Input::Term*& result)");
  PR_TERM_LIST* args = at->Arguments();
  ulong arity = PR_TERM_LIST::Length(args);
  Input::Symbol* pred = _signature.registerPredicate(at->Symbol()->Symbol(),
						     arity);
  if (!pred) return false;
  Input::TermList* terms;
  if (!readTerms(args,terms)) return false;
  result = new Input::Term(pred,terms);
  return true;
}; // bool Parser::readAtom(PR_TERM* at,Input::Term*& result)


bool Parser::readTerms(PR_TERM_LIST* terms,Input::TermList*& result)
{
  CALL("readTerms(PR_TERM_LIST* terms,Input::TermList*& result)");
  if (!terms) { result = 0; return true; };
  Input::TermList* tail;
  if (!readTerms(terms->Next(),tail)) return false;
  Input::Term* head;
  if (!readTerm(terms->Term(),head))
    {
      Input::destroyTermList(tail);
      return false;
    };
  result = new Input::TermList(head,tail);
  return true;
}; // bool Parser::readTerms(PR_TERM_LIST* terms,Input::TermList*& result)


bool Parser::readTerm(PR_TERM* term,Input::Term*& result)
{
  CALL("readTerm(PR_TERM* term,Input::Term*& result)");
  Input::Symbol* top;
  switch (term->Tag())
    {
    case PR_TERM::ATOM_FUNCTOR: // constant, no break here
    case PR_TERM::ATOM_NUMBER: // constant
      top = _signature.registerFunction(term->Symbol()->Symbol(),0UL);
      if (!top) return false;
      result = new Input::Term(top);
      return true;

    case PR_TERM::VAR:
      top = _signature.registerVariable(term->Symbol()->Symbol());
      if (!top) return false;
      result = new Input::Term(top);
      return true;

    case PR_TERM::COMPLEX:
      {
	PR_TERM_LIST* args = term->Arguments();
	ulong arity = PR_TERM_LIST::Length(args);
	top = _signature.registerFunction(term->Symbol()->Symbol(),arity);
	if (!top) return false;
	Input::TermList* terms;
	if (!readTerms(args,terms)) return false;
	result = new Input::Term(top,terms);
	return true;
      };
    default: return false;
    };

}; // bool Parser::readTerm(PR_TERM* term,Input::Term*& result)

bool Parser::readSymbolPrecedence(PR_LITERAL* item)
{
  CALL("readSymbolPrecedence(PR_LITERAL* item)");
  if (PR_TERM_LIST::Length(item->Arguments()) != 4) 
    {
      cout << "Exactly 4 arguments expected.\n";
      return false;
    };

  PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),3);
  PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),2);
  PR_TERM* arg3 = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* arg4 = PR_TERM_LIST::Nth(item->Arguments(),0);
  
  if ((arg1->Tag() != PR_TERM::ATOM_FUNCTOR) ||
      (strcmp(arg1->Symbol()->Symbol(),"predicate") &&       
       strcmp(arg1->Symbol()->Symbol(),"function")))
    {
      cout << "1st argument must be either 'predicate' or 'function'\n";
      return false;
    };
  
  if ((arg2->Tag() != PR_TERM::ATOM_FUNCTOR) &&
      (arg2->Tag() != PR_TERM::ATOM_NUMBER))
    {
      cout << "Symbolic atom or numeric constant expected as the 2nd argument.\n";
      return false;
    };
  if (arg3->Tag() != PR_TERM::ATOM_NUMBER) 
    {
      cout << "Integer expected as the 3rd argument.\n";
      return false;
    };

  if (arg4->Tag() != PR_TERM::ATOM_NUMBER) 
    {
      cout << "Integer expected as the 4th argument.\n";
      return false;
    };

  bool predicate = !strcmp(arg1->Symbol()->Symbol(),"predicate");
  long arity = strtol(arg3->Symbol()->Symbol(),0,10);
  long precedence = strtol(arg4->Symbol()->Symbol(),0,10);
  bool res = _signature.registerPrecedence(predicate,
					   arg2->Symbol()->Symbol(),
					   arity,
					   precedence);
  if (!res)
    {
      cout << "Inconsistent with some other settings.\n";
    };
  return res;
}; // bool Parser::readSymbolPrecedence(PR_LITERAL* item)




bool Parser::readHeaderPrecedence(PR_LITERAL* item)
{
  CALL("readHeaderPrecedence(PR_LITERAL* item)");
  if (PR_TERM_LIST::Length(item->Arguments()) != 3) 
    {
      cout << "Exactly 3 arguments expected.\n";
      return false;
    };

  PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),2);
  PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* arg3 = PR_TERM_LIST::Nth(item->Arguments(),0);

  if (arg1->Tag() != PR_TERM::COMPLEX ||
      (strcmp(arg1->Symbol()->Symbol(),"++") &&
       strcmp(arg1->Symbol()->Symbol(),"--")) ||
      PR_TERM_LIST::Length(arg1->Arguments()) != 1 ||
      arg1->Arguments()->Term()->Tag() != PR_TERM::ATOM_FUNCTOR)
    {
      cout << "Header (predicate name with polarity) expected as the 1st argument.\n";
      return false;
    };
  
  if (arg2->Tag() != PR_TERM::ATOM_NUMBER) 
    {
      cout << "Integer (arity) expected as the 2nd argument.\n";
      return false;
    };

  if (arg3->Tag() != PR_TERM::ATOM_NUMBER) 
    {
      cout << "Integer (precedence) expected as the 3rd argument.\n";
      return false;
    };

  long arity = strtol(arg2->Symbol()->Symbol(),0,10);
  long precedence = strtol(arg3->Symbol()->Symbol(),0,10);
  bool positive = !strcmp(arg1->Symbol()->Symbol(),"++");
  const char* predicateName = 
    arg1->Arguments()->Term()->Symbol()->Symbol();
  bool res = _signature.registerHeaderPrecedence(positive,
						 predicateName,
						 arity,
						 precedence);
  if (!res)
    {
      cout << "Inconsistent with some other settings.\n";
    };
  return res;
}; // bool Parser::readHeaderPrecedence(PR_LITERAL* item)


bool Parser::readSymbolWeight(PR_LITERAL* item)
{
  CALL("readSymbolWeight(PR_LITERAL* item)");

  if (PR_TERM_LIST::Length(item->Arguments()) != 4) 
    {
      cout << "Exactly 4 arguments expected.\n";
      return false;
    };

  PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),3);
  PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),2);
  PR_TERM* arg3 = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* arg4 = PR_TERM_LIST::Nth(item->Arguments(),0);
  

   if ((arg1->Tag() != PR_TERM::ATOM_FUNCTOR) ||
      (strcmp(arg1->Symbol()->Symbol(),"predicate") &&       
       strcmp(arg1->Symbol()->Symbol(),"function")))
    {
      cout << "1st argument must be either 'predicate' or 'function'\n";
      return false;
    };

  if ((arg2->Tag() != PR_TERM::ATOM_FUNCTOR) && 
      (arg2->Tag() != PR_TERM::ATOM_NUMBER))
    {
      cout << "Function symbol expected as the 2nd argument.\n";
      return false;
    };   

  if (arg3->Tag() != PR_TERM::ATOM_NUMBER)
    {
      cout << "Integer expected as the 3rd argument.\n";
      return false;
    };  
  
  if (arg4->Tag() != PR_TERM::ATOM_NUMBER)
    {
      cout << "Integer expected as the 4th argument.\n";
      return false;
    };  
  
  bool predicate = !strcmp(arg1->Symbol()->Symbol(),"predicate");

  long arity = strtol(arg3->Symbol()->Symbol(),0,10);

  Input::Symbol* symbol = _signature.registerSymbol(predicate,arg2->Symbol()->Symbol(),arity);

  VampireKernel::TermWeightType weightConstPart = 
    strtol(arg4->Symbol()->Symbol(),0,10);

  if (weightConstPart < (VampireKernel::TermWeightType)0)
    {
      cout << "Symbol weight < 0.\n";
      return false;
    };

  symbol->setWeightConstPart(weightConstPart);
  
  return true;
}; // bool Parser::readSymbolWeight(PR_LITERAL* item)



bool Parser::readSymbolInterpretation(PR_LITERAL* item)
{
  CALL("readSymbolInterpretation(PR_LITERAL* item)");
  
  if (PR_TERM_LIST::Length(item->Arguments()) != 5) 
    {
      cout << "5 arguments expected: input symbol type ('predicate' or 'function'), print name and arity, theory name and native symbol.\n";
      return false;
    };

  PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),4);  
  PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),3);
  PR_TERM* arg3 = PR_TERM_LIST::Nth(item->Arguments(),2);
  PR_TERM* arg4 = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* arg5 = PR_TERM_LIST::Nth(item->Arguments(),0);
  

   if ((arg1->Tag() != PR_TERM::ATOM_FUNCTOR) ||
      (strcmp(arg1->Symbol()->Symbol(),"predicate") &&       
       strcmp(arg1->Symbol()->Symbol(),"function")))
    {
      cout << "1st argument must be either 'predicate' or 'function'\n";
      return false;
    };


  if ((arg2->Tag() != PR_TERM::ATOM_FUNCTOR) ||
      (arg3->Tag() != PR_TERM::ATOM_NUMBER) ||
      (arg4->Tag() != PR_TERM::ATOM_FUNCTOR) ||
      (arg5->Tag() != PR_TERM::ATOM_FUNCTOR))
    {
      cout << "The 2nd,4th and 5th arguments are expected to be symbolic atoms. The 3rd argument must be integer.\n";
      return false;
    };

  bool predicate = !strcmp(arg1->Symbol()->Symbol(),"predicate");

  long arity = strtol(arg3->Symbol()->Symbol(),0,10);

  Input::Symbol* symbol = _signature.registerSymbol(predicate,
						    arg2->Symbol()->Symbol(),
						    arity);
  symbol->addInterpretation(arg4->Symbol()->Symbol(),
			    arg5->Symbol()->Symbol());

  return true;
}; // bool Parser::readSymbolInterpretation(PR_LITERAL* item)


bool Parser::readAnswerPredicateDeclaration(PR_LITERAL* item)
{
  CALL("readAnswerPredicateDeclaration(PR_LITERAL* item)");
  if (PR_TERM_LIST::Length(item->Arguments()) != 2) 
    {
      cout << "2 arguments expected: predicate print name and arity.\n";
      return false;
    };
  PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),1);
  PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),0);
  if (arg1->Tag() != PR_TERM::ATOM_FUNCTOR) return false;  
  if (arg2->Tag() != PR_TERM::ATOM_NUMBER) return false;
  long arity = strtol(arg2->Symbol()->Symbol(),0,10);
  return _signature.registerAnswerPredicate(arg1->Symbol()->Symbol(),arity);
}; // bool Parser::readAnswerPredicateDeclaration(PR_LITERAL* item)



bool Parser::readLiteralAnnotation(PR_TERM_LIST* lst,
				   bool preselected,
				   Input::Literal::Annotation& result)
{
  CALL("readLiteralAnnotation(PR_TERM_LIST* lst,bool preselected,Input::Literal::Annotation& result)");

  bool selectabilityCategorySet = false;
  bool selectabilityPrioritySet = false;
  bool selectabilityInheritanceLimitSet = false;
  bool deploymentModeSet = false;
  bool equalityModeSet = false;
  
  Input::Literal::Annotation::SelectabilityCategory 
    selectabilityCategory = Input::Literal::Annotation::DefaultSelectability;
  long selectabilityPriority = 0L;
  Input::Literal::Annotation::SelectabilityInheritanceLimit 
    selectabilityInheritanceLimit;
  Input::Literal::Annotation::DeploymentMode deploymentMode;
  Input::Literal::Annotation::EqualityMode equalityMode;

  for (; lst; lst = lst->Next())
    {
      PR_TERM* item = lst->Term();
      switch (item->Tag())
	{
	case PR_TERM::ATOM_FUNCTOR: 
	  if (strcmp(item->Symbol()->Symbol(),"rule_head") == 0)
	    {
	      if (deploymentModeSet)
		{
		  cout << "Bad literal annotation: duplicate deployment mode entries.\n";
		  return false;
		};
	      deploymentModeSet = true;
	      deploymentMode = Input::Literal::Annotation::RuleHead;
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"subquery") == 0)
	    {
	      if (deploymentModeSet)
		{
		  cout << "Bad literal annotation: duplicate deployment mode entries.\n";
		  return false;
		};
	      deploymentModeSet = true;
	      deploymentMode = Input::Literal::Annotation::Subquery;
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"high_sel") == 0)
	    {
	      if (selectabilityCategorySet)
		{
		  cout << "Bad literal annotation: duplicate selectability category entries.\n";
		  return false;
		};
	      selectabilityCategorySet = true;
	      selectabilityCategory = Input::Literal::Annotation::HighSelectability;
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"low_sel") == 0)
	    {
	      if (selectabilityCategorySet)
		{
		  cout << "Bad literal annotation: duplicate selectability category entries.\n";
		  return false;
		};
	      selectabilityCategorySet = true;
	      selectabilityCategory = Input::Literal::Annotation::LowSelectability;
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"add_res") == 0)
	    {
	      if (equalityModeSet)
		{
		  cout << "Bad literal annotation: duplicate equality mode entries.\n";
		  return false;
		};
	      equalityModeSet = true;
	      equalityMode = Input::Literal::Annotation::AddResolution;
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"res_only") == 0)
	    {
	      if (equalityModeSet)
		{
		  cout << "Bad literal annotation: duplicate equality mode entries.\n";
		  return false;
		};
	      equalityModeSet = true;
	      equalityMode = Input::Literal::Annotation::ResolutionOnly;
	    }
	  else
	    {
	      cout << "Bad literal annotation entry: " 
		   << item->Symbol()->Symbol() << "\n"; 
	      return false;
	    };
	  break;

	case PR_TERM::COMPLEX:
	  if (strcmp(item->Symbol()->Symbol(),"pri") == 0)
	    {
	      if (PR_TERM_LIST::Length(item->Arguments()) != 1)
		{
		  cout << "Bad literal annotation entry: pri(.) must be unary.\n";
		  return false;
		};
	      PR_TERM* priorityValue = PR_TERM_LIST::Nth(item->Arguments(),0);
	      if (priorityValue->Tag() == PR_TERM::ATOM_NUMBER)
		{ 
		  char* endptr;
		  long val = strtol(priorityValue->Symbol()->Symbol(),
				    &endptr,
				    10);
		  if ((*endptr) != '\0')
		    {
		      cout << "Bad literal annotation entry: priority value must be\n"
			   << "an integer, 'max' or 'min'.\n";
		      return false; 
		    };
		  if (selectabilityPrioritySet)
		    {
		      cout << "Bad literal annotation: duplicate priority values.\n";
		      return false;
		    };
		  
		  selectabilityPrioritySet = true;
		  selectabilityPriority = static_cast<long>(val);
		}
	      else if (priorityValue->Tag() == PR_TERM::ATOM_FUNCTOR)
		{
		  if (strcmp(priorityValue->Symbol()->Symbol(),"max") == 0)
		    {
		      if (selectabilityPrioritySet)
			{
			  cout << "Bad literal annotation: duplicate priority values.\n";
			  return false;
			};
		      selectabilityPrioritySet = true;
		      selectabilityPriority = 
			Input::Literal::Annotation::MaximalSelectabilityPriority;
		    }
		  else if (strcmp(priorityValue->Symbol()->Symbol(),"min") == 0)
		    {
		      if (selectabilityPrioritySet)
			{
			  cout << "Bad literal annotation: duplicate priority values.\n";
			  return false;
			};
		      selectabilityPrioritySet = true;
		      selectabilityPriority = 
			Input::Literal::Annotation::MinimalSelectabilityPriority;
		    }
		  else
		    {
		      cout << "Bad literal annotation entry: priority value must be\n"
			   << "an integer, 'max' or 'min'.\n";
		      return false; 
		    };
		}
	      else
		{
		  cout << "Bad literal annotation entry: priority value must be\n"
		       << "an integer, 'max' or 'min'.\n";
		  return false; 
		};
	    }
	  else if (strcmp(item->Symbol()->Symbol(),"inh") == 0)
	    {
	      if (PR_TERM_LIST::Length(item->Arguments()) != 1)
		{
		  cout << "Bad literal annotation entry: inh(.) must be unary.\n";
		  return false;
		};
	      
	      PR_TERM* limitValue = PR_TERM_LIST::Nth(item->Arguments(),0);
	      
	      if (limitValue->Tag() == PR_TERM::ATOM_NUMBER)
		{ 
		  char* endptr;
		  long val = strtol(limitValue->Symbol()->Symbol(),
				    &endptr,
				    10);
		  if ((*endptr) != '\0' ||
		      val < 0L)
		    {
		      cout << "Bad literal annotation entry: inheritance limit must be\n"
			   << "a nonnegative integer.\n";
		      return false; 
		    };
		  if (selectabilityInheritanceLimitSet)
		    {
		      cout << "Bad literal annotation: duplicate inheritance limit entries.\n";
		      return false;
		    };
		  
		  selectabilityInheritanceLimitSet = true;
		  selectabilityInheritanceLimit.makeFinite(static_cast<ulong>(val));
		}
	      else
		{
		  cout << "Bad literal annotation entry: inheritance limit must be\n"
		       << "a nonnegative integer.\n";
		  return false; 
		};
	    }
	  else 
	    {
	      cout << "Bad literal annotation entry: " 
		   << item->Symbol()->Symbol() << "\n"; 
	      return false;
	    };
	  break;

	default:	  
	  cout << "Bad literal annotation entry.\n"
	       << "Should be rule_head, subquery, high_sel, low_sel, pri(.) or inh(.) .\n";
	  return false;
	};
    }; // for (; lst; lst = lst->Next())

  if (!(preselected ||
	selectabilityCategorySet ||
	selectabilityPrioritySet ||
	selectabilityInheritanceLimitSet ||
	deploymentModeSet ||
	equalityModeSet))
    {
      result.makeDefault();
      return true;
    };
	

  if (!selectabilityPrioritySet) 
    selectabilityPriority = 0UL;

  if (!selectabilityInheritanceLimitSet)
    selectabilityInheritanceLimit.makeInfinite();

  if (!deploymentModeSet)
    deploymentMode = Input::Literal::Annotation::DefaultDeploymentMode;

  if (!equalityModeSet)
    equalityMode = Input::Literal::Annotation::DefaultEqualityMode;


  // check that preselection is consistent with the annotation
  if (preselected)
    {
      if (selectabilityCategorySet || selectabilityPrioritySet)
	{
	  cout << "Literal annotation conflicts with preselection.\n" 
	       << "Preselected literals have fixed selectability category and priority.\n";
	  return false;
	};      
      result.makeHighSelectability(Input::Literal::Annotation::MaximalSelectabilityPriority,
				   selectabilityInheritanceLimit,
				   deploymentMode,
				   equalityMode);
      return true;
    };

  // !preselected

  if (selectabilityCategorySet)
    {
      if (selectabilityCategory == Input::Literal::Annotation::LowSelectability)
	{
	  result.makeLowSelectability(selectabilityPriority,
				      selectabilityInheritanceLimit,
				      deploymentMode,
				      equalityMode);
	}
      else
	result.makeHighSelectability(selectabilityPriority,
				     selectabilityInheritanceLimit,
				     deploymentMode,
				     equalityMode);
      return true;
    };

  // !selectabilityCategorySet
  if (selectabilityInheritanceLimitSet)
    {
      cout << "Bad literal annotation: explicit inheritance limit value\n"
	   << "is not allowed with the default selectability category.\n";
      return false;
    };

  result.makeDefaultSelectability(selectabilityPriority,
				  deploymentMode,
				  equalityMode);
  return true;
}; // bool Parser::readLiteralAnnotation(PR_TERM_LIST* lst,bool preselected,Input::Literal::Annotation& result)

//======================================================================
