//
// File:         Statistics.hpp
// Description:  Statistics on inferences and clauses.
// Created:      Nov 17, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef STATISTICS_H
#define STATISTICS_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
using namespace std;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class Statistics
  {
  public:
    class Inferences
    {
    public:
      class Deduction
      {
      public:
	class RuleProfile
	{
	public:
	  RuleProfile() { reset(); };
	  void reset();
	  void absorb(const RuleProfile& stat);
	  bool somethingRegistered() const;
	  bool someNonredundantDiscarded() const;
	  ulong totalNonredundantDiscarded() const;
	  ulong totalNonredundantBranchesDiscarded() const;
	  ostream& outputDetailed(ostream& str) const;
	public:
	  ulong accomplished; // led to generating a new clause
	  ulong redundantDiscarded; 
	  ulong redundantBranchesDiscarded;
	  ulong discardedBySize;
	  ulong branchesDiscardedBySize;
	  ulong discardedParamodulationsFromVariables;
	  ulong branchesDiscardedWithParamodulationsFromVariables;
	  ulong discardedChainingsThroughVariables;
	  ulong branchesDiscardedWithChainingsThroughVariables;
	  ulong mainPremisesSubsumed;
	  ulong sidePremisesSubsumed;
	  ulong mainPremisesDemodulated;
	  ulong sidePremisesDemodulated;
	  ulong discardedForOtherReasons;
	  ulong branchesDiscardedForOtherReasons;
	}; // class RuleProfile
      public:
	Deduction() { reset(); };
	void reset();
	void absorb(const Deduction& stat);
	bool somethingRegistered() const
	{
	  return 
	    resolution.somethingRegistered() ||
	    forwardSuperposition.somethingRegistered() ||
	    backwardSuperposition.somethingRegistered() ||
	    equalityFactoring.somethingRegistered() ||
	    equalityResolution.somethingRegistered() ||
	    forwardChaining.somethingRegistered() ||
	    backwardChaining.somethingRegistered();
	};
	ulong totalAccomplished() const;
	ulong totalRedundantDiscarded() const;
	ulong totalRedundantBranchesDiscarded() const;
	ulong totalNonredundantDiscarded() const;
	ulong totalNonredundantBranchesDiscarded() const;
      public:
	RuleProfile resolution;
	RuleProfile forwardSuperposition;
	RuleProfile backwardSuperposition;
	RuleProfile equalityFactoring;
	RuleProfile equalityResolution;
	RuleProfile forwardChaining;
	RuleProfile backwardChaining;
      }; // class Deduction

      class Simplifying
      {
      public:
	Simplifying() { reset(); };
	void reset();
	void absorb(const Simplifying& stat);
	ulong total() const;
	ulong totalForward() const;
	ulong totalBackward() const;
      public:
	ulong propositionalTautologyDeletion;
	ulong equationalTautologyDeletion;
	ulong reductionInNonBasicPosition;
	ulong forwardSubsumption;
	ulong backwardSubsumption;
	ulong forwardSubsumptionResolution;
	ulong forwardDemodulation;
	ulong onBranchForwardDemodulation;
	ulong backwardDemodulation;
	ulong onBranchBackwardDemodulation;
	ulong destructiveEqualityResolution;
	ulong builtInTheory;
      }; // class Simplifying

      class PropositionalSplitting
      {
      public:
	PropositionalSplitting() { reset(); };
	void reset();
	void absorb(const PropositionalSplitting& stat);
	bool somethingRegistered() const;
      public:
	ulong accomplished;
	ulong components;
	ulong introducedComponentNames;
      }; // class PropositionalSplitting

      class NegativeEqualitySplitting
      {
      public:
	NegativeEqualitySplitting() { reset(); };
	void reset();
	void absorb(const NegativeEqualitySplitting& stat);
	bool somethingRegistered() const;
      public:
	ulong accomplished;
	ulong argumentsSplitOff;
	ulong introducedNames;
      }; // class NegativeEqualitySplitting

    public:
      Inferences() { reset(); };
      void reset();
      void absorb(const Inferences& stat);
      bool someNonredundantDiscarded() const
      {
	return 
	  deduction.totalNonredundantDiscarded() ||
	  deduction.totalNonredundantBranchesDiscarded();
      };
      ostream& outputDetailed(ostream& str) const;
      ostream& outputSummary(ostream& str) const;
    public:
      Deduction deduction;
      Simplifying simplifying;
      PropositionalSplitting propositionalSplitting;
      NegativeEqualitySplitting negativeEqualitySplitting;
    }; // class Inferences

    class Clauses
    {
      class Generated
      {
      public:
	Generated() { reset(); };
	void reset();
	void absorb(const Generated& stat);
	ulong total() const;
	ulong totalGeneratedByDeduction() const;
	ulong totalGeneratedBySimplification() const;
	ulong totalRedundant() const;
	ulong totalNonredundantDiscarded() const;
	ulong totalSimplified() const;

      public:
	ulong byResolution;
	ulong byForwardSuperposition;
	ulong byBackwardSuperposition;
	ulong byEqualityFactoring;
	ulong byEqualityResolution;
	ulong byForwardChaining;
	ulong byBackwardChaining;
	ulong byBackwardDemodulation;
	ulong byOnBranchBackwardDemodulation;
	ulong byPropositionalSplitting;
	ulong byNegativeEqualitySplitting;
	ulong byOther;

	ulong discardedPropositionalTautologies;
	ulong discardedEquationalTautologies;
	ulong reducedInNonBasicPosition;
	ulong subsumed;
	
	ulong discardedWithTooDeepInferences;
	ulong discardedTooLarge;
	ulong discardedWithTooLargeLiterals;
	ulong discardedWithTooManyLiterals;
	ulong discardedTooDeep;
	ulong discardedWithTooManyVariables;
	ulong discardedForOtherReasons;
	
	ulong simplifiedBySubsumptionResolution;
	ulong simplifiedByDemodulation;
	ulong simplifiedByOnBranchDemodulation;
	ulong simplifiedByDestructiveEqualityResolution;
	ulong simplifiedByBuiltInTheories;
	
	ulong transformedByPropositionalSplitting;
	ulong transformedByNegativeEqualitySplitting;

      }; // class Generated

      class Retained
      {
      public:
	Retained() { reset(); };
	void reset();
	void absorb(const Retained& stat);
	ulong totalSimplified() const;
	ulong totalNonredundantDiscarded() const
	{
	  return passiveRecycled + activeRecycled;
	};
      public:
	ulong total;
	ulong selected;
	ulong passiveSubsumed;
	ulong activeSubsumed;
	ulong passiveSimplifiedByDemodulation;
	ulong activeSimplifiedByDemodulation;
	ulong passiveSimplifiedByOnBranchDemodulation;
	ulong activeSimplifiedByOnBranchDemodulation;
	ulong passiveOrphansMurdered;
	ulong activeOrphansMurdered;
	ulong passiveRecycled;
	ulong activeRecycled;
      }; // class Retained
    public:
      Clauses() { reset(); };
      void reset();
      void absorb(const Clauses& stat);
      bool someNonredundantDiscarded() const 
      {
	return 
	  generated.totalNonredundantDiscarded() ||
	  retained.totalNonredundantDiscarded();
      };
      ostream& outputDetailed(ostream& str) const;
      ostream& outputSummary(ostream& str) const;
    public:
      Generated generated;
      Retained retained;
    }; // class Clauses
  public:
    Statistics() { reset(); };
    void reset();
    void absorb(const Statistics& stat);
    bool completenessViolated() const;
    ostream& outputDetailed(ostream& str) const;
    ostream& outputSummary(ostream& str) const;

  public:
    Inferences inferences;
    Clauses clauses;
  }; // class Statistics
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
}; // namespace VK

//===================================================================

#endif
