//
// File:         TempLitListExtras.cpp
// Description:  Utilities for processing temporary lists
//               of temporary literals.
// Created:      Oct 24, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       stable
// Revised:      Jun 11, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in weed(..,const SymbolInfo* commutativityTableForEqTaut,..). 
//               el2->literal() was not tested for being equality.
//=================================================================
#include "TempLitListExtras.hpp"
#include "SymbolInfo.hpp"
using namespace BK;
using namespace VK;
//===================================================================

TempLitListExtras::Context* TempLitListExtras::Context::_current; 
// ^ implicitely initialised by 0

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
//===================================================================

long 
TempLitListExtras::computeWeightBase(const BasicTempLitList& lst)
{
  CALL("computeWeightBase(const BasicTempLitList& lst)");
  long res = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    res += el->literal()->weightBase();
  return res;
}; // long TempLitListExtras::computeWeightBase(const BasicTempLitList& lst)

void
TempLitListExtras::computeSizeAndWeightBase(const BasicTempLitList& lst,
					    long& size,
					    long& weightBase)
{
  CALL("computeSizeAndWeightBase(..)");
  size = 0L;
  weightBase = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    {
      size += el->literal()->size();
      weightBase += el->literal()->weightBase();
    };
}; // void TempLitListExtras::computeSizeAndWeightBase(..)



void
TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(const BasicTempLitList& lst,
							 long& size,
							 long& weightBase,
							 long& numOfPosEq)
{
  CALL("computeSizeAndWeightBaseAndNumOfPosEq(..)");
  size = 0L;
  weightBase = 0L;
  numOfPosEq = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    {
      size += el->literal()->size();
      weightBase += el->literal()->weightBase();
      if (el->literal()->isEquality() &&
	  el->literal()->isPositive())
	++numOfPosEq;
    };
}; // void TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(..)


long 
TempLitListExtras::minimalPosEqLHSSize(const BasicTempLitList& lst)
{
  CALL("minimalPosEqLHSSize(const BasicTempLitList& lst)");
  for (const Element* el = lst.first(); el; el = el->next())
    if (el->literal()->isEquality() && 
	el->literal()->isPositive())
      {
	long res = el->literal()->minPosEqLHSSize();
	for (el = el->next(); el; el = el->next())
	  if (el->literal()->isEquality() && 
	      el->literal()->isPositive())
	    {
	      long tmp = el->literal()->minPosEqLHSSize();
	      if (tmp < res) res = tmp;
	    };
	return res;
      };
    return 0L;
}; // long TempLitListExtras::minimalPosEqLHSSize(const BasicTempLitList& lst)


bool 
TempLitListExtras::instantiationIsPossible(const BasicTempLitList& lst,
					   ulong var,
					   ulong instanceDepth,
					   long instanceSize,
					   bool checkInputNonBasicness)
{
  CALL("instantiationIsPossible(..)");
  ASSERT(instanceDepth);
  for (const Element* el = lst.first(); el; el = el->next())
    if (!el->literal()->instantiationPossible(var,
					      instanceDepth,
					      instanceSize,
					      checkInputNonBasicness))
      return false;
  return true;
}; // bool TempLitListExtras::instantiationIsPossible(..)



void
TempLitListExtras::passInOrder(BasicTempLitList& src,
			       BasicTempLitList& dest,
			       const VAR_WEIGHTING& varWeighting)
{
  CALL("passInOrder(..,const VAR_WEIGHTING& varWeighting)");
  DOP(ulong debugDestInitLength = dest.length());
  DOP(ulong debugSrcInitLength = src.length());
  if (src.isEmpty()) return;
  if (src.length() == 1UL) 
    {
      src.popTo(dest);
      return;
    };
  if (src.length() == 2UL)
    {
      Element* fst = src.first();
      Element* snd = src.last();
      if (fst->literal()->compareWithVarWeighting(*(snd->literal()),varWeighting) == 
	  BK::Greater)
	{
	  src.popTo(dest);
	  src.popTo(dest);
	}
      else // fst < snd
	{
	  src.passTo(fst,dest);
	  src.passTo(snd,dest);
	};
      return;
    };


  // general case, use the quicksort algorithm
  LitHandle pivot;
  src.popTo(pivot);
  BasicTempLitList greaterThanPivot(src.elementPool());
  
  for (Element* el = src.first(); el; /* nothing here */ )
    {
      Element* tmp = el->next();
      if (el->literal()->compareWithVarWeighting(*pivot,varWeighting) == 
	  BK::Greater) 
	{
	  src.passTo(el,greaterThanPivot);
	};
      el = tmp;
    };
  
  passInOrder(src,dest,varWeighting);
  dest.collect(pivot);
  passInOrder(greaterThanPivot,dest,varWeighting);
  ASSERT(src.isEmpty());
  ASSERT(greaterThanPivot.isEmpty());
  ASSERT(dest.length() == debugDestInitLength + debugSrcInitLength);
}; // void TempLitListExtras::passInOrder(..,const VAR_WEIGHTING& varWeighting)


void 
TempLitListExtras::renameVariables(const BasicTempLitList& lst,
				   INT_VAR_RENAMING& varRenaming)
{
  CALL("renameVariables(..)");
  for (const Element* el = lst.last(); el; el = el->previous()) 
    // ^ order is significant
    el->literal()->renameVariables(varRenaming);
}; // void TempLitListExtras::renameVariables(..)


void TempLitListExtras::weed(BasicTempLitList& lst,
			     bool& isTautology,
			     bool& isEqTautology)
{
  CALL("weed(BasicTempLitList& lst,bool& isTautology,bool& isEqTautology)");
  
  for (Element* el1 = lst.first(); el1; /* nothing here */)
    {
      Element* next = el1->next();

      LitHandle lit1 = el1->literal();
      if (lit1->isEquality())
	{
	  if (lit1->argsAreEqual())
	    {
	      if (lit1->isPositive())
		{
		  // equational tautology
		  isTautology = true;
		  isEqTautology = true;
		  return;
		}
	      else // redundant literal
		lst.remove(el1);
	    }
	  else // different arguments
	    {
	      lit1->orderEqArgs();
	      for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		if (el2->literal()->isEquality())
		  if (el2->literal()->equalEqLit(*lit1))
		    {
		      // el2 is redundant
		      lit1->properties().factoriseWith(el2->literal()->properties());
		      lst.remove(el2);
		    }
		  else
		    if (el2->literal()->complementaryEqLit(*lit1))
		      {
			// propositional tautology
			isTautology = true;
			isEqTautology = false;
			return;
		      };
	    };	    
	}
      else // nonequality literal
	{
	  if (lit1->header() == TERM::builtInTrue())
	    {
	      // propositional tautology
	      isTautology = true;
	      isEqTautology = false;
	      return;
	    };

	  if (lit1->header() == TERM::builtInFalse())
	    {
	      // redundant literal
	      lst.remove(el1);
	    }
	  else
	    for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
	      if (el2->literal()->equal(*lit1))
		{
		  // el2 is redundant
		  lit1->properties().factoriseWith(el2->literal()->properties());
		  lst.remove(el2);
		}
	      else
		if (el2->literal()->isComplementary(*lit1))
		  {
		    // propositional tautology 
		    isTautology = true;
		    isEqTautology = false;
		    return;
		  };
	}; // if (lit1->isEquality())

      el1 = next;
    }; // for (Element* el1 = lst.first(); el1; /* nothing here */)
  
  isTautology = false;
  isEqTautology = false;
}; // void TempLitListExtras::weed(BasicTempLitList& lst,bool& isTautology,bool& isEqTautology)



void TempLitListExtras::weed(BasicTempLitList& lst,
			     bool& isTautology,
			     bool& isEqTautology,
			     const SymbolInfo* commutativityTableForEqTaut,
			     const SymbolInfo* commutativityTableForPropTaut,
			     const SymbolInfo* commutativityTableForTrivialEqRes,
			     const SymbolInfo* commutativityTableForIdenticalLitFactoring,
			     BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("weed(..,const SymbolInfo* commutativityTableForEqTaut,..)");
  bool commutativityUsed;
  for (Element* el1 = lst.first(); el1; /* nothing here */)
    {
      Element* next = el1->next();

      LitHandle lit1 = el1->literal();
      if (lit1->isEquality())
	{
	  if (lit1->isPositive())
	    {
	      // check for equational tautology
	      if (commutativityTableForEqTaut)
		{
		  if (lit1->argsAreEqual(commutativityTableForEqTaut,commutativityUsed))
		    {
		      // equational tautology	
		      if (commutativityUsed)
			{
			  if ((lit1->arg1()->size() == 3L) &&
			      lit1->arg1()->next()->isVariable() &&
			      lit1->arg1()->next()->next()->isVariable())
			    {
			      // tautology that can not be discarded
			    }
			  else
			    {
			      // tautology that can be discarded
			      lit1->commutativityLawsForArgsAreEqual(commutativityTableForEqTaut,
								     usedCommutativityLaws);	
			      isTautology = true;
			      isEqTautology = true;
			      return;
			    };			  
			}
		      else
			{
			  isTautology = true;
			  isEqTautology = true;
			  return;
			};
		    };
		}
	      else
		if (lit1->argsAreEqual())
		  {
		    isTautology = true;
		    isEqTautology = true;
		    return;
		  };
	      
	      goto scan_for_equal_or_complementary;
	    }
	  else // lit1 is negative equality 
	    {
	      // check for trivial equality resolution
	      bool resolved;
	      if (commutativityTableForTrivialEqRes)
		{
		  resolved = 
		    lit1->argsAreEqual(commutativityTableForTrivialEqRes,commutativityUsed);
		  if (resolved)
		    lit1->commutativityLawsForArgsAreEqual(commutativityTableForTrivialEqRes,
							   usedCommutativityLaws);	
		}
	      else
		resolved = lit1->argsAreEqual();

	      if (resolved)
		{
		  lst.remove(el1);
		}
	      else
		{
		  // check for propositional tautology or possible factoring
		scan_for_equal_or_complementary:

		  for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		    {
		      if (el2->literal()->isEquality())
			if (el2->literal()->polarity() == lit1->polarity())
			  {
			    if (equalEq(el2->literal(),
					lit1,
					commutativityTableForIdenticalLitFactoring,
					usedCommutativityLaws))
			      {
				// el2 is redundant
				lit1->properties().factoriseWith(el2->literal()->properties());
				lst.remove(el2);
			      };
			  }
			else // different polarities
			  {
			    if (complementaryEq(el2->literal(),
						lit1,
						commutativityTableForPropTaut,
						usedCommutativityLaws))
			      {
				// propositional tautology 
				isTautology = true;
				isEqTautology = false;
				return;
			      };
			  };
		    }; // for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())		  
		};
	    };
	}
      else // nonequality literal
	{
	  if (lit1->header() == TERM::builtInTrue())
	    {
	      // propositional tautology
	      isTautology = true;
	      isEqTautology = false;
	      return;
	    };

	  if (lit1->header() == TERM::builtInFalse())
	    {
	      // redundant literal
	      lst.remove(el1);
	    }
	  else
	    for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
	      {
		if (el2->literal()->polarity() == lit1->polarity())
		  {
		    if (equal(el2->literal(),
			      lit1,
			      commutativityTableForIdenticalLitFactoring,
			      usedCommutativityLaws))
		      {
			// el2 is redundant
			lit1->properties().factoriseWith(el2->literal()->properties());
			lst.remove(el2);
		      };
		  }
		else // different polarities
		  {
		    if (complementary(el2->literal(),
				      lit1,
				      commutativityTableForPropTaut,
				      usedCommutativityLaws))
		      {
			// propositional tautology 
			isTautology = true;
			isEqTautology = false;
			return;
		      };
		  };
	      }; // for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())	      
	}; // if (lit1->isEquality())

      el1 = next;
    }; // for (Element* el1 = lst.first(); el1; /* nothing here */)
  
  isTautology = false;
  isEqTautology = false;
}; // void TempLitListExtras::weed(..,const SymbolInfo* commutativityTableForEqTaut,..)



inline
bool 
TempLitListExtras::complementary(const LitHandle& lit1,
				 const LitHandle& lit2,
				 const SymbolInfo* commutativityTable,
				 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("complementary(const LitHandle& lit1,...)");
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->isComplementary(*lit2,
				commutativityTable,
				commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForComplementary(*lit2,
						    commutativityTable, 
						    usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->isComplementary(*lit1);
}; // bool TempLitListExtras::complementary(const LitHandle& lit1,...)



inline
bool 
TempLitListExtras::equal(const LitHandle& lit1,
			 const LitHandle& lit2,
			 const SymbolInfo* commutativityTable,
			 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("equal(const LitHandle& lit1,...)");
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->equal(*lit2,
			commutativityTable,
			commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForEqual(*lit2,
					    commutativityTable, 
					    usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->equal(*lit1);
}; // bool TempLitListExtras::equal(const LitHandle& lit1,...)


inline
bool 
TempLitListExtras::complementaryEq(const LitHandle& lit1,
				 const LitHandle& lit2,
				 const SymbolInfo* commutativityTable,
				 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("complementaryEq(const LitHandle& lit1,...)");
  ASSERT(lit1->isEquality() && lit2->isEquality());
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->complementaryEqLit(*lit2,
				   commutativityTable,
				   commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForComplementaryEqLit(*lit2,
							 commutativityTable, 
							 usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->complementaryEqLit(*lit1);
}; // bool TempLitListExtras::complementaryEq(const LitHandle& lit1,...)



inline
bool 
TempLitListExtras::equalEq(const LitHandle& lit1,
			 const LitHandle& lit2,
			 const SymbolInfo* commutativityTable,
			 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("equalEq(const LitHandle& lit1,...)");
  ASSERT(lit1->isEquality() && lit2->isEquality());
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->equalEqLit(*lit2,
			   commutativityTable,
			   commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForEqualEqLit(*lit2,
						 commutativityTable, 
						 usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->equalEqLit(*lit1);
}; // bool TempLitListExtras::equalEq(const LitHandle& lit1,...)




ulong
TempLitListExtras::countDifferentVariables(const BasicTempLitList& lst)
{
  CALL("countDifferentVariables(const BasicTempLitList& lst)");
  if (lst.length() == 1UL)
    {
      return lst.first()->literal()->numOfDifferentVariables();
    }
  else
    {      
      static WeightPolynomial::Monomials vars;
      vars.reset();
      collectVariableOccurences(lst,vars);
      return vars.length();
    };
}; // ulong TempLitListExtras::countDifferentVariables(const BasicTempLitList& lst)

void 
TempLitListExtras::collectVariableOccurences(const BasicTempLitList& lst,
					     WeightPolynomial::Monomials& vars)
{
  CALL("collectVariableOccurences(..)");  
  for (const BasicTempLitList::Element* el = lst.first(); el; el = el->next())
    vars.add(el->literal()->variableOccurences());
}; // void TempLitListExtras::collectVariableOccurences(..)






//=================================================================
