//
// File:         TemporaryClauseCollector.cpp
// Description:  Structure for collecting temporary clauses.
//               Collected tautologies are discarded 
//               as well as clauses which do not satisfy
//               certain size/complexity restrictions.
// Created:      Nov 08, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "TemporaryClauseCollector.hpp"
#include "TemporaryLiteral.inl"
#include "PagedPassiveClauseStorage.inl"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//===================================================================

TemporaryClauseCollector::Context* TemporaryClauseCollector::Context::_current; 
// ^ implicitely initialised by 0

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
//===================================================================

//                  Basic functionality

TemporaryClauseCollector::TemporaryClauseCollector() :  
  OPEN_BR(1,"br 1") DCOMMA
  _markSubstitutionPositions(false),
  _showAnnouncedClauses(0),
  _showCompleteClauses(0),
  _preprocessingInProgress(false),
  _outputQueue(0),
  _rejectionReason(None),
  _statistics(0),
  _tautologyCheck(true),
  _literalFactoring(true),
  _sizeLimit(VampireKernelConst::MaxClauseSize),
  _literalSizeLimit(VampireKernelConst::MaxTermSize),
  _lengthLimit(VampireKernelConst::MaxlNumOfLits)
{    
#ifndef NO_DEBUG
  _name = "?";
#endif
  DOP(_debugTrash = true);
};

#ifndef NO_DEBUG
TemporaryClauseCollector::TemporaryClauseCollector(const char* name) :
  OPEN_BR(1,"br 1") DCOMMA 
  _markSubstitutionPositions(false),
  _showAnnouncedClauses(0),
  _showCompleteClauses(0),
  _preprocessingInProgress(false),
  _outputQueue(0),
  _rejectionReason(None),
  _statistics(0),
  _tautologyCheck(true),
  _literalFactoring(true),
  _sizeLimit(VampireKernelConst::MaxClauseSize),
  _literalSizeLimit(VampireKernelConst::MaxTermSize),
  _lengthLimit(VampireKernelConst::MaxlNumOfLits)
{    
  _name = name;
  DOP(_debugTrash = true);
};
#endif


TemporaryClauseCollector::~TemporaryClauseCollector()
{
  CALLM(_name,"destructor ~TemporaryClauseCollector()");
}; // TemporaryClauseCollector::~TemporaryClauseCollector()

void TemporaryClauseCollector::reset()
{
  CALLM(_name,"reset()");
  _markSubstitutionPositions = false; 
  _showAnnouncedClauses = 0;
  _showCompleteClauses = 0;
  _preprocessingInProgress = false;
  _outputQueue = 0;
  _rejectionReason = None;
  _newClause.clear();
  _newLiteral.clear();
  _statistics = 0;
  _tautologyCheck = true;
  _literalFactoring = true;
  _sizeLimit = VampireKernelConst::MaxClauseSize;
  _literalSizeLimit = VampireKernelConst::MaxTermSize;
  _lengthLimit = VampireKernelConst::MaxlNumOfLits;
  DOP(_debugTrash = true);
}; // void TemporaryClauseCollector::reset()


void TemporaryClauseCollector::openNewClause(ClauseBackground::Rule rule)
{
  CALLM(_name,"openNewClause(ClauseBackground::Rule rule)");
  registerGenerated(rule);
  _rejectionReason = None;
  if (_newClause.isNull())
    {
      TemporaryClause::Extras::freshClause(_newClause);
    }
  else
    _newClause->reset();
  _newClause->background().UsedRule(rule);
  if (_preprocessingInProgress)
    _newClause->background().UsedRule(ClauseBackground::Prepro);
  _unifierVarRenaming.Reset();
  
  _nonanswerSize = 0UL;
  _totalSize = 0UL;
  DOP(_debugTrash = false);
}; // void TemporaryClauseCollector::openNewClause(ClauseBackground::Rule rule)
  
void TemporaryClauseCollector::announceNewClause()
{
  CALL("announceNewClause()");
  if (_showAnnouncedClauses)
    _newClause->outputAnnounced((*_showAnnouncedClauses) << "% ") << "\n";
}; // void TemporaryClauseCollector::announceNewClause()

bool TemporaryClauseCollector::finaliseClause()
{
  CALLM(_name,"finaliseClause()");
  ASSERT(!_debugTrash);
  TempLitListExtras::preorientEqualities(*(_newClause->literals()));
  _newClause->resetSimplificationFlags();
  _newClause->markAsWeeded();
  if (_showCompleteClauses)
    _newClause->outputComplete((*_showCompleteClauses) << "% ") << "\n";
  _outputQueue->collect(_newClause);
  DOP(_debugTrash = true);
  return true; // may change in the future
}; // bool TemporaryClauseCollector::finaliseClause()



bool TemporaryClauseCollector::finaliseClause(ClauseHandle& recipient)
{
  CALLM(_name,"finaliseClause(recipient)");  
  ASSERT(!_debugTrash);
  TempLitListExtras::preorientEqualities(*(_newClause->literals()));
  _newClause->resetSimplificationFlags();
  _newClause->markAsWeeded();
  if (_showCompleteClauses)
    _newClause->outputComplete((*_showCompleteClauses) << "% ") << "\n";
  recipient.collectFrom(_newClause);
  DOP(_debugTrash = true);
  return true; // may change in the future
}; // bool TemporaryClauseCollector::finaliseClause(ClauseHandle& recipient)

//                  Extended functionality



bool TemporaryClauseCollector::pushLiteral(const TERM* lit,
					   const LiteralProperties& prop)
  {
    CALLM(_name,"pushLiteral(const TERM* lit,const LiteralProperties& prop)");
    
    ASSERT(!_debugTrash);
    ASSERT(!_newClause.isNull());
    TERM header = *lit;
    if (!(header.arity())) 
      {
	return pushPropLiteral(header,prop);
      };
    if (!openNewNonpropLiteral(header,prop)) 
      return false;
    TERM::Iterator iter;
    iter.Reset(lit);
    while (iter.Next()) 
      {
	if (iter.CurrentSymIsVar())
	  {
	    pushVar(iter.Symbol(),false);
	  }
	else
	  {
	    if (iter.Symbol().arity())
	      {
		if (!pushNonconstFunc(iter.Symbol(),false)) return false;
	      }
	    else 
	      pushConst(iter.Symbol(),false); 
	  }; 
      };
    return closeNonpropLiteral();
  }; // bool TemporaryClauseCollector::pushLiteral(const TERM* lit,const LiteralProperties& prop)

bool 
TemporaryClauseCollector::pushPseudoLiteral(const TERM& header,
					    const Flatterm* arg1,
					    const WeightPolynomial::Monomials& varsForPadding,
					    const LiteralProperties& prop)
{
  CALLM(_name,"pushPseudoLiteral(const TERM& header,const Flatterm* arg1,..)");
  ASSERT(!_debugTrash);
  ASSERT(header.arity() == 1 + varsForPadding.length());  
  ASSERT(!header.IsEquality());
  if (_newLiteral.isNull())
    {
      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	{
	  reject(None);
	  return false;
	};
    };
  if (!_newLiteral->pushPseudoLit(header,arg1,varsForPadding,prop)) 
    {
      reject(LiteralTooDeep);
      return false;
    };
  _totalSize += _newLiteral->size();
  if (_totalSize > VampireKernelConst::MaxClauseSize)
    {
      reject(ClauseTooLarge);
      return false;
    };
  _collectingNonanswerLiteral = !header.isAnswerHeader();
  
  if (_collectingNonanswerLiteral)
    {
      _nonanswerSize += _newLiteral->size();
      ASSERT(_nonanswerSize <= _totalSize);
      if (_nonanswerSize > _sizeLimit)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  return closeNonpropLiteral();
}; // bool TemporaryClauseCollector::pushPseudoLiteral(const TERM& header,const Flatterm* arg,..)


bool TemporaryClauseCollector::loadClause(const Clause* c,ClauseBackground::Rule rule)
{
  CALLM(_name,"loadClause(const Clause* c,ClauseBackground::Rule rule)");

  openNewClause(rule);
  announceNewClause();

  for(LiteralList::Iterator iter(c->LitList());
      iter.notEnd();
      iter.next())
    if (!(pushLiteral(iter.currentLiteral(),
		      iter.currentElement()->properties()))) 
      return false;       
  return true;
}; // bool TemporaryClauseCollector::loadClause(const Clause* c,ClauseBackground::Rule rule,bool retainNumber)



bool TemporaryClauseCollector::loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,ClauseBackground::Rule rule)
{
  CALLM(_name,"loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,ClauseBackground::Rule rule)");
  literals.reset();
  openNewClause(rule);
  announceNewClause();

  while (literals.nextLiteral())
    {
      LiteralProperties prop = literals.currentLiteralProperties();

      TERM sym;
      ALWAYS(literals.nextSymbol(sym)); // header
      if (!sym.arity()) 
	{
	  if (!pushPropLiteral(sym,prop)) return false;
	}
      else
	{
	  if (!(openNewNonpropLiteral(sym,prop))) return false; 
	  while (literals.nextSymbol(sym)) 
	    if (!pushSymbol(sym,false)) return false;
	  if (!closeNonpropLiteral()) return false;
	};
    };
  return true;
}; // bool TemporaryClauseCollector::loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,ClauseBackground::Rule rule)



bool TemporaryClauseCollector::pushLiteralWithUnifier(ulong index,
						      const TERM* lit,
						      const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithUnifier(ulong index,const TERM* lit,const LiteralProperties& prop)");
  
  ASSERT(!_debugTrash);
  TERM header = *lit;

  if (!header.arity()) return pushPropLiteral(header,prop);

  if (!openNewNonpropLiteral(header,prop)) return false;
 
  StdTermWithSubst term;
  term.Reset(index,lit);
  term.SkipSym();
  PrefSym sym;
  if (_markSubstitutionPositions)
    {
      while (!term.Empty())
	{
	  sym = term.CurrSym();
	  if (sym.isVariable()) // free variable, must be renamed
	    {
	      if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) return false; 
	    }
	  else
	    {
	      if (sym.Func().arity())
		{
		  if (!pushNonconstFunc(sym.Func(),
					term.currentPositionIsInSubstitution())) 
		    return false;
		}
	      else 
		{
		  pushConst(sym.Func(),
			    term.currentPositionIsInSubstitution());
		}; 
	    };
	  term.SkipSym();
	};
    }
  else // !_markSubstitutionPositions
    {
      while (!term.Empty())
	{
	  sym = term.CurrSym();
	  if (sym.isVariable()) // free variable, must be renamed
	    {
	      if (!pushVar(_unifierVarRenaming.Rename(sym.var()))) return false; 
	    }
	  else
	    {
	      if (sym.Func().arity())
		{
		  if (!pushNonconstFunc(sym.Func())) return false;
		}
	      else 
		{
		  pushConst(sym.Func());
		}; 
	    };
	  term.SkipSym();
	};
    };


  return closeNonpropLiteral();  

}; // bool TemporaryClauseCollector::pushLiteralWithUnifier(ulong index,const TERM* lit,const LiteralProperties& prop)



bool 
TemporaryClauseCollector::pushLiteralWithUnifier(PrefixSym* lit,
						 const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithUnifier(PrefixSym* lit,..)");
  
  ASSERT(!_debugTrash);
  ASSERT(Prefix::CorrectComplexTerm(lit));
  TERM header = lit->Head().Func();
  if (!header.arity()) 
    return pushPropLiteral(header,prop);
  
  if (!openNewNonpropLiteral(header,prop)) return false;
  
  Prefix::Iter iter;
  iter.Reset(lit);
  TERM sym;
  iter.SkipSym(); // skipping the header
  if (_markSubstitutionPositions)
    {
      do
	{
	  if (iter.CurrVar()) // free variable
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true))
		return false;   
	    }
	  else // functor
	    {
	      sym = iter.CurrSubterm()->Head().Func();
	      if (sym.arity())
		{
		  if (!pushNonconstFunc(sym,
					iter.currentPositionIsInSubstitution())) 
		    return false;
		}
	      else // constant
		pushConst(sym,
			  iter.currentPositionIsInSubstitution());
	    };
	  iter.SkipSym();
	}
      while (!iter.Empty());
    }
  else // !_markSubstitutionPositions
    {
      do
	{
	  if (iter.CurrVar()) // free variable
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar())))
		return false;   
	    }
	  else // functor
	    {
	      sym = iter.CurrSubterm()->Head().Func();
	      if (sym.arity())
		{
		  if (!pushNonconstFunc(sym)) return false;
		}
	      else // constant
		pushConst(sym);
	    };
	  iter.SkipSym();
	}
      while (!iter.Empty());
    };
  return closeNonpropLiteral();
}; // bool TemporaryClauseCollector::pushLiteralWithUnifier(PrefixSym* lit,..)

bool 
TemporaryClauseCollector::pushEqLiteralWithUnifier(const TERM& header,
						   PrefixSym* arg1,
						   PrefixSym* arg2,
						   const LiteralProperties& prop)
{
  CALLM(_name,"pushEqLiteralWithUnifier(const TERM& header,..)"); 

  ASSERT(!_debugTrash);
  ASSERT(header.IsEquality());   
  if (_markSubstitutionPositions)
    {
      return openNewNonpropLiteral(header,prop)
	&& pushTermWithUnifier(arg1,false)
	&& pushTermWithUnifier(arg2,false)
	&& closeNonpropLiteral();
    }
  else // !_markSubstitutionPositions
    {
      return openNewNonpropLiteral(header,prop)
	&& pushTermWithUnifier(arg1)
	&& pushTermWithUnifier(arg2)
	&& closeNonpropLiteral();
    };
}; // bool TemporaryClauseCollector::pushEqLiteralWithUnifier(const TERM& header,..)



bool 
TemporaryClauseCollector::pushLiteralWithReplacement(ulong index,
						     const TERM* lit,
						     const TERM* redex,
						     PrefixSym* replacingTerm,
						     const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithReplacement(ulong index, const TERM* lit,...)");
  
  ASSERT(!_debugTrash);
  TERM header = *lit;
  ASSERT(header.arity()); 
  if (header.IsEquality()) header.MakeUnordered();
  if (!openNewNonpropLiteral(header,prop)) 
    return false;
  
  StdTermWithSubst iter;
  iter.Reset(index,lit);
  iter.SkipSym();
  PrefSym sym;
  
  if (_markSubstitutionPositions)
    {
      while (!iter.Empty())
	{
	  if (iter.CurrPosInBase() == redex)
	    {
	      if (!pushTermWithUnifier(replacingTerm,false)) return false;
	      iter.SkipTerm();
	    }
	  else
	    { 
	      sym = iter.CurrSym();
	      if (sym.isVariable()) // free variable, must be renamed
		{
		  if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) 
		    return false;
		}
	      else
		{
		  if (sym.Func().arity())
		    {
		      if (!pushNonconstFunc(sym.Func(),
					    iter.currentPositionIsInSubstitution())) 
			return false;
		    }
		  else 
		    pushConst(sym.Func(),
			      iter.currentPositionIsInSubstitution()); 
		};
	      iter.SkipSym();
	    };  
	};
    }
  else // !_markSubstitutionPositions
    {      
      while (!iter.Empty())
	{
	  if (iter.CurrPosInBase() == redex)
	    {
	      if (!pushTermWithUnifier(replacingTerm)) return false;
	      iter.SkipTerm();
	    }
	  else
	    { 
	      sym = iter.CurrSym();
	      if (sym.isVariable()) // free variable, must be renamed
		{
		  if (!pushVar(_unifierVarRenaming.Rename(sym.var()))) 
		    return false;
		}
	      else
		{
		  if (sym.Func().arity())
		    {
		      if (!pushNonconstFunc(sym.Func())) return false;
		    }
		  else 
		    pushConst(sym.Func()); 
		};
	      iter.SkipSym();
	    };  
	};
    };
   
  return closeNonpropLiteral();
}; // bool TemporaryClauseCollector::pushLiteralWithReplacement(ulong index, const TERM* lit,...)

bool 
TemporaryClauseCollector::pushLiteralWithReplacement(PrefixSym* lit,
						     PrefixSym* redex,
						     const TERM& replacingTerm,
						     ulong index,
						     const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithReplacement(PrefixSym* lit,..)");
  
  ASSERT(!_debugTrash);
  TERM header = lit->Head().Func();
  ASSERT(header.arity());
  if (header.IsEquality()) header.MakeUnordered(); 
  if (!openNewNonpropLiteral(header,prop))
    return false;
  
  Prefix::Iter iter;
  iter.Reset(lit);
  TERM sym;
  iter.SkipSym(); // skipping the header
  
  if (_markSubstitutionPositions)
    {
      do
	{
	  if (iter.CurrVar()) // free variable
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true))
		return false; 
	      iter.SkipSym(); 
	    }
	  else // functor
	    {
	      if (iter.CurrSubterm() == redex)
		{
		  if (!pushTermWithUnifier(index,replacingTerm,false)) return false;
		  iter.SkipTerm();
		}
	      else
		{
		  sym = iter.CurrSubterm()->Head().Func();
		  if (sym.arity())
		    {
		      if (!pushNonconstFunc(sym,
					    iter.currentPositionIsInSubstitution())) 
			return false;
		    }
		  else // constant
		    pushConst(sym,
			      iter.currentPositionIsInSubstitution());
		  iter.SkipSym();
		}; 
	    };
	}
      while (!iter.Empty()); 
    }
  else // !_markSubstitutionPositions
    {
      do
	{
	  if (iter.CurrVar()) // free variable
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar())))
		return false; 
	      iter.SkipSym(); 
	    }
	  else // functor
	    {
	      if (iter.CurrSubterm() == redex)
		{
		  if (!pushTermWithUnifier(index,replacingTerm)) return false;
		  iter.SkipTerm();
		}
	      else
		{
		  sym = iter.CurrSubterm()->Head().Func();
		  if (sym.arity())
		    {
		      if (!pushNonconstFunc(sym)) return false;
		    }
		  else // constant
		    pushConst(sym);
		  iter.SkipSym();
		}; 
	    };
	}
      while (!iter.Empty()); 
    };
  return closeNonpropLiteral();
}; // TemporaryClauseCollector::pushLiteralWithReplacement(PrefixSym* lit,..)

bool TemporaryClauseCollector::pushPseudoLiteralWithUnifier(ulong index,
							    const TERM& header,
							    const TERM& arg)
{
  CALLM(_name,"pushPseudoLiteralWithUnifier(..)");
  
  ASSERT(!_debugTrash);
  ASSERT(header.Positive());
  ASSERT(header.arity() == 1UL);
  LiteralProperties prop;
  if (!openNewNonpropLiteral(header,prop)) 
    return false;
   
  StdTermWithSubst term;
  term.Reset(index,arg);
  PrefSym sym;
  do
    {
      sym = term.CurrSym();
      if (sym.isVariable()) // free variable, must be renamed
	{
	  if (!pushVar(_unifierVarRenaming.Rename(sym.var()),false)) 
	    return false;
	}
      else
	{
	  if (sym.Func().arity())
	    {
	      if (!pushNonconstFunc(sym.Func(),false)) return false;
	    }
	  else 
	    {
	      pushConst(sym.Func(),false);
	    }; 
	};
      term.SkipSym();
    }
  while (!term.Empty());
   
  return closeNonpropLiteral();  
}; // bool TemporaryClauseCollector::pushPseudoLiteralWithUnifier(..)


//=================================================================
