//
// File:         TemporaryClauseCollector.hpp
// Description:  Structure for collecting temporary clauses.
//               Collected tautologies are discarded 
//               as well as clauses which do not satisfy
//               certain size/complexity restrictions.
// Created:      Nov 08, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef TEMPORARY_CLAUSE_COLLECTOR_H
#define TEMPORARY_CLAUSE_COLLECTOR_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "ClauseBackground.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseQueue.hpp"
#include "Term.hpp"
#include "Clause.hpp"
#include "prefix.hpp"
#include "variables.hpp"
#include "PagedPassiveClauseStorage.hpp"
#include "Statistics.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class TemporaryClauseCollector
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BasicTempLitList::Element LitListElement;
    typedef BK::ObjectPool<TemporaryClause>::EntryHandle ClauseHandle;
  public:
    enum RejectionReason
    {
      None,
      PropositionalTautology,
      EquationalTautology,
      ClauseTooLarge,
      LiteralTooLarge,
      TooManyLiterals,
      TooManyVariables,
      LiteralTooDeep
    };
    class Context;
  public:
    
    //                  Basic functionality

    TemporaryClauseCollector();
#ifndef NO_DEBUG
    TemporaryClauseCollector(const char* name);
#endif
    ~TemporaryClauseCollector();
    void reset();
    void setMarkSubstitutionPositions(bool fl)
    {
      _markSubstitutionPositions = fl;
    };
    void setShowAnnouncedClauses(ostream* str)
    {
      _showAnnouncedClauses = str;
    };
    void setShowCompleteClauses(ostream* str)
    {
      _showCompleteClauses = str;
    };
    void setPreprocessingInProgress(bool fl)
    {
      _preprocessingInProgress = fl;
    };
    
    void setOutputQueue(TemporaryClauseQueue* q) { _outputQueue = q; };
    void setStatistics(Statistics* stat) { _statistics = stat; };

    void setTautologyCheck(bool fl) { _tautologyCheck = fl; };
    void setLiteralFactoring(bool fl) { _literalFactoring = fl; };
    void setSizeLimit(ulong l) 
    { 
      _sizeLimit = l; 
      if (_sizeLimit > VampireKernelConst::MaxClauseSize)
	_sizeLimit = VampireKernelConst::MaxClauseSize;
    };
    void setLiteralSizeLimit(ulong l) 
    { 
      _literalSizeLimit = l; 
      if (_literalSizeLimit > VampireKernelConst::MaxTermSize)
	_literalSizeLimit = VampireKernelConst::MaxTermSize;
    };
    void setLengthLimit(ulong l) 
    { 
      _lengthLimit = l;
      if (_lengthLimit > VampireKernelConst::MaxlNumOfLits)
	_lengthLimit = VampireKernelConst::MaxlNumOfLits;
    };


    RejectionReason rejectionReason() const { return _rejectionReason; };
    bool rejectedAsRedundant() const
    {
      return 
	(_rejectionReason == PropositionalTautology) ||
	(_rejectionReason == EquationalTautology);
    };

    void openNewClause(ClauseBackground::Rule rule);
    void openNewClause(ClauseBackground::Rule rule,Clause* mainAnc)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,Clause* mainAnc)");
      openNewClause(rule);
      addMainAncestor(mainAnc);
    };
    void openNewClause(ClauseBackground::Rule rule,
		       Clause* mainAnc1,
		       Clause* mainAnc2)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,Clause* mainAnc1,Clause* mainAnc2)");
      openNewClause(rule);
      addMainAncestor(mainAnc1);
      addMainAncestor(mainAnc2);
    };

    void openNewClause(ClauseBackground::Rule rule,
		       Clause* mainAnc1,
		       Clause* mainAnc2,
		       Clause* mainAnc3)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,..,Clause* mainAnc3)");
      openNewClause(rule);
      addMainAncestor(mainAnc1);
      addMainAncestor(mainAnc2);
      addMainAncestor(mainAnc3);
    };

    
    void addMainAncestor(Clause* cl) 
    { 
      CALLM(_name,"addMainAncestor(Clause* cl)");
      if (cl->isSubgoal()) _newClause->background().markAsSubgoal();
      _newClause->background().pushAncestor(cl,cl->inferenceDepth());
    };
    void announceNewClause();
    
    
    bool finaliseClause();
    bool finaliseClause(ClauseHandle& recipient);
    bool openNewNonpropLiteral(const TERM& header,
			       const LiteralProperties& prop);
    bool pushPropLiteral(const TERM& header,
			 const LiteralProperties& prop);
    bool openNewLiteral(const TERM& header,
			const LiteralProperties& prop);
    bool collectLiteral(LitHandle& lit);
    bool collectLiteralList(BasicTempLitList& lst);
    bool closeNonpropLiteral();
    bool closeLiteral();

    bool pushVar(ulong v);
    bool pushVar(ulong v,bool substPos);
    bool pushVar(const TERM& var);
    bool pushVar(const TERM& var,bool substPos);
    bool pushConst(const TERM& c);
    bool pushConst(const TERM& c,bool substPos);
    bool pushNonconstFunc(const TERM& f);
    bool pushNonconstFunc(const TERM& f,bool substPos);
    bool pushFunc(const TERM& f) 
    {
      CALLM(_name,"pushFunc(const TERM& f)");
      ASSERT(f.isComplex());
      return (f.arity())? pushNonconstFunc(f) : pushConst(f);
    };
    bool pushFunc(const TERM& f,bool substPos) 
    {
      CALLM(_name,"pushFunc(const TERM& f,bool substPos)");
      ASSERT(f.isComplex());
      return (f.arity())? pushNonconstFunc(f,substPos) : pushConst(f,substPos);
    };
    bool pushSymbol(const TERM& sym)
    {
      CALLM(_name,"pushSymbol(const TERM& sym)");
      if (sym.isVariable()) return pushVar(sym);
      return (sym.arity())? pushNonconstFunc(sym) : pushConst(sym);
    };

    bool pushSymbol(const TERM& sym,bool substPos)
    {
      CALLM(_name,"pushSymbol(const TERM& sym,bool substPos)");
      if (sym.isVariable()) return pushVar(sym,substPos);
      return (sym.arity())? pushNonconstFunc(sym,substPos) : pushConst(sym,substPos);
    };



    
    void setPersistenClauseNumber(ulong n)
    {
      _newClause->setPersistenClauseNumber(n);
    };
    
    void addRule(ClauseBackground::Rule r) 
    { 
      _newClause->background().UsedRule(r); 
    };
    void addRules(const ClauseBackground::Rules& r)
    {
      _newClause->background().UsedRules(r); 
    };
    void cancelRule(ClauseBackground::Rule r) 
    { 
      _newClause->background().CancelRule(r); 
    };    
    void makeVIP() { _newClause->background().MakeVIP(); };
    void markAsSupport() { _newClause->background().markAsSupport(); };
    void markAsSubgoal() { _newClause->background().markAsSubgoal(); };
    void markAsNonSubsuming() { _newClause->background().markAsNonSubsuming(); };
    void markAsNonSubsumable() { _newClause->background().markAsNonSubsumable(); };
    void addAncestor(Clause* cl) 
    { 
      CALLM(_name,"addAncestor(Clause* cl)");
      if (cl->isSubgoal()) _newClause->background().markAsSubgoal();
      _newClause->background().pushAncestor(cl,cl->inferenceDepth());
    };

    //                  Extended functionality


    bool pushLiteral(const TERM* lit,
		     const LiteralProperties& prop);
    bool pushPseudoLiteral(const TERM& header,
			   const Flatterm* arg1,
			   const WeightPolynomial::Monomials& varsForPadding,
			   const LiteralProperties& prop);
    bool loadClause(const Clause* c,ClauseBackground::Rule rule);
    bool loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,
		    ClauseBackground::Rule rule);


    bool pushLiteralWithUnifier(ulong index,
				const TERM* lit,
				const LiteralProperties& prop);
    bool pushLiteralWithUnifier(PrefixSym* lit,
				const LiteralProperties& prop); 
    bool pushEqLiteralWithUnifier(const TERM& header,
				  PrefixSym* arg1,
				  PrefixSym* arg2,
				  const LiteralProperties& prop);
    bool pushTermWithUnifier(ulong index,const TERM& t);
    bool pushTermWithUnifier(ulong index,const TERM& t,bool inSubst);
    bool pushComplexTermWithUnifier(ulong index,const TERM* t);
    bool pushComplexTermWithUnifier(ulong index,const TERM* t,bool inSubst);
    bool pushTermWithUnifier(PrefixSym* t);
    bool pushTermWithUnifier(PrefixSym* t,bool inSubst);
    bool pushComplexTermWithUnifier(PrefixSym* t);
    bool pushComplexTermWithUnifier(PrefixSym* t,bool inSubst);

    bool pushLiteralWithReplacement(ulong index,
				    const TERM* lit,
				    const TERM* redex,
				    PrefixSym* replacingTerm,
				    const LiteralProperties& prop);

    bool pushLiteralWithReplacement(PrefixSym* lit,
				    PrefixSym* redex,
				    const TERM& replacingTerm,
				    ulong index,
				    const LiteralProperties& prop);

    bool pushPseudoLiteralWithUnifier(ulong index,
				      const TERM& header,
				      const TERM& arg);

  private:
    void reject(const RejectionReason& reason);
    void registerGenerated(ClauseBackground::Rule rule);
  private:
    OPEN_BR_DECL(1);
    bool _markSubstitutionPositions;
    ostream* _showAnnouncedClauses;
    ostream* _showCompleteClauses;
    bool _preprocessingInProgress;
    TemporaryClauseQueue* _outputQueue;
    RejectionReason _rejectionReason;
    ClauseHandle _newClause;
    LitHandle _newLiteral;
    
    Statistics* _statistics;
    
    VAR_TO_INT_RENAMING _unifierVarRenaming;

    ulong _nonanswerSize;
    ulong _totalSize;

    bool _tautologyCheck;
    bool _literalFactoring;

    ulong _sizeLimit;
    ulong _literalSizeLimit;
    ulong _lengthLimit;
    ulong _inferenceDepthLimit;
    
    bool _collectingNonanswerLiteral;    
    bool _collectingNonpropLiteral;
    CLOSE_BR_DECL(1);

    DOP(bool _debugTrash);
#ifndef NO_DEBUG
    const char* _name;
#endif 
  }; // class TemporaryClauseCollector
}; // namespace VK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector::Context"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class TemporaryClauseCollector::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator() : _previous(Context::getCurrent())
      {
	Context::setCurrent(static_cast<Context*>(0));
      };
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; // class AutoActivator

  public:
    Context() {};
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
  private:
    static Context* _current;
  }; // class TemporaryClauseCollector::Context
    

}; // namespace VK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{

    
    //                  Basic functionality
  

  inline
  bool 
  TemporaryClauseCollector::openNewNonpropLiteral(const TERM& header,
						  const LiteralProperties& prop)
  {
    CALLM(_name,"openNewNonpropLiteral(..)");
    ASSERT(!_debugTrash);
    ASSERT(header.arity());
    (_totalSize += header.arity()) += 1;
    if (_totalSize > VampireKernelConst::MaxClauseSize)
      {
	reject(ClauseTooLarge);
	return false;
      };
    _collectingNonanswerLiteral = !header.isAnswerHeader();
    if (_collectingNonanswerLiteral)
      {
	(_nonanswerSize += header.arity()) += 1;
	ASSERT(_nonanswerSize <= _totalSize);
	if (_nonanswerSize > _sizeLimit)
	  {
	    reject(ClauseTooLarge);
	    return false;
	  };
      };
    if (_newLiteral.isNull())
      {
	if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	  {
	    reject(None);
	    return false;
	  };
      }
    _newLiteral->pushNonpropHeader(header,prop);
    return true;
  }; // void TemporaryClauseCollector::openNewNonpropLiteral(..)

  inline
  bool 
  TemporaryClauseCollector::pushPropLiteral(const TERM& header,
					    const LiteralProperties& prop)
  {
    CALLM(_name,"pushPropLiteral(..)");    
    ASSERT(!_debugTrash);
    if (header == TERM::builtInTrue() && _tautologyCheck)
      {
	reject(PropositionalTautology);
	return false;
      };
    if (header == TERM::builtInFalse() && _literalFactoring)
      {
	return true;
      };    
    TERM complementaryHeader = header;
    complementaryHeader.Inversepolarity();
    for (const LitListElement* el = _newClause->literals()->first();
	 el;
	 el = el->next())
      {
	if (el->literal()->header() == header && _literalFactoring)
	  {
	    el->literal()->properties().factoriseWith(prop);
	    return true;
	  };
	if (el->literal()->header() == complementaryHeader && 
	    _tautologyCheck)
	  {	    
	    reject(PropositionalTautology);
	    return false;
	  };
      };
    ++_totalSize;
    if (_totalSize > VampireKernelConst::MaxClauseSize)
      {
	reject(ClauseTooLarge);
	return false;
      };
    _collectingNonanswerLiteral = !header.isAnswerHeader();
    if (_collectingNonanswerLiteral)
      {
	++_nonanswerSize;
	ASSERT(_nonanswerSize <= _totalSize);
	if (_nonanswerSize > _sizeLimit)
	  {
	    reject(ClauseTooLarge);
	    return false;
	  };
      };
    
    if (_newClause->literals()->length() >= _lengthLimit)
      {	
	reject(TooManyLiterals);
	return false;
      };
    
    if (_newLiteral.isNull())
      {
	if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	  {
	    reject(None);
	    return false;
	  };
      };
    _newLiteral->pushPropLit(header,prop);
    _newClause->literals()->collect(_newLiteral);
    return true;
  }; // bool TemporaryClauseCollector::pushPropLiteral(..)

  inline 
  bool TemporaryClauseCollector::openNewLiteral(const TERM& header,
						const LiteralProperties& prop)
  {
    CALLM(_name,"openNewLiteral(..)");
    if (header.arity())
      {
	_collectingNonpropLiteral = true;
	return openNewNonpropLiteral(header,prop);
      }
    else
      {
	_collectingNonpropLiteral = false;
	return pushPropLiteral(header,prop);
      };

  }; // bool TemporaryClauseCollector::openNewLiteral(..)

  inline 
  bool TemporaryClauseCollector::collectLiteral(LitHandle& lit)
  {
    CALLM(_name,"collectLiteral(LitHandle& lit)");
    ASSERT(!_debugTrash);
    TERM header = lit->header();
    if (header.arity())
      {
	// nonpropositional literal
	_totalSize += lit->size();
	if (_totalSize > VampireKernelConst::MaxClauseSize)
	  {
	    reject(ClauseTooLarge);
	    return false;
	  };
	_collectingNonanswerLiteral = !header.isAnswerHeader();
	
	if (_collectingNonanswerLiteral)
	  {
	    _nonanswerSize += lit->size();
	    ASSERT(_nonanswerSize <= _totalSize);
	    if (_nonanswerSize > _sizeLimit)
	      {
		reject(ClauseTooLarge);
		return false;
	      };
	  };

	
	if (lit->isEquality())
	  {	
	    ASSERT(!lit->isAnswerLiteral());
	    ASSERT(_collectingNonanswerLiteral);
	    if (lit->argsAreEqual())
	      {	
		if (lit->isPositive() && _tautologyCheck)
		  {
		    reject(EquationalTautology);
		    return false;
		  }
		else // trivial equality resolution
		  if (_literalFactoring)
		    {
		      _totalSize -= lit->size();
		      _nonanswerSize -= lit->size();
		      return true;
		    }; 
	      };


	    for (LitListElement* el = _newClause->literals()->first();
		 el;
		 el = el->next())
	      {
		// look for identical or complementary literals
		if (el->literal()->isEquality())
		  {
		    if (lit->equalEqLit(*(el->literal())) &&
			_literalFactoring)
		      {
			// propositional factoring
			el->literal()->properties().factoriseWith(lit->properties());
			_totalSize -= lit->size();
			_nonanswerSize -= lit->size();
			return true;
		      };
          
		    if (lit->complementaryEqLit(*(el->literal())) &&
			_tautologyCheck)
		      {
			reject(PropositionalTautology);
			return false;
		      };
		  };
	      }; // for (LitListElement* el ...
    
	    if (lit->size() > _literalSizeLimit)
	      {	      
		reject(LiteralTooLarge);
		return false;
	      };
	  }
	else // nonpropositional nonequality literal
	  {
	    // look for identical or complementary literals
	    for (LitListElement* el = _newClause->literals()->first();
		 el;
		 el = el->next())
	      {  
		if (lit->equal(*(el->literal())) && _literalFactoring) 
		  {
		    // propositional factoring
		    el->literal()->properties().factoriseWith(lit->properties());
		    _totalSize -= lit->size();
		    if (_collectingNonanswerLiteral) 
		      {
			ASSERT(!lit->isAnswerLiteral());
			_nonanswerSize -= lit->size();
		      };
		    return true;
		  }; 
		if (lit->isComplementary(*(el->literal())) && _tautologyCheck)
		  {
		    reject(PropositionalTautology);
		    return false;
		  };
	      }; // for (LitListElement* el..

	    ASSERT(lit->size() <= VampireKernelConst::MaxTermSize);
	
	    if (_collectingNonanswerLiteral)
	      {
		ASSERT(!lit->isAnswerLiteral());	    
		if (lit->size() > _literalSizeLimit)
		  {	      
		    reject(LiteralTooLarge);
		    return false;
		  };
	      };
	  };
    
	if (_newClause->literals()->length() >= _lengthLimit)
	  {
	    reject(TooManyLiterals);
	    return false;
	  };
	_newClause->literals()->collect(lit);
	return true;
      }
    else // propositional literal
      {
	if (header == TERM::builtInTrue() && _tautologyCheck)
	  {
	    reject(PropositionalTautology);
	    return false;
	  };
	if (header == TERM::builtInFalse() && _literalFactoring)
	  {
	    return true;
	  };    
	TERM complementaryHeader = header;
	complementaryHeader.Inversepolarity();
	for (const LitListElement* el = _newClause->literals()->first();
	     el;
	     el = el->next())
	  {
	    if (el->literal()->header() == header &&
		_literalFactoring) 
	      {
		el->literal()->properties().factoriseWith(lit->properties());
		return true;
	      };
	    if (el->literal()->header() == complementaryHeader &&
		_tautologyCheck)
	      {	    
		reject(PropositionalTautology);
		return false;
	      };
	  };
	++_totalSize;
	if (_totalSize > VampireKernelConst::MaxClauseSize)
	  {
	    reject(ClauseTooLarge);
	    return false;
	  };
	_collectingNonanswerLiteral = !header.isAnswerHeader();
	if (_collectingNonanswerLiteral)
	  {
	    ++_nonanswerSize;
	    ASSERT(_nonanswerSize <= _totalSize);
	    if (_nonanswerSize > _sizeLimit)
	      {
		reject(ClauseTooLarge);
		return false;
	      };
	  };
    
	if (_newClause->literals()->length() >= _lengthLimit)
	  {	
	    reject(TooManyLiterals);
	    return false;
	  };
	
	_newClause->literals()->collect(lit);
	return true;
      }; // if (lit->arity())

  }; // bool TemporaryClauseCollector::collectLiteral(LitHandle& lit)

  inline 
  bool TemporaryClauseCollector::collectLiteralList(BasicTempLitList& lst)
  {
    CALLM(_name,"collectLiteralList(BasicTempLitList& lst)");  
    ASSERT(!_debugTrash);
    LitHandle lit;
    while (!lst.isEmpty())
      {
	lst.popTo(lit);
	if (!collectLiteral(lit))
	  return false;
      };
    return true;
  }; // bool TemporaryClauseCollector::collectLiteralList(BasicTempLitList& lst)

  inline bool TemporaryClauseCollector::closeNonpropLiteral()
  {
    CALLM(_name,"closeNonpropLiteral()");    
    ASSERT(!_debugTrash);
    ASSERT(!_newLiteral.isNull());
    _newLiteral->endOfLiteral();
    ASSERT(_newLiteral->arity());
    if (_newLiteral->isEquality())
      {	
	ASSERT(!_newLiteral->isAnswerLiteral());
	ASSERT(_collectingNonanswerLiteral);
	if (_newLiteral->argsAreEqual())
	  {	
	    if (_newLiteral->isPositive() && _tautologyCheck)
	      {
		reject(EquationalTautology);
		return false;
	      }
	    else // trivial equality resolution
	      if (_literalFactoring)
		{
		  _totalSize -= _newLiteral->size();
		  _nonanswerSize -= _newLiteral->size();
		  return true;
		}; 
	  };
	
      for (LitListElement* el = _newClause->literals()->first();
	   el;
	   el = el->next())
	{	

	  // look for identical or complementary literals
	  if (el->literal()->isEquality())
	    {
	      if (_newLiteral->equalEqLit(*(el->literal())) &&
		  _literalFactoring)
		{
		  // propositional factoring
		  el->literal()->properties().factoriseWith(_newLiteral->properties());
		  _totalSize -= _newLiteral->size();
		  _nonanswerSize -= _newLiteral->size();
		  return true;
		};
          
	      if (_newLiteral->complementaryEqLit(*(el->literal())) &&
		  _tautologyCheck)
		{
		  reject(PropositionalTautology);
		  return false;
		};
	    };
	}; // for (LitListElement* el ...
    


      if (_newLiteral->size() > _literalSizeLimit)
	{	      
	  reject(LiteralTooLarge);
	  return false;
	};
      ASSERT(_nonanswerSize <= _sizeLimit);
      }
    else // nonpropositional nonequality literal
      {
	// look for identical or complementary literals
	for (LitListElement* el = _newClause->literals()->first();
	     el;
	     el = el->next())
	  {  
	    if (_newLiteral->equal(*(el->literal()))) 
	      {
		// propositional factoring
		_totalSize -= _newLiteral->size();
		if (_collectingNonanswerLiteral) 
		  {
		    ASSERT(!_newLiteral->isAnswerLiteral());
		    _nonanswerSize -= _newLiteral->size();
		  };
		return true;
	      }; 
	    if (_newLiteral->isComplementary(*(el->literal())) &&
		_tautologyCheck)
	      {
		reject(PropositionalTautology);
		return false;
	      };
	  }; // for (LitListElement* el..

	ASSERT(_newLiteral->size() <= VampireKernelConst::MaxTermSize);
	
	if (_collectingNonanswerLiteral)
	  {
	    ASSERT(!_newLiteral->isAnswerLiteral());	    
	    if (_newLiteral->size() > _literalSizeLimit)
	      {	      
		reject(LiteralTooLarge);
		return false;
	      };
	    ASSERT(_nonanswerSize <= _sizeLimit);
	  };
      };
    
    if (_newClause->literals()->length() >= _lengthLimit)
      {
	reject(TooManyLiterals);
	return false;
      };
    _newClause->literals()->collect(_newLiteral);
    return true;
  }; // bool TemporaryClauseCollector::closeNonpropLiteral()

  inline bool TemporaryClauseCollector::closeLiteral()
  {
    CALLM(_name,"closeLiteral()");
    if (_collectingNonpropLiteral)
      return closeNonpropLiteral();
    return true;
  };
    
  inline
  bool TemporaryClauseCollector::pushVar(ulong v)
  {
    CALLM(_name,"pushVar(ulong v)");    
    ASSERT(!_debugTrash);
    // The test below is very expensive and
    // should be eliminated in the future.
    if (v >= VampireKernelConst::MaxNumOfVariables) 
      {
	reject(TooManyVariables);
	return false;
      };    
    _newLiteral->pushVar(v);  
    return true;
  }; // bool TemporaryClauseCollector::pushVar(ulong v)


  inline
  bool TemporaryClauseCollector::pushVar(ulong v,bool substPos)
  {
    CALLM(_name,"pushVar(ulong v,bool substPos)");    
    ASSERT(!_debugTrash);
    // The test below is very expensive and
    // should be eliminated in the future.
    if (v >= VampireKernelConst::MaxNumOfVariables) 
      {
	reject(TooManyVariables);
	return false;
      };    
    _newLiteral->pushVar(v,substPos);  
    return true;
  }; // bool TemporaryClauseCollector::pushVar(ulong v,bool substPos)


  inline
  bool TemporaryClauseCollector::pushVar(const TERM& var)
  {
    CALLM(_name,"pushVar(const TERM& var)");
    ASSERT(!_debugTrash);
    _newLiteral->pushVar(var); 
    return true;
  }; // bool TemporaryClauseCollector::pushVar(const TERM& var)

  inline
  bool TemporaryClauseCollector::pushVar(const TERM& var,bool substPos)
  {
    CALLM(_name,"pushVar(const TERM& var,bool substPos)");
    ASSERT(!_debugTrash);
    _newLiteral->pushVar(var,substPos); 
    return true;
  }; // bool TemporaryClauseCollector::pushVar(const TERM& var,bool substPos)


  inline
  bool TemporaryClauseCollector::pushConst(const TERM& c)
  {
    CALLM(_name,"pushConst(const TERM& c)");
    ASSERT(!_debugTrash);
    _newLiteral->pushConst(c);
    return true;
  }; // bool TemporaryClauseCollector::pushConst(const TERM& c)


  inline
  bool TemporaryClauseCollector::pushConst(const TERM& c,bool substPos)
  {
    CALLM(_name,"pushConst(const TERM& c,bool substPos)");
    ASSERT(!_debugTrash);
    _newLiteral->pushConst(c,substPos);
    return true;
  }; // bool TemporaryClauseCollector::pushConst(const TERM& c,bool substPos)


  inline
  bool TemporaryClauseCollector::pushNonconstFunc(const TERM& f)
  {
    CALLM(_name,"pushNonconstFunc(const TERM& f)");
    ASSERT(!_debugTrash);
    _totalSize += f.arity();    
    if (_totalSize > VampireKernelConst::MaxClauseSize)  
      {
	reject(ClauseTooLarge);
	return false;
      };
    
    if (_collectingNonanswerLiteral)
      {
	_nonanswerSize += f.arity();
	if (_nonanswerSize > _sizeLimit) 
	  {
	      
	    reject(ClauseTooLarge);
	    return false;
	  };
      };
    
    
    if (!_newLiteral->pushNonconstFun(f))
      {
	reject(LiteralTooDeep);
	return false;
      };
    
    if (_collectingNonanswerLiteral)
      {
	
	if (_newLiteral->size() > _literalSizeLimit)
	  {
	    reject(LiteralTooLarge);
	    return false;
	  }; 
      }
    else
      {
	if (_newLiteral->size() > VampireKernelConst::MaxTermSize)
	  {
	    reject(LiteralTooLarge);
	    return false;
	  }; 
      };
    return true;
  }; // bool TemporaryClauseCollector::pushNonconstFunc(const TERM& f)
    

  inline
  bool TemporaryClauseCollector::pushNonconstFunc(const TERM& f,bool substPos)
  {
    CALLM(_name,"pushNonconstFunc(const TERM& f,bool substPos)");
    ASSERT(!_debugTrash);
    _totalSize += f.arity();    
    if (_totalSize > VampireKernelConst::MaxClauseSize)  
      {
	reject(ClauseTooLarge);
	return false;
      };
    
    if (_collectingNonanswerLiteral)
      {
	_nonanswerSize += f.arity();
	if (_nonanswerSize > _sizeLimit) 
	  {
	      
	    reject(ClauseTooLarge);
	    return false;
	  };
      };
    
    
    if (!_newLiteral->pushNonconstFun(f,substPos))
      {
	reject(LiteralTooDeep);
	return false;
      };
    
    if (_collectingNonanswerLiteral)
      {
	
	if (_newLiteral->size() > _literalSizeLimit)
	  {
	    reject(LiteralTooLarge);
	    return false;
	  }; 
      }
    else
      {
	if (_newLiteral->size() > VampireKernelConst::MaxTermSize)
	  {
	    reject(LiteralTooLarge);
	    return false;
	  }; 
      };
    return true;
  }; // bool TemporaryClauseCollector::pushNonconstFunc(const TERM& f,bool substPos)
    

    //                  Extended functionality


  

  inline
  bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const TERM& t)
  {
    CALLM(_name,"pushTermWithUnifier(ulong index,const TERM& t)");
    ASSERT(!_debugTrash);
    if (t.isVariable())
      {
	bool freeVar;
	void* bind = Unifier::current()->variable(t.var(),index)->Unref(freeVar);
	if (freeVar) 
	  {
	    return 
	      pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)));
	  }
	else
	  return pushComplexTermWithUnifier(static_cast<PrefixSym*>(bind));
      }
    else
      {
	ASSERT(t.IsReference());
	return pushComplexTermWithUnifier(index,t.First());
      };
    
  }; // bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const TERM& t)


  inline
  bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const TERM& t,bool inSubst)
  {
    CALLM(_name,"pushTermWithUnifier(ulong index,const TERM& t,bool inSubst)");
    ASSERT(!_debugTrash);
    if (t.isVariable())
      {
	bool freeVar;
	void* bind = Unifier::current()->variable(t.var(),index)->Unref(freeVar);
	if (freeVar) 
	  {
	    return 
	      pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)),true);
	  }
	else
	  return pushComplexTermWithUnifier(static_cast<PrefixSym*>(bind),true);
      }
    else
      {
	ASSERT(t.IsReference());
	return pushComplexTermWithUnifier(index,t.First(),inSubst);
      };
    
  }; // bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const TERM& t,bool inSubst)



  inline
  bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const TERM* t)  
  {
    CALLM(_name,"pushComplexTermWithUnifier(ulong index,const TERM* t)");
    ASSERT(!_debugTrash);
    
    StdTermWithSubst term;
    term.Reset(index,t);
    PrefSym sym;
    do
      {
	sym = term.CurrSym();
	if (sym.isVariable()) // free variable, must be renamed
	  {
	    if (!(pushVar(_unifierVarRenaming.Rename(sym.var())))) return false; 
	  }
	else
	  {
	    if (sym.Func().arity())
	      {
		if (!(pushNonconstFunc(sym.Func()))) return false;
	      }
	    else pushConst(sym.Func()); 
	  };
	term.SkipSym();
      }
    while (!(term.Empty()));
    return true; 
    
  }; // bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const TERM* t)     


  inline
  bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const TERM* t,bool inSubst)  
  {
    CALLM(_name,"pushComplexTermWithUnifier(ulong index,const TERM* t,bool inSubst)");
    ASSERT(!_debugTrash);
    
    StdTermWithSubst term;
    term.Reset(index,t);
    PrefSym sym;
    do
      {
	sym = term.CurrSym();
	if (sym.isVariable()) // free variable, must be renamed
	  {
	    if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) return false; 
	  }
	else
	  {
	    if (sym.Func().arity())
	      {
		if (!(pushNonconstFunc(sym.Func(),
				       inSubst || term.currentPositionIsInSubstitution()))) 
		  return false;
	      }
	    else 
	      pushConst(sym.Func(),
			inSubst || term.currentPositionIsInSubstitution()); 
	  };
	term.SkipSym();
      }
    while (!(term.Empty()));
    return true; 
    
  }; // bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const TERM* t,bool inSubst)     


  inline
  bool TemporaryClauseCollector::pushTermWithUnifier(PrefixSym* t)
  {
    CALLM(_name,"pushTermWithUnifier(PrefixSym* t)");
    ASSERT(!_debugTrash);
    if (t->Head().isVariable())
      {
	bool free_var;
	void* bind = t->Head().var()->Unref(free_var);
	if (free_var) 
	  return pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)));
	return pushComplexTermWithUnifier(static_cast<PrefixSym*>(bind)); 
      };
    return pushComplexTermWithUnifier(t);
  }; // bool TemporaryClauseCollector::pushTermWithUnifier(PrefixSym* t)

  inline
  bool TemporaryClauseCollector::pushTermWithUnifier(PrefixSym* t,bool inSubst)
  {
    CALLM(_name,"pushTermWithUnifier(PrefixSym* t,bool inSubst)");
    ASSERT(!_debugTrash);
    if (t->Head().isVariable())
      {
	bool free_var;
	void* bind = t->Head().var()->Unref(free_var);
	if (free_var) 
	  return pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)),true);
	return pushComplexTermWithUnifier(static_cast<PrefixSym*>(bind),true); 
      };
    return pushComplexTermWithUnifier(t,inSubst);
  }; // bool TemporaryClauseCollector::pushTermWithUnifier(PrefixSym* t,bool inSubst)

  inline
  bool TemporaryClauseCollector::pushComplexTermWithUnifier(PrefixSym* t)
  {
    CALLM(_name,"pushComplexTermWithUnifier(PrefixSym* t)");
    ASSERT(!_debugTrash);
    Prefix::Iter iter;
    iter.Reset(t);
    TERM sym;
    do
      {
	if (iter.CurrVar()) // free variable
	  { 
	    if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()))) return false; 
	  }
	else // functor
	  {
	    sym = iter.CurrSubterm()->Head().Func();
	    if (sym.arity())
	      {
		if (!(pushNonconstFunc(sym))) return false;
	      }
	    else // constant
	      pushConst(sym);
	  };
	iter.SkipSym();
      }
    while (!(iter.Empty()));
    return true;
  }; // bool TemporaryClauseCollector::pushComplexTermWithUnifier(PrefixSym* t)

  inline
  bool TemporaryClauseCollector::pushComplexTermWithUnifier(PrefixSym* t,bool inSubst)
  {
    CALLM(_name,"pushComplexTermWithUnifier(PrefixSym* t,bool inSubst)");
    ASSERT(!_debugTrash);
    Prefix::Iter iter;
    iter.Reset(t);
    TERM sym;
    do
      {
	if (iter.CurrVar()) // free variable
	  { 
	    if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true)) return false; 
	  }
	else // functor
	  {
	    sym = iter.CurrSubterm()->Head().Func();
	    if (sym.arity())
	      {
		if (!(pushNonconstFunc(sym,
				       inSubst || iter.currentPositionIsInSubstitution()))) 
		  return false;
	      }
	    else // constant
	      pushConst(sym,
			inSubst || iter.currentPositionIsInSubstitution());
	  };
	iter.SkipSym();
      }
    while (!(iter.Empty()));
    return true;
  }; // bool TemporaryClauseCollector::pushComplexTermWithUnifier(PrefixSym* t,bool inSubst)


  inline
  void TemporaryClauseCollector::reject(const RejectionReason& reason)
  {
    CALLM(_name,"reject(const RejectionReason& reason)");
    ASSERT(!_debugTrash);
    DOP(_debugTrash = true);
    _rejectionReason = reason;
    
    switch (reason)
      {	
      case None:
	++(_statistics->clauses.generated.discardedForOtherReasons);
	break;
      case PropositionalTautology:
	++(_statistics->clauses.generated.discardedPropositionalTautologies);
	++(_statistics->inferences.simplifying.propositionalTautologyDeletion);
	break;
      case EquationalTautology:
	++(_statistics->clauses.generated.discardedEquationalTautologies);
	++(_statistics->inferences.simplifying.equationalTautologyDeletion);
	break;
      case ClauseTooLarge:
	++(_statistics->clauses.generated.discardedTooLarge);
	break;
      case LiteralTooLarge:
	++(_statistics->clauses.generated.discardedWithTooLargeLiterals);
	break;
      case TooManyLiterals:
	++(_statistics->clauses.generated.discardedWithTooManyLiterals);
	break;
      case TooManyVariables: 
	++(_statistics->clauses.generated.discardedWithTooManyVariables);
	break;
      case LiteralTooDeep:
	++(_statistics->clauses.generated.discardedTooDeep);
	break;
      };

  }; // void TemporaryClauseCollector::reject(const RejectionReason& reason)
  
  inline
  void 
  TemporaryClauseCollector::registerGenerated(ClauseBackground::Rule rule)
  {
    CALLM(_name,"registerGenerated(ClauseBackground::Rule rule)");
    

    ASSERT(_statistics);
    switch (rule)
      {
      case ClauseBackground::Input:
	++(_statistics->clauses.generated.byOther);
	break;
      case ClauseBackground::Prepro:
	++(_statistics->clauses.generated.byOther);
	break;
      case ClauseBackground::BinRes:
	++(_statistics->clauses.generated.byResolution);
	++(_statistics->inferences.deduction.resolution.accomplished);
	break;
      case ClauseBackground::ForwTransChain:
	++(_statistics->clauses.generated.byForwardChaining);
	++(_statistics->inferences.deduction.forwardChaining.accomplished);
	break;
      case ClauseBackground::BackTransChain:
	++(_statistics->clauses.generated.byBackwardChaining);
	++(_statistics->inferences.deduction.backwardChaining.accomplished);
	break;
      case ClauseBackground::ForwSup:
	++(_statistics->clauses.generated.byForwardSuperposition);
	++(_statistics->inferences.deduction.forwardSuperposition.accomplished);
	break;
      case ClauseBackground::BackSup:
	++(_statistics->clauses.generated.byBackwardSuperposition);
	++(_statistics->inferences.deduction.backwardSuperposition.accomplished);
	break;
      case ClauseBackground::EqRes:
	++(_statistics->clauses.generated.byEqualityResolution);
	++(_statistics->inferences.deduction.equalityResolution.accomplished);
	break;
      case ClauseBackground::EqFact:
	++(_statistics->clauses.generated.byEqualityFactoring);
	++(_statistics->inferences.deduction.equalityFactoring.accomplished);
	break;
      case ClauseBackground::BackDemod:
	++(_statistics->clauses.generated.byBackwardDemodulation);
	break;
      case ClauseBackground::Split:
	++(_statistics->clauses.generated.byPropositionalSplitting);
	break;
      case ClauseBackground::NegEqSplit:
	++(_statistics->clauses.generated.byNegativeEqualitySplitting);
	break;
      case ClauseBackground::Reanimation:
	++(_statistics->clauses.generated.byOther);
	break;
      case ClauseBackground::Clausification:
	++(_statistics->clauses.generated.byOther);
	break;
      case ClauseBackground::Unspecified:
	++(_statistics->clauses.generated.byOther);
	break;
#ifdef DEBUG_NAMESPACE
      default:
	ICP("ICP0");
#else
#  ifdef _SUPPRESS_WARNINGS_
      default:
	BK::RuntimeError::report("Wrong argument value in TemporaryClauseCollector::registerGenerated(ClauseBackground::Rule rule)");
#  endif
#endif
      };
  }; // void TemporaryClauseCollector::registerGenerated(ClauseBackground::Rule rule)

    
}; // namespace VK

//===================================================================

#endif
