//
// File:         backward_superposition.hpp
// Description:  Classes dealing with backward superposition.
// Created:      Mar 12, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 15, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
//====================================================
#ifndef BACKWARD_SUPERPOSITION_H
//==================================================================
#define BACKWARD_SUPERPOSITION_H
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"   
#include "DestructionMode.hpp"
#include "para_machine.hpp"
#include "SimplificationOrdering.hpp"
#include "weight_formula.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class ForwardProcessing;
  class Statistics;
}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BACKWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//================================================= 

namespace VK
{
class BACKWARD_SUPERPOSITION
{
 public:
  BACKWARD_SUPERPOSITION(OR_INDEX* unifIndex); 
  ~BACKWARD_SUPERPOSITION(); 

  void setStatistics(Statistics* stat)
  {
    _dummyClauseCollector.setStatistics(stat);
    _statistics = stat;
  };

  void setClauseCollector(TemporaryClauseCollector* cc) 
  {
    _clauseCollector = cc;
  };
  
  void setForwardProcessing(ForwardProcessing* fp)
  {
    _forwardProcessing = fp;
  };
  
  void setSimultaneous(long degree) { _simultaneousness = degree; };
  void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;     
    };
  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
  };
  void setOrientPosEqualitiesInSupport(bool fl)
  {
    _orientPosEqualitiesInSupport = fl;
  };
   
  void setMainPremiseSubsumption(bool fl)
  {
    _mainPremiseSubsumption = fl;
  };

  void setSidePremiseSubsumption(bool fl)
  {
    _sidePremiseSubsumption = fl;
  };

  
  void setMainSelectedLitDemodulation(bool fl)
  {
    _mainSelectedLitDemodulation = fl;
  };
  void setSideSelectedLitDemodulation(bool fl)
  {
    _sideSelectedLitDemodulation = fl;
  };

  
  void setClauseSizeLimit(long lim)
  {
    _clauseSizeLimit = lim;
  };
  
    

  bool SetQuery(Clause* cl); 
  bool CollectResult();
  bool NextInference(); 
  
private:
  void compileMainPremiseResidualSizeFormula();
  void recompileMainPremiseSelectedLitResidualSizeFormula();
  void loadMainPremise(Clause* cl);
  const TERM* findNextMainSelectedPosEq();
  void computeMainPremiseResidualWeight()
    {
      _mainPremiseResidualSize =
	_mainPremiseResidualSizeFormula.Compute() + 
	_mainPremiseSelectedLitResidualSizeFormula.Compute();
    }; 
  bool collectMainPremiseLiterals();
  bool LoadEq(const TERM* eq_lit,ulong eq_ind);
  bool Paramodulate();
  bool collectSidePremiseLiterals();

  bool LoadLHS(PrefixSym* t);
            
  bool NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
		     Clause*& candCl,
		     BackParaCandidates::LitRedexPair*& candLitRedexPair); 

  bool subsumeMainPremise(); 
  bool collectMainPremiseInstance();
  bool subsumeSidePremise(); 
  bool collectSidePremiseInstance();

  
  bool demodulateSelectedLits();
  bool collectMainPremiseSubtermsAvailableForParamodulationInto();
  bool collectSidePremiseSubtermsAvailableForParamodulationInto();


private:
  OPEN_BR_DECL(1);
  // main parameters
  Statistics* _statistics;
  TemporaryClauseCollector* _clauseCollector;
  ForwardProcessing* _forwardProcessing;
  OR_INDEX* _unifIndex; 
  long _simultaneousness;
  bool _noInfBetweenRulesFlag;
  bool _paramodulateFromVariables;
  bool _mainPremiseSubsumption;
  bool _sidePremiseSubsumption;
  bool _mainSelectedLitDemodulation;
  bool _sideSelectedLitDemodulation;
  bool _orientPosEqualitiesInSupport;
  bool _orientPosEqualitiesInMainPremise;
  long _clauseSizeLimit;


  // main premise data
  ulong _mainPremiseVarIndex;
  Clause* _mainPremise;

  LiteralList* _allMainPremiseLiterals;
  LiteralList::Element* _firstNonselectedMainLit;
  LiteralList::Iterator _mainSelectedLitIter;
  LiteralList::Element* _currentSelectedLitListElement;

  WEIGHT_FORMULA _mainPremiseResidualSizeFormula;
  WEIGHT_FORMULA _mainPremiseSelectedLitResidualSizeFormula;
  long _mainPremiseResidualSize;

  // current selected literal in the main premise
  const TERM* _currentMainPosEqLit;
  PrefixSym _currentMainPosEqArg1Mem[VampireKernelConst::MaxTermSize];
  PrefixSym _currentMainPosEqArg2Mem[VampireKernelConst::MaxTermSize];
  PrefixSym* _currentMainPosEqRHS;
  bool _currentMainPosEqIsUnordered;
  SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
  SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;

  // side premise data
  Clause* _sidePremise;
  ulong _sidePremiseVarIndex;
  SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;



  // retrieval
  PARA_MACH _retrievalOfUnifiable;
  BackParaCandidates::Iterator _candidates;
  BackParaCandidates::LitRedexPair* _candLitRedexPair;


  // control
  bool _newRetrievalAttempt;
  bool _isFromVariable;
  bool _swapArguments;
  bool _noRulesAsSidePremises;
  bool _mainSelectedLitDemodulationAttempted;
  bool _mainSelectedLitDemodulated;

  // misc
  TemporaryClause::Extras::Handle _dummyClause;
  TemporaryClauseCollector _dummyClauseCollector;
  CLOSE_BR_DECL(1);
}; // class BACKWARD_SUPERPOSITION

}; // namespace VK
//==============================================================
#endif 







 
