//
// File:         forward_superposition.cpp
// Description:  Representation of stored terms.
// Created:      Feb 14, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 04, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               The array FPARA_PREMISE::_inherentlySelected
//               has been made extendable.
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 05, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
//====================================================
#include "forward_superposition.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
#include "Clause.inl"

using namespace BK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "FORWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//=================================================



VK::FORWARD_SUPERPOSITION::FORWARD_SUPERPOSITION(OR_INDEX* unifIndex) : 
  OPEN_BR(1,"br 1") DCOMMA
  // main parameters 
  _statistics(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _noDuplicateRedexesFlag(false),
  _noInfBetweenRulesFlag(false),
  _useInputNonBasicness(false),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),


  // main premise data
  _mainPremiseVarIndex(0UL), 
  _mainSelectedLiterals(DOP("FORWARD_SUPERPOSITION::_mainSelectedLiterals")),
  _mainLitProperties(DOP("FORWARD_SUPERPOSITION::_mainLitProperties")),

  // current selected literal in the main premise
  _eligibleSubtermPositions("_eligibleSubtermPositions"),
  
  // side premise data
  _sidePremiseVarIndex(1UL),
    
  // retrieval
  _pseudoHeader((TERM::Arity)1,(TERM::Functor)0,Signature::PositivePolarity),
  _dummyClauseCollector(DOP("FORWARD_SUPERPOSITION::_dummyClauseCollector"))

{
  CALL("constructor FORWARD_SUPERPOSITION(OR_INDEX* unifIndex)");
  
  _mainPremiseResidualSizeFormula.SetIndex(_mainPremiseVarIndex);
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);  
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _storedConstraintCheck = 
    SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; 

VK::FORWARD_SUPERPOSITION::~FORWARD_SUPERPOSITION() 
{ 
  CALL("destructor ~FORWARD_SUPERPOSITION()"); 
  if (DestructionMode::isThorough())
    {
      ASSERT(_storedConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
    };
};

bool VK::FORWARD_SUPERPOSITION::Load(Clause* cl)
{
  CALL("Load(Clause* cl)");

  if (cl->isSupportClause())
    {
      // just in case we decide to change 
      // the treatment of support clauses
      _orientPosEqualitiesInMainPremise = _orientPosEqualitiesInSupport;
      _orientNegEqualitiesInMainPremise = _orientNegEqualitiesInSupport;
    }
  else
    {
      _orientPosEqualitiesInMainPremise = true;
      _orientNegEqualitiesInMainPremise = true;
    };
  

  DOP(_mainSelectedLiterals.unfreeze());
  if (!_mainSelectedLiterals.tryToExpand(cl->numOfAllLiterals()))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return false;
    };
  DOP(_mainSelectedLiterals.freeze());   

  

  DOP(_mainLitProperties.unfreeze());
  if (!_mainLitProperties.tryToExpand(cl->numOfAllLiterals()))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return false;
    };
  DOP(_mainLitProperties.freeze());   


  _equalityResolutionApplicable = false;
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _numOfSelectedMainLits = _allMainPremiseLiterals->numOfSelectedLits();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_numOfSelectedMainLits);

  LiteralList::Iterator iter(_allMainPremiseLiterals);  
  _mainSelectedLiterals.reset();
  _freeMainSelectedLitMemory = _mainSelectedLiteralMemory;
  for (ulong ln = 0; ln < _numOfSelectedMainLits; ln++)
    {
      ASSERT(iter.notEnd());
      _mainSelectedLiterals.push(_freeMainSelectedLitMemory);
      Prefix::StandardTermToPrefix(_mainPremiseVarIndex,iter.currentLiteral(),_freeMainSelectedLitMemory);
      ASSERT(_freeMainSelectedLitMemory - _mainSelectedLiteralMemory <= (long)VampireKernelConst::MaxClauseSize);
      _mainLitProperties[ln] = iter.currentElement()->properties();
      if ((iter.currentLiteral()->Negative())
	  && (iter.currentLiteral()->IsEquality())
	  && (iter.currentLiteral()->IsUnorderedEq()))
	_equalityResolutionApplicable = true;
       
      iter.next();   };
     
  compileMainPremiseResidualSizeFormula();
     
  resetIterationOverMainLits();


  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();
  PrefixSym* lit;
  LiteralProperties prop;
 try_next_lit:

  lit = nextSelectedMainLit(prop);

  if (!lit) 
    {
      if (_equalityResolutionApplicable)
	{
	  resetIterationOverMainLits();
	  _continuation = ContEqRes;
	  return true;
	}
      else
	return false;
    };

  if (LoadLiteral(lit,prop)) { _continuation = ContFSup; return true; };
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::Load(Clause* cl)
 


bool VK::FORWARD_SUPERPOSITION::NextInference()
{ 
  CALL("NextInference()");
  switch (_continuation)
    {
    case ContFSup:
      while (NextFSupInference())
	if (CollectResult()) return true;
      // no more superpositions, start equality resolution 
      if (!_equalityResolutionApplicable) return false;
      resetIterationOverMainLits();
      _continuation = ContEqRes;
      // no break here, proceed with the next case
    case ContEqRes:
      while (NextEqRes())
	{
	  if (CollectEqResResult())
	    {
	      UndoEqRes();         
	      return true;
	    };
	  UndoEqRes(); 
	}; 
      // no more inferences
      return false;
    };
  ICP("ICP1");
  return false;
}; // bool VK::FORWARD_SUPERPOSITION::NextInference()




void VK::FORWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();

  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());


     
  resetIterationOverMainLits();
  for (PrefixSym* m_lit = nextSelectedMainLitMemory(); 
       m_lit; 
       m_lit = nextSelectedMainLitMemory())
    {
      TERM header = m_lit->Head().Func();
      if (header.IsEquality() && header.IsOrderedEq()) _mainPremiseResidualSizeFormula.Compile(m_lit->Arg2()); 
    };
      
}; // void VK::FORWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()



bool VK::FORWARD_SUPERPOSITION::NextFSupInference()
{
  if (Paramodulate()) return true;  
  PrefixSym* lit;
  LiteralProperties prop;
 try_next_lit:

  lit = nextSelectedMainLit(prop);


  if (!lit) return false;
  if (LoadLiteral(lit,prop) && (Paramodulate())) 
    return true;
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::NextFSupInference()

bool VK::FORWARD_SUPERPOSITION::NextEqRes()
{
  CALL("NextEqRes()");
  PrefixSym* lit; 
 try_next_lit:
  lit = nextSelectedMainLitMemory();
  if (!lit) return false;
  TERM header =  lit->Head().Func();
  if (header.Negative() && header.IsEquality() && header.IsUnorderedEq())
    {
      PrefixSym* arg1 = lit+1;
      PrefixSym* arg2 = arg1->Bracket();
      if (UNIFICATION::Unify(arg1,arg2)) 
	{
	  return true;
	};
    };
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::NextEqRes()



bool VK::FORWARD_SUPERPOSITION::LoadLiteral(PrefixSym* lit,const LiteralProperties& prop)
{
  CALL("LoadLiteral(PrefixSym* lit,const LiteralProperties& prop)");

  if (!lit->Head().Func().arity()) return false; // propositional
  // here lit must be a plain nonpropositional literal with all brackets set
  _currentMainLiteral = lit;
  _eligibleSubtermPositions.reset();
  if (lit->Head().Func().IsEquality())
    {
      _eqArg1 = lit + 1;
      _eqArg2 = _eqArg1->Bracket();
      
      bool notToBeOriented = 
	(lit->Head().Func().Positive() && 
	 (!_orientPosEqualitiesInMainPremise)) ||
	(lit->Head().Func().Negative() && 
	 (!_orientNegEqualitiesInMainPremise));
	

      if (!notToBeOriented && lit->Head().Func().IsOrderedEq())
	{
	  _currentCompiledConstraint = 0;
	  loadEligibleSubtermPositions(_eqArg1);
	  _swapArguments = false;
	}
      else // unordered equality
	{ 
	  if (_eqArg1->Head().isVariable() || _eqArg2->Head().isVariable())
	    {   
	      _currentCompiledConstraint = 0;
	      if (_eqArg1->Head().isVariable())
		{
		  loadEligibleSubtermPositions(_eqArg2);
		}
	      else
		loadEligibleSubtermPositions(_eqArg1);
	      _swapArguments = false;
	    }
	  else // both arguments are complex 
	    {
	      _currentCompiledConstraint = 0;
	      if (!notToBeOriented)
		{
		  _compiledConstraint->load(_eqArg2,_eqArg1);
		  if (_compiledConstraint->rhsVarInstMayOrder())
		    {
		      _compiledConstraint->compileGreaterOrEqual();
		      _currentCompiledConstraint = _compiledConstraint;
		    };
		};
	      
	      loadEligibleSubtermPositions(_eqArg1);
	      _swapArguments = true;
	    };
	};
    }
  else // noneq. literal
    {
      _currentCompiledConstraint = 0;
      PrefixSym* endOfLit = lit->Bracket();
      PrefixSym* arg = lit + 1; 
      ulong argNum = 0UL;
      const bool* nonBasicArgs = lit->Head().Func().nonBasicArguments();
      while (arg != endOfLit)
	{
	  if (!_useInputNonBasicness ||
	      !nonBasicArgs ||
	      !nonBasicArgs[argNum])
	    loadEligibleSubtermPositions(arg);
	  arg = arg->Bracket();
	  ++argNum;
	};
      _swapArguments = false;
    };
 
  while (!_eligibleSubtermPositions.empty())
    {
      if (LoadTermToRewrite(_eligibleSubtermPositions.pop()))
	{
	  _currentMainLitProperties = prop;
	  return true;
	}
      else
	++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
    };
  return false;
}; // bool VK::FORWARD_SUPERPOSITION::LoadLiteral(PrefixSym* lit,const LiteralProperties& prop)


bool VK::FORWARD_SUPERPOSITION::LoadRestOfLit()
{
  CALL("LoadRestOfLit()");

  if (!_swapArguments) return false;
  _swapArguments = false;

  _currentCompiledConstraint = 0;
  _compiledConstraint->load(_eqArg1,_eqArg2);
  if (_compiledConstraint->rhsVarInstMayOrder())
    {
      _compiledConstraint->compileGreaterOrEqual();
      _currentCompiledConstraint = _compiledConstraint;
    };
  loadEligibleSubtermPositions(_eqArg2);
  return !_eligibleSubtermPositions.empty();
}; // bool VK::FORWARD_SUPERPOSITION::LoadRestOfLit()


inline
void VK::FORWARD_SUPERPOSITION::loadEligibleSubtermPositions(PrefixSym* term)
{
  CALL("loadEligibleSubtermPositions(PrefixSym* term)");
  if (term->Head().isVariable()) return;
  if (_noDuplicateRedexesFlag)
    {
      for (PrefixSym** prev =  _eligibleSubtermPositions.begin();
           prev != _eligibleSubtermPositions.end();
           ++prev)
	{
	  if (Prefix::identical(*prev,term)) return;
	};
    };

  _eligibleSubtermPositions.push(term);
  PrefixSym* endOfTerm = term->Bracket();
  PrefixSym* arg = term + 1; 
  ulong argNum = 0UL;
  const bool* nonBasicArgs = term->Head().Func().nonBasicArguments();
  while (arg != endOfTerm)
    {
      if (!_useInputNonBasicness ||
	  !nonBasicArgs ||
	  !nonBasicArgs[argNum])
	loadEligibleSubtermPositions(arg);
      arg = arg->Bracket();
      ++argNum;
    };
}; // void VK::FORWARD_SUPERPOSITION::loadEligibleSubtermPositions(PrefixSym* term)

bool VK::FORWARD_SUPERPOSITION::Paramodulate() 
{
  CALL("Paramodulate()");
 try_another_candidate:
  _currentCandidate = NextCandidate(_currentCompiledConstraint);
  if (_currentCandidate) 
    {
      _sidePremise = _currentCandidate->TheClause();
      if (_noRulesAsSidePremises && _sidePremise->IsRule())
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.discardedForOtherReasons);
	  goto try_another_candidate;
	};
      if (demodulateSelectedLits())
	goto try_another_candidate;

      if (subsumeSidePremise())
	goto try_another_candidate;
      return true; 
    }
  else // try another position
    {
    try_another_position:
      if (_eligibleSubtermPositions.empty()) 
	{
	  if (!LoadRestOfLit()) return false;
	};
      ASSERT(!_eligibleSubtermPositions.empty());
    
      PrefixSym* term_to_rewrite = _eligibleSubtermPositions.pop();
      if (!LoadTermToRewrite(term_to_rewrite))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
	  goto try_another_position;
	};
      _currentCandidate = NextCandidate(_currentCompiledConstraint);
      if (_currentCandidate) 
	{
	  _sidePremise = _currentCandidate->TheClause();
	  if (_noRulesAsSidePremises && _sidePremise->IsRule())
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.discardedForOtherReasons);
	    goto try_another_candidate;
	    };
	  if (demodulateSelectedLits())
	    goto try_another_candidate;
	  if (subsumeSidePremise())
	    goto try_another_candidate;
	  return true;
	};
      goto try_another_position;
    };
}; // bool VK::FORWARD_SUPERPOSITION::Paramodulate() 



VK::RES_PAIR_LIST* 
VK::FORWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr) 
{
  CALL("NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr)");
  RES_PAIR_LIST* result;

  switch (_paramodulationMode)
    {
    case Init:
      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaTree(_currentParamodulationTargetTerm->Head().Func()))))
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();

	  _mainSelectedLitDemodulationAttempted = false;
	  _paramodulationMode = ParaFromComplex;

	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;

	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A10");
	  return result;
	};
      goto try_para_from_var; 
     
    case ParaFromComplex:
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      if (result) return result;
    try_next_leaf:
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };  
	  if (subsumeMainPremise()) goto try_next_leaf;

	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A20");
	  return result;
	};
      goto try_para_from_var;
        
    case ParaFromVar:
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      if (result) return result;
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
	  _mainSelectedLitDemodulationAttempted = false;
	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A30"); 
	  return result;
	};
      return 0;
    };
     
 try_para_from_var:
  _paramodulationMode = ParaFromVar;
  if (!_retrievalOfUnifiable.CompilePseudoLiteral(_pseudoHeader,_currentParamodulationTargetTerm))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return 0;
    };
      
  if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaFromVarTree())))
    {
      _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
      _mainSelectedLitDemodulationAttempted = false;
      _retrievalOfUnifiable.ResetCandidatesFromLeaf();
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      ASSERT_IN(result,"A20");
      return result;
    };
  return 0; 
}; // RES_PAIR_LIST* VK::FORWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr)  



bool VK::FORWARD_SUPERPOSITION::CollectResult()
{
  CALL("CollectResult()");

  if (!_mainPremise->Active()) return false;
  
  if (!_sidePremise->Active()) return false;
 

  const SimplificationOrdering::StoredConstraint* constr = 
    _currentCandidate->orderingConstraint();
  if (constr)
    {
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantDiscarded);
	  return false;
	};
    };


  if (_mainPremiseResidualSize
      + _sidePremise->practicallyMinResidualSize() 
      > _clauseSizeLimit)
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.discardedBySize);
      return false;
    }; 

  _clauseCollector->openNewClause(ClauseBackground::ForwSup,
				  _mainPremise,
				  _sidePremise);
  _clauseCollector->announceNewClause();


  LiteralProperties::SLModifier rewLitSLMod = 
    LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier mainPremSLMod = 
    LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier sidePremSLMod = 
    LiteralProperties::DefaultSLModifier;

  
  if (_mainLitProperties[_nextMainSelectedLitNum].deploymentMode() !=
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_mainLitProperties[_nextMainSelectedLitNum].deploymentMode() ==
	  LiteralProperties::RuleHead)
	{
	  mainPremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_mainLitProperties[_nextMainSelectedLitNum].deploymentMode() ==
		 LiteralProperties::Subquery);
	  mainPremSLMod = LiteralProperties::PostponeSelection;
	};
    };

  ulong eqLitNum = _currentCandidate->LitNum();
  LiteralProperties eqProp = 
    _sidePremise->LitList()->nth(eqLitNum)->properties();
    

  if (eqProp.deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (eqProp.deploymentMode() == LiteralProperties::RuleHead)
	{
	  sidePremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(eqProp.deploymentMode() == LiteralProperties::Subquery);
	  sidePremSLMod = LiteralProperties::PostponeSelection;
	};
    };


  return 
    collectMainPremiseLiterals(mainPremSLMod) && 
    collectRewrittenLiteral(rewLitSLMod) &&
    collectSidePremiseLiterals(sidePremSLMod) &&
    _clauseCollector->finaliseClause();
}; // bool VK::FORWARD_SUPERPOSITION::CollectResult()   
 
bool VK::FORWARD_SUPERPOSITION::CollectEqResResult()
{
  CALL("CollectEqResResult()");

  if (!_mainPremise->Active()) return false;
  _clauseCollector->openNewClause(ClauseBackground::EqRes,_mainPremise);
  _clauseCollector->announceNewClause();
  return 
    collectMainPremiseLiterals(LiteralProperties::DefaultSLModifier) &&
    _clauseCollector->finaliseClause();
}; // bool VK::FORWARD_SUPERPOSITION::CollectEqResResult() 



bool VK::FORWARD_SUPERPOSITION::collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod)");
  LiteralProperties prop;
  prop.inheritFrom(_currentMainLitProperties,slmod);
  return 
    _clauseCollector->pushLiteralWithReplacement(_currentMainLiteral,
						 _currentParamodulationTargetTerm,
						 _currentCandidate->ReplacingTerm(),
						 _sidePremiseVarIndex,
						 prop);
}; // bool VK::FORWARD_SUPERPOSITION::collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod)


bool VK::FORWARD_SUPERPOSITION::collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod)");
  unsigned ln = 0;
  unsigned used_lit_num = _currentCandidate->LitNum();
  LiteralProperties prop;
  for (LiteralList::Iterator iter(_sidePremise->LitList());
       iter.notEnd();
       iter.next())
    {
      if (ln != used_lit_num) 
	{
	  prop.inheritFrom(iter.currentElement()->properties(),slmod);
	  if (!_clauseCollector->pushLiteralWithUnifier(_sidePremiseVarIndex,
							iter.currentLiteral(),
							prop)) 
	    return false;
	};
      ln++;
    }; 
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod)


bool VK::FORWARD_SUPERPOSITION::collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod)");
  LiteralProperties prop;
  // all the selected literals except the rewritten one
  for (ulong ln = _numOfSelectedMainLits; ln; ln--)
    if ((ln - 1) != _nextMainSelectedLitNum)
      {
	prop.inheritFrom(_mainLitProperties[ln-1],slmod);
	if (!_clauseCollector->pushLiteralWithUnifier(_mainSelectedLiterals.nth(ln-1),
						      prop)) 
	  return false;     
      };
  // all unselected
  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    {
      prop.inheritFrom(iter.currentElement()->properties(),slmod);
      if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						    iter.currentLiteral(),
						    prop)) 
	return false; 
    }; 
  return true; 
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod)

bool VK::FORWARD_SUPERPOSITION::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);

  if (collectMainPremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _mainPremise->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesSubsumed);
 discarded:
  _mainPremise->unblockForSubsumption();
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::subsumeMainPremise()


bool VK::FORWARD_SUPERPOSITION::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseInstance()



bool VK::FORWARD_SUPERPOSITION::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  ASSERT(_sidePremise == _currentCandidate->TheClause());
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 

  if (collectSidePremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _sidePremise->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.forwardSuperposition.sidePremisesSubsumed);
 discarded:
  _sidePremise->unblockForSubsumption();
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::subsumeSidePremise()


bool VK::FORWARD_SUPERPOSITION::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");  
  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;  
}; // bool VK::FORWARD_SUPERPOSITION::collectSidePremiseInstance()


bool VK::FORWARD_SUPERPOSITION::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_mainPremise->paramodulationsIntoProtected() &&
      _sidePremise->Unit())
    return false; // paramodulations into this clause are protected

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesDemodulated);
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
	  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() &&
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {	  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesDemodulated);
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (tautology)
		++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      return tautology;
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 



      if (collectSidePremiseSubtermsAvailableForParamodulationInto() &&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{	    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();

	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++(_statistics->inferences.deduction.forwardSuperposition.sidePremisesDemodulated);
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  bool tautology = 
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (tautology)
	    ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	  return tautology;
	};
    }
  else
    return false;
}; // bool VK::FORWARD_SUPERPOSITION::demodulateSelectedLits()




bool 
VK::FORWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");  
  LiteralProperties prop;
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_mainPremiseVarIndex,
								      TERM::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::FORWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_sidePremiseVarIndex,
								      TERM::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::FORWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()


//=======================================================================
