
/*
 * File XML.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#include <string>
#include <iostream>
#include "Allocator.hpp"

namespace XML
{


class Attribute {
protected:
  
  enum Type {
    TEXT,
    LONG,
    INTEGER,
    DOUBLE
  };
public:
  Attribute (Type type, const string& name);
  virtual ~Attribute () {}
  void writeList (ostream&);

  
  void* operator new (size_t);
  void operator delete (void* obj);
  void destroy();
  
  void setNext (Attribute* next) { _next = next; }
  
  void setPrevious (Attribute* previous) { _previous = previous; }
protected:
  Type _type;
  string _name;
  Attribute* _next;
  Attribute* _previous;

  virtual void writeValue (ostream&) = 0;
}; 



class IntegerAttribute :
  public Attribute 
{
public:
  IntegerAttribute (const string& name, int value);
  ~IntegerAttribute() {}
  void writeValue (ostream&);

  
  CLASS_NAME("XML::IntegerAttribute");
  USE_ALLOCATOR(IntegerAttribute);
private:
  int _value;
}; 



  class LongAttribute :
  public Attribute 
{
public:
  LongAttribute (const string& name, long value);
  ~LongAttribute() {}
  void writeValue (ostream&);

  
  CLASS_NAME("XML::LongAttribute");
  USE_ALLOCATOR(LongAttribute);
private:
  long _value;
}; 



  class DoubleAttribute :
  public Attribute 
{
public:
  DoubleAttribute (const string& name, double value);
  ~DoubleAttribute() {}
  void writeValue (ostream&);

  
  CLASS_NAME("XML::DoubleAttribute");
  USE_ALLOCATOR(DoubleAttribute);
private:
  double _value;
}; 



class TextAttribute :
  public Attribute 
{
public:
  TextAttribute (const string& name, const string& value);
  ~TextAttribute() {}
  void writeValue (ostream&);

  
  CLASS_NAME("XML::TextAttribute");
  USE_ALLOCATOR(TextAttribute);
private:
  string _value;
}; 



class Content {
protected:
  
  enum Type {
    PCDATA,
    EMPTY,
    TEXT,
    FLOAT,
    LONG,
    INTEGER,
    DEEP
  };

public: 
  virtual void write (ostream&) = 0;
  virtual void write (ostream&, int indent) = 0;
  explicit Content (Type ct);
  virtual ~Content () {}
  
  void* operator new (size_t);
  void operator delete (void* obj);
  void destroy ();
  
public:
  Type _type;
  Content* _next;
  Content* _previous;
protected:
  virtual void destroyAttributes () {}
  virtual void destroyContent () {}
}; 



class PCData
  : public Content
{
public: 
  void write (ostream&);
  void write (ostream&, int indent);
  explicit PCData (const string& text);
  ~PCData () {}

  CLASS_NAME("XML::PCData");
  USE_ALLOCATOR(PCData);
private:
  string _text;
}; 



class Element
  : public Content
{
public: 
  void append(const string& file); 
  virtual void write(ostream&); 
  virtual void write(ostream&, int indent); 
  Element(Type type, const string& name);
  virtual ~Element() {}
  static Element* generalProofAttemptInfo();
  void addAttribute(Attribute*);
  
  void* operator new (size_t);
  void operator delete (void* obj);

  
protected:
  string _name;
  Attribute* _firstAttribute;
  Attribute* _lastAttribute;
  void addAfter(Element* next);
  static void writeSpaces(ostream& str,int n);
  void destroyAttributes ();
private:
  virtual void writeContent (ostream&,int indent) = 0;
  virtual void writeFooter(ostream& str);
  
  virtual void writeEndOfHeader(ostream& str) {}
}; 



class EmptyElement
  : public Element
{
public: 
  explicit EmptyElement(const string& name);
  ~EmptyElement() {}

  
  CLASS_NAME("XML::EmptyElement");
  USE_ALLOCATOR(EmptyElement);
private:
  void writeContent (ostream&,int indent);
  
  void writeFooter(ostream& str) {}
  void writeEndOfHeader(ostream& str);
}; 



class TextElement
  : public Element
{
public: 
  TextElement(const string& name, const string& text);
  ~TextElement() {}

  
  CLASS_NAME("XML::TextElement");
  USE_ALLOCATOR(TextElement);
private:
  string _text;
  void writeContent (ostream&,int indent);
}; 



class FloatElement
  : public Element
{
public: 
  FloatElement(const string& name, float value);
  ~FloatElement() {}

  
  CLASS_NAME("XML::FloatElement");
  USE_ALLOCATOR(FloatElement);
private:
  float _value;
  void writeContent (ostream&,int indent);
}; 



class LongElement
  : public Element
{
public: 
  LongElement(const string& name, long value);
  ~LongElement() {}

  
  CLASS_NAME("XML::LongElement");
  USE_ALLOCATOR(LongElement);
private:
  long _value;
  void writeContent (ostream&,int indent);
}; 



class IntegerElement
  : public Element
{
public: 
  IntegerElement(const string& name, int value);
  ~IntegerElement() {}

  
  CLASS_NAME("XML::IntegerElement");
  USE_ALLOCATOR(IntegerElement);
private:
  int _value;
  void writeContent (ostream&,int indent);
}; 



class DeepElement
  : public Element
{
public: 
  explicit DeepElement (const string& name);
  void addChild (Content* subelement);
  ~DeepElement () {}
  void destroyContent ();

  
  CLASS_NAME("XML::DeepElement");
  USE_ALLOCATOR(DeepElement);
private:
  Content* _firstSubelement;
  Content* _lastSubelement;
  void writeContent (ostream&,int indent);
}; 


}; 
