
/*
 * File vampire.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "VampireKernel.hpp"
#include "RuntimeError.hpp"

#include "antlr/ANTLRException.hpp"

#include "Tracer.hpp"
#include "Timer.hpp"
#include "Random.hpp"
#include "Run.hpp"
#include "Options.hpp"
#include "XML.hpp"
#include "Unit.hpp"
#include "ProofChecker.hpp"


int main (int argc, char* argv [])
{
  TRACER ("main");

  
  Timer timer;
  timer.reset();
  timer.start();

  
  Random::resetSeed ();

  
  Run::setSignalHandlers();
  BK::RuntimeError::errorHook = runtimeErrorHook;
#ifndef NO_DEBUG
  BK::Debug::setErrorHook(debugSystemInternalErrorHook);
#endif

  try {
    
    Options::globalOptions.correct (argc, argv);
    GlobalAllocator::setMemoryLimitInMegabytes(Options::globalOptions.memoryLimit());

    
    string inputFileName = Options::globalOptions.inputFile ();
    FormulaList formulas;
    UnitList units;
    Run::readInput (inputFileName, formulas, units);
    if (formulas.isNonEmpty()) {
      units = UnitList(AXIOM, formulas);
    }
    else if (units.isEmpty()) { 
      USER_ERROR("cannot open input file or syntax error in the file");
    }

    Run::log->addAttribute(new XML::TextAttribute("mode",
						  Options::globalOptions.modeName()));
    Run::run(units,timer);
  }
  catch (AllocationFailureException& exception) {
    Run::log->addChild(new XML::TextElement("result",Run::toString(UNKNOWN)));
    Run::log->addChild(new XML::TextElement("termination_reason",
					    "allocation failure"));
    exception.cry(cout);
  }
  catch (VampireTerminationException& exception) {
    Run::log->addChild(new XML::TextElement("termination_reason",
					    exception.getReason()));
    Run::log->addChild(new XML::TextElement("result",
					    Run::toString(exception.getResult())));
    
    Run::reportAllocationFailure();
  }
  catch (AssertionViolationException& exception) {
    XML::DeepElement* explanation = new XML::DeepElement("abnormal_termination");
    explanation->addChild(exception.toXML());
    Run::log->addChild(explanation);
    
  }
  catch (UserErrorException& exception) {
    XML::DeepElement* explanation = new XML::DeepElement("abnormal_termination");
    explanation->addChild(exception.toXML());
    Run::log->addChild(explanation);
    exception.cry(cout);
    
  }
  catch (ProofCheckingException& exception) {
    XML::DeepElement* explanation = new XML::DeepElement("abnormal termination");
    explanation->addChild(exception.toXML());
    Run::log->addChild(explanation);
    exception.cry(cout);
    
  }
  catch(antlr::ANTLRException& e) { 
    XML::DeepElement* explanation = new XML::DeepElement("abnormal_termination");
    Run::log->addChild(explanation);
    explanation->addChild(new XML::TextElement("parser_exception",
					       e.toString()));
    cerr << e.toString() << "\n";
    
  }

  Run::saveLog();  
  return EXIT_SUCCESS;
} 
